//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: dp_tick.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  modified for R3D, moved to main engine 
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 28 (v220-28) 04/Oct/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8 JDoom.dll code
//
//
// DESCRIPTION:   Top-level tick stuff.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "doom/p_local.h"

#include "r_main.h"
#include "cvars.h"
#include "net_main.h"
#include "r3d_compat.h"
#include "r3d_mapinfo.h"
#include "doom\sndidx.h"
#include "s_main.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void R_SetMobjSectorLists(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL	ui_active;
extern int		waswaterdynamics;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int			lastlight;				// v220-31

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

// This is called at all times, no matter gamestate.
void P_RunPlayers(void)
{
	if (mbf_players[0].plr->ingame)
	{
		// Get the command for the player.
		Net_GetTicCmd(&mbf_players[0].cmd, 0);
		
		// The player thinks.
		if (   gamestate == GS_LEVEL
			&& !ma_inf.exitlevel)
		{
			P_PlayerThink(&mbf_players[0]);
		}
	}
}

static void DeadPlayerMessage()
{
// v220-35 show red mist if player's crushed by a plat (see T_MovePlane)
//         This avoids rendering issues if in space between
//         floor and ceiling (which is normally proscribed)
	if (ma_inf.no_light[0] && ma_inf.no_light[0] < 35 && ++ma_inf.no_light[0] == 35)
	{
		scpMessage.scriptMessage = "How careless of you getting crushed to nothing.\\n"
			"Your crushed bones and blood is all that's left\\nto mark that you ever existed.";
		scpMessage.delay_to_next_text = 0;
		scpMessage.delay_to_fade = 400;
		scpMessage.alpha = 1.0f;
	}

// v220-31 show special message if player is killed in a toxic fake sector
	if (ma_inf.no_light[1] && ma_inf.no_light[1] < 35 && ++ma_inf.no_light[1] == 35)
	{
		scpMessage.scriptMessage =
			"Highly reactive chemicals encountered\\n"
			" - Radiation suits are ineffective -\\n"
			" - God mode is ineffective -\\n";
		scpMessage.delay_to_next_text = 0;
		scpMessage.delay_to_fade = 280;
		scpMessage.alpha = 1.0f;
	}
}

//===========================================================================
// DP_Ticker
//	Called 35 times per second. 
//	The heart of play sim.
//===========================================================================
void DP_Ticker (void)
{
	// If the game is paused, nothing (much) will happen.
	if (paused) return;
	
	actual_leveltime++;
		
	// pause if in menu and thinkers have run
	if (  (menu_active || ui_active || con_freezegame)	// v220-28 add con_freezegame
		&& ppt.thinkers_have_run)
	{
		return;
	}
	
	P_RunThinkers ();
	DeadPlayerMessage();
	P_UpdateSpecials ();
	R_SetMobjSectorLists();

	if (!ppt.thinkers_have_run)
	{
		mobj_t *mo = ppt.viewplrmo;

		if (mo)
		{
			P_UnlinkThing(mo);
			P_SetThingPosition(mo);	// ensure player is linked in
			P_SetMpos(mo);
			ppt.thinkers_have_run = true;
			ge_inf.nextname[0] = 0;
			AS_SetFCO_Limit();
			if (ma_inf.startlevelmusic)
				S_LevelMusic();
		}

		// v220-5 set to a default of 35.
		// This ensures smoothing values will be used in the first second.
		stats.fpsnum = 35;
		stats.framecount = 0;
		ma_inf.loadgame = false;
		lastlight = 0;
	}
	else
	{
		if (waswaterdynamics != waterdynamics)
		{
			if (R3D_edit) waterdynamics = false;
			TS_ResetWater();
			P_SpawnFriction();	// v220-4 rerun - friction set depends on state of waterdynamics
			waswaterdynamics = waterdynamics;
		}

		if (cfg.useTorchLight)
		{
			sector_t *sc = ppt.viewplrmo->subsector->sector;
			int n, v;

			if (underfakefloor && sc->heightsec >= 0)
			{
				n = sectors[sc->heightsec].lightlevel;
			}
			else
			{
				n = sc->lightlevel;
			}

			v = (255 - n) & 0xfe;

			if (v > 128) v = 128;

			if (lastlight > v || n > 128)
			{
				if (lastlight >= 10) lastlight-=2;
			}
			else if (lastlight < v)
			{
				lastlight += 2;
			}
		}

		if (ge_inf.conprogress_disabled)
		{
			// v220-4 add spawn fog at map start option
			if (mapinfo->R3Dflags & R3Dnewmapfog && ppt.viewplrmo)
			{
				unsigned	an = ppt.viewplrmo->angle>>ANGLETOFINESHIFT;
				mobj_t		*mo = P_SpawnMobj(
								ppt.viewplrmo->x + 20*finecosine[an],
								ppt.viewplrmo->y + 20*finesine[an],
								ppt.viewplrmo->z, MT_TFOG, 0);

				P_SetMpos(mo);

				S_StartSound(sfx_telept, mo);
				
				ppt.viewplrmo->reactiontime = 18;	
			}

			ge_inf.loadlevel = ge_inf.conprogress_disabled = false;
		}
	}

	// For par times, among other things.
	leveltime++;
	// Although leveltime is signed, it would actually take 710
	// days, running the same map, before the value becomes -ve
	// Belt and braces must prevail, however, and I can't be
	// fagged to go round changing everything to a uint - gmj;
	if (leveltime < 0) leveltime = 0;
}

