//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 id Software, Inc., G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: m_menu.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D, Moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  Defunct code removed
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//  dead stuff removed. Support for more pwad graphics added. some tidying up
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll
//
// DESCRIPTION:
//	DOOM selection menu, options, episode etc.
//	Sliders and icons. Kinda widget stuff.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>
#include <io.h>
#include <direct.h>
#include <math.h>

#include "r3d_compat.h"

#include "doom/mbf_player.h"
#include "doom/Mn_def.h"
#include "doom/hu_stuff.h"
#include "doom/m_menu.h"
#include "doom/p_local.h"
#include "doom/p_saveg.h"
#include "doom/g_game.h"
#include "doom/d_englsh.h"
#include "doom/d_main.h"
#include "con_console.h"
#include "con_functions.h"
#include "dd_wad.h"
#include "gl_draw.h"
#include "gl_tex.h"
#include "net_main.h"
#include "s_main.h"
#include "r_model.h"
#include "sys_system.h"
#include "sys_timer.h"
#include "rend_sky.h"
#include "gl_main.h"

// OGL
#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

#define	CVAR(typ, x)	(*(typ*)Con_GetVariable(x)->ptr)

#define NUM_QUITMESSAGES	22				// QuitDOOM messages
#define	load_end			NUMSAVESLOTS

// TYPES -------------------------------------------------------------------

// Identify language to use, software localization.
typedef enum
{
  english,
  french,
  german,
  unknown

} Language_e;

typedef struct
{
	int	width, height;
} MenuRes_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

int		MI_SetEpisodes(void);
int		R3D_EpisodeNum(int num);
RDBOOL	return_to_menu;
void	ReplaceInString(char *str, char *occurance, char *replacewith);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void M_Episode(int choice);
static void M_ChooseSkill(int choice);
static void M_LoadGame(int choice);
static void M_SaveGame(int choice);
static void M_Options(int choice);
static void M_EndGame(int choice);
static void M_ReadThis(int choice);
static void M_ReadThis2(int choice);
static void M_QuitDOOM(int choice);

static void M_OpenDCP(int choice);
static void M_ChangeMessages(int choice);
static void M_AlwaysRun(int option);
static void M_LookSpring(int option);
#if 0		// $bulletrange
static void M_BulletRange(int option);
#endif
static void M_AllowJump(int option);
static void M_FallingDamage(int option);
static void M_WaterDynamics(int option);
static void M_UnderwaterDamage(int option);
static void M_HUDInfo(int option);
static void M_HUDScale(int option);
static void M_HUDRed(int option);
static void M_HUDGreen(int option);
static void M_HUDBlue(int option);
static void M_XhairType(int option);
static void M_XhairSize(int option);
static void M_AllowDoubleC(int option);
static void M_MouseLook(int option);
static void M_MouseLookInverse(int option);
static void M_MouseXSensi(int option);
static void M_MouseYSensi(int option);
static void M_JoyEnable(int option);
static void M_JoyXAxis(int option);
static void M_JoyYAxis(int option);
static void M_JoyZAxis(int option);
static void M_JoyLook(int option);
static void M_InverseJoyLook(int option);
static void M_POVLook(int option);
static void M_JoySensitivity(int option);
static void M_JoyDeadzone(int option);

static void M_FinishReadThis(int choice);
static void M_LoadSelect(int choice);
static void M_SaveSelect(int choice);
static void M_ReadSaveStrings(void);
static void M_QuickSave(void);
static void M_QuickLoad(void);

static void M_DrawMainMenu(void);
static void M_DrawNewGame(void);
static void M_NewGame(int choice);
static void M_DrawReadThis(void);
static void M_DrawEpisode(void);
static void M_DrawOptions(void);
static void M_DrawGameplay(void);
static void M_DrawHUD(void);
static void M_DrawMouseOpts(void);
static void M_DrawLoad(void);
static void M_DrawSave(void);

static void M_SetupNextMenu(Menu_t *menudef);
static void M_DrawSlider(Menu_t *menu, int index, int width, int dot, int offx);
static int  M_StringHeight(char *string, dpatch_t *font);
static void M_StartMessage(char *string,void *routine,RDBOOL input);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL		message_dontmesswithme;
extern RDBOOL		usejoystick;
extern RDBOOL		chat_on;		// in heads-up code
extern char			base_save_path[MAXPATHNAMESIZE];
extern int			epihaspwad[10];	// v220-6
extern int			episodeinf[9][3];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

//char gammamsg[10][81];
//flashcolor[3] = { 1, .7f, .1f };	defaults

// Show messages has default, 0 = off, 1 = on
int		showMessages = true;
int		waswaterdynamics = -1;

float	menuScale = .8f;

int		menu_active;

// we are going to be entering a savegame string
int		saveStringEnter;              
int		saveSlot;		// which slot to save in
int		saveCharIndex;	// which char we're editing

// used to store save description before editing
char	saveOldString[SAVESTRINGSIZE];  

#define SKULLXOFF		-32

char	savegamestrings[10][SAVESTRINGSIZE];
char	endstring[160];
char	*episodemsg;

int		MenuTime;
int		typein_time = 0;

Menu_t*	currentMenu;	// current menudef

RDBOOL	autoAim;		// b211-0

// PRIVATE DATA DEFINITIONS ------------------------------------------------

void    (*messageRoutine)(int response);

Language_e   language = english;	// Only english supported

// MENU TYPEDEFS

static MenuItem_t MainItems[] =						// DOOM MENU
{
	{ ITT_EFUNC, "NEW GAME"  , M_NewGame,  0, "M_NGAME",  -1 },
	{ ITT_EFUNC, "OPTIONS"   , M_Options,  0, "M_OPTION", -1 },
	{ ITT_EFUNC, "LOAD GAME" , M_LoadGame, 0, "M_LOADG",  -1 },
	{ ITT_EFUNC, "SAVE GAME" , M_SaveGame, 0, "M_SAVEG",  -1 },
	{ ITT_EFUNC, "READ THIS!", M_ReadThis, 0, "M_RDTHIS", -1 },
	{ ITT_EFUNC, "QUIT GAME" , M_QuitDOOM, 0, "M_QUITG",  -1 }
};

static Menu_t MainDef =
{
	97, 64,
	M_DrawMainMenu,
	6, MainItems,
	0, MENU_NONE,
	hu_font_b, //1, 0, 0,
	LINEHEIGHT,
	0, 6
};

static MenuItem_t EpisodeItems[] =				// EPISODE SELECT
{
	// Text defs TXT_EPISODE1..4 + 2 for MAPINF0.
	{ ITT_EFUNC, "K", M_Episode, 0 /*, "M_EPI1"*/ },
	{ ITT_EFUNC, "T", M_Episode, 1 /*, "M_EPI2"*/ },
	{ ITT_EFUNC, "I", M_Episode, 2 /*, "M_EPI3"*/ },
	{ ITT_EFUNC, "T", M_Episode, 3 /*, "M_EPI4"*/ },
	{ ITT_EFUNC, "I", M_Episode, 4 /*, "M_EPI5"*/ },
	{ ITT_EFUNC, "T", M_Episode, 5 /*, "M_EPI6"*/ },
	{ ITT_EFUNC, "I", M_Episode, 6 /*, "M_EPI7"*/ },
	{ ITT_EFUNC, "T", M_Episode, 7 /*, "M_EPI8"*/ },
	{ ITT_EFUNC, "T", M_Episode, 8 /*, "M_EPI9"*/ }
};

static Menu_t EpiDef =
{
	48, 63,
	M_DrawEpisode,
	9, EpisodeItems,
	0, MENU_MAIN,
	hu_font_b, //1, 0, 0,
	LINEHEIGHT,
	0, 9
};

static MenuItem_t NewGameItems[] =				// NEW GAME
{
	{ ITT_EFUNC, "I'M TOO YOUNG TO DIE.", M_ChooseSkill, 0, "M_JKILL", -1 },
	{ ITT_EFUNC, "HEY, NOT TOO ROUGH."	, M_ChooseSkill, 1, "M_ROUGH", -1 },
	{ ITT_EFUNC, "HURT ME PLENTY."		, M_ChooseSkill, 2, "M_HURT",  -1 },
	{ ITT_EFUNC, "ULTRA-VIOLENCE."		, M_ChooseSkill, 3, "M_ULTRA", -1 },
	{ ITT_EFUNC, "NIGHTMARE!"			, M_ChooseSkill, 4, "M_NMARE", -1 }
};

static Menu_t NewDef =
{
	48, 63,
	M_DrawNewGame,
	5, NewGameItems,
	2, MENU_EPISODE,
	hu_font_b, //1, 0, 0,
	LINEHEIGHT,
	0, 5
};

static MenuItem_t ReadItems1[] =				// OPTIONS MENU
{
	{ ITT_EFUNC, "", M_ReadThis2, 0 }
};

static Menu_t ReadDef1 =
{
	280, 185,
	M_DrawReadThis,
	1, ReadItems1,
	0, MENU_MAIN,
	hu_font_b, //1, 0, 0,
	LINEHEIGHT,
	0, 1
};

static MenuItem_t ReadItems2[] =
{
	{ ITT_EFUNC, "", M_FinishReadThis, 0 }
};

static Menu_t ReadDef2 =
{
	330, 175,
	M_DrawReadThis,
	1, ReadItems2,
	0, MENU_MAIN,
	hu_font_b, //1, 0, 0,
	LINEHEIGHT,
	0, 1
};

static MenuItem_t LoadItems[] =					// LOAD GAME MENU
{
	{ ITT_EFUNC, NULL, M_LoadSelect, 0 },
	{ ITT_EFUNC, NULL, M_LoadSelect, 1 },
	{ ITT_EFUNC, NULL, M_LoadSelect, 2 },
	{ ITT_EFUNC, NULL, M_LoadSelect, 3 },
	{ ITT_EFUNC, NULL, M_LoadSelect, 4 },
	{ ITT_EFUNC, NULL, M_LoadSelect, 5 },
	{ ITT_EFUNC, NULL, M_LoadSelect, 6 },
	{ ITT_EFUNC, NULL, M_LoadSelect, 7 },
	{ ITT_EFUNC, NULL, M_LoadSelect, 8 },
	{ ITT_EFUNC, NULL, M_LoadSelect, 9 }
};

static Menu_t LoadDef =
{
	80, 20,			//was 80, 54
	M_DrawLoad,
	NUMSAVESLOTS, LoadItems,
	0, MENU_MAIN,
	hu_font_b, //1, 0, 0,
	LINEHEIGHT,
	0, NUMSAVESLOTS
};

static MenuItem_t SaveItems[] =					// SAVE GAME MENU
{
	{ ITT_EFUNC, NULL, M_SaveSelect, 0 },
	{ ITT_EFUNC, NULL, M_SaveSelect, 1 },
	{ ITT_EFUNC, NULL, M_SaveSelect, 2 },
	{ ITT_EFUNC, NULL, M_SaveSelect, 3 },
	{ ITT_EFUNC, NULL, M_SaveSelect, 4 },
	{ ITT_EFUNC, NULL, M_SaveSelect, 5 },
	{ ITT_EFUNC, NULL, M_SaveSelect, 6 },
	{ ITT_EFUNC, NULL, M_SaveSelect, 7 },
	{ ITT_EFUNC, NULL, M_SaveSelect, 8 },
	{ ITT_EFUNC, NULL, M_SaveSelect, 9 }
};

static Menu_t SaveDef =
{
	80, 20,		//was 80, 54
	M_DrawSave,
	NUMSAVESLOTS, SaveItems,
	0, MENU_MAIN,
	hu_font_b, //1, 0, 0,
	LINEHEIGHT,
	0, NUMSAVESLOTS
};

static MenuItem_t OptionsItems[] =
{
	{ ITT_CPFUNC , "DISPLAY..."		, M_OpenDCP	, MENU_DISPLAY	},
	{ ITT_CPFUNC , "VIDEO..."		, M_OpenDCP	, MENU_VIDEO	},
	{ ITT_CPFUNC , "SOUND..."		, M_OpenDCP	, MENU_SOUND	},
	{ ITT_SETMENU, "INPUT OPTIONS...",NULL		, MENU_INPUT	},
	{ ITT_SETMENU, "CONTROLS..."	, NULL		, MENU_CONTROLS	},
	{ ITT_SETMENU, "GAMEPLAY..."	, NULL		, MENU_GAMEPLAY	},
	{ ITT_SETMENU, "HUD..."			, NULL		, MENU_HUD		},
	{ ITT_EFUNC  , "END GAME"		, M_EndGame	, MENU_MAIN		}
};

static Menu_t OptionsDef =
{
	108, 92,
	M_DrawOptions,
	8, OptionsItems,
	0, MENU_MAIN,
	hu_font_a, //1, 0, 0, 
	LINEHEIGHT_A,
	0, 8
};

static MenuItem_t GameplayItems[] =
{
	{ ITT_EFUNC, "JUMPING :"			, M_AllowJump, 0 },
	{ ITT_EFUNC, "MESSAGES :"			, M_ChangeMessages, 0 },
	{ ITT_EFUNC, "ALWAYS RUN :"			, M_AlwaysRun,  0 },
	{ ITT_EFUNC, "LOOKSPRING :"			, M_LookSpring, 0 },
#if 0		// $bulletrange
	{ ITT_EFUNC, "BULLET RANGE :"		, M_BulletRange, 0 },
#endif
	{ ITT_EFUNC, "FALLING DAMAGE :"		, M_FallingDamage, 0 },
	{ ITT_EFUNC, "WATER DYNAMICS :"		, M_WaterDynamics, 0 },
	{ ITT_EFUNC, "UNDERWATER DAMAGE :"	, M_UnderwaterDamage, 0 }
};

static Menu_t GameplayDef =
{
	70, 40,
	M_DrawGameplay,
#if 0		// $bulletrange
	8,
#else
	7,
#endif
	GameplayItems,
	0, MENU_OPTIONS,
	hu_font_a, 
	LINEHEIGHT_A,
	0,
#if 0		// $bulletrange
	8
#else
	7
#endif
};

static MenuItem_t HUDItems[] =
{
	{ ITT_EFUNC , "SHOW HEALTH :"	, M_HUDInfo		, HUD_HEALTH},
	{ ITT_EFUNC , "SHOW AMMO :"		, M_HUDInfo		, HUD_AMMO	},
	{ ITT_EFUNC , "SHOW KEYS :"		, M_HUDInfo		, HUD_KEYS	},
	{ ITT_EFUNC , "SHOW ARMOR :"	, M_HUDInfo		, HUD_ARMOR	},
	{ ITT_LRFUNC, "SCALE"			, M_HUDScale	, 0			},
	{ ITT_LRFUNC, "COLOR RED    "	, M_HUDRed		, 0			},
	{ ITT_LRFUNC, "COLOR GREEN"		, M_HUDGreen	, 0			},
	{ ITT_LRFUNC, "COLOR BLUE  "	, M_HUDBlue		, 0			},
	{ ITT_LRFUNC, "CROSSHAIR :"		, M_XhairType	, 0			},
	{ ITT_LRFUNC, "CROSSHAIR SIZE"	, M_XhairSize	, 0			}
};

static Menu_t HUDDef =
{
	70, 40,
	M_DrawHUD,
	10, HUDItems,
	0, MENU_OPTIONS,
	hu_font_a, 
	LINEHEIGHT_A,
	0, 10
};


static MenuItem_t InputItems[] =
{
	{ ITT_EFUNC,  "DOUBLE CLICK USE :", M_AllowDoubleC, 0 },
	{ ITT_EFUNC,  "MOUSE LOOK :", M_MouseLook, 0 },
	{ ITT_EFUNC,  "INVERSE MLOOK :", M_MouseLookInverse, 0 },
	{ ITT_LRFUNC, "X SENSITIVITY", M_MouseXSensi, 0 },
	{ ITT_LRFUNC, "Y SENSITIVITY", M_MouseYSensi, 0 },
	{ ITT_EMPTY,   NULL, NULL, 0 },
	{ ITT_EFUNC,  "ENABLE JOYSTICK :", M_JoyEnable, 0 },
	{ ITT_LRFUNC, "JOYSTICK X AXIS :", M_JoyXAxis, 0 },
	{ ITT_LRFUNC, "JOYSTICK Y AXIS :", M_JoyYAxis, 0 },
	{ ITT_LRFUNC, "JOYSTICK Z AXIS :", M_JoyZAxis, 0 },
	{ ITT_EFUNC,  "JOY LOOK :", M_JoyLook, 0 },
	{ ITT_EFUNC,  "INVERSE LOOK :", M_InverseJoyLook, 0 },
	{ ITT_EFUNC,  "POV LOOK :", M_POVLook, 0 },
	{ ITT_LRFUNC, "J SENSITIVITY", M_JoySensitivity, 0 },
	{ ITT_LRFUNC, "J DEADZONE   ", M_JoyDeadzone, 0 }
};

static Menu_t InputDef = 
{
	70, 40,
	M_DrawMouseOpts,
	15, InputItems,
	0, MENU_OPTIONS,
	hu_font_a, 
	LINEHEIGHT_A,
	0, 15
};

static Menu_t *menulist[] =
{
	&MainDef,
	&EpiDef,
	&NewDef,
	&OptionsDef,
	&GameplayDef,
	&HUDDef, 
	&ControlsDef,
	&InputDef,
	&LoadDef,
	&SaveDef,
	NULL
};

static float	mfSpeeds[2]		= { 1*.05f, 1*-.085f };
static float	mfAngle[2]		= { 93, 12 };
static float	mfPosAngle[2]	= { 35, 77 };
static float	mfAlpha			= 0;
static float	mfPos[2][2];

static int		numepi;

static short	itemOn;				// menu item skull is on
static short	skullAnimCounter;	// skull animation counter
static short	whichSkull;			// which skull to draw

static int		quickSaveSlot;	// -1 = no quicksave slot picked!

static int		messageToPrint;	//  1 = message to be printed
static char*	messageString;	// ...and the message string

// message x & y
static int		messx;			
static int		messy;

static int		skullnum[2];	// v220-13

static int		messageLastMenuActive;
static RDBOOL	messageNeedsInput;	// timed message = no input from user   

// graphic name of skulls
// warning: initializer-string for array of chars is too long
static char		skullName[2][9] = {"M_SKULL1", "M_SKULL2"};

static char		*endmsg[NUM_QUITMESSAGES + 1];
static char		*yesno[3] = { "NO", "YES", "MAYBE?" };
#if 0		// $bulletrange
static char		*bullet_dist[3] = { "DOOM", "DOOM x 2", "Doom x 4" };
#endif
static char		*mapset = "SET BY MAP";

// Alpha level for the entire menu. Used primarily by M_WriteText2().
static float	menu_alpha  = 1;
static int		menu_color  = 0;
static float	skull_angle = 0;

// v220-22
static int		fiddlefactor = 0;
// v220-35 dfont type set depending on pwad entries
static dpatch_t *usefont = NULL;

// CODE --------------------------------------------------------------------

void M_SetNumItems(Menu_t *menu, int num)
{
	menu->itemCount = menu->numVisItems = num;
}

//
// M_ReadSaveStrings
//  read the strings from the savegame files
//
static void M_ReadSaveStrings(void)
{
    int     i;

	for (i = 0; i < load_end; i++)
	{
		SV_SaveGameFile(i);

		if (!SV_GetSaveDescription(savegamestrings[i])) 
		{
			LoadItems[i].type = ITT_EMPTY;
		}
		else
			LoadItems[i].type = ITT_EFUNC;
	}
}

//
// Draw border for the savegame description
//   v220-13 replace draw rectangle
static void M_DrawSaveLoadBorder(int x, int y)
{
	static int m_lw = 0, m_cw;
	static int m_ln, m_cn, m_rn;
	int			i, oy = y + 10;

	if (!m_lw)
	{
		m_ln = W_GetNumForName("M_LSLEFT");
		GL_SetPatch(m_ln, 0); m_lw = lumptexinfo[m_ln].origw;
		m_cn = W_GetNumForName("M_LSCNTR");
		GL_SetPatch(m_cn, 0); m_cw = lumptexinfo[m_cn].origw;
		m_rn = W_GetNumForName("M_LSRGHT");
		GL_SetPatch(m_rn, 0);
	}

    GL_DrawPatch(x - m_lw, oy, m_ln);

	for (i = 0; i <= (SAVESTRINGSIZE + 1); i++)
		GL_DrawPatch(x + i*m_cw,oy, m_cn);

    GL_DrawPatch(x + i*m_cw, oy, m_rn);
}

//
// M_LoadGame & Cie.
//
static void M_DrawLoad(void)
{
	int x = LoadDef.x;
	int y = LoadDef.y;
    int	i;
	
    GL_DrawPatch(72, 0, W_CheckIWADNumForName("M_LOADG"));	//was 72, 28

    for (i = 0; i < load_end; i++)
    {
		M_DrawSaveLoadBorder(x ,y + LINEHEIGHT*i);
		M_WriteText2(x, y + 3 + LINEHEIGHT*i, savegamestrings[i],
			hu_font_a, 1, 0, 0);
    }
}

//
// User wants to load this game
//
static void M_LoadSelect(int choice)
{
	SV_SaveGameFile(choice);
	gameaction = ga_loadgame;
	menu_active = 0;
}

//
// Selected from DOOM menu
//
static void M_LoadGame (int choice)
{
    if (R3D_edit)
    {
		S_LocalMenuSound(sfx_oof);
		return;
    }

    M_SetupNextMenu(&LoadDef);
    M_ReadSaveStrings();
}

//
//  M_SaveGame & Cie.
//
static void M_DrawSave(void)
{
    int	i;
	int x = LoadDef.x;
	int y = LoadDef.y;
	
    GL_DrawPatch(72, 0, W_CheckIWADNumForName("M_SAVEG"));	//was 72, 28

    for (i = 0; i < load_end; i++)
    {
		M_DrawSaveLoadBorder(x, y + LINEHEIGHT*i);
		M_WriteText2(x, y + 3 + LINEHEIGHT*i, savegamestrings[i],
			hu_font_a, 1, 0, 0);
    }
    if (saveStringEnter)
    {
		i = M_StringWidth(savegamestrings[saveSlot], usefont);
		M_WriteText(x + i, y + 3 + LINEHEIGHT*saveSlot, "_");
    }
}

//
// M_Responder calls this when user is finished
//
void M_DoSave(int slot)
{
    G_SaveGame(slot, savegamestrings[slot]);
	menu_active = 0;
	
    // PICK QUICKSAVE SLOT YET?
    if (quickSaveSlot == -2)
		quickSaveSlot = slot;
}

//
// User wants to save. Start string input for M_Responder
//
static void M_SaveSelect(int choice)
{
	// we are going to be intercepting all chars
	saveStringEnter = 1;
	saveSlot = choice;
	
	strcpy(saveOldString,savegamestrings[choice]);

	if (!strcmp(savegamestrings[choice], "EMPTY SLOT"))
		savegamestrings[choice][0] = 0;

	saveCharIndex = strlen(savegamestrings[choice]);
}

//
// Selected from DOOM menu
//
static void M_SaveGame (int choice)
{
    if (R3D_edit || gamestate != GS_LEVEL)
    {
		S_LocalMenuSound(sfx_oof);
		return;
    }
	
    if (  !usergame
		|| mbf_players[0].plr->mo->health <= 0)
    {
		M_StartMessage(SAVEDEAD, NULL, false);
		return;
    }

    M_SetupNextMenu(&SaveDef);
    M_ReadSaveStrings();
}

//
//      M_QuickSave
//
char    tempstring[80];

void M_QuickSaveResponse(int ch)
{
	if (R3D_edit) return;

    if (ch == 'y')
    {
		M_DoSave(quickSaveSlot);
		S_LocalMenuSound(sfx_swtchx);
    }
}

static void M_QuickSave(void)
{
	if (R3D_edit || !usergame || gamestate != GS_LEVEL)
    {
		S_LocalMenuSound(sfx_oof);
		return;
    }
	
	if (mbf_players[0].plr->mo->health <= 0)
	{
		M_StartMessage(SAVEDEAD, NULL, false);
		return;
	}
	
    if (quickSaveSlot < 0)
    {
		M_StartControlPanel();
		M_ReadSaveStrings();
		M_SetupNextMenu(&SaveDef);
		quickSaveSlot = -2;	// means to pick a slot now
		return;
    }
    sprintf(tempstring,QSPROMPT,savegamestrings[quickSaveSlot]);
    M_StartMessage(tempstring,M_QuickSaveResponse,true);
}

//
// M_QuickLoad
//
void M_QuickLoadResponse(int ch)
{
    if (ch == 'y')
    {
		M_LoadSelect(quickSaveSlot);
		S_LocalMenuSound(sfx_swtchx);
    }
}

static void M_QuickLoad(void)
{
    if (R3D_edit)
    {
		S_LocalMenuSound(sfx_oof);
		return;
    }

    if (quickSaveSlot < 0)
    {
		M_StartMessage(QSAVESPOT,NULL,false);
		return;
    }
    sprintf(tempstring,QLPROMPT,savegamestrings[quickSaveSlot]);
    M_StartMessage(tempstring,M_QuickLoadResponse,true);
}

//
// Read This Menus - v220-22 tidied up
//
static void M_DrawReadThis(void)
{
	// Make static so search is done once
	static int picnum = -1;
	
	if (picnum < 0)
	{
		if ( (picnum = W_CheckNumForName("HELP1")) < 0)
			picnum = W_CheckNumForName("HELP");
	}

	GL_DrawPagePatch(picnum);
}

//
//	M_NewGame
//
static void M_NewGame(int choice)
{	
    if (commercial_game && !numepi)
		M_SetupNextMenu(&NewDef);
    else
		M_SetupNextMenu(&EpiDef);
}

static void M_DrawNewGame(void)
{
    GL_DrawPatch(96,14,W_GetNumForName("M_NEWG"));
    GL_DrawPatch(54,38,W_GetNumForName("M_SKILL"));
}

//
//	M_DrawMainMenu
//
static void M_DrawMainMenu(void)
{
	static int n = 0;

	if (!n) n = W_GetNumForName("M_DOOM");

    GL_DrawPatch(94, 2, n);
}

//
//	M_Episode
//
static void M_DrawEpisode(void)
{
	GL_DrawPatch(96, 14, W_GetNumForName("M_NEWG"));
	M_DrawTitle(episodemsg, 40);
}

static int     epi_choice;

static void M_VerifyNightmare(int ch)
{
    if (ch != 'y')
	return;
	
	if (commercial_game)
		G_DeferedInitNew(sk_nightmare, 0, 1);
	else
	{
		int e = epi_choice + 1;

		if (numepi) e = R3D_EpisodeNum(e);

		G_DeferedInitNew(sk_nightmare, e, 1);
	}

	menu_active = 0;
}

static void M_ChooseSkill(int choice)
{
    if (choice == sk_nightmare)
    {
		M_StartMessage(NIGHTMARE,M_VerifyNightmare,true);
		return;
    }
	
	if (commercial_game)
	{
		if (numepi)
			G_DeferedInitNew(choice, 0, episodeinf[epi_choice][0]);
		else
			G_DeferedInitNew(choice, 0, 1);
	}
	else
	{
		int e = epi_choice + 1;
		int n = 1;

		if (numepi) e = R3D_EpisodeNum(e);

		if (haslevelExMx[e*10]) n = 0;
	
		G_DeferedInitNew(choice, e, n);
	}

	menu_active = 0;
}

static void M_Episode(int choice)
{
    if(gamemode == shareware && choice)
    {
		M_StartMessage(SWSTRING, NULL, false);
		M_SetupNextMenu(&ReadDef1);
		return;
    }

    // Yet another hack...
    if(gamemode == registered && choice > 2)
    {
		Con_Message("M_Episode: 4th episode requires UltimateDOOM\n");
		choice = 0;
    }
	
    epi_choice = choice;
    M_SetupNextMenu(&NewDef);
}

//
// M_Options
//
char    detailNames[2][9]	= {"M_GDHIGH","M_GDLOW"};
char	msgNames[2][9]		= {"M_MSGOFF","M_MSGON"};

static void M_DrawOptions(void)
{
	M_DrawTitle("OPTIONS", 78);
}

static void M_DrawGameplay(void)
{
	Menu_t *menu = &GameplayDef;
	RDBOOL temp = R3D_edit || R3D_ScriptedMap;

	M_DrawTitle("GAMEPLAY OPTIONS", menu->y-20);

	M_WriteMenuText(menu, 0, temp || ma_inf.noscriptset & NSF_SCPLIGHT ? mapset : yesno[cfg.jumpEnabled != 0]);
	M_WriteMenuText(menu, 1, yesno[showMessages != 0]); 
	M_WriteMenuText(menu, 2, yesno[cfg.alwaysRun != 0]);
	M_WriteMenuText(menu, 3, yesno[cfg.lookSpring != 0]);
	// v220-16
#if 0		// $bulletrange
	M_WriteMenuText(menu, 4, bullet_dist[cfg.bulletRange]);
	// v220-4
	M_WriteMenuText(menu, 5, temp || ma_inf.noscriptset & NSF_FALLINGDAMAGE ? mapset : yesno[cfg.fallingDamage != 0]);
	M_WriteMenuText(menu, 6, temp || ma_inf.noscriptset & NSF_WATERDYNAMICS ? mapset : yesno[waterdynamics != 0]);
	M_WriteMenuText(menu, 7, temp || ma_inf.noscriptset & NSF_UWATERDAMAGE ? mapset : yesno[cfg.underwaterDamage != 0]);
#else
	M_WriteMenuText(menu, 4, temp || ma_inf.noscriptset & NSF_FALLINGDAMAGE ? mapset : yesno[cfg.fallingDamage != 0]);
	M_WriteMenuText(menu, 5, temp || ma_inf.noscriptset & NSF_WATERDYNAMICS ? mapset : yesno[waterdynamics != 0]);
	M_WriteMenuText(menu, 6, temp || ma_inf.noscriptset & NSF_UWATERDAMAGE ? mapset : !waterdynamics ? "OFF" : yesno[cfg.underwaterDamage != 0]);
#endif
}

//===========================================================================
// M_DrawHUD
//===========================================================================
static void M_DrawHUD(void)
{
	Menu_t *menu = &HUDDef;
	char *xhairnames[NUM_XHAIRS + 1] = { "NONE", "CROSS", "ANGLES", "SQUARE",
		"OPEN SQUARE", "DIAMOND", "V" };
	
	M_DrawTitle("HUD OPTIONS", menu->y - 20);

	M_WriteMenuText(menu, 0, yesno[cfg.hudShown[HUD_HEALTH]]); 
	M_WriteMenuText(menu, 1, yesno[cfg.hudShown[HUD_AMMO]]); 
	M_WriteMenuText(menu, 2, yesno[cfg.hudShown[HUD_KEYS]]); 
	M_WriteMenuText(menu, 3, yesno[cfg.hudShown[HUD_ARMOR]]); 
	M_DrawSlider   (menu, 4, 10, (int)(10*(cfg.hudScale - 0.3f)/0.5f + .5f), 0);
	M_DrawSlider   (menu, 5, 11, (int)(cfg.hudColor[0]*10 + .5f), 0);
	M_DrawSlider   (menu, 6, 11, (int)(cfg.hudColor[1]*10 + .5f), 0);
	M_DrawSlider   (menu, 7, 11, (int)(cfg.hudColor[2]*10 + .5f), 0);
	M_WriteMenuText(menu, 8, xhairnames[cfg.xhairType]); 
	M_DrawSlider   (menu, 9, 9, cfg.xhairSize, 0);
}

static void M_DrawMouseOpts(void)
{
	Menu_t *menu = &InputDef;
	char *axisname[5] = { "NOT USED", "MOVE", "TURN", "STRAFE", "LOOK" };

	M_DrawTitle("INPUT OPTIONS", menu->y - 20);

	M_WriteMenuText(menu, 0, yesno[cfg.dclickuse]);
	M_WriteMenuText(menu, 1, yesno[cfg.usemlook]);
	M_WriteMenuText(menu, 2, yesno[cfg.mlookInverseY]);
	M_DrawSlider(menu, 3, 21, cfg.mouseSensiX/2, 0);
	M_DrawSlider(menu, 4, 21, cfg.mouseSensiY/2, 0);

	if (InputDef.itemCount > 5)	// Don't draw joystick stuff unless one is connected
	{
		M_WriteMenuText(menu, 6, yesno[cfg.joyEnable]);
		if (InputDef.itemCount > 7)	// Don't draw joystick stuff unless enabled
		{
			M_WriteMenuText(menu, 7, axisname[cfg.joyaxis[0]]);
			M_WriteMenuText(menu, 8, axisname[cfg.joyaxis[1]]);
			M_WriteMenuText(menu, 9, axisname[cfg.joyaxis[2]]);
			M_WriteMenuText(menu, 10, yesno[cfg.usejlook]);
			M_WriteMenuText(menu, 11, yesno[cfg.jlookInverseY]);
			M_WriteMenuText(menu, 12, yesno[cfg.povLookAround]);
			M_DrawSlider(menu, 13, 10, cfg.joySensitivity, 0);
			M_DrawSlider(menu, 14, 21, cfg.joyDeadzone/4, 0);
		}
	}
}

static void M_Options(int choice)
{
    M_SetupNextMenu(&OptionsDef);
}

static void M_OpenDCP(int choice)
{
	menu_active = 0;

	switch (choice)
	{
	case MENU_DISPLAY: Con_Execute("panel Display", true); break;
	case MENU_VIDEO  : Con_Execute("panel Video"  , true); break;
	case MENU_SOUND  : Con_Execute("panel Audio"  , true); break;
	default: Con_Execute("panel", true); break;
	}
}

//
//      Toggle messages on/off
//
static void M_ChangeMessages(int choice)
{
    showMessages = 1 - showMessages;
	P_SetMessage(mbf_players/* + 0*/, !showMessages? MSGOFF : MSGON);
    message_dontmesswithme = true;
}

void M_SendMessage(char *msg)	// R3D
{
	static char buff[MAX_LINELEN];

	buff[0] = 0;

	if (msg)
	{
		strncpy(buff, msg, MAX_LINELEN - 2);
		buff[MAX_LINELEN - 1] = 0;
		P_SetMessage(mbf_players/* + 0*/, msg);
	}
}

static void M_AlwaysRun(int option)
{
	cfg.alwaysRun = !cfg.alwaysRun;
}

static void M_LookSpring(int option)
{
	cfg.lookSpring = !cfg.lookSpring;
}

#if 0		// $bulletrange
static void M_BulletRange(int option)
{
	if (cfg.bulletRange == 0)
	{
		cfg.bulletRange = 1;
		useBulletRange = MISSILERANGE;
	}
	else if (cfg.bulletRange == 1)
	{
		cfg.bulletRange = 2;
		useBulletRange = MAXBULLETRANGE;
	}
	else
	{
		cfg.bulletRange = 0;
		useBulletRange = STANDARDRANGE;
	}
}
#endif

static void M_AllowJump(int option)
{
	if (R3D_edit || R3D_ScriptedMap || ma_inf.noscriptset & NSF_SCPLIGHT) return;

	cfg.jumpEnabled = !cfg.jumpEnabled;

	if (cfg.jumpEnabled && !cfg.jumpPower)
		cfg.jumpPower = 9.f;

	ge_inf.flags &= ~GEF_JUMP;
	if (cfg.jumpEnabled) ge_inf.flags |= GEF_JUMP;
}

static void M_FallingDamage(int option)
{
	if (R3D_edit || R3D_ScriptedMap || ma_inf.noscriptset & NSF_FALLINGDAMAGE) return;

	cfg.fallingDamage = !cfg.fallingDamage;
	ge_inf.flags &= ~GEF_FALLINGDAMAGE;
	if (cfg.fallingDamage) ge_inf.flags |= GEF_FALLINGDAMAGE;
}

static void M_WaterDynamics(int option)
{
	if (R3D_edit || R3D_ScriptedMap || ma_inf.noscriptset & NSF_WATERDYNAMICS) return;

	waswaterdynamics = waterdynamics;
	waterdynamics = !waterdynamics;
	ge_inf.flags &= ~GEF_WATERDYNAMICS;
	if (waterdynamics) ge_inf.flags |= GEF_WATERDYNAMICS;
	if (!waterdynamics && ge_inf.mapset & GEF_UWDAMAGE)
	{
		ge_inf.flags &= ~GEF_UWDAMAGE;
		cfg.underwaterDamage= false;
	}
}

static void M_UnderwaterDamage(int option)
{
	if (R3D_edit || R3D_ScriptedMap || ma_inf.noscriptset & NSF_UWATERDAMAGE) return;

	cfg.underwaterDamage = !cfg.underwaterDamage;
	ge_inf.flags &= ~GEF_UWDAMAGE;
	if (cfg.underwaterDamage) ge_inf.flags |= GEF_UWDAMAGE;
}

static void M_HUDInfo(int option)
{
	cfg.hudShown[option] = (byte)(cfg.hudShown[option] == 0);
}

static void M_FloatMod10(float *variable, int option)
{
	int val = (int)((*variable + .05f) * 10);

	if(option == RIGHT_DIR) 
	{
		if(val < 10) val++;
	}
	else if(val > 0) val--;

	*variable = val/10.0f;
}

static void M_HUDScale(int option)
{
	// cfg.hudScale can vary from 0.3f to 0.8f
	float	value = (cfg.hudScale - 0.3f) / 0.5f;

	M_FloatMod10(&value, option);

	cfg.hudScale = (value + 0.3f) * 0.5f + 0.15f;
}

static void M_HUDRed(int option)
{
	M_FloatMod10(&cfg.hudColor[0], option);
}

static void M_HUDGreen(int option)
{
	M_FloatMod10(&cfg.hudColor[1], option);
}

static void M_HUDBlue(int option)
{
	M_FloatMod10(&cfg.hudColor[2], option);
}

static void M_XhairType(int option)
{
	if(option == RIGHT_DIR)
	{
		if(cfg.xhairType < NUM_XHAIRS) cfg.xhairType++;
	}
	else if(cfg.xhairType > 0) cfg.xhairType--;
}

static void M_XhairSize(int option)
{
	if(option == RIGHT_DIR)
	{
		if(cfg.xhairSize < 8) cfg.xhairSize++;
	}
	else if(cfg.xhairSize > 0) cfg.xhairSize--;
}

//
// M_EndGame
//
void M_EndGameResponse(int ch)
{
    if (ch != 'y')
	return;
		
    currentMenu->lastOn = itemOn;
	menu_active = 0;
    D_StartTitle ();
}

static void M_EndGame(int choice)
{
    choice = 0;
    if (!usergame)
    {
		S_LocalMenuSound(sfx_oof);
		return;
    }
	
    M_StartMessage(ENDGAME, M_EndGameResponse, true);
}

//
// M_ReadThis
//
static void M_ReadThis(int choice)
{
    choice = 0;
    M_SetupNextMenu(&ReadDef1);
}

static void M_ReadThis2(int choice)
{
    choice = 0;
    M_SetupNextMenu(&ReadDef2);
}

static void M_FinishReadThis(int choice)
{
    choice = 0;
    M_SetupNextMenu(&MainDef);
}

//
// M_QuitDOOM
//


int quitsounds[8] =
{
    sfx_pldeth,
    sfx_dmpain,
    sfx_popain,
    sfx_slop,
    sfx_telept,
    sfx_posit1,
    sfx_posit3,
    sfx_sgtatk
};

int quitsounds2[8] =
{
    sfx_vilact,
    sfx_getpow,
    sfx_boscub,
    sfx_slop,
    sfx_skeswg,
    sfx_kntdth,
    sfx_bspact,
    sfx_sgtatk
};

void M_QuitResponse(int ch)
{
	if (ch != 'y') return;

	if (sfx_avail)
	{
		int quitsound;
		
		// option quitsound - if not present teleport sound is used
		if (ArgExists("-qs"))
		{
			if (commercial_game)
				quitsound = quitsounds2[(gametic>>2)&7];
			else
				quitsound = quitsounds[(gametic>>2)&7];
		}
		else
		{
			quitsound = sfx_telept;
		}
			
		S_LocalMenuSound(quitsound);
		// 1.5 sec delay to exit prevents the exit sound being truncated.
		Sleep (1500);
	}
	
	Sys_Quit (false);
}

static void M_QuitDOOM(int choice)
{
	Con_Open(false);

	// v220-3 immediate quit with F10 removed as key could be hit accidentally
	if (R3D_edit && choice != -1) Sys_Quit(false);

	// We pick index 0 which is language sensitive,
	//  or one at random, between 1 and maximum number.

/*	only English supported
	if (language != english )
		sprintf(endstring,"%s\n\n%s", endmsg[0], DOSY );
	else*/
		sprintf(endstring,"%s\n\n%s", endmsg[(gametic%(NUM_QUITMESSAGES+1))], DOSY);

	M_StartMessage(endstring, M_QuitResponse, true);
}

static void M_AllowDoubleC(int option)
{
	cfg.dclickuse = !cfg.dclickuse;
}

static void M_MouseLook(int option)
{
	cfg.usemlook = !cfg.usemlook;
}

static void M_MouseLookInverse(int option)
{
	cfg.mlookInverseY = !cfg.mlookInverseY;
}

static void M_MouseXSensi(int option)
{
	if(option == RIGHT_DIR)
	{
		if(cfg.mouseSensiX < 39)
		{
			cfg.mouseSensiX+=2;
		}
	}
	else if(cfg.mouseSensiX > 1)
	{
		cfg.mouseSensiX-=2;
	}
}

static void M_MouseYSensi(int option)
{
	if(option == RIGHT_DIR)
	{
		if(cfg.mouseSensiY < 39)
		{
			cfg.mouseSensiY+=2;
		}
	}
	else if(cfg.mouseSensiY > 1)
	{
		cfg.mouseSensiY-=2;
	}
}

static void M_JoyEnable(int option)
{
	if (ge_inf.nojoystick) return;
	cfg.joyEnable = !cfg.joyEnable;
	InputDef.itemCount = cfg.joyEnable ? 15 : 7;	
}

static void M_JoyXAxis(int option)
{
	if(option == RIGHT_DIR)
	{
		if(cfg.joyaxis[0] < 4) cfg.joyaxis[0]++;
	}
	else
	{
		if(cfg.joyaxis[0] > 0) cfg.joyaxis[0]--;
	}
}

static void M_JoyYAxis(int option)
{
	if(option == RIGHT_DIR)
	{
		if(cfg.joyaxis[1] < 4) cfg.joyaxis[1]++;
	}
	else
	{
		if(cfg.joyaxis[1] > 0) cfg.joyaxis[1]--;
	}
}

static void M_JoyZAxis(int option)
{
	if(option == RIGHT_DIR)
	{
		if(cfg.joyaxis[2] < 4) cfg.joyaxis[2]++;
	}
	else
	{
		if(cfg.joyaxis[2] > 0) cfg.joyaxis[2]--;
	}
}

static void M_JoyLook(int option)
{
	cfg.usejlook = !cfg.usejlook;
}

static void M_InverseJoyLook(int option)
{
	cfg.jlookInverseY = !cfg.jlookInverseY;
}

static void M_POVLook(int option)
{
	cfg.povLookAround = !cfg.povLookAround;
}

static void M_JoySensitivity(int option)
{
	if (option == RIGHT_DIR)
	{
		if (cfg.joySensitivity < 9)
		{
			cfg.joySensitivity += 1;
		}
	}
	else if (cfg.joySensitivity > 0)
	{
		cfg.joySensitivity -= 1;
	}
}

static void M_JoyDeadzone(int option)
{
	if(option == RIGHT_DIR)
	{
		if (cfg.joyDeadzone < 80)
		{
			cfg.joyDeadzone += 4;
		}
	}
	else if (cfg.joyDeadzone > 1)
	{
		cfg.joyDeadzone -= 4;
	}
}


//
//      Menu Functions
//

// Height is in pixels, and determines the scale of the whole thing.
// [left/right:6, middle:8, thumb:5]
void M_DrawThermo2(int x, int y, int thermWidth, int thermDot, int height)
{
	float	ht = (float)height;
	float	scale = ht/13.0f;	// 13 is the normal scale.
    float	xx = (float)x;
    float	yy = (float)y;

	GL_SetPatch(W_GetNumForName("M_THERML"), 0);
	GL_DrawRect(xx, yy, 6*scale, ht, 1.0f, 1.0f, 1.0f, menu_alpha);

    xx += 6.0f*scale;
	GL_SetPatch(W_GetNumForName("M_THERM2"), 0); // in Risen3D.wad
	GL_DrawRectTiled(xx, yy, 8.0f*thermWidth*scale, ht, 8*scale, ht);

	xx += 8.0f*thermWidth*scale;
	GL_SetPatch(W_GetNumForName("M_THERMR"), 0);
	GL_DrawRect(xx, yy, 6*scale, ht, 1.0f, 1.0f, 1.0f, menu_alpha);

	xx = (float)x + (6.0f+thermDot*8.0f)*scale;
	GL_SetPatch(W_GetNumForName("M_THERMO"), 0);
	GL_DrawRect(xx, yy, 6*scale, ht, 1.0f, 1.0f, 1.0f, menu_alpha);
}


static void M_DrawSlider(Menu_t *menu, int index, int width, int dot, int offx)
{
	if(!offx)
	{
		if (menu->items[index].mntext)
		{
			offx = M_StringWidth(menu->items[index].mntext, menu->font);
			offx /= 4;
			offx *= 4;
		}
	}
	M_DrawThermo2(menu->x+6+offx, menu->y+menu->itemHeight*index,
		width, dot, menu->itemHeight-1);
}

static void M_StartMessage(char* string, void* routine, RDBOOL input)
{
    messageLastMenuActive = menu_active;
    messageToPrint = 1;
    messageString = string;
    messageRoutine = (void (__cdecl *)(int))routine;
    messageNeedsInput = input;
    menu_active = 1;
	typein_time = 0;
    return;
}

//
// Find string width from hu_font chars
//
int M_StringWidth(char* string, dpatch_t *font)
{
	unsigned i;
    int c, w = 0;
	
    for (i = 0; i < strlen(string); i++)
    {
		c = toupper(string[i]) - HU_FONTSTART;
		if (c < 0 || c >= HU_FONTSIZE)
			w += 4;
		else
			w += font[c].width;
    }
    return w;
}

//
//      Find string height from hu_font chars
//
static int M_StringHeight(char* string, dpatch_t *font)
{
	unsigned i;
    int h, height = font[0].height;
	
    h = height;
    for (i = 0;i < strlen(string);i++)
		if (string[i] == '\n')
			h += height;
	return h;
}

// Used to ensure the map name chars have been cached
void M_CacheText_hu_font_b(char* string)
{
	dpatch_t	*font = hu_font_b;
    char		*ch = string;

    while(1)
    {
		int c = *ch++;

		if (!c) break;

		c = toupper(c) - HU_FONTSTART;

		if (c >= 0 && c < HU_FONTSIZE)
			GL_SetPatch(font[c].lump, 0);
	}
}

//
//      Write a string possibly using a colored, custom font
//
void M_WriteText2(int x, int y, char* string, dpatch_t *font,
							float red, float green, float blue)
{
    char	*ch;
    int		c, cx, cy;
	int		count = 0, maxcount;

	// Disable type-in?
	if (cfg.menuEffects > 0 || typein_time >= 0xffff)
		maxcount = 0xffff;
	else
		maxcount = typein_time << 1;

	if (red >= 0) glColor4f(red, green, blue, menu_alpha);

    ch = string;
    cx = x;
    cy = y;
	
    while(1)
    {
		if ( !(c = *ch++) || ++count > maxcount) break;

		if (c == '\n') { cx = x; cy += 12; continue; }
		
		if (red >= 0)
		{
			if (count == maxcount || ge_inf.textbrighten) 
				glColor3f(1.0f, 1.0f, 1.0f);
			else if (count + 1 == maxcount)
				glColor4f((1+red)/2.0f, (1+green)/2.0f, (1+blue)/2.0f, menu_alpha);
		}

		if ( (c = toupper(c) - HU_FONTSTART) < 0 || c >= HU_FONTSIZE)
		{
			cx += 4;
		}
		else
		{
			GL_DrawPatch_CS((float)cx, (float)cy, font[c].lump, 0);
			cx += font[c].width;
		}
    }
}

// Menu text is white.
void M_WriteMenuText(Menu_t *menu, int index, char *text)
{
	int off = 0;

	if(menu->items[index].mntext)
		off = M_StringWidth(menu->items[index].mntext, menu->font) + 4;

	M_WriteText2(menu->x + off,
		menu->y + menu->itemHeight*index, text, 
		menu->font, 1, 1, 1);
}

void M_DrawTitle(char *text, int y)
{
	M_WriteText2(160 - M_StringWidth(text, hu_font_b)/2,
		y, text, hu_font_b,
		cfg.menuColor[0], cfg.menuColor[1], cfg.menuColor[2]);
}

//
//      Write a string using the hu_font
//
void M_WriteText (int x, int y, char* string)
{
	M_WriteText2(x, y, string, hu_font, 1, 1, 1);
}

//
// CONTROL PANEL
//

void Set_Menu(MenuType_e menu)
{
	M_SetupNextMenu(menulist[menu]);
}

//
// M_Responder
//

RDBOOL M_Responder (event_t* ev)
{
    static  int     joywait = 0;
    static  int     mousewait = 0;
    static  int     mousey = 0;
    static  int     lasty = 0;
    static  int     mousex = 0;
    static  int     lastx = 0;

	MenuItem_t		*item;
    int             ch;
    int             i;
	int				firstVI, lastVI;	// first and last visible item

	// -----------------------------

	if (Edit_CheckEditstate(ev))
		return true;

    // Pop-up menu?
    if (!menu_active)
    {
		if (return_to_menu && !chat_on)
		{
			return_to_menu = false;
			M_StartControlPanel();
			return true;
		}
    }

	if(ev->data1 == DDKEY_RSHIFT)
	{
		shiftdown = (ev->type == ev_keydown || ev->type == ev_keyrepeat);
	}

    ch = -1;
	
    if (ev->type == ev_joystick && joywait < Sys_GetTime())
    {
		if (ev->data3 == -1)
		{
			ch = DDKEY_UPARROW;
			joywait = Sys_GetTime() + 5;
		}
		else if (ev->data3 == 1)
		{
			ch = DDKEY_DOWNARROW;
			joywait = Sys_GetTime() + 5;
		}
		
		if (ev->data2 == -1)
		{
			ch = DDKEY_LEFTARROW;
			joywait = Sys_GetTime() + 2;
		}
		else if (ev->data2 == 1)
		{
			ch = DDKEY_RIGHTARROW;
			joywait = Sys_GetTime() + 2;
		}
		
		if (ev->data1&1)
		{
			ch = DDKEY_ENTER;
			joywait = Sys_GetTime() + 5;
		}
		if (ev->data1&2)
		{
			ch = DDKEY_BACKSPACE;
			joywait = Sys_GetTime() + 5;
		}
    }
    else
    {
		if (ev->type == ev_keydown || ev->type == ev_keyrepeat)
		{
			ch = ev->data1;
		}
    }
    
    if (ch == -1)
		return false;
	
    
    // Save Game string input
    if (saveStringEnter)
    {
		switch(ch)
		{
		case DDKEY_BACKSPACE:
			if (saveCharIndex > 0)
			{
				saveCharIndex--;
				savegamestrings[saveSlot][saveCharIndex] = 0;
			}
			break;
			
		case DDKEY_ESCAPE:
			saveStringEnter = 0;
			strcpy(&savegamestrings[saveSlot][0],saveOldString);
			break;
			
		case DDKEY_ENTER:
			saveStringEnter = 0;

			if (savegamestrings[saveSlot][0])
				M_DoSave(saveSlot);
			break;
			
		default:
			if ( (ch = toupper(ch)) != 32)
			{
				if (ch-HU_FONTSTART < 0 || ch-HU_FONTSTART >= HU_FONTSIZE)
					break;
			}

			if (ch >= 32 && ch <= 127 &&
				saveCharIndex < SAVESTRINGSIZE-1 &&
				M_StringWidth(savegamestrings[saveSlot], hu_font) <
				(SAVESTRINGSIZE-2)*8)
			{
				savegamestrings[saveSlot][saveCharIndex++] = (char)ch;
				savegamestrings[saveSlot][saveCharIndex] = 0;
			}
			break;
		}
		return true;
    }
    
    // Take care of any messages that need input
    if (messageToPrint)
    {
		if (messageNeedsInput == true &&
			!(ch == ' ' || ch == 'n' || ch == 'y' || ch == DDKEY_ESCAPE))
			return false;
		
		menu_active = messageLastMenuActive;
		messageToPrint = 0;
		if (messageRoutine) messageRoutine(ch);
		
		menu_active = 0;
		S_LocalMenuSound(sfx_swtchx);
		return true;
    }
	
    // Pop-up menu?
    if (!menu_active)
    {
		if (ch == DDKEY_ESCAPE && !chat_on)
		{
			M_StartControlPanel();
			return true;
		}
		return false;
    }

	firstVI = currentMenu->firstItem; 
	lastVI = firstVI + currentMenu->numVisItems-1;	
	if(lastVI > currentMenu->itemCount-1) 
		lastVI = currentMenu->itemCount-1;
	item = &currentMenu->items[itemOn];
	currentMenu->lastOn = itemOn;

    // Keys usable within menu
    switch (ch)
    {
	case DDKEY_DOWNARROW:
		i = 0;
		do
		{
			if(itemOn+1 > lastVI)
			{
				itemOn = (short)firstVI;
			}
			else
			{
				itemOn++;
			}
		} while(currentMenu->items[itemOn].type == ITT_EMPTY
			&& i++ < currentMenu->itemCount);
		menu_color = 0;
		S_LocalMenuSound(sfx_pstop);
		return(true);
		
	case DDKEY_UPARROW:
		i = 0;
		do
		{
			if(itemOn <= firstVI)
			{
				itemOn = (short)lastVI;
			}
			else
			{
				itemOn--;
			}
		} while(currentMenu->items[itemOn].type == ITT_EMPTY
			&& i++ < currentMenu->itemCount);
		menu_color = 0;
		S_LocalMenuSound(sfx_pstop);
		return(true);
		
	case DDKEY_LEFTARROW:
		if(item->type == ITT_LRFUNC && item->func != NULL)
		{
			item->func(LEFT_DIR);
			S_LocalMenuSound(sfx_stnmov);
		}
		else
		{
			// Let's try to change to the previous page.
			if(currentMenu->firstItem - currentMenu->numVisItems >= 0)
			{
				currentMenu->firstItem -= currentMenu->numVisItems;
				itemOn = (short)(itemOn - currentMenu->numVisItems);

				//Ensure cursor points to editable item		Anon $09 add next 3 lines
				firstVI = currentMenu->firstItem;
				while(currentMenu->items[itemOn].type == ITT_EMPTY
					&& itemOn > firstVI) itemOn--;

				// Make a sound, too.
				S_LocalMenuSound(sfx_stnmov);
			}
		}
		return(true);
		
	case DDKEY_RIGHTARROW:
		if(item->type == ITT_LRFUNC && item->func != NULL)
		{
			item->func(RIGHT_DIR);
			S_LocalMenuSound(sfx_stnmov);
		}
		else
		{
			// Move on to the next page, if possible.
			if(currentMenu->firstItem + currentMenu->numVisItems < 
				currentMenu->itemCount)
			{
				currentMenu->firstItem += currentMenu->numVisItems;
				itemOn = (short)(itemOn + currentMenu->numVisItems);
				if(itemOn > currentMenu->itemCount-1)
					itemOn = (short)(currentMenu->itemCount-1);
				S_LocalMenuSound(sfx_stnmov);						
			}
		}
		return(true);
		
	case DDKEY_ENTER:
		if(item->type == ITT_SETMENU)
		{
			M_SetupNextMenu(menulist[item->option]);
			S_LocalMenuSound(sfx_pistol);
		}
		else if(item->func != NULL)
		{
			currentMenu->lastOn = itemOn;

			if(item->type == ITT_LRFUNC)
			{
				item->func(RIGHT_DIR);
				S_LocalMenuSound(sfx_stnmov);
			}
			else if(item->type == ITT_EFUNC)
			{
				item->func(item->option);
				S_LocalMenuSound(sfx_pistol);
			}
			else if(item->type == ITT_CPFUNC)
			{
				return_to_menu = true;
				item->func(item->option);
			}
		}
		return(true);
		
	case DDKEY_ESCAPE:
		currentMenu->lastOn = itemOn;
		menu_active = 0;
		S_LocalMenuSound(sfx_swtchx);
		return(true);
		
	case DDKEY_BACKSPACE:
		currentMenu->lastOn = itemOn;
		if(currentMenu->prevMenu == MENU_NONE)
		{
			currentMenu->lastOn = itemOn;
			menu_active = 0;
			S_LocalMenuSound(sfx_swtchx);
		}
		else
		{
			//					SetMenu(currentMenu->prevMenu);
			currentMenu = menulist[currentMenu->prevMenu];
			itemOn = (short)currentMenu->lastOn;
			S_LocalMenuSound(sfx_swtchn);
			typein_time = 0;
		}
		return(true);
		
	default:
		for(i = firstVI; i <= lastVI; i++)
		{
			if (   currentMenu->items[i].mntext
				&& currentMenu->items[i].type != ITT_EMPTY)
			{
				if (toupper(ch) == toupper(currentMenu->items[i].mntext[0]))
				{
					itemOn = (short)i;
					return(true);
				}
			}
		}
		break;
	}
	
    return false;
}



//
// M_StartControlPanel
//
void M_StartControlPanel (void)
{
    // intro might call this repeatedly
    if (menu_active) return;
    
	Con_Open(false);
    menu_active = 1;
	menu_color = 0;
	skull_angle = 0;
    currentMenu = &MainDef;         // JDC
    itemOn = (short)currentMenu->lastOn;   // JDC
	typein_time = 0;
	S_LocalMenuSound(sfx_swtchn);
}

void M_DrawBackground(void)
{
	int i;

	if(cfg.menuEffects > 1) return;

	if(cfg.menuFog == 0)
	{
		glDisable(GL_TEXTURE_2D);
		glColor4f(mfAlpha, mfAlpha/2, .0f, mfAlpha/3);
		glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
		GL_DrawRectTiled(0, 0, SCREENWIDTH, SCREENHEIGHT, 1, 1);
		glEnable(GL_TEXTURE_2D);
	}

	GL_Bind(G_LoadMenuFog());
	glColor3f(mfAlpha, mfAlpha, mfAlpha);
	glMatrixMode(GL_TEXTURE);
	for(i = 0; i < 2; i++)
	{
		if(i || cfg.menuFog == 1)
		{
			glColor3f(mfAlpha, mfAlpha, mfAlpha);
			glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
		}
		else
		{
			glColor3f(mfAlpha/5, mfAlpha/3, mfAlpha/2);
			glBlendFunc(GL_SRC_ALPHA, GL_SRC_ALPHA);
		}
		glLoadIdentity();
		glTranslatef(mfPos[i][VX]/320, mfPos[i][VY]/200, 0);
		glRotatef(mfAngle[i], 0, 0, 1);
		glTranslatef(-mfPos[i][VX]/320, -mfPos[i][VY]/200, 0);
		if(cfg.menuFog == 0)
			GL_DrawRectTiled(0, 0, SCREENWIDTH, SCREENHEIGHT, 270/8, 4*225);
		else
			GL_DrawRectTiled(0, 0, SCREENWIDTH, SCREENHEIGHT, 270, 225);
	}
	glLoadIdentity();	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

static void msgToPrint()
{
    char		string[40];
    int			start;
	short		i, x, y;

    // Horiz. & Vertically center string and print it.
	ReplaceInString(messageString, "\\n", "\n");

	start = 0;
	y = (short)(100 - M_StringHeight(messageString, usefont)/2);

	while( *(messageString + start) )
	{
		memset(string, 0, 40);

		for (i = 0; (unsigned)i < strlen(messageString+start);i++)
		{
			if (*(messageString + start + i) == '\n')
			{
				// Prevent users defining quit messages
				// that could crash the engine.
				strncpy(string, messageString+start, i > 39 ? 39 : i);
				start += i + 1;
				break;
			}
		}

		if (!string[0])
		{
			strncpy(string, messageString+start, i > 39 ? 39 : i);
			start += i;
		}
		
		x = (short)(160 - M_StringWidth(string, usefont)/2);
		M_WriteText2(x, y, string, hu_font_a, cfg.menuColor[0], 
			cfg.menuColor[1], cfg.menuColor[2]);
		y = (short)(y + usefont[0].height + 1);
	}
}

static RDBOOL menuNeeded()
{
	// b211-0 Simplify and correct help screen display
    if (menu_active)
	{
		if (currentMenu->drawFunc == M_DrawReadThis)
		{
			currentMenu->drawFunc();
			return false;	// Nothing more needed
		}
	}

	// Draw menu background.
	if (mfAlpha) M_DrawBackground();

	// Setup matrix.
	if (messageToPrint || menu_active)
	{
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		// Scale by the menuScale.
		glTranslatef(160, 100, 0);
		glScalef(cfg.menuScale, cfg.menuScale, 1.0f);
		glTranslatef(-160, -100, 0);

		if (messageToPrint)
		{
			msgToPrint();
			// Restore original matrix.	
			glMatrixMode(GL_MODELVIEW);
			glPopMatrix();
			return false;
		}

		return true;
	}

	return false;
}

//
// M_Drawer
// Called after the view has been rendered,
// but before it has been blitted.
//
void M_Drawer (void)
{
	if (menuNeeded())
	{
		short		c, i, x, y;
		short		max;
		int			w, h, off_x, off_y, cpos;
		// R3D
		int			offset;
		int			skulln;
		float		lt, scale;
		float		r, g, b;


		if (currentMenu->drawFunc)
			currentMenu->drawFunc();         // call Draw routine
    
		// DRAW MENU
		x = (short)currentMenu->x;
		y = (short)currentMenu->y + fiddlefactor;
		max = (short)currentMenu->itemCount;

		if (fiddlefactor && currentMenu->items == MainItems)
			y += fiddlefactor;
		
		offset = currentMenu->itemHeight;
		off_y = y + offset/2;
		cpos = itemOn - currentMenu->firstItem;

		for (c = 0, i = (short)currentMenu->firstItem; 
			i < max && i < currentMenu->firstItem + currentMenu->numVisItems; 
			i++, c++)
		{
			int n;

			if (itemOn == i)
			{
				if ( (lt = menu_color/100.f) < 0.5f) lt = 1.f - lt;
				off_y += (offset * c);
			}
			else
			{
				lt = 1.0f;
			}

			if ( (n = currentMenu->items[i].mnlumpnum) > 0)
			{
				if (currentMenu->items[i].id == LMP_PATCH)
				{
					GL_DrawPatchLitAlpha(x, y, lt, 1.f, n);
				}
				else	// hi-res
				{
					GL_SetPatch(n, 0);

					r = cfg.menuColor[0] * lt;
					g = cfg.menuColor[1] * lt;
					b = cfg.menuColor[2] * lt;

					w = lumptexinfo[n].origw;
					h = lumptexinfo[n].origh;

					GL_DrawRect(
						(float)x,
						(float)y,
						(float)w,
						(float)h,
						r,
						g,
						b,
						1.0f);
				}
			}
			else
			if (currentMenu->items[i].mntext)
			{
				float r, g, b;

				// Which color?
				if(currentMenu->items[i].type == ITT_EMPTY)
				{
					r = 1.0f;
					g =  .7f;
					b =  .3f;
				}
				else if (haspwads
					&& currentMenu->drawFunc == M_DrawEpisode
					&& !epihaspwad[i + 1])
				{
					r = .3f;
					g = .3f;
					b = .3f;
				}
				else
				{
					r = cfg.menuColor[0] * lt;
					g = cfg.menuColor[1] * lt;
					b = cfg.menuColor[2] * lt;
				}

				M_WriteText2(x, 
						y + currentMenu->itemHeight - currentMenu->font[0].height - 1, 
						currentMenu->items[i].mntext,
						currentMenu->font, r, g, b);
			}

			// v220-16
			if (!c && currentMenu == &ControlsDef)
				M_WriteText2(x + 129, 
					y + currentMenu->itemHeight - currentMenu->font[0].height - 1, 
					"[F4 IS RESERVED]",
					currentMenu->font, 153.f/255.f, 102.f/255.f,  31.f/255.f);

			y += offset;
		}
		
		// DRAW SKULL
		skulln = skullnum[ (whichSkull & 1) ];	// Can only be 0 or 1
		scale = currentMenu->itemHeight / (float) LINEHEIGHT;
		w = (int)(lumptexinfo[skulln].origw*scale); // skull size
		h = (int)(lumptexinfo[skulln].origh*scale);
		off_x = x + w/2 + (int)(SKULLXOFF*scale);

		glTranslatef((float)off_x, (float)off_y, 0);
		if (skull_angle) glRotatef(skull_angle, 0, 0, 1);

		GL_SetPatch(skulln, 0);
		GL_DrawRect((float)(-w/2), (float)(-h/2), (float)w, (float)h, 1, 1, 1, menu_alpha);

		// Restore original matrix (current matrix was set in menuNeeded())
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}
}

//
// M_SetupNextMenu
//
static void M_SetupNextMenu(Menu_t *menudef)
{
	if(!menudef) return;
    currentMenu = menudef;
    itemOn = (short)currentMenu->lastOn;
	menu_color = 0;
	skull_angle = 0;
	typein_time = 0;
}

//
// M_Ticker
//
void M_Ticker (void)
{
	typein_time++;

	if(menu_active)
	{
		int	i;

		for (i = 0; i < 2; i++) 
		{
			if(cfg.menuFog == 1)
			{
				mfAngle[i] += mfSpeeds[i]/4;
				mfPosAngle[i] -= mfSpeeds[!i];
				mfPos[i][VX] = (float)(160 + 120.f*(float)cos((double)(mfPosAngle[i]/180.f*PI)));
				mfPos[i][VY] = (float)(100 + 100.f*(float)sin((double)(mfPosAngle[i]/180.f*PI)));
			}
			else
			{
				mfAngle[i] += mfSpeeds[i]/2;
				mfPosAngle[i] -= 3*mfSpeeds[!i];
				mfPos[i][VX] = (float)(320 + 320.f*(float)cos((double)(mfPosAngle[i]/180.f*PI)));
				mfPos[i][VY] = (float)(240 + 240.f*(float)sin((double)(mfPosAngle[i]/180.f*PI)));
			}
		}

		if(mfAlpha < 1) mfAlpha += .1f;
		else
		if(mfAlpha > 1) mfAlpha = 1;
	}
	else 
	{
		if(mfAlpha > 0) mfAlpha -= .1f;
		else
		if(mfAlpha < 0) mfAlpha = 0;
	}

    if (--skullAnimCounter <= 0)
    {
		whichSkull ^= 1;
		skullAnimCounter = 8;
    }

	if (menu_active)
	{
		float rewind = 20;

		MenuTime++;

		menu_color += cfg.flashspeed;
		if(menu_color >= 100) menu_color -= 100;

		if(cfg.turningSkull && currentMenu->items[itemOn].type == ITT_LRFUNC)
			skull_angle += 5;
		else if(skull_angle != 0)
		{
			if(skull_angle <= rewind || skull_angle >= 360-rewind)
				skull_angle = 0;
			else if(skull_angle < 180) 
				skull_angle -= rewind;
			else 
				skull_angle += rewind;
		}
		if(skull_angle >= 360) skull_angle -= 360;
	}
}

//
// M_Init
//
// v220-1a extend pwad graphics support. Some tidying up
void M_Init (void)
{
	int	i, w, maxw;

	usefont = ge_inf.hadpwadplaypal ? hu_font : hu_font_a;

	if (ge_inf.nojoystick) InputDef.itemCount = 5;
	else
	if (!cfg.joyEnable) InputDef.itemCount = 7;
	
	numepi = MI_SetEpisodes(); // MAPINFO episode check/setup
	M_SetNumItems(&EpiDef, numepi);

	// Quit messages.	
	endmsg[0] = GET_TXT(TXT_QUITMSG);

	for(i = 1; i <= NUM_QUITMESSAGES; i++)
		endmsg[i] = GET_TXT(TXT_QUITMESSAGE1 + i-1);
	
	for(i = 0, maxw = 0; i < 9; i++)
	{
		EpisodeItems[i].mntext = GET_TXT(TXT_EPISODE1 + i);
		w = M_StringWidth(EpisodeItems[i].mntext, hu_font_b);
		if(w > maxw) maxw = w;
	}

	// Center the episodes menu appropriately.
	EpiDef.x = 160 - maxw/2 + 12;
	// "Choose Episode"
	episodemsg = GET_TXT(TXT_ASK_EPISODE);

    currentMenu = &MainDef;
    menu_active = 0;
    itemOn = (short)currentMenu->lastOn;
    whichSkull = 0;
    skullAnimCounter = 10;

    messageToPrint = 0;
    messageString = NULL;
    messageLastMenuActive = menu_active;
    quickSaveSlot = -1;
	
    switch ( gamemode )
    {
	case commercial:
		// does not have a "read this!" entry
		MainItems[4].mntext = MainItems[5].mntext;
		MainItems[4].func = MainItems[5].func;
		MainItems[4].mnlumpname = MainItems[5].mnlumpname;

		M_SetNumItems(&MainDef, 5);
		MainDef.y = 64 + 8;

		NewDef.prevMenu = MENU_MAIN; //&MainDef;

		ReadDef1.drawFunc = M_DrawReadThis;
		ReadDef1.x = 330;
		ReadDef1.y = 165;

		ReadItems1[0].func = M_FinishReadThis;
		break;
	case shareware:
		M_SetNumItems(&EpiDef, 1);
		MainDef.y = 64;
		break;
	case registered:
		M_SetNumItems(&MainDef, 6);
		// We need to remove the fourth episode.
		M_SetNumItems(&EpiDef, 3);
		MainDef.y = 64;
		break;
	case retail:
		// We are fine.
		if (!numepi) M_SetNumItems(&EpiDef, 4);
	default:
		break;
    }

	skullnum[0] = W_GetNumForName(skullName[0]);
	skullnum[1] = W_GetNumForName(skullName[1]);

	// fiddle factor for hr2final
	if (!_stricmp(analysename, "hr2final")) fiddlefactor = -3;

	int		id, k, n;
	Menu_t	*mt;

	if (episodeinf[0][1] >= 0)	// See if a MAPINFO image has been set
	{
		mt = &EpiDef;

		for (i = 0; i < 9; i++)
		{
			if (episodeinf[i][1] >= 0) mt->items[i].mnlumpnum = episodeinf[i][1];
		}
	}

	for (i = 0; i < 2; i++)
	{
		// The following are the only two where custom patches are looked
		// for because most of the other R3D menus have entries with no
		// corresponding Doom lump.
		if (!i) mt = &MainDef; else mt = &NewDef;

		for (k = 0; k < mt->itemCount; k++)
		{
			char *nm = mt->items[k].mnlumpname;

			if (  (n = W_CheckPWADNumForName(nm)) > 0
				&& ((id = DoomLumpIdentify(n, NULL, NULL,DLI_GRAPHIC)) == LMP_PATCH
				|| id == LMP_PNG
				|| id == LMP_JPG))
			{
				mt->items[k].mnlumpnum = n;
				mt->items[k].id = id;
			}
			else if (!(hiresPatches && stricmp(nm, "M_NMARE"))	// v220-35 use m_nmare hi-res patch
				&& (n = W_CheckIWADNumForName(nm)) > 0)
			{
				mt->items[k].mnlumpnum = n;
				mt->items[k].id = LMP_PATCH;
			}
		}
	}
}

int CCmdMenuAction(int argc, char **argv)
{
	if(!_stricmp(argv[0], "HelpScreen")) // F1
	{
		M_StartControlPanel ();
		
		if (gamemode == commercial)		// Everything other than Doom
		{
			currentMenu = &ReadDef2;
		}
		else
		{
			currentMenu = &ReadDef1;
		}
		
		itemOn = 0;
	}
	else if(!_stricmp(argv[0], "SaveGame")) // F2
	{
		M_StartControlPanel();
		M_SaveGame(0);
	}
	else if(!_stricmp(argv[0], "LoadGame")) // F3
	{
		M_StartControlPanel();
		M_LoadGame(0);
	}
	else if(!_stricmp(argv[0], "SoundMenu")) // F4
	{
		M_OpenDCP(MENU_SOUND);
	}
	else if(!_stricmp(argv[0], "QuickSave")) // F6
	{
		S_LocalMenuSound(sfx_swtchn);
		M_QuickSave();
	}
	else if(!_stricmp(argv[0], "EndGame")) // F7
	{
		S_LocalMenuSound(sfx_swtchn);
		M_EndGame(0);
	}
	else if(!_stricmp(argv[0], "ToggleMsgs")) // F8
	{
		S_LocalMenuSound(sfx_swtchn);
		M_ChangeMessages(0);
	}
	else if(!_stricmp(argv[0], "QuickLoad")) // F9
	{
		S_LocalMenuSound(sfx_swtchn);
		M_QuickLoad();
	}
	else if(!_stricmp(argv[0], "quit")) // F10
	{
		S_LocalMenuSound(sfx_swtchn);
		M_QuitDOOM(-1);
	}
	else if(!_stricmp(argv[0], "ToggleGamma")) // F11
	{
		M_OpenDCP(MENU_DISPLAY);
	}
	return true;
}

void M_Close()
{
	currentMenu->lastOn = itemOn;
	menu_active = 0;
}
