//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 id Software, Inc.,
//  Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman,
//  Jaakko Kernen, G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_ceilng.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//		T_MoveCeiling revised (R3D flags SF_CEILSTOPSOUND
//      removed and SF2_STNMOVCEIL added)
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
//
// DESCRIPTION:  Ceiling animation (lowering, crushing, raising)
//
//-----------------------------------------------------------------------------

#include "doom/p_local.h"

#include "s_main.h"
#include "sys_timer.h"
#include "m_winmem.h"
#include "r3d_defs.h"

//
// CEILINGS
//

// the list of ceilings moving currently, including crushers
ceilinglist_t *activeceilings;


//
// T_MoveCeiling
//
void Con_Message(char *message, ...);

void T_CeilInStasis(void *data)
{
	return;
}

void T_MoveCeiling (ceiling_t* ceiling)
{
	result_e	res;
	int			dir = ceiling->direction;

	// v220-22 Sound for this flagged here else it may never play due to
	// the chance that res == pastdest could always be true when ceiling
	// to floor differentials are small.
	if (dir != 1 && dir != -1) return;	// in stasis

	res = T_MovePlane(	ceiling->sector,
						ceiling->speed,
						dir == 1 ? ceiling->topheight :ceiling->bottomheight,
						dir == 1 ? 0 : ceiling->crush,
						1,
						ceiling->direction);

	// R3D scripted sounds can be enabled/disabled
	if ( !(ceiling->sector->specFlags2 & SF2_NOMOVSOUND) )
	{
		switch(ceiling->type)
		{
		case silentCrushAndRaise:
			if (res == pastdest)
			{
				if (ceiling->topheight - ceiling->bottomheight <= 8 * FRACUNIT)
				{
					ceiling->sector->specFlags2 |= SF2_STNMOVCEIL;
				}
				else
				{
					S_SectorSound(ceiling->sector, sfx_pstop);
				}
			}
		case genSilentCrusher:
			break;
		default:
			ceiling->sector->specFlags2 |= SF2_STNMOVCEIL;
		}
	}

	if (ceiling->direction == 1)	// UP
	{
		if (res == pastdest)
		{
			switch(ceiling->type)
			{
			case raiseToHighest:
			case genCeiling:
				P_RemoveActiveCeiling(ceiling);
				break;

			// movers with texture change, change the texture then get removed
			case genCeilingChgT:
			case genCeilingChg0:
				ceiling->sector->special = ceiling->newspecial;
				//jff 3/14/98 transfer old special field as well
				ceiling->sector->oldspecial = ceiling->oldspecial;
				ceiling->sector->ceilingpic = ceiling->texture;
				P_RemoveActiveCeiling(ceiling);
				break;

			case genCeilingChg:
				ceiling->sector->ceilingpic = ceiling->texture;
				P_RemoveActiveCeiling(ceiling);
				break;

			case silentCrushAndRaise:
			case genSilentCrusher:
			case genCrusher:
			case fastCrushAndRaise:
			case crushAndRaise:
				ceiling->direction = -1;
				break;
				
			default:
				break;
			}
		}
	}
	else	// DOWN
	{			
		if (res == pastdest)
		{
			switch(ceiling->type)
			{
			case genSilentCrusher:
			case genCrusher:
				if (ceiling->oldspeed<CEILSPEED*3)
					ceiling->speed = ceiling->oldspeed;
				ceiling->direction = 1; //jff 2/22/98 make it go back up!
				break;
				
			// make platform stop at bottom of all crusher strokes
			// except generalized ones, reset speed, start back up
			case silentCrushAndRaise:
			case crushAndRaise:
				ceiling->speed = CEILSPEED;
			case fastCrushAndRaise:
				ceiling->direction = 1;
				break;
				
			// in the case of ceiling mover/changer, change the texture
			// then remove the active ceiling
			case genCeilingChgT:
			case genCeilingChg0:
				ceiling->sector->special = ceiling->newspecial;
				//jff add to fix bug in special transfers from changes
				ceiling->sector->oldspecial = ceiling->oldspecial;
			case genCeilingChg:
				ceiling->sector->ceilingpic = ceiling->texture;
				P_RemoveActiveCeiling(ceiling);
				break;
				
			// in other cases, just remove the active ceiling
			case lowerAndCrush:
			case lowerToFloor:
			case lowerToLowest:
			case lowerToMaxFloor:
			case genCeiling:
				P_RemoveActiveCeiling(ceiling);
				break;
				
			default:
				break;
			}
		}
		else // ( res != pastdest )
		{
			if (res == crushed)
			{
				switch(ceiling->type)
				{
				case genCrusher:  
				case genSilentCrusher:
					if (ceiling->oldspeed < CEILSPEED*3)
						ceiling->speed = CEILSPEED / 8;
					break;
				case silentCrushAndRaise:
				case crushAndRaise:
					ceiling->speed = CEILSPEED / 8;
					break;
				case lowerAndCrush:
					if ( !(ceiling->sector->specFlags2 & SF2_CEILNOSLOW) )
						ceiling->speed = CEILSPEED / 8;
					break;
					
				default:
					break;
				}
			}
		}
	}
	
	if (ceiling->sector->specFlags2 & SF2_STNMOVCEIL)
		S_SectorSoundRepeat(r3ds.stnmov_id, ceiling->sector);
}

//
// EV_DoCeiling
// Move a ceiling up/down and all around!
//
int EV_DoCeiling(line_t* line, ceiling_e type)
{
	ceiling_t	*ceiling;
	sector_t	*sec;
	int			secnum = -1;
	int			rtn = 0;
	
	//	Reactivate in-stasis ceilings...for certain types.
	switch(type)
	{
	case fastCrushAndRaise:
	case silentCrushAndRaise:
	case crushAndRaise:
		rtn = P_ActivateInStasisCeiling(line);
		break;
	default:
		break;
	}
	
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		sec = &sectors[secnum];
		
		// if ceiling already moving, don't start a second function on it
		if (P_SectorActive(ceiling_special,sec))  //jff 2/22/98
			continue;
		
		// new ceiling thinker
		rtn = 1;
		ceiling = (ceiling_t*)W_Calloc(sizeof(*ceiling), THINKID, 0);
		ceiling->thinker.function = (void (__cdecl *)(void))T_MoveCeiling;
		P_AddFNThinker (&ceiling->thinker);
		sec->ceilingdata = ceiling;
		ceiling->sector = sec;
		ceiling->crush = 0;
		
		switch(type)
		{
		case fastCrushAndRaise:
			ceiling->crush = 1;
			ceiling->topheight = sec->ceilingheight;
			ceiling->bottomheight = sec->floorheight + (8*FRACUNIT);
			ceiling->direction = -1;
			ceiling->speed = CEILSPEED * 2;
			break;
			
		case silentCrushAndRaise:
		case crushAndRaise:
			ceiling->crush = 1;
			ceiling->topheight = sec->ceilingheight;
		case lowerAndCrush:
		case lowerToFloor:
			ceiling->bottomheight = sec->floorheight;
			if (type != lowerToFloor)
				ceiling->bottomheight += 8*FRACUNIT;
			ceiling->direction = -1;
			ceiling->speed = CEILSPEED;
			break;
			
		case raiseToHighest:
			ceiling->topheight = P_FindHighestCeilingSurrounding(sec);
			ceiling->direction = 1;
			ceiling->speed = CEILSPEED;
			break;
			
		case lowerToLowest:
			ceiling->bottomheight = P_FindLowestCeilingSurrounding(sec);
			ceiling->direction = -1;
			ceiling->speed = CEILSPEED;
			break;
			
		case lowerToMaxFloor:
			ceiling->bottomheight = P_FindHighestFloorSurrounding(sec);
			ceiling->direction = -1;
			ceiling->speed = CEILSPEED;
			break;
		}
		
		ceiling->tag = sec->tag;
		ceiling->type = type;
		P_AddActiveCeiling(ceiling);
	}
	return rtn;
}

//////////////////////////////////////////////////////////////////////
//
// Active ceiling list primitives
//
/////////////////////////////////////////////////////////////////////

// jff 2/22/98 - modified Lee's plat code to work for ceilings
//
// The following were all rewritten by Lee Killough
// to use the new structure which places no limits
// on active ceilings. It also avoids spending as much
// time searching for active ceilings. Previously a 
// fixed-size array was used, with NULL indicating
// empty entries, while now a doubly-linked list
// is used.

//
// P_ActivateInStasisCeiling()
//
// Reactivates all stopped crushers with the right tag
//
// Passed the line reactivating the crusher
// Returns true if a ceiling reactivated
//
//jff 4/5/98 return if activated
int P_ActivateInStasisCeiling(line_t *line)
{
	ceilinglist_t	*cl;
	int				rtn = 0;
	
	for (cl=activeceilings; cl; cl=cl->next)
	{
		ceiling_t *ceiling = cl->ceiling;

		if (   ceiling->tag == line->tag
			&& ceiling->thinker.function == (void (__cdecl *)(void))T_CeilInStasis	// Function added
			&& ceiling->direction == 0)	// Not now strictly needed
		{
			ceiling->direction = ceiling->olddirection;
			ceiling->thinker.function = (void (__cdecl *)(void))T_MoveCeiling;

			//jff 4/5/98 return if activated
			rtn = 1;
		}
	}
	return rtn;
}

//
// EV_CeilingCrushStop()
//
// Stops all active ceilings with the right tag
//
// Passed the linedef stopping the ceilings
// Returns true if a ceiling put in stasis
//
int EV_CeilingCrushStop(line_t* line)
{
	ceilinglist_t	*cl;
	int				rtn = 0;
	
	for (cl=activeceilings; cl; cl=cl->next)
	{
		ceiling_t *ceiling = cl->ceiling;

		if (ceiling->direction != 0 && ceiling->tag == line->tag)
		{
			ceiling->olddirection = ceiling->direction;
			ceiling->direction = 0;
			ceiling->thinker.function = (void (__cdecl *)(void))T_CeilInStasis;

			// R3D stop stnmov sound
			if (ceiling->sector->specFlags2 & SF2_STNMOVCEIL)
				S_StopSound(r3ds.stnmov_id ,(mobj_t*)&ceiling->sector->soundorg);

			rtn = 1;
		}
	}
	return rtn;
}

//
// P_AddActiveCeiling()
//
// Adds a ceiling to the head of the list of active ceilings
//
void P_AddActiveCeiling(ceiling_t* ceiling)
{
	ceilinglist_t *list = (ceilinglist_t*)W_Calloc(sizeof(ceilinglist_t), LEVELID, 67);

	list->ceiling = ceiling;
	ceiling->list = list;

	if ( (list->next = activeceilings) )	// i.e. if not NULL when set
		list->next->prev = &list->next;

	list->prev = &activeceilings;
	activeceilings = list;
}

//
// P_RemoveActiveCeiling()
//
// Removes a ceiling from the list of active ceilings
// Stops any r3ds.stnmov_id sound at the origin
//
void P_RemoveActiveCeiling(ceiling_t* ceiling)
{
	ceilinglist_t *list = ceiling->list;

	if (ceiling->sector->specFlags2 & SF2_STNMOVCEIL)
	{
		ceiling->sector->specFlags2 &= ~SF2_STNMOVCEIL;
		S_StopSound(r3ds.stnmov_id, (mobj_t*)&ceiling->sector->soundorg);
	}

	ceiling->sector->planes[PLN_CEILING].speed = 0;
	ceiling->sector->ceilingdata = NULL;  //jff 2/22/98
	P_RemoveThinker(&ceiling->thinker);

	if ((*list->prev = list->next))
		list->next->prev = list->prev;

	W_Free(list, LEVELID, 68);
}

void P_NullActiveCeilings(void)
{
	// all LEVELID mem has already been reclaimed
	activeceilings = NULL;
}
