//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_enemy.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 14 (v220-14) 01/May/2012 gmj
// Revision 2.2.0 build 28 (v220-28) 04/Oct/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 34 (v220-34) 12/Nov/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
// DESCRIPTION:
//	Enemy thinking, AI.
//	Action Pointer Functions
//	that are associated with states/frames. 
//
//-----------------------------------------------------------------------------


// HEADER FILES ------------------------------------------------------------

#include "doom/mbf_data.h"
#include "doom/p_local.h"
#include "doom/p_start.h"
#include "doom/tables.h"
#include "p_data.h"
#include "doom/g_game.h"

#include "AcFnLink.h"
#include "con_console.h"
#include "m_misc.h"
#include "p_maputil.h"
#include "p_sight.h"
#include "r_data.h"
#include "r_main.h"
#include "net_main.h"
#include "s_main.h"

#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "r3d_mapinfo.h"

// TYPES -------------------------------------------------------------------

#define	TRACEANGLE 0xc000000;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

void P_SetMobjAbort();

void __cdecl A_Fall (mobj_t *actor);
void __cdecl A_ReFire(mbf_player_t* player, pspdef_t* psp );
void __cdecl A_SpawnFly (mobj_t* mo);
void __cdecl A_Fire (mobj_t* actor);
void __cdecl A_FaceTarget(mobj_t *actor);
void __cdecl A_PainDie(mobj_t *actor);

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// CODE --------------------------------------------------------------------

//=========================================================
// ENEMY THINKING
//  Enemies are allways spawned with targetplayer = -1,
//  threshold = 0. Most monsters are spawned unaware of
//  all players, but some can be made preaware
//=========================================================

static void P_RecursiveSound(sector_t *sec, int soundblocks, mobj_t *soundtarget)
{
	int i;
	
	// wake up all monsters in this sector
	if (sec->validcount == validcount && sec->soundtraversed <= soundblocks+1)
		return;             // already flooded
	
	sec->validcount = validcount;
	sec->soundtraversed = soundblocks+1;
	if (sec->siflags & SIF_SOUNDBLOCK) return;
	P_SetTarget(&sec->soundtarget, soundtarget);
	
	for (i = 0; i < sec->linecount; i++)
	{
		line_t		*ld = sec->lines[i];
		sector_t	*other = ld->frontsector == sec ? ld->backsector : ld->frontsector;
		
		if (!other || other == sec || ld->specType >= R3D_40) continue;
		
		if (  MIN_OF(ld->frontsector->ceilingheight, ld->backsector->ceilingheight)
			- MAX_OF(ld->frontsector->floorheight  , ld->backsector->floorheight  ) <= 0) // closed door
		{
			continue;
		}
		
		if (ld->ml_flags & ML_SOUNDBLOCK || ld->lnflags2 & LNF2_OCCLUDE)
		{
			if (!soundblocks) P_RecursiveSound(other, 1, soundtarget);
		}
		else
			P_RecursiveSound(other, soundblocks, soundtarget);
	}
}

// If the player fires a weapon then
// alert other monsters to the player.
void P_NoiseAlert(mobj_t* target)
{
	validcount++;
	P_RecursiveSound(target->subsector->sector, 0, target);
}

// Validate target and check it's an active mobj thinker
// b211-0 Revised
static RDBOOL validActorTarget(mobj_t* actor, int from)
{
	return actor->target
		&& actor->target->thinker.function == (void (__cdecl *)(void))P_MobjThinker
		&& actor->target->health > 0;
}

RDBOOL P_CheckMeleeRange (mobj_t* actor)
{
    mobj_t	*pl;
    fixed_t	dist;
	
	pl = actor->target;
	
	// killough 7/18/98: friendly monsters don't attack other friends
	if (actor->flags3 & pl->flags3 & MF3_FRIEND) return false;

	// Make 3d (JK)
	dist = P_ApproxDistance(P_ApproxDistance (pl->x-actor->x, pl->y-actor->y),
		(pl->z + (pl->height>>1)) - (actor->z + (actor->height>>1)));

	return (dist < MELEERANGE - 20*FRACUNIT + pl->info->radius)
		&& P_CheckSight(actor, actor->target);
}

//
// P_HitFriend()
//
// killough 12/98
// This function tries to prevent shooting at friends

static RDBOOL P_HitFriend(mobj_t *actor)
{
	return actor->target &&
		(P_AimLineAttack(actor, 
		R_PointToAngle2(actor->x, actor->y,
		actor->target->x, actor->target->y),
		P_ApproxDistance(actor->x-actor->target->x, 
		actor->y-actor->target->y), 0),
		mobjtarget) && mobjtarget != actor->target &&
		!((mobjtarget->flags3 ^ actor->flags3) & MF3_FRIEND);
}

//
// P_CheckMissileRange
//
RDBOOL P_CheckMissileRange (mobj_t* actor)
{
    fixed_t	dist;
	
    if ( actor->flags & MF_JUSTHIT )
    {
		// the target just hit the enemy,
		// so fight back!
		actor->flags &= ~MF_JUSTHIT;

	// killough 7/18/98: no friendly fire at corpses
	// killough 11/98: prevent too much infighting among friends
		return 
			!(actor->flags3 & MF3_FRIEND) || 
			(actor->target->health > 0 &&
			(!(actor->target->flags3 & MF3_FRIEND) ||
			(actor->target->player ? cfg.inFight || M_RandomD() > 128 :
			!(actor->target->flags & MF_JUSTHIT) && M_RandomD() > 128)));
    }
	
	// killough 7/18/98: friendly monsters don't attack other friendly
	// monsters or players (except when attacked, and then only once)
	if (actor->flags3 & actor->target->flags3 & MF3_FRIEND)
		return false;

    if (actor->reactiontime)
		return false;	// do not attack yet
	
    if (!P_CheckSight(actor, actor->target) )
		return false;
		
    // OPTIMIZE: get this from a global checksight
    dist = P_ApproxDistance ( actor->x-actor->target->x,
			     actor->y-actor->target->y) - 64*FRACUNIT;
    
    if (!actor->info->meleestate)
		dist -= 128*FRACUNIT;	// no melee attack, so fire more

    dist >>= 16;

    if (actor->type == MT_VILE)
    {
		if (dist > 14*64) return false;	// too far away
    }
	
    else if (actor->type == MT_UNDEAD)
    {
		if (dist < 196) return false;	// close for fist attack
		dist >>= 1;
    }
	
    else if (
		   actor->type == MT_CYBORG
		|| actor->type == MT_SPIDER
		|| actor->type == MT_SKULL)
    {
		dist >>= 1;
    }
    
    if (dist > 200) dist = 200;
		
    if (actor->type == MT_CYBORG && dist > 160) dist = 160;
		
    if (M_RandomD() < dist) return false;
		
	if (actor->flags3 & MF3_FRIEND && P_HitFriend(actor))
		return false;
	
    return true;
}

//
// P_IsOnLift
//
// killough 9/9/98:
//
// Returns true if the object is on a lift. Used for AI,
// since it may indicate the need for crowded conditions,
// or that a monster should stay on the lift for a while
// while it goes up or down.
//

// v220-31 streamline and add moving flats detection
static RDBOOL P_IsOnLift(mobj_t *actor)
{
	return (actor->misc & MJF_ONLIFT
		|| (actor->subsector->sector->specFlags & SF_MOVSECTOR) > 0);
}

//
// P_IsUnderDamage
//
// killough 9/9/98:
//
// Returns nonzero if the object is under damage based on
// their current position. Returns 1 if the damage is moderate,
// -1 if it is serious. Used for AI.
//

static int ceilres;

static int PIT_CEILING(sector_t *sec, void *data)
{
	ceiling_t *cl = (ceiling_t*)sec->ceilingdata;

	if (cl && 
		cl->thinker.function == (void (__cdecl *)(void))T_MoveCeiling)
	{
		ceilres |= cl->direction;
	}

	return true;
}

static int P_IsUnderDamage(mobj_t *actor)
{
	ceilres = 0;

	P_ThingSectorsIterator(actor, (int (__cdecl *)(struct sector_s *,void *))PIT_CEILING, NULL);

	return ceilres;
}

//
// P_Move
// Move in the current direction,
// returns false if the move is blocked.
//
static fixed_t	xspeed[8] = {FRACUNIT,47000,0,-47000,-FRACUNIT,-47000,0,47000};
static fixed_t	yspeed[8] = {0,47000,FRACUNIT,47000,0,-47000,-FRACUNIT,-47000};

extern	line_t	**spechit;
extern	int		numspechit;

static RDBOOL P_Move (mobj_t* actor)
{
	fixed_t	tryx, tryy, deltax, deltay;
	RDBOOL	try_ok;
	int		friction, movefactor;    // killough 10/98
	int		speed;
	
	if (actor->movedir == DI_NODIR)
		return false;
	
#ifdef RANGECHECK
	if ((unsigned)actor->movedir >= 8)
		Con_Error("Weird actor->movedir: mapnum %i", MAPNUM(actor));
#endif

	gdt.pt_oldx = actor->x;
	gdt.pt_oldy = actor->y;
	gdt.pt_oldz = actor->z;
	actor->servoz = 0;
	
	// killough 10/98: make monsters get affected by ice and sludge too:
	speed = actor->info->speed >> FRACBITS;
	// Set friction and movefactor
	movefactor = P_GetMoveFactor(actor, &friction);
	
	if (friction < ORIG_FRICTION &&     // sludge
		!(speed = ((ORIG_FRICTION_FACTOR - (ORIG_FRICTION_FACTOR-movefactor)/2)
		* speed) / ORIG_FRICTION_FACTOR))
	{
		speed = 1;      // always give the monster a little bit of speed
	}

	fixed_t	origx, origy;

	tryx = (origx = actor->x) + (deltax = speed * xspeed[actor->movedir]);
	tryy = (origy = actor->y) + (deltay = speed * yspeed[actor->movedir]);

	try_ok = P_TryMove(actor, tryx, tryy, 1);

	// killough 12/98: rearrange, fix potential for stickiness on ice
	// v220-31 allow enemies to slide on ice rather than staying stationary
	if (try_ok && friction > ORIG_FRICTION)
	{
		actor->x = origx;
		actor->y = origy;
		movefactor *= FRACUNIT / ORIG_FRICTION_FACTOR;
		actor->momx += FixedMul(deltax, movefactor);
		actor->momy += FixedMul(deltay, movefactor);
	}
	
	if (!try_ok)
	{
		int	i, good;

		if (actor->flags & MF_FLOAT && gdt.floatok)
		{
			if (gdt.adjustz)
			{
				// must adjust height
				// v220-28 mods to prevent rapid bouncing
				if (   actor->z < gdt.tmfloorz
					&& actor->z + actor->height < gdt.tmceilingz + FLOATSPEED	// v220-28
					&& checkStepupClear(actor, actor->z + FLOATSPEED))
				{
					actor->srvo[VZ] = (short)(-FLOATSPEED >> 8);
					actor->z += FLOATSPEED;
				}
				else if (actor->z > gdt.tmfloorz + FLOATSPEED	// v220-28
					&& checkStepupClear(actor, actor->z - FLOATSPEED))
				{
					actor->srvo[VZ] = (short)(FLOATSPEED >> 8);
					actor->z -= FLOATSPEED;
				}
			}

			actor->flags |= MF_INFLOAT;
			return true;
		}
		
		if (!numspechit) return false;
		
      // if the special is not a door that can be opened, return false
      //
      // killough 8/9/98: this is what caused monsters to get stuck in
      // doortracks because it was thought that the monster freed itself
      // by opening a door, even if it was moving towards the doortrack,
      // and not the door itself.
      //
      // killough 9/9/98: If a line blocking the monster is activated,
      // return true 90% of the time. If a line blocking the monster is
      // not activated, but some other line is, return false 90% of the
      // time. A bit of randomness is needed to ensure it's free from
      // lockups, but for most cases, it returns the correct result.
      //
      // Do NOT simply return false 1/4 of the time (causes monsters to
      // back out when they shouldn't and creates secondary stickiness).

		actor->movedir = DI_NODIR;
		good = 0;

		i = numspechit;

		while (--i >= 0)
		{
			if (   spechit[i]
				// If the actor was not blocked then it could now be on the wrong side of the line
				&& !P_CheckOnLineSide(actor->x, actor->y, spechit[i])
				&& P_UseSpecialLine(actor, spechit[i], 0, 0))
			{
				good |= ((spechit[i] == gdt.blockline) ? 1 : 2);
			}
		}

		numspechit = 0;
			
		return good && ((M_RandomD() >= 230) ^ (good & 1));
	}
	else
		actor->flags &= ~MF_INFLOAT;
	
	actor->srvo[VX] = (gdt.pt_oldx - actor->x) >> 8;
	actor->srvo[VY] = (gdt.pt_oldy - actor->y) >> 8;

	if (gdt.adjustz)
		actor->srvo[VZ] = (gdt.pt_oldz - actor->z) >> 8;

    return true; 
}


//
// P_SmartMove
//
// killough 9/12/98: Same as P_Move, except smarter
//

static RDBOOL P_SmartMove(mobj_t *actor)
{
	mobj_t	*target = actor->target;
	int		on_lift, under_damage;
	
	// killough 9/12/98: Stay on a lift if target is on one
	on_lift = target && target->health > 0
		&& target->subsector->sector->tag == actor->subsector->sector->tag
		&& P_IsOnLift(actor);
	
	under_damage = P_IsUnderDamage(actor);

	if (!P_Move(actor)) return false;
	
	// killough 9/9/98: avoid crushing ceilings or other damaging areas
	if (
		(on_lift && M_RandomD() < 230 &&				// Stay on lift
		!P_IsOnLift(actor))
		||
		(!under_damage &&							// Get away from damage
		(under_damage = P_IsUnderDamage(actor)) &&
		(under_damage < 0 || M_RandomD() < 200))
		)
	{
		actor->movedir = DI_NODIR;	// avoid the area (most of the time anyway)
	}
	
	return true;
}

//
// TryWalk
// Attempts to move actor on
// in its current (ob->moveangle) direction.
// If blocked by either a wall or an actor
// returns FALSE
// If move is either clear or blocked only by a door,
// returns TRUE and sets...
// If a door is in the way,
// an OpenDoor call is made to start it opening.
//
RDBOOL P_TryWalk (mobj_t* actor)
{	
	// v220-14
	if (actor->subsector->ssflags & SSF_NORENDER)
	{
		// Ensure things are on the floor in control sectors
		// else switch tp specials may not be activated (see $swtp).
		// Why? It would seem, from cchest3 map28, that floating things
		// get blocked by a dropoff in the original Doom engine code.
		// This aberrant behaviour must have been unwittingly resolved
		// by the changes made in the R3D code prior to this point.

		if (!actor->momz && actor->z > actor->floorz)
		{
			actor->momz = -FRACUNIT;
		}

		if (!P_Move(actor)) return false;
		actor->movecount = 15;
		return true; 
	}

	if (!P_SmartMove(actor)) return false;

    actor->movecount = (M_RandomD() & 15);
    return true;
}

//*****************************************************************************
//  killough $dropoff_fix Replaces original DOOM code with following 4 functions
//
// P_DoNewChaseDir
//
// Most of P_NewChaseDir(), except for what
// determines the new direction to take
//

static void P_DoNewChaseDir(mobj_t *actor, fixed_t deltax, fixed_t deltay)
{
	int xdir, ydir, tdir;
	int olddir = actor->movedir;
	int turnaround = olddir;

	if (turnaround != DI_NODIR)         // find reverse direction
		turnaround ^= DI_WEST;

	xdir = 
		deltax >  10*FRACUNIT ? DI_EAST :
		deltax < -10*FRACUNIT ? DI_WEST : DI_NODIR;

	ydir = 
		deltay < -10*FRACUNIT ? DI_SOUTH :
		deltay >  10*FRACUNIT ? DI_NORTH : DI_NODIR;

	// try direct route
	if (   xdir != DI_NODIR
		&& ydir != DI_NODIR
		&& turnaround != 
			(actor->movedir = deltay < 0 ?
			deltax > 0 ?
			DI_SOUTHEAST : DI_SOUTHWEST :
			deltax > 0 ?
			DI_NORTHEAST : DI_NORTHWEST)
		&& P_TryWalk(actor))
	{
		return;
	}

	// try other directions
	if (M_RandomD() > 200 || abs(deltay)>abs(deltax))
		tdir = xdir, xdir = ydir, ydir = tdir;

	if ((xdir == turnaround ? xdir = DI_NODIR : xdir) != DI_NODIR &&
		(actor->movedir = xdir, P_TryWalk(actor)))
		return;         // either moved forward or attacked

	if ((ydir == turnaround ? ydir = DI_NODIR : ydir) != DI_NODIR &&
		(actor->movedir = ydir, P_TryWalk(actor)))
		return;

	// there is no direct path to the player, so pick another direction.
	if (olddir != DI_NODIR && (actor->movedir = olddir, P_TryWalk(actor)))
		return;

	// randomly determine direction of search
	if (M_RandomD() & 1)
	{
		for (tdir = DI_EAST; tdir <= DI_SOUTHEAST; tdir++)
		{
			if (tdir != turnaround && (actor->movedir = tdir, P_TryWalk(actor)))
				return;
		}
	}
	else
	{
		for (tdir = DI_SOUTHEAST; tdir != DI_EAST-1; tdir--)
		{
			if (tdir != turnaround && (actor->movedir = tdir, P_TryWalk(actor)))
				return;
		}
	}
  
	if ((actor->movedir = turnaround) != DI_NODIR && !P_TryWalk(actor))
		actor->movedir = DI_NODIR;
}

//===========================================================================
// PIT_AvoidDropoff - Killough
//
// In Doom, actors were never allowed to hang over dropoffs
// and would remain stuck if involuntarily forced over one.
// This logic, combined with p_map.c (P_TryMove), allows
// monsters to free themselves without making them tend to
// hang over dropoffs.
//
// Extended for R3D - (3D lines and slopes)
//===========================================================================

static fixed_t	dropoff_deltax, dropoff_deltay;

// b211-0  Revised to fix actors getting stuck on edge of moving plats
// v220-31 Support added for Things on 3D lifts
static RDBOOL PIT_AvoidDropoff(line_t *line, mobj_t *th)
{
	sector_t	*fs = line->frontsector, *bs = line->backsector;

	if (fs == bs) return true;	// pointless checking this line

	angle_t		angle;
	int			side = P_PointOnLineSide(th->x, th->y, line);

	if (line->specType >= R3D_40)
	{
		if (th->misc & MJF_ONLIFT)
		{
			if (side)
			{
				angle = R_PointToAngle2(0, 0, line->dx, line->dy);	// front side drop off
				dropoff_deltax -= finesine[angle >> ANGLETOFINESHIFT]*32;
				dropoff_deltay += finecosine[angle >> ANGLETOFINESHIFT]*32;
				return true;
			}
		}
		return true;	// static 3D lines use a different method
	}

	// b211-0 Use generic height $slopes

	RDBOOL	sf = (!side && fs->specFlags2 & SF2_LOSLOPE) ? true : false;
	RDBOOL	sb = ( side && bs->specFlags2 & SF2_LOSLOPE) ? true : false;
	fixed_t	front = sf ? SL_PointHeight(fs, th->x, th->y, LOSLOPE) : fs->floorheight;
	fixed_t	back  = sb ? SL_PointHeight(bs, th->x, th->y, LOSLOPE) : bs->floorheight;

	// The monster must contact one of the two floors,
	// and the other must be a tall drop off (more than 24).
	if ((sb || back == th->z) && front < th->z - FRACUNIT*24)
		angle = R_PointToAngle2(0, 0, line->dx, line->dy);	// front side drop off
	else if ((sf || front == th->z) && back < th->z - FRACUNIT*24)
		angle = R_PointToAngle2(line->dx, line->dy, 0, 0);	// back side drop off
	else
		return true;

	// Move away from drop off at a standard speed.
	// Multiple contacted linedefs are cumulative (e.g. hanging over corner)
	dropoff_deltax -= finesine[angle >> ANGLETOFINESHIFT]*32;
	dropoff_deltay += finecosine[angle >> ANGLETOFINESHIFT]*32;
	
	return true;
}

// Driver for above
static RDBOOL P_AvoidDropoff(mobj_t *actor)
{
	dropoff_deltax = dropoff_deltay = 0;

	// check lines (use Doomsday line linker code)
	// b211-0 Add actor - delete previous static mobj_t global
	P_ThingLinesIterator(actor, (int (__cdecl *)(struct line_s *,void *))PIT_AvoidDropoff, actor);

   // true if movement prescribed
	return dropoff_deltax != 0 || dropoff_deltay != 0;
}

//===========================================================================

RDBOOL ignore_dropoff;	// v220-31

void P_NewChaseDir (mobj_t*	actor)
{
	mobj_t	*target = actor->target;
	fixed_t	deltax, deltay;

	actor->strafecount = 0;

	if (target)	// b211-0 Might have been removed
	{
		deltax = target->x - actor->x;
		deltay = target->y - actor->y;
	}
	else
	{
		actor->movecount = 0;
		return;
	}

	if (   actor->floorz - actor->dropoffz > MAXSTEP
		&& actor->z <= actor->floorz
		&& !(actor->flags & (MF_DROPOFF|MF_FLOAT))
		&& P_AvoidDropoff(actor)) // Move away from dropoff
	{
		// b211-0 Added - If walking on the spot give it some momentum
		if (  !actor->momx && !actor->momy
			&& actor->state->action_num == nmChase)
		{
			actor->momx = dropoff_deltax>>6;
			actor->momy = dropoff_deltay>>6;
		}

		// v220-33 adjust z if caught between a step and a dropoff
		if (   actor->movedir == DI_NODIR
			&& actor->state->action_num == nmChase
			&& gdt.tmfloorz - actor->z > 0 && gdt.tmfloorz - actor->z <= MAXSTEP)
		{
			 actor->floorz = actor->z = gdt.tmfloorz;
		}

		ignore_dropoff = !actor->subsector->sector->floordata;	// v220-35
		P_DoNewChaseDir(actor, dropoff_deltax, dropoff_deltay);
		ignore_dropoff = false;

		// $dropoff_fix killough 9/8/98
		// If moving away from drop off, set movecount to 1 so that 
		// small steps are taken to get monster away from drop off.

		actor->movecount = 1;
		return;
	}
	else if (target)	// b211-0
	{
		fixed_t dist = P_ApproxDistance(deltax, deltay);
		
		// Move away from friends when too close, except
		// in certain situations (e.g. a crowded lift)
		
		if (   actor->flags3 & target->flags3 & MF3_FRIEND
			&& 128*FRACUNIT > dist // distfriend << FRACBITS
			&& !P_IsOnLift(target) && !P_IsUnderDamage(actor))
		{
			deltax = -deltax, deltay = -deltay;
		}
		else
		{
			if (   target->health > 0
				&& actor->info->missilestate
				&& (actor->flags3 ^ target->flags3) & MF3_FRIEND
				&& actor->type != MT_SKULL)
			{
				RDBOOL backaway = !target->info->missilestate && dist < MELEERANGE*2;

				if (!backaway)
					backaway =  target->player && dist < MELEERANGE*3
							&& (target->player->readyweapon == wp_fist
							|| 	target->player->readyweapon == wp_chainsaw);

				if (backaway)
				{
					// Back away from melee attacker
					actor->strafecount = M_RandomD() & 15;
					deltax = -deltax, deltay = -deltay;
				}
			}
		}
	}

	P_DoNewChaseDir(actor, deltax, deltay);

	// If strafing, set movecount to strafecount so that old Doom
	// logic still works the same, except in the strafing part
	
	if (actor->strafecount)
		actor->movecount = actor->strafecount;
}

//*****************************************************************************

// 
// Friendly monsters, by Lee Killough 7/18/98
//
// Friendly monsters go after other monsters first, but 
// also return to owner if they cannot find any targets.
// A marine's best friend :)  killough 7/18/98, 9/98
//

static RDBOOL current_allaround;

static RDBOOL P_IsVisible(mobj_t *actor, mobj_t *mo, RDBOOL allaround)
{
	if (!allaround)
	{
		angle_t an = R_PointToAngle2(actor->x, actor->y, 
			mo->x, mo->y) - actor->angle;

		if (an > ANG90 && an < ANG270 &&
			P_ApproxDistance(mo->x - actor->x, mo->y - actor->y) > MELEERANGE)
			return false;
	}
	return P_CheckSight(actor, mo);
}

static RDBOOL PIT_FindTarget(mobj_t *mo, mobj_t *actor)
{
	if (!(mo->health > 0
		&& (((mo->flags3 ^ actor->flags3) & MF3_FRIEND		// Invalid target
		|| (mo->ddflags ^ actor->ddflags) & DDF_SCPINFIGHT)	// scripted NOINFIGHTALL
		&& (mo->flags & MF_COUNTKILL || mo->type == MT_SKULL))))
		return true;

	// If the monster is already engaged in a one-on-one attack
	// with a healthy friend, don't attack around 60% the time
	{
		mobj_t *targ = mo->target;

		if (targ && targ->target == mo &&
			M_RandomD() > 100 &&
			(targ->flags3 ^ mo->flags3) & MF3_FRIEND &&
			targ->health*2 >= targ->info->spawnhealth)
			return true;
	}
	
	if (!P_IsVisible(actor, mo, current_allaround)) return true;
	
	P_SetTarget(&actor->lastenemy, actor->target);  // Remember previous target
	P_SetTarget(&actor->target, mo);                // Found target
	
	// Move the selected monster to the end of its associated
	// list, so that it gets searched last next time.
	{
		int			thclass = (mo->flags3 & MF3_FRIEND) ? th_friends : th_enemies;
		thinker_t	*cap = &thinkerclasscap[thclass];

		(mo->thinker.cprev->cnext = mo->thinker.cnext)->cprev = mo->thinker.cprev;
		(mo->thinker.cprev = cap->cprev)->cnext = &mo->thinker;
		(mo->thinker.cnext = cap)->cprev = &mo->thinker;
	}
	
	return false;
}

// 
// Friendly monsters, by Lee Killough 7/18/98
//
// Friendly monsters go after other monsters first, but 
// also return to owner if they cannot find any targets.
// A marine's best friend :)  killough 7/18/98, 9/98
//
static RDBOOL P_LookForMonsters(mobj_t *actor, RDBOOL allaround)
{
	thinker_t	*cap, *th;
	int			thclass;
	
	if (   actor->lastenemy
		&& actor->lastenemy->health > 0
		&& !(actor->lastenemy->flags3 & actor->flags3 & MF3_FRIEND)) // not friends
	{
		P_SetTarget(&actor->target, actor->lastenemy);
		P_SetTarget(&actor->lastenemy, NULL);
		return true;
	}
	
	// Search the threaded list corresponding to this object's potential targets
	thclass = (actor->flags3 & MF3_FRIEND) ? th_enemies : th_friends;
	cap = &thinkerclasscap[thclass];

	// Search for new enemy
	if (cap->cnext != cap)        // Empty list? bail out early
	{
		int x = MAPBS(actor->x, bmaporgx);
		int y = MAPBS(actor->y, bmaporgy);
		int d;

		current_allaround = allaround;

		validcount++;

		// Search first in the immediate vicinity.
		if (!P_BlockThingsIterator(x, y, (int (__cdecl *)(struct mobj_s *,void *))PIT_FindTarget, actor))
			return true;
		
		for (d = 1; d < 5; d++)
		{
			int i = 1 - d;

			do
			{
				if (!P_BlockThingsIterator(x+i, y-d, (int (__cdecl *)(struct mobj_s *,void *))PIT_FindTarget, actor) ||
					!P_BlockThingsIterator(x+i, y+d, (int (__cdecl *)(struct mobj_s *,void *))PIT_FindTarget, actor))
					return true;
			}
			while (++i < d);

			do
			{
				if (!P_BlockThingsIterator(x-d, y+i, (int (__cdecl *)(struct mobj_s *,void *))PIT_FindTarget, actor) ||
					!P_BlockThingsIterator(x+d, y+i, (int (__cdecl *)(struct mobj_s *,void *))PIT_FindTarget, actor))
					return true;
			}
			while (--i + d >= 0);
		}

		{   // Random number of monsters, to prevent patterns from forming
			int n = (M_RandomD() & 31) + 15;
			
			for (th = cap->cnext; th != cap; th = th->cnext)
			{
				if (--n < 0)
				{ 
					// Only a subset of the monsters were searched. Move all of
					// the ones which were searched so far, to the end of the list.
					
					(cap->cnext->cprev = cap->cprev)->cnext = cap->cnext;
					(cap->cprev = th->cprev)->cnext = cap;
					(th->cprev = cap)->cnext = th;
					break;
				}
				else
				{
					if (!PIT_FindTarget((mobj_t *) th, actor))   // If target sighted
						return true;
				}
			}
		}
	}
	
	return false;  // No monster found
}

//
// P_LookForPlayers
// If allaround is false, only look 180 degrees in front.
// Returns true if a player is targeted.
// v220-5 modified further to reflect the fact that there is only one player.
//
RDBOOL P_LookForPlayers(mobj_t* actor, RDBOOL allaround)
{
	mbf_player_t *pl = &mbf_players[0];
	angle_t		 an;
	fixed_t		 dist;
	
	// Check the player is valid
	if ( !(pl->plr->ingame && plhealth.health > 0) )
		return false;
	
	if (actor->flags3 & MF3_FRIEND)
	{
		// killough 9/9/98: friendly monsters go about players differently
		
		// Go back to a player, no matter whether it's visible or not
		// Check for a visible player first
		if (P_IsVisible(actor, pl->plr->mo, allaround))
		{
			P_SetTarget(&actor->target, pl->plr->mo);
			
			// killough 12/98:
			// get out of refiring loop, to avoid hitting player accidentally
			
			if (actor->info->missilestate)
			{
				P_SetMobjState(actor, actor->info->seestate);
				actor->flags &= ~MF_JUSTHIT;
			}
			
			return true;
		}

		return false;
	}

	if (!P_IsVisible (actor, pl->plr->mo, allaround))
		return false;	// out of sight
	
	if (!allaround)
	{
		an = R_PointToAngle2 (actor->x,
			actor->y, 
			pl->plr->mo->x,
			pl->plr->mo->y)
			- actor->angle;
		
		if (an > ANG90 && an < ANG270)
		{
			dist = P_ApproxDistance (pl->plr->mo->x - actor->x,
				pl->plr->mo->y - actor->y);
			// if real close, react anyway
			if (dist > MELEERANGE) return false;	// behind back
		}
	}
		
	P_SetTarget(&actor->target, pl->plr->mo);
	
	// killough 9/9/98: give monsters a threshold towards getting players
	// (we don't want it to be too easy for a player with dogs :)
	actor->threshold = 0; // was 60;
	
	return true;
}

// killough 2/ 7/98: fixed lost soul bug (LSs left behind when PEs are killed)
//                   Adapted for friendly monsters
// b211-0 Reformatted for clarity
// v220-3 notfriends test added
// v220-4 made void - changed scp_killcount to massacre_count and the way this is used in stats
int P_Massacre(RDBOOL notfriends)
{
	int i, mask, killcount = mbf_players[0].killcount;

	con_freezegame = false;

	for (i = 0; i < 2; i++)
	{
		thinker_t *th = &thinkercap;

		// killough 7/20/98: kill friendly monsters only if no others to kill
		if (i && (killcount || notfriends)) break;

		mask = i ? 0 : MF3_FRIEND;

		while ((th = th->next) != &thinkercap)
		{
			if (th->function == (void (__cdecl *)(void))P_MobjThinker)
			{
				mobj_t	*mo = (mobj_t*)th;

				if (  !(mo->flags3 & mask)
					&& (mo->flags & MF_COUNTKILL || mo->type == MT_SKULL))
				{
					if (mo->health > 0)
					{
						P_DamageMobj(mo, NULL, NULL, 10000, 9);
					}
					if (mo->type == MT_PAIN)
					{
						A_PainDie(mo);    // killough 2/8/98
						P_SetMobjState(mo, S_PAIN_DIE6);
					}
				}
			}
		}
	}

	// v220-4 track scripted or player cheat massacres

	i = mbf_players[0].killcount - killcount;	// Get difference between start and finish counts
	ma_inf.massacre_count += i;					// Add it to the current massacre total
	mbf_players[0].killcount = killcount;		// Restore player killcount

	return i;	// Return number killed
}


#pragma optimize("g", off)

//
// A_KeenDie
// DOOM II special, map 32.
// Uses special tag 666.
//
void __cdecl A_KeenDie (mobj_t* mo)
{
	thinker_t	*th;
	line_t		junk;
	
	A_Fall (mo);
	
	// scan the remaining thinkers
	// to see if all Keens are dead
	for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
	{
		if (th->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t *mo2 = (mobj_t *)th;
			
			if (mo2 != mo
				&& mo2->type == mo->type
				&& mo2->health > 0)
			{
				// other Keen not dead
				return;		
			}
		}
	}
	
	junk.tag = 666;
	junk.lnflags2 = 0;
	EV_DoDoor(&junk, openNormal);
}


//
// ACTION ROUTINES
//

void P_PlaySeesound(mobj_t *mo)
{
	int		sound;
	
	if (mo->flags2 & MF2_NOSEESOUND) return;

	switch (mo->info->seesound)
	{
	case 0:
		return;

	case sfx_posit1:
	case sfx_posit2:
	case sfx_posit3:
		sound = sfx_posit1 + (M_RandomD() % 3);
		break;
		
	case sfx_bgsit1:
	case sfx_bgsit2:
		sound = sfx_bgsit1 + (M_RandomD() % 2);
		break;
		
	case sfx_spisit:	// MT_SPIDER
	case sfx_cybsit:	// MT_CYBORG
		sound = mo->info->seesound;
		if (ma_inf.mapEXM8) sound |= HSF_NO_ATTENUATION;
		break;

	default:
		sound = mo->info->seesound;
		break;
	}

	S_StartSound(sound, mo);
}

static RDBOOL P_LookForTargets(mobj_t *actor, int allaround)
{
	return actor->flags3 & MF3_FRIEND ?
		P_LookForMonsters(actor, allaround) || P_LookForPlayers (actor, allaround):
		P_LookForPlayers (actor, allaround) || P_LookForMonsters(actor, allaround);
}

//
// P_HelpFriend
//
// killough 9/8/98: Help friends in danger of dying
//

static RDBOOL P_HelpFriend(mobj_t *actor)
{
	thinker_t	*cap, *th;
	int			thclass;
	
	// If less than 33% health, self-preservation rules
	if (actor->health*3 < actor->info->spawnhealth)
		return false;
	
	current_allaround = true;
	
	// Possibly help a friend under 50% health
	thclass = (actor->flags3 & MF3_FRIEND) ? th_friends : th_enemies;
	cap = &thinkerclasscap[thclass];
	
	for (th = cap->cnext; th != cap; th = th->cnext)
	{
		if (th->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t *mo = (mobj_t *) th;

			if (mo->health*2 >= mo->info->spawnhealth)
			{
				if (M_RandomD() < 180)
					break;
			}
			else
			{
				if (mo->flags & MF_JUSTHIT &&
					mo->target && 
					mo->target != actor->target &&
					!PIT_FindTarget(mo->target, actor))
				{
					// Ignore any attacking monsters, while searching for friend
					actor->threshold = BASETHRESHOLD;
					return true;
				}
				
			}
		}
	}
	return false;
}

// A_Look - Stay in state until a player is sighted.
void __cdecl A_Look(mobj_t *actor)
{
	// prevent 8113 line triggered monsters from being woken by enemy fire
	if (actor->ddflags & DDF_HOLDSTATE) return;	// v220-34

	if (actor->subsector->sector->siflags & SIF_SOUNDBLOCK)	// v220-33
		actor->subsector->sector->siflags &= ~SIF_SOUNDBLOCK;

	actor->threshold = actor->pursuecount = 0;	// any shot will wake up

	// killough 7/18/98:
	// Friendly monsters go after other monsters first, but 
	// also return to player, without attacking them, if they
	// cannot find any targets. A marine's best friend :)

	mobj_t	*targ;
	
	if ( !(actor->flags3 & MF3_FRIEND
		&& P_LookForTargets(actor, false))
		&& !((targ = actor->subsector->sector->soundtarget)
		&& targ->flags & MF_SHOOTABLE
		&& (P_SetTarget(&actor->target, targ),!(actor->flags & MF_AMBUSH)
		|| P_CheckSight(actor, targ)))
		&& (actor->flags3 & MF3_FRIEND
		|| !P_LookForTargets(actor, false)))
	{
		return;
	}

	P_PlaySeesound(actor);
	P_SetMobjState (actor, actor->info->seestate);
}

// A_Chase
// Actor has a melee attack so tries to close as fast as possible
void __cdecl A_Chase(mobj_t *actor)
{
	RDBOOL valid = validActorTarget(actor, 0);

    if (actor->reactiontime) actor->reactiontime--;
				
    // modify target threshold
    if  (actor->threshold)
    {
		if (!valid)
			actor->threshold = 0;
		else
			actor->threshold--;
    }

	// turn towards movement direction if not there yet
	// killough 9/7/98: keep facing towards target if strafing or backing out
	
	if (actor->strafecount)
	{
		A_FaceTarget(actor);
	}
	else if (actor->movedir < 8)
	{
		int	delta = (actor->angle &= (7<<29)) - (actor->movedir << 29);

		if (delta > 0)
			actor->angle -= ANG90/2;
		else if (delta < 0)
			actor->angle += ANG90/2;
	}

	if (!valid || !(actor->target->flags & MF_SHOOTABLE))
	{
		// v220-13
		if (ma_inf.playerdead && actor->health > 0)
		{
			// Scripted types between 6070 and 6099 do not revert to the spawnstate
			if (actor->info->doomednum >= 6070 && actor->info->doomednum <= 6099)
			{
				return;
			}
		}
		
		if (!P_LookForTargets(actor,true)) // look for a new target
			P_SetMobjState(actor, actor->info->spawnstate); // no new target
		return;
	}
    
	// do not attack twice in a row
	if (actor->flags & MF_JUSTATTACKED)
	{
		actor->flags &= ~MF_JUSTATTACKED;

		if (actor->ddflags & DDF_AGGRESSIVE)
			return;

		if (gameskill != sk_nightmare && !fastparm)
			P_NewChaseDir (actor);

		return;
	}

	// check for melee attack
	if (actor->info->meleestate && P_CheckMeleeRange (actor))
	{
		if (actor->info->attacksound)
			S_StartSound (actor->info->attacksound, actor);
		
		P_SetMobjState (actor, actor->info->meleestate);

		if (!actor->info->missilestate)
			actor->flags |= MF_JUSTHIT;   // killough 8/98: remember an attack

		return;
	}

	// check for missile attack
	if (actor->info->missilestate)
	{
		if (!actor->movecount || gameskill == sk_nightmare || fastparm || actor->ddflags & DDF_AGGRESSIVE)
		{
			if (P_CheckMissileRange(actor))
			{
				P_SetMobjState(actor, actor->info->missilestate);
				actor->flags |= MF_JUSTATTACKED;
				return;
			}
		}
	}

	if (!actor->threshold)
	{
		// killough 7/18/98, 9/9/98: new monster AI
		// killough 9/8/98: Help friends in need
		if (ge_inf.isfriendly && P_HelpFriend(actor)) return;

		// Look for new targets if current one is bad or is out of view
		if (actor->pursuecount)
			actor->pursuecount--;
		else
		{
			// Our pursuit time has expired. We're going to think about changing targets
			actor->pursuecount = BASETHRESHOLD;
			
			// Unless we have a live target and it's not friendly
			// and we can see it, try to find a new one; return if sucessful
			
			if (
				  !(
						   actor->target
						&& actor->target->health > 0
						&& (
								(
									(actor->target->ddflags ^ actor->ddflags) & DDF_SCPINFIGHT
									||
									(actor->target->flags3 ^ actor->flags3) & MF3_FRIEND
									||
									(
										cfg.inFight && !(actor->flags3 & MF3_FRIEND)
										&& !(actor->target->ddflags & (DDMF_NOINFIGHT|DDF_SCPINFIGHT))
									)
								)
								&& P_CheckSight(actor, actor->target)
							)
					)
					&& P_LookForTargets(actor, true)
				)
			{
				return;
			}
			
			// (Current target was good, or no new target was found.)

			// If monster is a missile-less friend, give up pursuit and
			// return to player, if no attacks have occurred recently.
			
			if (!actor->info->missilestate && actor->flags3 & MF3_FRIEND)
			{
				if (actor->flags & MF_JUSTHIT)			// if recent action,
					actor->flags &= ~MF_JUSTHIT;		// keep fighting
				else if (P_LookForPlayers(actor, true)) // else return to player
					return;
			}
		}
	}

	if (actor->strafecount) actor->strafecount--;

	// chase towards player
	if (--actor->movecount < 0 || !P_SmartMove(actor))
		P_NewChaseDir(actor);
	
	// make active sound
	if (actor->info->activesound && M_RandomD() < 3)
		S_StartSound(actor->info->activesound, actor);
}

void __cdecl A_FaceTarget(mobj_t *actor)
{
	if (!actor->target) return;

	actor->turntime = true; // $visangle-facetarget
    actor->flags &= ~MF_AMBUSH;
    actor->angle = R_PointToAngle2 (actor->x,
				    actor->y,
				    actor->target->x,
				    actor->target->y);
    
    if (actor->target->flags & MF_SHADOW)
		actor->angle += ( (M_RandomD() - M_RandomD()) << 21);
}

// b211-0 DBshotgun for actors
void __cdecl A_dbAttack(mobj_t *actor)
{
	angle_t	angle;
    int		damage;
    int		slope;
    int		i;
	
    if (!validActorTarget(actor, 4)) return;

	angle = actor->angle;
	
    S_StartSound(sfx_dshtgn, actor);
    A_FaceTarget (actor);
    slope = P_AimLineAttack (actor, angle, STANDARDRANGE, 0);	// v220-14 was MISSILERANGE - this was too far
	
    for (i=0 ; i<20 ; i++)
    {
		angle = angle + (angle_t)((M_RandomD() - M_RandomD()) << 19);	// v220-33 was << 20
		damage = 5 * ((M_RandomD() % 3) + 1);
		// v220-14 use STANDARDRANGE
		P_LineAttack (actor, angle,STANDARDRANGE, slope + ((M_RandomD() - M_RandomD()) << 5), damage);
    }
}

void __cdecl A_PosAttack(mobj_t *actor)
{
	angle_t	angle;
	int		damage;
	int		slope;
	
	if (!validActorTarget(actor, 5)) return;
	
	A_FaceTarget (actor);

	if (actor->flags4 & MF4_BADASSZOMBIE)	// COD support
	{
		P_SpawnMissile (actor, actor->target, MT_ROCKET);
		return;
	}

	angle = actor->angle;
	slope = P_AimLineAttack (actor, angle, useBulletRange, 0);
	
	S_StartSound(sfx_pistol, actor);
	angle += (angle_t)((M_RandomD() - M_RandomD()) << 20);
	damage = ((M_RandomD() % 5) + 1) * 3;
	P_LineAttack (actor, angle, useBulletRange, slope, damage);
}

void __cdecl A_SPosAttack(mobj_t *actor)
{
    angle_t	angle;
    int		slope;
    int		i;
	
    if (!validActorTarget(actor, 6)) return;
	
	if (useBulletRange >= MISSILERANGE)
		// Double the distance over which the chaingun sound can be heard
		S_StartSound(sfx_shotgn|HSF_HALFATTN, actor);
	else
		S_StartSound(sfx_shotgn, actor);

    A_FaceTarget(actor);
    angle = actor->angle;
    slope = P_AimLineAttack (actor, angle, useBulletRange, 0);
	
    for (i = 0; i < 3; i++)
    {
		int		damage = ((M_RandomD() % 5) + 1) * 3;

		angle += (angle_t)((M_RandomD() - M_RandomD()) << 20);
		P_LineAttack (actor, angle, useBulletRange, slope, damage);
    }
}

void __cdecl A_CPosAttack(mobj_t *actor)
{
	if (!validActorTarget(actor, 7)) return;
	
	S_StartSound(sfx_shotgn, actor);
	
	A_FaceTarget(actor);
	
	int		damage = ((M_RandomD() % 5) + 1) * 3;
	angle_t	angle = actor->angle;
	int		slope = P_AimLineAttack (actor, angle, useBulletRange, 0);
	angle += (angle_t)((M_RandomD() - M_RandomD()) << 20);
	
	P_LineAttack (actor, angle, useBulletRange, slope, damage);
}

void __cdecl A_CPosRefire(mobj_t *actor)
{	
  // keep firing unless target got out of sight
	A_FaceTarget(actor);
	
	// killough 12/98: Stop firing if a friend has gotten in the way
	if (actor->flags3 & MF3_FRIEND && P_HitFriend(actor))
	{
		P_SetMobjState(actor, actor->info->seestate);
		return;
	}
	
	// killough 11/98: prevent refiring on friends continuously
	if (M_RandomD() < 40)
	{
		if (actor->target && actor->flags3 & actor->target->flags3 & MF3_FRIEND)
			P_SetMobjState(actor, actor->info->seestate);

		return;
	}

	if (  !validActorTarget(actor, 8)
		|| actor->target->health <= 0
		|| !P_CheckSight(actor, actor->target))
	{
		P_SetMobjState(actor, actor->info->seestate);
	}
}

void __cdecl A_SpidRefire(mobj_t* actor)
{
	// keep firing unless target got out of sight
	A_FaceTarget(actor);
	
	// killough 12/98: Stop firing if a friend has gotten in the way
	if (actor->flags3 & MF3_FRIEND && P_HitFriend(actor))
	{
		P_SetMobjState(actor, actor->info->seestate);
		return;
	}
	
	if (M_RandomD() < 10) return;
	
	// killough 11/98: prevent refiring on friends continuously
	if (  !validActorTarget(actor, 9)
		|| actor->target->health <= 0
		|| actor->flags & actor->target->flags3 & MF3_FRIEND
		|| !P_CheckSight(actor, actor->target))
	{
		P_SetMobjState(actor, actor->info->seestate);
	}
}

void __cdecl A_BspiAttack(mobj_t *actor)
{	
    if (!validActorTarget(actor, 10)) return;
		
    A_FaceTarget (actor);

    // launch a missile
    P_SpawnMissile (actor, actor->target, MT_ARACHPLAZ);
}

void __cdecl A_TroopAttack(mobj_t *actor)
{	
    if (!validActorTarget(actor, 11)) return;
	
    A_FaceTarget (actor);

    if (P_CheckMeleeRange (actor))
    {
		int	damage = ((M_RandomD() % 8) + 1) * 3;
		actor->flags2 |= MF2_MELEE;
		S_StartSound(sfx_claw, actor);
		P_DamageMobj (actor->target, actor, actor, damage, 10);
		return;
    }
	
	actor->flags2 &= ~MF2_MELEE;
	
    // launch a missile
    P_SpawnMissile (actor, actor->target, MT_TROOPSHOT);
}

void __cdecl A_SargAttack(mobj_t *actor)
{
    if (!validActorTarget(actor, 12)) return;
		
    A_FaceTarget(actor);

    if (P_CheckMeleeRange (actor))
    {
		int	damage = ((M_RandomD() % 10) + 1) * 4;
		P_DamageMobj (actor->target, actor, actor, damage, 11);	// sarge attack
    }
}

void __cdecl A_HeadAttack(mobj_t *actor)
{
	if (!validActorTarget(actor, 13)) return;
	
	A_FaceTarget(actor);
	
	if (P_CheckMeleeRange (actor))
	{
		int	damage = ((M_RandomD() % 6) + 1) * 10;
		actor->flags2 |= MF2_MELEE;
		P_DamageMobj (actor->target, actor, actor, damage, 12);
		return;
	}
	
	actor->flags2 &= ~MF2_MELEE;
	
	P_SpawnMissile (actor, actor->target, MT_HEADSHOT); // launch a missile
}

void __cdecl A_CyberAttack(mobj_t *actor)
{
	// v220-4 terminate firing sequence if target's not valid
    if (!validActorTarget(actor, 14))	return;
	
    A_FaceTarget(actor);

	P_SpawnMissile (actor, actor->target, MT_ROCKET);
}

void __cdecl A_BruisAttack(mobj_t *actor)
{	
    if (!validActorTarget(actor, 15)) return;
		
    if (P_CheckMeleeRange(actor))
    {
		int damage = ((M_RandomD() % 8) + 1) * 10;
		actor->flags2 |= MF2_MELEE;
		S_StartSound(sfx_claw, actor);
		P_DamageMobj (actor->target, actor, actor, damage, 13);
		return;
    }
    
	actor->flags2 &= ~MF2_MELEE;
	
    // launch a missile
    P_SpawnMissile (actor, actor->target, MT_BRUISERSHOT);
}

void __cdecl A_SkelMissile(mobj_t *actor)
{	
    mobj_t*	mo;
	
    if (!validActorTarget(actor, 16)) return;
		
    A_FaceTarget (actor);

    actor->z += 16*FRACUNIT;	// so missile spawns higher
    mo = P_SpawnMissile (actor, actor->target, MT_TRACER);
    actor->z -= 16*FRACUNIT;	// back to normal

	if (!mo) return;	// Missile was removed

    mo->x += mo->momx;
    mo->y += mo->momy;
	P_SetTarget(&mo->tracer, actor->target);	// killough 11/98
}

void __cdecl A_Tracer(mobj_t *actor)
{
	angle_t	exact;
	fixed_t	dist;
	fixed_t	slope;
	mobj_t*	dest;
	mobj_t*	th;
	
	if (gametic & 3) return;

	// spawn a puff of smoke behind the skel's missile
	// v220-35 not wanted
//	if (!r_use_particles) P_SpawnPuff(actor->x, actor->y, actor->z, 0);

	th = P_SpawnMobj (
		actor->x - actor->momx,
		actor->y - actor->momy,
		actor->z, MT_SMOKE, 0);
	
	th->momz = FRACUNIT;
	th->tics -= (M_RandomD() & 3);
	if (th->tics < 1) th->tics = 1;
	
	// adjust direction
	dest = actor->tracer;
	
	if (!dest || dest->health <= 0)
		return;
	
	// change angle	
	exact = R_PointToAngle2 (actor->x, actor->y, dest->x, dest->y);
	
	if (exact != actor->angle)
	{
		if (exact - actor->angle > 0x80000000)
		{
			actor->angle -= TRACEANGLE;
			if (exact - actor->angle < 0x80000000)
				actor->angle = exact;
		}
		else
		{
			actor->angle += TRACEANGLE;
			if (exact - actor->angle > 0x80000000)
				actor->angle = exact;
		}
	}
	
	exact = actor->angle>>ANGLETOFINESHIFT;
	actor->momx = FixedMul (actor->info->speed, finecosine[exact]);
	actor->momy = FixedMul (actor->info->speed, finesine[exact]);
	
	// change slope
	dist = P_ApproxDistance (dest->x - actor->x,
		dest->y - actor->y);
	
	dist = dist / actor->info->speed;
	
	if (dist < 1)
		dist = 1;
	slope = (dest->z+40*FRACUNIT - actor->z) / dist;
	
	if (slope < actor->momz)
		actor->momz -= FRACUNIT/8;
	else
		actor->momz += FRACUNIT/8;

	P_SetMpos(th);
}


void __cdecl A_SkelWhoosh(mobj_t *actor)
{
    if (!validActorTarget(actor, 17)) return;

    A_FaceTarget (actor);

    S_StartSound(sfx_skeswg, actor);
}

void __cdecl A_SkelFist(mobj_t *actor)
{
    int		damage;

    if (!validActorTarget(actor, 17)) return;
		
    A_FaceTarget (actor);
	
    if (P_CheckMeleeRange (actor))
    {
		damage = ((M_RandomD() % 10) + 1) * 6;
		S_StartSound(sfx_skepch, actor);
		P_DamageMobj(actor->target, actor, actor, damage, 14);
    }
}

//
// PIT_VileCheck
// Detect a corpse that could be raised.
// v220-31 tweaked
//
static mobj_t	*corpsehit;
static fixed_t	viletryx;
static fixed_t	viletryy;

RDBOOL PIT_VileCheck(mobj_t *thing, void *data)
{
    if ( !(thing->flags & MF_CORPSE)			// not a dead monster
		|| thing->tics != -1					// not lying still yet
		|| thing->info->raisestate == S_NULL	// has no raise state
		|| thing->flags3 & MF3_AMODEL)			// or is an amodel
	{
		return true;
	}

	int maxdist = thing->info->radius + mobjinfo[MT_VILE].radius;
	
	if (   abs(thing->x - viletryx) > maxdist
		|| abs(thing->y - viletryy) > maxdist)
	{
		return true;							// not actually touching
	}
	
	corpsehit = thing;
	corpsehit->momx = corpsehit->momy = 0;
	
	// temporarily save current vars
	int height = corpsehit->height;
	int radius = corpsehit->radius;
	int flags  = corpsehit->flags;
	int flags2 = corpsehit->flags2;

	// Change to active state before calling P_CheckPosition
	corpsehit->health = corpsehit->info->spawnhealth;
	corpsehit->height = corpsehit->info->height;
	corpsehit->radius = corpsehit->info->radius;
	corpsehit->flags = corpsehit->info->flags;
	corpsehit->flags2 &= ~(MF2_CORPSEFADE|MF2_TELEPCORPSE|MF2_CANSLOPE);

	// v220-5 check not blocked by another Thing (including vertically i.e. none above or below)	
	if ( !P_CheckPosition(corpsehit, corpsehit->x, corpsehit->y, corpsehit->z, -1) )
	{
		// reset changed vars back
		corpsehit->height = height;
		corpsehit->radius = radius;
		corpsehit->flags = flags;
		corpsehit->flags2 = flags2;
		corpsehit->health = 0;

		return true;	// doesn't fit here
	}
	
	return false;		// got one, so stop checking
}

// Check for resurrecting a body
void __cdecl A_VileChase(mobj_t *actor)
{	
    int			xl, xh, yl, yh;    
    int			bx, by, bbox[4];

    if (actor->movedir != DI_NODIR)
    {
		// check for corpses to raise
		viletryx =
			actor->x + actor->info->speed*xspeed[actor->movedir];
		viletryy =
			actor->y + actor->info->speed*yspeed[actor->movedir];

		bbox[BOXLEFT] = bbox[BOXRIGHT] = viletryx;
		bbox[BOXBASE] = bbox[BOXTOP  ] = viletryy;
		
		M_getBlockOffsets(bbox, MAXRADIUS*2, &xl, &xh, &yl, &yh);

		validcount++;

		for (bx=xl ; bx<=xh ; bx++)
		{
			for (by=yl ; by<=yh ; by++)
			{
				// PIT_VileCheck checks whether object is a resurrectable corpse
				if (!P_BlockThingsIterator(bx, by, (int (__cdecl *)(struct mobj_s *,void *))PIT_VileCheck, NULL))
				{
					// got one!
					ma_inf.mobjsResurrected++;

					mobj_t *temp = actor->target;
					actor->target = corpsehit;
					A_FaceTarget (actor);
					actor->target = temp;
					
					P_SetMobjState (actor, S_VILE_HEAL1);
					S_StartSound(sfx_slop, corpsehit);
					
					P_SetMobjState (corpsehit,corpsehit->info->raisestate);

					// Any MF2_TELEPCORPSE types will now be treated normally
					// i.e. will fade when next killed if corpsetime is set
					corpsehit->corpsetics = 0;
					corpsehit->inverseAlpha = 0;
					corpsehit->flags3 = corpsehit->info->flags3 & MF3_TRANSFERMASK;	// v220-31

					// killough 7/18/98: 
					// friendliness is transferred from AV to raised corpse
					if (actor->flags3 & MF3_FRIEND)
						corpsehit->flags3 |= MF3_FRIEND;
					else
						corpsehit->flags3 &= ~MF3_FRIEND;

					// killough 8/29/98: reset targets and add to appropriate thread
					P_SetTarget(&corpsehit->target, NULL);
					P_SetTarget(&corpsehit->tracer, NULL);
					P_SetTarget(&corpsehit->lastenemy, NULL);
					P_UpdateThinker(&corpsehit->thinker);
					return;
				}
			}
		}
    }

    A_Chase (actor);	// Return to normal attack
}

void __cdecl A_VileStart(mobj_t *actor)
{
	S_StartSound(sfx_vilatk, actor);
}

// Keep fire in front of player unless out of sight
void __cdecl A_StartFire(mobj_t *actor)
{
	S_StartSound(sfx_flamst, actor);
    A_Fire(actor);
}

void __cdecl A_FireCrackle(mobj_t *actor)
{
	S_StartSound(sfx_flame, actor);
    A_Fire(actor);
}

void __cdecl A_Fire(mobj_t *actor)
{
    mobj_t		*dest = actor->tracer;
    unsigned	an;
	
    if (!dest) return;
	
    // don't move it if the vile lost sight
    if (!P_CheckSight(dest, actor->target) ) return;
	
    an = dest->angle >> ANGLETOFINESHIFT;
	
    P_UnlinkThing (actor);
    actor->x = dest->x + FixedMul (24*FRACUNIT, finecosine[an]);
    actor->y = dest->y + FixedMul (24*FRACUNIT, finesine[an]);
    actor->z = dest->z;
    P_SetThingPosition (actor);
	P_SetMpos(actor);

}

mobjtype_t subfnGetRandomType()
{
	// Randomly select monster to spawn.
	int r = M_RandomD();

	// Probability distribution (kind of :),
	// decreasing likelihood.
	if ( r<50 ) return MT_TROOP;
	if (r<90) return MT_SERGEANT;
	if (r<120) return MT_SHADOWS;
	if (r<130) return MT_PAIN;
	if (r<160) return MT_HEAD;
	if (r<162) return MT_VILE;
	if (r<172) return MT_UNDEAD;
	if (r<192) return MT_BABY;
	if (r<222) return MT_FATSO;
	if (r<246) return MT_KNIGHT;
	return MT_BRUISER;		
}

// Spawn the hellfire
void __cdecl A_VileTarget(mobj_t *actor)
{
    mobj_t	*fog;
	
    if (!validActorTarget(actor, 18)) return;

	mobj_t	*targ = actor->target;
	
	if (actor->flags4 & MF4_VILESPAWNER)
	{
		mobjtype_t t = subfnGetRandomType();
		//  Spawn teleport fog.
		fog = P_SpawnMobj(targ->x, targ->y, targ->z, MT_SPAWNFIRE, 0);
		P_SetMpos(fog);
		S_StartSound(sfx_telept, fog);
		
		mobj_t*	mo = P_SpawnMobj(targ->x, targ->y, targ->z, t, 0);
		mo->flags |= (MF_SHADOW|MF_NOCLIP);
		mo->ddflags |= DDMF_SHADOW;
		mo->flags2 |= MF2_NOSHADOW;

		if (t == MT_PAIN) mo->flags4 = MF4_TRANSPAIN;

		// friendliness is transferred
		mo->flags3 = (mo->flags3 & ~MF3_FRIEND) | (mo->flags3 & MF3_FRIEND);

		// killough 8/29/98: add to appropriate thread
		P_UpdateThinker(&mo->thinker);
		P_SetMpos(mo);

		if (P_LookForPlayers(mo, true) )
			P_SetMobjState(mo, mo->info->seestate);

		return;
	}

    A_FaceTarget (actor);

	fog = P_SpawnMobj(targ->x, targ->y, targ->z, MT_FIRE, 0);
    
	P_SetTarget(&actor->tracer, fog);   // killough 11/98
	P_SetTarget(&fog->target, actor);
	P_SetTarget(&fog->tracer, actor->target);

    A_Fire (fog);
}

void __cdecl A_VileAttack(mobj_t *actor)
{	
	mobj_t*	fire;
	int		an;

	if (!validActorTarget(actor, 19)) return;

	A_FaceTarget (actor);

	if (!P_CheckSight(actor, actor->target) ) return;

	S_StartSound(sfx_barexp, actor);
	P_DamageMobj (actor->target, actor, actor, 20, 15);
	actor->target->momz = 1000*FRACUNIT/actor->target->info->mass;

	an = actor->angle >> ANGLETOFINESHIFT;

	if ( !(fire = actor->tracer) ) return;

	// move the fire between the vile and the player
	fire->x = actor->target->x - FixedMul (24*FRACUNIT, finecosine[an]);
	fire->y = actor->target->y - FixedMul (24*FRACUNIT, finesine[an]);	
	P_RadiusAttack (fire, actor, 70 );
}

//
// Mancubus attack,
// firing three missiles (bruisers)
// in three different directions?
// Doesn't look like it. 
//
#define	FATSPREAD	(ANG90/8)

void __cdecl A_FatRaise(mobj_t *actor)
{
    A_FaceTarget (actor);
	S_StartSound(sfx_manatk, actor);
}

void __cdecl A_FatAttack1(mobj_t *actor)
{
	mobj_t*	mo;
	int		an;
	
	if (!validActorTarget(actor, 20)) return;
	
	A_FaceTarget (actor);
	// Change direction  to ...
	actor->angle += FATSPREAD;

	if (actor == actor->target) return;

	if ( !P_SpawnMissile (actor, actor->target, MT_FATSHOT) )
		return;	// Missile was removed
	
	if ( !(mo = P_SpawnMissile (actor, actor->target, MT_FATSHOT)) )
		return;	// Missile was removed

	mo->angle += FATSPREAD;
	an = mo->angle >> ANGLETOFINESHIFT;
	mo->momx = FixedMul (mo->info->speed, finecosine[an]);
	mo->momy = FixedMul (mo->info->speed, finesine[an]);
}

void __cdecl A_FatAttack2(mobj_t *actor)
{
	mobj_t*	mo;
	int		an;
	
	if (!validActorTarget(actor, 21)) return;
	
	if (actor == actor->target) return;

	A_FaceTarget (actor);
	// Now here choose opposite deviation.
	actor->angle -= FATSPREAD;

	if ( !P_SpawnMissile (actor, actor->target, MT_FATSHOT) )
		return;	// Missile was removed
	
	if ( !(mo = P_SpawnMissile (actor, actor->target, MT_FATSHOT)) )
		return;	// Missile was removed

	mo->angle -= FATSPREAD*2;
	an = mo->angle >> ANGLETOFINESHIFT;
	mo->momx = FixedMul (mo->info->speed, finecosine[an]);
	mo->momy = FixedMul (mo->info->speed, finesine[an]);
}

void __cdecl A_FatAttack3(mobj_t *actor)
{
    mobj_t*	mo;
    int		an;

    if (!validActorTarget(actor, 22)) return;

    A_FaceTarget (actor);
    
	if ( !(mo = P_SpawnMissile (actor, actor->target, MT_FATSHOT)) )
		return;	// Missile was removed
	
	mo->angle -= FATSPREAD/2;
	an = mo->angle >> ANGLETOFINESHIFT;
	mo->momx = FixedMul (mo->info->speed, finecosine[an]);
	mo->momy = FixedMul (mo->info->speed, finesine[an]);

	if ( !(mo = P_SpawnMissile (actor, actor->target, MT_FATSHOT)) )
		return;	// Missile was removed

    mo->angle += FATSPREAD/2;
    an = mo->angle >> ANGLETOFINESHIFT;
    mo->momx = FixedMul (mo->info->speed, finecosine[an]);
    mo->momy = FixedMul (mo->info->speed, finesine[an]);
}

//
// SkullAttack
// Fly at the player like a missile.
//
#define	SKULLSPEED		(20*FRACUNIT)

void __cdecl A_SkullAttack(mobj_t *actor)
{
    mobj_t*		dest;
    angle_t		an;
    int			dist;

    if (!validActorTarget(actor, 23)) return;
		
    dest = actor->target;	
    actor->flags |= MF_SKULLFLY;

    S_StartSound(actor->info->attacksound, actor);
    A_FaceTarget (actor);
    an = actor->angle >> ANGLETOFINESHIFT;
    actor->momx = FixedMul (SKULLSPEED, finecosine[an]);
    actor->momy = FixedMul (SKULLSPEED, finesine[an]);
    dist = P_ApproxDistance (dest->x - actor->x, dest->y - actor->y);
    dist = dist / SKULLSPEED;
    
    if (dist < 1)
	dist = 1;
    actor->momz = (dest->z+(dest->height>>1) - actor->z) / dist;
	P_SetMpos(actor);
}

//
// A_PainShootSkull
// Spawn a lost soul and launch it at the target
void A_PainShootSkull(mobj_t *actor, angle_t angle)
{ 
    mobj_t*	newmobj;
    angle_t	an;
    fixed_t	x, y, z;
    int		prestep;

	if (compat[c_skulllimit])	// v220-3
	{
		thinker_t*	th = thinkercap.next;
		int			count = 20;	// max of 20 skulls in level

		for ( ;th != &thinkercap; th = th->next)
		{
			if (   th->function == (void (__cdecl *)(void))P_MobjThinker
				&& ((mobj_t *)th)->type == MT_SKULL)
			{
				if (--count < 0) return;
			}
		}
	}

    an = angle >> ANGLETOFINESHIFT;
    
    prestep = 4*FRACUNIT + 
			3*(actor->info->radius + mobjinfo[MT_SKULL].radius)/2;
    
    x = actor->x + FixedMul (prestep, finecosine[an]);
    y = actor->y + FixedMul (prestep, finesine[an]);
    z = actor->z + 8*FRACUNIT;
		
	// Check whether the Lost Soul is being fired through a 1-sided
	// wall or an impassible line, or a "monsters can't cross" line.
	// If it is, then we don't allow the spawn. - phares
	
	if (Check_Sides(actor, x, y)) return;
	
	newmobj = P_SpawnMobj(x, y, z, MT_SKULL, 0);

	// killough 7/20/98: PEs shoot lost souls with the same friendliness
	newmobj->flags3 = (newmobj->flags3 & ~MF3_FRIEND) | (actor->flags3 & MF3_FRIEND);

	// killough 8/29/98: add to appropriate thread
	P_UpdateThinker(&newmobj->thinker);
	
	// Check to see if the new Lost Soul's z value is above the ceiling
	// of its new sector, or below the floor. If so, kill it. - phares
	
	if (   (newmobj->z > (newmobj->subsector->sector->ceilingheight - newmobj->height))
		|| (newmobj->z < newmobj->subsector->sector->floorheight)
		|| !P_TryMove (newmobj, newmobj->x, newmobj->y, 2))  // killough $dropoff_fix
	{
		P_RemoveMobj(newmobj, 29);	// v220-4 don't kill it, remove it. This type is never resurrected
		return;
	}

	if (actor->flags4 & MF4_TRANSPAIN)	// COD make skulls translucent
	{
		newmobj->flags |= MF_SHADOW;
		newmobj->ddflags |= DDMF_SHADOW;
		newmobj->flags2 |= MF2_NOSHADOW;
	}
		
	P_SetMpos(newmobj);
	P_SetTarget(&newmobj->target, actor->target);
    A_SkullAttack (newmobj);
}

// Spawn a lost soul and launch it at the target
void __cdecl A_PainAttack(mobj_t *actor)
{
    if (!validActorTarget(actor, 24)) return;

    A_FaceTarget (actor);
    A_PainShootSkull (actor, actor->angle);
}

void __cdecl A_PainDie(mobj_t *actor)
{
    A_Fall (actor);
    A_PainShootSkull (actor, actor->angle+ANG90);
    A_PainShootSkull (actor, actor->angle+ANG180);
    A_PainShootSkull (actor, actor->angle+ANG270);
}

void __cdecl A_Scream(mobj_t *actor)
{
	int sound = actor->info->deathsound & ~HSF_FLAG_MASK;
	
	switch (sound)
	{
	case 0:
		return;
		
	case sfx_barexp:
		if (   actor->flags2 & MF2_NORENDER
			&& !(actor->flags3 & MF3_TOUCHY))
			return;
		break;
		
	case sfx_podth1:
	case sfx_podth2:
	case sfx_podth3:
		sound = sfx_podth1 + (M_RandomD() % 3);
		break;
		
	case sfx_bgdth1:
	case sfx_bgdth2:
		sound = sfx_bgdth1 + (M_RandomD() % 2);
		break;
		
	default:
		break;
	}

	actor->ddflags |= DDF_DEATHSOUND;	// flag as a death sound to prevent sound being stopped in P_RemoveMobj
	S_StartSound(sound, actor);
}

void __cdecl A_XScream(mobj_t *actor)
{
    S_StartSound(sfx_slop, actor);	
}

void __cdecl A_Pain(mobj_t *actor)
{
	if (actor->flags2 & MF2_NOAIRSUPPLY)
		return;

	int snd;

	if (sthing && (snd = actor->info->painsound) & HSF_SNDTHING)	// v220-33
	{
		soundthing_t *sth = &sthing[ (snd & ~HSF_SNDTHING) ];
		if (sth->oldsoundid) S_StartSound(sth->oldsoundid, actor);
	}
	else
		S_StartSound(actor->info->painsound, actor);
}

void __cdecl A_Fall(mobj_t *actor)
{
	if ( !(actor->flags3 & MF3_ALWAYSBLOCK) )	// $v71
	{
		// actor is on ground, it can be walked over
		// So change this if corpse objects are meant to be obstacles.
		actor->flags &= ~MF_SOLID;
	}
}

void __cdecl A_Explode(mobj_t *thingy)
{
    P_RadiusAttack(thingy, thingy->target, 128);
}

//
// A_BossDeath
// Check for special effects - Rewritten to use R3Dflags
//

// make sure there is a live player for victory and all bosses are dead
RDBOOL subfn_ChkValidEndConditions(mobj_t *mo)
{
    thinker_t	*th;

	if ( !(mbf_players[0].plr->ingame && plhealth.health > 0) )
			return false;
	
	// scan the remaining thinkers to see if all bosses are dead
	for (th = thinkercap.next; th != &thinkercap; th = th->next)
	{
		if (th->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t *mo2 = (mobj_t*)th;

			if (mo2->type == mo->type
				&& mo2 != mo
				&& mo2->health > 0)
			{
				return false;	// other boss not dead
			}
		}
	}

	return true;
}

// v220-19 scripted dmednum things 6085 to 6086
void Chk608xDeath(mobj_t *mo)
{
	int i;

	for (i = 5; i < MAXDM_NUM; i++)	// 7305 to 7309
	{
		if (deathmonitor[i].type == mo->type)
		{
			if (subfn_ChkValidEndConditions(mo))
				subfn_UseSpecialLine(ppt.viewplrmo, deathmonitor[i].linenum, 13);

			break;
		}
	}
}
		
void __cdecl A_BossDeath(mobj_t *mo)
{
	int		R3Dflags = mapinfo->R3Dflags;
    line_t	junk;

	junk.tag = 666;	// default
	junk.lnflags2 = 0;

	// This can now fade
	mo->flags2 &= ~MF2_NOFADE;

	// v220-4 scripted deathmonitor
	if (R3D_ScriptedMap && mo->type)
	{
		int i;

		for (i = 0; i < 5; i++)	// 7300 to 7304
		{
			if (deathmonitor[i].type == mo->type)
			{
				if (subfn_ChkValidEndConditions(mo))
					subfn_UseSpecialLine(ppt.viewplrmo, deathmonitor[i].linenum, 13);

				break;
			}
		}
	}

	if (R3Dflags & specialaction_mask)
	{
		int type = mo->type;

		if (   R3Dflags & map07special
			&& (type == MT_FATSO || type == MT_BABY)
			&& subfn_ChkValidEndConditions(mo))
		{
			if (R3Dflags & specialaction_killmonsters)
				P_Massacre(false);

			if (type == MT_FATSO)
				EV_DoFloor(&junk, lowerFloorToLowest);
			else
			{
				junk.tag = 667;
				EV_DoFloor(&junk, raiseToTexture);
			}
		}
		else
		{
			if (  ((R3Dflags & spiderbabyspecial       && type == MT_BABY)
				|| (R3Dflags & mancubusspecial         && type == MT_FATSO)
				|| (R3Dflags & baronspecial            && type == MT_BRUISER)
				|| (R3Dflags & cyberdemonspecial       && type == MT_CYBORG)
				|| (R3Dflags & spidermastermindspecial && type == MT_SPIDER))
				&& subfn_ChkValidEndConditions(mo))
			{
				if (R3Dflags & specialaction_killmonsters) P_Massacre(false);
				if (R3Dflags & specialaction_lowerfloor  ) EV_DoFloor(&junk, lowerFloorToLowest);
				if (R3Dflags & specialaction_opendoor    ) EV_DoDoor(&junk, openNormal);
				if (R3Dflags & specialaction_blazedoor   ) EV_DoDoor(&junk, blazeOpen);
				if (R3Dflags & specialaction_exitlevel   ) G_ExitLevel();
			}
		}
	}
}

void __cdecl A_Hoof(mobj_t *mo)
{
	if (ma_inf.mapEXM8)
		S_StartSound(sfx_hoof|HSF_NO_ATTENUATION, mo);
	else
		S_StartSound(sfx_hoof, mo);

    A_Chase (mo);
}

void __cdecl A_Metal(mobj_t *mo)
{
	if (ma_inf.mapEXM8)
		S_StartSound(sfx_metal|HSF_NO_ATTENUATION, mo);
	else
		S_StartSound(sfx_metal, mo);

    A_Chase (mo);
}

void __cdecl A_BabyMetal(mobj_t *mo)
{
	S_StartSound(sfx_bspwlk, mo);
    A_Chase (mo);
}

void __cdecl A_BrainAwake(mobj_t *mo)
{
	S_StartSound(sfx_bossit, NULL);
}

void __cdecl A_BrainPain(mobj_t *mo)
{
	S_StartSound(sfx_bospn, NULL);
}

void __cdecl A_BrainScream(mobj_t *mo)
{
    int		x;
    int		y;
    int		z;
    mobj_t*	th;
	
    for (x=mo->x - 196*FRACUNIT ; x< mo->x + 320*FRACUNIT ; x+= FRACUNIT*8)
    {
		y = mo->y - 320*FRACUNIT;
		z = 128 + M_RandomD() * 2 * FRACUNIT;
		th = P_SpawnMobj (x,y,z, MT_ROCKET, 0);
		th->momz = M_RandomD() * 512;
		
		P_SetMobjState (th, S_BRAINEXPLODE1);
		
		th->tics -= (M_RandomD() & 7);
		if (th->tics < 1)
			th->tics = 1;
    }
	
	P_SetMpos(th);

	S_StartSound(sfx_bosdth, NULL);
}

void __cdecl A_BrainExplode(mobj_t *mo)
{
    int		x;
    int		y;
    int		z;
    mobj_t*	th;
	
    x = mo->x + (M_RandomD() - M_RandomD()) * 2048;
    y = mo->y;
    z = 128 + M_RandomD() * 2 * FRACUNIT;
    th = P_SpawnMobj (x,y,z, MT_ROCKET, 0);
    th->momz = M_RandomD() * 512;

	P_SetMpos(th);
    P_SetMobjState (th, S_BRAINEXPLODE1);

    th->tics -= (M_RandomD() & 7);
    if (th->tics < 1)
		th->tics = 1;
}

void __cdecl A_BrainDie(mobj_t *mo)
{
    G_ExitLevel ();
}

// Modified for no braintargets or for multiple bosses sharing targets
void __cdecl A_BrainSpit(mobj_t *mo)
{
    mobj_t	*targ, *newmobj;
	fixed_t	v;

	// v220-3 add nomonsters
	if (nomonsters || !numbraintargets)
	{
		return;
	}
	
    braineasy ^= 1;
    if (gameskill <= sk_easy && !braineasy) return;
	
    // shoot a cube at current target
    targ = braintargets[braintargeton++];
    braintargeton %= numbraintargets;
	
    // spawn brain missile
	// Note: this type normally has MF_NOCLIP set so won't
	//       have been removed by P_CheckMissileSpawn().
	//       The following test is done in case flags have been dehacked
    if ( !(newmobj = P_SpawnMissile(mo, targ, MT_SPAWNSHOT)) )
		return;

	P_SetTarget(&newmobj->target, targ);

	newmobj->reactiontime = 1;	// default

	if (newmobj->state->tics)	// defensive coding - should not be 0
	{
		// v220-5 fix Doom bug where momy could be 0
		if (abs(newmobj->momx) > abs(newmobj->momy))
			v = (targ->x - mo->x) / newmobj->momx;
		else
			v = (targ->y - mo->y) / newmobj->momy;

		// reactiontime is used as a tics counter in A_SpawnFly.
		// Counting these down should time out the cube when it
		// reaches the target.
		newmobj->reactiontime = v / newmobj->state->tics;
	}

	// killough 7/18/98: brain friendliness is transferred
	newmobj->flags3 = (newmobj->flags3 & ~MF3_FRIEND) | (mo->flags3 & MF3_FRIEND);

	// killough 8/29/98: add to appropriate thread
	P_UpdateThinker(&newmobj->thinker);

	S_StartSound(sfx_bospit, NULL);
}

// travelling cube sound
void __cdecl A_SpawnSound(mobj_t *mo)	
{
    S_StartSound(sfx_boscub, mo);
    A_SpawnFly(mo);
}

void __cdecl A_SpawnFly(mobj_t *mo)	// cube spawner
{
	mobj_t*	newmobj;
	mobj_t*	fog;
	mobj_t*	targ;

	if (--mo->reactiontime) return;	// still flying

	// R3D something has gone seriously wrong if no target is set
	if ( !(targ = mo->target) ) P_SetMobjAbort();

	// First spawn teleport fog.
	fog = P_SpawnMobj (targ->x, targ->y, targ->z, MT_SPAWNFIRE, 0);
	P_SetMpos(fog);
	S_StartSound(sfx_telept, fog);
	
	newmobj	= P_SpawnMobj (targ->x, targ->y, targ->z, subfnGetRandomType(), 0);
	
	// killough 7/18/98: brain friendliness is transferred
	newmobj->flags3 = (newmobj->flags3 & ~MF3_FRIEND) | (mo->flags3 & MF3_FRIEND);
	
	// killough 8/29/98: add to appropriate thread
	P_UpdateThinker(&newmobj->thinker);
	P_SetMpos(newmobj);
	
	if (P_LookForTargets(newmobj, true))
		P_SetMobjState(newmobj, newmobj->info->seestate);
	
	// telefrag anything in this spot
	P_TeleportMove(newmobj, newmobj->x, newmobj->y, 0, false, true);
	P_SetMpos(newmobj);	// v220-5
	
	// remove self (i.e., cube).
	P_RemoveMobj(mo, 31);
}

void __cdecl A_PlayerScream(mobj_t *mo)
{
	int sound;

	if (mo->flags2 & MF2_NOAIRSUPPLY)
		return;

	if (commercial_game && mo->health < -50)
	{
		// IF THE PLAYER DIES
		// LESS THAN -50% WITHOUT GIBBING
		sound = sfx_pdiehi;
	}
	else
		sound = sfx_pldeth;    // Default death sound.
    
    S_StartSound(sound, mo);
}

// MBF additional DEH accessible code pointers

// killough 11/98: kill an object
void __cdecl A_Die(mobj_t *actor)
{
	if (compat[c_nocodep]) return;

	P_DamageMobj(actor, NULL, NULL, actor->health, 16);
}

// A_Detonate
// killough 8/9/98: same as A_Explode, except that the damage is variable
void __cdecl A_Detonate(mobj_t *mo)
{
	if (compat[c_nocodep]) return;

	P_RadiusAttack(mo, mo->target, mo->info->damage);
}

mobj_t *P_SpawnMCPMissile(mobj_t* source, mobj_t* dest, mobjtype_t type);

// killough 9/98: a mushroom explosion effect, sorta :)
// Original idea: Linguica
void __cdecl A_Mushroom(mobj_t *actor)
{
	fixed_t misc1, misc2;
	int		i, j, n;

	//-------------------------------

	if (compat[c_nocodep]) return;

	// Mushroom parameters are part of code pointer's state
	// Note: misc vals usually used for pspr offsets
	misc1 = actor->state->misct[0] ? actor->state->misct[0] : FRACUNIT*4;
	misc2 = actor->state->misct[1] ? actor->state->misct[1] : FRACUNIT/2;

	n = actor->info->damage;

	A_Explode(actor);               // make normal explosion

	actor->flags2 |= MF2_NOFADE;	// disable corpsefade

	for (i = -n; i <= n; i += 8)    // launch mushroom cloud
	{
		for (j = -n; j <= n; j += 8)
		{
			mobj_t target = *actor, *mo;

			// Aim in many directions from source
			target.x += i << FRACBITS;
			target.y += j << FRACBITS;

			// Aim fairly high
			target.z += P_ApproxDistance(i,j) * misc1;

			// Launch fireball
			if ( (mo = P_SpawnMCPMissile(actor, &target, MT_FATSHOT)) )
			{
				// Slow down a bit
				mo->momx = FixedMul(mo->momx, misc2);
				mo->momy = FixedMul(mo->momy, misc2);
				mo->momz = FixedMul(mo->momz, misc2);

				// Make debris fall under gravity
				mo->flags &= ~MF_NOGRAVITY;
			}
		}
	}
}

//
// killough 11/98
//
// The following were inspired by Len Pitre
//
// A small set of highly-sought-after code pointers
//

void __cdecl A_Spawn(mobj_t *mo)
{
	if (compat[c_nocodep]) return;

	int s1 = mo->state->misct[0];
	int s2 = mo->state->misct[1];

	int n = s1 - 1;

	if (n > 0 && n < DH_NUMOBJECTS)
	{
		mobj_t *newmobj = 
			P_SpawnMobj(mo->x, mo->y, (s2 << FRACBITS) + mo->z, (mobjtype_t)n, 0);

		newmobj->flags3 |= (newmobj->flags3 & ~MF3_FRIEND) | (mo->flags3 & MF3_FRIEND);
	}
}

void __cdecl A_Turn(mobj_t *mo)
{
	if (compat[c_nocodep]) return;

	mo->angle += (unsigned int)(((unsigned __int64) mo->state->misct[0] << 32) / 360);
}

void __cdecl A_Face(mobj_t *mo)
{
	if (compat[c_nocodep]) return;

	mo->angle = (unsigned int)(((unsigned __int64) mo->state->misct[0] << 32) / 360);
}

void __cdecl A_Scratch(mobj_t *mo)
{
	if (compat[c_nocodep] || !mo->target) return;

	A_FaceTarget(mo);

	(P_CheckMeleeRange(mo)) ?
		mo->state->misct[1] ? S_StartSound(mo->state->misct[1], mo) : (void) 0,
		P_DamageMobj(mo->target, mo, mo, mo->state->misct[0], 17) : (void) 0;
}

void __cdecl A_PlaySound(mobj_t *mo)
{
	if (compat[c_nocodep]) return;

	// v220-35 attenuate these sounds by doubling actual distance from the source if source is not NULL
	S_StartSoundAtVolume(mo->state->misct[0], mo->state->misct[1] ? NULL : mo, 0.8f, mo->state->misct[1] ? 1.f : 2.f);
}

void __cdecl A_RandomJump(mobj_t *mo)
{
	if (compat[c_nocodep]) return;

	if (M_RandomD() < mo->state->misct[1])
		P_SetMobjState(mo, mo->state->misct[0]);
}

//
// This allows linedef effects to be activated inside deh frames.
//

void __cdecl A_LineEffect(mobj_t *mo)
{
	static line_t	junk;
	ddplayer_t		plr;
	mbf_player_t	player;
	mbf_player_t	*oldplayer;
	mobj_t			pmo;
	
	//------------------------------

	if (compat[c_nocodep] || R3D_maptype) return;
	
	junk = *lines;

	if (!(junk.special = (short)mo->state->misct[0])) return;

	junk.tag = (short)mo->state->misct[1];
	junk.ml_flags = 0;

	oldplayer = mo->player;
	mo->player = &player;

	plr.mo =  &pmo;
	player.plr = &plr;
	pmo.health = 100;

	if (!P_UseSpecialLine(mo, &junk, 0, 0))
		P_CrossSpecialLine(&junk, mo);

	mo->state->misct[0] = junk.special;
	mo->player = oldplayer;
}

// End of MBF additional DEH accessible code pointers


// R3D
void __cdecl A_NullHandler(mobj_t* mo)
{
	// do nothing
}


// v220-13 modified
RDBOOL CheckAction(state_t *st)
{
	int n = st->action_num;

	return  n == nmCPosAttack
		||  n == nmCPosRefire
		||  n == nmdbAttack
		||  n == nmPosAttack
		||  n == nmSPosAttack;
}

// Spawn a BFG explosion on every monster in view
void __cdecl A_BFGSpray(mobj_t *mo) 
{
	mobj_t	*ltmo;
    int		i, k, damage;

    // offset angles from the attack angle
    for (i = 0; i < 40; i++)
    {
		angle_t an = mo->angle - ANG90/2 + ANG90/40*i;

		// mo->target is the originator (player) of the missile
		if ((P_AimLineAttack (mo->target, an, MEDIUMRANGE, MF3_FRIEND),
			!mobjtarget))
			P_AimLineAttack (mo->target, an, MEDIUMRANGE, 0);
		
		if (!mobjtarget) continue;

		ltmo = P_SpawnMobj (mobjtarget->x,
			mobjtarget->y,
			mobjtarget->z + (mobjtarget->height>>2),
			MT_EXTRABFG, 0);
		
		P_SetMpos(ltmo);

		for (k = 0, damage = 0; k < 15; k++)
			damage += ((M_RandomD() & 7) + 1);
		
		P_DamageMobj (mobjtarget, mo->target,mo->target, damage, 18);
    }
}

void __cdecl A_OpenShotgun2(mbf_player_t *player, pspdef_t *psp)
{
    S_StartSound(sfx_dbopn, player->plr->mo);
}

void __cdecl A_LoadShotgun2(mbf_player_t *player, pspdef_t *psp)
{
    S_StartSound(sfx_dbload, player->plr->mo);
}

#pragma optimize("", on)