//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_floor.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//		T_MoveFloor revised (R3D flag SF_FLOORSTOPSOUND removed)
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
//
// DESCRIPTION:
//	Floor animation: raising stairs.
//
//-----------------------------------------------------------------------------

#include "r3d_compat.h"

#include "p_data.h"
#include "doom/p_spec.h"
#include "r_data.h"
#include "r_main.h"
#include "s_main.h"
#include "con_console.h"

RDBOOL P_ChangeSector(sector_t *sector, sector_t *sec3D, RDBOOL crunch);

int		plat_direction;

#pragma optimize("g", off)

//-----------------------------------------------------------------------------

//
// P_FindModelFloorSector()
//
// Passed a floor height and a sector number, return a pointer to a
// sector with that floor height across the lowest numbered two sided
// line surrounding the sector.
//
// Note: If no sector at that height bounds the sector passed, return NULL
//
// jff 02/03/98 Add routine to find numeric model floor
//  around a sector specified by sector number
// jff 3/14/98 change first parameter to plain height to allow call
//  from routine not using floormove_t
//
// killough 11/98: reformatted

sector_t *P_FindModelFloorSector(fixed_t floordestheight, int secnum)
{
	sector_t	*sec;
	int			i;
	
	for (i = 0; i < sectors[secnum].linecount; i++)
	{
		if ( twoSided(secnum, i) )
		{
			if (getSide(secnum, i, 0)->sector - sectors == secnum)
				sec = getSector(secnum, i, 1);
			else
				sec = getSector(secnum, i, 0);
			
			if (sec->floorheight == floordestheight)
				return sec;
		}
	}

	return NULL;
}

//=============================================================================

// v220-10
floormove_t *AssignFloor(sector_t *s, floor_e type, RDBOOL crush)
{
	floormove_t *floor = (floormove_t*)W_Calloc(sizeof(floormove_t), THINKID, 0);

	s->floordata = floor;
	floor->sector = s;
	floor->type = type;	//jff 2/26/98 should always be set
	floor->crush = crush;
	floor->thinker.function = (void (__cdecl *)(void))T_MoveFloor;
	P_AddFNThinker (&floor->thinker);
	return floor;
}

//
// FLOORS
//

// R3D check for linked 3D sectors - gmj
// Revised v220-31
static int P_CheckLinkedSectors(sector_t *sec, RDBOOL crush)
{
	int *listsec = sec->listsec, s;

	if (listsec)
	{
		while ((s = *(int*)listsec++) >= 0)
		{
			if (P_ChangeSector(sec, &sectors[s], -1))	// 3D lifts always crush
			{
				return true;;
			}
		}
		return false;
	}

	return P_ChangeSector(sec, NULL, crush);
}

//
// Move a plane (floor or ceiling) and check for crushing
// Revised v220-31
result_e T_MovePlane(sector_t*	sector,
					 fixed_t	speed,
					 fixed_t	dest,
					 int		crush,
					 int		floorOrCeiling,
					 int		direction )
{
	fixed_t	destheight;
	fixed_t lastpos, offset;	//jff 02/04/98 used to keep floors/ceilings from moving thru each other
	fixed_t *pspeed = &sector->planes[floorOrCeiling].speed;
	
	// store dest and speed (dest is fixed, speed may need adjusting)
	// these values are needed for 3D lifts
	sector->planes[floorOrCeiling].target = dest;
	*pspeed = speed;
	
	// R3D - global plat_direction added
	// If a floor plat is moving up then set to -1 else if
	// a ceil plat moving up when flagged down then set to 1
	plat_direction = 0;

	switch(floorOrCeiling)
	{
	case 0:

		// FLOOR
		switch(direction)
		{
		case -1:
			// DOWN

			// v220-35 stop all plat movement if player is dead
			if (   crushed
				&& ppt.viewplrmo->flags & MF_CORPSE
				&& ppt.viewplrmo->subsector->sector == sector
				&& sector->floorheight - speed < ppt.viewplrmo->floorz + 8*FRACUNIT)
			{
				if (!ma_inf.no_light[0]) ma_inf.no_light[0] = 1;
				return crushed;
			}

			if (sector->floorheight - speed < dest)
			{
				*pspeed = dest - sector->floorheight;
				lastpos = sector->floorheight;
				sector->floorheight = dest;

				if (P_CheckLinkedSectors(sector, crush))
				{
					sector->floorheight = lastpos;
					P_CheckLinkedSectors(sector, crush);
				}
				return pastdest;
			}
			else
			{
				plat_direction = -1;
				lastpos = sector->floorheight;
				sector->floorheight -= speed;

				if (P_CheckLinkedSectors(sector, crush))
				{
					sector->floorheight = lastpos;

					if (ma_inf.modifier & MODF_FLOORFIX)
					{
						P_ChangeSector(sector, NULL, crush);
						return crushed;
					}

					P_CheckLinkedSectors(sector, crush);
					plat_direction = 0;

					return crushed;
				}
				plat_direction = 0;
			}
			break;

		case 1:
			// Moving a floor up
			// offset by floorslope height
			offset = sector->stptr >= 0 && sector->floordata ?
							slopes[sector->stptr].slope[0].slopeheight * FRACUNIT : 0;

			// jff 02/04/98 keep floor from moving thru ceilings
			destheight = (ma_inf.modifier & MODF_FLOORFIX || dest < sector->ceilingheight) ?
					dest : sector->ceilingheight;

			destheight -= offset;

			// v220-35 stop all plat movement if player is dead
			if (   crushed
				&& ppt.viewplrmo->flags & MF_CORPSE
				&& ppt.viewplrmo->subsector->sector == sector
				&& sector->floorheight - speed < ppt.viewplrmo->floorz + 8*FRACUNIT)
			{
				if (!ma_inf.no_light[0]) ma_inf.no_light[0] = 1;
				return crushed;
			}
			
			if (sector->floorheight + speed > destheight)
			{
				lastpos = sector->floorheight;
				*pspeed = destheight - sector->floorheight;
				sector->floorheight = destheight;

				if (P_CheckLinkedSectors(sector, crush))
				{
					sector->floorheight = lastpos;
					
					P_CheckLinkedSectors(sector, crush);
				}

				return pastdest;
			}
			else
			{
				// COULD GET CRUSHED
				lastpos = sector->floorheight;
				sector->floorheight += speed;
				
				if (P_CheckLinkedSectors(sector, crush))
				{
					if (crush && ma_inf.modifier & MODF_FLOORFIX)
						return crushed;

					sector->floorheight = lastpos;
					
					P_CheckLinkedSectors(sector, crush);

					return crushed;	// 3D lift or crush
				}
			}
		}

		break;
		
	case 1:
		// CEILING
		switch(direction)
		{
		case -1:
			// moving a ceiling down

			// offset by ceilingslope height
			offset = sector->stptr >= 0 && sector->ceilingdata ?
							slopes[sector->stptr].slope[1].slopeheight * FRACUNIT : 0;

			// jff 02/04/98 keep ceiling from moving thru floors
			destheight = (ma_inf.modifier & MODF_FLOORFIX || dest > sector->floorheight + offset) ?
					dest : sector->floorheight + offset;
			
			// v220-35 stop all plat movement if player is dead
			if (   crushed
				&& ppt.viewplrmo->flags & MF_CORPSE
				&& ppt.viewplrmo->subsector->sector == sector
				&& sector->floorheight - speed < ppt.viewplrmo->floorz + 8*FRACUNIT)
			{
				if (!ma_inf.no_light[0]) ma_inf.no_light[0] = 1;
				return crushed;
			}
			
			if (sector->ceilingheight - speed < destheight)
			{
				*pspeed = destheight - sector->ceilingheight;
				// v220-9 destheight check. If ceiling is actually going to
				//        a higher destination then set to 1
				if (destheight > sector->ceilingheight)
					plat_direction = 1;

				lastpos = sector->ceilingheight;
				sector->ceilingheight = destheight;

				if (P_ChangeSector(sector, NULL, crush))
				{
					sector->ceilingheight = lastpos;

					P_ChangeSector(sector, NULL, crush);
				}

				plat_direction = 0;

				return pastdest;
			}
			else
			{
				// COULD GET CRUSHED
				lastpos = sector->ceilingheight;
				sector->ceilingheight -= speed;
				
				if (P_ChangeSector(sector, NULL, crush))
				{
					if (crush == true) return crushed;

					sector->ceilingheight = lastpos;

					//	sector->ceilingheight = lastpos;
					P_ChangeSector(sector, NULL, crush);
					return crushed;
				}
			}
			break;
			
		case 1:
			// UP

			if (sector->ceilingheight + speed > dest)
			{
				lastpos = sector->ceilingheight;
				sector->ceilingheight = dest;
				*pspeed = dest - sector->ceilingheight;

				if (P_ChangeSector(sector, NULL, crush))
				{
					sector->ceilingheight = lastpos;

					P_ChangeSector(sector, NULL, crush);
				}
				
				return pastdest;
			}
			else
			{
				sector->ceilingheight += speed;
				P_ChangeSector(sector, NULL, crush);
			}

			break;
		}
		break;	
	}
	return ok;
}

//
// MOVE A FLOOR TO IT'S DESTINATION (UP OR DOWN)
//
void T_MoveFloor(floormove_t* floor)
{
	result_e	res;
	short		newspecial = floor->newspecial;
	RDBOOL		hassound = !(floor->sector->specFlags2 & SF2_NOMOVSOUND);
	RDBOOL		fakewater = hassound && floor->sector->specFlags2 & SF2_FAKEMOVSECTOR;

	res = T_MovePlane(
		floor->sector,
		floor->speed,
		floor->floordestheight,
		floor->crush,
		0,
		floor->direction);

	if (res != pastdest)
	{
		if (hassound)
		{
			// R3D change sound if a rising/lowering fake water sector
			if (fakewater)
				S_SectorSoundRepeat(r3ds.fakewt_id, floor->sector);
			else
				S_SectorSoundRepeat(r3ds.stnmov_id, floor->sector);
		}

		return;
	}

	if (floor->direction == 1)
	{
		switch(floor->type)
		{
		case donutRaise:
			floor->sector->special = newspecial;
			floor->sector->floorpic = floor->texture;
			break;

		case genFloorChgT:
		case genFloorChg0:
			floor->sector->special = newspecial;
			//jff add to fix bug in special transfers from changes
			floor->sector->oldspecial = floor->oldspecial;
			floor->sector->floorpic = floor->texture;
			break;

		case genFloorChg:
			floor->sector->floorpic = floor->texture;
			break;
		default:
			break;
		}
	}
	else if (floor->direction == -1)
	{
		switch(floor->type)
		{
		case lowerAndChange:
			floor->sector->special = newspecial;
			floor->sector->floorpic = floor->texture;
			break;

		case genFloorChgT:
		case genFloorChg0:
			floor->sector->special = newspecial;
			//jff add to fix bug in special transfers from changes
			floor->sector->oldspecial = floor->oldspecial;
			floor->sector->floorpic = floor->texture;
			break;

		case genFloorChg:
			floor->sector->floorpic = floor->texture;
			break;

		default:
			break;
		}
	}

	if (hassound)
	{
		// Stop repeating floor sound
		if (fakewater)
			S_StopSound(r3ds.fakewt_id, (mobj_t*)&floor->sector->soundorg);
		else
		{
			S_StopSound(r3ds.stnmov_id, (mobj_t*)&floor->sector->soundorg);
 			S_SectorSound(floor->sector, sfx_pstop);
		}
	}

	floor->sector->planes[PLN_FLOOR].speed = 0;
	floor->sector->floordata = NULL; //jff 2/22/98

	P_RemoveThinker(&floor->thinker);
	
	//jff 2/26/98 implement stair retrigger lockout while still building
	// note this only applies to the retriggerable generalized stairs
	
	if (floor->sector->stairlock == -2)		// if this sector is stairlocked
	{
		sector_t *sec = floor->sector;
		sec->stairlock = -1;				// thinker done, promote lock to -1
		
		while (sec->prevsec >= 0 && sectors[sec->prevsec].stairlock != -2)
			sec = &sectors[sec->prevsec];	// search for a non-done thinker

		if (sec->prevsec == -1)				// if all thinkers previous are done
		{
			sec = floor->sector;

			// search forward
			while (sec->nextsec >= 0 && sectors[sec->nextsec].stairlock != -2) 
				sec = &sectors[sec->nextsec];

			if (sec->nextsec == -1)			// if all thinkers ahead are done too
			{
				while (sec->prevsec >= 0)	// clear all locks
				{
					sec->stairlock = 0;
					sec = &sectors[sec->prevsec];
				}
				sec->stairlock = 0;
			}
		}
	}
}

//
// T_MoveElevator()
//
// Move an elevator to it's destination (up or down)
// Called once per tick for each moving floor.
//
// Passed an elevator_t structure that contains all pertinent info about the
// move. See P_SPEC.H for fields.
// No return.
//
// jff 02/22/98 added to support parallel floor/ceiling motion
//
void T_MoveElevator(elevator_t* elevator)
{
	result_e	  res;

	if (elevator->direction < 0)		// moving down
	{
		// v220-31 ensure a lowering R3D type continues to crush
		if (  (elevator->type == elevateR3DDown || elevator->sector->listsec)
			&& elevator->sector->floordata
			&& elevator->direction < 0)
		{
			res = T_MovePlane
				(
				elevator->sector,
				elevator->speed,
				elevator->floordestheight,
				0,
				0,						  // move ceiling
				elevator->direction
				);
		}
		else
		{
			res = T_MovePlane			  //jff 4/7/98 reverse order of ceiling/floor
				(
				elevator->sector,
				elevator->speed,
				elevator->ceilingdestheight,
				0,
				1,							// move floor
				elevator->direction
				);
			if (res==ok || res==pastdest) // jff 4/7/98 don't move ceil if blocked
			{
				T_MovePlane
				(
				elevator->sector,
				elevator->speed,
				elevator->floordestheight,
				0,
				0,						  // move ceiling
				elevator->direction
				);
			}
		}
	}
	else // up
	{
		res = T_MovePlane			  //jff 4/7/98 reverse order of ceiling/floor
			(
			elevator->sector,
			elevator->speed,
			elevator->floordestheight,
			0,
			0,							// move ceiling
			elevator->direction
			);
		if (res==ok || res==pastdest) // jff 4/7/98 don't move floor if blocked
		{
			T_MovePlane
			(
			elevator->sector,
			elevator->speed,
			elevator->ceilingdestheight,
			0,
			1,						  // move floor
			elevator->direction
			);
		}
	}

	// If its an R3D linked 3D floor sector then fix the ceilingheight
	if (elevator->sector->R3Dceilheight != DDMININT) elevator->sector->ceilingheight = elevator->sector->R3Dceilheight;
	
	if (res == pastdest)			// if destination height achieved
	{
		// only make floor stop sound if it has a move sound
		if (S_IsPlaying(r3ds.stnmov_id, (mobj_t*) &elevator->sector->soundorg))
		{
			S_SectorSound(elevator->sector, sfx_pstop);
			S_StopSound(r3ds.stnmov_id, (mobj_t*) &elevator->sector->soundorg);
		}
		
		elevator->sector->planes[0].speed = 0;
		elevator->sector->floordata = NULL; 	//jff 2/22/98
		elevator->sector->ceilingdata = NULL;	//jff 2/22/98
		P_RemoveThinker(&elevator->thinker);	// remove elevator from actives
		return;
	}
	
	// make floor move sound
	if (!(elevator->sector->specFlags2 & SF2_NOMOVSOUND))
		S_SectorSoundRepeat(r3ds.stnmov_id, elevator->sector);
}

//
// HANDLE FLOOR TYPES
//

int EV_DoFloor(line_t *line, floor_e floortype)
{
	floormove_t	*floor;
	sector_t	*sec;
	int		 	secnum = -1, rtn = 0;
	int			floorspeed = FLOORSPEED;
	
	while ((secnum = P_FindSectorFromTag(line->tag, secnum)) >= 0)
	{
		sec = &sectors[secnum];
		
		// ALREADY MOVING?	IF SO, KEEP GOING...
		if (P_SectorActive(floor_special, sec)) //jff 2/23/98
			continue;

		switch(floortype)
		{
		default:
			floorspeed = FLOORSPEED;
			break;
		// R3D - Note: lowerFloor32Turbo and raiseFloor32Turbo
		//             never had specials assigned.
		case turboLower:
		case lowerFloor32Turbo:
		case raiseFloor32Turbo:
		case raiseFloorTurbo:
			floorspeed = FLOORSPEED*4;
			break;
		}

		if (sec->specFlags2 & SF2_FAKEMOVSECTOR)	// R3D
			floorspeed >>= 2;	// Floors 4*slower
		
		// new floor thinker
		floor = AssignFloor(sec, floortype, floortype == raiseFloorCrush);
		floor->speed = floorspeed;

		rtn = 1;

		switch(floortype)
		{
		case lowerFloor:
			floor->direction = -1;
			floor->floordestheight = P_FindHighestFloorSurrounding(sec);
			break;

			//jff 02/03/30 support lowering floor by 24 absolute
		case lowerFloor24:
			floor->direction = -1;
			floor->floordestheight = floor->sector->floorheight + 24 * FRACUNIT;
			break;

			//jff 02/03/30 support lowering floor by 32 absolute (fast)
		case lowerFloor32Turbo:
			floor->direction = -1;
			floor->floordestheight = floor->sector->floorheight + 32 * FRACUNIT;
			break;

		case lowerFloorToLowest:
			floor->direction = -1;
			floor->floordestheight = P_FindLowestFloorSurrounding(sec);
			break;

			//jff 02/03/30 support lowering floor to next lowest floor
		case lowerFloorToNearest:
			floor->direction = -1;
			floor->floordestheight = P_FindNextLowestFloor(sec,floor->sector->floorheight);
			break;

		case turboLower:
			floor->direction = -1;
			floor->floordestheight = P_FindHighestFloorSurrounding(sec);
			if (floor->floordestheight != sec->floorheight)
				floor->floordestheight += 8*FRACUNIT;
			break;

		case raiseFloorCrush:
		case raiseFloor:
			floor->direction = 1;
			floor->floordestheight = P_FindLowestCeilingSurrounding(sec);
			if (floor->floordestheight > sec->ceilingheight)
				floor->floordestheight = sec->ceilingheight;
			if (floortype == raiseFloorCrush)
				floor->floordestheight -= 8*FRACUNIT;
			break;

		case raiseFloorTurbo:
			floor->direction = 1;
			floor->floordestheight = P_FindNextHighestFloor(sec,sec->floorheight);
			break;

		case raiseFloorToNearest:
			floor->direction = 1;
			floor->floordestheight = P_FindNextHighestFloor(sec,sec->floorheight);
			break;

		case raiseFloor24:
			floor->direction = 1;
			floor->floordestheight = floor->sector->floorheight + 24 * FRACUNIT;
			break;

			// jff 2/03/30 support straight raise by 32 (fast)
		case raiseFloor32Turbo:
			floor->direction = 1;
			floor->floordestheight = floor->sector->floorheight + 32 * FRACUNIT;
			break;

		case raiseFloor512:
			floor->direction = 1;
			floor->floordestheight = floor->sector->floorheight + 512 * FRACUNIT;
			break;

		case raiseFloor24AndChange:
			floor->direction = 1;
			floor->floordestheight = floor->sector->floorheight + 24 * FRACUNIT;
			sec->floorpic = line->frontsector->floorpic;
			sec->special = line->frontsector->special;
			//jff 3/14/98 transfer both old and new special
			sec->oldspecial = line->frontsector->oldspecial;
			break;

		case raiseToTexture: 
			floor->direction = 1;

			{
				int texval = P_FindShortestTextureAround(secnum);

				if (compat[c_model])
					floor->floordestheight = floor->sector->floorheight + (texval<<FRACBITS);
				else
				{
					floor->floordestheight = targetMAX(floor->sector->floorheight, texval);
				}
			}
			break;

		case lowerAndChange:
			floor->direction = -1;
			floor->floordestheight = P_FindLowestFloorSurrounding(sec);
			floor->texture = sec->floorpic;

			// jff 1/24/98 make sure floor->newspecial gets initialized
			// in case no surrounding sector is at floordestheight
			floor->newspecial = sec->special; 
			//jff 3/14/98 transfer both old and new special
			floor->oldspecial = sec->oldspecial;

			//jff 5/23/98 use model subroutine to unify fixes and handling
			if ( ( sec = P_FindModelFloorSector(floor->floordestheight, secnum)) )
			{
				floor->texture = sec->floorpic;
				floor->newspecial = sec->special;
				//jff 3/14/98 transfer both old and new special
				floor->oldspecial = sec->oldspecial;
			}
			break;

		default:
			break;
		}
	}
	return rtn;
}

//
// EV_DoChange()
//
// Handle pure change types. These change floor texture and sector type
// by trigger or numeric model without moving the floor.
//
// The linedef causing the change and the type of change is passed
// Returns true if any sector changes
//
// jff 3/15/98 added to better support generalized sector types
//
int EV_DoChange(line_t* line, change_e changetype )
{
	int                   secnum;
	int                   rtn;
	sector_t*             sec;
	sector_t*             secm;
	
	secnum = -1;
	rtn = 0;
	// change all sectors with the same tag as the linedef
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		sec = &sectors[secnum];
		
		rtn = 1;
		
		// handle trigger or numeric change type
		switch(changetype)
		{
		case trigChangeOnly:
			sec->floorpic = line->frontsector->floorpic;
			sec->special = line->frontsector->special;
			sec->oldspecial = line->frontsector->oldspecial;
			break;
		case numChangeOnly:
			secm = P_FindModelFloorSector(sec->floorheight,secnum);
			if (secm) // if no model, no change
			{
				sec->floorpic = secm->floorpic;
				sec->special = secm->special;
				sec->oldspecial = secm->oldspecial;
			}
			break;
		default:
			break;
		}
	}
	return rtn;
}


/*
 * EV_BuildStairs()
 *
 * Handles staircase building. A sequence of sectors chosen by algorithm
 * rise at a speed indicated to a height that increases by the stepsize
 * each step.
 *
 * Passed the linedef triggering the stairs and the type of stair rise
 * Returns true if any thinkers are created
 * 
 * cph 2001/09/21 - compatibility nightmares again
 * There are three different ways this function has, during its history, stepped
 * through all the stairs to be triggered by the single switch
 * - original Doom used a linear P_FindSectorFromTag, but failed to preserve
 * the index of the previous sector found, so instead it would restart its
 * linear search from the last sector of the previous staircase
 * - MBF/PrBoom with comp_stairs fail to emulate this, because their
 * P_FindSectorFromTag is a chained hash table implementation. Instead they
 * start following the hash chain from the last sector of the previous
 * staircase, which will (probably) have the wrong tag, so they miss any further
 * stairs
 * - Boom fixed the bug, and MBF/PrBoom without comp_stairs work right
 */
int P_FindSectorFromTagWithLowerBound(int tag, int start, int min)
{
	/* Emulate original Doom's linear lower-bounded P_FindSectorFromTag
	* as needed */
	do
	{
		start = P_FindSectorFromTag(tag,start);
	} while (start >= 0 && start <= min);
	return start;
}

int EV_BuildStairs(line_t* line, stair_e type)
{
	sector_t	*sec;
	int			i, secnum;
	int			s = -1,mins = -1;
	int			tag = line->tag;
	int			rtn = 0;

	// set up the speed and stepsize according to the stairs type (turbo16 or build8)
	if (type != turbo16) type = build8;	// Insurance
	fixed_t	stairsize = type == turbo16 ? FRACUNIT*16  : FRACUNIT*8;
	fixed_t speed = type == turbo16 ? FLOORSPEED*4 : FLOORSPEED/4;

	// v220-5 support type with no line tag
	if (!tag)
	{
		if (line->backsector) tag = line->backsector->tag;
		if (!tag) return 0;
		s = line->backsector - sectors;
		goto manual_stair;
	}

	// start a stair at each sector tagged with tag
	while ((s = P_FindSectorFromTagWithLowerBound(tag, s, mins)) >= 0)
	{
manual_stair:;
		sec = &sectors[s];
		secnum = s;

		// don't start a stair if the first step's floor is already moving
		if (!P_SectorActive(floor_special, sec))
		{
			floormove_t	*floor;
			int			texture, height;
			RDBOOL		repeat = true;
			
			rtn = 1;

			// create new floor thinker for first step
			floor = AssignFloor(sec, buildStair, false);
			floor->direction = 1;
			floor->speed = speed;
			height = sec->floorheight + stairsize;
			floor->floordestheight = height;

			texture = sec->floorpic;

			// Find next sector to raise
			//   1. Find 2-sided line with same sector side[0] (lowest numbered)
			//   2. Other side is the next sector to raise
			//   3. Unless already moving, or different texture, then stop building

			while (repeat)      // repeat until no next step is found
			{
				repeat = false;

				for (i = 0; i < sec->linecount; i++)
				{
					sector_t *tsec = sec->lines[i]->frontsector;

					if (secnum != tsec - sectors
						|| !(sec->lines[i]->ml_flags & ML_TWOSIDED)
						// if sector's floor is different texture, look for another
						|| (tsec = sec->lines[i]->backsector)->floorpic != texture)
					{
						continue;
					}

					if (compat[c_stairs]) height += stairsize;

					// if sector's floor already moving, look for another - jff 2/22/98
					if (P_SectorActive(floor_special, tsec))
						continue;

					if (!compat[c_stairs])
					{
						height += stairsize;
					
						// v220-13 If floor height unchanged then exit
						// v220-31 made contingent on compat test
						if (tsec->floorheight == height) return rtn;
					}
					
					sec    = tsec;
					secnum = tsec - sectors;
					
					// create and initialize a thinker for the next step
					floor = AssignFloor(sec, buildStair, type != build8);
					floor->direction = 1;
					floor->speed = speed;
					floor->floordestheight = height;

					repeat = true;
					break;
				}
			}
		}
		if (compat[c_stairs])  s = -1, mins = secnum;
	}
	return rtn;
}
  
int EV_DoDonut(line_t*	line)
{
	sector_t*		s1;
	sector_t*		s2;
	sector_t*		s3;
	int			secnum;
	int			rtn;
	int			i;
	floormove_t*	floor;
	
	secnum = -1;
	rtn = 0;
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		s1 = &sectors[secnum];
		
		// do not start the donut if the pillar is already moving
		if (P_SectorActive(floor_special,s1)) //jff 2/22/98
			continue;
		
		s2 = getNextSector(s1->lines[0],s1);
		if (!s2) continue;                    // note lowest numbered line around

		// do not start the donut if the pool is already moving
		if (P_SectorActive(floor_special,s2))
			continue;                           //jff 5/7/98
		
		for (i = 0; i < s2->linecount; i++)
		{
			if (compat[c_model])
			{
				if (!(s2->lines[i]->ml_flags & ML_TWOSIDED) ||
					(s2->lines[i]->backsector == s1))
					continue;
			}
			else
			if (!s2->lines[i]->backsector || s2->lines[i]->backsector == s1)
				continue;
			
			rtn = 1; //jff 1/26/98 no donut action - no switch change on return
			
			s3 = s2->lines[i]->backsector;      // s3 is model sector for changes
			
			//	Spawn rising slime
			floor = AssignFloor(s2, donutRaise, false);
			floor->direction = 1;
			floor->speed = FLOORSPEED / 2;
			floor->texture = s3->floorpic;
			floor->newspecial = 0;
			floor->floordestheight = s3->floorheight;
			
			//	Spawn lowering donut-hole
			floor = AssignFloor(s1, lowerFloor, false);
			floor->direction = -1;
			floor->speed = FLOORSPEED / 2;
			floor->floordestheight = s3->floorheight;
			break;
		}
	}
	return rtn;
}


//
// EV_DoElevator
//
// Handle elevator linedef types
//
// Passed the linedef that triggered the elevator and the elevator action
//
// jff 2/22/98 new type to move floor and ceiling in parallel
//
// Revised v220-31 Speed change using 17000 type Things added (search elevspeed)
int EV_DoElevator(line_t *line, elevator_e elevtype)
{
	elevator_t	*elevator;
	sector_t	*sec;
	int			secnum = -1;
	int			rtn = 0;
	
	// v220-31 3D floor change vars
	RDBOOL	use3Dheights;
	fixed_t gotofloorheight;

	// -------------------------------------------

	// act on all sectors with the same tag as the triggering linedef
	while ((secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
	{
		use3Dheights = false;

		sec = &sectors[secnum];
		if (!(sec->specFlags & SF_MOVSECTOR)) continue;
		
		// If either floor or ceiling is already activated, skip it
		if (sec->floordata || sec->ceilingdata) //jff 2/22/98
			continue;
		
		int speed = ELEVATORSPEED;	// default

		if (elevtype != elevateCurrent)
		{
			// v220-31 see if it's been changed
			if (elevspeedset || (elevheightset && line->tag))
			{
				for (int i = 0; i < 99; i++)
				{
					if (!elevspeed[i][ELV_HEIGHT])
					{
						if (elevspeed[i][ELV_SPEED] && elevspeed[i][ELV_SECN] == secnum)
						{
							speed = elevspeed[i][ELV_SPEED];
							if (elevtype == elevateDown)
								gotofloorheight = P_FindNextLowestFloor(sec,sec->floorheight);
							else
								gotofloorheight = P_FindNextHighestFloor(sec,sec->floorheight);

							use3Dheights = true;
							if (elevtype == elevateDown)
								elevtype = elevateR3DDown;
							break;
						}
					}
					else if (elevheightset && line->tag && elevspeed[i][ELV_TAG] == line->tag)
					{
						fixed_t elv_height = elevspeed[i][ELV_HEIGHT];

						if (elevtype == elevateDown || elevtype == elevateR3DDown)
						{
							if (!P_CheckCompositeLower(line->tag, elv_height, elevspeed[i][ELV_LOLIMIT])) return false;

							gotofloorheight = sec->floorheight - elv_height;
							elevtype = elevateR3DDown;
						}
						else
						{
							if (!P_CheckCompositeRaise(line->tag, elv_height, elevspeed[i][ELV_HILIMIT])) return false;

							gotofloorheight = sec->floorheight + elv_height;

							if (sec->stptr >= 0)
								gotofloorheight += slopes[sec->stptr].slope[0].slopeheight * FRACUNIT;
						}

						speed = elevspeed[i][ELV_SPEED];
						use3Dheights = true;
						break;
					}
				}
			}
		}

		// create and initialize new elevator thinker
		rtn = 1;
		elevator = (elevator_t*)W_Calloc(sizeof(*elevator), THINKID, 0);
		elevator->type = elevtype;
		sec->floordata = elevator; //jff 2/22/98
		sec->ceilingdata = elevator; //jff 2/22/98
		elevator->thinker.function = (void (__cdecl *)(void))T_MoveElevator;
		P_AddFNThinker (&elevator->thinker);
		
		// set up the fields according to the type of elevator action
		switch(elevtype)
		{
			// elevator down to next floor
		case elevateDown:
		case elevateR3DDown:
			elevator->direction = -1;
			elevator->sector = sec;
			elevator->speed = speed;
			elevator->floordestheight = use3Dheights ? gotofloorheight :
				P_FindNextLowestFloor(sec,sec->floorheight);
			elevator->ceilingdestheight = use3Dheights ? sec->ceilingheight :
				elevator->floordestheight + sec->ceilingheight - sec->floorheight;
			break;
			
			// elevator up to next floor
		case elevateUp:
			elevator->direction = 1;
			elevator->sector = sec;
			elevator->speed = speed;
			elevator->floordestheight = use3Dheights ? gotofloorheight :
				P_FindNextHighestFloor(sec,sec->floorheight);
			elevator->ceilingdestheight = use3Dheights ? sec->ceilingheight :
				elevator->floordestheight + sec->ceilingheight - sec->floorheight;
			break;
			
			// elevator to floor height of activating switch's front sector
			// v220-31 Note: not supported with 3D lifts
		case elevateCurrent:
			elevator->sector = sec;
			elevator->speed = speed;
			elevator->floordestheight = line->frontsector->floorheight;
			elevator->ceilingdestheight =
				elevator->floordestheight + sec->ceilingheight - sec->floorheight;
			elevator->direction = elevator->floordestheight > sec->floorheight?  1 : -1;
			break;
			
		default:
			break;
		}
	}
	return rtn;
}

#pragma optimize("", on)
