//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_lights.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
//  Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION:
//	Handle Sector base lighting effects.
//	Muzzle flash?
//
//-----------------------------------------------------------------------------

#include "r3d_compat.h"

#include "doom/p_spec.h"

#include "con_console.h"
#include "m_misc.h"

//
// FIRELIGHT FLICKER
//

//
// T_FireFlicker
//
void T_FireFlicker (fireflicker_t* flick)
{
    int	amount;
	
    if (--flick->count)
		return;
	
    amount = (M_RandomD() & 3) * 16;
    
    if (flick->sector->lightlevel - amount < flick->minlight)
		flick->sector->lightlevel = flick->minlight;
    else
		flick->sector->lightlevel = flick->maxlight - amount;
	
    flick->count = 4;
}

//
// T_LightFlash
// Do flashing lights.
//
void T_LightFlash (lightflash_t* flash)
{
    if (--flash->count)
		return;
	
    if (flash->sector->lightlevel == flash->maxlight)
    {
		flash-> sector->lightlevel = flash->minlight;
		flash->count = (M_RandomD() & flash->mintime) + 1;
    }
    else
    {
		flash-> sector->lightlevel = flash->maxlight;
		flash->count = (M_RandomD() & flash->maxtime) + 1;
    }
	
}

//
// T_StrobeFlash
//
// Strobe light flashing action routine, called once per tick
//
void T_StrobeFlash (strobe_t* flash)
{
    if (--flash->count) return;
	
    if (flash->sector->lightlevel == flash->minlight)
    {
		flash->sector->lightlevel = flash->maxlight;
		flash->count = flash->brighttime;
    }
    else
    {
		flash->sector->lightlevel = flash->minlight;
		flash->count = flash->darktime;
    }

}
    
//
// T_Glow()
//
// Glowing light action routine, called once per tick
//
void T_Glow(glow_t*	g)
{
	switch(g->direction)
	{
	case -1:
		// DOWN
		g->sector->lightlevel -= GLOWSPEED;

		if (g->sector->lightlevel <= g->minlight)
		{
			if (g->minlight == g->maxlight)	// v220-33 don't dither around median
			{
				g->sector->lightlevel = g->minlight;
			}
			else
			{
				g->sector->lightlevel += GLOWSPEED;
				g->direction = 1;
			}
		}
		break;
		
	case 1:
		// UP
		g->sector->lightlevel += GLOWSPEED;

		if (g->sector->lightlevel >= g->maxlight)
		{
			if (g->minlight == g->maxlight)	// v220-33 don't dither around median
			{
				g->sector->lightlevel = g->maxlight;
			}
			else
			{
				g->sector->lightlevel -= GLOWSPEED;
				g->direction = -1;
			}
		}
		break;
	}
}

//////////////////////////////////////////////////////////
//
// Sector lighting type spawners
//
// After the map has been loaded, each sector is scanned
// for specials that spawn thinkers
//
//////////////////////////////////////////////////////////

//
// P_SpawnFireFlicker
//
void P_SpawnFireFlicker (sector_t*	sector)
{
	fireflicker_t*	flick;

	// Note that we are resetting sector attributes.
	// Nothing special about it during gameplay.
	sector->special &= ~31;

	flick = (fireflicker_t*)W_Calloc(sizeof(*flick), THINKID, 0);
	flick->thinker.function = (void (__cdecl *)(void))T_FireFlicker;
	flick->sector = sector;
	flick->maxlight = sector->lightlevel;
	flick->minlight = P_FindMinSurroundingLight(sector,sector->lightlevel)+16;
	flick->count = 4;
	P_AddFNThinker(&flick->thinker);
}

//
// P_SpawnLightFlash
// After the map has been loaded, scan each sector
// for specials that spawn thinkers
//
void P_SpawnLightFlash (sector_t*	sector)
{
    lightflash_t*	flash;

    // nothing special about it during gameplay
    sector->special &= ~31;
	
    flash = (lightflash_t*)W_Calloc(sizeof(*flash), THINKID, 0);
    flash->thinker.function = (void (__cdecl *)(void))T_LightFlash;
    flash->sector = sector;
    flash->maxlight = sector->lightlevel;
    flash->minlight = P_FindMinSurroundingLight(sector,sector->lightlevel);
    flash->maxtime = 64;
    flash->mintime = 7;
    flash->count = (M_RandomD() & flash->maxtime) + 1;
    P_AddFNThinker (&flash->thinker);
}

//
// P_SpawnStrobeFlash
// After the map has been loaded, scan each sector
// for specials that spawn thinkers
//
void P_SpawnStrobeFlash(sector_t* sector, int fastOrSlow, int inSync)
{
	strobe_t *flash;
	
	flash = (strobe_t*)W_Calloc(sizeof(*flash), THINKID, 0);
	flash->sector = sector;
	flash->darktime = fastOrSlow;
	flash->brighttime = STROBEBRIGHT;
	flash->thinker.function = (void (__cdecl *)(void))T_StrobeFlash;
	flash->maxlight = sector->lightlevel;
	flash->minlight = P_FindMinSurroundingLight(sector, sector->lightlevel);
	
	if (flash->minlight == flash->maxlight) flash->minlight = 0;

	// nothing special about it during gameplay
	sector->special &= ~31;

	flash->count = inSync ? 1 : (M_RandomD() & 7) + 1;
	P_AddFNThinker (&flash->thinker);
}

//
// P_SpawnGlowingLight()
//
// Spawns a glowing light (smooth oscillation from min to max) thinker
//
// Passed the sector that spawned the thinker
//
void P_SpawnGlowingLight(sector_t* sector)
{
	glow_t	*glow = (glow_t*)W_Calloc(sizeof(glow_t), THINKID, 0);

	glow->maxlight = sector->lightlevel;
	glow->sector = sector;
	glow->minlight = P_FindMinSurroundingLight(sector, sector->lightlevel);
	glow->direction = -1;
	glow->thinker.function = (void (__cdecl *)(void))T_Glow;
	P_AddFNThinker(&glow->thinker);

	sector->special &= ~31;
}

//////////////////////////////////////////////////////////
//
// Linedef lighting function handlers
//
//////////////////////////////////////////////////////////

//
// EV_StartLightStrobing()
//
// Start strobing lights (usually from a trigger)
//
void EV_StartLightStrobing(line_t*	line)
{
    sector_t	*sector;
    int			secnum = -1;
	
    while ( (secnum = P_FindSectorFromTag(line->tag,secnum)) >= 0)
    {
		sector = &sectors[secnum];
		
		// if already doing a lighting function, don't start a second
		if (P_SectorActive(lighting_special,sector)) //jff 2/22/98
		  continue;

		P_SpawnStrobeFlash (sector, SLOWDARK, 0);
    }
}

//
// EV_TurnTagLightsOff()
//
// Turn line's tagged sector's lights to min adjacent neighbor level
//
void EV_TurnTagLightsOff(line_t* line)
{
	sector_t	*sector, *tsec;
	int			i, min, secnum = -1;;
	
	// search sectors for those with same tag as activating line
	
	// killough 10/98: replaced inefficient search with fast search
	while ( (secnum = P_FindSectorFromTag(line->tag, secnum)) >= 0 )
	{
		sector = &sectors[secnum];
		min = sector->lightlevel;
		
		// find min neighbor light level
		for (i = 0; i < sector->linecount; i++)
		{
			if (  (tsec = getNextSector(sector->lines[i], sector))
				&& tsec->lightlevel < min)
			{
				min = tsec->lightlevel;
			}
		}

		sector->lightlevel = min;
	}
}

//
// EV_LightTurnOn()
//
// Turn sectors tagged to line lights on to specified or max neighbor level
//
void EV_LightTurnOn(line_t* line, int bright)
{
	int i, tbright;
	
	// search all sectors for ones with same tag as activating line
	
	// killough 10/98: replace inefficient search with fast search
	for (i = -1; (i = P_FindSectorFromTag(line->tag,i)) >= 0;)
	{
		sector_t	*sector = sectors + i;

		// v220-31 deal with sector type 6 dynlights
		if (sector->special == 6)
		{
			for (int k = 0; k < trig5505.count; k++)
			{
				if (trig5505.seclist[k] == sector)
				{
					for (int m = 0; m < secmo[k].count; m++)
					{
						secmo[k].list[m]->torchtic = 255;
					}

					secmo[k].timer = secmo[k].period;
					break;
				}
			}
			continue;
		}

		tbright = bright; //jff 5/17/98 search for maximum PER sector
		
		if (!bright) // bright = 0 means search for highest light level surrounding sector
		{
			sector_t *temp;

			for (int k = 0; k < sector->linecount; k++)
			{
				if ((temp = getNextSector(sector->lines[k],sector)) &&
					temp->lightlevel > tbright)
					tbright = temp->lightlevel;
			}
		}
				
		sector->lightlevel = tbright;

		if (compat[c_model]) bright = tbright;
	}
}

// killough 10/98:
//
// EV_LightTurnOnPartway()
//
// Turn sectors tagged to line lights on to specified or max neighbor level
//
// Passed the activating line, and a light level fraction between 0 and 1.
// Sets the light to min on 0, max on 1, and interpolates in-between.
// Used for doors with gradual lighting effects.
//
// line must be set (not NULL)

void EV_LightTurnOnPartway(line_t *line, fixed_t level)
{
	int i;

	// clip at extremes 
	if (level < 0) level = 0;
	else
	if (level > FRACUNIT) level = FRACUNIT;
	
	// search all sectors for ones with same tag as activating line
	for (i = -1; (i = P_FindSectorFromTag(line->tag,i)) >= 0;)
	{
		sector_t *temp, *sector = sectors+i;
		int j, bright = 0, min = sector->lightlevel;
		
		// v220-31 eliminate special case
		if (sector->special == 6) continue;
		
		for (j = 0; j < sector->linecount; j++)
		{
			if ((temp = getNextSector(sector->lines[j],sector)))
			{
				if (temp->lightlevel > bright)
					bright = temp->lightlevel;
				if (temp->lightlevel < min)
					min = temp->lightlevel;
			}
		}
			
		sector->lightlevel =   // Set level in-between extremes
			(level * bright + (FRACUNIT-level) * min) >> FRACBITS;
	}
}

//v220-33 COD light support
void EV_LightSector(line_t *line, int value)
{
	int i, tmp;

	// search all sectors for ones with same tag as activating line
	for (i = -1; (i = P_FindSectorFromTag(line->tag,i)) >= 0;)
	{		
		// v220-31 eliminate special case
		if (sectors[i].special != 6)
		{
			tmp = sectors[i].lightlevel + value;

			// clip at extremes 
			if (tmp <= 0  ) tmp = 0;
			if (tmp >  255) tmp = 255;
			
			sectors[i].lightlevel = tmp;
		}
	}
}
