//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//  Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_map.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 2 17/Jun/2008 gmj (b220-2)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 11 (v220-11) 08/Aug/2011 gmj
// Revision 2.2.0 build 12 (v220-12) 01/Oct/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 14 (v220-14) 01/May/2012 gmj
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj BFG specific handling removed
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
// Revision 2.2.0 build 28 (v220-28) 04/Oct/2014 gmj
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 32 (v220-32) 18/Mar/2017 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
// DESCRIPTION:
//	Movement, collision handling.
//	Shooting and aiming.
//
//-----------------------------------------------------------------------------


// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "doom/mbf_data.h"
#include "doom/p_local.h"
#include "doom/tables.h"
#include "doom/m_cheat.h"
#include "con_console.h"
#include "dd_wad.h"
#include "net_main.h"
#include "m_bams.h"
#include "p_maputil.h"
#include "p_sight.h"
#include "r_data.h"
#include "r_main.h"
#include "r_things.h"
#include "s_main.h"
#include "m_misc.h"
#include "doom\d_englsh.h"

#include "r3d_edit.h"
#include "r3d_MAINspec.h"
#include "r3d_slopes.h"

// MACROS ------------------------------------------------------------------

#define MORE_FRICTION_MOMENTUM 15000       // mud factor based on momentum

// TYPES -------------------------------------------------------------------

typedef struct {
	line_t	*ld;
	int		tpspec;
} splist_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	P_TouchDEHKeyThing(mobj_t *mo);
void	P_ClearThingSRVO(mobj_t *mo, RDBOOL clrmpos);
void	ForceNoClipStatus(void);
RDBOOL	checkLineSpawnedTP(fixed_t x, fixed_t y, fixed_t r);


// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

void	G_ExitLevel ();

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int		plat_direction;		// b211-0
extern int		locked_door;		// v220-7
extern RDBOOL	resetNextViewer;	// v220-35 used with P_TeleportMove

// shared only with p_maputil.cpp
extern fixed_t	opentop;
extern fixed_t	openbottom;
extern fixed_t	openrange;
extern fixed_t	lowfloor;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// keep track of special lines as they are hit,
// 1/11/98 killough: removed limit on special lines crossed

plane3d_t	plane3d;
line_t		**spechit = NULL;	// new code -- killough
RDBOOL		make_noway_sound;
int			numspechit  =  0;
int			numsplist  =  0;
RDBOOL		killthing;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// All following defs added for R3D

static struct adjustp_t
{
	subsector_t	*contacted;
	fixed_t		x , y , z;
	fixed_t		floorz;
	fixed_t		ceilz;
	short		zflat;

} dpm;

// v220-4 moline
//        This is used to keep track of all contacted lines
//        for an actor whilst it is being processed
static struct molinelist_t
{
	line_t	**plist;
	int		count;
	int		maxcount;

} mll;

static mobj_t	*mobj_mblist[MAXMODELBOX];
static int		mobj_mblistcount;

// v220-4 molist
//        This is used to keep track of all contacted mobjs
//        for an actor whilst it is being processed
#define	MAXMOLISTNUM 64		// should be more than enough
static mobj_t	dummy3D;	// v220-8
static mobj_t	*molist[MAXMOLISTNUM];
static int		molistnum;

static int		spechit_max = 0;
static int		splist_max = 0;

// Used with walkover specials
static splist_t *splist = NULL;

// CODE --------------------------------------------------------------------

// v220-10 Spechits handling revised (also see check_spechits)
//         Array added to return a short set in the range 0-6
//         for line specials in the range 0 to 269
//

// v220-14 Type 5 now 6 and type 6 now 7
//         New type 5 introduced
/*
	returns;
	0) use type
	1) walkover type
	2) walkover teleport to sector with TPman
	3) walkover teleport line to line type
	4) walkover exit special
	5) teleport switch types
	6) gun type (only rejected using this array ATM)
	7) scrollers general
	9) to be ignored in all cases
*/

/*
	Addendum
	gun specials 24, 46, 47, 197, 198

	These do not have their specials set to 0 as
	they will be needed if restoring a saved game;
	special == 223	// friction
	special >= 214 && special <= 218	// accelerative scrollers
	special >= 245 && special <= 249	// displacement scrollers
	special >= 250 && special <= 253	// scroll effects

*/

// v220-29 optimisation seems to create a race hazard causing line to line tp'ing
// to occasionally fail. No failures have yet been observed when not optimised.
#pragma optimize("g",off)

byte walkCheck[] = {
						// NOTE: Prior to v220-12 case 10 was not set
	9,					// cases following are all those other than use specials
	0,1,1,1,1,1,0,1,	// case 2: case 3: case 4: case 5: case 6: case 8:
	0,1,0,1,1,0,0,1,	// case 10: case 12: case 13: case 16:
	1,0,1,0,0,1,0,6,	// case 17: case 19: case 22: case 24:
	1,0,0,0,0,1,0,0,	// case 25: case 30:
	0,0,1,1,1,1,2,1,	// case 35: case 36: case 37: case 38: case 39: case 40:
	0,0,0,1,0,6,6,7,	// case 44: case 46: case 47: case 48:
	0,0,0,4,1,1,0,1,	// case 52: case 53: case 54: case 56:
	1,1,1,0,0,0,0,0,	// case 57: case 58: case 59:
	0,0,0,0,0,0,0,1,	// case 72:
	1,1,1,1,1,0,1,1,	// case 73: case 74: case 75: case 76: case 77: case 79: case 80:
	1,1,1,1,7,1,1,1,	// case 81: case 82: case 83: case 84: case 85: case 86: case 87: case 88:
	1,1,1,1,1,1,1,1,	// case 89: case 90: case 91: case 92: case 93: case 94: case 95: case 96:
	2,1,0,1,0,0,0,1,	// case 97: case 98: case 100: case 104:
	1,1,1,1,1,1,0,0,	// case 105: case 106: case 107: case 108: case 109: case 110:
	0,0,0,0,0,0,1,1,	// case 119: case 120:
	1,0,0,4,2,2,0,1,	// case 121: case 124: case 125: case 126: case 128:
	1,1,0,0,0,0,0,0,	// case 129: case 130:
	0,0,0,0,1,1,1,1,	// case 141: case 142: case 143: case 144:
	1,1,1,1,1,1,1,1,	// case 145: case 146: case 147: case 148: case 149: case 150: case 151: case 152:
	1,1,1,1,1,0,0,0,	// case 153: case 154: case 155: case 156: case 157:
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,5,0,0,	// case 174
	0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,
	0,0,5,0,6,6,1,1,	// case 195: case 199: case 197: case 198: case 200:
	1,1,0,0,0,0,2,2,	// case 201: case 202: case 207: case 208:
	5,5,0,1,0,7,7,7,	// case 209: case 210: case 212: case 214: case 215: case 216:
	7,7,1,1,0,0,9,0,	// case 217: case 218: case 219: case 220:
	0,0,1,1,0,0,1,1,	// case 227: case 228: case 231: case 232:
	0,0,1,1,0,0,1,1,	// case 235: case 236: case 239: case 240:
	0,0,3,3,7,7,7,7,	// case 243: case 244: case 245: case 246: case 247: case 248:
	7,7,7,7,7,7,7,1,	// case 249: case 250: case 251: case 252: case 253: case 254: case 255: case 256:
	1,0,0,0,0,3,3,3,	// case 257: case 262: case 263: case 264:
	3,3,3,2,2,0,0,0,	// case 265: case 266: case 267: case 268: case 269:
	0,1,1				// case 274: case 275:
};

short identifySpecial(line_t *line)
{
	short special = line->special;

	if (special <=  0) return 9;
	if (special < 276) return walkCheck[special];
	
	// R3D specific types
	if (R3D_maptype && special >= 8064 && special <= 8145)
	{
		if (special == 8093) return 0;
		if (special == 8106) return 0;	// specifically a use line sound type
		if (special == 8107) return 9;	// special dummy type

		if (special == 8113) return 1;	// v220-13 release holdstate (walkover)
		if (special == 8114) return 6;	// v220-13 release holdstate (gun)

		// The next three are ambient sound triggers
		if (special == 8091) return 6;
		if (special == 8090) return gdt.tmthing->player ? 0 : 9;
		if (special == 8092) return gdt.tmthing->player ? 1 : 9;

		// v220-31 walkover once sound for standard map type (set in SCP_ZOFF)
		if (special == 8140) return gdt.tmthing->player ? 1 : 9;	// 0x1FCC
		// v220-31 walkover many sound for standard map type (set in SCP_ZOFF)
		if (special == 8141) return gdt.tmthing->player ? 1 : 9;	// 0x1FCD
		if (special == 8142) return gdt.tmthing->player ? 0 : 9;	// 0x1FCE
		if (special == 8143) return gdt.tmthing->player ? 0 : 9;	// 0x1FCF
		if (special == 8144) return gdt.tmthing->player ? 9 : 1;	// 0x1FD0
		if (special == 8145) return 0;	// USE type to trigger ext_LockedMessage

		if (R3D_ScriptedMap && !line->tag)
		{
			if (special >= 8064 && special <= 8066) return 0;
			if (special >= 8080 && special <= 8082) return 0;
			if (special >= 8096 && special <= 8098) return 0;

			if (special >= 8070 && special <= 8073) return 1;
			if (special >= 8120 && special <= 8139) return 1;

			if (special >= 8067 && special <= 8069) return 6;
			if (special >= 8083 && special <= 8085) return 6;
		}
	}

	// check for generalized linedef types
	if (special >= GenCrusherBase && (ushort)special < (ushort)GenLimit)
	{
		int val  = (special & TriggerType) >> TriggerTypeShift;

		if (val == WalkOnce || val == WalkMany) return 1;
		if (val == GunOnce  || val == GunMany ) return 6;
		
		return 0;	// assume use type
	}

	return 9;	// special not recognised
}

static void AddSpecialToList(line_t *ld)
{
	int		newside;
	int		oldside;
	short	ws;

	// Note: gun types [6] are dealt with elsewhere
	if (ld->special <= 0 || (ws = identifySpecial(ld)) >= 6) return;

	oldside = P_CheckOnLineSide(gdt.oldx, gdt.oldy, ld);

	// v220-6 specials put onto two lists
	// 1)	use specials
	if (!ws)
	{
		// Use lines are only activated if on side 0
		if (!oldside)
		{
			if (numspechit >= spechit_max)
			{
				spechit_max += 16;
				spechit = (line_t**)W_Realloc(spechit, sizeof(*spechit) * spechit_max, STATICID, 226);
			}

			spechit[numspechit++] = ld;	// Used in P_Move
		}

		return;
	}

	if (ws == 5)
	{
		// $swtp: if contacted a special line, add it to the list
		// These lines are only activated if on side 0 and not
		// inhibited by a dropoff unless it's the player
		if (   !oldside
			&& (gdt.player
			|| (gdt.tmthing->z - gdt.tmdropoffz <= MAXSTEP)) )
		{
			if (numspechit >= spechit_max)
			{
				spechit_max += 16;
				spechit = (line_t**)W_Realloc(spechit, sizeof(*spechit) * spechit_max, STATICID, 226);
			}

			spechit[numspechit++] = ld;	// Used in P_Move
		}

		return;
	}

	// 2)	walk specials

	// v220-29 map design error fix in R3D_WADanalyse
	if (ws == 1 && ge_inf.tdcount)
	{
		for (int i = 0; i < ge_inf.tdcount; i++)
		{
			if (ge_inf.tdnum[i] == ld - lines)
			{
				// Start Thinker directly, special not added to list
				P_CrossSpecialLine(ld, gdt.tmthing);
				return;
			}
		}
	}

	newside = P_CheckOnLineSide(gdt.tmx , gdt.tmy , ld);

	// v220-29 [rev 2] reject a line where a Thing has moved from being on the line to the front side
	if (oldside + newside == 2) return;

	// Check line will be crossed (if other checks allow this - search check_spechits)
	// v220-29 revamped
	if (oldside != newside)
	{
		if (  (ws == 2 || ws == 3)			// line to TPthing
			&& oldside != 0
			&& !(oldside == 2 && newside == 1))	// v220-29 add oldside, newside test
		{
			return;	
		}

		if (numsplist >= splist_max)
		{
			splist_max += 16;
			splist = (splist_t*)W_Realloc(splist, sizeof(splist_t) * splist_max, STATICID, 226);
		}

		// Account for Thing straddling more than one tp line
		if (ws == 2 || ws == 3)	// include line to line tp's
		{
			splist[numsplist].tpspec = true;

			if (   (gdt.tmthing->x >= ld->bbox[BOXLEFT] && gdt.tmthing->x <= ld->bbox[BOXRIGHT])
				|| (gdt.tmthing->y >= ld->bbox[BOXBASE] && gdt.tmthing->y <= ld->bbox[BOXTOP  ]))
			{
				// null any others
				for (int i = 0; i < numsplist; i++)
				{
					if (splist[i].tpspec) splist[i].ld = NULL;
				}
			}
		}
		else
		{
			splist[numsplist].tpspec = false;
			for (int i = 0; i < numsplist; i++)
			{
				if (splist[i].ld == ld) return;
			}
		}

		splist[numsplist++].ld = ld;
	}
}

#pragma optimize("", on)

// ============= v220-10 end spechits revision


//
// TELEPORT MOVE PIT_StompThing
// 

// v220-4 add vertical overlap check
static RDBOOL	check3doverlap;
static fixed_t	overlaph;
static fixed_t	overlapz;
static fixed_t	exitLevel;

static RDBOOL PIT_StompThing (mobj_t* thing, int *telefrag)
{
    fixed_t	blockdist;

    if ( !(thing->flags & MF_SHOOTABLE)		// Can't shoot it? Can't stomp it!
		|| thing == gdt.tmthing				// don't clip against self
		|| (thing->dplayer
		&& thing->dplayer->mo == thing
		&& thing->player->cheats & CF_NOCLIP))	// b220-1
	{
		return true;
	}
    
    blockdist = thing->radius + gdt.tmthing->radius;
    
    if (   abs(thing->x - gdt.tmx) > blockdist
		|| abs(thing->y - gdt.tmy) > blockdist )
    {
		return true;	// didn't hit it
    }

	// v220-35 check eprite's valid (could be a dehacked thing
	exitLevel = thing->info->doomednum == 88 && thing->sprite == SPR_BBRN;

	if (check3doverlap)
	{
		if ( !(thing->z + thing->height >= overlapz
			&& overlaph >= thing->z))
		{
			return true;	// didn't hit it
		}
	}
    
	// monsters don't stomp things except levels having boss
	// killough 8/9/98: make consistent across all levels
	if ( !(*telefrag) ) return false;

	// v220-4 attempt to emulate ZDoom behaviour
	if (ma_inf.modifier & MODF_KILL)
	{
		RDBOOL exception = gdt.tmthing->dplayer && gdt.tmthing->dplayer->mo == gdt.tmthing;

		if (!exception)
		{
			switch (thing->type)
			{
			case MT_BABY:
			case MT_FATSO:
			case MT_BRUISER:
			case MT_CYBORG:
			case MT_SPIDER:
				exception = thing->type == gdt.tmthing->type;
				break;
			default:
				break;
			}
		}

		if (exception)	// reverse kill
		{
			P_DamageMobj(gdt.tmthing, thing, thing, 10000, 20);
			return false;	// prevent tp
		}
	}

    P_DamageMobj(thing, gdt.tmthing, gdt.tmthing, 10000, 21);
	
	return true;
}

static int PIT_CheckTPLine(line_t *li, mobj_t *mo)
{
	if (  li->lnflags2 & LNF2_3DTELEPLINE
	   && PointIsOnLine(mo->x, mo->y, li, 1))
	{
		mo->z = mo->floorz = mo->subsector->sector->floorheight + li->d3d->offset3D;
		mo->flags2 |= MF2_3DTELEPLINE;
		return false;
	}
	return true;
}

//===========================================================================
//	Also stomps on things.
//===========================================================================
RDBOOL P_TeleportMove(mobj_t* thing, fixed_t x, fixed_t y, fixed_t z, RDBOOL is3d, RDBOOL boss)
{
	// v220-4 check if an actor could get spawned in a 1's line.
	// v220-5 MF_SPECIAL is used to eliminate the player
	//        Pickups could well get stuck as the difference between their
	//        info->radius and true radius is exaggerated to allow them to
	//        be picked up even if the player's radius does not overlap the
	//        thing's true radius.
	if (  (thing->flags & (MF_SOLID|MF_CORPSE|MF_SPECIAL)) == MF_SOLID
		&& !checkLineSpawnedTP(x, y, thing->radius))
	{
		return false;
	}

//	if (ge_inf.realtekCompat)
	if (thing->dplayer && thing->dplayer->mo == thing)
		Sfx_StopSound(0, NULL); // v220-33

	gdt.tmthing = thing;
	gdt.tmflags = thing->flags;
	
	gdt.tmx = x;
	gdt.tmy = y;
	
	gdt.newsubsec = R_PointInSubsector (x,y);
	
	gdt.player = thing->dplayer && thing->dplayer->mo == thing;
	// $unstuck: floorline used with player
	gdt.blockline = gdt.floorline = gdt.ceilingline = NULL;

	// The base floor/ceiling is from the subsector that contains the point.
	if ( (check3doverlap = is3d) )
	{
		overlapz = z; overlaph = z + thing->height;
	}

	gdt.tmfloorz = gdt.tmdropoffz = SL_PointHeight(gdt.newsubsec->sector, gdt.tmx, gdt.tmy, LOSLOPE);
	gdt.dropoffz3D = gdt.same3D = DDMAXINT;
	gdt.tmceilingz = SL_PointHeight(gdt.newsubsec->sector, gdt.tmx, gdt.tmy, HISLOPE);
	
	numspechit = 0;
	numsplist = 0;
	
	// v220-4
	if (   ma_inf.modifier & MODF_KILL
		&& gdt.tmthing->dplayer
		&& gdt.tmthing->dplayer->mo == gdt.tmthing
		&& gdt.tmthing->health <= 0)
	{
		return false;
	}

	// Only stomp using a solid thing
	if (   !R3D_edit
		&& thing->flags & MF_SOLID
		&& !(gdt.player && thing->player->cheats & CF_NOCLIP))
	{
		// killough 8/9/98: whether to telefrag at exit
		// v220-4 add ma_inf.modifier test
		// v220-13 ma_inf.hasboss is now preset if ma_inf.modifier & MODF_KILL is true.
		int telefrag = thing->player || boss || ma_inf.hasboss;
		
		int xl, xh, yl, yh, bx, by;

		//-------------------------------
		
		// kill anything occupying the position
		M_setBoxOffsets(gdt.tmbbox, x, y, thing->radius);
		M_getBlockOffsets(gdt.tmbbox, MAXRADIUS, &xl, &xh, &yl, &yh);
		
		validcount++;
		
		for(bx = xl; bx <= xh; bx++)
		{
			for(by = yl; by <= yh; by++)
			{
				if (!P_BlockThingsIterator(bx, by, (int (__cdecl *)(struct mobj_s *,void *))PIT_StompThing, &telefrag))
				{
					thing->momx = thing->momy = 0;	// v220-15 kill momentum
					return false;
				}
			}
		}
	}

	if (exitLevel)	// Boss was stomped
	{
		ma_inf.exitlevel = 0x0080003C;	// R3D Exit but not immediately (delay is low byte in tics)
		exitLevel = false;
	}
	
	// the move is ok, so link the thing into its new position
	P_UnlinkThing (thing);

	thing->x = x;
	thing->y = y;
	thing->floorz = gdt.tmfloorz;
	thing->ceilingz = gdt.tmceilingz;	
	thing->dropoffz = gdt.tmdropoffz;		 // killough $unstuck
	
	P_SetThingPosition(thing);
	P_ClearThingSRVO(thing, true);

	gdt.pt_oldx = x;
	gdt.pt_oldy = y;
	gdt.pt_oldz = thing->floorz;
	
	if (thing->flags & MF_NOGRAVITY) thing->momz = 0;

	if (ppt.viewplrmo == thing) resetNextViewer = true;

	// R3D - Check if x,y is on an $Axxxxx line
	if (R3D_maptype)
		P_ThingLinesIterator(thing, (int (__cdecl *)(struct line_s *,void *))PIT_CheckTPLine, thing);
	
	return true;
}


//
// MOVEMENT ITERATOR FUNCTIONS
//

// Returns the friction associated with a particular mobj - killough 8/28/98
// b211-0 Added PIT_Friction()
static int PIT_Friction(sector_t *sec, fixed_t *data)
{
	if ( sec->special & FRICTION_MASK &&
		(sec->friction < data[0] || data[0] == ORIG_FRICTION)
		&& (data[2] <= sec->floorheight
		|| (sec->heightsec >= 0
		&& data[2] <= sectors[sec->heightsec].floorheight)))
	{
		data[0] = sec->friction, data[1] = sec->movefactor;
	}
	return true;
}

// b211-0 P_GetFriction() rewritten
fixed_t P_GetFriction(mobj_t *mo, fixed_t *frictionfactor)
{
	fixed_t data[3];

	if (mo->misc & (MJF_LINE3D|MJF_FLAT3D))
	{
		if (frictionfactor) *frictionfactor = ORIG_FRICTION_FACTOR;
		return ORIG_FRICTION;
	}

	data[0] = ORIG_FRICTION;
	data[1] = ORIG_FRICTION_FACTOR;
	data[2] = mo->z;

	// Get the friction value where on the floor, non-floating, alive
	// and sector linked. Normally the actor's friction value is kept
	// at ORIG_FRICTION unless on an icy or muddy floor.
	// v220-9 don't set if waterdynamics is in force
	if ( !(mo->flags & (MF_NOCLIP|MF_NOGRAVITY)) && !(waterdynamics && waterlevel) )
	{
		P_ThingSectorsIterator(mo, (int (__cdecl *)(struct sector_s *,void *))PIT_Friction, data);
	}

	if (frictionfactor) *frictionfactor = data[1];

	// v220-10 modify friction slightly for a dead player
	if (   mo->health <= 0
		&& mo->dplayer
		&& mo->dplayer->mo == mo
		&& data[0] == ORIG_FRICTION
		&& data[1] == ORIG_FRICTION_FACTOR)
	{
		// Note: increasing the value actually reduces "stickiness"
		data[0] += 0x0800;
	}

	return data[0];
}

// Returns the addition factor for player x,y movements - phares 3/19/98
// rewritten
int P_GetMoveFactor(mobj_t *mo, int *frictionp)
{
	int movefactor, friction;
	
	// If the floor is icy or muddy, it's harder to get moving. This is where
	// the different friction factors are applied to 'trying to move'. In
	// p_mobj.c, the friction factors are applied as you coast and slow down.
	
	if ((friction = P_GetFriction(mo, &movefactor)) < ORIG_FRICTION)
	{
		// phares 3/11/98: you start off slowly, then increase as
		// you get better footing
		
		int momentum = P_ApproxDistance(mo->momx, mo->momy);
		
		if (momentum > MORE_FRICTION_MOMENTUM<<2)
			movefactor <<= 3;
		else if (momentum > MORE_FRICTION_MOMENTUM<<1)
			movefactor <<= 2;
		else if (momentum > MORE_FRICTION_MOMENTUM)
			movefactor <<= 1;
	}
	
	if (frictionp)
		*frictionp = friction;
	
	return movefactor;
}

// $unstuck: used to test intersection between thing and line
// assuming NO movement occurs -- used to avoid sticky situations.
static int untouched(line_t *ld, fixed_t x, fixed_t y)
{
	fixed_t utbbox[4];
	fixed_t radius = gdt.tmthing->radius;

	return (utbbox[BOXRIGHT] = M_AddLIMIT(x,  radius)) <= ld->bbox[BOXLEFT ]
		|| (utbbox[BOXLEFT ] = M_AddLIMIT(x, -radius)) >= ld->bbox[BOXRIGHT]
		|| (utbbox[BOXTOP  ] = M_AddLIMIT(y,  radius)) <= ld->bbox[BOXBASE ]
		|| (utbbox[BOXBASE ] = M_AddLIMIT(y, -radius)) >= ld->bbox[BOXTOP  ]
		|| P_BoxOnLineSide(utbbox, ld) != -1;
}

static RDBOOL setSkullState(mobj_t *th)
{
	int	se, sn = th->state - states;

	th->flags &= ~MF_SKULLFLY;
	th->momx = th->momy = th->momz = 0;

	if (   sn != th->info->spawnstate
		&& sn != th->info->seestate
		&& (se = th->info->seestate) > 0)
	{
		P_SetMobjState (th, se);
	}

	return false;
}

// v220-4 set line list
void P_AddXLines(line_t *ld)
{
	if (mll.count >= mll.maxcount)
	{
		mll.maxcount += 32;
		mll.plist = (line_t**)W_Realloc(mll.plist, mll.maxcount * sizeof(line_t*), STATICID, 0);
	}

	if (gdt.player)
	{
		extern byte*lnamap;
		extern int linedetect;

		if (!lnamap[ld - lines])
		{
			lnamap[ld - lines] = true;
			linedetect++;
		}
	}

	mll.plist[mll.count++] = ld;
}

static RDBOOL PIT_SetLineList(line_t* ld, void *data)
{
	if (   gdt.tmbbox[BOXRIGHT] <= ld->bbox[BOXLEFT ]
		|| gdt.tmbbox[BOXLEFT ] >= ld->bbox[BOXRIGHT]
		|| gdt.tmbbox[BOXTOP  ] <= ld->bbox[BOXBASE ]
		|| gdt.tmbbox[BOXBASE ] >= ld->bbox[BOXTOP  ]
		|| P_BoxOnLineSide(gdt.tmbbox, ld) != -1)
	{
		return true;	// ignore
	}

	// v220-33 ignore lines flagged unblocked if a missile is involved
	if (   gdt.tmthing->flags & MF_MISSILE
		&& ld->lnflags2 & LNF2_UNBLOCKED
		&& ld->frontsector->floorheight == ld->backsector->floorheight)
	{
		return true;	// v220-16 emulate Doom bug
	}
		
	if (ld->backsector)
	{
		if (ld->ml_flags & ML_BLOCKING) gdt.mlblocking = true;

		// v220-6 deal with crossing an exit game line where the exit is blocked.
		//        If set then gdt.mlblocking must also be set. This combination, for
		//        example, differentiates between the exit line in Herian2 map03 where
		//        it must be crossed and Strain10 map07 where it cannot be crossed.

		if (gdt.player)
		{
			if (  (ld->special == 52 || ld->special == 124)
				&& P_PointOnLineSide(gdt.oldx, gdt.oldy, ld) == 0
				&& P_PointOnLineSide(gdt.tmx, gdt.tmy, ld) == 1
				// v220-11 The player's centre must be on the exit line (approximately)
				&& XYonLine(gdt.oldx, gdt.oldy, ld, FRACUNIT/2))
			{
				gdt.exitgame = true;
			}
		}
		else
		{
			// v220-10  grenades/pipebombs pass straight through
			// textureless 3D support lines on mov floor sectors
			if (   ld->specType == R3D_40
				&& (ge_inf.grenade_id && gdt.tmthing->type == ge_inf.grenade_id)
				&& ld->frontsector->specFlags & SF_MOVFLOOR
				&& !origTEX[ld - lines])
			{
				return true; // ignore
			}
		}
	}
	else
	{
		// A Thing should never be facing a NULL backsector in its current position
		if (P_PointOnLineSide(gdt.oldx, gdt.oldy, ld))
		{
			return true; // ignore
		}
	}

	P_AddXLines(ld);

	return true;
}

// Adjusts tmfloorz and tmceilingz as lines are contacted
static RDBOOL PIT_CheckLine (line_t* ld)
{
	if (ld->lnflags2 & LNF2_3DTELEPLINE)
	{
		return true; // ignore
	}

	// A line has been hit
	gdt.tmthing->wallhit = true;

	// The moving thing's destination position will cross the given line.
	// If this should not be allowed, return false.
	// If the line is special, keep track of it
	// to process later if the move is proven ok.
	// NOTE: specials are NOT sorted by order,
	// so two special lines that are only 8 pixels apart
	// could be crossed in either order.
	
	if (  !ld->backsector
		|| ld->lnflags2 & (LNF2_BLOCKALL|LNF2_OCCLUDE)) // scripted blocking line types
	{
		// v220-3 If a pickup is stuck in a 1's wall on a moving plat
		//        then make its floorheight the same as the plat.
		if (   gdt.tmflags & (MF_DROPPED|MF_SPECIAL)
			&& gdt.tmthing->subsector->sector->specFlags & SF_MOVFLOOR)
		{
			gdt.tmthing->z = gdt.tmthing->floorz = gdt.tmthing->dropoffz = gdt.tmthing->subsector->sector->floorheight;
			return false;
		}

		// Store blocking or occluding lines with a special
		if (ld->special && ld->backsector) AddSpecialToList(ld);

		// Check for being blocked in a fake water sector
		// The line needs to be set for swimming up
		if (gdt.player && waterlevel && waterlevel < 4)
		{
			gdt.plrlineblocked = ld;
		}

		// $unstuck: killough 7/24/98: allow player to move out
		// of 1s wall, to prevent sticking
		// v220-3 applies to all sentient things (previously was just the player)
		if (sentient(gdt.tmthing) && !untouched(ld, gdt.oldx, gdt.oldy))
		{
			return FixedMul(gdt.tmx - gdt.oldx, ld->dy) > FixedMul(gdt.tmy - gdt.oldy, ld->dx);
		}

		return false;
	}

	int ordinance = (gdt.tmthing->flags & MF_MISSILE || gdt.tmthing->flags3 & MF3_BOUNCES);
	
	// killough 8/10/98: allow bouncing objects to pass through as missiles
	if (!ordinance)
	{
		if (ld->ml_flags & ML_BLOCKING)	// explicitly blocking (nearly) everything
		{
			if (gdt.tmflags & MF_SKULLFLY) return setSkullState(gdt.tmthing);

			// killough 8/1/98: allow escape
			return gdt.tmthing->dplayer && !untouched(ld, gdt.oldx, gdt.oldy);
		}
		
		// v220-6 exclude pickups
		if ( !(gdt.tmflags & (MF_SPECIAL|MF_DROPPED)) )
		{
			// killough 8/9/98: monster-blockers don't affect friends
			if (   ld->ml_flags & ML_BLOCKMONSTERS
				&& !gdt.tmthing->player
				&& !(gdt.tmthing->flags3 & MF3_FRIEND))
			{
				if (gdt.tmflags & MF_SKULLFLY) return setSkullState(gdt.tmthing);

				return false;			// block monsters only
			}
			
			// v220-4 scripted map block friend only
			if (ld->scpFlags & SCPF_BLOCKFRIEND && !gdt.player && gdt.tmthing->flags3 & MF3_FRIEND)
			{
				if (gdt.tmflags & MF_SKULLFLY) return setSkullState(gdt.tmthing);

				return false;			// block friends only
			}
		}
	}

	gdt.tmthing->wallhit = false;

	if (ld->specType >= R3D_40)	// 3D lines
	{
		int		lift = ld->frontsector->liftsec3D;
		fixed_t z_offset = lift >= 0 ? sectors[lift].planes[0].speed : 0;
		fixed_t	cf = ld->backsector->ceilingheight;
		fixed_t	lf = ld->backsector->floorheight;
		fixed_t	height = lift >= 0 ? sectors[lift].floorheight : lf;
		fixed_t	bottomStep, topStep;

		fixed_t	zp = gdt.tmthing->z + z_offset;
		fixed_t	zh = zp + gdt.tmthing->height;

		if (   ld->ml_flags & ML_SECRET
			&& ld->spflags & SPF_NOTEX
			&& (gdt.tmflags & MF_MISSILE || (ge_inf.grenade_id && gdt.tmthing->type == ge_inf.grenade_id)) )
		{
			bottomStep = topStep = height;
		}
		else
		{
			bottomStep = height  + ld->d3d->offset3D;
			topStep = bottomStep + ld->d3d->totalheight;
		}

		// v220-33 plasma types can go through masked 3D textures
		if (   gdt.tmflags & MF_MISSILE
			&& ld->specType == R3D_40
			&& ld->lnflags2 & (LNF2_MIDTEXMASKED|LNF2_LOWTEXMASKED)
			&& (gdt.tmthing->sprite == 18	// imp
			||  gdt.tmthing->sprite == 19	// cacodemon
			||  gdt.tmthing->sprite == 20	// plasma
			||  gdt.tmthing->sprite == 33	// revenant
			||  gdt.tmthing->sprite == 36	// fatso
			||  gdt.tmthing->sprite == 41) )// baron/hellknight
		{
			if (ld->lnflags2 & LNF2_LOWTEXMASKED)
				topStep = bottomStep = height;
			if (ld->lnflags2 & LNF2_MIDTEXMASKED)
				topStep = bottomStep;
		}

		// v220-31 revised
		if (ld->ovflags & OVF_3DLADDERLINE)
		{
			bottomStep += ld->d3d->lineheightlow;	// offset from base (if set)

			if (gdt.player && waterlevel != 4)	// i.e. not flying
			{
				gdt.laddertop = topStep;
				gdt.ladderbase = bottomStep;

				if (gdt.onladder) return false;	// Don't go into line

				// v220-3 Only get on ladder if angle between
				//        player line is within +- 30 degrees
				angle_t ang = R_PointToAngle2(0, 0, ld->dx, ld->dy) - gdt.tmthing->angle + ANG90;

				if (  (ang < ANG30 || ang > (unsigned) -ANG30)
					&& gdt.tmz < gdt.laddertop
					&& gdt.moheight > gdt.ladderbase
					&& !P_PointOnLineSide(gdt.tmx, gdt.tmy, ld))	// must be facing front of line
				{
					gdt.onladder = true;
					gdt.tmthing->floorz = gdt.tmz;
					return false;
				}
			}

			lf = gdt.ssec->sector->floorheight;
		}

		// v220-31 ammended
		if (  !(ld->ovflags & OVF_3DLADDERLINE) )
		{
			if (ordinance)
			{
				if (  (origTEX[ld - lines] & 0x33
					|| (!origTEX[ld - lines] && ld->specType == R3D_40))
					&& gdt.moheight >= bottomStep
					&& gdt.tmz < topStep)
				{
					if (gdt.tmthing->flags & MF_MISSILE) gdt.explodemissile = true;
					gdt.blockmove = true;
					return false;
				}
			}
			// If not player simulate fence for blocking lines
			else if (!gdt.player
				&& sentient(gdt.tmthing)		// Only sentient objects
				&& !(gdt.tmflags & MF_FLOAT)	// except those that float
				&& (ld->specType == R3D_41 || ld->specType == R3D_43))
			{
				topStep += FRACUNIT*256;	// invisible fence
			}
		}

		if (gdt.moheight <= bottomStep)
		{
			openbottom = lf;
			opentop = bottomStep;
		}
		else
		{
			openbottom = topStep;
			opentop = cf;

			if (   ld->special
				&& gdt.player
				&& abs(gdt.tmz - openbottom) <= MAXSTEP)
			{
				AddSpecialToList(ld);
			}
		}

		if (opentop <= openbottom)
		{
			opentop = openbottom;
		}

		// adjust floor / ceiling heights
		if (opentop < gdt.tmceilingz)
		{
			gdt.tmceilingz  = opentop;
			gdt.ceilingline = ld;
		}
		
		if (openbottom > gdt.tmfloorz)
		{
			gdt.tmfloorz  = openbottom; 
			gdt.floorline = ld;    // killough $unstuck: remember floor linedef
		}

		gdt.tmthing->misc |= MJF_LINE3D;

		if (lf < gdt.tmdropoffz) gdt.tmdropoffz = lf;

		if (   zp <= bottomStep
			&& zh > bottomStep)
		{
			if (z_offset)
			{
				gdt.tmceilingz = bottomStep;
				gdt.blocked3D = true;
			}
			else if (topStep - zp > 24*FRACUNIT)
			{
				gdt.blockmove = true;
				return false;
			}
		}

		if (ld->frontsector == ld->backsector && ld->specType == R3D_40 && origTEX[ld - lines] == 0x11)
		{
			gdt.same3D = topStep;
		}

		if (zp >= topStep && (zp > gdt.dropoffz3D || gdt.dropoffz3D == DDMAXINT))
			gdt.dropoffz3D = topStep;

		openrange = opentop - openbottom;
	}
	else	// 2D lines
	{
		// set openrange, opentop, openbottom
		P_LineOpening (ld, DDMININT, gdt.tmx, gdt.tmy);

		// adjust floor / ceiling heights
		if (opentop < gdt.tmceilingz)
		{
			gdt.tmceilingz = opentop;
			gdt.ceilingline = ld;
		}
		
		if (openbottom > gdt.tmfloorz)
		{
			gdt.tmfloorz = openbottom; 
			gdt.floorline = ld;    // killough $unstuck: remember floor linedef
		}
		
		if (lowfloor < gdt.tmdropoffz)
		{
			gdt.tmdropoffz = lowfloor;
		}

		// check for Things that don't climb stairs
		if (gdt.moheight > opentop || gdt.tmz < openbottom)
		{
			// v220-31 revised
			if (gdt.tmflags & MF_MISSILE)
			{
				gdt.explodemissile = true;
				return false;
			}
			else
			// v220-13 support dehacked spawnshot
			if (   gdt.tmthing->type == MT_SPAWNSHOT
				&& gdt.tmthing->flags3 & MF3_BOUNCES
				&& !(gdt.tmthing->flags & MF_NOCLIP))
			{
				gdt.blockmove = true;
				return false;
			}
		}

		if (!openrange) gdt.blockmove = true;

		AddSpecialToList(ld);
	}

	return true;
}

//
// PIT_CheckThing
//

static RDBOOL PIT_CheckThing(mobj_t* thing, int *pcptest)
{
	extern int	specialSprite;

	fixed_t		blockdist;
	RDBOOL		solid, touch_vertically = false;	// b211-0
	RDBOOL		overlap = false;
	int			damage;

	if (gdt.tmthing == thing) return true;

	if (specialSprite)	// v220-16 DEHACKED fixes
	{
		if (specialSprite & 0x00000001)	// flags are set in R3D_setparms
		{
			if (   gdt.tmflags & MF_PICKUP
				&& thing->type == MT_MISC25
				&& thing->sprite == SPR_BFUG)
			{
				// Note: cannot be preset as does not initially exist until
				// paired sprite is removed (DEHACKED - don't you just love it?)
				thing->flags |= MF_SPECIAL;
				thing->pickupRadius = 40*FRACUNIT;
				specialSprite &= ~0x00000001;
			}
		}
	}

	if (thing->ddflags & DDF_AMBIENT) return true;

	// VDOORs don't require processing as they only
	// rise and no Thing can be above them.
	if (thing->flags2 & MF2_VDOOR) return true;

	// b220 R3D proprietory - amodel door support 
	if (thing->modelboxnum > 0)
	{
		int		num = thing->modelboxnum - 1;
		fixed_t *mbox = MB_GetBox(num, NULL);

		if (mbox)
		{
			// Check tmthing is within its bounds
			if (   gdt.tmbbox[BOXRIGHT] <= mbox[BOXLEFT ]
				|| gdt.tmbbox[BOXLEFT ] >= mbox[BOXRIGHT]
				|| gdt.tmbbox[BOXTOP  ] <= mbox[BOXBASE ]
				|| gdt.tmbbox[BOXBASE ] >= mbox[BOXTOP  ])
			{
				return true;	// didn't hit it
			}

			for (int n = 0; n < mobj_mblistcount; n++)
				if (mobj_mblist[n] == thing) return true;

			mobj_mblist[mobj_mblistcount++] = thing;
		}

		return true;
	}
	// v220-14
	else if (thing->health <= 0 && (!(thing->flags3 & MF3_AMODEL) || thing->ddflags & DDF_IGNORE))
	{
		return true;
	}

	if (gdt.tmflags & MF_PICKUP && thing->flags & MF_SPECIAL)
	{
		blockdist = gdt.tmthing->radius + thing->pickupRadius;
	}
	else
	{
		blockdist = gdt.tmthing->radius + thing->radius;
		if (blockdist > 16*FRACUNIT) blockdist -= 4*FRACUNIT;
	}

	if (   abs(thing->x - gdt.tmx) >= blockdist
		|| abs(thing->y - gdt.tmy) >= blockdist )
	{
		return true;	// didn't hit it
	}

	// R3D v220-35 More DEHACKED buggering about
	if (hasDEH && thing->dplayer && !gdt.player && gdt.tmthing->flags & MF_SPECIAL)
	{
		int state = gdt.tmthing->info->spawnstate;

		if (state < DH_CHECKNUM && DEHCheck[gdt.tmthing->info->spawnstate] & DEHF_STATE)
		{
			if (    gdt.tmthing->sprite == SPR_BKEY
				||  gdt.tmthing->sprite == SPR_YKEY
				||  gdt.tmthing->sprite == SPR_RKEY
				||  gdt.tmthing->sprite == SPR_BSKU
				||  gdt.tmthing->sprite == SPR_YSKU
				||  gdt.tmthing->sprite == SPR_RSKU)
				
			{
				P_TouchDEHKeyThing(gdt.tmthing);
				return true;	// gdt.tmthing will have been removed
			}
		}
	}

	if (   !(thing->flags & (MF_SOLID|MF_SPECIAL|MF_SHOOTABLE))
		&& !(thing->flags3 & MF3_TOUCHY) )
	{
		if (   gdt.player
			&& thing->flags & MF_NOCLIP
			&& thing->flags3 & MF3_NEARCLIP)
		{
			reduceNearClip = true;
		}

		return true;
	}

	if (gdt.solid)
	{
		overlap =  (thing->flags3 & MF3_AMODEL) != 0
				|| (thing->flags & (MF_SOLID|MF_CORPSE|MF_SPECIAL|MF_NOCLIP|MF_MISSILE)) == MF_SOLID;
	}

	if (overlap && molistnum < MAXMOLISTNUM)
	{
		// v220-4 used to test for list entry for stepup/falldown check
		molist[molistnum++] = thing;
	}

	// v220-4 exit if blocked. We have only returned to populate the molist
	if (gdt.thingblocked) return true;

	// NOTE: don't use gdt.tmz - this could be DDMAXINT
	touch_vertically = thing->z + thing->height >= gdt.z && gdt.z + gdt.tmheight >= thing->z;

	solid = (thing->flags & MF_SOLID) > 0;	// MF_NOCLIP must be ignored

	if (!touch_vertically && gdt.solid && solid)
	{
		if (gdt.z + gdt.tmheight <= thing->z)
		{
			if (gdt.tmceilingz > thing->z)
				gdt.tmceilingz = thing->z;
		}
		else	// if not below then must be above
		{
			if (gdt.tmfloorz < thing->z + thing->height)
				gdt.tmfloorz = thing->z + thing->height;
		}
		return true;
	}

	if ((*pcptest) == -1)	// We don't want to blow anything up as yet
	{
		// do nothing
	}
	else
	{
		// v220-35 smooth player stepup on decorations.(even when
		//			scrolling). For compatibility only Things
		//			defined using the TFOG sprite are affected.
		if (   touch_vertically
			&& ppt.viewplrmo == thing
			&& gdt.tmthing->height <= MAXSTEP
			&& (!R3D_ScriptedMap || (gdt.tmthing->sprite == SPR_TFOG && !gdt.tmthing->info->deathstate))			
			&& (thing->z == gdt.z || thing->onmobj == gdt.tmthing)
			&& gdt.z + gdt.tmheight < gdt.newsubsec->sector->ceilingheight)
		{
			return true;
		}

		// killough 11/98:
		// modified b211-0
		//
		// TOUCHY flag, for mines or other objects which die on contact with solids.
		// If a solid object of a different type comes in contact with a touchy
		// thing, and the touchy thing is not the sole one moving relative to fixed
		// surroundings such as walls, then the touchy thing dies immediately.

		if (   touch_vertically								// touches vertically
			&& thing->flags3 & MF3_TOUCHY					// touchy object
			&& gdt.tmflags & MF_SOLID						// solid object touches it
			&& thing->health > 0							// touchy object is alive
			&& (thing->flags3 & MF3_ARMED					// Thing is an armed mine
			|| sentient(thing)								// ... or a sentient thing
			// b211-0 Make grenades kill immediately on being touched
			|| (thing->thinker.spawnid						// grenade or pipebomb
			&& thing->thinker.spawnid						// where gdt.tmthing
			!= gdt.tmthing->thinker.id))					// doesn't own it
			&& (thing->type != gdt.tmthing->type			// only different species
			|| thing->type == MT_PLAYER)					// ... or different players
			&& (thing->type ^ MT_PAIN)						// PEs and lost souls
			|  (gdt.tmthing->type ^ MT_SKULL)				// are considered same
			&& (thing->type ^ MT_SKULL)						// (but Barons & Knights
			|  (gdt.tmthing->type ^ MT_PAIN))				// are intentionally not)
		{
			P_DamageMobj(thing, NULL, NULL, thing->health, 22);	// kill object
			return true;
		}

		// check for skulls slamming into things
		if (gdt.tmflags & MF_SKULLFLY)
		{
			// b211-0 check it's hitting something
			if (touch_vertically)
			{
				if ( !(thing->type == gdt.tmthing->type
					&& thing->ddflags & DDMF_NOINFIGHT) )	// v220-13
				{
					damage = ((M_RandomD() % 8) + 1) * gdt.tmthing->info->damage;	
					P_DamageMobj (thing, gdt.tmthing, gdt.tmthing, damage, 23);
				}
				else
				{
					return setSkullState(gdt.tmthing);		// stop moving
				}
			}
		}
	}

	// missiles can hit other things
	if (   gdt.tmflags & MF_MISSILE
		|| (gdt.tmthing->flags3 & MF3_BOUNCES && !(gdt.tmflags & MF_SOLID)) )
	{
		// b211-0 See if it went over / under
		if (!touch_vertically) return true;
		
		if (gdt.tmthing->target)
		{
			// Check species to species attack
			if (    gdt.tmthing->target->type == thing->type
				|| (gdt.tmthing->target->type == MT_KNIGHT  && thing->type == MT_BRUISER)
				|| (gdt.tmthing->target->type == MT_BRUISER && thing->type == MT_KNIGHT ) )
			{
				if (thing == gdt.tmthing->target)	// Already set
					return true;

				if (thing->type != MT_PLAYER)
					return false;	// Explode, but do no damage.
			}
		}

		// R3D - Grenades bouncing off top of architectural models
		if (   solid
			&& gdt.tmthing->flags3 & MF3_BOUNCES
			&& thing->flags3 & MF3_AMODEL
			&& gdt.tmthing->momz < 0)
		{
			// Reverse momz and reduce by 20%
			gdt.tmthing->momz = -FixedMul(0xCCCC, gdt.tmthing->momz);
			return true;
		}
		
		// killough 8/10/98: if moving thing is not a missile, no damage
		// is inflicted, and momentum is reduced if object hit is solid.
		if ( !(gdt.tmflags & MF_MISSILE) )
		{
			if (!(thing->flags & MF_SOLID))
			{
				return true;
			}
			else
			{
				gdt.tmthing->momx = -gdt.tmthing->momx;
				gdt.tmthing->momy = -gdt.tmthing->momy;

				if (!(gdt.tmflags & MF_NOGRAVITY))
				{
					gdt.tmthing->momx >>= 2;
					gdt.tmthing->momy >>= 2;
				}

				return false;
			}
		}

		if ( !(thing->flags & MF_SHOOTABLE) )
		{
			return !solid;		// didn't do any damage
		}

		// damage / explode
		damage = ((M_RandomD() % 8) + 1)*gdt.tmthing->info->damage;
		P_DamageMobj (thing, gdt.tmthing, gdt.tmthing->target, damage, 24);
		
		// don't traverse any more
		return false;				
	}

	// check for special pickup
	if (thing->flags & MF_SPECIAL)		// removable thing
	{
		if (   gdt.tmflags & MF_PICKUP	// thing can pick up items
			&& touch_vertically			// and is touching thing
			&& gdt.tmthing->health > 0)	// and is not dead
		{
			P_TouchSpecialThing(thing);
		}
		return !solid;
	}

	if (overlap && solid)
	{
		int tht = thing->z + thing->height;

		// v220-4 Only players can step onto models the full maxstep
		//        height. Other actors are limted to half that height.
		int testheight = gdt.player ? MAXSTEP : MAXSTEP>>1;

		// See if tmfloorz can be adjusted
		if (   thing->flags3 & MF3_AMODEL
			&& thing->z < gdt.moheight
			&& tht > gdt.tmfloorz
			&& !(thing->flags & MF_SPAWNCEILING))
		{
			if (gdt.tmfloorz < tht) gdt.tmfloorz = tht;
			if (gdt.tmthing->z + testheight > tht) return true;
		}

		if (   gdt.player
			&& thing->type == cfg.jumpkill
			&& gdt.tmthing->z <= tht + 4*FRACUNIT
			&& sentient(thing)
			&& (gdt.tmthing->player->jumptics
			|| (!gdt.tmthing->player->jumptics && gdt.tmthing->old_z > gdt.tmthing->z))
			&& thing->z == thing->floorz)
		{
			P_DamageMobj(thing, NULL, NULL, thing->health, 25);
		}
		else
		if (   gdt.z >= tht								// mo standing/falling
			|| (gdt.moheight > tht && gdt.z >= tht - testheight))	// or stepping up onto things
		{
			if (gdt.z <= tht)
			{
				{
					gdt.tmthing->onmobj = thing;
					if (gdt.tmfloorz < tht) gdt.tmfloorz = tht;
				}
			}
			return true;	// mo can move
		}

		if (thing->z + thing->momz >= gdt.moheight)
			return true;	// mo can move
	}

	if (solid && gdt.tmflags & MF_SOLID)
	{
		// Amodels block even if MF_NOCLIP is set
		if (thing->flags3 & MF3_AMODEL)
		{
			if (touch_vertically)
			{
				if (!gdt.player)	// Actors move away from an amodel
				{
					gdt.tmthing->momx = -gdt.tmthing->momx;
					gdt.tmthing->momy = -gdt.tmthing->momy;
				}

				return false;
			}

			return true;
		}

		if (thing->flags & MF_NOCLIP) return true;

		if (   thing->misc & MJF_FLAT3D
			&& !(thing->flags & (MF_FLOAT|MF_SKULLFLY)) )
		{
			return !touch_vertically;
		}

		// v220-16 support for pushable things
		// v220-28 modified to fix bug
		if (touch_vertically && thing->flags2 & MF2_PUSHABLE)
		{
			// The other thing's momentum
			fixed_t momx = gdt.tmthing->momx;
			fixed_t momy = gdt.tmthing->momy;

			if (momx|momy) // check if momentum needs to be transferred
			{
				if (gdt.tmthing->player && gdt.tmthing->player->cmd.buttons & BT_USE)
				{
					// item is pushable and other thing is the player
					fixed_t maxm = MAX_OF(abs(momx), abs(momy));

					if (maxm <= FRACUNIT)
					{
						thing->momx = momx;
						thing->momy = momy;
					}
					else	// don't push object too far
					{
						fixed_t frac = FixedDiv(FRACUNIT, maxm);
						gdt.tmthing->momx = thing->momx = FixedMul(momx, frac);
						gdt.tmthing->momy = thing->momy = FixedMul(momy, frac);
					}
				}
				else if (gdt.tmthing->flags2 & MF2_PUSHABLE)
				{
					// both things are pushable
					thing->momx = FixedMul(momx, 0xC000);
					thing->momy = FixedMul(momy, 0xC000);
				}
			}
		}

		return false;
	}

	return true;
}


// This routine checks for Lost Souls trying to be spawned      // phares
// across 1-sided lines, impassible lines, or "monsters can't   //   |
// cross" lines. Draw an imaginary line between the PE          //   V
// and the new Lost Soul spawn spot. If that line crosses
// a 'blocking' line, then disallow the spawn. Only search
// lines in the blocks of the blockmap where the bounding box
// of the trajectory line resides. Then check bounding box
// of the trajectory vs. the bounding box of each blocking
// line to see if the trajectory and the blocking line cross.
// Then check the PE and LS to see if they're on different
// sides of the blocking line. If so, return true, otherwise
// false.

static int pe_x; // Pain Elemental position for Lost Soul checks
static int pe_y; // Pain Elemental position for Lost Soul checks
static int ls_x; // Lost Soul position for Lost Soul checks
static int ls_y; // Lost Soul position for Lost Soul checks

static RDBOOL PIT_CrossLine(line_t *ld, int *data)
{
  return 
    !((ld->ml_flags ^ ML_TWOSIDED) & (ML_TWOSIDED|ML_BLOCKING|ML_BLOCKMONSTERS))
    || data[BOXLEFT ] > ld->bbox[BOXRIGHT]
    || data[BOXRIGHT] < ld->bbox[BOXLEFT ]   
    || data[BOXTOP  ] < ld->bbox[BOXBASE ]
    || data[BOXBASE ] > ld->bbox[BOXTOP  ]
    || P_PointOnLineSide(pe_x, pe_y, ld) == P_PointOnLineSide(ls_x, ls_y, ld);
}

RDBOOL Check_Sides(mobj_t *actor, int x, int y)
{
	int bx, by, xl, xh, yl, yh, box[4];
	
	pe_x = actor->x;
	pe_y = actor->y;
	ls_x = x;
	ls_y = y;
	
	// Here is the bounding box of the trajectory
	
	box[BOXLEFT ] = pe_x < x ? pe_x : x;
	box[BOXRIGHT] = pe_x > x ? pe_x : x;
	box[BOXTOP  ] = pe_y > y ? pe_y : y;
	box[BOXBASE ] = pe_y < y ? pe_y : y;
	
	// Determine which blocks to look in for blocking lines	
	M_getBlockOffsets(box, 0, &xl, &xh, &yl, &yh);

	validcount++; // prevents checking same line twice
	for (bx = xl ; bx <= xh ; bx++)
	{
		for (by = yl ; by <= yh ; by++)
		{
			if (!P_BlockLinesIterator(bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_CrossLine, box))
				return true;
		}
	}																//   ^
	return(false);													//   |
}																	// phares

//-------------------------------------

//
// MOVEMENT CLIPPING
//

// v220-31 3D flats handling
static int subfn_check3Dflats(RDBOOL if_moving)
{
	int *flatslist;

	// If necessary find floor and ceiling thing is between
	if ( (flatslist = gdt.newsubsec->flats3DArray) )
	{
		int		i, count = *(int*)flatslist++;
		fixed_t	z_offset = sectors[flatslist[1]].planes[0].speed;

	//	if ( if_moving && !z_offset) return true;
	//	if (!if_moving &&  z_offset) return true;

		fixed_t	fhstore[8], chstore[8], fpstore[8];

		fixed_t	zp = gdt.tmthing->z + z_offset;
		fixed_t	zh = zp + gdt.tmthing->height;

		fhstore[0] = gdt.tmfloorz;
		fpstore[0] = gdt.fpic;

		// Remember - floor's higher than ceiling
		// Store in floor/ceil pairs where floor's below the ceiling
		for (i = 0; i < count; i++)
		{
			sector_t *lsec = &sectors[flatslist[0]];
			sector_t *rsec = &sectors[flatslist[1]];

			chstore[i] = rsec->floorheight + flatslist[2];
			fhstore[i + 1] = chstore[i] + flatslist[3];
			fpstore[i + 1] = lsec->floorpic;

			flatslist += FLIST_ENTRIES;
		}

		chstore[i] = gdt.tmceilingz;

		fixed_t offset = gdt.player && gdt.plrisjumping ? gdt.tmthing->z - gdt.tmthing->floorz : 0;

		count++;

		for (i = 0; i < count; i++)
		{
			if (fhstore[i] - zp - offset < MAXSTEP && zh - offset >= fhstore[i] && zh - offset <= chstore[i])
			{
				gdt.tmfloorz = gdt.tmdropoffz = fhstore[i];
				gdt.tmceilingz = chstore[i];
				gdt.tmthing->flat3Dlevel[1] = i;
				gdt.fpic = fpstore[i];
				// MJF_FLAT3D Used to suppress autoaim
				if (i) gdt.tmthing->misc |= MJF_FLAT3D;
				goto found_ok;
			}
		}

		gdt.blocked3D = true;
		if (gdt.tmthing->flags & MF_MISSILE) gdt.explodemissile = true;

		// Floor crush
		if (gdt.tmthing->z <= chstore[0])
		{
			gdt.tmfloorz = gdt.tmdropoffz = fhstore[0];
			gdt.tmceilingz = chstore[0];
		}
		else	// ceiling crush
		{
			gdt.tmfloorz = gdt.tmdropoffz = fhstore[count - 1];
			gdt.tmceilingz = chstore[count - 1];
		}

		return false;

found_ok:;

		gdt.tmthing->floorz = gdt.tmfloorz;
		gdt.tmthing->ceilingz = gdt.tmceilingz;

		if (gdt.moheight > gdt.tmceilingz && gdt.tmthing->z > gdt.tmfloorz)
		{
			gdt.tmthing->z = gdt.tmceilingz - gdt.tmthing->height;
		}

		if (gdt.tmthing->z < gdt.tmfloorz) gdt.tmthing->z = gdt.tmfloorz;
		gdt.tmz = gdt.tmthing->z;
		if (i) gdt.tmthing->misc |= MJF_FLAT3D;

		if (gdt.blocked3D) return false;
	}

	return true;
}

static RDBOOL compHeights(int *flatslist1, int *flatslist2)
{
	if (!flatslist1 || !flatslist2) return false;
	if (flatslist1 == flatslist2) return true;
	int c1 = *(int*)flatslist1++;
	int c2 = *(int*)flatslist2++;
	if (c1 != c2) return false;

	// ignore flats texture sec and flags
	for (c2 = 0; c2 < c1; c2++)
	{
		if (flatslist1[1] != flatslist2[1]) return false;
		if (flatslist1[2] != flatslist2[2]) return false;
		if (flatslist1[3] != flatslist2[3]) return false;
		flatslist1 += FLIST_ENTRIES;
		flatslist2 += FLIST_ENTRIES;
	}
	return true;
}

//
// P_CheckPosition
// This is purely informative, nothing is modified
// (except things picked up).
// 
// in:
//  a mobj_t (can be valid or invalid)
//  a position to be checked
//   (doesn't need to be related to the mobj_t->x,y)
//
// during:
//  special things are touched if MF_PICKUP
//  early out on solid lines?
//
// out:
//  newsubsec
//  floorz
//  ceilingz
//  tmdropoffz
//   the lowest point contacted
//   (monsters won't move to a drop off)
//  speciallines[]
//  numspeciallines
//
// test =  0, normal checking
// test =  1. check pickup behaviour on edge of move sectors
// test = -1, check 3D flats and Thing blocking only

RDBOOL P_CheckPosition(mobj_t* thing, fixed_t x, fixed_t y, fixed_t z, int pcptest)
{
	int		xl, xh;
	int		yl, yh;
	int		bx, by;
	int		i, ret;

	// Remember current position's slope ptr
	int		radius;

	//-----------------------------------

	numspechit = 0;
	numsplist = 0;
	molistnum = 0;
	mll.count = 0;

	thing->onmobj = NULL;
	thing->wallhit   = false;

	gdt.tmthing = thing;

/*	cleared using memset
	gdt.blocked3D
	gdt.blockmove
	gdt.exitgame
	gdt.explodemissile
	gdt.floatok		// $dropoff_fix: felldown
	gdt.mlblocking
	gdt.thingblocked
	// $unstuck: floorline, ceilingline used with gdt.player
	gdt.ceilingline
	gdt.floorline
    gdt.blockline
*/
	memset(&gdt.blocked3D, 0, 7*sizeof(int) + 3*sizeof(line_t*));
	
	if (thing->flags2 &  MF2_DONTDRAW) return false;
	if (thing->flags3 & (MF3_AMODEL|MF3_STATIC)) return false;
	
	// Set the rest of the gamedata array for the current thing
	gdt.player = thing->dplayer && thing->dplayer->mo == thing;

	// b211-0 Only reset if it's the player in case scanned elsewhere
	if (gdt.player) gdt.plrlineblocked = NULL;

	gdt.oldx = thing->x;
	gdt.oldy = thing->y;

	gdt.loslope = thing->subsector->sector->specFlags2 & SF2_LOSLOPE;

	// set to DDMAXINT if not on a slope
	gdt.oldslopez = gdt.loslope ? thing->z : DDMAXINT;

	gdt.tmx = x;
	gdt.tmy = y;
	gdt.tmz = z;
	gdt.tmheight = thing->height;
	gdt.moheight = z + thing->height;

	gdt.tmflags = thing->flags;
	gdt.solid = (gdt.tmflags  & (MF_SOLID|MF_CORPSE|MF_SPECIAL|MF_NOCLIP|MF_MISSILE)) == MF_SOLID;
	
	gdt.ssec = thing->subsector;
	gdt.newsubsec = R_PointInSubsector (x,y);

	gdt.fpic = gdt.newsubsec->ssfloorpic;
	
	gdt.fz = gdt.tmfloorz = gdt.tmdropoffz = gdt.newsubsec->sector->specFlags2 & SF2_LOSLOPE ?
		SL_PointHeight(gdt.newsubsec->sector, x, y, LOSLOPE) : gdt.newsubsec->sector->floorheight;

	gdt.z = gdt.loslope && thing->z == thing->floorz ? gdt.fz : thing->z;

	gdt.dropoffz3D = gdt.same3D = DDMAXINT;
	gdt.nextlowest = DDMININT;
	gdt.nexthighest = DDMAXINT;
	gdt.tmceilingz = SL_PointHeight(gdt.newsubsec->sector, gdt.tmx, gdt.tmy, HISLOPE);

	// $added v220-13
	if (thing->flags & MF_MISSILE && (z < gdt.fz || z > gdt.tmceilingz))
	{
		if (gdt.loslope && z < gdt.fz) thing->z += 16*FRACUNIT;
		gdt.explodemissile = true;
		return false;
	}

	// if (P_IsCamera(gdt.tmthing)) return true;

	ret = !(thing->flags & MF_SOLID && !(thing->flags & MF_NOCLIP));

	if (   gdt.fz + gdt.tmheight >= gdt.tmceilingz
		&& gdt.newsubsec->sector->specFlags2 & (SF2_LOSLOPE|SF2_HISLOPE))
	{
		return ret;
	}

	gdt.tmstptr = gdt.newsubsec->sector->stptr;

	gdt.tmthing->misc &= ~MJF_FLAT3D;	// b211-0

	// $slopes - gradient check
	if (   gdt.tmstptr >= 0
		&& !(gdt.player && waterlevel)
		&& !(thing->flags & MF_FLOAT)
		&& slopes[gdt.tmstptr].slope[0].gradient > 0x32000)
	{
		// going up
		if (gdt.tmfloorz > gdt.oldslopez)
		{
			return ret;
		}
		// going down
		if (
			// player can go down
			!gdt.player

			// missiles/skulls in flight are unaffected
			&& !(thing->flags & (MF_MISSILE|MF_SKULLFLY))

			// else check for too big a drop
			&& thing->subsector->sector->floorheight
			- gdt.newsubsec->sector->floorheight > MAXSTEP)
		{
			return ret;
		}
	}

	if (gdt.tmflags & MF_NOCLIP) return ret;

	gdt.compHeights = compHeights(gdt.newsubsec->flats3DArray, thing->subsector->flats3DArray);

	if (!gdt.compHeights)
	{
		thing->flat3Dlevel[1] = -1;

		if (gdt.tmstptr >= 0)
		{
			// stop missiles
			if (thing->flags & MF_MISSILE && thing->z <= gdt.fz)
			{
				// Set it slightly above the floor
				thing->z = gdt.fz + 4*FRACUNIT;
				gdt.explodemissile = true;

				return ret;
			}
			// check if a thing has been forced under a sloped floor
			else if (thing->z - gdt.fz < -MAXSTEP && thing->subsector == gdt.newsubsec)
			{
				thing->z = gdt.fz;
				if (thing->momz < 0) thing->momz = 0;
			}
		}
	}
	
	// No further movement
	if (thing->misc & MJF_SCP_SPLASH) return false;

	if (!subfn_check3Dflats(false)) return false;

	if (pcptest == -2) return true;

	// v220-14 Specifically prevent a Thing on 3D flats from stepping down > than its own height
	if (   thing->flags2 & MF2_NOSTEPDOWN
		&& thing->floorz == thing->z
		// since gdt.fz is the lowest floor under 3d flats then if the thing
		// is moving within a flats array then it can move 
		&& !gdt.compHeights
		&& thing->z - gdt.fz > thing->height)
	{
		return false;
	}

	// Check Things first, possibly picking Things up. The bounding box is
	// extended by the Thing's radius as mobj_t's are grouped into mapblocks,
	// based on their origin point, and can overlap into adjacent blocks.

	M_setBoxOffsets(gdt.tmbbox, x, y, thing->radius);
	M_getBlockOffsets(gdt.tmbbox, MAXRADIUS, &xl, &xh, &yl, &yh);

	validcount++;
	thing->validcount = validcount;	// don't check self
	mobj_mblistcount = 0;

	// Check all normal things and set up list for door things
	for (bx = xl ; bx <= xh ; bx++)
	{
		for (by = yl ; by <= yh ; by++)
		{
			if (!P_BlockThingsIterator(bx, by, (int (__cdecl *)(struct mobj_s *,void *))PIT_CheckThing, &pcptest))
			{
				gdt.thingblocked = true;
			}
		}
	}

	if (pcptest == -1) return !gdt.thingblocked;	// Signals to only check PIT_CheckThing.
	if (gdt.player) gdt.plrceilblocked = false;	// check lines

	radius = thing->radius;	// default

	if (pcptest == 1)
	{
		// v220-5 check_pickup is set true when checking the position of dropped
		//        or special Things that have not moved in the x,y plane and are
		//        on a moving plat. A small nominal value for their radius is used
		//        to keep them hooked to the plat by ignoring lines close by.
		radius = FRACUNIT;
	}
	else
	// v220-10 Use true radius for pickups on a non-scrolling mov floor. This
	//         prevents a Thing being affected by an adjacent floor sector.
	if (   thing->flags & (MF_SPECIAL|MF_DROPPED)
		&& (thing->subsector->sector->specFlags & (SF_MOVFLOOR|SF_MOVFIXED)) == SF_MOVFLOOR
		&& !(thing->misc & MJMISC_TRANS_TC))
	{
		radius = R_MaxRadius(thing, false, 5);
	}
	else if (radius >= FRACUNIT)
	{
		// v220-17 prevent dead Things getting stuck
		if (thing->health <= 0)
		{
			if (radius < 2*FRACUNIT && thing->flags3 & MF3_FALLING && thing->z > thing->floorz)
			{
				radius = 2*FRACUNIT;
			}
		}
		// v220-21 if radius is too small then lines may not be detected correctly
		else if (radius < 4*FRACUNIT && (thing->momx || thing->momy))
		{
			radius = 4*FRACUNIT;
		}
	}

	M_setBoxOffsets(gdt.tmbbox, x, y, radius);
	M_getBlockOffsets(gdt.tmbbox, 0, &xl, &xh, &yl, &yh);

	validcount++;

	// Set thing's line list
	for (bx = xl ; bx <= xh ; bx++)
	{
		for (by = yl ; by <= yh ; by++)
			P_BlockLinesIterator (bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_SetLineList, NULL);
	}

	opentop = gdt.tmceilingz;
	openbottom = gdt.tmfloorz;

	if (gdt.player)
	{
		gdt.ladderbase = DDMAXINT;
		gdt.laddertop = DDMININT;
	}

	for (i = 0; i < mll.count; i++)
	{
		// Check for blocking. Also adds line specials to list
		if (!PIT_CheckLine(mll.plist[i]))
		{
			gdt.blockline = mll.plist[i];
			return false;
		}
	}

	if (gdt.blocked3D) return false;

	if (!mll.count)
	{
		if (!subfn_check3Dflats(true)) return false;
	}

	// Do this last to ensure any line specials are added to list
	// b220 R3D proprietory - interactive amodel/doors support 
	for (i = 0; i < mobj_mblistcount; i++)
	{
		mobj_t	*mo = mobj_mblist[i];
		line_t	*dl;
		int		k, num = mo->modelboxnum - 1;

		if (num < 0) continue;

		if (mo->flags3 & MF3_INTERREACT)
		{
			for (k = 0; k < 4; k++)
			{
				if ( (dl = MB_GetLine(num, k, gdt.tmbbox)) )
				{
					fixed_t bottomStep = dl->backsector->floorheight + dl->d3d->offset3D;
					fixed_t topStep = bottomStep + dl->d3d->totalheight;

					if (bottomStep > gdt.moheight)
					{
						if (gdt.tmceilingz > bottomStep)
							gdt.tmceilingz = bottomStep;
						continue;
					}

					if (bottomStep > gdt.tmz + MAXSTEP)
					{
						gdt.blockline = dl;
						gdt.tmfloorz = topStep;
						return false;
					}

					if (gdt.tmfloorz < topStep)
					{
						gdt.tmfloorz = topStep;
					}
				}
			}
		}
		// Doors are expected to block (i.e. no step up)
		else if (MB_TestDoor(num))
		{
			return false;
		}
	}

	if (gdt.player) gdt.plrlineblocked = gdt.floorline;

	// Check for slope z positioning
	if (   gdt.tmthing->z < gdt.fz
		&& gdt.tmstptr >= 0
		&& gdt.newsubsec->sector->specFlags2 & SF2_LOSLOPE)
	{
		gdt.tmthing->z = gdt.tmthing->floorz = gdt.fz;
	}

	return !gdt.thingblocked;
}

RDBOOL checkStepupClear(mobj_t *chkmo, fixed_t z)
{
	int i;

	if ( !(chkmo->flags & MF_SOLID)
		|| chkmo->flags2 & MF2_NORENDER )
	{
		return true;
	}

	for (i = 0; i < molistnum; i++)
	{
		if (   molist[i]->thinker.function == (void (__cdecl *)(void))P_MobjThinker
			&& molist[i] != chkmo
			&& molist[i]->flags & MF_SOLID
			&& !(molist[i]->flags2 & MF2_NORENDER))	// exclude a coincident Thing
		{
			if (   z + chkmo->height > molist[i]->z
				&& z < molist[i]->z + molist[i]->height)
			{
				return false;	// mo's overlap
			}
		}
	}
	return true;
}

//
// P_TryMove
//  Attempt to move to a new position,
//  crossing special lines unless MF_TELEPORT is set.
//
//  killough $dropoff_fix
//

extern RDBOOL ignore_dropoff;

RDBOOL P_TryMove(mobj_t* thing, fixed_t x, fixed_t y, int from)
{
	int grad, maxstep = MAXSTEP;	// b220-1
	
	gdt.slopefht = DDMAXINT;		// b220-1

	// v220-6
	if (gdt.exitgame && gdt.mlblocking) G_ExitLevel ();

	// v220-4 Eliminate things that are not sector or blockmap linked
	if (  (thing->flags & (MF_NOBLOCKMAP|MF_NOSECTOR)) == (MF_NOBLOCKMAP|MF_NOSECTOR)
		|| thing->flags2 & MF2_DONTDRAW
		|| thing->flags3 & (MF3_AMODEL|MF3_STATIC))
	{
		return false;
	}

	if (!P_CheckPosition(thing, x, y, thing->z, 0) || gdt.blockmove)
	{
		if (thing->flags2 & MF2_PUSHABLE)
		{
			int	sndid = AS_GetPushSound(thing->info->doomednum);

			if (sndid && SFXBF_playing(thing)) S_StopSound(sndid, thing);
		}

		return false;
	}
	else if (thing->flags2 & MF2_PUSHABLE && (thing->momx || thing->momy))
	{
		int	sndid = AS_GetPushSound(thing->info->doomednum);

		if (sndid)
		{
			if (!thing->momz && !SFXBF_playing(thing)) S_SoundRepeat(sndid|HSF_REPEAT, thing, 1.0f, DEF_ATTN);
			else
			if (thing->momz && SFXBF_playing(thing)) S_StopSound(sndid, thing);
		}
	}

	if (gdt.dropoffz3D < DDMAXINT) gdt.tmdropoffz = gdt.dropoffz3D;

	// b220-1 increase step-up depending on a slope's gradient
	if (   gdt.tmstptr >= 0
		&& thing->subsector->sector->specFlags2 & SF2_LOSLOPE)
	{
		if (   gdt.floorline
			&& gdt.floorline->ml_flags & ML_TWOSIDED
			&& (gdt.floorline->frontsector->stptr < 0
			||  gdt.floorline->backsector->stptr < 0)
			&& (grad = slopes[gdt.tmstptr].slope[0].gradient) < 0x1E000)
		{
			maxstep += FixedDiv(grad, 0x5000);
		}

		gdt.slopefht = gdt.fz;
	}

	if ( !(thing->flags & MF_NOCLIP) )
	{
		RDBOOL valid;

		if (  !gdt.blockline
			&& !(thing->flags & (MF_DROPOFF|MF_FLOAT|MF_NOGRAVITY))
			&& thing->momz <= 0				// v220-14 don't care if rising or about to be raised
			&& thing->z == gdt.tmfloorz)	// and must be on the floor
		{
			// v220-10 check validity of dropoff
			if (gdt.tmdropoffz <= gdt.tmfloorz)
			{
				if (   gdt.fz == thing->z  	// at the required height
					|| thing->subsector->sector->specFlags & SF_MOVFIXED)	// or a fixed plat
				{
					// reset anything to do with falling
					thing->dropoffz = thing->z;
					gdt.tmdropoffz = gdt.tmfloorz;
					thing->momz = 0;
					thing->gear = 0;
					thing->flags3 &= ~MF3_FALLING;
				}
			}

			// b220-0 Allow decorations to move on scrolling floors in sky sectors with
			//        a low ceiling that would otherwise cause them to remain stuck 
			if (!sentient(thing) && thing->misc & MJMISC_TRANS_TC)
			{
				gdt.tmceilingz += gdt.newsubsec->sector->skyfix;
			}
		}

		if ( (valid = gdt.tmceilingz - gdt.tmfloorz >= thing->height) )
		{
			gdt.floatok = true;

			if ( !(thing->flags & MF_TELEPORT) )
			{
				if (   gdt.tmceilingz - thing->z < thing->height
					|| gdt.tmfloorz - thing->z > maxstep)
				{
					valid = false;
				}
				else
				{
					if (thing->z < gdt.tmfloorz)
						valid = checkStepupClear(thing, gdt.tmfloorz);
				}
			}
		}

		if (!valid)
		{
			if (   gdt.player					// if player
				&& ppt.viewer_fake_sector >= 0	// is in a fake sector
				&& waterlevel					// and in water,
				&& waterlevel < 4				// not flying,
				&& gdt.ceilingline				// and is blocked by a low ceiling
				&& gdt.tmceilingz > ppt.fake_floorheight	// which is higher than the fake floor height
				&& gdt.tmceilingz < gdt.tmz + gdt.tmheight	// and has been hit by player
				&& !untouched(gdt.ceilingline, gdt.tmx, gdt.tmy))
			{
				gdt.plrceilblocked = true;
			}

			// killough 8/ 1/98: Possibly allow escape if otherwise stuck
			return gdt.player
				&& !(gdt.ceilingline && untouched(gdt.ceilingline, gdt.oldx, gdt.oldy))
				&& !(gdt.floorline && untouched(gdt.floorline, gdt.oldx, gdt.oldy));
		}

		// killough 3/15/98: Allow certain objects to drop off
		// killough 7/24/98, 8/1/98: 
		// Prevent monsters from getting stuck hanging off ledges
		// killough 10/98: Allow dropoffs in controlled circumstances
		// killough 11/98: Improve symmetry of clipping on stairs

		if ( !(thing->flags & (MF_DROPOFF|MF_FLOAT)) )
		{
			if (gdt.tmfloorz - thing->z > MAXSTEP && gdt.nexthighest == DDMAXINT) return false;

			if (   thing->misc & MJF_ONLIFT
				&& thing->z == gdt.tmfloorz
				&& (gdt.same3D == DDMAXINT || abs(gdt.tmfloorz - gdt.same3D) > MAXSTEP)
				&& gdt.dropoffz3D < DDMAXINT	// must have contacted a line
				&& gdt.dropoffz3D > MAXSTEP)
			{
				return false;
			}
			else
			// v220-10 check validity of dropoff
			if (   sentient(thing)
				&& gdt.tmfloorz - gdt.tmdropoffz > maxstep
				// thing can be swept off edges by scroll carry unless in a fake sector
				&& !(thing->misc & MJMISC_TRANS_TC && thing->subsector->sector->heightsec < 0))
			{
			/*	removed v220-33
				if (!R3D_maptype)
				{
					if (   ( ignore_dropoff && thing->z == gdt.tmfloorz)
						|| (!ignore_dropoff && gdt.fz == gdt.tmfloorz && thing->z - gdt.fz < MAXSTEP))
					{
						goto done_ok;
					}

					return false;	// standard Doom behaviour
				}
				*/

				// v220-31 revised code
				fixed_t zh = SL_PointHeight(R_PointInSubsector(x, y)->sector, x, y, LOSLOPE);
				int s1 = gdt.ssec->sector->specFlags2 & SF2_LOSLOPE;
				int s2 = gdt.newsubsec->sector->specFlags2 & SF2_LOSLOPE;
				RDBOOL ret = true;

				if (s1 || s2)	// slope checks
				{
					if (   zh == gdt.tmfloorz		// at the required height?
						|| zh == thing->z
						|| (gdt.tmfloorz == thing->floorz
						&& thing->z == thing->floorz
						&& abs(thing->z - zh) < maxstep))
					{
						ret = false;
					}
				}
				// stair check
				else
				{
					if (   ( ignore_dropoff && thing->z == gdt.tmfloorz)
						|| (!ignore_dropoff && gdt.fz == gdt.tmfloorz && thing->z - gdt.fz <= MAXSTEP))
					{
						goto done_ok;
					}

					// Things can occasionally get stuck on steps so
					// a check is made that it spans two viable steps
					if (gdt.nexthighest < DDMAXINT || gdt.nextlowest > DDMININT)
					{
						if ( (s1 = gdt.tmfloorz - gdt.tmdropoffz) > 2*maxstep) return false;

						s2 = gdt.tmfloorz - gdt.nextlowest;

						// check if going down
						if (   gdt.nextlowest > DDMININT
							&& s2 > 0
							&& s2 <= maxstep
							&& s1 - s2 <= maxstep)
						{
							ret = false;
						}
						else	// check if going up
						if (   gdt.nexthighest == gdt.tmfloorz
							&& gdt.nextlowest == gdt.tmdropoffz
							&& gdt.tmz > gdt.nextlowest
							&& gdt.tmz < gdt.nexthighest
							&& gdt.tmz - gdt.nextlowest <= maxstep
							&& gdt.nexthighest - gdt.tmz <= maxstep)
						{
							ret = false;
						}
					}
					else if (thing->z > gdt.tmfloorz)
					{
						ret = false;
					}
				}

				if (ret) return false;
			}
		}
done_ok:;
		// killough 3/15/98: Allow certain objects to drop off
		// killough 7/24/98, 8/1/98: 
		// Prevent monsters from getting stuck hanging off ledges
		// killough 10/98: Allow dropoffs in controlled circumstances
		// killough 11/98: Improve symmetry of clipping on stairs

		// v220-10
		if (   thing->flags & (MF_SPECIAL|MF_DROPPED)			// if a pickup,
			&& compat[c_carry]									// where old doom code not used,
			&& gdt.tmfloorz - thing->z > 4*FRACUNIT				// is blocked by a height of 4
			&& thing->subsector->sector->floordata				// whilst on an active floor plat
			&& thing->misc & MJMISC_TRANS_TC)					// with a scroll carry floor
		{
			return false;										// then don't move it
		}

		if (   thing->flags3 & MF3_BOUNCES   // killough 8/13/98
			&& !(thing->flags & (MF_MISSILE|MF_NOGRAVITY))
			&& !sentient(thing) && gdt.tmfloorz - thing->z > 16*FRACUNIT)
		{
			return false; // too big a step up for bouncers under gravity
		}
	  
		// killough $dropoff_fix: prevent falling objects from going up too many steps
		if (   thing->flags3 & MF3_FALLING
			&& gdt.tmfloorz - thing->z >
			FixedMul(thing->momx, thing->momx) + 
			FixedMul(thing->momy, thing->momy))
		{
			return false;
		}
	}

	if (gdt.player) gdt.plrlineblocked = NULL;
	thing->floorz = gdt.tmfloorz;
	thing->ceilingz = gdt.tmceilingz;	
	thing->dropoffz = gdt.tmdropoffz; // killough $dropoff_fix: keep track of dropoffs

	// the move is ok
	if (thing->x != gdt.tmx || thing->y != gdt.tmy)
	{
		RDBOOL	oldstptr = false;	// Assume not on a slope

		if (   R3D_slopetypes
			&& gdt.tmstptr >= 0
			&& !(thing->flags & (MF_MISSILE|MF_NOGRAVITY|MF_FLOAT))
			&& thing->subsector->sector->specFlags2 & SF2_LOSLOPE)	// Low slope only
		{
			// Only if still on same slope (quicker than comparing
			// direction of gradients and an acceptable compromise)
			oldstptr = thing->subsector->sector->stptr == gdt.tmstptr;
		}
		
		P_UnlinkThing (thing);
		
		thing->x = gdt.tmx;
		thing->y = gdt.tmy;
		
		P_SetThingPosition (thing);

		// Slope servoz setting - also see in P_ZMovement()
		if (oldstptr && gdt.oldslopez < DDMAXINT)
			thing->servoz = gdt.tmfloorz - gdt.oldslopez;
	}

	// check_spechits
	// check if any special lines were hit	
	if (numsplist > 0 && !(thing->flags & (MF_TELEPORT|MF_NOCLIP)) )
	{
		RDBOOL	tp = false;
		int		i = -1;

		while (++i < numsplist)	// scan non-teleport types first
		{
			if (splist[i].ld)
			{
				if (splist[i].tpspec) { tp = true; continue; }

				P_CrossSpecialLine(splist[i].ld, thing);
				S_SetLineTriggeredAmbient(splist[i].ld->ambNumLink, 1);
			}
		}

		if (tp)
		{
			i = -1;

			while (++i < numsplist)	// scan teleport types
			{
				if (splist[i].tpspec && splist[i].ld)
				{
					P_CrossSpecialLine(splist[i].ld, thing);

					if (splist[i].ld->ambNumLink >= 0)	// v220-35
						S_SetLineTriggeredAmbient(splist[i].ld->ambNumLink, 2);

					if (!numsplist) break;	// If tp successful then will have been cleared
				}
			}
		}

		numsplist = 0;
	}

	if (   thing->flags2 & MF2_SPECNOMOVE
		&& ((thing->x>>16) != thing->spawnpoint.x || (thing->y>>16) != thing->spawnpoint.y))
	{
		thing->flags2 &= ~MF2_SPECNOMOVE;
	}

	if (thing->flags & MF_MISSILE || thing->flags3 & MF3_BOUNCES)
	{
		thing->old_x = gdt.oldx;
		thing->old_y = gdt.oldy;
	}
	return true;
}

static fixed_t thbox[4];

static RDBOOL PIT_CheckLines(line_t *ld, mobj_t *thing)
{
	int stepup;

	if (   thbox[BOXRIGHT] <= ld->bbox[BOXLEFT ]
		|| thbox[BOXLEFT ] >= ld->bbox[BOXRIGHT]
		|| thbox[BOXTOP  ] <= ld->bbox[BOXBASE ]
		|| thbox[BOXBASE ] >= ld->bbox[BOXTOP  ]
		|| P_BoxOnLineSide(thbox, ld) != -1)
	{
		// Line does not cross the thing's bounding box.
		return true;
	}

	if (ld->specType >= R3D_40) return false;
	if (!ld->backsector) return false;
	if (ld->frontsector->ceilingheight < thing->z + thing->height) return false;
	if (ld->backsector->ceilingheight < thing->z + thing->height) return false;
	if (ld->frontsector->floorheight > thing->z || ld->backsector->floorheight > thing->z)
	{
		stepup = MAX_OF(ld->frontsector->floorheight, ld->backsector->floorheight) - thing->z;

		if (stepup > MAXSTEP) return false;
		if (ld->frontsector->ceilingheight - stepup < thing->z + thing->height) return false;
		if (ld->backsector->ceilingheight - stepup < thing->z + thing->height) return false;
		return true;
	}
	return true;
}

void ED_MovePlayer(int x, int y)
{
	mbf_player_t	*plyr = &mbf_players[0];
	mobj_t			*mo = ppt.viewplayer->mo;
	int				clipstatus = plyr->cheats & CF_NOCLIP;
	int				ret = true;

	Sfx_StopSound(0, NULL); // v220-35

	P_UnlinkThing(mo);
	mo->x = x;
	mo->y = y;
	mo->momx = mo->momy = 1;
	P_SetThingPosition (mo);
	mo->z = mo->floorz = mo->subsector->sector->floorheight;

	if (R3D_edit) return;

	if (clipstatus) plyr->cheats &= ~CF_NOCLIP;

	// spawned in a Thing?
	if (!P_CheckPosition(mo, 1, 1, mo->z, -1)) // moving it starts things going
	{
		ret = false; goto endmove;
	}

	int		bx, by, xl, yl, xh, yh;

	M_setBoxOffsets(thbox, mo->x, mo->y, mo->radius);
	M_getBlockOffsets(thbox, mo->radius, &xl, &xh, &yl, &yh);

	validcount++;

	for(bx = xl; bx <= xh; bx++)
	{
		for(by = yl; by <= yh; by++)
		{
			if (!P_BlockLinesIterator(bx, by, (int (__cdecl *)(struct line_s *,void *))PIT_CheckLines, mo))
			{
				ret = false; goto endmove;
			}
		}
	}

endmove:;

	if (clipstatus) plyr->cheats |= CF_NOCLIP;	// reset state on entry

	if (ret)	// no clip can be set to off
	{
		if (clipstatus) cht_NoClipFunc(plyr);	// reset back with message
	}
	else	// no clip should be set to on
	{
		if (!clipstatus) cht_NoClipFunc(plyr);	// reset back with message
	}
}

//-----------------------------------------------------------
//
// Routines for wall slide

static fixed_t		bestslidefrac;
static line_t		*bestslideline;

static fixed_t		tmxmove;
static fixed_t		tmymove;

//
// P_HitSlideLine
// Adjusts the xmove / ymove
// so that the next move will slide along the wall.
//
void P_HitSlideLine (line_t* ld, RDBOOL isplayer)
{
	RDBOOL	icyfloor;  // is floor icy?								// phares
	//   |
	// Under icy conditions, if the angle of approach to the wall	//   V
	// is more than 45 degrees, then you'll bounce and lose half
	// your momentum. If less than 45 degrees, you'll slide along
	// the wall. 45 is arbitrary and is believable.
	
	// killough 10/98: only bounce if hit hard (prevents wobbling)
	icyfloor =
		slidemo->z <= slidemo->floorz
		&& P_ApproxDistance(tmxmove, tmymove) > 4*FRACUNIT
		&& P_GetFriction(slidemo, NULL) > ORIG_FRICTION;
	
	// Check for the special cases of horiz or vert walls.
	if (ld->slopetype == ST_HORIZONTAL)
	{
		if (icyfloor && (abs(tmymove) > abs(tmxmove)))
		{
			if (isplayer) S_StartSound(sfx_oof, slidemo);
			tmxmove /= 2; // absorb half the momentum
			tmymove = -tmymove/2;
		}
		else
			tmymove = 0; // no more movement in the Y direction
		return;
	}
	
	if (ld->slopetype == ST_VERTICAL)
	{
		if (icyfloor && (abs(tmxmove) > abs(tmymove)))
		{
			if (isplayer) S_StartSound(sfx_oof, slidemo);
			tmymove /= 2; // absorb half the momentum
			tmxmove = -tmxmove/2;
		}															//   |
		else														// phares
			tmxmove = 0; // no more movement in the X direction
		return;
	}
	
	// The wall is angled. Bounce if the angle of approach is		// phares
	// less than 45 degrees.										// phares
	
	if (icyfloor)
	{
		int			side = P_PointOnLineSide (slidemo->x, slidemo->y, ld);
		angle_t		lineangle = R_PointToAngle2 (0,0, ld->dx, ld->dy);
		// +10 prevents sudden path reversal due to rounding error - phares
		angle_t		moveangle = R_PointToAngle2 (0,0, tmxmove, tmymove) + 10;
		angle_t		deltaangle;
		
		if (side) lineangle += ANG180;
				
		deltaangle = moveangle-lineangle;
		
		if (deltaangle > ANG45 && deltaangle < ANG90+ANG45)
		{
			fixed_t movelen = P_ApproxDistance(tmxmove, tmymove);

			moveangle = lineangle - deltaangle;
			movelen /= 2; // absorb
			moveangle >>= ANGLETOFINESHIFT;							//   ^
			tmxmove = FixedMul (movelen, finecosine[moveangle]);	//   |
			tmymove = FixedMul (movelen, finesine[moveangle]);		// phares
			if (isplayer) S_StartSound(sfx_oof, slidemo);
			return;
		}
	}

	// v220-5 Deal with a Thing sticking on long lines with small angles
	//        giving inaccurate results. As such using angles has been
	//        supplanted with vectors. The method used here follows the
	//        neat one devised by Rand Heit in ZDoom. Noted with thanks.
	{
		divline_t	dl, dv;
		fixed_t		a , b;

		dl.x = ld->v1->x; dl.dx = ld->dx;
		dl.y = ld->v1->y; dl.dy = ld->dy;
		
		dv.x = slidemo->x; dv.dx =  ld->dy;
		dv.y = slidemo->y; dv.dy = -ld->dx;

		a = P_InterceptVector(&dl, &dv);

		dv.dx = tmxmove; dv.dy = tmymove;

		if (   (b = P_InterceptVector(&dv, &dl)) 
			&& (a = P_InterceptVector(&dl, &dv) - a) )
		{
			a = FixedDiv(a, b);
			tmxmove = FixedMul(a, dl.dx);
			tmymove = FixedMul(a, dl.dy);
		}
		else
		{
			tmxmove = tmymove = 0;
		}
	}
}

//
// PTR_SlideTraverse
//
static RDBOOL PTR_SlideTraverse (intercept_t* in)
{
	line_t*	li;
	
	if (!in->isaline)
		Con_ErrorD("PTR_SlideTraverse: not a line?");
	
	li = in->d.line;
	
	if (!(li->ml_flags & ML_TWOSIDED))
	{
		if (P_PointOnLineSide (slidemo->x, slidemo->y, li))
			return true;	// don't hit the back side
	}
	else if (li->ml_flags & ML_BLOCKING
		|| li->lnflags2 & (LNF2_BLOCKALL|LNF2_LOWTRIWALL)
		|| (li->ml_flags & ML_BLOCKMONSTERS
		&& !(slidemo->type == MT_PLAYER || slidemo->flags3 & MF3_FRIEND)) )
	{
		// blocks
	}
	else
	{
		// set openrange, opentop, openbottom
		P_LineOpening (li, slidemo->z + slidemo->height, slidemo->x, slidemo->y);
		
		if (   openrange >= slidemo->height					// fits
			&& (opentop - slidemo->z >= slidemo->height)	// mobj not too high
			&& (openbottom - slidemo->z <= MAXSTEP))		// can step up
		{
			if (slidemo->z >= openbottom)
				return true;
			
			// b220-0 If stepping up check slidemo won't be blocked
			if (checkStepupClear(slidemo, openbottom))
			{
				return true;    // this line doesn't block movement
			}
		}
	}
	
	// the line blocks movement, see if it is closer than the best so far
	if (in->frac < bestslidefrac)
	{
		bestslidefrac = in->frac;
		bestslideline = li;
	}
	
	return true;
}

//
// P_SlideMove
// The momx / momy move is bad, so try to slide
// along a wall.
// Find the first line hit, move flush to it,
// and slide along it
//
// This is a kludgy mess.
//
// killough 3/15/98: Allow objects to drop off ledges
//     by setting the fourth parm of P_TryMove() true

void P_SlideMove (mobj_t* mo)
{
	RDBOOL	isplayer = mo->dplayer && mo->dplayer->mo == mo;
	fixed_t leadx, leady, trailx, traily;
	fixed_t tryx = mo->momx, tryy = mo->momy;
	fixed_t	radius = mo->radius, useradius;
	int		hitcount = 3;

	// v220-5 voodoo dolls or player exiting tp never slide
	if (!mo->type && (!isplayer || mo->reactiontime > 0)) return;

	slidemo = mo; // the Thing that's sliding

	do
	{
		if (!--hitcount)
			goto stairstep;   // don't loop forever
		
		bestslidefrac = FRACUNIT+1;
		is_sliding = true;
		
		// trace along the three leading corners
		useradius = tryx > 0 ? radius : -radius;
		leadx = mo->x + useradius, trailx = mo->x - useradius;
		useradius = tryy > 0 ? radius : -radius;
		leady = mo->y + useradius, traily = mo->y - useradius;

		P_PathTraverse(leadx, leady, tryx, tryy, true,
			PT_ADDLINES, PTR_SlideTraverse);
		P_PathTraverse(trailx, leady, tryx, tryy, true,
			PT_ADDLINES, PTR_SlideTraverse);
		P_PathTraverse(leadx, traily, tryx, tryy, true,
			PT_ADDLINES, PTR_SlideTraverse);
		
		is_sliding = false;

		// move up to the wall
		if (bestslidefrac == FRACUNIT+1)	// the move must have hit the middle
		{
stairstep:

			if (!P_TryMove (mo, mo->x, mo->y + tryy, 5))
				P_TryMove (mo, mo->x + tryx, mo->y, 5);

			break;
		}

		// fudge a bit to make sure it doesn't hit
		if ((bestslidefrac -= 0x800) > 0)
		{
			fixed_t newx = FixedMul(tryx, bestslidefrac);
			fixed_t newy = FixedMul(tryy, bestslidefrac);
			
			// killough 3/15/98: Allow objects to drop off ledges
			if ( !P_TryMove(mo, mo->x + newx, mo->y + newy, 5) )
				goto stairstep;
		}

		// Now continue along the wall.
		// First calculate remainder.
		bestslidefrac = FRACUNIT - (bestslidefrac + 0x800);
		if (bestslidefrac <= 0) break;
		if (bestslidefrac > FRACUNIT) bestslidefrac = FRACUNIT;		
		
		tryx = tmxmove = FixedMul(tryx, bestslidefrac);
		tryy = tmymove = FixedMul(tryy, bestslidefrac);

		P_HitSlideLine(bestslideline, isplayer); // clip the moves		

		mo->momx = tmxmove;
		mo->momy = tmymove;
	}
	while (!P_TryMove(mo, mo->x + tmxmove, mo->y + tmymove, 5));
}

//
// P_LineAttack
//
mobj_t*		mobjtarget;	// who got hit (or NULL)

RDBOOL		CheckLineOpenings(line_t *ld);
RDBOOL		P_BulletLineOpening(line_t *ld);
lineopen_t	lop;

static int		aim_flags_mask;
static int		la_damage;

//
// PTR_AimTraverse
// Sets mobjtarget and aimslope when a target is aimed at.
//
RDBOOL PTR_AimTraverse (intercept_t* in)
{
	line_t		*li;
	mobj_t		*th;
	fixed_t		tempslope;
	fixed_t		thingtopslope;
	fixed_t		thingbottomslope;
	
	if (in->isaline)
	{
		li = in->d.line;

		if ( !(li->ml_flags & ML_TWOSIDED) )
			return false;		// stop

		lop.front = li->frontsector;
		lop.back  = li->backsector;
		lop.dist  = FixedMul (lop.attackrange, in->frac);
		lop.frac  = in->frac;
		
		return !CheckLineOpenings(li);
	}
	
	// shoot a thing	
	if (   (th = in->d.thing) == shootthing	// can't shoot itself
		|| !(th->flags & MF_SHOOTABLE)		// corpse for example
		|| th->flags3 & MF3_AMODEL)			// v220-15 amodels, although shootable, are purely decorative
	{
		return true;
	}

	// check angles to see if the thing can be aimed at
	lop.dist = FixedMul (lop.attackrange, in->frac);
	thingtopslope = FixedDiv (th->z + th->height - lop.shootz, lop.dist);

	if (thingtopslope < lop.bottomslope)
		return true;			// shot over the thing
	
	thingbottomslope = FixedDiv (th->z - lop.shootz, lop.dist);

	if (thingbottomslope > lop.topslope)
		return true;			// shot under the thing

	// killough 7/19/98, 8/2/98:
	// friends don't aim at friends (except players), at least not first
	if ((th->flags3 & shootthing->flags3 & aim_flags_mask) && !th->player)
		return true;

	//  v220-24 prevent autoaim selecting thing that can't be seen
	//          Only applies for the player (or voodoo dolls)
	if (   lop.autoaim
		&& shootthing->player
		&& !P_CheckSight(shootthing, th))
	{
		return false;	// Can't see it, can't shoot it
	}

	// this thing can be hit!
	if (thingtopslope > lop.topslope)
		thingtopslope = lop.topslope;
	
	if (thingbottomslope < lop.bottomslope)
		thingbottomslope = lop.bottomslope;
	
	// v220-5 if possible aim higher up the thing
	tempslope = FixedDiv (th->z + FixedMul(th->height, 0xC000) - lop.shootz, lop.dist);

	if (tempslope <= lop.topslope && tempslope >= lop.bottomslope)
	{
		lop.aimslope = tempslope;
	}
	else
	{
		lop.aimslope = (thingtopslope+thingbottomslope)/2;
	}

	mobjtarget = th;
	
	return false;	// flag no further
}

// v220-16 dpm.floorz used to determine whether bullets,
//         when firing up, should go through a fake floor
RDBOOL CalcXYZ(fixed_t infrac)
{
	fixed_t		frac, dx, dy, dz, wz;
	int			divisor = 2;

	// Adustment to position a bit closer.
	frac = FixedDiv (4*FRACUNIT,lop.attackrange);

	if (plane3d.set)
	{
		frac = plane3d.frac - frac;
		dpm.x = plane3d.x, dpm.y = plane3d.y, dpm.z = plane3d.z;
	}
	else
	{
		frac = infrac - frac;
		dpm.x = trace.x + FixedMul (trace.dx, frac);
		dpm.y = trace.y + FixedMul (trace.dy, frac);
		dpm.z = lop.shootz + FixedMul (lop.aimslope, FixedMul(frac, lop.attackrange));
	}

	dpm.contacted = R_PointInSubsector(dpm.x, dpm.y);

	// Is it a sky hack wall? If the hitpoint is above 
	// the visible line, no puff must be shown.
	if (   lop.back 
		&& lop.front->ceilingpic == skyflatnum
		&& lop.back->ceilingpic == skyflatnum
		&& (dpm.z > lop.front->ceilingheight
			|| dpm.z > lop.back->ceilingheight))
	{
		return false;
	}

	dpm.ceilz  = dpm.contacted->dlceilheight;
	dpm.floorz = dpm.contacted->dlfloorheight;

	// Should we backtrack to hit a plane instead?
	if (!(dz = dpm.z - lop.shootz) ) return true;

	dx = dpm.x - trace.x;
	dy = dpm.y - trace.y;

	if (dpm.contacted->flats3DArray)
	{
		dpm.ceilz  = dpm.contacted->sector->ceilingheight;

		if (   dpm.z < lop.shootz	// looking down
			&& (wz = dpm.contacted->s_waterplane) > DDMININT
			&& shootthing->z > wz)
		{
			dpm.floorz = wz;
		}
		else
		{
			dpm.floorz = dpm.contacted->sector->floorheight;
		}
	}
	else
	{
		dpm.ceilz  = dpm.contacted->dlceilheight;
		dpm.floorz = dpm.contacted->dlfloorheight;

		// looking up?
		if (   dpm.z > lop.shootz
			&& dpm.contacted->sector->heightsec >= 0)
		{
			dpm.ceilz = dpm.contacted->sector->ceilingheight;
		}
	}

	// Find the approximate hitpoint by stepping back and
	// forth using smaller and smaller steps. 
	while(divisor <= 512 && (dpm.z > dpm.ceilz || dpm.z < dpm.floorz))
	{
		// Take a step backwards.
		dpm.x -= dx / divisor;
		dpm.y -= dy / divisor;
		dpm.z -= dz / divisor;
		
		// Divisor grows.
		divisor += divisor;
		
		while(dpm.z <= dpm.ceilz && dpm.z >= dpm.floorz)
		{
			dpm.x += dx / divisor;
			dpm.y += dy / divisor;
			dpm.z += dz / divisor;
		}

		dpm.contacted = R_PointInSubsector(dpm.x, dpm.y);

		if (dpm.contacted->flats3DArray)
		{
			dpm.ceilz  = dpm.contacted->sector->ceilingheight;

			if (   dpm.z < lop.shootz	// looking down
				&& (wz = dpm.contacted->s_waterplane) > DDMININT
				&& shootthing->z > wz)
			{
				dpm.floorz = wz;
			}
			else
			{
				dpm.floorz = dpm.contacted->sector->floorheight;
			}
		}
		else
		{
			dpm.ceilz  = dpm.contacted->dlceilheight;
			dpm.floorz = dpm.contacted->dlfloorheight;

			// looking up?
			if (   dpm.z > lop.shootz
				&& dpm.contacted->sector->heightsec >= 0)
			{
				dpm.ceilz = dpm.contacted->sector->ceilingheight;
			}
		}
	}

	if (dpm.z > dpm.ceilz ) return (dpm.contacted->ssceilpic != skyflatnum);
	if (dpm.z < dpm.floorz) return (dpm.contacted->ssfloorpic != skyflatnum);
	return true;
}

// R3D
int crossSlopedSector(line_t *li, fixed_t infrac)
{
	fixed_t	x = trace.x + FixedMul(trace.dx, infrac);
	fixed_t	y = trace.y + FixedMul(trace.dy, infrac);

	int		side    = P_PointOnLineSide(trace.x, trace.y, li);
	int		sidenum = li->sidenum[side];

	if (sidenum >= 0)	// should always be true
	{
		fixed_t	z = lop.shootz + FixedMul (lop.aimslope, FixedMul(infrac, lop.attackrange));

		sector_t *s1 = sides[sidenum].sector;
		sector_t *s2 = li->frontsector == s1 ? li->backsector : li->frontsector;

		int		txn = sides[sidenum].texFlags;
		int		fht = s1->floorheight;
		int		cht = s1->ceilingheight;

		RDBOOL	slo = z < fht;
		RDBOOL	shi = z > cht;

		RDBOOL	s2valid = (s2 && s2->stptr >= 0);

		fixed_t	z0 = (s2valid && !slo) ? SL_PointHeight(s2, x, y, LOSLOPE) : fht;
		fixed_t	z1 = (s2valid && !shi) ? SL_PointHeight(s2, x, y, HISLOPE) : cht;

		RDBOOL	reslo = !slo && z <= z0;
		RDBOOL	reshi = !shi && z >= z1;

		if (s2)
		{
			if (s1->stptr < 0)
			{
				if (slo || shi) return -1;							// Hit floor/ceiling in front off slope line

				if (!reslo && !reshi)								// Nothing was hit
				{
					lop.lastx = x, lop.lasty = y, lop.lastz = z;
					return 1;
				}
			}

			if (txn & T_MID && (reslo || reshi))					// Hit slope's triwall
			{
				fixed_t frac = infrac - FixedDiv(4*FRACUNIT, lop.attackrange);
				lop.lastx = trace.x + FixedMul(trace.dx, frac);
				lop.lasty = trace.y + FixedMul(trace.dy, frac);
				lop.lastz = z;

				return 0;
			}
		}

		if (s1->stptr < 0)
		{
			lop.lastx = x, lop.lasty = y, lop.lastz = z;
			return 1;
		}

		reslo = slo || (s1->specFlags2 & SF2_LOSLOPE && z <= SL_PointHeight(s1, x, y, LOSLOPE));
		reshi = shi || (!reslo && s1->specFlags2 & SF2_HISLOPE && z >= SL_PointHeight(s1, x, y, HISLOPE));

		if (reslo || reshi)											// Hit slope's floor/ceiling
		{
			// Calculate intersect point of the bullet slope with the sloped plane
			float xA = (float)lop.lastx, yA = (float)lop.lasty, zLA = (float)lop.lastz;
			float xB = (float)x, yB = (float)y, zLB = (float)z;
			float zRA, zRB, xyzDiv;

			if (reslo)
			{
				zRA = (float)SL_PointHeight(s1, lop.lastx, lop.lasty, LOSLOPE);
				zRB = (float)SL_PointHeight(s1, x, y, LOSLOPE);
			}
			else
			{
				zRA = (float)SL_PointHeight(s1, lop.lastx, lop.lasty, HISLOPE);
				zRB = (float)SL_PointHeight(s1, x, y, HISLOPE);
			}
			
			xyzDiv = zLA - zLB - zRA + zRB;

			// if xyzDiv == 0 then bullet slope is parallel with trace line of equal z heights across slope
			if (FABS(xyzDiv) > 0.01f)
			{
				xyzDiv = 1 / xyzDiv;

				lop.lastx = (int)((-xA*zLB + xB*zLA + xA*zRB - xB*zRA) * xyzDiv);
				lop.lasty = (int)((-yA*zLB + yB*zLA + yA*zRB - yB*zRA) * xyzDiv);
				lop.lastz = (int)((zLA*zRB - zLB*zRA) * xyzDiv);
				
				if (R3D_edit) return -2;

				// Check if hitting a faked floor
				if (lop.lastz < R_PointInSubsector(lop.lastx, lop.lasty)->dlfloorheight) return -1;

				lop.lastz += reslo ? 2*FRACUNIT : -4*FRACUNIT;

				return 0;
			}
		}

		if (!s2 || z < R_PointInSubsector(x, y)->dlfloorheight) return -1;

		lop.lastx = x, lop.lasty = y, lop.lastz = z;

		return z > s2->floorheight && z < s2->ceilingheight;
	}
	return 0;
}

// R3D
static RDBOOL checkSplash(fixed_t x, fixed_t y, fixed_t z, subsector_t *ss)
{
	if (z - 2*FRACUNIT <= ss->dlfloorheight)
	{
		int flags = R_GetFlatFlags(ss->ssfloorpic);

		if (flags & FLT_SPLASH)
		{
			P_SpawnSplash(x, y, ss->dlfloorheight, cfg.splashvol, flags, false);
			return true;
		}
	}
	return false;
}

// Only called in edit mode
static int CheckOVF_SPECLine(line_t *ld, int snum)
{
	if (snum >= 0)
	{
		side_t *ls = &sides[snum];

		if (ls->lsiflags & LS_LOWTEX &&
			lop.editz >= ls->ll &&
			lop.editz <= ls->lh)
		{
			return true;
		}
		else if (ls->lsiflags & LS_MIDTEX &&
			lop.editz >= ls->ml + ls->editml &&
			lop.editz <= ls->mh - ls->editmh)
		{
			return true;
		}
		else if (ls->lsiflags & LS_HITEX &&
			lop.editz >= ls->hl &&
			lop.editz <= ls->hh)
		{
			return true;
		}
	}

	return false;
}

//
// PTR_ShootTraverse
//
// R3D - Modified for fake sectors, sloped sectors and 3D lines
//
RDBOOL PTR_ShootTraverse (intercept_t* in)
{
	line_t		*li; 
	mobj_t		*th;

	//----------------------------------

	if (in->isaline)
	{
		RDBOOL	ricoh = la_damage
						&& r_use_particles
						&& cfg.ricochet
						&& shootthing->player
						&& (shootthing->player->readyweapon == wp_pistol
						||  shootthing->player->readyweapon == wp_chaingun);

		int	snum;
		li = in->d.line;

		lop.front  = li->frontsector;
		lop.back   = NULL;
		lop.frac   = in->frac;
		lop.dist   = FixedMul (lop.attackrange, in->frac);
				
		lop.editz = lop.shootz + FixedMul (lop.aimslope, FixedMul(in->frac, lop.attackrange));	// gives same result

		plane3d.set = false;

		if (lop.has3Dflats)
		{
			fixed_t x = trace.x + FixedMul (trace.dx, in->frac);
			fixed_t y = trace.y + FixedMul (trace.dy, in->frac);

			dummy3D.x = x, dummy3D.y = y, dummy3D.z = lop.editz;
			dummy3D.subsector = R_PointInSubsector(x, y);
			dummy3D.height = FRACUNIT;

			P_CheckSight(shootthing, &dummy3D);

			if (plane3d.set)
			{
				if (R3D_edit) return false;

				goto calcxyz;
			}
		}

		if (R3D_edit)
		{
			snum = li->sidenum[P_PointOnLineSide(viewx, viewy, li)];

			if (snum < 0) return false;

			// Check if facing a flat determining sector
			if (li->specType >= R3D_41 && li->specType <= R3D_43)
			{
				int		on = snum == *li->sidenum ? li->sidenum[1] : -1;
				int		liftsec = li->frontsector->liftsec3D;	// Lift sector
				int		height  = liftsec >= 0 ?
							li->frontsector[liftsec].smoothfloor : li->backsector->smoothfloor;
				fixed_t	base = height + li->d3d->offset3D;
				fixed_t	top = base + li->d3d->lineheightlow;

				if (on >= 0 && sides[on].sector->sector3D >= 0)
				{
					if (lop.editz >= base && lop.editz <= top)
					{
						if (   origTEX[li - lines] & 0x01
							&& sides[snum].ll == base
							&& sides[snum].lh == top)
						{
							Edit_Textures(li, lop.editz, 1);
							return false;
						}

						return false;
					}
				}

				// Must be facing the backsector
				if (lop.editz >= base && lop.editz <= top)
				{
					return false;
				}
			}

			if ((snum == *li->sidenum || snum == li->sidenum[1]) && CheckOVF_SPECLine(li, snum))
			{
				Edit_Textures(li, lop.editz, 2);
				return false;
			}
		}

		// v220-35 ensure wall is not triggered if shot is under or over
		if (li->special == 8143)
		{
			if (   lop.editz < li->frontsector->floorheight
				|| (li->backsector ? lop.editz > li->backsector->floorheight : lop.editz > li->frontsector->ceilingheight))
			{
				return true;
			}
		}

		// In Doom a gun trigger works even if the bullet passes across
		// the line. This behaviour is not allowed in an R3D scripted
		// line or slope map where the line must block the bullet with
		// the.exception of special 8114.
		if (   la_damage
			&& li->special
			&& li->specType < R3D_40
			&& (!R3D_maptype || li->special == 8114))
		{
			P_ShootSpecialLine(shootthing, li, 0);
		}

		// crosses a two sided line
		if (li->ml_flags & ML_TWOSIDED)
		{		
			lop.back = li->backsector;
			lop.dist = FixedMul (lop.attackrange, lop.frac);

			if (li->frontsector->stptr >= 0 || li->backsector->stptr >= 0)
			{
				switch(crossSlopedSector(li, in->frac))
				{
				case  1: break;			// Nothing hit, check the line opening	
				case -1: goto calcxyz;	// Standard floor
				case -2: return false;	// Edit flats mode
				case  0:
					if (R3D_edit && la_damage)
					{
						Edit_Textures(li, lop.editz, 3);
					}
					else
					{
						subsector_t *ss = R_PointInSubsector(lop.lastx, lop.lasty);

						if (  !(ss->sector->specFlags2 & SF2_LOSLOPE)
							&& lop.lastz > ss->dlfloorheight
							&& lop.lastz - 4*FRACUNIT <= ss->dlfloorheight)
						{
							goto calcxyz;	// Standard floor
						}

						// v220-3 stop bullets hitting sloped sky ceilings
						if (   ss->sector->specFlags2 & SF2_HISLOPE
							&& ss->sector->ceilingpic == skyflatnum
							&& lop.lastz > SL_LowCeilHeight(ss->sector))
						{
							return false;	// Hit a sky ceiling
						}

						if (la_damage)	// i.e. not a test trace
						{
							if (!checkSplash(lop.lastx, lop.lasty, lop.lastz, ss))
								P_SpawnPuff(lop.lastx, lop.lasty, lop.lastz, ricoh);
						}
					}
					mobjtarget = NULL;
					return false;
				}
			}	
			
			if (P_BulletLineOpening(li))
			{
				if (la_damage)
				{
					if (R3D_edit)
					{
						if (Edit_Textures(li, lop.editz, 4))
							return false;
					}

					if (   R3D_maptype
						&& li->specType < R3D_40
						&& !(li->lnflags2 & LNF2_OWNS_SWITCH) )
					{
						if (P_ShootSpecialLine (shootthing, in->d.line, 1))
							S_SetLineTriggeredAmbient(in->d.line->ambNumLink, 3);
					}
					else
					{
						if (li->special)
						{
							// Test for impact model switch R3D_40 line
							TS_Activate(shootthing, li, MS_GUN);
						}
					}
				}

				// v220-31 ensure same sided line gets checked
				if (   !R3D_edit
					&& li->specType < R3D_40
					&& li->frontsector == li->backsector
					&& lop.lastz > li->frontsector->floorheight
					&& lop.lastz < li->frontsector->ceilingheight)
				{
					return true;
				}

				goto calcxyz;
			}

			return true;
		}
		else	// single sided sloped line handling
		{
			if (li->frontsector->stptr >= 0)
			{
				switch(crossSlopedSector(li, in->frac))
				{
				case  1: return true;	// Nothing hit ??
				case -1: goto calcxyz;	// Standard floor
				case -2: return false;	// Edit flats mode
				case  0:
					if (!R3D_edit && la_damage)	// i.e. not a test trace
					{
						P_SpawnPuff(lop.lastx, lop.lasty, lop.lastz, ricoh);
					}
					mobjtarget = NULL;
					return false;
				}
			}
		}

calcxyz:;

		if (R3D_edit)
		{
			if (la_damage)
			{
				if (Edit_Textures(li, lop.editz, 6))
					return false;
			}

			return false;
		}
		
		mobjtarget = NULL;

		if (!CalcXYZ(in->frac)) return false;	// don't bother going further

		if (plane3d.set)
		{
			if (plane3d.floorpic >= 0)
			{
				int flags;

				if ( (flags = R_GetFlatFlags(plane3d.floorpic)) && flags & FLT_SPLASH)
				{
					P_SpawnSplash(dpm.x, dpm.y, plane3d.z - 2*FRACUNIT, cfg.splashvol, flags|F3D_3DFLAT, false);
				}
				else
				{
					P_SpawnPuff (dpm.x, dpm.y, plane3d.z, ricoh);
				}

				return false;
			}
			
			if (plane3d.ceilpic >= 0)
			{
				P_SpawnPuff (dpm.x, dpm.y, plane3d.z, ricoh);
				return false;
			}
		}

		if (dpm.z <= dpm.contacted->dlfloorheight)
		{
			dpm.z = dpm.contacted->dlfloorheight;

			if (!R3D_edit)
			{
				if (la_damage && checkSplash(dpm.x, dpm.y, dpm.z, dpm.contacted))
				{
					return false;
				}
			}
			dpm.z += 2*FRACUNIT;
		}
		else if (dpm.z >= dpm.contacted->dlceilheight)
		{
			if (dpm.ceilz > dpm.contacted->dlceilheight)
				dpm.z = dpm.ceilz - 2*FRACUNIT;
			else
				dpm.z = dpm.contacted->dlceilheight - 2*FRACUNIT;
		}
		else if (la_damage)
		{
			if (R3D_edit) return false;
			
			if (   R3D_maptype
				&& li->special
				&& li->specType < R3D_40)
			{
				if (P_ShootSpecialLine (shootthing, li, 2))
					S_SetLineTriggeredAmbient(li->ambNumLink, 4);
			}
		}
		
		if (la_damage)
		{
			if (shootthing->player && ma_inf.plr_melee_ht)	// punch or chainsaw
			{
				// Don't spawn sparks if;
				if (  !gdt.plrblocked			// 1) player not against wall
					||  dpm.z < shootthing->z	// 2) lower than player's feet
					|| (dpm.z > shootthing->z	// 3) beyond player's reach when looking up
					   + shootthing->height + 4*FRACUNIT))
				{
					return false;
				}
			}

			if (!checkSplash(dpm.x, dpm.y, dpm.z, dpm.contacted))
				if (!(li->tyflags & TYF_SKYWALL))
					P_SpawnPuff(dpm.x, dpm.y, dpm.z, ricoh);
		}
		
		return false;	// stop here.
	}

	// shoot a thing
	if (   R3D_edit
		|| !(th = in->d.thing)				// should never be true
		||   th == shootthing				// can't shoot self
		|| !(th->flags & MF_SHOOTABLE))		// corpse or something
	{
		mobjtarget = NULL;
		return true;
	}

	RDBOOL	isheadshot = enemyheadshotidx
			&& !lop.autoaim
			&& shootthing->dplayer
			&& shootthing->player->readyweapon == wp_pistol;

	// check angles to see if the thing can be aimed at
	lop.dist = FixedMul (lop.attackrange, in->frac);
	
	// check for fist/chainsaw height interaction
	// v220-8 modified
	if (shootthing->player && ma_inf.plr_melee_ht)
	{
		if (   lop.dist > MELEERANGE
			|| th->z + th->height < shootthing->z
			|| shootthing->z + ma_inf.plr_melee_ht < th->z
			|| shootthing->z + ma_inf.plr_melee_ht > th->z + th->height)
		{
			mobjtarget = NULL;
			return true;	// out of reach
		}
	}
	else if (!isheadshot || !la_damage)
	{
		if (FixedDiv(th->z + th->height - lop.shootz , lop.dist) < lop.aimslope)
		{
			mobjtarget = NULL;
			return true;		// shot over the thing
		}

		if (FixedDiv(th->z - lop.shootz, lop.dist) > lop.aimslope)
		{
			mobjtarget = NULL;
			return true;		// shot under the thing
		}
	}

	killthing = false;

	// v220-35 added test for case where no lines have been hit
	if (plane3d.set)
	{
		if (   (shootthing->z < plane3d.z && th->z >= plane3d.z)
			|| (th->z < plane3d.z && shootthing->z >= plane3d.z))
		{
			return false;
		}
	}
	
	if (la_damage)
	{
		fixed_t		frac;

		// hit thing - position a bit closer?
		if (lop.dist > th->radius + shootthing->radius + 10*FRACUNIT)
			frac = in->frac - FixedDiv (10*FRACUNIT, lop.attackrange);
		else
			frac = in->frac;

		dpm.x = trace.x + FixedMul (trace.dx, frac);
		dpm.y = trace.y + FixedMul (trace.dy, frac);
		dpm.z = lop.shootz + FixedMul(isheadshot ? lop.origslope : lop.aimslope, FixedMul(frac, lop.attackrange));

		if (isheadshot && (dpm.z > th->z + th->height || dpm.z < th->z))
		{
			mobjtarget = NULL;
			return true;		// shot over the thing
		}

		// Spawn bullet puffs or blood spots,
		// depending on target type.
		if (th->flags & MF_NOBLOOD)
			P_SpawnPuff (dpm.x, dpm.y, dpm.z, 0); // No ricochet if hitting an enemy
		else
			P_SpawnBlood (dpm.x, dpm.y, dpm.z, la_damage, th->type);

		// v220-33 headshot
		if (isheadshot)
		{
			fixed_t zht = dpm.z - th->z;

			for (int i = 0; i < enemyheadshotidx; i++)
			{
				if (th->type == enemyheadshottype[i][0])
				{
					if (   zht >= enemyheadshottype[i][1]
						&& zht <= enemyheadshottype[i][2])
					{
						killthing = true;
					}
					break;
				}
			}
		}

		P_DamageMobj(th, shootthing, shootthing, la_damage, 26);	// $bullets

		killthing = false;
	}

	return false;	// Stop here
}

//
// P_AimLineAttack
//
fixed_t P_AimLineAttack(mobj_t* t1, angle_t angle, fixed_t	distance, int mask)
{
	fixed_t	dx;
	fixed_t	dy;
	
	angle >>= ANGLETOFINESHIFT;
	shootthing = t1;
	
	dx = (distance>>FRACBITS)*finecosine[angle];
	dy = (distance>>FRACBITS)*finesine[angle];
	
	if (t1->dplayer)
	{
		// The slope is determined by lookdir.
		float x = (float)(t1->dplayer->lookdir * RATIO * PI / 180);
		
		lop.aimslope = lop.origslope = (int)((FRACUNIT * RATIO_FACTOR) * tan(x));
		lop.autoaim = autoAim;
		lop.shootz = R3D_edit ? viewz : viewz - 5*FRACUNIT;		// Eye height
	}
	else	// enemies must autoaim
	{
		lop.aimslope = 0;
		lop.autoaim = true;
		lop.shootz = t1->z + (t1->height>>1) + 8*FRACUNIT;
	}
	
	lop.topslope = 60*FRACUNIT;	// Can't shoot outside view angles.  Doom default was 100
	lop.bottomslope = -lop.topslope;
	
	lop.attackrange = distance;
	mobjtarget = NULL;
	
	if (R3D_edit)
	{
		aim_flags_mask = 0;
		lop.autoaim = false;
		
		P_PathTraverse (t1->x, t1->y, dx, dy, true,
			PT_ADDLINES,
			PTR_AimTraverse );
	}
	else
	{
		// killough 8/2/98: prevent friends from aiming at friends
		aim_flags_mask = mask;
tryagain:;		
		P_PathTraverse (t1->x, t1->y, dx, dy, true,
			PT_ADDLINES|PT_ADDTHINGS,
			PTR_AimTraverse );
		
		if (lop.autoaim)
		{
			lop.autoaim = false;
			
			if (mobjtarget)
			{
				if (!mobjtarget->info->seestate)	// shootable decoration
					goto tryagain;	// v220-35 only hit decoration if aiming at it
				return lop.aimslope;
			}
		}
	}
	
	if (t1->dplayer)
	{
		return lop.autoaim ? lop.aimslope : lop.origslope;
	}
	
	return 0;
}

//
// P_LineAttack
// When damage == 0 it is just a test trace that will leave mobjtarget set.
//
void P_LineAttack(mobj_t* t1, angle_t angle, fixed_t distance, fixed_t slope, int damage)
{
	fixed_t	dx;
	fixed_t	dy;
	
	angle >>= ANGLETOFINESHIFT;
	shootthing = t1;
	la_damage = damage;
	
	dx = (distance >> FRACBITS)*finecosine[angle];
	dy = (distance >> FRACBITS)*finesine[angle];
		
	if (R3D_edit)
	{
		lop.shootz = viewz;
	}
	else if (t1->player)
	{
		lop.autoaim = autoAim && mobjtarget;	// b211-0
		lop.shootz = viewz - 5*FRACUNIT;		// Eye height
	}
	else
	{
		lop.autoaim = true;	// always true for enemies
		lop.shootz = t1->z + (t1->height>>1) + 8*FRACUNIT;
	}
	
	lop.attackrange = distance;
	lop.aimslope  = slope;
	lop.lastx = t1->x;
	lop.lasty = t1->y;
	lop.lastz = lop.shootz;
	
	P_PathTraverse (t1->x, t1->y, dx, dy, true,
		R3D_edit ? PT_ADDLINES : PT_ADDLINES|PT_ADDTHINGS,
		PTR_ShootTraverse );
	
	lop.autoaim = false;
}


//
// USE LINES
//

static 	mobj_t	*usething;
static 	int		allow_specials;

static RDBOOL OcclusionTest(line_t *ld, int side)
{
	// if coincident with a trigger line keep checking
	if (ld->lnflags2 & LNF2_TRIGCOINCIDENT) return true;

	if (   ld->ml_flags & ML_TWOSIDED
		&& !(ld->lnflags2 & LNF2_BLOCKALL)
		&& ld->sidenum[side] >= 0
		// v220-4 any point on the thing must be able to see beyond the line
		// v220-13 make <= (was <)
		&& sides[ld->sidenum[side]].sector->floorheight <= usething->z + usething->height
		&& sides[ld->sidenum[side]].sector->ceilingheight > usething->z)
	{
		if (   MAX_OF(ld->frontsector->floorheight  , ld->backsector->floorheight  )
			<  MIN_OF(ld->frontsector->ceilingheight, ld->backsector->ceilingheight))
		{
			return true;		// if not closed off keep checking
		}
	}

	return false;
}

RDBOOL	PTR_UseTraverse (intercept_t* in)
{
	line_t	*ld = in->d.line;
	int side = P_PointOnLineSide(usething->x, usething->y, ld);

	if (ld->specType >= R3D_40)
	{
		if (!ld->special)
		{
			return !(ld->lnflags2 & LNF2_BLOCKALL);
		}
		
		if (side) return false;

		if (ld->lnflags2 & LNF2_LINESOUND && ld->linesound)
		{
			S_StartSound (ld->linesound, usething);
			make_noway_sound = false;
		}

		return P_UseSpecialLine(usething, ld, 0, 1);
	}

	// Check if blocking only
	if (R3D_maptype && (ld->lnflags2 & LNF2_BLOCKALL || ld->special == 8106))
	{
		if (ld->lnflags2 & LNF2_LINESOUND)
		{
			S_StartSound (ld->linesound ? ld->linesound : sfx_noway, usething);
			make_noway_sound = false;
		}

		TS_SetMessage(NULL, ld->ltext, false);

		if (ld->special == 8089 || ld->special == 8106)
			return false;	// Block or message only
	}
	// v220-3 add ML_BLOCKING type
	else if (ld->ml_flags & ML_BLOCKING && ld->linesound)
	{
		if (   ld->special != 8093
			// type 8093 sound if encountered with either side but
			// if it's the frontside this is dealt with elsewhere
			|| P_CheckOnLineSide(usething->x, usething->y, ld))
		{
			S_StartSound (ld->linesound, usething);
		}

		make_noway_sound = false;

		TS_SetMessage(NULL, ld->ltext, false);
	}

	if (ld->special)	// i.e. has not been cleared
	{
		make_noway_sound = false;

		if (R3D_ScriptedMap)
		{
			int res = P_UseSpecialLine(usething, ld, side, 2);

			// special scripted type handling
			if (ld->lnflags2 & LNF2_BLOCKALL) return false;

			if (ld->ml_flags & ML_PASSUSE) return true;

			// if player is on line then continue (side indeterminate)
			if (usething->player && PointIsOnLine(usething->x, usething->y, ld, 1))
			{
				return true;
			}

			// if special was a non-blocking gun or walkover type return true
			return (!res && OcclusionTest(ld, side));
		}

		// v220-15 ignore non-use types that are straddled by the player
		if (   usething->player		// ignore monsters
			&& identifySpecial(ld)	// not a use type
			// if line is straddled then in range and not blocking
			&& XYonLine(usething->x, usething->y, ld, usething->radius - FRACUNIT/8) )
		{
			return true;
		}

		if (allow_specials)
		{
			locked_door = false;

			if (P_UseSpecialLine(usething, ld, side, 3))
				//WAS can't use for than one special line in a row
				//jff 3/21/98 NOW multiple use allowed with enabling line flag	
				return (ld->ml_flags & ML_PASSUSE) > 0;

			if (locked_door) return false;

			// v220-7  If allow_specials was set and P_UseSpecialLine failed
			// and usething is the player and the player's centre is on this
			// line then assume the current type is a walkover special
			// Note this replaces walkoverSpecial() settting used in v220-6
			if (   usething->player
				&& !(ld->ml_flags & ML_PASSUSE)
				&& PointIsOnLine(usething->x, usething->y, ld, 1))
			{
				goto occlusion_test;
			}
		}
		
		allow_specials = side || (ld->ml_flags & ML_PASSUSE) > 0;

		// continue even if P_UseSpecialLine returns false
	}

occlusion_test:;

	// Occlusion test
	if (allow_specials && OcclusionTest(ld, side)) return true;

	// bail out - no further checking required
	if (!ld->linesound && make_noway_sound) // Was S_StartSound (sfx_noway, usething);
		SetSFXnoway(usething, 1);	// Now queued
	
	return false;
}


RDBOOL PTR_NoWayTraverse(intercept_t* in)
{
	line_t *ld = in->d.line;

	return ld->special || !(					// Ignore specials
		ld->ml_flags & ML_BLOCKING || (			// Always blocking
		// Find openings
		P_LineOpening(ld, usething->z + usething->height, usething->x, usething->y),
		openrange <= 0 ||						// No opening
		openbottom > usething->z + MAXSTEP ||	// Too high it blocks
		opentop < usething->z+usething->height	// Too low it blocks
		)
		);
}

//
// P_UseLines
// Looks for special lines in front of the player to activate.
//
void P_UseLines(mbf_player_t*	player) 
{
    int		angle;
    fixed_t	x1;
    fixed_t	y1;
    fixed_t	dx;
    fixed_t	dy;
	
    usething = player->plr->mo;
		
    angle = player->plr->mo->angle >> ANGLETOFINESHIFT;

    x1 = player->plr->mo->x;
    y1 = player->plr->mo->y;
    dx = (USERANGE>>FRACBITS)*finecosine[angle];
    dy = (USERANGE>>FRACBITS)*finesine[angle];
	
	// old code:
	//
	// P_PathTraverse ( x1, y1, x2, y2, PT_ADDLINES, PTR_UseTraverse );
	//

	make_noway_sound = true;
	allow_specials = true;
	
	// This added test makes the "noway" sound work on 2s lines -- killough:
	// R3D b211-0 make_noway_sound test added.  This can be changed by PTR_UseTraverse
	if (P_PathTraverse( x1, y1, dx, dy, true, PT_ADDLINES|PT_3DSWITCH, PTR_UseTraverse)
		&& make_noway_sound)
	{
		if (!P_PathTraverse( x1, y1, dx, dy, true, PT_ADDLINES, PTR_NoWayTraverse ))
			S_StartSound(sfx_noway, usething);
	}
}


//
// RADIUS ATTACK
//

typedef struct {
	mobj_t*		bombsource;
	mobj_t*		bombspot;
	int			bombdamage;
} bs_t;

//
// PIT_RadiusAttack
// "bombsource", if set, is the creature
// that caused the explosion at "bombspot".
//
RDBOOL PIT_RadiusAttack(mobj_t* thing, bs_t *data)
{
	fixed_t	dist, dx, dy;
	
	if (   !(thing->flags & MF_SHOOTABLE)
		&& !(thing->flags3 & MF3_BOUNCES))
		return true;
	
	if (data->bombspot->flags3 & MF3_BOUNCES)
	{
		// killough 8/10/98: allow grenades to hurt anyone, unless
		// fired by Cyberdemons, in which case it won't hurt Cybers.
		if (   data->bombsource	// R3D check it's valid
			&& data->bombsource->type == MT_CYBORG
			&& thing->type == MT_CYBORG)
		{
			return true;
		}
	}
	else if (thing->type == MT_CYBORG || thing->type == MT_SPIDER)
	{
		return true;	// These take no damage
	}
	
	dx = abs(thing->x - data->bombspot->x);
	dy = abs(thing->y - data->bombspot->y);
	dist = dx > dy ? dx : dy;

	// v220-5 Make 3D unless it's the boss brain and the bombspot is a rocket
	if ( !(thing->type == MT_BOSSBRAIN && data->bombspot->type == MT_ROCKET) )
	{
		if (data->bombspot->z < thing->z)
			dist = P_ApproxDistance(dist, thing->z - data->bombspot->z);
		else if (data->bombspot->z > thing->z + thing->height)
			dist = P_ApproxDistance(dist, thing->z + thing->height - data->bombspot->z);
	}
	
	dist = (dist - thing->radius) >> FRACBITS;
	
	if (dist < 0) dist = 0;
	
	if (   dist < data->bombdamage					// must be in range
		&& P_CheckSight(thing, data->bombspot) )	// and must be in direct path
	{
		P_DamageMobj(thing, data->bombspot, data->bombsource, data->bombdamage - dist, 27);	// bomb damage
	}
	
	return true;
}

//
// P_RadiusAttack
// Source, if set, is the creature that caused the explosion at spot.
//
void P_RadiusAttack(mobj_t *spot, mobj_t *source, int damage)
{
	bs_t	data;
	fixed_t dist = (damage+MAXRADIUS)<<FRACBITS;
	int yh = MAPBS3(spot->y, bmaporgy,  dist);
	int yl = MAPBS3(spot->y, bmaporgy, -dist);
	int xh = MAPBS3(spot->x, bmaporgx,  dist);
	int xl = MAPBS3(spot->x, bmaporgx, -dist);
	int x, y;
	
	data.bombspot = spot;
	data.bombdamage = damage;
    data.bombsource = source;
	validcount++;

	for (y = yl ; y <= yh ; y++)
	{
		for (x = xl ; x <= xh ; x++)
		{
			P_BlockThingsIterator(x, y, (int (__cdecl *)(struct mobj_s *,void *))PIT_RadiusAttack, &data);
		}
	}
}


//===========================================================================
// SECTOR HEIGHT CHANGING
//  R3D - modified for 3D lifts plus a few other things
//===========================================================================
// After modifying a sector's floor or ceiling height this
// routine should be called to adjust the positions of all
// things that touch the sector.
//
// If any Thing doesn't fit, true will be returned.
// If crunch is true, a Thing will take damage	whilst being crushed.
// If crunch is false, the sector's height should be reset and
// P_ChangeSector recalled to undo the changes.
//

static RDBOOL	crushchange;
static RDBOOL	nofit;

//
// P_ThingHeightClip
// Takes a valid thing and adjusts the thing->floorz,
// thing->ceilingz, and possibly thing->z.
// This is called for all nearby monsters
// whenever a sector changes height.
// If the thing doesn't fit,
// the z will be set to the lowest value
// and false will be returned.
//

static struct msecv_t {
	sector_t *check_sector;
	sector_t *moving_sector;
} msecv;

static RDBOOL P_ThingHeightClip (mobj_t* thing)
{
	RDBOOL		onfloor = thing->z == thing->floorz;
	RDBOOL		chkfit;
	RDBOOL		pcptest;
	fixed_t		oldfloorz = thing->floorz;

	if (   thing->flags3 & MF3_AMODEL
	//	|| P_IsCamera(thing)
		|| thing->modelboxnum > 0
		|| thing->flags2 & MF2_VDOOR)
	{
		return true;	// These don't require processing
	}
	
	// These are changed in P_MoveSectorThingsIterator
	if (thing->flags & MF_SPAWNCEILING)
	{
		return !crushchange || (thing->ceilingz >> 2) - (thing->floorz >> 2) >= (thing->height >> 2);
	}
	
	pcptest = (thing->flags2 & MF2_SPECNOMOVE) > 0;	// must be 1 if true

	// v220-5 Don't allow check_pickup if spanning sectors with different floordata.
	if (   pcptest
		&& thing->subsector->sector->specFlags & SF_MOVFLOOR
		&& thing->subsector->sector != msecv.check_sector)
	{
		if (   thing->subsector->sector->floorheight == thing->z
			&& !(msecv.check_sector->specFlags & SF_MOVFLOOR
			&& msecv.check_sector->floordata != thing->subsector->sector->floordata))
			return true;

		thing->flags2 &= ~MF2_SPECNOMOVE; pcptest = 0;
	}

	P_CheckPosition(thing, thing->x, thing->y, thing->z, pcptest);

	// v220-31	Gib dead enemies in a height of 8 or less,
	//			NOTE: in addition height cannot be left at 8 as this will not work with 3d levels.
	// v220-32	Remove crushchange check
	if (/*   crushchange : now all moving plats
		&&*/ (thing->flags & (MF_CORPSE|MF_COUNTKILL)) == (MF_CORPSE|MF_COUNTKILL)
		&& thing->state != &states[S_GIBS]
		&& thing->z < gdt.tmceilingz
		&& gdt.tmceilingz - gdt.tmfloorz <= 8*FRACUNIT)
	{
			P_SetMobjState (thing, S_GIBS);
			S_StartSound(sfx_slop, thing);
			return true;
	}
	
	if (gdt.blocked3D)
	{
		if (thing->z <= gdt.tmfloorz) return false;
		if (plat_direction != -1 && !(thing->flags & MF_NOGRAVITY) ) return false;
		if (plat_direction == -1)
		{
			if (gdt.tmceilingz - msecv.moving_sector->planes[0].speed > gdt.tmfloorz + thing->height)
			{
				// v220-31 allow Thing to be pushed down until feet hits the floor
				onfloor = false;
				gdt.blocked3D = false;
				// adjust immediately
				gdt.tmthing->z = gdt.tmceilingz - msecv.moving_sector->planes[0].speed - thing->height;
			}
		}
	}

	thing->floorz   = gdt.tmfloorz;
	thing->ceilingz = gdt.tmceilingz;
	thing->dropoffz = gdt.tmdropoffz;	// killough $dropoff_fix: remember dropoffs

	if (onfloor && msecv.moving_sector->specFlags & SF_MOVFLOOR)
	{
		// v220-7 test that got left out.
		if (  !crushchange					// not crushing
			&& thing->floorz > oldfloorz)	// and rising
		{
			// v220-31 check Things's ceilingz where instantaneous rise is being used.
			fixed_t	cz = (msecv.moving_sector->specFlags & (SF_MOVINVALID|SF_NOLOTEX)) == (SF_MOVINVALID|SF_NOLOTEX)
					&& !msecv.check_sector->listsec ?
						   thing->subsector->sector->ceilingheight : thing->ceilingz;

			if (thing->floorz + thing->height > cz) return false;
		}

		// v220-5 if moving up into a Thing then change its z position
		if (   thing->subsector->sector == msecv.moving_sector
			&& msecv.moving_sector->specFlags & SF_MOVINVALID
			&& abs(thing->floorz - oldfloorz) > 24 * FRACUNIT)	// further test added v220-31
		{
			int i;

			for (i = 0; i < molistnum; i++)
			{
				if (   molist[i]->thinker.function == (void (__cdecl *)(void))P_MobjThinker
					&& sentient(molist[i])	// Also excludes player (player has no seestate)
					&& molist[i]->z < thing->floorz + thing->height)
				{
					molist[i]->z = thing->floorz + thing->height;
				}
			}

			// v220-20 if moving up adjust immediately
			if (   (thing->z < thing->floorz && thing->ceilingz - thing->floorz > thing->height)
				|| (thing->subsector->sector == msecv.check_sector
				&& (!thing->player || thing->z - thing->floorz > 512*FRACUNIT)) )
			{
				thing->z = thing->floorz;	// R3D - Adjust immediately
			}
		}
		else if (thing->ceilingz - thing->floorz > thing->height)
		{
			if (oldfloorz != thing->floorz)
			{
				// v220-8 restrict this test to 3D floors only
				if (   msecv.check_sector->listsec
					&& !checkStepupClear(thing, thing->floorz))
				{
					return false;
				}

				thing->z = thing->floorz;	// R3D - Adjust immediately
				thing->misc |= MJF_ONLIFT;	// needs z-smoothing
			}
			else
			if (msecv.check_sector != msecv.moving_sector	// 3D lift
				&& thing->z + thing->height >= thing->ceilingz)
			{
				return false;
			}
		}

		// killough $dropoff_fix:
		// Possibly upset balance of objects hanging off ledges
		if (thing->flags3 & MF3_FALLING && thing->gear >= MAXGEAR)
			thing->gear = 0;

		// v220-4 update the flame z immediately - don't wait for the thinker to run
		if (thing->corelink) thing->corelink->z = thing->z;
	}
	else if (thing->z > thing->floorz)	// don't adjust a floating monster unless forced to
	{
		if (thing->z + thing->height > thing->ceilingz)
			thing->z = thing->ceilingz - thing->height;
	}

	// v220-10 don't go below the floor
	if (  !thing->info->seestate	// Some have ridiculous radius values
		&& thing->z < thing->subsector->sector->floorheight
		&& thing->subsector->sector == msecv.moving_sector
		&& !msecv.check_sector->listsec)
	{
		thing->z = thing->subsector->sector->floorheight;
	}

	if (R3D_edit || thing->flags3 & MF3_AMODEL) return true;
	
	// Check things stuck in the ceiling
	if (   !gdt.blocked3D
		&& !msecv.check_sector->listsec
		&& plat_direction == -1)
	{
		return true;	// 2D Floors lowering are never stopped
	}

	// b211-0 If the ceiling to floor difference is > 32767 then an overflow
	//        will occur. This is  avoided by using __int64.
	// v220-31 change chkfit from > to >=

	if (gdt.blocked3D)
		chkfit = false;
	// v220-4 Use the sector's ceiling height with floors that jump in
	//        one tic to new height as ceilingz may not be correctly set
	else if (msecv.check_sector->specFlags & SF_MOVINVALID)
		chkfit = (__int64)(thing->subsector->sector->ceilingheight) -
				 (__int64)(thing->floorz) >= (__int64)(thing->height);
	else
		chkfit = (__int64)(thing->ceilingz) - (__int64)(thing->floorz) >= (__int64)(thing->height);

	return chkfit;
}

static mobj_t *blockthing;

// v220-4 handle AMODELS with MF3_CHECKPOS flag
static RDBOOL PIT_BlockCheck(mobj_t* thing, void *data)
{
	fixed_t		blockdist;
	
	if (   thing->flags3 & MF3_AMODEL
		|| !(thing->flags & MF_SOLID)
	//	|| P_IsCamera(thing)
		|| thing->modelboxnum > 0
		|| thing->flags2 & MF2_VDOOR)	// These don't require processing
	{
		return true;	// self or democam
	}

	blockdist = blockthing->radius + thing->radius;

	if (   abs(thing->x - blockthing->x) >= blockdist
		|| abs(thing->y - blockthing->y) >= blockdist )
	{
		return true;	// didn't hit it
	}

	if (   thing->z + thing->height >= blockthing->z
		&& blockthing->z + blockthing->height >= thing->z)
	{
		nofit = true;
	}

	return true;
}


RDBOOL PIT_ChangeSector(mobj_t *thing, sector_t *sec3D)
{
	if (   thing->flags2 & MF2_CORPSEFADE			// Don't bother if faded out
		|| (sec3D && thing->flags3 & MF3_STATIC))	// if a 3D lift and static then it is not on a moving flat
	{
		return true;	// keep checking
	}

	// v220-4 Check for an amodel linked to a plat fouling a Thing.
	if (thing->flags3 & MF3_CHECKPOS)
	{
		int		xl, xh;
		int		yl, yh;
		int		bx, by;
		int		box[4];

		blockthing = thing;

		M_setBoxOffsets(box, thing->x, thing->y, thing->radius);
		M_getBlockOffsets(box, MAXRADIUS, &xl, &xh, &yl, &yh);

		validcount++;
		blockthing->validcount = validcount;	// don't check against self

		// Check all normal things
		for (bx = xl ; bx <= xh ; bx++)
		{
			for (by = yl ; by <= yh ; by++)
			{
				if (!P_BlockThingsIterator(bx, by, (int (__cdecl *)(struct mobj_s *,void *))PIT_BlockCheck, NULL))
				{
					return false;
				}
			}
		}
		return true;
	}

	// Don't check things that aren't blocklinked (supposedly immaterial). 
	if (thing->flags & MF_NOBLOCKMAP || thing->flags2 & MF2_DONTDRAW)
	{
		return true;	// keep checking
	}

	// v220-31 if MF3_TOUCHY is under a 3D lift floor then ignore else the blast will kill Things
	//         above the floor. If someone wants to put an exploding barrel under the floor
	//         then that's their decision. Note; given any enemies will have already been crushed
	//         by virtue of their greater height then doing this only affects enemies near to but
	//         not under the lift. By this point they are pretty much protected (visually) given
	//         there is no space to speak of between the lift floor and ground it is acceptable.
	//         Perhaps it would be more acceptable if the lift got damaged - hmm...
	if (   thing->flags3 & MF3_TOUCHY
		&& thing->subsector->sector->liftsec3D >= 0
		&& thing->ceilingz - thing->floorz <= 8*FRACUNIT)
	{
		S_StartSound(sfx_barexp, thing);
		P_SetMobjState (thing, S_GIBS);
		thing->flags &= ~MF_SOLID;
		thing->flags3 &= ~MF3_TOUCHY;
		thing->height = 0;
		thing->radius = 0;
		return true;
	}

	// v220-33 reduce player corpse height to 0 (needed for heightless detection with 3D floors)
	if (thing->subsector->sector->liftsec3D >= 0 && thing->height && thing->flags & MF_CORPSE) thing->height = 0;
	// Things, corpses, gibs you name it all get their heights set here.
	if (P_ThingHeightClip(thing)) return true;

	// crunch bodies to giblets
	if (thing->health <= 0)
	{
		if ( !(thing->flags2 & MF2_TELEPCORPSE)
			&& thing->state != &states[S_GIBS]
			&& thing->floorz + 8*FRACUNIT > thing->ceilingz)
		{
			thing->height = 0;
			thing->radius = 0;
			thing->flags &= ~MF_SOLID;
			thing->movecount = 0;	// in case respawned
			P_SetMobjState (thing, S_GIBS);
			if (thing->flags & MF_COUNTKILL)
				S_StartSound(sfx_slop, thing);
		}
		
		return true;		// keep checking
	}

	// crunch dropped items
	if (thing->flags & MF_DROPPED)
	{
		P_RemoveMobj(thing, 33);
		return true;		// keep checking	
	}
	
	// killough 11/98: kill touchy things immediately
	if (   thing->flags3 & MF3_TOUCHY
		&& (thing->flags3 & MF3_ARMED || sentient(thing)))
	{
		P_DamageMobj(thing, NULL, NULL, thing->health, 28);  // kill object
		return true;   // keep checking
	}

	if ( !(thing->flags & MF_SHOOTABLE) )
	{
		// assume it is bloody gibs or something
		return true;			
	}

	nofit = thing->type == MT_BARREL || plat_direction != 1;
		
	// v220-35 If it's a barrel don't wait. A fast change
	//         could mean this will stay intact.
	if ( (crushchange || gdt.blocked3D) && (thing->type == MT_BARREL || !(leveltime & 3)) )
	{
		P_DamageMobj(thing, NULL, NULL, 10, 29);

		if (   !(thing->flags2 & MF2_TELEPCORPSE)	// purely a 'nicety'
			&& !(thing->flags & MF_NOBLOOD)
			&& thing->type != MT_SHADOWS)
		{
			// spray blood in a random direction
			mobj_t*	mo = P_SpawnMobj (thing->x, thing->y, 
				thing->z + thing->height/2, MT_BLOOD, 0);
			
			mo->momx = (M_RandomD() - M_RandomD()) << 12;
			mo->momy = (M_RandomD() - M_RandomD()) << 12;

			P_SetMpos(mo);
		}
	}
	
	// keep checking (crush other things)	
    return true;	
}

//===========================================================================
// P_ChangeSector
//===========================================================================
RDBOOL	P_ChangeSector(sector_t *sector, sector_t *sec3D, RDBOOL crunch)
{
	nofit = false;
	crushchange = crunch;

	msecv.moving_sector = sector;
	msecv.check_sector  = sec3D ? sec3D : sector;
	
	// re-check heights for all things near the moving sector
	P_MoveSectorThingsIterator(msecv.check_sector, (int (__cdecl *)(struct mobj_s *,void *))PIT_ChangeSector, sec3D);
	
	return nofit;
}

