//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_maputl.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
// DESCRIPTION:
//	Movement/collision utility functions,
//	as used by function in p_map.c. 
//	BLOCKMAP Iterator functions,
//	and some PIT_* functions to use for iteration.
//
//-----------------------------------------------------------------------------

#include <math.h>

#include "doom/p_local.h"
#include "doom/p_start.h"
#include "doom/tables.h"

#include "r3d_defs.h"
#include "r_main.h"
#include "p_maputil.h"

const double	pidiv = 0xffffffff / (DPI*2);

//
// THING POSITION SETTING
//

//===========================================================================
// P_SetThingPosition
//	Links a thing into both a block and a subsector based on it's x,y.
//	Sets thing->subsector properly.
//===========================================================================
void P_SetThingPosition(mobj_t* thing)
{
	P_LinkThing(thing, 
		(thing->flags & MF_NOSECTOR   ? 0 : DDLINK_SECTOR) |
		(thing->flags & MF_NOBLOCKMAP ? 0 : DDLINK_BLOCKMAP));
}

//
// killough $dropoff_fix
//
// Apply "torque" to objects hanging over ledges, so that they
// fall off. It's not really torque, since Doom has no concept of
// rotation, but it's a convincing effect which avoids anomalies
// such as lifeless objects hanging more than halfway off of ledges,
// and allows objects to roll off of the edges of moving lifts, or
// to slide up and then back down stairs, or to fall into a ditch.
// If more than one linedef is contacted, the effects are cumulative,
// so balancing is possible.
//

static int	origmom;
static int	averageheight;

//===========================================================================
// PIT_ApplyTorque
//===========================================================================
static RDBOOL PIT_ApplyTorque(line_t *ld, mobj_t *mo)
{
	fixed_t		dist;
	int			side = 0;

	if (mo->misc & MJF_ONLIFT)
	{
		if (ld->specType < R3D_40) return false;
		if ( (side = P_CheckOnLineSide(mo->x, mo->y, ld)) != 2)
			return false;
	}
	else
	if ( !(ld->ml_flags & ML_TWOSIDED)
		// R3D eliminate lines with same front and back floor height
		|| ld->frontsector->floorheight == ld->backsector->floorheight	// quo vadis?
		//  v220-17  or where the mo is above those floor heights
		|| mo->z > MAX_OF(ld->frontsector->floorheight, ld->backsector->floorheight)
		|| ld->specType >= R3D_40
		// v220-4 if centered on line then no movement (i.e. == 2)
		|| (side = P_CheckOnLineSide(mo->x, mo->y, ld)) == 2
		|| sides[ld->sidenum[side]].sector->floorheight >= mo->z)
	{
		return !(side == 2 && !origmom);
	}

	dist = (ld->dx >> FRACBITS) * (mo->y >> FRACBITS)
		 - (ld->dy >> FRACBITS) * (mo->x >> FRACBITS)
		 - (ld->dx >> FRACBITS) * (ld->v1->y >> FRACBITS)
		 + (ld->dy >> FRACBITS) * (ld->v1->x >> FRACBITS);	// lever arm

	int ffht = ld->frontsector->floorheight;
	int bfht = ld->specType < R3D_40 ? ld->backsector->floorheight : mo->z;

	if (   (dist < 0 && ffht < mo->z && bfht == mo->z)
		|| (dist > 0 && bfht < mo->z && ffht == mo->z) )
	{		
		// At this point, we know that the object straddles a two-sided
		// linedef, and that the object's center of mass is above-ground.
		fixed_t		x = abs(ld->dx), y = abs(ld->dy);
		angle_t		an;

		// v220-17 Stop sliding where suspended between two lines each with the same
		//         height as the Thing on the side opposite the Thing's centre. This
		//         only takes into account lines with no dx or dy and will fail if a
		//         line is split but it covers most cases whilst keeping it simple..
		if (!x || !y)
		{
			if (side) ffht = ld->specType < R3D_40 ? bfht : mo->z;

			if (mo->z == ffht)
			{
				int ht = ffht >> 16;

				if (ht < 0) ht += 0x10000; else ht -= 0x10000;	// deal with floorheight == 0

				if (averageheight == DDMAXINT)
				{
					if (!x)
					{
						if (ld->v1->x > mo->x) averageheight = ht;
						else
						if (ld->v1->x < mo->x) averageheight = -ht;
					}

					if (!y)
					{
						if (ld->v1->y > mo->y) averageheight = ht;
						else
						if (ld->v1->y < mo->y) averageheight = -ht;
					}
				}
				else
				{
					if (!x)
					{
						if (ld->v1->x > mo->x) averageheight += ht;
						else
						if (ld->v1->x < mo->x) averageheight -= ht;
					}

					if (!y)
					{
						if (ld->v1->y > mo->y) averageheight += ht;
						else
						if (ld->v1->y < mo->y) averageheight -= ht;
					}
				}
			}

			if (!averageheight) return false;
		}
		
		if (y > x) { fixed_t t = x; x = y; y = t; }	// swap x,y

		// v220-5 tantoangle table removed.
		an = (angle_t)(atan2((double)(FixedDiv(y,x) >> 5), (double)SLOPERANGE) * pidiv);
		y = finesine[ ((an + ANG90) >> ANGLETOFINESHIFT) ];
		
		// Momentum is proportional to distance between the
		// object's center of mass and the pivot linedef.
		//
		// It is scaled by 2^(OVERDRIVE - gear). When gear is increased,
		// the momentum gradually decreases to 0 for the same amount of
		// pseudotorque, so that oscillations are prevented, yet it has
		// a chance to reach equilibrium.
		
		dist = FixedDiv(FixedMul(dist, (mo->gear < OVERDRIVE) ?
				y << -(mo->gear - OVERDRIVE) :
				y >> +(mo->gear - OVERDRIVE)), x);
		
		// Apply momentum away from the pivot linedef.
		x = FixedMul(ld->dy, dist);
		y = FixedMul(ld->dx, dist);
		
		// Avoid moving too fast all of a sudden (step into "overdrive")
		dist = FixedMul(x,x) + FixedMul(y,y);
		
		while (dist > FRACUNIT*4 && mo->gear < MAXGEAR)
				++mo->gear, x >>= 1, y >>= 1, dist >>= 1;
		
		mo->momx -= x;
		mo->momy += y;
	}

	return true;
}

//===========================================================================
// P_ApplyTorque
//  killough $dropoff_fix
//  Only called if the object is non-sentient, not a player and at rest.
//  Applies "torque" to the object based on all contacted linedefs.
//  Modified to use Doomsday's line linker code.
//  Revised v220-31 for 3D lifts
//===========================================================================
void P_ApplyTorque(mobj_t *mo)
{
	// Remember the current state, for gear-change
	int isfalling = mo->flags3 & MF3_FALLING;
	
	origmom = mo->momx || mo->momy;
	averageheight = DDMAXINT;

	if (!P_ThingLinesIterator(mo, (int (__cdecl *)(struct line_s *,void *))PIT_ApplyTorque, mo))
	{
		mo->momx = mo->momy = 0;
	}

	// If any momentum, mark object as 'falling' using engine-internal flags
	if ((mo->momx || mo->momy) && mo->subsector->sector->stptr < 0)
	{
		mo->flags3 |= MF3_FALLING;
		isfalling = true;
	}
	else  // Clear the engine-internal flag indicating falling object.
	{
		mo->flags3 &= ~MF3_FALLING;
	}

	// If the object has been moving, step up the gear.
	// This helps reach equilibrium and avoid oscillations.
	//
	// Doom has no concept of potential energy, much less
	// of rotation, so we have to creatively simulate these 
	// systems somehow :)

	if (!isfalling)	// If not falling for a while,
	{
		mo->gear = 0;					// reset it to full strength
	}
	else if (mo->gear < MAXGEAR)		// else, if not at max gear,
	{
		mo->gear++;						// move up a gear
	}
}



