//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc.,
//  Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman,
//  Jaakko Kernen, G Jackson
// Copyright (C) 2006 - 2020 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_pspr.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
// DESCRIPTION:
//	Weapon sprite animation, weapon objects.
//	Action functions for weapons.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <math.h>

#include "doom/d_englsh.h"
#include "doom/p_local.h"
#include "doom/tables.h"

#include "dd_main.h"
#include "AcFnLink.h"
#include "m_misc.h"
#include "s_main.h"
#include "r_things.h"
#include "net_main.h"
#include "p_maputil.h"
#include "con_console.h"
#include "r3d_compat.h"
#include "m_bams.h"

// MACROS ------------------------------------------------------------------

#define LOWERSPEED		FRACUNIT*6
#define RAISESPEED		FRACUNIT*6

#define WEAPONBOTTOM	128*FRACUNIT
#define WEAPONTOP		32*FRACUNIT

// plasma cells for a bfg attack
#define BFGCELLS		40

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	P_SetState(mobj_t *mo, int statenum);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int	customPunchRestState;

// v220-35 smoke stuff
extern int pgsmokestatenum;
extern int pgsmokestatenumX;
extern int cgsmokestatenum;
extern int cgsmokestatenumX;
extern int sgsmokestatenum;
extern int sgsmokestatenumX;
extern int dbsmokestatenum;
extern int dbsmokestatenumX;
extern int gnsmokestatenum;
extern int gnsmokestatenumX;
extern int srsmokestatenum;
extern int srsmokestatenumX;
extern int igsmokestatenum;
extern int igsmokestatenumX;
extern int wnsmokestatenum;
extern int wnsmokestatenumX;
extern int aksmokestatenum;
extern int aksmokestatenumX;

// v220-35 weapon sounds
extern int stsndArray[MAXSTSOUNDS][2];
extern int pistoldamage;
extern int plasmadamage;
extern int bfgdamage;
extern int punchDamage;
extern int bfgsound;
extern int pistolsound;
extern int plasmasound;
extern int punchsound1;
extern int punchsound2;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

fixed_t		bulletslope;
int	customPunch = 0;	// COD support

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int	usegundamage = 0;
static int repeattime = 0;

// CODE --------------------------------------------------------------------

// Sumtin 'ere screws up poor ol' MSVC's head...
#pragma optimize("g",off)

//
// P_BulletSlope
// Sets a slope so that a near miss is at the
// approximate height of the intended target.
//

// Try to find a Thing or Line target
// If not the trace will terminate with a blocking line.
// This is only called by the player
void P_BulletSlope(mobj_t *mo, fixed_t dist)
{
	angle_t	an = mo->angle;
	int		mask = 0;

	// killough 8/2/98: make autoaiming prefer enemies												
	// v220-16 Don't bother if there are no friends
	if (autoAim && ge_inf.isfriendly && !R3D_edit) mask = MF3_FRIEND;

	while (1)
	{
		bulletslope = P_AimLineAttack(mo, an, dist, mask);
		if (mobjtarget || !autoAim) return;
		bulletslope = P_AimLineAttack(mo, an += 1<<26, dist, mask);
		if (mobjtarget) return;
		bulletslope = P_AimLineAttack(mo, an -= 2<<26, dist, mask);
		if (mobjtarget) return;
		if (!mask) { bulletslope = 0; return; }
		mask = 0;
		an = mo->angle;
	}
}

void P_GunShot(mobj_t *mo, RDBOOL accurate)
{
	angle_t	angle = mo->angle;
	int		damage;
	
	damage = 5 * ((M_RandomD() % 3) + 1) + usegundamage;
	
	if (!accurate)
		angle += (angle_t)((M_RandomD() - M_RandomD()) << 18);
	
	P_LineAttack (mo, angle, useBulletRange, bulletslope, damage);
}

void P_ShotAmmo(mbf_player_t *player)
{
	weaponinfo_t *win = weaponinfo + player->readyweapon;

	// If the weapon takes no ammo, do nothing.
	if(win->ammo == am_noammo) return;
	if (hasDEH && !win->pershot) win->pershot = 1;	// v220-35
	player->ammo[win->ammo] -= win->pershot;
	// Don't let it fall below zero.
	if(player->ammo[win->ammo] < 0) player->ammo[win->ammo] = 0;
}

//
// P_BringUpWeapon
// Starts bringing the pending weapon up from the bottom of the screen.
//
void P_BringUpWeapon(mbf_player_t *player)
{
    statenum_t	newstate;
	
    if (player->pendingweapon == wp_nochange)
		player->pendingweapon = player->readyweapon;
	
    if (player->pendingweapon == wp_chainsaw)
		S_StartSound(sfx_sawup, player->plr->mo);
	
    newstate = (statenum_t)weaponinfo[player->pendingweapon].upstate;
	
    player->pendingweapon = wp_nochange;
    player->psprites[ps_weapon].sy = WEAPONBOTTOM;
	
    P_SetPsprite (player, ps_weapon, newstate);
}

//
// P_CheckAmmo
// Returns true if there is enough ammo to shoot.
// If not, selects the next weapon to use.
//
RDBOOL P_CheckAmmo(mbf_player_t *player)
{
    ammotype_t	ammo  = weaponinfo[player->readyweapon].ammo;
    int			count = weaponinfo[player->readyweapon].pershot;
	
    // Some do not need ammunition anyway.
    // Return if current ammunition sufficient.
    if (ammo == am_noammo || player->ammo[ammo] >= count)
		return true;
	
    // Out of ammo, pick a weapon to change to.
    // Preferences are set here.
    do
    {
		if (player->weaponowned[wp_plasma]
			&& player->ammo[am_cell] >= weaponinfo[wp_plasma].pershot
			&& (gamemode != shareware) )
		{
			player->pendingweapon = wp_plasma;
		}
		else if (commercial_game
			&& player->weaponowned[wp_supershotgun] 
			&& player->ammo[am_shell] > weaponinfo[wp_supershotgun].pershot)
		{
			player->pendingweapon = wp_supershotgun;
		}
		else if (player->weaponowned[wp_chaingun]
			&& player->ammo[am_clip] >= weaponinfo[wp_chaingun].pershot)
		{
			player->pendingweapon = wp_chaingun;
		}
		else if (player->weaponowned[wp_shotgun]
			&& player->ammo[am_shell] >= weaponinfo[wp_shotgun].pershot)
		{
			player->pendingweapon = wp_shotgun;
		}
		else if (player->ammo[am_clip] >= weaponinfo[wp_pistol].pershot)
		{
			player->pendingweapon = wp_pistol;
		}
		else if (player->weaponowned[wp_chainsaw])
		{
			player->pendingweapon = wp_chainsaw;
		}
		else if (player->weaponowned[wp_missile]
			&& player->ammo[am_misl] >= weaponinfo[wp_missile].pershot)
		{
			player->pendingweapon = wp_missile;
		}
		else if (player->weaponowned[wp_bfg]
			&& player->ammo[am_cell] > weaponinfo[wp_bfg].pershot
			&& (gamemode != shareware) )
		{
			player->pendingweapon = wp_bfg;
		}
		else
		{
			// If everything fails.
			player->pendingweapon = wp_fist;
		}
    } while (player->pendingweapon == wp_nochange);
	
    // Now set appropriate weapon overlay.
    P_SetPsprite (player, ps_weapon,
		(statenum_t)weaponinfo[player->readyweapon].downstate);
	
    return false;	
}

void P_FireWeapon(mbf_player_t *player)
{
	statenum_t	newstate;
	
	if (!P_CheckAmmo (player))
		return;
	
	P_SetMobjState (player->plr->mo, S_PLAY_ATK1);
	newstate = (statenum_t)weaponinfo[player->readyweapon].atkstate;
	P_SetPsprite (player, ps_weapon, newstate);
	P_NoiseAlert (player->plr->mo);
}

// Player died, so put the weapon away.
void P_DropWeapon(mbf_player_t *player)
{
	P_SetPsprite (player,
		ps_weapon,
		(statenum_t)weaponinfo[player->readyweapon].downstate);
}

static void playWeaponReadySound(int weaponid)
{
	int sndid = 0;

	switch (weaponid)
	{
	case wp_pistol:
	case wp_missile:
		sndid = r3ds.wprdy1_id;
		break;
	case wp_supershotgun:
	case wp_bfg:
		sndid = r3ds.wprdy2_id;
		break;
	case wp_shotgun:
	case wp_plasma:
	case wp_chaingun:
		sndid = r3ds.wprdy3_id;
		break;
	}

	if (sndid && cfg.weaponsounds && BDMFX)
		S_LocalSoundAtVolume(sndid, NULL, 1);
}

void P_SetPsprite(mbf_player_t *player, int position, statenum_t stnum) 
{
    pspdef_t	*psp = &player->psprites[position];
    state_t		*state;

    do
    {
		if (!stnum)
		{
			// object removed itself
			psp->state = NULL;
			break;	
		}
		
		// COD support
		if (   stnum == S_PUNCH
			&& customPunch > 0
			&& customPunchRestState >= 0)
		{
			stnum = (statenum_t)customPunchRestState;
		}
		
		state = &states[stnum];
		psp->state = state;
		psp->tics = state->tics;	// could be 0
		
		// Call action routine.
		// Modified handling.
		// v220-13 more rigorous function handling used
		if (state->action_num)	// NULL is ignored
		{
			if (state->action_num <= nmWeapSound)
			{
				actionlinks[state->action_num].plfn(player, psp);
			}
			else
			{
				if (!(state->stflags & ST_NOTPSPR))
				{
					state->stflags |= ST_NOTPSPR;
					pr_OFinsert("DEH error: invalid player action %s ignored\n", actionlinks[state->action_num].name);
				}
				break;
			}

			if (!psp->state) break;
		}
		
		stnum = (statenum_t)psp->state->nextstate;
		
    } while (!psp->tics);	// an initial state of 0 could cycle through
}

// v220-35 mod for model chaingun
void P_SetMsprite(mbf_player_t *player, int position, statenum_t stnum, int newstnum) 
{
    pspdef_t	*psp = &player->psprites[position];
    state_t		*state;
    state_t		*nstate = &states[newstnum];

    do
    {
		if (!stnum)
		{
			// object removed itself
			psp->state = NULL;
			break;	
		}
		
		state = &states[stnum];
		psp->state = state;
		psp->tics = state->tics;	// could be 0
		
		// Call action routine.
		// Modified handling.
		// v220-13 more rigorous function handling used
		if (nstate->action_num)
		{
			if (state->action_num <= nmWeapSound)
			{
				actionlinks[nstate->action_num].plfn(player, psp);
			}

			if (!psp->state) break;
		}
		
		stnum = (statenum_t)psp->state->nextstate;
		
    } while (!psp->tics);	// an initial state of 0 could cycle through
}

//
// A_WeaponReady
// The player can fire the weapon
// or change to another weapon at this time.
// Follows after getting weapon up,
// or after previous attack/fire sequence.
//
void __cdecl A_WeaponReady (mbf_player_t *player, pspdef_t *psp)
{
	statenum_t	newstate;
	int			a, b;
	
	// get out of attack state
	if (   player->plr->mo->state == &states[S_PLAY_ATK1]
		|| player->plr->mo->state == &states[S_PLAY_ATK2] )
	{
		P_SetMobjState (player->plr->mo, S_PLAY);
	}

	if (   player->readyweapon == wp_chainsaw
		&& psp->state == &states[S_SAW])
	{
		S_StartSound(sfx_sawidl, player->plr->mo);
	}
	
	// check for change
	//  if player is dead, put the weapon away
	if (player->pendingweapon != wp_nochange || player->playerstate == PST_DEAD)
	{
		// change weapon
		//  (pending weapon should already be validated)
		newstate = (statenum_t)weaponinfo[player->readyweapon].downstate;
		playWeaponReadySound(player->readyweapon);
		P_SetPsprite (player, ps_weapon, newstate);
		xydelay = 20;
		return;	
	}
	
	// check for fire
	//  the missile launcher and bfg do not auto fire
	if (player->cmd.buttons & BT_ATTACK)
	{
		if (   !player->attackdown
			|| (player->readyweapon != wp_missile
			&& (player->readyweapon != wp_bfg || bfgdamage)) )
		{
			player->attackdown = true;
			P_FireWeapon (player);		
			return;
		}
	}
	else
		player->attackdown = false;

	a = FixedMul((int)(FRACUNIT*cfg.bobWeapon), mbf_players[0].bob);
	b = (128 * leveltime) & FINEMASK;
	
	// Set the player movement x and y swing
	psp->sx =    FRACUNIT + FixedMul(a, finecosine[b]);
	psp->sy = 32*FRACUNIT + FixedMul(a, finesine[b & (FINEANGLES/2-1)]);
}

// Player can re-fire without lowering weapon entirely.
void __cdecl A_ReFire(mbf_player_t *player, pspdef_t *psp)
{
	// check for fire
	//  (if a weaponchange is pending, let it go through instead)
	if (  (player->cmd.buttons & BT_ATTACK) 
		&& player->pendingweapon == wp_nochange
		&& plhealth.health)
	{
		player->refire++;
		P_FireWeapon (player);
	}
	else
	{
		player->refire = 0;
		P_CheckAmmo (player);
	}
}

void __cdecl A_CheckReload(mbf_player_t *player, pspdef_t *psp)
{
    P_CheckAmmo (player);
}

// Lowers current weapon and changes weapon at bottom.
void __cdecl A_Lower(mbf_player_t *player, pspdef_t *psp)
{	
	psp->sy += LOWERSPEED;
	
	// Is already down.
	if (psp->sy < WEAPONBOTTOM ) return;
	
	// Player is dead.
	if (player->playerstate == PST_DEAD)
	{
		psp->sy = WEAPONBOTTOM;
		return;	// don't bring weapon back up
	}
	
	// The old weapon has been lowered off the screen,
	// so change the weapon and start raising it
	player->readyweapon = player->pendingweapon; 
	
	P_BringUpWeapon (player);
}

void __cdecl A_Raise(mbf_player_t *player, pspdef_t *psp)
{
    statenum_t	newstate;
	
    psp->sy -= RAISESPEED;
	
    if (psp->sy > WEAPONTOP )
		return;
    
    psp->sy = WEAPONTOP;
    
    // The weapon has been raised all the way,
    //  so change to the ready state.
    newstate = (statenum_t)weaponinfo[player->readyweapon].readystate;
	playWeaponReadySound(player->readyweapon);
	
	player->plr->mo->thinker.function = (void (__cdecl *)(void))P_MobjThinker;	// Change ref $winmem
    P_SetPsprite (player, ps_weapon, newstate);
}

void __cdecl A_GunFlash(mbf_player_t *player, pspdef_t *psp) 
{
    P_SetMobjState (player->plr->mo, S_PLAY_ATK2);
    P_SetPsprite (player,ps_flash,(statenum_t)weaponinfo[player->readyweapon].flashstate);
}


//========== WEAPON ATTACKS ============

void __cdecl A_Punch(mbf_player_t *player, pspdef_t *psp) 
{
    angle_t	angle;
    int		damage;
    int		slope;
	
    if (customPunch > 0)
	{
		damage = ge_inf.customdamage;
	}
	else
	{
		damage = ((M_RandomD() % 10) + 1) << 1;

		if (player->powers[pw_strength] || cfg.berserkBug || punchDamage)
			damage *= 10;
	}
	
    angle = player->plr->mo->angle;
    angle += (M_RandomD() - M_RandomD()) << 18;

	slope = P_AimLineAttack(player->plr->mo, angle, MELEERANGE+1, MF3_FRIEND);

	if (!mobjtarget)
		slope = P_AimLineAttack (player->plr->mo, angle, MELEERANGE+1, 0);

    P_LineAttack (player->plr->mo, angle, MELEERANGE+1, slope, damage);

    // turn to face target
    if (mobjtarget)
    {
		// COD support
		if (customPunch > 0)
		{
			if (--customPunch == 5)
			{
				scpMessage.scriptMessage = CUSTOMSPECIAL2;
				scpMessage.delay_to_next_text = 0;
				scpMessage.delay_to_fade = 100;
				scpMessage.alpha = 1.0f;
			}
		}

		S_LocalSoundAtVolume(punchsound1 ? punchsound1 : r3ds.r3dp1_id, NULL, 1);

		player->plr->mo->angle = R_PointToAngle2 (player->plr->mo->x,
			player->plr->mo->y,
			mobjtarget->x,
			mobjtarget->y);

		player->plr->flags |= DDPF_FIXANGLES;
    }
	else
		S_LocalSoundAtVolume(punchsound2 ? punchsound2 : r3ds.r3dp2_id, NULL, 1);	// air swing
}

void __cdecl A_Saw(mbf_player_t *player, pspdef_t *psp) 
{
	mobj_t	*mo;
    angle_t	angle;
    int		damage;
    int		slope;
	
	mo = player->plr->mo;

    damage = 2*((M_RandomD() % 10) + 1);
    angle  = mo->angle;
    angle += (M_RandomD() - M_RandomD()) << 18;
    
	slope = P_AimLineAttack (mo, angle, MELEERANGE+1, MF3_FRIEND);

    if (!mobjtarget)
		slope = P_AimLineAttack (mo, angle, MELEERANGE+1, 0);

    P_LineAttack (mo, angle, MELEERANGE+1, slope, damage);
	
    if (!mobjtarget)
    {
		S_StartSound(sfx_sawful, mo);
		return;
    }
	S_StartSound(sfx_sawhit, mo);
	
    // turn to face target
    angle = R_PointToAngle2 (mo->x, mo->y,
			     mobjtarget->x, mobjtarget->y);
    if (angle - mo->angle > ANG180)
    {
		if (angle - mo->angle < -ANG90/20)
			mo->angle = angle + ANG90/21;
		else
			mo->angle -= ANG90/20;
    }
    else
    {
		if (angle - mo->angle > ANG90/20)
			mo->angle = angle - ANG90/21;
		else
			mo->angle += ANG90/20;
    }
    mo->flags |= MF_JUSTATTACKED;
}

// v220-35 weapon smoke ----------------------------------

static mobj_t smoke;
static mobj_t smoke1;

int		W_CheckPWADNumForName(char *name);

static int defaultRadius = 20*FRACUNIT;

// yoffset of 0 is used with all standard models
typedef struct
{
	int		nmyoffset;
	int		fpyoffset;
	int		apyoffset;
	int		nmzoffset;
	int		fpzoffset;
	int		apzoffset;

} smokeOffset_t;

typedef enum
{
	wpt_pistol,
	wpt_shotgun,
	wpt_dbshotgun,
	wpt_IngramR,		// shared with the chaingun
	wpt_IngramL,
	wpt_gnlauncher,
	wpt_winchester,
	wpt_ak47,
	wpt_srmgun,
	WPT_NUM

} weapontype_e;

static smokeOffset_t smokeOffset[WPT_NUM];

void LoadSmokeOffsets()
{
	char	buf[512];
	int		lump, found = false;
		
	if (!smoke.info ) smoke.info  = (mobjinfo_t*)W_Calloc(sizeof(mobjinfo_t), STATICID, 0);
	if (!smoke1.info) smoke1.info = (mobjinfo_t*)W_Calloc(sizeof(mobjinfo_t), STATICID, 0);

	if (ge_inf.hasscp_defn && (lump = W_CheckPWADNumForName("R3D_WPSO")) >= 0)
	{
		SR_LoadLump(lump, "R3D_WPSO");
		found = true;
	}
	else
	{
		sprintf(buf, "%sR3D_WPSO", ddBinDir.path);
		found = SR_LoadFile(buf, "R3D_WPSO", NULL);
	}

	if (found)
	{
		char	*nextstr;
		int		err;

		if ( !(nextstr = SR_NextSTR(false)) ) goto entryerror;
		defaultRadius = SR_NumForString(&err); if (err) goto entryerror;
		defaultRadius *= FRACUNIT;

		for (int i = 0; i < WPT_NUM; i++)
		{
			if ( !(nextstr = SR_NextSTR(false)) ) break;
			smokeOffset[i].nmyoffset = SR_NumForString(&err); if (err) goto entryerror;
			if ( !(nextstr = SR_NextSTR(false)) ) break;
			smokeOffset[i].fpyoffset = SR_NumForString(&err); if (err) goto entryerror;
			if ( !(nextstr = SR_NextSTR(false)) ) break;
			smokeOffset[i].apyoffset = SR_NumForString(&err); if (err) goto entryerror;
			if ( !(nextstr = SR_NextSTR(false)) ) break;
			smokeOffset[i].nmzoffset = SR_NumForString(&err); if (err) goto entryerror;
			if ( !(nextstr = SR_NextSTR(false)) ) break;
			smokeOffset[i].fpzoffset = SR_NumForString(&err); if (err) goto entryerror;
			if ( !(nextstr = SR_NextSTR(false)) ) break;
			smokeOffset[i].apzoffset = SR_NumForString(&err); if (err) goto entryerror;
		}
		SR_EndScript();

		return;
	}
	else
	{
		Con_Error("ERROR: R3D_WPSO file not found.\nTo fix re-install v220-35");
		return;
	}
entryerror:;
	SR_SCPErr("invalid entry");
}

RDBOOL P_SetSmokeState(mobj_t *mobj, int state);

// Calculate polar co-ords for smoke gen point on an imaginery sphere
// with a fixed radius. The two angles required are those of the player.
// y_rotate is added to the player's angle for alignment purposes.
// z_offset is the player's viewheight plus any offset required.

static void P_ProjectSmoke3D(mobj_t *mo, angle_t angle, float lookdir, 
								angle_t y_rotate, fixed_t z_offset)
{
	angle_t pitch = (angle_t)((lookdir * RATIO)/360 * ANGMAX);
	int x, y, z;

	angle += y_rotate;					// adjust y rotation
	angle >>= ANGLETOFINESHIFT;			// angle in the x,y plane
	pitch >>= ANGLETOFINESHIFT;			// angle in the z plane

	// Find x,y intersect on sphere's circumference
	x = FixedMul(defaultRadius, finecosine[angle]);
	y = FixedMul(defaultRadius, finesine[angle]);

	// Find x,y,z intersect on sphere with the player's line of sight
	x = FixedMul(x, finecosine[pitch]);
	y = FixedMul(y, finecosine[pitch]);
	z = FixedMul(defaultRadius, finesine[pitch]);

	// The x y z point is on a sphere with central co-ords of 0 0 0
	// so these can now just be added to the smoke's actual position.
	mo->x += x;
	mo->y += y;
	mo->z += (z + z_offset);	// add z_offset
}

static void P_WeaponSmoke(mbf_player_t *player, int stnum, int stnumx, weapontype_e type)
{
	int addz = 0;

	if (abs(player->plr->mo->momx) > 4*FRACUNIT || abs(player->plr->mo->momy) > 4*FRACUNIT)
		return;

	if (r_use_particles && cfg.weapsmoke && stnum && stnumx && !underwater)
	{
		angle_t y_offset;
		int		z_offset;
		mobj_t	*sm = player->plr->mo;

		if (ge_inf.switch_flags & SWF_FP)
		{
			y_offset = smokeOffset[type].fpyoffset;
			z_offset = smokeOffset[type].fpzoffset;
		}
		else if (ge_inf.switch_flags & SWF_AP)
		{
			y_offset = smokeOffset[type].apyoffset;
			z_offset = smokeOffset[type].apzoffset;
		}
		else
		{
			y_offset = smokeOffset[type].nmyoffset;
			z_offset = smokeOffset[type].nmzoffset;
		}

		if (type != wpt_IngramL)
		{
			smoke.momx = sm->momx;
			smoke.momy = sm->momy;
			smoke.momz = sm->momz;
			smoke.x = sm->x;
			smoke.y = sm->y;
			smoke.z = sm->z + player->plr->viewheight;
			smoke.thinker = sm->thinker;

			P_ProjectSmoke3D(&smoke, sm->angle, player->plr->lookdir, y_offset * 11930464, z_offset<<FRACBITS);

			smoke.info->spawnstate = stnum;
			smoke.info->deathstate = stnumx;

			//	initialise the generator
			P_SetState(&smoke, stnumx);
			// run the states
			P_SetSmokeState (&smoke, stnum);
		}
		else
		{
			smoke1.momx = sm->momx;
			smoke1.momy = sm->momy;
			smoke1.momz = sm->momz;
			smoke1.x = sm->x;
			smoke1.y = sm->y;
			smoke1.z = sm->z + player->plr->viewheight;
			smoke1.thinker = sm->thinker;

			P_ProjectSmoke3D(&smoke1, sm->angle, player->plr->lookdir, y_offset * 11930464, z_offset<<FRACBITS);

			smoke1.info->spawnstate = stnum;
			smoke1.info->deathstate = stnumx;

			//	initialise the generator
			P_SetState(&smoke1, stnumx);
			// run the states
			P_SetSmokeState (&smoke1, stnum);
		}
	}
}

// ---------- smoke end

void __cdecl A_FireMissile(mbf_player_t *player, pspdef_t *psp) 
{
	// v220-35 grenade launcher smoke
	if (ge_inf.grenade_id)
		P_WeaponSmoke(player, gnsmokestatenum, gnsmokestatenumX, wpt_gnlauncher);
	P_ShotAmmo(player);
    P_SpawnPlayerMissile (player->plr->mo, MT_ROCKET);
}

void __cdecl A_FireBFG(mbf_player_t *player, pspdef_t *psp) 
{
	P_ShotAmmo(player);

	if (bfgdamage)
	{
		mobj_t	*mo = player->plr->mo;
		
		P_SetMobjState (mo, S_PLAY_ATK2);
		// v220-35 ak47 smoke
		if (ge_inf.switch_flags & SWF_AP)
			P_WeaponSmoke(player, aksmokestatenum, aksmokestatenumX, wpt_ak47);

		if (autoAim)
		{
			P_BulletSlope (mo, useBulletRange);
		}
		else
		{
			float x = (float)(player->plr->lookdir * RATIO * PI / 180);

			bulletslope = (fixed_t)((FRACUNIT * RATIO_FACTOR) * tan(x));
		}

		P_LineAttack (mo, player->plr->mo->angle, useBulletRange, bulletslope, bfgdamage);
	}
	else
	    P_SpawnPlayerMissile (player->plr->mo, MT_BFG);
}

void __cdecl A_FirePlasma(mbf_player_t *player, pspdef_t *psp) 
{
	P_ShotAmmo(player);

	if (plasmadamage)
	{
		mobj_t *mo = player->plr->mo;
		
		S_StartSound(plasmasound, mo);
		
		P_SetMobjState (mo, S_PLAY_ATK2);		
		// v220-35 winchester smoke
		if (ge_inf.switch_flags & SWF_AP)
			P_WeaponSmoke(player, wnsmokestatenum, wnsmokestatenumX, wpt_winchester);

		P_SetPsprite (player, ps_flash,
			  (statenum_t)weaponinfo[player->readyweapon].flashstate);
		P_BulletSlope (mo, useBulletRange);
		usegundamage = plasmadamage;
		P_GunShot (mo, (RDBOOL)!player->refire);
		usegundamage = 0;
	}
	else
	{
		P_SetPsprite (player, ps_flash,
			  (statenum_t)(weaponinfo[player->readyweapon].flashstate+(M_RandomD() & 1)) );

		P_SpawnPlayerMissile (player->plr->mo, MT_PLASMA);
	}
}

void __cdecl A_FirePistol(mbf_player_t *player, pspdef_t *psp) 
{
	mobj_t *mo = player->plr->mo;
	
	if (R3D_edit)
	{
		if (Edit_CheckTNUM()) return;
		bulletslope = P_AimLineAttack (mo, mo->angle, STANDARDRANGE, 0);
		P_LineAttack(mo, mo->angle, STANDARDRANGE, bulletslope, 1);
		return;
	}
	
	if (abs(repeattime - leveltime) > 3)
	{
		repeattime = leveltime;
		S_StartSound(pistolsound ? pistolsound : sfx_pistol, mo);
	}
	
	P_SetMobjState (mo, S_PLAY_ATK2);

	// v220-35 gunsmoke
	P_WeaponSmoke(player, pgsmokestatenum, pgsmokestatenumX, wpt_pistol);

	P_ShotAmmo(player);

	if (enemyheadshotidx && !autoAim) player->refire = 0;	// $changex

	P_SetPsprite (player, ps_flash,
		(statenum_t)weaponinfo[player->readyweapon].flashstate);
	
	P_BulletSlope (mo, useBulletRange);
	usegundamage = pistoldamage;
	P_GunShot (mo, (RDBOOL)!player->refire);
	usegundamage = 0;
}

void __cdecl A_FireShotgun(mbf_player_t *player, pspdef_t *psp) 
{
	mobj_t	*mo = player->plr->mo;
    int		i;

    S_StartSound(sfx_shotgn, mo);
	
    P_SetMobjState (mo, S_PLAY_ATK2);

	// v220-35 shotgun smoke
	P_WeaponSmoke(player, sgsmokestatenum, sgsmokestatenumX, wpt_shotgun);

	P_ShotAmmo(player);
    P_SetPsprite (player, ps_flash,
		(statenum_t)weaponinfo[player->readyweapon].flashstate);

    P_BulletSlope (mo, useBulletRange);
	
	// b211-0 replace with; (adds z variance)
	for (i = 0 ; i < 7 ; i++)
	{
		P_LineAttack (mo,
			player->plr->mo->angle + ((M_RandomD() - M_RandomD()) << 18),
			useBulletRange,
			bulletslope + ((M_RandomD() - M_RandomD()) << 3),
			5*((M_RandomD() % 3) + 1));
	}
}

void __cdecl A_FireShotgun2(mbf_player_t *player, pspdef_t *psp) 
{
    int		i;
    angle_t	angle;
    int		damage;
	mobj_t *mo = player->plr->mo;

    S_StartSound(sfx_dshtgn, mo);

    P_SetMobjState (mo, S_PLAY_ATK2);
	
	if (srsmokestatenum)
	{
		// v220-35 shotgun smoke
		P_WeaponSmoke(player, srsmokestatenum, srsmokestatenumX, wpt_srmgun);
		P_ShotAmmo(player);
		P_SetPsprite (player, ps_flash,
			(statenum_t)weaponinfo[player->readyweapon].flashstate);

		P_BulletSlope (mo, useBulletRange);

		// b211-0 replace with; (adds z variance)
		for (i = 0 ; i < 7 ; i++)
		{
			P_LineAttack (mo,
				player->plr->mo->angle + ((M_RandomD() - M_RandomD()) << 18),
				useBulletRange,
				bulletslope + ((M_RandomD() - M_RandomD()) << 3),
				10*((M_RandomD() % 3) + 1));
		}
		return;
	}

	// v220-35 shotgun smoke
	P_WeaponSmoke(player, dbsmokestatenum, dbsmokestatenumX, wpt_dbshotgun);

	P_ShotAmmo(player);
    P_SetPsprite (player, ps_flash,
		(statenum_t)weaponinfo[player->readyweapon].flashstate);
	
    P_BulletSlope (mo, STANDARDRANGE);
	
    for (i=0 ; i<20 ; i++)
    {
		damage = 5*((M_RandomD() % 3) + 1);
		angle = player->plr->mo->angle;
		angle += (M_RandomD() - M_RandomD()) << 19;
		// v220-14 use STANDARDRANGE
		P_LineAttack (mo, angle, STANDARDRANGE, bulletslope + ((M_RandomD() - M_RandomD()) << 5), damage);
    }
}

void __cdecl A_FireCGun(mbf_player_t *player, pspdef_t *psp) 
{
	static int modeltype = 0;
	static int ch1st = 0;
	static int st1 = 0;
	static int st2 = 0;
	static int flip = 0;

	mobj_t *mo = player->plr->mo;

	// -----------------

	if (!ch1st)
	{
		ch1st = Def_GetStateNum("CHAIN1");

		if (ch1st != S_CHAIN1)
		{
			st1 = Def_GetStateNum("CHAINFLASH1");
			st2 = Def_GetStateNum("CHAINFLASH2");	

			modeltype = (st1 != S_CHAINFLASH1 && st2 != S_CHAINFLASH2);

			if (!modeltype && (st1 != S_CHAINFLASH1 || st2 != S_CHAINFLASH2))
				Con_Error("DEF ERROR: chainflash model error");
		}
	}
	
    P_SetMobjState (mo, S_PLAY_ATK2);

	flip = !flip;
	S_StartSound(sfx_chgun, mo);

	// v220-35 chain gun smoke
	if (ge_inf.switch_flags & SWF_AP)
	{
		P_WeaponSmoke(player, igsmokestatenum, igsmokestatenumX, wpt_IngramR);
		P_WeaponSmoke(player, igsmokestatenum, igsmokestatenumX, wpt_IngramL);
	}
	else
		P_WeaponSmoke(player, cgsmokestatenum, cgsmokestatenumX, wpt_IngramR);

	P_ShotAmmo(player);

	if (modeltype)
	{
		statenum_t stnum = (statenum_t)(psp->state - states);

		P_SetMsprite (player, ps_flash,
			(statenum_t)weaponinfo[player->readyweapon].flashstate,
			(stnum == ch1st ? st1 : st2) );
	}
	else
	{
		P_SetPsprite (player, ps_flash,
			(statenum_t)(weaponinfo[player->readyweapon].flashstate
									+ psp->state - &states[S_CHAIN1]) );
	}

    P_BulletSlope(mo, useBulletRange);
	usegundamage = 0;	// in case still set
    P_GunShot(mo, (RDBOOL)!player->refire || enemyheadshotidx);
}

void __cdecl A_BFGsound(mbf_player_t *player, pspdef_t *psp)
{
	S_StartSound(bfgsound ? bfgsound : sfx_bfg, player->plr->mo);
}

void __cdecl A_WeapSound(mbf_player_t *player, pspdef_t *psp)
{
	int i;

	for (i = 0; i < MAXSTSOUNDS; i++)
	{
		if (   stsndArray[i][0] == psp->state - states
			&& stsndArray[i][1])
		{
			S_StartSound(stsndArray[i][1], player->plr->mo);
		}
	}
}

void __cdecl A_Light0(mbf_player_t *player, pspdef_t *psp)
{
	player->plr->extralight = 0;
}

void __cdecl A_Light1 (mbf_player_t *player, pspdef_t *psp)
{
	player->plr->extralight = 1;
}

void __cdecl A_Light2 (mbf_player_t *player, pspdef_t *psp)
{
	player->plr->extralight = 2;
}

void __cdecl A_CloseShotgun2(mbf_player_t *player, pspdef_t *psp )
{
    S_StartSound(sfx_dbcls, player->plr->mo);
    A_ReFire(player,psp);
}

//==================================================

// Called at start of level for each player.
void P_SetupPsprites (mbf_player_t* player) 
{
    int	i;
	
    // remove all psprites
    for (i = 0; i < NUMPSPRITES; i++)
		player->psprites[i].state = NULL;
		
    // spawn the gun
    player->pendingweapon = player->readyweapon;
    P_BringUpWeapon (player);
}

// Called every tic by player think routine.
void P_MovePsprites (mbf_player_t* player) 
{
    pspdef_t	*psp = &player->psprites[0];
    state_t		*state;
    int			i;
	
    for (i = 0; i < NUMPSPRITES; i++, psp++)
    {
		// a null state means not active
		if ( (state = psp->state) )	
		{
			// drop tic count and possibly change state
			if (psp->tics != -1)	// a -1 tic count never changes
			{
				psp->tics--;
				if (!psp->tics)
					P_SetPsprite (player, i, (statenum_t)psp->state->nextstate);
			}				
		}
    }
    
    player->psprites[ps_flash].sx = player->psprites[ps_weapon].sx;
    player->psprites[ps_flash].sy = player->psprites[ps_weapon].sy;
}

#pragma optimize("", on)
