//-----------------------------------------------------------------------------
//
// Copyright (C) 1988 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Rand Phares, Ty Halderman, Jaakko Kernen, Jack Vermeulen
// Copyright (C) 2006 - 2018 G Jackson

//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_setup.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
//  Extended map node support beyond gNd2 is gNd4 specific
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 25 (v220-25) 13/May/2014 gmj
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//  Various revisions plus some functions moved to
//  p_setupNns.cpp as was getting ridiculously bloated.
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
//
// DESCRIPTION:
//	Do all the WAD I/O, get map description,
//	set up initial state and misc. LUTs.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------
#include <io.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <direct.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/timeb.h>

#include "..\bsp\rd_shared.h"
#include "doom/p_local.h"
#include "doom/p_start.h"
#include "doom/g_game.h"
#include "doom/f_infine.h"
#include "doom/hu_stuff.h"
#include "doom/st_stuff.h"
#include "doom/p_setup.h"
#include "doom/am_map.h"
#include "sys_timer.h"
#include "sys_system.h"
#include "p_data.h"
#include "dd_wad.h"
#include "m_misc.h"
#include "m_bams.h"
#include "p_maputil.h"
#include "net_main.h"
#include "s_logic.h"
#include "s_main.h"
#include "r_world.h"
#include "r_things.h"
#include "r_model.h"
#include "con_console.h"
#include "con_bar.h"
#include "gl_tex.h"

#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "r3d_compat.h"
#include "r3d_mapinfo.h"
#include "r3d_modelbox.h"
#include "r3d_nodefile.h"
#include "r3d_setupFns.h"
#include "rend_dyn.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// Map level types. The following data structures define the
// persistent format used in the lumps of the WAD files.

// Ensure alignment is maintained (R3Dglbsp uses same directive).
// In practice it is only mapsubsector4_t that needs this since
// the compiler bases alignment on the last term of the array.

#pragma pack(2)

// v1 segs
typedef struct
{
	short		v1;
	short		v2;
	short		angle;		
	short		linedef;
	short		side;
	short		offset;
} mapseg_t;

// With gl segs angle and offset are not
// sent as these can easily be computed.
// gNd2 segs
typedef struct
{
	ushort	v1, v2;
	ushort	linedef, side;
	ushort	partner;

} glseg_t;

// gNd4 segs
typedef struct
{
	uint	v1, v2;
	ushort	linedef, side;

} glseg4_t;

// gNd5 segs
typedef struct
{
	uint	v1, v2;
	ushort	linedef, side;
	int		partner;

} glseg5_t;

// v1/gNd2 nodes
typedef struct
{
	// Partition line from (x,y) to x+dx,y+dy)
	short	x;
	short	y;
	short	dx;
	short	dy;
	short	bbox[2][4];
	ushort	children[2];

} mapnode_t;

// gNd4/5 nodes
typedef struct
{
	short	x;  // x coordinate
	short	y;  // y coordinate
	short	dx;
	short	dy;
	short	bbox[2][4];
	uint	children[2];

} mapnode4_t;

// gNd2 subsectors
typedef struct
{
	ushort	numSegs;
	ushort	firstseg;	// Index of first seg, segs are stored sequentially
} mapsubsector_t;

// gNd4 subsectors
typedef struct
{
	ushort	numSegs;
	uint	firstseg;
} mapsubsector4_t;

// gNd5 subsectors
typedef struct
{
	int	numSegs;
	int	firstseg;
} mapsubsector5_t;

// sector
typedef	struct
{
	short	floorheight;
	short	ceilingheight;
	char	floorpic[8];
	char	ceilingpic[8];
	short	lightlevel;
	short	special;
	short	tag;
} mapsector_t;

// linedef
typedef struct
{
	ushort	v1, v2;
	short	flags;
	short	special;
	short	tag;
	ushort	sidenum[2];		// sidenum[1] will be 0xFFFF if one sided

} maplinedef_t;

// linedef
typedef struct
{
	ushort	v1, v2;
	short	flags;
	short	special;
	short	tag;
	int		sidenum[2];		// sidenum[1] will be 0xFFFF if one sided

} maplinedefE_t;

#pragma pack()

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	S_InitMUSScript(char *mapid);
void	S_CheckLineTriggeredAmbients(void);
RDBOOL	chkCODambient(int tnum);
void	setCODflag(int id);
void	P_SpawnMapThing (mapthing_t* mthing, int mapnum, int *typecheck, int codflags, RDBOOL is_iwad);
void	R_ResetVisSprites(uint num);
void	S_CalcSectorReverbs(void);
void	Def_SetThingDimensions(void);
void	Rend_LsiArrayInit(void);
void	DR_DrawTextLoading(void);
void	P_SetupWaterSectors(void);
void	R_SetupSky(void);
void	H_SetTHINGsLODstate(void);
void	flushMouse(void);
void	TP_CheckTPSpawnedSectors(RDBOOL removeThing);
void	S_SetLISTge_inf(int listnum, int lumpnum);	// v220-13
void	R3D_SetParms(char *mapid);
void	S_InitAMBIENTS(void);
void	SCP_precacheModels(void);	// v220-31
void	LG_FakeMissing(void);		// v220-31
void	P_InitLineHit(int lcount);	// v220-33
void	WA_CheckAnalyseFile(char *name, int from);
void	printoutAmbients(void);

int		bsp_BuildNodes(int type, int factor, int noprune, char *infile, char *outfile);	// v220-6
int		zdbsp_main (const char *mapname, bool gl, int AA, bool noprune, char *inFname, char *outFname);	// v220-14

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int		wallfornamecount;
extern int		numspechit;
extern int		numsplist;
extern byte		*linkSectors;
extern char		*addflags;
extern int		mapcheat;
extern ushort	rndindex;

// v220-5
extern int		numplayerstarts;

// v220-10
extern float	dehamb_vol, dehamb_attn;	// v220-10 dehacked ambients

// v220-25
extern float	stat_nodestime;

// v220-35
extern int		pu_timer;
extern int		pu_counter;
extern int		wasskyflatnum;
extern float	setskycolour;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

texname3D_t *texname3D;

int		*ssinitial;

RDBOOL	R3D_lightextns;	// v220-31
RDBOOL	R3D_slopetypes;
RDBOOL	R3D_linemap;
RDBOOL	R3D_ScriptedMap;
RDBOOL	R3D_Secret_Pickups;
RDBOOL	R3D_scpzoff;
RDBOOL	R3D_maptype;
RDBOOL	shortListQuads;

int		scriptedMapError;
int		actual_leveltime;
int		leveltime;
int		numthings;
int		corpsetime;

char	*origTEX;
char	new_mapid[8];	// The map being loaded

byte	*spawnArray;
byte	*asyflags;	// Used to identify AASHITTY texture

scpzoff_t	scpzoff[900];
ushort		scplinesound[100][2];
int			fakeidentifier[100];	// v220-3

// v220-6
long		nodesBuildFinish;
int			doom2Map;
int			mapEternal;

// v220-14
char	*show_wadname;

// v220-16
editnode_t edn;

// v220-31
int			pseudofake[100][PF_NUM];
int			elevspeed[100][ELV_NUM];
RDBOOL		elevspeedset;
RDBOOL		elevheightset;
RDBOOL		scpofftagsset;
int			gamenum;	// made global

// v220-31 SectorHook support to trigger sounds
SH_t		*SHarray;
int			SHcontainer[SH_MAXSOUNDS];
int			last_player_sector;

// v220-33
int NFnumlinedefs, NFnumsidedefs, NFnumvertexes, NFnumsectors;

// v220-35
mapsidedef_t **sideptrlist = NULL;
mapsidedef_t *newsidelist = NULL;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

const char rebuild[] = {"\nFIX: restart\n"};

static maplinedefE_t *mldtmp;
static void *linedata;
static void *sidedata;
static byte mapHasRun[100];
static int	map_lumpnum;
static int	preanalysed;
static char	nodefile[512];

// v220-6 Declarations for nodes build stuff -------------------------------

#define LINE_PTR		0
#define SIDE_PTR		1
#define VTX_PTR			2
#define SEG_PTR			3
#define SSEC_PTR		4
#define NODE_PTR		5
#define SECT_PTR		6
#define GLVTX_PTR		7
#define GLSEG_PTR		8
#define GLSSEC_PTR		9
#define GLNODE_PTR		10

const char	*datanames[] = {
	"LINEDEFS",
	"SIDEDEFS",
	"VERTEXES",
	"SEGS",
	"SSECTORS",
	"NODES",
	"SECTORS",
	"GL_VERT",
	"GL_SEGS",
	"GL_SSECT",
	"GL_NODES",
	NULL
};

lump_header_t lhdr[12], rhdr, nhdr;

wad_header_t nbhdr;

typedef struct 
{
	uint	pos;
	uint	length;

} data_hdr_t;

data_hdr_t dhdr[12];

DFILE	*fnode = NULL;

static uint header_offset;
static uint linf_maxsize;
static uint totalnum_segs;


// CODE --------------------------------------------------------------------

static int getLumpByName(char *name)
{
	int i, lim = map_lumpnum + 9;

	for (i = map_lumpnum + 1; i < lim && i < numlumps; i++)
	{
		if (!stricmp(name, lumpinfo[i].name))
		{
			return i;
		}
	}

	return -1;
}

static char *lump_names[] = { NULL, NULL, "LINEDEFS", "SIDEDEFS", "VERTEXES", "SEGS", "SSECTORS", "NODES", "SECTORS", NULL };

// v220-6 write lumps to temporary file for use by nodes build
static void dumpLUMP(DFILE *file, RDBOOL musthave, int offset)
{
	lumpinfo_t	*linf;
	int			lump;

	if (offset == -1)	// write lumps and directory
	{
		byte	*data = (byte*)malloc(linf_maxsize);
		int		i;

		nbhdr.dir_start = header_offset;
		F_FileWrite(&nbhdr, sizeof(wad_header_t), file, 0);

		linf = &lumpinfo[map_lumpnum];
		lseek(linf->handle, linf->position, SEEK_SET);
		read(linf->handle, data, linf->size);
		F_FileWrite(data, linf->size, file, 1);

		for (i = ML_LINEDEFS; i < ML_REJECT; i++)
		{
			lump = getLumpByName(lump_names[i]);

			if (lump >= 0)
			{
				linf = &lumpinfo[lump];

				lseek(linf->handle, linf->position, SEEK_SET);
				read(linf->handle, data, linf->size);
				F_FileWrite(data, linf->size, file, 1);
			}
			else
			{
				data[0] = 0;
				F_FileWrite(data, 1, file, 1);
			}
		}

		for (i = ML_LABEL; i < ML_REJECT; i++)
		{
			if (i != ML_THINGS)
				F_FileWrite(&lhdr[i], sizeof(lump_header_t), file, 2);
		}

		M_Free(data);
	}
	else
	{
		if (lump_names[offset])
			lump = getLumpByName(lump_names[offset]);
		else
			lump = map_lumpnum;

		if (lump >= 0)
		{
			linf = &lumpinfo[lump];

			if (offset == ML_LABEL)
			{
				strncpy(nbhdr.type, "PWAD", 4);
				nbhdr.num_entries = 0;
				header_offset = sizeof(wad_header_t);
				linf_maxsize = 0;
			}

			lhdr[offset].start = header_offset;
			lhdr[offset].length = linf->size;
			memcpy(lhdr[offset].name, linf->name, 8);

			if (linf->size > linf_maxsize) linf_maxsize = linf->size;

			if (offset == ML_LINEDEFS) NFnumlinedefs = lhdr[offset].length/sizeof(maplinedef_t);
			if (offset == ML_SIDEDEFS) NFnumsidedefs = lhdr[offset].length/sizeof(mapsidedef_t);
			if (offset == ML_VERTEXES) NFnumvertexes = lhdr[offset].length/(2 * sizeof(short));
			if (offset == ML_SECTORS ) NFnumsectors = lhdr[offset].length/sizeof(mapsector_t);

			nbhdr.num_entries++;
			header_offset += linf->size;
		}
		else
		{
			char data[1] = { 0 };

			if (musthave) Con_Error("Error: lump %s missing", lump_names[offset]);

			lhdr[offset].start = header_offset;
			lhdr[offset].length = 1;
			memcpy(lhdr[offset].name, lump_names[offset], 8);

			nbhdr.num_entries++;
			header_offset += 1;
		}
	}
}

// Called when loading a game.  Used to compare saved sizes
// with sizes that would be expected when loading the wad - R3D
// v220-7 chkthings added
void setCHECKsize(int *chksecs, int *chklines, int *chkthings, int lump)
{
	char *name[] = {"THINGS", "LINEDEFS", "SECTORS", NULL};
	int i, k, lim;
	if (lump == -2) lump = map_lumpnum;

	lim = lump + 9;

	for (i = 0; name[i]; i++)
	{
		for (k = lump + 1; k < lim && k < numlumps; k++)
		{
			if (!stricmp(name[i], lumpinfo[k].name))
			{
				if (i == 0)
				{
					*chkthings = W_LumpLength(k) / sizeof(mapthing_t);
					goto next_name;
				}

				if (i == 1)
				{
					*chklines  = W_LumpLength(k) / sizeof(maplinedef_t);
					goto next_name;
				}

				if (i == 2)
				{
					*chksecs  = W_LumpLength(k) / sizeof(mapsector_t);
					goto next_name;
				}
			}
		}
next_name:;
	}
}

void delNodesBuild()
{
	if (nodefile[0] && (_access(nodefile, 0)) != -1L)
		remove(nodefile);
}

//-----------------------------------------------------

static fixed_t getOffset(seg_t *sg, line_t *ld, int side)
{
	fixed_t dx, dy;

	if (!ld->v1 || !ld->v2)
	{
		delNodesBuild();
		Con_Error("ERROR: %s has bad errors%s", new_mapid, rebuild);
	}

	dx = side ? sg->v1->x - ld->v2->x : sg->v1->x - ld->v1->x;
	dy = side ? sg->v1->y - ld->v2->y : sg->v1->y - ld->v1->y;

	return (fixed_t)((float)FRACUNIT * M_AccurateDistance(dx, dy));
}

// gets data from internal lumps or external nodes build file
static void *getDATA(data_hdr_t *ptr)
{
	void *data = NULL;

	if (fnode)
	{
		if ( (data = malloc(ptr->length)) == NULL)
			Con_ErrorD("failed to allocate data");

		if (   lseek(fnode->handle, ptr->pos, SEEK_SET) == -1L 
			|| read(fnode->handle, data, ptr->length) != (int)ptr->length)
		{
			Con_ErrorD("failed to read node data");
		}
	}

	return data;
}

void initNodesFile()
{
	if (fnode)
	{
		F_Close(fnode);
		Con_ErrorD("nodes file was open and should be closed");
	}

	if (!nodefile[0] || (_access(nodefile, 4)) == -1)
		Con_Error("nodes file \"%s\" not found", nodefile);

	// edn.data is alloced here
	R3D_ReadNodesFile(nodefile, false);

	if (!edn.data)
		Con_ErrorD("failed to extract nodes file");

	if ( !(fnode = F_OpenData(edn.data, edn.length)) )
		Con_ErrorD("failed to open nodes file");
}

byte *NodesData(char *name, uint *size)
{
	byte	*data = NULL;
	int		sz;
	int		i;

	for (i = 0; datanames[i]; i++)
	{
		if (!strncmp(name, datanames[i], strlen(datanames[i])))
		{
			if ( (sz = dhdr[i].length) )
			{
				*size = (uint)sz;

				if ( (data = (byte*)malloc(sz)) )
				{
					F_Seek(fnode, dhdr[i].pos, SEEK_SET);

					if (F_Read(data, sz, fnode) != sz)
					{
						M_Free(data); data = NULL;
					}
				}
			}
			break;
		}
	}
	return data;
}

// modified v220-27
void updateNodesBuild(byte *sidedata, uint timewrite)
{
	if (!fnode) return;	// Should have been set by a previous call to initNodesFile()

	remove(nodefile);	// Always written if R3D_edit irrespective of stat_nodestime

	if (   stat_nodestime > 1.0f
		&& timewrite	// if 0 then mapwad was not found which should not happen
		&& edn.length	// further safety checks
		&& edn.data)
	{
		// Update the side data
		memcpy(edn.data + dhdr[SIDE_PTR].pos, sidedata, dhdr[SIDE_PTR].length);

		edn.edit = R3D_edit;
		edn.filetime = timewrite;	// Set to wad's new filetime
		
		if (R3D_WriteNodesFile(nodefile))
			pr_OFinsert("Failed to update nodes file\n   %s\n", nodefile);
	}

	// Do this after writing nodefile as closing reclaims edn.data
	F_Close(fnode); fnode = NULL; edn.data = NULL;
}

//==========================================================================
// P_LoadVertexes - Modifed for R3D
//==========================================================================
static void P_LoadVertexes(data_hdr_t *normal, data_hdr_t *gl)
{
	void	*data;
	short	*ml;
	int		glverts = 0;
	int		i;
	
	if ( !(numvertexes = normal->length / (2 * sizeof(short))) )
		Con_ErrorD("Map has no vertexes!");

	// set defaults for normal nodes
	glverts = 0;

	// If a gllump is sent check for gl nodes and type
	if (fnode)
	{
		char id[4];

		glverts = (gl->length - 4) / (2*sizeof(fixed_t));

		if (glverts > 0 )
		{
			lseek(fnode->handle, gl->pos, SEEK_SET);
			read(fnode->handle, id, 4);

			// Which version?  Not interested in v1 or v3
			if (!strncmp(id, "gNd5", 4))
			{   
				ma_inf.nodesFormat = 5;
			}
			else if (!strncmp(id, "gNd4", 4))
			{   
				ma_inf.nodesFormat = 4;
			}
			else if (!strncmp(id, "gNd2", 4) )
			{
				ma_inf.nodesFormat = 2;
			}

			if (!ma_inf.nodesFormat)
			{
				delNodesBuild();
				Con_Error("ERROR: %s. No GL verts v2, v4 or v5 found%s", new_mapid, rebuild);
			}
		}
		else
			ma_inf.nodesFormat = 2;	// Default when there is a gl lump but no nodes
	}

	// load the normal verts
	vertexes = (vertex_t*)W_Calloc((numvertexes + glverts)*sizeof(vertex_t), LEVELID, 0);

	data = getDATA(normal);

	ml = (short*)data;
	
	for(i = 0; i < numvertexes; i++)
	{
		vertexes[i].x = ((short)*ml++) << FRACBITS;
		vertexes[i].y = ((short)*ml++) << FRACBITS;
	}

	M_Free(data);
	
	if (glverts)
	{
		// There are additional vertices in gllump.
		fixed_t *glv;

		data = getDATA(gl);
		GLVertStart = (uint)numvertexes;
		numvertexes += glverts;

		// Load the glverts (skip the header)
		glv = (fixed_t*) ((byte*)data + 4);
		
		for (; i < numvertexes; i++)
		{
			vertexes[i].x = *glv++;
			vertexes[i].y = *glv++;
		}
		
		M_Free(data);
	}
}

//==========================================================================
// P_LoadSegsGL2
//==========================================================================
static void P_LoadSegsGL2(data_hdr_t *gl)
{
	glseg_t		*gls;
	seg_t		*sg;
	void		*data;
	int			i;

	if ( !(numsegs = gl->length / sizeof(glseg_t)) )
	{
		delNodesBuild();
		Con_Error("ERROR: %s has no GL_SEGS!%s", new_mapid, rebuild);
	}

	sg = segs = (seg_t*)W_Calloc(numsegs * sizeof(seg_t), LEVELID, 72);

	data = getDATA(gl);

	gls = (glseg_t*)data;

	for (i = 0; i < numsegs; i++, sg++, gls++)
	{
		ushort lnm = gls->linedef;
		ushort gn1 = gls->v1;
		ushort gn2 = gls->v2;

		uint n1 = gn1 & 0x8000 ? GLVertStart + (gn1 & ~0x8000) : gn1;
		uint n2 = gn2 & 0x8000 ? GLVertStart + (gn2 & ~0x8000) : gn2;

		if (n1 >= (uint)numvertexes || n2 >= (uint)numvertexes)
		{
			delNodesBuild();
			Con_Error("ERROR: %s has bad seg%s", new_mapid, rebuild);
		}

		sg->v1 =  &vertexes[n1];
		sg->v2 =  &vertexes[n2];

		if (lnm == 0xFFFF)
		{
			ma_inf.minisegs++;
		}
		else
		{
			ushort	sid = (gls->side > 0);	// Ensure 0 or 1
			line_t	*ldef = &lines[lnm];
			int		sdef  = ldef->sidenum[sid];

			sg->linedef = ldef;

			if (sdef >= 0 && sdef < numsides)
			{
				sg->sidedef = &sides[sdef];
				sg->frontsector = sg->sidedef->sector;
				sg->offset = getOffset(sg, ldef, sid);
			}

			if (ldef->ml_flags & ML_TWOSIDED
				&& (sdef = ldef->sidenum[sid^1]) >= 0 && sdef < numsides)
			{
				sg->backsector = sides[sdef].sector;
			}
			else
			{
				ldef->ml_flags &= ~ML_TWOSIDED;
				sg->backsector = 0;
			}
		}
	}

	M_Free(data);
}

//==========================================================================
// P_LoadSegsGL4
//==========================================================================
static void P_LoadSegsGL4(data_hdr_t *gl)
{
	glseg4_t	*gls;
	seg_t		*li;
	void		*data;
	uint		n1, n2;
	int			i;

	line_t	*ldef;
	int		lnum;
	int		side;
	int		sdef;

	//--------------------------------------------

	numsegs = gl->length / sizeof(glseg4_t);

	if (!numsegs)
	{
		delNodesBuild();
		Con_Error("ERROR: %s has no GL_SEGS!%s", new_mapid, rebuild);
	}

	li = segs = (seg_t*)W_Calloc(numsegs * sizeof(seg_t), LEVELID, 73);

	data = getDATA(gl);

	gls = (glseg4_t*)data;

	for (i = 0; i < numsegs; i++, li++, gls++)
	{
		n1 = gls->v1 & 0x80000000 ? GLVertStart + (gls->v1 & ~0x80000000) : gls->v1;
		n2 = gls->v2 & 0x80000000 ? GLVertStart + (gls->v2 & ~0x80000000) : gls->v2;

		lnum = (ushort)gls->linedef;
		side = ((ushort)gls->side != 0);

		li->v1 = &vertexes[n1];
		li->v2 = &vertexes[n2];

		if ((ushort)lnum == (ushort)0xFFFF)
		{
			ma_inf.minisegs++;
		}
		else
		{
			ldef = &lines[lnum];
			sdef = ldef->sidenum[side];

			li->linedef = ldef;

			if (sdef >= 0 && sdef < numsides)
			{
				li->sidedef = &sides[sdef];
				li->frontsector = li->sidedef->sector;
				li->offset = getOffset(li, ldef, side);
			}

			if (ldef->ml_flags & ML_TWOSIDED
				&& (sdef = ldef->sidenum[side^1]) >= 0 && sdef < numsides)
			{
				li->backsector = sides[sdef].sector;
			}
			else
			{
				ldef->ml_flags &= ~ML_TWOSIDED;
				segs[i].backsector = 0;
			}
		}
	}

	M_Free(data);
}

//==========================================================================
// P_LoadSegsGL5
//==========================================================================
static void P_LoadSegsGL5(data_hdr_t *gl)
{
	glseg5_t	*gls;
	seg_t		*sg;
	void		*data;
	uint		n1, n2;
	int			i;

	line_t	*ldef;
	int		lnum;
	int		side;
	int		sdef;

	//--------------------------------------------

	numsegs = gl->length / sizeof(glseg5_t);

	if (!numsegs)
	{
		delNodesBuild();
		Con_Error("ERROR: %s has no GL_SEGS!%s", new_mapid, rebuild);
	}

	sg = segs = (seg_t*)W_Calloc(numsegs * sizeof(seg_t), LEVELID, 73);

	data = getDATA(gl);

	gls = (glseg5_t*)data;

	for (i = 0; i < numsegs; i++, sg++, gls++)
	{
		n1 = gls->v1 & 0x80000000 ? GLVertStart + (gls->v1 & ~0x80000000) : gls->v1;
		n2 = gls->v2 & 0x80000000 ? GLVertStart + (gls->v2 & ~0x80000000) : gls->v2;

		lnum = (ushort)gls->linedef;
		side = ((ushort)gls->side != 0);

		sg->v1 = &vertexes[n1];
		sg->v2 = &vertexes[n2];

		if ((ushort)lnum == (ushort)0xFFFF)
		{
			ma_inf.minisegs++;
		}
		else
		{
			ldef = &lines[lnum];
			sdef = ldef->sidenum[side];

			sg->linedef = ldef;

			if (sdef >= 0 && sdef < numsides)
			{
				sg->sidedef = &sides[sdef];
				sg->frontsector = sg->sidedef->sector;
				sg->offset = getOffset(sg, ldef, side);
			}

			if (ldef->ml_flags & ML_TWOSIDED
				&& (sdef = ldef->sidenum[side^1]) >= 0 && sdef < numsides)
			{
				sg->backsector = sides[sdef].sector;
			}
			else
			{
				ldef->ml_flags &= ~ML_TWOSIDED;
				sg->backsector = 0;
			}
		}
	}

	M_Free(data);
}

//==========================================================================
// P_LoadSubsectors - Modifed for R3D
//==========================================================================
static void P_LoadSubsectors(data_hdr_t *ptr)	// could be gl or normal
{
	mapsubsector_t	*ms;
	subsector_t		*ss;
	void			*data;
	int				i;
	
	if (!(numsubsectors = ptr->length / sizeof(mapsubsector_t)))
		Con_ErrorD("ERROR: map has no subsectors!");
	
	ss = subsectors = (subsector_t*)W_Calloc(numsubsectors*sizeof(subsector_t), LEVELID, 74);
	
	data = getDATA(ptr);

	ms   = (mapsubsector_t*)data;
	
	for (i = 0; i < numsubsectors; i++, ss++, ms++)
	{
		ss->linecount = (int)((ushort)ms->numSegs);
		ss->firstline = (int)((ushort)ms->firstseg);
		totalnum_segs += ss->linecount;
	}
	
	M_Free (data);
}


//==========================================================================
// P_LoadSubsectorsGL4
//==========================================================================
static void P_LoadSubsectorsGL4 (data_hdr_t *gl)
{
	mapsubsector4_t	*ms;   // from the level
	subsector_t		*ss;   // internal
	void			*data;
	int				i;
	
	if (!(numsubsectors = gl->length / sizeof(mapsubsector4_t)))
	{
		delNodesBuild();
		Con_Error("ERROR: %s has no subsectors%s", new_mapid, rebuild);
	}

	ss = subsectors = (subsector_t*)W_Calloc(numsubsectors*sizeof(subsector_t), LEVELID, 75);

	data = getDATA(gl);

	ms   = (mapsubsector4_t*)data;
	
	for (i = 0; i < numsubsectors ; i++, ss++, ms++)
	{
		ss->linecount = (int)((ushort)ms->numSegs);
        ss->firstline = (int)ms->firstseg;
		totalnum_segs += ss->linecount;
	}
	
	M_Free (data);
}

//==========================================================================
// P_LoadSubsectorsGL5
//==========================================================================
static void P_LoadSubsectorsGL5 (data_hdr_t *gl)
{
	mapsubsector5_t	*ms;   // from the level
	subsector_t		*ss;   // internal
	void			*data;
	int				i;
	
	if (!(numsubsectors = gl->length / sizeof(mapsubsector5_t)))
	{
		delNodesBuild();
		Con_Error("ERROR: %s has no subsectors%s", new_mapid, rebuild);
	}

	ss = subsectors = (subsector_t*)W_Calloc(numsubsectors*sizeof(subsector_t), LEVELID, 75);

	data = getDATA(gl);

	ms   = (mapsubsector5_t*)data;
	
	for (i = 0; i < numsubsectors ; i++, ss++, ms++)
	{
		ss->linecount = (int)ms->numSegs;
        ss->firstline = (int)ms->firstseg;
		totalnum_segs += ss->linecount;
	}
	
	M_Free (data);
}

//==========================================================================
// P_LoadSectors - Modifed for R3D
//==========================================================================
static void P_LoadSectors (data_hdr_t *normal)
{
	mapsector_t	*ms;
	sector_t	*sec;
	skyline_t	*sk;
	void		*data;
	int			i;
	int			showspec = ArgExists("-secspec");

	if (!(numsectors = normal->length / sizeof(mapsector_t)))
		Con_Error("ERROR: %s has no sectors!", new_mapid);
	
    sec = sectors = (sector_t*)W_Calloc(numsectors*sizeof(sector_t), LEVELID, 76);
	sk = skl = (skyline_t*)W_Calloc(numsectors*sizeof(skyline_t), LEVELID, 76);
	
	data = getDATA(normal);

    ms = (mapsector_t*)data;

	for (i = 0; i < numsectors; i++, sec++, ms++, sk++)
    {
		sk->sky_line = -1;

		sec->floorheight   = ms->floorheight<<FRACBITS;
		sec->ceilingheight = ms->ceilingheight<<FRACBITS;
		sec->smoothfloor   = sec->floorheight;
		sec->smoothceil    = sec->ceilingheight;

		sec->floorpic   = R_FlatNumForName(ms->floorpic);
		sec->ceilingpic = R_FlatNumForName(ms->ceilingpic);
		sec->lightlevel = ms->lightlevel;

		if (sec->lightlevel > 255) sec->lightlevel = 255;

		sec->walllight = sec->lightlevel;

		sec->special   = ms->special;
		sec->tag       = ms->tag;
		sec->thinglist = NULL;

		if (showspec && sec->special) pr_OFinsert("S%i sp %i 0x%4X\n", i, sec->special, sec->special);

		// killough ptrs to other sectors
		sec->heightsec		= -1;	// sector used to get floor and ceiling height
		sec->oversec		= -1;	// sector referencing a movsec that goes under a fake sector
		sec->floorlightsec	= -1;
		sec->ceilinglightsec= -1;
		sec->topmap			=  0;	// killough 4/4/98: colormaps:
		sec->bottommap		=  0;
		sec->midmap			=  0;
		sec->nextsec		= -1;	//jff 2/26/98 add fields to support locking out
		sec->prevsec		= -1;	// stair retriggering until build completes
		sec->sidesky		=  0;	// killough 10/98: sky textures coming from sidedefs
		sec->skyvar			= -1;	// v220-5 sky scroll rotate values index

		// R3D flags/ptrs/values
		sec->specFlags		=  0;
		sec->sector3D		= -1;	// Sector for 3D structures
		sec->R3Dfloorsec	= -1;
		sec->R3Dceilsec		= -1;
		sec->R3Dfloorheight	= DDMAXINT;
		sec->R3Dceilheight	= DDMININT;
		sec->liftsec3D		= -1;	// $smoothplane If influenced by another sector's height
		sec->floodline		= -1;
		sec->skyfix			=  0;
		sec->sky_offy		=  0;
		sec->listsec		= NULL;	// List of linked lift sectors
		sec->stptr			= -1;
		sec->ceiloffset		= DDMAXINT;
		sec->R3Dlowsec		= -1;
		sec->withheight		= DDMAXINT;	// $v16added
		sec->texwithheight[0]	=  0;	// $v16added
		sec->texwithheight[1]	=  0;	// $v16added
 		// v220-33
		sec->scalingValueX	= 1.f;
		sec->scalingValueY	= 1.f;
   }
	
    M_Free(data);
}

static void BADNodes(RDBOOL ovf)
{
	char *type = !ma_inf.nodesFormat ? "NORMAL" :
				  ma_inf.nodesFormat == 2 ? "GL2":
				  ma_inf.nodesFormat == 4 ? "GL4": "GL5";

	delNodesBuild();

	if (ovf)
		Con_Error("ERROR: %s nodes %s\nsubsectors out of range.%s\n",
			new_mapid, type, rebuild);

	Con_Error("ERROR: %s nodes %s\n"
		"Number of subsectors [%i] should equal number of nodes [%i] + 1%s\n",
		new_mapid, type, numsubsectors, numnodes, rebuild);
}

//==========================================================================
// P_LoadNodes
//==========================================================================
static void P_LoadNodes(data_hdr_t *ptr)	// could be normal or gl
{
	mapnode_t	*mn;
	node_t		*no;
	void		*data;
	int			i, j, k;

	numnodes = ptr->length / sizeof(mapnode_t);

	if (numsubsectors != numnodes + 1) BADNodes(false); 

	// Use numsubsectors which is never 0
	no = nodes = (node_t*)W_Calloc(numsubsectors * sizeof(node_t), LEVELID, 77);

	data = getDATA(ptr);

	mn = (mapnode_t*)data;

	for (i = 0 ; i < numnodes; i++, no++, mn++)
	{
		no->x  = (short)(mn->x )<<FRACBITS;
		no->y  = (short)(mn->y )<<FRACBITS;
		no->dx = (short)(mn->dx)<<FRACBITS;
		no->dy = (short)(mn->dy)<<FRACBITS;
		
		for (j = 0; j < 2; j++)
		{
			ushort child = mn->children[j];

			// Load values converting to extended type
			if (child == 0xFFFF)
			{
				no->children[j] = NFx_SUBSECTOR;
			}
			else if (child & 0x8000)
			{
				child &= ~0x8000;
				if (child >= numsubsectors) BADNodes(true);
				no->children[j] = (int)((ushort)child)|NFx_SUBSECTOR;
			}
			else
			{
				if (child >= numnodes) BADNodes(true);
				no->children[j] = (int)((ushort)child);
			}

			for (k = 0; k < 4; k++)
				no->bbox[j][k] = (int)((short)(mn->bbox[j][k])) << FRACBITS;
		}
	}
	M_Free (data);
}

//==========================================================================
// P_LoadNodesGLE
//==========================================================================
static void P_LoadNodesGLE(data_hdr_t *gl)
{
	mapnode4_t	*mn;
	node_t		*no;
	void		*data;
	int			i, j, k;

	numnodes = gl->length / sizeof(mapnode4_t);

	if (numsubsectors != numnodes + 1) BADNodes(false);

	// Use numsubsectors which is never 0
	no = nodes = (node_t*)W_Calloc(numsubsectors * sizeof(node_t), LEVELID, 78);	

	data = getDATA(gl);

	mn = (mapnode4_t*)data;

	for (i = 0; i < numnodes; i++, no++, mn++)
	{
		no->x  = (short)(mn->x )<<FRACBITS;
		no->y  = (short)(mn->y )<<FRACBITS;
		no->dx = (short)(mn->dx)<<FRACBITS;
		no->dy = (short)(mn->dy)<<FRACBITS;
		
		for (j = 0; j < 2; j++)
		{
			no->children[j] = (uint)mn->children[j];

			if (no->children[j] == 0xFFFFFFFF)
			{
				no->children[j] = NFx_SUBSECTOR;
			}
			else if (no->children[j] & NFx_SUBSECTOR)
			{
				if ((no->children[j] & ~NFx_SUBSECTOR) >= (uint)numsubsectors)
					BADNodes(true);
			}
			else
			{
				if (no->children[j] >= numnodes) BADNodes(true);
			}

			for (k = 0; k < 4; k++)
				no->bbox[j][k] = (int)((short)(mn->bbox[j][k])) << FRACBITS;
		}
	}
	M_Free (data);
}

//==================================================
// P_CheckThings15500 - R3D
//
// currently only 15500 to 15509 are used
// initially 15502 was force stretched sky but this is now determined automatically
// 15505 and 15506 are now redundant (were used up to v19 and v5 respectively)
// type 15504 aggregates types 15500 + 15501 + 15503 
//==================================================
static void P_CheckThings15500 (int lump)
{
    mapthing_t	*mt;
    void		*data;
	int			i;

	numthings = W_LumpLength (lump) / sizeof(mapthing_t);

	if (!numthings) Con_ErrorD("Map has no THINGS!");

	waterdynamics = !R3D_edit && BDMFX;

	R3D_lightextns = false;
	R3D_linemap    = false;
	R3D_slopetypes = false;
	R3D_Secret_Pickups = false;

	ge_inf.mapset = 0;
	data = M_CacheLumpNum(lump, NULL, false);
	mt = (mapthing_t*)data;

	for (i = 0 ; i < numthings ; i++, mt++)
	{
		if (mt->doomednum >= 15500 && mt->doomednum < 15550)
		{
			int n = mt->doomednum;

			if (R3D_edit)
			{
				waterdynamics = water_trans = false;
				cfg.fallingDamage = cfg.underwaterDamage = false;
			}

			if (n == 15500 || n == 15501 || n == 15504 || n == 15507)
			{
				if (!R3D_edit)
				{
					waterdynamics = true;
					water_trans = true;
					water_opacity = 170;
					ge_inf.mapset |= GEF_WATERTRANS;
					ma_inf.noscriptset |= NSF_WATERTRANS;
				}
			}

			if (n == 15501 || n == 15504 || n == 15507)
			{
				R3D_slopetypes = true;
				R3D_linemap = true;

				if (!R3D_edit)
				{
					cfg.jumpPower = 9;
					cfg.jumpEnabled = true;
					ge_inf.mapset |= GEF_WATERDYNAMICS;
					ma_inf.noscriptset |= NSF_WATERDYNAMICS;
				}
			}

			if (n == 15503) R3D_slopetypes = true;

			if (n == 15507 && !R3D_edit)
			{
				cfg.fallingDamage = cfg.underwaterDamage = true;
				ge_inf.mapset |= (GEF_FALLINGDAMAGE|GEF_UWDAMAGE);
				ma_inf.noscriptset |= (NSF_FALLINGDAMAGE|NSF_UWATERDAMAGE);
			}

			if (n == 15508 && R3D_ScriptedMap) R3D_Secret_Pickups = true;

			if (n == 15509 && !R3D_edit)
			{
				if (!CFX)
				{
					if (!Sys_CriticalMessage("WARNING: Map calls for lighting effects but CFX is disabled\n"
						"If continuing special lighting effects will also remain disabled", true)) Sys_Quit(false);
				}
				else
				{
					R3D_lightextns = true;
				}
			}
		}
	}

	R3D_maptype = R3D_slopetypes || R3D_ScriptedMap || R3D_linemap;

	if (R3D_maptype && gamemode == shareware)
		Con_ErrorD("Risen3D script/slopes/3Dlines cannot be run with a shareware iwad");

	M_Free(data);
}

static void P_Init_origTEX()
{
	line_t	*ld;
	int		i, numnewsides = 0;

 	origTEX = (char*)W_Calloc(numlines, LEVELID, 84);

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		int s0 = ld->sidenum[0];
		int s1 = ld->sidenum[1];

		origTEX[i] = (char)sides[s0].texFlags;

		// Ignore the ML_TWOSIDED flag
		if (s1 >= 0)
			origTEX[i] |= (sides[s1].texFlags << 4);
	}
}

//==================================================
// P_LoadThings - Modified for R3D
//  b211-0 typecheck added
//==================================================
void P_LoadThings (int lump)
{
	void		*data = W_MallocLump(lump, STATICID, 402);
	mapthing_t	*mt = (mapthing_t*)data;

	int			typecheck;
	int			codna[4] = { 0, 0, 0, 0 };
    int			i, codflags, dmnum;

	//-------------------------------------

	spawnArray = (byte*)W_Calloc(numthings, STATICID, 19);

	memset(fakeidentifier, 0xFF, sizeof(fakeidentifier));	// v220-3
	memset(ltarray, 0xFF, sizeof(linetrig_t) * LT_MAXCOUNT);

	musChange = false;
	dmcount = 0;
	ambientfound = 0;
	memset(&ambientcount, 0, sizeof(int)*64);
		
	if (gamemode == shareware && R3D_ScriptedMap)
		Con_Error("Scripted maps cannot be run with the shareware version\n"
				  "You will need to get the registered version of Doom2");

	if (ge_inf.MBFfriendallow)
	{
		for (i = 0, mt = (mapthing_t*)data; i < numthings; i++, mt++)
		{
			if (mt->options & 0xff00)
			{
				pr_OFinsert("Options have unused bits set. MBF extensions ignored.\n");
				ge_inf.MBFfriendallow = false;
				break;
			}
		}
	}

    for (i = 0, mt = (mapthing_t*)data; i < numthings; i++, mt++)
    {
		dmnum = mt->doomednum;
		codflags = 0;

		// Do not spawn cool, new monsters if shareware
		if (gamemode == shareware)
		{
			if (dmnum >= 5500) continue;

			switch(dmnum)
			{
			  case 68:	// Arachnotron
			  case 64:	// Archvile
			  case 88:	// Boss Brain
			  case 89:	// Boss Shooter
			  case 69:	// Hell Knight
			  case 67:	// Mancubus
			  case 71:	// Pain Elemental
			  case 65:	// Former Human Commando
			  case 66:	// Revenant
			  case 84:	// Wolf SS
				  continue;
			}
		}

		// Unfortunately the link to the COD engine code was broken (Jun 2007)
		// so all of the following has been derived by creating test maps which
		// were then run with the COD DOS engine to try to establish what's what.
		// And if that doesn't deserve a medal then I don't know what does....
		if (ge_inf.addon_flags & AF_COD)		// COD engine support
		{
			// COD ambients
			if (dmnum >= 4988 && dmnum <= 4996)
			{
				if (R3D_edit || !sfx_avail || !chkCODambient(dmnum - 4988))
					continue;
			}

			// these come from DD_DEFNS in an addon.wad
			// 4997 MISC87, 4998 MISC88, 4999 MISC89, 5017 COD_VILE

			if (dmnum >= 5000 && dmnum <= 5044)
			{
				int num = 0, chkcase = dmnum - 5000;

				// 5000 to 5002 are not COD specific
				// 5006 to 5008 are not COD specific
				// 5012, 5013 are not COD specific
				// 5016 is not COD specific
				// 5018 to 5028 are not COD specific

				chkcase = dmnum - 5000;

				switch(chkcase)
				{
				case  0: case 1: case 2: break;
				case  3: codflags = 1; num = 3004; break;	// zombie->lost soul				
				case  4: codflags = 1; num =    9; break;	// shotgun guy->lost soul
				case  5: codflags = 1; num =   65; break;	// chaingun bob->lost soul
				case  6: case 7: case 8: break;
				case  9: codflags = 2; num =    7; break;	// spiderboss (health = 1500)
				case 10: codflags = 6; num = 9056; break;	// Invisible knight
				case 11: codflags = 4; num = 3003; break;	// baron spawned dead comes alive when hit
				case 12: case 13: break;
				case 14: codflags = 8; num =   16; break;	// cyberdemon (health = 4000)
				case 15: codflags = 5; num = 3004; break;	// hardass zombie
				case 16: break;
				case 17: codflags = 7; num =   64; break;	// stealth archvile (defined in codaddon)
				case 18: case 19: case 20: case 21: case 22: case 23:
				case 24: case 25: case 26: case 27: case 28:
					break;
				// types that trigger floor when all killed
				case 29: codflags = 3; num = 3004; break;	// zombie
				case 30: codflags = 3; num =    9; break;	// shotgun guy
				case 31: codflags = 3; num =   58; break;	// spectre
				case 32: codflags = 3; num = 3001; break;	// imp
				case 33: codflags = 3; num = 3002; break;	// demon
				case 34: codflags = 3; num = 3006; break;	// lost soul
				case 35: codflags = 3; num = 3005; break;	// cacodemon
				case 36: codflags = 3; num = 3003; break;	// baron
				case 37: codflags = 3; num =   16; break;	// cyberdemon
				case 38: codflags = 3; num =    7; break;	// spiderboss
				case 39: codflags = 3; num =   65; break;	// chaingun bob
				case 40: codflags = 3; num =   69; break;	// hell knight
				case 41: codflags = 3; num =   68; break;	// arachnotron
				case 42: codflags = 3; num =   71; break;	// pain elemental
				case 43: codflags = 3; num =   66; break;	// revenant
				case 44: codflags = 3; num =   67; break;	// mancubus
				default: break;
				}

				if (num > 0) mt->doomednum = num;
			}
		}

		typecheck = -1;

		// Now spawn what's allowed
		P_SpawnMapThing(mt, i, &typecheck, codflags, lumpinfo[lump].handle == IWADhandle);

		if (!typecheck)
		{
			pr_OFinsert("T%i (doomednum 0 BOOM spare) not spawned\n", i);
		}
		else if (typecheck > 0 && !R3D_edit)
		{
			ma_inf.unspawned++;
			pr_OFinsert("T%i doomednum %i not spawned\n", i, typecheck);
		}
    }

	if (scriptedMapError) Sys_Quit(false);
	if (musChange) pr_OFdirect("Map has MUSCHANGE things.\n");

	W_Free(data, STATICID, 529);
}

static scpzoff_t *getSPOTNUM(char *ins)
{
	int num = SR_NextNum() - 16000;		// SPOT NUM

	// Check number in range
	if (num < 0 || num > 899)
		SR_SCPErr("%s : 16xxx thing out of range 16000 to 16899", ins);

	return &scpzoff[num];
}

// Note: do not put \n's in the SCP_ERR message
// If adding Things 17100 or above ammend range check in P_SpawnScriptThing()
static void Load_SCP_ZOFF(char *mapid)	// R3D
{
	scpzoff_t *sc;
	int		num;
	char	*nextstr;

	memset(scpzoff, 0, sizeof(scpzoff));
	memset(scplinesound, 0, sizeof(scplinesound));
	memset(elevspeed, 0, sizeof(elevspeed));
	memset(pseudofake, 0, sizeof(pseudofake));
	elevspeedset = elevheightset = false;
	scpofftagsset = false;

	R3D_scpzoff = false;

	if ( W_OpenLatestScriptLump("SCP_ZOFF", mapid) )
	{
		RDBOOL	val2, val3, val4, inMap = false;

		R3D_scpzoff = true;

		while( (nextstr = SR_NextSTR(false)) )
		{
			if (inMap || !_stricmp(nextstr, mapid))
			{
				if (!inMap)
				{
					inMap = true;
					nextstr = SR_NextSTR(false);
					if (!nextstr) break;
				}

				val3 = val4 = 0;

				if (   (val2 = !_stricmp(nextstr, "SETLINESOUNDBV"))
					|| (val3 = !_stricmp(nextstr, "SETLINETRIGGERSOUND"))
					|| (val4 = !_stricmp(nextstr, "SETLINESOUNDOFF"))
					|| !_stricmp(nextstr, "SETLINESOUND"))
				{
					num = SR_NextNum() - 16900;		// SCPLINESOUND NUM

					// Check number in range
					if (num < 0 || num > 99)
						SR_SCPErr("16xxx thing out of range 16900 to 16999");

					nextstr = SR_NextSTR(true);

					if ( !(scplinesound[num][0] = R3D_AddNewSound(nextstr)) )
					{
						SR_SCPErr("sound lump \"%s\" not found or invalid", nextstr);
					}

					if (val2 || val3)
					{
						int val = SR_NextNum();

						if (val2)
						{
							if (val < 1 || val > 100)
							{
								SR_SCPErr("bv value out of range 1 to 99", val);
							}
							scplinesound[num][1] = val;
						}
						else	// val3 must be set
						{
							if (!val && val != 1)
							{
								SR_SCPErr("repeat value must be 0 or 1", val);
							}
							scplinesound[num][1] = 100 + val;
						}
					}
					else if (val4)
					{
						scplinesound[num][1] = 102;
					}
				}
				else if (!_stricmp(nextstr, "LIGHTSCHEME"))
				{
					int val;

					val = SR_NextNum(); cfg.useTorchLight = val != 0;
					val = SR_NextNum(); cfg.setVtxLight = val != 0 && !cfg.useTorchLight;
					val = SR_NextNum(); cfg.sineCompress = val != 0;
					
					ma_inf.noscriptset |= NSF_SCPLIGHT;
				}
				else if (!_stricmp(nextstr, "SETOFFSET5"))
				{
					// Apply a virtual distance offset of 5 away from the player
					// for sorting purposes.  Does not change the actual position.
					sc = getSPOTNUM(nextstr);
					if (R3D_ScriptedMap)
						sc->zoflags |= ZO_OFFSET5;
					else
						SR_SCPErr("SETOFFSET5 can only be used with a scripted map");
				}
				else if (  (val2 = !_stricmp(nextstr, "SETDEFDLI"))
					|| !_stricmp(nextstr, "SETDEF"))
				{
					sc = getSPOTNUM(nextstr);
					num = SR_NextNum();					// THING num to spawn

					// Check number in range
					if (   (num >= 16000 && num < 17100)
						|| (num >=  7700 && num <  8000))
					{
						SR_SCPErr("SCP_ZOFF spawn number %i is invalid", num);
					}

					sc->thingnum = num;

					num = SR_NextNum();			// Light level

					if (num <   0) num =   0;
					if (num > 255) num = 255;

					sc->lightlevel = num;

					num = SR_NextNum();			// Height offset (can be negative)

					sc->z_off = (short)num;

					if (val2)
					{
						float intensity = SR_NextFloat();

						if (intensity <  0.05f) intensity = 0.05f;
						if (intensity >= 1.0f ) intensity = 0;	// Since 1 means no change and 0 is ignored

						if (intensity)
						{
							sc->intensity = (fixed_t)(intensity*FRACUNIT);
						}
					}
				}
				else if (!_stricmp(nextstr, "SETELEVSPEED"))
				{
					num = SR_NextNum() - 17000;		// SETELEVSPEED NUM

					if (num < 0 || num > 99)
						SR_SCPErr("num %i out of range 17000 to 17099", 17000 + num);

					float val = SR_NextFloat();

					if (val < 0.1f || val > 6.f)
						SR_SCPErr("%i speed %0.f out of range 0.1 to 6", 17000 + num, val);

					elevspeed[num][ELV_SPEED] = (int)(val * FRACUNIT);	// speed
				}
				else if (!_stricmp(nextstr, "SETELEVSPEEDHEIGHT"))
				{
					num = SR_NextNum() - 17000;		// SETELEVSPEEDHEIGHT NUM

					if (num < 0 || num > 99)
						SR_SCPErr("num %i out of range 17000 to 17099", 17000 + num);

					float val = SR_NextFloat();

					if (val < 0.1f || val > 6.f)
						SR_SCPErr("%i speed %0.f out of range 0.1 to 6", 17000 + num, val);

					elevspeed[num][ELV_SPEED] = (int)(val * FRACUNIT);	// speed

					int n = SR_NextNum();	// height amount to add to floorheight

					if (n > 32000 || n < -32000 || !n)	
						SR_SCPErr("3rd term must not be 0 and must be in range -32000 to 32000");

					elevspeed[num][ELV_HEIGHT] = (fixed_t)(n << FRACBITS);

					n = SR_NextNum();	// get tag num

					if (n <= 0 || n > 32767)	
						SR_SCPErr("tag must be in range 1 to 32767");

					elevspeed[num][ELV_TAG] = n;	// tag to use

					n = SR_NextNum();	// get n-up

					if (n < 0 || n > 128)
						SR_SCPErr("n-up must be in range 0 to 128");

					elevspeed[num][ELV_HILIMIT] = n;

					n = SR_NextNum();	// get n-down

					if (n < 0 || n > 128)
						SR_SCPErr("n-down must be in range 0 to 128");

					elevspeed[num][ELV_LOLIMIT] = n;

					if (!elevspeed[num][ELV_HILIMIT] && !elevspeed[num][ELV_LOLIMIT])
						SR_SCPErr("both limits are zero!");
				}
				else if (!_stricmp(nextstr, "SETPSEUDOFAKE"))
				{
					num = SR_NextNum() - 17000;		// SETPSEUDOFAKE NUM

					if (num < 0 || num > 99)
						SR_SCPErr("num %i out of range 17000 to 17099", 17000 + num);

					int n = SR_NextNum();	// get tag num

					if (n <= 0 || n > 32767)	
						SR_SCPErr("tag must be in range 1 to 32767");

					pseudofake[num][PF_TAG] = n;

					n = SR_NextNum();	// get damage flag

					if (n < 0 || n > 1)	
						SR_SCPErr("damage flag expected to be 0 or 1");

					pseudofake[num][PF_DAMAGE] = n;
					pseudofake[num][PF_FILTER] = SR_NextNum();
				}
				else
				{
					if (!_stricmp(nextstr, "END") || M_ValidateID(nextstr))
						break;

					SR_SCPErr("Unknown keyword \"%s\"", nextstr);
				}
			}
		}
		SR_EndScript();
	}
}

// killough 4/4/98, 4/11/98: highest referencing special linedef's type,
// or lump number of special effect. Allows texture names to be overloaded
// for other functions.
// b211-0 'int special' removed from side_t; tempspec array used instead
static int *tempspec;

//==================================================
// subfn_LoadLineDefs
//	called from P_SetSidePtrArray() v220-35
//==================================================

static void subfn_LoadLineDefs(maplinedefE_t *mld)
{
	line_t			*ld;
	int				i;
	ushort			tempv1, tempv2;
	int				showspec = ArgExists("-linspec");
	int				error_count = 0;
	
	//--------------------------------

	tempspec = (int*)W_Calloc(numsides*sizeof(int), STATICID, 0);

	for (i = 0, ld = lines ; i < numlines ; i++, ld++)
	{
		tempv1 = mld[i].v1;
		tempv2 = mld[i].v2;

		if (tempv1 >= numvertexes || tempv2 >= numvertexes)
		{ 
			delNodesBuild();
			Con_Error("ERROR: invalid vertex line=%d v1=%hu v2=%hu (max=%d)",i,tempv1,tempv2,numvertexes);
		}

		ld->ml_flags   = mld[i].flags & ML_LINEMASK;	// screen out flags we don't recognise
		ld->special	   = mld[i].special;
		ld->tag		   = mld[i].tag;

		ld->v1	       = &vertexes[tempv1];
		ld->v2	       = &vertexes[tempv2];

		if (tempv1 >= numvertexes || tempv2 >= numvertexes)
		{ 
			delNodesBuild();
			Con_Error("ERROR: invalid vertex line=%d v1=%hu v2=%hu (max=%d)",i,tempv1,tempv2,numvertexes);
		}

		ld->sidenum[0] = mld[i].sidenum[0];
		ld->sidenum[1] = mld[i].sidenum[1];
		
		// set sidedef 1 number since unsigned won't set correctly
		if (ld->sidenum[0] == -1)
		{
			if (ge_inf.devmap)
				pr_OFinsert( "MAP ERROR: L%i side 0 has no sidedef. Setting sidedef 0 to 0\n", i);
			ld->sidenum[0] = 0;
			error_count++;
		}
		
		if (   ld->sidenum[0] >= numsides
			|| ld->sidenum[1] >= numsides)
		{ 
			delNodesBuild();
			Con_Error("ERROR: invalid sidedef line=%d sd1=%d sd2=%d",i,ld->sidenum[0],ld->sidenum[1]);
		}

		if (ge_inf.devmap || R3D_edit) ld->ml_flags &= ~ML_DONTDRAW;
		if (ld->ml_flags & ML_MAPPED) ld->ovflags |= OVF_MAPPED;

		if (showspec && ld->special) pr_OFinsert("L%i spec %i 0x%.4X\n", i, ld->special, ld->special);
		
		ld->adjacentsec = -1;	// b211-0
		ld->ambNumLink = 0xFF;	// v220-35
		
		// Set up the vertexes, flags and box limits
		ld->dx = ld->v2->x - ld->v1->x;
		ld->dy = ld->v2->y - ld->v1->y;
		
		if (!ld->dx)
			ld->slopetype = ST_VERTICAL;
		else if (!ld->dy)
			ld->slopetype = ST_HORIZONTAL;
		else
		{
			if (FixedDiv (ld->dy , ld->dx) > 0)
				ld->slopetype = ST_POSITIVE;
			else
				ld->slopetype = ST_NEGATIVE;
		}
		
		if (ld->v1->x < ld->v2->x)
		{
			ld->bbox[BOXLEFT ] = ld->v1->x;
			ld->bbox[BOXRIGHT] = ld->v2->x;
		}
		else
		{
			ld->bbox[BOXLEFT ] = ld->v2->x;
			ld->bbox[BOXRIGHT] = ld->v1->x;
		}
		
		if (ld->v1->y < ld->v2->y)
		{
			ld->bbox[BOXBASE] = ld->v1->y;
			ld->bbox[BOXTOP ] = ld->v2->y;
		}
		else
		{
			ld->bbox[BOXBASE] = ld->v2->y;
			ld->bbox[BOXTOP ] = ld->v1->y;
		}
		
		// killough 4/4/98: support special sidedef interpretation in P_LoadSideDefs2()
		// b211-0 made transient as only needed in P_LoadSideDefs2()
		if (ld->sidenum[0] != -1 && ld->special)
			tempspec[ld->sidenum[0]] = ld->special;
	}

	if (error_count > 10)
	{
		char buff[96];

		sprintf(buff, "%i FATAL MAP ERRORS detected.\n"
			"See Risen3D.out for details.\n"
			"MAP may still be able to be run.\n", error_count);

		if (!Sys_CriticalMessage(buff, true))
		{
			delNodesBuild();
			Sys_Quit(false);
		}
	}
}

// v220-35
static void P_SetSidePtrArray(data_hdr_t *lnptr, data_hdr_t *sdptr)
{
extern int		num_xlsame;
	if (!(numlines = lnptr->length / sizeof(maplinedef_t)) )
		Con_Error("ERROR: %s has no lines!", new_mapid);

	if (numlines >= 0xFFFF)
		Con_Error("ERROR: %s has invalid number of lines!", new_mapid);

	if (!(numsides = sdptr->length / sizeof(mapsidedef_t)) )
		Con_Error("ERROR: %s has no sides!", new_mapid);

	lines = (line_t*)W_Calloc(numlines*sizeof(line_t), LEVELID, 79);

	int				i, n;
	int				error = 0;
	mapsidedef_t	*msd;
	maplinedef_t	*mld;

	sidedata = getDATA(sdptr);
	msd = (mapsidedef_t*)sidedata;

	linedata = getDATA(lnptr);
	mld = (maplinedef_t*)linedata;

	ma_inf.mapsides = numsides;
	ma_inf.numsidecheck = numlines;	// Each line should have at least one side
	
	mldtmp = (maplinedefE_t*)malloc(sizeof(maplinedefE_t) * numlines);

	// set up pointers
	sideptrlist = (mapsidedef_t**)calloc(sizeof(mapsidedef_t*), numlines * 2);
	
	for (i = n = 0; i < numlines ; i++)
	{
		ushort s0 = mld[i].sidenum[0];
		ushort s1 = mld[i].sidenum[1];

		mldtmp[i].v1 = mld[i].v1;
		mldtmp[i].v2 = mld[i].v2;
		mldtmp[i].flags = mld[i].flags;
		mldtmp[i].special = mld[i].special;
		mldtmp[i].tag = mld[i].tag;
		mldtmp[i].sidenum[0] = s0;
		mldtmp[i].sidenum[1] = s1;

		// check for lines with no sides (e.g. aaliens map21)
		if (s0 == 0xFFFF && s1 == 0xFFFF)
		{
			mldtmp[i].sidenum[0] = -1;
			mldtmp[i].sidenum[1] = -1;
			ma_inf.numsidecheck--;
			ma_inf.sd_err_check++;
			continue;
		}

		// also catches s0 = 0xFFFF which is illegal
		if (s0 >= numsides) { error = 1; break; }

		sideptrlist[s0] = &msd[s0];

		if (s1 != 0xFFFF)
		{
			if (s1 >= numsides) { error = 2; break; }
			sideptrlist[s1] = &msd[s1];
			ma_inf.numsidecheck++;
		}
		else
			mldtmp[i].sidenum[1] = -1;
	}

	if (error)
	{
		free(sideptrlist);
		free(mldtmp);
		delNodesBuild();
		Con_Error("FATAL MAP ERROR with sidenum on line %i side %i in %s\n"
			"This is not an R3D fault: contact the map author.",
			i, error - 1, new_mapid);
	}

	if (ma_inf.mapsides > ma_inf.numsidecheck)
	{
		// Account for xlines
		if (ma_inf.mapsides - ma_inf.numsidecheck == 2 * num_xlsame)
		{
			ma_inf.numsidecheck = ma_inf.mapsides;
		}
		else
		{
			pr_OFinsert("WARNING: mismatched sides\n");
		}
	}

	if (ma_inf.numsidecheck > numsides)
	{
		int n;

		for (i = n = 0; i < numlines ; i++)
		{
			int s0 = mldtmp[i].sidenum[0];
			int s1 = mldtmp[i].sidenum[1];

			mldtmp[i].sidenum[0] = n;
			sideptrlist[n++] = &msd[s0];

			if (s1 >= 0)
			{
				if (s1 >= numsides) { error = true; break; }
				mldtmp[i].sidenum[1] = n;
				sideptrlist[n++] = &msd[s1];
			}
		}

		numsides = ma_inf.numsidecheck;
	}

	sides = (side_t*)W_Calloc(numsides * sizeof(side_t), LEVELID, 80);

	if (R3D_linemap|R3D_slopetypes)
		texname3D = (texname3D_t*)calloc(numsides, sizeof(texname3D_t));

	subfn_LoadLineDefs(mldtmp);
}

//
// P_LoadLineDefs2
//
// killough 4/4/98: delay using sidedefs until they are loaded
static void P_LoadLineDefs2()
{
	int i, badflagcount = 0;
	line_t *ld = lines;

	for (i = 0; i < numlines; i++, ld++)
	{
		ld->frontsector = sides[ld->sidenum[0]].sector;
		ld->backsector  = ld->sidenum[1] != -1 ? sides[ld->sidenum[1]].sector : NULL;

		// This error is fixable
		if (ld->sidenum[1] == -1 && ld->ml_flags & ML_TWOSIDED)
		{
			ld->ml_flags &= ~ML_TWOSIDED;  // Reset flag
			badflagcount++;

			if (R3D_edit)
				pr_OFinsert("  WARNING: ML_TWOSIDED flag on single sided line %i reset.\n",numlines - i);
		}

		if (   R3D_linemap
			&& (ld->tag == -1 || ld->tag == -2)
			&& ld->ml_flags & ML_TWOSIDED)
		{
			continue;
		}

		// v220-4 provision to allow block friends only
		if (ld->special == 8112 && R3D_ScriptedMap)	// 0x1FB0
		{
			ld->scpFlags |= SCPF_BLOCKFRIEND;
			ld->special = 0;
			continue;
		}

		if (ld->special == 260)
		{
			int k, tag = ld->tag;

			if (!tag)
			{
				ld->spflags |= SPF_MIDTRANSLUCENT;	// if tag == 0 affect this linedef only
			}
			else for (k = 0; k < numlines; k++)	// else affect all matching linedefs
			{
				// Note: don't clear tags as they may be needed (e.g. silentline tp)
				if (lines[k].tag == tag)
					lines[k].spflags |= SPF_MIDTRANSLUCENT;
			}
			ld->special = 0;
		}

		// These next line specials are used only to set spflags
		// After the flag is set ld->special is zeroed.
		// This prevents their use as trigger lines (in UseSpecialLine)
		if (R3D_linemap && (ld->special == 277 || ld->special == 278))
		{
			int k, mask = SPF_LOWTRANSLUCENT, tag = ld->tag;
				
			if (ld->special == 278) mask |= SPF_MIDTRANSLUCENT;

			if (!tag)
			{
				ld->spflags |= mask;	// if tag == 0 affect this linedef only
			}
			else for (k = 0; k < numlines; k++)	// else affect all matching linedefs
			{
				if (lines[k].tag == tag) lines[k].spflags |= mask;
			}

			ld->special = 0;
		}
	}

	if (badflagcount && !R3D_edit)
		pr_OFinsert("MAP ERROR: ML_TWOSIDED flag on %i single sided ldef(s) reset.\n", badflagcount);
}

// Filter stuff (v220-13 revised)
int MakeFilterARGB(int lump, int *lim)
{
	uint	i, r, g, b, count;
	int		w, h;
	int		pixsize;
	int		lidx = DoomLumpIdentify(lump, &w, &h, DLI_GRAPHIC);
	byte	val[4], *map;
	RDBOOL	use_palbuffer = false;

	if (lim) *lim = 0;

	if (lidx == LMP_PNG || lidx == LMP_JPG)
	{
#include "gl_png.h"
#include "gl_jpg.h"

		if (w < 64 || h < 64) return -1;	// only called for flats

		if (lidx == LMP_PNG)
		{
			if ((map = PNG_Load(NULL, lump, &w, &h, &pixsize, 10)) == NULL)
				return -1;
		}
		else
		{
			if ((map = JPG_Load(NULL, lump, &w, &h, &pixsize, 10)) == NULL)
				return -1;
		}

		count = (w * h);
	}
	else if (lumpinfo[lump].groupMarker == LGT_COLORMAPS)
	{
		if ( !(count = lumpinfo[lump].size) ) return -1;
		use_palbuffer = true;
	}
	else if (lumpinfo[lump].lumptype == LMP_FLAT)
	{
		if ( !(count = w * h) ) return -1;
		use_palbuffer = true;
	}
	else
	{
		return -1;
	}

	if (use_palbuffer)
	{
		if ( (map = (byte*)M_CacheLumpNum(lump, NULL, false)) == NULL)
			return -1;
	}

	r = g = b = 0;

	for (i = 0; i < count; i++)
	{
		if (use_palbuffer)
			memcpy(val, palbuffer + 3*map[i], 3);
		else
			memcpy(val, map + ((pixsize) * i), pixsize);

		r += val[0];
		g += val[1];
		b += val[2];

		if (lim && val[0] > 128 && val[1] > 128 && val[2] > 128)
			(*lim) += 1;
	}

	M_Free(map);

	// Use a fixed value for the alpha
	return (0x70000000|((r/count)<<16)|((g/count)<<8)|(b/count));
}

// Is there a 'C' routine for this?
static int hextodec(char c)
{
	char v = toupper(c);

	if (v >= '0' && v <= '9') return v - '0';
	if (v >= 'A' && v <= 'F') return v - 'A' + 10;

	return -1;
}

int R_ColorFilterForName(char *name, int *type)
{
	int i, v, f = 0, cn;

	*type = false;

	if (!name[0] || name[0] == '-')
	{
		return -1;
	}

	// COLORMAP predefined to return 0
	if (!strnicmp(name, "COLORMAP", 8))
	{
		return 0;
	}

	// for WATERMAP use preset value
	if (!strnicmp(name, "WATERMAP", 8))
	{
		return 0x80004fa5;
	}

	// for BLUMAP use a slightly bluer preset value
	if (!strnicmp(name, "BLUMAP", 8))
	{
		return 0xA0004fa5;
	}

	if ((cn = W_CheckColormapForName(name)) >= 0)
	{
		return MakeFilterARGB(cn, NULL);
	}

	// No match - check if a direct HEX definition
	for (i = 0; i < 8; i++)
	{
		if ((v = hextodec(name[i])) >= 0)
			f |= (v<<((7 - i) * 4));
		else
		{
			*type = true;
			return -1;
		}
	}

	return f;
}

// killough 4/4/98: delay using texture names until
// after linedefs are loaded, to allow overloading.
static void P_LoadSideDefs2()
{
	mapsidedef_t mpt[1];
	sector_t	 *sec;
	side_t		 *sd = sides;
	int			 num, error_count = 0, bad_error = 0;
	int			 filter;
	int			 fog_ignore = 0;
	int			 tran_ignore = 0;
	RDBOOL		 type;
	char		 texname[8];
	char		 basename[10];
	char		 *ptr;
	int			 i, k;

	memset(basename, 0, 10);

	if (commercial_game)
		strcpy(basename, "AASHITTY");
	else
		strcpy(basename, "AASTINKY");

	asyflags = (byte*)W_Calloc(numsides, STATICID, 0);

	for (i = 0; i < numsides; i++, sd++)	// ma_inf.mapsides
	{
		RDBOOL	setNormal = true;

		if (!sideptrlist[i])	// Deal with bad sidedef so map can run
		{
			if (ge_inf.devmap) pr_OFinsert("BAD SIDEDEF %i", i);

			if (i + 1 < numsides && sideptrlist[i + 1])
			{
				sideptrlist[i] = sideptrlist[i + 1];
				if (ge_inf.devmap) pr_OFinsert(" next side has been substituted referencing sector %i\n", sideptrlist[i]->sector);
			}
			else if (sideptrlist[0])
			{
				sideptrlist[i] = sideptrlist[0];
				if (ge_inf.devmap) pr_OFinsert(" side 0 has been substituted referencing sector %i\n", sideptrlist[0]->sector);
			}
			else
			{
				Con_Error("MAP has bad sidedefs and cannot be run");
			}

			if (ge_inf.devmap) pr_OFdirect("\n");
		}

		memcpy(mpt, sideptrlist[i], sizeof(mapsidedef_t));

		// Mod if sector error so map can be run
		if (mpt->sector >= numsectors)
        {
			int lnum = -1;

			for (k = 0; k < numlines; k++)
			{
				if (lines[k].sidenum[0] == i || lines[k].sidenum[1] == i)
				{
					lnum = i; break;
				}
			}

			if (lnum >= 0)
			{
				if (R3D_edit || verbose)
					pr_OFinsert("WARNING: sidedef %i used with line %i has invalid sector %i : sector 0 substituted\n",
								i, lnum, mpt->sector);

				bad_error++;
			}

			error_count++;
			sd->sector = sectors;

			continue;
        }

		if (!sd) Con_Error("BAD SIDE %i", i);

		sd->orig_textureoffset = mpt->textureoffset;
		sd->textureoffset = (mpt->textureoffset)<<FRACBITS;
		sd->orig_rowoffset= mpt->rowoffset;
		sd->rowoffset     = (mpt->rowoffset)<<FRACBITS;
		sd->sector  = sec = &sectors[mpt->sector];
		sd->smoothoff = -1;	// v220-4 indicate unset

		// v220-6 Ensure any names starting with - are truncated
		if (mpt->bottomtexture[0] == '-') mpt->bottomtexture[1] = 0;
		if (mpt->midtexture[0] == '-'   ) mpt->midtexture[1] = 0;
		if (mpt->toptexture[0] == '-'   ) mpt->toptexture[1] = 0;

		if (R3D_slopetypes)
		{
			slsid_t	*sld = NULL;
			byte	type;

			if (   mpt->bottomtexture[0] == '$'
				&& mpt->bottomtexture[7]
				&& (!(type = mpt->bottomtexture[1] - 'A') || type == 1))	// Must be 'A' or 'B'
			{
				strncpy(texname, &mpt->bottomtexture[2], 6);
				texname[6] = 0;

				if ( (num = atoi(texname)) > 0x7FFF) num = 0xFFFF;

				if (num <= 0) num = 0xFFFE;

				mpt->bottomtexture[0] = '-';	// Put string "-" in mpt
				mpt->bottomtexture[1] = 0;

				if (numsidarray >= maxnumsidarray)
				{
					maxnumsidarray += 32;
					slsidarray = (slsid_t*)W_Realloc(slsidarray, maxnumsidarray * sizeof(slsid_t), LEVELID, 81);
				}
				
				sld = &slsidarray[numsidarray++];

				sld->sidenum = i;
				sld->offsetlow = num;
				sld->offsethigh = 0;
				sld->type[0] = type;
				if (type == 1) sd->lsiflags |= LS_LOSLOPEB;
			}

			if (   mpt->toptexture[0] == '$'
				&& mpt->toptexture[7]
				&& (!(type = mpt->toptexture[1] - 'A') || type == 1))
			{
				strncpy(texname, &mpt->toptexture[2], 6);
				texname[6] = 0;

				if ( (num = atoi(texname)) > 0x7FFF) num = 0xFFFF;

				if (num <= 0) num = 0xFFFE;

				mpt->toptexture[0] = '-';	// Put string "-" in mpt
				mpt->toptexture[1] = 0;

				if (!sld)
				{
					if (numsidarray >= maxnumsidarray)
					{
						maxnumsidarray += 32;
						slsidarray = (slsid_t*)W_Realloc(slsidarray, maxnumsidarray * sizeof(slsid_t), LEVELID, 82);
					}
					
					sld = &slsidarray[numsidarray++];
					sld->sidenum = i;
					sld->offsetlow = 0;
				}

				sld->offsethigh = num;
				sld->type[1] = type;
				if (type == 1) sd->lsiflags |= LS_HISLOPEB;
			}
		}

		// 3D line definitions
		if (texname3D && tempspec[i] != 242) // 242 dealt with below
		{
			if (mpt->toptexture[0] == '$')
			{
				texname3D_t *tx = &texname3D[i];
				strncpy(tx->hiname, mpt->toptexture, 8);
				mpt->toptexture[0] = '-';
				mpt->toptexture[1] = 0;
			}

			if (mpt->midtexture[0] == '$')
			{
				texname3D_t *tx = &texname3D[i];
				strncpy(tx->midname, mpt->midtexture, 8);
				mpt->midtexture[0] = '-';
				mpt->midtexture[1] = 0;
			}

			if (mpt->bottomtexture[0] == '$')
			{
				texname3D_t *tx = &texname3D[i];
				strncpy(tx->loname, mpt->bottomtexture, 8);
				mpt->bottomtexture[0] = '-';
				mpt->bottomtexture[1] = 0;
			}
		}

		// killough 4/4/98: allow sidedef texture names to be overloaded
		// killough 4/11/98: refined to allow colormaps to work as wall
		//                   textures valid if invalid as colormaps

		if (tempspec[i] == 242)	// variable colormap via 242 linedef
		{
			filter = R_ColorFilterForName(mpt->bottomtexture, &type);

			if (filter != -1)
			{
				sec->bottommap = filter;
				sd->bottomtexture = 0;
			}
			else
			{
				sd->bottomtexture = type ? R_TextureNumForName(mpt->bottomtexture, false) : 0;
			}

			num = 0;

			if (mpt->midtexture[0] == '$'
				&& mpt->midtexture[7]
				&& mpt->midtexture[1] == 'F')
			{
				strncpy(texname, &mpt->midtexture[2], 6);
				texname[6] = 0;

				if ( (num = atoi(texname)) > 0x7FFF) num = 0x7FFF;

				sd->lsiflags |= LS_FAKESOUND;

				sd->sector->floorheight -= (num * FRACUNIT);
			}
			else
			{
				if (mpt->toptexture)
				{
					filter = R_ColorFilterForName(mpt->toptexture, &type);

					if (filter != -1)
					{
						sec->topmap = filter;
						sd->toptexture = 0;
					}
					else
					{
						sd->toptexture = type ? R_TextureNumForName(mpt->toptexture, false) : 0;
					}
				}

				if (mpt->midtexture)
				{
					filter = R_ColorFilterForName(mpt->midtexture, &type);

					if (filter != -1)
					{
						sec->midmap = filter;
						sd->midtexture = 0;
					}
					else
					{
						sd->midtexture = type ? R_TextureNumForName(mpt->midtexture, false) : 0;
					}
				}
			}
			setNormal = false;
		}
		else if (tempspec[i] == 260)
		{
			// killough 4/11/98: apply translucency to 2s normal texture
		/*	sd->midtexture = !strncmp("TRANMAP", mpt->midtexture, 8) ?
				(sd->special = W_CheckNumForName(mpt->midtexture)) < 0 ||
				W_LumpLength(sd->special) != 65536 ?
				sd->special=0, R_TextureNumForName(mpt->midtexture, false) :
				(sd->special++, 0) : (sd->special=0);

			sd->toptexture = R_TextureNumForName(mpt->toptexture, false);
			sd->bottomtexture = R_TextureNumForName(mpt->bottomtexture, false);
			break;*/

			// v220-5 TRANMAP ignored, code would not work in any case as the only
			//        time I've seen these they've had names like TRAN15 or TRAN50
			ptr = &mpt->midtexture[0];
			if (!strncmp(ptr, "TRAN", 4)) tran_ignore++;
			tempspec[i] = 0;
		}
		else if (tempspec[i] == 282)
		{
			ptr = mpt->toptexture;

			// v220-5 Assume this is a line special 282 fog texture
			if (ptr[0] == '#')
			{
				// These appear in hg.wad map05 and don't make much sense. I assume it's
				// formatted as RGB HEX following the # with the end letter being alpha
				// in the range 0 to Z?. Despite this several types with different
				// end letters border the same sector (see S618) so maybe it means
				// something else. The special is also an unsupported type.
				// On top of all this prBoom 2.5.1.4, ZDoom 2.7.1 and GZDoom 1.8.04 all
				// produce different effects, Furthermore in ZDoom S16 is also affected
				// despite no definition for this sector being provided.
				// IMHO using a fake sector would have been better.
				mpt->toptexture[0] = '-';
				mpt->toptexture[1] = 0;
				fog_ignore++;
			}
		}
		else if (tempspec[i] == 8113 || tempspec[i] == 8114)
		{
			ma_inf.holdstate = R3D_ScriptedMap;	// v220-13
		}

		if (setNormal)
		{
			sd->midtexture = R_TextureNumForName(mpt->midtexture, false);
			sd->toptexture = R_TextureNumForName(mpt->toptexture, false);
			sd->bottomtexture = R_TextureNumForName(mpt->bottomtexture, false);
		}

		// R3D - There is no provision in the Doom engine, following loading, to
		// distinguish a NULL texture, named '-', from texture num 0 named "AASHITTY"
		// in Doom2 or "AASTINKY" in Doom. This will cause the missing texture stuff
		// to fail unless dealt with.
		sd->texFlags = 0;

		if (sd->toptexture)
			sd->texFlags = T_HI;
		else if (!strnicmp(mpt->toptexture, basename, 8))
			asyflags[i] |= ASY_H;

		if (sd->midtexture)
			sd->texFlags |= T_MID;
		else if (!strnicmp(mpt->midtexture, basename, 8))
			asyflags[i] |= ASY_M;

		if (sd->bottomtexture)
			sd->texFlags |= T_LOW;
		else if (!strnicmp(mpt->bottomtexture, basename, 8))
			asyflags[i] |= ASY_L;
	}

	W_Free(tempspec, STATICID, 0);
	
	if (error_count)
	{
		if (bad_error)
			pr_OFinsert("\nWARNING: %i invalid sidedef sectors found of which %i are referenced.\n", error_count, bad_error);
		else
			pr_OFinsert("\nADVISORY: %i invalid but unreferenced sidedef%s found.\n", error_count, plural(error_count));
	}

	if (tran_ignore)
		pr_OFinsert("ADVISORY: %i unsupported TRAN maps with L special 260 found\n", tran_ignore);

	if (fog_ignore)
		pr_OFinsert("ADVISORY: %i unsupported fog textures with L special 282 found\n", fog_ignore);

	free(mldtmp);
	free(linedata);
	free(sidedata);
	free(sideptrlist);
}

//
// P_GroupLines
// b211-0 changed
// Builds sector line lists and subsector sector numbers.
// Sets bounding boxes for sectors.
//
static void addLineToSector(sector_t *sec, line_t *li)
{
	M_AddLineToBox (sec->blockbox, li);
	*sec->lines++ = li;
}

static void P_GroupLines (void)
{
	line_t		**linebuffer;
	subsector_t	*ss;
	sector_t	*sec;
	line_t		*li;
	seg_t		*seg;
	int			i, j, total;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		ss->sector = NULL;

		ss->sslightsec	= -1;
		ss->floorsec	= -1;
		ss->ceilsec		= -1;
		ss->lcsec		= -1;
		ss->lfsec		= -1;
		ss->viewgroup	=  0;

		for (j = 0, seg = &segs[ss->firstline]; j < ss->linecount; j++, seg++)
		{
			if (seg->frontsector)
			{
				ssinitial[i] = seg->frontsector - sectors;
				ss->sector = seg->frontsector;
				break;
			}
		}

		if (subsectors[i].sector == NULL)
			Con_Error("P_GroupLines: Subsector %i a part of no sector", i);
	}

	// set seg parms
	for (i = 0, seg = segs; i < numsegs; i++, seg++)
	{
		seg->fv1[VX] = FIX2FLT(seg->v1->x);
		seg->fv1[VY] = FIX2FLT(seg->v1->y);
		seg->fv2[VX] = FIX2FLT(seg->v2->x);
		seg->fv2[VY] = FIX2FLT(seg->v2->y);
		seg->rtp_num = 0;
		seg->zptr1 = 0;
		seg->zptr2 = 0;
		seg->length = M_AccurateDistance(seg->v2->x - seg->v1->x, seg->v2->y - seg->v1->y);							
/* $angle not used
		seg->angle = bamsAtan2((seg->v2->y - seg->v1->y)>>FRACBITS, 
			(seg->v2->x - seg->v1->x)>>FRACBITS) << 16;
			*/
	}
	
	// count number of lines in each sector
	for (i = 0, li = lines; i < numlines; i++, li++)
	{
		li->frontsector->linecount++;

		if (!li->backsector) li->ml_flags |= ML_BLOCKING;

		if (li->backsector && li->backsector != li->frontsector)
			li->backsector->linecount++;
	}
	
	// compute total number of lines and clear bounding boxes
	for (total = 0, i = 0; i < numsectors; i++)
	{
		total += sectors[i].linecount;
		M_ClearBox(sectors[i].blockbox);
	}
	
	// create sector sequential buffer of line pointers
	linebuffer = (line_t**)W_Calloc(total * sizeof(line_t*), LEVELID, 83);
		
	// point the sectors line list ptr to its position in the line buffer
	for (i = 0; i < numsectors; i++)
	{
		sectors[i].lines = linebuffer;
		linebuffer += sectors[i].linecount;
	}

	// Enter the lines using the same logic as counting the lines for each sector
	for (i = 0, li = lines; i < numlines; i++, li++)
	{
		addLineToSector(li->frontsector, li);

		if (li->backsector && li->backsector != li->frontsector)
			addLineToSector(li->backsector, li);
	}
	
	// Reset line ptr and calculate sound org/blockbox
	for (i = 0, sec = sectors; i < numsectors; i++, sec++) 
	{
		fixed_t	*bbox = sec->blockbox;
		int		xl, xh, yl, yh;

		// reset pointer back to the correct list position
		// (was incremented when adding the lines)
		sec->lines -= sec->linecount;

		// b211-0 Added for analysis
		if (sec->linecount > ma_inf.maxseclinecount)
			ma_inf.maxseclinecount = sec->linecount;

		// set the degenmobj_t to the middle of the bounding box
		sec->soundorg.x = meanV(bbox[BOXRIGHT], bbox[BOXLEFT]);
		sec->soundorg.y = meanV(bbox[BOXTOP  ], bbox[BOXBASE]);

		M_getBlockOffsets(bbox, MAXRADIUS, &xl, &xh, &yl, &yh);

		// adjust bounding box to map blocks
		sec->blockbox[BOXTOP  ] = yh >= bmapheight ? bmapheight - 1 : yh;		
		sec->blockbox[BOXBASE ] = yl < 0 ? 0 : yl;
		sec->blockbox[BOXRIGHT] = xh >= bmapwidth ? bmapwidth - 1 : xh;
		sec->blockbox[BOXLEFT ] = xl < 0 ? 0 : xl;
	}
}

static void LS_CheckPlayerFriendliness()
{
	if (ge_inf.isfriendly)
	{
		mobj_t	*iter = mbf_players[0].plr->mo;

		if ( !(iter->flags3 & MF3_FRIEND) )
		{
			iter->flags3 |= MF3_FRIEND;
			P_UpdateThinker(&iter->thinker);
		}
	}
}

// B211-0 Relink all Things
static void checkOwners()
{
	thinker_t *th;

	for (th = thinkercap.next; th != &thinkercap; th = th->next)
	{
		if ((void (__cdecl *)(void))th->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			P_UnlinkThing((mobj_t *) th);
			P_SetThingPosition((mobj_t *) th);
		}
	}
}

static char currentMapid[8];

void P_SetMapID(char *id, int ep, int gm)
{
	if (commercial_game)
		sprintf(id, "map%02i", gm);
    else
		sprintf(id,"E%iM%i", ep, gm);
}

char *P_GetCurrentMapid()
{
	return currentMapid;
}

// v220-4 new routine added
static void Set_SsecReferences()
{
	subsector_t	*ss;
	seg_t		*sg;
	int			h, i, k;
	
	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (ss->ssflags & SSF_NORENDER) goto skip_flag;

		// v220-31 needed for automap
		for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef && !(sg->linedef->ml_flags & ML_DONTDRAW) ) goto skip_flag;
		}

		// check the sector
		{
			sector_t *sc = ss->sector;

			for (k = 0; k < sc->linecount; k++)
			{
				if (   !(sc->lines[k]->ml_flags & ML_DONTDRAW)
					&& !(sc->lines[k]->spflags & SPF_CONTROLLINE))
				{
					goto skip_flag;
				}
			}
		}

		ss->ssflags |= SSF_NOAMFLOOR;

skip_flag:;

		// Floor defaults
		ss->dlfloorheight = ss->sector->floorheight;
		ss->ssfloorpic    = ss->sector->floorpic;
		ss->s_waterplane = DDMININT;

		// v220-31 clear stuff that may have got set in control ssecs
		if (ss->ssflags & SSF_NORENDER)
		{
			ss->ceilsec = ss->lcsec = -1;
			ss->floorsec = ss->lfsec = -1;
			continue;
		}

		if (ss->ceilsec >= 0)
		{
			ss->dlceilheight = sectors[ss->ceilsec].ceilingheight;
			ss->ssceilpic    = sectors[ss->ceilsec].ceilingpic;
		}
		else
		{
			ss->dlceilheight  = ss->sector->ceilingheight;
			ss->ssceilpic     = ss->sector->ceilingpic;
		}

		if ( (h = ss->sector->heightsec) >= 0)
		{
			ss->s_waterplane = sectors[h].floorheight;
		}
		else if (ss->floorsec >= 0)
		{
			ss->dlfloorheight = sectors[ss->floorsec].floorheight;
			ss->ssfloorpic    = sectors[ss->floorsec].floorpic;

			if ( (h = sectors[ss->floorsec].heightsec) >= 0)
			{
				ss->s_waterplane = sectors[h].floorheight;
			}
		}
	}
}

// v220-4 written to allow other modifier list headers
//        to be added later should this be required.
// v220-10 modified for new entries
// v220-13 modified for new entry

typedef struct {
	char	*str;
	int		val;
	int		flag;
} modvar_t;

modvar_t mval[] = {
	{"[nodeFactor12]",		MODF_FACTOR12,	  0 },
	{"[nodeFactor32]",		MODF_FACTOR32,	  0 },
	{"[zdNodesGL]",			MODF_ZDNODG,	  0 },
	{"[zdNodes]",			MODF_ZDNODO,	  0 },
	{"[gwaOnly]",			MODF_GWAONLY,	  0 },
	{"[120modifier]",		MODF_120,		  0 },
	{"[stptelefrag]",		MODF_KILL,		  0 },
	{"[DoomFloorBug]",		MODF_FLOORFIX,	  0 },
	{"[AdjustScroll]",		MODF_SCROLL,	  0 },
	{"[modifyNodes]",		MODF_MODNODES,	  0 },
	{"[noskymask]",			MODF_NOSKYMASK,	  0 },
	{"[dehAmbients]",		MODF_DEHAMBIENT, -1 },
	{"[ignoreThingError]",	MODF_ITE,		 -1 },
	{"[DoomStairs]",		MODF_DMSTAIRS,	 -1 },
	{"[DoomModelSectors]",	MODF_MSECTORS,	 -1 },
	{"[carryFix_off]",		MODF_CFO,		 -1 },
	{"[noHudSpriteScale]",	0,				  1 },
	{NULL,					0,	0 }
};

static void Check_Modifier(char *mapid, int modidx)
{
	RDBOOL	matched;
	char	fname[512], *nextstr;
	char	wadname[256];
	int		i;

	//------------------------------

	if (modidx < 0) return;

	show_wadname = filenameRecords[modidx].filename;

	_splitpath(show_wadname, NULL, NULL, wadname, NULL);

	// iwad defaults
	if (!(ge_inf.flags & GEF_COMPSTAIRS)) compat[c_stairs] = 1;
	if (!(ge_inf.flags & GEF_COMPMODEL )) compat[c_model ] = 0;

	sprintf(fname, "%sR3D_modify", ddBinDir.path);

	if (!SR_LoadFile(fname, "R3D_modify", NULL))
	{
		Con_ErrorD("ERROR: the file Risen3D\\bin\\R3D_modify is missing");
	}

	for (i = 0; mval[i].str; i++, SR_RewindScript())
	{
		RDBOOL	found = false;
		
		nextstr = SR_NextSTR(false);

		if (_stricmp("Version", nextstr) || SR_NextNum() != 4)
			Con_Error("ERROR:\tthe R3D_modify file loaded is an old version\n"
					  "\tPlease re-install the latest version of "VERSIONSTR);
		
		while ( (nextstr = SR_NextSTR(false)) )
		{
			if (!_stricmp(nextstr, mval[i].str))
			{
				found = true; break;
			}
		}

		if (!found) continue;

		while ( (nextstr = SR_NextSTR(false)) )
		{
			uint	numchar = 0;

			if (!nextstr || nextstr[0] == '[') break;

			if (nextstr[0] == '\\') numchar = strlen(nextstr + 1);

			if (numchar && strlen(wadname) >= numchar)
				matched = !_strnicmp(wadname, nextstr + 1, numchar);	// wad name
			else if (!numchar)
			{
				if (*nextstr == '#')	// v220-35 string in wad name
				{
					matched = strstrCI(wadname, nextstr + 1);
				//	pr_OFinsert("%s %s %i\n", wadname, nextstr + 1, matched);
				}
				else
				{
					matched = !_stricmp(wadname, nextstr);	// wad name
				}
			}

			if (mval[i].flag == -1)	// mapless entries
			{
				if (matched)
				{
					switch(mval[i].val)
					{
					case MODF_DMSTAIRS: compat[c_stairs] = 0; break;							
					case MODF_MSECTORS: compat[c_model ] = 1; break;
					case MODF_CFO     : compat[c_carry ] = 0; break;
					case MODF_ITE: ma_inf.modifier |= MODF_ITE; break;
					case MODF_DEHAMBIENT:
						dehamb_vol = SR_NextFloat();
						if (dehamb_vol < 0) dehamb_vol = 0;
						if (dehamb_vol > 1.0f) dehamb_vol = 1.0f;
						dehamb_attn = SR_NextFloat();
						if (dehamb_attn < 0) dehamb_attn = 0;
						if (dehamb_attn > 4.0f) dehamb_attn = 4.0f;
						break;
					}

					break;
				}
			}
			else if (mval[i].flag == 1)	// noHudSpriteScale
			{
				if (matched) ge_inf.noHudSpriteScale = true;
			}
			else	// nodebuild entries (should) have a mapid
			{
				// even if not matched strings must be read
				nextstr = SR_NextSTR(true);

				// v220-29 check for missing mapid in the script
				if (_stricmp("all", nextstr))
				{
					if (strlen(nextstr) == 5)
					{
						int num = atoi(nextstr + 3);

						if (   strnicmp(nextstr, "map", 3)
							|| num < 0
							|| num > 99)
						{
							SR_SCPErr("INVALID map with %s\n", wadname);
						}
					}
					else if (strlen(nextstr) != 4
						|| strnicmp(nextstr    , "E", 1)
						|| strnicmp(nextstr + 2, "M", 1) )
					{
						SR_SCPErr("INVALID map with %s\n", wadname);
					}
				}

				if (   matched
					&& (!_stricmp(nextstr, mapid)		// mapid
					||  !_stricmp("all", nextstr)) )	// all maps in pwad
				{
					ma_inf.modifier |= mval[i].val;
					break;
				}
			}
		}
	}

	SR_EndScript();
}

static int getEntries(char *name, int sz, RDBOOL needed, char *msg)
{
	int lump = getLumpByName(name);
	int in = lump >= 0 ? lumpinfo[lump].size / sz : 0;

	if (needed && !in)
	{
		int		modidx = lumpinfo[map_lumpnum].modidx; 
		char	*n = filenameRecords[modidx].filename;

		Con_Error ("Wad %s %s has no %s\n", n, new_mapid, msg);
	}

	return in;
}

int CheckTexLumpByName(char *name)
{
	int w, h, lump = getLumpByName(name);
	int type = DoomLumpIdentify(lump, &w, &h, DLI_GRAPHIC);

	if (type >= LMP_PNG && type <= LMP_PATCH && w && h)
		return lump;
	return -1;
}

static RDBOOL checkMapType(RDBOOL isPwad)
{
	uint doom2m02[] = {  97,  501,  523,  715,   797 };
	uint doom2m18[] = { 108,  806,  774,  986,  1132 };
	uint doom2m19[] = { 251, 1252, 1255, 1821,  2117 };
	uint doom2m21[] = {  61,  396,  327,  538,   640 };
	uint doom2m22[] = { 140,  590,  657,  940,   995 };
	uint doom2m27[] = { 187,  864,  946, 1407,  1522 };
	uint doom2m30[] = {  22,   82,   93,  131,   137 };
	uint etIIIm10[] = { 664, 2777, 2807, 4056,  4713 };
	uint etIIIm28[] = { 960, 7437, 6240, 8612, 10831 };

	uint	vn[5];
	uint	*ptr = NULL;
	char	*mw = filenameRecords[lumpinfo[map_lumpnum].modidx].filename;
	char	*mname = lumpinfo[map_lumpnum].name;

	//-----------------------------------------

	doom2Map = mapEternal = 0;

	// rough check on whether map can be built
	vn[4] = getEntries("SEGS", sizeof(mapseg_t), 0, "segs");

#ifdef _DEBUG
	// v220-16 allow cmd line -zdnodg to force building zdbsp nodes
	if (ma_inf.modifier & MODF_ZDNODG)
	{
		// v220-14 This can only be dealt with using zdbsp
		if (!(ma_inf.modifier & MODF_ZDNODONLY))
		{
			if ( (ma_inf.modifier & (MODF_ZDNODG|MODF_ZDNODO|MODF_GWAONLY)) != MODF_ZDNODG)
				pr_OFdirect("Map size requires zdbsp. User settings ignored\n");
		}

		ma_inf.modifier &= ~(MODF_ZDNODG|MODF_ZDNODO|MODF_GWAONLY);
		ma_inf.modifier |= (MODF_ZDNODG|MODF_ZDNODONLY);
			
		return false;
	}
#endif

	if (vn[4] > 0xFFFE) return false;	// v220-35 do not force zdbsp nodes

	// Check for any special nodes build modifications.
	if (!isPwad && gamemission == doom2)
	{
		//  The Doom2 iwad has some significant errors that
		//  are best fixed at the time of building the nodes.
		if (!_stricmp(mname, "MAP02"))	// added v220-35
		{
			ptr = doom2m02; doom2Map = 2;
		}
		else
		if (!_stricmp(mname, "MAP18"))
		{
			ptr = doom2m18; doom2Map = 18;
		}
		else
		if (!_stricmp(mname, "MAP19"))
		{
			ptr = doom2m19; doom2Map = 19;
		}
		else
		if (!_stricmp(mname, "MAP21"))
		{
			ptr = doom2m21; doom2Map = 21;
		}
		else
		if (!_stricmp(mname, "MAP22"))
		{
			ptr = doom2m22; doom2Map = 22;
		}
		else
		if (!_stricmp(mname, "MAP27"))
		{
			ptr = doom2m27; doom2Map = 27;
		}
		else
		if (!_stricmp(mname, "MAP30"))
		{
			ptr = doom2m30; doom2Map = 30;
		}
	}
	else
	if (isPwad && mw && gamemission == doom2)
	{
		char szFname[256];

		_splitpath(mw, NULL, NULL, szFname, NULL);

		if (  !_stricmp(szFname, "ETERNALL")
			&& pwadchecksum[1] == 0x00D83EAE)
		{
			// Most pwad faults can be fixed by the analyser
			// In the case of Eternal III then some are a bit intractable
			// and are best fixed before building the nodes. Since this is
			// a classic pwad the trouble of doing this was felt to be
			// worthwhile. Early pwads are the ones most likely to have
			// intractable faults as rendering resolutions were poor at the
			// time making faults difficult to spot.
			if (!_stricmp(mname, "MAP10"))
			{
				ptr = etIIIm10; mapEternal = 10;
			}
			else
			if (!_stricmp(mname, "MAP28"))
			{
				ptr = etIIIm28; mapEternal = 28;
			}
		}
	}

	if (ptr)
	{
		int i;

		// Get the remaining sizes
		vn[0] = getEntries("SECTORS" , sizeof(mapsector_t) , 1, "sectors");
		vn[1] = getEntries("VERTEXES", 2*sizeof(short)	   , 1, "vtxs");
		vn[2] = getEntries("LINEDEFS", sizeof(maplinedef_t), 1, "lines");
		vn[3] = getEntries("SIDEDEFS", sizeof(mapsidedef_t), 1, "sides");

		for (i = 0; i < 5; i++)
		{
			if (ptr[i] != vn[i]) { doom2Map = mapEternal = 0; break; }
		}

		// v220-8 disable nodes correction in edit mode
		if (R3D_edit && (doom2Map|mapEternal))
		{
			Sys_MessageBox( "NOTE: internal nodes fix that is normally\n"
							"applied to this map is disabled in edit mode\n", false);
			doom2Map = mapEternal = 0;
		}
	}

	return (R3D_edit || ((doom2Map|mapEternal) > 0) );
}

//===========================================================================
// P_SetupLevel
//  v220-6 Nodes build now carried out on a per map basis.
//  If a nodes build takes longer than 1.0 second a compressed backup
//  file is created using routines  added to r3d_anifile.c
//
//===========================================================================
void P_SetupLevel(int *gameepisode, int *gamemap)
{
	static int oldge = -1, oldgm = -1;
	static RDBOOL wasScriptedMap = false;

	int		i, k, error;
	int		factor = 10;	// default
	int		use_analyse_file;
	int		current_pwad;
	int		ge_infmod = ge_inf.modifier;
	char	analysisfile[MAXPATHNAMESIZE];
	long	nodetime = 0;

	//----------------------------

	fnode = NULL;
	
	Con_Open(false);
	
	nodefile[0] = 0;

	ge_inf.conprogress_disabled = (mapinfo && mapinfo->R3Dflags & nointermission);

	if (ge_inf.loadlevel) ge_inf.loadlevel = ge_inf.conprogress_disabled;	
	if (ge_inf.loadlevel && R3D_ScriptedMap) DR_DrawTextLoading();

	ZoomReset(true);	// Sets fieldOfView
	ResetProgHeader();
	PH_ResetChanges();

	if (!BDMFX)
	{
		water_trans = false;
		useDynLights = false;
		forceDynLights = false;
	}

	wminfo.maxfrags = 0;

	gametic = 0;
	curfilter = fakefilter = 0;	
	firstNetUpdate = true;
	current_skyvarnum = -1;
	cfg.berserkBug = cfg.noberserkBug = 0;
	anisl.numentries = 0;
	numLuminous = 0;
	playerdamageidx = enemyheadshotidx = 0;
	setskycolour = -1.f;
	skyflatnum = wasskyflatnum;

	// Reset all the counters and other data.
	for (i = 0; i < MAXPLAYERS; i++)
	{
		players[i].ingame = false;
		players[i].flags &= ~DDPF_CAMERA;
	}

	players[0].ingame = true;
	numplayerstarts = 0;
	fogval.isset = false;

    for (i = 0 ; i < MAXPLAYERS ; i++)
    {
		mbf_players[i].killcount
			= mbf_players[i].secretcount 
			= mbf_players[i].itemcount = 0;
    }
	
    // Initial height of PointOfView
    // will be set by player think.
    mbf_players[0].plr->viewz = DDMININT; 
	
	if (oldge != *gameepisode && oldgm != *gamemap)
	{
		GL_DeleteAllSecondaryAlphaTex();

		// If 3D textures were added then resize arrays (i.e remove them).
		// This allows each map to have its own unique texture sizing
		// without stacking up otherwise unwanted textures from other maps.
		GL_Delete3DTextures();
	}

    S_LevelChange();	// Stop sounds
	W_CheckHeap(1);		// Check before clearing
	if (W_FreeALL(LEVELID) ) Con_ErrorD("LEVEL  : memory corrupt");

	// reset nodefix and vertexfix arrays *after* freeing LEVELID
	// and *before* scanning the R3D_NODEfix file.
	nodefix = NULL;
	vertexfix = NULL;
	numnodefix = numvertexfix = 0;
	currentnodefix = currentvertexfix = -1;

	SHarray = NULL;
	memset(SHcontainer, 0, sizeof(SHcontainer));
	last_player_sector = -1;

	linkSectors = NULL;
	memset(reverbThing, 0, sizeof(reverbThing));

	// Set the skill bit
	ge_inf.skill  = gameskill < sk_medium ?
			MTF_EASY : gameskill >= sk_hard ?
			MTF_HARD : MTF_NORMAL;

	Init_SlopeArrays();

	// v220-4 Init the scroll smooth array
	memset(&scsm, 0, sizeof(scsm_t));
	
	// Clear brain data.
	memset(braintargets, 0, sizeof(braintargets));
	numbraintargets = 0;
	braintargeton = 0;
	braineasy = 0;
	
	memset(&ma_inf, 0, sizeof(maininfo_t));	// info collated during analysis

	G_GetFriendStats();
	
	memset(th7400 , 0, sizeof(th7400));
	memset(th7700 , 0, sizeof(th7700));
	memset(th7800 , 0, sizeof(th7800));
	memset(th7900 , 0, sizeof(th7900));
	memset(redir77, 0, sizeof(redir77));
	memset(&T78_disable, 0, sizeof(T78_disable));
	memset(deathmonitor, 0, sizeof(deathmonitor));
	memset(t79list, 0, sizeof(t79list));

	// Clear the game data/player parm arrays
	memset(&gdt, 0, sizeof(gamedata_t));
	memset(&ppt, 0, sizeof(plrparm_t));

	// v220-35 wadani reset
	memset(&wadani, 0xff, sizeof(wadani));
	wadani.mobj = NULL;
	
	// v220-35
	numdoorspots = 0;

	mapcheat = 0;
	ma_inf.mapchange = 1;

	P_SetMapID(new_mapid, *gameepisode, *gamemap);

	// Modified to work properly (see also cht_WarpFunc and CCmdCheatWarp)
	// b211-0 Revised
    if ( (map_lumpnum = W_CheckNumForName(new_mapid)) < 0)
	{
		char msg[64];

		sprintf(msg, "ERROR: %s not found: defaulting to ", new_mapid);

		*gameepisode = 1;
		*gamemap = 1;

		P_SetMapID(new_mapid, *gameepisode, *gamemap);

		if ( (map_lumpnum = W_CheckNumForName(new_mapid)) < 0)
			Con_Error("ERROR: map %s does not exist", new_mapid);

		strcat(msg, new_mapid);

		Sys_MessageBox(msg, true);
	}

	if (gamemode == shareware && lumpinfo[map_lumpnum].handle != IWADhandle)
	{
		Con_Error(
			"Under the iD software license 3rd party maps cannot be run using\n"
			"the shareware iwad. To do so please first purchase the game from;\n"
			DOOMWADSUPPLIER);
	}

	if ((gamenum = (*gameepisode)*10 + (*gamemap)) < 1 || gamenum > 99)
		Con_Error("Invalid game num %i", gamenum);

	MI_VerifyMapInfo(new_mapid);	// Sky parms also reset here
	strcpy(currentMapid, new_mapid);
	ma_inf.map_lumpnum = map_lumpnum;

	current_pwad = hasPWADlevel[gamenum];

	AS_ResetMobjInfo();
	
	if ( (R3D_ScriptedMap = AS_CheckSCP_DEFN(new_mapid)) )
	{
		static	char headerMessage[64];

		if (!R3D_edit)
		{
			if (!loadModels)
				Con_ErrorD("Scripted maps must have models enabled");
		}

		sprintf(headerMessage, "SCRIPTED MAP ERROR REPORT for %s\n\n", new_mapid);

		SetProgHeader(headerMessage);
	}

	if (wasScriptedMap^R3D_ScriptedMap)
		Def_SetThingDimensions();

	wasScriptedMap = R3D_ScriptedMap;

	scriptedMapError = 0;

	pr_OFinsert("\nLOADING: %s\n", new_mapid);

	totalkills = totalfriends = totalitems = totalsecret = 0;
	
	if (ge_inf.first_map)
		ge_inf.first_map = !(mapinfo->R3Dflags & R3Dhasmapinfo);

	ma_inf.iwadmap = lumpinfo[map_lumpnum].handle == IWADhandle;

	show_wadname = mainIwadname;

	// v220-4 check type, if a pwad, against external list
	if (current_pwad)
		Check_Modifier(new_mapid, lumpinfo[map_lumpnum].modidx);

	if (ma_inf.iwadmap)
	{
		// Is this an original doom E?M8 iwad map?
		// This is used to identify whether MT_SPIDER or MT_CYBORG
		// sounds should be played with no attenuation.
		ma_inf.mapEXM8 = !commercial_game && *gamemap == 8;
	}
	else if (!_stricmp("id Software", mapinfo->author))
	{
		// If a pwad lump and author not set then clear the default
		mapinfo->author[0] = 0;
	}

	// For development only
	if (ge_inf.devmap)
	{
		if (ArgExists("-xlnodes"  )) ma_inf.modifier |=  MODF_MODNODES;
		if (ArgExists("-noxlnodes")) ma_inf.modifier &= ~MODF_MODNODES;
	}

	PH_InitXLines();

	leveltime = actual_leveltime = 0;

	pr_OFdirect ("WAD CHECKS\n");

	// v220-3 check for behavior and if found abort
	if (   map_lumpnum + ML_BEHAVIOR < numlumps
		&& !strnicmp(W_LumpName(map_lumpnum + ML_BEHAVIOR), "BEHAVIOR",8))
	{
		Con_Error("ERROR: %s has a BEHAVIOR lump.\n"
				  "This type is not supported by Risen3D", new_mapid);
	}

	if (R3D_edit && lumpinfo[map_lumpnum + ML_LINEDEFS].handle == IWADhandle)
	{
		if (!Sys_CriticalMessage("IWAD editing is not allowed.\n"
			"Any changes made cannot be saved.\n", true))
		{
			Sys_Quit(false);
		}

		ma_inf.iwadedit = true;
	}

	// Pre init some required info
 	P_CheckThings15500(map_lumpnum + ML_THINGS);

	// Load map data in following order
	totalnum_segs = 0;
	GLVertStart = 0;

	memset(dhdr, 0, sizeof(dhdr));
	NFnumlinedefs = NFnumsidedefs = NFnumvertexes = NFnumsectors = 0;

	sprintf(analysisfile, "%sR3Danfiles\\%s%s.R3Dan", ddRuntimeDir.path, analysename, new_mapid);

	// ============= v220-6 Internal nodes build changes =========

	timecache[ITC_NODESBUILD] = 0;

	// v220-6 build or fetch the gl nodes
	// v220-14 modified to allow option of using zdbsp
	{
		int		fileFactor = 16;
		int		useGLBSP = false;
		int		file_exists;
		long	file_time_write;
		float	nodesBuildTime = 0.f;
		char	*intempFname;
		char	*outtempFname;
		uint	checksum;
		DFILE	*fin = NULL;

		// -------------------------------

		// get the date the map we're dealing with was created
		// Note this must reference the actual wad being loaded as
		// this may be one that takes precedence over the main wad
		file_time_write = RetrieveFileDateByModidx(lumpinfo[map_lumpnum].modidx, &checksum);

		intempFname  = (char*)W_Calloc(512, STATICID, 0);
		outtempFname = (char*)W_Calloc(512, STATICID, 0);

		sprintf(nodefile, "%sgwa\\%s.%s", ddRuntimeDir.path, analysename, new_mapid);
		sprintf(intempFname , "%sgwa\\intemp.wad" , ddRuntimeDir.path); 
		sprintf(outtempFname, "%sgwa\\outtemp.wad", ddRuntimeDir.path);

		if (ma_inf.modifier & MODF_FACTOR12)
			fileFactor = 12;
		else if (ma_inf.modifier & MODF_FACTOR32)
			fileFactor = 32;
		else
			ma_inf.modifier |= MODF_FACTOR16;

		if (R3D_edit)
			ma_inf.modifier |= MODF_NOPRUNE;

		// Set any development override conditions
		if ((ma_inf.modifier & (MODF_ZDNODG|MODF_ZDNODO)) && (ge_infmod & (MODF_ZDNODG|MODF_ZDNODO)) )
		{
			ge_infmod &= ~(MODF_ZDNODG|MODF_ZDNODO);
			Sys_MessageBox("ge_inf.modifier overruled by R3D_modify", false);
		}

		ma_inf.modifier |= ge_infmod;

		useGLBSP = checkMapType(current_pwad);
		file_exists = (_access(nodefile, 4)) != -1;

		if (useGLBSP && ma_inf.modifier & (MODF_ZDNODG|MODF_ZDNODO))
		{
			ma_inf.modifier &= ~(MODF_ZDNODG|MODF_ZDNODO);

			if (R3D_edit)
				pr_OFdirect("EDIT MODE: zdbsp requested but reverting to glbsp\n");
			else
				Sys_MessageBox("ZD nodes request ignored. Glbsp must be used with this map", true);
		}

		if (   ma_inf.modifier & MODF_GWAONLY
			&& ma_inf.modifier & (MODF_ZDNODG|MODF_ZDNODO))
		{
			ma_inf.modifier &= ~(MODF_ZDNODG|MODF_ZDNODO);
			pr_OFdirect("Only glbsp can be used with gl only\n");
		}

		edn.data = NULL;

		if (file_exists)
		{
			R3D_ReadNodesFile(nodefile, true);

			if (  !edn.data
				|| edn.modifier != (ma_inf.modifier & MODF_NODEMASK)
				|| edn.filetime != file_time_write)
			{
				if (edn.data) { M_Free(edn.data); edn.data = NULL; }
				remove(nodefile);
			}
		}

		if (edn.data)
		{
			if ( !(fin = F_OpenFileForWrite(outtempFname, "b")) )
				Con_Error("Failed to open nodes file\n%s", outtempFname);

			pr_OFinsert("READING NODES file: %s\n", nodefile);

			pr_OFinsert("File created using %s: factor %i: gl only? %s: pruned? %s\n",
				edn.modifier & MODF_ZDNODG ? "ZDBSP" : edn.modifier & MODF_ZDNODO ? "ZDBSP & GLBSP" : "GLBSP",
				fileFactor,
				edn.modifier & MODF_GWAONLY ? "YES" : "NO",
				edn.modifier & MODF_NOPRUNE ? "NO" : "YES");

			// write the decompressed nodes data to the outfile
			if (!F_FileWrite(edn.data, edn.length, fin, -1))
			{
				remove(nodefile);
				Con_ErrorD("Failed to recreate nodes file\nFIX: restart");
			}

			F_Close(fin);
			M_Free(edn.data);

			if ( !(fnode = F_OpenFile(outtempFname, "b")) )
				Con_Error("Failed to open nodes file\n%s", outtempFname);
		}
		else
		{
			struct _timeb	timebuffer;
			float	buildtime;
			bool	noprune;

			if ( !(fin = F_OpenFileForWrite(intempFname, "b")) )
				Con_Error("Failed to create nodes file\n%s", outtempFname);

			// Create input wad.
			//
			// When building normal and gl nodes the segs, ssectors and nodes are not needed.
			// They have been left, however, in case only gl nodes are wanted

			dumpLUMP(fin, 1, ML_LABEL);	// mapname
			dumpLUMP(fin, 1, ML_LINEDEFS);
			dumpLUMP(fin, 1, ML_SIDEDEFS);
			dumpLUMP(fin, 1, ML_VERTEXES);
			dumpLUMP(fin, 0, ML_SEGS);
			dumpLUMP(fin, 0, ML_SSECTORS);
			dumpLUMP(fin, 0, ML_NODES);
			dumpLUMP(fin, 1, ML_SECTORS);
			dumpLUMP(fin, 1, -1);	// create wad from lumps

			F_Close(fin);
			
			// Nodes are normally "pruned" in that vertexes are merged,
			// empty sectors removed etc. The former is required for analysis
			// especially for checking for closed sectors without the over-
			// head of comparing vertex co-ords. In edit mode, however, the
			// original pwad must not (or at least should not) be changed
			// other than texture co-ordinates for which the mode exists.
			noprune = (ma_inf.modifier & MODF_NOPRUNE) ? 1 : 0;

			if (ma_inf.modifier & MODF_GWAONLY)
				pr_OFrepeat("\n=====  R3DGLBSP NODES BUILD GL ONLY", "=", 27, 1);
			else if (ma_inf.modifier & MODF_ZDNODO)
				pr_OFrepeat("\n=====  R3D ZDBSP STANDARD NODES BUILD", "=", 24, 1);
			else if (ma_inf.modifier & MODF_ZDNODG)
				pr_OFrepeat("\n=====  R3D ZDBSP NODES BUILD", "=", 33, 1);
			else
				pr_OFrepeat("\n=====  R3DGLBSP NODES BUILD", "=", 34, 1);

			pr_OFinsert("Using factor %i, prune? %s\n", fileFactor, noprune ? "NO" : "YES");
	
			if (ma_inf.modifier & MODF_ZDNODO)
			{
				Con_InitProgress("WAIT building zdbsp standard nodes", 100); // $progress XA Start

				Con_TimingStart();
				if (zdbsp_main(new_mapid, 0, fileFactor, noprune, intempFname, intempFname))
					Con_Error("ERROR: failed to build zd standard nodes\n");
				buildtime = Con_TimingEnd(NULL);

				nodesBuildTime = buildtime;

				pr_OFinsert("  Time to build standard nodes: %7.3f sec\n", buildtime);

				ma_inf.zdnormal = true;	// allow gNd4 nodes when only building gl nodes

				Con_InitProgress("WAIT building glbsp gl nodes", 100); // $progress XA Start
				pr_OFrepeat("\n=====  R3D GLBSP GL NODES BUILD", "=", 30, 1);

				Con_TimingStart();
				if (bsp_BuildNodes(0, fileFactor, noprune, intempFname, outtempFname))
				{
					Con_Error("Failed to build gl nodes for %s", new_mapid);
				}
				buildtime = Con_TimingEnd(NULL);
				pr_OFinsert("  Time to build glbsp gl nodes: %7.3f sec\n", buildtime);
			}
			else if (ma_inf.modifier & MODF_ZDNODG)
			{
				Con_InitProgress("WAIT building zdbsp game nodes", 200); // $progress XA Start

				Con_TimingStart();

				// Support for R3D_NODEfix
				if (!(ma_inf.modifier & MODF_GWAONLY) && !R3D_edit)
				{
					WA_CheckAnalyseFile("R3D_NODEfix", 2);
				}

				if (zdbsp_main(new_mapid, 1, fileFactor, noprune, intempFname, outtempFname))
					Con_Error("ERROR: failed to build zd nodes\n");

				buildtime = Con_TimingEnd(NULL);
				pr_OFinsert("  Time to build zdbsp standard and gl nodes: %7.3f sec\n", buildtime);
			}
			else
			{
				if (R3D_edit)
					Con_InitProgress("WAIT building glbsp edit nodes", 100); // $progress XA Start
				else
					Con_InitProgress("WAIT building glbsp game nodes", 100); // $progress XA Start

				Con_TimingStart();

				// Support for R3D_NODEfix
				if (!(ma_inf.modifier & MODF_GWAONLY) && !R3D_edit)
				{
					WA_CheckAnalyseFile("R3D_NODEfix", 2);
				}

				if (bsp_BuildNodes(
						!(ma_inf.modifier & MODF_GWAONLY),	// 0 gl only, 1 build gl and normal
						fileFactor, noprune, intempFname, outtempFname)
					)
				{
					Con_Error("Failed to build gl nodes for %s", new_mapid);
				}

				buildtime = Con_TimingEnd(NULL);
				pr_OFinsert("  Time to build glbsp standard and gl nodes: %7.3f sec\n", buildtime);
				if (useGLBSP) pr_OFinsert("NOTE: internal R3D map correction applied\n");
			}

			_ftime( &timebuffer );
			nodetime = timebuffer.time;

			nodesBuildTime += buildtime;

			timecache[ITC_NODESBUILD] = stat_nodestime = nodesBuildTime;

			pr_OFrepeat(NULL, "=", 61, 2);

			if ( !(fnode = F_OpenFile(outtempFname, "b")) )
				Con_Error("Failed to open nodes file;\n%s", outtempFname);

			Con_HideProgress();			// $progress XA End

			if (nodesBuildTime > 1.0f || R3D_edit)
			{
				int		error;

				F_Seek(fnode, 0, SEEK_END);
				edn.length = F_Tell(fnode);

				edn.data = (byte*)M_Malloc(edn.length, 0);
				F_Seek(fnode, 0, SEEK_SET);

				if (F_Read(edn.data, edn.length, fnode) != edn.length)
					Con_Error("Failed to read from;\n%s", outtempFname);

				edn.nodetime = nodetime;
				edn.filetime = file_time_write;
				edn.modifier = ma_inf.modifier & MODF_NODEMASK;
				edn.edit = R3D_edit;

				error = R3D_WriteNodesFile(nodefile);

				if (error) pr_OFinsert("Failed to make nodes file\n   %s\n", nodefile);
			}

			if ((_access( analysisfile, 4 )) != -1)
				remove(analysisfile);
			if ((_access( intempFname, 4 )) != -1)
				remove(intempFname);
		}

		W_Free(intempFname , STATICID, 0);
		W_Free(outtempFname, STATICID, 0);

		F_Seek(fnode, 0, SEEK_SET);
		F_Read(&nbhdr, sizeof(nbhdr), fnode);
		F_Seek(fnode, nbhdr.dir_start, SEEK_SET);

		for (i = 0; i < nbhdr.num_entries; i++)
		{
			F_Read(&rhdr, sizeof(lump_header_t), fnode);

			for (k = 0; datanames[k]; k++)
			{
				if (!strncmp(rhdr.name, datanames[k], strlen(datanames[k])))
				{
					dhdr[k].pos = rhdr.start;
					dhdr[k].length = rhdr.length;
				}
			}
		}
	}

	pr_OFdirect("Loading level.\n");
	
	unknownname.count = wallfornamecount = 0;	// v220-35

	P_LoadVertexes(&dhdr[VTX_PTR], &dhdr[GLVTX_PTR]);
	P_LoadSectors (&dhdr[SECT_PTR]);

	// v220-35
	P_SetSidePtrArray(&dhdr[LINE_PTR], &dhdr[SIDE_PTR]);
	P_LoadSideDefs2();	
	P_LoadLineDefs2();
	
	if (ma_inf.nodesFormat == 5)
	{
		P_LoadSubsectorsGL5(&dhdr[GLSSEC_PTR]);
		P_LoadNodesGLE(&dhdr[GLNODE_PTR]);
		P_LoadSegsGL5(&dhdr[GLSEG_PTR]);
	}
	else if (ma_inf.nodesFormat == 4)
	{
		P_LoadSubsectorsGL4(&dhdr[GLSSEC_PTR]);
		P_LoadNodesGLE(&dhdr[GLNODE_PTR]);
		P_LoadSegsGL4(&dhdr[GLSEG_PTR]);
	}
	else if (ma_inf.nodesFormat == 2)
	{
		P_LoadSubsectors(&dhdr[GLSSEC_PTR]);
		P_LoadNodes(&dhdr[GLNODE_PTR]);
		P_LoadSegsGL2(&dhdr[GLSEG_PTR]);
	}
	else if (!ma_inf.nodesFormat)
		Con_ErrorD("NodesFormat unknown");

	if (totalnum_segs != (uint)numsegs)
		Con_Error("Bad nodes build: segs needed %i != segs created %i\nRebuild nodes",
		totalnum_segs, numsegs);

	ssinitial = (int *)W_Calloc(numsubsectors* sizeof(int), STATICID, 324);

	P_GroupLines();

	WA_CheckAnalyseFile("R3D_presetVars", 0);

	if (fnode)
	{
		char fname[512];
		
		F_Close(fnode); fnode = NULL; edn.data = NULL;

		// v220-10 remove outtemp.wad as some people like to play
		// around and, although the wad entries comply with wad
		// specifications, missing entries not needed by Risen3D
		// can upset progs that expect them to be there.
		sprintf(fname, "%sgwa\\outtemp.wad", ddRuntimeDir.path);

		if ((_access(fname, 4 )) != -1)
			remove(fname);
	}

	Load_SCP_ZOFF(new_mapid);	// Check for any z-off, light changes with 16000 types
	P_Init_origTEX();

	vtoflags = (byte*)W_Calloc(numvertexes, STATICID, 0); // b211-0
	vtocount = (ushort*)W_Calloc(numvertexes * sizeof(ushort), STATICID, 0); // b211-0

	// - dlBlockLinks initialized
	// - necessary GL data generated
	R_SetupLevel();

	if (!mapHasRun[gamenum])
	{
		PH_CheckIntegrity();
		mapHasRun[gamenum] = 1;
	}

	if (gameskill == sk_nightmare)
	{
		respawnmonsters = true;
	}
	else if (gameskill != sk_nightmare && (ArgExists("-respawn") || respawnmonsters))
	{
int FR_StandardMessage(const char *msg);
		char buff[96];
		
		sprintf(buff,	"WARNING: respawning is active.\n"
						"Do you want this disabled? Y : N\n");

		respawnmonsters = !FR_StandardMessage(buff);
	}

	// v220-4 advise wad has missing texture entries
	if (    !unknownname.printed
		&& ((ArgExists("-listloresfm") && R3D_ScriptedMap && unknownname.count > 0)
		|| (!R3D_ScriptedMap && unknownname.count - unknownname.flat_count > 4)))
	{
		char buff[756];
		
		unknownname.printed = true;

		sprintf(buff,	"WAD ERROR: %s requires %i entries with no corresponding lump.\n"
						"Ensure the correct IWAD is being used.\n"
						"See : %s after exiting.\n",
						new_mapid, unknownname.count, LOGoutfile);

		if (!Sys_CriticalMessage(buff, true))
		{
extern RDBOOL bypassAdvisory;
			bypassAdvisory = false;
			Sys_Quit(0);
		}
	}
	
    bodyqueslot = 0;
	numspechit = 0;
	numsplist = 0;

	// ======== Risen3D analysis calls =============
	
	movsec_list = (int*)W_Calloc((numsectors + 1)*sizeof(int), LEVELID, 87);

	timecache[ITC_ANALYSE] = 0;
	timecache[ITC_SECREVERB] = 0;

	use_analyse_file = !R3D_edit && !ArgExists ("-noAni");

	i = _access( analysisfile, 4 );

	if (!use_analyse_file || i == -1)
	{
		if (i != -1)
		{
			remove(analysisfile);
			pr_OFdirect("Analysis file has been deleted.\n");
		}

		preanalysed = false;
		R3D_MAINAnalyseMap(map_lumpnum + ML_THINGS);
	}
	else if ( (error = R3D_CheckAnalysedFile(analysisfile, nodetime)) > 0)
	{
		pr_OFinsert("Analysis file is incompatible and will be rebuilt [%i].\n", error);
		remove(analysisfile);
		preanalysed = false;
		R3D_MAINAnalyseMap(map_lumpnum + ML_THINGS);
	}
	else
	{
		R3D_MainSetupInit(map_lumpnum + ML_THINGS, false);

		if ((error = R3D_ReadAnalysedFile(analysisfile)) != 0)
		{
			remove(analysisfile);
			Con_Error("ERROR %i: Bad analyse file %s\n\n"
					  "File has been deleted - please restart",
					  error, analysisfile);
		}

		pr_OFdirect("Using preset analysis file.\n");
		preanalysed = true;
		LS_InitTagLists();	// Update hash table
	}

	if (gamemode != shareware)
	{
		S_InitMUSScript(new_mapid);
	}
	
	// set up world state
	AM_Init ();		// reset the automap arrays
	GL_ClearAllHudIgnoreFlags();
	P_SpawnSpecials();
	P_CreateDynLightStructures(false);	// v220-31
	R_PrecacheLevel(CACHE_GRAPHICS);

	if (!preanalysed)
	{
		extern float	stat_analysetime;
		float starttime = Sys_GetSeconds();

		// Stuff done following graphics cacheing
		LS_PostSetupRoutines();
		PH_CheckMobjFitsSector();

		// Must be done only after all other analysis is complete
		FW_SetFogwallGroups();	// v220-3

		timecache[ITC_ANALYSE] += Sys_GetSeconds() - starttime;

		// Must follow R_SetupLevel but precede Init_ActionScript
		if (!R3D_edit) S_CalcSectorReverbs();

		stat_analysetime = timecache[ITC_ANALYSE];

		if (   use_analyse_file
			&& (stat_analysetime + timecache[ITC_SECREVERB] > 3.0f) )
		{
			R3D_WriteAnalysedFile(analysisfile);
		}
	}

	Set_SsecReferences();
	checkOwners();
	H_SetTHINGsLODstate();	// v220-9

	W_Free(ssinitial, STATICID, 0); ssinitial = NULL;

	R3D_MAINfreeArrays();

	W_Free(vtoflags , STATICID, 0); vtoflags = NULL;
	W_Free(vtocount , STATICID, 0); vtocount = NULL;
	W_Free(asyflags , STATICID, 0); asyflags = NULL;
	W_Free(addflags , STATICID, 0); addflags = NULL;

	if (dbfile_active) Sys_Quit(false);

	// v220-7 check line specials
	if (!ArgCheck("-notag_fix"))
	{
		sector_t	*bs;
		line_t		*ld;
		int			i, v, tag;

		for (i = 0, ld = lines; i < numlines; i++, ld++)
		{
			if (   !lines[i].tag
				&& (v = lines[i].special)
			// validate special
				&& (v == 114		// invalid type
				||  v == 99			// locked door types
				|| (v >= 133 && v <= 137))
			// validate backsector)
				&& (bs = ld->backsector)
				&& bs != ld->frontsector
				&& bs->floorheight == bs->ceilingheight
				&& bs->linecount > 2
				&& bs->specFlags & SF_SECCLOSED)
			{
				RDBOOL		found = false;

				if (v == 114)
				{
					pr_OFinsert("Wrong special used with L%i - changed from 114 to 117 [bs %i fl %.8X]\n", ld - lines, bs - sectors, bs->specFlags);
					ma_inf.maperror++;
					ld->special = 117;
				}
				else	// locked door type
				if ( ((tag = ld->backsector->tag) || ((tag = getNextTag()) > 0)) )
				{
					ld->tag = tag;

					if (!ld->backsector->tag)
					{
						pr_OFinsert("Error: L%i and S%i fl %.8X had no tag, They have now been tagged %i\n", ld - lines, bs - sectors, bs->specFlags, tag);
						ld->backsector->tag = tag;
					}
					else
						pr_OFinsert("Error: L%i had no tag, This has now been tagged to the backsector tag\n", ld - lines);

					LS_InitTagLists();				// Update hash table
					ma_inf.maperror++;
				}
			}
		}
	}

	// =========== End Risen3D analysis

	// R3D - remove mobjs depending on skill level
	PH_CheckSpawnedbySkill();
	if (!numthings) Con_ErrorD("Map has no THINGS");

	LG_FakeMissing();
	TP_CheckTPSpawnedSectors(true);	// v220-9 reset as ssec sectors may have changed

	if (!preanalysed)
		pr_OFinsert("\n=== End analysis [%i map error%s found]\n\n", ma_inf.maperror, plural(ma_inf.maperror));

	PH_setThingAngles(new_mapid, current_pwad);	// v220-9
	P_SpawnPlayers();

	if ( !(ppt.viewplayer = mbf_players[0].plr)
		|| ppt.viewplayer->mo != ppt.viewplrmo)
	{
		Con_Error("Could not initialise console player");
	}

	if (!sslines) R3D_Set_sslinesArray();

	if (sslines)
	{
		ma_inf.thingheight = LS_CheckThingsHeight();	// R3D - check spawned things heights
		FreeSSLines();
	}

	ST_Start ();		// wake up the status bar
	HU_Start ();		// wake up the heads up text

	TS_ResetWater();	// R3D - Reset water vars

	if (ma_inf.midtexerror)
	{
		// v220-8 report 1's lines with missing mid tex (if in visible part of the map)
		//        Must not precede TP_CheckTPSpawnedSectors

		pr_OFinsert(
			"\nMID TEX ERROR: %i visible 1's line%s no midtex.\n"
			"   Tex has been derived from a connected 1's line\n%s",
			ma_inf.midtexerror, ma_inf.midtexerror > 1 ? "s have" : " has",
			R3D_edit ? "" : "   Using edit mode will list the affected lines\n");

		if (R3D_edit)
		{
			for (i = 0; i < numlines; i++)
			{
				if (lines[i].scpFlags & SCPF_REPERROR)
					pr_OFinsert(" L%i : %s used\n", i, texArray[sides[lines[i].sidenum[0]].midtexture]->name);
			}
			pr_OFinsert("\n");
		}
	}

	TS_CheckLinkedModels();
	R_ResetVisSprites(numthings);
	PH_SpriteOffsetAdjust();

	Init_ActionScript(new_mapid);

	LS_CheckPlayerFriendliness();

	R_SetupSky();	// v220-5 moved here

	R_PrecacheLevel(CACHE_SPRITES);
	MB_Init(false);	// Initialise amodel doors
	P_CheckInheritance5505();	// v220-31

	P_InitParticledefSectors();
	
	PH_CheckHoldState();

	compat[c_resurrect] = (byte)(ArgExists("-c_resurrect_on") != 0);

	PH_CheckStress();	// $stresschange

	autoAim = false;
	gdt.fpic = sectors[0].floorpic;	// preset to a valid texture

	// R3D functions added
	PH_CheckMobjShadows();
	S_CheckLineTriggeredAmbients();
	LS_SetSectorSlopeFlags();
	W_CheckHeap(2);
	LS_PrintGeneralInfo(preanalysed);

	PH_PrintModelStats();

	if (ge_inf.oldgamenum != gamenum)
	{
		ma_inf.startlevelmusic = true;

		currentMapMusLumpnum = -1;

		S_SetLISTge_inf(W_GetMapIndex(mapinfo->id), mapinfo->muslump);

		if (   ge_inf.UseList & 2
			&& (cfg.usemapmusic && R3D_ScriptedMap)
			|| (mapinfo->R3Dflags & R3Dmusiconce) )
		{
			pr_OFinsert("LIST music override: %s\n", mapinfo->R3Dflags & R3Dmusiconce ? "mapmusic once" : "by user");
			ge_inf.UseList = 1;
		}

		ge_inf.oldgamenum = gamenum;
	}
	
	oldge = *gameepisode, oldgm = *gamemap;

	TS_StartListSounds();

	if (brief_disabled < 0)
		brief_disabled = 1;
	else
		brief_disabled = 0;

	P_SetupWaterSectors();		// v220-5 water modulation
	R_PrecacheNoAlphaTex();		// v220-6 pre-cache unmasked variants of masked textures 

	PH_SetSSECMaxDimensions();
	PH_SetXLines();				// v220-25 if needed setup up xline stuff
	PH_SetBlockOcclude();		// v220-16 set blocking and occlude flags

	// Set the mapinfo parameters
	MI_ActionMapInfo(new_mapid);	// Fog values initialised here

	shortListQuads = R3D_slopetypes || outsidefog;

	ma_inf.world_starttime = 70 + rand() % 105;

	S_InitAMBIENTS();	// v220-6 moved here
	Rend_LsiArrayInit();
	R3D_SetParms(new_mapid);
	xydelay = 20;
	flushMouse();
	ma_inf.checkPlayer = true;
	rndindex = 0;
	PH_CheckTorchGlow();
	PH_ThingFlags(new_mapid);
	P_SetupDynlightList();		// v220-31

	// v220-31 SCP_ZOFF support
	if (scpofftagsset) PH_SetSCPOffsets();
	// v220-31 Precache any models that are to be spawned but are not yet cached
	if (R3D_ScriptedMap) SCP_precacheModels();

	PH_CheckSSECrefs();				// v220-33
	if (R3D_linemap)
		PH_CheckExternal3DLiftSector();	// v220-33
	PH_CheckYoffInheritance();		// v220-35
	P_InitLineHit(numlines);

	if (ma_inf.checkforcesky)
	{
		for (i = 0; i < numlines; i++)
		{
			if (lines[i].ovflags & OVF_NOOCCLUDE)
			{
				lines[i].ovflags &= ~OVF_NOOCCLUDE;
				lines[i].ovflags &= ~OVF_NOSKYWALL;
				lines[i].frontsector->specFlags &= ~SF_NOOCCLUDE;
			}
		}
	}

	pu_timer = pu_counter = 0;		// v220-35
	if (ArgCheck("-pramb")) printoutAmbients();
}
