//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Chi Hoang, Lee Killough,
//   Jim Flynn, Randy Heit, Rand Phares, Ty Halderman, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_spec.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
//  Limited support added for ZDOOM ANIMDEFS lump
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 2 17/Jun/2008 gmj (b220-2)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code, BOOM v2.01,
//           ZDOOM v1.22 December 12, 1999 (dis. marked v1.23beta33)
//
//
// DESCRIPTION:
//  Implements special effects:
//  Texture animation, height or lighting changes
//  according to adjacent sectors, respective
//  utility functions, etc.
//  Line Tag handling. Line and Sector triggers.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "doom/mbf_data.h"
#include "doom/p_local.h"
#include "doom/p_start.h"
#include "doom/g_game.h"
#include "doom/tables.h"
#include "doom/p_spec.h"
#include "doom/d_englsh.h"

#include "con_console.h"
#include "dd_wad.h"
#include "m_misc.h"
#include "p_maputil.h"
#include "p_sight.h"
#include "r_things.h"
#include "r_data.h"
#include "r_main.h"
#include "net_main.h"
#include "s_main.h"
#include "s_mus.h"
#include "rend_sky.h"
#include "sys_timer.h"
#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "r3d_sfx.h"
#include "r3d_slopes.h"

// MACROS ------------------------------------------------------------------

#define MAXANIMS 32					// no longer a strict limit -- killough
#define MAX_ANIM_FRAMES 32

#define ZOOMXHAIRSIZE	((int)(1 + (setFOV - fieldOfView)/2))	// R3D


// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

void P_InitSwitchList(void);
int P_Activate1_1801x_line(mobj_t* thing,line_t *line);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void P_SpawnScrollers(void);							// killough
static void P_SpawnPushers(void);							// phares 3/20/98

static void P_SetupFakeTeleport(void);
static void	P_SpawnPhasedLightingSequence(sector_t *sec);
static void	P_SpawnPhasedSectorLight(sector_t *sec);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern linklist_t	link_list;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

float		fieldOfView	 = 90.0f;
float		lodFactor;

// ZOOM stuff - R3D
float		setFOV		 = 90.0f;
float		targetFOV;
int			inZoomMode;

int		anim_group_count = 1;
int		bloodglowcurrent_state;

// v220-5 skyvars (rotate, scroll)
int		numskyvar;
int		maxnumskyvar = 0;
skyvar_t *skyvars = NULL;

float	skyAngle;
float	skyAngleGlobal;

// v220-6 smooth scroll
scsm_t	scsm;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

typedef struct
{
	short 	basepic;
	short	numframes;
	byte 	istexture;
	byte	uniqueframes;
	byte	countdown;
	byte	curframe;
	byte 	speedmin[MAX_ANIM_FRAMES];
	byte	speedmax[MAX_ANIM_FRAMES];
	short	framepic[MAX_ANIM_FRAMES];
} anim_t;


static anim_t*	lastanim = NULL;
static anim_t*	anims = NULL;
static int		maxanims = 0;
static RDBOOL	showflatsset;

// CODE --------------------------------------------------------------------

static anim_t *checkAnimArray(anim_t *current)
{
	if (!anims || current - anims >= maxanims)
	{
		size_t newmax = maxanims + MAXANIMS;
		size_t temp = anims ? current - anims : 0;

		anims = (anim_t*)W_Realloc(anims, newmax * sizeof(anim_t), STATICID, 230);
		maxanims = newmax;
		return &anims[temp];	// Not current as mem base address can change
	}
	return current;
}

// ParseAnim:
//  With guidance from ZDOOM v1.23 ParseAnim() code (with thanks)
static void ParseAnim(RDBOOL istex)
{
	anim_t	dummy, *place;
	short	picnum;
	byte	min, max;
	char	*nextstr = SR_NextSTR(true);
	int		nextnum;

	picnum = (short)(istex ? R_CheckTextureNumForName (nextstr)
		: W_CheckFlatNumForName (nextstr));// - first texture/flat;

	if (picnum == -1)
	{
		// Base pic is not present but script must still be parsed
		place = &dummy;
	}
	else
	{
		for (place = anims; place < lastanim; place++)
		{
			if (place->basepic == picnum && place->istexture == istex)
			{
				break;
			}
		}

		if (place == lastanim)
		{
			lastanim++;
			place = checkAnimArray(place);
			lastanim = place + 1;
		}
	}

	place->uniqueframes = true;
	place->curframe = 0;
	place->numframes = 0;
	place->basepic = picnum;
	place->istexture = istex;
	memset (place->speedmin, 1, MAX_ANIM_FRAMES * sizeof(*place->speedmin));
	memset (place->speedmax, 1, MAX_ANIM_FRAMES * sizeof(*place->speedmax));

	while ( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, "allowdecals")) continue;
		
		if (_stricmp(nextstr, "pic"))
		{
			SR_KeepString(); break;
		}

		if (place->numframes >= MAX_ANIM_FRAMES)
		{
			SR_SCPErr("Animation has too many frames");
		}

		min = max = 1;
		
		nextnum = SR_NextNum();
		place->framepic[place->numframes] = nextnum + picnum - 1;

		nextstr = SR_NextSTR(true);

		if (!_stricmp(nextstr, "tics"))
		{
			nextnum = SR_NextNum();
			
			if (nextnum <   0) nextnum =   0;
			if (nextnum > 255) nextnum = 255;

			min = max = nextnum;
		}
		else if (!_stricmp(nextstr, "rand"))
		{
			nextnum = SR_NextNum(); min = nextnum >=   0 ? nextnum :   0;
			nextnum = SR_NextNum(); max = nextnum <= 255 ? nextnum : 255;
		}
		else
		{
			SR_SCPErr("Unspecified duration");
		}

		place->speedmin[place->numframes] = min;
		place->speedmax[place->numframes] = max;
		place->numframes++;
	}

	if (place->numframes < 2)
		SR_SCPErr("Animation requires 2 frames minimum");

	place->countdown = place->speedmin[0];
}

char *getanimType(int flag)
{
	int fl = flag & FLT_SPLASH;

	if (fl == FLT_WATER) return "water";
	if (fl == FLT_LAVA ) return "lava";
	if (fl == FLT_SLIME) return "slime";
	if (fl == FLT_BLOOD) return "blood";
	if (fl == FLT_NUKE ) return "nuke";
	if (fl == FLT_BLKWATER) return "blackwater";
	if (fl) return "error - multiset";
	return "none";
}

//
// P_LoadANIMDEFS
//
// Limited support for ZDOOM ANIMDEFS lump
//
static void P_LoadANIMDEFS ()
{
	int		lump, lastlump = 0;
	char	*nextstr;

	while ((lump = W_FindScriptLump ("ANIMDEFS", &lastlump)) != -1)
	{
		SR_LoadLump(lump, "ANIMDEFS");

		while ( (nextstr = SR_NextSTR(false)) )
		{
			RDBOOL type = !_stricmp(nextstr, "texture");

			if (type || !_stricmp(nextstr, "flat"))
			{
				ParseAnim(type); continue;
			}

			if (!_stricmp(nextstr, "switch")) continue;	// unsupported

			if (!_stricmp(nextstr, "warp"))	// unsupported but get strings
			{
				nextstr = SR_NextSTR(true);

				if (_stricmp(nextstr, "flat") && _stricmp(nextstr, "texture"))
					SR_SCPErr(NULL);
					
				SR_NextSTR(true);
			}
		}

		SR_EndScript();
	}
}

static int isANIM(int picnum)
{
	char *name = lumpinfo[picnum].name;

	uint len = strlen(name) - 1;

	if (   len < 3
		|| name[len] < '0'
		|| name[len] > '9')
	{
		return 0;
	}

	if (   name[len - 1] >= '0'
		&& name[len - 1] <= '9')
	{
		return atoi(name + len - 1);
	}

	return atoi(name + len);
}

void subfn_SetAnimType(int picnum, int tex)
{
	if (tex)
	{
		if(picnum < 0 || picnum >= numtextures) return;
		texArray[picnum]->anim_group = anim_group_count;
		texArray[picnum]->txf_flags |= TXF_ANIMPIC;
	}
	else if (picnum < 0 || picnum >= numlumps)
	{
		Con_Error("ERROR: %i invalid picnum", picnum);
	}
	else
	{
		char	*name = lumpinfo[picnum].name;
		flat_t	*flat = name ? R_GetFlat(picnum) : NULL;
		
		if (!flat) Con_Error("ERROR: picnum %i; lumpname \"%s\" unrecognised flat name", picnum, name);

		flat->group = anim_group_count;
		flat->flflags |= TXF_ANIMPIC;
	}
}

#define TYPE_UNSET -1
#define TYPE_WATER	0
#define TYPE_LAVA	1
#define TYPE_SLIME	2
#define TYPE_BLOOD	3
#define TYPE_NUKAGE	4
#define TYPE_ROCK	5

static int identifyType(int picnum)
{
	char *name = lumpinfo[picnum].name;

	if (!strnicmp(name, "FWATER", 6)) return TYPE_WATER;
	if (!strnicmp(name, "LAVA"  , 4)) return TYPE_LAVA;
	if (!strnicmp(name, "SLIME" , 5)) return TYPE_SLIME;
	if (!strnicmp(name, "BLOOD" , 5)) return TYPE_BLOOD;
	if (!strnicmp(name, "NUKAGE", 6)) return TYPE_NUKAGE;
	if (!strnicmp(name, "RROCK" , 5)) return TYPE_ROCK;

	return TYPE_UNSET;
}

static RDBOOL inPWAD(int picnum)
{
	char *name = lumpinfo[picnum].name;

	return (W_CheckPWADNumForName(name) >= 0);
}

// Only called if all flats are present in the pwad
int identifySplashFlats(int picnum, int type)
{
	flat_t	*flat = R_GetFlat(picnum);
	int		filter, count_128 = 0;

	//----------------------------

	// Try to determine type
	if ( (filter = MakeFilterARGB(picnum, &count_128)) != -1)
	{
		int b = filter & 0xFF;
		int g = (filter & 0xFF00)>>8;
		int r = (filter & 0xFF0000)>>16;

		if (count_128) return TYPE_UNSET;

		if (type == TYPE_ROCK)	// pwad custom tex used in Herian2
		{
			if (   b >=   5 && b <=   6
				&& r >= 141 && r <= 146
				&& g >=  24 && g <=  28)
			{
				return TYPE_LAVA;	// RROCK changes to LAVA
			}
		}

		if (type == TYPE_WATER && b > 40 && r < 9 && g < 9) return TYPE_WATER;

		if (type != TYPE_ROCK && type != TYPE_SLIME)
		{
			if (   (b > 127 && r < 7  && g < 10)	// Pwad custom tex 1
				|| (b > 200 && r < 42 && g < 48))	// ditto
			{
				return TYPE_WATER;
			}
		}

		if (type == TYPE_WATER) return TYPE_UNSET;

		// BLOOD to SLIME
		if (type == TYPE_BLOOD && r > 32 && r < 50 && r > g && g > b) return TYPE_SLIME;

		if (type == TYPE_SLIME)
		{
			if (r < 64 && r > 20 && r > g && g > b) return TYPE_SLIME;	// SLIME
			return TYPE_UNSET;
		}

		if (r && !b && !g) return TYPE_BLOOD;

		if (   (b < 15 && r > 94 && g < 33)		// General (RED dominant)
			|| (b == 45 && r > 61 && r < 78 && g == 45))	// Another pwad custom tex used
		{
			return TYPE_LAVA;
		}

		if (g > r && r > b) return TYPE_NUKAGE;
		if (r > 90 && r > g + 40 && r > b + 40) return TYPE_LAVA;
	}

	return TYPE_UNSET;
}

// v220-10 splash flat management added

typedef struct {
	char	*str;
	int		val;
	int		flag;
} modflat_t;

modflat_t mflat[] = {
	{"[water]"		, MODF_WATER,		FLT_WATER },
	{"[lava]"		, MODF_LAVA,		FLT_LAVA },
	{"[slime]"		, MODF_SLIME,		FLT_SLIME },
	{"[blood]"		, MODF_BLOOD,		FLT_BLOOD },
	{"[nuke]"		, MODF_NUKE,		FLT_NUKE },
	{"[blackwater]"	, MODF_BLKWATER,	FLT_BLKWATER },
	{NULL,		0,	0 }
};

static void setREFanim(int picnum, int flag, int add)
{
	flat_t *fl = R_GetFlat(picnum);

	if (!fl) return;

	fl->flflags &= ~(FLT_SPLASH|FLT_GLOW);

	if (add > 0) fl->flflags |= (flag|FLT_GLOW);

	if (showflatsset)
	{
		pr_OFinsert("%s FLAT %s type %s\n",
			add > 0 ? "SET" : "UNSET",
			lumpinfo[fl->lump].name,
			getanimType(fl->flflags));
	}
}

static void Check_Flats()
{
	RDBOOL	matched;
	char	fname[512], *nextstr;
	int		i, k;

	//------------------------------

	sprintf(fname, "%sR3D_setflats", ddBinDir.path);

	if (!SR_LoadFile(fname, "R3D_setflats", NULL))
	{
		Con_ErrorD("ERROR: the file Risen3D\\bin\\R3D_setflats is missing");
	}

	for (i = 0; mflat[i].str; i++, SR_RewindScript())
	{
		RDBOOL	found = false;
		
		nextstr = SR_NextSTR(false);

		if (_stricmp("Version", nextstr) || SR_NextNum() != 2)
			Con_Error("ERROR: the R3D_setflats file loaded is an old version");
		
		while ( (nextstr = SR_NextSTR(false)) )
		{
			if (!_stricmp(nextstr, mflat[i].str))
			{
				found = true; break;
			}
		}

		if (!found) continue;

		while ( (nextstr = SR_NextSTR(false)) )
		{
			uint	numchar = 0;
			int		num;
			int		flag = 0;
			int		add = 0;
			int		term = 0;

			if (nextstr[0] == '[') { term = 1; break; }

			if (nextstr[0] == '\\') numchar = strlen(nextstr + 1);

			if (numchar && strlen(analysename) >= numchar)
				matched = !_strnicmp(analysename, nextstr + 1, numchar);
			else
				matched = !_stricmp(analysename, nextstr);

			flag = mflat[i].flag;

			nextstr = SR_NextSTR(false);

			if (!_stricmp("SET"  , nextstr)) add =  1;
			else
			if (!_stricmp("UNSET", nextstr)) add = -1;

			if (!add) SR_SCPErr("Operator SET or UNSET is missing");

			nextstr = SR_NextSTR(false);

			if (_stricmp("{", nextstr)) SR_SCPErr("missing {");

			while (nextstr = SR_NextSTR(false))
			{
				if (!_stricmp("}", nextstr)) { term = 1;  break; }

				if (matched)
				{
					int bp;

					if ( (num = W_CheckFlatNumForName(nextstr)) < 0 )
					{
						SR_SCPWarn("\"%s\" is not a valid flat name", nextstr);
						break;
					}
					else if (!pwadHandle(lumpinfo[num].handle))
					{
						SR_SCPWarn("\"%s\" is not a pwad flat name", nextstr);
						break;
					}
					else
					{
						anim_t	*anim, *refan = NULL;

						if (showflatsset) pr_OFinsert("-------- %s\n", nextstr);

						for (anim = anims ; anim < lastanim ; anim++)
						{
							if (!anim->istexture && anim->numframes > 2)
							{
								bp = anim->uniqueframes ? 0 : anim->basepic;

								for (k = bp; k < bp + anim->numframes; k++)
								{
									if (num == (bp ? k : anim->framepic[k]))
									{
										refan = anim; break;
									}
								}
							}

							if (refan) break;
						}

						if (refan)
						{
							for (k = bp; k < bp + refan->numframes; k++)
								setREFanim(bp ? k : refan->framepic[k], flag, add);
						}
						else
						{
							SR_SCPWarn("flat \"%s\" is not animated", nextstr);
						}
					}
				}
			}

			if (!term) SR_SCPErr("Missing }");

			if (matched) continue;	// there may be set and unset entries under this header
		}
	}

	if (showflatsset) pr_OFdirect("  ============== END SPLASH ANIM FLATS\n\n");

	SR_EndScript();
}

// ---------------------------------------------------------------------------

static void P_SetAnimTypeArray(void)
{
	flat_t	*fl;
    anim_t	*anim;
	int		picnum, lnum, type;
    int		bp, count, i;

	bloodglowcurrent_state = r_bloodglow;

	// Check entries and set anim flag
    for (anim = anims ; anim < lastanim ; anim++, anim_group_count++)
    {
		if (anim->uniqueframes)
		{
			for (i = 0; i < anim->numframes; i++)
				subfn_SetAnimType(anim->framepic[i], anim->istexture);
		}
		else for (i = anim->basepic; i < anim->basepic + anim->numframes; i++)
			subfn_SetAnimType(i, anim->istexture);
    }

	if (R3D_edit) return;

    for (anim = anims ; anim < lastanim ; anim++)
    {
		if (!anim->istexture && anim->numframes > 2)
		{
			lnum  = anim->uniqueframes ? anim->framepic[0] : anim->basepic;
			type = identifyType(lnum);

			if (type > TYPE_UNSET && isANIM(lnum) < 8)
			{
				RDBOOL	pwad = inPWAD(lnum);
				
				count = 1;

				if (anim->uniqueframes)
				{
					for (i = 1; i < anim->numframes; i++)
					{
						if (identifyType(anim->framepic[i]) != type) break;
						if (inPWAD(anim->framepic[i]) != pwad) break;
						count++;
					}
				}
				else for (i = anim->basepic + 1; i < anim->basepic + anim->numframes; i++)
				{
					if (identifyType(i) != type) break;
					if (inPWAD(i) != pwad) break;
					count++;
				}

				if (count != anim->numframes) continue;

				if (pwad)
				{
					// Type can change with pwad flats
					RDBOOL newtype = identifySplashFlats(lnum, type);

					if (newtype == TYPE_UNSET) continue;

					count = 1;

					if (anim->uniqueframes)
					{
						for (i = 1; i < anim->numframes; i++)
						{
							if (identifySplashFlats(anim->framepic[i], type) != newtype) break;
							count++;
						}
					}
					else for (i = anim->basepic + 1; i < anim->basepic + anim->numframes; i++)
					{
						if (identifySplashFlats(i, type) != newtype) break;
						count++;
					}

					if (count != anim->numframes) continue;

					type = newtype;
				}

				// Set the flats
				bp = anim->uniqueframes ? 0 : anim->basepic;

				for (i = bp; i < bp + anim->numframes; i++)
				{
					fl = R_GetFlat(bp ? i : anim->framepic[i]);

					switch(type)
					{
					case TYPE_WATER	: fl->flflags |= (FLT_WATER|FLT_GLOW); break;
					case TYPE_LAVA	: fl->flflags |= (FLT_LAVA |FLT_GLOW); break;
					case TYPE_SLIME	: { r_bloodglow ?  fl->flflags |= (FLT_SLIME|FLT_GLOW) : fl->flflags |= FLT_SLIME; } break;
					case TYPE_BLOOD	: { r_bloodglow ?  fl->flflags |= (FLT_BLOOD|FLT_GLOW) : fl->flflags |= FLT_BLOOD; } break;
					case TYPE_NUKAGE: fl->flflags |= (FLT_NUKE |FLT_GLOW); break;
					}
				}
			}
		}
	}

	if (!showflatsset) return;

	for (anim = anims ; anim < lastanim ; anim++)
	{
		if (!anim->istexture && anim->numframes > 2)
		{
			RDBOOL	pr_separator = false;

			bp = anim->uniqueframes ? 0 : anim->basepic;

			for (i = bp; i < bp + anim->numframes; i++)
			{
				picnum = bp ? i : anim->framepic[i];

				if (pwadHandle(lumpinfo[picnum].handle) && (fl = R_GetFlat(picnum)) )
				{
					pr_separator = true;
					pr_OFinsert("%8s type %s\n", lumpinfo[picnum].name, getanimType(fl->flflags));
				}
			}

			if (pr_separator) pr_OFdirect("----\n");
		}
	}

	pr_OFdirect(" ============== END AUTO\n");
}

// Info in Risen3D.wad ANIMDEF; each def is 23 bytes long
// texiden[0] endname[1] startname[10] speed[19] (bigendian int)
void P_InitANIMATED(void)
{
	byte	*animdefs = (byte*)M_lumpCACHE(W_GetNumForName("ANIMATED"), NULL, true, false);
	byte	*anim_p;
	char	*ep, *sp;
	short	startpic, endpic;
	RDBOOL	istex;

	if (!animdefs) Con_ErrorD("ANIMATED lump missing from Risen3D.wad");

	// Init animation
	for (anim_p = animdefs; *anim_p != 255; anim_p += 23)
	{
		istex = *anim_p & 1;
		ep = (char*)anim_p + 1; sp = ep + 9;

		startpic = istex ? R_CheckTextureNumForName(sp) : W_CheckFlatNumForName(sp);
		endpic   = istex ? R_CheckTextureNumForName(ep) : W_CheckFlatNumForName(ep);

		// different episode ?
		if (startpic == -1 || endpic == -1) continue;

		lastanim = checkAnimArray(lastanim);

		lastanim->basepic   = startpic;
		lastanim->numframes = endpic - startpic + 1;
		lastanim->istexture = *anim_p;
		lastanim->uniqueframes = false;
		lastanim->curframe = 0;

		if (lastanim->numframes < 2)
			Con_Error ("P_InitPicAnims: bad cycle from %s to %s", sp, ep);

		lastanim->speedmin[0] = lastanim->speedmax[0] = lastanim->countdown =
			anim_p[19]|(anim_p[20] <<  8)|(anim_p[21] << 16)|(anim_p[22] << 24);

		lastanim->countdown--;

		lastanim++;
	}

	M_unlockCACHE(animdefs);

	if (!lastanim) Con_ErrorD("invalid ANIMATED lump loaded");
}

//
// P_InitPicAnims
//
void P_InitPicAnims (void)
{
	if (haspwads && ArgExists("-sfset")) showflatsset = true;

	// Identify switch texture pairs (SW1/SW2) with a group number.
	// This will allow them to be precached at the same time.
	P_InitSwitchList();

	// An ANIMATED lump is contained in the Risen3D.wad
	// It can be overridden by a pwad ANIMATED lump
	P_InitANIMATED();
	if (!R3D_edit) P_LoadANIMDEFS();
	P_SetAnimTypeArray();
	if (haspwads && !R3D_edit) Check_Flats();
}

// v220-5 vertical water movement
watersec_t	*wsec;
int	watersec_count;

static int FindWaterThings(int secn, int idx, RDBOOL set)
{
	thinker_t	*th;
	int			n = 0;

	for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
	{
		mobj_t	*iter = (mobj_t*)th;

		if (   iter->thinker.function == (void (__cdecl *)(void))P_MobjThinker
			&& iter->subsector->sector->heightsec == secn
			&& iter->flags3 & MF3_MODULATE)
		{
			if (set)
			{
				wsec[idx].wmo[n].z = &iter->z;
				wsec[idx].wmo[n].startz = iter->z;
			}

			n++;
		}
	}

	return n;
}

void P_SetupWaterSectors()
{
	watersec_count = 0;

	if (R3D_ScriptedMap)
	{
		int		c, i;
		int		*s = (int*)malloc(numsectors * sizeof(int));
		int		*t = (int*)malloc(numsectors * sizeof(int));

		for (c = i = 0; i < numsectors; i++)
		{
			if (sectors[i].wvar && sectors[i].wvar->modulate)
			{
				sectors[i].specFlags2 |= SF2_NOFAKESMOOTH;
				t[c] = FindWaterThings(i, 0, false);
				s[c++] = i;
			}
		}

		if ( (watersec_count = c) )
		{
			wsec = (watersec_t*)W_Malloc(c * sizeof(watersec_t), LEVELID, 53);

			for (i = 0; i < c; i++)
			{
				wsec[i].secn = s[i];
				wsec[i].waterthing_count = t[i];
				wsec[i].period = sectors[s[i]].wvar->period;
				wsec[i].start_val = (float)(sectors[s[i]].wvar->phase);
				wsec[i].floorht = sectors[s[i]].floorheight;

				if (t[i])
				{
					wsec[i].wmo = (wmo_t*)W_Malloc(t[i] * sizeof(wmo_t), LEVELID, 54);

					FindWaterThings(s[i], i, true);
				}
			}
		}

		M_Free(s);
		M_Free(t);
	}
}

//
// P_UpdateSpecials
// Animate planes, scroll walls, etc.
//

void P_UpdateSpecials (void)
{
	int	i;

// CHECK scpMessage

	if (scpMessage.delay_to_next_text > 0)
	{
		scpMessage.delay_to_next_text--;
	}
	else if (scpMessage.delay_to_fade > 0)
	{
		scpMessage.delay_to_fade--;

		if (scpMessage.delay_to_fade < 11)
		{
			scpMessage.alpha = (float)scpMessage.delay_to_fade * 0.1f;
		}
	}

	TS_CheckMessageQueue();

// CHECK ZOOM MODE - R3D
	if (!inZoomMode)
	{
		if (setFOV != fieldOfView)
		{
			if (fieldOfView > setFOV)
			{
				fieldOfView = setFOV;
			}
			else
			{
				fieldOfView += 6;
				cfg.xhairSize = ZOOMXHAIRSIZE;

				if (fieldOfView >= setFOV)
				{
					fieldOfView = setFOV;
					RestoreXhair();
				}
			}
		}
	}
	else if (inZoomMode && !zoomweapontype)
	{
		ZoomReset(false);
	}
	else if (inZoomMode > 0)
	{
		if (fieldOfView > targetFOV)
		{
			if ((fieldOfView -= 4) < targetFOV) fieldOfView = targetFOV;

			cfg.xhairType = 1;
			cfg.xhairColor[0] = 255;
			cfg.xhairColor[1] = 0;
			cfg.xhairColor[2] = 0;
			cfg.xhairSize = ZOOMXHAIRSIZE;
		}
		else
			inZoomMode = -1;
	}

	lodFactor = 1.0f/(256.0f * current_w/640.0f / (fieldOfView/90.0f));

// ANIMATE FLATS AND TEXTURES GLOBALLY

	if (!R3D_edit)
	{
		anim_t			*anim;
		flat_t			*fl;
		int				pic;

		for (anim = anims ; anim < lastanim ; anim++)
		{
			if (--anim->countdown == 0)
			{
				int speedframe;

				anim->curframe = (anim->curframe + 1) % anim->numframes;
				
				speedframe = (anim->uniqueframes) ? anim->curframe : 0;

				if (anim->speedmin[speedframe] == anim->speedmax[speedframe])
					anim->countdown = anim->speedmin[speedframe];
				else
					anim->countdown = (M_RandomD() %
						(anim->speedmax[speedframe] - anim->speedmin[speedframe])) +
						anim->speedmin[speedframe];

				if (anim->uniqueframes)
				{
					pic = anim->framepic[anim->curframe];

					if (anim->istexture)
					{
						for (i = 0; i < anim->numframes; i++)
							texturetranslation[anim->framepic[i]] = pic;
					}
					else
					{
						for (i = 0; i < anim->numframes; i++)
						{
							fl = R_GetFlat(anim->framepic[i]);
							fl->flat_translation = pic;	// flat pic numbers are lump numbers
						}
					}
				}
				else
				{
					for (i = 0; i < anim->numframes; i++)
					{
						pic = anim->basepic + (anim->curframe + i) % anim->numframes;

						if (anim->istexture)
							texturetranslation[anim->basepic + i] = pic;
						else
						{
							fl = R_GetFlat(anim->basepic + i);
							fl->flat_translation = pic;
						}
					}
				}
			}
		}
	}

// DO BUTTONS
    for (i = 0; i < MAXBUTTONS; i++)
	{
		if (buttonlist[i].btimer)
		{
			buttonlist[i].btimer--;

			if (!buttonlist[i].btimer)
			{
				int snum = buttonlist[i].line->sidenum[0];

				switch(buttonlist[i].where)
				{
				case button_top:
					sides[snum].toptexture = (short)buttonlist[i].btexture;
					break;
					
				case button_middle:
					sides[snum].midtexture = (short)buttonlist[i].btexture;
					break;
					
				case button_bottom:
					sides[snum].bottomtexture = (short)buttonlist[i].btexture;
					break;
				}
				S_StartSound(sfx_swtchn,(mobj_t *)&buttonlist[i].soundorg);
				memset(&buttonlist[i],0,sizeof(button_t));
			}
		}
	}
	
// Check if light level compression's been changed - R3D
	{
		extern float rend_lightcompress;
		static int oldlightcompress = 0;

		if (oldlightcompress != cfg.lightCompress)
		{
			rend_lightcompress = (float)(cfg.lightCompress)/255.0f;
			oldlightcompress = cfg.lightCompress;
		}
	}

// Countdown delay to playing world ambients - R3D
	if (ma_inf.world_starttime > 0) ma_inf.world_starttime--;

// v220-5 Modulate fake water heights
	if (watersec_count)
	{
		fixed_t	mod, floor_mod;
		int		k;

		for (i = 0; i < watersec_count; i++)
		{
			sector_t *s = &sectors[wsec[i].secn];

			wsec[i].start_val += 360.0f/(float)wsec[i].period;
			if (wsec[i].start_val > 360.0f) wsec[i].start_val = 0.0f;

			floor_mod = (fixed_t)(sin(wsec[i].start_val / 180.f * PI) * FRACUNIT);
			mod = FixedMul(floor_mod, s->wvar->modulate);

			wsec[i].floorht = s->floorheight + mod;

			for (k = 0; k < wsec[i].waterthing_count; k++)
				*wsec[i].wmo[k].z = wsec[i].wmo[k].startz + mod;
		}
	}

// Update sky rotate angle - R3D
	if (skyAngleGlobal)	// mapinfo overrides
	{
		skyAngle += skyAngleGlobal;
		if (skyAngle >  360.0f) skyAngle -= 360.0f;
		if (skyAngle < -360.0f) skyAngle += 360.0f;
	}
	else
	{
// v220-5 sky rotate
		for (i = 0; i < numskyvar; i++)
		{
			float rot = skyvars[i].rotate_inc;

			if (rot)
			{
				skyvars[i].rotate_angle += rot;
				if (skyvars[i].rotate_angle >  360.0f) skyvars[i].rotate_angle -= 360.0f;
				if (skyvars[i].rotate_angle < -360.0f) skyvars[i].rotate_angle += 360.0f;
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////////
// UTILITIES
//

//
// Returns a side_t* derived from;
// the current sector, the line number, and the side (0 or 1).
// line must be known to be two sided before calling
//
side_t *getSide(int secnum, int lnum, int side)
{
    return &sides[sectors[secnum].lines[lnum]->sidenum[side]];
}

//
// Returns a sector_t* derived from;
// the current sector, the line number and the side (0 or 1).
//
sector_t *getSector(int secnum, int lnum, int side )
{
    return sides[sectors[secnum].lines[lnum]->sidenum[side]].sector;
}

//
//  Returns true if a line is two-sided derived from the sector and line number.
//
int twoSided(int secnum, int lnum)
{
	return compat[c_model] ?
		(sectors[secnum].lines[lnum])->ml_flags & ML_TWOSIDED
		:
		(sectors[secnum].lines[lnum])->sidenum[1] != -1;
}

//
// getNextSector()
// Return sector next to current if not floor's sector else NULL
//
sector_t* getNextSector(line_t* line, sector_t* sec)
{
/*	return  line->frontsector != sec ? line->frontsector :
			line->backsector  != sec ? line->backsector  : NULL;*/

	//jff 1/26/98 check unneeded since line->backsector already
	//returns NULL if the line is not two sided, and does so from
	//the actual two-sidedness of the line, rather than its 2S flag
	//
	//jff 5/3/98 don't retn sec unless compatibility
	// fixes an intra-sector line breaking functions
	// like floor->highest floor
	
	return	compat[c_model] && !(line->ml_flags & ML_TWOSIDED) ?
			NULL : line->frontsector == sec ?
			compat[c_model] || line->backsector != sec ?
			line->backsector : NULL : line->frontsector;
}

//
// P_FindLowestFloorSurrounding()
//
fixed_t	P_FindLowestFloorSurrounding(sector_t* sec)
{
	sector_t	*other;
	line_t		*check;
	fixed_t		floor = sec->floorheight;
	int			i;
	
	for (i = 0; i < sec->linecount; i++)
	{
		check = sec->lines[i];
		other = getNextSector(check,sec);
		
		if (!other) continue;
		
		if (other->floorheight < floor)
			floor = other->floorheight;
	}

	return floor;
}

//
// P_FindHighestFloorSurrounding()
//
fixed_t	P_FindHighestFloorSurrounding(sector_t *sec)
{
	sector_t	*other;
	//jff 1/26/98 Fix initial value for floor to not act differently
	//            in sections of wad that are below -500 units
	fixed_t		floor = compat[c_model] ? -500*FRACUNIT : -32000*FRACUNIT;
	int			i;
	
	for (i = 0; i < sec->linecount; i++)
	{
		if ((other = getNextSector(sec->lines[i],sec)) &&
			other->floorheight > floor)
		{
			floor = other->floorheight;
		}
	}

	return floor;
}

//
// P_FindNextHighestFloor
// v220-31 don't include linked sector or 3D lines (also see P_FindNextLowestFloor)
//         Mod needed to allow 3D lifts to be freed from the constraint
//         of having to return to their sector's floorheight
//
fixed_t P_FindNextHighestFloor(sector_t* sec, fixed_t currentheight)
{
	sector_t	*other;
	sector_t	*chk = sec->listsec ? sec : NULL;	// v220-31
	int			i, height;
	
	for (i = 0; i < sec->linecount; i++)
	{
		if (   sec->lines[i]->backsector != chk				// v220-31
			&& sec->lines[i]->specType < R3D_40				// v220-31
			&& (other = getNextSector(sec->lines[i],sec))
			&& (height = other->floorheight) > currentheight)
		{
			while (++i < sec->linecount)
			{
				if (   sec->lines[i]->backsector != chk		// v220-31
					&& sec->lines[i]->specType < R3D_40		// v220-31
					&& (other = getNextSector(sec->lines[i],sec))
					&& other->floorheight < height
					&& other->floorheight > currentheight)
				{
					height = other->floorheight;
				}
			}
			return height;
		}
	}
	return currentheight;
}

//
// P_FindNextLowestFloor()
// v220-31 don't include linked sector or 3D lines (also see P_FindNextHighestFloor)
//
fixed_t P_FindNextLowestFloor(sector_t *sec, int currentheight)
{
	sector_t	*other;
	sector_t	*chk = sec->listsec ? sec : NULL;	// v220-31
	int			i, height;
	
	for (i = 0; i < sec->linecount; i++)
	{
		if (   sec->lines[i]->backsector != chk				// v220-31
			&& sec->lines[i]->specType < R3D_40				// v220-31
			&& (other = getNextSector(sec->lines[i],sec))
			&& (height = other->floorheight) < currentheight)
		{
			while (++i < sec->linecount)
			{
				if (   sec->lines[i]->backsector != chk		// v220-31
					&& sec->lines[i]->specType < R3D_40		// v220-31
					&& (other = getNextSector(sec->lines[i],sec))
					&& other->floorheight > height
					&& other->floorheight < currentheight)
				{
					height = other->floorheight;
				}
			}
			return height;
		}
	}
	return currentheight;
}

//
// P_FindLowestCeilingSurrounding()
//

fixed_t P_FindLowestCeilingSurrounding(sector_t* sec)
{
	sector_t	*other;
	fixed_t		height = DDMAXINT;
	int			i;
	
	for (i = 0; i < sec->linecount; i++)
	{
		if (  (other = getNextSector(sec->lines[i],sec))
			&& other != sec
			&& other->ceilingheight < height
			// b211-0 ZDoom map compatibility
			&& !(other->specFlags & SF_MOVCEIL
			&& sec->specFlags & SF_MOVCEIL		// v220-3 additional test
			&& other->ceilingheight == sec->floorheight))
		{
			height = other->ceilingheight;
		}
	}
	
	return height > 32000*FRACUNIT ? 32000*FRACUNIT : height;
}


//
// P_FindHighestCeilingSurrounding()
//
fixed_t	P_FindHighestCeilingSurrounding(sector_t* sec)
{
	sector_t	*other;
	fixed_t		height = DDMININT;
	int			i;
	
	for (i = 0; i < sec->linecount; i++)
	{
		if ((other = getNextSector(sec->lines[i],sec)) &&
			other->ceilingheight > height)
		{
			height = other->ceilingheight;
		}
	}
		
	return height < -32000*FRACUNIT ? -32000*FRACUNIT : height;
}

// Find the next sector with the same tag as a linedef.
// Rewritten by Lee Killough to use chained hashing to improve speed
int P_FindSectorFromTag(int tag, int start)
{
	start = (start >= 0 ? sectors[start].nexttag :
		sectors[(unsigned) tag % (unsigned) numsectors].firsttag);
	while (start >= 0 && sectors[start].tag != tag)
		start = sectors[start].nexttag;
	return start;
}

// killough 4/16/98: Same thing, only for linedefs
int P_FindLineFromLineTag(const line_t *line, int start)
{
	start = (start >= 0 ? lines[start].nexttag :
		lines[(unsigned) line->tag % (unsigned) numlines].firsttag);
	while (start >= 0 && lines[start].tag != line->tag)
		start = lines[start].nexttag;
	return start;
}

// v220-31
int P_CheckCompositeCeilings(int tag, fixed_t amount)
{
	int sn;

	for (sn = -1; (sn = P_FindSectorFromTag(tag, sn)) >= 0;)
	{
		if (sectors[sn].specFlags & SF_MOVSECTOR)
		{
			fixed_t sh = sectors[sn].stptr >= 0 ? slopes[sectors[sn].stptr].slope[0].slopeheight * FRACUNIT : 0;

			if (sectors[sn].ceilingheight - (sectors[sn].floorheight + sh) < amount)
			return sn;
		}
	}

	return -1;
}

// v220-31
fixed_t P_GetLowestCompositeFloor(int tag)
{
	int sn, lf = DDMAXINT;

	for (sn = -1; (sn = P_FindSectorFromTag(tag, sn)) >= 0;)
	{
		if (   sectors[sn].specFlags & SF_MOVSECTOR
			&& sectors[sn].floorheight < lf)
		{
			lf = sectors[sn].floorheight;
		}
	}

	return lf;
}

// v220-31
RDBOOL P_CheckCompositeLower(int tag, fixed_t amount, fixed_t lolimit)
{
	fixed_t lf = P_GetLowestCompositeFloor(tag);
	return (__int64)lf - (__int64)amount >= (__int64)lolimit;
}

// v220-31
fixed_t P_GetHighestCompositeFloor(int tag)
{
	int sn, hf = DDMININT;

	for (sn = -1; (sn = P_FindSectorFromTag(tag, sn)) >= 0;)
	{
		if (   sectors[sn].specFlags & SF_MOVSECTOR
			&& sectors[sn].floorheight > hf)
		{
			hf = sectors[sn].floorheight;
		}
	}

	return hf;
}

// v220-31
RDBOOL P_CheckCompositeRaise(int tag, fixed_t amount, fixed_t hilimit)
{
	fixed_t hf = P_GetHighestCompositeFloor(tag);
	return (__int64)hf + (__int64)amount <= (__int64)hilimit;
}

//
// Find minimum light from an adjacent sector
//
int P_FindMinSurroundingLight(sector_t *sector, int max)
{
    sector_t	*check;
    int			i, min = max;
	
    for (i = 0; i < sector->linecount; i++)
    {
		if (  (check = getNextSector(sector->lines[i], sector))
			&& check->lightlevel < min)
		{
			min = check->lightlevel;
		}
    }
    return min;
}

//
// EVENTS
// Events are operations triggered by using, crossing,
// or shooting special lines, or by timed thinkers.
//

//
// P_CrossSpecialLine - Walkover Trigger Dispatcher
//
// Called every time a thing origin is about
//	to cross a line with a non 0 special, whether a walkover type or not.
//
// jff 02/12/98 all W1 lines were fixed to check the result from the EV_
//	function before clearing the special. This avoids losing the function
//	of the line, should the sector already be active when the line is
//	crossed.
//
// Rationalised
//

// v220-10 reformatted
void P_CrossSpecialLine(line_t *line, mobj_t *thing)
{
	extern int	plasmadamage, bfgdamage;
	ushort special = (ushort)line->special;
	short *p = &line->special;

	if (special >= GenLimit) return;

	if (thing->player && wadani.mobj != NULL)
	{
		if (wadani.trigset[0] < 0 && wadani.triglinenum[0] == line - lines)
		{
			wadani.trigset[0] = 1;
		}
		else if (wadani.trigset[1] < 0 && wadani.triglinenum[1] == line - lines)
		{
			wadani.trigset[1] = 1;
		}
		else if (wadani.trigset[2] < 0 && wadani.triglinenum[2] == line - lines)
		{
			wadani.trigset[2] = 1;
		}
	}

	//	Things that should never trigger lines
	if (!thing->player)
	{
		if (line->xtraflags & XT_PLRTRIGONLY) return;	// R3D trigger type

		// Things that should NOT trigger specials...
		switch(thing->type)
		{
		case MT_PLASMA:
			if (plasmadamage) break;	// v220-35
			return;
		case MT_BFG:
			if (bfgdamage) break;	// v220-35
			return;
		case MT_BRUISERSHOT:
		case MT_TROOPSHOT:
		case MT_HEADSHOT:
		case MT_ROCKET:
		case MT_GRENADEX:	// v220-35
//		case MT_GRENADE:	// b211-0 for R3D grenade
		case MT_PIPEBOMB:	// b211-0 for R3D pipebomb
		case MT_MINE:		// R3D
			return;
		default:
			break;
		}
	}

	if (R3D_maptype)
	{
		if (special == 8092) return;

		// v220-31 walkover once sound for standard map type (set in SCP_ZOFF)
		if (special == 0x1FCC || special == 0x1FD0)
		{
			if (line->linesound)
			{
				S_StartSound (line->linesound, thing);
				line->special = 0;	// play once
			}
			return;
		}

		// v220-31 walkover many sound for standard map type (set in SCP_ZOFF)
		if (special == 0x1FCD)
		{
			S_StartSound (line->linesound, thing);
			return;
		}

		if (line->xtraflags & XT_TURNLSOFF)
		{
			if (gdt.tmthing->player)
			{
				for (int i = 0; i < MAXLINESOUNDS; i++)
				{
					if (soundstarts[i] && line->linesound == soundstarts[i])
					{
						S_StopSound(line->linesound, NULL);
						soundstarts[i] = 0;
						break;
					}
				}
			}
		}

		if (line->xtraflags & XT_LIGHTFLAGS)
		{
			cfg.useTorchLight =  cfg.setVtxLight = cfg.sineCompress = 0;

			if (line->xtraflags & XT_TURNTORCHON)
			{
				cfg.useTorchLight = 1;
			}
			else
			{
				if (line->xtraflags & XT_TURNVTXON ) cfg.setVtxLight = 1;
			}

			if (line->xtraflags & XT_TURNSINEON) cfg.sineCompress = 1;
		}


		// v220-4 Music line triggers
		if (special >= 0x1FB8 && special < 0x1FCC)	// 8120 to 8139
		{
			special -= 0x1FB8;

			if (music_list[special])
			{
				global_looped = 0;
				PlayScriptedFile(music_list[special]);
			}

			line->special = 0;
			return;
		}
		
		// Scripted line check - Note 0x1FB0 (8112) is cleared on loading so will never appear here
		if (special >= 8064 && special <= 8113)
		{
			if (TS_Activate(thing, line, MS_WALKOVER)) return;
		}
	}

	// Add check here for generalized linedef types. Pointer to
	// line function is NULL by default, set non-null if line
	// special is walkover generalized linedef type. jff 02/04/98
	// check each range of generalized linedefs
	// NOTE: with changes made in v220-10 this is only called for walk types
	if (special >= GenCrusherBase)
	{
		int  (*linefunc)(line_t *line) = NULL;

		if (special >= GenFloorBase)
		{
			// tag required, FloorModel allows monsters if FloorChange is 0
			if (!line->tag || (!thing->player && (!(special & FloorModel) || special & FloorChange)) ) return;

			linefunc = EV_DoGenFloor;
		}
		else if (special >= GenCeilingBase)
		{
			// tag required, CeilingModel allows monsters if CeilingChange is 0
			if (!line->tag || (!thing->player && (!(special & CeilingModel) || special & CeilingChange)) ) return;

			linefunc = EV_DoGenCeiling;
		}
		else if (special >= GenDoorBase)
		{
			// tag required, DoorMonster allows monsters if door not secret
			if (!line->tag || (!thing->player && (!(special & DoorMonster) || line->ml_flags & ML_SECRET)) ) return;

			linefunc = EV_DoGenDoor;
		}
		else if (special >= GenLockedBase)
		{
			if (!thing->player) return;	// monsters are not allowed to unlock doors

			// check it can be unlocked (tag 0 allowed)
			if (!P_CanUnlockGenDoor(line, thing->player, false)) return;

			linefunc = EV_DoGenLockedDoor;
		}
		else if (special >= GenLiftBase)
		{
			// tag required, LiftMonster allows monsters
			if (!line->tag || (!thing->player && !(special & LiftMonster)) ) return;

			linefunc = EV_DoGenLift;
		}
		else if (special >= GenStairsBase)
		{
			// tag required, StairMonster allows monsters
			if (!line->tag || (!thing->player && !(special & StairMonster)) ) return;

			linefunc = EV_DoGenStairs;
		}
		else	// GenCrusherBase
		{
			// tag required, CrusherMonster allows monsters
			if (!line->tag || (!thing->player && !(special & CrusherMonster)) ) return;

			linefunc = EV_DoGenCrusher;
		}
		
		if (linefunc) // if it was a valid generalized type
		{
			switch((line->special & TriggerType) >> TriggerTypeShift)
			{
			case WalkOnce:
				// clear special if a walk once type
				if (linefunc(line)) line->special = 0;
				break;
			case WalkMany:
				linefunc(line);
				break;
			default:	// shouldn't happen (worth erroring out?)
				break;
			}
		}

		return;
	}

	if (!P_CheckTag(line)) return;	//jff 2/27/98 disallow zero tag on some types

	if (thing->player)
	{
		if (special == 51 || special == 52)
		{
			// killough 10/98: prevent dead voodoo dolls causing an exit
			if (thing->dplayer && thing->dplayer->mo != thing && thing->health <= 0) return;
		}
		else if (special >= 124)
		{
			if (special == 124)
			{
				// killough 10/98: prevent dead players from causing an exit
				if (plhealth.health <= 0) return;
			}
			else	// monster only teleports
			{
				if (special == 125 || special == 126) return;
				if (special >= 264 && special <= 269) return;
			}
		}
	}
	else
	{
		// Most functions are restricted to the player but some are allowed for monsters
		switch(special)
		{
		case 4:		// raise door
		case 10:	// plat down-wait-up-stay trigger
		case 88:	// plat down-wait-up-stay retrigger

			// Teleportation
		case 39:	// once
		case 97:	// teleport retrigger
		case 125:	// monster only once
		case 126:	// monster only
			// silent line teleporters
		case 207:	// once
		case 208:	// WR silent teleporter (normal kind)
			// Line to line teleporters
		case 243:	// once
		case 244:	// WR silent teleporter (linedef-linedef kind)
		case 262:	// reversed once
		case 263:	// reversed
		case 264:	// reversed monster only once
		case 265:	// reversed monster only
		case 266:	// monster only once
		case 267:	// monster only
			// silent line teleporters
		case 268:	// monster only once
		case 269:	// monster only
			break;

		case 120:
			if (ma_inf.modifier & MODF_120) break;	// See Risen3D\bin\R3D_modifier.txt

		default: return;	// player only
		}
	}
	
	// Dispatch on the line special value to the line's action routine
	// If a once only function, and successful, clear the line special
	
	switch (special)
	{
		// Regular walk once triggers
		
	case 2: if (EV_DoDoor(line, openNormal)) { *p = 0; }; break;				// Open Door
	case 3: if (EV_DoDoor(line, closeNormal)) { *p = 0; }; break;				// Close Door
	case 4: if (EV_DoDoor(line, normal)) { *p = 0; }; break;					// Raise Door
	case 5: if (EV_DoFloor(line, raiseFloor)) { *p = 0; }; break;				// Raise Floor
	case 6: if (EV_DoCeiling(line, fastCrushAndRaise)) { *p = 0; }; break;		// Fast Ceiling Crush & Raise
	case 8: if (EV_BuildStairs(line,build8)) { *p = 0; }; break;				// Build Stairs
	case 10: if (EV_DoPlat(line,downWaitUpStay,0)) { *p = 0; }; break;			// PlatDownWaitUp
	case 12: EV_LightTurnOn(line,0); *p = 0; break;								// Light Turn On - brightest near		
	case 13: EV_LightTurnOn(line,255); *p = 0; break;							// Light Turn On 255	
	case 16: if (EV_DoDoor(line,close30ThenOpen)) { *p = 0; }; break;			// Close Door 30
	case 17: EV_StartLightStrobing(line); *p = 0; break;						// Start Light Strobing	
	case 19: if (EV_DoFloor(line,lowerFloor)) { *p = 0; }; break;				// Lower Floor
	case 22: if (EV_DoPlat(line,raiseToNearestAndChange,0)) { *p = 0; }; break;	// Raise floor to nearest height and change texture
	case 25: if (EV_DoCeiling(line,crushAndRaise)) { *p = 0; }; break;			// Ceiling Crush and Raise
	case 30: if (EV_DoFloor(line,raiseToTexture)) { *p = 0; }; break;			// Raise floor to shortest texture height on either side of lines.
	case 35: EV_LightTurnOn(line,35); *p = 0; break;							// Lights Very Dark	
	case 36: if (EV_DoFloor(line,turboLower)) { *p = 0; }; break;				// Lower Floor (TURBO)
	case 37: if (EV_DoFloor(line,lowerAndChange)) { *p = 0; }; break;			// LowerAndChange
	case 38: if (EV_DoFloor(line, lowerFloorToLowest)) { *p = 0; }; break;		// Lower Floor To Lowest
	case 39: if (EV_Teleport(line, thing)) { *p = 0; }; break;					// TELEPORT!
	case 40: if (EV_DoCeiling(line, raiseToHighest)) { *p = 0; }; break;		// RaiseCeilingLowerFloor
	case 44: if (EV_DoCeiling(line, lowerAndCrush)) { *p = 0; }; break;			// Ceiling Crush
	case 52: G_ExitLevel();														// EXIT!
	case 53: if (EV_DoPlat(line,perpetualRaise,0)) { *p = 0; }; break;			// Perpetual Platform Raise
	case 54: EV_StopPlat(line); *p = 0; break;									// Platform Stop	
	case 56: if (EV_DoFloor(line,raiseFloorCrush)) { *p = 0; }; break;			// Raise Floor Crush
	case 57: if (EV_CeilingCrushStop(line)) { *p = 0; }; break;					// Ceiling Crush Stop
	case 58: if (EV_DoFloor(line,raiseFloor24)) { *p = 0; }; break;				// Raise Floor 24
	case 59: if (EV_DoFloor(line,raiseFloor24AndChange)) { *p = 0; }; break;	// Raise Floor 24 And Change
	case 100: if (EV_BuildStairs(line,turbo16)) { *p = 0; }; break;				// Build Stairs Turbo 16
	case 104: EV_TurnTagLightsOff(line); *p = 0; break;							// Turn lights off in sector(tag)		
	case 108: if (EV_DoDoor(line,blazeRaise)) { *p = 0; }; break;				// Blazing Door Raise (faster than TURBO!)
	case 109: if (EV_DoDoor (line,blazeOpen)) { *p = 0; }; break;				// Blazing Door Open (faster than TURBO!)
	case 110: if (EV_DoDoor (line,blazeClose)) { *p = 0; }; break;				// Blazing Door Close (faster than TURBO!)
	case 119: if (EV_DoFloor(line,raiseFloorToNearest)) { *p = 0; }; break;		// Raise floor to nearest surr. floor
	case 121: if (EV_DoPlat(line,blazeDWUS,0)) { *p = 0; }; break;				// Blazing PlatDownWaitUpStay
	case 124: ma_inf.exitlevel = 0x8000100A; /*G_SecretExitLevel ();*/ break;	// Exit to secret level 2
	case 125:if (EV_Teleport(line, thing)) { *p = 0; }; break;					// TELEPORT MonsterONLY
	case 130:if (EV_DoFloor(line,raiseFloorTurbo)) { *p = 0; }; break;			// Raise Floor Turbo
	case 141:if (EV_DoCeiling(line,silentCrushAndRaise)) { *p = 0; }; break;	// Silent Ceiling Crush & Raise
		
		// Regular walk many retriggerable	
	case 72: EV_DoCeiling( line, lowerAndCrush); break;						// Ceiling Crush
	case 73: EV_DoCeiling(line,crushAndRaise); break;						// Ceiling Crush and Raise	
	case 74: EV_CeilingCrushStop(line); break;								// Ceiling Crush Stop	
	case 75: EV_DoDoor(line,closeNormal); break;							// Close Door
	case 76: EV_DoDoor(line,close30ThenOpen); break;						// Close Door 30
	case 77: EV_DoCeiling(line,fastCrushAndRaise); break;					// Fast Ceiling Crush & Raise
	case 79: EV_LightTurnOn(line,35); break;								// Lights Very Dark
	case 80: EV_LightTurnOn(line,0); break;									// Light Turn On - brightest near
	case 81: EV_LightTurnOn(line,255); break;								// Light Turn On 255	
	case 82: EV_DoFloor( line, lowerFloorToLowest ); break;					// Lower Floor To Lowest	
	case 83: EV_DoFloor(line,lowerFloor); break;							// Lower Floor	
	case 84: EV_DoFloor(line,lowerAndChange); break;						// LowerAndChange	
	case 86: EV_DoDoor(line,openNormal); break;								// Open Door
	case 87: EV_DoPlat(line,perpetualRaise,0); break;						// Perpetual Platform Raise	
	case 88: EV_DoPlat(line,downWaitUpStay,0); break;						// PlatDownWaitUp	
	case 89: EV_StopPlat(line); break;										// Platform Stop
	case 90: EV_DoDoor(line,normal); break;									// Raise Door
	case 91: EV_DoFloor(line,raiseFloor); break;							// Raise Floor
	case 92: EV_DoFloor(line,raiseFloor24); break;							// Raise Floor 24
	case 93: EV_DoFloor(line,raiseFloor24AndChange); break;					// Raise Floor 24 And Change
	case 94: EV_DoFloor(line,raiseFloorCrush); break;						// Raise Floor Crush
	case 95: EV_DoPlat(line,raiseToNearestAndChange,0); break;				// Raise floor to nearest height and change texture.
	case 96: EV_DoFloor(line,raiseToTexture); break;						// Raise floor to shortest texture height on either side of lines.
	case 97: EV_Teleport( line, thing ); break;								// TELEPORT!
	case 98: EV_DoFloor(line,turboLower); break;							// Lower Floor (TURBO)
	case 105: EV_DoDoor (line,blazeRaise); break;							// Blazing Door Raise (faster than TURBO!)
	case 106: EV_DoDoor (line,blazeOpen); break;							// Blazing Door Open (faster than TURBO!)
	case 107: EV_DoDoor (line,blazeClose); break;							// Blazing Door Close (faster than TURBO!)
	case 120: EV_DoPlat(line,blazeDWUS,0); break;							// Blazing PlatDownWaitUpStay.
	case 126: EV_Teleport( line, thing ); break;							// TELEPORT MonsterONLY.
	case 128: EV_DoFloor(line,raiseFloorToNearest); break;					// Raise To Nearest Floor
	case 129: EV_DoFloor(line,raiseFloorTurbo); break;						// Raise Floor Turbo

		// Extended walk triggers
		
		// jff 1/29/98 added new linedef types to fill all functions out so that
		// all have varieties SR, S1, WR, W1
		
		// killough 1/31/98: "factor out" compatibility test, by
		// adding inner switch qualified by compatibility flag.
		
		// killough 2/16/98: Fix problems with W1 types being cleared too early
		
	default:
		switch (special)
		{
			// Extended walk once triggers
	
		case 142: if (EV_DoFloor(line,raiseFloor512)) { *p = 0; }; break;			// Raise Floor 512
		case 143: if (EV_DoPlat(line,raiseAndChange,24)) { *p = 0; }; break;		// Raise Floor 24 and change
		case 144: if (EV_DoPlat(line,raiseAndChange,32)) { *p = 0; }; break;		// Raise Floor 32 and change
		case 145: if (EV_DoCeiling( line, lowerToFloor)) { *p = 0; }; break;		// Lower Ceiling to Floor
		case 146: if (EV_DoDonut(line)) { *p = 0; }; break;							// Lower Pillar, Raise Donut
		case 199: if (EV_DoCeiling(line,lowerToLowest)) { *p = 0; }; break;			// Lower ceiling to lowest surrounding ceiling
		case 200: if (EV_DoCeiling(line,lowerToMaxFloor)) { *p = 0; }; break;		// Lower ceiling to highest surrounding floor
		case 207: if (EV_SilentTeleport(line, thing)) { *p = 0; }; break;			// killough 2/16/98: W1 silent teleporter (normal kind)
		case 153: if (EV_DoChange(line,trigChangeOnly)) { *p = 0; }; break;			// Texture/Type Change Only (Trig)
		case 239: if (EV_DoChange(line,numChangeOnly)) { *p = 0; }; break;			// Texture/Type Change Only (Numeric)
		case 219: if (EV_DoFloor(line,lowerFloorToNearest)) { *p = 0; }; break;		// Lower floor to next lower neighbor
		case 227: if (EV_DoElevator(line,elevateUp)) { *p = 0; }; break;			// Raise elevator next floor
		case 231: if (EV_DoElevator(line,elevateDown)) { *p = 0; }; break;			// Lower elevator next floor
		case 235: if (EV_DoElevator(line,elevateCurrent)) { *p = 0; }; break;		// Elevator to current floor
		case 243: if (EV_SilentLineTeleport(line, thing, false)) { *p = 0; }; break;// killough 2/16/98: W1 silent teleporter (linedef-linedef kind)
		case 262: if (EV_SilentLineTeleport(line, thing, true)) { *p = 0; }; break; //jff 4/14/98 add silent line-line reversed
		case 264: if (EV_SilentLineTeleport(line, thing, true)) { *p = 0; }; break;	//jff 4/14/98 add monster-only silent line-line reversed
		case 266: if (EV_SilentLineTeleport(line, thing, false)) { *p = 0; }; break; //jff 4/14/98 add monster-only silent line-line
		case 268: if (EV_SilentTeleport(line, thing)) { *p = 0; }; break;			//jff 4/14/98 add monster-only silent
			
			//jff 1/29/98 end of added W1 linedef types
			
			// Extended walk many retriggerable
			
			//jff 1/29/98 added new linedef types to fill all functions
			//out so that all have varieties SR, S1, WR, W1
			
		case 147: EV_DoFloor(line,raiseFloor512); break;						// Raise Floor 512
		case 148: EV_DoPlat(line,raiseAndChange,24); break;						// Raise Floor 24 and Change
		case 149: EV_DoPlat(line,raiseAndChange,32); break;						// Raise Floor 32 and Change
		case 150: EV_DoCeiling(line,silentCrushAndRaise); break;				// Start slow silent crusher
		case 151: EV_DoCeiling( line, raiseToHighest);
				  EV_DoFloor( line, lowerFloorToLowest); break;					// EV_DoCeiling(raiseToHighest), EV_DoFloor(lowerFloortoLowest)
		case 152: EV_DoCeiling( line, lowerToFloor ); break;					// Lower Ceiling to Floor
		case 256: EV_BuildStairs(line,build8); break;							// Build stairs, step 8	(jff 3/16/98 renumber was 153)
		case 257: EV_BuildStairs(line,turbo16); break;							// Build stairs, step 16	(jff 3/16/98 renumber was 154)
		case 155: EV_DoDonut(line); break;										// Lower Pillar, Raise Donut
		case 156: EV_StartLightStrobing(line); break;							// Start lights strobing
		case 157: EV_TurnTagLightsOff(line); break;								// Lights to dimmest near
		case 201: EV_DoCeiling(line,lowerToLowest); break;						// Lower ceiling to lowest surrounding ceiling
		case 202: EV_DoCeiling(line,lowerToMaxFloor); break;					// Lower ceiling to highest surrounding floor
		case 208: EV_SilentTeleport(line, thing); break;						// killough 2/16/98: WR silent teleporter (normal kind)
		case 212: EV_DoPlat(line,toggleUpDn,0); break;							// Toggle floor between C and F instantly. jff 3/14/98 create instant toggle floor type
		case 154: EV_DoChange(line,trigChangeOnly); break;						// Texture/Type Change Only (jff 3/16/98 was 216)
		case 240: EV_DoChange(line,numChangeOnly); break;						// Texture/Type Change Only (Numeric)
		case 220: EV_DoFloor(line,lowerFloorToNearest); break;					// Lower floor to next lower neighbor
		case 228: EV_DoElevator(line,elevateUp); break;							// Raise elevator next floor
		case 232: EV_DoElevator(line,elevateDown); break;						// Lower elevator next floor
		case 236: EV_DoElevator(line,elevateCurrent); break;					// Elevator to current floor
		case 244: EV_SilentLineTeleport(line, thing, false); break;				// killough 2/16/98: WR silent teleporter (linedef-linedef kind)
		case 263: EV_SilentLineTeleport(line, thing, true); break;				//jff 4/14/98 add silent line-line reversed
		case 265: EV_SilentLineTeleport(line, thing, true); break;				//jff 4/14/98 add monster-only silent line-line reversed
		case 267: EV_SilentLineTeleport(line, thing, false); break;				//jff 4/14/98 add monster-only silent line-line
		case 269: EV_SilentTeleport(line, thing); break;						// monster only silent

			// v220-33 COD specials
		case 274: EV_LightSector(line, 15); { *p = 0; }; break;					// raise sector light level by 15
		case 275: EV_LightSector(line, -15); { *p = 0; }; break;				// lower sector light level by 15
		}
		break;
	}
}

//
// P_ShootSpecialLine - IMPACT SPECIALS
// Called when a thing shoots a special line.
//
// Rationalised
//
RDBOOL P_ShootSpecialLine (mobj_t* thing, line_t* line, int from)
{
	short special = line->special;
	
	if (special <= 0) return false;

	if (R3D_maptype)
	{
		if (R3D_ScriptedMap && special == 8143)
		{
			if (from == 2) P_Activate1_1801x_line(thing, line);
			return false;
		}
		
		if (special == 8091)	// R3D ambient gun trigger - no other action
		{
			if (thing->player)
				S_SetLineTriggeredAmbient(line->ambNumLink, 5);
			return false;
		}

		if (special == 8114)
		{
			TS_Activate(thing, line, MS_GUN);
			return false;
		}

		// TS_Activate() is only called if line has blocking flag set.
		// If an unblocked R3D_40 type then it is dealt with separately
		// in doom\p_map.c
		if (line->lnflags2 & LNF2_BLOCKALL)
		{
			TS_Activate(thing, line, MS_GUN);
			return false;
		}
	}

	if ((ushort)special >= (ushort)GenLimit) return false;
	
	//jff 02/04/98 add check here for generalized linedef	
	// Separate out the general types
	if (special >= GenCrusherBase)
	{
		int (*linefunc)(line_t *line) = NULL;
		int trigtype = line->special & TriggerType;
		
		// Check walkover trigger type before checking specials
		if (trigtype != GunOnce && trigtype != GunMany)
		{
			return false;
		}
		
		if (special >= GenFloorBase)
		{
			//jff 2/27/98 most gun generalized types require tag
			if (!line->tag) return false;
			
			// FloorModel is "Allow Monsters" if FloorChange is 0
			if (!thing->player && (special & FloorChange || !(special & FloorModel)) )
				return false;
			
			linefunc = EV_DoGenFloor;
		}
		else if (special >= GenCeilingBase)
		{
			if (!line->tag) return false;
			
			// CeilingModel is "Allow Monsters" if CeilingChange is 0
			if (!thing->player && (special & CeilingChange || !(special & CeilingModel)) )
				return false;
			
			linefunc = EV_DoGenCeiling;
		}
		else if (special >= GenDoorBase)
		{
			if (!line->tag) return false;
			
			// monsters may be disallowed, they can't open secret doors in any event
			if (!thing->player && (!(special & DoorMonster) || line->ml_flags & ML_SECRET) )
				return false;
			
			linefunc = EV_DoGenDoor;
		}
		else if (special >= GenLockedBase)
		{
			// monsters disallowed from unlocking doors
			// untagged gun doors can't be unlocked, so don't report
			if (!thing->player || !line->tag) return false;
			
			if (!P_CanUnlockGenDoor(line, thing->player, false)) return false;
			
			linefunc = EV_DoGenLockedDoor;
		}
		else if (special >= GenLiftBase)
		{
			if (!thing->player && !(special & LiftMonster) )
				return false; // monsters disallowed

			linefunc = EV_DoGenLift;
		}
		else if (special >= GenStairsBase)
		{
			// normal monsters disallowed
			if (!line->tag || (!thing->player && !(special & StairMonster)) )
				return false;
			
			linefunc = EV_DoGenStairs;
		}
		else	// GenCrusherBase
		{
			// normal monsters disallowed
			if (!line->tag || (!thing->player && !(special & StairMonster)) )
				return false;
			
			linefunc = EV_DoGenCrusher;
		}
		
		if (linefunc)
		{
			if (trigtype == GunOnce)
			{
				if (linefunc(line))
				{
					P_ChangeSwitchTexture(line, 0);
					return true;
				}
			}
			else // GunMany
			{
				if (linefunc(line))
				{
					P_ChangeSwitchTexture(line, 1);
					return true;
				}
			}
		}

		return false;
	}

	// GR Open door on impact weapon can also be monster activated
	if (!thing->player && line->special != 46) return false;

	if (!P_CheckTag(line))  //jff 2/27/98 disallow zero tag on some types
		return false;

	switch(line->special)
	{
	case 24: // 24 G1 raise floor to highest adjacent
		if (EV_DoFloor(line,raiseFloor))
		{
			P_ChangeSwitchTexture(line,0);
			return true;
		}
		break;
		
	case 46: // 46 GR open door, stay open
		EV_DoDoor(line,openNormal);
		P_ChangeSwitchTexture(line,1);
		return true;
		
	case 47: // 47 G1 raise floor to nearest and change texture and type
		if (EV_DoPlat(line,raiseToNearestAndChange,0))
		{
			P_ChangeSwitchTexture(line,0);
			return true;
		}
		break;
		
		//jff 1/30/98 added new gun linedefs here
	case 197: // Exit to next level
		
		// killough 10/98: prevent zombies from exiting levels
		if (thing->dplayer && thing->dplayer->mo != thing && thing->health <= 0)
			break;
		P_ChangeSwitchTexture(line,0);
		G_ExitLevel();
		return true;
		
	case 198: // Exit to secret level
		
		// killough 10/98: prevent zombies from exiting levels
		if (thing->dplayer && thing->dplayer->mo != thing && thing->health <= 0) break;
		
		P_ChangeSwitchTexture(line,0);
	//	G_SecretExitLevel();

		ma_inf.exitlevel = 0x0800100A;	// R3D add delay so switch change can be seen
		return true;
		//jff end addition of new gun linedefs
	}
	return false;
}

//
// P_PlayerInSpecialSector
// Called per tic if the player's origin is in a special sector
//
void P_PlayerInSpecialSector(mbf_player_t* player)
{
    sector_t *sector = player->plr->mo->subsector->sector;
	RDBOOL	 damage = !(leveltime&0x1f);

	// v220-31 check for pseudofake damage
	if (   damage
		&& !ge_inf.devmap
		&& ppt.viewer_fake_sector >= 0
		&& (sectors[ppt.viewer_fake_sector].specFlags2 & (SF2_PSEUDOFAKE|SF2_PSFAKEDAMAGE)) == (SF2_PSEUDOFAKE|SF2_PSFAKEDAMAGE)
		&& player->plr->mo->z <= sectors[ppt.viewer_fake_sector].floorheight)
	{
		RDBOOL resetgodmode = false;

		if (player->cheats & CF_GODMODE)
		{
			player->cheats &= ~CF_GODMODE;	// Annul godmode in case set
			resetgodmode = true;
		}

		P_DamageMobj(player->plr->mo, NULL, NULL, 20, 30);  // damage player

		if (!ma_inf.no_light[1] && ppt.viewplrmo->flags & MF_CORPSE)
		{
			ma_inf.no_light[1] = 1;
		}

		damage = false;

		if (resetgodmode) player->cheats |= CF_GODMODE;
	}

	if (!sector->special) return;
	
    // Falling, not all the way down yet? Unless in a water sector...
    if (player->plr->mo->z != SL_PointHeight(sector, player->plr->mo->x, player->plr->mo->y, LOSLOPE)
		&& (!waterlevel || waterlevel == 4))
	{
		return;
	}

	RDBOOL	 noironfeet = !player->powers[pw_ironfeet];
	
	// Has hit ground.
	// jff add < 32 to handle old vs generalized types
	// $sectorspecials
	if (sector->special < 32) // regular sector specials
	{
		switch (sector->special)
		{
		case 5: // HELLSLIME DAMAGE
			if (damage && noironfeet)
				P_DamageMobj (player->plr->mo, NULL, NULL, 10, 31);
			break;
				
		case 7: // NUKAGE DAMAGE
			if (damage && noironfeet)
				P_DamageMobj (player->plr->mo, NULL, NULL, 5, 32);
			break;
				
		case 16:
			// SUPER HELLSLIME DAMAGE
		case 4:
			// STROBE HURT
			if (damage)
			{
				if (noironfeet || (M_RandomD() < 5) )
					P_DamageMobj (player->plr->mo, NULL, NULL, 20, 33);
			}
			break;
			
		case 9:
			// SECRET SECTOR
			player->secretcount++;
			sector->special = 0;
			sector->siflags |= SIF_WASSECRET;	// v220-15

			if (cfg.secretMsg)
			{
				P_SetMessage(player, STSTR_SECRETFOUND);
				S_ConsoleSound(sfx_itmbk, NULL, player - mbf_players);
			}
			break;
			
		case 11:
			// SUPER DAMAGE (e.g. E1M8 finale)
			// Exit on health < 11, take 20 damage per 31 ticks

			if (!ge_inf.devmap) player->cheats &= ~CF_GODMODE;	// Annul godmode in case set
			
			if (damage)
				P_DamageMobj(player->plr->mo, NULL, NULL, 20, 34);

			if (plhealth.health < 11) G_ExitLevel();

			break;

		case 18:	// v220-33 COD instant death support
			if (!ge_inf.devmap) player->cheats &= ~CF_GODMODE;
			P_DamageMobj (player->plr->mo, NULL, NULL, 1000, 35);
			break;

		case 19:	// v220-33 COD healing support
			if (plhealth.health < 100 && !(leveltime & 0x1f)) plhealth.health++;
			break;
			
		default:
			break;
		};
	}
	else
	{
		if (damage)
		{
			int d = 0;

			switch ((sector->special & DAMAGE_MASK) >> DAMAGE_SHIFT)
			{
			case 1: if (noironfeet) d = 5;
				break;
			case 2: if (noironfeet) d = 10;
				break;
			case 3:
				// maybe take damage even with suit
				if (noironfeet || (M_RandomD() < 5)) d = 20;
				break;
			default: break; // no damage
			}

			if (d) P_DamageMobj (player->plr->mo, NULL, NULL, d, 36);
		}

		if (sector->special & SECRET_MASK)
		{
			player->secretcount++;

			sector->special &= ~SECRET_MASK;
			sector->siflags |= SIF_WASSECRET;	// v220-15

			if (sector->special < 32) sector->special = 0; // clear standard specials

			if (cfg.secretMsg)	// Added v2.1.00 build_62
			{
				P_SetMessage(player, STSTR_SECRETFOUND);
				S_ConsoleSound(sfx_itmbk, NULL, player - mbf_players);
			}
		}
	}
}

//
// P_CanUnlockGenDoor()
//
// Passed a generalized locked door linedef and a player, returns whether
// the player has the keys necessary to unlock that door.
//
// Note: The linedef passed MUST be a generalized locked door type
//       or results are undefined.
//
// jff 02/05/98 routine added to test for unlockability of
//  generalized locked doors
//
// killough 11/98: reformatted

RDBOOL P_CanUnlockGenDoor(line_t *line, mbf_player_t *player, RDBOOL type)
{
	// does this line special distinguish between skulls and keys?
	int skulliscard = (line->special & LockedNKeys) >> LockedNKeysShift;

	RDBOOL red_card = player->cards[it_redcard    ];
	RDBOOL blu_card = player->cards[it_bluecard   ];
	RDBOOL ylw_card = player->cards[it_yellowcard ];
	RDBOOL red_skul = player->cards[it_redskull   ];
	RDBOOL blu_skul = player->cards[it_blueskull  ];
	RDBOOL ylw_skul = player->cards[it_yellowskull];

	RDBOOL ret = true;

	// determine for each case of lock type if player's keys are adequate. killough 3/20/98
	switch((line->special & LockedKey) >> LockedKeyShift)
	{
	case AnyKey:
		if (   !red_card
			&& !blu_card
			&& !ylw_card
			&& !red_skul
			&& !blu_skul
			&& !ylw_skul)
		{
			P_SetMessage(player, PD_ANYKEY);
			ret = false;
		}
		break;
	case RCard:
		if (!red_card && (!skulliscard || !red_skul))
		{
			P_SetMessage(player, skulliscard ? PD_REDO : PD_REDKC);
			ret = false;
		}
		break;
	case BCard:
		if (!blu_card && (!skulliscard || !blu_skul))
		{
			P_SetMessage(player, skulliscard ? PD_BLUEO : type ? PD_BLUEKCA : PD_BLUEKC);
			ret = false;
		}
		break;
	case YCard:
		if (!ylw_card && (!skulliscard || !ylw_skul))
		{
			P_SetMessage(player, skulliscard ? PD_YELLOWO : type ? PD_YELLOWKCA : PD_YELLOWKC);
			ret = false;
		}
		break;
	case RSkull:
		if (!red_skul && (!skulliscard || !red_card))
		{
			P_SetMessage(player, skulliscard ? PD_REDO : type ? PD_REDSKA : PD_REDSK);
			ret = false;
		}
		break;
	case BSkull:
		if (!blu_skul && (!skulliscard || !blu_card))
		{
			P_SetMessage(player, skulliscard ? PD_BLUEO : type ? PD_BLUESKA : PD_BLUESK);
			ret = false;
		}
		break;
	case YSkull:
		if (!ylw_skul && (!skulliscard || !ylw_card))
		{
			P_SetMessage(player, skulliscard ? PD_YELLOWO : type ? PD_YELLOWSKA : PD_YELLOWSK);
			ret = false;
		}
		break;
	case AllKeys:
		if (!skulliscard)
		{
			if (   !red_card
				|| !blu_card
				|| !ylw_card
				|| !red_skul
				|| !blu_skul
				|| !ylw_skul)
			{
				P_SetMessage(player, PD_SIXKEYS);
				ret = false;
			}
		}
		else
		{
			if ((!red_card && !red_skul) ||
				(!blu_card && !blu_skul) ||
				(!ylw_card && !ylw_skul))
			{
				P_SetMessage(player, PD_THREEKEYS);
				ret = false;
			}
		}
		break;
	}
	if (!ret) S_StartSound(sfx_oof,player->plr->mo);
	return ret;
}

//
// P_SectorActive()
//
// Passed a linedef special class (floor, ceiling, lighting) and a sector
// returns whether the sector is already busy with a linedef special of the
// same class. If old demo compatibility true, all linedef special classes
// are the same.
//
// killough 11/98: reformatted

int P_SectorActive(special_e t,sector_t *sec)
{
	switch (t)	// return whether thinker of same type is active
	{
	case floor_special:
		return (int)sec->floordata;
	case ceiling_special:
		return (int)sec->ceilingdata;
	case lighting_special:
		return (int)sec->lightingdata;
	default:
		return 1; // don't know which special, must be active, shouldn't be here
	}
}


//
// P_CheckTag()
//
// Passed a line, returns true if the tag is non-zero or the line special
// allows no tag without harm. If compatibility, all linedef specials are
// allowed to have zero tag.
//
// Note: Only line specials activated by walkover, pushing, or shooting are
//       checked by this routine.
//
// jff 2/27/98 Added to check for zero tag allowed for regular special types

int P_CheckTag(line_t *line)
{
	if (line->tag || line->lnflags2 & LNF2_DOORTYPE)
		return 1;

	if (line->special > 260)
	{
	/* these are dealt with separately e.g.
		case 8090:	// R3D ambient trigger - push
		case 8091:	// R3D ambient trigger - gun
		case 8092:	// R3D ambient trigger - walkover
		case 8093:	// R3D line trigger - push
	*/
		return 0;
	}

	switch (line->special)
	{

/*	case 48:  // Scrolling wall specials are reset to zero on start up
	case 85:
*/
	case 7:		// stairs build8;
	case 8:
	case 256:
	case 258:

	case 100:	// stairs turbo16
	case 127:
	case 257:
	case 259:
		return (line->backsector && line->backsector->tag);

	case 1:   // VerticalDoor door specials
	case 26:
	case 27:
	case 28:
	case 31:
	case 32:
	case 33:
	case 34:
	case 117:
	case 118:

	case 139:  // Lighting specials
	case 170:
	case 79:
	case 35:
	case 138:
	case 171:
	case 81:
	case 13:
	case 192:
	case 169:
	case 80:
	case 12:
	case 194:
	case 173:
	case 157:
	case 104:
	case 193:
	case 172:
	case 156:
	case 17:
		
	case 195:  // Thing teleporters
	case 174:
	case 97:
	case 39:
	case 126:
	case 125:
	case 210:
	case 209:
	case 208:
	case 207:
		
	case 11:  // Exits
	case 52:
	case 197:
	case 51:
	case 124:
	case 198:
		return 1;

	default:
		return 0;
	}
}

//  Following loading scan for specials that spawn thinkers
void P_SpawnSpecials (void)
{
	sector_t	*sector;
	char		badhdr[] = {"\n===== SECTORS with invalid specials ======\n"};
	int			i;
	int			badcount = 0;
	int			mask;
	short		val;
	RDBOOL		badHeader = false, codwarn = false;;
	RDBOOL		listSecret = ArgCheck("-listSecret");

		// $sectorspecials	Init

	for (i = 0, sector = sectors; i < numsectors ; i++, sector++)
	{
		if (!sector->special) continue;

		// secret sectors
		if (sector->special == 9 || sector->special & SECRET_MASK)	// jff 3/15/98 count extended
		{
			if (listSecret) pr_OFinsert("SECRET SECTOR %i\n", i);
			totalsecret++;
		}

		if (sector->special == 21)
			P_SpawnPhasedSectorLight(sector);		// phased single
		else if (sector->special == 22)
			P_SpawnPhasedLightingSequence(sector);	// phased group


		val = sector->special & 0x1f;

		switch (val)
		{
		case 0: break;
		case 1: P_SpawnLightFlash (sector); break;					// random off

			// STROBE
		case 4:	// strobe fast/death slime
			sector->special |= (3 << DAMAGE_SHIFT);	//jff 3/14/98 put damage bits in
		case 2:		// strobe fast
		case 3:		// strobe slow
		case 12:	// sync strobe slow
		case 13:	// sync strobe fast
			mask = (val == 3 || val == 12) ? SLOWDARK : FASTDARK;
			P_SpawnStrobeFlash(sector, mask, (val == 12 || val == 13));	// v220-24 set sync
			break;

		case 6: if (R3D_lightextns) break;	// v220-31 do nothing See P_CheckInheritance5505()	
			goto set_default;

		case 8:  P_SpawnGlowingLight(sector); break;		// glowing light
		case 10: P_SpawnDoorCloseIn30(sector); break;		// door close in 30 seconds
		case 14: P_SpawnDoorRaiseIn5Mins(sector); break;	// door raise in 5 minutes
		case 17: P_SpawnFireFlicker(sector); break;			// fire flickering
		case 18:											// v220-33 COD instant death
		case 19: if ( !(ge_inf.addon_flags & AF_COD) )		// v220-33 COD healing support
					codwarn = true;
			break;
		// remaining valid specials
		case 5: case 7: case 9: case 11: case 16:
		// Phased lighting
		case 21: case 22: case 23: case 24:
			break;

		default:
set_default:;
			if (R3D_maptype && val == 20) break;	// v220-33
			// v220-6 check for invalid fields
			if (!badHeader) { Con_Message(badhdr); badHeader = true; }
			Con_Message(" sec %5i special 0x%.4X - unrecognised\n", i, sector->special);
			badcount++;
			sector->special = 0;		// value error
			break;
		}
	}
	if (codwarn) pr_OFdirect("WARNING: COD sector specials 18 or 19 detected\n");
	if (badcount) pr_OFrepeat(NULL, "=", 42, 2);

	for (i = 0; i < MAXBUTTONS; i++) memset(&buttonlist[i], 0, sizeof(button_t));

	P_NullActiveCeilings();
	P_NullActivePlats();
	P_SetupFakeTeleport();		// R3D
	P_SpawnScrollers();			// killough 3/7/98: Add generalized scrollers
	if (!R3D_edit)
	{
		P_SpawnFriction();			// phares 3/12/98: New friction model using linedefs
		P_SpawnPushers();			// phares 3/20/98: New pusher model using linedefs
	}
}

// R3D - Teleport fake sectors
static void P_SetupFakeTeleport()
{
	thinker_t	*th;
	int			i;

	for (i = 0; i < 100; i++)
	{
		tpfake[i].secnum[0] = -1;
		tpfake[i].secnum[1] = -1;
	}

	for (th = thinkercap.next; th != &thinkercap; th = th->next)
	{
		if (th->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t	*iter = (mobj_t*)th;
			if (iter->type == MT_TPUNDERFAKE ||
				iter->type == MT_TPOVERFAKE)
			{
				RDBOOL	invalid = false;
				int		under  = iter->type == MT_TPUNDERFAKE ? 1 : 0;
				int		mapnum = MAPNUM(iter);
				int		num    = TIDNUM(iter);

				//-------------------------------

				// Program error check
				if (num > 99) Con_Error("Program error [18]: num %i > 99 [mapnum %i]", num, mapnum);

				if (tpfake[num].secnum[under] >= 0)
				{
					Con_Message("Error: TPFAKE thing mapnum %i - more than one in map\n", mapnum);
					invalid = true;
				}

				if (iter->subsector->sector->heightsec < 0)
				{
					Con_Message("Error: TPFAKE thing not in a fake sector : mapnum %i\n", mapnum);
					invalid = true;
				}

				if (invalid) continue;

				if (!under) iter->subsector->sector->specFlags2 |= SF2_FAKETPSEC;

				tpfake[num].secnum[under] = iter->subsector->sector - sectors;
				tpfake[num].x_pos [under] = iter->x;
				tpfake[num].y_pos [under] = iter->y;
			}
		}
	}

	for (i = 0; i < 100; i++)
	{
		int a = tpfake[i].secnum[0], b = tpfake[i].secnum[1];

		if ( a >= 0 || b >= 0 )
		{
			int c = 0;

			if (a >= 0 && b <  0) c = 14300;
			if (a <  0 && b >= 0) c = 14200;

			if (c)
			{
				Con_Message("Error : TPFAKE thing %i has no pair\n", c + i);
				tpfake[i].secnum[0] = -1;
				tpfake[i].secnum[1] = -1;
			}
		}
	}
}
// End teleport fake sectors


// killough 2/28/98:
//
// This function, [WAS with the help of r_plane.c and r_bsp.c], supports
// generalized scrolling floors and walls, with optional mobj-carrying prop-
// erties, e.g. conveyor belts, rivers, etc. A linedef with a special type
// affects all tagged sectors the same way, by creating scrolling and/or
// object-carrying properties. Multiple linedefs may be used on the same sector
// and are cumulative, although the special case of scrolling a floor and carrying
// things on it, requires only one linedef. The linedef's direction determines
// the scrolling direction, and the linedef's length determines the scrolling
// speed. This was designed so that an edge around the sector could be used to
// control the direction of the sector's scrolling, which is usually what is
// desired.
//
// Process the active scrollers.
//
// This is the main scrolling code
// killough 3/7/98
void T_Scroll(scroll_t *s)
{
	fixed_t dx = s->dx, dy = s->dy;

	if (s->control != -1)
	{	// compute scroll amounts based on a sector's height changes
		fixed_t height = sectors[s->control].floorheight +
			sectors[s->control].ceilingheight;
		fixed_t delta = height - s->last_height;
		s->last_height = height;
		dx = FixedMul(dx, delta);
		dy = FixedMul(dy, delta);
	}
	
	// killough 3/14/98: Add acceleration
	if (s->accel)
	{
		dx += s->vdx;
		dy += s->vdy;
		s->vdx = dx;
		s->vdy = dy;
	}
	
	if (!(dx | dy))
	{
		if (s->type != sc_carry || !ma_inf.ambtrigger) return;	// pointless continuing

		if ( !(leveltime & 7) )
		{
			for (int i = 0; i < MAXAMBIENT; i++)
			{
				if (   scrollAMB[i].ambnum >= 0
					&& scrollAMB[i].secnum == s->affectee)
				{
					if (scrollAMB[i].count < 16)
						S_StopAmbient(scrollAMB[i].mo, scrollAMB[i].ambnum);
					else
					{
						 scrollAMB[i].count--;
					}
				}
			}
		}

		return;
	}

	// v220-10 ZDoom scroll compatibility
	if (ma_inf.modifier & MODF_SCROLL)
	{
		if (dx && abs(dx) <= 0x300) dx = (dx > 0 ? STOPSPEED : -STOPSPEED);
		if (dy && abs(dy) <= 0x300) dy = (dy > 0 ? STOPSPEED : -STOPSPEED);
	}
	
	switch (s->type)
	{
	case sc_side:					// killough 3/7/98: Scroll wall texture
		scsm.s[sides[s->affectee].smoothoff].sdx += dx;
		scsm.s[sides[s->affectee].smoothoff].sdy += dy;
		break;
		
	case sc_floor:					// killough 3/7/98: Scroll floor texture
		scsm.f[sectors[s->affectee].idxf].sdx += dx;
		scsm.f[sectors[s->affectee].idxf].sdy -= dy;
		// If floor scroll in sky sector then sky rotated
		break;
		
	case sc_ceiling:			   // killough 3/7/98: Scroll ceiling texture
		scsm.c[sectors[s->affectee].idxc].sdx += dx;
		scsm.c[sectors[s->affectee].idxc].sdy -= dy;
		break;
		
	case sc_carry:
		{
			sector_t	*sec = sectors + s->affectee;
			fixed_t		height, waterheight;  // killough 4/4/98: add waterheight
			int			i, c = 0;
		
			if (ma_inf.ambtrigger && !(leveltime & 7) )
			{
				for (i = 0; i < MAXAMBIENT; i++)
				{
					if (   scrollAMB[i].ambnum >= 0
						&& scrollAMB[i].secnum == s->affectee)
					{
						scrollAMB[i].count = 16;
						S_StartAmbient(scrollAMB[i].ambnum);
					}
				}
			}

			// killough 3/7/98: Carry things on floor
			// killough 3/27/98: fix carrier bug
			// killough 4/4/98: Underwater, carry things even w/o gravity
			
			height = sec->floorheight;
			waterheight = (sec->heightsec >= 0
				&& sectors[sec->heightsec].floorheight > height ?
					sectors[sec->heightsec].floorheight : DDMININT);
			
			validcount++;

			// Only sector linked things are placed on the list
			P_SectorTouchingThingsList(sec, false);
			
			// R3D - add linked lists
			for (i = 0; i < link_list.listcnt; i++)
			{
				mobj_t	*mo = link_list.mobj_list[i];

				// v220-10 prevent additive carry across a line with different sectors
				if (   compat[c_carry]
					&& mo != ppt.viewplrmo
					&& mo->floorz == mo->subsector->sector->floorheight 
					&& mo->lastdx == dx
					&& mo->lastdy == dy)
				{
					continue;
				}
				
				// Move objects only if on floor or underfakefloor,
				// non-floating, clipped and sector linked.
				if (   !(mo->flags & MF_NOCLIP)
					&& mo->z == mo->floorz
					&& (!(mo->flags & MF_NOGRAVITY
					|| mo->z > height)
					|| mo->z < waterheight))
				{
					mo->misc |= MJMISC_TRANS_TC;
					mo->lastdx = dx, mo->lastdy = dy;
					mo->momx += dx;
					mo->momy += dy;

					// v220-35 Move player if standing on a decoration model being floor scrolled
					if (   R3D_ScriptedMap
						&& ppt.viewplrmo
						&& ppt.viewplrmo->onmobj == mo
						&& !mo->info->deathstate)	// decoration model
					{
						ppt.viewplrmo->momx += dx;
						ppt.viewplrmo->momy += dy;
						ppt.viewplrmo->misc |= MJMISC_TRANS_TC;
					}
				}
			}

			link_list.listcnt = 0;
		}
		
		break;
			
	case sc_carry_ceiling:	// to be added later (never happened - gmj)
		break;
	}
}

//
// Add_Scroller()
//
// Add a generalized scroller to the thinker list.
//
// type: the enumerated type of scrolling: floor, ceiling, floor carrier,
//   wall, floor carrier & scroller
//
// (dx,dy): the direction and speed of the scrolling or its acceleration
//
// control: the sector whose heights control this scroller's effect
//   remotely, or -1 if no control sector
//
// affectee: the index of the affected object (sector or sidedef)
//
// accel: non-zero if this is an accelerative effect
//
// Risen3D v220-4 scsm struct added for smooth scroll

static void Add_Scroller(int type, fixed_t dx, fixed_t dy,
                         int control, int affectee, int accel)
{
	if (type == sc_floor)
	{
		if (scsm.fnum >= scsm.maxfnum)
		{
			scsm.maxfnum += 16;
			scsm.f = (scvar_t*)W_ReallocCLR(scsm.f, scsm.maxfnum * sizeof(scvar_t), LEVELID, 0);
		}

		sectors[affectee].idxf = scsm.fnum;
		scsm.f[scsm.fnum].sdx = scsm.f[scsm.fnum].oldx = scsm.f[scsm.fnum].dx = sectors[affectee].flooroffx;
		scsm.f[scsm.fnum].sdy = scsm.f[scsm.fnum].oldy = scsm.f[scsm.fnum].dy = sectors[affectee].flooroffy;
		scsm.f[scsm.fnum++].s = affectee;

		sectors[affectee].siflags |= SIF_FLOOR;
	}

	if (type == sc_ceiling)
	{
		if (scsm.cnum >= scsm.maxcnum)
		{
			scsm.maxcnum += 16;
			scsm.c = (scvar_t*)W_ReallocCLR(scsm.c, scsm.maxcnum * sizeof(scvar_t), LEVELID, 0);
		}

		sectors[affectee].idxc = scsm.cnum;
		scsm.c[scsm.cnum].sdx = scsm.c[scsm.cnum].oldx = scsm.c[scsm.cnum].dx = sectors[affectee].ceiloffx;
		scsm.c[scsm.cnum].sdy = scsm.c[scsm.cnum].oldy = scsm.c[scsm.cnum].dy = sectors[affectee].ceiloffy;
		scsm.c[scsm.cnum++].s = affectee;

		sectors[affectee].siflags |= SIF_CEILING;
	}

	if (type == sc_side)
	{
		if (R3D_edit) return;
		sides[affectee].lsiflags |= LS_SIDESCROLL;
		if (sides[affectee].smoothoff >= 0) return;

		if (scsm.snum >= scsm.maxsnum)
		{
			scsm.maxsnum += 16;
			scsm.s = (scvar_t*)W_ReallocCLR(scsm.s, scsm.maxsnum * sizeof(scvar_t), LEVELID, 0);
		}

		sides[affectee].smoothoff = scsm.snum;
		scsm.s[scsm.snum].sdx = scsm.s[scsm.snum].oldx = scsm.s[scsm.snum].dx = sides[affectee].textureoffset;
		scsm.s[scsm.snum].sdy = scsm.s[scsm.snum].oldy = scsm.s[scsm.snum].dy = sides[affectee].rowoffset;
		scsm.s[scsm.snum++].s = affectee;
	}

	{
		scroll_t *sc = (scroll_t*)W_Calloc(sizeof(scroll_t), THINKID, 0);

		sc->thinker.function = (void (__cdecl *)(void))T_Scroll;
		sc->type = type;
		sc->dx = dx;
		sc->dy = dy;
		sc->accel = accel;
		sc->vdx = sc->vdy = 0;
		sc->affectee = affectee;

		if ((sc->control = control) != -1)
		{
			sc->last_height =
				sectors[control].floorheight + sectors[control].ceilingheight;
		}

		// v220-24 revised SIF_CARRY handling
		if (type == sc_carry) sectors[affectee].siflags |= SIF_CARRY;

		P_AddFNThinker(&sc->thinker);
	}
}

// Adds wall scroller. Scroll amount is rotated with respect to wall's
// linedef first, so that scrolling towards the wall in a perpendicular
// direction is translated into vertical motion, while scrolling along
// the wall in a parallel direction is translated into horizontal motion.
//
// killough 5/25/98: cleaned up arithmetic to avoid drift due to roundoff
//
// killough 10/98:
// fix scrolling aliasing problems, caused by long linedefs causing ovf

static void Add_WallScroller(fixed_t dx, fixed_t dy, const line_t *l,
                             int control, int accel)
{
	fixed_t x = abs(l->dx), y = abs(l->dy), d;
	angle_t an;

	if (y > x) { d = x; x = y; y = d; }	// swap x,y

	an = (angle_t)(atan2((double)(FixedDiv(y,x) >> 5), (double)SLOPERANGE) * pidiv);
	d = (an + ANG90) >> ANGLETOFINESHIFT;
	d = FixedDiv(x, finesine[d]);

	// killough 10/98: Use long long arithmetic
	x = (fixed_t)(((__int64)dy * -(__int64)l->dy - (__int64)dx * (__int64)l->dx) / (__int64)d);
	y = (fixed_t)(((__int64)dy *  (__int64)l->dx - (__int64)dx * (__int64)l->dy) / (__int64)d);

	Add_Scroller(sc_side, x, y, control, *l->sidenum, accel);
}

// Amount (dx,dy) vector linedef is shifted right to get scroll amount
#define SCROLL_SHIFT 5

// Factor to scale scrolling effect into mobj-carrying properties = 3/32.
// (This is so scrolling floors and objects on them can move at same speed.)
#define CARRYFACTOR ((fixed_t)(FRACUNIT*.09375f))

RDBOOL checkTP_special(int special)
{
	if (R3D_ScriptedMap) return false;

	switch(special)
	{
	default: return false;

	case 39:	// teleport trigger
	case 97:	// teleport retrigger
	case 125:	// teleport monsteronly trigger
	case 126:	// as 125
	case 207:	// silent thing teleporters
	case 208:	// as 207
	case 268:	// add monster-only silent
	case 269:	// as 268

	// Line to line teleporters
	case 243:	// silent line-line teleporter
	case 244:	// make fit within DCK's 256 linedef types
	case 266:	// add monster-only silent line-line
	case 267:	// as 266

	// Reversed types
	case 262:	// add monster only
	case 263:	// silent thing,line,line rev types
	case 264:	// plus player/monster silent line
	case 265:	// reversed types

	// switch types
	case 174:
	case 195:
	case 209:
	case 210:

		return true;
	}
}

// Check if an adjacent sector has teleport lines
static RDBOOL checkSectorTPspecials(int secnum)
{
	sector_t *f, *b, *s = &sectors[secnum];
	int		 i, k;

	for (i = 0; i < s->linecount; i++)
	{
		line_t *ld = s->lines[i];

		if (   ld->ml_flags & ML_TWOSIDED
			&& (f = ld->frontsector) != (b = ld->backsector)
			&& f->floorheight == b->floorheight)
		{
			if (f == s) f = b;

			for (k = 0; k < f->linecount; k++)
			{
				line_t *li = f->lines[k];

				if (   li->special
					&& li->frontsector == li->backsector
					&& checkTP_special(li->special))
				{
					return true;
				}
			}
		}
	}
	return false;
}

// R3D- Stop monsters on scrolling floors from fading
// out/being removed when also under a crushing ceiling.
static void setMonsterNofade(int secnum)
{
	mobj_t* thing;
	
	for (thing = sectors[secnum].thinglist;thing; thing = thing->snext)
	{
		if (  !(thing->flags2 & MF2_TELEPCORPSE)
			&& thing->thinker.function == (void (__cdecl *)(void))P_MobjThinker
			&& thing->flags & MF_COUNTKILL
			&& thing->flags & MF_SOLID
			&& !(thing->flags & (MF_FLOAT|MF_SKULLFLY))
			&& thing->subsector->sector->specFlags & SF_MOVCEIL
			&& thing->type != MT_PLAYER		// I assume zombies should be excluded
			&& checkSectorTPspecials(secnum))
		{
			thing->flags2 |= MF2_TELEPCORPSE;
		}
	}
}

// Initialize the scrollers v220-33 revamped
static void P_SpawnScrollers(void)
{
	extern byte		walkCheck[];
	line_t	*l = lines;
	float	inc = 0.f;
	int		snum;
	int		*notagscroll = NULL;
	int		maxnumnotagscroll = 0;
	int		numnotagscroll = 0;
	int		i;
	
	for (i = 0; i < numlines; i++, l++)
	{
		int		control = -1, accel = 0;	// no control sector or acceleration
		int		s, special = l->special;

		if (special > 255 || (byte)walkCheck[special] != 7) continue;

		// v220-35 fix unwanted scrolling walls bug
		if (!l->tag && !(special == 48 || special == 85 || special == 255))
		{
			pr_OFinsert("MAP ERROR: L%i special %i is untagged\n", i, l->special);
			ma_inf.maperror++;
			continue;
		}
		
		fixed_t	dx = l->dx >> SCROLL_SHIFT;	// direction and speed of scrolling
		fixed_t	dy = l->dy >> SCROLL_SHIFT;

		// killough 3/7/98: Types 245-249 are same as 250-254 except that the
		// first side's sector's heights cause scrolling when they change, and
		// this linedef controls the direction and speed of the scrolling. The
		// most complicated linedef since donuts, but powerful :)
		//
		// killough 3/15/98: Add acceleration. Types 214-218 are the same but
		// are accelerative.
		
		if (special >= 245 && special <= 249)		// displacement scrollers
		{
			special += 250-245;
			control = sides[*l->sidenum].sector - sectors;
		}
		else
		if (special >= 214 && special <= 218)		// accelerative scrollers
		{
			accel = 1;
			special += 250-214;		// +36 -> 250 to 254
			control = sides[*l->sidenum].sector - sectors;
		}
		
		switch (special)
		{			
		case 250:	// scroll ceiling: accelerative type 214 and displacement type 245
			if (!l->tag)
			{
				Add_Scroller(sc_ceiling, -dx, dy, control, l->frontsector - sectors, accel);
			}
			else
			{
				for (s = -1; (s = P_FindSectorFromTag(l->tag,s)) >= 0;)
					Add_Scroller(sc_ceiling, -dx, dy, control, s, accel);
			}
			break;
			
		case 251:	// scroll floor: accelerative type 215 and displacement type 246
		case 253:   // carry objects on floor: accelerative type 217 and displacement type 248
			if (!l->tag)
			{
				setMonsterNofade(l->frontsector - sectors);
				Add_Scroller(sc_floor, -dx, dy, control, l->frontsector - sectors, accel);
			}
			else
			{
				for (s = -1; (s = P_FindSectorFromTag(l->tag,s)) >= 0;)
				{
					setMonsterNofade(s);
					Add_Scroller(sc_floor, -dx, dy, control, s, accel);
				}
			}
			if (special == 251) break;	// floor only
			// fall through
			
		case 252:	// carry objects on floor: accelerative type 216 and displacement type 247
			dx = FixedMul(dx,CARRYFACTOR);
			dy = FixedMul(dy,CARRYFACTOR);
			if (!l->tag)
			{
				Add_Scroller(sc_carry, dx, dy, control, l->frontsector - sectors, accel);
			}
			else
			{
				for (s = -1; (s = P_FindSectorFromTag(l->tag,s)) >= 0;)
					Add_Scroller(sc_carry, dx, dy, control, s, accel);
			}
			break;
			
			// killough 3/1/98: scroll wall according to linedef
			// (same direction and speed as scrolling floors)
			// changed v220-35
		case 254: // accelerative type 218 and displacement type 249
			
			inc = 0.f;

			for (s = -1; (s = P_FindLineFromLineTag(l,s)) >= 0;)
			{
				if (s != i)
					Add_WallScroller(dx, dy, lines+s, control, accel);
			}

			// v220-35	support rotation periods using the dx value
			//			Note that dx, if set, takes precedance over dy

			// v220-5 sky rotation
			for (s = -1; (s = P_FindSectorFromTag(l->tag, s)) >= 0;)
			{
				int		k;
				float	rotate = 0.f;

				if (s >= 0 && (sectors[s].floorpic == skyflatnum || sectors[s].ceilingpic == skyflatnum))
				{
					if (sectors[s].sidesky & PL_SKYFLAT) rotate = dx ? FIX2FLT(dx) / 180.0f : FIX2FLT(dy) / 180.0f;
				}

				if (rotate)
				{
					// see if one set has the same values
					for (k = 0; k < numskyvar; k++)
					{
						if (skyvars[k].rotate_inc == rotate) break;
					}

					if (k < numskyvar)
					{
						sectors[s].skyvar = k;	// found match
					}
					else
					{
						if (numskyvar >= maxnumskyvar)
						{
							maxnumskyvar += 16;

							skyvars = (skyvar_t*)W_ReallocCLR(skyvars, maxnumskyvar * sizeof(skyvar_t), STATICID, 0);
						}

						// v220-29 reverse direction for flipped sky
						if (sectors[s].specFlags2 & SF2_HASTYPE272)
						{
							skyvars[numskyvar].rotate_inc = -(FABS(rotate));
						}
						else
						{
							skyvars[numskyvar].rotate_inc = FABS(rotate);
						}

						sectors[s].skyvar = numskyvar++;
					}
				}
			}
			break;
				
		case 255:    // killough 3/2/98: scroll according to sidedef offsets
			s = *l->sidenum;
			Add_Scroller(sc_side, -sides[s].textureoffset,
				sides[s].rowoffset, -1, s, accel);
			break;
			
		case 48:
		case 85:
			if (sides[*l->sidenum].lsiflags & LS_SIDESCROLL) break;	// already applied
			// This allows a tag to be set without affecting floor scroll having been applied.
			if (numnotagscroll == maxnumnotagscroll)
			{
				maxnumnotagscroll += 32;
				notagscroll = (int*)realloc(notagscroll, maxnumnotagscroll * sizeof(int));
			}
			notagscroll[numnotagscroll++] = i;
			break;

		default:
			break;
		}
	}

	for (i = 0; i < numnotagscroll; i++)	// Could be tagged to something else
	{
		snum = *lines[notagscroll[i]].sidenum;

		if (sides[snum].lsiflags & LS_SIDESCROLL) continue;

		if (lines[notagscroll[i]].special == 48)  // scroll first side +ve
			Add_Scroller(sc_side,  FRACUNIT, 0, -1, snum, 0);
		else	// type 85 scroll first side -ve
			Add_Scroller(sc_side, -FRACUNIT, 0, -1, snum, 0);
	}

	free(notagscroll);
}

// killough 3/7/98 -- end generalized scroll effects

////////////////////////////////////////////////////////////////////////////
//
// FRICTION EFFECTS
//
// phares 3/12/98: Start of friction effects
//
// As the player moves, friction is applied by decreasing the x and y
// momentum values on each tic. By varying the percentage of decrease,
// we can simulate muddy or icy conditions. In mud, the player slows
// down faster. In ice, the player slows down more slowly.
//
// The amount of friction change is controlled by the length of a linedef
// with type 223. A length < 100 gives you mud. A length > 100 gives you ice.
//
// Also, each sector where these effects are to take place is given a
// new special type _______. Changing the type value at runtime allows
// these effects to be turned on or off.
//
// Sector boundaries present problems. The player should experience these
// friction changes only when his feet are touching the sector floor. At
// sector boundaries where floor height changes, the player can find
// himself still 'in' one sector, but with his feet at the floor level
// of the next sector (steps up or down). To handle this, Thinkers are used
// in icy/muddy sectors. These thinkers examine each object that is touching
// their sectors, looking for players whose feet are at the same level as
// their floors. Players satisfying this condition are given new friction
// values that are applied by the player movement code later.
//
// killough 8/28/98:
//
// Completely redid code, which did not need thinkers, and which put a heavy
// drag on CPU. Friction is now a property of sectors, NOT objects inside
// them. All objects, not just players, are affected by it, if they touch
// the sector's floor. Code simpler and faster, only calling on friction
// calculations when an object needs friction considered, instead of doing
// friction calculations on every sector during every tic.
//
// Although this -might- ruin Boom demo sync involving friction, it's the only
// way, short of code explosion, to fix the original design bug. Fixing the
// design bug in Boom's original friction code, while maintaining demo sync
// under every conceivable circumstance, would double or triple code size, and
// would require maintenance of buggy legacy code which is only useful for old
// demos. Doom demos, which are more important IMO, are not affected by this
// change.
//
// Forget demos, only demos recorded with R3D are supported
//
/////////////////////////////
//
// Initialize the sectors where friction is increased or decreased

// v220-4 P_SpawnFriction made global so friction values can be reset
//        if the wd command is used. This is because friction is now
//        ignored in fake water sectors if waterdynamics is active so
//        changing wd will require friction values to be re-calculated.

void P_SpawnFriction(void)
{
	int i;
	line_t *li = lines;
	
	// killough 8/28/98: initialize all sectors to normal friction first
	for (i = 0; i < numsectors; i++)
	{
		sectors[i].friction = ORIG_FRICTION;
		sectors[i].movefactor = ORIG_FRICTION_FACTOR;
	}
	
	for (i = 0 ; i < numlines ; i++, li++)
	{
		if (li->special == 223)
		{
			int length = P_ApproxDistance(li->dx, li->dy) >> FRACBITS;
			int friction = (0x1EB8*length)/0x80 + 0xD000;
			int movefactor, s;
			
			// The following check might seem odd. At the time of movement,
			// the move distance is multiplied by 'friction/0x10000', so a
			// higher friction value actually means 'less friction'.
			
			if (friction > ORIG_FRICTION)		// ice
				movefactor = ((0x10092 - friction)*(0x70))/0x158;
			else
				movefactor = ((friction - 0xDB34)*(0xA))/0x80;
			
			// killough 8/28/98: prevent odd situations
			if (friction > FRACUNIT) friction = FRACUNIT; else if (friction < 0) friction = 0;	
			if (movefactor < 32) movefactor = 32;

			for (s = -1; (s = P_FindSectorFromTag(li->tag, s)) >= 0 ; )
			{
				// killough 8/28/98:
				//
				// Instead of spawning thinkers, which are slow and expensive,
				// modify the sector's own friction values. Friction should be
				// a property of sectors, not objects which reside inside them.
				// Original code scanned every object in every friction sector
				// on every tic, adjusting its friction, putting unnecessary
				// drag on CPU. New code adjusts friction of sector only once
				// at level startup, and then uses this friction value.
				
				// v220-4 If waterdynamics is active then don't change default
				//        friction values in fake water sectors
				if ( !(waterdynamics
					&& sectors[s].heightsec >= 0
					&& sectors[sectors[s].heightsec].specFlags2 & SF2_HASWATERFLATS) )
				{
					sectors[s].friction = friction;
					sectors[s].movefactor = movefactor;
				}
			}
		}
	}
}

//
// phares 3/12/98: End of friction effects
//
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
//
// PUSH/PULL EFFECT
//
// phares 3/20/98: Start of push/pull effects
//
// This is where push/pull effects are applied to objects in the sectors.
//
// There are four kinds of push effects
//
// 1) Pushing Away
//
//    Pushes you away from a point source defined by the location of an
//    MT_PUSH Thing. The force decreases linearly with distance from the
//    source. The force is applied within a circle, ignoring sector
//    boundaries, with the center being that of the MT_PUSH Thing.
//    The force is only applied if MT_PUSH can see the target object.
//
// 2) Pulling toward
//
//    Same as 1) except you're pulled toward an MT_PULL point source.
//
// 3) Wind
//
//    Pushes you in a constant direction. Full force above ground, half
//    force on the ground, nothing if you're below it (water).
//
// 4) Current
//
//    Pushes you in a constant direction. No force above ground, full
//    force if on the ground or below it (water).
//
// The magnitude of the force is controlled by the length of a controlling
// linedef. The force vector for types 3 & 4 is determined by the angle
// of the linedef, and is constant.
//
// For each sector where these effects occur, the sector special type has
// to have the PUSH_MASK bit set. If this bit is turned off by a switch
// at run-time, the effect will not occur. The controlling sector for
// types 1 & 2 is the sector containing the MT_PUSH/MT_PULL Thing.


#define PUSH_FACTOR 7

/////////////////////////////
//
// Add a push thinker to the thinker list

static void Add_Pusher(int type, int x_mag, int y_mag, mobj_t* source, int affectee)
{
	pusher_t *p = (pusher_t*)W_Calloc(sizeof(*p), THINKID, 0);
	p->thinker.function = (void (__cdecl *)(void))T_Pusher;
	p->source = source;
	p->type = type;
	p->x_mag = x_mag>>FRACBITS;
	p->y_mag = y_mag>>FRACBITS;
	p->magnitude = P_ApproxDistance(p->x_mag,p->y_mag);

	if (source) // point source exist?
	{
		p->radius = (p->magnitude)<<(FRACBITS+1); // where force goes to zero
		p->x = p->source->x;
		p->y = p->source->y;
	}

	p->affectee = affectee;
	P_AddFNThinker(&p->thinker);
}

/////////////////////////////
//
// PIT_PushThing determines the angle and magnitude of the effect.
// The object's x and y momentum values are changed.
//
// tmpusher belongs to the point source (MT_PUSH/MT_PULL).
//
// killough 10/98: allow effect with things in addition to players

RDBOOL PIT_PushThing(mobj_t* thing, pusher_t *p)
{
	if (thing->flags3 & MF3_AMODEL) return true;

	if (   (sentient(thing) || (thing->flags & MF_SHOOTABLE))
		&& !(thing->flags & MF_NOCLIP))
	{
		angle_t pushangle;
		fixed_t speed;
		fixed_t sx = p->x;
		fixed_t sy = p->y;
		int		sn;
		
		speed = (p->magnitude -
			((P_ApproxDistance(thing->x - sx,thing->y - sy)
			>>FRACBITS)>>1))<<(FRACBITS-PUSH_FACTOR-1);
		
		// killough 10/98: make magnitude decrease with square
		// of distance, making it more in line with real nature,
		// so long as it's still in range with original formula.
		//
		// Removes angular distortion, and makes effort required
		// to stay close to source, grow increasingly hard as you
		// get closer, as expected. Still, it doesn't consider z :(
		
		if (speed > 0)
		{
			int x = (thing->x - sx) >> FRACBITS;
			int y = (thing->y - sy) >> FRACBITS;
			speed = (int)(((__int64)p->magnitude << 23) / (x*x+y*y+1));
		}
		
		// If speed <= 0, you're outside the effective radius. You also have
		// to be able to see the push/pull source point.
		
		if (speed > 0 && P_CheckSight(thing, p->source))
		{
			int max = waterdynamics ? FRACUNIT/2 : FRACUNIT;

			pushangle = R_PointToAngle2(thing->x, thing->y, sx, sy);

			if (p->source->type == MT_PUSH) pushangle += ANG180; // away

			pushangle >>= ANGLETOFINESHIFT;

			if (   speed > max
				&& (sn = thing->subsector->sector->heightsec) >= 0
				&& thing->z < sectors[sn].floorheight)
			{
				speed = max;
			}

			thing->momx += FixedMul(speed, finecosine[pushangle]);
			thing->momy += FixedMul(speed, finesine[pushangle]);
		}
	}
	return true;
}

/////////////////////////////
//
// T_Pusher - apply force effects (wind, current, push, pull)
//

void T_Pusher(pusher_t *p)
{
	mobj_t		*thing;
	sector_t	*sec = sectors + p->affectee;
	int			xl, xh, yl, yh, bx, by;
	int			xspeed, yspeed;
	int			i, ht;
	int			tmbbox[4];

	// Be sure the special sector type is still turned on. If so, proceed.
	// Else, bail out; the sector type has been changed on us.
	
	if (!(sec->special & PUSH_MASK)) return;
	
	// For constant pushers (wind/current) there are 3 situations:
	//
	// 1) Affected Thing is above the floor.
	//
	//    Apply the full force if wind, no force if current.
	//
	// 2) Affected Thing is on the ground.
	//
	//    Apply half force if wind, full force if current.
	//
	// 3) Affected Thing is below the ground (underfakefloor effect).
	//
	//    Apply no force if wind, full force if current.

	if (p->type == p_push)
	{
		// Seek out all pushable things within the force radius
		// of this point pusher. Blockmap used; not sectors.
		// p->radius is where force goes to zero

		M_setBoxOffsets(tmbbox, p->x, p->y, p->radius);		
		M_getBlockOffsets(tmbbox, MAXRADIUS, &xl, &xh, &yl, &yh);

		validcount++;

		for (bx = xl; bx <= xh; bx++)
		{
			for (by = yl; by <= yh; by++)
			{
				P_BlockThingsIterator(bx, by, (int (__cdecl *)(struct mobj_s *,void *))PIT_PushThing, p);
			}
		}
		return;
	}
	
	// constant pushers p_wind and p_current

	if (sec->heightsec >= 0)				// special water sector?
	{
		ht = sectors[sec->heightsec].floorheight;
	}
	else
		ht = sec->floorheight;

	validcount++;

	P_SectorTouchingThingsList(sec, false);
	
	for (i = 0; i < link_list.listcnt; i++)
	{
		thing = link_list.mobj_list[i];

		if (   !thing->player
			|| (thing->flags & (MF_NOGRAVITY | MF_NOCLIP)))
		{
			continue;
		}

		if (p->type == p_wind)				// WIND
		{
			if (sec->heightsec < 0)			// NOT special water sector
			{
				if (thing->z > thing->floorz) // above ground
				{
					xspeed = p->x_mag;		// full force
					yspeed = p->y_mag;
				}
				else // on ground
				{
					xspeed = (p->x_mag)>>1;	// half force
					yspeed = (p->y_mag)>>1;
				}
			}
			else							// special water sector
			{
				if (thing->z > ht)			// above ground
				{
					xspeed = p->x_mag;		// full force
					yspeed = p->y_mag;
				}
				else if (viewz < ht			// underfakefloor
					|| waterlevel & 3)		// waterdynamics
				{
					continue;				// no force
				}
				else // wading in water
				{
					xspeed = (p->x_mag)>>1; // half force
					yspeed = (p->y_mag)>>1;
				}
			}
		}
		else								// CURRENT
		{
			if (sec->heightsec < 0)			// NOT special water sector
			{
				if (thing->z > ht)			// above ground
					continue;				// no force
				else
				{							// on ground
					xspeed = p->x_mag;		// full force
					yspeed = p->y_mag;
				}
			}
			else							// special water sector
			{
				if (thing->z > ht)			// above ground
					continue;				// no force
				else // underfakefloor
				{
					xspeed = p->x_mag;		// full force
					yspeed = p->y_mag;
				}
			}
		}
		thing->momx += xspeed<<(FRACBITS-PUSH_FACTOR);
		thing->momy += yspeed<<(FRACBITS-PUSH_FACTOR);
	}

	link_list.listcnt = 0;
}

/////////////////////////////
//
// P_GetPushThing() returns a pointer to an MT_PUSH or MT_PULL thing,
// NULL otherwise.

mobj_t* P_GetPushThing(int s)
{
	mobj_t* thing;
	
	for (thing = sectors[s].thinglist; thing; thing = thing->snext)
	{
		if (thing->type == MT_PUSH || thing->type == MT_PULL)
			return thing;
	}
	return NULL;
}

// Initialize sectors where pushers are present
static void P_SpawnPushers(void)
{
	line_t	*l = lines;
	int		i, s;

	if (R3D_edit) return;
	
	for (i = 0 ; i < numlines ; i++, l++)
	{
		switch(l->special)
		{
		case 224: // wind
			for (s = -1; (s = P_FindSectorFromTag(l->tag,s)) >= 0 ; )
				Add_Pusher(p_wind, l->dx, l->dy, NULL, s);
			break;
		case 225: // current
			for (s = -1; (s = P_FindSectorFromTag(l->tag,s)) >= 0 ; )
				Add_Pusher(p_current, l->dx, l->dy, NULL, s);
			break;
		case 226: // push/pull
			for (s = -1; (s = P_FindSectorFromTag(l->tag,s)) >= 0 ; )
			{
				mobj_t	*thing = P_GetPushThing(s);
				// No MT_P* means no effect - Only one supported per sector
				if (thing)
					Add_Pusher(p_push,l->dx,l->dy,thing,s);
			}
		default:
			break;
		}
	}
}

//
// phares 3/20/98: End of Pusher effects
//

////////////////////////////////////////////////////////////////////////////

// R3D Phased lighting
//
// Added to support ZDOOM type phased lighting specials 21 to 24
//

// Initialize the sectors for phased lighting.
#define	TP_TICKNUM 7

static int phased_count;

// linear progression
static float adjustll[TP_TICKNUM + 2] = {1.0f, 0.7875f, 0.675f, 0.5625f, 0.45f, 0.3375f, 0.225f, 0.1125f, 0.0};

void T_SinglePhased(phased_t *d)
{
	short ll, origll;

	--d->phase_tics;

	if (d->phase_tics < 0) d->phase_tics = 70;
	if (d->phase_tics < 35) return;	// Dark period

	origll = (short)d->slist[1];

	// increment new
	if (d->phase_tics > 52)
		ll = (short)(origll + (float)((255 - origll)*(70 - d->phase_tics))/17.0f);
	else
		ll = (short)(origll + (float)((255 - origll)*(d->phase_tics - 35))/17.0f);

	sectors[d->slist[0]].lightlevel = ll;
}

// Single sector phased lighting special 21
static void	P_SpawnPhasedSectorLight(sector_t *sec)
{
	phased_t *d;

	d = (phased_t*)W_Calloc(sizeof(*d), THINKID, 0);
	d->slist = (int*) W_Malloc(2*sizeof(int), THINKID, 55);

	d->slist[0] = sec - sectors;
	d->slist[1] = (int)sec->lightlevel;

	d->count = 1;

	d->phase_tics = 0;
	d->thinker.function = (void (__cdecl *)(void))T_SinglePhased;

	P_AddFNThinker(&d->thinker);
}

void T_Phased(phased_t *d)
{
	short	ll, origll;
	int		lptr;

	--d->phase_tics;

	if (d->phase_tics < 0)
	{
		if (++d->ptr >= d->count) d->ptr = 0;
		d->phase_tics = TP_TICKNUM;
	}

	// increment new
	origll = (short)(d->slist[d->ptr + d->count]);
	ll = (short)(origll + (float)(255 - origll)*adjustll[d->phase_tics + 1]);
	sectors[d->slist[d->ptr]].lightlevel = ll;

	lptr = d->ptr - 1;
	if (lptr < 0) lptr = d->count - 1;

	// decrement previous
	origll = (short)(d->slist[lptr + d->count]);
	ll = (short)(origll + (float)(255 - origll)*adjustll[TP_TICKNUM - d->phase_tics]);
	sectors[d->slist[lptr]].lightlevel = ll;
}

void subfn_ScanPhasedSectors(int secnum, int nexttype, ushort *hitsec)
{
	int i;

	hitsec[secnum] = phased_count;

	for (i = 0; i < sectors[secnum].linecount; i++)
	{
		int newsec = sectors[secnum].lines[i]->frontsector - sectors;

		if (sectors[newsec].special == nexttype && !hitsec[newsec])
		{
			phased_count++;
			subfn_ScanPhasedSectors(newsec, nexttype == 23 ? 24 : 23, hitsec);
			break;
		}
	}
}

// Single sector phased lighting specials 22/23/24
// 22 - Start sector, 23 - intermediate sector(s), 24 - end sector
static void	P_SpawnPhasedLightingSequence(sector_t *sec)
{
	phased_t	*d;
	ushort		*hitsec = (ushort*)M_Calloc(numsectors * sizeof(ushort), 325);
	int			i;

	phased_count = 1;
	hitsec[sec - sectors] = phased_count;

	subfn_ScanPhasedSectors(sec - sectors, 23, hitsec);

	if (phased_count < 2) return;

	d = (phased_t*)W_Calloc(sizeof(*d), THINKID, 0);
	d->slist = (int*) W_Malloc(2*phased_count*sizeof(int), THINKID, 56);

	d->count = phased_count;

	for (i = 0; i < numsectors; i++)
	{
		if (hitsec[i])
		{
			int ptr = phased_count - hitsec[i];
			// Store backwards
			d->slist[ptr] = i;
			d->slist[ptr + phased_count] = (int)sectors[i].lightlevel;
			sectors[i].special = 0;	// Job done
		}
	}

	d->phase_tics = TP_TICKNUM - 1;
	d->ptr = 0;	// start of list
	d->thinker.function = (void (__cdecl *)(void))T_Phased;

	P_AddFNThinker(&d->thinker);

	M_Free(hitsec);
}
// End phased lighting

void ZoomReset(int resetting)
{
	if (resetting)
	{
		RestoreXhair();
		zoomweapontype = false;
		fieldOfView = setFOV;
	}

	targetFOV = setFOV;

	if (inZoomMode)
	{
		inZoomMode = 0;
		if (!resetting) S_LocalSoundAtVolume(r3ds.zoomout_id, NULL, 1);
	}
}

int CCmdZoom(int argc, char **argv)
{
	if (   zoomweapontype
		&& !R3D_edit
		&& !automap_active)
	{
		if (!inZoomMode && setFOV > 20.f)
		{
			SaveXhair();
			inZoomMode = 1;
			targetFOV = (setFOV - 10.f)/2.f;
			S_LocalSoundAtVolume(r3ds.zoomin_id, NULL, 1);
		}
		else
		{
			ZoomReset(false);
		}
	}
	return true;
}
