//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen, G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_svtexarc.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D, moved to main engine
//
// Based on: Doomsday v1.7.8 JDoom.dll code
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "r3d_compat.h"

#include "doom/p_svtexarc.h"
#include "dd_wad.h"
#include "r_data.h"

#include "con_console.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// Savegame I/O:
void	SV_Write(void *data, int len);
void	SV_WriteShort(short val);
void	SV_Read(void *data, int len);
short	SV_ReadShort();

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

texarchive_t flat_archive;
texarchive_t tex_archive;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
// SV_PrepareTexture
//	Called for every texture and flat in the level before saving by
//	Sv_InitTextureArchives.
//==========================================================================
void SV_PrepareTexture(int tex, RDBOOL isflat, texarchive_t *arc)
{
	int c;
	char *ptr = isflat ? R_LumpNameForNum(tex) : R_TextureNameForNum(tex);

	for (c = 0; c < arc->count; c++)
	{
		if(!_stricmp(arc->table[c].name, ptr))
		{
			return;	// already registered
		}
	}
	
	if (arc->count >= MAX_ARCHIVED_TEXTURES)
		Con_ErrorD("Out of tex archive slots");

	strncpy(arc->table[arc->count++].name, ptr, 8);
}

//==========================================================================
// SV_InitTextureArchives
//	Initializes the texture and flat archives (translation tables).
//	Must be called before saving. The tables are written before any
//	world data is saved.
//==========================================================================
void SV_InitTextureArchives(void)
{
	int			i;
	sector_t	*sec;
	side_t		*sid;

	// Init flats.
	flat_archive.count = 0;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		SV_PrepareTexture(sec->floorpic  , true, &flat_archive);
		SV_PrepareTexture(sec->ceilingpic, true, &flat_archive);
	}

	// Init textures.
	tex_archive.count = 0;

	for (i = 0, sid = sides; i < numsides; i++, sid++)
	{
		SV_PrepareTexture(sid->midtexture   , false, &tex_archive);
		SV_PrepareTexture(sid->toptexture   , false, &tex_archive);
		SV_PrepareTexture(sid->bottomtexture, false, &tex_archive);
	}
}

//==========================================================================
// SV_SearchArchive
//==========================================================================
ushort SV_SearchArchive(texarchive_t *arc, char *name)
{
	int i;

	for (i = 0; i < arc->count; i++)
	{
		if (!_stricmp(arc->table[i].name, name))
			return i;
	}

	return 0;	// Not found ??
}

//==========================================================================
// SV_TextureArchiveNum
//	Returns the archive number of the given texture.
//	It will be written to the savegame file.
//==========================================================================
ushort SV_TextureArchiveNum(int texnum)
{
	char name[9];
	
	strncpy(name, R_TextureNameForNum(texnum), 8);
	name[8] = 0;
	return SV_SearchArchive(&tex_archive, name);
}

//==========================================================================
// SV_TextureArchiveNum
//	Returns the archive number of the given flat.
//	It will be written to the savegame file.
//==========================================================================
ushort SV_FlatArchiveNum(int flatnum)
{
	return SV_SearchArchive(&flat_archive, R_LumpNameForNum(flatnum));
}

//==========================================================================
// SV_GetArchiveFlat
//==========================================================================
int SV_GetArchiveFlat(int archivenum)
{
	char *name = flat_archive.table[archivenum].name;

	int i = W_CheckFlatNumForName(name);

	if (i >= 0) return i;

	// If not found check the PNAMES list
	for (i = 0; i < numlumps; i++)
	{
		if (lumpinfo[i].groupMarker == LGT_PNAMES
			&& !strnicmp(lumpinfo[i].name, name, 8))
		{
			return i;
		}
	}

	// Just get the flat (will be a substituted type).
	i = R_FlatNumForName(name);

	if (i < 0) Con_Error("LOADGAME: tex %8.8s not found", name);

	return i;
}

//==========================================================================
// SV_GetArchiveTexture
//==========================================================================
int SV_GetArchiveTexture(int archivenum)
{
	return R_TextureNumForName(tex_archive.table[archivenum].name, false);
}

//==========================================================================
// SV_WriteTexArchive
//==========================================================================
void SV_WriteTexArchive(texarchive_t *arc)
{
	int i;
	
	SV_WriteShort((short)arc->count);

	for(i=0; i<arc->count; i++)
		SV_Write(arc->table[i].name, 8);
}

//==========================================================================
// SV_ReadTexArchive
//==========================================================================
void SV_ReadTexArchive(texarchive_t *arc)
{
	int i;

	arc->count = SV_ReadShort();

	if (arc->count < 0 || arc->count > MAX_ARCHIVED_TEXTURES)
		Con_ErrorD("Bad tex archive. Savegame is not compatible");

	for (i = 0; i < arc->count; i++)
	{
		SV_Read(arc->table[i].name, 8);
		arc->table[i].name[8] = 0;
	}
}

//==========================================================================
// SV_WriteTextureArchive
//==========================================================================
void SV_WriteTextureArchive(void)
{
	SV_WriteTexArchive(&flat_archive);
	SV_WriteTexArchive(&tex_archive);
}

//==========================================================================
// SV_ReadTextureArchive
//==========================================================================
void SV_ReadTextureArchive(void)
{
	SV_ReadTexArchive(&flat_archive);
	SV_ReadTexArchive(&tex_archive);
}
