//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 id Software, Inc.,
//  Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman,
//  Jaakko Kernen, G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_user.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  BOOM and R3D extensions added, reformatted, moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8 JDoom.dll code and BOOM v2.01
//
// DESCRIPTION:
//	Player related stuff.
//	Bobbing POV/weapon, movement.
//	Pending weapon.
//
//-----------------------------------------------------------------------------


#include <math.h>

#include "doom/p_local.h"
#include "doom/p_view.h"
#include "doom/tables.h"
#include "doom/g_game.h"
#include "con_console.h"
#include "net_main.h"
#include "p_maputil.h"
#include "r_data.h"
#include "r_main.h"
#include "sys_timer.h"
#include "r3d_compat.h"

int maxhealth;		// 100
int healthlimit;	// 200
int armorpoints[2];	// Green and blue points.

// Index of the special effects (INVUL inverse) map.
#define INVERSECOLORMAP		32
#define	FLYMOVEZMAX			16*FRACUNIT

//
// Movement.
//

RDBOOL	CheckStrafe();
void	TS_PlayerInWaterSector(mobj_t *mo);

//
// P_Thrust
// Imparts momentum in the player's forward direction.
//
void P_Thrust(mbf_player_t*	player, angle_t angle, fixed_t move) 
{
	mobj_t *mo = player->plr->mo;

    angle >>= ANGLETOFINESHIFT;

    mo->momx += FixedMul(move, finecosine[angle]); 
    mo->momy += FixedMul(move, finesine[angle]);
}

//===========================================================================
// P_CheckPlayerJump
//	Services jump command.
//===========================================================================
static void P_CheckPlayerJump(mbf_player_t *player, RDBOOL onground, RDBOOL testjump)
{
	ticcmd_t *cmd = &player->cmd;

	if (R3D_edit || waterlevel == 4)
	{
		gdt.plrisjumping = false;
		player->jumptics = 0;
		return;
	}

	// Try to jump onto ledge - see also $jumping
	if (   gdt.plrisjumping
		&& !onground
		&& !player->plr->mo->momx
		&& !player->plr->mo->momy)
	{
		P_Thrust(player, player->plr->mo->angle, FRACUNIT>>1);
		return;
	}

	if (   !(cmd->buttons & BT_JUMP)
		|| cmd->buttons & (BT_SPECIAL|BT_CHANGE))
	{
		return;
	}

	if (player->jumptics <= 0)
	{
		if (gdt.onladder)
		{
			gdt.onladder = false;
			player->plr->mo->momz = mapgravity;
			player->plr->mo->momx = player->plr->mo->momy = 0;
			P_Thrust(player, player->plr->mo->angle, 2*FRACUNIT);
			player->jumptics = 24;
			return;
		}

		if (   player->plr->mo->flags3 & MF3_ONGROUND
			&& cfg.jumpEnabled)
		{
			// Jump, then!
			gdt.plrisjumping = true;
			player->plr->mo->momz = (int)(FRACUNIT * cfg.jumpPower);
			player->jumptics = 24;
			return;
		}
	}

	if (   testjump
		&& cfg.jumpEnabled
		&& ppt.viewer_fake_sector >= 0
		&& waterlevel
		&& waterlevel < 4)
	{
		if (   (waterlevel > 1 && waterlevel < 3)
			|| gdt.plrlineblocked)
		{
			P_Thrust(player, player->plr->mo->angle, FRACUNIT>>3);
		}

		if (   !gdt.plrlineblocked
			&& (waterlevel == 3 || viewz < ppt.fake_floorheight + 5*FRACUNIT))
		{
			player->plr->mo->momz = FRACUNIT;
		}
		
		player->jumptics = 0;
	}
}

//
// P_MovePlayer
//
static int subfn_GRAVITY(int w)
{
	if (w < 3 ) return  mapgravity;
	if (w == 4) return 0;
	return mapgravity/2;
}

void P_MovePlayer (mbf_player_t* player)
{
	mobj_t		*plrmo = player->plr->mo;
    ticcmd_t	*cmd;
	RDBOOL		onground;

    cmd = &player->cmd;

	// Change the angle if possible.v220-35 add DDPF_DELAY
	if ( !(player->plr->flags & (DDPF_FIXANGLES|DDPF_DELAY)) )
	{
		plrmo->angle = cmd->cmd_angle;
		player->plr->lookdir = (float)cmd->cmd_lookdir/(float)DDMAXSHORT * 110.f;
	}

	player->plr->flags &= ~DDPF_DELAY;	// v220-35

	// b211-0 autoAim now set here
	if (player->readyweapon == wp_fist|| player->readyweapon == wp_chainsaw)
	{
		float lookang;

		ma_inf.plr_melee_ht = plrviewheight;

		// v220-8 set melee ht depending on player's up/dowm angle
		if ( (lookang = player->plr->clLookDir) < -1.0f)
		{
			ma_inf.plr_melee_ht += (int)(lookang * 0xA02D);	// equates to 40*FRACUNIT
			if (ma_inf.plr_melee_ht < PLR_VIEWMINHT) ma_inf.plr_melee_ht = PLR_VIEWMINHT;
		}
		else if (lookang > 1.0f)
		{
			ma_inf.plr_melee_ht += (int)(lookang * 0x2D8A);	// equates to 10*FRACUNIT
			if (ma_inf.plr_melee_ht > plrviewheight + PLR_VIEWMAXOFFSET)
				ma_inf.plr_melee_ht = plrviewheight + PLR_VIEWMAXOFFSET;
		}

		autoAim = false;
	}
	else
	{
		ma_inf.plr_melee_ht = 0;
		autoAim = !(cfg.usemlook || cmd->cmd_lookdir || R3D_edit);
	}

	// v220-4
	onground = (plrmo->z <= plrmo->floorz)
		|| (plrmo->onmobj && plrmo->z == plrmo->onmobj->z + plrmo->onmobj->height);

    // Dont let the player control movement if not onground.
	if (onground)
	{
		plrmo->flags3 |= MF3_ONGROUND;
		// If onground then has finished jumping
		player->jumptics = 0;
	}

	if (!onground || (waterlevel > 1 && waterlevel < 4))
	{
		plrmo->flags3 &= ~MF3_ONGROUND;
	}

/*	if (player->plr->flags & DDPF_CAMERA) // $democam
	{
		// Cameramen have a 3D thrusters!
		P_Thrust3D(player, player->plr->mo->angle,
			player->plr->lookdir, cmd->forwardmove*2048,
			cmd->sidemove*2048);
		return;
	}*/

	if (waterlevel > 2) plrmo->momz = 0;

	if ( !(cmd->forwardmove || cmd->sidemove) )
	{
		if (!onground && waterlevel == 3)	// Constant sink rate if submerged
			plrmo->momz = -subfn_GRAVITY(waterlevel);

		P_CheckPlayerJump(player, onground, true);
		return;
	}
	else if (onsurface && !onground && !gdt.plrlineblocked)	// or on surface
	{
		if (abs(plrmo->momx) + abs(plrmo->momy) > 6*FRACUNIT)
		{
			plrmo->momz = -12*FRACUNIT;
			plrmo->momx /= 6;
			plrmo->momy /= 6;
		}
	}

	if (onground || waterlevel)
	{
		int divfactor = waterlevel + 1 - onground;
		int friction, movefactor = P_GetMoveFactor(plrmo, &friction);

		if (divfactor > 0 && waterlevel != 4)
			movefactor /= divfactor;

		if (movefactor)
		{
			if (waterlevel < 3 && onground)
			{
				if (cmd->sidemove)
					P_Thrust(player, plrmo->angle-ANG90,cmd->sidemove*movefactor);
				if (cmd->forwardmove)
					P_Thrust(player, plrmo->angle,cmd->forwardmove*movefactor);
			}
			else if (waterlevel)
			{
				// Speed of diving/swimming/flying depends on the magnitude of clLookDir
				// 0 degrees gives fastest swimming speed and does not change sinking
				// If not moving and not flying then sinking is set at -GRAVITY
				// 110/-110 degrees gives fastest rising/sinking and no forward speed
				// Values between 0 and 110/-110 give a proportional rate.

				int		dive = abs((int)(player->plr->clLookDir)*2383);
				int		swim = 2048 - (abs(dive)>>7);
				int		increment;

				RDBOOL	dir = (cmd->forwardmove > 0)^(player->plr->clLookDir >= 0);
				RDBOOL	move = true;

				//-------------------------------------

				if (dir) dive = -dive;
				movefactor = (movefactor*swim)>>11;
				increment  = cmd->forwardmove*movefactor;

				if (cmd->sidemove)
				{
					P_Thrust(player,
						plrmo->angle-ANG90,
						cmd->sidemove*(waterlevel == 4 ? 2048 : movefactor));
				}

				if (increment)
				{
					static int		lastchange = 0;
					static uint		flytic;

					uint			realtime = Sys_GetRealTime();
					uint			timediff;
					int				flymovez;

					timediff = realtime - flytic;
					flytic = realtime;

					if (timediff > 1000) timediff = 1000;	// 1 fps

					flymovez = 512*timediff;

					// Strafe when flying/swimming means going up/down
					if (waterlevel > 2 && CheckStrafe())
					{
						if (increment > 0)
						{
							if (plrmo->z + plrmo->height < plrmo->ceilingz)
							{
								// smooth direction change
								if (lastchange < 0) lastchange >>= 1;

								if (waterlevel == 4)
								{
									lastchange += flymovez;
									if (lastchange > FLYMOVEZMAX)
										lastchange = FLYMOVEZMAX;
								}
								else
								{
									lastchange = 2*FRACUNIT;
								}
							}
							else
								lastchange = 0;
						}
						else
						{
							if (onground)
							{
								lastchange = 0;
							}
							else
							{
								if (lastchange > 0) lastchange >>= 1;

								if (waterlevel == 4) // flying
								{
									lastchange -= flymovez;
									if (lastchange < -FLYMOVEZMAX)
										lastchange = -FLYMOVEZMAX;
								}
								else
								{
									lastchange = -2*FRACUNIT;
								}
							}
						}

						plrmo->momz = lastchange;

						move = false;
					}
					else
					{
						P_Thrust(player, plrmo->angle,increment);
					}
				}

				if (move)
				{
					if (   (waterlevel > 2 && !onground)
						|| (waterlevel > 1 && player->plr->clLookDir>0))
					{
						plrmo->momz = (waterlevel == 4 ? dive<<1 : dive);
					}
					else if (!onground && !gdt.plrlineblocked)
					{
						plrmo->momz = -subfn_GRAVITY(waterlevel);
					}
				}
			}
			else if (!onground && !gdt.plrlineblocked)
			{
				plrmo->momz = -subfn_GRAVITY(waterlevel);
			}
		}
	}

	if (plrmo->state == &states[S_PLAY])
		P_SetMobjState(plrmo, S_PLAY_RUN1);

	P_CheckPlayerJump(player, onground, true);
}	

//
// P_DeathThink
// Fall on your face when dying.
// Decrease POV height to floor height.
// v220-5 check if in a super hell slime sector
//
#define ANG5   	(ANG90/18)

void P_DeathThink (mbf_player_t* player)
{
    angle_t		angle;
    angle_t		delta;
	RDBOOL		onground;

    P_MovePsprites (player);
	
    // fall to the ground
    if (player->plr->viewheight > 6*FRACUNIT)
		player->plr->viewheight -= FRACUNIT;

    if (player->plr->viewheight < 6*FRACUNIT)
		player->plr->viewheight = 6*FRACUNIT;

    player->plr->deltaviewheight = 0;
	player->plr->flags |= DDPF_FIXANGLES | DDPF_FIXPOS;
    onground = (player->plr->mo->z <= player->plr->mo->floorz);
    P_CalcHeight (player);

	// In netgames we won't keep tracking the killer.
    if(player->attacker && player->attacker != player->plr->mo)
    {
		angle = R_PointToAngle2(player->plr->mo->x,
			player->plr->mo->y,
			player->attacker->x,
			player->attacker->y);
		
		delta = angle - player->plr->mo->angle;
		
		if (delta < ANG5 || delta > (unsigned)-ANG5)
		{
			// Looking at killer,
			//  so fade damage flash down.
			player->plr->mo->angle = angle;
			if(player->damagecount) player->damagecount--;
		}
		else if (delta < ANG180)
			player->plr->mo->angle += ANG5;
		else
			player->plr->mo->angle -= ANG5;
    }
    else if (player->damagecount)
		player->damagecount--;
	
    if (player->cmd.buttons & BT_USE)
		player->playerstate = PST_REBORN;
}

#define LIST_BEGIN		-1
#define LIST_END		-2

weapontype_t P_PlayerFindWeapon(mbf_player_t *player, RDBOOL next)
{
	static int sw_list[] = 
	{ 
		wp_fist, wp_pistol, wp_shotgun, wp_chaingun, wp_missile, wp_chainsaw
	};
	static int reg_list[] = 
	{ 
		wp_fist, wp_pistol, wp_shotgun, wp_chaingun, wp_missile, wp_plasma, 
		wp_bfg, wp_chainsaw
	};
	static int dm2_list[] =
	{
		wp_fist, wp_pistol, wp_shotgun, wp_supershotgun, wp_chaingun, 
		wp_missile,	wp_plasma, wp_bfg, wp_chainsaw
	};

	int *list, num, i;

	if (gamemode == shareware)
	{
		list = sw_list; num = 6;
	}
	else if(commercial_game)
	{
		list = dm2_list; num = 9;
	}
	else
	{
		list = reg_list; num = 8;
	}

	// Find the current position in the weapon list.
	for(i = 0; i < num; i++) { if(list[i] == player->readyweapon) break; }

	// Locate the next or previous weapon owned by the player.
	for(;;)
	{
		// Move the iterator.
		if(next) i++; else i--;
		if(i < 0) i = num-1;
		else if(i > num-1) i = 0;
		// Have we circled around?
		if(list[i] == player->readyweapon) break;
		// A valid weapon?
		if(player->weaponowned[list[i]]) break;
	}
	return (weapontype_t)list[i];
}

//===========================================================================
// P_PlayerThink
//===========================================================================
void P_PlayerThink (mbf_player_t* player)
{
	mobj_t			*plrmo = player->plr->mo;
	ticcmd_t		*cmd;
	weapontype_t	newweapon, oldweapon;
	
	// Selector 0 = Generic (used by default)
	// Selector 1 = Fist
	// Selector 2 = Pistol
	// Selector 3 = Shotgun
	// Selector 4 = Fist
	// Selector 5 = Chaingun
	// Selector 6 = Missile
	// Selector 7 = Plasma
	// Selector 8 = BFG
	// Selector 9 = Chainsaw
	// Selector 10 = Super shotgun
	plrmo->selector = plrmo->selector & ~DDMOBJ_SELECTOR_MASK 
		| (player->readyweapon + 1);
	
//	P_CameraThink(player); // $democam
	
	// Done here as may have been set in P_CameraThink()
	if (player->cheats & CF_NOCLIP)
		plrmo->flags |= MF_NOCLIP;
	else
		plrmo->flags &= ~MF_NOCLIP;

	// chain saw run forward
	cmd = &player->cmd;

	if (plrmo->flags & MF_JUSTATTACKED)
	{
		cmd->cmd_angle = plrmo->angle;	// Don't turn.
		// The client must know of this.
		player->plr->flags |= DDPF_FIXANGLES;
		cmd->forwardmove = (char)0xc64; // = 0xc800/0x200;
		cmd->sidemove = 0;
		plrmo->flags &= ~MF_JUSTATTACKED;
	}
	
	if (player->playerstate == PST_DEAD)
	{
		P_DeathThink (player);
		return;
	}
	
	// Move around.
	// Reactiontime is used to prevent movement
	//  for a bit after a teleport.
	if (plrmo->reactiontime)
		plrmo->reactiontime--;
	else
	{
		if (ma_inf.playerteleport) ma_inf.playerteleport--;
		P_MovePlayer (player);
	}
	
	P_CalcHeight (player);
	P_PlayerInSpecialSector (player);

	// v220-3
	if (thinker_uwater.breath_tics)
		TS_PlayerInWaterSector (plrmo);

	if (player->jumptics > 0) player->jumptics--;
	
	oldweapon = player->pendingweapon;
	
	if(cmd->buttons & BT_SPECIAL)
	{
		// There might be a special weapon change.
		if(cmd->buttons & (BTS_NEXTWEAPON | BTS_PREVWEAPON) &&
			!(cmd->buttons & BTS_PAUSE))
		{
			player->pendingweapon = P_PlayerFindWeapon(player, 
				(RDBOOL)((cmd->buttons & BTS_NEXTWEAPON) != 0));
		}
		cmd->buttons = 0;
	}
	
	// Check for weapon change.
	if (cmd->buttons & BT_CHANGE)
	{
		// The actual changing of the weapon is done
		//  when the weapon psprite can do it
		//  (read: not in the middle of an attack).
		newweapon = (weapontype_t)((cmd->buttons & BT_WEAPONMASK)>>BT_WEAPONSHIFT);
		
		if(!commercial_game && newweapon == wp_supershotgun)
		{
			// In non-Doom II, supershotgun is the same as normal shotgun.
			newweapon = wp_shotgun;
		}
		
		if (   player->weaponowned[newweapon]
			&& newweapon != player->readyweapon)
		{
			// Do not go to plasma or BFG in shareware,
			//  even if cheated.
			if (   (newweapon != wp_plasma && newweapon != wp_bfg)
				|| (gamemode != shareware) )
			{
				player->pendingweapon = newweapon;
			}
		}
	}
	
	// check for use
	if (cmd->buttons & BT_USE)
	{
		if (!player->usedown)
		{
			P_UseLines (player);
			player->usedown = true;
		}
	}
	else
		player->usedown = false;
	
	// cycle psprites
	P_MovePsprites (player);
	
	// Counters, time dependent power ups.
	
	// Strength counts up to diminish fade.
	if (player->powers[pw_strength])
		player->powers[pw_strength]++;	
	
	if (player->powers[pw_invulnerability])
		player->powers[pw_invulnerability]--;
	
	if (player->powers[pw_invisibility])
	{
		if (! --player->powers[pw_invisibility] )
			player->plr->mo->flags &= ~MF_SHADOW;
	}
	
	if (player->powers[pw_infrared])
		player->powers[pw_infrared]--;
	
	if (player->powers[pw_ironfeet])
		player->powers[pw_ironfeet]--;
	
	if (player->damagecount)
		player->damagecount--;
	
	if (player->bonuscount)
		player->bonuscount--;
}

//===========================================================================
// P_SetMessage
//===========================================================================
void P_SetMessage(mbf_player_t *pl, char *msg)
{
void TS_SCPmessage(mbf_player_t *pl, char *msg);
	extern RDBOOL	con_active;

	if (appStatus != 2)
	{
		Con_Message("   %s\n", msg);
		return;
	}

	if (cfg.olddoommessage && !R3D_ScriptedMap)
		pl->hint_msg = msg;
	else
		TS_SCPmessage(pl, msg);

	if (pl - mbf_players == 0 && (cfg.echoMsg || con_active))
		Con_FPrintf(CBLF_CYAN, "%s\n", msg);
}

//===========================================================================
// P_playSetMessage
//===========================================================================
void P_playSetMessage(mbf_player_t *pl, char *msg)
{
	pl->hint_msg = msg;

	if (pl == &mbf_players[0] && cfg.echoMsg)
		Con_playFPrintf(false, CBLF_CYAN, "%s\n", msg);
}

int CCmdTorch(int argc, char **argv)
{
	if ( !(R3D_ScriptedMap || ma_inf.noscriptset & NSF_SCPLIGHT) )
	{
		if (cfg.useTorchLight)
		{
			cfg.useTorchLight = false;
			cfg.setVtxLight = true;
			P_SetMessage(&mbf_players[0], "Torchlight OFF");
		}
		else
		{
			cfg.useTorchLight = true;
			cfg.setVtxLight = false;
			P_SetMessage(&mbf_players[0], "Torchlight ON");
		}
	}
	else
	{
		P_SetMessage(&mbf_players[0], "TORCHLIGHT: Fixed by script");
	}
	return true;
}

int CCmdWalkRun(int argc, char **argv)
{
	cfg.alwaysRun = !cfg.alwaysRun;
	return true;
}
