//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: dd_loop.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 28 (v220-28) 04/Oct/2014 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:	Primary game loop
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <windows.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_network.h"
#include "de_refresh.h"
#include "de_play.h"
#include "de_render.h"
#include "de_graphics.h"
#include "de_audio.h"
#include "de_ui.h"
#include "de_misc.h"

#include "cvars.h"
#include "r3d_compat.h"
#include "r3d_sfx.h"
#include "doom/g_game.h"
#include "doom/m_menu.h"
#include "doom/mbf_player.h"

#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	D_DoAdvanceDemo (void);
void	D_Display (void);
void	W_MakeGammaRamp(float gamma, float contrast, float bright);
void	W_SetOriginalGammaRamp(void);
void	DR_SCP_Drawer(void);
int		UI_OpenPanel(void);	// v220-4
void	Airbar_Drawer(void);
void	setClampPeriod(void);
void	P_CheckPlayerCanMove(mobj_t *mo);
RDBOOL	glMessageBox(const char *msg, int centre, int exit_type, int prendmsg);
void	setPauseState(RDBOOL setpause, RDBOOL mustpause);
void	TS_PUticker(void);
void	TS_PUdisplay(void);
void	GL_CheckVSYNC(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void DD_TryRunTics (void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL	resetNextViewer;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		appStatus = -2;
int		appTerminate = false;
int		useBulletRange;		// v220-16
uint	clamp_period;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

void MessagePump()
{
	MSG		msg;
	int		gotmessage = false;

	// Start by checking Windows messages. This is the message pump.
	while (PeekMessage(&msg, hWndMain, 0, 0, PM_REMOVE) > 0)
	{
		if (msg.message == WM_PAINT)
			GetMessage(&msg, hWndMain, 0, 0);

		TranslateMessage(&msg);
		DispatchMessage(&msg);
		gotmessage = true;
	}

	// v220-31 Overcomes problems with screen savers etc.
	if (!gotmessage) PostMessage(hWndMain, WM_USER, 0, 0);
}

// v220-4 only define if not building the release code
//   R3D has been extensively tested with this enabled
//   to ensure no errors were being generated.
//   NOTE: Enabling this will cause any benefit of using
//         NVidia's 'Threaded optimisation' to be lost.

// #define CHECK_ERROR_GL

//==========================================================================
// DD_GameLoop
//==========================================================================

void DD_GameLoop(void)
{
	static uint lastfpstic = 0;
	static uint fpstic = 0;
	uint		nowtime;

#ifdef CHECK_ERROR_GL
	RDBOOL	gl_err;
#endif
	//---------------------------

	_flushall();
	setClampPeriod();
	GL_CheckVSYNC();

#if 0		// $bulletrange
	// v220-16 initialise bullet range
	if (cfg.bulletRange == 0)
		useBulletRange = STANDARDRANGE;
	else if (cfg.bulletRange == 1)
		useBulletRange = MISSILERANGE;
	else
		useBulletRange = MAXBULLETRANGE;
#else
	useBulletRange = MISSILERANGE;
#endif

	GL_CheckError("startup");	// v220-4 Check for errors during startup

	while(1)
	{
		if (appTerminate)
		{
			setPauseState(false, false);	// FMOD must be unpaused when shutdown
			if (ui_active) UI_End();
			break;
		}

		MessagePump();

		if (ui_active || menu_active || con_freezegame)
		{
			if (!ma_inf.suspendsound)
			{
				ma_inf.suspendsound = true;
				Sfx_SuspendRepeatingChannels();
			}
		}
		else if (ma_inf.suspendsound)
		{
			ma_inf.suspendsound = false;
			Sfx_ReleaseRepeatingChannels();
		}
		
		if (paused)
		{
			if (ui_active)
			{
				MC_WinMidiTime();	// v220-13 $midi
				Sys_Sleep(20);
			}
			else
			{
				stats.fpsnum = 10;
				Sys_Sleep(100);
			}
		}
		else
		{
			if (open_panel_request) UI_OpenPanel();

			S_StartFrame();		// Sound

			DD_TryRunTics();	// Will run at least one tic.

			if (cMode.w_checkwin > 0)
				cMode.w_checkwin = 0;

			// Check for color adjustment changes.
			if (update_gamma)
				W_MakeGammaRamp(vid_gamma, vid_contrast, vid_bright);

			nowtime = Sys_GetRealTime();

			// On each second boundary calculate the frames.
			if (nowtime - lastfpstic >= 1000)
			{
				lastfpstic = nowtime;
				stats.fpsnum = stats.framecount;
				stats.framecount = 0;
			}

			// v220-33 Hand time to any other threads that may be running.
			if (nowtime - fpstic < 2) { Sleep(0); continue; }

			// v220-13 User configurable frame rate cap
			if (!vidVsync && nowtime - fpstic < clamp_period) continue;

			fpstic = Sys_GetRealTime();
		}

		if (ui_active)	// Control Panel
		{
			UI_Drawer();
			Sys_Sleep(10);
		}
		else
		{
			D_Display();		// Draw the game graphics.

			// Increment the frame counter.
			stats.framecount++;

			GL_DrawFilter();

			// Draw Menu
			if (menu_active)
			{
				M_Drawer();
			}
			else if (gamestate == GS_LEVEL && ppt.thinkers_have_run)
			{
				if (ma_inf.checkPlayer && !(ppt.viewplrmo->flags & MF_NOCLIP))
				{
					ma_inf.checkPlayer = false;
					P_CheckPlayerCanMove(ppt.viewplrmo);
				}

				if (R3D_edit)	// R3D
					Edit_DrawerEditWindow();
				else
				{
					Airbar_Drawer();	// e.g. FPS information.
					DR_SCP_Drawer();	// b211-0 scp messages
					TS_PUdisplay();		// v220-35 display all messages
				}

			}

			if (sound_info && sfx_avail && !automap_active)
				Sfx_DebugInfo();	// debug info

			// Draw console.
			if (!menu_active && ConsoleY != 0) Con_Drawer();
		}

		// Flush buffered stuff to screen (blits everything).
		GL_DoUpdate();

#ifdef CHECK_ERROR_GL
		if ( (gl_err = GL_CheckError("")) )
			Con_ErrorD("gl Error - see Risen3D.out");
#endif
	}

	GL_CheckError("exit");	// v220-4 Check for errors on exit
	Sys_Quit(false);
}

//===========================================================================
// DD_TryRunTics
//	Run at least one tic.
//  v220-28 add con_freezegame
//===========================================================================
static void DD_TryRunTics (void)
{
	int		counts;

	// We'll tick away all the tics we can.

	realtics = 0;	// Zero the real tics counter.

	Net_Update();	// Read input devices

	counts = (ui_active || con_freezegame) ? realtics : availabletics;
				
	// Run the tics.
	while(counts--)
	{
		if (thinkercap.next != &thinkercap)	// Check initialized
			P_CheckPtcPlanes();				// New ptcgens for planes?

		if (ui_active)
		{
			availabletics = 0;
			UI_Ticker();	// User interface ticks.
		}
		else if (con_freezegame)	// v220-28 pause game
		{
			availabletics = 0;
			Con_Ticker();	// Console ticks
		}
		else
		{
			ddplayer_t *pl = players;

			if (advancedemo) D_DoAdvanceDemo();

			M_Ticker();				// Menu ticker
			G_Ticker();				// Game ticker
			Con_Ticker();			// Console.
			TS_PUticker();			// v220-35 pick ups ticker

			// We can't send FixAngles messages to ourselves, so it's done here.
			if (pl->ingame)
			{
				if (resetNextViewer)	// not tp specific as also used at game start
					pl->flags |= DDPF_DELAY;

				if (pl->flags & (DDPF_LOCAL|DDPF_FIXANGLES) == (DDPF_LOCAL|DDPF_FIXANGLES))
				{
					pl->flags &= ~DDPF_FIXANGLES;
					pl->clAngle = pl->mo->angle;
					pl->clLookDir = pl->lookdir;
				}
			}
		}

		// This tick is done, advance time.
		systics++;

		if (!ui_active && !con_freezegame) 
		{
			if (availabletics > 0) availabletics--;

			gametic++;
		}
	}
}

