//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson, Jaakko Kernen, Jack Vermeulen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: dd_wad.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  modified/extended for Risen3D
//  Runtime (un)loading, replacement etc. is not supported
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 7 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:	Routines for loading, interrogating  wad files
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <time.h>
#include <io.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/timeb.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_misc.h"

#include "r_data.h"
#include "r_things.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

#define TYPE_UNDEF	0
#define TYPE_IWAD	1
#define TYPE_DETAIL	2

// TYPES -------------------------------------------------------------------

typedef struct
{
	char identification[4];
	int numlumps;
	int infotableofs;
} wadinfo_t;

typedef struct
{
	int filepos;
	int size;
	char name[8];
} filelump_t;

typedef struct
{
	char	*start, *end;	// Start and end markers.
} grouping_t; 

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

RDBOOL	chkLinkedType(char *nptr);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char		detail_wad[MAXFNAMEX];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int			Detailhandle;
int			RISEN3DWADhandle;
int			IWADhandle;
int			pwad_nodesNotRebuilt;

lumpinfo_t	*lumpinfo = NULL;
int			numlumps = 0;

// The file filenameRecords.
int				numfilenameRecords = 0;
char			actualwadname[MAXFNAME];
filerecord_t	*filenameRecords = NULL;

// v220-22 moved here
int			has_detail;	// Whether detail textures are loaded

// PRIVATE DATA DEFINITIONS ------------------------------------------------

typedef struct {
	char *markerstart;
	char *markerend;
} marker_name_t;

// Note: the values must align with LGT_NONE etc.
static marker_name_t groups[] = 
{
	{""			, ""		},	// 0 = none
	{"F_START"	, "F_END"	},	// 1 Flats
	{"S_START"	, "S_END"	},	// 2 Sprites
	{"P_START"	, "P_END"	},	// 3 Pnames (helper settings)
    {"TX_START"	, "TX_END"	},	// 4 TX hi-res textures (jpg or png) for both TEXTURES and FLATS
	{"C_START"	, "C_END"	},	// 5 Colormaps
	{"HI_START"	, "HI_END"	}	// 6 Support for ZDoom Marker (similar to TX_START)
};

static int		iwad_file_idx;
static short	filerecnum = -1;

// CODE --------------------------------------------------------------------

//==========================================================================
//
// File Record Handling
//
//==========================================================================

// Allocate a new file record.
static filerecord_t *W_RecordNew()
{
	filenameRecords = (filerecord_t*)W_ReallocCLR(filenameRecords, sizeof(filerecord_t) * (++numfilenameRecords), STATICID, 0);
	return &filenameRecords[numfilenameRecords - 1];
}

// v220-7 check the sent date does not exceed any of the map files stored
RDBOOL MapFileDateGreaterThan(long time_write)
{
	int i;

	for (i = 0; i < numfilenameRecords; i++)
	{
		if (   filenameRecords[i].numlevels
			&& filenameRecords[i].time_write > time_write)
		{
			return true;
		}
	}

	return false;
}

// v220-6 get the file write time from modidx
int RetrieveFileDateByModidx(int ref, uint *csm)
{
	actualwadname[0] = 0;

	if (ref >= 0 && ref < numfilenameRecords)
	{
		_splitpath(filenameRecords[ref].filename, NULL, NULL, actualwadname, NULL);
		*csm = filenameRecords[ref].checksum;
		return filenameRecords[ref].time_write;
	}

	Con_Error("ERROR: ref %i\n", ref);
	return 0;	// happy compiler
}

// get the file write time by file number offset in filenameRecords - jav
long RetrieveFileDateByNum(int fnum)
{
	return filenameRecords[fnum].time_write;
}

// get the file write time by file name searching filenameRecords - jav
long RetrieveFileDateByName(char *filename)
{
    int i;

	for(i = 0; i < numfilenameRecords; i++)
    {    
		if (!_stricmp(filenameRecords[i].filename, filename))
		{
			return filenameRecords[i].time_write;
		}
    }
    
    return 0;
}

// get the file index in filenameRecords searching by name - jav
int RetrieveFileNumberByName(char *filename)
{
    int i;

	for (i = 0; i < numfilenameRecords; i++)
    {    
		if (!_stricmp(filenameRecords[i].filename, filename))
			return i;
    }
    
	// jav this should never occur unless we screwed up
    Con_Error("Filename : %s not found", filename);

    return -1;
}


//==========================================================================
//
// Lumpinfo inserting, filling, removal and other operations.
//
//==========================================================================

//===========================================================================
// W_FillLumpInfo
//	Writes the correct data into a lumpinfo_t entry.
//===========================================================================
static void W_FillLumpInfo
	(int liIndex, filelump_t *flump, filerecord_t *rec, int groupTag)
{
	lumpinfo_t	*li = lumpinfo + liIndex;

	li->handle   = rec->handle;
	li->position = flump->filepos;
	li->size     = flump->size;
	li->modidx   = filerecnum;

	strncpy(li->name, flump->name, 8);
	li->name[8] = 0;

	li->groupMarker = groupTag;

	if (!li->size)
	{
		if (groupTag == LGT_SPRITES && chkLinkedType(flump->name))
		{
			// no change
		}
		else
			li->groupMarker = LGT_MARKER;
	}
	else	// Trap wad error with colormaps
	if (   groupTag == LGT_PNAMES
		&& DoomLumpIdentify(liIndex, NULL, NULL, DLI_GRAPHIC) == LMP_FLAT)
	{
		char	liname[9];
		int		len;

		strncpy(liname, li->name, 8);
		liname[8] = 0;
		_strupr(liname);

		len = strlen(liname);

		// This is likely to be a colormap even without this test.
		// See if the name has MAP at the end.
		if (len > 3 && strstr(liname, "MAP") == &liname[len - 3])
		{
			pr_OFinsert("WAD ERROR FIXED: name %8.8s size %i between wrong markers\n", liname, li->size);
			li->groupMarker = LGT_COLORMAPS;
		}
	}

	li->lumptype = LMP_UNDEF;
}

// Moves the rest of the lumps forward.
static void W_InsertAndFillLumpRange(int toIndex, filelump_t *lumps, int num,
							  filerecord_t *rec, int groupTag)
{
	int		i, count = numlumps - toIndex;

	// First, if necessary, move the lumps
	if (num && count > 0 && toIndex >= 0 && toIndex < numlumps)
	{
		memmove(&lumpinfo[toIndex + num], &lumpinfo[toIndex], sizeof(lumpinfo_t) * count);
	}

	for (i = 0; i < num; i++)	// Add the new lumps
	{
		W_FillLumpInfo(toIndex+i, lumps+i, rec, groupTag);
	}

	numlumps += num;	// Update the number of lumps.
}

// Reallocates lumpinfo and lumpcache. 
static void W_ResizeLumpStorage(int numitems)
{
	lumpinfo = (lumpinfo_t*)W_ReallocCLR(lumpinfo, sizeof(lumpinfo_t) * numitems, STATICID, 203);
}

// Look for the named lump, starting from the specified index.
static int W_ScanForName(int comp, int startfrom)
{
	int			i;
	lumpinfo_t	*lump_p;

	if (startfrom < 0 || startfrom > numlumps-1 || comp == -1) return -1;

	// Start from the beginning.
	for(i = startfrom, lump_p = lumpinfo + startfrom; i < numlumps; i++, lump_p++)
	{
		switch (comp)
		{
		case LGT_FLATS:
			if (!strnicmp(lump_p->name, "FF_END", 8) || !strnicmp(lump_p->name, "F_END", 8)) return i;
			break;
		case LGT_SPRITES:
			if (!strnicmp(lump_p->name, "SS_END", 8) || !strnicmp(lump_p->name, "S_END", 8)) return i;
			break;
		case LGT_PNAMES:
			if (!strnicmp(lump_p->name, "PP_END", 8) || !strnicmp(lump_p->name, "P_END", 8)) return i;
			break;
		case LGT_TX:
			if (!strnicmp(lump_p->name, "TX_END", 8) || !strnicmp(lump_p->name, "HI_END", 8)) return i;
			break;
		case LGT_COLORMAPS:
			if (!strnicmp(lump_p->name, "CC_END", 8) || !strnicmp(lump_p->name, "C_END", 8)) return i;
			break;
		default:
			break;
		}
	}
	return -1;
}

//==========================================================================
//
// W_InsertLumps
//
// Inserts the lumps in the fileinfo/record to their correct places in
// the lumpinfo. Also maintains lumpinfo/filenameRecords that all data is valid.
//
// The current placing of the lumps is that flats and sprites are added
// to previously existing flat and sprite groups. All other lumps are
// just appended to the end of the list.
//
//==========================================================================

static void W_InsertLumps(filelump_t *fileinfo, filerecord_t *rec, int type)
{
	int			i, k, to, num;
	filelump_t	*flump = fileinfo;
	int			inside = LGT_NONE;	// Not inside any group.
	int			groupFirst = 0;		// First lump in the current group.
	int			maxNumLumps = numlumps + rec->numlumps; // This must be enough.
	int			csm = 0;
	char		flname[10];

	// Allocate more memory for the lumpinfo.
	W_ResizeLumpStorage(maxNumLumps);

	for (i = 0; i < rec->numlumps; i++, flump++)
	{
		if (type == TYPE_DETAIL)
		{
			W_FillLumpInfo(numlumps++, flump, rec, LGT_DETAIL);
			continue;
		}

		strncpy(flname, flump->name, 8);
		flname[8] = 0;

		if (type == TYPE_IWAD)
		{
			csm += flump->size;
			for(k = 0; k < 8; k++) csm += flump->name[k];
		}
		
		if (inside == LGT_NONE)
		{
			// We are currently not inside any group.
			if (!_stricmp(flname, "FF_START") || !_stricmp(flname, "F_START"))
				inside = LGT_FLATS;
			else if (!_stricmp(flname, "SS_START") || !_stricmp(flname, "S_START"))
				inside = LGT_SPRITES;
			else if (!_stricmp(flname, "PP_START") || !_stricmp(flname, "P_START"))
				inside = LGT_PNAMES;
			else if (!_stricmp(flname, "TX_START") || !_stricmp(flname, "HI_START"))
				inside = LGT_TX;
			else if (!_stricmp(flname, "CC_START") || !_stricmp(flname, "C_START"))
				inside = LGT_COLORMAPS;

			if (inside != LGT_NONE)
			{
				// We have entered a group! Go to the next lump.
				groupFirst = i+1;
				continue; 
			}

			W_FillLumpInfo(numlumps++, flump, rec, inside);
		}
		else 
		{
			// We are currently inside a group; check for end marker.
			switch (inside)
			{
			case LGT_FLATS:
				if (!strnicmp(flname, "FF_END", 8) || !strnicmp(flname, "F_END", 8)) goto endmarker_found;
				break;
			case LGT_SPRITES:
				if (!strnicmp(flname, "SS_END", 8) || !strnicmp(flname, "S_END", 8)) goto endmarker_found;
				break;
			case LGT_PNAMES:
				if (!strnicmp(flname, "PP_END", 8) || !strnicmp(flname, "P_END", 8)) goto endmarker_found;
				break;
			case LGT_TX:
				if (!strnicmp(flname, "TX_END", 8) || !strnicmp(flname, "HI_END", 8)) goto endmarker_found;
				break;
			case LGT_COLORMAPS:
				if (!strnicmp(flname, "CC_END", 8) || !strnicmp(flname, "C_END", 8)) goto endmarker_found;
				break;
			default:
				break;
			}

			continue;

endmarker_found:;

			// This is how many lumps we'll add.
			num = i - groupFirst;

			// Find the existing group.
			to = W_ScanForName(inside, 0);

			if (to < 0)
			{
				// There is no existing group. Include the start
				// and end markers in the range of lumps to add.
				groupFirst--;
				num += 2;
				to = numlumps;
			}

			W_InsertAndFillLumpRange(to, fileinfo + groupFirst, num, rec, inside);

			inside = LGT_NONE;	// Exit this group.
		}
	}

	if (type == TYPE_IWAD) pwadchecksum[0] = csm;

	rec->checksum = csm;

	// It may be that all lumps weren't added. Make sure we don't have
	// excess memory allocated (=synchronize the storage size with the 
	// real numlumps).
	W_ResizeLumpStorage(numlumps);

	// Update the record with the number of lumps that were loaded.
	rec->numlumps -= maxNumLumps - numlumps;
}


//==========================================================================
//
// W_AddFile
//
// Files with a .wad extension are wadlink files with multiple lumps,
// other files are single lumps with the base filename for the lump name.
//
// Returns true if the operation is successful.
//
//==========================================================================

RDBOOL W_AddFile(char *fpath, char *fname, RDBOOL pwadHasLevels)
{
	wadinfo_t		header;
	int				handle, length;
	filelump_t		*fileinfo, singleinfo;
	filelump_t		*freeFileInfo = NULL;
	filerecord_t	*rec;

	struct _finddata_t	filedata;
	struct _timeb		timebuffer;
	int    hFile;

	char	filename[MAXPATHNAMESIZE];	// Added for implicit
	char	szFname[MAXFNAME];
	char	szExt[MAXEXT];

	int		i;
	int		type = TYPE_UNDEF;

	//---------------------------------------------

	sprintf(filename, "%s%s", fpath, fname);

	if (!Dir_IsAbsolute(filename))	// Check not a base path
		sprintf(filename, "%s%s", ddRuntimeDir.path, fname);

	Dir_ConvertSlashes(filename);

	if ( (handle = open(filename, _O_RDONLY|_O_BINARY)) == -1 )
		Con_Error("ERROR: %s not found", filename);

	// Do not enter a duplicate.
	for (i = 0; i < numfilenameRecords; i++)
	{
		if (!_stricmp(filenameRecords[i].filename, filename))
		{
			Con_Error("AddFile: Duplicate file %s defined", filename);
		}
	}

	_splitpath(filename, NULL, NULL, szFname, szExt);

	if (!pwadHasLevels && !_stricmp(szExt,".wad"))
	{
		char *ptr = (char*)calloc(sizeof(filename), 1);

		strcpy (ptr, filename);

		for (i = 0; filename[i]; i++)
			ptr[i] = toupper(filename[i]);

		if (strstr(ptr, "\\DATA\\RISEN3D.WAD"))
		{
			RISEN3DWADhandle = handle;
		}
		else
		if (!stricmp(ptr, detail_wad))
		{
			has_detail = 1;	// has_detail must be either 1 (as first file loaded) or 0
			type = TYPE_DETAIL;
			Detailhandle = handle;
		}

		M_Free(ptr);
	}

	// Get a new file record.
	filerecnum++;
	rec = W_RecordNew();
	strcpy(rec->filename, filename);
	rec->handle = handle;

	// This will have been preset to -1 if an add_on file has been sent
	if (ge_inf.addon_handle == -1) ge_inf.addon_handle = handle;

	if (verbose) pr_OFinsert("AddFile   : %s\n", rec->filename);

    // date check added - jav	
	if ( (hFile = _findfirst(rec->filename , &filedata )) != -1 )
	{
		rec->time_write = filedata.time_write;
		rec->time_create = filedata.time_create;
		rec->numlevels = pwadHasLevels;
		_findclose( hFile );

		if (pwadHasLevels)
		{
			_ftime( &timebuffer );

			if (rec->time_write > timebuffer.time)
			{
				changeFileDate(filename);	// change date - gj
				rec->time_write = timebuffer.time;
			}
		}
	}
	else
	{
		// Terminate program if user is sending invalid filenames
		Con_Error("AddFile: Could not access file %s", rec->filename);
    }

	if (!_stricmp(filename + strlen(filename) - 4, ".wad"))
	{
		// WAD
		read(handle, &header, sizeof(header));

		if (!strnicmp(header.identification, "IWAD", 4))
		{
			type = TYPE_IWAD;
			IWADhandle = handle;
			iwad_file_idx = RetrieveFileNumberByName(filename);
		}
		else if (!strnicmp(header.identification, "PWAD", 4))
		{
			rec->pwad = true;
		}
		else
		{
			rec->pwad = false;
		}

		if (!(length = header.numlumps*sizeof(filelump_t)))
			Con_ErrorD("AddFile: file has no lumps");

		freeFileInfo = fileinfo = (filelump_t*)M_Malloc(length, 7);

		lseek(handle, header.infotableofs, SEEK_SET);
		read(handle, fileinfo, length);
		rec->numlumps = header.numlumps;
	}
	else
	if (ge_inf.devmap || !_stricmp(filename + strlen(filename) - 4, ".lmp"))
	{ 
		// Single lump file
		fileinfo = &singleinfo;
		singleinfo.filepos = 0;
		singleinfo.size = filelength(handle);
		M_ExtractFileBase(filename, singleinfo.name);
		rec->numlumps = 1;
	}
	
	// Insert the lumps to lumpinfo, into their rightful places.
	W_InsertLumps(fileinfo, rec, type);

	if (freeFileInfo) { M_Free(freeFileInfo); freeFileInfo = NULL; }

	return true;
}

void closeWADhandles()
{
	for (int i = 0; i < numfilenameRecords; i++)
	{
		if (filenameRecords[i].handle > 0)
		{
			close(filenameRecords[i].handle);
			filenameRecords[i].handle = 0;
		}
	}
}

//===============================================================================
// FREEDOOMcheck
//===============================================================================
void FREEDOOMcheck(void)
{
	RDBOOL	freedoom1check = false;
	RDBOOL	freedoom2check = false;

	int	i;

	// Note: analysename not set at this point
	for (i = 0; i < numfilenameRecords; i++)
	{
		char	*fptr = filenameRecords[i].filename;

		if (strstr(fptr, "freedmPWAD") || strstr(fptr, "freedoom1PWAD"))
			freedoom1check = true;
		if (strstr(fptr, "freedoom2PWAD")) freedoom2check = true;
	}

	pr_OFdirect("\n");

	if (freedoom1check)
	{
		if (gamemission != doom) Con_ErrorD("Wrong IWAD selected - use Doom\n");
		if (gamemode != registered && CFX)
		{
			Con_ErrorD(	"Sorry but Risen3D can only run freedoom with the\n"
						"shareware version of doom in basic doom mode\n");
		}
	}

	if (freedoom2check && gamemission != doom2)
		Con_ErrorD("Wrong IWAD selected - use Doom2\n");
}

//==========================================================================
//
// Check routines modified for - R3D
//

int W_CheckTXNumForName(char *name)
{
	lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards

	// v220-22 exclude search in basic doom mode with the cod addon wad
	if (!CFX && ge_inf.addon_flags & AF_COD) return -1;

	while (--lump_p >= lumpinfo)
	{
		if ((lump_p->groupMarker == LGT_TX) &&
			!strnicmp(lump_p->name, name, 8))
		{
			return lump_p - lumpinfo;
		}
	}

	lump_p = lumpinfo + numlumps; // Scan backwards

	while (--lump_p >= lumpinfo)
	{
		if ((lump_p->groupMarker == LGT_HI) &&
			!strnicmp(lump_p->name, name, 8))
		{
			return lump_p - lumpinfo;
		}
	}

	return -1;
}

int W_CheckNumForName(char *name)
{
	int	lump;

	if (!name[0] || strlen(name) > 8) return -1;

	if ( (lump = W_CheckTXNumForName(name)) < 0)
	{
		lumpinfo_t *lump_p = lumpinfo + numlumps;

		while(--lump_p >= lumpinfo)
		{
			if (lump_p->groupMarker != LGT_DETAIL &&
				!strnicmp(lump_p->name, name, 8))
			{
				return lump_p - lumpinfo;
			}
		}
	}

	return lump;
}

int W_CheckDetailNumForName(char *name)
{
	int i;

	if (!name || !name[0]) return -1;

	for (i = 0; i < numlumps; i++)
	{
		if (lumpinfo[i].groupMarker == LGT_DETAIL
			&& !strnicmp(lumpinfo[i].name, name, 8))
		{
			return i;
		}
	}

	return -1;
}

int W_CheckSpriteNumForName(char *name)
{
	lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards

	while (--lump_p >= lumpinfo)
	{
		if (lump_p->groupMarker == LGT_SPRITES
			&& !strnicmp(lump_p->name, name, 8))
		{
			return lump_p - lumpinfo;
		}
	}

	return -1;
}

int W_CheckFlatNumForName(char *name)
{
	lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards

	while (--lump_p >= lumpinfo)
	{
		if (lump_p->groupMarker == LGT_FLATS &&
			!strnicmp(lump_p->name, name, 8))
		{
			return lump_p - lumpinfo;
		}
	}

	return -1;
}

int W_CheckTextureMissingNumForName(char *name)
{
	lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards
	int	lump = -1;

	while (--lump_p >= lumpinfo)
	{
		if (   !strnicmp(lump_p->name, name, 8)
			&& DoomLumpIdentify(lump_p - lumpinfo, NULL, NULL, DLI_GRAPHIC) == LMP_PATCH)
		{
			return lump_p - lumpinfo;
		}
	}

	return lump;
}

//==========================================================================
// Returns -1 if name not found.
//==========================================================================
int W_CheckTextureNONTXNumForName(char *name)
{
	lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards

	while (--lump_p >= lumpinfo)
	{
		if (lump_p->groupMarker != LGT_DETAIL &&
			lump_p->groupMarker != LGT_FLATS &&
			lump_p->groupMarker != LGT_TX &&
			lump_p->groupMarker != LGT_HI &&
			!strnicmp(lump_p->name, name, 8))
		{
			return lump_p - lumpinfo;
		}
	}

	return -1;
}

//==========================================================================
// Returns -1 if name not found.
//==========================================================================
int W_CheckTextureNumForName(char *name)
{
	int	lump = -1;

	if ( (lump = W_CheckTXNumForName(name)) < 0)
	{
		return W_CheckTextureNONTXNumForName(name);
	}

	return lump;
}

//==========================================================================
//
// Added for R3D
// W_FindScriptLump
// W_CheckRisen3DWADNumForName
// W_CheckTextureNONTXNumForName
// W_CheckOriginalPatch
// W_CheckIWADNumForName
// W_IWADtype
//
//==========================================================================

RDBOOL pwadHandle(int handle)
{
	return handle != IWADhandle
		&& handle != Detailhandle
		&& handle != RISEN3DWADhandle;
}

//==========================================================================
// Returns -1 if name not found.
//==========================================================================
int W_FindScriptLump (char *name, int *lastlump)
{
	if (haspwads)
	{
		lumpinfo_t *lump_p = lumpinfo + *lastlump;

		while (lump_p < lumpinfo + numlumps)
		{
			if (pwadHandle(lump_p->handle) &&
				!strnicmp(lump_p->name, name, 8))
			{
				int lump = lump_p - lumpinfo;
				*lastlump = lump + 1;
				return lump;
			}
			lump_p++;
		}
	}

	*lastlump = numlumps;
	return -1;
}

//==========================================================================
// Returns -1 if name not found.
//==========================================================================
RDBOOL W_OpenLatestScriptLump (char *name, char *mapid)
{
	if (haspwads)
	{
		lumpinfo_t	*lump_p = lumpinfo + numlumps; // Scan backwards

		if (!mapid || !mapid[0]) return false;

		while (--lump_p >= lumpinfo)
		{
			if (  !strnicmp(lump_p->name, name, 8)
				&& pwadHandle(lump_p->handle))
			{
				char	*nextstr;

				SR_LoadPWADSCPLump(lump_p - lumpinfo);

				while( (nextstr = SR_NextSTR(false)) )
				{
					if (!_stricmp(nextstr, mapid))
					{
						SR_RewindScript();
						return true;
					}
				}

				SR_EndScript();
			}
		}
	}

	return false;
}

//==========================================================================
// Returns -1 if name not found.
//==========================================================================
int W_CheckIWADNumForName(char *name)
{
	lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards

	// Scan backwards in case name is duplicated in iwad
	// E.g. SW18_7 (DOOM2 iwad).
	while (--lump_p >= lumpinfo)
	{
		if (lump_p->handle == IWADhandle &&
			!strnicmp(lump_p->name, name, 8))
		{
			return lump_p - lumpinfo;
		}
	}

	return -1;
}

//==========================================================================
// EXITS if name not found.
//==========================================================================
int W_CheckRisen3DWADNumForName(char *name, RDBOOL eror)
{
	lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards

	while (--lump_p >= lumpinfo)
	{
		if (lump_p->handle == RISEN3DWADhandle &&
			!strnicmp(lump_p->name, name, 8))
		{
			return lump_p - lumpinfo;
		}
	}

	if (eror)
	{
		Con_Error("ERROR lump \"%s\" not found;\n"WADV, name);
	}

	return -1;
}

//==========================================================================
// Scans pwad for R3D_RCKT, R3D_GREN or R3D_PBMB
// Lumps only scanned if an SCP_DEFN lump is present
// Sets val where found to 1, 2 or 3 else 0
// Returns tic val if R3D_GREN is found
//==========================================================================
int W_CheckPwadSwitch(int *type)
{
	*type = 0;

	if (haspwads && W_CheckNumForName("SCP_DEFN") >= 0)
	{
		lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards

		while (--lump_p >= lumpinfo)
		{
			if (pwadHandle(lump_p->handle))
			{
				if (!strnicmp(lump_p->name, "R3D_RCKT", 8))
				{
					*type = 1; break;
				}
				if (!strnicmp(lump_p->name, "R3D_GREN", 8))
				{
					char str[6];
					int  n;

					*type = 2;

					if (!lump_p->size || lump_p->size > 5) return 0;

					lseek(lump_p->handle, lump_p->position, SEEK_SET);

					// only a number (in text form) should be present
					read(lump_p->handle, (void*)str, lump_p->size);
					str[lump_p->size + 1] = 0;
					
					// grenade_tics;
					return (n = atoi(str)) > 0 ? n : 0;
				}
				if (!strnicmp(lump_p->name, "R3D_PBMB", 8))
				{
					*type = 3; break;
				}
			}
		}
	}

	return 0;
}

//==========================================================================
// Looks for a standard Doom patch, Returns -1 if name not found.
//==========================================================================
int W_CheckOriginalPatch(char *name)
{
	lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards

	while (--lump_p >= lumpinfo)
	{
		if (   lump_p->groupMarker != LGT_TX
			&& lump_p->groupMarker != LGT_HI
			&& lump_p->groupMarker != LGT_DETAIL
			&& !strnicmp(lump_p->name, name, 8))
		{
			int lump = lump_p - lumpinfo;

			if (DoomLumpIdentify(lump, NULL, NULL, DLI_GRAPHIC) == LMP_PATCH)
				return lump;
		}
	}

	return -1;
}

// Search PWAD for name
//  returns -1 if not found
int W_CheckPWADNumForName(char *name)
{
	if (haspwads)
	{
		lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards
		
		// Scan backwards in case name is duplicated
		while (--lump_p >= lumpinfo)
		{
			if (  !strnicmp(lump_p->name, name, 8)
				&& pwadHandle(lump_p->handle))
			{
				return lump_p - lumpinfo;
			}
		}
	}

	return -1;
}

int W_GetPWADsound(char *name)
{
	int type, lump = W_CheckPWADNumForName(name);

	if (lump >= 0 && (type = DoomLumpIdentify(lump, NULL, NULL, DLI_SOUND)) == LMP_SND || type == LMP_WAV)
	{
		return lump;
	}
	return -1;
}

// Check texture source v220-3
RDBOOL W_CheckPWADtexture(char *name)
{
	if (haspwads)
	{
		int idx = R_CheckTextureNumForName(name);
		int lmp = idx >= 0 ? texArray[idx]->patches[0].patch : -1;

		return (lmp >= 0 && pwadHandle(lumpinfo[lmp].handle));
	}

	return false;
}

// Search for colormap
int W_CheckColormapForName(char *name)
{
	lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards
	
	// Scan backwards in case name is duplicated
	while (--lump_p >= lumpinfo)
	{
		if (  !strnicmp(lump_p->name, name, 8)
			&& lump_p->groupMarker == LGT_COLORMAPS)
		{
			return lump_p - lumpinfo;
		}
	}

	return -1;
}

// Search PWAD for lo-res graphics lumps only
int W_CheckPWADNumForLoResName(char *name)
{
	if (haspwads)
	{
		lumpinfo_t *lump_p = lumpinfo + numlumps; // Scan backwards
		
		// Scan backwards in case name is duplicated
		while (--lump_p >= lumpinfo)
		{
			if (   pwadHandle(lump_p->handle)
				&& lump_p->groupMarker != LGT_TX
				&& lump_p->groupMarker != LGT_HI
				&& lump_p->groupMarker != LGT_DETAIL
				&& !strnicmp(lump_p->name, name, 8))
			{
				int	lump = lump_p - lumpinfo;
				int type = DoomLumpIdentify(lump, NULL, NULL, DLI_GRAPHIC);
				
				if (type == LMP_PATCH || type == LMP_FLAT)
					return lump;
			}
		}
	}

	return -1;
}

RDBOOL subfn_IwadName(char *fname, char *name)
{
	int len1 = strlen(fname);
	int len2 = strlen(name);

	if (len2 > len1) return 0;
	return (!_stricmp(fname + len1 - len2, name));
}

int W_IWADtype()
{
	// Note has_detail will be 1 if a detail.wad was loaded else
	// it will be 0. If the detail.wad is used it will always be
	// the first wad loaded. Since the iwad is always the next to
	// be loaded has_detail can be used to index it.

	if (   numfilenameRecords > has_detail
		&& filenameRecords[has_detail].handle == IWADhandle)
	{
		char *fname = filenameRecords[has_detail].filename;

		if (subfn_IwadName(fname, "doom.wad"))
			return 1;
		if (subfn_IwadName(fname, "doom1.wad"))
			return 1;
		if (subfn_IwadName(fname, "doom2.wad"))
			return 2;
		if (subfn_IwadName(fname, "plutonia.wad"))
			return 3;
		if (subfn_IwadName(fname, "tnt.wad"))
			return 4;
	}

	return 0;
}

//==========================================================================
//
// W_GetNumForName
//
// Calls W_CheckNumForName. Risen3D terminates if not found.
//
//==========================================================================

int	W_GetNumForName (char *name)
{
	int	i = W_CheckNumForName(name);

	if (i >= 0) return i;

	Con_Error("W_GetNumForName: \"%s\" not found", name);
	return -1;	// Happy compiler
}

//==========================================================================
//
// W_LumpLength
//
// Returns the buffer size needed to load the given lump.
//
//==========================================================================

int W_LumpLength(int lump)
{
	if(lump >= numlumps)
	{
		int a = 1, b = 0, c;
		c = a/b;
		Con_Error("W_LumpLength: %i >= numlumps", lump);
	}
	return lumpinfo[lump].size;
}

//===========================================================================
// W_LumpName
//	Get the name of the given lump.
//===========================================================================
char *W_LumpName(int lump)
{
	if (lump < 0 || lump >= numlumps)
	{
		Con_Error("W_LumpName: %i >= numlumps", lump);
	}
	return lumpinfo[lump].name;
}

//==========================================================================
//
// W_ReadLump
//
// Loads the lump into the given buffer, which must be >= W_LumpLength().
//
//==========================================================================

// Does not return if there's an error
void CheckLumpRange(int lump, int num)
{
	if(lump < 0 || lump >= numlumps)
		Con_Error("CheckLump [%i]: Bad lump number %i [maxnum %i]", num, lump, numlumps - 1);
}

// The pointer must be manually freed by the caller with free()
void *M_CacheLumpNum(int lump, size_t *length, RDBOOL error)
{
	lumpinfo_t	*l;
	size_t		a, c;
	void		*ptr;

	CheckLumpRange(lump, 1);

	l = lumpinfo + lump;
	if (length) *length = l->size;

	if (!(c = l->size))
	{
		if (error) Con_Error("lump \"%.8s\" num %i has no data",
								lumpinfo[lump].name, lump);
		return NULL;
	}

	// Align and allow margin for any error with code using this
	// (e.g. sound rate conversion, ApplyDEH etc.)
	a = (c & 1) ? 3 : 4;

	ptr = M_Calloc(l->size + a, 8);

	lseek(l->handle, l->position, SEEK_SET);
	c = read(l->handle, ptr, l->size);

	if(c < l->size)
	{
		Con_Error("M_CacheLumpNum: only read %i of %i on lump %i", c, l->size, lump);
	}
	
	return ptr;
}

// The pointer must be manually freed by the caller with free()
void *M_CacheLumpName(char *name, size_t *length, RDBOOL error)
{
	int lump = W_CheckNumForName(name);

	if (lump < 0)
	{
		if (error)
			Con_Error("M_CacheLumpName: \"%s\" not found", name);
		return NULL;
	}

	return M_CacheLumpNum(lump, length, error);
}

//==========================================================================
//
// W_GetLumpNamePtr
//
//==========================================================================

char *W_GetLumpNamePtr(int lumpnum)
{
	static char invalid[] = {"INVALID"};

	if (lumpnum < 0 || lumpnum >= numlumps)
		return invalid;

	return lumpinfo[lumpnum].name;
}

//===========================================================================
// W_LumpSourceFile
//	Returns the name of the WAD file where the given resides.
//	Returns a valid filename or an empty string.
//===========================================================================
char *W_LumpSourceFile(int lump)
{
	int i;
	lumpinfo_t *l;

	CheckLumpRange(lump, 2);

	l = lumpinfo + lump;
	for(i = 0; i < numfilenameRecords; i++)
	{
		if(filenameRecords[i].handle == l->handle)
			return filenameRecords[i].filename;
	}

	return "";
}

int W_GetMapIndex(char *mapid)
{
	if (commercial_game)
		return atoi(&mapid[3]);

	return (mapid[1] - '0')*9 + mapid[3] - '0';
}


