//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: gl_pcx.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:  pcx file load
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_graphics.h"

#include "m_winmem.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//===========================================================================
// PCX_HeaderGetInfo
//	Returns true if successful.
//===========================================================================
int PCX_HeaderGetInfo(void *imageData, int *w, int *h)
{
	pcx_t *hdr = (pcx_t*) imageData;

	if (    hdr->xmin >= 0
		&&  hdr->xmax >= 0
		&&  hdr->ymin >= 0
		&&  hdr->ymax >= 0
		&&  hdr->manufacturer   == 0x0a
		&&  hdr->version        == 0x05
		&&  hdr->encoding       == 0x01
		&&  hdr->bits_per_pixel == 0x08 )       // must be 8
	{
		if(w) *w = hdr->xmax - hdr->xmin + 1;
		if(h) *h = hdr->ymax - hdr->ymin + 1;
		return true;
	}

	return false;
}

//===========================================================================
// PCX_GetSize
//===========================================================================
int PCX_GetSize(char *fn, int *w, int *h)
{
	DFILE *file = F_Open(fn, "fb");
	pcx_t hdr;

	if(!file) return false;
	F_Read(&hdr, sizeof(hdr), file);
	F_Close(file);
	return PCX_HeaderGetInfo(&hdr, w, h);
}

//===========================================================================
// PCX_MemoryLoad
//	Returns true if the data is a PCX image (probably).
//===========================================================================
int PCX_MemoryLoad(byte *imgdata, int len, int buf_w, int buf_h, 
				   byte *outBuffer)
{
	pcx_t	*pcx = (pcx_t*) imgdata;
	byte	*raw = &pcx->data, *palette;
	int		x, y;
	int		dataByte, runLength;
	byte	*pix;

	// Check the format.
	if (   pcx->manufacturer != 0x0a
		|| pcx->version != 5
		|| pcx->encoding != 1
		|| pcx->bits_per_pixel != 8)
	{
		Con_Message("PCX_Load: unsupported format.\n");
		return false;
	}

	// Check that the PCX is not larger than the buffer.
	if(pcx->xmax >= buf_w || pcx->ymax >= buf_h)
	{
		Con_Message("PCX_Load: larger than expected.\n");
		return false;
	}

	palette = (byte*)M_Malloc(768, 14);
	memcpy(palette, ((byte*) pcx) + len - 768, 768); // Palette's at the end.

	pix = outBuffer;

	for (y = 0; y <= pcx->ymax; y++, pix += (pcx->xmax + 1)*3)
	{
		for (x = 0; x <= pcx->xmax; )
		{
			dataByte = *raw++;
			
			if((dataByte & 0xC0) == 0xC0)
			{
				runLength = dataByte & 0x3F;
				dataByte = *raw++;
			}
			else
				runLength = 1;
			
			while(runLength-- > 0)
			{
				memcpy(pix + x++ * 3, palette + dataByte*3, 3);
			}
		}
	}

	if(raw - (byte *)pcx > len)
		Con_ErrorD("PCX_Load: corrupt image");

	M_Free(palette);
	return true;
}

//===========================================================================
// PCX_Load
//	PCX loader, partly borrowed from the Q2 utils source (lbmlib.c). 
//===========================================================================
void PCX_Load(char *fn, int buf_w, int buf_h, byte *outBuffer)
{
	DFILE	*file = F_Open(fn, "fb");
	byte	*raw;
	int		len;

	if(!file) 
	{
		Con_Message("PCX_Load: can't find %s.\n", fn);
		return;
	}

	// Load the file.
	F_Seek(file, 0, SEEK_END);		// Seek to end.
	len = F_Tell(file);				// How long?
	F_Seek(file, 0, SEEK_SET);
	raw = (byte*)M_Malloc(len, 15);
	F_Read(raw, len, file);
	F_Close(file);

	// Parse the PCX file.
	if(!PCX_MemoryLoad(raw, len, buf_w, buf_h, outBuffer))
		Con_Message("PCX_Load: error loading %s.\n", fn);

	M_Free(raw);
}
