//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen, Jack Vermeulen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: gl_png.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified for R3D
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:
//  Portable Network Graphics high-level handling.
//  You'll need libpng and zlib.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------
 
#include <setjmp.h>

#include "de_base.h"
#include "de_console.h"
#include "de_graphics.h"
#include "de_misc.h"

#include "r3d_compat.h"
#include "..\..\ext\png.h"

int palpix;

static const char	*current_png_file;
static const char	*type;
static char			err_lumpnum[20];
static int			local_from;

// CODE --------------------------------------------------------------------

void PNGAPI user_error_fn(png_structp png_ptr, png_const_charp error_msg)
{
	Con_Error("PNG-Error [%s %s %s] called from %i: %s", type, err_lumpnum, current_png_file, local_from, error_msg);
}

void PNGAPI user_warning_fn(png_structp png_ptr, png_const_charp warning_msg)
{
	pr_OFinsert("PNG-Warning [%s %s %s]: %s\n", type, err_lumpnum, current_png_file, warning_msg);
}

//===========================================================================
// my_read_data - libpng calls this to read from files.
//===========================================================================
void PNGAPI my_read_data(png_structp read_ptr,
						 png_bytep data, 
						 png_size_t length)
{
	F_Read(data, length, (DFILE*)png_get_io_ptr(read_ptr));
}

//===========================================================================
// subfn_PNG_Load
//	Reads the given PNG image and returns a pointer to a planar RGB or
//	RGBA buffer. Width and height are set, and pixelSize is either 3 (RGB)
//	or 4 (RGBA). The caller must free the allocated buffer with M_Free.
//	width, height and pixelSize can't be NULL. Handles 1-4 channels.
//===========================================================================
static byte *subfn_PNG_Load(DFILE *dfile,
					 int *width, int *height, int *pixelSize)
{
	png_structp	png_ptr;
	png_infop	png_info;
	png_infop	end_info;

	png_bytep	*rows, pixel;
	byte		*buff = NULL;
	int			i, k, off;

	// Init libpng.	
	if (!(png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 
							NULL, user_error_fn, user_warning_fn))
		|| !(png_info = png_create_info_struct(png_ptr))
		|| !(end_info = png_create_info_struct(png_ptr))
		|| setjmp(png_jmpbuf(png_ptr)))
	{
		png_destroy_read_struct(&png_ptr, &png_info, &end_info);
		return NULL;	
	}
	
	// jav support gray scale
	if (   png_info->color_type == PNG_COLOR_TYPE_GRAY
		|| png_info->color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
	{
		png_set_gray_to_rgb(png_ptr);
	}

	// read whole file
	png_set_read_fn(png_ptr, dfile, my_read_data);

	//  Decompress the file
	png_read_png(png_ptr, png_info, PNG_TRANSFORM_IDENTITY, NULL);
	
	// Check if it can be used	
	if(png_info->bit_depth != 8)
	{
		if (png_info->bit_depth == 16)
		{
			png_set_strip_16(png_ptr);  // jav strip down to 8
		}
		else
		{
			pr_OFdirect("subfn_PNG_Load: Bit depth must be 8\n");
			png_destroy_read_struct(&png_ptr, &png_info, &end_info);
			return NULL;	
		}
	}
	
	// Save image info required by the caller	
	*width     = png_info->width;
	*height    = png_info->height;
	*pixelSize = png_info->channels;
	
	// Paletted images have three color components per pixel.	
	if (*pixelSize == 1) *pixelSize = 3;
	if (*pixelSize == 2) *pixelSize = 4; // With alpha channel.
	
	buff = (byte*)M_Malloc(4 * png_info->width * png_info->height, 16); // Use 4 for safety - jav
	rows = png_get_rows(png_ptr, png_info);

	// Get palette entry 0
	if(png_info->channels == 1)
	{
		palpix =  png_info->palette[0].red
				+ png_info->palette[0].green
				+ png_info->palette[0].blue;
	}

	for(i = 0; i < *height; i++)
	{
		if(png_info->channels >= 3)
		{
			memcpy(buff + i * (*pixelSize) * png_info->width,
      			   rows[i], 
                  (*pixelSize) * png_info->width);
		}
		else // Paletted image.
		{
			for(k = 0; k < *width; k++)
			{
				pixel = buff + ((*pixelSize) * (i*png_info->width + k));
				off = k * png_info->channels;
				pixel[0] = png_info->palette[rows[i][off]].red;
				pixel[1] = png_info->palette[rows[i][off]].green;
				pixel[2] = png_info->palette[rows[i][off]].blue;
				if(png_info->channels == 2) // Alpha data.
				{
                   pixel[3] = rows[i][off + 1];
                }
			}
		}
	}
	png_destroy_read_struct(&png_ptr, &png_info, &end_info);

	return buff;
}

//===========================================================================
// PNG_Load
//  If fileName is set then lump is ignored.
//  If a lump is used then it must be pre-validated as being a png - R3D
//===========================================================================
byte *PNG_Load(char *fileName, int lump,
			   int *width, int *height, int *pixelSize, int from)
{
	DFILE *file = fileName ? F_Open(fileName, "fb" ) : F_OpenLump(lump);
	byte  *buffer = NULL;

	if (!file) return NULL;

	local_from = from;

	current_png_file = fileName ? fileName : lumpinfo[lump].name;
	type = fileName ? "file" : "lump";
	err_lumpnum[0] = 0;

	if (!fileName) itoa(lump, err_lumpnum, 10);

	buffer = subfn_PNG_Load(file, width, height, pixelSize);	
	F_Close(file);

	if (!buffer)
		Con_Error("PNG file [%s %s %s] called from %i is bad", type, err_lumpnum, current_png_file, local_from);

	return buffer;
}
