//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen, Jack Vermeulen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: gl_tex.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D, OpenGL routines moved to main engine
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 21 (v220-21) 02/Jun/2013 gmj
// Revision 2.2.0 build 24 (v220-24) 15/Feb/2014 gmj
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:   Texture handling for OpenGL
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <io.h>

#include <stdlib.h>
#include <direct.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_refresh.h"
#include "de_graphics.h"
#include "de_render.h"
#include "de_misc.h"

#include "r3d_compat.h"
#include "r3d_sl3D.h"

#include "ui_main.h"

#include "r3d_OGLfunc.h"
#include "r3d_edit.h"

// MACROS ------------------------------------------------------------------

#define MAX_NUM_SKINNAMES	4096

#define RGB18(r, g, b) ((r)+((g)<<6)+((b)<<12))

// TYPES -------------------------------------------------------------------

// A translated sprite.
#pragma pack(1)
typedef struct
{
	DGLuint	tex;
	int		patch;
	byte	*table;
} transspr_t;

// Sky texture topline colors.
typedef struct
{
	int		texidx;
	int		use;
	float	rgb[3];
} skycol_t;

// Model skin.
typedef struct
{
	DGLuint	tex;
	char	*path;
	ushort	hasalpha;	// v220-9
} skintex_t;

// Detail texture instance.
typedef struct dtexset_s
{
	struct	dtexset_s *next;
	DGLuint	tex;
	int		lump;
} dtexset_t;

// Used to check runs of non masked source pixels
typedef struct
{
	byte	topdelta;		// 0xFF is the last post in a column
	byte	length;
} column_t;

#pragma pack()

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	P_Remove3DSwitches(void);
int		PH_CheckTexLine(int idx, int width, int *lnum, int *lcnt);
int		PH_CheckMidTex(int idx);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

void	GL_Delete3Dflats(void);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static RDBOOL	GL_ColorKey(byte *color);
static RDBOOL	GL_BufferTexture(texture_t *tex, byte *buffer, int width,
									int height);

static void		averageColorIdx(rgbcol_t *sprcol, byte *data, int w, int h);
static void		averageColorRGB(rgbcol_t *col, byte *data, int w, int h);

static byte *GL_LoadHighResTexture(texture_t *tex, int *width, int *height, int *pixelsize);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL	s3tcAvailable;
extern int		palpix;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

RDBOOL			filloutlines;

// Convert a 18-bit RGB (666) value to a playpal index. 
// FIXME: 256kb - Too big?
byte			pal18to8[262144];	

// Properties of the current texture.
RDBOOL			rendtexmask;	
int				rendtexflags;
detailinfo_t	*rendtexdetail;
texscale_t		*ctex;
DGLuint			curtex;

skycol_t		*skytop_colors = NULL;
int				num_skytop_colors = 0;

DGLuint			dltexname;	// Name of the dynamic light texture.
DGLuint			glowtexname;
DGLuint			menuFogTexture;	

int				r_detail;	// Render detail textures (if available).
int				skinCount;	// v220-31 for sripted precache stats
int				reduceTo = 8192;	// v220-33

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static GLint	miplevel = GL_LINEAR_MIPMAP_LINEAR;	// default

static RDBOOL	texInited = false;	// Init done.
static int		maxnum_skytop_colors = 0;

// Skinnames will only *grow*. It will never be destroyed, not even
// at resets. The skin textures themselves will be deleted, though.
// This is because we want to have permanent ID numbers for skins,
// and the ID numbers are the same as indices to the skinnames array.
// Created in r_model.c, when registering the skins.
static int			numskinnames = 0;
static skintex_t	skinnames[MAX_NUM_SKINNAMES];

// The translated sprites.
static transspr_t *transsprites;
static int		numtranssprites;

// Track if a detail tex has already been set
static dtexset_t *dtset;
static ded_detailtexture_t *ded_det;

static texscale_t	ctex_dummy;

// CODE --------------------------------------------------------------------

//===========================================================================
// PalIdxToRGB
//===========================================================================
void PalIdxToRGB(byte *pal, int idx, byte *rgb)
{
	int c;

	for(c = 0; c < 3; c++) // Red, green and blue.
		rgb[c] = pal[idx*3 + c];
}

//===========================================================================
// CeilPow2
//	Finds the power of 2 that is equal to or greater than 
//	the specified number.
//===========================================================================
int CeilPow2(int num)
{
	int cumul;
	for(cumul = 1; num > cumul; cumul <<= 1);
	return cumul;
}

//===========================================================================
// FloorPow2
//	Finds the power of 2 that is less than or equal to
//	the specified number.
//===========================================================================
int FloorPow2(int num)
{
	int fl = CeilPow2(num);

	return (fl > num ? fl >> 1 : fl);
}

//===========================================================================
// WeightPow2
//	Weighted rounding. Weight determines the point where the number 
//	is still rounded down (0..1).
//===========================================================================
static int WeightPow2(int num, float weight)
{
	int		fp2 = FloorPow2(num);
	float	frac = (num - fp2) / (float) fp2;

	return (frac <= weight ? fp2 : fp2 << 1);
}

/*
//===========================================================================
// RoundPow2
//	Finds the power of 2 that is nearest the specified number.
//	In ambiguous cases, the smaller number is returned.
//===========================================================================
int RoundPow2(int num)
{
	int cp2 = CeilPow2(num), fp2 = FloorPow2(num);
	return (cp2-num >= num-fp2)? fp2 : cp2;
}
*/

//===========================================================================
// LookupPal18to8
//	Prepare the pal18to8 table.
//	A time-consuming operation (64 * 64 * 64 * 256!).
//===========================================================================
static void LookupPal18to8(byte *palette)
{
	int		r, g, b, i;
	byte	palRGB[3];
	uint	diff, smallestDiff, closestIndex;
		
	for(r = 0; r < 64; r++)
	{
		for(g = 0; g < 64; g++)
		{
			for(b = 0; b < 64; b++)
			{
				// We must find the color index that most closely
				// resembles this RGB combination.
				smallestDiff = 0xFFFFFFFF;
				for(i = 0; i < 256; i++)
				{
					memcpy(palRGB, palette + 3*i, 3);
					diff = (palRGB[0]-(r<<2))*(palRGB[0]-(r<<2)) 
						+ (palRGB[1]-(g<<2))*(palRGB[1]-(g<<2)) 
						+ (palRGB[2]-(b<<2))*(palRGB[2]-(b<<2));
					if(diff < smallestDiff) 
					{
						smallestDiff = diff;
						closestIndex = i;
					}
				}
				pal18to8[RGB18(r, g, b)] = (byte)closestIndex;
			}
		}
	}
}

//===========================================================================
// GL_InitTextureManager - called once on startup
//===========================================================================
void GL_InitTextureManager(void)
{
	RDBOOL	Pal18to8_loaded = false;
	int		i;

	if (!filloutlines) Con_Message( "filloutlines : disabled by user\n");

	// b211-0 set up the no texture ctex dummy
	ctex_dummy.rendh = ctex_dummy.rendw = 1.0f;
	ctex_dummy.texscalex = ctex_dummy.texscaley = 1.0f;
	ctex_dummy.high = 0.f;

	transsprites = NULL;
	numtranssprites = 0;

	// Detail tex tracking array
	dtset = NULL;

	// Do we need to generate a pal18to8 table?
	if(ArgCheck("-dump_pal18to8"))
	{	
		FILE *file;

		LookupPal18to8(palbuffer);
		Pal18to8_loaded = true;

		if ( (file = fopen("Pal18to8.lmp", "wb")) != NULL)
		{
			fwrite(pal18to8, sizeof(pal18to8), 1, file);								
			fclose(file);
		}
	}

	// Load the pal18to8 table from the lump PAL18TO8.
	// We need it when resizing textures.
	if ( (i = W_CheckNumForName("PAL18TO8")) >= 0 )
	{
		size_t	size;
		byte	*ptr = (byte*)M_CacheLumpNum(i, &size, false);

		if (size == sizeof(pal18to8))
		{
			memcpy(pal18to8, ptr, size);
			Pal18to8_loaded = true;
		}
		M_Free(ptr);
    }

	if (!Pal18to8_loaded) LookupPal18to8(palbuffer);

	// System textures loaded in GL_LoadSystemTextures.
	dltexname = glowtexname = 0;
}

//===========================================================================
// GL_PrepareGlowTexture
//===========================================================================
static DGLuint GL_PrepareGlowTexture(void)
{
	if (!glowtexname)
	{
		// Generate the texture.
		uint size;
		byte *data = (byte*)M_CacheLumpName("WDLIGHT", &size, true);

		if (size != 4*32) Con_ErrorD("Invalid WDLIGHT texture");

		GL_NewTexture(&glowtexname);
		GL_TexImage(GL_LUMINANCE, 4, 32, 0, data);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

		M_Free(data);
	}

	return glowtexname;
}

//===========================================================================
// GL_LoadSystemTextures
//	Prepares all the system textures (dlight, ptcgens).
//===========================================================================
void GL_LoadSystemTextures(void)
{
	if (texInited) return;

	dltexname = GL_PrepareLightTexture();
	glowtexname = GL_PrepareGlowTexture();

	GL_SetFlat(console_flat);	// re init the console flat first

	// Load particle textures.
	PG_InitTextures();

	// Initialization done.
	texInited = true;
}

void GL_DeleteAllSecondaryAlphaTex()
{
	int i;

	for(i = 0; i < numtextures; i++)		// 2) Textures
		GL_DeleteTexIndex(&texArray[i]->noalphatex, "texArray alpha");
}

//===========================================================================
// GL_ShutdownTextureManager
//	Runtime textures are not loaded until precached or actually needed.
//	They may be cleared, in which case they will be reloaded when needed.
//===========================================================================
void GL_ShutdownTextureManager(void)
{
	dtexset_t *dtex;
	int i;

	if (!texInited) return;

	// OGL texture references are stored on several different lists.
	// These must be carefully managed to ensure no errors are made.

	// 1) Rendering lists reference OGL textures, so these must be reset.
	RL_DeleteLists();

	// Delete all OGL map textures;
	GL_Delete3Dflats();						// 1) 3Dflats textures

	for (i = 0; i < numlumps; i++)	// 2) Standard flats textures
	{
		flat_t *fl = flatsarray[i];

		if (fl)
			GL_DeleteTexIndex(&fl->gltex, lumpinfo[fl->lump].name);
	}

	GL_DeleteAllSecondaryAlphaTex();		// 3) Secondary alpha textures

	for(i = 0; i < numtextures; i++)		// 4) Wall textures
		GL_DeleteTexIndex(&texArray[i]->gltex, texArray[i]->name);

	for(i = 0; i < numspritelumps; i++)		// 5) Sprite textures
		GL_DeleteTexIndex(&spritelumps[i].tex, "spritelumps");
	
	GL_ResetHudsprites();					// 6) hires hudsprites

	if (transsprites)						// 7) The translated sprite textures.
	{
		for(i = 0; i < numtranssprites; i++)
			GL_DeleteTexIndex(&transsprites[i].tex, "transsprites");
		W_Free(transsprites, STATICID, 503);
		transsprites = NULL;
		numtranssprites = 0;
	}

	for(i = 0; i < numskinnames; i++)		// 8) Model skin textures
		GL_DeleteTexIndex(&skinnames[i].tex, "skinnames");

	while(dtset)							// 9) Detail textures
	{
		dtex = dtset->next;
		GL_DeleteTexIndex(&dtset->tex, "dtset");
		M_Free(dtset);
		dtset = dtex;
	}

	for(i = 0; i < numlumps; i++)			// 11) Any raw lumptexinfo textures
		GL_DeleteTexIndex(&lumptexinfo[i].ltitex, "lumptexinfo 1");

	Rend_DeleteSkyAddOnTexures(true);		// 12) v220-3 skybox, cap textures

	GL_DeleteTexIndex(&dltexname, "dltexname");		// 13) system textures
	GL_DeleteTexIndex(&glowtexname, "glowtexname");
	GL_DeleteTexIndex(&menuFogTexture, "menuFogTexture");
	PG_ShutdownTextures();	// Delete the particle textures.

	// error check: just in case not 0
	if (skin_inf.graphicstexmemoryused || skin_inf.skintexmemoryused)
	{
		Con_Printf("POST DELETE %i %i\n", skin_inf.graphicstexmemoryused, skin_inf.skintexmemoryused);

		skin_inf.graphicstexmemoryused = skin_inf.skintexmemoryused = 0;
		memset(skin_inf.textureidx, 0, sizeof(uint) * 4096);	// MAXTEXINDEX
		memset(skin_inf.textype, 0, sizeof(byte) * 4096);		// MAXTEXINDEX
	}

	num_skytop_colors = 0;
	texInited = false;
}

//===========================================================================
// GL_ConvertBuffer
//	in/outformat:
//	1 = palette indices
//	2 = palette indices followed by alpha values
//	3 = RGB 
//	4 = RGBA
//===========================================================================
static void GL_ConvertBuffer(int width, int height,
							 int informat, int outformat, 
							 byte *in, byte *out)
{
	int		inSize  = informat  == 2 ? 1 : informat;
	int		outSize = outformat == 2 ? 1 : outformat;
	int		a, i, numPixels = width * height;
	
    if (!in || !out)
    {
        pr_OFinsert("GL_ConvertBuffer no buffers %p and %p\n",in,out);
        return;
    }

	if(informat == outformat) 
	{
		// No conversion necessary.
		memcpy(out, in, numPixels * informat);
		return;
	}
	// Conversion from pal8(a) to RGB(A).
	if(informat <= 2 && outformat >= 3)
	{
		for (i = 0; i < numPixels; i++, in += inSize, out += outSize) 
		{
			memcpy(out, palbuffer + 3*(*in), 3);

			// Will the alpha channel be necessary?
			a = 0;
			if (informat  == 2) a = in[numPixels*inSize];
			if (outformat == 4) out[3] = (byte)a;
		}
	}
	// Conversion from RGB(A) to pal8(a), using pal18to8.
	else if(informat >= 3 && outformat <= 2)
	{
		for(i = 0; i < numPixels; i++, in += inSize, out += outSize)
		{
			// Convert the color value.
			*out = pal18to8[RGB18(in[0]>>2, in[1]>>2, in[2]>>2)];
			// Alpha channel?
			a = 0;
			if(informat == 4) a = in[3];
			if(outformat == 2) out[numPixels*outSize] = (byte)a;
		}
	}
	else if(informat == 3 && outformat == 4)
	{
		for(i = 0; i < numPixels; i++, in += inSize, out += outSize)
		{
			memcpy(out, in, 3);
			out[3] = 0;
		}
	}
}

//===========================================================================
// scaleLine
//	Len is in measured in out units. Comps is the number of components per 
//	pixel, or rather the number of bytes per pixel (3 or 4). The strides 
//	must be byte-aligned anyway, though; not in pixels.
//	FIXME: Probably could be optimized.
//===========================================================================
static void scaleLine(byte *in,     int inStride, 
                      byte *out,    int outStride,
					  int   outLen, int inLen, int colorcomps)
{
	int		i, c;
	float	inToOutScale;
	
	inToOutScale = outLen / (float) inLen;
	
	if(inToOutScale > 1)	
	{   
		// Magnification is done using linear interpolation.
		fixed_t		inPosDelta;
		fixed_t		inPos;
		byte		*col1, *col2;
		int			weight, invWeight;
		
		inPosDelta = (FRACUNIT*(inLen-1))/(outLen-1);
		inPos      = inPosDelta;
		
		// The first pixel.
		
		memcpy(out, in, colorcomps);
		out += outStride;
		
		// Step at each out pixel between the first and last ones.
		
		for(i=1; i < outLen-1; i++, out += outStride, inPos += inPosDelta)
		{
			col1      = in + (inPos >> FRACBITS) * inStride;
			col2      = col1 + inStride;
			weight    = inPos & 0xffff;
			invWeight = 0x10000 - weight;
			
			out[0] = (byte)((col1[0]*invWeight + col2[0]*weight) >> 16);
			out[1] = (byte)((col1[1]*invWeight + col2[1]*weight) >> 16);
			out[2] = (byte)((col1[2]*invWeight + col2[2]*weight) >> 16);
			
			if (colorcomps == 4)
			{
				out[3] = (byte)((col1[3]*invWeight + col2[3]*weight) >> 16);
			}
		}
		// The last pixel.
		memcpy(out, in + (inLen-1)*inStride, colorcomps);
	}
	else if(inToOutScale < 1)
	{
		// Minification needs to calculate the average of each of 
		// the pixels contained by the out pixel.
		
		uint	cumul[4] = {0, 0, 0, 0};
		uint	count  = 0;
		int		outpos = 0;
		
		for(i = 0; i < inLen; i++, in += inStride)
		{
			if((int) (i*inToOutScale) != outpos)
			{
				outpos = (int) (i*inToOutScale);
				
				for(c = 0; c < colorcomps; c++) 
				{
					out[c]   = (byte)(cumul[c] / count);
					cumul[c] = 0;
				}
				count = 0;
				out  += outStride;
			}
			
			for(c = 0; c < colorcomps; c++) cumul[c] += in[c];

			count++;
		}
		// Fill in the last pixel, too.
		if(count) for(c = 0; c < colorcomps; c++) out[c] = (byte)(cumul[c] / count);
	}
	else	// No need for scaling.
	{
		for(i = 0; i < outLen; i++, out += outStride, in += inStride)
		{	
			memcpy(out, in, colorcomps);
		}
	}
}

//===========================================================================
// ScaleBuffer32
//===========================================================================
static void ScaleBuffer32(byte *in, int inWidth, int inHeight,
						  byte *out, int outWidth, int outHeight, int comps)
{
	int		i;
	byte	*temp = (byte*)M_Calloc(outWidth * inHeight * comps, 210);

	// First scale horizontally, to outWidth, into the temporary buffer.
	for(i = 0; i < inHeight; i++)
	{
		scaleLine(in + inWidth*comps*i, comps, temp + outWidth*comps*i, comps, outWidth, 
			inWidth, comps);
	}
	// Then scale vertically, to outHeight, into the out buffer.
	for(i = 0; i < outWidth; i++)
	{
		scaleLine(temp + comps*i, outWidth*comps, out + comps*i, outWidth*comps, outHeight,
			inHeight, comps);
	}
	M_Free(temp);
}			

//===========================================================================
// GL_DownMipmap32
//	Works within the given data, reducing the size of the picture to half 
//	its original. Width and height must be powers of two.
//===========================================================================
void GL_DownMipmap32(byte *in, int width, int height, int comps)
{
	byte	*out = in;
	int		x, y, c, outW = width >> 1, outH = height >> 1;

	if (!outW || !outH)
	{
		Con_ErrorD("GL_DownMipmap32: invalid texture");
	}

	for (y = 0; y < outH; y++, in += width*comps)
	{
		for (x = 0; x < outW; x++, in += comps*2)
		{
			for (c = 0; c < comps; c++, out++) 
				*out = (byte)((in[c] + in[comps + c] + in[comps*width + c] + in[comps*(width+1) + c]) >> 2);
		}
	}
}	

//===========================================================================
// GL_UploadTexture
//	Can be rather time-consuming.
//	Returns the name of the texture. 
//	The texture parameters will NOT be set here.
//	'data' contains indices to the playpal. If 'alphachannel' is true,
//	'data' also contains the alpha values (after the indices).
//===========================================================================
static void GL_UploadTexture(
		byte *data,
		int width,
		int height,
		RDBOOL alpha, 
		RDBOOL genMipmap,
		RDBOOL RGBData,
		int noStretch,
		DGLuint *texIndex,
		int from
		)
{
	byte	*buffer = NULL, *rgbaOriginal;

	// The best quality is always used. All textures are
	// sized *upwards*, so no details are lost. This takes
	// more memory, but naturally looks better.
	int		levelWidth  = CeilPow2(width);
	int		levelHeight = CeilPow2(height);

	// Number of color components in the destination image.
	int		i, comps = alpha ? 4 : 3;

	//-----------------------------------------------

	// Hardware limitations may force us to modify the
	// preferred texture size.
	if (   levelWidth  > maxTexSize
		|| levelHeight > maxTexSize)
	{
		noStretch = 0;
		if(levelWidth  > maxTexSize) levelWidth  = maxTexSize;
		if(levelHeight > maxTexSize) levelHeight = maxTexSize;
	}

	// Get the RGB(A) version of the original texture.
	rgbaOriginal = (byte*)M_Calloc(width * height * 4, 211);

	// Convert a paletted source image to truecolor so it can be scaled.
	if(RGBData)
	{
		memcpy(rgbaOriginal, data, width * height * comps);
	}
	else
	{
		GL_ConvertBuffer(width, height, alpha ? 2 : 1, comps, 
												data, rgbaOriginal);
	}
	
	// Prepare the RGB(A) buffer for the texture: we want a buffer with 
	// power-of-two dimensions. It will be the mipmap level zero.
	// The buffer will be modified by the mipmap generation (if done here).

	if (noStretch == 1)
	{
		buffer = (byte*)M_Calloc(levelWidth * levelHeight * 4, 212);

		// Copy the image into a buffer with power-of-two dimensions.
		for(i = 0; i < height; i++) // Copy line by line
		{
			memcpy(buffer + levelWidth*comps*i, rgbaOriginal + width*comps*i, 
				comps*width);
		}

		M_Free(rgbaOriginal);
	}
	else if (noStretch == 2 && levelWidth >= reduceTo && levelHeight >= reduceTo)
	{
		levelWidth /= 2, levelHeight /= 2;
		buffer = (byte*)M_Calloc(levelWidth * levelHeight * 4, 213);

		ScaleBuffer32(rgbaOriginal, width, height, buffer, levelWidth, 
			levelHeight, comps);

		M_Free(rgbaOriginal);

		skin_inf.downscaledskintexcount++;
	}
	else
	{
		// Stretch to fit into power-of-two.
		if (width != levelWidth || height != levelHeight)
		{
			buffer = (byte*)M_Calloc(levelWidth * levelHeight * 4, 213);

			ScaleBuffer32(rgbaOriginal, width, height, buffer, levelWidth, 
				levelHeight, comps);

			M_Free(rgbaOriginal);
		}
		else
		{
			// We can use the input data as-is.
			buffer = rgbaOriginal;
		}
	}
	
	// Get and bind a new texture name.
	GL_NewTexture(texIndex);

	uint size = levelWidth * levelHeight * comps;
	uint idx = *texIndex;

	skin_inf.textureidx[idx] = size;
	skin_inf.textype[idx] = (from == 11 || from == 12);

	// Generate mipmaps, compress if required, etc.
	if (!GL_TexImage(alpha ? GL_RGBA : GL_RGB,
			levelWidth, levelHeight, genMipmap, buffer))
	{
		Con_Error("GL_TexImage: TexImage failed w %i h %i alpha:%i",
			levelWidth, levelHeight, alpha);
	}

	if (skin_inf.textype[idx])
	{
		skin_inf.skintexmemoryused += size;
		skin_inf.skintexcount++;
	}
	else
	{
		skin_inf.graphicstexmemoryused += size;
		skin_inf.graphicstexcount++;
	}


	M_Free(buffer);
}

//===========================================================================
// GL_GetDetail
//===========================================================================
static dtexset_t *GL_GetDetail(int lump)
{
	dtexset_t *dt;

	for(dt = dtset; dt; dt = dt->next)
	{
		if (dt->lump == lump) return dt;
	}

	// Set the array for a new detail tex
	dt = (dtexset_t*)M_Malloc(sizeof(dtexset_t), 17);
	dt->next = dtset;	// this one
	dtset = dt;			// set to next one
	dt->lump = lump;	// store the lump
	dt->tex = 0;		// Not initialised yet
	return dt;
}

//===========================================================================
// GL_LoadDetailTexture
//	Detail textures are 128x128 grayscale 8-bit raw data.
//===========================================================================
static DGLuint GL_LoadDetailTexture(int num)
{
	dtexset_t	*found = GL_GetDetail(num);
	byte		*imgdata;
	int			w = 0, h = 0, valid = false;
	uint		size;

	if (found->tex) return found->tex;	// already loaded

	if ((imgdata = (byte*)M_CacheLumpNum(num, &size, false)) != NULL)	// has data
	{
		// First try loading it as a PCX image.
		if (PCX_HeaderGetInfo(imgdata, &w, &h))
		{
			if (w == h && (w == 256 || w == 128 || w == 64))
			{
				byte *outbuffer = (byte*)M_Calloc(w * h * 4, 214);

				if (PCX_MemoryLoad(imgdata, size, w, h, outbuffer))
				{
					GL_NewTexture(&found->tex);
					valid = GL_TexImage(GL_RGB, w, h, true, outbuffer);
				}

				M_Free(outbuffer);
			}
		}
		else // It must be a raw image.
		{
			int size = lumpinfo[num].size;

			valid = true;

			// How big is it?
			if		(size == 256 * 256) w = h = 256;
			else if (size == 128 * 128) w = h = 128;
			else if (size ==  64 *  64) w = h =  64;
			else
			{
				valid = false;
			}

			if (valid)
			{
				 GL_NewTexture(&found->tex);
				// Make mipmap levels.
				valid = GL_TexImage(GL_LUMINANCE, w, h, true, imgdata);
			}
		}

		M_Free(imgdata);
	}

	if (!valid)
		Con_Error("GL_LoadDetailTexture [w %i h %i name %s]: Must be 256x256 or 128x128 or 64x64", w, h, lumpinfo[num].name);

	// Set texture parameters.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	
	return found->tex;
}

//===========================================================================
// GL_PrepareDetailTexture
//===========================================================================
static DGLuint GL_PrepareDetailTexture(int index, RDBOOL is_wall_texture)
{
	DGLuint	gltex;
	int		i;

	// Search through the assignments. 
	for(i = defs.count.details.num - 1; i >= 0; i--)
	{
		detailtex_t *det = &dettex[i];

		// Check detail lump's been set (see Def_PostInit())
		if (det->detail_lump < 0) continue;

		// Check the index sent matches
		if (   ( is_wall_texture && index == det->wall_texture)
			|| (!is_wall_texture && index == det->flat_lump   ))
		{
			ded_det = defs.dettex + i;

			gltex = GL_LoadDetailTexture(det->detail_lump);

			if (compat[c_dtex] && gltex)
			{
				if (is_wall_texture)
					Con_Message("DETTEX %8.8s linked to wall %8.8s\n", lumpinfo[det->detail_lump].name, texArray[index]->name);
				else
					Con_Message("DETTEX %8.8s linked to flat %8.8s\n", lumpinfo[det->detail_lump].name, lumpinfo[index].name);
			}

			return gltex;
		}
	}
	return 0; // There is no detail texture for this.
}

int findTEXNAMEfromPATCH(char *name, int *lumpnum, int *patchnum)
{
	int i, k;

	for (i = 0; i < numtextures; i++)
	{
		for (k = 0; k < texArray[i]->patchcount; k++)
		{
			if (!_stricmp(lumpinfo[texArray[i]->patches[k].patch].name, name))
			{
				*patchnum = k;
				*lumpnum = texArray[i]->patches[k].patch;
				return i;
			}
		}
	}
	return -1;
}

static byte *noAlphaTex(void *buffer, int height, int width, int *pxsize)
{
	int		i, k;
	int		tpos = 0, fpos = 0;
	byte	*newbuff, *oldbuff;

	newbuff = (byte*)M_Calloc(height * width * 4, 215);
	oldbuff = (byte*)buffer;

	for (i = 0; i < height; i++)
	{
		for (k = 0; k < width; k++)
		{
			newbuff[tpos++] = oldbuff[fpos++];
			newbuff[tpos++] = oldbuff[fpos++];
			newbuff[tpos++] = oldbuff[fpos++];
			fpos++;	// skip the alpha byte
		}
	}

	M_Free(buffer);
	*pxsize = 3;

	return newbuff;
}

//===========================================================================
// GL_LoadHighResFlat
//	The returned buffer must be freed with M_Free.
//  This is called if a substitution is required of a flat in an iwad
//  with a hi-res texture.  With hi-res external types it is only allowed
//  if a pwad has not already overriden an iwad texture.
//===========================================================================
static byte *GL_LoadHighResFlat(char *name, int *width, int *height, int *pixelsize)
{
	int		lump = W_CheckTXNumForName(name);
	char	*path;
	
	// v220-5 check for an embedded hi_res lump (png and jpg are supported)
	if (lump >= 0)
	{
		if (lumpinfo[lump].lumptype == LMP_PNG)
		{
			byte *ptr = PNG_Load(NULL, lump, width, height, pixelsize, 9);
			return ptr;
		}

		if (lumpinfo[lump].lumptype == LMP_JPG)
		{
			byte *ptr = JPG_Load(NULL, lump, width, height, pixelsize, 9);
			return ptr;
		}
	}

	if ( (path = IW_HiresTexFile(true, name)) )
		return GL_LoadImage(path, width, height, pixelsize, false, false);

	return NULL;
}

//===========================================================================
// GL_SetTexFlat (was GL_BindTexFlat)
//  b211-0 rewritten v220-3 revised
//===========================================================================
static void GL_SetTexFlat(flat_t *fl, int chkdet)
{
	if (!fl->gltex)
	{
		void	*imgptr, *tmpptr;
		size_t	size;
		int		width, height;

		int		lump = fl->lump;
		int		pixelsize = 3;
		int		texnum = -1;

		RDBOOL	RGBData = false;
		RDBOOL	masked  = false;

		//--------------------------------------

		if (   lump == skyflatnum	// The sky 'flat' (actually a texture) is dealt with separately
			|| lump < 0
			|| lump >= numlumps)
		{
			return;
		}

		// Check for a replacement high resolution version.
		if ( (imgptr = GL_LoadHighResFlat(lumpinfo[lump].name, &width, 
											&height, &pixelsize)) != NULL)
		{
			RGBData = true;

			masked = pixelsize == 4;

			// Hi-res flats are scaled to the standard flat size
			fl->width  = 64;
			fl->height = 64;
		}
		else
		{
			int	type = DoomLumpIdentify(lump, &width, &height, DLI_GRAPHIC);

			switch(type)
			{
			case LMP_FLAT:
			case LMP_FLAT320:
				tmpptr  = M_CacheLumpNum(lump, &size, false);
				imgptr = M_Calloc(4 * width * height, 216);
				memcpy(imgptr, tmpptr, size);
				M_Free(tmpptr);

				// Standard flat size
				fl->width  = 64;
				fl->height = 64;
				break;
			case LMP_PATCH:	// Load as texture
				{
					int	w, h, lumpnum, patchnum;
					int	texnum = findTEXNAMEfromPATCH(lumpinfo[lump].name, &lumpnum, &patchnum);

					fl->flflags |= FLT_FLATFROMTEX;

					// Load complete texture.  This is only done for patch number 0 and then
					// only if the width of the p0 tex is the same as the whole texture
					if (   patchnum == 0 && texnum >= 0
						&& DoomLumpIdentify(lumpnum, &w, &h, DLI_GRAPHIC) == LMP_PATCH
						&& w == texArray[texnum]->width)
					{
						texture_t *tex = texArray[texnum];

						fl->width  = (int)((float)tex->width  * texscales[texnum].texscalex);
						fl->height = (int)((float)tex->height * texscales[texnum].texscaley);

						// Check for hires version - R3D
						if ( (imgptr = GL_LoadHighResTexture(tex, &width, &height, &pixelsize)) )
						{
							RGBData = true;

							if (pixelsize == 4)
							{
								fl->flflags |= FLT_3DMASKED;
								imgptr = noAlphaTex(imgptr, height, width, &pixelsize);
								masked = false;
							}
						}
						else
						{
							width = tex->width, height = tex->height;
							imgptr = M_Calloc(4 * width * height, 217);
							masked = GL_BufferTexture(tex, (byte*)imgptr, width, height);
							if (masked) fl->flflags |= FLT_3DMASKED;
						}
					}
					else	// just load the single patch
					{
						texture_t	dummy;

						dummy.width  = (short)width;
						dummy.height = (short)height;

						dummy.patchcount = 1;

						dummy.patches[0].patch = lump;
						dummy.patches[0].originx = 0;
						dummy.patches[0].originy = 0;

						memcpy(dummy.name, lumpinfo[lump].name, 9);

						imgptr = M_Calloc(4 * width * height, 218);
						masked = GL_BufferTexture(&dummy, (byte*)imgptr, width, height);

						fl->width  = width;
						fl->height = height;
					}
				}
				break;
			case LMP_PNG:
			case LMP_JPG:
				if (type == LMP_PNG)
					imgptr = PNG_Load(NULL, lump, &width, &height, &pixelsize, 0);
				else
					imgptr = JPG_Load(NULL, lump, &width, &height, &pixelsize, 0);

				RGBData = true;

				if ( (masked = pixelsize == 4) )
				{
					// We don't want a masked flat
					int		i, k;
					int		tpos = 0, fpos = 0;

					byte	*newflat = (byte*)M_Calloc(height * width * 4, 219);
					byte	*oldflat = (byte*)imgptr;

					for(i = 0; i < height; i++)
					{
						for(k = 0; k < width; k++)
						{
							newflat[tpos++] = oldflat[fpos++];
							newflat[tpos++] = oldflat[fpos++];
							newflat[tpos++] = oldflat[fpos++];
							fpos++;	// skip the alpha byte
						}
					}

					M_Free(imgptr);
					imgptr = newflat;
					pixelsize = 3;
					masked = false;
				}

				// Embedded textures between FF_START and FF_END markers
				// are always scaled to the standard flat size of 64x64
				// as there is no other reference that can be used.
				fl->width  = 64;
				fl->height = 64;

				break;
			default:
				GL_SetNoTexture();	
				return;
			}
		}

		if (masked) fl->flflags |= FLT_MASKED;

		// Is there a detail texture for this?
		if (   chkdet
			&& !RGBData
			&& (ge_inf.customdtex
			||  W_CheckPWADNumForName(lumpinfo[lump].name) < 0)
			&& (fl->detail.tex = GL_PrepareDetailTexture(fl->lump, false)) > 0)
		{
			// The width and height could be used for something meaningful.
			fl->detail.width  = 128;
			fl->detail.height = 128;
			fl->detail.scale  = ded_det->scale;
			fl->detail.strength = ded_det->strength;
			fl->detail.maxdist  = ded_det->maxdist;
		}
		else
			fl->detail.tex = 0;

		// Load the texture.
		GL_UploadTexture((byte*)imgptr, width, height, pixelsize == 4, true, RGBData, 0, &fl->gltex, 1);

		// Average color for glow planes.
		if(RGBData)
			averageColorRGB(&fl->color, (byte*)imgptr, width, height);
		else
			averageColorIdx(&fl->color, (byte*)imgptr, width, height);

		M_Free(imgptr);

		// Set the parameters.
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, miplevel);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
}
// v220-15
RDBOOL CheckFlatSet(int idx)
{
	if (idx == skyflatnum) return false;
	if (!flatsarray[idx]) return false;

	return flatsarray[R_GetFlat(idx)->flat_translation] != NULL;
}

RDBOOL textrans;

//===========================================================================
// GL_PrepareFlat
//	Returns the OpenGL name of the texture.
//	(idx is really a lumpnum)
//===========================================================================
DGLuint GL_PrepareFlat(int idx)
{
	flat_t *flat = R_GetFlat(idx);

	// Get the translated one?
	if (flat->flat_translation != idx)
		flat = R_GetFlat(flat->flat_translation);

	GL_SetTexFlat(flat, has_detail);

	ctex = &ctex_dummy;
	ctex->rendw = (float)flat->width;
	ctex->rendh = (float)flat->height;

	rendtexdetail = (r_detail && flat->detail.tex > 0) ? &flat->detail : NULL;
	rendtexflags = 0;
	// Standard flats are always treated as not being masked.
	// If wanting to change this then use: (flat->flflags & FLT_MASKED) > 0;
	rendtexmask = false;
	textrans  = R3D_ScriptedMap ?
		(flat->flflags & FLT_SPLASH) > 0 : (flat->flflags & FLT_WATER) > 0;

	return flat->gltex;
}

//===========================================================================
// GL_GetFlatColor
//	'fnum' is really a lump number.
//===========================================================================
void GL_GetFlatColor(int fnum, unsigned char *rgb)
{
	flat_t *flat = R_GetFlat(fnum);
	memcpy(rgb, flat->color.rgb, 3);
}

//===========================================================================
// GL_SetFlat
//===========================================================================
void GL_SetFlat(int idx)
{
	flat_t *flat = R_GetFlat(idx);

	GL_SetTexFlat(flat, false);

	GL_Bind(flat->gltex);
}

//===========================================================================
// v220-5 CheckPatch
//===========================================================================
RDBOOL CheckPatch(patch_t *patch, int origx, int origy, int texwidth, uint size)
{
	column_t	*column;
	void		*start = (byte*)patch + patch->columnofs[0];
	int			col = 0;
	int			width = patch->width;

	for(; col < width; col++)
	{
		column = (column_t*)((byte*)patch + patch->columnofs[col]);
		
		if ((uint)column - (uint)start > size)
			return false;

		// Step through the posts in a column
		while (column->topdelta != 0xff)
		{			
			column = (column_t *)((byte *)column + column->length + 4);

			if ((uint)column - (uint)start > size)
				return false;
		}
	}
	return true; // valid patch
}

//===========================================================================
// DrawRealPatch
//	The buffer must have room for the alpha values.
//	Returns true if the buffer really has alpha information.
//	If !checkForAlpha, returns false.
//	Origx and origy set the origin of the patch.
//
//  Modified to allow taller masked textures
//
//===========================================================================
static RDBOOL DrawRealPatch(byte *buffer,
							int texwidth,
							int texheight,
							patch_t *patch,
							int origx,
							int origy,
							unsigned char *transtable,
							RDBOOL checkForAlpha)
{
	int			count;
	int			col = 0;
	int			width = patch->width;
	int			bufsize = texwidth * texheight;
	int			x = origx, y;
	int			top;	// Keep track of pos for clipping.

	byte		*destTop	  = buffer + origx;
	byte		*destAlphaTop = destTop + bufsize;
	byte		*dest1, *dest2;
	byte		*source = NULL;

	for(; col < width; col++, destTop++, destAlphaTop++, x++)
	{
		column_t *column = (column_t*)((byte*)patch + patch->columnofs[col]);
		top = -1;
		
		// Step through the posts in a column
		while(column->topdelta != 0xff)
		{
			source = (byte *)column + 3;
			
			if( x < 0 || x >= texwidth ) break; // Out of bounds. 
			
			if (column->topdelta <= top)
				top += column->topdelta;
			else
				top = column->topdelta;
			
			if ( (count = column->length) > 0)     
			{
				y     = origy + top;
				dest1 = destTop      + y*texwidth;
				dest2 = destAlphaTop + y*texwidth;	
				
				while(count--)
				{
					byte palidx = *source++;	// get byte of data (index into palette)
					
					if (transtable) palidx = transtable[palidx];	// Check for translation.
					
					if (y >= 0 && y < texheight)// Check bounds.
					{
						*dest1 = palidx;
						*dest2 = 0xFF;
					}
					// One row down.
					dest1 += texwidth; 				
					dest2 += texwidth; 
					y++;
				}
			}
			column = (column_t *)((byte *)column + column->length + 4);
		}
	}

	if (checkForAlpha)
	{
		// 'Small' textures are allowed to have one alpha pixel
		int	i, singlepixAllow = texwidth < 128 || texheight < 128;
		int ret = false;

		// Scan through the RGBA buffer and check for sub-0xff alpha.
		source = buffer + bufsize;

		for (i = 0, count = 0; i < bufsize; i++)
		{
			// v220-13 do an alpha cut
			if(source[i] <= 0x80) 
			{
				source[i] = 0;

				if (!ret && singlepixAllow) ret = true;

				// To qualify there must be more than 1 apha pixel
				// Some textures have only one which is an error
				// E.g. ZZFACE3, SKINEDGE
				if (!ret && count++ > 1) ret = true; // Has alpha data.
			}
			else
				source[i] = 0xFF;
		}

		return ret;
	}

	return false; // Doesn't have alpha data.
}

//===========================================================================
// TranslatePatch
//	Translate colors in the specified patch.
//===========================================================================
static void TranslatePatch(patch_t *patch, byte *transTable)
{
	int			count;
	int			col = 0;
	column_t	*column;
	byte		*source;
	int			w = patch->width;

	for(; col < w; col++)
	{
		column = (column_t *)((byte *)patch + patch->columnofs[col]);
		// Step through the posts in a column
		while(column->topdelta != 0xff)
		{
			source = (byte*) column + 3;
			count = column->length;
			while(count--)
			{
				*source = transTable[*source];
				source++;
			}
			column = (column_t *)((byte *)column + column->length+4);
		}
	}
}

/* v220-16 not used
//===========================================================================
// GL_IsColorKeyed
//	Returns true if the given path name refers to an image, which should
//	be color keyed. Color keying is done for both (0,255,255) and 
//	(255,0,255).
//===========================================================================
static RDBOOL GL_IsColorKeyed(const char *path)
{
	char buf[512];

	strcpy(buf, path);
	strlwr(buf);
	return strstr(buf, "-ck.") != NULL;
}

//===========================================================================
// GL_ColorKey
//	Returns true if the given color is either (0,255,255) or (255,0,255).
//===========================================================================
static RDBOOL GL_ColorKey(byte *color)
{
	return   color[CB] == 0xff 
		&& ((color[CR] == 0xff && color[CG] == 0   )
		||  (color[CR] == 0    && color[CG] == 0xff));
}

//===========================================================================
// GL_DoColorKeying
//	Buffer must be RGBA. Doesn't touch the non-keyed pixels.
//===========================================================================
static void GL_DoColorKeying(byte *rgbaBuf, int width)
{
	int i;

	for(i = 0; i < width; i++, rgbaBuf += 4)
	{
		if(GL_ColorKey(rgbaBuf))
			rgbaBuf[3] = rgbaBuf[2] = rgbaBuf[1] = rgbaBuf[0] = 0;
	}
}
*/

//===========================================================================
// GL_LoadImage
//	Loads JPG, PNG, TGA and PCX images. The returned buffer must be freed 
//	with M_Free. Color keying is done if "-ck." is found in the filename.
//===========================================================================
byte *GL_LoadImage(char *imagefn, int *width, int *height,
				   int *pixelsize, RDBOOL usemodelpath, RDBOOL ignorealpha)
{
	char		filename[MAXPATHNAMESIZE];
	char		ext[MAXEXT];
	byte		*buffer = NULL;

	_chdir(ddRuntimeDir.path);

	if (usemodelpath)	// Only true for shiny model skins
	{
		if (!R_FindGraphicFile(imagefn, filename))
			return NULL; // Not found.
	}
	else
		strcpy(filename, imagefn);

	_splitpath(filename, NULL, NULL, NULL, (char*)ext);

	if(!_stricmp(ext, ".jpg"))
	{
		buffer = JPG_Load(filename, 0, width, height, pixelsize, 1);
	}
	else if(!_stricmp(ext, ".png"))
	{
		buffer = PNG_Load(filename, 0, width, height, pixelsize, 1);
	}
	else if(!_stricmp(ext, ".tga"))	// v220-3
	{
		buffer = TGA_LoadFile888(filename, width, height, pixelsize);
	}
	else if(!_stricmp(ext, ".pcx"))
	{
		if (!PCX_GetSize(filename, width, height)) return NULL;

		buffer = (byte*)M_Calloc(4 * (*width) * (*height), 220);
		PCX_Load(filename, *width, *height, buffer);
		*pixelsize = 3; // PCXs can't be masked.
	}

	if (!buffer) return NULL;

	if (verbose)
	{
		// Can't use Con_Message as calls startupscreen thus causing
		// endless loop if loading png's for TITLEPIC etc.
		pr_OFinsert("LoadImage: %s (%ix%i)\n", filename, *width, *height); 
	}

	// v220-16 colour keying is not used
	// How about some color-keying?
/*	if (GL_IsColorKeyed(filename))
	{
		byte		*ckdest, *in, *out;
		int			i, numpx; 

		numpx = (*width) * (*height);

		// We must allocate a new buffer if the loaded image
		// has three color components.
		if(*pixelsize < 4)
		{
			ckdest = (byte*)M_Calloc(4 * numpx, 222);

			for(in = buffer, out = ckdest, i = 0; i < numpx; 
				i++, in += *pixelsize, out += 4)
			{
				if(GL_ColorKey(in))
					memset(out, 0, 4);	// Totally black.
				else
				{
					memcpy(out, in, 3); // The color itself.
					out[CA] = 255;		// Opaque.
				}
			}
			M_Free(buffer);
			buffer = ckdest;
		}
		else // We can do the keying in-buffer.
		{
			// This preserves the alpha values of non-keyed pixels.
			for(i = 0; i < *height; i++)
				GL_DoColorKeying(buffer + 4*i*(*width), *width);
		}
		// Color keying is done; now we have 4 bytes per pixel.
		*pixelsize = 4;
	}*/

	return buffer;
}

//===========================================================================
// GL_LoadImageCK
//	First sees if there is a color-keyed version of the given image. If
//	there is it is loaded. Otherwise the 'regular' version is loaded.
//===========================================================================
static byte *GL_LoadImageCK(char *name, int *width, int *height, int *pixelsize, 
							RDBOOL usemodelpath, RDBOOL ignorealpha)
{
	char fn[256];
	char *ptr;

	strcpy(fn, name);
	ptr = strrchr(fn, '.');

	// Check for type with name body ending in -ck.
	// If found then this takes precedence
	if(ptr)
	{
		char buf[32];

		strcpy(buf, "-ck");
		strcat(buf, ptr);

		ptr[0] = 0;

		strcat(fn, buf);

		if( (_access( fn, 4 )) != -1 )
		{
			//	pr_OFinsert("CK image %s found\n", fn);
			return GL_LoadImage(fn, width, height, pixelsize, usemodelpath, false);
		}
	}
	return GL_LoadImage(name, width, height, pixelsize, usemodelpath, false);
}

//===========================================================================
// GL_LoadHighResExternal v220-5 renamed from GL_LoadHighRes
//	Name must end in \0.
//  Prefix is either or "-Flat" or ""
//  File has embedded -ck if colour keying's required (R3D not supported)
//  Revised b220-1
//===========================================================================

static byte *GL_LoadHighResExternal(char *name, int *width, int *height, int *pixelsize)
{
	char	*path;

	if ( (path = IW_HiresTexFile(false, name)) )
		return GL_LoadImage(path, width, height, pixelsize, false, false);

	return NULL;
}

//===========================================================================
// GL_LoadHighResTexture
//	The returned buffer must be freed with M_Free.
//===========================================================================
static byte *GL_LoadHighResTexture
	(texture_t *tex, int *width, int *height, int *pixelsize)
{
	char *name;

	if (tex->txf_flags & TXF_NOHIRES) return NULL;

	name = tex->oldname[0] ? tex->oldname : tex->name;

	// v220-16 check the original iwad dimensions (if they exist)
	// If there's a dimension mismatch then the defaults have been frigged
	if (!IW_CheckIWADTexDimensions(name, tex->width, tex->height))
	{
		return NULL;
	}

	return GL_LoadHighResExternal(name, width, height, pixelsize);
}

static int temp_oy;	// v220-31 used for texture repositioning

//===========================================================================
// GL_BufferTexture
//	Renders the given texture into the buffer.
//===========================================================================
static RDBOOL GL_BufferTexture(texture_t *tex, byte *buffer, int width, int height)
{
	patch_t		*patch;
	RDBOOL		alpha = false;
	int			i, ca, ox, oy;

	temp_oy = 0;

	// Draw all the patches. The final patch determines
	// alphaChannel even if previous patches have none.
	for(i = 0; i < tex->patchcount; i++)
	{
		patch = (patch_t*)M_CacheLumpNum(tex->patches[i].patch, NULL, false);

		if (!patch) Con_ErrorD("Invalid patch: Check the correct IWAD is being used");

		ox = tex->patches[i].originx;
		oy = tex->patches[i].originy;
		if (tex->patchcount == 1) temp_oy = oy;
		ca = (i == tex->patchcount-1);	// check alpha

		// Draw the patch in the buffer.
		alpha = DrawRealPatch(buffer, width, height, patch, ox, oy, 0, ca);

		M_Free(patch);
	}

	return alpha;
}

static void scaleTex(texture_t *tex, int idx, int lump, int width, int height)
{
	int		origwidth, origheight;
	int		origlump = W_CheckTextureNONTXNumForName(lumpinfo[lump].name);

	if (DoomLumpIdentify(origlump, &origwidth, &origheight, DLI_GRAPHIC) == LMP_PATCH)
	{
		patch_t		hdrcache;
		lumpinfo_t	*l = lumpinfo + origlump;

		lseek(l->handle, l->position, SEEK_SET);
		read(l->handle, &hdrcache, 8);

		texscales[idx].texscalex = (float)origwidth /(float)width;
		texscales[idx].texscaley = (float)origheight/(float)height;

		tex->patches[0].originx = hdrcache.xoffset;
		tex->patches[0].originy = hdrcache.yoffset;
	}
}

// Only png and jpg types are supported from a lump.
static byte *checkForHIRESTexture(texture_t *tex, int idx,
						 int *width, int *height, int *pixelsize,
						 RDBOOL needsscale)
{
	int		lump = tex->patches[0].patch;
	int		type = DoomLumpIdentify(lump, width, height, DLI_GRAPHIC);
	byte	*buffer = NULL;

	if (type == LMP_PNG)
	{
		buffer = PNG_Load(NULL, lump, width, height, pixelsize, 2);

		if (needsscale) scaleTex(tex, idx, lump, *width, *height);
	}
	else
	if (type == LMP_JPG)
	{
		buffer = JPG_Load(NULL, lump, width, height, pixelsize, 2);

		if (needsscale) scaleTex(tex, idx, lump, *width, *height);
	}

	return buffer;
}

static void noAlphaError(int idx, int num, int flags)
{
	Con_Error("Program error [%i]: GL_PrepareTextureNoAlpha tex %s flags %.8X",
		num, idx > 0 ? texArray[idx]->name : "TEX UNDEF");
}

//===========================================================================
//  b211-0 GL_PrepareTextureNoAlpha
//   GL_PrepareTexture must have been called first
//===========================================================================
uint GL_PrepareTextureNoAlpha(int idx)
{
	texture_t	*tex;

	idx  = texturetranslation[idx];
	if (!idx || idx >= numtextures) noAlphaError(-1, 3, 0);

	tex  = texArray[idx];

	if (!tex->gltex || !(tex->txf_flags & TXF_MASKED))
	{
		noAlphaError(idx, 0, tex->txf_flags);
	}
	
	if (tex->txf_flags & TXF_FLAT)
		return tex->gltex;

	ctex = &texscales[idx];

	if (!tex->noalphatex)
	{
		RDBOOL		RGBData = false;
		int			pixsize, width, height;
		byte		*buffer;

		buffer = checkForHIRESTexture(tex, idx, &width, &height, &pixsize, false);

		// Check if an embedded png texture exists.
		if (buffer != NULL)
		{
			RGBData = true;
			if (pixsize != 4) noAlphaError(idx, 1, tex->txf_flags);
			buffer = noAlphaTex(buffer, height, width, &pixsize);
		}
		// Try to load a high resolution version of this texture.
		else if ( (buffer = GL_LoadHighResTexture(tex, &width, &height, &pixsize)) != NULL)
		{
			RGBData = true;
			if (pixsize != 4) noAlphaError(idx, 2, tex->txf_flags);
			buffer = noAlphaTex(buffer, height, width, &pixsize);
		}
		else
		{
			width  = tex->width;
			height = tex->height;
			buffer = (byte*)M_Calloc(4 * width * height, 223);
			
			GL_BufferTexture(tex, buffer, width, height);
		}
		
		GL_UploadTexture(buffer, width, height, false, true, RGBData, 0, &tex->noalphatex, 2);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, miplevel);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		M_Free(buffer);
	}

	rendtexflags  = tex->txf_flags & ~TXF_MASKED;
	rendtexdetail = NULL;
	rendtexmask = false;

	return tex->noalphatex;
}

//===========================================================================
// ColorOutlines
//  This is only used with doom alpha wall textures and sprites (low res)
//
//	Remove "outlines" by setting pixels that have zero alpha to an adjacent
//  colour that has non-zero alpha. This may seem somewhat bizarre given the
//  zero-alpha pixel is never seen but helps prevent black outlines being
//  generated when the pixels are mip-mapped.
//
//===========================================================================
static void ColorOutlines(byte *buffer, int width, int height, RDBOOL alpha)
{
	if (alpha && filloutlines)
	{
		int		numpels = width*height;
		byte	*ptr;
		int		i, k, a, b, p;


		// Scan all pixels
		for (k = 0; k < height; k++)
		{
			for (i = 0; i < width; i++)
			{
				p = i + k * width;	// offset to pixel

				if (!buffer[numpels + p]) continue;// skip zero values
				
				// Scan pixels for fully transparent neighbours
				for (b = -1; b <= 1; b++)
				{
					if (b && (k + b < 0 || k + b >= height)) continue;

					for (a = -1; a <= 1; a++)
					{
						if (!a && !b) continue;	// would point to self
						if (a && (i + a < 0 || i + a >= width)) continue;

						ptr = buffer + p + a + b * width;

						// If fully transparent to the solid pixel's colour
						if (!ptr[numpels]) *ptr = buffer[p];
					}
				}
			}
		}
	}
}

// v220-17
static void ColorOutlinesRGBA(byte *buffer, int width, int height, RDBOOL alpha)
{
	if (alpha && filloutlines)
	{
		int		numpels = width*height;
		byte	*ptr;
		int		i, k, a, b, p;

		for (k = 0; k < height; k++)
		{
			for (i = 0; i < width; i++)
			{
				p = (i + k * width) * 4;	// offset to pixel

				// skip zero alpha values
				if (!buffer[p + 3]) continue;
				
				// Scan pixels for fully transparent neighbours
				for (b = -1; b <= 1; b++)
				{
					if (b && (k + b < 0 || k + b >= height)) continue;

					for (a = -1; a <= 1; a++)
					{
						if (!a && !b) continue;	// would point to self
						if (a && (i + a < 0 || i + a >= width)) continue;

						ptr = buffer + p + (a + b * width) * 4;

						if (!ptr[3])	// A fully transparent pixel?
						{
							// Set to the solid pixel's colour
							ptr[0] = buffer[p];
							ptr[1] = buffer[p + 1];
							ptr[2] = buffer[p + 2];
						}
					}
				}
			}
		}
	}
}

// v220-31 routine used to unclamp 3D ladderline textures
void GL_ReloadTextureUnclamped(int idx)
{
	texture_t	*tex;

	idx  = texturetranslation[idx];

	if (!idx || idx >= numtextures) return;

	tex  = texArray[idx];
	ctex = &texscales[idx];

	if (tex->gltex) GL_DeleteTexIndex(&tex->gltex, "reload");

	tex->txf_flags |= TXF_NOCLAMP;
	
	GL_PrepareTexture(idx);
}

//===========================================================================
// GL_PrepareTexture
//	Returns the OGL texture index.
//===========================================================================
uint GL_PrepareTexture(int idx)
{
	texture_t	*tex;

	if (!idx || idx >= numtextures)
	{	
		// No texture?
		ctex = &ctex_dummy;
		ctex->rendw = ctex->rendh = 1.0f;
		rendtexmask = false;
		rendtexflags = 0;
		rendtexdetail = NULL;
		return 0;
	}

	idx  = texturetranslation[idx];
	if (!idx || idx >= numtextures) noAlphaError(-1, 3, 0);
	tex  = texArray[idx];
	ctex = &texscales[idx];

	if (!tex->gltex)
	{
		// Check for flat used as a wall texture
		// This is identified in Add_NewTexture()
		if (tex->txf_flags & TXF_FLAT)
		{
			flat_t dummy;

			memset(&dummy, 0, sizeof(flat_t));
			dummy.lump = tex->patches[0].patch;

			GL_SetTexFlat(&dummy, false);	// Load texture from flat

			if (dummy.flflags & TXF_MASKED)
				tex->txf_flags |= TXF_MASKED;

			tex->gltex = dummy.gltex;
			ctex->rendw = (float)dummy.width;
			ctex->rendh = (float)dummy.height;
			rendtexflags = 0;
			rendtexmask = false;
			rendtexdetail = NULL;

			return tex->gltex;
		}

		RDBOOL		alpha = false;
		RDBOOL		RGBData = false;
		int			pixsize = 0;
		int			width, height;
		int			clampT = 0;
		byte		*buffer;

		int scalewidth  = (int)(SHORT(tex->width ));
		int scaleheight = (int)(SHORT(tex->height));

		// If reloading these are the only flags that need clearing
		tex->txf_flags &= ~(TXF_NOALPHACUT|TXF_CLAMPS);
		tex->detail.tex = 0;
		ctex->high = ctex->high2 = ctex->offset = 0.f;

		// Check if an embedded png/jpg texture exists.
		if (   (buffer = checkForHIRESTexture(tex, idx, &width, &height, &pixsize, false)) != NULL
				// or, if not , that an external high resolution version of this texture exists.
			|| (buffer = GL_LoadHighResTexture(tex, &width, &height, &pixsize)) != NULL)
		{
			RGBData = true;
			tex->txf_flags |= TXF_HIRES;
			alpha = (pixsize == 4);
		
			// v220-13 see if the alphacut flag should be set
			if (alpha)
			{
				byte	*ptr;
				int		sz = height * width;
				int		i, c;

				for (c = i = 0, ptr = buffer; i < sz; i++)
				{
					if (!ptr[3])
					{
						ptr[0] = ptr[1] = ptr[2] = 0;	// ensure set to black
						if (i < width) clampT++;		// only top row checked
					}
					else
					{
						if (ptr[3] != 0xFF) c++;	// variable alpha
					}

					ptr += 4;
				}

				clampT = clampT > 8;

				if (c > 1) tex->txf_flags |= TXF_NOALPHACUT;
			}
		}
		else if (!buffer)	// standard image
		{
			int		hialpharows = 0;
			int		loalpharows = 0;
			int		cmt = -1;

			width  = scalewidth;
			height = scaleheight;

			buffer = (byte*)M_Calloc(4 * width * height, 224);
			alpha  = GL_BufferTexture(tex, buffer, width, height);

			if (alpha)
			{
				byte	*ptr;
				int		sz = height * width;
				int		mismatch = 0;
				int		offschanged = 0;
				int		i, k;

				ptr = buffer + sz;

				for (i = 0; i < height; i++)
				{
					for (k = 0; k < width; k++)
					{
						if (*(ptr++)) goto end_hirowcheck;
					}
					hialpharows++;
				}
end_hirowcheck:;

				for (i = height - 1; i >= 0; i--)
				{
					ptr = buffer + sz + i * width;

					for (k = 0; k < width; k++)
					{
						if (*(ptr++)) goto end_lorowcheck;
					}
					loalpharows++;
				}	
end_lorowcheck:;
			   ctex->loalpha = loalpharows<<FRACBITS;
			   ctex->hialpha = hialpharows<<FRACBITS;

				// v220-33 check for GL_TEXTURE_WRAP_S clamping with lores doom textures
				if (!R3D_edit && !(tex->txf_flags & TXF_ANIMPIC) )
				{
					for (i = 0, ptr = buffer + sz; i < height; i++)
					{
						if ( (*ptr && !*(ptr + width -1)) || (!*ptr && *(ptr + width -1)) ) mismatch++;
						
						ptr += width;
					}

					// check whether texture can be clamped - this is only needed if there is a mismatch
					if (mismatch > 1)
					{
						int *lnum = (int*)malloc(numlines * sizeof(int));
						int lcnt = 0;

						if (PH_CheckTexLine(idx, width, lnum, &lcnt))
						{
							line_t	*ld;
							int		off;

							tex->txf_flags |= TXF_CLAMPS;

							// if necessary adjust the texture x offset
							for (i = 0; i < lcnt; i++)
							{
								ld = &lines[lnum[i]];

								for (k = 0; k < 2; k++)
								{
									if (   sides[ld->sidenum[k]].midtexture == idx
										&& (off = (int)sides[ld->sidenum[k]].textureoffset>>16) )
									{
										if (off < 0)
										{
											while (off < 0) off += width;

											sides[ld->sidenum[k]].orig_textureoffset = (short)off; offschanged++;
											sides[ld->sidenum[k]].lsiflags |= LS_ORIGTEXOFF;
										}
										else if (off >= width)
										{
											while (off >= width) off -= width;
											sides[ld->sidenum[k]].orig_textureoffset = (short)off; offschanged++;
											sides[ld->sidenum[k]].lsiflags |= LS_ORIGTEXOFF;
										}
									}
								}
							}
						}

						free(lnum);
					}
				}

				// v220-31   Where invoked, ColorOutlines reduces hard edges on textures, but
				//           can cause other problems with pixel blending. This is dealt with
				//           by shifting the image up or down by one row.
				if (hialpharows > 1 && !loalpharows)
				{
					// Shift all rows up one and copy the top row into the bottom row
					byte* tempbuffer = (byte*)M_Calloc(4 * width * height, 224);

					ptr = tempbuffer + sz - width;	// point to the bottom row

					// copy the texture's top row into tempbuffer's bottom row
					memcpy(ptr, buffer, width);				// data
					// since the alpha pixels are all set to 0 then there is no need to copy these

					// copy the rest shifted one row up
					ptr = buffer + width;	// point to 2nd row down

					memcpy(tempbuffer, ptr, width * (height - 1));			// data
					memcpy(tempbuffer + sz, ptr + sz, width * (height - 1));// alpha

					M_Free(buffer);
					buffer = tempbuffer;

					ctex->offset = -1.f;	// compensate for pixel shift

					int viewable = height - hialpharows, newheight = CeilPow2(viewable + 1);

					// v220-33 Check for textures with an unseemly high number of transparent rows.
					// These need to be reduced to prevent problems with glow upsetting zbias calcs
					// by the graphics card. Notably Herian2 map27 where lines of this type are
					// close together and where viewing through both e.g. L587. L716 etc.
					// Only resize if used exclusively as a mid tex

					if (temp_oy >= 0 && height > newheight && (cmt = PH_CheckMidTex(idx)) == 1)
					{
						// trim off the empty rows as these cause problems
						int		newsize = width * newheight;
						int		diff = height - newheight;
						
						tempbuffer = (byte*)M_Calloc(4 * newsize, 224);

						ptr = buffer + diff*width;			// data
						memcpy(tempbuffer, ptr, newsize);

						ptr = buffer + sz + diff*width;		// alpha
						memcpy(tempbuffer + newsize, ptr, newsize);

						if (!R3D_edit && ArgExists("-alphaInf"))
							pr_OFinsert("  TEXTURE %s will be reformatted: current height %i, current hialpharows %i, current temp_oy %i\n",
							texArray[idx]->name, height, hialpharows, temp_oy);

						if (ge_inf.codlev && !strcmp(texArray[idx]->name, "ADEL_Z61") )
						{
							// Do nothing. The patch y should be ignored?? Seems this is an issue confined to a texture of this type
							// (i.e with a patched y of half the texture height) plus different results occur in different doom ports.
						}
						else
						{
							ctex->high2 = (float)-diff;		// flag clamped sensitive
						}
						hialpharows -= diff;				// for stats
						height = scaleheight = newheight;	// Change height
						M_Free(buffer);
						buffer = tempbuffer;
					}
					else
					{
						if (temp_oy > 0 && (hialpharows % 4)) ctex->high = (float)temp_oy; // attempt to hack what doom does
					}
				}
				else if (!hialpharows && loalpharows > 1)
				{
					// Shift all rows down one and copy the bottom row into the top row
					byte* tempbuffer = (byte*)M_Calloc(4 * width * height, 224);

					ptr = buffer + sz - width;	// point to the bottom row

					// copy the texture's bottom row into tempbuffer's top row
					memcpy(tempbuffer, ptr, width);				// data
					// since the alpha pixels are all set to 0 then there is no need to copy these

					// copy the rest shifted one row down
					ptr = tempbuffer + width;	// point to 2nd row down

					memcpy(ptr, buffer, width * (height - 1));			// data
					memcpy(ptr + sz, buffer + sz, width * (height - 1));// alpha

					M_Free(buffer);
					buffer = tempbuffer;

					ctex->offset = 1.f;	// compensate for pixel shift
				}
				else
				{
					// v220-33 making temp_oy negative will indicate that line pegging flags should be ignored
					//         Note that true negative values are vanishingly rare and are not usually applied
					if (temp_oy <= loalpharows) ctex->high = (float)(-temp_oy);
				}

				if (hialpharows || loalpharows) ColorOutlines(buffer, width, height, alpha);

				clampT = !(tex->txf_flags & TXF_NOCLAMP) && (cmt == 1 || (cmt == -1 && PH_CheckMidTex(idx)));

				if (!R3D_edit && ArgExists("-alphaInf"))	// debug aid
				{
					int msmonly = ArgCheckWith("-alphaInf", 1) && ArgNext()[0] == '1';

					if ((msmonly && mismatch) || !msmonly)
						pr_OFinsert("%8s pY %3i cth %3.0f cth2 %3.0f [w/h %4i x %4i] [alpharows hi %3i lo %3i] [mismatch %3i cto %3.0f xoffs changed %3i clampS %i clampT %i]\n",
							texArray[idx]->name, temp_oy, ctex->high, ctex->high2, width, height, hialpharows, loalpharows, mismatch, ctex->offset, offschanged, (tex->txf_flags & TXF_CLAMPS) > 0, clampT);
				}
			}
			else
			{
				// Only used with lores non-alpha textures
				if (has_detail
					&& (ge_inf.customdtex
					||  W_CheckPWADNumForName(lumpinfo[tex->patches[0].patch].name) < 0)
					&& (tex->detail.tex = GL_PrepareDetailTexture(idx, true)) > 0)
				{
					// The width and height are fixed
					tex->detail.width = tex->detail.height = 128;
					tex->detail.scale	 = ded_det->scale;
					tex->detail.strength = ded_det->strength;
					tex->detail.maxdist  = ded_det->maxdist;
				}
			}
		}
		
		GL_UploadTexture(buffer, width, height, alpha, true, RGBData, 0, &tex->gltex, 3);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (alpha && !RGBData) ? GL_LINEAR : miplevel);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		if (alpha)	// v220-16 revised
		{
			tex->txf_flags |= TXF_MASKED;
			
			if (clampT && !(tex->txf_flags & TXF_NOCLAMP) )	// v220-27 TXF_NOCLAMP test added
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			else
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

			if (tex->txf_flags & TXF_CLAMPS)
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			else
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		}

		ctex->rendw = (float)scalewidth  * ctex->texscalex;
		ctex->rendh = (float)scaleheight * ctex->texscaley;

		M_Free(buffer);
	}

	rendtexflags = tex->txf_flags;
	rendtexmask = (rendtexflags & TXF_MASKED) > 0;
	rendtexdetail = (r_detail && tex->detail.tex > 0) ? &tex->detail : NULL;

	return tex->gltex;
}

//===========================================================================
// GL_PrepareSkyBox v220-3
//===========================================================================
void GL_PrepareSkyBox(byte *buffer, int width, int height, RDBOOL alpha, DGLuint *address)
{
	GL_UploadTexture(buffer, width, height, alpha, true, true, 0, address, 4);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, miplevel);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

//===========================================================================
// GL_BufferSkyTexture
//  alpha with sky textures ignored
//	Draws the given sky texture in a buffer. The returned buffer must be 
//	freed by the caller. Idx must be a valid texture number.
//===========================================================================
static byte *GL_BufferSkyTexture(int idx, int *width, int *height, int *pixsize)
{
	texture_t	*tex = texArray[idx];	
	byte		*imgdata;
	int			i, ox, oy;

	*width  = tex->width;
	*height = tex->height;
	*pixsize = 0;

	if (tex->patchcount > 1)
	{
		imgdata = (byte*)M_Calloc(4 * tex->width * tex->height, 225);

		for(i = 0; i < tex->patchcount; i++)
		{
			int w = tex->width, h = tex->height;
			patch_t	*patch = (patch_t*)M_CacheLumpNum(tex->patches[i].patch, NULL, true);

			ox = tex->patches[i].originx;
			oy = tex->patches[i].originy;

			DrawRealPatch(imgdata, (*width), (*height), patch, ox, oy, 0, false);

			M_Free(patch);
		}
	}
	else 
	{
		if ( (imgdata = checkForHIRESTexture(tex, idx, width, height, pixsize, true)) != NULL)
		{
			tex->txf_flags |= TXF_HIRES;
		}
		else
		{
			patch_t	*patch;
			int		w, h, lump = tex->patches[0].patch;
			
			// Get the true width and height
			DoomLumpIdentify(lump, &w, &h, DLI_GRAPHIC);

			if (!w || !h)
			{
				Con_ErrorD("Missing sky texture\n"
					"Try restarting with -tnt on the command line");
			}

			patch = (patch_t*)M_CacheLumpNum(lump, NULL, true);

			// Allocate a large enough buffer. 
			imgdata = (byte*)M_Calloc(w * h * 4, 226);

			DrawRealPatch(imgdata, w, h, patch, 0, 0, 0, false);

			*width = w, *height = h;

			M_Free(patch);
		}
	}

	return imgdata;
}

//===========================================================================
// GL_PrepareSky
//	Sky textures are usually 256 pixels wide.
//  v220-29 remove skywidth
//===========================================================================
uint GL_PrepareSky(int idx)
{
	texture_t	*tex;

	if (idx >= numtextures) return 0;

	idx = texturetranslation[idx];
	tex = texArray[idx];
	ctex = &texscales[idx];

	if (!tex->gltex)
	{
		int		pixelsize, width, height;
		RDBOOL	RGBData, alpha;
		byte *buffer = checkForHIRESTexture(tex, idx,
										 &width, &height, &pixelsize, true);

		// Check if an embedded png texture exists.
		if (buffer != NULL)
		{
			RGBData = true;
			alpha = (pixelsize == 4);
		}
		// Try to load a high resolution version of this texture.
		else if( (buffer = GL_LoadHighResTexture(tex, &width, &height, &pixelsize)) != NULL)
		{
			// High resolution texture loaded.
			RGBData = true;
			alpha = (pixelsize == 4);
		}
		else
		{
			RGBData = false;
			buffer = GL_BufferSkyTexture(idx, &width, &height, &pixelsize);
			if (tex->txf_flags & TXF_DELMIRROR)	// v220-35 lose the bottom half
			{
				height = tex->height >>= 1;
			}
			alpha = false;
		}

		// Upload it.
		GL_UploadTexture(buffer, width, height, alpha, true, RGBData, 0, &tex->gltex, 5);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, miplevel);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		M_Free(buffer);

		ctex->rendw = (float)tex->width;
		ctex->rendh = (float)tex->height;
	}

	rendtexdetail = NULL;
	return tex->gltex;
}

//===========================================================================
// firstRowAverageRGB
//===========================================================================
static void firstRowAverageRGB(byte *imgdata, int width, int height,
												int pixsize, byte *rgb)
{
	byte	*buff = imgdata;
	int		i, c;
	int		integerRGB[3] = {0,0,0};
	byte	col[4];

	for (i = 0; i < width; i++)
	{
		if (pixsize)
		{
			byte *p = buff + ((pixsize) * i);
			memcpy(col, p, pixsize);
		}
		else
			memcpy(col, palbuffer + 3*buff[i], 3);

		for(c = 0; c < 3; c++)
			integerRGB[c] += col[c];
	}

	for(c = 0; c < 3; c++)
		rgb[c] = (byte)(integerRGB[c] / width);
}

//===========================================================================
// GL_SetSkyTopColor v220-5
//	Returns the sky fadeout color of the given texture.
//===========================================================================
void GL_SetSkyTopColor(skylayer_t *slayer)
{
	int			texidx = slayer->texture;
	skycol_t	*skycol;
	int			i;

	if (texidx < 0 || texidx > numtextures)
	{ // Default to black.
		slayer->fadeout.rgb[0] = 0;
		slayer->fadeout.rgb[1] = 0;
		slayer->fadeout.rgb[2] = 0;
		return;
	}

	// See if a skytop color has already been set
	for (i = 0, skycol = skytop_colors; i < num_skytop_colors; i++, skycol++)
	{
		if (skycol->texidx == texidx)
		{
			slayer->fadeout.use = skycol->use;
			memcpy(slayer->fadeout.rgb, skycol->rgb, sizeof(slayer->fadeout.rgb));
			return;
		}
	}

	// None found so create a new entry
	{
		int		width, height, pixsize;
		byte	*imgdata;
		byte	rgb[3];

		if (num_skytop_colors >= maxnum_skytop_colors)
		{
			maxnum_skytop_colors += 16;
			skytop_colors = (skycol_t*)W_ReallocCLR(skytop_colors,
				sizeof(skycol_t) * maxnum_skytop_colors, STATICID, 205);
		}

		skycol = &skytop_colors[num_skytop_colors++];
		skycol->texidx = texidx;

		// Calculate the color.
		imgdata = GL_BufferSkyTexture(texidx, &width, &height, &pixsize);
		firstRowAverageRGB(imgdata, width, height, pixsize, rgb);

		M_Free(imgdata); // Free the temp buffer created by GL_BufferSkyTexture.

		for (i = 0; i < 3; i++)
			skycol->rgb[i] = (float)rgb[i] / 255.0f;

		// v220-3 always use if from linedef
		if (slayer->texture != skyflatnum)
		{
			skycol->use = true;
		}
		else
		{
			// Determine if it should be used.
			for (skycol->use = false, i = 0; i < 3; i++)
			{
				if ((float)skycol->rgb[i] > slayer->fadeout.limit)
				{
					// Colored fadeout is needed.
					skycol->use = true;
					break;
				}
			}
		}

		slayer->fadeout.use = skycol->use;
		memcpy(slayer->fadeout.rgb, skycol->rgb, sizeof(slayer->fadeout.rgb));
	}
}

//===========================================================================
// GL_NewTranslatedSprite
//===========================================================================
static transspr_t *GL_NewTranslatedSprite(int splnum, byte *table)
{
	transspr_t *trspr;
	
	transsprites = (transspr_t*)W_Realloc(transsprites, sizeof(transspr_t) 
							* (numtranssprites + 1), STATICID, 206);

	trspr = &transsprites[numtranssprites++];
	trspr->patch = splnum;
	trspr->table = table;
	trspr->tex = 0;
	return trspr;
}

//===========================================================================
// GL_GetTranslatedSprite
//===========================================================================
static transspr_t *GL_GetTranslatedSprite(int splnum, unsigned char *table)
{
	int i;

	for(i = 0; i < numtranssprites; i++)
	{
		if(transsprites[i].patch == splnum && transsprites[i].table == table)
		{
			return transsprites + i;
		}
	}
	return 0;
}

//===========================================================================
// amplify
//	The given RGB color is scaled uniformly so that the highest component 
//	becomes one.
//===========================================================================
static void amplify(byte *rgb)
{
	int i, max = 0;

	for(i = 0; i < 3; i++) if(rgb[i] > max) max = rgb[i];

	if (!max || max == 255) return;

	for(i = 0; i < 3; i++) rgb[i] = (byte)(rgb[i] * 255.0f / max);
}

//===========================================================================
// averageColorIdx
//	Used by dynamic lights. The resulting average color is 
//	amplified to be as bright as possible.
//===========================================================================
static void averageColorIdx(rgbcol_t *col, byte *data, int w, int h)
{
	int		i, count = w * h;
	uint	r, g, b;
	byte	rgb[3];

	// First clear them.
	memset(col->rgb, 0, 3);

	if (!count) return;

	r = g = b = 0;

	for(i = 0; i < count; i++)
	{
		memcpy(rgb, palbuffer + 3*data[i], 3);
		r += rgb[0];
		g += rgb[1];
		b += rgb[2];
	}

	col->rgb[0] = (byte)(r/count);
	col->rgb[1] = (byte)(g/count);
	col->rgb[2] = (byte)(b/count);

	// Make it glow (average colors are used with dynlights).
	amplify(col->rgb);
}

//===========================================================================
// averageColorRGB
//===========================================================================
static void averageColorRGB(rgbcol_t *col, byte *data, int w, int h)
{
	int  i, count = w*h;
	uint cumul[3];

	if(!count) return;
	memset(cumul, 0, sizeof(cumul));
	for(i = 0; i < count; i++)
	{
		cumul[0] += *data++;
		cumul[1] += *data++;
		cumul[2] += *data++;
	}
	for(i = 0; i < 3; i++) col->rgb[i] = (byte)(cumul[i]/count);
	amplify(col->rgb);
}

//==========================================================================
// GL_CalcLuminance
//	Calculates the properties of a dynamic light that the given sprite
//	frame casts. 
//==========================================================================
static void GL_CalcLuminance(int splnum, byte *buffer, int width, int height)
{
	spritelump_t	*slump = spritelumps + splnum;
	rgbcol_t		*sprcol = &slump->color;

	int		average[3] = { 0, 0, 0 };
	int		lowavg[3]  = { 0, 0, 0 };

	int		limit = 0xc0, collimit = 0xc0, poslimit = 0xe0;
	int		cnt = 0, avcnt = 0, lowcnt = 0, poscnt = 0;
	int		c, i, k;

	byte	rgb[3], *src, *alphasrc;

	//--------------------------------------

	slump->flarex = slump->flarey = 0;
	src = buffer;
	alphasrc = buffer + width*height;

	for(k = 0; k < height; k++)
	{
		for(i = 0; i < width; i++, src++, alphasrc++)
		{
			if(*alphasrc != 0xFF) continue;

			// Bright enough?
			memcpy(rgb, palbuffer + (*src * 3), 3);

			if(rgb[0] > poslimit || rgb[1] > poslimit || rgb[2] > poslimit)
			{
				// Include in the count.
				slump->flarex += i;
				slump->flarey += k;
				poscnt++;
			}
			// Bright enough to affect size?
			if(rgb[0] > limit || rgb[1] > limit || rgb[2] > limit) cnt++;
			// How about the color of the light?
			if(rgb[0] > collimit || rgb[1] > collimit || rgb[2] > collimit)
			{
				avcnt++;
				for(c=0; c<3; c++) average[c] += rgb[c];
			}
			else
			{
				lowcnt++;
				for(c=0; c<3; c++) lowavg[c] += rgb[c];
			}
		}
	}

	if(!poscnt)
	{
		slump->flarex = slump->width/2.0f;
		slump->flarey = slump->height/2.0f;
	}
	else
	{
		// Get the average.
		slump->flarex /= poscnt;
		slump->flarey /= poscnt;
	}
	// The color.
	if(!avcnt)
	{
		if(!lowcnt) 
		{
			// Doesn't the thing have any pixels??? Use white light.
			memset(sprcol->rgb, 0xff, 3);
		}
		else
		{
			// Low-intensity color average.
			for(c = 0; c < 3; c++) sprcol->rgb[c] = (byte)(lowavg[c] / lowcnt);
		}
	}
	else
	{
		// High-intensity color average.
		for(c = 0; c < 3; c++) sprcol->rgb[c] = (byte)(average[c] / avcnt);
	}

	amplify(sprcol->rgb);	// Amplify color.

	// How about the size of the light source?
	slump->lumsize = (2*cnt+avcnt)/3.0f / 70.0f;

	if(slump->lumsize > 1.0f) slump->lumsize = 1.0f;
}

//==========================================================================
// GL_SetTexCoords
//  Calculates texture coordinates based on the given dimensions.
//  The coordinates are calculated as width/CeilPow2(width), or 1
//  if CeilPow2 exceeds maxTexSize.
//==========================================================================
static void GL_SetTexCoords(float *tc, int width, int height)
{
	int pw = CeilPow2(width), ph = CeilPow2(height);

	if(pw > maxTexSize || ph > maxTexSize)
		tc[VX] = tc[VY] = 1.0f;
	else
	{
		tc[VX] = (float) width  / (float) pw;
		tc[VY] = (float) height / (float) ph;
	}
}

//===========================================================================
// GL_PrepareTranslatedSprite
//===========================================================================
static uint GL_PrepareTranslatedSprite(int splnum, byte *table)
{
	transspr_t *tspr = GL_GetTranslatedSprite(splnum, table);

	if(!tspr)
	{
		// There's no name for this patch, load it in.
		int			width, height, lump = spritelumps[splnum].lump;
		int			type = DoomLumpIdentify(lump, &width, &height, DLI_GRAPHIC);
		int			bufsize = width * height;
		patch_t		*patch;
		byte		*buffer;	// Currently only loading doom type tex
		RDBOOL		alpha;

        if (!bufsize || type != LMP_PATCH)  
        {  
            Con_Error("GL_PrepareTranslatedSprite Invalid Sprite format %8.8s pindex %d lump %d",
                       W_LumpName(lump),
                       splnum,
                       lump);
        }

		patch  = (patch_t*)M_lumpCACHE(lump, NULL, MLC_LOCK, MLC_NOERR);
		buffer = (byte*)M_Calloc(4 * bufsize, 227);	// Currently only loading doom type tex
		
		alpha = DrawRealPatch(buffer, width, height, patch, 0, 0, table, true);
		spritelumps[splnum].alpha = alpha;

		// Calculate light source properties.
		GL_CalcLuminance(splnum, buffer, patch->width, patch->height);

		ColorOutlines(buffer, patch->width, patch->height, alpha);

		tspr = GL_NewTranslatedSprite(splnum, table);
		GL_UploadTexture(buffer, patch->width, patch->height, alpha, true, false, 1, &tspr->tex, 6);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, miplevel);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);


		M_unlockCACHE(patch);
		M_Free(buffer);

		// Determine coordinates for the texture.
		GL_SetTexCoords(spritelumps[splnum].tc, width, height);
	}

	return tspr->tex;
}

//===========================================================================
// GL_PrepareSprite
//===========================================================================

typedef struct {
	char		*name;
	short		iskey;
	DGLuint		tex;
	int			lump;
	ushort		hsflags;
	ushort		i_width;
	ushort		i_height;
	ushort		p_width;
	ushort		p_height;
	float		tcI[2];
	float		tcP[2];

} hudsprite_t;

hudsprite_t hudsprites[] = {
	{ ""          },		// 0 - Empty won't be set/referenced
	{ "ammoa0", 0 },		// 1
	{ "arm1a0", 0 },		// 2
	{ "arm2a0", 0 },		// 3
	{ "bkeya0", 1 },		// 4  offsets needed in HUD for keys
	{ "bskua0", 2 },		// 5
	{ "cella0", 0 },		// 6
	{ "rkeya0", 1 },		// 7
	{ "rocka0", 0 },		// 8
 	{ "rskua0", 2 },		// 9
	{ "sboxa0", 0 },		// 10
 	{ "stima0", 0 },		// 11
	{ "ykeya0", 1 },		// 12
 	{ "yskua0", 2 },		// 13
 	{ "media0", 1 },		// 14
	{ NULL        }
};

int ammo_sprite[NUMAMMO] =
{
	SPR_AMMO,
	SPR_SBOX,
	SPR_CELL,
	SPR_ROCK
};

int healthSPR = SPR_STIM;

void GL_ResetHudsprites()
{
	int i = 0;

	while (hudsprites[++i].name)
		GL_DeleteTexIndex(&hudsprites[i].tex, "hudsprites");

	for(i = 0; i < numspritelumps; i++)
		spritelumps[i].hudspr = 0;
}

void GL_ClearAllHudIgnoreFlags()
{
	int i = 0;

	while (hudsprites[++i].name)
		hudsprites[i].hsflags &= ~HS_IGNORE;
}

// v220-31
int NGL_getFlags()
{
	int flags = 0;

	for (int i = 1; hudsprites[i].name; i++)
	{
		if (hudsprites[i].hsflags & HS_IGNORE)
			flags |= (1 << i);
	}

	return flags;
}

// v220-31
void NGL_setFlags(int flags)
{
	for (int i = 1; hudsprites[i].name; i++)
	{
		if (flags & (1 << i))
			hudsprites[i].hsflags |= HS_IGNORE;
		else
			hudsprites[i].hsflags &= ~HS_IGNORE;	;
	}
}

// b211-0 key matching for scripted map
RDBOOL GL_SetIgnoreFlag(int val)
{
int matchKey(int val);
	int index = matchKey(val & T79_ALLKEYMASK);

	if (index)
	{
		hudsprites[index].hsflags |= HS_IGNORE;
		return true;
	}
	return false;
}

// b211-0 key clearing for scripted map. index must be pre-validated
void GL_ClearIgnoreFlag(int index)
{
	hudsprites[index].hsflags &= ~HS_IGNORE;
}

RDBOOL GL_GetHSsize(int index, int *w, int *h, int *f, float *ttc)
{
	*f = hudsprites[index].hsflags;

	if (*f & (HS_PNGTEX|HS_JPGTEX))
	{
		ttc[VX] = ttc[VY] = 1.0f;

		*w = hudsprites[index].i_width;
		*h = hudsprites[index].i_height;
		return true;
	}

	if (*f & HS_PWADNAME)
	{
		ttc[VX] = hudsprites[index].tcP[VX];
		ttc[VY] = hudsprites[index].tcP[VY];

		*w = hudsprites[index].p_width;
		*h = hudsprites[index].p_height;

		return true;
	}

	if (*f & HS_IWADNAME)
	{
		ttc[VX] = hudsprites[index].tcI[VX];
		ttc[VY] = hudsprites[index].tcI[VY];
		*w = hudsprites[index].i_width;
		*h = hudsprites[index].i_height;
		return true;
	}

	return false;
}

// b211-0 Not a gl function per se as justs sets an array but the array is used in
// conjunction with GL_PrepareSprite so placing it here keeps everything together.
// This is only called once on startup.
void GL_SetHudsprites()
{
	int		i, type, width, height;

	// b211-0 Wads, if found to have hud problems, can be dealt with here.

	// Alter the ammo_sprites for arctic_se/batman.
	if (ge_inf.arctlev)
	{
		ammo_sprite[1] = SPR_SHEL;
		ammo_sprite[3] = SPR_SBOX;
		healthSPR = SPR_MEDI;
	}
	else if (ge_inf.batman)
	{
		ammo_sprite[1] = SPR_AMMO;
		ammo_sprite[3] = SPR_LAUN;
		hudsprites[8].name = "LAUNA0";
	}
	
	// Check if grenade has been selected
	if (ge_inf.grenade_id)
	{
		hudsprites[8].name = "GHUDA0";
		// This has no iwad entry so values have to be set
		hudsprites[8].i_width  =  7;
		hudsprites[8].i_height = 12;
	}

	// Set the array with IWAD and PWAD (if exists) sprite values
	// and check for png textures in the data\patches\hud directory
	i = 0;

	while (hudsprites[++i].name)
	{
		char	*name = hudsprites[i].name;
		int		n_iwad = W_CheckIWADNumForName(name);
		int		n_pwad = hiresPatches ?
					W_CheckPWADNumForName(name) : W_CheckPWADNumForLoResName(name);
		int		n_chkpwad;

		hudsprites[i].hsflags = 0;
		hudsprites[i].lump = 0;

		// no skull keys in doom1
		if (ge_inf.doomwad && hudsprites[i].iskey == 2) continue;

		if (n_iwad >= 0)
		{
			// Precheck validity of sprite found
			type = DoomLumpIdentify(n_iwad, &width, &height, DLI_GRAPHIC);

			if (width && height && type == LMP_PATCH)
			{
				hudsprites[i].i_width  = width;
				hudsprites[i].i_height = height;

				if (height > 16)
				{
					float factor = 16.f/(float)height;

					hudsprites[i].i_height = (int)((float)height * factor + 0.5f);
					hudsprites[i].i_width  = (int)((float)width * factor + 0.5f);
				}

				hudsprites[i].lump = n_iwad;
				hudsprites[i].hsflags |= (HS_IWADNAME|HS_DOOMTEX);
				GL_SetTexCoords(hudsprites[i].tcI, width, height);
			}
		}

		if (   hiresPatches
			&& !ge_inf.codlev	// codlev has its own specific hudsprites
			&& n_iwad >= 0
			&& (n_chkpwad = W_CheckPWADNumForLoResName(name)) >= 0)
		{
			int w, h, typepwad = DoomLumpIdentify(n_chkpwad, &w, &h, DLI_GRAPHIC);

			if (   typepwad == LMP_PATCH
				&& (hudsprites[i].i_width != w
				|| hudsprites[i].i_height != h))
			{
				hudsprites[i].p_width  = w;
				hudsprites[i].p_height = h;

				if (h > 16)
				{
					float factor = 16.f/(float)h;

					hudsprites[i].p_height = (int)((float)h * factor + 0.5f);
					hudsprites[i].p_width  = (int)((float)w * factor + 0.5f);
				}

				hudsprites[i].lump     = n_chkpwad;
				hudsprites[i].hsflags  = (HS_DOOMTEX|HS_PWADNAME|HS_CUSTONAME);
				GL_SetTexCoords(hudsprites[i].tcP, w, h);
				continue;
			}
		}

		if (n_pwad >= 0)
		{			// Precheck validity of sprite found
			type = DoomLumpIdentify(n_pwad, &width, &height, DLI_GRAPHIC);

			if (width && height && (type == LMP_PATCH  || type == LMP_PNG  || type == LMP_JPG))
			{
				if (type == LMP_JPG)
					hudsprites[i].hsflags |= HS_JPGTEX;
				else if (type == LMP_PNG)
					hudsprites[i].hsflags |= HS_PNGTEX;
				else
					hudsprites[i].hsflags |= HS_DOOMTEX;

				hudsprites[i].p_width  = width;
				hudsprites[i].p_height = height;

				if (height > 16)
				{
					float factor = 16.f/(float)height;

					hudsprites[i].p_height = (int)((float)height * factor + 0.5f);
					hudsprites[i].p_width  = (int)((float)width * factor + 0.5f);
				}

				hudsprites[i].lump     = n_pwad;
				hudsprites[i].hsflags |= HS_PWADNAME;
				GL_SetTexCoords(hudsprites[i].tcP, width, height);
			}
		}
	}
}

// b211-0 Clip, if found, top, bottom, left and/or right fully transparent rows/columns
// v220-7 corrected
byte *chkAlphaColumns(byte *buffer, int *width, int *height, char *name)
{
	int		c, w, h;
	int		ml = 0;
	int		mr = 0;
	int		mt = 0;
	int		mb = 0;
	int		nh, nw;
	int		wd = *width;
	int		ht = *height;
	int		pos, of = wd * ht;

	// check left columns
	for (pos = of, w = 0; w < wd; w++, ml++, pos++)
	{
		for (c = h = 0; h < ht; h++)
		{
			if (buffer[pos + h * wd] != 0xFF) c++;
		}

		if (c != ht) break;
	}

	if (ml == wd) return buffer;

	// check right columns
	for (pos = of + wd - 1, w = 0; w < wd; w++, mr++, pos--)
	{
		for (c = h = 0; h < ht; h++)
		{
			if (!buffer[pos + h * wd]) c++;
		}

		if (c != ht) break;
	}

	// check top rows
	for (pos = of, h = 0; h < ht; h++, mt++, pos += wd)
	{
		for (c = w = 0; w < wd; w++)
		{
			if (!buffer[pos + w]) c++;
		}

		if (c != wd) break;
	}


	// check bottom rows
	for (pos = of + of - 1, h = 0; h < ht; h++, mb++, pos -= wd)
	{
		for (c = w = 0; w < wd; w++)
		{
			if (!buffer[pos - w]) c++;
		}

		if (c != wd) break;
	}

	if (!ml && !mr && !mt && !mb) return buffer;

	if ( (nw = wd - (ml + mr)) > 0 && (nh = ht - (mt + mb)) > 0)
	{
		byte	*newbuf = (byte*)M_Calloc(nw * nh * 4, 0);
		byte	*p1 = newbuf;
		byte	*p2 = newbuf + nw * nh;

		pos = ml + mt * wd;

		for (h = 0; h < nh; h++, pos += wd)
		{
			for (w = 0; w < nw; w++)
			{
				*p1++ = buffer[pos + w];
				*p2++ = buffer[pos + w + of];
			}
		}

		M_Free(buffer);
		buffer  = newbuf;
		*width  = nw;
		*height = nh;
	}

	return buffer;
}

// v220-35 If hsindex is set then these are hud sprites called
//         from ST_drawHUDSprite
uint GL_PrepareSprite(int splnum, int *hsindex, int callnum)
{
	DGLuint		*tex = NULL;
	int			lump;
	ushort		hsnum = 0;

	if (splnum < 0 || splnum >= numspritelumps)
	{
		return 0;
	}

	lump = spritelumps[splnum].lump;

	if (!lumpinfo[lump].size) return 0;

	tex = &spritelumps[splnum].tex;	// Set the default

	if (hsindex)	// This can only be set in ST_drawHUDSprite()
	{
		hsnum = spritelumps[splnum].hudspr;
		
		*hsindex = 0;	// default until set otherwise

		// First set the ptr in spritelumps[splnum].hudspr
		if (!hsnum)
		{
			int i = 0;

			spritelumps[splnum].hudspr = 0x8000;	// flag as searched, index unset

			while (hudsprites[++i].name)			// Find the index
			{
				if (!_stricmp(hudsprites[i].name, lumpinfo[lump].name))	// index found
				{
					spritelumps[splnum].hudspr |= i;// OR the index

					tex = &hudsprites[i].tex;		// redirect the tex ptr
					hsnum = *hsindex = i;
					break;
				}
			}
		}
		else if (hsnum & 0x8000) // If flag is set then get the tex ptr
		{
			hsnum &= ~0x8000;

			if (hsnum)
			{
				tex = &hudsprites[hsnum].tex;	// redirect the tex ptr
				*hsindex = hsnum;			// in case texture needs to be reloaded
			}
		}
	}

	if ( !(*tex) )
	{
		extern char	*defhudstr;

		int		width, height;
		int		type = DoomLumpIdentify(lump, &width, &height, DLI_GRAPHIC);
		int		w, h, p;	// width, height, pixelsize
		int		n_hires;
		byte	*buffer = NULL;
		RDBOOL	alpha = false;

		//----------------------
		palpix = -1;	// v220-18 will be set to 0 if png has single layer and palette entry 0 = 0,0,0

		if (!width || !height || type != LMP_PATCH)
		{
			if (verbose)
			{
				pr_OFinsert("INVALID SPRITE %8.8s: %s %s : w %i h %i size %i\n",
					lumpinfo[lump].name,
					typedesc[type],
					grpdesc[lumpinfo[lump].groupMarker],
					width, height, lumpinfo[lump].size);
			}

			lumpinfo[lump].size = 0;
			return 0;
		}

		if (   !hsindex
			&& !(ge_inf.switch_flags & (SWF_NC|SWF_XA))
			&&  (n_hires = W_CheckTXNumForName(lumpinfo[lump].name)) >= 0 && n_hires != lump)
		{
			if (lumpinfo[n_hires].lumptype == LMP_PNG)
				buffer = PNG_Load(NULL, n_hires, &w, &h, &p, 4);
			else if (lumpinfo[n_hires].lumptype == LMP_JPG)
				buffer = JPG_Load(NULL, n_hires, &w, &h, &p, 4);
		}
		else
		if (hsindex && *hsindex && !(hudsprites[*hsindex].hsflags & HS_CUSTONAME) )
		{
			int		flags = hudsprites[*hsindex].hsflags;
			char	*codname = NULL;

			if (hiresPatches && ge_inf.codlev)
			{
				if (*hsindex == 5)
					codname = "cod_bskua0";
				else
				if (*hsindex == 9)
					codname = "cod_rskua0";
				else
				if (*hsindex == 13)
					codname = "cod_yskua0";
			}

			if (ge_inf.switch_flags & SWF_AP && defhudstr && *hsindex == 6)
			{
				char path[MAXPATHNAMESIZE];

				sprintf(path, "%s%s.png", defhudstr, "cella02");	// hud sprite for alternative weapons
				buffer = PNG_Load(path, 0, &w, &h, &p, 1);
				hudsprites[*hsindex].hsflags |= HS_PNGTEX;
			}
			else
			if (flags & HS_JPGTEX)
			{
				buffer = JPG_Load(NULL, hudsprites[*hsindex].lump, &w, &h, &p, 4);
			}
			else if (flags & HS_PNGTEX)
			{
				buffer = PNG_Load(NULL, hudsprites[*hsindex].lump, &w, &h, &p, 4);
			}
			// check for a hi-res texture in the data HUD folder
			else if (defhudstr && hiresPatches && *hsindex > 0 && *hsindex < 15)
			{
				char path[MAXPATHNAMESIZE];

				sprintf(path, "%s", defhudstr);

				if ( !(_chdir(path)) )
				{
					if (codname)
						strcat(path, codname);
					else
						strcat(path, hudsprites[*hsindex].name);

					// All hud textures are, or should be, png types
					strcat(path, ".png");

					if ((_access(path, 0 )) != -1)
					{
						buffer = PNG_Load(path, 0, &w, &h, &p, 1);
						hudsprites[*hsindex].hsflags |= HS_PNGTEX;
					}
					_chdir(ddRuntimeDir.path);
				}
			}

			// v220-35 separate treatment else lores sprite will be adversely affected
			if (buffer)
			{
				alpha = (p == 4);

				if (!alpha && !palpix)	// If !palpix then signifies black is transparent
				{
					int		numpels = w*h;
					byte	*ptr, *nptr, *newptr;
					int		i;

					nptr = newptr = (byte*)malloc(numpels * 4);

					for (i = 0, ptr = buffer; i < numpels; i++, ptr += 3, nptr += 4)
					{
						nptr[0] = ptr[0];
						nptr[1] = ptr[1];
						nptr[2] = ptr[2];

						if (!ptr[0] && !ptr[1] && !ptr[2])
						{
							nptr[3] = 0;
							alpha = true;
						}
						else
							nptr[3] = 255;
					}

					M_Free(buffer);
					buffer = newptr;
				}

				GL_UploadTexture(buffer, w, h, alpha, true, true, 0, tex, 7);

				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, miplevel);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

				if (alpha)
				{
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
				}
				M_Free(buffer);return *tex;
			}
		}

		if (buffer)		// If true this is a hi-res texture
		{
			spritedef_t		*sprdef = sprites;
			float			*tc = spritelumps[splnum].tc;
			int				numpels = w*h;
			byte			*ptr;

			tc[0] = tc[1] = 1.0f;
			alpha = p == 4;

			if (!alpha && !palpix)	// If !palpix then signifies black is transparent
			{
				byte	*nptr, *newptr = (byte*)malloc(numpels * 4);
				int		i;

				nptr = newptr;

				for (i = 0, ptr = buffer; i < numpels; i++, ptr += 3, nptr += 4)
				{
					nptr[0] = ptr[0];
					nptr[1] = ptr[1];
					nptr[2] = ptr[2];

					if (!ptr[0] && !ptr[1] && !ptr[2])
					{
						nptr[3] = 0;
						alpha = true;
					}
					else
						nptr[3] = 255;
				}

				M_Free(buffer);
				buffer = newptr;
			}

			ColorOutlinesRGBA(buffer, w, h, alpha);

			GL_UploadTexture(buffer, w, h, alpha, true, true, 0, tex, 8);
		}
		else
		{
			patch_t	*patch = (patch_t*)M_CacheLumpNum(lump, NULL, false);
			int		bufsize	= 4 * width * height;
			RDBOOL	coordset = false;

			buffer = (byte*)M_Calloc(bufsize, 228);

			alpha = DrawRealPatch(buffer, width, height, patch, 0, 0, 0, true);

			if (isarctic)
			{
				// v220-33 Hack to improve OGL rendering of arctic's trees, formed using
				//  two textures, by removing the black outlines caused by wrapping with
				//  pixel blend on the top and bottom of each texture.
				if (!strcmp(lumpinfo[lump].name, "COL5B0"))
				{
					// Shift all rows down two leaving both the top and bottom rows transparent.
					// The bottom two rows are lost but since it is offset into the ground by 5
					// we can get away with this without having to increase the texture's size.
					byte	*tempbuffer = (byte*)M_Calloc(bufsize, 224);
					byte	*ptr;
					int		sz = height * width;

					ptr = buffer + width;	// point one row up

					memcpy(tempbuffer + width, ptr, width * (height - 2));			// data
					memcpy(tempbuffer + width + sz, ptr + sz, width * (height - 2));// alpha

					M_Free(buffer);
					buffer = tempbuffer;
				}
				else if ( (isarctic == 1 && !strcmp(lumpinfo[lump].name, "PVISA0"))
					|| (isarctic == 2 && !strcmp(lumpinfo[lump].name, "VILE\\0")) )
				{
					// Shift all rows bar the top one up one leaving the bottom row transparent
					// This can be done without increasing the texture size because there are
					// several fully transparent rows at the top of the texture.
					byte	*tempbuffer = (byte*)M_Calloc(bufsize, 224);
					byte	*ptr;
					int		sz = height * width;

					// copy the rest shifted one row up
					ptr = buffer + width;	// point one row up

					memcpy(tempbuffer, ptr, width * (height - 1));			// data
					memcpy(tempbuffer + sz, ptr + sz, width * (height - 1));// alpha

					M_Free(buffer);
					buffer = tempbuffer;
				}
			}

			// We may need to clip alpha columns and/or rows - (e.g. arctic_se)
			if (   alpha
				&& hsindex
				&& *hsindex
				&& hudsprites[*hsindex].hsflags & HS_DOOMTEX
				&& hudsprites[*hsindex].hsflags & HS_PWADNAME)
			{
				byte *ptr = buffer;

				buffer = chkAlphaColumns(buffer, &width, &height,lumpinfo[lump].name);

				if (ptr != buffer)
				{
					hudsprites[*hsindex].p_height = height;
					hudsprites[*hsindex].p_width  = width;
					// co-ords will need resetting
					GL_SetTexCoords(hudsprites[*hsindex].tcP, width, height);
				}
				coordset = true;
			}

			// Calculate light source properties.
			GL_CalcLuminance(splnum, buffer, width, height);
			
			ColorOutlines(buffer, width, height, alpha);

			// Determine coordinates for the texture.
			if (!coordset) GL_SetTexCoords(spritelumps[splnum].tc, width, height);
			GL_UploadTexture(buffer, width, height, alpha, true, false, 1, tex, 9);

			M_Free(patch);
		}

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, miplevel);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		spritelumps[splnum].alpha = alpha;

		if (alpha)
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		}

		M_Free(buffer);
	}

	return *tex;
}

//===========================================================================
// GL_SetSprite
//===========================================================================
void GL_SetSprite(int splnum, int *hsindex, int callnum)
{
	GL_Bind(GL_PrepareSprite(splnum, hsindex, callnum));
}

//===========================================================================
// GL_SetTranslatedSprite
//===========================================================================
void GL_SetTranslatedSprite(int splnum, byte *trans)
{
	GL_Bind(GL_PrepareTranslatedSprite(splnum, trans));
}

// Only checks types embedded in a pwad
static byte* GL_CheckHighResPatch(char *lname, int lump,
								int *pixelsize, int *width, int *height)
{
	void	*ptr;
	int		txlump;

	// First check between TX_ or HI_ markers , there must be
	if (  (txlump = W_CheckTXNumForName(lname)) > 0
		&& lumpinfo[txlump].lumptype == LMP_PNG)
	{
		ptr = PNG_Load(NULL, txlump, width, height, pixelsize, 5);
		return (byte*)ptr;
	}

	if (txlump > 0 && lumpinfo[txlump].lumptype == LMP_JPG)
	{
		ptr = JPG_Load(NULL, txlump, width, height, pixelsize, 5);
		return (byte*)ptr;
	}

	// v220-21 check handling where not between accepted markers
	if (!R3D_ScriptedMap)
	{
		if (  (txlump = W_CheckPWADNumForName(lname)) > 0
			&& lumpinfo[txlump].lumptype == LMP_PNG)
		{
			ptr = PNG_Load(NULL, txlump, width, height, pixelsize, 5);
			return (byte*)ptr;
		}

		if (txlump > 0 && lumpinfo[txlump].lumptype == LMP_JPG)
		{
			ptr = JPG_Load(NULL, txlump, width, height, pixelsize, 5);
			return (byte*)ptr;
		}
	}

	return NULL;
}

//===========================================================================
//	GL_SetPatch - rewritten for R3D
//	  No mipmaps are generated for regular doom patches
//	  alpha ignored with some Doom patches
//	  Only hi-res textures are compressed (if optioned)
//    hi-res textures are scaled by the hardware.
//    v220-8 nearest switch added,
//           This is mainly set true by calls from the wi stuff
//    v220-13 patch type now explicitly checked
//            hi-res textures must have a same named pwad or iwad lump
//            for scaling purposes (their width and height is used).
//    v220-21 revised
//===========================================================================
RDBOOL GL_SetPatch(int lump, RDBOOL nearest)	
{
	if (lump < 0 || lump >= numlumps || !gl_inited) return false;

	if (!lumptexinfo[lump].ltitex)
	{
		lumpinfo_t *lump_p;

		RDBOOL	alpha = false;
		RDBOOL	doomalpha = false;
		RDBOOL	RGBdata = true;

		patch_t	*patch = NULL;

		byte	*buffer = NULL;
		byte	*tempptr = NULL;
		char	lname[10];

		int		w, h, width, height, pixsize = 2;
		int		type = DoomLumpIdentify(lump, &width, &height, DLI_GRAPHIC);
		int		numpels = width * height;
		int		plump = -1, ilump = -1;

		//--------------------------------------------
		
		if (type == LMP_IGNORE) return false; // This won't do!

		strncpy(lname, lumpinfo[lump].name, 8);
		lname[8] = 0;

		// Check for a PWAD patch
		lump_p = lumpinfo + numlumps; // Scan backwards

		if (type == LMP_PATCH && pwadHandle(lumpinfo[lump].handle))
		{
			plump = lump;
		}
		else
		{
			while (--lump_p >= lumpinfo)
			{
				if (   pwadHandle(lump_p->handle)
					&& !stricmp(lump_p->name, lname))
				{
					int lnum = lump_p - lumpinfo;

					if (DoomLumpIdentify(lnum, &width, &height, DLI_GRAPHIC) == LMP_PATCH)
					{
						numpels = width * height;
						plump = lnum;
						break;
					}
				}
			}
		}

		// If unset look for an IWAD patch
		if (plump < 0)
		{
			if (type == LMP_PATCH && lumpinfo[lump].handle == IWADhandle)
			{
				ilump = lump;
			}
			else
			{
				lump_p = lumpinfo + numlumps; // Scan backwards

				while (--lump_p >= lumpinfo)
				{
					if (   lump_p->handle == IWADhandle
						&& !stricmp(lump_p->name, lname))
					{
						int lnum = lump_p - lumpinfo;

						if (DoomLumpIdentify(lnum, &width, &height, DLI_GRAPHIC) == LMP_PATCH)
						{
							numpels = width * height;
							ilump = lnum;
							break;
						}
					}
				}
			}
		}

		// Finally check the RISEN3DWADhandle
		if (ilump < 0 && plump < 0)
		{
			lump_p = lumpinfo + numlumps; // Scan backwards

			while (--lump_p >= lumpinfo)
			{
				if (   lump_p->handle == RISEN3DWADhandle
					&& !stricmp(lump_p->name, lname))
				{
					int lnum = lump_p - lumpinfo;

					if (DoomLumpIdentify(lnum, &width, &height, DLI_GRAPHIC) == LMP_PATCH)
					{
						numpels = width * height;
						ilump = lnum;
						break;
					}
				}
			}
		}

		// A standard patch must exist even if a hi-res patch
		// exists as it will be needed for scaling purposes.
		if (!numpels || (ilump < 0 && plump < 0))
		{
			pr_OFinsert("tex \"%.8s\" not found or invalid\n", lname);
			lumpinfo[lump].lumptype = LMP_IGNORE;
			return false;
		}

		// NOTE: some pwads have duplicated patches inserted. To stop this from
		// preventing standard hi-res replacement the switch -nodup can be used.
		if (ge_inf.nodup && plump >= 0 && ilump < 0)
		{
			lump_p = lumpinfo + numlumps; // Scan backwards

			while (--lump_p >= lumpinfo)
			{
				if (   lump_p->handle == IWADhandle
					&& !stricmp(lump_p->name, lname))
				{
					int lnum = lump_p - lumpinfo;

					if (DoomLumpIdentify(lnum, &w, &h, DLI_GRAPHIC) == LMP_PATCH)
					{
						ilump = lnum;
						break;
					}
				}
			}

			if  (ilump >= 0 && w == width && h == height)
			{
				byte	*b1 = (byte*)M_Calloc(4 * numpels, 229);
				byte	*b2 = (byte*)M_Calloc(4 * numpels, 229);
				RDBOOL	a1, a2;
				int		c, i;

				patch = (patch_t*)M_CacheLumpNum(ilump, NULL, true);
				a1 = DrawRealPatch(b1, width, height, patch, 0, 0, 0, true);
				M_Free(patch);
				patch = (patch_t*)M_CacheLumpNum(plump, NULL, true);
				a2 = DrawRealPatch(b2, width, height, patch, 0, 0, 0, true);
				M_Free(patch);

				if (a1 == a2)
				{
					for (c = i = 0; i < w * h; i++)
					{
						if (b1[i] != b2[i]) { c++; break; }
					}

					if (!c)
					{
						pr_OFinsert("identical pwad/iwad patch \"%s\"\n", lname);
						plump = -1;
					}
				}

				M_Free(b1), M_Free(b2);
			}
		}

		// Check if this patch has been replaced with a custom high-res type
		if ( (buffer = GL_CheckHighResPatch(lname, lump, &pixsize, &w, &h)) )
		{
			width = w; height = h;
		}

		if (plump >= 0)
		{
			patch = (patch_t*)M_CacheLumpNum(plump, NULL, true);
		}
		else	// at least ilump must be set
		{
			patch = (patch_t*)M_CacheLumpNum(ilump, NULL, true);

			if (!buffer)	// Look for standard replacement hi-res texture
			{
				char	*path;

				if (   (path = IW_HiresPatchFile(lname))
					&& (buffer = GL_LoadImage(path, &w, &h, &pixsize, false, false)) )
				{
					width = w; height = h;
				}
			}

			// v220-35 Check for no replacement patch in a pwad requiring a standard
			// patch to be substituted. These patches must use the iwad playpal.
			// Note that the iwadpalbuffer is only set if a pwad playpal was found.
			if (!buffer && iwadpalbuffer)
			{
				tempptr = palbuffer;
				palbuffer = iwadpalbuffer;
			}
		}

		// The size/offset information.
		lumptexinfo[lump].offx = (short)(-patch->xoffset);
		lumptexinfo[lump].offy = (short)(-patch->yoffset);
		lumptexinfo[lump].origw = patch->width;
		lumptexinfo[lump].origh = patch->height;
		lumptexinfo[lump].width  = (ushort)width;
		lumptexinfo[lump].height = (ushort)height;

		if (buffer)
		{
			alpha = pixsize == 4;
		}
		else
		{
			numpels = width * height;
			if (!numpels) return false;

			// Note: filloutlines does not apply with patches
			buffer = (byte*)M_Calloc(4 * numpels, 229);
			RGBdata = false;

			alpha = DrawRealPatch(buffer, width, height, patch, 0, 0, 0, true);

			// Ignore alpha with certain patches
			if (   !_stricmp(lname, "TITLEPIC")
				|| !_stricmp(lname, "CREDIT")
				|| !_stricmp(lname, "HELP1")
				|| !_stricmp(lname, "HELP"))
			{
				alpha = false;
			}
			else if (alpha && !nearest)
			{
				// v220-13 add the load/save bar patches
				if (   !_stricmp(lname, "M_LSLEFT")
					|| !_stricmp(lname, "M_LSCNTR")
					|| !_stricmp(lname, "M_LSRGHT"))
				{
					nearest = true;	// prevents lines on a vertical edges
				}
			}
			
			doomalpha = alpha;
		}

		M_Free(patch);

		// v220-8 R3D pixblt removed
		// Limits below those where this would have been called are no longer supported

		// Generate a texture.
		GL_UploadTexture(buffer, width, height, alpha, RGBdata, RGBdata, 0, &lumptexinfo[lump].ltitex, 10);

		M_Free(buffer);

		if (!lumptexinfo[lump].ltitex)
			Con_Error("FAILED TO LOAD %s type %i", lname, lumpinfo[lump].lumptype);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, nearest ? GL_NEAREST : GL_LINEAR);

		// v220-4 if doomalpha clamp rather than repeat with this type
		if (doomalpha)
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		}
		else
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		}
		if (tempptr) palbuffer = tempptr;
	}

	GL_Bind(lumptexinfo[lump].ltitex);
	return true;
}

//===========================================================================
// GL_SetNoTexture
//===========================================================================
void GL_SetNoTexture(void)
{
	if (gl_inited) GL_Bind(0);
	curtex = 0;
}

//===========================================================================
// GL_PrepareLightTexture
//	The dynamic light map is a 64x64 grayscale 8-bit image.
//===========================================================================
DGLuint GL_PrepareLightTexture(void)
{
	if (!dltexname)
	{
		uint size;
		byte *data = (byte*)M_CacheLumpName("DLIGHT", &size, true);
		
		if (size != 64*64) Con_ErrorD("Invalid DLIGHT texture");

		GL_NewTexture(&dltexname);
		// No mipmapping or resizing is needed, upload directly.
		GL_TexImage(GL_LUMINANCE, 64, 64, 0, data);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

		M_Free(data);
	}

	return dltexname;
}

//===========================================================================
// GL_DeleteTexture
//	Deletes a texture. Only for textures (not for sprites, flats, etc.).
//===========================================================================
void GL_DeleteTexture(int texidx)
{
	if (texidx < 0 || texidx >= numtextures) return;

	GL_DeleteTexIndex(&texArray[texidx]->gltex, "texArray 2");
}


//===========================================================================
// GL_GetSkinTexIndex
//===========================================================================
int GL_GetSkinTexIndex(const char *skin)
{
	skintex_t	*st;
	char		realpath[MAXPATHNAMESIZE];
	int			i;

	if (!skin || !skin[0])
		Con_ErrorD("GL_GetSkinTexIndex: invalid skin");

	// Convert the given skin file to a full pathname.
	_fullpath(realpath, skin, MAXPATHNAMESIZE - 1);

	for (i = 0; i < numskinnames; i++)
	{
		if (!_stricmp(skinnames[i].path, realpath))
			return i;
	}

	if (numskinnames >= MAX_NUM_SKINNAMES)
		Con_ErrorD("GL_GetSkinTexIndex: Out of skin name slots");

	st = &skinnames[numskinnames];
	st->tex = 0; // Not yet prepared.

	// Allocate the skintex_t path.
	st->path = (char*)W_Calloc(strlen(realpath) + 1, STATICID, 102);
	strcpy(st->path, realpath);

	if(verbose) pr_OFinsert("SkinTex: %s => %i\n", skin, numskinnames);

	return numskinnames++;	
}

//===========================================================================
// GL_GetSkinTexByIndex
//===========================================================================
static skintex_t *GL_GetSkinTexByIndex(int id)
{
	if(id >= 0 && id < numskinnames) return skinnames + id;
	return NULL; // Not recognised
}

RDBOOL chkSkinIndex(model_t *mdl, int skin)
{
	skintex_t	*st;

	if (   mdl->info.numSkins > 0
		&& (st = GL_GetSkinTexByIndex(mdl->skins[skin].id))
		&& !st->tex)
	{
		return true;
	}

	return false;
}

//===========================================================================
// GL_PrepareSkin
//===========================================================================
uint GL_PrepareSkin(model_t *mdl, int skin, ushort *skinalpha)
{
	skintex_t	*st;

	if (skin < 0 || mdl->info.numSkins < 1 || skin >= mdl->info.numSkins)
	{
		Con_Error("Invalid model %s skin %i numskins %i",
			mdl->fullpathName ? mdl->fullpathName : "NULL",
			skin, mdl->info.numSkins);
	}

	if ( !(st = GL_GetSkinTexByIndex(mdl->skins[skin].id)) )
		return 0;

	// If the texture has already been loaded we don't need to do anything.
	if (!st->tex)
	{
		int		width, height, pxsize;
		int		downscale = 0;

		// Load the model skin texture.
		byte *image = GL_LoadImage(mdl->skins[skin].name, &width, &height, &pxsize, false, false);

		if (!image)
			Con_Error("GL_PrepareSkin: \"%s\" not found", mdl->skins[skin].name);

		skinCount++;

		// v220-9 Convert lod image to alpha if pxsize == 3
		//        The aim is to render black as transparent
		if (pxsize == 3 && strstrCI(mdl->skins[skin].name, "\\MD2\\LOD\\"))
		{
			uint	i, sz = width * height;
			byte	*newimage = (byte*)calloc(sz * 4, 1);
			byte	*p1, *p2;

			p1 = image, p2 = newimage;

			for (i = 0 ; i < sz; i++, p1 += 3, p2 += 4)
			{
				memcpy(p2, p1, 3);

				// If rgb then set to opaque
				if (p1[0] || p1[1] || p1[2]) p2[3] = 0xFF;
			}

			M_Free(image);
			image = newimage;
			pxsize = 4;
		}

		if (skin_inf.downscaleskintex && width >= reduceTo && height >= reduceTo)
		{
			downscale = 2;
		}

		GL_UploadTexture(image, width, height, pxsize == 4, true, true, downscale, &st->tex, 11);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, miplevel);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		if (pxsize == 4)
		{
			st->hasalpha = true;
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		}
				
		// We don't need the image data any more.
		M_Free(image);
	}

	if (skinalpha && st->hasalpha) *skinalpha |= RF_ALPHA;

	return st->tex;
}

//===========================================================================
// GL_PrepareShinySkin
//===========================================================================
uint GL_PrepareShinySkin(modeldef_t *md, int sub)
{
	skintex_t	*st = GL_GetSkinTexByIndex(md->sub[sub].shinyskin);
	int			pxsize, width, height;
	byte		*image;

	if(!st) return 0;	// Does not have a shiny skin.
	if(!st->tex)
	{
		int downscale = 0;
		// Load in the texture. 
		image = GL_LoadImageCK(st->path, &width, &height, &pxsize, true, false);

		if (!image)
		{
			Con_Error("GL_PrepareShinySkin: %s not found", st->path);
		}

		if (skin_inf.downscaleskintex && width >= reduceTo && height >= reduceTo)
		{
			downscale = 2;
		}

		GL_UploadTexture(image, width, height, pxsize == 4, true, true, downscale, &st->tex, 12);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, miplevel);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

		// We don't need the image data any more.
		M_Free(image);
	}
	return st->tex;
}

//--------------------------------------------------------------------------

int CCmdResetTextures(int argc, char **argv)
{
	if (appStatus != 2) return false;

	GL_ShutdownTextureManager();
	GL_LoadSystemTextures();

	if (gamestate == GS_LEVEL)
	{
		R_PrecacheLevel(CACHE_GRAPHICS|CACHE_SPRITES);
	}
	Con_Message("GL textures reloaded.\n");
	return true;
}

//---------------------------------------------------------------------------
//
//	3D flats code
//
//---------------------------------------------------------------------------

// 3D masked flats
typedef struct
{
	int		lumpnum;
	int		snum;
	float	texw;
	float	texh;
	uint	gl_texid;
	int		alpha;

} flat3D_t;

static int		flat3Dcount = 0;
static int		flat3Dslots = 0;
static flat3D_t	*flat3Dlist = NULL;

// v220-31 check for a hires wall tex that has been used as a floor/ceil tex
static RDBOOL GL_CheckScaledHiresWallTex(flat3D_t *mflat)
{
	int			lumpnum, patchnum;
	int			texnum = findTEXNAMEfromPATCH(lumpinfo[mflat->lumpnum].name, &lumpnum, &patchnum);

	if (texnum >= 0)
	{
		texture_t	*tex;
		void		*imgptr;
		int			w, h, width, height;
		int			pixelsize;

		// Check for hires version - R3D
		if (  (tex = texArray[texnum])
			&& DoomLumpIdentify(lumpnum, &w, &h, DLI_GRAPHIC) == LMP_PATCH
			&& w == texArray[texnum]->width)
		{
			if ( !(imgptr = checkForHIRESTexture(tex, texnum, &width, &height, &pixelsize, false)) )
				imgptr = GL_LoadHighResTexture(tex, &width, &height, &pixelsize);

			if (imgptr)
			{
				mflat->alpha = pixelsize == 4;
				mflat->texw = (float)w;
				mflat->texh = (float)h;

				GL_UploadTexture((byte*)imgptr, width, height, pixelsize == 4, true, true, 0, &mflat->gl_texid, 13);

				// Set the parameters.
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, miplevel);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

				M_Free(imgptr);
				return true;
			}
		}
	}

	return false;
}

//===========================================================================
// GL_Set3DFlatTexture
//  Where scaled then a new entry must be generated
//===========================================================================
static void GL_Set3DFlatTexture(flat3D_t *mflat)
{
	// check for a high res replacement
	if ( GL_CheckScaledHiresWallTex(mflat) ) return;

	flat_t	*fl = R_GetFlat(mflat->lumpnum);
	int		width  = fl->width;
	int		height = fl->height	;

	patch_t	*patch  = (patch_t*)M_CacheLumpNum(mflat->lumpnum, NULL, true);
	byte	*buffer = (byte*)M_Calloc(4 * width * height, 233);

	//---------------------------------------

	mflat->alpha = DrawRealPatch(buffer, width, height, patch, 0, 0, 0, true);
	mflat->texw = (float)width;
	mflat->texh = (float)height;

	GL_UploadTexture(buffer, width, height, mflat->alpha, true, false, 0, &mflat->gl_texid, 14);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	M_Free(buffer);
	M_Free(patch);
}

//===========================================================================
// GL_Delete3Dflats - R3D
//===========================================================================
void GL_Delete3Dflats()
{
	if (flat3Dlist)
	{
		int i;

		for(i = 0; i < flat3Dcount; i++)
		{
			GL_DeleteTexIndex(&flat3Dlist[i].gl_texid, "flat3Dlist");
		}

		W_Free(flat3Dlist, STATICID, 505);
	}

	flat3Dcount = 0;
	flat3Dslots = 0;
	flat3Dlist = NULL;
}

//===========================================================================
// GL_Prepare3DFlat - uses the actual widths and heights
//	Sets the OpenGL texture. Returns true if masked
//===========================================================================
RDBOOL GL_Prepare3DFlat(rendpoly_t *poly, int lumpnum, int snum)
{
	flat_t	*flat = R_GetFlat(lumpnum);

	poly->flags = 0;
	poly->detail = NULL;

	// Get the translated one?
	if (flat->flat_translation != lumpnum)
		flat = R_GetFlat(flat->flat_translation);

	// Scaled textures have their own list, plus alpha
	// is preserved as these will be 3D flats textures
	if (flat->flflags & FLT_FLATFROMTEX)
	{
		flat3D_t	*mflat = flat3Dlist;
		int			i;

		for (i = 0; i < flat3Dcount; i++, mflat++)
		{
			if (mflat->lumpnum == lumpnum && mflat->snum == snum)
			{
				poly->texw = mflat->texw;
				poly->texh = mflat->texh;
				poly->tex  = mflat->gl_texid;
				return mflat->alpha;
			}
		}

		if (flat3Dcount == flat3Dslots)
		{
			flat3Dslots += 16;

			flat3Dlist = (flat3D_t*)W_Realloc(flat3Dlist,
				sizeof(flat3D_t) * flat3Dslots, STATICID, 207);
		}

		mflat = &flat3Dlist[flat3Dcount++];
		mflat->lumpnum = lumpnum;
		mflat->snum = snum;
		
		GL_Set3DFlatTexture(mflat);

		mflat->texw *= sectors[snum].scalingValueX;
		mflat->texh *= sectors[snum].scalingValueY;

		poly->texw = mflat->texw;
		poly->texh = mflat->texh;
		poly->tex  = mflat->gl_texid;

		return mflat->alpha;
	}
	
	GL_SetTexFlat(flat, has_detail);

	poly->tex  = flat->gltex;
	poly->texw = (float)flat->width;
	poly->texh = (float)flat->height;
	poly->detail = flat->detail.tex ? &flat->detail : NULL;

	return (flat->flflags & FLT_MASKED) != 0;
}

//===========================================================================
// GL_Delete3DTextures
//	Deletes textures on restarting a level
//===========================================================================

void GL_Delete3DTextures()
{
	P_Remove3DSwitches();

	if (newtexcount > 0)
	{
		int i;

		if (numtextures - numstandardtextures != newtexcount)
			Con_ErrorD("Program error [1]: corrupt texArray");

		pr_OFinsert("%5i 3D scaled or wall flat textures removed.\n", newtexcount);

		for (i = numtextures - newtexcount; i < numtextures; i++)
		{
			if (texArray[i]->txf_flags & TXF_DELETE)
			{
				GL_DeleteTexture(i);
				newtexcount--;
			}
		}

		if (newtexcount)
			Con_ErrorD("Program error [2]: corrupt texArray");

		numtextures = numstandardtextures;

		texturetranslation =
			(int*)W_Realloc(texturetranslation, (numtextures + 1) * sizeof(int), STATICID, 215);
		texArray  =
			(texture_t**)W_Realloc(texArray, numtextures * sizeof(texture_t*), STATICID, 216);
		texscales =
			(texscale_t*)W_ReallocCLR(texscales, numtextures * sizeof(texscale_t), STATICID, 217);

	}

	GL_Delete3Dflats();
}
