//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: m_args.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 3 (v220-6) 17/Dec/2008 gmj
//  -texdir option removed
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION: Argument handler
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "de_base.h"
#include "de_console.h"
#include "de_misc.h"

// MACROS ------------------------------------------------------------------

#define MAX_ARG_ABBREVIATED	64

// TYPES -------------------------------------------------------------------

enum // Parser modes.
{
	PM_NORMAL,
	PM_NORMAL_REC,
	PM_COUNT
};

typedef struct
{
	char long_name[32];
	char short_name[8];
} argname_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

int ExternalInfo(int type, char *wadspath);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static argname_t	*argnames[MAX_ARG_ABBREVIATED];
static char			**args;
static int			arg_pos;
static int			num_args = 0;
static int			num_abbr_names = 0;
static int			last_match;

// CODE --------------------------------------------------------------------

//==========================================================================
// ArgParse
//	Parses the given command line. 
//==========================================================================
int ArgParse(int mode, const char *cmdline)
{
	char	word[512], *ptr, *response, *cmd = (char *) cmdline;	
	FILE	*file;
	RDBOOL	quote = false;
	RDBOOL	is_response;
	int		count = 0;

	// Init normal mode?
	if(mode == PM_NORMAL) arg_pos = 0;

	for(; *cmd; )
	{
		// Must allow: -cmd "echo ""this is a command"""
		// But also: @"\Program Files\test.rsp\"
		// And also: @\"Program Files"\test.rsp
		// Hello" My"Friend means "Hello MyFriend"

		// Skip initial whitespace.
		cmd = M_SkipWhite(cmd);

		// Check for response files.
		if(*cmd == '@')
		{
			is_response = true;
			cmd = M_SkipWhite(cmd + 1);
		}
		else
		{
			is_response = false;
		}

		for(ptr = word; *cmd && (quote || !isspace(*cmd)); cmd++)
		{
			if(!quote)
			{
				// We're not inside quotes.
				if(*cmd == '\"') // Quote begins.
				{
					quote = true;
					continue;
				}
			}
			else
			{
				// We're inside quotes.
				if(*cmd == '\"') // Quote ends.
				{
					if(cmd[1] == '\"') // Doubles?
					{
						// Normal processing, but output only one quote.
						cmd++;
					}
					else
					{
						quote = false;
						continue;
					}
				}
			}
			*ptr++ = *cmd;
		}
		// Append NULL to the word.
		*(ptr++) = 0;

		// Word has been extracted, examine it.
		if(is_response) // Response file?
		{
			// Try to open it.
			if((file = fopen(word, "rt")) != NULL)
			{
				uint sz;

				// How long is it?
				fseek(file, 0, SEEK_END);
				sz = ftell(file);
				rewind(file);
				response = (char*)calloc(1, sz + 1);
				fread(response, 1, sz, file);
				fclose(file);
				count += ArgParse(mode == PM_COUNT? PM_COUNT
					: PM_NORMAL_REC, response);
				free(response);
			}
		}
		else if(!strcmp(word, "--")) // End of arguments.
		{
			return count;
		}
		else if(word[0]) // Make sure there *is* a word.
		{
			// Increase argument count.
			count++;
			if(mode != PM_COUNT)
			{
				// Allocate memory for the argument and copy it to the list.
				args[arg_pos] = (char*)malloc(strlen(word) + 1);
				strcpy(args[arg_pos++], word);
			}
		}
	}
	return count;
}

//==========================================================================
// ArgAbbreviate
//	Registers a short name for a long arg name.
//	The short name can then be used on the command line and CheckParm 
//	will know to match occurances of the short name with the long name.
//==========================================================================
static void ArgAbbreviate(char *longname, char *shortname)
{
	argname_t *name;

	if (num_abbr_names >= MAX_ARG_ABBREVIATED) return;

	argnames[num_abbr_names++] = name = (argname_t*)calloc(sizeof(argname_t), 1);
	strncpy(name->long_name , longname , sizeof(name->long_name )-1);
	strncpy(name->short_name, shortname, sizeof(name->short_name)-1);
}

// NOTE: Limit 64 (see MAX_ARG_ABBREVIATED)
static void Check_ArgAbbreviate()
{
	// Standard doom options
	ArgAbbreviate("-file"			, "-f"  );	//  1
	ArgAbbreviate("-skill"			, "-sk" );	//  2
	ArgAbbreviate("-warp"			, "-wp" );	//  3
	ArgAbbreviate("-nomonsters"		, "-nm" );	//  4

	// Window selection/sizing (removed)
/*	ArgAbbreviate("-height"			, "-h"  );
	ArgAbbreviate("-width"			, "-w"  );
	ArgAbbreviate("-nocenter"		, "-noc");
*/

	// Sound selection
	ArgAbbreviate("-sfxchan"		, "-ch" );	//  5
	ArgAbbreviate("-sfx_3d"			, "-3d" );	//  6

	// General management
	ArgAbbreviate("-config"			, "-cfg");	//  7
	ArgAbbreviate("-cparse"			, "-cp" );	//  8
	ArgAbbreviate("-defs"			, "-d"  );	//  9
	ArgAbbreviate("-command"		, "-cmd");	// 10
	ArgAbbreviate("-modeldir"		, "-md" );	// 11
	ArgAbbreviate("-nohightex"		, "-nht");	// 12
	ArgAbbreviate("-fontdir"		, "-fd" );	// 13
	ArgAbbreviate("-hirescustom"	, "-hcu");	// 14
	ArgAbbreviate("-hirescommon"	, "-hco");	// 15
	ArgAbbreviate("-nohirespatches"	, "-np" );	// 16
	ArgAbbreviate("-parse"			, "-p"  );	// 17
	ArgAbbreviate("-texdir2"		, "-td2");	// 18
	ArgAbbreviate("-verbose"		, "-v"  );	// 19

	// OGL options
	ArgAbbreviate("-anifilter"		, "-ani");	// 20
	ArgAbbreviate("-outlines"		, "-ol" );	// 21
	ArgAbbreviate("-texcomp"		, "-tc" );	// 22

	// R3D options
	ArgAbbreviate("-detail"			, "-dt" );	// 23
	ArgAbbreviate("-nocorefx"		, "-nc" );	// 24
	ArgAbbreviate("-waterdynamics"	, "-wd" );	// 25
	ArgAbbreviate("-use_sky_box"	, "-usb");	// 26	v220-3
	ArgAbbreviate("-force_sky_box"	, "-fsb");	// 27	v220-3
	// Editing options
	ArgAbbreviate("-edit_ambient"	, "-txa");	// 28
	ArgAbbreviate("-noobjects"		, "-txn");	// 29
	ArgAbbreviate("-texedit"		, "-txe");	// 30

	// Command line compatibility options
	ArgAbbreviate("-c_model_on"		, "-mfo");	// 31
	ArgAbbreviate("-c_resurrect_on"	, "-ro" );	// 32
	ArgAbbreviate("-c_stairs_off"	, "-stx");	// 33
	ArgAbbreviate("-c_nuke24_off"	, "-nnx");	// 34
	ArgAbbreviate("-c_skulllim_on"	, "-sko");	// 35
}

//==========================================================================
// ArgInit
//	Initializes the command line arguments list.
//==========================================================================
int ArgInit()
{
	const char *cmdline = GetCommandLine();

	if (!cmdline[0]) return 0;

	// Clear the names register.
	num_abbr_names = 0;
	memset(argnames, 0, sizeof(argnames));

	// First count the number of arguments.
	num_args = ArgParse(PM_COUNT, cmdline);

	// Allocate memory for the list.
	args = (char**)calloc(num_args, sizeof(char*));

	// Then create the actual list of arguments.
	ArgParse(PM_NORMAL, cmdline);

	// Register some abbreviations for command line options.
	Check_ArgAbbreviate();

	return num_args;
}

//==========================================================================
// ArgShutdown
//	Frees the memory allocated for the command line.
//==========================================================================
void ArgShutdown(void)
{
	int i;

	if (num_args)
	{
		for(i = 0; i < num_args; i++) free(args[i]);
		free(args);
		args = NULL;
		num_args = 0;
	}

	if (num_abbr_names)
	{
		for(i = 0; i < num_abbr_names; i++) free(argnames[i]);
		memset(argnames, 0, sizeof(argnames));
		num_abbr_names = 0;
	}
}

//==========================================================================
// Argc
//	Returns the number of arguments on the command line.
//==========================================================================
int Argc(void)
{
	return num_args;
}

//==========================================================================
// Argv
//	Returns a pointer to the i'th argument.
//==========================================================================
char * Argv(int i)
{
	if(i < 0 || i >= num_args) Con_Error("Argv: There is no arg %i", i);
	return args[i];
}

//==========================================================================
// ArgvPtr
//	Returns a pointer to the i'th argument's pointer.
//==========================================================================
char ** ArgvPtr(int i)
{
	if(i < 0 || i >= num_args) Con_Error("ArgvPtr: There is no arg %i", i);
	return &args[i];
}

//==========================================================================
// ArgGetLastMtch
//==========================================================================
int ArgGetLastMtch()
{
	return last_match;
}

//==========================================================================
// ArgGetLastMtch
//==========================================================================
void ArgSetLastMtch(int num)
{
	last_match = num;
}

//==========================================================================
// ArgNext
//==========================================================================
char * ArgNext(void)
{
	if(!last_match || last_match >= Argc()-1) return NULL;
	return args[++last_match];
}

//==========================================================================
// ArgCheck
//	Checks for the given parameter in the program's command line arguments.
//	Returns the argument number (1 to argc-1) or 0 if not present.
//==========================================================================
int ArgCheck(char *check)
{
	int i, k;

	last_match = 0;

	for(i = 1; i < num_args; i++)
	{
		if (!_stricmp(check, args[i])) 
		{
			last_match = i;
			return i;
		}
		
		// Check the short argnames for this arg.
		for(k = 0; k < num_abbr_names; k++)
		{
			if (!_stricmp(check, argnames[k]->long_name)
			&&  !_stricmp(argnames[k]->short_name, args[i]))
			{
				last_match = i;
				return i;
			}
		}
	}
	return 0;
}

//==========================================================================
// ArgCheckWith
//	Checks for the given parameter in the program's command line arguments
//	and it is followed by N other arguments.
//	Returns the argument number (1 to argc-1) or 0 if not present.
//==========================================================================
int ArgCheckWith(char *check, int num)
{
	int i = ArgCheck(check);

	if (!i || i + num >= num_args) return 0;
	return i;
}

//==========================================================================
// ArgIsOption
//	Returns true if the given argument begins with a hyphen.
//==========================================================================
int ArgIsOption(int i)
{
	if(i < 0 || i >= num_args) return false;
	return args[i][0] == '-';
}

//==========================================================================
// ArgExists
//	Returns true if the given parameter exists in the program's command
//	line arguments, false if not.
//==========================================================================
int ArgExists(char *check)
{
	return ArgCheck(check) != 0;
}
