//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
//
// $Log: m_nodepile.c
// 
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified for R3D
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:
//  Sets up link nodes for mobjs 
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "de_base.h"
#include "de_console.h"
#include "de_misc.h"

#include "m_winmem.h"

// MACROS ------------------------------------------------------------------

#define NP_MAX_NODES	0x20000

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

nodepile_t	thingnodes, linenodes;	// all kinds of wacky links

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//===========================================================================
// NP_Init
//   initial must be at least two, node zero is never used.
//===========================================================================
void NP_Init(nodepile_t *pile, int initial)
{
	pile->nodes = (linknode_t*)W_Calloc(sizeof(linknode_t) * initial, NODESID, 103);
	pile->count = initial;
	pile->pos   = 1;	// Index #1 is the first.
}

//===========================================================================
// NP_New
//	Pos always has the index of the next node to check when allocating
//	a new node. Pos shouldn't be accessed outside this routine because
//	its value may prove to be outside the valid range.
//===========================================================================
nodeindex_t NP_New(nodepile_t *pile, void *ptr)
{
	linknode_t	*nodes = pile->nodes;
	nodeindex_t	idx = 0;
	int			c, i, n = pile->count;

	// Use the pos value as the starting point. It has a
	// better chance of quickly happening upon an unused node.
	// Note: the value in pile->pos *must* be between 1 and n - 1
	//       Since it is inited to 1 it should never be less than 1

	if ( (c = pile->pos) >= n) c = 1;	// Wrap back to the start

	// Scan for an unused node
	for (i = c; i < n; i++)
	{
		if (!nodes[i].object_ptr) { idx = i; break; }
	}

	if (!idx)
	{	// No free node found from pos to n - 1
		// Check from 1 to pos - 1
		for (i = 1; i < c; i++)
		{
			if (!nodes[i].object_ptr) { idx = i; break; }
		}
	}

	if (!idx)
	{	// No free nodes exist - create some more
		idx = n;

		if (n >= NP_MAX_NODES)
		{
			// This can happen *theoretically* but is highly unlikely
			Con_Error("Program error: linknodes limit exceeded.\n");
		}

		if (n < 1024) n *= 2; else n += 1024;

		if (n > NP_MAX_NODES) n = NP_MAX_NODES;

		pile->nodes = nodes =
			(linknode_t*)W_ReallocCLR(nodes, n * sizeof(linknode_t), NODESID, 0);
		pile->count = n;
	}

	nodes[idx].object_ptr = ptr;

	// Make it point to itself by default (a root, basically).
	nodes[idx].next = nodes[idx].prev = idx;
	pile->pos = idx + 1;

	return idx;
}

//===========================================================================
// NP_Link
//	Links the node to the beginning of the ring. 
//===========================================================================
void NP_Link(nodepile_t *pile, nodeindex_t node, nodeindex_t root)
{
	linknode_t *nd = pile->nodes;
	
	nd[node].prev = root;
	nd[node].next = nd[root].next;
	nd[root].next = nd[nd[node].next].prev = node;
}

//===========================================================================
// NP_Unlink
//  Aim : d->n->p = d->p, d->p->n = d->n
//===========================================================================
void NP_Unlink(nodepile_t *pile, nodeindex_t node)
{
	linknode_t *nd = pile->nodes;

	// Original code was;
	// nd[ nd[nd[node].next].prev = nd[node].prev ].next = nd[node].next;

	// R3D - Make it clearer and give the compiler a chance :)
	// The next 4 lines replace the above (commented out) original:

	nodeindex_t p = nd[node].prev;
	nodeindex_t n = nd[node].next;

	nd[n].prev = p;
	nd[p].next = n;

	// Make it link to itself (a root, basically).
	nd[node].next = nd[node].prev = node;
}
