//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: m_winmem.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//
//
// DESCRIPTION:  Win memory list functions (replaces the old zone memory
//       functions) cacheing functions and general win memory allocation. 
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <string.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"

#include "m_winmem.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// max memory allowed for a single allocation
#define MAX_MEMLIMIT	0x2000000	// 32MB

// TYPES -------------------------------------------------------------------

#pragma pack(4)

// level allocation instance.
typedef struct memlevel_s
{
	struct	memlevel_s *prev;
	struct	memlevel_s *next;
	uint	size;
	uint	id;
	int		from;
	void	*ptr;
} memheader_t;

typedef struct
{
	char	*type;
	uint	id;
}winm_t;

#pragma pack()

// Names and id's for the win mem lists
winm_t wmtypes[] = {
	"THINK ",	THINKID,
	"MODEL ",	MODELID,
	"PARTCL",	PARTCLID,
	"LIST  ",	LISTID,
	"LEVEL ",	LEVELID,
	"STATIC",	STATICID,
	"NODES ",	NODESID,
	"MBOX  ",	MBOXID,
	"TEMP  ",	TEMPID,
	NULL,		0
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void CrashShutdown(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// The head and tail of the winmem ptr lists
static memheader_t mlist[NUM_MCLASS];

static RDBOOL	mem_error[NUM_MCLASS];
static int		memcounter[NUM_MCLASS];

static void		*cacheptr;
static void		*transientCachePtr;
static size_t	cachesize;
static RDBOOL	cachelocked;

static char		errbuff[64];

// CODE --------------------------------------------------------------------
void subfn_MemError(size_t size, char *type, int from)
{
	if (!size)
		Con_Error("Unset size %s from %i", type, from);
	else if (size > MAX_MEMLIMIT)
		Con_Error("Excessive memory call from %s %i [%u bytes requested %u max allowed]",
		type, from, size, MAX_MEMLIMIT);

	Con_Error("Insufficient Windows memory [%u bytes requested from %s %i]",
		size, type, from);
}

//===========================================================================
// M_Malloc
//===========================================================================
void *M_Malloc(size_t size, int from)
{
	void *ptr;
	
	if (size & 1) size++;	// Align (note: would fail with 0xFFFFFFFF
							//        but this is somewhat academic)

	if (size > MAX_MEMLIMIT) subfn_MemError(size, "M_Malloc", from);

	if (!size ||  !(ptr = malloc(size)) ) subfn_MemError(size, "M_Malloc", from);

	return ptr;
}

//===========================================================================
// M_Calloc
//===========================================================================
void *M_Calloc(size_t size, int from)
{
	void *ptr;
	
	if (size & 1) size++;	// Align

	if (size > MAX_MEMLIMIT) subfn_MemError(size, "M_Calloc", from);

	if (!size || !(ptr = calloc(size, 1)) ) subfn_MemError(size, "M_Calloc", from);

	return ptr;
}

//===========================================================================
// M_Realloc
//  NOTE: The caller must verify that size > 0
//===========================================================================
void *M_Realloc(void *ptr, size_t size, int from)
{
	if (size & 1) size++;	// Align

	if (size > MAX_MEMLIMIT) subfn_MemError(size, "M_Realloc", from);

	ptr = realloc(ptr, size);

	if (!size || !ptr)
	{
		subfn_MemError(size, "M_Realloc", from);
	}

	return ptr;
}

//===========================================================================
// M_Free
//===========================================================================
void M_Free(void *ptr)
{
	if (!ptr) return;

	if (ptr == cacheptr)
	{
		M_FreelumpCACHE();
		return;
	}
	free(ptr);
}

//===========================================================================
//
// M_lumpCACHE - single transient cache.
//
//  This is for transient memory usage.  It is not constantly
//  alloced/dealloced but just grows to the largest size requested.
//  The memory is automatically reclaimed on exit.
//
//  NOTE: If locked the cacheptr must ONLY be freed with M_FreelumpCACHE
//        else M_Free should be used.  This allows for memory to be freed
//		  if, for example, a known large allocation has been made.  The
//        main use of locking, however, was for debugging.  It could now
//        all be safely replaced but haven't got round to it.
//       
//
//===========================================================================

void *M_lumpCACHE(int lump, size_t *length, int locked, int error)
{
	lumpinfo_t	*l;
	size_t		c, size;

	if (cachelocked) Con_ErrorD("[1] CACHE locked");

	CheckLumpRange(lump, 3);

	l = lumpinfo + lump;

	if (length) *length = l->size;

	if (!l->size)
	{
		if (error)
			Con_Error("lump \"%.8s\" num %i has no data", l->name, lump);
		return NULL;
	}

	// add 1 byte to be sure. Useful for strings.
	size = l->size + 1;

	if (size & 1) size++;	// Align

	if (cachesize < size)
	{
		cachesize = size;
		cacheptr = M_Realloc(cacheptr, cachesize, 31);
	}

	lseek(l->handle, l->position, SEEK_SET);
	c = read(l->handle, cacheptr, l->size);

	if (c < l->size)
	{
		Con_Error("M_lumpCACHE: only read %i of %i on lump %i",
			c, l->size, lump);
	}

	memset((byte*)cacheptr + l->size, 0, size - l->size);

	cachelocked = locked;

	return cacheptr;
}

void *M_LockedCACHE(size_t size)
{
	if (cachelocked) Con_ErrorD("[2] CACHE locked");

	size += 1;	// Add one byte

	if (size & 1) size++;	// Align

	if (cachesize < size)
	{
		cachesize = size;
		cacheptr = M_Realloc(cacheptr, cachesize, 32);
	}

	memset(cacheptr, 0, cachesize);

	cachelocked = true;

	return cacheptr;
}

void M_unlockCACHE(void *ptr)
{
	if (ptr && cachelocked && ptr != cacheptr)
		Con_Error("CACHE ptrs do not match %p %p", ptr, cacheptr);
	cachelocked = false;
}

// Only called on exit
void M_FreelumpCACHE()
{
	if (cacheptr)
	{
		free(cacheptr);		// NOT M_Free as can be called from M_Free
		cacheptr = NULL;
		cachesize = 0;
		cachelocked = false;
	}
}

//===========================================================================
//
// Risen3D windows memory routines. These, together with thinker
// handling changes, replace the original zone memory functions.
//
//===========================================================================

static int getCLASS(uint id)
{
	int i;

	for (i = 0; i < NUM_MCLASS; i++)
		if (wmtypes[i].id == id) return i;

	pr_OFinsert("getCLASS error: unknown class id %.8X called\n", id);
	CrashShutdown();
	exit(-1);
}

static uint getWID(int num)
{
	if (num >= 0 && num < NUM_MCLASS)
		return wmtypes[num].id;

	pr_OFinsert("getWID error: unknown index %i\n", num);
	CrashShutdown();
	exit(-1);
}


//===========================================================================
// W_InitWINM
//  Called once on entry
//===========================================================================
void W_InitWINM()
{
	int	i;

	for (i = 0; i < NUM_MCLASS; i++)
	{
		memheader_t	*hed = &mlist[i];
		hed->prev = hed->next = hed;
	}
}

static void W_InitID(int id)
{
	memheader_t	*hed = &mlist[getCLASS(id)];
	hed->prev = hed->next = hed;
}

//===========================================================================
// W_Malloc
//===========================================================================
void *W_Malloc(size_t size, uint id, int from)
{
	void		*ptr;
	memheader_t	*mpt, *hed;

	int		numclass;

	//--------------------------------------

	if (!size)
	{
		int i;

		for (i = 0; i < 8; i++) { if (wmtypes[i].id == id) break; }
		Con_Error("W_Malloc type %s [%i] called with size zero", wmtypes[i].type , from);
	}

	numclass = getCLASS(id);
	ptr = malloc(sizeof(memheader_t) + size);

	if (!ptr)
	{
		sprintf(errbuff, "M_Malloc %s\n", wmtypes[numclass].type);
		subfn_MemError(size, errbuff, -1);
	}

	mpt = (memheader_t*)ptr;
	hed = &mlist[numclass];

	mpt->id   = id;
	mpt->from = from;
	mpt->size = size;

	// Link into list
	hed->prev->next = mpt;
	mpt->next = hed;
	mpt->prev = hed->prev;
	hed->prev = mpt;

	memcounter[numclass]++;

	mpt->ptr = (void*)((byte*)ptr + sizeof(memheader_t));
	return mpt->ptr;
}

//===========================================================================
// W_MallocLump
//===========================================================================
void *W_MallocLump(int lump, uint id, int from)
{
	lumpinfo_t	*l;
	void		*ptr;
	size_t		r, n;

	CheckLumpRange(lump, 4);

	l = lumpinfo + lump;

	if (!(r = l->size))
	{
		Con_Error("W_MallocLump: lump from %i \"%.8s\" num %i has no data", from, l->name, lump);
	}

	if (r & 1) r++;	// align
	r += 4;			// add an int of slack (in case of string termination/bad wav files etc)

	ptr = W_Calloc(r, id, from);

	lseek(l->handle, l->position, SEEK_SET);

	if ( (n = read(l->handle, ptr, l->size)) < l->size)
	{
		Con_Error("W_MallocLump: only read %i of %i on lump %i",
												n, l->size, lump);
	}

	return ptr;
}

//===========================================================================
// W_Calloc
//===========================================================================
void *W_Calloc(size_t size, uint id, int from)
{
	void  *ptr = W_Malloc(size, id, from);
	memset(ptr, 0, size);
	return ptr;
}

//===========================================================================
// W_Free
//===========================================================================
void W_Free(void *ptr, uint id, int from)
{
#ifdef _DEBUG
	int			nclass;
	memheader_t *hed, *mcurrent;

	if (!ptr) return;

	nclass = getCLASS(id);

	if (mem_error[nclass]) return;

	hed = &mlist[nclass];
	mcurrent = hed->next;

	while(mcurrent != hed)
	{
		if (mcurrent->id != id)
		{
			mem_error[nclass] = true;
			Con_Error("W_Free [%i] %s: header has bad id", from, wmtypes[nclass].type);
		}
		if (mcurrent->ptr == ptr)
		{
			// Unlink from list
			if (mcurrent->next) mcurrent->next->prev = mcurrent->prev;
			if (mcurrent->prev) mcurrent->prev->next = mcurrent->next;

			memcounter[nclass]--;

			ptr = (void*)((byte*)ptr - sizeof(memheader_t));
			free(ptr);
			ptr = NULL;

			return;
		}
		mcurrent = mcurrent->next;
	}
	mem_error[nclass] = true;
	Con_Error("W_Free [%i] %s: ptr [%p] not found in linked list",
											from, wmtypes[nclass].type, ptr);
#else	// release code

	memheader_t	*mpt;
	int			nclass;

	if (!ptr) return;

	nclass = getCLASS(id);
	ptr = (void*)((byte*)ptr - sizeof(memheader_t));

	mpt = (memheader_t*)ptr;

	if (mpt->id != id)
	{
		mem_error[nclass] = true;
		Con_Error("W_Free %s: from %i header has bad id", wmtypes[nclass].type, from);
	}

	// Unlink from list
	if (mpt->next) mpt->next->prev = mpt->prev;
	if (mpt->prev) mpt->prev->next = mpt->next;

	memcounter[nclass]--;

	free(ptr);
	ptr = NULL;

	return;
#endif
}

//===========================================================================
// subfn_Realloc
//===========================================================================
static void *subfn_Realloc(void *ptr, size_t size, uint id, int from, RDBOOL clear)
{
#ifdef _DEBUG
	memheader_t *hed, *mpt, *mcurrent;
	int			nclass;
	size_t		oldsize;

	nclass = getCLASS(id);
	hed = &mlist[nclass];
	mcurrent = hed->next;

	while (mcurrent != hed)
	{
		if (mcurrent->id != id)
		{
			mem_error[nclass] = true;
			Con_Error("W_Realloc [%i] %s: header has bad id",
									from, wmtypes[nclass].type);
		}

		if (mcurrent->ptr == ptr)
		{
			ptr = (void*)((byte*)ptr - sizeof(memheader_t));
			oldsize = ((memheader_t*)ptr)->size;

			ptr = M_Realloc(ptr, size + sizeof(memheader_t), 33);

			mpt = (memheader_t*)ptr;
			if (mpt->next) mpt->next->prev = mpt;
			if (mpt->prev) mpt->prev->next = mpt;
			mpt->size = size;
			mpt->from = -from;	// Indicate changed for debugging
			mpt->ptr = (void*)((byte*)ptr + sizeof(memheader_t));

			// Clear the newly allocated memory?
			if (clear && size > oldsize)
			{
				memset((byte*)mpt->ptr + oldsize, 0, size - oldsize);
			}

			return mpt->ptr;
		}
		mcurrent = mcurrent->next;
	}

	mem_error[nclass] = true;
	Con_Error("W_Realloc [%i] %s: ptr [%p] not found in linked list",
											from, wmtypes[nclass].type, ptr);
	return NULL; // happy compiler

#else	// release code

	memheader_t *mpt;
	size_t		oldsize;

	ptr = (void*)((byte*)ptr - sizeof(memheader_t));
	mpt = (memheader_t*)ptr;

	if (mpt->id != id)
	{
		int nclass = getCLASS(id);
		mem_error[nclass] = true;
		Con_Error("W_Realloc %s: header has bad id", wmtypes[nclass].type);
	}

	oldsize = mpt->size;

	ptr = M_Realloc(ptr, size + sizeof(memheader_t), 34);

	mpt = (memheader_t*)ptr;
	if (mpt->next) mpt->next->prev = mpt;
	if (mpt->prev) mpt->prev->next = mpt;
	mpt->size = size;
	mpt->from = from;	// for debugging
	mpt->ptr = (void*)((byte*)ptr + sizeof(memheader_t));

	// Clear the newly allocated memory?
	if (clear && size > oldsize)
	{
		memset((byte*)mpt->ptr + oldsize, 0, size - oldsize);
	}

	return mpt->ptr;
#endif
}

//===========================================================================
// W_Realloc
//===========================================================================
void *W_Realloc(void *ptr, size_t size, uint id, int from)
{
	if (!ptr) return W_Malloc(size, id, from);
	return subfn_Realloc(ptr, size, id, from, false);
}

// Reallocs and clears any extended memory
void *W_ReallocCLR(void *ptr, size_t size, uint id, int from)
{
	if (!ptr) return W_Calloc(size, id, from);
	return subfn_Realloc(ptr, size, id, from, true);
}

//===========================================================================
// W_FreeALL
//===========================================================================
static RDBOOL subfnFreeALL(uint id)
{
	int			nclass;
	memheader_t *hed, *mpt, *next, *mcurrent;
	void		*ptr;
	uint		bytesfreed = 0;
	int			setfrom;
	int			ptrsset;

	nclass = getCLASS(id);

	if (mem_error[nclass]) return true;

	ptrsset = memcounter[nclass];

	hed = &mlist[nclass];
	mcurrent = hed->next;

	while(mcurrent != hed)
	{
		memcounter[nclass]--;

		if (memcounter[nclass] < 0)
		{
			sprintf(errbuff, "W_FreeALL %s: list has bad count\n", wmtypes[nclass].type);
			mem_error[nclass] = true;

			if (appStatus == 3)	exit(-1);

			Con_ErrorD(errbuff);
		}

		next = mcurrent->next;

		ptr = (void*)((byte*)mcurrent->ptr - sizeof(memheader_t));

		mpt = (memheader_t*)ptr;

		if (mpt->id != id)
		{
			sprintf(errbuff, "W_FreeALL %s: header has bad id\n", wmtypes[nclass].type);
			mem_error[nclass] = true;

			if (appStatus == 3)	exit(-1);

			Con_ErrorD(errbuff);
		}

		bytesfreed += mpt->size;
		setfrom = mpt->from;	// for debugging

		// Unlink from list
		if (mcurrent->next) mcurrent->next->prev = mcurrent->prev;
		if (mcurrent->prev) mcurrent->prev->next = mcurrent->next;

		free(ptr);

		mcurrent = next;
	}

	if (memcounter[nclass])
	{
		mem_error[nclass] = true;
		pr_OFinsert("ERROR %s: of %i mem ptrs %i were not cleared\n",
						wmtypes[nclass].type, ptrsset, memcounter[nclass]);
	}
	else if (bytesfreed)	//  b211-0 Added
	{
		if (id != TEMPID)
			pr_OFinsert("%s ptrs [%5i]: bytes freed %9i\n", wmtypes[nclass].type, ptrsset, bytesfreed);
	}

	W_InitID(id);

	return mem_error[nclass];
}

// An id of 0 clears everything - 0 is only used on shutdown
RDBOOL W_FreeALL(uint id)
{
	int i;

	if (id) return subfnFreeALL(id);

	for (i = 0; i < NUM_MCLASS; i++)
	{
		if (subfnFreeALL(getWID(i)))
			return true;
	}

	return false;
}

// Check validity of list allocations
// Not terribly useful but better than nothing
void W_CheckHeap(int from)
{
	int			i;

	ma_inf.memsize = 0;

	for ( i = 0; i < NUM_MCLASS; i++)
	{
		memheader_t *hed = &mlist[i];
		memheader_t	*mcurrent = hed->next;
		uint		id = getWID(i);
		int			memcount = 0;

		ma_inf.mempointers[i] = 0;

		while(mcurrent != hed)
		{
			if (mcurrent->id != id)
			{
				mem_error[i] = true;
				Con_Error("W_CheckHeap %s: header has bad id [from %i]", wmtypes[i].type, from);
			}
			memcount++;
			ma_inf.memsize += mcurrent->size + 40;
			mcurrent = mcurrent->next;
		}
		if (memcount != memcounter[i])
				Con_Error("W_CheckHeap %s: list has bad count [from %i]", wmtypes[i].type, from);

		ma_inf.mempointers[i] = memcount;
	}
}
