//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: net_main.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified as required for Risen3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  minor changes - this code will ultimately be replaced
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
//  The game is currently single-player mode only. As such most
//  of the network code has now been removed. The residual code
//  now only deals with input events.
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:
//  Client/server networking. 
//  Player number zero is always the server.
//  In single-player games there is only the server present.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h> // for atoi()

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_network.h"
#include "de_play.h"
#include "de_graphics.h"
#include "de_misc.h"
#include "de_ui.h"

#include "r_main.h"

#include "r3d_compat.h"
#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

#define NETBUFFER_MAXDATA	1024	// This only buffers tic commands
#define LOCALTICS			10		// Built ticcmds are stored here.
#define	BACKUPTICS			70		// Two seconds worth of tics.
#define szTICCMD			sizeof(ticcmd_t)

// TYPES -------------------------------------------------------------------

typedef struct {
	ticcmd_t	lastcmd;		
	ticcmd_t	ticcmds[BACKUPTICS];
	int			numtics;
	int			firsttic;
	int			time;
	byte		runTime;
	byte		align;
} ticset_t;

typedef struct
{
	int		length;		// Number of bytes in the data buffer.
	byte	*nbcursor;	// Points into the data buffer.
	byte	data[NETBUFFER_MAXDATA]; // The data buffer
	byte	align;
} netbuffer_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void G_BuildTiccmd (ticcmd_t* cmd); 

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------


ddplayer_t		players[MAXPLAYERS];

RDBOOL			firstNetUpdate = true;

int				realtics;
int				availabletics;
int             gametic;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// Local packets are stored into this buffer.
static ticset_t		sender;
static netbuffer_t	netbuffer;
static netbuffer_t	reboundstore;
static ticcmd_t		localticcmds[BACKUPTICS];

static int	coordTimer = 0;
static int	lasttime;
static int	numlocal;		// Number of cmds in the buffer.

// CODE --------------------------------------------------------------------

//===========================================================================
// Net_Init
//===========================================================================
void Net_Init(void)
{
	numlocal = 0;	// Nothing in the buffer.
	memset(&netbuffer, 0, sizeof(netbuffer));
}

//===========================================================================
// Net_ClearCmds
//===========================================================================
void Net_ClearCmds()
{
	// Invalidate old cmds.
	memset(&sender, 0, sizeof(ticset_t));
}

//===========================================================================
// Net_TimeDelta
//  Returns delta based on 'now' (- future, + past).
//===========================================================================
static int Net_TimeDelta(byte now, byte then)
{
	int delta = (now >= then) ? now - then : 256 - then + now;

	// The time can be in the future. We'll allow one second.
	return delta > 220 ? delta - 256 : delta;
}

//===========================================================================
// Net_Update
//	Builds ticcmds for console player, sends out a packet.
//===========================================================================
void Net_Update()
{
	ticcmd_t	*cmd;
	int			newtics;
	int			start, num, i, time;

	int			nowtime = Sys_GetTime();

	//-------------------------------------

	if (firstNetUpdate)
	{
		firstNetUpdate = false;
		lasttime = nowtime;
	}

	if ( (newtics = nowtime - lasttime) > 0)
	{
		lasttime = nowtime;

		// Realtics counts the actual time that has passed.
		realtics += newtics;

		// Build new ticcmds for player 0
		for (i = 0; i < newtics; i++)
		{
			DD_ProcessEvents();	

			if (!ui_active && numlocal < LOCALTICS)
			{
				// Place the new ticcmd in the local ticcmds buffer.
				ticcmd_t *cmd = &localticcmds[numlocal++];

				G_BuildTiccmd(cmd);

				// Set the time stamp. Only the lowest byte is stored.
				cmd->time = (byte)(gametic + availabletics);

				// Availabletics counts the tics that have cmds.
				availabletics++;
			}
		}

		// Send the proper number of tics to the server.
		// Since we are the server the tics will rebound back
		netbuffer.nbcursor = netbuffer.data;
		netbuffer.length = 0;

		// Check enough space is available
		if ( (netbuffer.nbcursor - netbuffer.data + 1 + numlocal * szTICCMD) >= NETBUFFER_MAXDATA)
			Con_Error("Net buffer length exceeded");

		*netbuffer.nbcursor++ = (byte)numlocal;

		// Copy the tics to the outgoing packet.
		for (i = 0; i < numlocal; i++)
		{
			memcpy(netbuffer.nbcursor, &localticcmds[i], szTICCMD);
			netbuffer.nbcursor += szTICCMD;
		}

		// Update the length of the message.
		netbuffer.length = netbuffer.nbcursor - netbuffer.data;
		reboundstore = netbuffer;

		numlocal = 0;		// Clear the buffer counter
	}

	netbuffer = reboundstore;
	netbuffer.nbcursor = netbuffer.data;

	start = sender.firsttic + sender.numtics;
	
	// Add the tics into the client's ticcmd buffer, if there is room.
	// If the buffer overflows, the rest of the cmds will be forgotten.
	num = (byte)*netbuffer.nbcursor++;

	if (sender.numtics + num > BACKUPTICS) 
		num = BACKUPTICS - sender.numtics;

	sender.numtics += num;	// Increase the counter.

	// Copy as many as fits - all the bytes sent must be read
	for (i = start; num > 0; num--, i++)
	{
		if (i >= BACKUPTICS) i -= BACKUPTICS;

		cmd = &sender.ticcmds[i];
		memcpy(cmd, netbuffer.nbcursor, szTICCMD);
		netbuffer.nbcursor += szTICCMD;

		// Check the time stamp.
		time = gametic - Net_TimeDelta((byte)gametic, cmd->time);	// sender->enterTime

		if (time > sender.time) sender.time = time;
	}
}

//===========================================================================
// Net_GetTicCmd
//  v220-8 simplified
//===========================================================================
int Net_GetTicCmd(void *pCmd, int player)
{
	ticcmd_t	*cmd = (ticcmd_t*)pCmd;
	int			doMerge = false;

	while(1)	// Run round until no tics or nothing to merge
	{
		if (!sender.numtics) 
		{
			// No player commands left so use the last command.
			memcpy(cmd, &sender.lastcmd, szTICCMD);

			if (Net_TimeDelta((byte)gametic, sender.runTime) > 0)
				sender.runTime++;		// Advance runtime

			return false; 
		}

		// There's at least one command in the buffer.
		// There will be one less tic in the buffer after this.
		sender.numtics--;
		memcpy(cmd, &sender.ticcmds[sender.firsttic++], szTICCMD);

		// Make sure the firsttic iterator is in range.
		if (sender.firsttic >= BACKUPTICS) sender.firsttic = 0;

		// Do we need to do any merging? We're only interested in buttons.
		if (doMerge	&& !(cmd->buttons & (BT_SPECIAL|BT_CHANGE)) )
		{
			int lastbuttons = sender.lastcmd.buttons;

			if ( !(lastbuttons & BT_SPECIAL) && lastbuttons & BT_CHANGE)
			{
				// Use the old weapon change and old Attack and Use buttons if  set
				cmd->buttons |= (lastbuttons & (BT_CHANGE | BT_WEAPONMASK | BT_ATTACK | BT_USE));
			}
		}

		// This is the new last command.
		memcpy(&sender.lastcmd, cmd, szTICCMD);

		// Check that the command is valid (not if its tick has already passed).
		if ( (doMerge = (Net_TimeDelta(sender.runTime, cmd->time) < 0)) ) break;
	}

	sender.runTime = cmd->time;
	return true;
}
