//-----------------------------------------------------------------------------
//
//  Copyright (C) 1998 - 2005 id Software Inc.,
//   Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman,
//  Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_sight.c
//
// Revision R3D v2.1.0 build 62 04/Oct/2006 gmj
//  modified/extended for R3D
//  REJECT useage removed
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  3D flats occlusion added
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  3D line processing revised
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION:   LineOfSight/Visibility checks.
//
//-----------------------------------------------------------------------------


// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_play.h"
#include "de_refresh.h"
#include "r3d_compat.h"
#include "r3d_edit.h"
#include "r3d_defs.h"
#include "r3d_sl3D.h"
#include "r3d_slopes.h"
#include "r3d_modelbox.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

void P_PlaySeesound(mobj_t *mo);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char		*origTEX;
extern int		modelboxcnt;


// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// killough 4/19/98:
// Convert LOS info to struct for reentrancy and efficiency of data locality

static struct los_t
{
	divline_t	strace;						// from t1 to t2
	fixed_t		sightzstart, t2x, t2y;		// eye z of looker
	fixed_t		t2head;						// for 3D flats
	fixed_t		topslope, bottomslope;		// slopes to top and bottom of target
	fixed_t		bbox[4];
	fixed_t		zdiff;
	fixed_t		highlimit, lowlimit;

} los;

static subsector_t	**sseclist3D;
static int			ssecscount3D;
static int			maxssecscount3D;
static subsector_t	**listplanessec;
static int			planesseccount;
static int			maxplanesseccount;
static int			*fseclist;
static int			fseclistcount;
static int			maxfseclistcount;
static sector_t		*t1_sector;
static sector_t		*t2_sector;
static RDBOOL		samesubsector;
static int			*linehit;
static int			linehitnumber;
static int			linehitcount;
static int			allowplanes;


// CODE --------------------------------------------------------------------

void P_InitLineHit(int lcount)
{
	if (!linehit)
	{
		linehitnumber = lcount;
		linehit = (int*)W_Malloc(lcount*sizeof(int), STATICID, 3);
	}
	else if (linehitnumber < lcount)
	{
		linehitnumber = lcount;
		linehit = (int*)W_Realloc(linehit, lcount*sizeof(int), STATICID, 91);
	}

	memset(linehit, 0, lcount*sizeof(int));
	linehitcount = 0;
}

// P_DivlineSide
//  return 2 (on line) 0 (facing side 0) 1 (facing side 1)
int P_DivlineSide(fixed_t x, fixed_t y, const divline_t *node)
{
	int ret = asmPointOnSide(y - node->y, node->dx, node->x - x, node->dy);

	return (!ret ? 2 : ret > 0);
}

// v220-33
// Functions added add3DSSEC(), CheckSSECplanes(), checkHitFlat(), checkFlats3D()
static void add3DSSEC(subsector_t *ss)
{
	if (los.zdiff > FRACUNIT)
	{
		if (ss->flats3DArray && ss->ssvalidcount != ss_validcount)
		{
			ss->ssvalidcount = ss_validcount;

			if (ssecscount3D >= maxssecscount3D)
			{
				maxssecscount3D += 32;
				sseclist3D = (subsector_t**)W_Realloc(sseclist3D, maxssecscount3D * sizeof(subsector_t*), STATICID, 0);
			}
			sseclist3D[ssecscount3D++] = ss;
		}
	}
	
	if (!allowplanes) return;

	// Note lfsec and 3Dflats do not co-exist
	if (ss->lfsec >= 0 || ss->floorsec >= 0 || ss->lcsec >= 0)
	{
		// check a plane lies between the low and high sight points
		// and select the plane, if more than one, closest to the lowpoint
		if (ss->ssvalidcount != ss_validcount)
		{
			int		val = ss->lcsec;	// takes priority
			fixed_t diff = DDMAXINT;
			fixed_t hgt;

			ss->ssvalidcount = ss_validcount;

			if (val >= 0)
			{
				hgt = sectors[val].ceilingheight;

				// player above Thing
				if (hgt > los.sightzstart && hgt < los.t2head)
				{
					diff = hgt - los.lowlimit;
				}
			}
			
			if ( (val = ss->floorsec) >= 0)
			{
				hgt = sectors[val].floorheight;

				if (hgt > los.sightzstart && hgt < los.t2head)
				{
					if (hgt - los.lowlimit < diff)
					{
						diff = hgt - los.lowlimit;
					}
				}
			}
			
			if ( (val = ss->lfsec) >= 0)
			{
				hgt = sectors[val].floorheight;

				if (hgt > los.sightzstart && hgt < los.t2head)
				{
					if (hgt - los.lowlimit < diff)
					{
						diff = hgt - los.lowlimit;
					}
				}
			}
			
			if (diff != DDMAXINT)
			{
				for (int i = 0; i < fseclistcount; i++)
				{
					if (fseclist[i] == los.lowlimit + diff) goto addlosplane;
				}

				if (fseclistcount >= maxfseclistcount)
				{
					maxfseclistcount += 8;	// Probably will never encounter more than 4 - usually just 1
					fseclist = (int*)W_Realloc(fseclist, maxfseclistcount * sizeof(int), STATICID, 0);
				}

				fseclist[fseclistcount++] = los.lowlimit + diff;

addlosplane:;
				if (planesseccount >= maxplanesseccount)
				{
					maxplanesseccount += 32;
					listplanessec = (subsector_t**)W_Realloc(listplanessec, maxplanesseccount * sizeof(subsector_t*), STATICID, 0);
				}
				listplanessec[planesseccount++] = ss;
			}
		}
	}
}

static RDBOOL CheckSSECplanes()
{
	for (int i = 0; i < fseclistcount; i++)
	{
		fixed_t frac = FixedDiv(abs(los.sightzstart - fseclist[i]), los.zdiff);

		if (frac > plane3d.frac) return false;

		fixed_t x = los.strace.x + FixedMul(los.strace.dx, frac);
		fixed_t y = los.strace.y + FixedMul(los.strace.dy, frac);
		subsector_t *chk = R_PointInSubsector(x, y);

		for (int k = 0; k < planesseccount; k++)
		{
			if (listplanessec[k] == chk) return true;
		}
	}
	return false;
}

// Check if flat is between the trajectory from t1's firing height and t2's z height - if so it will block sight
static RDBOOL checkHitFlat(fixed_t planeheight, RDBOOL lookup)
{
	fixed_t frac = FixedDiv(abs(los.sightzstart - planeheight), los.zdiff);

	if (frac > plane3d.frac) return false;

	fixed_t x = los.strace.x + FixedMul(los.strace.dx, frac);
	fixed_t y = los.strace.y + FixedMul(los.strace.dy, frac);
	int		c, *flatslist;

	if ( !(flatslist = R_PointInSubsector(x, y)->flats3DArray) ) return false;

	c = *(int*)flatslist++;

	// Find matching flat height
	while (--c >= 0)
	{
		// floorheight + offset + depth
		int cmp = sectors[flatslist[1]].floorheight + flatslist[2];

		if (!lookup) cmp += flatslist[3];

		if (cmp == planeheight)
		{
			plane3d.set = true;
			plane3d.x = x;
			plane3d.y = y;
			plane3d.frac = frac;

			if (lookup)
			{
				lop.editz = cmp;
				plane3d.z = cmp - 2*FRACUNIT;
				plane3d.floorpic = -1;
				plane3d.ceilpic = sectors[flatslist[0]].ceilingpic;
			}
			else
			{
				lop.editz = cmp;
				plane3d.z = cmp;
				plane3d.floorpic = sectors[flatslist[0]].floorpic;
				plane3d.ceilpic = -1;
			}
			return true;
		}

		flatslist += FLIST_ENTRIES;
	}

	return false;
}

// flatslist must be set - R3D
static RDBOOL checkFlats3D()
{
	fixed_t	cmp;
	int		i;

	for (i = 0; i < ssecscount3D; i++)
	{
		subsector_t	*ss = sseclist3D[i];
		int			*flatslist = ss->flats3DArray;
		int			count = *(int*)flatslist++;

		// Find nearest matching flat.
		// Search floor flats from highest to lowest.
		if (los.t2head < los.sightzstart) // Looking down, so can only be a floor flat
		{
			flatslist += (count - 1)*FLIST_ENTRIES;	// start at the top

			while (--count >= 0)
			{
				// floorheight + offset + depth
				cmp = sectors[flatslist[1]].floorheight + flatslist[2] + flatslist[3];

				if (los.sightzstart > cmp && los.t2head < cmp
					&& checkHitFlat(cmp, false))	// floor
				{
					return true;
				}

				flatslist -= FLIST_ENTRIES;
			}
		}
		else	// Looking up - search ceiling flats from lowest to highest.
		{
			// Remember - floor's higher than ceiling
			while (--count >= 0)	// start at the bottom
			{
				// floorheight + offset
				cmp = sectors[flatslist[1]].floorheight + flatslist[2];

				if (   los.sightzstart < cmp && los.t2head > cmp
					&& checkHitFlat(cmp, true))		// ceil
				{
					return true;
				}

				flatslist += FLIST_ENTRIES;
			}
		}
	}

	return false;
}

//
// P_CrossSubsector
// Returns true
//  if strace crosses the given subsector successfully.
//
static RDBOOL P_CrossSubsector(subsector_t *ssec)
{
	sector_t		*front, *back;
	vertex_t		*v1, *v2;
	divline_t		divl;
	line_t			*line;
	seg_t			*seg;
	fixed_t			frac, slope;
	int				i;
	
	add3DSSEC(ssec);

	// First check standard 2D line types.  This will adjust the
	// slopes as necessary based on standard openings between the mobj's
	// b211-0 modified to fix MBF bug	// $v72

	for (i = 0, seg = &segs[ssec->firstline]; i < ssec->linecount; i++, seg++)
	{
		if ( (line = seg->linedef) && linehit[line - lines] != linehitcount)
		{
			fixed_t ot, ob;
			
			linehit[line - lines] = linehitcount;

			if (   line->lnflags2 & LNF2_UNBLOCKED
				&& line->frontsector->floorheight == line->backsector->floorheight)
			{
				continue;	// v220-16 emulate doom bug
			}

			// OPTIMIZE: killough 4/20/98: Added quick bounding-box rejection test
			
			if (line->bbox[BOXLEFT ] > los.bbox[BOXRIGHT] ||
				line->bbox[BOXRIGHT] < los.bbox[BOXLEFT ] ||
				line->bbox[BOXBASE ] > los.bbox[BOXTOP  ] ||
				line->bbox[BOXTOP  ] < los.bbox[BOXBASE ])
			{
				continue;
			}
			
			v1 = line->v1;
			v2 = line->v2;

			// line isn't crossed?
			if (P_DivlineSide(v1->x, v1->y, &los.strace) ==
				P_DivlineSide(v2->x, v2->y, &los.strace))
			{
				continue;
			}
			
			divl.dx = v2->x - (divl.x = v1->x);
			divl.dy = v2->y - (divl.y = v1->y);
			
			// line isn't crossed?
			if (P_DivlineSide(los.strace.x, los.strace.y, &divl) ==
				P_DivlineSide(los.t2x, los.t2y, &divl))
			{
				continue;
			}

			// stop because it is not two sided anyway
			if (!(line->ml_flags & ML_TWOSIDED)) return false;	// stop
			if (line->xtraflags & XT_OCCLUDE) return false;		// v220-33

			// Check it's not a  blocking line - R3D
			if (   line->lnflags2 & LNF2_OCCLUDE
				|| (line->lnflags2 & LNF2_BLOCKALL
				&& !(line->lnflags2 & LNF2_ALLOWSIGHT)) )
			{
				return false;
			}

			// Check for R3D 3D line types
			if (line->specType >= R3D_40)
			{
				// b220-1 processed immediately

				// 3D lines normally only block if they have textures and where,
				// if present, the bottom texture's not masked or if the mid-
				// lineheight sent was > 1 with no midtex. (1 is reset to 0 and
				// is used to prevent the flag being set at load time).
				if ( !(line->lnflags2 & (LNF2_LOWTEXMASKED|LNF2_3DNOBLOCK)) )	
				{
					int sec3d, lo, hi, z;

					frac = P_InterceptVector(&los.strace, &divl);
			
					sec3d = line->frontsector->liftsec3D;
					lo = line->d3d->offset3D + (sec3d >= 0 ?
								sectors[sec3d].floorheight : line->backsector->floorheight);
					hi = lo + (line->lnflags2 & LNF2_MIDTEXMASKED ?
								line->d3d->lineheightlow : line->d3d->totalheight);
					z = los.sightzstart + FixedMul(los.t2head - los.sightzstart, frac);

					if (z >= lo && z <= hi)
					{
						plane3d.frac = frac;
						return false;
					}
				}
				continue;
			}

			front = line->frontsector;
			back  = line->backsector;

			ot = front->ceilingheight;
			ob = front->floorheight;
			
			if (front != back)
			{
				ot = MIN_OF(ot, back->ceilingheight);	
				ob = MAX_OF(ob, back->floorheight  );
			}

			// quick test for totally closed sectors
			if (ob >= ot) return false;				// stop

			frac = P_InterceptVector(&los.strace, &divl);

			// Check for R3D slopes - v220-33
			if (line->lnflags2 & (LNF2_LOWTRIWALL|LNF2_HITRIWALL))
			{
				fixed_t		x = los.strace.x + FixedMul (los.strace.dx, frac);
				fixed_t		y = los.strace.y + FixedMul (los.strace.dy, frac);

				if (line->lnflags2 & LNF2_LOWTRIWALL)
				{
					sector_t	*s = line->frontsector->stptr >= 0 ? line->frontsector : line->backsector;
					ob = MAX_OF((SL_PointHeightZ(s, x, y, LOSLOPE)>>FRACBITS)<<FRACBITS, ob);
				}
				if (line->lnflags2 & LNF2_HITRIWALL)
				{
					sector_t	*s = line->frontsector->stptr >= 0 ? line->frontsector : line->backsector;
					ot = MIN_OF((SL_PointHeightZ(s, x, y, HISLOPE)>>FRACBITS)<<FRACBITS, ot);
				}
			}
	
			if ( (slope = FixedDiv(ob - los.sightzstart , frac)) > los.bottomslope)
			{
				los.bottomslope = slope;
			}
			
			if ( (slope = FixedDiv(ot - los.sightzstart , frac)) < los.topslope)
			{
				los.topslope = slope;
			}
			
			if (los.topslope <= los.bottomslope)
			{
				plane3d.frac = frac;
				return false;	// stop
			}
		}
	}

	return true;
}

// P_CrossBSPNode
//  Returns true if strace crosses the given node successfully.
//
//  killough 4/20/98: rewritten to remove tail recursion, clean up, and optimize
//
static RDBOOL P_CrossBSPNode(int bspnum)
{
	while (!(bspnum & NFx_SUBSECTOR))
	{
		node_t	*bsp = &nodes[bspnum];
		// Ensure side is either 0 or 1
		int		side = P_DivlineSide(los.strace.x, los.strace.y, (divline_t *)bsp) & 1;

		if (side == P_DivlineSide(los.t2x, los.t2y, (divline_t *) bsp))
		{
			bspnum = bsp->children[side];		// doesn't touch the other side
		}
		else if (!P_CrossBSPNode(bsp->children[side]))
		{
			return false;						// cross the starting side
		}
		else
		{
			bspnum = bsp->children[side^1];		// cross the ending side
		}
	}

	return P_CrossSubsector(&subsectors[bspnum & ~NFx_SUBSECTOR]);
}

// P_CheckSight
//	Returns true if a straight line between t1 and t2 is unobstructed.
//	Look from eyes of t1 to any part of t2 (start from middle of t1).
//
RDBOOL P_CheckSight (mobj_t *t1, mobj_t *t2)
{
	RDBOOL	ret = false;
	fixed_t	t1h, t2h;

	plane3d.set = false;
	plane3d.frac = FRACUNIT;
	plane3d.floorpic = plane3d.ceilpic = -1;

	// v220-31 needs at least one node
	if (!numnodes) return true;

	// If either is unlinked, they can't see each other.
	if (   !t1->subsector
		|| !t2->subsector
		|| t1->subsector->viewgroup != t2->subsector->viewgroup)
	{
		return false;
	}

	if (!ppt.inmap)
	{
		if (t1->dplayer && t1->dplayer->mo == t1) return false;
		if (t2->dplayer && t2->dplayer->mo == t2) return false;
	}

	if (t1->flags2 & MF2_CORPSEFADE || t2->flags2 & MF2_CORPSEFADE)
		return false;

	t1_sector = t1->subsector->sector;
	t2_sector = t2->subsector->sector;
	// v220-33 exploding barrels fix
	t1h = (t1->type == MT_BARREL && !t1->height) ? 0x260000 : t1->height;
	t2h = (t2->type == MT_BARREL && !t2->height) ? 0x260000 : t2->height;

	// Make fake floors and ceilings block monster view
	// v220-3 Allow sight if water is translucent
	if (t1_sector->heightsec >= 0 && !(t1->subsector->ssflags & SSF_FAKETRANS))
	{
		sector_t *sec = &sectors[t1_sector->heightsec];

		if (   (t1->z + t1h < sec->floorheight && t2->z >= sec->floorheight)
			|| (t1->z >= sec->ceilingheight && t2->z + t1h < sec->ceilingheight))
		{
			return false;
		}
	}

	if (t2_sector->heightsec >= 0 && !(t2->subsector->ssflags & SSF_FAKETRANS))
	{
		sector_t *sec = &sectors[t2_sector->heightsec];

		if (   (t2->z + t2h < sec->floorheight && t1->z >= sec->floorheight)
			|| (t2->z >= sec->ceilingheight && t1->z + t2h < sec->ceilingheight))
		{
			return false;
		}
	}

	samesubsector = t1->subsector == t2->subsector;
	ssecscount3D = planesseccount = fseclistcount = 0;

	// An unobstructed LOS is possible.
	// Now look from firing point of t1 to any part of t2.

	los.sightzstart = t1->z + t1h - (t1h>>2);
	los.t2head		= t2->dplayer ? viewz : t2->z + t2h;	// For 3D lines/flats - R3D
	los.zdiff       = abs(los.sightzstart - los.t2head);
	los.strace.dx   = (los.t2x = t2->x) - (los.strace.x = t1->x);
	los.strace.dy   = (los.t2y = t2->y) - (los.strace.y = t1->y);
	los.lowlimit    = MIN_OF(los.sightzstart, los.t2head);
	los.highlimit   = MAX_OF(los.sightzstart, los.t2head);

	allowplanes = t2->player
		&& t1->state
		&& t1->flags & MF_AMBUSH
		&& !t1->subsector->sector->soundtarget
		&& los.sightzstart <= los.t2head
		&& !(t1->flags & (MF_FLOAT | MF_INFLOAT | MF_SKULLFLY | MF_NOGRAVITY));

	ss_validcount++;
	add3DSSEC(t1->subsector);

	if (samesubsector)
	{
		ret = true;	// shortcut for melee situations
	}
	else
	if (   !t1->subsector->flats3DArray
		&& !t2->subsector->flats3DArray
		&& t1->x + t1->radius >= t2->x - t2->radius
		&& t1->x - t1->radius <= t2->x + t2->radius
		&& t1->y + t1->radius >= t2->y - t2->radius
		&& t1->y - t1->radius <= t2->y + t2->radius )
	{
		return true;	// overlap
	}
	else
	{
		los.bottomslope = t2->z - los.sightzstart;
		los.topslope    = los.bottomslope + (t2->dplayer ? viewz - t2->z : t2h);

		if (t1->x > t2->x)
		{
			los.bbox[BOXRIGHT] = t1->x; los.bbox[BOXLEFT] = t2->x;
		}
		else
		{
			los.bbox[BOXRIGHT] = t2->x; los.bbox[BOXLEFT] = t1->x;
		}
		
		if (t1->y > t2->y)
		{
			los.bbox[BOXTOP] = t1->y; los.bbox[BOXBASE] = t2->y;
		}
		else
		{
			los.bbox[BOXTOP] = t2->y; los.bbox[BOXBASE] = t1->y;
		}
		linehitcount++;
		ret = P_CrossBSPNode(numnodes-1);	// the head node is the last node output
	}

	// Check if a 3d plane was hit
	if (ssecscount3D && checkFlats3D() )
	{
		// blocked - so plane3d values will have been set
		ret = false;
	}

	// v220-33 plane blocking
	if (planesseccount && CheckSSECplanes())
	{
		t1->subsector->sector->siflags |= SIF_SOUNDBLOCK;
		ret = false;
	}

	// v220-33 3D door blocking
	// Note: actual position is not needed - just whether LOS is blocked
	if (ret && modelboxcnt)
	{
		// Model line blocking - R3D
		for (int i = 0; i < modelboxcnt; i++)
		{
			mobj_t *mo = MB_GetMBmo(i);

			if (!mo || (mo->flags3 & MF3_INTERREACT) ) continue;

			fixed_t		*mbox = MB_GetBox(i, NULL);
			vertex_t	*v1, *v2;
			divline_t	divl;
			line_t		*line;
			fixed_t		lo, hi, frac, z;
			
			if (   los.bbox[BOXRIGHT] <= mbox[BOXLEFT ]
				|| los.bbox[BOXLEFT ] >= mbox[BOXRIGHT]
				|| los.bbox[BOXTOP  ] <= mbox[BOXBASE ]
				|| los.bbox[BOXBASE ] >= mbox[BOXTOP  ])
			{
				continue;
			}

			for (int k = 0; k < 4; k++)
			{
				line = MB_GetLine(i, k, NULL);

				v1 = line->v1;
				v2 = line->v2;

				// line isn't crossed?
				if (P_DivlineSide(v1->x, v1->y, &los.strace) ==
					P_DivlineSide(v2->x, v2->y, &los.strace))
				{
					continue;
				}
				
				divl.dx = v2->x - (divl.x = v1->x);
				divl.dy = v2->y - (divl.y = v1->y);
				
				// line isn't crossed?
				if (P_DivlineSide(los.strace.x, los.strace.y, &divl) ==
					P_DivlineSide(los.t2x, los.t2y, &divl))
				{
					continue;
				}

				frac = P_InterceptVector(&los.strace, &divl);

				if (frac < 0) return true;	// behind source

				lo = line->frontsector->floorheight;
				hi = lo + line->d3d->totalheight;

				z = los.sightzstart + FixedMul(los.t2head - los.sightzstart, frac);

				if (z >= lo && z <= hi) return false;
			}
		}
	}

	// Mod to when seesound is played - R3D
	if (   ret
		&& t1->flags2 & MF2_NOSEESOUND
		&& t2->player)
	{
		t1->flags2 &= ~MF2_NOSEESOUND;
		P_PlaySeesound(t1);
	}

	return ret;
}

// =======================================================================


// For use with P_CrossSubsector2
static fixed_t		hiBlock;
static fixed_t		loBlock;

static RDBOOL P_CrossSubsector2(subsector_t *ssec)
{
	sector_t		*front, *back;
	vertex_t		*v1, *v2;
	divline_t		divl;
	line_t			*line;
	seg_t			*seg;
	int				i;
	
	for (i = 0, seg = &segs[ssec->firstline]; i < ssec->linecount; i++, seg++)
	{
		if ( (line = seg->linedef) && linehit[line - lines] != linehitcount)
		{
			linehit[line - lines] = linehitcount;
			
			if (   line->lnflags2 & LNF2_UNBLOCKED
				&& line->frontsector->floorheight == line->backsector->floorheight)
			{
				continue;	// v220-16 emulate doom bug
			}
			
			if (line->bbox[BOXLEFT ] > los.bbox[BOXRIGHT] ||
				line->bbox[BOXRIGHT] < los.bbox[BOXLEFT ] ||
				line->bbox[BOXBASE ] > los.bbox[BOXTOP  ] ||
				line->bbox[BOXTOP  ] < los.bbox[BOXBASE ])
			{
				continue;
			}
			
			v1 = line->v1;
			v2 = line->v2;

			// line isn't crossed?
			if (P_DivlineSide(v1->x, v1->y, &los.strace) ==
				P_DivlineSide(v2->x, v2->y, &los.strace))
			{
				continue;
			}
			
			divl.dx = v2->x - (divl.x = v1->x);
			divl.dy = v2->y - (divl.y = v1->y);
			
			// line isn't crossed?
			if (P_DivlineSide(los.strace.x, los.strace.y, &divl) ==
				P_DivlineSide(los.t2x, los.t2y, &divl))
			{
				continue;
			}

			// stop because it is not two sided anyway
			if (!(line->ml_flags & ML_TWOSIDED)) return false;	// stop
			if (line->lnflags2 & LNF2_BLOCKALL) return false;	// stop
			if (line->xtraflags & XT_OCCLUDE) return false;		// v220-33
			if (line->specType >= R3D_40) continue;

			front = line->frontsector;
			if ( (back = line->backsector) == front) continue;
			
			fixed_t ot = front->ceilingheight;
			fixed_t ob = front->floorheight;

			ot = MIN_OF(ot, back->ceilingheight);	
			ob = MAX_OF(ob, back->floorheight  );

			if (ob >= ot) return false;			// stop

			// At last! Don't get thing through a wall with upper/lower textures.
			// E.g hc20 map02 T99 (red key could be obtained by hitting L11094)
			if ((origTEX[line - lines] & 0x11) && ob > hiBlock) return false;
			if ((origTEX[line - lines] & 0x44) && ot < loBlock) return false;
		}
	}

	return true;
}

// P_CrossBSPNode2
static RDBOOL P_CrossBSPNode2(int bspnum)
{
	while (!(bspnum & NFx_SUBSECTOR))
	{
		node_t	*bsp = &nodes[bspnum];
		// Ensure side is either 0 or 1
		int		side = P_DivlineSide(los.strace.x, los.strace.y, (divline_t *)bsp) & 1;

		if (side == P_DivlineSide(los.t2x, los.t2y, (divline_t *) bsp))
		{
			bspnum = bsp->children[side];		// doesn't touch the other side
		}
		else if (!P_CrossBSPNode2(bsp->children[side]))
		{
			return false;						// cross the starting side
		}
		else
		{
			bspnum = bsp->children[side^1];		// cross the ending side
		}
	}

	return P_CrossSubsector2(&subsectors[bspnum & ~NFx_SUBSECTOR]);
}


// v220-16 P_CheckBlocked. Used to establish that a Thing can be picked up
//  Only a sentient player and a special Thing are sent. The distance
//  between them is small so not many subsectors will need to be checked.
//  LNF2_BLOCKALL is only set by a scripted map.
//  t1 must be the special Thing and t2 must be the player
//
RDBOOL P_CheckUnblocked(mobj_t *t1, mobj_t *t2)
{
	sector_t	*ts;

	if (!numnodes) return true;

	loBlock = t1->z;
	hiBlock = loBlock + t1->height;

	ts = t1->subsector->sector;

	// Account for thing being forced into the floor or ceiling
	if (loBlock < ts->floorheight  ) loBlock = ts->floorheight;
	if (hiBlock > ts->ceilingheight) hiBlock = ts->ceilingheight;

	los.strace.dx   = (los.t2x = t2->x) - (los.strace.x = t1->x);
	los.strace.dy   = (los.t2y = t2->y) - (los.strace.y = t1->y);

	if (t1->x > t2->x)
		los.bbox[BOXRIGHT] = t1->x, los.bbox[BOXLEFT] = t2->x;
	else
		los.bbox[BOXRIGHT] = t2->x, los.bbox[BOXLEFT] = t1->x;
	
	if (t1->y > t2->y)
		los.bbox[BOXTOP] = t1->y, los.bbox[BOXBASE] = t2->y;
	else
		los.bbox[BOXTOP] = t2->y, los.bbox[BOXBASE] = t1->y;

	linehitcount++;

	return (P_CrossBSPNode2(numnodes-1));	// the head node is the last node output
}
