//-----------------------------------------------------------------------------
//
//  Copyright (C) 2008 - 2012 G Jackson (uses code similar to p_sight.c)
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_soundsight.c
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
//  Added P_CheckXLines
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
//   P_CheckXLines removed (method changed)
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//   SSF_OCCLUDESEC removed
//
// DESCRIPTION:   Sound source visibility check
//
//-----------------------------------------------------------------------------


// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "de_base.h"
#include "de_play.h"
#include "de_refresh.h"
#include "m_winmem.h"
#include "s_sfx.h"
#include "r3d_edit.h"
#include "r3d_defs.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

int P_DivlineSide(fixed_t x, fixed_t y, const divline_t *node);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static struct slos_t
{
	fixed_t		t2x, t2y;		// start position
	divline_t	strace;			// from t1 to t2
	fixed_t		box[4];
} slos;

// CODE --------------------------------------------------------------------


// Returns true if strace crosses the given subsector successfully.
static RDBOOL px_Subsector(subsector_t *ssec)
{
	sector_t		*front, *back;
	vertex_t		*v1, *v2;
	divline_t		divl;
	line_t			*line;
	seg_t			*seg;
	int				i;
	
	for (i = 0, seg = &segs[ssec->firstline] ; i < ssec->linecount; i++, seg++)
	{
		if ( (line = seg->linedef) && line->validcount != validcount)
		{
			fixed_t ot, ob;
			
			line->validcount = validcount;	

			if (line->bbox[BOXLEFT ] > slos.box[BOXRIGHT] ||
				line->bbox[BOXRIGHT] < slos.box[BOXLEFT ] ||
				line->bbox[BOXBASE ] > slos.box[BOXTOP  ] ||
				line->bbox[BOXTOP  ] < slos.box[BOXBASE ])
			{
				continue;
			}
			
			// Bypass 3D line types
			if (line->specType >= R3D_40)
				continue;

			v1 = line->v1;
			v2 = line->v2;

			// line isn't crossed?
			if (P_DivlineSide(v1->x, v1->y, &slos.strace) ==
				P_DivlineSide(v2->x, v2->y, &slos.strace))
			{
				continue;
			}
			
			divl.dx = v2->x - (divl.x = v1->x);
			divl.dy = v2->y - (divl.y = v1->y);
			
			// line isn't crossed?
			if (P_DivlineSide(slos.strace.x, slos.strace.y, &divl) ==
				P_DivlineSide(slos.t2x, slos.t2y, &divl))
			{
				continue;
			}

			if ( !(line->ml_flags & ML_TWOSIDED)		// single sided
				|| line->lnflags2 & LNF2_OCCLUDE	// sight occlusion
				|| (line->lnflags2 & LNF2_BLOCKALL
				&& !(line->lnflags2 & (LNF2_ALLOWSIGHT|LNF2_LINESOUND))) )
			{
				return false;
			}

			front = line->frontsector;
			back  = line->backsector;

			ot = front->ceilingheight;
			ob = front->floorheight;

			if (front != back)
			{
				ot = MIN_OF(ot, back->ceilingheight);	
				ob = MAX_OF(ob, back->floorheight  );
			}

			// sector totally closed?
			if (ob >= ot) return false;
		}
	}
	return true;
}

static RDBOOL (*bspFunc)(subsector_t *ssec);

static RDBOOL px_BSPNode(int bspnum)
{
	while (!(bspnum & NFx_SUBSECTOR))
	{
		node_t	*bsp = &nodes[bspnum];
		// Ensure side is either 0 or 1
		int		side = P_DivlineSide(slos.strace.x, slos.strace.y, (divline_t *)bsp) & 1;

		if (side == P_DivlineSide(slos.t2x, slos.t2y, (divline_t *) bsp))
		{
			bspnum = bsp->children[side];			// doesn't touch the other side
		}
		else if (!px_BSPNode(bsp->children[side]))
		{
			return false;							// cross the starting side
		}
		else
		{
			bspnum = bsp->children[side^1];			// cross the ending side
		}
	}

	return bspFunc(&subsectors[bspnum & ~NFx_SUBSECTOR]);
}

static void setSlos(mobj_t *t1, mobj_t *t2)
{
	slos.strace.dx = (slos.t2x = t2->x) - (slos.strace.x = t1->x);
	slos.strace.dy = (slos.t2y = t2->y) - (slos.strace.y = t1->y);

	slos.box[BOXRIGHT] = MAX_OF(t1->x, t2->x);
	slos.box[BOXLEFT ] = MIN_OF(t1->x, t2->x);
	slos.box[BOXTOP  ] = MAX_OF(t1->y, t2->y);
	slos.box[BOXBASE ] = MIN_OF(t1->y, t2->y);
}

// Trace from t1 to t2.
// Returns true if t2 is visible to t1
RDBOOL P_SoundSight(mobj_t *t1, mobj_t *t2)
{
	sector_t	*s1, *s2;

	if (!numnodes) return true;

	// If either is unlinked, they can't see each other.
	if (  !sfx_avail	// b211-0 Must be first test
		|| t1->subsector->viewgroup != t2->subsector->viewgroup)
	{
		return false;
	}

	if (t1->flags3 & MF3_GLOBALAMBIENT)
		return true;

	s1 = t1->subsector->sector;
	s2 = t2->subsector->sector;

	if (   s1 == s2
		|| (s1->sightgroup > 0
		&&  s1->sightgroup == s2->sightgroup) )
	{
		return true;
	}

	// Make fake floors and ceilings block monster view
	// v220-3 Allow sight if water is translucent
	if (s1->heightsec >= 0 && !(t1->subsector->ssflags & SSF_FAKETRANS))
	{
		sector_t *sec = &sectors[s1->heightsec];

		if (   (t1->z + t1->height < sec->floorheight && t2->z >= sec->floorheight)
			|| (t1->z >= sec->ceilingheight && t2->z + t1->height < sec->ceilingheight))
		{
			return false;
		}
	}

	if (s2->heightsec >= 0 && !(t2->subsector->ssflags & SSF_FAKETRANS))
	{
		sector_t *sec = &sectors[s2->heightsec];

		if (   (t2->z + t2->height < sec->floorheight && t1->z >= sec->floorheight)
			|| (t2->z >= sec->ceilingheight && t1->z + t2->height < sec->ceilingheight))
		{
			return false;
		}
	}
	
	setSlos(t1, t2);

	validcount++;
	bspFunc = px_Subsector;

	return px_BSPNode(numnodes - 1);
}
