//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_OGL_tex.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified for R3D, moved to main engine
//
// Based on: Doomsday v1.7.8 drOpenGL.dll code
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>

#include "m_winmem.h"

#include "dd_types.h"

#include "r3d_OGLfunc.h"
#include "m_args.h"
#include "dglib.h"

#include "dd_share.h"
#include "con_console.h"
#include "gl_tex.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

#define MAXTEXINDEX 4096

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void Con_Error (char *error, ...);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//===========================================================================
// GL_DeleteAllTextures - R3D
//	Only called on exit. Deletes all OGL textures
//===========================================================================
void GL_DeleteAllTextures()
{
	int		i, texIndexCount = 0;
	uint	size = 0;

	GL_DeleteFontLists();

	for (i = 1; i <= MAXTEXINDEX; i++)
	{
		if (glIsTexture(i))
		{
			DGLuint dummy = i;
			glDeleteTextures(1, &dummy);
			texIndexCount++;
			size += skin_inf.textureidx[i];
		}
	}

	pr_OFinsert(" # %i gl textures deleted [Total size %.2fMB]\n",
		texIndexCount, float(size)/1000000.f);
}

//===========================================================================
// GL_DeleteTexIndex - R3D
//	Delete an existing texture index
//===========================================================================
void GL_DeleteTexIndex(DGLuint *texIndex, char *type)
{
	DGLuint	index = *texIndex;

	if (!index) return;

	if (index >= MAXTEXINDEX)
		Con_Error("Invalid index %i called from %s", index, type);

	if (glIsTexture(index))
	{
		glDeleteTextures(1, texIndex);

		if (skin_inf.textype[index])
		{
			skin_inf.skintexmemoryused -= skin_inf.textureidx[index];
			skin_inf.skintexcount--;
		}
		else
		{
			skin_inf.graphicstexmemoryused -= skin_inf.textureidx[index];
			skin_inf.graphicstexcount--;
		}
	}

	skin_inf.textureidx[index] = 0;
	skin_inf.textype[index] = 0;

	*texIndex = 0;
}

//===========================================================================
// GL_NewTexture - R3D
//	Create a new texture index
//===========================================================================
void GL_NewTexture(DGLuint *texIndex)
{
	// Generate a new texture name and bind it.
	glGenTextures(1, texIndex);

	if ( !DGLuint(*texIndex) )
		Con_ErrorD("Failed to generate texname");

	if (DGLuint(*texIndex) >= MAXTEXINDEX)
		Con_ErrorD("Max textures allowed exceeded");

	curtex = DGLuint(*texIndex);

	glBindTexture(GL_TEXTURE_2D, *texIndex);
}

//===========================================================================
// Power2
//===========================================================================
static int Power2(int num)
{
	int cumul;
	for(cumul = 1; num > cumul; cumul <<= 1);
	return cumul;
}

//===========================================================================
// DG_TexImage
//	'width' and 'height' must be powers of two.
//===========================================================================
RDBOOL GL_TexImage(int format, int width, int height, int genMips, void *data)
{
	int		mipLevel = 0;
	byte	*buffer = (byte*)data, *pixel, *in;

	if (!data) return false;

	// Check that the texture dimensions are valid.
	if (width != Power2(width) || height != Power2(height))
		return false;

	if (width > maxTexSize || height > maxTexSize)
		return false;

	// Automatic mipmap generation?
	if (glx.extGenMip && genMips)
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);

	// Form the texture
	{
		int		alphachannel = (format == GL_RGBA || format == GL_LUMINANCE_ALPHA);
		int		colorComps   = (format == GL_LUMINANCE ? 1 : alphachannel ? 4 : 3);
		int		numPixels    = width * height;
		int		putFormat    = GL_RGBA;
		int		i;

		// Convert to either RGB or RGBA, if necessary.
		if (format == GL_RGBA)
		{
			// do nothing
		}
		// An Nvidia driver bug? Very small RGBA textures don't load properly.
		else if (format == GL_RGB && width > 2 && height > 2)
		{
			putFormat = GL_RGB;
		}
		else // Needs converting.
		{
			buffer = (byte*)M_Calloc(numPixels * 4, 266);
			// buffer and data addresses must not be changed
			pixel  = buffer;
			in     = (byte*)data;
			
			switch(format)
			{
			case GL_RGB:
				for(i = 0; i < numPixels; i++, pixel += 4)
				{
					pixel[CR] = *in++;
					pixel[CG] = *in++;
					pixel[CB] = *in++;
					pixel[CA] = 255;
				}

				break;

			case GL_LUMINANCE:
				putFormat = GL_RGB;

				for(i = 0; i < numPixels; i++, pixel += 3)
					pixel[CR] = pixel[CG] = pixel[CB] = in[i];

				break;

			case GL_LUMINANCE_ALPHA:
				for(i = 0; i < numPixels; i++, pixel += 4)
				{
					pixel[CR] = pixel[CG] = pixel[CB] = in[i];
					pixel[CA] = in[numPixels+i];
				}

				break;

			default:
				M_Free(buffer);
				M_Free(data);
				Con_Error("LoadTexture: Unknown format %x", format);
				break;
			}
		}

		if (!glx.extGenMip && genMips)	// build mipmap
		{
			gluBuild2DMipmaps(GL_TEXTURE_2D, colorComps, 
				width, height, putFormat, GL_UNSIGNED_BYTE, buffer);
		}
		else		
		{
			glTexImage2D(GL_TEXTURE_2D, mipLevel, colorComps, 
				width, height, 0, putFormat, GL_UNSIGNED_BYTE, buffer);
		}

		if (data != buffer) M_Free(buffer);
	}

	// Should anisotropic filtering be used?
	if (anisotropic > 1)
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, (float)anisotropic);

	return true;
}

//===========================================================================
// GL_Bind
//===========================================================================
void GL_Bind(DGLuint texture)
{
	// Does it need binding
	if (texture != curtex)
	{
		glBindTexture(GL_TEXTURE_2D, texture);
		curtex = texture;
	}
}
