//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 id Software, Inc.
// Copyright (C) 2012 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_iwadtext.cpp	First written 2012 (v220-16) gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
// Revision 2.2.0 build 26 (v220-26) 05/Jul/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION:	Validates the use of external hi-res textures
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <direct.h>
#include <stdlib.h>
#include <stdio.h>

#include "de_base.h"
#include "m_misc.h"
#include "r3d_MAINspec.h"

// MACROS ------------------------------------------------------------------

#define	HIRES_LIST_CUSTOM		0
#define	HIRES_LIST_TX_SHARED	1
#define	HIRES_LIST_TX_SPECIFIC	2
#define	HIRES_LIST_PT_SHARED	3
#define	HIRES_LIST_PT_SPECIFIC	4
#define	HIRES_LIST_HUD			5

// TYPES -------------------------------------------------------------------

typedef struct
{
	short	originx;
	short	originy;
	short	patch;
	short	stepdir;
	short	colormap;
} mappatch_t;

typedef struct
{
	char		name[8];
	short		masked;	
	byte		ScaleX;
	byte		ScaleY;
	short		width;
	short		height;
	byte		columndirectory[4];
	short		patchcount;
	mappatch_t	patches[1];
} maptexture_t;

typedef struct
{
	char	name[10];		// for switch changing, etc
	short	width;
	short	height;
	short	patchcount;

} IWADtexture_t;

typedef struct {
	char	*hireslist;
	int		num;
	int		maxnum;	// b211-0
} hires_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	p2swd(char *msg, ...);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern byte	*palbuffer;
extern byte	*iwadpalbuffer;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// hi-res tex/patch paths
int		hiresPatches;
int		hiresTextures;
char	*customTexPath;
char	*defhudstr;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int	nohirestexcount;
static char	testpath[512];

static int				numIWADtextures;
static IWADtexture_t	**texIWADArray;

static char	*externPatchPath[2];
static char	*externTexPath[2];

static hires_t wall_list_custom;
static hires_t flat_list_custom;
static hires_t hud_list_external;
static hires_t wall_list_ext1;
static hires_t wall_list_ext2;
static hires_t flat_list_ext1;
static hires_t flat_list_ext2;
static hires_t patch_list_ext1;
static hires_t patch_list_ext2;

// v220-21 palette change indices
extern byte		palptr[256];

// CODE --------------------------------------------------------------------

static RDBOOL patchCountCheck(char *name, int patchcount)
{
	int i;
  
	for (i = 0; i < numIWADtextures; i++)
	{
		if (!strnicmp(name, texIWADArray[i]->name, 8))
		{
			return texIWADArray[i]->patchcount == patchcount;
		}
	}

	return false;
}

static void addToList(char *gname, int idx)
{
	hires_t *type;
	hires_t *check = NULL;
	char	*nameptr;
	char	fname[MAXFNAME];
	RDBOOL	isflat;

	_splitpath(gname, NULL, NULL, fname, NULL);

	isflat = !strnicmp(fname, "FLAT-", 5);

	nameptr = &fname[(isflat ? 5 : 0)];

	if (!(*nameptr) || strlen(nameptr) > 8) return;

	// v220-26 revised texture patches check
	if (!R3D_ScriptedMap && !isflat && idx == HIRES_LIST_TX_SHARED)
	{
		int id = R_CheckTextureNumForName(nameptr);

		if (id >= 0)
		{
			int i, n = texArray[id]->patchcount;

			// 1) Check not repatched.
			//    Strictly speaking patch names should be checked
			//    but in general, where an author has repatched a
			//    texture, the count is usually different e.g.
			//    Sw1brcom in vrack3 (test 2 would also detect this)
			//    zzwolf10 in mordeth map06
			//    Sw1starg in sid map01 (amongst many others)
			if (!patchCountCheck(nameptr, n))
			{
				nohirestexcount++; return;
			}

			// 2) search to find if any of the texture's patches
			//    have a specific pwad texture assigned.
			for (i = 0; i < n; i++)
			{
				int  lp = texArray[id]->patches[i].patch;

				if (lp >= 0 && lp < numlumps)
				{
					char *pname = lumpinfo[lp].name;

					if (W_CheckPWADNumForName(pname) >= 0)
					{
						nohirestexcount++; return;
					}
				}
			}
		}

		goto endcheck;
	}

	// Don't override custom pwad textures with external hires textures
	switch(idx)
	{
	case HIRES_LIST_CUSTOM:	break;	// always loaded
	case HIRES_LIST_TX_SHARED:
	case HIRES_LIST_TX_SPECIFIC:
		if (   ( isflat && W_CheckPWADNumForName(nameptr) >= 0)
			|| (!isflat && W_CheckPWADtexture(nameptr)) )
		{
			nohirestexcount++; return;
		}
		break;
	case HIRES_LIST_PT_SHARED:
	case HIRES_LIST_PT_SPECIFIC:
	case HIRES_LIST_HUD:
		if (W_CheckPWADNumForName(nameptr) >= 0)
		{
			nohirestexcount++; return;
		}
		break;
	default: return;
	}
endcheck:;
	switch(idx)
	{
	case HIRES_LIST_CUSTOM:
		type = (isflat ? &flat_list_custom : &wall_list_custom);
		break;
	case HIRES_LIST_TX_SHARED:
		type = (isflat ? &flat_list_ext1 : &wall_list_ext1);
		check = (isflat ? &flat_list_ext2 : &wall_list_ext2);
		break;
	case HIRES_LIST_TX_SPECIFIC:
		type = (isflat ? &flat_list_ext2 : &wall_list_ext2);
		break;
	case HIRES_LIST_PT_SHARED:
		type = &patch_list_ext1;
		check = &patch_list_ext2;
		break;
	case HIRES_LIST_PT_SPECIFIC:
		type = &patch_list_ext2;
		break;
	case HIRES_LIST_HUD:
		type = &hud_list_external;
		break;
	default: return;
	}

	// Check duplicate in main list against secondary list
	// It is not actually necessary to stop the file being
	// put in the primary list other than marginally less
	// memory will be used (and fewer entries to search).
	if (check)
	{
		int i;

		for (i = 0; i < check->num; i++)
		{
			if (!strnicmp(check->hireslist + i*8, nameptr, 8))
			{
				return;
			}
		}
	}

	if (type->num >= type->maxnum)
	{
		type->maxnum += 16;
		type->hireslist = (char*)W_ReallocCLR(type->hireslist, type->maxnum * 8 + 2, STATICID, 201);
	}

	strncpy(type->hireslist + type->num * 8, nameptr, 8);
	type->num++;
}

// v220-33 If hires textures are found then returns the number found. This is used to indicate
// that all hires directories should be scanned even if hiresTextures or hiresPatches are unset
static int InitHiResList(char *path, int idx)
{
	// v220-16 , "*.tga", "*.pcx" removed
	char	*formats[] = { "*.jpg", "*.png", NULL };
	char	currentpath[MAXPATH];
	struct	_finddata_t g_file;
	long	hFile;
	int		count = 0;

	if (!path) return 0;

	pr_OFinsert("HI-RES PATH ADDED:   %s\n", path);

	_getcwd(currentpath, MAXPATH);

	if (_chdir(path) == 0)
	{
		int  i;

		for (i = 0; formats[i]; i++)
		{
			if ( (hFile = _findfirst(formats[i], &g_file)) != -1L )
			{
				do 
				{ addToList(g_file.name, idx); count++;
				}
				while(!_findnext(hFile, &g_file));

				_findclose( hFile );
			}
		}
	}

	_chdir(currentpath);

	return count;
}

void IW_SetHiresTexPaths()
{
	RDBOOL	error;
	RDBOOL	hiresexists = false;
	char	userCustomPath[MAXFNAME];
	char	txpath[256];
	int		iwadidx = W_IWADtype();

	if (strstrCI(analysename, "freedoom")) return;

	hiresTextures = BDMFX && !ArgExists("-nohightex");
	hiresPatches  = BDMFX && !ArgExists("-nohirespatches");

	if (!BDMFX && !ArgExists("-nohightex"))
		pr_OFdirect("High-res textures are disabled in basic doom mode\n");
	if (!BDMFX && !ArgExists("-nohirespatches"))
		pr_OFdirect("High-res patches are disabled in basic doom mode\n");
	if (!hiresPatches && ge_inf.switch_flags & SWF_AP)
		Con_Error("PROBLEM: Hi-res HUD patches must be selected for use with alternative weapons");

	if (!hiresTextures && !hiresPatches) return;

	userCustomPath[0] = 0;

	// v220-31 Check for user defined custom.hires directory name
	if (ArgCheckWith("-uchd", 1))
	{
		char *ptr = ArgNext();

		// Custom.hires directory to search
		if (ptr && stricmp(ptr, analysename) && strlen(ptr) < MAXFNAME)
			strcpy(userCustomPath, ptr);
	}

	// v220-31 add support for -uchd switch
	if (haspwads && analysename)
	{
		char *useptr = NULL;

		sprintf(testpath, "%sData\\Custom.hires\\%s\\", ddRuntimeDir.path, analysename);

		RDBOOL valid = (_access(testpath, 4)) != -1L;

		// check for custom textures. If any exist these are always loaded
		if (valid)
		{
			useptr = testpath;

			if (userCustomPath[0])
			{
				pr_OFdirect("hires texture path using -uchd rejected as custom path already exists\n");
				userCustomPath[0] = 0;
			}
		}
		else if (userCustomPath[0])
		{
			sprintf(testpath, "%sData\\Custom.hires\\%s\\", ddRuntimeDir.path, userCustomPath);

			if ( (valid = (_access(testpath, 4)) != -1L) )
			{
				useptr = testpath;
				pr_OFdirect("hires texture path using -uchd accepted\n");
			}
		}

		if (useptr)
		{
			char	skypath[300];

			customTexPath = (char*)W_Calloc(strlen(testpath) + 1, STATICID, 0);
			strcpy(customTexPath, testpath);
			hiresexists = InitHiResList(customTexPath, HIRES_LIST_CUSTOM);

			// v220-31 autodetect if a custom skybox exists but
			// v220-35 only if -usb is present
			if (ArgExists("-usb") && !(ge_inf.use_sky_box & GE_SKYBOXCHECK) )
			{
				for (int i = 1; i < 4; i++)
				{
					sprintf(skypath, "%ssky%i\\", testpath, i);
					if ((_access(skypath, 4)) != -1L) ge_inf.use_sky_box |= GE_SKYBOXCHECK;
				}
			}
		}
	}
	
	if (hiresexists)
	{
		hiresTextures = true;	// Force settings on
		if (!CFX) hiresPatches = true;	// If not BDM force settings on
	}
	else
	{
		if (!hiresTextures && !hiresPatches) return;
	}

	if (ArgCheckWith("-td", 1))
	{
		strcpy(txpath, ArgNext());
		Dir_ValidDir(txpath);
	}
	else
	{
		sprintf(txpath, "%sData\\", ddRuntimeDir.path);
	}

	error = false;

	if (hiresTextures)
	{
		sprintf(testpath, "%sTextures\\", txpath);

		if (!(error = (_access(testpath, 4)) == -1L))
		{
			externTexPath[0] = (char*)W_Calloc(strlen(testpath) + 1, STATICID, 0);
			strcpy(externTexPath[0], testpath);

			if (iwadidx == 1 || iwadidx == 3 || iwadidx == 4)
			{
				switch(iwadidx)
				{
				case 1: strcat(testpath, "doom1\\"     ); break;
				case 3: strcat(testpath, "doom2-plut\\"); break;
				case 4: strcat(testpath, "doom2-tnt\\" ); break;
				}

				if (!(error = (_access(testpath, 4)) == -1L))
				{
					externTexPath[1] = (char*)W_Calloc(strlen(testpath) + 1, STATICID, 0);
					strcpy(externTexPath[1], testpath);
				}
			}

			if (!error)
			{
				//  Add specific first
				InitHiResList(externTexPath[1], HIRES_LIST_TX_SPECIFIC);
				InitHiResList(externTexPath[0], HIRES_LIST_TX_SHARED);
			}
		}
	}

	error = false;
	
	if (hiresPatches)
	{
		sprintf(testpath, "%sPatches\\Hud\\", txpath);

		if ( (_access(testpath, 4)) != -1)
		{
			defhudstr = (char*)W_Calloc(strlen(testpath) + 1, STATICID, 0);
			strcpy(defhudstr, testpath);
			InitHiResList(defhudstr, HIRES_LIST_HUD);
		}

		sprintf(testpath, "%sPatches\\", txpath);

		if (!(error = (_access(testpath, 4)) == -1L))
		{
			externPatchPath[0] = (char*)W_Calloc(strlen(testpath) + 1, STATICID, 0);
			strcpy(externPatchPath[0], testpath);

			if (iwadidx >= 1 && iwadidx <= 4)
			{
				switch(iwadidx)
				{
				case 1: strcat(testpath, "doom\\"    ); break;
				case 2: strcat(testpath, "doom2\\"   ); break;
				case 3: strcat(testpath, "plutonia\\"); break;
				case 4: strcat(testpath, "tnt\\"     ); break;
				}

				if (!(error = (_access(testpath, 4)) == -1L))
				{
					externPatchPath[1] = (char*)W_Calloc(strlen(testpath) + 1, STATICID, 0);
					strcpy(externPatchPath[1], testpath);
				}
			}

			if (!error)
			{
				//  Add specific first
				InitHiResList(externPatchPath[1], HIRES_LIST_PT_SPECIFIC);
				InitHiResList(externPatchPath[0], HIRES_LIST_PT_SHARED);
			}
		}
	}

	if (nohirestexcount)
		pr_OFinsert("Unique textures: %i pwad textures/patches have external hires replacements suppressed.\n",
		nohirestexcount);
}

static void loadPatch(byte *buffer, int w, int h, patch_t *patch)
{
	typedef struct { byte topdelta, length; } column_t;

	column_t	*column;
	int			top;
	int			c, i, x, y;

	byte		*source, *dest;
	byte		*destTop = buffer;

	// ---------------------------

	for (i = x = 0; i < patch->width; i++, destTop++, x++)
	{
		column = (column_t*)((byte*)patch + patch->columnofs[i]);
		top = -1;
		
		// Step through the posts in a column
		while (column->topdelta != 0xff)
		{
			source = (byte*)column + 3;
			
			if( x < 0 || x >= w ) break;
			
			if (column->topdelta <= top)
				top += column->topdelta;
			else
				top = column->topdelta;
			
			if ( (c = column->length) > 0)     
			{
				y = top; dest = destTop + y * w;
				
				while (c--)
				{
					if (y >= 0 && y < h) *dest = *source++;	// get index into palette
										
					dest += w; y++;	// One row down.
				}
			}
			column = (column_t*)((byte*)column + column->length + 4);
		}
	}
}

RDBOOL IW_CheckSpritePalette(int lump)
{
	int		w, h;
	RDBOOL	changed = false;

	if (   iwadpalbuffer
		&& DoomLumpIdentify(lump, &w, &h, DLI_GRAPHIC) == LMP_PATCH
		&& w
		&& h)
	{
		patch_t *patch = (patch_t*)M_CacheLumpNum(lump, NULL, true);
		int		k, sz = w * h;
		byte	*buf = (byte*)M_Calloc(sz, 0);

		loadPatch(buf, w, h, patch);

		for (k = 0; k < sz; k++)
		{
			if (palptr[buf[k]]) { changed = true; break; }
		}
		
		free(patch);
		free(buf);
	}

	return changed;
}

// External or custom hi-res texture/flat names are stored in pre-set
// hi-res lists. They are differentiated by having "flat-" preceding
// the name. If flat == true then checks for a custom hi-res flat.
// The custom list must be searched first as these are always loaded
// whereas the other lists have been filtered to remove pwad custom
// flats/textures.

// If found in list then ptr to the path is set
static char	*hirespathptr;

static RDBOOL checkHiRes(char *name, RDBOOL isflat, int idx)
{
	hires_t *type;
	int		w, h, sz;
	int		i, k, dtyp, lump;

	switch(idx)
	{
	case HIRES_LIST_CUSTOM:
		type = (isflat ? &flat_list_custom : &wall_list_custom);
		hirespathptr = customTexPath;
		break;
	case HIRES_LIST_TX_SHARED:
		type = (isflat ? &flat_list_ext1 : &wall_list_ext1);
		hirespathptr = externTexPath[0];
		break;
	case HIRES_LIST_TX_SPECIFIC:
		type = (isflat ? &flat_list_ext2 : &wall_list_ext2);
		hirespathptr = externTexPath[1];
		break;
	case HIRES_LIST_PT_SHARED:
		type = &patch_list_ext1;
		hirespathptr = externPatchPath[0];
		break;
	case HIRES_LIST_PT_SPECIFIC:
		type = &patch_list_ext2;
		hirespathptr = externPatchPath[1];
		break;
	case HIRES_LIST_HUD:
		type = &hud_list_external;
		hirespathptr = defhudstr;
		break;
	default: return false;
	}

	for (i = 0; i < type->num; i++)
	{
		if (!strnicmp(type->hireslist + i*8, name, 8))
		{
			// v220-26 eliminate certain types from the palette check
			if (   !strnicmp("TITLEPIC", name, 8)
				|| !strnicmp("INTERPIC", name, 8)
				|| !strnicmp("CREDIT", name, 8)
				|| !strnicmp("HELP", name, 8))
			{
				return true;
			}

			// v220-21 palette check
			if (   iwadpalbuffer
				&& (lump = W_CheckNumForName(name)) >= 0
				&& lumpinfo[lump].handle == IWADhandle
				&& ((dtyp = DoomLumpIdentify(lump, &w, &h, DLI_GRAPHIC)) == LMP_FLAT
				||   dtyp == LMP_PATCH)
				&& (sz = w * h))
			{
				// check to see if any of the image's pixels point to a
				// palette value changed from the original doom palette

				byte		*buf = (byte*)M_Calloc(sz, 0);
				patch_t		*patch = NULL;
				RDBOOL		changed = false;
				RDBOOL		valid = false;

				if (dtyp == LMP_FLAT)
				{
					lumpinfo_t	*l = lumpinfo + lump;
					
					lseek(l->handle, l->position, SEEK_SET);
					valid = read(l->handle, (void*)buf, (uint)sz) == sz;
				}
				else
				{
					patch = (patch_t*)M_CacheLumpNum(lump, NULL, true);
					loadPatch(buf, w, h, patch);
					valid= true;
				}

				if (valid)
				{
					for (k = 0; k < sz; k++)
					{
						if (palptr[buf[k]]) { changed = true; break; }
					}
				}

				free(buf);
				free(patch);
				return !changed;
			}

			return true;
		}
	}

	return false;
}

static RDBOOL getHiresFile(RDBOOL flat, char *name)
{
	char	*formats[] = { "jpg", "png", NULL };
	int		i;

	for (i = 0; formats[i]; i++)
	{
		if (flat)
			sprintf(testpath, "%sflat-%s.%s", hirespathptr, name, formats[i]);
		else
			sprintf(testpath, "%s%s.%s", hirespathptr, name, formats[i]);

		if ( (_access(testpath, 4)) != -1L) return true;
	}

	return false;
}

// Public interface --------------------------------

char *IW_HiresTexFile(RDBOOL flat, char *name)
{
	// Order must not be changed
	if (   checkHiRes(name, flat, HIRES_LIST_CUSTOM)
		|| checkHiRes(name, flat, HIRES_LIST_TX_SPECIFIC)
		|| checkHiRes(name, flat, HIRES_LIST_TX_SHARED)  )
	{
		if (getHiresFile(flat, name)) return testpath;
	}

	return NULL;
}

char *IW_HiresPatchFile(char *name)
{
	// Order must not be changed
	if (   checkHiRes(name, 0, HIRES_LIST_PT_SPECIFIC)
		|| checkHiRes(name, 0, HIRES_LIST_PT_SHARED  ) )
	{
		if (getHiresFile(0, name)) return testpath;
	}

	return NULL;
}

// =================================================================

// NOTE: called after loading an iwad and before any pwads are loaded
void IW_InitIWADTextures()
{
	maptexture_t	*mtexture;
	IWADtexture_t	*texture;

	int			i ;
	int			*maptex, *maptex1;
	char		name[10];	// Needs 9 but made 10 to align
	int			offset;
	uint		maxoff;
	int			numtextures1;
	int			numtextures2 = 0;
	int			*directory;

	int			*maptex2 = NULL;
	uint		maxoff2 = 0;

	float		dotinc, dot = 0.f;

	//---------------------------------

	// Load the map texture definitions from TEXTURE1/2.
	i = W_GetNumForName ("TEXTURE1");
	maptex = maptex1 = (int*)M_CacheLumpNum(i, &maxoff, false);
	numtextures1	 = *maptex;
	directory		 = maptex + 1;

	// Only needed by doom.wad
	if ((i = W_CheckNumForName ("TEXTURE2")) != -1)
	{
		maptex2 = (int*)M_CacheLumpNum (i, &maxoff2, true);
		numtextures2 = *maptex2;
	}

	numIWADtextures = numtextures1 + numtextures2;

	// Assign memory including any specials found.
	texIWADArray  = (IWADtexture_t**)W_Calloc((numIWADtextures)*4, STATICID, 110);

	p2swd("Loading IWAD textures: %i textures.\n", numIWADtextures);

	dotinc = numIWADtextures/72.f;

	for (i = 0 ; i < numIWADtextures ; i++, directory++)
	{
		if (i > (int)dot)
		{
			p2swd(".");
			dot += dotinc;
		}

		if (i == numtextures1)
		{	// Start looking in second texture file.
			maptex = maptex2;
			maxoff = maxoff2;
			directory = maptex+1;
		}

		offset = *directory;

		if ((uint)offset > maxoff)
			Con_ErrorD("R_InitTextures: bad IWAD texture directory");

		mtexture = (maptexture_t*) ( (byte*)maptex + offset);

		if (mtexture->patchcount < 1)
		{
			strncpy (name, mtexture->name, 8);
			name[8] = 0;
			Con_Error ("R_InitTextures: invalid IWAD patchcount with \"%s\"", name);
		}

		texture = texIWADArray[i] = (IWADtexture_t*)W_Calloc(sizeof(IWADtexture_t) 
			+ sizeof(texpatch_t)*(mtexture->patchcount-1), STATICID, 123);

		memcpy (texture->name, mtexture->name, 8);
		texture->width		= mtexture->width;
		texture->height		= mtexture->height;
		texture->patchcount	= mtexture->patchcount;
	}

	p2swd("\n");

	M_Free(maptex1);
	M_Free(maptex2);
}

// Unless a wall texture is explicitly required then whether it is
// valid is checked by comparing dimensions against the original,
// unadulterated, iwad values. This may seem odd but one map known
// has juggled some of the references around in an attempt to try
// to centre a wall switch. Ho humm....
RDBOOL IW_CheckIWADTexDimensions(char *name, short width, short height)
{
	int i;
  
	if (checkHiRes(name, 0, 0)) return true;

	for (i = 0; i < numIWADtextures; i++)
	{
		if (!strnicmp(name, texIWADArray[i]->name, 8))
		{
			return width  == texIWADArray[i]->width
				&& height == texIWADArray[i]->height;
		}
	}

	return false;
}
