//-----------------------------------------------------------------------------
//
// Copyright (C) 2012 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_misContained.cpp	First written 2012 (v220-16) gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//    MC_CheckLoopsContainingSSlines removed
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
//
// DESCRIPTION: nested missing texture loop analysis
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <memory.h>

#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "m_misc.h"
#include "p_maputil.h"
#include "dd_wad.h"

// MACROS ------------------------------------------------------------------

#define SEC_MARKER	0x80000000

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void traceLoFromMarked(line_t *ld, int fht, int *c);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

static RDBOOL checkSec(int secnum)
{
	int i;

	for (i = 0; i < osx.scount; i++)
	{
		if (countsec[i] == secnum) return true;
	}

	return false;
}

static void markOuterSec(int secnum)
{
	int i;

	for (i = 0; i < osx.scount; i++)
	{
		if (countsec[i] == secnum)
		{
			countsec[i] |= SEC_MARKER;
			return;
		}
	}
}

static void addVtxSectors(vertex_t *v)
{
	vertexowner_t	*vo = &vertexowners[v - vertexes];
	int				i;

	for (i = 0; i < vo->num; i++)
	{
		if (!checkSec(vo->list[i]))
			countsec[osx.scount++] = vo->list[i];
	}
}

static void setLineType(line_t *ld, byte type, int num)
{
	lineflags[ld - lines] = 0;	// could have already been set
	Settype(ld, type, num);
	SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_FIXABS, num);
}

static void SetLinesLFG(sector_t *s, int num)
{
	line_t	*ld;
	int		i;

	for (i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (lineflags[ld - lines] & LFG_FIXABS) continue;

		if (!origTEX[ld - lines])
			setLineType(ld, R3D_3F, num);
		else
			setLineType(ld, R3D_36, num);

		SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_FIXABS, num);
	}

	Setsecdone(s, num);
}

static RDBOOL secReferenced(sector_t *s)
{
	sector_t	*ch;
	line_t		*ld;
	int			i, secnum;
	int			lf, mask = LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE;

	osx.scount = osx.lcount = 0;

	for (i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (!ld->backsector) return false;
		if (ld->frontsector == ld->backsector) continue;

		lf = lineflags[ld - lines];

		if (  ( lf && !((lf & mask) == mask) )
			|| (origTEX[ld - lines] && !ld->mtflags)
			|| !Vtxs2(ld)
			|| !(ld->lnflags2 & LNF2_OWNEQUALTOTWO))
		{
			return false;
		}

		ch = ld->frontsector == s ? ld->backsector : ld->frontsector;
		
		secnum = ch - sectors;

		if (!checkSec(secnum))
		{
			if (   SecCtrl(secnum)
				|| ch->specFlags & 0x30000
				|| ch->heightsec >= 0)
			{
				return false;
			}

			countsec[osx.scount++] = secnum;
		}

		linearray[osx.lcount++] = ld - lines;
	}

	return true;
}

static void setLowHeightSectors(int secnum)
{
	subsector_t *ss;
	sector_t	*s = &sectors[secnum], *rf, *ch;
	line_t		*ld;
	int			c, i, k;

	if (s->ceilingheight - s->floorheight > FRACUNIT) return;

	for (rf = NULL, c = i = 0; i < s->linecount; i++)
	{
		ld = s->lines[i];

		ch = ld->frontsector == s ? ld->backsector : ld->frontsector;

		if (!rf) { rf = ch; if (rf->specFlags & SF_MOVSECTOR) return; }
		else if (rf != ch
			&& (rf->floorheight != ch->floorheight
			||  rf->ceilingheight != ch->ceilingheight))
		{
			return;
		}

		linearray[c++] = ld - lines;
	}

	if (!rf) return;

	ss_validcount++;

	for (i = 0; i < c; i++)
	{
		ssline_t *sl = &sslines[linearray[i]];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->sector == s && ss->ssvalidcount != ss_validcount)
			{
				ss->ssvalidcount = ss_validcount;
				SetOverride(ss, rf - sectors, 49001);
			}
		}

		setLineType(&lines[linearray[i]], R3D_3F, 49001);
	}
}

// v220-16 Check for structures that generate false floors and ceilings
void MC_CheckDoubleLowMove()
{
	sector_t	*s;
	line_t		*ld;
	short		fpic, cpic;
	fixed_t		h;
	int			c, i, k, secnum;
	int			*seclist;

	// ------------------------------

	for (i = 0; i < numsectors; i++)
	{
		if (   (sectors[i].specFlags & 0x10018) == 0x10018
			&& sectors[i].linecount > 6
			&& !secdone[i]
			&& (osx.tag = sectors[i].tag)
			&& sectors[i].heightsec < 0)
		{
			int		fheight = DDMININT;

			if (!secReferenced(&sectors[i]) || osx.scount != 2 || osx.lcount < 6 ) continue;

			// Two sectors means we expect this thing to have two loops
			if ( !U_checkDoubleLoop(&sectors[i], 0) ) continue;
			if (osx.inner->tag != osx.tag)
			{
				M_Free(osx.intList); osx.intList = NULL;
				continue;
			}

			for (k = 0; k < osx.outer->linecount; k++)
			{
				ld = osx.outer->lines[k];
				
				addVtxSectors(ld->v1);
				addVtxSectors(ld->v2);
			}

			osx.listcnt = 0;
			osx.intList = (int*)malloc(sizeof(int) * MAX_LISTCOUNT);

			for (k = 0; k < osx.scount; k++)
			{
				secnum = countsec[k];

				// First mark heightless sectors outside
				// and get fheight
				if (!M_FitFPBox(sectors[secnum].sibounds, osx.outer->sibounds))
				{
					if ((s = &sectors[secnum])->specFlags & 0x10000) continue;

					// Mark any near heightless sectors
					h = s->ceilingheight - s->floorheight;

					if (h >= 0 && h <= 2*FRACUNIT)
					{
						if (osx.listcnt < MAX_LISTCOUNT)
							osx.intList[osx.listcnt++] = secnum;
					}
					else
					{
						if (s->floorheight > fheight) fheight = s->floorheight;
					}
				}
			}

			// Form a list of outer sectors and set fpic and cpic
			// cpic can only be obtained from an outer sector
			seclist = (int*)malloc(osx.scount * sizeof(int));

			fpic = sectors[i].floorpic;

			for (c = k = cpic = 0; k < osx.scount; k++)
			{
				secnum = countsec[k];

				if (!M_FitFPBox(sectors[secnum].sibounds, osx.outer->sibounds))
				{
					seclist[c++] = secnum;
				}
				else
				{
					s = &sectors[secnum];
					h = s->ceilingheight - s->floorheight;

					if (    s->ceilingheight == fheight
						&& (s->specFlags & 0x1000C) == 0xC)
					{
						cpic = s->ceilingpic;
					}
				}
			}

			// Mark sectors outside the box
			for (k = 0; k < c; k++) markOuterSec(seclist[k]);
 
			if (cpic && fpic)
			{
				sector_t	*rf = NULL, *ch;
				int			m;

				// Check unequal sectors are similar
				for (rf = NULL, k = 0; k < osx.listcnt; k++)
				{
					s = &sectors[osx.intList[k]];

					for ( m = 0; m < s->linecount; m++)
					{
						ld = s->lines[m];

						ch = ld->frontsector == s ? ld->backsector : ld->frontsector;

						if (!ch) { rf = NULL; break; }
						if (!rf) rf = ch;
						else if (rf != ch
							&& (rf->floorheight != ch->floorheight
							|| rf->ceilingheight != ch->ceilingheight
							|| rf->floorpic != ch->floorpic
							|| rf->ceilingpic != ch->ceilingpic))
						{
							rf = NULL; break;
						}
					}

					if (!rf) break;
				}

				if (rf)
				{
					float *bnd = osx.outer->sibounds;
					validcount++;
					hitssecptr = 0;

					// get all the subsectors inside the outer sector's bounds
					for (k = 0; k < numsubsectors; k++)
					{
						float x = subsectors[k].midpoint.x;
						float y = subsectors[k].midpoint.y;

						if (   x > bnd[BOXLEFT ]
							&& x < bnd[BOXRIGHT]
							&& y > bnd[BOXBASE ]
							&& y < bnd[BOXTOP  ])
						{
							int found = false;
							// check this is not an outer sector (in case box is not square)
							for (m = 0; m < c; m++)
							{
								if(seclist[m] == subsectors[k].sector - sectors)
								{
									found = true; break;
								}
							}

							if (!found)
								hitssec[hitssecptr++] = k;
						}
					}

					for (k = 0; k < hitssecptr; k++)
					{
						subsector_t *ss = &subsectors[hitssec[k]];
						seg_t		*sg = &segs[ss->firstline];

						Setr3dsslfsec(ss, osx.outer - sectors, 49002);
						Setr3dsscs(ss, i, 49002);

						if (ss->sector->withheight == DDMAXINT)
						{
							ss->sector->withheight = fheight;
							ss->sector->texwithheight[0] = fpic;
							ss->sector->texwithheight[1] = cpic;
						}

						for (m = 0; m < ss->linecount; m++)
						{
							if ( (ld = sg->linedef) && ld->validcount != validcount)
							{
								float	tst;
								byte	type = R3D_3F;

								ld->validcount = validcount;

								if (origTEX[ld - lines])
								{
									// If the line's vertexes are on the box edge then it must be drawn
									if (   ((tst = FIX2FLT(ld->v1->x)) <= bnd[BOXLEFT] || tst >= bnd[BOXRIGHT])
										&& ((tst = FIX2FLT(ld->v1->y)) <= bnd[BOXBASE] || tst >= bnd[BOXTOP  ])
										&& ((tst = FIX2FLT(ld->v2->x)) <= bnd[BOXLEFT] || tst >= bnd[BOXRIGHT])
										&& ((tst = FIX2FLT(ld->v2->y)) <= bnd[BOXBASE] || tst >= bnd[BOXTOP  ]) )
									{
										type = R3D_36;
									}
								}
								
								setLineType(ld, type, 49003);
							}
						}
					}

					for (k = 0; k < osx.scount; k++)
					{
						secnum = countsec[k];

						if (secnum < 0)
						{
							setLowHeightSectors( (secnum & ~SEC_MARKER) );
						}
						else if (sectors[secnum].withheight == DDMAXINT)
						{
							sectors[secnum].withheight = fheight;
							sectors[secnum].texwithheight[0] = fpic;
							sectors[secnum].texwithheight[1] = cpic;
							SetLinesLFG(&sectors[secnum], 49004);
						}
					}
				}
			}

			M_Free(seclist); seclist = NULL;
			M_Free(osx.intList); osx.intList = NULL;
		}
	}
}

// =========================================================================

static void subfn_checkFlatRings(sector_t *sc)
{
	sector_t	*ch, *os;
	line_t		*ld;
	int			c, i, k, box[4], rc;

	// ------------------------------

	CLR_hitline();

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| ld->spflags & SPF_CONTROLLINE
			|| origTEX[ld - lines])
		{
			return;
		}

		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if ( !(ch->specFlags & SF_MOVSECTOR)
			&& ch->ceilingheight > ch->floorheight
			&& ch->ceilingpic != skyflatnum)
		{
			hitline[ld - lines] = 1; c++;
		}
	}

	if (c < 3) return;

	while(1)
	{
		int lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];
			if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
		}

		if (lnum < 0) return;

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL) && rc > 2)
		{
			for (i = 0, os = NULL; i < sc->linecount; i++)
			{
				ld = sc->lines[i];

				if (hitline[ld - lines] == 4)
				{
					ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
					if (!os) os = ch;
					else if (ch != os
						&& (ch->floorheight != sc->floorheight
						|| ch->ceilingheight != sc->ceilingheight))
					{
						os = NULL; break;
					}
				}
			}

			for (i = 0; i < sc->linecount; i++)
			{
				ld = sc->lines[i];

				if (hitline[ld - lines] == 4)
				{
					hitline[ld - lines] = 0;

					if (os)
					{
						ssline_t *sl = &sslines[ld - lines];

						for (k = 0; k < sl->num; k++)
						{
							subsector_t *ss = &subsectors[sl->list[k]];

							if (ss->sector == sc && ss->ceilsec < 0)
							{
								SetOverride(ss, os - sectors, 49005);
							}
						}

						setLineType(ld, R3D_3F, 49006);
					}
				}
			}
		}
		else
			hitline[lnum] = 0;
	}
}

void MC_CheckFlatRings(void)
{
	int		i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& (sectors[i].specFlags & 0x3001C) == 0x1C
			&& sectors[i].ceilingpic != skyflatnum
			&& sectors[i].floorheight == sectors[i].ceilingheight
			&& sectors[i].heightsec < 0)
		{
			subfn_checkFlatRings(&sectors[i]);
		}
	}
}

// =========================================================================

typedef struct {
	__int64	area;
	int		secn;
	int		lcnt;
	int		scnt;
	int		*llist;
	int		*slist;

} nested_t;

static int __cdecl AreaSort(const void *e1, const void *e2)
{
	if (((nested_t*)e1)->area > ((nested_t*)e2)->area) return -1;
	if (((nested_t*)e1)->area < ((nested_t*)e2)->area) return  1;
	return 0;
}

static void subfn_setNT(nested_t *nt)
{
	subsector_t *ss;
	sector_t	*sc, *ch;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			c, i, k;

	if (secdone[nt->secn]) return;

	// transfer the stored nt variables
	sc = &sectors[nt->secn];
	osx.scount = 0;
	countsec[osx.scount++] = nt->secn;
	hitssecptr = nt->scnt;
	memcpy(hitssec, nt->slist, hitssecptr * sizeof(int));

	validcount++;

	CLR_hitline();

	// Identify boundary lines as these could be shared
	for (i = 0; i < nt->lcnt; i++) hitline[nt->llist[i]] = true;

	// Set all subsectors in the relevant area and put all lines
	// found, other than the boundary lines, in linearray
	for (c = i = 0; i < hitssecptr; i++)
	{
		sector_t	*lfsec1 = NULL, *lfsec2 = NULL;
		sector_t	*lcsec1 = NULL, *lcsec2 = NULL;

		ss = &subsectors[hitssec[i]];
		sg = &segs[ss->firstline];

		if (!checkSec(ss->sector - sectors)) countsec[osx.scount++] = ss->sector - sectors;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) )
			{
				if (ld->validcount != validcount)
				{
					ld->validcount = validcount;

					if (hitline[ld - lines] == 0) linearray[c++] = ld - lines;
				}

				// Since lines are shared between subsectors we still need to look at them
				// all even if validcount is set. Note that different lines, even in the
				// same subsector, may not convey the correct information. The dependancy
				// here is that at least one will.

				if (ld->mtflags & R3D_MISLOTEX)
				{
					sector_t *tmp1 = ld->frontsector->floorheight > ld->backsector->floorheight ?
																ld->frontsector : ld->backsector;
					sector_t *tmp2 = ld->frontsector == tmp1 ? ld->backsector : ld->frontsector;

					if (!lfsec1) lfsec1 = tmp1;
					else if (tmp1->floorheight > lfsec1->floorheight) lfsec1 = tmp1;

					if (!lfsec2) lfsec2 = tmp2;
					else if (tmp2->floorheight < lfsec2->floorheight) lfsec2 = tmp2;
				}

				if (ld->mtflags & R3D_MISHITEX)
				{
					sector_t *tmp1 = ld->frontsector->ceilingheight < ld->backsector->ceilingheight ?
																ld->frontsector : ld->backsector;
					sector_t *tmp2 = ld->frontsector == tmp1 ? ld->backsector : ld->frontsector;

					if (!lcsec1) lcsec1 = tmp1;
					else if (tmp1->ceilingheight < lcsec1->ceilingheight) lcsec1 = tmp1;

					if (!lcsec2) lcsec2 = tmp2;
					else if (tmp2->ceilingheight > lcsec2->ceilingheight) lcsec2 = tmp2;
				}
			}
		}

		if (lfsec1) Setr3dssfs(ss, lfsec1 - sectors, 49007);
		if (lfsec2) Setr3dsslfsec(ss, lfsec2 - sectors, 49007);

		if (lcsec1) Setr3dsscs(ss, lcsec1 - sectors, 49007);
		if (lcsec2) Setr3dsslcsec(ss, lcsec2 - sectors, 49007);
	}

	// check for same sided line sharing
	for (i = 0; i < c; i++)
	{
		if (lines[linearray[i]].frontsector == lines[linearray[i]].backsector)
		{
			subsector_t *copyss = NULL;
			
			sl = &sslines[linearray[i]];

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];

				if (ss->floorsec >= 0) { copyss = ss; break; }
			}

			if (!copyss) continue;

			for (k = 0; k < sl->num; k++)
			{
				ss = &subsectors[sl->list[k]];

				if (ss->floorsec < 0)
				{
					Setr3dssfs(ss, copyss->floorsec, 49016);
					Setr3dsslfsec(ss, copyss->lfsec, 49016);
				}
			}
		}
	}

	// Check if the lines' of subsectors outside the area 
	// (adjoining boundary lines) should be set to R3D_16
	for (i = 0; i < nt->lcnt; i++)
	{
		ld = &lines[nt->llist[i]];

		// get the sector adjoining
		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if (!ch) continue;	// insurance
		if (ch->specFlags & SF_MOVSECTOR) continue;	// leave as is
		if ( !(ch->specFlags & (SF_NOLOTEX|SF_NOHITEX)) ) continue;

		// check all lines in the adjoining subsector
		sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->sector == ch)
			{
				int n;

				for (n = 0, sg = &segs[ss->firstline]; n < ss->linecount; n++, sg++)
				{
					if ( (ld = sg->linedef) && ld->mtflags & R3D_MISLOTEX)
					{
						Settype(ld, R3D_16, 49008);
					}
				}
			}
		}
	}

	// Set line types (excluding boundary lines)
	for (i = 0; i < c; i++)
	{
		byte type = R3D_36;

		if (hitline[linearray[i]]) continue;

		ld = &lines[linearray[i]];

		if ( !origTEX[ld - lines] )
		{
			type = R3D_3F;
		}
		else if (origTEX[ld - lines] & 0x22 && !(origTEX[ld - lines] & 0x55))
		{
			type = R3D_03;
		}
		
		if ( !(origTEX[ld - lines] & 0x11) )
		{
			ld->lnflags2 |= LNF2_NORENDLOW;
		}

		if ( !(origTEX[ld - lines] & 0x44) )
		{
			ld->lnflags2 |= LNF2_NORENDHI;
		}

		setLineType(ld, type, 49009);
	}

	// mark the sectors inside as being done
	for (i = 0; i < osx.scount; i++) SetSecidxDone(countsec[i], 49010);
}

#define MAX_NESTED	1024

// v220-16 check for complex structures of nested loops with loop
//         lines contained by textureless missing texture lines
//         Also deals with simpler cases
void MC_Nested()
{
	nested_t	nt[MAX_NESTED];
	sector_t	*sc;
	line_t		*ld;
	int			c, d, i, k;
	int			nt_count = 0;

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (   !secdone[i]
			&& !SecCtrl(i)
			&& !(sc->specFlags & 0x30414)
			&& sc->heightsec < 0
			&& sc->stptr < 0)
		{
			sector_t	*is = NULL, *ch;
			int			inum;

			for (c = d = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (!(ld->ml_flags & ML_TWOSIDED)) goto next_check; 
				if (lineflags[ld - lines]) goto next_check;

				if (ld->frontsector == ld->backsector) continue;

				if (origTEX[ld - lines] == 0x02 || origTEX[ld - lines] == 0x20)
				{
					if (ld->mtflags & R3D_MISTEXANY)
						linearray[d++] = ld - lines;

					continue;
				}

				if (ld->mtflags & R3D_MISTEXANY)
				{
					ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

					if (!is)
					{
						if (ch->specFlags & SF_MOVSECTOR) goto next_check;
						is = ch;
					}
					else if (is != ch) goto next_check;

					inum = ld - lines; c++;
					continue;
				}

				linearray[d++] = ld - lines;
			}

			if (c && d > 3 && is)
			{
				int		rc, chksec[1], box[2][4];

				CLR_hitline();

				for (k = 0; k < d; k++) hitline[linearray[k]] = 1;

				if (subfn_CheckForRing(hitline, sc, linearray[0], &rc, box[0], NULL, NULL) && rc == d)
				{
					CLR_conarray();

					for (k = 0; k < is->linecount; k++)
					{
						ld = is->lines[k];

						if (ld->mtflags & R3D_MISTEXANY)
							conarray[ld - lines] = 1;
					}

					chksec[0] = i;

					if (   subfn_CheckForRing(conarray, is, inum, &rc, box[1], NULL, NULL)
						&& LS_MarkSubsecsInMultiSecLoop(1, chksec, d, box[0], NULL, false, MSIA_NONE))
					{
						if (nt_count == MAX_NESTED)
						{
							pr_OFinsert("WARNING: MC_Nested - array limit of %i insufficient\n", MAX_NESTED);
							break;
						}

						// Store the results
						nt[nt_count].area = ((box[0][BOXRIGHT] - box[0][BOXLEFT]) >> FRACBITS)
										  * ((box[0][BOXTOP  ] - box[0][BOXBASE]) >> FRACBITS);
						nt[nt_count].secn = i;
						nt[nt_count].lcnt = d;
						nt[nt_count].scnt = hitssecptr;
						nt[nt_count].llist = (int*)W_Malloc(d * sizeof(int), TEMPID, 20);
						nt[nt_count].slist = (int*)W_Malloc(hitssecptr * sizeof(int), TEMPID, 21);

						memcpy(nt[nt_count].llist, linearray, d * sizeof(int));
						memcpy(nt[nt_count].slist, hitssec, hitssecptr * sizeof(int));

						nt_count++;
					}
				}
			}
		}
next_check:;
	}

	if (nt_count)
	{
		if (nt_count > 1) qsort(nt, nt_count, sizeof(nested_t), AreaSort);

		// scan results starting with the largest found as this may
		// contain others that also comply
		for (i = 0; i < nt_count; i++) subfn_setNT(&nt[i]);

		W_FreeALL(TEMPID);
	}
}

#undef MAX_NESTED


// =========================================================================

static void subfn_DoubleLoop(sector_t *sc)
{
	line_t	*ld;
	int		i, k;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		if (ss->sector != sc && (ss->sector->specFlags & 0x3041C) != 0x1C) return;

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef && !hitline[sg->linedef - lines]) return;
		}	
	}

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];

		Setr3dsslcsec(ss, sc - sectors, 49010);
	}

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (!origTEX[ld - lines]) Settype(ld, R3D_2A, 49010);

		SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS, 49010);
	}
}

// v220-16 Deals with double nested loop structures with different heights
void MC_DoubleLoop()
{
	sector_t	*sc;
	int			i;

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (   !secdone[i]
			&& !SecCtrl(i)
			&& sc->ceilingpic != skyflatnum
			&& sc->linecount > 5
			&& !(sc->specFlags & 0xFFFFC)
			&& sc->heightsec < 0
			&& sc->stptr < 0)
		{
			sector_t	*chk;
			int			mh = 0, rs = 0, lnum = -1;
			int			rc, box[4];
			line_t		*ld;
			int			k;

			CLR_hitline();

			for (k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (lineflags[ld - lines]) { mh = 0; break; }
				if (ld->mtflags & R3D_MISLOTEX) { mh = 0; break; }
				if (ld->mtflags & R3D_MISHITEX)
				{
					if (origTEX[ld - lines] & 0x55) { mh = 0; break; }
					chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;
					if (   chk
						&& (chk->ceilingheight > sc->ceilingheight
						|| chk->floorheight != sc->floorheight
						|| (chk->specFlags & 0x3041C) != 0x1C))
					{
						mh = 0; break;
					}

					hitline[ld - lines] = 8;
					mh++;
				}
				else
				{
					lnum = ld - lines;
					hitline[ld - lines] = 1;
					rs++;
				}
			}

			if (mh < 3 || rs < 3) continue;

			if (   subfn_CheckForRing(hitline, sc, lnum, &rc, box, NULL, NULL)
				&& rs == rc
				&& LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, box, -1, MSIA_NONE))
			{
				subfn_DoubleLoop(sc);
			}
		}
	}
}

// =========================================================================

static void getLinesIterator(int x, int y, int *box)
{
	int			offset;
	const long	*list;
	
	if (x < 0 || y < 0 || x >= bmapwidth || y >= bmapheight)
		return;

	offset = y*bmapwidth + x;

	offset = *(blockmap + offset);
	list = blockmaplump + offset;

	// skip starting delimiter
	while( *(++list) >= 0)
	{
		line_t *ld = &lines[*list];

		if (ld->validcount == validcount)
			continue;		// line has already been checked

		ld->validcount = validcount;

		if (M_FitBox(ld->bbox, box))
			linearray[osx.lcount++] = ld - lines;
	}
}


// Checks for same sided lines in box and gets all subsectors
static int subfn_GetSsecsInBox(int *box, float *fpbox)
{
	subsector_t	*ss;
    int			xl, xh, yl, yh, bx, by;
	int			samesec = -1;

	M_getBlockOffsets(box, 0, &xl, &xh, &yl, &yh);

	osx.lcount = 0;
	osx.scount = 0;
	hitssecptr = 0;
	validcount++;
	ss_validcount++;

	for (bx=xl ; bx<=xh ; bx++)
	{
		for (by=yl ; by<=yh ; by++)
		{
			getLinesIterator(bx, by, box);
		}
	}

	for (xl = 0; xl < osx.lcount; xl++)
	{
		ssline_t *sl = &sslines[linearray[xl]];

		for (xh = 0; xh < sl->num; xh++)
		{
			ss = &subsectors[sl->list[xh]];

			if (ss->ssvalidcount == ss_validcount) continue;

			ss->ssvalidcount = ss_validcount;

			if (M_CompareFPxyBox(ss->midpoint.x, ss->midpoint.y, fpbox))
			{
				line_t *ld = &lines[linearray[xl]];

				if (ld->frontsector == ld->backsector)
				{
					int secnum = ld->frontsector - sectors;

					if (samesec < 0) samesec = secnum;
					else if (samesec != secnum) return -1;
				}

				if (!checkSec(ss->sector - sectors)) countsec[osx.scount++] = ss->sector - sectors;
				hitssec[hitssecptr++] = sl->list[xh];
			}
		}
	}

	return samesec;
}

// v220-16 Look for false floors bounded by a re-entrant sector
void MC_FalseFloorInBox()
{
	sector_t	*sc = sectors;
	line_t		*ld;
	int			i;

	for (i = 0; i < numsectors; i++, sc++)
	{
		if (   !secdone[i]
			&& sc->linecount > 3
			&& (sc->specFlags & 0xF041C) == 0x41C
			&& !SecCtrl(i)
			&& sc->heightsec < 0
			&& sc->stptr < 0)
		{
			sector_t	*ch, *os = NULL;
			int			c, d, k, rc, box[4];

			M_ClearBox(box);

			for (c = d = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];
				if (lineflags[ld - lines]) goto end_check;

				if (ld->ml_flags & ML_TWOSIDED)
				{
					ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

					if (!os)
					{
						if (ch->floorheight >= sc->floorheight) goto end_check;
						if (ch->specFlags & SF_NOLOTEX) goto end_check;
						if (!(ch->specFlags & SF_SECCLOSED)) goto end_check;

						os = ch;
					}
					else if (os != ch)
					{
						goto end_check;
					}

					if (ld->mtflags == 0x30 || ld->mtflags == 0xC0)
					{
						if (ch->ceilingheight < sc->ceilingheight) goto end_check;
					}
					else if (ld->mtflags != 0x10 && ld->mtflags != 0x40)	// same side missing upper and lower
					{
						goto end_check;
					}

					d++;
				}
				else
				{
					M_AddLineToBox(box, ld);
					c++;
				}
			}

			if (   c > 1
				&& d > 1
				&& c + d == sc->linecount
				)
			{
				float	fpbox[4];

				// check os is inside the box.
				// This also proves sc's missing tex lines all have at least one vertex inside the box
				M_SetFPBoxFromIntBox(fpbox, box);

				if (!M_FitFPBox(os->sibounds, fpbox)) goto end_check;

				CLR_hitline();

				for (k = 0; k < sc->linecount; k++)
				{
					ld = sc->lines[k]; hitline[ld - lines] = 1;
				}

				if (subfn_CheckForRing(hitline, sc, sc->lines[0] - lines, &rc, box, NULL, NULL) && rc == sc->linecount)
				{
					int maxht = DDMININT, samesec, inherit = -1;

					samesec = subfn_GetSsecsInBox(box, fpbox);

					if (samesec < 0) goto end_check;

					// Using zdnodes the containing sector cannot be identified from the subsectors
					for (k = 0; k < osx.scount; k++)
					{
						if (   sectors[samesec].floorheight > sectors[countsec[k]].floorheight
							&& M_CompareFPBox(sectors[samesec].sibounds, sectors[countsec[k]].sibounds))
						{
							inherit = countsec[k];
							break;
						}
					}

					if (inherit < 0) goto end_check;

					for (k = 0; k < osx.scount; k++)
					{
						if (   countsec[k] != i
							&& sectors[countsec[k]].floorheight > maxht)
						{
							maxht = sectors[countsec[k]].floorheight;
						}
					}

					if (maxht == sc->floorheight)
					{
						for (k = 0; k < hitssecptr; k++)
						{
							subsector_t *ss = &subsectors[hitssec[k]];

							if (ss->sector->floorheight == sc->floorheight)
								Setr3dssfs(ss, inherit, 49017);
							Setr3dsslfsec(ss, i, 49017);
							Setr3dsslcsec(ss, i, 49017);
						}

						for (k = 0; k < osx.lcount; k++)
						{
							ld = &lines[linearray[k]];
							if (!origTEX[ld - lines])
								Settype(ld, R3D_3F, 49012);
							SetLFG(ld, LFG_NEWSET|LFG_FIXABS, 49012);
						}

						// mark the sectors inside as being done
						for (k = 0; k < osx.scount; k++) SetSecidxDone(countsec[k], 49012);
					}
				}
			}
		}
end_check:;
	}
}

// Look for ceiling inheritance
// Done last else could be confused with other types
void MC_ContiguousMissingHi()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& sectors[i].linecount > 3
			&& (sectors[i].specFlags & 0xF0414) == 0x14
			&& !SecCtrl(i)
			&& sectors[i].heightsec < 0
			&& sectors[i].stptr < 0)
		{
			sector_t	*sc = &sectors[i], *ch, *os = NULL;
			line_t		*ld;
			int			c, k, m, rc;

			for (c = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (lineflags[ld - lines]) goto end_check;
				if (origTEX[ld - lines] & 0x33) goto end_check;
				if (!(ld->ml_flags & ML_TWOSIDED)) goto end_check;
				if (ld->specType) goto end_check;
				if (ld->mtflags & R3D_MISHITEX) c++;

				ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

				if (!os) os = ch;
				else if (os != ch && os->ceilingheight != ch->ceilingheight) goto end_check;
			}

			if (!os || c < 2) goto end_check;

			CLR_hitline();

			for (k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];
				hitline[ld - lines] = 1;
			}

			if (subfn_CheckForRing(hitline, sc, sc->lines[0] - lines, &rc, NULL, NULL, NULL) && rc == sc->linecount)
			{
				ssline_t *sl;

				ss_validcount++;

				for (k = 0; k < sc->linecount; k++)
				{
					ld = sc->lines[k];
					sl = &sslines[ld - lines];

					for (m = 0; m < sl->num; m++)
					{
						subsector_t *ss = &subsectors[sl->list[m]];

						if (ss->ssvalidcount != ss_validcount)
						{
							ss->ssvalidcount = ss_validcount;

							if (ss->sector == sc && ss->ceilsec < 0)
								Setr3dsscs(ss, os - sectors, 49013);
							Setr3dsslcsec(ss, sc - sectors, 49013);
						}
					}

					if (!origTEX[ld - lines])
						Settype(ld, R3D_3F, 49013);
					SetLFG(ld, LFG_NEWSET, 49013);
				}

				Setsecdone(sc, 49013);	
			}
		}

end_check:;
	}
}

// ============================================================

static void subfn_MissingLoMove(sector_t *sc)
{
	sector_t	*ch, *os = NULL;
	line_t		*ld;
	int			i, k;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;
		if (lineflags[ld - lines]) return;
		if (!(ld->ml_flags & ML_TWOSIDED)) continue;
		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
		if ((ch->specFlags & 0xC) == 0xC) continue;
		if (ch->specFlags & 0x30004) return;
		if (ch->ceilingheight != sc->ceilingheight || ch->ceilingpic != sc->ceilingpic) return;
		if (!os) os = ch; else if (os != ch) return;
	}

	if (os)
	{
		ssline_t	*sl;

		for (i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (ld->spflags & SPF_CONTROLLINE) continue;

			sl = &sslines[ld - lines];

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (   ss->sector != os
					&& ss->lfsec < 0
					&& ss->floorsec < 0)
				{
					Setr3dsslfsec(ss, os - sectors, 48201);
				}
			}
		}
	}
}

// v220-16 Check for stuff that defies all logic
void MC_MissingLoMove()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& sectors[i].linecount > 3
			&& (sectors[i].specFlags & 0xFFFFC) == 0x98C9C
			&& SecCtrl(i)
			&& sectors[i].heightsec < 0
			&& sectors[i].stptr < 0)
		{
			subfn_MissingLoMove(&sectors[i]);
		}
	}
}

// ============================================================

static void addMarkedSecLines(sector_t *sc, int fht, int *c)
{
	int i;

	for (i = 0; i < sc->linecount; i++)
	{
		line_t *ld = sc->lines[i];

		if (ld->validcount == validcount) continue;

		ld->validcount = validcount;

		if (!origTEX[ld - lines] && ld->mtflags & R3D_MISLOTEX)
		{
			linearray[(*c)++] = ld - lines;
			traceLoFromMarked(ld, fht, c);
		}
	}
}

static void traceLoFromMarked(line_t *ld, int fht, int *c)
{
	vertexowner_t	*vo;
	int				i;

	countsec[ld->frontsector - sectors] = 1;
	countsec[ld->backsector - sectors] = 1;

	vo = &vertexowners[ld->v1 - vertexes];

	for (i = 0; i < vo->num; i++)
	{
		if (countsec[vo->list[i]]) continue;
		if (sectors[vo->list[i]].floorheight != fht) continue;
		countsec[vo->list[i]] = 2;
		addMarkedSecLines(&sectors[vo->list[i]], fht, c);
	}

	vo = &vertexowners[ld->v2 - vertexes];

	for (i = 0; i < vo->num; i++)
	{
		if (countsec[vo->list[i]]) continue;
		if (sectors[vo->list[i]].floorheight != fht) continue;
		countsec[vo->list[i]] = 2;
		addMarkedSecLines(&sectors[vo->list[i]], fht, c);
	}
}

void MC_CheckNonContiguous(sector_t *sc)
{
	sector_t	*os, *ch;
	line_t		*ld;
	int			c, i, lnum, rc;

	for (i = 0, os = NULL; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		lnum = ld - lines;

		if (ld->specType != R3D_00) return;
		if (!(ld->ml_flags & ML_TWOSIDED)) return;

		if (ld->mtflags & R3D_MISLOTEX)
		{
			if (lineflags[lnum] || origTEX[lnum]) return;
			ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
			if (!os)
			{
				if (ch->specFlags & 0x30004) return;
				os = ch;
			}
			else if (os != ch && ch->floorheight != os->floorheight)
				return;
		}
		else if (lineflags[lnum] != LFG_MARKED)
		{
			return;
		}
	}

	if (!os) return;

	CLR_hitline();

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		hitline[ld - lines] = 1;
	}

	if (subfn_CheckForRing(hitline, sc, sc->lines[0] - lines, &rc, NULL, NULL, NULL) && rc == sc->linecount)
	{
		int box[4];

		SET_countsec(0);

		countsec[sc - sectors] = 1;
		countsec[os - sectors] = 1;

		validcount++;

		// Trace other missinglo lines from owners of marked lines
		for (c = i = 0; i < sc->linecount; i++)
		{
			ld = sc->lines[i];

			if (lineflags[ld - lines] & LFG_MARKED)
			{
				traceLoFromMarked(ld, os->floorheight, &c);
			}
			else
				linearray[c++] = ld - lines;
		}

		if (nf_chkLineListRing(c, box, 0, NULL) == c)
		{
			RDBOOL	found;
			int		*chksec = (int*)malloc(c * sizeof(int));
			int		k, cscnt = 0;

			// Find all subsecs in the marked area
			for (i = 0; i < c; i++)
			{
				ld = &lines[linearray[i]];
				ch = ld->frontsector->floorheight == sc->floorheight ? ld->frontsector : ld->backsector;

				if (ch->floorheight == sc->floorheight)
				{
					for (k = 0, found = false; k < cscnt; k++)
					{
						if (chksec[k] == ch - sectors) { found = true; break; }
					}

					if (!found) chksec[cscnt++] = ch - sectors;
				}
			}

			if (cscnt && LS_MarkSubsecsInMultiSecLoop(cscnt, chksec, c, box, NULL, 0, MSIA_NONE))
			{
				for (i = 0; i < hitssecptr; i++)
				{
					Setr3dsslfsec(&subsectors[hitssec[i]], os - sectors, 49014);
				}

				for (i = 0; i < c; i++)
				{
					ld = &lines[linearray[i]];
					Settype(ld, R3D_16, 49014);
					SetLFG(ld, LFG_NEWSET, 49014);
				}

				Setsecdone(sc, 49014);
			}

			free(chksec);
		}
	}
}

// ============================================================

static void subfn_CheckIfHeightsecInvalid(sector_t *sc)
{
	sector_t	*ch, *os;
	ssline_t	*sl;
	line_t		*ld;
	RDBOOL		test0, test1;
	int			mttest[2] = { 0, 0 };
	int			c, i, k;

	for (c = i = 0, os = NULL; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
		if (!ch) continue;
		if (ch == sc) continue;
		if (ch->heightsec >= 0) return;

		if (!os) os = ch;
		else if (os != ch) return;

		if (ld->mtflags & R3D_MISLOTEX) mttest[0]++;	// missing textures
		if (ld->mtflags & R3D_MISHITEX) mttest[1]++;
		c++;
	}

	if (!os) return;

	test0 = mttest[0] == c && sectors[sc->heightsec].floorheight > os->floorheight;
	test1 = mttest[1] == c && sectors[sc->heightsec].ceilingheight < os->ceilingheight;

	if (!test0 && !test1) return;

	for (i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		sl = &sslines[ld - lines];

		for (k = 0; k < sl->num; k++)
		{
			subsector_t *ss = &subsectors[sl->list[k]];

			if (ss->sector == sc)
			{
				if (test0 && ss->floorsec < 0)
					Setr3dssfs(ss, os - sectors, 49015);

				if (test1 && ss->ceilsec < 0)
					Setr3dsscs(ss, os - sectors, 49015);
			}
		}

		SetLFG(ld, LFG_FIXABS|LFG_NOCHANGE|LFG_FIXED, 49015);
	}

	Setsecdone(sc, 49015);
}

// v220-16 heightsec errors, especially with ceiling heights, coupled
//     with poor software rendering can make a fault difficult to see.
void MC_CheckIfHeightsecInvalid()
{
	int i;

	for (i = 0; i < numsectors; i++)
	{
		if (  !secdone[i]
			&& sectors[i].heightsec >= 0
			&& !(sectors[i].specFlags & SF_MOVSECTOR)
			&& sectors[i].ceilingheight > sectors[i].floorheight
			&& !SecCtrl(i)
			&& sectors[i].stptr < 0)
		{
			subfn_CheckIfHeightsecInvalid(&sectors[i]);
		}
	}
}

// ===============================================================

static RDBOOL noLoTexInSsec(RDBOOL usehitline)
{
	int i, k;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (usehitline && sg->linedef && hitline[sg->linedef - lines]) continue;
			if (sg->linedef && origTEX[sg->linedef - lines] & 0x11)
				return false;
		}
	}

	return true;
}

static sector_t *mc_getOwner(subsector_t *ss, int fht)
{
	subsector_t	*os;
	ssline_t	*sl;
	seg_t		*sg = &segs[ss->firstline], *og;
	int			i, k, m;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (sg->linedef)
		{
			sl = &sslines[sg->linedef - lines];

			for (k = 0; k < sl->num; k++)
			{
				if ( (os = &subsectors[sl->list[k]]) != ss)
				{
					og = &segs[os->firstline];

					for (m = 0; m < os->linecount; m++, og++)
					{
						if (   og->backsector
							&& og->backsector != sg->frontsector
							&& og->backsector->floorheight == fht)
						{
							return og->backsector;
						}
					}
				}
			}
		}
	}

	return NULL;
}

static void mc_setSameSecSectors(int flidx, int c)
{
	subsector_t	*ss;
	sector_t	*os, *st, *sd = NULL;
	RDBOOL		v;
	int			i, k;

	for (i = 0; i < c; i++)
	{
		st = &sectors[countsec[i]];

		if ((st->specFlags & 0xC) == 0xC)
		{
			// get the dominant floor (this is the one that would have been used by the visplane)
			if (st->specFlags & SF_MOVFLOOR)
			{
				if (!sd) sd = st; else if (st->floorheight > sd->floorheight) sd = st;
			}
			else
			{
				for (k = 0, v = true; k < st->linecount; k++)
				{
					if (st->lines[k]->frontsector != st->lines[k]->backsector)
					{
						v = false; break;
					}
				}

				if (!v) continue;

				for (k = 0; k < hitssecptr; k++)
				{
					ss = &subsectors[hitssec[k]];

					if (ss->sector == st)
					{
						// Ceiling ownership has to be checked
						if ( (os = mc_getOwner(ss, sectors[flidx].floorheight)) )
							Setr3dsscs(ss, os - sectors, 48202);

						Setr3dssfs(ss, flidx, 48202);
					}
				}
			}
		}
	}

	if (sd)
	{
		for (i = 0; i < hitssecptr; i++)
		{
			ss = &subsectors[hitssec[i]];

			if (   ss->sector->floorheight < sd->floorheight
				|| ss->floorsec >= 0)
			{
				Setr3dsslfsec(ss, sd - sectors, 48203);
			}
		}
	}
}

static void setLoTexSsecs(int lfidx, int lfht, int *count, int usehitline)
{
	line_t	*ld;
	int		i, k;

	validcount++;

	for (i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];
		RDBOOL		setfs = ss->sector->floorheight == lfht;

		ss->ssflags |= SSF_NOCHANGE;

		if (!setfs && !usehitline) Setr3dsslfsec(ss, lfidx, 49019);

		addSec(ss->sector - sectors, count);

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;

				if (usehitline && hitline[ld - lines])
				{
					Forcetype(ld, R3D_36, 49019);
				}
				else
				{
					byte txn = origTEX[ld - lines];

					ld->lnflags2 |= LNF2_NORENDLOW;

					if (!txn)
					{
						Settype(ld, R3D_3F, 49019);
					}
					else if (txn & 0x22 && ld->frontsector == ld->backsector)
					{
						Settype(ld, R3D_03, 49019);
					}
					else if ( ((txn == 0x02 && ld->mtflags == R3D_MISLTEX1)
						||     (txn == 0x20 && ld->mtflags == R3D_MISLTEX2))
						&& !(ld->frontsector->specFlags & SF_MOVFLOOR)
						&& !(ld->backsector->specFlags  & SF_MOVFLOOR))
					{
						Settype(ld, R3D_1D, 49019);
					}
					else if (!ld->mtflags)
					{
						Settype(ld, R3D_36, 49019);
					}
				}

				SetLFG(ld, LFG_NEWSET|LFG_FIXED, 49019);

				if (   setfs
					&& !usehitline
					&& sg->backsector
					&& sg->backsector->floorheight > lfidx)
				{
					Setr3dssfs(ss, sg->backsector - sectors, 49019);
					setfs = false;
				}
			}
		}
	}
}

// should only be called with condition set to 1 or 2
static RDBOOL subfn_getLineStrings(sector_t *sc, int sccnt, int c, int condition)
{
	RDBOOL	ret = false;
	int		n1, n2;
	int		i, ep = 0, mp = 0;
	int		*endp = (int*)malloc(c * sizeof(int));
	int		*midp = (int*)malloc(c * sizeof(int));

	vertex_t	*v1, *v2, *vs1, *ve1, *vs2, *ve2;
	int			*ls1, *ls2;
	int			lc1, lc2;
	int			repeat;

	// ---------------------------------

	CLR_vtxcheck();

	for (i = 0; i < c; i++)
	{
		vtxcheck[lines[linearray[i]].v1 - vertexes]++;
		vtxcheck[lines[linearray[i]].v2 - vertexes]++;
	}

	// find the end points
	for (i = 0; i < c; i++)
	{
		n1 = vtxcheck[lines[linearray[i]].v1 - vertexes];
		n2 = vtxcheck[lines[linearray[i]].v2 - vertexes];

		if (n1 == 2 && n2 == 2)
		{
			midp[mp++] = linearray[i];
		}
		else
		if ((n1 == 1 && n2 == 2) || (n2 == 1 && n1 == 2))
		{
			endp[ep++] = linearray[i];
		}
		else
		{
			ep = 0; break;
		}
	}

	if (ep != 4) { free(endp); free(midp); return false; }
	
	// ep = 4 indicates two line strings

	ls1 = (int*)malloc(c * sizeof(int));
	ls2 = (int*)malloc(c * sizeof(int));
	lc1 = 0, lc2 = 0;

	// check for two contiguous strings

	// string1 - set the first endpoint and v1
	ls1[lc1++] = endp[0];
	v1 = lines[endp[0]].v1;
	
	if (vtxcheck[v1 - vertexes] == 1)
	{
		vs1 = v1;
		v1 = lines[endp[0]].v2;
	}
	else
	{
		vs1 = lines[endp[0]].v2;
	}

	endp[0] = -1;
	repeat = true;

	while (repeat)
	{
		for (i = 0, repeat = false; i < mp; i++)
		{
			if (midp[i] < 0) continue;

			if (lines[midp[i]].v1 == v1)
			{
				v1 = lines[midp[i]].v2;
				ls1[lc1++] = midp[i];
				midp[i] = -1;
				repeat = true;
				break;
			}
			else
			if (lines[midp[i]].v2 == v1)
			{
				v1 = lines[midp[i]].v1;
				ls1[lc1++] = midp[i];
				midp[i] = -1;
				repeat = true;
				break;
			}
		}
	}

	// set the second end point
	for (i = 1, ve1 = NULL; i < ep; i++)
	{
		if (endp[i] < 0) continue;

		if (lines[endp[i]].v1 == v1)
		{
			ls1[lc1++] = endp[i];
			ve1 = lines[endp[i]].v2;
			endp[i] = -1;
			break;
		}
		else
		if (lines[endp[i]].v2 == v1)
		{
			ls1[lc1++] = endp[i];
			ve1 = lines[endp[i]].v1;
			endp[i] = -1;
			break;
		}
	}

	if (!ve1) goto tidyup;

	// string 2 - set the first endpoint and v1
	for (i = 0, vs2 = NULL; i < ep; i++)
	{
		if (endp[i] >= 0)
		{
			v1 = lines[endp[i]].v1;
			ls2[lc2++] = endp[i];

			if (vtxcheck[v1 - vertexes] == 1)
			{
				vs2 = v1;
				v1 = lines[endp[i]].v2;
			}
			else
			{
				vs2 = lines[endp[i]].v2;
			}
			endp[i] = -1;
			break;
		}
	}

	if (!vs2) goto tidyup;

	repeat = true;

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < mp; i++)
		{
			if (midp[i] < 0) continue;

			if (lines[midp[i]].v1 == v1)
			{
				v1 = lines[midp[i]].v2;
				ls2[lc2++] = midp[i];
				midp[i] = -1;
				repeat = true;
				break;
			}
			else
			if (lines[midp[i]].v2 == v1)
			{
				v1 = lines[midp[i]].v1;
				ls2[lc2++] = midp[i];
				midp[i] = -1;
				repeat = true;
				break;
			}
		}
	}

	// set the second end point
	for (i = 0, ve2 = NULL; i < ep; i++)
	{
		if (endp[i] >= 0 && lines[endp[i]].v1 == v1)
		{
			ls2[lc2++] = endp[i];
			ve2 = lines[endp[i]].v2;
			endp[i] = -1;
			break;
		}
		else
		if (endp[i] >= 0 && lines[endp[i]].v2 == v1)
		{
			ls2[lc2++] = endp[i];
			ve2 = lines[endp[i]].v1;
			endp[i] = -1;
			break;
		}
	}

	// check the 2nd endpoint was set and that the lines add up
	if (ve2 && lc1 + lc2 == ep + mp)
	{
		extern int vtxc_x;
		extern int vtxc_y;

		sector_t	*os;
		line_t		*lb, *ld, ref;
		int			mv1, mv2;
		int			complete;
		int			lc = 0;
		int			usehitline = condition == 2;
		int			box[4];

		if (usehitline) CLR_hitline();

		M_ClearBox(box);

		// First add the lines we know about to the box
		for (i = 0; i < lc1; i++)
		{
			ld = &lines[ls1[i]];

			if (usehitline)
			{
				sidelist[lc] = ld->frontsector == sc ? ld->sidenum[0] : ld->sidenum[1];
				hitline[ld - lines] = 1;
			}
			else
				sidelist[lc] = ld->frontsector != sc ? ld->sidenum[0] : ld->sidenum[1];

			linearray[lc++] = ls1[i];
			M_AddLineToBox(box, ld);
		}

		for (i = 0; i < lc2; i++)
		{
			ld = &lines[ls2[i]];

			if (usehitline)
			{
				sidelist[lc] = ld->frontsector == sc ? ld->sidenum[0] : ld->sidenum[1];
				hitline[ld - lines] = 1;
			}
			else
				sidelist[lc] = ld->frontsector != sc ? ld->sidenum[0] : ld->sidenum[1];

			linearray[lc++] = ls2[i];
			M_AddLineToBox(box, ld);
		}

		// set the global box centre for subfn_FindVtxConnected (needed where condition != 0)
		vtxc_x = (box[BOXLEFT]>>1) + (box[BOXRIGHT]>>1);
		vtxc_y = (box[BOXBASE]>>1) + (box[BOXTOP  ]>>1);

		// Second find the lines that close the loop
		if (abs(ve1->x - vs2->x) + abs(ve1->y - vs2->y) < abs(vs1->x - vs2->x) + abs(vs1->y - vs2->y))
		{
			v1 = ve1; ve1 = vs1; vs1 = v1;
		}

		ref.frontsector = sc;
		ref.specType = 0;
		ref.v1 = vs1; ref.v2 = vs2;
		ref.bbox[BOXLEFT ] = MIN_OF(vs1->x, vs2->x);
		ref.bbox[BOXRIGHT] = MAX_OF(vs1->x, vs2->x);
		ref.bbox[BOXBASE ] = MIN_OF(vs1->y, vs2->y);
		ref.bbox[BOXTOP  ] = MAX_OF(vs1->y, vs2->y);

		if ( !(lb = subfn_FindVtxConnected(&ref, condition)) ) goto tidyup;

		mv1 = ref.v1 == lb->v1 || ref.v1 == lb->v2;
		mv2 = ref.v2 == lb->v1 || ref.v2 == lb->v2;

		if (lb->frontsector->floorheight == sc->floorheight)
		{
			sidelist[lc] = lb->sidenum[0]; os = lb->frontsector;
		}
		else
		{
			sidelist[lc] = lb->sidenum[1]; os = lb->backsector;
		}

		linearray[lc++] = lb - lines;
		M_AddLineToBox(box, lb);
		if (usehitline) hitline[lb - lines] = 1;

		if (!mv1 || !mv2)
		{
			// More than 1 bridging line so rest must be found
			if (mv1)
			{
				v2 = ref.v2; v1 = ref.v1 == lb->v1 ? lb->v2 : lb->v1;
			}
			else
			{
				v2 = ref.v1; v1 = ref.v2 == lb->v1 ? lb->v2 : lb->v1;
			}

			for (i = 0, complete = 0; i < os->linecount; i++)
			{
				RDBOOL found = false;
				ld = os->lines[i];

				if (ld != lb)
				{
					if (ld->v1 == v1) { v1 = ld->v2; found = true; }
					else
					if (ld->v2 == v1) { v1 = ld->v1; found = true; }

					if (found)
					{
						sidelist[lc] = ld->frontsector == os ? ld->sidenum[0] : ld->sidenum[1];
						linearray[lc++] = ld - lines;
						M_AddLineToBox(box, ld);
						if (usehitline) hitline[ld - lines] = 1;
						if (v1 == v2) { complete = 1; break; }
					}
				}
			}

			if (!complete)
			{
				goto tidyup;
			}
		}

		ref.v1 = ve1; ref.v2 = ve2;
		ref.bbox[BOXLEFT ] = MIN_OF(ve1->x, ve2->x);
		ref.bbox[BOXRIGHT] = MAX_OF(ve1->x, ve2->x);
		ref.bbox[BOXBASE ] = MIN_OF(ve1->y, ve2->y);
		ref.bbox[BOXTOP  ] = MAX_OF(ve1->y, ve2->y);

		if ( !(lb = subfn_FindVtxConnected(&ref, condition)) ) goto tidyup;

		mv1 = ref.v1 == lb->v1 || ref.v1 == lb->v2;
		mv2 = ref.v2 == lb->v1 || ref.v2 == lb->v2;

		if (lb->frontsector->floorheight == sc->floorheight)
		{
			sidelist[lc] = lb->sidenum[0]; os = lb->frontsector;
		}
		else
		{
			sidelist[lc] = lb->sidenum[1]; os = lb->backsector;
		}

		linearray[lc++] = lb - lines;
		M_AddLineToBox(box, lb);
		if (usehitline) hitline[lb - lines] = 1;

		if (!mv1 || !mv2)
		{
			// More than 1 bridging line so rest must be found
			if (mv1)
			{
				v2 = ref.v2; v1 = ref.v1 == lb->v1 ? lb->v2 : lb->v1;
			}
			else
			{
				v2 = ref.v1; v1 = ref.v2 == lb->v1 ? lb->v2 : lb->v1;
			}

			for (i = 0, complete = 0; i < os->linecount; i++)
			{
				RDBOOL found = false;
				ld = os->lines[i];

				if (ld != lb)
				{
					if (ld->v1 == v1) { v1 = ld->v2; found = true; }
					else
					if (ld->v2 == v1) { v1 = ld->v1; found = true; }

					if (found)
					{
						sidelist[lc] = ld->frontsector == os ? ld->sidenum[0] : ld->sidenum[1];
						linearray[lc++] = ld - lines;
						M_AddLineToBox(box, ld);
						if (usehitline) hitline[ld - lines] = 1;
						if (v1 == v2) { complete = 1; break; }
					}
				}
			}

			if (!complete)
			{
				goto tidyup;
			}
		}
		
		// Get all subsectors in the loop
		if (LS_MarkSsecsUsingSidelist(lc, box, MSIA_NONE) && noLoTexInSsec(usehitline))
		{
			setLoTexSsecs(sc - sectors, sc->floorheight, &sccnt, usehitline);
			if (usehitline) mc_setSameSecSectors(sc - sectors, sccnt);
			addSec(sc - sectors, &sccnt);

			for (i = 0; i < sccnt; i++) Setsecdone(&sectors[countsec[i]], 49020);

			ret = true;
		}
	}

tidyup:;
	free(ls1);
	free(ls2);
	free(endp);
	free(midp);
	return ret;
}

// v220-17. Rare but awkward condition where an open sector has two strands
// that enclose an area with no low textures. The two strands should both
// reference an outer lower floor sector(s) with the same height as the
// sector being tested. The test sec should be rendered beneath.

void MC_FixedMLoContained()
{
	line_t	*ld;
	int		i;

	for (i = 0; i < numsectors; i++)
	{
		if (   !secdone[i]
			&& (sectors[i].specFlags & 0xFFFFC) == 0x404
			&& !SecCtrl(i)
			&& sectors[i].heightsec < 0)
		{
			sector_t	*ch, *sc = &sectors[i];
			int			sccnt, tmpsccnt, c, k;
			int			fht = sc->floorheight, nfht = DDMAXINT;
			short		fpc = sc->floorpic;

			for (sccnt = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (lineflags[ld - lines]) { sccnt = 0; break; };

				if (   ld->ml_flags & ML_TWOSIDED
					&& !ld->mtflags
					&& ld->frontsector != ld->backsector)
				{
					if (ld->backsector != sc) { sccnt = 0; break; };

					ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

					if (addSec(ch - sectors, &sccnt))
					{
						if (ch->floorheight != fht) { sccnt = 0; break; };
						if (ch->floorpic    != fpc) { sccnt = 0; break; };
						if (ch->specFlags & SF_MOVFLOOR) { sccnt = 0; break; };
						if (ch->specFlags & SF_NOLOTEX && !(ch->specFlags & SF_SSLINETRUE)) { sccnt = 0; break; };
					}
				}
			}

			if ( !(tmpsccnt = sccnt) ) continue;

			for (c = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (   ld->ml_flags & ML_TWOSIDED
					&& ld->mtflags
					&& ld->frontsector != ld->backsector)
				{
					if (ld->backsector != sc) { sccnt = 0; break; };

					linearray[c++] = ld - lines;
					ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;
					if (addSec(ch - sectors, &sccnt))
					{
						if ((ch->specFlags & (SF_MOVFLOOR|SF_NOLOTEX)) != SF_NOLOTEX) { sccnt = 0; break; };

						if (nfht == DDMAXINT)
						{
							nfht = ch->floorheight;
							if (nfht <= sc->floorheight) { sccnt = 0; break; };
						}
						else if (nfht != ch->floorheight)
						{ sccnt = 0; break; };
					}
				}
			}

			if (sccnt > tmpsccnt) subfn_getLineStrings(sc, sccnt, c, 1);
		}
	}
}

// v220-18 Move version of MC_FixedMLoContained()
//         Currently only called for donut types
RDBOOL mc_MovsecMLoContained(sector_t *sc)
{
	sector_t	*ch;
	line_t		*ld;
	int			sccnt;
	int			c, i;

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];
		if (lineflags[ld - lines]) { c = 0; break; }
		if (ld->mtflags & R3D_MISLOTEX) c++;
	}

	if (c < 3) return false;

	for (c = i = 0; i < sc->linecount; i++)
	{
		ld = sc->lines[i];

		if (   ld->mtflags & R3D_MISLOTEX
			|| ld->frontsector == ld->backsector)
		{
			continue;
		}

		ch = ld->frontsector == sc ? ld->backsector : ld->frontsector;

		if (ch && ch->specFlags & SF_MOVFLOOR) continue;

		linearray[c++] = ld - lines;
	}

	if (c < 4) return false;

	sccnt = 0; return (subfn_getLineStrings(sc, sccnt, c, 2));
}


// v220-27 Look for same sided lines outside a box with the same sector (rare)
// Note: Line types will already have been set at this point.
//       Sector and line done flags are left unset.
void MC_CheckSSLinesContained()
{
	subsector_t	*ss;
	sector_t	*sc, *fs, *bs, *ch, *os;
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			box[4];
	int			c, d, i, k, m, n;

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (secdone[i]) continue;
		if (sc->heightsec >= 0) continue;

		for (c = d = k = 0, os = NULL; k < sc->linecount; k++)
		{
			ld = sc->lines[k];

			if (ld->spflags & SPF_CONTROLLINE) continue;
			if (lineflags[ld - lines]) goto nextsector;
			if ((fs = ld->frontsector) == (bs = ld->backsector)) { d++; continue; }
			ch = fs == sc ? bs : fs;
			if (ch)
			{
				if (!os) os = ch; else if (os != ch) goto nextsector;
			}
			
			linearray[c++] = ld - lines;
		}

		if (   c > 2
			&& d
			&& os
			&& sc->floorheight > os->floorheight
			&& !(os->specFlags & SF_MOVFLOOR)
			&& M_CompareFPBox(sc->sibounds, os->sibounds)
			&& nf_chkLineListRing(c, box, 0, NULL) == c)
		{
			for (k = 0; k < c; k++)
			{
				sl = &sslines[linearray[k]];

				for (m = 0; m < sl->num; m++)
				{
					ss = &subsectors[sl->list[m]];

					if (ss->floorsec >= 0 || ss->ceilsec >= 0) goto nextsector;

					sg = &segs[ss->firstline];

					for (n = d = 0; n < ss->linecount; n++, sg++)
					{
						if (sg->frontsector != os && sg->frontsector != sc) goto nextsector;
						if (!sg->backsector) continue;
						if (sg->backsector != os && sg->backsector != sc) goto nextsector;
					}
				}
			}

			for (d = k = 0; k < sc->linecount; k++)
			{
				ld = sc->lines[k];

				if (ld->spflags & SPF_CONTROLLINE) continue;

				if (ld->frontsector == ld->backsector)
				{
					if (!M_FitBox(ld->bbox, box)) linearray[d++] = ld - lines;
				}
			}

			if (!d) goto nextsector;

			hitssecptr = 0;
			ss_validcount++;

			for (k = 0; k < d; k++)
			{
				sl = &sslines[linearray[k]];

				for (m = 0; m < sl->num; m++)
				{
					ss = &subsectors[sl->list[m]];

					if (ss->sector == os) continue;
					if (ss->sector != sc) goto nextsector;
					if (ss->ssvalidcount == ss_validcount) continue;

					ss->ssvalidcount = ss_validcount;
					hitssec[hitssecptr++] = sl->list[m];
				}
			}

			for (k = 0; k < hitssecptr; k++)
			{
				SetOverride(&subsectors[hitssec[k]], os - sectors, 49021);
			}
		}
nextsector:;
	}
}