//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_modelbox.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//  Door lines changed to 3D line types to get over limitation
//  of infinite height blocking
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
//
// DESCRIPTION: sets boundary lines with a four sided model identified
//              in a definition file with the flag MF2_HULLCOLLIDE.
//              These are then used with the doom logical code. 
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "con_console.h"
#include "dd_defns.h"
#include "m_winmem.h"
#include "m_misc.h"
#include "p_maputil.h"
#include "p_data.h"
#include "r_data.h"
#include "r_main.h"
#include "r_model.h"
#include "s_sfx.h"
#include "doom/p_local.h"
#include "r3d_modelbox.h"
#include "r3d_compat.h"
#include "r3d_utils.h"

// MACROS ------------------------------------------------------------------

#define MB_SET			0x0001
#define MB_CLOSING		0x0002
#define MB_ACTIVE		0x0004
#define MB_SUSPENDED	0x0008
#define MB_SWING		0x0010

// TYPES -------------------------------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

RDBOOL		R3D_amodeldoors;
int			modelboxcnt;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

#pragma pack(2)

typedef struct {
	thid_t		mobjid;
	line_t		cline[4];
	line3d_t	l3d[4];
	vertex_t	ownv[8];
	vertex_t	start;
	fixed_t		boxcentre[2];
	fixed_t		offset;
	fixed_t		old_posx, old_posy;
	fixed_t		new_posx, new_posy;
	fixed_t		c, s, x, y;
	fixed_t		mbox[4];
	int			vidp[4];
	short		mbflags;
} modelbox_t;

#pragma pack()

typedef struct {
	int		*bbox;
	mobj_t	*doorcrunch[32];
	line_t	*doorline[4];
	fixed_t	changed[2];
	RDBOOL	opening;
	int		numdoorcrunch;
	int		base;
	int		top;
	int		xl, xh, yl, yh;
} doorparms_t;

modelbox_t	*modelboxptr[MAXMODELBOX];

// CODE --------------------------------------------------------------------

//===========================================================================
//  Set any Things designated as a model box
//===========================================================================
void MB_Init(RDBOOL load)
{
	thinker_t	*th;
	mobj_t		*mo;

	if (modelboxcnt)
		W_FreeALL(MBOXID);

	memset(modelboxptr, 0, sizeof(modelboxptr));	// For debug purposes
	modelboxcnt = 0;
	R3D_amodeldoors = false;

	if (!R3D_ScriptedMap) return;

	for (th = thinkercap.next; th != &thinkercap; th = th->next)
	{
		if (   th->function == (void (__cdecl *)(void))P_MobjThinker
			&& ((mo = (mobj_t*)th)->flags2 & MF2_HULLCOLLIDE
			|| (load && mo->flags3 & MF3_INTERREACT))
			&& mo->flags3 & MF3_AMODEL)
		{
			// Remember current state
			state_t *st  = mo->state;
			int		tics = mo->tics;

			// Check that the mo's modelboxnum is unset
			if (mo->modelboxnum)
				Con_ErrorD("MB_Init: multiple reference");

			// Set starting conditions
			mo->state  = states + mo->info->spawnstate;
			mo->frame  = mo->state->frame;
			mo->tics   = 0;

			// Init in starting position
			MB_Update(mo);

			mo->flags &= ~MF_SHOOTABLE;	// Ensure unset else will be selected by P_BlockStandardThingsIterator
			mo->flags2 |= MF2_NOOCCLUDE;

			if (mo->flags2 & MF2_HULLCOLLIDE)
			{
				if (tics || mo->state != st)
				{
					mo->state = st;
					mo->frame = st->frame;
					mo->tics  = tics;

					// Set current position
					MB_Update(mo);
				}
			}
		}
	}
}

//-------------------------------------
// checkDoorPosition
//	Routine to test interaction between a Thing and a model door

static RDBOOL PIT_CheckDoorThing(mobj_t* thing, doorparms_t *dpt)
{
	fixed_t		blockdist, box[4];
	int			i;
	RDBOOL		type, ret = true;
	RDBOOL		untested;
	
	if (thing->modelboxnum > 0) return true;
	if (thing->state - states == S_GIBS) return true;

	if ( !(thing->flags & (MF_SPECIAL|MF_DROPPED)) )
	{
		//	|| P_IsCamera(thing)	democam
		if ( !(thing->flags & (MF_SOLID|MF_CORPSE))
			|| thing->flags & MF_NOCLIP
			|| thing->flags2 & MF2_DONTDRAW
			|| thing->flags3 & (MF3_AMODEL|MF3_STATIC))
		{
			return true;	// self, don't care or active door
		}
	}

	if (thing->z > dpt->top || thing->z + thing->height < dpt->base)
		return true;

	blockdist = thing->info->radius;

	if ( (type = (thing->flags & (MF_SHOOTABLE|MF_CORPSE)) > 0) )
		blockdist += 4*FRACUNIT;

	if (   (box[BOXRIGHT] = thing->x + blockdist) < dpt->bbox[BOXLEFT ]
		|| (box[BOXLEFT ] = thing->x - blockdist) > dpt->bbox[BOXRIGHT]
		|| (box[BOXTOP  ] = thing->y + blockdist) < dpt->bbox[BOXBASE ]
		|| (box[BOXBASE ] = thing->y - blockdist) > dpt->bbox[BOXTOP  ])
	{
		return true;
	}

	for (i = 0, untested = true; i < 4; i++)
	{
		line_t *ld = dpt->doorline[i];

		if (   box[BOXRIGHT] <= ld->bbox[BOXLEFT ]
			|| box[BOXLEFT ] >= ld->bbox[BOXRIGHT]
			|| box[BOXTOP  ] <= ld->bbox[BOXBASE ]
			|| box[BOXBASE ] >= ld->bbox[BOXTOP  ]
			|| P_BoxOnLineSide (box, ld) != -1)
		{
			continue;
		}

		// v220-3 add door carry
		if (untested)
		{
			if (   thing->z == dpt->top
				&& !(thing->flags & (MF_FLOAT|MF_SKULLFLY))
				&& ((dpt->changed[0] < 0 && thing->momx > dpt->changed[0])
				||  (dpt->changed[0] > 0 && thing->momx < dpt->changed[0])
				||  (dpt->changed[1] < 0 && thing->momy > dpt->changed[1])
				||  (dpt->changed[1] > 0 && thing->momy < dpt->changed[1])) )
			{
				thing->momx += dpt->changed[0]>>3;
				thing->momy += dpt->changed[1]>>3;
				return true;
			}
			untested = false;
		}

		if (dpt->opening)
		{
			// Opening slide doors don't crunch things
			return true;
		}

		if (P_PointOnLineSide(thing->x, thing->y, ld))
			continue;

		if (thing->flags & (MF_CORPSE|MF_SPECIAL|MF_DROPPED))
			dpt->doorcrunch[dpt->numdoorcrunch++] = thing;
		else if (type)
			ret = false;

		break;
	}

	return ret;
}

// Thing must have been validated as a door thing prior to calling
static RDBOOL checkDoorPosition(mobj_t* thing, doorparms_t *dpt)
{
	int		bx, by;
	int		i, res = true;

	validcount++;

	for (bx = dpt->xl ; bx <= dpt->xh ; bx++)
	{
		for (by = dpt->yl ; by <= dpt->yh ; by++)
		{
			if (!P_BlockThingsIterator(bx, by, (int (__cdecl *)(struct mobj_s *,void *))PIT_CheckDoorThing, dpt))
			{
				res = false;
				break;
			}
		}
	}

	// door carry test only
	if (dpt->opening) return true;
	
	for (i = 0; i < dpt->numdoorcrunch; i++)
	{
		mobj_t *mo = dpt->doorcrunch[i];

		if (mo->flags & (MF_DROPPED|MF_SPECIAL))
		{
			P_RemoveMobj(mo, 4);
			res = true;
		}
		else if (mo->health > 0)
		{
			P_DamageMobj(mo, NULL, NULL, mo->health, 4);	// kill object
			res = true;
		}
		else if (mo->state - states != S_GIBS)
		{
			// crunch bodies to giblets
			P_SetMobjState (mo, S_GIBS);

			mo->height = 0;
			mo->radius = 0;
			mo->flags &= ~MF_SOLID;
			res = true;
		}
	}

	return res;
}

static void subfn_DoorThinker(mobj_t* thing, int num)
{
	modelbox_t	*mb = modelboxptr[num];
	doorparms_t	dpt;

	//-------------------------------------

	// v220-33 Height may need to be restored
	thing->height = mb->cline[0].d3d->offset3D + mb->cline[0].d3d->totalheight;

	// R3D - check modelbox doors
	if (thing->tics > 0)
	{
		int		i;

		thing->tics--;
		MB_Update(thing);

		mb->mbflags |= MB_ACTIVE;

		dpt.bbox = MB_GetBox(num, dpt.changed);

		dpt.numdoorcrunch = 0;

		for (i = 0; i < 4; i++) dpt.doorline[i] = MB_GetLine(num, i, NULL);

		// all sides should have the same height and offset
		dpt.base = dpt.doorline[0]->frontsector->floorheight + dpt.doorline[0]->d3d->offset3D;
		dpt.top  = dpt.base + dpt.doorline[0]->d3d->totalheight;

		// a sliding door that is opening cannot be blocked
		dpt.opening = !(mb->mbflags & (MB_CLOSING|MB_SWING));

		M_getBlockOffsets(dpt.bbox, MAXRADIUS, &dpt.xl, &dpt.xh, &dpt.yl, &dpt.yh);

		if (!checkDoorPosition(thing, &dpt))
		{
			thing->tics++;
			MB_Update(thing);

			if (mb->mbflags & MB_SWING)
			{
				// prevent door going partially into the blocking Thing
				if (   thing->tics < thing->state->tics
					&& thing->state->tics - thing->tics < 4
					&& !checkDoorPosition(thing, &dpt))
				{
					thing->tics++;
					MB_Update(thing);
				}
			}

			mb->mbflags |= MB_SUSPENDED;

			// Suspend any sound
			Sfx_SuspendChannel(thing->info->attacksound, thing, true);
		}
		else
		{
			if (mb->mbflags & MB_SUSPENDED)
			{
				mb->mbflags &= ~MB_SUSPENDED;
				// Resume any sound
				Sfx_SuspendChannel(thing->info->attacksound, thing, false);
			}
		}
		return;
	}

	if (thing->tics < 0)
	{
		if (mb->mbflags & MB_ACTIVE)
		{
			if (mb->mbflags & MB_CLOSING)
				mb->offset = 0;

			mb->mbflags &= ~(MB_CLOSING|MB_ACTIVE);

			if (mb->mbflags & MB_SUSPENDED)
			{
				mb->mbflags &= ~MB_SUSPENDED;
				// Resume any sound
				Sfx_SuspendChannel(thing->info->attacksound, thing, false);
			}
		}
	}

	if (!thing->tics)
	{
		thing->tics--;
	}
}

//===========================================================================
// MB_RunModelBoxThinkers
//===========================================================================
void MB_RunModelBoxThinkers()
{
	int i;

	for (i = 0; i < modelboxcnt; i++)
	{
		mobj_t *mo = SetReferencedMobj(modelboxptr[i]->mobjid);

		if (mo && !(mo->flags3 & MF3_INTERREACT) )
			subfn_DoorThinker(mo, i);
	}
}

//===========================================================================
// MB_GetMBmo		added v220-33
//===========================================================================
mobj_t *MB_GetMBmo(int num)
{
	if (num < modelboxcnt)
		return (mobj_t*)SetReferencedMobj(modelboxptr[num]->mobjid);
	return NULL;
}

//===========================================================================
// MB_GetBox
//===========================================================================
fixed_t *MB_GetBox(int num, int *changed)
{
	if (num < 0 || num >= modelboxcnt) return NULL;

	if (changed)
	{
		changed[0] = modelboxptr[num]->new_posx - modelboxptr[num]->old_posx;
		changed[1] = modelboxptr[num]->new_posy - modelboxptr[num]->old_posy;
	}

	return modelboxptr[num]->mbox;
}

static RDBOOL checkMBox(int *b1, int *b2, line_t *dl)
{
	return b1[BOXRIGHT] <= b2[BOXLEFT ]
		|| b1[BOXLEFT ] >= b2[BOXRIGHT]
		|| b1[BOXTOP  ] <= b2[BOXBASE ]
		|| b1[BOXBASE ] >= b2[BOXTOP  ]
		|| P_BoxOnLineSide(b1, dl) != -1;
}

//===========================================================================
// MB_GetLine
//===========================================================================
line_t *MB_GetLine(int num, int lptr, int *box)
{
	if (num >= 0 && num < modelboxcnt)
	{
		line_t *dl = &modelboxptr[num]->cline[lptr];

		if (!box || !checkMBox(box, dl->bbox, dl))
			return dl;
	}
	else
	{
		Con_Error("Program error [5]: MB_GetLine %i [%i]", num, modelboxcnt);
	}

	return NULL;
}

//===========================================================================
// MB_TestDoor
//===========================================================================
RDBOOL	MB_TestDoor(int num)
{
	mobj_t	*th = gdt.tmthing;
	int		i, ht = th->z - th->subsector->sector->floorheight;
	RDBOOL	untested = true;

	for (i = 0; i < 4; i++)
	{
		line_t *dl = &modelboxptr[num]->cline[i];

		if (checkMBox(gdt.tmbbox, dl->bbox, dl) )
		{
			continue;
		}

		if (untested)
		{
			int top = dl->backsector->floorheight + dl->d3d->offset3D + dl->d3d->totalheight;

			if (th->z >= top)
			{
				if (top > gdt.tmfloorz)
					gdt.tmdropoffz = gdt.tmfloorz = top;

				break;
			}
			untested = false;
		}

		if (P_PointOnLineSide(th->x, th->y, dl))
			continue;

		gdt.blockline = dl;

		return true;
	}

	return false;
}

//===========================================================================
// MB_Remove - removes the amodel just set
//  Only for use with an interreact thing
//===========================================================================
void MB_Remove(mobj_t *thing)
{
	if (   thing->flags3 & MF3_INTERREACT
		&& thing->modelboxnum == modelboxcnt)
	{
		W_Free(modelboxptr[--modelboxcnt], MBOXID, 0);
	}
	else
		Con_ErrorD("Program error: MB_Remove");
}

//===========================================================================
// MB_Update - sets the current vertex values bounding an amodel
//  b220 amodel door support v220-3 modified to use MBOXID
//===========================================================================
void MB_Update(mobj_t *thing)
{
	modeldef_t		*mf;
	modeldef_t		*nextmf;
	submodeldef_t	*smf;
	model_t			*mdl;
	modelbox_t		*mb;

	dmd_frame_t		*frame;
	dmd_frame_t		*nextframe = NULL;

	vertex_t		v[4], vt[4];

	dmd_vertex_t	*vtx, *nextvtx;
	
	float			inv, interpos, endpos;
	float			mVtx[2];

	fixed_t			*mbox;

	int				i, n, mbnum;
	int				numVertices;

	RDBOOL			door_init = false;
	RDBOOL			doorblock;
	RDBOOL			swingtype = (thing->flags3 & MF3_SWINGDOOR) != 0;

	//----------------------------------------------

	interpos = R_CheckModelFor(thing, &mf, &nextmf);

	if (!mf) return;
		
	doorblock = (thing->flags2 & MF2_HULLCOLLIDE) > 0;

	if ( (mbnum = thing->modelboxnum - 1) < 0)
	{
		if (modelboxcnt >= MAXMODELBOX)
			Con_Error("MB_Update: limit of %i exceeded", MAXMODELBOX);

		mbnum = modelboxcnt++;
		modelboxptr[mbnum] = (modelbox_t*)W_Calloc(sizeof(modelbox_t), MBOXID, 211);
		thing->modelboxnum = modelboxcnt;	// i.e. mbnum + 1
		mb = modelboxptr[mbnum];
		mb->mobjid = thing->thinker.id;

		for (i = n = 0; i < 4; i++)
		{
			mb->cline[i].frontsector = mb->cline[i].backsector = thing->subsector->sector;
			mb->cline[i].ml_flags = ML_TWOSIDED;
			mb->cline[i].lnflags2 = LNF2_OWNEQUALTOONE;
			mb->cline[i].specType = R3D_40;
			mb->cline[i].spflags = SPF_NOTEX;
			mb->cline[i].xtraflags |= XT_MBLINE;
			mb->cline[i].v1 = &mb->ownv[n++];
			mb->cline[i].v2 = &mb->ownv[n++];

			if (!doorblock)
			{
				mb->l3d[i].offset3D = thing->z - thing->subsector->sector->floorheight;
				mb->l3d[i].lineheightlow = mb->l3d[i].totalheight = thing->height;
			}

			mb->l3d[i].scale3Dlow = FRACUNIT;
			mb->l3d[i].scale3Dmid = FRACUNIT;
			mb->cline[i].d3d = &mb->l3d[i];
		}

		door_init = true;

		R3D_amodeldoors = true;
	}
	else
	{
		if (!doorblock)
			Con_Error("MB_Update: Wrong type");

		mb = modelboxptr[mbnum];

		if (thing->tics == -1 && thing->health > 0)
		{
			if (mb->mbflags & MB_SET) return;
			mb->mbflags |= MB_SET;
		}
		else
		{
			mb->mbflags &= ~MB_SET;
		}
	}

	mbox = mb->mbox;

	smf = &mf->sub[0];
	mdl = modellist[smf->model];

	numVertices = mdl->info.numVertices;
	frame = R_GetVisibleFrame(mf, 0, thing->thinker.id, 1);

	if (door_init)
	{
		angle_t	angle = thing->angle >> ANGLETOFINESHIFT;
		float	minz, maxz;
		int		snum = thing->state - states;
		RDBOOL	err = false;

		//-----------------------------------------------------

		if (numVertices != 8 && ge_inf.devmap)
			pr_OFinsert("WARNING: Door model spawned from mapnum %i has %i vtxs not 8\n", MAPNUM(thing), numVertices);

		if (doorblock && (swingtype || strstr(defs.states[snum].id, "swing")) )	// "swing" for backward compatibility
		{
			mb->mbflags |= MB_SWING;
		}

		mb->c = finecosine[angle];
		mb->s = finesine[angle];
		mb->x = thing->x + (fixed_t)(mf->offset[VX] * FRACUNIT);
		mb->y = thing->y + (fixed_t)(mf->offset[VZ] * FRACUNIT);

		for (i = 0, vtx = frame->vertices; i < numVertices; i++, vtx++)
		{
			if (!i || vtx->vertex[VZ] < minz) minz = vtx->vertex[VZ];
			if (!i || vtx->vertex[VZ] > maxz) maxz = vtx->vertex[VZ];
		}

		// b220-1 set the 3D line height
		n = (fixed_t)((maxz - minz) * mf->scale[VY]) * FRACUNIT;

		// All four lines should have the same height
		for (i = 0; i < 4; i++)
		{
			if (doorblock)
				mb->cline[i].d3d->totalheight = n;
		}

		// Check for four lower vertexes
		minz += 0.4f;	// Allow for small errors

		for (i = n = 0, vtx = frame->vertices; i < numVertices; i++, vtx++)
		{
			if (vtx->vertex[VZ] < minz)
			{
				if (n > 3) { err = true; break; }
				mb->vidp[n++] = i;
			}
		}

		if (err || n != 4)
			Con_Error("Non-compliant interactive model spawned from Thing mapnum %i\n", MAPNUM(thing));

		mbox[BOXLEFT] = mbox[BOXBASE ] = DDMAXINT;
		mbox[BOXTOP ] = mbox[BOXRIGHT] = DDMININT;
	}

	// Check for interpolation
	if (mf->interrange[0] == 0 && mf->interrange[1] == 1.0f)
	{
		if (mf->internext) 
			endpos = mf->internext->intermark;
		else
			endpos = 1.0f;

		// Scale interpos
		interpos = (interpos - mf->intermark) / (endpos - mf->intermark);		
	}

	if (nextmf && nextmf->sub[0].model == smf->model)
	{
		nextframe = R_GetVisibleFrame(nextmf, 0, thing->thinker.id, 2);
		nextvtx = nextframe->vertices;
	}

	inv = 1.0f - interpos;

	for (i = 0; i < 4; i++)
	{
		vtx = frame->vertices + mb->vidp[i];

		mVtx[0] = vtx->vertex[VX];
		mVtx[1] = vtx->vertex[VY];

		if (!door_init && nextframe)
		{
			// Calculate interpolated vertices for the bounding lines
			nextvtx = nextframe->vertices + mb->vidp[i];
			mVtx[0] = mVtx[0] * inv + nextvtx->vertex[VX] * interpos;
			mVtx[1] = mVtx[1] * inv + nextvtx->vertex[VY] * interpos;
		}

		vt[i].x = (fixed_t)(mVtx[0] * mf->scale[VX] * FRACUNIT);
		vt[i].y = (fixed_t)(mVtx[1] * mf->scale[VZ] * FRACUNIT);

		if (door_init)
		{
			// Set the model space vertex box
			if (vt[i].x < mbox[BOXLEFT ]) mbox[BOXLEFT ] = vt[i].x;
			else
			if (vt[i].x > mbox[BOXRIGHT]) mbox[BOXRIGHT] = vt[i].x;

			if (vt[i].y < mbox[BOXBASE ]) mbox[BOXBASE ] = vt[i].y;
			else
			if (vt[i].y > mbox[BOXTOP  ]) mbox[BOXTOP  ] = vt[i].y;
		}
	}

	// transform vertexes to world space
	for (i = 0; i < 4; i++)
	{
		v[i].x = mb->x + FixedMul(vt[i].x, mb->c) - FixedMul(vt[i].y, mb->s);
		v[i].y = mb->y + FixedMul(vt[i].y, mb->c) + FixedMul(vt[i].x, mb->s);

		if (!i)
		{
			mbox[BOXLEFT] = mbox[BOXRIGHT] = v[0].x;
			mbox[BOXTOP ] = mbox[BOXBASE ] = v[0].y;
		}
		else
		{
			if (v[i].x < mbox[BOXLEFT ]) mbox[BOXLEFT ] = v[i].x;
			else
			if (v[i].x > mbox[BOXRIGHT]) mbox[BOXRIGHT] = v[i].x;

			if (v[i].y < mbox[BOXBASE ]) mbox[BOXBASE ] = v[i].y;
			else
			if (v[i].y > mbox[BOXTOP  ]) mbox[BOXTOP  ] = v[i].y;
		}
	}

	if (doorblock)
	{
		if (door_init)
		{
			mb->boxcentre[0] = meanV(mbox[BOXLEFT], mbox[BOXRIGHT]);
			mb->boxcentre[1] = meanV(mbox[BOXTOP ], mbox[BOXBASE ]);
			mb->offset = 0;
			mb->old_posx = mb->new_posx = mb->boxcentre[0];
			mb->old_posy = mb->new_posy = mb->boxcentre[1];
		}
		else
		{
			mb->old_posx = mb->new_posx;
			mb->old_posy = mb->new_posy;

			mb->new_posx = meanV(mbox[BOXLEFT], mbox[BOXRIGHT]);
			mb->new_posy = meanV(mbox[BOXTOP ], mbox[BOXBASE ]);

			if (!(mb->mbflags & MB_CLOSING))
			{
				fixed_t	offset = mb->offset;	// remember current offset
				double dx = mb->boxcentre[0] - mb->new_posx;
				double dy = mb->boxcentre[1] - mb->new_posy;

				mb->offset = (fixed_t)sqrt(dx * dx + dy * dy);

				if (mb->offset < offset)
					mb->mbflags |= MB_CLOSING;
			}
		}
	}

	// Set up the vertexes, flags and box limits
	for (i = 0; i < 4; i++)
	{
		line_t		*ld = &mb->cline[i];
		vertex_t	*v1 = ld->v1, *v2 = ld->v2;
		int			n = (i + 1) % 4;

		v1->x = v[i].x;
		v1->y = v[i].y;
		v2->x = v[n].x;
		v2->y = v[n].y;

		ld->dx = v2->x - v1->x;
		ld->dy = v2->y - v1->y;

		// Smallest dimension must be at least 1 FRACUNIT
		// to prevent an ofv with the doom logical code
		if (!ld->dx && abs(ld->dy) < FRACUNIT)
			ld->dy = ld->dy < 0 ? -FRACUNIT : FRACUNIT;
		else if (!ld->dy && abs(ld->dx) < FRACUNIT)
			ld->dx = ld->dx < 0 ? -FRACUNIT : FRACUNIT;

		if (!ld->dx)
			ld->slopetype = ST_VERTICAL;
		else if (!ld->dy)
			ld->slopetype = ST_HORIZONTAL;
		else
		{
			if (FixedDiv (ld->dy , ld->dx) > 0)
				ld->slopetype = ST_POSITIVE;
			else
				ld->slopetype = ST_NEGATIVE;
		}

		if (v1->x < v2->x)
		{
			ld->bbox[BOXLEFT ] = v1->x;
			ld->bbox[BOXRIGHT] = v2->x;
		}
		else
		{
			ld->bbox[BOXLEFT ] = v2->x;
			ld->bbox[BOXRIGHT] = v1->x;
		}

		if (v1->y < v2->y)
		{
			ld->bbox[BOXBASE] = v1->y;
			ld->bbox[BOXTOP ] = v2->y;
		}
		else
		{
			ld->bbox[BOXBASE] = v2->y;
			ld->bbox[BOXTOP ] = v1->y;
		}
	}
}
