//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_movsec.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// DESCRIPTION: sets up the movsec array
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "r3d_MAINspec.h"
#include "r3d_compat.h"
#include "doom\p_local.h"

// MACROS ------------------------------------------------------------------

// Used with movflags
#define MOVISVALID		0x01
#define MOVINVALID		0x02
#define MOVOTHER		0x04	// Any function other than P_FindLowestFloorSurrounding

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

void Check_MovSpecial(line_t *ld);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		*movsec_list;	// $smoothplane

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static byte	*movflags;
static byte *tagfield;


// CODE --------------------------------------------------------------------

void LS_BuildMovsecList()
{
	int i, s;

	for (i = 0, s = 1; i < numsectors; i++)
	{
		if (sectors[i].specFlags & SF_MOVSECTOR)
			movsec_list[s++] = i;
	}

	movsec_list[0] = s - 1;

	M_Free(movflags);
	movflags = NULL;
}

// This ensures the sector's walls will be updated in the renderer
void subfn_CheckMovSet(line_t *ld)
{
	int secnum = -1;

	while ((secnum = P_FindSectorFromTag(ld->tag,secnum)) >= 0)
	{
		sectors[secnum].specFlags |= SF_MOVSECTOR;
	}
}

// ===== LS_IdentifyMovsecList - done at the start

// Set up mov_sector list
void LS_IdentifyMovsecList()
{
	sector_t	*sc;
	line_t		*ld;
	int			i, k;

	// v220-5 add special cases of timed doors
	for (i = 0; i < numsectors; i++)
	{
		if (sectors[i].special == 10 || sectors[i].special == 14)
			sectors[i].specFlags |= SF_MOVCEIL;
	}

	// check for end specials
	{
		line_t	temp;

		temp.tag = 666;
		subfn_CheckMovSet(&temp);
		temp.tag = 667;
		subfn_CheckMovSet(&temp);
	}

	movflags = (byte*)M_Calloc(numsectors, 280);
	tagfield = (byte*)M_Calloc(65536, 280);

	SET_countsec(0);

	for (i = 0; i < numlines; i++)
		if (lines[i].tag) tagfield[(ushort)lines[i].tag] = true;

	for (i = 0; i < numsectors; i++)
	{
		if (sectors[i].tag)
		{
			tagfield[(ushort)sectors[i].tag] = true;

			if (   ge_inf.addon_flags & AF_COD
				&& sectors[i].tag >= 668
				&& sectors[i].tag <= 681)
			{
				sectors[i].specFlags |= SF_MOVFLOOR;
			}
		}
	}

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->special && (ld->tag || ld->backsector))
			Check_MovSpecial(ld);
	}

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (movflags[i] & MOVINVALID && !(movflags[i] & MOVISVALID))
			sc->specFlags |= SF_MOVINVALID;

		if (sc->specFlags & SF_MOVFIXED)
		{
			if (movflags[sc - sectors] & MOVOTHER)
			{
				sc->specFlags &= ~SF_MOVFIXED;
				continue;
			}

			for (k = 0; k < sc->linecount; k++)
			{
				line_t	 *li = sc->lines[k];
				sector_t *s = li->frontsector == sc ? li->backsector : li->frontsector;

				if (!s) continue;

				if (s->specFlags & SF_MOVFLOOR)
				{
					sc->specFlags &= ~SF_MOVFIXED;
					s->specFlags &= ~SF_MOVFIXED;
				}
			}

			if (sectors[i].specFlags & SF_MOVFIXED)
			{
				sectors[i].specFlags &= ~(SF_MOVFLOOR|SF_MOVLOWER|SF_MOVINVALID);
			}
		}
	}

	M_Free(tagfield);
}

// Checks for single mid tex line in moving sector
void LS_Check_MovSpecial_Valid()
{
	sector_t *sec = sectors, *osec;
	line_t *ld;
	int i, k, valid;

	for (i = 0; i < numsectors; i++, sec++)
	{
		if (!secdone[i] && sec->specFlags & SF_MOVFLOOR)
		{
			for (osec = NULL, valid = 0, k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (ld->spflags & SPF_CONTROLLINE)
					continue;

				if (!(ld->ml_flags & ML_TWOSIDED))
				{
					valid = 0;
					break;
				}

				if (ld->frontsector == ld->backsector)
				{
					valid++;
					if (valid > 1) break;
					if (!(ld->spflags & SPF_NOULTEX) ||
						ld->spflags & SPF_NOTEX)
					{
						valid = 0;
						break;
					}
					continue;
				}
				if (ld->backsector != sec ||
					!(ld->spflags & SPF_NOTEX))
				{
					valid = 0;
					break;
				}
				if (!osec) osec = ld->frontsector;
				if (osec != ld->frontsector)
				{
					valid = 0;
					break;
				}
			}
			if (!osec || valid != 1)
				continue;

			for (k = 0; k < sec->linecount; k++)
			{
				if (!(sec->lines[k]->spflags & SPF_CONTROLLINE))
				{
					Setspec(sec->lines[k], osec - sectors, 12002);

					if (!origTEX[sec->lines[k] - lines])
					{
						Settype(sec->lines[k], R3D_01, 12002);
						SetLFG(sec->lines[k], LFG_FIXED|LFG_LINESSECSET|LFG_NOCHANGE, 12001);
					}
				//	Settype(sec->lines[k], R3D_32, 12002);
				}
			}
			Setsecdone2(sec, 12002);
		}
	}
}

// Checks specials type 83 and 91
void LS_Check_MovSpecial_X()
{
	int i, k;

	SET_countsec(0);

	for (i = 0; i < numlines; i++)
	{
		if (lines[i].special == 83 || lines[i].special == 91)
		{
			int s;
			for (s = -1; (s = P_FindSectorFromTag(lines[i].tag, s)) >= 0;)
			{
				// For plats activated by line->special 91 they must be crossing
				// a TYF_MIDTEXLINE to qualify for analysis in r3d_contained.c
				sector_t *sec = &sectors[s];

				if (countsec[s]) continue;
				countsec[s] = 1;

				if (    sec->specFlags & SF_SECCLOSED
					&& !(sec->specFlags & SF_MOVINVALID))
				{
					for (k = 0; k < sec->linecount; k++)
					{
						line_t *ld = sec->lines[k];

						if (   LSS(ld)
							&& !ld->special
							&& ld->spflags & SPF_NOTEX
							&& Vtxs2(ld)
							&& !(ld->spflags & SPF_CONTROLLINE)
							&& subfn_FindLineCrossingMidTex(ld, LC_NORMAL) != NULL)
						{
							sectors[s].specFlags &= ~SF_MOVTYPE_X;
							break;
						}
					}
				}
			}
		}
	}
}

fixed_t getHeight(sector_t *sec, int Targ, int Dirn)
{
	int ht, sfht = sec->floorheight, dir = Dirn? 1 : -1;

	// find the destination height
	switch(Targ)
	{
	case FtoHnF:
		ht = P_FindHighestFloorSurrounding(sec);
		sec->siflags |= SIF_MOVHIGH1;
		break;
	case FtoLnF:
		ht = P_FindLowestFloorSurrounding(sec);
		if (ht == sec->floorheight) sec->specFlags |= SF_MOVFIXED;
		break;
	case FtoNnF:
		if (Dirn)	// i.e. either set or 0
		{
			ht = P_FindNextHighestFloor(sec, sfht);
			sec->siflags |= SIF_MOVHIGH2;
		}
		else
		{
			ht = P_FindNextLowestFloor (sec, sfht);
		}
		break;
	case FtoLnC:
		ht = P_FindLowestCeilingSurrounding(sec);
		break;
	case FtoC:
		ht = sec->ceilingheight;
		break;
	case FbyST:
		ht = targetMAX(sfht, dir * P_FindShortestTextureAround(sec - sectors));
		break;
	case Fby24:
		ht = sfht + dir * 24*FRACUNIT;
		break;
	case Fby32:
		ht = sfht + dir * 32*FRACUNIT;
		break;
	default:
		ht = sfht;
		break;
	}

	return ht;
}

static RDBOOL getTrigger(line_t *ld, int offset)
{
	int val  = (unsigned)ld->special - offset;
	int Trig = (val & TriggerType) >> TriggerTypeShift;
	return (Trig == PushOnce || Trig == PushMany);
}

static void chkGenFloorBase(line_t *ld)
{
	int val = ld->special;
	int Targ = (val & FloorTarget   ) >> FloorTargetShift;
	int	Dirn = (val & FloorDirection) >> FloorDirectionShift;
	int s, ht, manual = false;

	if (getTrigger(ld, 0))
	{
		if (!ld->backsector) return;

		s = ld->backsector - sectors;
		manual = true;
		goto manual_floor;
	}

	if (!ld->tag) return;
	
	for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
	{
manual_floor:;
		
		ht = getHeight(&sectors[s], Targ, Dirn);

		if (Targ != FtoLnF) movflags[s] |= MOVOTHER;

		// v220-5 revised
		if (   ( Dirn && sectors[s].floorheight > ht)
			|| (!Dirn && sectors[s].floorheight < ht))
		{
			movflags[s] |= MOVINVALID;
		}
		else
		{
			movflags[s] |= MOVISVALID;
			if (!Dirn) sectors[s].specFlags |= SF_MOVLOWER;
		}

		sectors[s].specFlags |= SF_MOVFLOOR;
			
		if (manual) return;
	}
}

static RDBOOL validSkySector(uint v, sector_t *sec)
{
	int i;

	if (   sec->ceilingpic != skyflatnum
		|| !(v & CeilingChange))
	{
		return false;
	}

	for (i = 0; i < sec->linecount; i++)
	{
		line_t	 *ld = sec->lines[i];

		sector_t *os = ld->frontsector == sec ? ld->backsector : ld->frontsector;

		if (os && os->ceilingpic == skyflatnum) return true;
	}
	return false;
}

static void setGeneralCeil(line_t *ld, int offset)
{
	uint		value = (uint)ld->special - GenCeilingBase;
	int Crsh = (value & CeilingCrush    ) >> CeilingCrushShift;
	int Dirn = (value & CeilingDirection) >> CeilingDirectionShift;


	if (getTrigger(ld, offset))
	{
		if (ld->backsector)
		{
			ld->backsector->specFlags |= SF_MOVCEIL;

			if (!Dirn && Crsh)	// lowerAndCrush type
				ld->backsector->specFlags2 |= SF2_CEILNOSLOW;
		}
	}

	if (ld->tag)
	{
		int s;

		for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
		{
			sectors[s].specFlags |= SF_MOVCEIL;

			if (!Dirn && Crsh)
				sectors[s].specFlags2 |= SF2_CEILNOSLOW;

			if ( !(sectors[s].specFlags2 & SF2_MOVSKYCEIL)
				&& validSkySector((uint)ld->special - offset, &sectors[s]))
			{
				sectors[s].specFlags2 |= SF2_MOVSKYCEIL;
			}
		}
	}
}

static void setGeneralFloor(line_t *ld, int offset)
{
	if (getTrigger(ld, offset))
	{
		if (ld->backsector)
		{
			ld->backsector->specFlags |= SF_MOVFLOOR;
			movflags[ld->backsector - sectors] |= MOVOTHER;
		}
	}

	if (ld->tag)
	{
		int s;

		for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
		{
			sectors[s].specFlags |= SF_MOVFLOOR;
		}
	}
}

static void setGeneralStair(line_t *ld)
{
	int			secnum = -1;
	int			osecnum;
	int			newsecnum;
	int			texture;
	int			i, rtn;
	RDBOOL		manual, mov_valid;
	
	sector_t*	sec;
	sector_t*	tsec;
	
	// v220-5 fix value (was int)
	uint		value = (uint)ld->special - GenStairsBase;
	int			Igno = (value & StairIgnore) >> StairIgnoreShift;

	//--------------------------------------
	
	rtn = 0;
	
	// check if a manual trigger, if so only do the sector on the backside
	manual = false;

	if (getTrigger(ld, GenStairsBase))
	{
		if (!(sec = ld->backsector)) return;

		secnum = sec - sectors;
		manual = true;
		goto manual_stair;
	}
	
	if (!ld->tag) return;
	
	// if not manual do all sectors tagged the same as the line
	while ((secnum = P_FindSectorFromTag(ld->tag, secnum)) >= 0)
	{		
		sec = &sectors[secnum];
manual_stair:;	   
						
		osecnum = secnum;

		sec->specFlags |= SF_MOVFLOOR;
		movflags[secnum] |= MOVOTHER;
		texture = sec->floorpic;

		// Find next sector to raise
		// 1.	  Find 2-sided line with same sector side[0]
		// 2.	  Other side is the next sector to raise
		do
		{
			for (i = 0, mov_valid = false; i < sec->linecount; i++)
			{
				if ( !((sec->lines[i])->backsector) ) continue;
				
				tsec = (sec->lines[i])->frontsector;
				
				if (secnum != tsec-sectors) continue;
				
				tsec = (sec->lines[i])->backsector;
				newsecnum = tsec - sectors;
				
				if (!Igno && tsec->floorpic != texture)
					continue;
		
				// Check if already set
				if (tsec->specFlags & SF_MOVFLOOR)
					continue;

				sec = tsec;

				secnum = newsecnum;
				
				sec->specFlags |= SF_MOVFLOOR;
				movflags[secnum] |= MOVOTHER;

				mov_valid = true;
				break;
			}

		} while(mov_valid);

		if (manual) return;

		secnum = osecnum;
	}
}

static int nextFreeTag()
{
	static ushort i = 0;

	while (++i)
	{
		if (!tagfield[i]) { tagfield[i] = true; return i; }
	}
	return 0;
}

static void AddLineAndSectorTag(line_t *ld, int flags)
{
	if (!P_CheckTag(ld)) return;	// v220-35 added

	sector_t *s = ld->backsector;

	if (!s->tag)
	{
		int c, i;

		s->tag = nextFreeTag();

		for (c = i = 0; i < s->linecount; i++)
		{
			if (   s->lines[i]->special == ld->special
				&& !s->lines[i]->tag)
			{
				s->lines[i]->tag = s->tag; c++;
			}
		}

		if (c) LS_InitTagLists();	// Update hash table
	}
	else
	{
		s->specFlags |= flags;
		ld->tag = s->tag;
	}
}

RDBOOL lineHasTag(int tag)
{
	int i;

	for (i = 0; i < numlines; i++)
		if (lines[i].tag == tag) return true;
	return false;
}

// b211-0 Extended checks for generalised types added
void Check_MovSpecial(line_t *ld)
{
	int		s, value = (ushort)ld->special;

	if (value <= 0) return;

	if (!ld->tag && !(ld->ml_flags & ML_TWOSIDED)) return;

	if (value >= GenCrusherBase && value < GenLimit)
	{
		// Don't change order
		if (value >= GenFloorBase)
		{
			chkGenFloorBase(ld);
		}
		else if (value >= GenCeilingBase)
		{
			setGeneralCeil(ld, GenCeilingBase);
		}
		else if (value >= GenDoorBase)
		{
			setGeneralCeil(ld, GenDoorBase);
		}
		else if (value >= GenLockedBase)
		{
			setGeneralCeil(ld, GenLockedBase);
		}
		else if (value >= GenLiftBase)
		{
			setGeneralFloor(ld, GenLiftBase);
		}
		else if (value >= GenStairsBase)
		{
			setGeneralStair(ld);
		}
		else	// GenCrusherBase
		{
			setGeneralCeil(ld, GenCrusherBase);
		}
		return;
	}

	switch(value)
	{
	// EV_DoFloor
	// lowerFloor, turboLower (P_FindHighestFloorSurrounding -1)
	case 36:
	case 45:	// v220-13 moved here
	case 70:
	case 71:
	case 19:
	case 83:
	case 98:
	case 102:
		if (!ld->tag)
		{
			AddLineAndSectorTag(ld, SF_MOVFLOOR);
		}

		if (ld->tag)
		{
			for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
			{
				sector_t *sec = &sectors[s];

				sec->specFlags |= SF_MOVFLOOR;

				if (value == 83) sec->specFlags |= SF_MOVTYPE_X;

				movflags[s] |= MOVOTHER;

				if (P_FindHighestFloorSurrounding(sec) > sec->floorheight)
				{
					movflags[s] |= MOVINVALID;
					sec->siflags |= SIF_MOVHIGH1;

					if (value == 36 || value == 70 || value == 71)
						sec->specFlags2 |= SF2_MOV8ABOVEFLR;
				}
				else
				{
					movflags[s] |= MOVISVALID;
					sec->specFlags |= SF_MOVLOWER;
				}
			}
		}
		break;

	// lowerFloorToLowest, lowerAndChange (P_FindLowestFloorSurrounding -1)
	case 23:
	case 38:
	case 60:
	case 82:
	case 84:
		if (!ld->tag)
		{
			AddLineAndSectorTag(ld, SF_MOVFLOOR);
		}

		if (ld->tag)
		{
			int ht;

			for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
			{
				sector_t *sec = &sectors[s];

				if ( (ht = P_FindLowestFloorSurrounding(sec)) > sec->floorheight)
				{
					sec->specFlags |= SF_MOVFLOOR;
					movflags[s] |= MOVINVALID;
				}
				else
				{
					movflags[s] |= MOVISVALID;
					sec->specFlags |= (SF_MOVFLOOR|SF_MOVLOWER);
				}
				
				if (ht == sec->floorheight)
					sec->specFlags |= SF_MOVFIXED;
			}
		}

		break;

	// raiseFloor, raiseFloorCrush (P_FindLowestCeilingSurrounding 1)
	case 5:
    case 24:
	case 55:
	case 56:
	case 64:
	case 65:
	case 91:
	case 94:
	case 101:
		if (!ld->tag)
		{
			AddLineAndSectorTag(ld, SF_MOVFLOOR);
		}

		if (ld->tag)
		{
			for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
			{
				sector_t *sec = &sectors[s];
				sec->specFlags |= SF_MOVFLOOR;

				if (value == 91) sec->specFlags |= SF_MOVTYPE_X;	// $v68 was in wrong place

				movflags[s] |= MOVOTHER;

				if (P_FindLowestCeilingSurrounding(sec) < sec->floorheight)
					movflags[s] |= MOVINVALID;
				else
					movflags[s] |= MOVISVALID;
			}
		}
		break;

// lowerFloorToNearest (P_FindNextLowestFloor -1)
	case 219:
	case 220:
	case 221:
	case 222:
// lowerAndChange
	case 37:
	case 159:
	case 177:
		if (!ld->tag)
		{
			AddLineAndSectorTag(ld, SF_MOVFLOOR|SF_MOVLOWER);
		}

		if (ld->tag)
		{
			for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
			{
				sectors[s].specFlags |= (SF_MOVFLOOR|SF_MOVLOWER);
				movflags[s] |= MOVOTHER;
			}
		}
		break;

// raiseFloorToNearest  (P_FindNextHighestFloor 1)
	case 18:
	case 69:
	case 119:
	case 128:
// raiseFloor24
	case 58:
	case 92:
	case 161:
	case 180:
// raiseFloor24AndChange
	case 59:
	case 93:
	case 160:
	case 179:
// raiseToTexture
	case 30:
	case 96:
	case 158:
	case 176:
// raiseFloorTurbo
	case 129:
	case 130:
	case 131:
	case 132:
// raiseFloor512
	case 140:
	case 142:
	case 147:
	case 178:

// EV_DoPlat
	case 10:	// plat down-wait-up-stay trigger
	case 14:	// door raise in 5 minutes
	case 15:	// Raise Floor 24 and change texture
	case 20:	// Raise Plat next highest floor and change texture
	case 21:	// PlatDownWaitUpStay
	case 22:	// Raise floor to nearest height and change texture
    case 47:	// G1 raise floor to nearest and change texture and type
	case 53:	// Perpetual Platform Raise
	case 62:	// PlatDownWaitUpStay
	case 66:	// Raise Floor 24 and change texture
	case 67:	// Raise Floor 32 and change texture
	case 68:	// Raise Plat to next highest floor and change texture
	case 87:	// Perpetual Platform Raise
	case 88:	// PlatDownWaitUp
	case 95:	// Raise floor to nearest height
	case 120:	// Blazing PlatDownWaitUpStay
	case 121:	// Blazing PlatDownWaitUpStay
	case 122:	// Blazing PlatDownWaitUpStay
	case 123:	// Blazing PlatDownWaitUpStay
	case 143:	// Raise Floor 24 and change
	case 144:	// Raise Floor 32 and change
	case 148:	// Raise Floor 24 and Change
	case 149:	// Raise Floor 32 and Change
	case 162:	// Moving floor min n to max n
	case 163:	// Stop Moving floor
	case 181:	// Moving floor min n to max n
	case 182:	// Stop Moving floor
	case 211:	// Toggle Floor Between C and F Instantly
	case 212:	// Toggle floor between C and F instantly
		if (!ld->tag)
		{
			AddLineAndSectorTag(ld, SF_MOVFLOOR);
		}

		if (ld->tag)
		{
			for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
			{
				sectors[s].specFlags |= SF_MOVFLOOR;
				sectors[s].siflags |= SIF_MOVHIGH2;
				movflags[s] |= MOVOTHER;
			}
		}
		break;

// EV_DoCeiling
	case 6:		// fastCrushAndRaise
	case 25:	// crushAndRaise
	case 40:	// raiseToHighest
	case 41:	// lowerToFloor
	case 43:	// lowerToFloor
	case 44:	// lowerAndCrush
	case 49:	// crushAndRaise
	case 72:	// lowerAndCrush
	case 73:	// crushAndRaise
	case 77:	// fastCrushAndRaise
	case 141:	// silentCrushAndRaise
	case 145:	// lowerToFloor
	case 150:	// silentCrushAndRaise
	case 152:	// lowerToFloor
	case 164:	// fastCrushAndRaise
	case 165:	// silentCrushAndRaise
	case 167:	// lowerAndCrush
	case 183:	// fastCrushAndRaise
	case 184:	// crushAndRaise
	case 185:	// silentCrushAndRaise
	case 187:	// lowerAndCrush
	case 199:	// lowerToLowest
	case 200:	// lowerToMaxFloor
	case 201:	// lowerToLowest
	case 202:	// lowerToMaxFloor
	case 203:	// lowerToLowest
	case 204:	// lowerToMaxFloor
	case 205:	// lowerToLowest
	case 206:	// lowerToMaxFloor
		if (!ld->tag)
		{
			AddLineAndSectorTag(ld, SF_MOVCEIL);
		}

		if (ld->tag)
		{
			for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
			{
				sectors[s].specFlags |= SF_MOVCEIL;
			}
		}
		break;

// Triggers both floor and ceiling
	case 151:
	case 166:
	case 186:

// EV_DoElevator
	case 227:
	case 228:
	case 229:
	case 230:
	case 231:
	case 232:
	case 233:
	case 234:
	case 235:
	case 236:
	case 237:
	case 238:
		if (!ld->tag)
		{
			AddLineAndSectorTag(ld, SF_MOVCEIL|SF_MOVFLOOR);
		}

		if (ld->tag)
		{
			for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
			{
				sectors[s].specFlags |= (SF_MOVCEIL|SF_MOVFLOOR);
				movflags[s] |= MOVOTHER;

				if (value >= 227 && value <= 234)
					sectors[s].specFlags2 |= SF2_ELEVATOR3D;
				else if (value >= 235 && value <= 238)
					sectors[s].specFlags2 |= SF2_ELEVATOR;
			}
		}
		break;

// EV_DoDonut
	case 9:
	case 146:
	case 155:
	case 191:
		{
			sector_t	*s1 = NULL, *s2 = NULL, *s3 = NULL;
			int			i, secnum = -1;

			while ((secnum = P_FindSectorFromTag(ld->tag, secnum)) >= 0)
			{
				s1 = &sectors[secnum];
				
				if (s1->specFlags & SF_MOVFLOOR) continue;

				s1->specFlags |= SF_MOVFLOOR;
				movflags[secnum] |= MOVOTHER;
				
				s2 = getNextSector(s1->lines[0], s1);

				if (!s2 || s2->specFlags & SF_MOVFLOOR) continue;
				
				s2->specFlags |= SF_MOVFLOOR;
				movflags[s2 - sectors] |= MOVOTHER;

				for (i = 0;i < s2->linecount;i++)
				{
					if (!(s2->lines[i]->ml_flags & ML_TWOSIDED) ||
						s2->lines[i]->backsector == s1)
						continue;
					
					s3 = s2->lines[i]->backsector;      // s3 is model sector for changes
					s3->specFlags |= SF_MOVFLOOR;
					movflags[s3 - sectors] |= MOVOTHER;

					break;
				}
			}
		}
		break;

// EV_BuildStairs
	// build8;
	case 7:
	case 8:
	case 256:
	case 258:
	// turbo16
	case 100:
	case 127:
	case 257:
	case 259:
	{
		// v220-3 fixed bug where all stair sectors were not being marked
		sector_t	*sec;
		int			i, secnum;
		int			s = -1,mins = -1;
		int			tag = ld->tag;
		byte		*chksec = (byte*)M_Calloc(numsectors, 0);

		// v220-5 add support for type where line's tag is 0
		if (!tag)
		{
			if (ld->backsector) tag = ld->backsector->tag;
			if (!tag) { M_Free(chksec); return; }
			s = ld->backsector - sectors;
			goto manual_stair;
		}

		// start a stair at each sector tagged the same as the linedef
		while ((s = P_FindSectorFromTagWithLowerBound(tag, s, mins)) >= 0)
		{
manual_stair:;
			sec = &sectors[s];
			secnum = s;

			if (!chksec[s])
			{
				RDBOOL	repeat = true;
				int		texture = sec->floorpic;

				chksec[s] = true;

				sec->specFlags |= SF_MOVFLOOR;
				movflags[s] |= MOVOTHER;

				// Find next sector to raise
				//   1. Find 2-sided line with same sector side[0] (lowest numbered)
				//   2. Other side is the next sector to raise
				//   3. Unless set or different texture terminate loop

				while (repeat)      // repeat until no next step is found
				{
					repeat = false;

					for (i = 0; i < sec->linecount; i++)
					{
						sector_t *tsec = sec->lines[i]->frontsector;

						if (   secnum != tsec - sectors
							|| !(sec->lines[i]->ml_flags & ML_TWOSIDED)
							// if sector's floor is different texture, look for another
							|| (tsec = sec->lines[i]->backsector)->floorpic != texture
							|| chksec[tsec - sectors])
						{
							continue;
						}

						sec    = tsec;
						secnum = tsec - sectors;
						sec->specFlags |= SF_MOVFLOOR;
						movflags[secnum] |= MOVOTHER;
						chksec[secnum] = true;
						repeat = true;
						break;
					}
				}
			}
			if (compat[c_stairs])  s = -1, mins = secnum;
		}
		M_Free(chksec);
	}
	break;

// EV_DoDoor
	case 2:
	case 3:
	case 4:
	case 16:
	case 29:
	case 42:
    case 46:
	case 50:
	case 61:
	case 63:
	case 75:
	case 76:
	case 86:
	case 90:
	case 103:
	case 105:
	case 106:
	case 107:
	case 108:
	case 109:
	case 110:
	case 111:
	case 112:
	case 113:
	case 114:
	case 115:
	case 116:
	case 175:
	case 196:
	// EV_DoLockedDoor
	case 99:
	case 133:
	case 134:
	case 135:
	case 136:
	case 137:
	// EV_VerticalDoor
	case 1:
	case 26:
	case 27:
	case 28:
	case 31:
	case 32:
	case 33:
	case 34:
	case 117:
	case 118:
		// Doors are only checked for missing
		// sector tags, not line tags
		if (ld->tag)
		{
			RDBOOL found = false, manual_door = false;

			// Check it's a manual door
			switch(value)
			{
				case 1: case 26: case 27: case 28:
				case 31: case 32: case 33: case 34:
					manual_door = true;
					break;
				default:
					break;
			}

			for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0;)
			{
				found = true;
				sectors[s].specFlags |= SF_MOVCEIL;
				secspflags[s] |= SECSPF_DOORSECTOR;
			}

			// Check for errors
			if (   found
				&& ld->backsector
				&& ld->backsector->tag
				&& ld->backsector->tag != ld->tag
				&& !lineHasTag(ld->backsector->tag) )
			{
				if (manual_door)
				{
					// Cover problem where a sector tag has been matched
					// but the where the push line backs a sector with a
					// different tag.

					pr_OFinsert("MAP ERROR FIXED: L%i tag %i tags the wrong sector FIX: tag %i has been substituted\n",
								ld - lines, ld->tag, ld->backsector->tag);
					ma_inf.maperror++;
					s = ld->backsector - sectors;
					ld->tag = ld->backsector->tag;
					sectors[s].specFlags |= SF_MOVCEIL;
					secspflags[s] |= SECSPF_DOORSECTOR;
				}
			}
			else if (ld->backsector && !ld->backsector->tag)
			{
				if (!manual_door)
				{
					// v220-5 if the tag's unmatched then it's an error
					if (!found)
					{
						pr_OFinsert("ERROR: L%i has special %i with an unmatched tag %i FIX: ", ld - lines, ld->special, ld->tag);

						ld->tag = 0;	// clear tag
						ma_inf.tagerrors++;

						if (ld->backsector && ld->backsector->floorheight == ld->backsector->ceilingheight)
						{
							pr_OFdirect("door special backs heightless sector so only the tag will be ignored\n");
							goto set_movceil;
						}
						else
						{
							pr_OFdirect("special and tag will be ignored\n");
							ld->special = 0;
						}
					}
					else if ( !(ld->backsector->specFlags & SF_MOVCEIL)
						&& ld->backsector->ceilingheight == ld->backsector->floorheight)
					{
						goto set_movceil;	// v220-7 add to the movsec list just in case
					}

					return;
				}

				if (found && ld->frontsector == ld->backsector)
					goto set_movceil;

				// Untagged or incorrectly tagged manual doors
				pr_OFinsert("ERROR: L%i special %i tag %i is invalid", ld - lines, ld->special, ld->tag);
				ma_inf.tagerrors++;

				if (!found && ld->frontsector == ld->backsector)	// v220-4 Map error
				{
					pr_OFdirect(" FIX: special will be ignored\n");
					ld->special = 0, ld->tag = 0;
					return;
				}

				// Cover problem where a push type door's sector
				// is correctly untagged but where the push line
				// is tagged to another sector

				pr_OFdirect(" FIX: backsector tag set to line's tag\n");

				ld->backsector->tag = ld->tag;
set_movceil:;
				s = ld->backsector - sectors;
				sectors[s].specFlags |= SF_MOVCEIL;
				secspflags[s] |= SECSPF_DOORSECTOR;
				LS_InitTagLists();	// Update hash table
			}
		}
		else
		{
			ld->lnflags2 |= LNF2_DOORTYPE;	// v220-15
			ld->backsector->specFlags |= SF_MOVCEIL;
		}
		break;
	}
}

//=====================================================

static RDBOOL useSpecial(line_t *ld, short sp)
{
	// Buttons (retriggerable switches) excluding doors
	if (sp == 21 || sp == 62 || sp == 211)
	{
		int s;

		for (s = -1; (s = P_FindSectorFromTag(ld->tag, s)) >= 0 ; )
		{
			if (sectors[s].specFlags & SF_MOVFLOOR)
			{
				int i;

				for (i = 0; i < sectors[i].linecount; i++)
				{
					line_t *li = sectors[s].lines[i];

					if (   li->frontsector - sectors == s
						|| li->backsector  - sectors == s)
					{
						return true;
					}
				}
			}
		}
	}

	return false;
}

// Some ports have dispensed with the ML_PASSUSE flag meaning that maps
// written for them don't set the flag. This routine attempts to find
// those cases and, if needed, restore the flag.
void LS_SetPASSUSEflags()
{
	line_t	*ld;
	int		i;

	// First check walkover types
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (!ld->tag || ld->special <= 0 || !(ld->ml_flags & ML_TWOSIDED))
			continue;

		if (useSpecial(ld, ld->special) )
		{
			// Make certain use types in sector shared with
			// or adjacent with the sector to be activated.
			ld->ml_flags |= ML_PASSUSE;
		}
	}
}
