//-----------------------------------------------------------------------------
//
// Copyright (C) 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: p_setupFns.cpp
//
// First added 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//	R3D functions that were in or were to be added to p_setup.cpp
//	moved here for better clarity
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
//
// DESCRIPTION: Helper functions for p_setup
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------
#include <io.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <direct.h>
#include <stdio.h>
#include <stdlib.h>

#include "Doom\p_local.h"
#include "r3d_MAINspec.h"
#include "r3d_modelbox.h"
#include "r3d_utils.h"
#include "dd_types.h"
#include "r_things.h"
#include "r_model.h"
#include "p_maputil.h"
#include "p_think.h"
#include "m_misc.h"
#include "r3d_edit.h"
#include "r3d_xline.h"
#include "sys_system.h"
#include "dd_wad.h"
#include "r3d_setupFns.h"
#include "r3d_arrays.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

typedef struct {
	mobj_t	**list;
	int		num;
	int		maxnum;
} cotype_t;
	
// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	modelError(int lump, int num);
RDBOOL	CheckPatch(patch_t *patch, int origx, int origy, int texwidth, uint size);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern byte		*spawnArray;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		torch_count;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

void PH_ResetChanges()
{
	cfg.useTorchLight = (ge_inf.flags & GEF_TORCHLIGHT) > 0;
	cfg.setVtxLight = (ge_inf.flags & GEF_VTXLIGHT) > 0;
	cfg.sineCompress = (ge_inf.flags & GEF_SINECOMPRESS) > 0;
	cfg.fallingDamage = (ge_inf.flags & GEF_FALLINGDAMAGE) > 0;
	cfg.underwaterDamage = (ge_inf.flags & GEF_UWDAMAGE) > 0;
	cfg.jumpEnabled = (ge_inf.flags & GEF_JUMP) > 0;
	cfg.footsteps = (ge_inf.flags & GEF_FOOTSTEPS) > 0;
	cfg.weaponsounds = (ge_inf.flags & GEF_WEAPRSOUND) > 0;
	cfg.weapsmoke = (ge_inf.flags & GEF_WEAPSMOKE) > 0;
	cfg.splashsprite = (ge_inf.flags & GEF_SPLASHSPRITE) > 0;

	waterdynamics = (ge_inf.flags & GEF_WATERDYNAMICS) > 0;
	useDynLights = (ge_inf.flags & GEF_DYNLIGHTS) > 0;
	water_trans = (ge_inf.flags & GEF_WATERTRANS) > 0;

	cfg.jumpPower = ge_inf.jumpPower;
	water_opacity = ge_inf.was_water_opacity;
	modelLights = ge_inf.num_modelLight;
}

void PH_PrintModelStats()
{
	if (!ArgCheckWith("-prmv", 1)) return;

	modeldef_t	*mf;
	thinker_t	*thinker;
	char		*stopstring;
	char		*ptr = ArgNext();
	int			num = strtol(ptr, &stopstring, 10);
	RDBOOL		found = false;
		
	if (*stopstring != 0 || num < 0)
		Con_Error("PRMV: sent an invalid string %s\n", ptr);

	for (thinker = thinkercap.next; thinker != &thinkercap; thinker = thinker->next)
	{
		if ((void (__cdecl *)(void))thinker->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t	*iter = (mobj_t*)thinker;

			if (MAPNUM(iter) == num)
			{
				R_CheckModelFor(iter, &mf, NULL);

				if (mf)
					pr_OFinsert("PRMV T%i: file %s [solid? %s, amodel? %s]\n"
					"     def->radius %i, radius given %i\n"
					"     max scaled mesh values: radius %i minh %i maxh %i zoff %i\n\n",
						num, modellist[mf->sub[0].model]->modelName,
						iter->flags & MF_SOLID ? "YES" : "NO",
						iter->flags3 & MF3_AMODEL ? "YES" : "NO",
						iter->info->radius>>16,
						iter->radius>>16,
						mf->xyradius>>16, mf->zmin>>16, mf->zmax>>16, mf->zoff>>16);
				else
					pr_OFinsert("T%i has no model\n", num);

				found = true;
				break;
			}
		}
	}

	if (!found)
		pr_OFinsert("PRMV: T%i does not exist\n", num);
}

typedef struct { char *name; short xoff, yoff; } KEENinf_t;

KEENinf_t KEENinf[] = {
	{"KEENA0",  6, 67},
	{"KEENB0",  8, 67},
	{"KEENC0",  9, 67},
	{"KEEND0", 11, 67},
	{"KEENE0", 13, 67},
	{"KEENF0", 13, 67},
	{"KEENG0", 14, 67},
	{"KEENH0", 17, 67},
	{"KEENI0", 17, 67},
	{"KEENJ0", 17, 67},
	{"KEENK0", 17, 67},
	{"KEENL0", 17, 67},
	{"KEENM0",  8, 67},
	{NULL, 0, 0}
};

// Note: arctic offsets take pixel texture shifts in gl_tex into account
void PH_SpriteOffsetAdjust()
{
	static		byte *sprlumpmem;
	static		int  numsprlumpmem;
	thinker_t	*it;
	mobj_t		*mo;
	int			i, k;

	// v220-33 for added function to adjust sprite offsets
	spritedef_t		*sprdef;
	spriteframe_t	*sprframe;
	int				numf, splnum;

	if (!sprlumpmem)
	{
		sprlumpmem = (byte*)calloc(numspritelumps, 1);
		numsprlumpmem = 0;
	}
	
	// v220-23 adjust offsets
	for (it = thinkercap.next; it != &thinkercap; it = it->next)
	{
		if (it->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mo = (mobj_t*)it;

			if (mo->flags2 & (MF2_DONTDRAW|MF2_NORENDER) || mo->ddflags & DDMF_DONTDRAW) continue;

			sprdef = &sprites[mo->sprite];

			// modified $v2.0.11
			if (  !(numf = sprites[mo->sprite].numframes)
				|| (mo->frame & FF_FRAMEMASK) >= numf )
			{
#ifdef RANGECHECK
				pr_OFinsert("INVALID sprite %i T%i dmednum %i %.4X [FM %i NF %i]\n",
					mo->sprite ,MAPNUM(mo) ,mo->info->doomednum, mo->frame & FF_FRAMEMASK, numf);
#endif
				if (mo->sprite == SPR_PLAY && mo->type == MT_PLAYER)
					Con_ErrorD("TERMINAL ERROR: sprite with state SPR_PLAY is invalid");

				P_RemoveMobj(mo, 36);	// v220-35 just delete it
				continue;
			}

			if (mo->flags3 & MF3_AMODEL) continue;

			if (R3D_ScriptedMap && sprdef->sprflags & SPR_NOMODEL)
				modelError(spritelumps[splnum].lump, MAPNUM(mo));	// sprites are definately not allowed!

			sprframe = &sprdef->spriteframes[ (mo->frame & FF_FRAMEMASK) ];

			// Only interested in sprites with no rotation
			if (sprframe->rotate == 1) continue;

			if (R3D_ScriptedMap && sprdef->sprflags & SPR_NOMODEL)
				modelError(spritelumps[splnum].lump, MAPNUM(mo));	// sprites are definately not allowed!

			if ( (splnum = sprframe->lump[0]) < 0 || splnum >= numspritelumps)
			{
				pr_OFinsert("INVALID sprite %i T%i dmednum %i %.4X [FM %i NF %i]\n",
					mo->sprite ,MAPNUM(mo) ,mo->info->doomednum, mo->frame & FF_FRAMEMASK, numf);
				mo->flags2 |= MF2_DONTDRAW;
				continue;
			}

			// If already identified then continue
			if (sprlumpmem[splnum] == 1) { mo->flags2 |= MF2_NOFITBOTTOM; continue; }

			// Specific R3D support for arctic_se
			// Some other ports will not run it despite it being a vanilla Doom2 TC.
			if (isarctic)
			{
				// Fix fir tree base (same for both arctgfx wads
				if (mo->type == 101 && !strcmp(lumpinfo[spritelumps[splnum].lump].name, "COL5B0"))
				{
					mo->flags2 |= MF2_NOFITBOTTOM;
					continue;
				}

				if (isarctic == 2)
				{
					if (mo->type == 133)	// defining it specifically as an ambient improves the sound quality 
					{
						mo->ddflags |= DDF_AMBIENT;	// ambient sound
						continue;
					}
				}
			}

			if (sprlumpmem[splnum]) continue;

			// v220-35 correct keen sprite yoffsets unless changed
			if (mo->type == MT_KEEN)
			{
				int slnum;

				sprlumpmem[splnum] = 2;

				if (numf != 13) continue;

				// check all frames before making changes
				for (i = 0; i < numf; i++)
				{
					slnum = sprdef->spriteframes[i].lump[0];

					if (slnum >= 0 && slnum < numspritelumps)
					{
						if (   stricmp(lumpinfo[spritelumps[slnum].lump].name, KEENinf[i].name)
							|| spritelumps[slnum].slxoffset !=  KEENinf[i].xoff
							|| spritelumps[slnum].slyoffset !=  KEENinf[i].yoff)
						{
							goto nochange;
						}
					}
					else
					{
						goto nochange;
					}
				}

				for (i = 0; i < numf; i++)
				{
					slnum = sprdef->spriteframes[i].lump[0];
					spritelumps[slnum].slyoffset = 71;
				}
nochange:;
				continue;
			}

			if (mo->flags & MF_NOGRAVITY) continue;
			if (mo->ddflags & DDMF_FITTOP || mo->flags2 & (MF2_NOFITBOTTOM|MF2_GLUECEIL)) continue;
			if (spritelumps[splnum].height <= spritelumps[splnum].slyoffset) continue;

			// v220-35 Check for low alpha rows with sprite patches with yoffsets
			if (numf == 1 && spritelumps[splnum].slyoffset)
			{
				typedef struct { byte topdelta; byte length;} column_t;
				int		lump = spritelumps[splnum].lump, width, height;
				int		type = DoomLumpIdentify(lump, &width, &height, DLI_GRAPHIC);
				size_t	sz = width * height;
				patch_t	*patch;

				if (   type == LMP_PATCH
					&& width >= 16 && height >= 16	// otherwise don't bother
					&& (patch = (patch_t*)M_CacheLumpNum(lump, NULL, false))
					&& width == patch->width && height == patch->height
					&& CheckPatch(patch, patch->xoffset, patch->yoffset, width, sz))
				{
					int		col, count, x, y;
					int		loalpharows = 0;
					byte	*ptr, *buffer = (byte*)M_Calloc(2 * sz, 0);
					byte	*destTop = buffer;
					byte	*destAlphaTop = destTop + sz;
					byte	*dest1, *dest2;
					byte	*source = NULL;

					for(col = x = 0; col < width; col++, destTop++, destAlphaTop++, x++)
					{
						column_t *column = (column_t*)((byte*)patch + patch->columnofs[col]);
						int top = -1;
						
						// Step through the posts in a column
						while(column->topdelta != 0xff)
						{
							source = (byte *)column + 3;
							
							if( x < 0 || x >= width ) break; // Out of bounds. 
							
							if (column->topdelta <= top)
								top += column->topdelta;
							else
								top = column->topdelta;
							
							if ( (count = column->length) > 0)     
							{
								y     = top;
								dest1 = destTop      + y*width;
								dest2 = destAlphaTop + y*width;	
								
								while(count--)
								{
									byte palidx = *source++;	// get byte of data (index into palette)
																		
									if (y >= 0 && y < height)// Check bounds.
									{
										*dest1 = palidx;
										*dest2 = 0xFF;
									}
									// One row down.
									dest1 += width; 				
									dest2 += width; 
									y++;
								}
							}
							column = (column_t *)((byte *)column + column->length + 4);
						}
					}

					for (i = height - 1; i >= 0; i--)
					{
						ptr = buffer + sz + i * width;

						for (k = 0; k < width; k++)
						{
							if (*(ptr++)) goto end_lorowcheck;
						}
						loalpharows++;
					}	
end_lorowcheck:;
					M_Free(patch);
					M_Free(buffer);

					if (loalpharows)
					{
						sprlumpmem[splnum] = 1;
						mo->flags2 |= MF2_NOFITBOTTOM;
						continue;
					}
				}
			}

			// If we get here set all yoffs the same as the height as yoffs are meaningless
			for (i = 0; i < numf; i++)
			{
				spriteframe_t *sf = &sprdef->spriteframes[i];

				for (k = 0; k < 8; k++)	// probably all the same lump
				{
					if (sf->lump[k] >= 0 && sf->lump[k] < numspritelumps && !sprlumpmem[sf->lump[k]])
					{
						sprlumpmem[sf->lump[k]] = 2;
						spritelumps[sf->lump[k]].slyoffset = spritelumps[sf->lump[k]].height;
					}
				}
			}

			sprlumpmem[splnum] = 2;
		}
	}
}

// NOTE: sector fit can't be done until analysis is complete

// v220-6 if thing is on a same sector line then check the thing fits in the line's sector
// v220-10 checks if a shadow is viable
// NOTE: the sslines array has to be set
void PH_CheckMobjFitsSector()
{
	thinker_t	*it;
	mobj_t		*mo;
	int			i, lnum;

	// Hopefully using a design where Things are made to pop-up by putting them
	// in an undersized missing tex movfloor sector is not employed in a 3D map
	if (R3D_maptype) return;

	for (it = thinkercap.next; it != &thinkercap; it = it->next)
	{
		if (it->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mo = (mobj_t*)it;

			if (  !mo->info->seestate
				&& (mo->flags2 & MF2_HSETMASK) != MF2_HSETMASK
				&& (lnum = FindIfThingIsOnLine(mo, false)) >= 0
				&& lines[lnum].specType < R3D_40
				&& lines[lnum].frontsector == lines[lnum].backsector
				&& !(lines[lnum].frontsector->specFlags & SF_MOVFLOOR)
				&& !thingFitsInSector(mo, lines[lnum].frontsector))
			{
				sector_t *ns, *sc = mo->subsector->sector;
				ssline_t *sl;

				if (sc == lines[lnum].frontsector)
				{					
					sl = &sslines[lnum];

					for (i = 0; i < sl->num; i++)
					{
						subsector_t *ss = &subsectors[sl->list[i]];

						if (ss->floorsec >= 0)
						{
							ns = &sectors[ss->floorsec];
							mo->z = mo->floorz = mo->dropoffz = ns->floorheight;
							mo->flags2 |= (MF2_SPECNOMOVE|MF2_HEIGHTSET|MF2_NOFITBOTTOM);
							break;
						}
					}
				}
			}
		}
	}
}

// v220-31	Checks if shadows have been inhibited by a 15003 Thing
void PH_CheckMobjShadows()
{
	thinker_t	*it;
	mobj_t		*mo;

	if (R3D_slopetypes) return;	// No shadows with slopes

	for (it = thinkercap.next; it != &thinkercap; it = it->next)
	{
		if (it->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mo = (mobj_t*)it;

			if (mo->flags2 & MF2_NOSHADOW) continue;

			// check for viable shadows
			if (mo->flags & (MF_NOSECTOR|MF_NOBLOCKMAP))
			{
				mo->flags2 |= MF2_NOSHADOW;
			}
			else
			if (   mo->ddflags & (DDMF_DONTDRAW|DDMF_ALWAYSLIT)
				|| mo->flags2 & (MF2_NORENDER|MF2_DONTDRAW))
			{
				mo->flags2 |= MF2_NOSHADOW;
			}
			else
			if (   !(mo->flags & MF_SHOOTABLE)
				&& !thingFitsInSector(mo, NULL))
			{
				mo->flags2 |= MF2_NOSHADOW;
			}
		}
	}
}

// v220-10 set subsec maxdimension
void PH_SetSSECMaxDimensions()
{
	int i;
	
	for (i = 0; i < numsubsectors; i++)
	{
		subsector_t	*ss = &subsectors[i];
		sector_t	*s = ss->sector;

		ss->maxdimension = MAX_OF(
			s->sibounds[BOXRIGHT] - s->sibounds[BOXLEFT],
			s->sibounds[BOXTOP  ] - s->sibounds[BOXBASE]);
	}
}

void PH_SetSCPOffsets()
{
	int	i, s, flags, tag, sec;

	// v220-31 check pseudofake flats
	for (i = 0; i < 99; i++)
	{
		if ( (tag = pseudofake[i][PF_TAG]) && (sec = pseudofake[i][PF_REFSEC]) >= 0)
		{
			flags = SF2_PSEUDOFAKE;

			if (pseudofake[i][PF_DAMAGE]) flags |= SF2_PSFAKEDAMAGE;

			// If we get here then all is OK
			if (R_CheckFlatFlag(sectors[sec].floorpic, FLT_WATER))
				sectors[sec].specFlags2 |= SF2_HASWATERFLATS;

			for (s = -1; (s = P_FindSectorFromTag(tag,s)) >= 0;)
			{
				sectors[s].heightsec = sec;
				sectors[sec].specFlags2 |= flags;
				sectors[sec].bottommap = pseudofake[i][PF_FILTER];
			}
		}
	}
}

void PH_SetXLines()
{
	if (!num_xlsame) return;

	xlinemem_t	*lm;
	int			*xlarray = (int*)malloc(num_xlsame * sizeof(int));
	int			i, k, n;

	for (i = 0; i < num_xlsame; i++)
	{
		lm = &xlinemem[i];
		lm->dx = lm->endx - lm->startx;
		lm->dy = lm->endy - lm->starty;
		lm->length = M_AccurateDistanceFP(lm->dx, lm->dy);
		lm->texn = R_TextureNumForName(xlinemem[i].midtex, false);
		lm->floorheight = FIX2FLT(sectors[lm->secn].floorheight);
		lm->ceilheight = FIX2FLT(sectors[lm->secn].ceilingheight);

		lm->box[BOXLEFT ] = (fixed_t)((lm->startx < lm->endx ? lm->startx : lm->endx  ) * 65536.f);
		lm->box[BOXRIGHT] = (fixed_t)((lm->startx < lm->endx ? lm->endx   : lm->startx) * 65536.f);
		lm->box[BOXBASE ] = (fixed_t)((lm->starty < lm->endy ? lm->starty : lm->endy  ) * 65536.f);
		lm->box[BOXTOP  ] = (fixed_t)((lm->starty < lm->endy ? lm->endy   : lm->starty) * 65536.f);
	}

	// setup subsec xline lists
	for (i = 0; i < numsubsectors; i++)
	{
		for (k = n = 0; k < num_xlsame; k++)
		{
			lm = &xlinemem[k];

			if (lm->box[BOXLEFT ] > subsectors[i].ssbox[BOXRIGHT] ||
				lm->box[BOXRIGHT] < subsectors[i].ssbox[BOXLEFT ] ||
				lm->box[BOXBASE ] > subsectors[i].ssbox[BOXTOP  ] ||
				lm->box[BOXTOP  ] < subsectors[i].ssbox[BOXBASE ])
			{
				continue;
			}

			xlarray[n++] = k;
		}

		if (n)
		{
			subsectors[i].xlines = (int*)W_Malloc((n + 1) * sizeof(int), LEVELID, 22);
			memcpy(subsectors[i].xlines, xlarray, n * sizeof(int));
			subsectors[i].xlines[n] = -1;	// list terminator
		}
	}

	free(xlarray);
}

// v220-25 initialise the xline vars.
void PH_InitXLines()
{
	use_xl_lines = (!R3D_edit && ma_inf.modifier & MODF_MODNODES);
	xlinemem = NULL;
	xlisrendered = NULL;
	num_xlsame = 0;	
}

// v220-16 Always done after analysis so flags can be relied on to
//         overcome problems with backward savegame compatibility.
void PH_SetBlockOcclude()
{
	line_t	*ld = lines;
	int		i;

	for (i = 0; i < numlines; i++, ld++)
	{
		ld->lnflags2 &= ~LNF2_REASSIGN;	// clear flags to be reassigned

		// v220-16 needed for floor flats
		if (ld->ml_flags & ML_DONTDRAW)
		{
			sector_t *fs = ld->frontsector, *bs = ld->backsector;

			if (   fs->ceilingheight - fs->floorheight > FRACUNIT
				&& !(fs->special == 9 || fs->special & SECRET_MASK))
			{
				if (!bs || fs == bs)
				{
					ld->ml_flags |= ML_DRAWFLATS;
				}
				else if (bs->ceilingheight - bs->floorheight > FRACUNIT
					&& !(bs->special == 9 || bs->special & SECRET_MASK))
				{
					ld->ml_flags |= ML_DRAWFLATS;
				}
			}
		}

		if (   ld->ml_flags & ML_TWOSIDED
			&& ld->specType < R3D_40
			&& !(ld->lnflags2 & LNF2_BLOCKALL)
			&& ld->frontsector->heightsec < 0
			&& ld->backsector->heightsec < 0)
		{
			if (ld->frontsector == ld->backsector)
			{
				if (   ld->frontsector->ceilingheight > ld->frontsector->floorheight
					&& !(ld->frontsector->specFlags & SF_MOVCEIL))	// v220-33 Seems to be the case with ZDoom maps
				{
					// Partially emulate Doom bug (i.e. don't apply if sector ceiling can move)
					ld->lnflags2 |= LNF2_UNBLOCKED;	// Emulate Doom bug
				}
			}
			else if (  ld->frontsector->floorheight == ld->backsector->floorheight
					&& !(ld->mtflags & R3D_INVLOTEX)	// v220-35
					&& !(ld->frontsector->specFlags & SF_MOVCEIL)
					&& !(ld->backsector->specFlags & SF_MOVCEIL))
			{
				// Interestingly, in the Doom engine, if a line has the same front and back
				// ceiling heights and front and back floor heights it is ignored for the purposes of blocking.
				// NOTE: 'flat' sectors should block IMHO so a height check is done.
				//       Unfortunately a MO floor cannort be used at this pont to invalidate the condition
				if (   ld->frontsector->ceilingheight == ld->backsector->ceilingheight
					// Only true unless closed
					&& ld->frontsector->ceilingheight -  ld->frontsector->floorheight > FRACUNIT
					&& ld->backsector->ceilingheight  -  ld->backsector->floorheight  > FRACUNIT
					&& !(ld->lnflags2 & (LNF2_LOWTRIWALL|LNF2_HITRIWALL)))
				{
					ld->lnflags2 |= LNF2_UNBLOCKED;	// Emulate Doom bug
				}

				// v220-16 Set skyocclude flag.
				// This prevents having to run the full logic for each seg when rendering
				if (   ld->frontsector->ceilingpic == skyflatnum
					&& ld->backsector->ceilingpic == skyflatnum
					&& ld->lnflags2 & LNF2_OWNEQUALTOTWO)
				{
					if ( !(origTEX[i] & 0x02)
						&& ld->backsector->ceilingheight - ld->backsector->floorheight <= FRACUNIT)
					{
						ld->lnflags2 |= LNF2_SKYOCCLUDE;
					}
					else
					if ( !(origTEX[i] & 0x20)
						&& ld->frontsector->ceilingheight - ld->frontsector->floorheight <= FRACUNIT)
					{
						ld->lnflags2 |= LNF2_SKYOCCLUDE;
					}
				}
			}
		}
	}
}

static int subfn_FindNumEnemiesInSector(sector_t *s)
{
	thinker_t	*thinker;
	int			number = 0;

	for (thinker = thinkercap.next; thinker != &thinkercap; thinker = thinker->next)
	{
		if ((void (__cdecl *)(void))thinker->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t	*iter = (mobj_t*)thinker;
			if (iter->subsector->sector == s && iter->info->seestate) number++;
		}
	}

	return number;
}

void PH_CheckStress()
{
extern int	corpsetime;
extern int	zdistptr;

	if (ma_inf.activemodels > 4000 || ma_inf.activethinkers > 8000)
	{

		int		c, i, maxn, n, limit = ma_inf.activemodels > 4000 ? 4000 : 8000;

		for (c = i = maxn = 0; i < numsectors; i++)
		{
			if ( (n = subfn_FindNumEnemiesInSector(&sectors[i])) > 1000)
			{
				maxn += n; if (++c > 4) break;
			}
		}

		if (maxn > 4000)
		{
			char	buf[512];
			char	*cpt = "\nTo reduce stress corpsefade has been set to 2 seconds\n\n";
			char	*ldt = "Distance to lod models should be set to fastest.\n"
						   "(See Control Panel - Stress Settings)\n";

			sprintf(buf, "HIGH STRESS GAME DETECTED:\n"
						 "%i %s enemies exist in just %i sectors\n"
						 "This %s cause excessive loading of the CPU.\n",
						 maxn, ma_inf.activemodels ? "model" : "sprite", c,
						 ma_inf.activemodels ? "will" : "may");

			if (!ge_inf.devmap && ma_inf.activemodels)
			{
				strcat(buf, "\nPlease restart using basic doom mode.\n");
				pr_OFdirect(buf);
				Sys_MessageBox(buf, false);
				Sys_Quit(false);
			}

			if (!corpsetime || corpsetime > 2)
			{
				strcat(buf,	cpt);
				corpsetime = 2;
			}

			if (ma_inf.activemodels)
			{
				if (zdistptr > 1) strcat(buf,	ldt);
				strcat(buf, "More than 50 dropped weapons will be removed after 30 seconds\n");
				ma_inf.activedropped = 50;
			}

			pr_OFdirect(buf);
			Sys_MessageBox(buf, false);
		}
	}
	else
	{
		corpsetime = cfg.corpseTime;
	}

	corpsetime *= TICSPERSEC;
}

// In order to be able to analyse that mobjs' heights are correct
// then, for an analysis file to work with any skill setting,
// all map objects must be analysed.  This routine, therefore, is
// required to allow the later removal of any that should not be
// present as dictated by the skill level selected.
//
void PH_CheckSpawnedbySkill()
{
	thinker_t	*thinker, *next;
	cotype_t	ct[DH_NUMOBJECTS];
	int			sndthing[64];
	int			i, k, m;
	// v220-7
	mobj_t		**noflags_th = (mobj_t**)W_Calloc(numthings * sizeof(mobj_t*), TEMPID, 0);
	int			noflags_count = 0;
	//----------------------------------

	memset(sndthing, 0, sizeof(sndthing));
	torch_count = 0;
	
	// Check spawned things are valid
	for (thinker = thinkercap.next; thinker != &thinkercap; thinker = thinker->next)
	{
		if ((void (__cdecl *)(void))thinker->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t	*iter = (mobj_t*)thinker;
			int		mapnum = MAPNUM(iter);

			if (mapnum < 0 || mapnum >= numthings)	// sanity check
			{
				Con_Error("THING ERROR : mapnum %i", mapnum);
			}
		}
	}

	memset(ct, 0, sizeof(ct));
	thinker = thinkercap.next;

	while(thinker != &thinkercap)
	{
		next = thinker->next;

		if ((void (__cdecl *)(void))thinker->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t *iter = (mobj_t *) thinker;

			// Retrieve thing's map number
			int mapnum = MAPNUM(iter);

			ma_inf.numthinkers++;

			if (spawnArray[mapnum] == false || wadani.mapnummoremove == mapnum)
			{
				if (iter->flags3 & MF3_SECRET) totalsecret--;
				P_UnlinkThing(iter);
				P_RemoveThinker(thinker);
				ma_inf.thinkersremoved++;
			}
			else
			{
				int		sprnum = states[iter->info->spawnstate].sprite;
				int		dehacked = 0;
				RDBOOL	istorch = iter->type >= MT_MISC41 && iter->type <= MT_MISC46;

				if (mapnum >= 0 && wadani.mapnummorespawn == mapnum)
				{
					wadani.mobj = iter;
				}

				if (iter->type == MT_SOUNDSOURCE)
				{
					int n = TIDNUM(iter);

					if (n >= 0 && n < 63) sndthing[n]++;
				}

				// v220-16 preset pickup Things useradius
				if (iter->flags & MF_SPECIAL)
					iter->pickupRadius = iter->info->radius;

				if (iter->type == MT_BOSSSPIT)	// v220-13 was MT_BOSSBRAIN
				{
					ma_inf.hasboss = true;
				}
				else	// v330-13 setting numbraintargets moved here from p_enemy.c
				if (iter->type == MT_BOSSTARGET)
				{
					if (numbraintargets < MAXBRAINTARGETS)
					{
						braintargets[numbraintargets] = iter;
						numbraintargets++;
					}
					else
					{
						pr_OFinsert("WARNING: brain targets exceeds limit of %i.\n"
							"Target mapnum %i ignored\n",
							MAXBRAINTARGETS, MAPNUM(iter));
					}
				}
				else
				if (loadModels && iter->type == MT_BARREL)
				{
					sector_t *s = iter->subsector->sector;

					// suppresses barrels being drawn where, if a sprite, they are not visible e.g. in hc20 map07
					if (  (s->specFlags & 0x1041C) == 0x1001C
						&& s->ceilingheight - s->floorheight < 10*FRACUNIT)
					{
						iter->flags2 |= MF2_NORENDER;
					//	iter->flags3 |= MF3_TOUCHY;
					}
				}

				if (   iter->type < DH_NUMOBJECTS
					&& (iter->flags & MF_SPECIAL || iter->type == MT_BARREL))
				{
					cotype_t *cp = &ct[iter->type];

					if (cp->num >= cp->maxnum)
					{
						cp->maxnum += 128;
						cp->list = (mobj_t**)W_Realloc(cp->list, cp->maxnum * sizeof(mobj_t*), TEMPID, 0);
					}

					cp->list[cp->num++] = iter;
				}

				if (iter->flags & MF_COUNTKILL) // $stresschange was MF_SHOOTABLE
				{
					ma_inf.activethinkers++;
					if (states[iter->info->spawnstate].model) ma_inf.activemodels++;
				}

				// Check keys and torches with reset flags
				if (   hasDEH
					&& sprnum >= 0 && sprnum < DH_NUMSPRITES
					&& iter->type < DH_NUMOBJECTS
					&& (DEHCheck[iter->type] & DEHF_THING || DEHCheck[sprnum] & DEHF_SPRITE))
				{
					// changed v220-27
					if ( (iter->flags & ~MF_AMBUSH) == (MF_SPAWNCEILING|MF_NOGRAVITY) )
					{
						iter->ddflags |= DDMF_FITTOP;
					}

					// v220-9 check true height of dehacked thing
					//        Only done if no flags are set
					if (!iter->flags && DEHCheck[iter->type] & DEHF_THING)
					{
						fixed_t ht = GetSpriteMaxHeight(iter->sprite);

						if (ht > iter->height)
						{
							iter->height = iter->info->height = ht;
						}
					}

					// dehacked torches may be stacked
					if (   istorch
						&& !(iter->flags)
						&& !(iter->subsector->sector->siflags & SIF_CARRY)
						&& !(iter->subsector->sector->specFlags & SF_MOVFLOOR))
					{
						noflags_th[noflags_count++] = iter;
					}
				}
				else if (istorch)
				{
					// R3D - Only count torches that have not been dehacked
					iter->flags2 |= MF2_TORCH;
					torch_count++;
				}
			}
		}
		thinker = next;
	}

	// v220-13
	if (ma_inf.hasboss && !numbraintargets) ma_inf.hasboss = false;
	if (ma_inf.modifier & MODF_KILL) ma_inf.hasboss = true;

	// v220-7 check DEHacked things with flags set to 0
	for (i = 0; i < noflags_count; i++)
	{
		mobj_t *iter = noflags_th[i];

		if (!(iter->flags3 & MF3_STATIC))
		{
			for (k = i + 1; k < noflags_count; k++)
			{
				mobj_t *other = noflags_th[k];

				if (   other->subsector == iter->subsector
					&& other->x == iter->x
					&& other->y == iter->y)
				{
					fixed_t ht = GetSpriteMaxHeight(iter->sprite);
					other->z = other->floorz = other->dropoffz = iter->z + ht;
					other->flags3 |= MF3_STATIC;
					pr_OFinsert("T%i is on T%i Ht %i\n", MAPNUM(other), MAPNUM(iter), ht>>16);
				}
			}
		}
	}

	// Find coincident pickups or exploding objects of the same type
	// Not rendering superimposed objects can help speed things up
	for (i = 0; i < DH_NUMOBJECTS; i++)
	{
		if (ct[i].num > 1)
		{
			cotype_t *cp = &ct[i];
			byte	 *done = (byte*)M_Calloc(cp->num * sizeof(byte), 0);
			int		 n = cp->num - 1;

			for (k = 0; k < n; k++)
			{
				if (!done[k])
				{
					int x = cp->list[k]->x;
					int y = cp->list[k]->y;
					int z = cp->list[k]->z;

					for (m = k + 1; m < cp->num; m++)
					{
						if (  !done[m]
							&& cp->list[m]->x == x
							&& cp->list[m]->y == y
							&& cp->list[m]->z == z)
						{
							cp->list[m]->flags2 |= MF2_NORENDER;
							done[m] = true;
						}
					}
				}
			}
			M_Free(done);
		}
	}

	W_FreeALL(TEMPID);
	W_Free(spawnArray, STATICID, 531);	// Initialised in P_LoadThings
}


// R3D - inhibit torch glow if an excessive number are in the map
void PH_CheckTorchGlow()
{
	if (R3D_edit || R3D_maptype || !CFX) return;

	if (compat[c_torches] && (unsigned)(torch_count * numsubsectors) > 128 * 5000)
	{
		thinker_t	*th;
		mobj_t		*mo;

		pr_OFinsert("NOTE: To avoid speed problems dynamic lighting has been\n"
					"      automatically inhibited for all torches. The global\n"
					"      command line switch -c_torches_off prevents this test.\n");


		for (th = thinkercap.next; th != &thinkercap; th = th->next)
		{
			if ((void (__cdecl *)(void))th->function == (void (__cdecl *)(void))P_MobjThinker)
			{
				mo = (mobj_t*) th;

				switch (mo->type)
				{
				// tall torches
				case MT_MISC41:
				case MT_MISC42:
				case MT_MISC43:
				// short torches
				case MT_MISC44:
				case MT_MISC45:
				case MT_MISC46:
				// candlestik
				case MT_MISC49:
				// candelabra
				case MT_MISC50:
					mo->flags2 |= MF2_NOTLIT;
					break;
				default:
					break;
				}
			}
		}
	}
}

// v220-33 check texture is only used as a midtex
int PH_CheckMidTex(int idx)
{
	int i;

	for (i = 0; i < numsides; i++)
	{
		if (sides[i].bottomtexture == idx || sides[i].toptexture == idx) return 0;
	}

	return 1;	// must be 1
}

// v220-33	Checks that a texture can be clamped for all lines sharing a texture in a map.
//			In general this has been found to be true. If this was not the case then
//			another identical image would need to be created and stored to allow for lines
//			with a texture that cannot be clamped co-existing with any that can. This would
//			lead to unwanted complexity so a texture is only clamped if this is possible
//			for all sides referencing it. Note that this is only used with lo-res textures.
int PH_CheckTexLine(int idx, int width, int *lnum, int *lcnt)
{
	line_t	*ld;
	int		len;
	int		i, off1, off2, s1, s2, canclamp = false;
	
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->sidenum[1] < 0 || !(origTEX[i] & 0x22)) continue;

		s1 = sides[ld->sidenum[0]].midtexture == idx;
		s2 = sides[ld->sidenum[1]].midtexture == idx;

		if (s1 || s2)
		{			
			if ( !(ld->specType == R3D_00 || ld->specType == R3D_03 || ld->specType == R3D_10 || ld->specType == R3D_1D || ld->specType == R3D_36) )
			{
				if (ge_inf.devmap && ArgExists("-alphaInf"))	// debug aid
					Con_Printf("CLAMPS for texture %8s aborted on L%i with type R3D%.2X\n", texArray[idx]->name, i, ld->specType);
				canclamp = false; break;
			}
			
			// scrolling textures cannot be clamped
			if (   (s1 && sides[ld->sidenum[0]].lsiflags & LS_SIDESCROLL)
				|| (s2 && sides[ld->sidenum[1]].lsiflags & LS_SIDESCROLL))
			{
				canclamp = false; break;
			}

			if ( (len = (int)Get_LineLength(ld)) <= width)
			{
				// Check if the applied offset is viable
				off1 = s1 ? sides[ld->sidenum[0]].textureoffset>>16 : 0;
				off2 = s2 ? sides[ld->sidenum[1]].textureoffset>>16 : 0;

				if (off1 < 0)
				{
					// negative x offsets cannot be used when clamped
					// so check whether they can be converted to a positive value
					while (off1 < 0) off1 += width;

					if (len > (width - off1) ) { canclamp = false; break; }
				}
				
				if (off1 >= width)
				{
					// positve x offsets greater than a textures width cannot be used when clamped
					// so check whether they can be converted to be within range
					while (off1 >= width) off1 -= width;

					if ( (width - off1) > len) { canclamp = false; break; }
				}
				else
				if (off1 + len > width) { canclamp = false; break; }

				if (off2 < 0)
				{
					while (off2 < 0) off2 += width;

					if (len > (width - off2) ) { canclamp = false; break; }
				}
				
				if (off2 >= width)
				{
					while (off2 >= width) off2 -= width;

					if ( (width - off2) > len) { canclamp = false; break; }
				}
				else
				if (off2 + len > width) { canclamp = false; break; }

				lnum[ (*lcnt)++ ] = i;
				canclamp = true;
			}
			else
			{
				canclamp = false; break;
			}
		}
	}
	
	return canclamp;
}

// v220-9 change thing angles for iwads as defined in R3D_thdir.ded
void PH_setThingAngles(char *mapid, RDBOOL isPwad)
{
	RDBOOL	found;
	char	fname[512];
	char	*nextstr;

	// Note: do not change the order
	char	*angle[] = { "E", "NE", "N", "NW", "W", "SW", "S", "SE", NULL };
	short	sp_ang[] = { 0x0000, 0x002d, 0x005a, 0x0087, 0x00b4, 0x00e1, 0x010e, 0x013b };

	int		c = 0;

	//------------------------------------

	if (!loadModels) return;

	sprintf(fname, "%sData\\thd\\%s\\%s.thd",
		ddRuntimeDir.path, isPwad ? "pwad" : "iwad", isPwad ? analysename : mainIwadname);
	
	if (!SR_LoadFile(fname, "R3D_THDIR", NULL)) return;

	// 1) find the map entry
	found = false;

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, mapid)) { found = true; break; }
	}

	if (!found) { SR_EndScript(); return; }

	// 2) Read the Tnn entries
	found = false;

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, "{")) { found = true; break; }
	}

	if (!found) { SR_SCPWarn("Missing {"); SR_EndScript(); return; }

#ifdef _DEBUG
	pr_OFinsert("LOADING THD: %s\n", fname);
#endif

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, "}")) break;

		if (nextstr[0] == 'T' || nextstr[0] == 't')
		{
			thinker_t	*th;
			int			i, n = atoi(nextstr + 1);

			nextstr = SR_NextSTR(true);

			for (i = 0; angle[i]; i++)
			{
				if (!_stricmp(nextstr, angle[i]) ) break;
			}

			if (angle[i])
			{
				for (th = thinkercap.next ; th != &thinkercap ; th = th->next)
				{
					mobj_t *it = (mobj_t*)th;

					if (   it->thinker.function == (void (__cdecl *)(void))P_MobjThinker
						&& MAPNUM(it) == n)
					{
						it->startangle = 0;
						it->visangle = (i << 13);
						it->angle = (i << 29);
						it->spawnpoint.angle = sp_ang[i];
						c++;
						break;
					}
				}
			}
			else
			{
				SR_SCPWarn("Invalid angle \"%s\" with T%i", nextstr, n);
			}
		}
		else
		{
			SR_SCPWarn("Invalid Tn entry \"%s\"", nextstr); break;
		}
	}

	SR_EndScript();
}

// v220-22 check references
void PH_CheckIntegrity()
{
extern int use_asmnode;

	subsector_t *s2, *ss;
	fixed_t		x;
	fixed_t		y;
	int			error_c0 = 0;
	int			error_c1 = 0;
	int			mismatch_c0 = 0;
	int			mismatch_c1 = 0;
	int			oneD_count = 0;
	int			i;

	use_asmnode = true;	// default

	pr_OFdirect("Checking engine integrity with nodes;\n");

	for (i = 0; i < numsubsectors; i++)
	{
		ss = &subsectors[i];
		x = (fixed_t)(ss->midpoint.x * FRACUNIT);
		y = (fixed_t)(ss->midpoint.y * FRACUNIT);

		if ( (s2 = R_PointInSubsector(x, y)) != ss)
		{
			mismatch_c0++;

			// Check if boxes overlap
			if (   ss->ssbox[BOXLEFT ] > s2->ssbox[BOXRIGHT]
				|| ss->ssbox[BOXRIGHT] < s2->ssbox[BOXLEFT ]
				|| ss->ssbox[BOXBASE ] > s2->ssbox[BOXTOP  ]
				|| ss->ssbox[BOXTOP  ] < s2->ssbox[BOXBASE ])
			{
				if (   ss->ssbox[BOXLEFT] == ss->ssbox[BOXRIGHT]
					|| ss->ssbox[BOXBASE] == ss->ssbox[BOXTOP  ])
				{
					oneD_count++;
					continue;
				}

				error_c0++;
			}
		}
	}

	if (error_c0)
	{
		use_asmnode = false;

		for (i = 0; i < numsubsectors; i++)
		{
			ss = &subsectors[i];
			x = (fixed_t)(ss->midpoint.x * FRACUNIT);
			y = (fixed_t)(ss->midpoint.y * FRACUNIT);

			if ( (s2 = R_PointInSubsector(x, y)) != ss)
			{
				mismatch_c1++;

				// Check if boxes overlap
				if (   ss->ssbox[BOXLEFT ] > s2->ssbox[BOXRIGHT]
					|| ss->ssbox[BOXRIGHT] < s2->ssbox[BOXLEFT ]
					|| ss->ssbox[BOXBASE ] > s2->ssbox[BOXTOP  ]
					|| ss->ssbox[BOXTOP  ] < s2->ssbox[BOXBASE ])
				{
					if (   ss->ssbox[BOXLEFT] == ss->ssbox[BOXRIGHT]
						|| ss->ssbox[BOXBASE] == ss->ssbox[BOXTOP  ])
					{
						continue;
					}
					error_c1++;
				}
			}
		}

		// WARNING:	using zdbsp, unlike glbsp, can give an error count of 0.
		//			This does not mean that zdbsp nodes are the better choice.
		//			With par E1M6, for example, zdbsp nodes are incompatable
		//			with R3D's analysis code.

		// Select asm type
		if (error_c0 <= error_c1) use_asmnode = true;

		if (ge_inf.showconsistency)
		{
			char msg[256];

			sprintf(msg, "WARNING checkIntegrity:\n"
						 "error_c0    %i\n"
						 "error_c1    %i\n"
						 "mismatch_c0 %i\n"
						 "mismatch_c1 %i\n"
						 "oneD_count  %i\n\n"
						 "use_asmnode  %i\n",
								error_c0, error_c1,
								mismatch_c0, mismatch_c1,
								oneD_count, use_asmnode);

			if (!Sys_CriticalMessage(msg, true)) Sys_Quit(false);
		}
	}
}

void PH_CheckHoldState()
{
	// v220-13 check if any hold state flags need to be set
	if (ma_inf.holdstate)
	{
		line_t	*ld;
		int		i;

		for (i = 0, ld = lines; i < numlines; i++, ld++)
		{
			if (ld->backsector)
			{
				if (ld->special == 8113 || ld->special == 8114)
					TS_setHoldState(ld);
			}
		}
	}
}

static RDBOOL controlSEC(sector_t *sc)
{
	for (int i = 0; i < sc->linecount; i++)
	{
		if (sc->lines[i]->spflags & SPF_CONTROLLINE)
			return true;
	}
	return false;
}

// v220-33 rationalise settings
void PH_CheckSSECrefs()
{
	subsector_t *ss;
	sector_t	*sc;
	int			i;

	for (i = 0; i < numsubsectors; i++)
	{
		ss = &subsectors[i];
		sc = ss->sector;

		if (sc->heightsec >= 0) continue;

		if (!(sc->specFlags & SF_MOVFLOOR))
		{
			if (   ss->floorsec >= 0
				&& !(sectors[ss->floorsec].specFlags & SF_MOVFLOOR)
				&& sectors[ss->floorsec].heightsec < 0
				&& sc->floorheight == sectors[ss->floorsec].floorheight
				&& sc->floorpic == sectors[ss->floorsec].floorpic
				&& !controlSEC(sc))
			{
				ss->floorsec = -1;
			}
		}

		if (!(sc->specFlags & SF_MOVCEIL))
		{
			if (   ss->ceilsec >= 0
				&& !(sectors[ss->ceilsec].specFlags & SF_MOVCEIL)
				&& sectors[ss->ceilsec].heightsec < 0
				&& sc->ceilingheight == sectors[ss->ceilsec].ceilingheight
				&& sc->ceilingpic == sectors[ss->ceilsec].ceilingpic
				&& !controlSEC(sc))
			{
				ss->ceilsec = -1;
			}
		}
	}
}

void PH_CheckExternal3DLiftSector()
{
	sector_t	*sc, *chk, *os;
	line_t		*ld;
	int			c, d, i, k;

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (!(sc->specFlags2 & SF2_ELEVATOR3D)) continue;

		if (ge_inf.l3dreport) pr_OFinsert("PH_CheckExternal3DLiftSector  LIFTSEC %i sf2 0x%.8X\n", i, sc->specFlags2);

		for (c = d = k = 0, os = NULL; k < sc->linecount; k++)
		{
			ld = sc->lines[k];
			if (ld->spflags & SPF_CONTROLLINE) continue;
			if (ld->specType >= R3D_40)  { d++; continue; }

			chk = ld->frontsector == sc ? ld->backsector : ld->frontsector;

			if (!chk) continue;

			if (!os) os = chk; else if (os != chk) { c = 0; break; }

			c++;
		}

		if (os && c > 2 && d)
		{
			// remember the starting height
			sc->R3Dceilheight = sc->ceilingheight;
			if (ge_inf.l3dreport) pr_OFinsert("  LIFTSEC %i OUTER S%i count %i\n", sc - sectors, os - sectors, c);
		}
	}
}

// v220-35 Lump used to set MF3_TOUCHY|MF3_BOUNCES|MF3_TRANSLUCENT
void PH_ThingFlags(char *mapid)
{
	if (haspwads && SR_LoadPWADLumpName("SETFLAGS") )
	{
		thinker_t	*thinker;
		mobj_t		*m;
		char		*nextstr;

		while ( (nextstr = SR_NextSTR(false)) )
		{
			if (!stricmp(nextstr, mapid))
			{
				if ( !(nextstr = SR_NextSTR(false)) ) break;

				if (!stricmp(nextstr, "{"))
				{
					while ( (nextstr = SR_NextSTR(false)) )
					{
						if (!stricmp(nextstr, "}")) goto endscript;	// end of this map set

						int		i = atoi(nextstr);				// get thing mapnum
						char	*type = SR_NextSTR(false);		// get string of flag names

						for (thinker = thinkercap.next; thinker != &thinkercap; thinker = thinker->next)
						{
							if (   thinker->function == (void (__cdecl *)(void))P_MobjThinker
								&& (m = (mobj_t*) thinker)
								&& MAPNUM(m) == i)
							{
								if (strstr(type, "TOUCHY")) m->flags3 |= MF3_TOUCHY;
								if (strstr(type, "ARMED")) m->flags3 |= MF3_ARMED;
								if (strstr(type, "BOUNCES")) m->flags3 |= MF3_BOUNCES;
								if (strstr(type, "TRANSLUCENT")) m->flags3 |= MF3_TRANSLUCENT;
							}
						}
					}
				}
				break;
			}
		}
endscript:;
		SR_EndScript();
	}
}

// If a rowoffset has been set for one sky sector then it must be set for all
void PH_CheckYoffInheritance()
{
	int		i, k;

	if (!ma_inf.checkSkyOffy) return;

	for (i = 0; i < numsectors; i++)
	{
		if (sectors[i].sidesky & PL_SKYFLAT && sectors[i].sky_offy)
		{
			for (k = 0; k < numsectors; k++)
			{
				if (sectors[k].sidesky & PL_SKYFLAT && !sectors[k].sky_offy)
				{
					if (sectors[k].sidesky == sectors[i].sidesky)
						sectors[k].sky_offy = sectors[i].sky_offy;
				}
			}
		}
	}
}
