//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2011 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_skyloops.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
//
// DESCRIPTION: looks for line rings that determine that all lines and
//              subsectors inside neither have skywalls nor occlude.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "m_misc.h"
#include "r3d_MAINspec.h"

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// CODE --------------------------------------------------------------------


// b211-0 Marks lines in moving floor secs with sky sectors
//        where all their lines form a single complete loop.
void LS_MovSkyLoops()
{
	sector_t	*sec;
	int			i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->specFlags & SF_SECCLOSED
			&& !(sec->specFlags & SF_MOVCEIL)
			&& sec->ceilingpic == skyflatnum
			&& !IsLiquidFloorFlat(sec)
			&& sec->linecount > 2)
		{
			int k, valid = sec->specFlags & SF_MOVFLOOR;

			for (k = 0; k < sec->linecount; k++)
			{
				line_t *ld = sec->lines[k];

				if ( !(ld->ml_flags & ML_TWOSIDED)
					|| ld->frontsector == ld->backsector
					|| ld->frontsector->specFlags & SF_MOVCEIL
					|| ld->backsector->specFlags  & SF_MOVCEIL)
				{
					valid = false;
					break;
				}

				if (!valid
					&& (ld->frontsector->specFlags & SF_MOVFLOOR
					||  ld->backsector->specFlags  & SF_MOVFLOOR))
				{
					valid = true;
				}

			}

			// Check for line loops with missing lower textures
			// If valid set the rend flag to inhibit drawing if below the backsec height					
			if (!valid)	// rev170f
			{
				int count = 0, bbox[4];

				CLR_hitline();

				for (k = 0; k < sec->linecount; k++)
				{
					line_t *ld = sec->lines[k];

					if (ld->ml_flags & ML_TWOSIDED
						&& Vtxs2(ld)
						&& ld->backsector->ceilingpic == ld->frontsector->ceilingpic
						&& (IsLiquidFloorFlat(ld->frontsector)
						^   IsLiquidFloorFlat(ld->backsector))
						&& ((!(origTEX[ld - lines] & 0x01) && ld->frontsector->floorheight < ld->backsector->floorheight)
						||  (!(origTEX[ld - lines] & 0x10) && ld->frontsector->floorheight > ld->backsector->floorheight)))
					{
						hitline[ld - lines] = 1;
						count++;
					}
				}
				if (count < 3) continue;

				for (k = 0; k < sec->linecount; k++)
				{
					int m, lnum = sec->lines[k] - lines;

					if (hitline[lnum] == 1
						&& subfn_CheckForRing(hitline, sec, lnum, NULL, bbox, NULL, NULL)
						&& !LS_LinesInBox(sec, hitline, NULL, bbox, NULL, false))
					{
						for (m = 0; m < sec->linecount; m++)
						{
							lnum = sec->lines[m] - lines;

							if (hitline[lnum] == 4)
							{
								hitline[lnum] = 8;
								lines[lnum].lnflags2 |= LNF2_RENDVIEWH;
							}
						}
					}
				}
			}
		}
	}
}

// Look for a closed ring of lines with sky flats on each sector
// and with a lower sky height inside.  This is used to suppress
// the drawing of sky walls with anything inside. 
void LS_SetSkyLoops()
{
	sector_t	*fs, *bs;
	line_t		*ld;
	int			rc, box[4];
	int			c, d, i, k, m;
	int			txn;
	RDBOOL		repeat;

	CLR_hitline();

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& ld->lnflags2 & LNF2_OWNEQUALTOTWO
			&&  !( (txn = origTEX[i]) & 0x22)
			&& Vtxs2(ld)
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& fs->ceilingpic == skyflatnum
			&& bs->ceilingpic == skyflatnum)
		{
			int a = fs->ceilingheight - fs->floorheight;
			int b = bs->ceilingheight - bs->floorheight;
			
			if (abs(a - b) <= MAX_SECHEIGHT)
			{
				hitline[i] = 1; c++;
			}
		}
	}

	if (c > 2)
	{
		for (i = 0; i < numlines; i++)
		{
			if (   hitline[i] == 1
				&& (fs = lines[i].frontsector)
				&& (bs = lines[i].backsector)
				&& subfn_CheckForRing(hitline, fs, i, &rc, box, NULL, NULL)  )
			{
				if (rc == 4)
				{
					if (!LS_MarkSubsecsInAreaGeneral(fs, hitline, 0, NULL, box, -1, MSIA_TESTAREA))
						fs = lines[i].backsector, bs = lines[i].frontsector;

					if ( (bs->specFlags & 0xC) == 0xC) continue;	// v220-4 $skyloops

					if (   fs->ceilingheight <= bs->ceilingheight
						&& LS_MarkSubsecsInAreaGeneral(fs, hitline, 0, NULL, box, -1, MSIA_NONE))
					{
						RDBOOL	valid = true;
						int		ht = fs->ceilingheight;

						validcount++;

						// Check this is a valid loop
						for (k = 0; k < hitssecptr; k++)
						{
							if (subsectors[hitssec[k]].sector->ceilingheight > ht)
							{
								valid = false; break;
							}
						}

						if (valid)	// Check if a midtex extends beyond ht
						{
							for (k = 0; k < hitssecptr; k++)
							{
								subsector_t *ss = &subsectors[hitssec[k]];
								seg_t *sg = &segs[ss->firstline];

								for (m = 0; m < ss->linecount; m++, sg++)
								{
									if (sg->linedef && sg->linedef->validcount != validcount)
									{
										sg->linedef->validcount = validcount;

										if (sg->linedef->ml_flags & ML_TWOSIDED
											&& sg->sidedef->midtexture)
										{
											if (  texArray[sg->sidedef->midtexture]->height*FRACUNIT
												+ sg->frontsector->floorheight
												+ sg->sidedef->rowoffset > ht)
											{
												valid = false; break;
											}
										}
									}
								}
								if (!valid) break;
							}
						}

						if (valid)
						{
							int ht = MAX_OF(fs->ceilingheight + fs->skyfix, bs->ceilingheight + bs->skyfix);

							validcount++;

							for (c = k = 0; k < hitssecptr; k++)
							{
								subsector_t *ss = &subsectors[hitssec[k]];
								seg_t		*sg = &segs[ss->firstline];

								ss->ssflags |= SSF_NOOCCLUDE;

								if (ss->sector->ceilingpic == skyflatnum
									&& ss->sector->ceilingheight + ss->sector->skyfix < ht)
								{
									ss->sector->skyfix = ht - ss->sector->ceilingheight;
								}

								for (m = 0; m < ss->linecount; m++, sg++)
								{
									if (sg->linedef && sg->linedef->validcount != validcount)
									{
										sector_t *f, *b;
										int mh;

										sg->linedef->validcount = validcount;

										f = sg->linedef->frontsector;

										if (sg->linedef->ovflags & OVF_NOSKYWALL) continue;

										if (sg->linedef->ml_flags & ML_TWOSIDED)
										{
											b = sg->linedef->backsector;
											mh = MIN_OF(f->ceilingheight + f->skyfix,
														b->ceilingheight + b->skyfix);

										}
										else
											mh = f->ceilingheight + f->skyfix;

										if (mh < ht) linearray[c++] = sg->linedef - lines;
									}
								}
							}

							for (k = 0; k < c; k++)
							{
								SetNoskywall(linearray[k], 22001);
							}
						}
					}
				}

				for (k = 0; k < fs->linecount; k++)
				{
					int lnum = fs->lines[k] - lines;

					if (hitline[lnum] == 4) hitline[lnum] = 8;
				}
			}
		}
	}

	CLR_hitline();
	CLR_conarray();

	// eliminate rings of one sided lines that do not
	// access sectors other than their own
	for (i = 0, fs = sectors; i < numsectors; i++, fs++)
	{
		if (fs->ceilingpic != skyflatnum) continue;

		for (c = k = 0; k < fs->linecount; k++)
		{
			ld = fs->lines[k];

			if ( !(ld->ml_flags & ML_TWOSIDED)
				&& vertexowners[ld->v1 - vertexes].num == 1
				&& vertexowners[ld->v2 - vertexes].num == 1)
			{
				conarray[ld - lines] = 1;
				linearray[c++] = ld - lines;
			}
		}

		if (c > 2)
		{
			while (1)
			{
				int lnum = -1;

				for (k = 0; k < c; k++)
				{
					if (conarray[linearray[k]] == 1)
					{
						lnum = linearray[k];
						break;
					}
				}

				if (lnum < 0) break;

				if (subfn_CheckForRing(conarray, fs, lnum, &rc, NULL, NULL, NULL))
				{
					for (k = 0; k < c; k++)
					{
						if (conarray[linearray[k]] == 4)
							conarray[linearray[k]] = 8;
					}
				}
				else
				{
					conarray[lnum] = 0;
				}
			}
		}
	}

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (  !ld->backsector
			&& conarray[i] != 8
			&& vtocount[ld->v1 - vertexes] == 2
			&& vtocount[ld->v2 - vertexes] == 2
			&& ld->frontsector->ceilingheight - ld->frontsector->floorheight > 3*FRACUNIT)
		{
			if (ld->frontsector->skyfix && ld->frontsector->ceilingpic == skyflatnum)
				hitline[i] = 1;
			linearray[c++] = i;
			hitline[i] |= 2;
		}
	}

	if (c < 3) return;

	repeat = true;

	while (repeat)
	{
		repeat = false;

		for (i = 0; i < c; i++)
		{
			if ( hitline[linearray[i]] & 1 && (d = nf_chkLineListRing(c, box, i, hitline)) )
			{
				int		chksec[2];
				int		chkseccnt = 0;
				RDBOOL	valid = true;

				for (i = k = m = 0; i < c; i++)
				{
					if (hitline[linearray[i]] & 4)
					{
						if (valid)
						{
							int secnum = lines[linearray[i]].frontsector - sectors;

							if (sectors[secnum].ceilingpic != skyflatnum
								&& ++m > 1)
							{
								valid = false;
								break;
							}

							if (!chkseccnt)
								chksec[chkseccnt++] = secnum;
							else if (chksec[0] != secnum)
							{
								if (chkseccnt == 1)
									chksec[chkseccnt++] = secnum;
								else if (chksec[1] != secnum)
								{
									valid = false;
									break;
								}
							}
						}

						linearray[k++] = linearray[i];
					}
				}

				if (valid && k != d) Con_ErrorD("Program error [19]");

				if (   valid
					&& LS_MarkSubsecsInMultiSecLoop(chkseccnt, chksec, d, box, NULL, false, MSIA_MSREVERSED|MSIA_TESTAREA))
				{
					for (i = 0; i < k; i++)
						SetNoskywall(linearray[i], 22002);
				}

				for (i = k = 0; i < c; i++)
				{
					if (hitline[linearray[i]] & 4)
						hitline[linearray[i]] = 0;
					else
						linearray[k++] = linearray[i];
				}

				if ( (c = k) > 2) repeat = true;
				break;
			}
			else
				hitline[linearray[i]] = 0;
		}
	}

	// v220-18 check for any that got missed
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->ml_flags & ML_TWOSIDED
			&& (fs = ld->frontsector) != (bs = ld->backsector)
			&& fs->ceilingpic == skyflatnum
			&& bs->ceilingpic == skyflatnum
			&& !(fs->specFlags & SF_MOVCEIL)
			&& !(bs->specFlags & SF_MOVCEIL))
		{
			if (!fs->skyfix && bs->skyfix)
			{
				int diff = bs->ceilingheight + bs->skyfix - fs->ceilingheight;

				if (diff > 0) fs->skyfix = diff;
			}
			else
			if (fs->skyfix && !bs->skyfix)
			{
				int diff = fs->ceilingheight + fs->skyfix - bs->ceilingheight;

				if (diff > 0) bs->skyfix = diff;
			}
		}
	}
}

// v220-10 check any unset skies from the subsectors
void LS_CheckAdjacentSkyFromSubsector()
{
	subsector_t *ss;
	sector_t	*s;
	seg_t		*sg;
	int			i, k;
	int			isset = 0;

	if (R3D_maptype) return;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (ss->sector->ceilingpic == skyflatnum && !(ss->sector->specFlags & SF_MOVSECTOR))
		{
			fixed_t	ht1 = ss->sector->ceilingheight + ss->sector->skyfix;
			fixed_t	ht2;

			for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
			{
				if (   sg->frontsector
					&& sg->frontsector != ss->sector
					&& sg->frontsector->ceilingpic == skyflatnum)
				{
					ht2 = sg->frontsector->ceilingheight + sg->frontsector->skyfix;

					if (ht2 > ht1)
					{
						if (ss->sector->skyfix < ht2 - ht1) ss->sector->skyfix = ht2 - ht1;
					}
					else if (ht1 > ht2)
					{
						if (sg->frontsector->skyfix < ht1 - ht2) sg->frontsector->skyfix = ht1 - ht2;
					}
				}
			}

			// check for sky floor override
			for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
			{
				if (   sg->linedef
					&& !(sg->linedef->mtflags & R3D_INVLOTEX)
					&& (s = sg->backsector)
					&& s != ss->sector
					&& !(s->specFlags & SF_MOVSECTOR) 
					&& s->floorpic == skyflatnum
					&& s->ceilingpic == ss->sector->floorpic
					&& s->floorheight == ss->sector->floorheight
					&& s->ceilingheight > ss->sector->ceilingheight)
				{
					Setr3dssfs(ss, s - sectors, 22003);
					isset |= 1;
					break;
				}
			}
		}
		else if (ss->sector->floorpic == skyflatnum)
		{
			// check for sky ceiling override
			for (k = 0, sg = &segs[ss->firstline], s = NULL; k < ss->linecount; k++, sg++)
			{
				if (   sg->linedef
					&& !(sg->linedef->mtflags & R3D_INVHITEX) 
					&& (s = sg->backsector)
					&& s != ss->sector
					&& !(s->specFlags & SF_MOVSECTOR) 
					&& s->floorpic == ss->sector->ceilingpic
					&& s->ceilingpic == skyflatnum
					&& s->ceilingheight == ss->sector->ceilingheight)
				{
					Setr3dsscs(ss, s - sectors, 50007);
					isset |= 2;
					break;
				}
			}

		}
	}

	if (!(isset & 2)) return;	// Not bothering with floor ATM (no maps known needing it)

	// check for line same sector sided subsec inheritance
	for (i = 0; i < numlines; i++)
	{
		if ( (s = lines[i].frontsector) == lines[i].backsector
			&& s->ceilingpic != skyflatnum
			&& s->floorpic == skyflatnum)
		{
			ssline_t *sl = &sslines[i];
			int		 ceilsec = -1;

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];

				if (   ss->sector->ceilingpic != skyflatnum
					&& ss->sector->floorpic == skyflatnum)
				{
					if (   ss->ceilsec >= 0
						&& sectors[ss->ceilsec].ceilingpic == skyflatnum
						&& sectors[ss->ceilsec].floorpic == ss->sector->ceilingpic)
					{
						ceilsec = ss->ceilsec;
						break;
					}
				}
			}

			if (ceilsec >= 0)
			{
				for (k = 0; k < sl->num; k++)
				{
					subsector_t *ss = &subsectors[sl->list[k]];

					if (ss->ceilsec != ceilsec) Setr3dsscs(ss, ceilsec, 22004);
				}
			}
		}
	}
}

static int ssecLinesValidForNoSkywall(sector_t *sec)
{
	line_t	*ld;
	int		c, i, k;

	validcount++;

	for (c = i = 0; i < hitssecptr; i++)
	{
		subsector_t *ss = &subsectors[hitssec[i]];
		seg_t		*sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( (ld = sg->linedef) && ld->validcount != validcount)
			{
				ld->validcount = validcount;

				linearray[c++] = ld - lines;
				if (hitline[ld - lines] == 4) continue;
				// This type of construction is quite specific.
				// No sectors inside must equal the loop sector.
				if ( !(ld->ml_flags & ML_TWOSIDED)
					|| ld->frontsector == sec
					|| ld->backsector == sec
					|| (ld->frontsector->ceilingpic != skyflatnum
					&& ld->backsector->ceilingpic != skyflatnum))
				{
					return 0;
				}
			}
		}
	}

	return c;
}

void LS_ResetUnwantedSkyfix()
{
	sector_t	*sec;
	line_t		*ld;
	int			c, i, k, m;

	// Eliminate need to draw a skywall with some line types

	// Mark all 2's skysector lines that form loops in a heightless sector
	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		if (sec->ceilingpic != skyflatnum || sec->stptr >= 0) continue;

		if (sec->ceilingheight == sec->floorheight)
		{
			CLR_hitline();

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   ld->ml_flags & ML_TWOSIDED
					&& ld->frontsector != ld->backsector
					&& ld->frontsector->ceilingpic == ld->backsector->ceilingpic)
				{
					sector_t *s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (s->ceilingheight >= sec->ceilingheight)
					{
						c++; break;
					}
				}
			}

			if (c) continue;

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   ld->ml_flags & ML_TWOSIDED
					&& ld->frontsector != ld->backsector
					&& ld->frontsector->ceilingpic == ld->backsector->ceilingpic)
				{
					hitline[ld - lines] = 1; c++;
				}
			}

			if (c < 3) continue;

			for (k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (hitline[ld - lines] == 1)
				{
					if (subfn_CheckForRing(hitline, sec, ld - lines, NULL, NULL, NULL, NULL))
					{
						for (m = 0; m < sec->linecount; m++)
						{
							int lnum = sec->lines[m] - lines;

							if (hitline[lnum] == 4)
							{
								SetNoskywall(lnum, 22005);
								lines[lnum].ovflags |= OVF_NOOCCLUDE;
								hitline[lnum] = 8;
							}
						}
					}
				}
			}
		}
		// v220-10 check for specific ring that indicates all lines contained
		//         inside its boundary should not have a skywall set
		else if ( (sec->specFlags & 0x1C) == 0x1C)
		{
			sector_t	*osec = NULL;

			for (c = k = 0; k < sec->linecount; k++)
			{
				ld = sec->lines[k];

				if (   !origTEX[ld - lines]
					&& !(ld->ovflags & OVF_NOSKYWALL)
					&& Vtxs2(ld)
					&& ld->backsector
					&& ld->frontsector != ld->backsector
					&& ld->frontsector->ceilingpic == ld->backsector->ceilingpic
					&& ld->frontsector->floorheight == ld->backsector->floorheight
					&& ld->frontsector->ceilingheight == ld->backsector->ceilingheight)
				{
					sector_t *s = ld->frontsector == sec ? ld->backsector : ld->frontsector;

					if (!osec) osec = s;
					else if (osec != s) { c = 0; break; }

					c++;
				}
				else
				{
					c = 0; break;
				}
			}

			if (c == sec->linecount)
			{
				CLR_hitline();

				for (k = 0; k < sec->linecount; k++) hitline[sec->lines[k] - lines] = 1;

				while(1)
				{
					int		rc, box[4];

					int lnum = -1;

					for (k = 0; k < sec->linecount; k++)
					{
						ld = sec->lines[k];

						if (hitline[ld - lines] == 1) { lnum = ld - lines; break; }
					}

					if (lnum < 0) break;

					if (subfn_CheckForRing(hitline, sec, lnum, &rc, box, NULL, NULL) && rc > 2)
					{
						int		lineset = 0;

						if (LS_MarkSubsecsInAreaGeneral(sec, hitline, 0, NULL, box, -1, MSIA_NONE))
						{
							lineset = ssecLinesValidForNoSkywall(sec);
						}

						for (k = 0; k < sec->linecount; k++)
						{
							if (hitline[sec->lines[k] - lines] == 4) hitline[lnum] = 0;
						}

						if (lineset)
						{
							for (k = 0; k < lineset; k++)
								SetNoskywall(linearray[k], 22006);
						}
					}
					else
						hitline[lnum] = 0;
				}

			}
		}
	}

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		sector_t *fs, *bs;
		int		 h1, h2;

		if ( !(ld->ml_flags & ML_TWOSIDED)
			&& origTEX[i] == 0x02
			&& (fs = ld->frontsector)->skyfix
			&& fs->floorheight >= fs->ceilingheight
			&& !(fs->specFlags & SF_MOVSECTOR)
			&& fs->heightsec < 0)	// v220-10 test added
		{
			origTEX[i] = 0;
			sides[ld->sidenum[0]].texFlags &= ~T_MID;
			sides[ld->sidenum[0]].midtexture = 0;
			continue;
		}

		if (   !(ld->ml_flags & ML_TWOSIDED)
			|| ld->ovflags & OVF_NOSKYWALL
			|| (fs = ld->frontsector) == (bs = ld->backsector)
			|| fs->specFlags & SF_MOVSECTOR
			|| bs->specFlags & SF_MOVSECTOR)
		{
			continue;
		}

		// b220-0
		if (   (fs->ceilingpic == skyflatnum)^(bs->ceilingpic == skyflatnum)
			&& ld->lnflags2 & LNF2_OWNGTRTHANTWO
			&& (fs->specFlags & 0x1C) == 0x1C
			&& (bs->specFlags & 0x1C) == 0x1C )
		{
			if (ld->specType == R3D_16)
				Settype(ld, R3D_14, 22007);
			SetSkywall(ld, 22007);
			continue;
		}

		if (fs->ceilingpic != skyflatnum || bs->ceilingpic != skyflatnum)
			continue;

		h1 = fs->ceilingheight - fs->floorheight;
		h2 = bs->ceilingheight - bs->floorheight;

		if (!origTEX[i] && ((h1 <= 0) ^ (h2 <= 0)) )
		{
			if (   !(fs->specFlags & SF_MULTILOOP)
				&& !(bs->specFlags & SF_MULTILOOP))
			{
				SetSkywall(ld, 22008);
			}
		}
		else if (h1 > 64*FRACUNIT && h2 > 64*FRACUNIT)
		{
			if (   ld->lnflags2 & LNF2_OWNEQUALTOTWO
				&& !(origTEX[i] & 0x66)
				&& origTEX[i] & 0x11
				&& h1 > MAX_SECHEIGHT
				&& h2 > MAX_SECHEIGHT
				&& abs(h1 - h2) > MAX_SECHEIGHT
				&& Vtxs2(ld))
			{
				SetSkywall(ld, 22009);
			}
			else if (fs->floorpic != skyflatnum
				&& bs->floorpic != skyflatnum
				&& fs->ceilingheight + fs->skyfix
				== bs->ceilingheight + bs->skyfix)
			{
				SetNoskywall(i, 22010);
			}
		}
	}
}
