//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_sl3D.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  Allow ladder line with no mid textures
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
//  Support for rotating lines removed.
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//  3D lift sectors and 3D flats revised
//
// DESCRIPTION: 3D lines
//  Translates map definitons.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <math.h>
#include "common_vars.h"
#include "r_data.h"
#include "r3d_MAINspec.h"
#include "r3d_edit.h"
#include "r3d_compat.h"
#include "m_misc.h"
#include "m_bams.h"
#include "sys_system.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

typedef struct
{
	int		parmsec;	// Sector with flat pic and light info
	int		rendsec;	// Rendering sector
	int		offset;		// Height above rendsec's floorheight	
	int		depth;		// Line depth
	int		floorheight;// Used for sorting
	int		flags;
} flats3D_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

void P_Add3DSwitch(int texnum1, int texnum2);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// rtparm_t	*rtparm;	// v220-13 $rotate removed
int		newtexcount;
int		flatListsCreated;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int	newgroupcount;
static line3d_t *line3dt_array;

// CODE --------------------------------------------------------------------

static void sl3dErrorReportExit(int errors)
{
	Con_Error("%i map design errors were found for 3D lines or structures\n"
		"These have been printed to \\Risen3D\\R3Ddebug.out\n",
		errors);
}

static int subfn_Get3DPatchNum(float scale, short num, RDBOOL midtex)
{
	int i, pnum = texArray[num]->patches->patch;

	// Check the texture definitions for a match
	for (i = numstandardtextures; i < numtextures; i++)
	{
		if (   texArray[i]->patches->patch == pnum
			&& texscales[i].rendh == (float)texArray[i]->height * scale)
		{
			return i;
		}
	}
	return -1;
}

static void subfn_Add3DTexture(int pnum, float scale, int groupnum)
{
	texture_t	*texture;
	texpatch_t	*patch, *p_org;
	int			i, numtex = numtextures;
	int			patchcount = texArray[pnum]->patchcount;

	//----------------------------------------------

	if (patchcount <= 0)
		Con_Error("Scaled tex specified [%8.8s] has no patches", texArray[pnum]->name);

	if (newtexcount > 0x44A7)	// 26*26*26 - 1
		Con_Error("Scaled new patch limit of %i for 3D lines exceeded", 0x44A7);

	R_SetNewTexMemory(patchcount-1, scale, texArray[pnum]->width, texArray[pnum]->height);

	texture = texArray[numtex];

	// Used for startup texture matching
	// May be changed when a GL texture is created
	texscales[numtex].rendh = (float)texArray[pnum]->height * scale;

	strncpy(texture->oldname, texArray[pnum]->name, 8);
	texture->oldname[8] = 0;
	sprintf(texture->name, "3DTX%c%c%c%c", 0x19,
		'A' + (newtexcount / 26*26) % 26,
		'A' + (newtexcount /    26) % 26,
		'A' +  newtexcount          % 26);

	newtexcount++;	// Keep track of number of 3D textures added

	texture->width		= texArray[pnum]->width;
	texture->height		= texArray[pnum]->height;
	texture->txf_flags	= texArray[pnum]->txf_flags;
	texture->txf_flags	|= (TXF_3DTEX|TXF_DELETE);
	texture->anim_group = groupnum;
	texture->patchcount	= patchcount;

	p_org = &texArray[pnum]->patches[0];
	patch = &texture->patches[0];

	for (i = 0 ; i < patchcount ; i++, patch++, p_org++)
	{
		patch->patch   = p_org->patch;
		patch->originx = p_org->originx;
		patch->originy = p_org->originy;
	}
}

static int Add3DTexture(int pnum, float scale, int linenum)
{
	int		i, n = texArray[pnum]->anim_group, texnum = 0, texnum2 = 0;

	if(n)
	{
		for(i = 0; i < numstandardtextures; i++)
		{
			if(texArray[i]->anim_group == n)
			{

				if (texArray[i]->txf_flags & TXF_SWITCH)
				{
					if (!texnum)
					{
						texnum = numtextures;
						subfn_Add3DTexture(i, scale, newgroupcount + 1);
					}
					else if (!texnum2)
					{
						texnum2 = numtextures;
						subfn_Add3DTexture(i, scale, newgroupcount + 1);
					}
					else
					{
						texnum = 0; break;
					}
				}
				else if (texArray[i]->txf_flags & TXF_ANIMPIC)
				{
					PrintProgInfo("ERROR 3D line %i texture %8s: Scaling animated textures is not supported.\n",
						linenum, texArray[i]->oldname);
					break;
				}
			}
		}

		if (!texnum || !texnum2 || texnum == texnum2)
			Con_Error("Program error [6]: 3D switch on linenum %i", linenum);

		newgroupcount++;
		P_Add3DSwitch(texnum, texnum2);
	}
	else
	{
		texnum = numtextures;
		subfn_Add3DTexture(pnum, scale, 0);
	}

	return texnum;
}

static short subfn_AdjustLightValues(int ll)
{
	if (ll <   0) return   0;
	if (ll > 255) return 255;
	return (short)ll;
}

static RDBOOL getNum(char *buf, int *n, int lnum)
{
	char	*stopstring;

	*n = strtol(buf, &stopstring, 10);

	if (*stopstring != 0 || *n < 0)
	{
		PrintProgInfo("ERROR: 3D line %i has an invalid definition\n", lnum);
		return false;
	}

	return true;
}

static RDBOOL matchLiftSec(int count, int num, int *ls)
{
	int i;

	for (i = 0; i < count; i++)
	{
		if (ls[i] == num) return true;
	}
	return false;
}

static RDBOOL Subfn_setLinked3Dsectors(sector_t *sec, sector_t *lsec)
{
	sector_t	*msec = sec;
	char		ee[] = {"3D LIFT MAP ERROR: "};
	line_t		*ld;
	int			lowestfloor, nextfloor;
	int			i;
	int			error = false;

	// NOTE: Depending on map sector numbering lsec may not be the controlling sector
	//       but the linked sector or could just be the same as sec. It is purely used
	//       to confirm that the structure is a 3D lift sector.

	int		*liftsectors = (int*)M_Calloc(numsectors, sizeof(int));
	int		lcount = 0, checkcount = 0;

	// Find lowest floor
	// Note: this is done *before* assigning listsec to sec
	//       else lowest floor will not necessarily be returned
	for (nextfloor = sec->floorheight;;)
	{
		lowestfloor = P_FindNextLowestFloor(sec, nextfloor);
		if (nextfloor == lowestfloor) break;
		nextfloor = lowestfloor;
	}

	liftsectors[lcount++] = sec - sectors;

	if (lsec)
	{
		liftsectors[lcount++] = lsec - sectors;
		sec = lsec;
	}

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (ld->specType >= R3D_40)
		{
			if (ld->frontsector != sec && ld->frontsector != ld->backsector)
			{
				PrintProgInfo("%sInvalid line [%i] frontsec [%i] should be %i\n",
					ee, ld - lines, ld->frontsector - sectors, sec - sectors);
				error++;
				goto tidyup;
			}

			if (!matchLiftSec(lcount, ld->backsector - sectors, liftsectors) )
			{
				if (ld->backsector->floorheight != lowestfloor)
				{
					PrintProgInfo("%sInvalid sector [%i] floorheight [%i] should be %i\n",
						ee, ld->backsector - sectors, ld->backsector->floorheight>>16,
						lowestfloor>>16);
					error++;
					goto tidyup;
				}

				liftsectors[lcount++] = ld->backsector - sectors;
			}
		}
	}

	// If lcount < 2 then there's a map error as the list should contain the
	// controlling lift sector and at least one other that it references.
	if (lcount < 2)
	{
		PrintProgInfo("%s3D lift sector %i references no 3D sectors\n", ee, sec-sectors);
		error++;
		goto tidyup;
	}

	if (sec == lsec) sec = msec;

	if (lcount)
	{
		if (ge_inf.l3dreport) pr_OFinsert("LISTSEC for S%i :", sec - sectors);

		for (i = 0; i < lcount; i++)
		{
			if (i && sectors[liftsectors[i]].specFlags & SF_MOVSECTOR)
			{
				PrintProgInfo("%s3D lift sector %i can be triggered to change height.\n"
				"All elements forming a 3D lift can only be controlled using an external mov sector\n", ee, liftsectors[i]);
				error++;
			}

			sectors[liftsectors[i]].liftsec3D = sec - sectors;
		}

		sec->listsec = (int*)W_Malloc(lcount*sizeof(int), LEVELID, 23);

		// Don't store first entry as this is the controlling lift sector
		// so checking it for mobj's in p_floor.cpp would be a waste of time
		if (--lcount) memcpy(sec->listsec, &liftsectors[1], lcount*sizeof(int));

		sec->listsec[lcount] = -1;	// set end marker

		if (ge_inf.l3dreport)
		{
			int	*listsec = sec->listsec;

			while ((lcount = *(int*)listsec++) >= 0) pr_OFinsert(" %i", lcount);
		}

		if (ge_inf.l3dreport) pr_OFinsert("\n");
	}

tidyup:;
	M_Free(liftsectors);

	return error;
}

// v220-31 lift sector inheritance checks - called by the analyser

// With 3D lift structures another 3D line structure may have been plugged in.
// LS_3DLiftSectorSSecCheck initially only changes ssecs in the containing sector
// i.e. those between it and the 3D flats outer boundary. As such no internal
// ssecs have yet had their ssec sector references changed. In fact there is no
// need if no internal structures have been 'plugged in'
//
// Having reached this point the internal ssecs, within the lift's boundary, must
// then be checked in case of plugged in structures. Where these exist we can use
// the fact that any structure will have coincident (shared vertex) lines forming
// closed loops which can be used to find the internal ssecs for each loop. Each
// loop's internal ssec must now be set to reference their containing loop's 3D
// line's back sectors (which are all the same value - or should be).


// msecnum = moving sector
// c3dlinemem is an array of the coincident lines

static void subfn_3DLiftInternalSSecCheck(int msecnum, int*c3dlinemem, int c3dlcount, int lac)
{
	subsector_t *ss;
	sector_t	*ms = &sectors[msecnum];	// The moving sector that all lines are tied to
	sector_t	*mbsec = NULL;				// The main bodies sector - to be identified
	sector_t	*cs = NULL;					// the containing sector  - to be identified
	ssline_t	*sl;
	line_t		*ld;
	seg_t		*sg;
	int			c, i, k, m;
	int			errors = 0;

	c = hitssecptr = 0;
	ss_validcount++;

	// linearray holds all lines referencing the 3D lift sector that are not coincident
	// First find the containing sector and check for consistency
	for (c = i = 0; i < lac; i++)
	{
		ld = &lines[linearray[i]];
		sl = &sslines[linearray[i]];

		// The frontsector of a 3D liftline is tied to the liftsector for height calculations.
		// The subsector needs to point to the containing sector NOT the lift sector.
		// ld->frontsector is the lift sector so we need to scan all the 3D lines using it's
		// frontsector side to find the containing sector.

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];
			if (ss->ssvalidcount == ss_validcount) continue;
			ss->ssvalidcount = ss_validcount;
			if (ss->sector == ld->backsector) continue;

			if (ss->sector == ms)
			{
				RDBOOL storessec = true;
				
				for (sg = &segs[ss->firstline], m = 0; m < ss->linecount; m++, sg++)
				{
					if ( !sg->linedef || !sg->frontsector) continue;

					if (hitline[sg->linedef - lines] & 2)
					{
						storessec = false;	// internal ssec, dealt with further down
						break;
					}
				}

				if (!storessec) continue;
				
				for (sg = &segs[ss->firstline], m = 0; m < ss->linecount; m++, sg++)
				{
					if ( !sg->linedef || !sg->frontsector) continue;

					if (sg->frontsector == ms)
					{
						if (   sg->backsector
							&& sg->frontsector != sg->backsector
							&& sg->frontsector->liftsec3D == sg->backsector->liftsec3D
							&& sg->linedef->specType >= R3D_40)
						{
							if (!mbsec) mbsec = sg->backsector;
							else if (sg->backsector != mbsec)
							{
								PrintProgInfo("Error: L%i is in the wrong sector\n", sg->linedef - lines);
								errors++;
							}
						}
					}
					else if (sg->frontsector->liftsec3D == -1)
					{
						if (!cs) cs = sg->frontsector;
						else if (sg->frontsector != cs)
						{
							PrintProgInfo("Error: L%i is in the wrong sector\n", sg->linedef - lines);
							errors++;
						}
					}
				}

				hitssec[hitssecptr++] = ss - subsectors;
			}
		}
	}

	if (errors) sl3dErrorReportExit(errors);

	if (!cs) return;

	if (hitssecptr)
	{
		if (ge_inf.l3dreportc) pr_OFinsert("LIFTSEC %i changing %i external ssec sectors to S%i ====\n",
			msecnum, hitssecptr, cs - sectors);

		for (i = 0; i < hitssecptr; i++)
		{
			ss = &subsectors[hitssec[i]];

			if (ge_inf.l3dreportc) pr_OFinsert("  changing Ssec %5i  sector from %i to %i\n",
										ss - subsectors, ss->sector - sectors, cs - sectors);

			ss->ssflags &= ~SSF_BASESSECFIX;	// just in case
			ChangeSSECsec(ss, cs, 41015);
			ss->ssflags |= SSF_BASESSECFIX;
		}
	}

	if (!mbsec || !c3dlcount) return;

	// Check all coincident line ssecs - none of these should have ms set as the sector

	if (ge_inf.l3dreportc) pr_OFinsert("  LIFTSEC %i checking internal ssecs ====\n", msecnum);

	ss_validcount++;

	for (i = 0; i < c3dlcount; i++)
	{
		sl = &sslines[c3dlinemem[i]];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];
			if (ss->ssvalidcount == ss_validcount) continue;
			ss->ssvalidcount = ss_validcount;

			if (ss->sector == ms)
			{
				sector_t *inherit = mbsec;	// default (ssec may only have one line)

				// First see if a suitable sector can be found. inner nested sectors
				// will not necessarily own the default sector.
				for (m = 0, sg = &segs[ss->firstline]; m < ss->linecount; m++, sg++)
				{
					if (sg->frontsector && sg->frontsector != ms)
					{
						inherit = sg->frontsector; break;
					}
				}

				if (ge_inf.l3dreportc) pr_OFinsert("    changing internal Ssec %5i sector from %i to %i\n",
					ss - subsectors, ss->sector - sectors, inherit - sectors);

				ss->ssflags &= ~SSF_BASESSECFIX;	// just in case
				ChangeSSECsec(ss, inherit, 41016);
				ss->ssflags |= SSF_BASESSECFIX;
			}
		}
	}
}

// v220-31 Identical flat lists can be created for subsecs that prevent
//         address comparisons to work where trying to establish that a
//         Thing is still on the same flats when moving. This routine
//         looks for identical lists and eliminates duplicates after re-
//         setting a ssecs ptr to the shared list.
static void subfn_Collate3DFlats()
{
	int			**flatArrays = (int**)malloc(numsubsectors * sizeof(int));;
	byte		*sseclist = (byte*)calloc(numsubsectors, 1);
	int			*flatslist, *flatslist2;
	int			flatListsRemoved = 0;
	int			i, k, m;

	// compare lists
	for (i = 0; i < numsubsectors; i++)
	{
		if (sseclist[i]) continue;

		if ( !(flatslist = subsectors[i].flats3DArray) ) continue;

		int *ptr = flatslist;
		int c1 = *(int*)ptr++;

		sseclist[i] = true;

		for (k = i; k < numsubsectors; k++)
		{
			if (sseclist[k]) continue;
			if (!(flatslist2 = subsectors[k].flats3DArray) ) continue;
			if (flatslist2 == flatslist) continue;

			int c2 = *(int*)flatslist2++;

			if (c1 != c2) continue;

			int *cmp = ptr;

			for (m = 0; m < c1; m++)
			{
				if (*(int*)cmp++ != *(int*)flatslist2++) goto end_check;
				if (*(int*)cmp++ != *(int*)flatslist2++) goto end_check;
				if (*(int*)cmp++ != *(int*)flatslist2++) goto end_check;
				if (*(int*)cmp++ != *(int*)flatslist2++) goto end_check;
				if (*(int*)cmp++ != *(int*)flatslist2++) goto end_check;
			}

			for (m = 0; m < flatListsRemoved; m++)
			{
				if (flatArrays[m] == subsectors[k].flats3DArray) goto found;
			}

			flatArrays[flatListsRemoved++] = subsectors[k].flats3DArray;
found:;
			sseclist[k] = true;
			subsectors[k].flats3DArray = subsectors[i].flats3DArray;
end_check:;
		}
	}

	for (i = 0; i < flatListsRemoved; i++)
	{
		W_Free(flatArrays[i], LEVELID, 0);
	}

	free(flatArrays);
	free(sseclist);
}

// Public functions --------------------------------------------------

#if 0
// Currently not called
void LS_CheckR3D_4BLines()
{
	subsector_t *ss;
	int			i, k, errors = 0;

	validcount++;

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		seg_t *sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			line_t	*ld = sg->linedef;

			if (   ld
				&& ld->specType == R3D_4B
				&& ld->validcount != validcount
				&& (sg->sidedef->bottomtexture || sg->sidedef->toptexture))
			{
				int *flatslist = ss->flats3DArray;

				if (!flatslist)
				{
					PrintProgInfo("Lit line %i is in subsec with no flats list\n", ld - lines);
					ld->validcount = validcount;
					errors++;
				}
				else if (*(int*)flatslist != 1)
				{
					PrintProgInfo("Lit line %i subsec has flats list with more than one 3Dflat pair\n", ld - lines);
					ld->validcount = validcount;
					errors++;
				}
			}
		}
	}

	if (errors) sl3dErrorReportExit(errors);
}
#endif

static RDBOOL count3(sector_t *s)
{
	int c, i;

	for (c = i = 0; i < s->linecount; i++)
	{
		if (s->lines[i]->specType < R3D_40) c++;
	}

	return c == 3;
}

// Must be called first and before LS_InitTagLists
void LS_Init_Neg_3D_tags()
{
	line_t	*ld;
	int		c, i, r;

	if (!texname3D) return;

	for (c = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (   ld->tag == -2
			|| (texname3D[ld->sidenum[0]].hiname[0] == '$'
			&& strlen(texname3D[ld->sidenum[0]].hiname) == 7)
			|| texname3D[ld->sidenum[0]].midname[0] == '$')
		{
			c++;
		}
	}

	if (c)
	{
		// Create array of 3D line data and set line d3d pointers.

		line3dt_array = (line3d_t*)W_Calloc(c * sizeof(line3d_t), LEVELID, 17);

		for (c = i = 0, ld = lines; i < numlines; i++, ld++)
		{
			if (   ld->tag == -2
				|| (texname3D[ld->sidenum[0]].hiname[0] == '$'
				&& strlen(texname3D[ld->sidenum[0]].hiname) == 7)
				|| texname3D[ld->sidenum[0]].midname[0] == '$')
			{
				ld->d3d = &line3dt_array[c++];

				 // special effects
				ld->d3d->walllowlightlevel	= -1;
				ld->d3d->wallmidlightlevel	= -1;
			}
		}
	}

	for (i = r = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->tag < 0 && ld->tag >= -2 &&
			ld->ml_flags & ML_TWOSIDED)
		{
			side_t	*sd1 = &sides[ld->sidenum[0]];
			side_t	*sd2 = &sides[ld->sidenum[1]];
			int		value = ld->special;

			if (!sd1->bottomtexture ||
				sd1->bottomtexture != sd2->bottomtexture)
				continue;

			// Rotating
			if (ld->tag == -1)
			{
				// v220-13 $rotate removed
				ld->special = 0;
				ld->tag = 0;
			}
			else
			// Scrolling translucent
			if (ld->tag == -2)
			{
				SettypePA(ld, R3D_4A, 16001);
				ld->lnflags2 |= LNF2_LOWTEXMASKED;	// Actually translucent but this is for blocking code calcs
				ld->d3d->lineheightlow = (value << FRACBITS);
				ld->special = 255;
			}
			ld->tag = 0;
		}
	}
}

// only called if texname3D is set
int LS_IdentifyR3D_40Lines(int analyse)
{
	texname3D_t *tx1, *tx2;
	line_t	*ld;
	side_t	*sd1, *sd2;
	float	scale3Dlow, scale3Dmid;
	int		count, i;
	int		num;
	int		n1, n2, spf;
	int		errors = 0;
	byte	type3Ds1, type3Ds2;
	byte	*marksec = (byte*)M_Calloc(numsectors, 294);
	char	buff[6];

	RDBOOL	topset, typeset1, typeset2;
	int		offset3D, heightlow, heightmid;

	// For 3D lift sectors
	int		*mseclist = NULL;
	int		msecmax = 0;
	int		msecc = 0;

	// ----------------------------------------

	for (count = i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (!(ld->ml_flags & ML_TWOSIDED) || ld->specType) continue;

		typeset1 = typeset2 = false;

		// Check for 3D definitions
		sd1 = &sides[ld->sidenum[0]];
		sd2 = &sides[ld->sidenum[1]];
		tx1 = &texname3D[ld->sidenum[0]];
		tx2 = &texname3D[ld->sidenum[1]];

		typeset1 = tx1->hiname[0] == '$';
		typeset2 = tx2->hiname[0] == '$';

		topset = typeset1 && strlen(tx1->hiname) == 7;

		// Check for wall light levels on side 0 only
		if (tx1->midname[0] == '$' && strlen(tx1->midname) == 8)
		{
			int  ht;
			byte typemid = (byte)(tx1->midname[4] - 'A' + 1);

			if (topset)	// indicates standard line types
			{
				PrintProgInfo("ERROR: 3D light level line has top and mid def $ ids %i\n", i);
				errors++;
				continue;
			}

			// Note that 3D lines are always offset to the backsector's
			// floorheight.  This line type must always have the higher
			// floor height on the backsector.  As such it can be checked
			// for blocking either as a standard or 3D line.
			if (typemid == 1
				&& (ht = ld->backsector->floorheight - ld->frontsector->floorheight) > 0)
			{
				SettypePA(ld, R3D_4B, 16001);

				buff[3] = 0;
				strncpy(buff, tx1->midname + 1, 3);

				if (getNum(buff, &num, i))
				{
					ld->d3d->walllowlightlevel = subfn_AdjustLightValues(num);

					strncpy(buff, tx1->midname + 5, 3);
					getNum(buff, &num, i);
					ld->d3d->wallhilightlevel = subfn_AdjustLightValues(num);
				}

				ld->d3d->lineheightlow = ht;

				count++;
			}
			continue;
		}

		if (!typeset1 && !typeset2) continue;

		if ((typeset1 ^ typeset2) || ((typeset1 || typeset2) && !ld->d3d))
		{
			PrintProgInfo("ERROR: 3D line %i has missing or invalid definitions\n", ld - lines);
			errors++;
			continue;
		}

		heightlow = heightmid = 0;

		if (topset)
		{
			type3Ds1 = (byte)(tx1->hiname[1] - 'A' + 1);
			strncpy(buff, tx1->hiname + 2, 5);
			buff[5] = 0;
			getNum(buff, &num, i);
			offset3D = num<<16;
			typeset1 = false;
		}
		
		if (typeset2 && strlen(tx2->hiname) == 8)
		{
			buff[3] = 0;
			type3Ds2 = (byte)(tx2->hiname[4] - 'A' + 1);
			strncpy(buff, tx2->hiname + 1, 3);
			
			if (getNum(buff, &num, i))
			{
				heightlow = num<<16;

				strncpy(buff, tx2->hiname + 5, 3);
				getNum(buff, &num, i);
				heightmid = num<<16;
			}
			typeset2 = false;
		}

		if (typeset1 || typeset2)
		{
			char	*err = "invalid";

			if (   typeset1 && typeset2
				&& strlen(tx1->hiname) == 8 && strlen(tx2->hiname) == 7)
			{
				type3Ds2 = (byte)(tx2->hiname[1] - 'A' + 1);
				type3Ds1 = (byte)(tx1->hiname[4] - 'A' + 1);
			
				if (type3Ds1 >= 1 && type3Ds1 <= 3 &&
					type3Ds2 >= 1 && type3Ds2 <= 5)
				{
					err = "transposed";
				}
			}

			PrintProgInfo("ERROR: 3D line %i has %s hi definitions\n", ld - lines, err);
			errors++;
			continue;
		}

		if (strlen(tx2->midname) == 8 && tx2->midname[0] == '$')
		{
			byte typemid = (byte)(tx2->midname[4] - 'A' + 1);

			if (typemid == 1)
			{
				buff[3] = 0;
				strncpy(buff, tx2->midname + 1, 3);

				if (getNum(buff, &num, i))
				{
					ld->d3d->walllowlightlevel = subfn_AdjustLightValues(num);
					strncpy(buff, tx2->midname + 5, 3);
					getNum(buff, &num, i);
					ld->d3d->wallmidlightlevel = subfn_AdjustLightValues(num);
				}
			}
			else
				continue;
		}

		// Mark any sectors referenced by 3D lnes that are mov floor sectors
		// Note that these should only be elevator types
		n1 = ld->frontsector - sectors;
		n2 = ld->backsector - sectors;

		if (n1 != n2 && !marksec[n1])
		{
			marksec[n1] = true;

			if (sectors[n1].specFlags & SF_MOVSECTOR)
			{
				if ( !( (spf = sectors[n1].specFlags2) & (SF2_ELEVATOR3D|SF2_ELEVATOR)) )
				{
					if (ge_inf.devmap)
						PrintProgInfo("WARNING: sector used with L%i is not an elevator type\n", i);
				}
				else if (spf & SF2_ELEVATOR)
				{
					PrintProgInfo("ERROR: Elevator sector used with L%i has invalid trigger type in range 235 to 238\n", i);
					errors++;
				}
				else
				{
					if (msecc >= msecmax)	// OK, can't be >, but defensive coding never hurt
					{
						msecmax += 16;
						mseclist = (int*)M_Realloc(mseclist, msecmax*sizeof(int), 37);
					}
					mseclist[msecc++] = n1;
				}
			}
		}

		if (n1 != n2 && !marksec[n2])
		{
			marksec[n2] = true;
			if (sectors[n2].specFlags & SF_MOVSECTOR)
			{
				PrintProgInfo("ERROR: A move sector cannot be used with a 3D line's backsector (unless it's the same as the frontsector)\n", i);
				errors++;
			}
		}

		if (   !sd1->midtexture
			&& !sd1->bottomtexture
			&& !sd2->midtexture
			&& !sd2->bottomtexture)
		{
			ld->spflags |= SPF_NOTEX;
		}

		// Special lines
		if (type3Ds1 >= 1 && type3Ds1 <= 3 &&
			type3Ds2 >= 1 && type3Ds2 <= 5)
		{
			int		lowtexheight = 0;
			int		midtexheight = 0;
			int		tbh = 0, tmh = 0;

			scale3Dlow = scale3Dmid = 1.0f;

			// identify 3D switches
			if (ld->tag && ld->special && sd1->bottomtexture)
				ld->ovflags |= OVF_3DSWITCH;

			// FIXME - needs an algo to set lighting on back of line
			if (   ld->d3d->walllowlightlevel == -1
				&& ld->d3d->wallmidlightlevel == -1
				&& ld->backsector->sector3D >= 0
				&& ld->backsector->heightsec == -1)
			{
				ld->d3d->walllowlightlevel =
				ld->d3d->wallmidlightlevel =
				subfn_AdjustLightValues(sectors[ld->backsector->sector3D].floorheight>>16);
			}

			// Get texture heights if present
			// First check for ladder. only midtex allowed
			if (   type3Ds2 == 5
				&& ld->backsector == ld->frontsector
				&& sd1->midtexture == sd2->midtexture)
			{
				ld->ovflags |= OVF_3DLADDERLINE;
				tbh = heightlow;
				tmh = heightmid;
				SettypePA(ld, R3D_40, 16002);
			}
			else
			{
				if (sd1->bottomtexture)
					tbh = texArray[sd1->bottomtexture]->height << FRACBITS;

				if (sd1->midtexture)
					tmh = texArray[sd1->midtexture]->height << FRACBITS;

				if ((type3Ds2 == 2 || type3Ds2 == 4) && heightlow && tbh)
					scale3Dlow = (float)(heightlow) / (float)(tbh);
				else if (heightlow)
					tbh = heightlow;

				if ((type3Ds2 == 3 || type3Ds2 == 4) && heightmid && tmh)
					scale3Dmid = (float)(heightmid) / (float)(tmh);
				else if (heightmid)
					tmh = heightmid;

				if (type3Ds1 == 2)
					SettypePA(ld, R3D_41, 16003);
				else
					SettypePA(ld, R3D_40, 16004);

				if (ld->backsector->sector3D >= 0)
					SettypePA(ld, (byte)(ld->specType + 2), 16003);	// R3D_42, R3D_43
			}

			// Don't allow enlargement
			if (scale3Dlow > 1.0f) scale3Dlow = 1.0f;
			if (scale3Dmid > 1.0f) scale3Dmid = 1.0f;

			// Rounded up to nearest integer by adding 0.5f
			lowtexheight = (int)(scale3Dlow * (float)(tbh) + 0.5f);
			midtexheight = (int)(scale3Dmid * (float)(tmh) + 0.5f);

			ld->d3d->totalheight = lowtexheight + midtexheight;

			ld->d3d->lineheightlow = lowtexheight;
			ld->d3d->lineheightmid = midtexheight;

			ld->d3d->scale3Dlow = (fixed_t)(scale3Dlow*FRACUNIT);
			ld->d3d->scale3Dmid = (fixed_t)(scale3Dmid*FRACUNIT);
			ld->d3d->offset3D = offset3D;

			count++;
		}
	}

	if (!errors)
	{
		sector_t	*sc, *sd;
		int			k, m;

		//Set up any lift sectors found
		for (i = 0; i < msecc; i++)
		{
			if (Subfn_setLinked3Dsectors(&sectors[mseclist[i]], NULL) ) { errors++; break; }
		}

		// check for linked lift type with 3D lines only
		for (i = 0, sc = sectors; i < numsectors; i++, sc++)
		{
			if (sc->specFlags2 & SF2_ELEVATOR3D)
			{
				for (k = 0; k < msecc; k++)
				{
					if (mseclist[k] == i) goto next_msec;
				}

				sc->specFlags2 &= ~SF2_ELEVATOR3D;	// until proven otherwise

				// search for a connecting line
				for (k = 0; k < sc->linecount; k++)
				{
					if (sc->lines[k]->backsector == sc )
					{
						sd = sc->lines[k]->frontsector;

						for (m = 0; m < sd->linecount; m++)
						{
							if (sd->lines[m]->specType >= R3D_40
								&& count3(sd))
							{
								if (Subfn_setLinked3Dsectors(sc, sd)) { errors++; break; }
								sc->specFlags2 |= SF2_ELEVATOR3D;	// linked to 3D lines
								goto next_msec;
							}
						}
					}
				}
			}
next_msec:;
		}
	}

	M_Free(mseclist);
	M_Free(marksec);

	if (errors) sl3dErrorReportExit(errors);

	return count;
}

void LS_Set3DSpecialConditions()
{
	if (!R3D_linemap) return;

	subsector_t *ss;
	line_t		*ld;
	int			i;

	if (flatListsCreated)
	{
		for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
		{
			// Set scaled doom type texture 3D flats
			if (ss->flats3DArray)
			{
				sector_t	*lsec;
				int			*flatslist = ss->flats3DArray;
				int			k, flats3D_count = *(int*)flatslist++;
				int			flags = 0, movflags = F3D_MARKER;

				if (ss->flats3DArray[4] & F3D_MARKER) continue;

				for (k = 0; k < flats3D_count; k++)
				{
					lsec = &sectors[flatslist[0]];	// Sector with flats/lightlevel definitions

					if (sectors[flatslist[1]].liftsec3D >= 0)	// Sector to check for movement
						movflags |= (F3D_MOVFLAG << k);

					// Precache
					R_GetFlatFlags(lsec->ceilingpic);
					R_GetFlatFlags(lsec->floorpic);

					flatslist += FLIST_ENTRIES;
				}

				ss->flats3DArray[5] |= movflags;
			}
		}
	}

	if (R3D_edit) return;

	// Set 3D lines masked flags
	// These determine whether a bullet can pass a masked mid-texture
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->specType >= R3D_40)
		{
			if (origTEX[i] & 0x33)	// lower and/or middle textures present
			{
				side_t	*sd1 = &sides[ld->sidenum[0]];
				side_t	*sd2 = &sides[ld->sidenum[1]];
				int		type = 0;

				// v220-31 Make bottom alpha tex only line with a mid height of 1 (FRACUNIT) block
				if (   origTEX[i] == 0x11
					&& ld->d3d->lineheightmid == FRACUNIT
					&& texArray[sd1->bottomtexture]->txf_flags & TXF_MASKED
					&& texArray[sd2->bottomtexture]->txf_flags & TXF_MASKED)
				{
					ld->d3d->totalheight -= FRACUNIT;	// adjust back
					ld->d3d->lineheightmid = 0;
				}
				else
				if (  (origTEX[i] & 0x11) == 0x11
					&& texArray[sd1->bottomtexture]->txf_flags & TXF_MASKED
					&& texArray[sd2->bottomtexture]->txf_flags & TXF_MASKED)
				{
					ld->lnflags2 |= LNF2_LOWTEXMASKED;
					type++;
				}
				if (  (origTEX[i] & 0x22) == 0x22
					&& texArray[sd1->midtexture]->txf_flags & TXF_MASKED
					&& texArray[sd2->midtexture]->txf_flags & TXF_MASKED)
				{
					ld->lnflags2 |= LNF2_MIDTEXMASKED;
					type++;
				}
				if (type == 2)	ld->lnflags2 |= LNF2_3DNOBLOCK;
			}
			// Note - lines with no textures and with no mid height must not block
			else if (!(ld->d3d->lineheightmid))
			{
				ld->lnflags2 |= LNF2_3DNOBLOCK;
			}
		}
	}
	
	if (flatListsCreated) subfn_Collate3DFlats();
}

// Check scaled textures - each unique scale will need a new texture
void LS_Check3DScaledTexs()
{
	line_t	*ld;
	int		i, texnum;

	newgroupcount = anim_group_count;
	newtexcount = 0;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if (ld->specType >= R3D_40 && ld->specType != R3D_4A)
		{
			float	slow = FIX2FLT(ld->d3d->scale3Dlow);
			float	smid = FIX2FLT(ld->d3d->scale3Dmid);
			side_t	*sd1 = &sides[ld->sidenum[0]];
			side_t	*sd2 = &sides[ld->sidenum[1]];
			short	tnum;

			if (slow && slow != 1.0f)
			{
				if ((tnum = sd1->bottomtexture) > 0)
				{
					texnum = subfn_Get3DPatchNum(slow, tnum, false);
					if (texnum < 0)
					{
						texnum = Add3DTexture(tnum, slow, i);
					}
					sd1->bottomtexture = texnum;
				}
				if ((tnum = sd2->bottomtexture) > 0)
				{
					texnum = subfn_Get3DPatchNum(slow, tnum, false);
					if (texnum < 0)
					{
						texnum = Add3DTexture(tnum, slow, i);
					}
					sd2->bottomtexture = texnum;
				}
			}
			if (smid && smid != 1.0f)
			{
				if ((tnum = sd1->midtexture) > 0)
				{
					texnum = subfn_Get3DPatchNum(smid, tnum, true);
					if (texnum < 0)
					{
						texnum = Add3DTexture(tnum, smid, i);
					}
					sd1->midtexture = texnum;
				}
				if ((tnum = sd2->midtexture) > 0)
				{
					texnum = subfn_Get3DPatchNum(smid, tnum, true);
					if (texnum < 0)
					{
						texnum = Add3DTexture(tnum, smid, i);
					}
					sd2->midtexture = texnum;
				}
			}
		}
	}
}

void LS_SetSSFlats3DArray(void)
{
	subsector_t *ssec = subsectors;
	int			i, k, errors = 0;

	byte		*marksec  = (byte*)M_Calloc(numsectors, 295);
	byte		*markline = (byte*)M_Calloc(numlines  , 296);
	flats3D_t	*flats3D  = (flats3D_t*)M_Calloc(MAX_FLATS3D*2*sizeof(flats3D_t), 297);

	// ---------------------------------------------

	for (i = 0; i < MAX_FLATS3D; i++)
		flats3D[i].parmsec = flats3D[i].rendsec = -1;

	for (i = 0; i < numsubsectors; i++, ssec++)
	{
		seg_t	*seg = &segs[ssec->firstline];
		int		secnum, count = 0;

		for(k = 0; k < ssec->linecount; k++, seg++)
		{
			line_t	*ld = seg->linedef;

			if (   ld != NULL
				&& ld->specType >= R3D_40
				&& ld->specType <= R3D_43
				&& ld->backsector == seg->frontsector
				&& (secnum = ld->backsector->sector3D) >= 0)
			{
				sector_t	*sec3D = &sectors[secnum];
				int			k, found = false;

				if (count >= MAX_FLATS3D)
					Con_Error("Too many stacked 3D flats in sector %i"
							  "Max allowed is %i pairs.", secnum, MAX_FLATS3D);

				if (!marksec[secnum])
				{
					// Sector conveying flats textures does not have floor or ceiling
					// heights used.  So these have been hijacked to store lightlevels
					sec3D->floorlight3D = subfn_AdjustLightValues(sec3D->floorheight   >> 16);
					sec3D->ceillight3D  = subfn_AdjustLightValues(sec3D->ceilingheight >> 16);
					marksec[secnum] = 1;
				}

				flats3D[count].parmsec = secnum;
				flats3D[count].offset  = ld->d3d->offset3D;
				flats3D[count].depth   = ld->d3d->lineheightlow;
				flats3D[count].flags   = 0;

				// Check if same flat
				if (count)
				{
					for (k = 0; k < count; k++)
					{
						if (   flats3D[k].parmsec == secnum
							&& flats3D[k].offset  == flats3D[count].offset)
						{
							found = true;

							if (   !markline[ld - lines]
								&& flats3D[k].depth != flats3D[count].depth)
							{
								pr_OFinsert("WARNING: Flats3D L%5i has side 2 depth $%.3i---- when expecting $%.3i----\n",
									ld - lines, flats3D[count].depth>>16, flats3D[k].depth>>16); 
								markline[ld - lines] = 1;
							}
							break;
						}
					}

					if (found) continue;
				}

				if (ld->frontsector->liftsec3D >= 0)
					flats3D[count].rendsec = ld->frontsector->liftsec3D;
				else
					flats3D[count].rendsec = ld->backsector - sectors;

				// Transfer any set splash flags
				flats3D[count].flags |= (R_GetFlatFlags(sec3D->floorpic) & FLT_SPLASH);

				// Heights are used for sorting below
				flats3D[count].floorheight =
					sectors[flats3D[count].rendsec].floorheight + ld->d3d->offset3D;
				count++;
			}
		}
			
		if (count > 1 && (ma_inf.modifier & MODF_ZDNODG) )
		{
			M_Free(flats3D);
			M_Free(markline);
			M_Free(marksec);
			Con_Error("PROBLEM: Zdbsp nodes are not compatible with more\n"
				"than 1 R3D stacked floors. Count %i in sector %i\n"
				"R3D glbsp nodes must be used.", count, ssec->sector - sectors);
		}

		if (count)
		{
			// v220-31 height ordering method revised
			int *flatslist = (int*)W_Malloc((count*FLIST_ENTRIES + 1) * sizeof(int), LEVELID, 24);
			int done[MAX_FLATS3D];

			memset(done, 0, sizeof(done));
			flatListsCreated++;

			ssec->flats3DArray = flatslist;

			*(int*)flatslist++ = count;

			// Sort list into ascending ceiling heights
			while (1)
			{
				flats3D_t *ptr = flats3D, *tempptr = NULL;
				int index, lowest = DDMAXINT;

				for (k = 0; k < count; k++, ptr++)
				{
					if (ptr->parmsec < 0 || ptr->rendsec < 0) Con_Error("ERROR: flatslist creation with ssec %i", i);

					int height = sectors[ptr->rendsec].floorheight + ptr->offset;

					if (!done[k] && height < lowest)
					{
						lowest = height;
						tempptr = ptr;
						index = k;
					}
				}

				if (!tempptr) break;

				done[index] = true;

				*(int*)flatslist++ = tempptr->parmsec;
				*(int*)flatslist++ = tempptr->rendsec;
				*(int*)flatslist++ = tempptr->offset;
				*(int*)flatslist++ = tempptr->depth;
				*(int*)flatslist++ = tempptr->flags;
			}
		}
	}

	M_Free(flats3D);
	M_Free(markline);
	M_Free(marksec);

	if (errors) sl3dErrorReportExit(errors);
}

void LS_3DLiftSectorSSecCheck()
{
	sector_t	*sc;
	int			*liftsecnum = (int*)malloc(numsectors*sizeof(int));
	int			*c3dlinemem = (int*)malloc(numlines*sizeof(int)/2);
	int			i, k, lac;
	int			liftseccount = 0;
	int			c3dlcount = 0;		// coincident line count

	for (i = 0, sc = sectors; i < numsectors; i++, sc++)
	{
		if (sc->liftsec3D >= 0) // At this point just use this test
		{
			for (k = 0; k < liftseccount; k++)
			{
				if (liftsecnum[k] == sc->liftsec3D)
					goto find_nextls;
			}

			liftsecnum[liftseccount++] = sc->liftsec3D;
		}
find_nextls:;
	}

	if (R3D_edit && liftseccount && !ge_inf.devmap)
	{
		Sys_MessageBox(	"Note: If line vertexes are unmerged in the map being edited then \n"
						"rendering errors not normally encountered may be seen especially \n"
						"with 3D Lift sectors which are present in this map.\n\n"
						"Vertexes are normally merged when the nodes are built hiding the \n"
						"fact that they may not be merged in the map but this cannot be \n"
						"done for a map being edited\n", false);
	}

	for (i = 0; i < liftseccount; i++)
	{
		line_t	*ld, *li;
		
		if (ge_inf.l3dreportc) pr_OFinsert("\nLIFTSEC %i checking coincident lines ================================\n", liftsecnum[i]);

		CLR_hitline();

		for (c3dlcount = k = 0, ld = lines; k < numlines; k++, ld++)
		{
			if ( !(hitline[k] & 2)
				&& ld->frontsector->liftsec3D == liftsecnum[i])
			{
				if ( (li = subfn_FindCoincident3DLine(ld, 16)) )
				{
					// store first found. If there are others then this is of no consequence
					// as it is only the vertexes needed later on and they are shared
					c3dlinemem[c3dlcount++] = k;

					hitline[k] |= 2;
					hitline[li - lines] |= 2;	// prevent testing this one

					if (ge_inf.l3dreportc) pr_OFinsert("COINCIDENT L%i L%i\n", k, li - lines);
				}
			}
		}

		for (lac = k = 0, ld = lines; k < numlines; k++, ld++)
		{
			if ( !(hitline[k] & 2)
				&& ld->frontsector->liftsec3D == liftsecnum[i])
			{
				linearray[lac++] = k;
			}
		}

		if (lac < 3) continue;	// Some weird map

		subfn_3DLiftInternalSSecCheck(liftsecnum[i], c3dlinemem, c3dlcount, lac);
	}

	if (ge_inf.l3dreport) pr_OFdirect("End liftsec checks\n\n");

	free(c3dlinemem);
	free(liftsecnum);
}

int LS_SetFlatPointer(fixed_t floor, fixed_t ht, RDBOOL hi, char *ptr, int *flatslist)
{
	int c = *(int*)flatslist++;
	int ceil = DDMININT;
	int n = 0;

	ht += sectors[flatslist[1]].floorheight;

	while(n < c)
	{
		ceil = sectors[flatslist[1]].floorheight + flatslist[2];

		if (ht >= floor && ht < ceil + flatslist[3]) break;

		floor = sectors[flatslist[1]].floorheight + flatslist[2] + flatslist[3];

		if (++n == c) break;

		flatslist += FLIST_ENTRIES;
	}

	*ptr = n;

	return hi ? ceil : floor;
}

// To be called after floor smoothing has been carried out (e.g. when rendering)
void LS_GetFlatPointer(int n, sector_t *sc, fixed_t *floor, fixed_t *ceil, int *flatslist)
{
	int c = *(int*)flatslist++;

	if (n <= 0)
	{
		*floor = sc->floorheight;
		*ceil =  sectors[flatslist[1]].smoothfloor + flatslist[2];
	}
	else if (n >= (c - 1))
	{
		flatslist += (c - 1)*FLIST_ENTRIES;
		*floor = sectors[flatslist[1]].smoothfloor + flatslist[2] + flatslist[3];
		*ceil = sc->ceilingheight;
	}
	else
	{
		flatslist += n*FLIST_ENTRIES;
		*floor = sectors[flatslist[1]].smoothfloor + flatslist[2] + flatslist[3];
		flatslist += FLIST_ENTRIES;
		*ceil =  sectors[flatslist[1]].smoothfloor + flatslist[2];
	}
}

// ---------------------------------------
