//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_ssecchk.c
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 31 (v220-31) 11/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION: early checks for subsector->sector errors
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <memory.h>

#include "r3d_MAINspec.h"
#include "m_misc.h"
#include "p_maputil.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

// B211-0
// Check if a moving sector is passing through a solid fixed sector.
// If so then adopt the fixed not the moving sector for rendering.
// Yeah, it happens e.g. prwrsv07.wad ssec bounded by lines 1641, 1772, 1773
static void chkDiffsecs(subsector_t *ss)
{
	if (ss->sector->specFlags & SF_MOVSECTOR)
	{
		seg_t	*sg = &segs[ss->firstline];
		int		i;

		for (i = 0; i < ss->linecount; i++, sg++)
		{
			if (   sg->backsector
				&& !(sg->frontsector->specFlags & SF_MOVSECTOR)
				&& !(sg->backsector->specFlags  & SF_MOVSECTOR)
				&& sg->frontsector->floorheight > sg->backsector->floorheight)
			{
				ss->ssflags |= SSF_NOCHANGE;
				return;
			}
		}
	}
}

// v220-5 added test
static RDBOOL ssecValid(sector_t *s, subsector_t *ss, RDBOOL hasmove)
{
	seg_t	*sg = &segs[ss->firstline];
	int		i, ret = false;

	if (s->specFlags & SF_NOLOTEX) return false;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if (sg->frontsector != sg->backsector) return false;
		if (hasmove && sg->frontsector->specFlags & SF_MOVFLOOR)
			ret = true;
	}
	return hasmove ? ret : true;
}

static int		unresolved, sgarraycnt;
static RDBOOL	badssecmessage = true;

// Check ssec->sector assignments
static void chkSSECassignments(subsector_t	*ss)
{
	seg_t		*sg;
	int			i, k;

	int secptr  [6] = { -1, -1, -1, -1, -1, -1 };
	int linptr  [6] = { -1, -1, -1, -1, -1, -1 };
	int seccount[6] = { 0, 0, 0, 0, 0, 0 };
	int segsec  [6] = { 0, 0, 0, 0, 0, 0 };
	int segnobk [6] = { 0, 0, 0, 0, 0, 0 };
	int segmtht [6] = { DDMAXINT, DDMAXINT,
						DDMAXINT, DDMAXINT,
						DDMAXINT, DDMAXINT };

	int num;
	int cnt = ss->linecount;
	int maxcount = 0, segcount = 0;

	RDBOOL	heightfix = true, fixchange = false;
	RDBOOL	firstown = false;
	RDBOOL	mset = false;

	line_t	*li, *chk = NULL;
	seg_t	*longest = NULL;
	RDBOOL	do_length_count = true;

	//-----------------------------------------------------------

	sg = &segs[ss->firstline];

	// v220-31 eliminate ssecs with R3D linemap types
	for (i = 0; i < cnt; i++, sg++)
	{
		if (  !sg->linedef	// should never be true
			|| sg->linedef->specType >= R3D_40)
		{
			return;
		}
	}

	sg = &segs[ss->firstline];

	if (sg->frontsector != sg->backsector)
	{
		longest = sg;

		// v220-21 added
		if (cnt == 4 && ss->sector == sg->frontsector)
		{
			sector_t *s = sg->frontsector, *chk = NULL;
			int		 c, d;

			// v220-22 special case types
			if (ss->sector->specFlags & SF_MOVCEIL)
			{
				line_t	*la[4];

				for (c = d = i = 0; i < cnt; i++, sg++)
				{
					if (   sg->frontsector == sg->backsector
						|| sg->frontsector != s
						|| !sg->backsector)
					{
						chk = NULL; break;
					}

					if (sg->backsector->ceilingheight == s->ceilingheight) { chk = NULL; break; }
					if (origTEX[sg->linedef - lines] & 0x44) { chk = NULL; break; }
					if (!(origTEX[sg->linedef - lines] & 0x11)) d++;
					if (!chk) chk = sg->backsector; else if (chk != sg->backsector) { chk = NULL; break; }
					la[c++] = sg->linedef;
				}

				if (   chk
					&& chk->ceilingpic != skyflatnum
					&& (!(chk->specFlags & SF_NOHITEX) || chk->specFlags & SF_SSLINETRUE) )
				{
					int rc, lnum;

					sg = &segs[ss->firstline];
					lnum = sg->linedef - lines;
					CLR_hitline();

					for (i = 0; i < cnt; i++, sg++)
					{
						if (hitline[sg->linedef - lines]) { chk = NULL; break; }
						hitline[sg->linedef - lines] = 1;
					}

					if (subfn_CheckForRing(hitline, s, lnum, &rc, NULL, NULL, NULL) && rc == 4)
					{
						Setr3dsslcsec(ss, chk - sectors, 31001);
						if (d == 4 && chk->floorheight > s->floorheight)
							Setr3dsslfsec(ss, chk - sectors, 31001);

						for (i = 0; i < 4; i++) Settype(la[i], R3D_36, 31001);
						return;
					}
				}
			}
			else
			{
				for (i = 0; i < cnt; i++, sg++)
				{
					if (   sg->frontsector == sg->backsector
						|| sg->frontsector != s
						|| !sg->backsector)
					{
						chk = NULL; break;
					}

					if (sg->backsector->ceilingheight > s->ceilingheight) { chk = NULL; break; }
					if (sg->backsector->ceilingheight == s->ceilingheight) continue;
					if (origTEX[sg->linedef - lines] & 0x44) { chk = NULL; break; }
					// v220-23 Check if a missing texture what side this is on
					if (sg->backsector->ceilingheight < sg->frontsector->ceilingheight) { chk = NULL; break; }
					if (!chk) chk = sg->backsector; else { chk = NULL; break; }
				}

				if (   chk
					&& chk->ceilingpic != skyflatnum
					&& !(chk->specFlags & SF_MOVCEIL)
					&& (!(chk->specFlags & SF_NOHITEX) || chk->specFlags & SF_SSLINETRUE) )
				{
					int rc, lnum;

					sg = &segs[ss->firstline];
					lnum = sg->linedef - lines;
					CLR_hitline();

					for (i = 0; i < cnt; i++, sg++)
					{
						if (hitline[sg->linedef - lines]) { chk = NULL; break; }
						hitline[sg->linedef - lines] = 1;
					}

					if (subfn_CheckForRing(hitline, s, lnum, &rc, NULL, NULL, NULL) && rc == 4)
					{
						Setr3dsslcsec(ss, chk - sectors, 31002);
						return;
					}
				}
			}
		}

		sg = &segs[ss->firstline];
	}
	// v220-6 fix bad map assignment to lines connecting one's lines
	//        This is where the subsec only has one line preventing
	//        analysis from the ssec's lines
	else if (cnt == 1
		&& vertexowners[sg->linedef->v1 - vertexes].num == 2
		&& vertexowners[sg->linedef->v2 - vertexes].num == 2
		&& vertexowners[sg->linedef->v1 - vertexes].list[0] == vertexowners[sg->linedef->v2 - vertexes].list[0]
		&& vertexowners[sg->linedef->v1 - vertexes].list[1] == vertexowners[sg->linedef->v2 - vertexes].list[1]
		&& (li = subfn_FindLineConnected(sg->linedef, 0))
		&& li->frontsector != ss->sector
		&& li->frontsector->floorheight   <= ss->sector->floorheight
		&& li->frontsector->ceilingheight >= ss->sector->ceilingheight)
	{
		ss->sector = li->frontsector;
		return;
	}
	
	for (i = 0; i < cnt; i++, sg++)
	{
		if (sg->linedef)
		{
			if (   !chk
				&& sg->backsector
				&& cnt > 4
				// use blocking to see if this is somewhat suspect. If there was
				// any doubt in the author's mind or there were problems with any
				// original nodes builds then this would likely be set.
				&& sg->linedef->ml_flags & ML_BLOCKING
				&& sg->backsector->floorheight < sg->frontsector->floorheight
				&& !(origTEX[sg->linedef - lines] & 0x11))
			{
				chk = sg->linedef;
			}

			if (do_length_count && lxarray[sg->linedef - lines]) do_length_count = false;

			if (i && do_length_count)
			{
				if (sg->frontsector != sg->backsector && (!longest || sg->length > longest->length))
					longest = sg;
			}

			if (sg->linedef->frontsector != sg->linedef->backsector)
				heightfix = false;

			if (mset && Vtxs1(sg->linedef)) continue;
		}

		if (sg->frontsector)
		{
			int secnum = sg->frontsector - sectors;

			segcount++;

			for (k = 0; k < 6; k++)
			{
				if (secptr[k] < 0 || secptr[k] == secnum)
				{
					if (k >= maxcount) maxcount = k + 1;

					if (   !k
						&& (sg->frontsector == sg->backsector
						||  sg->linedef->lnflags2 & LNF2_OWNBOTHGTRTWO))
					{
						firstown = true;
					}
					else if (k && !firstown
						&& sg->linedef->lnflags2 & LNF2_OWNBOTHGTRTWO)
					{
						goto exitinnerloops;
					}

					seccount[k]++;

					if (secptr[k] < 0)
					{
						if (   vtxarray[sg->linedef->v1 - vertexes] > 1
							&& vtxarray[sg->linedef->v2 - vertexes] > 1)
						{
							mset = true;
						}

						secptr[k] = secnum;
					}

					if (linptr[k] < 0) linptr[k] = sg->linedef - lines;

					if (!sg->backsector)
						segnobk[k] = 1;

					if (sg->frontsector == sg->backsector)
					{
						if (sg->sidedef->midtexture)
						{
							int ht = texArray[sg->sidedef->midtexture]->height;
							if (ht < segmtht[k]) segmtht[k] = ht;
						}
						else
							segmtht[k] = 0;

						segsec[k]++;
					}
					else
					{
						segmtht[k] = DDMININT;
					}
					break;
				}
			}

			if (maxcount == 6) break;	// Work with what we've got
		}
	}

	if (chk)
	{
		seg_t *same = NULL;

		for (i = 0, sg = &segs[ss->firstline]; i < cnt; i++, sg++)
		{
			if (sg->frontsector == sg->backsector)
			{
				same = sg; break;
			}
		}

		if (same)
		{
			sector_t *s = same->frontsector;

			same = NULL;

			for (i = 0, sg = &segs[ss->firstline]; i < cnt; i++, sg++)
			{
				if (   sg->frontsector != sg->backsector
					&& sg->backsector
					&& sg->frontsector == s)
				{
					same = sg; break;
				}
			}

			if (same)
			{
				// NOTE: even if remaining unchanged this test must be retained
				//       to prevent changes being following on.

				if (ss->sector != same->frontsector)
					ChangeSSECsec(ss, same->frontsector, 31003);

				return;
			}
		}
	}

	if (maxcount == 6 && badssecmessage)
	{
		badssecmessage = false;
		// more than 4 usually indicates a map error (e.g. 2's line that should be a 1's line)
		pr_OFinsert("Map warning: Subsector %i sector %i references more than 6 sectors\n", ss - subsectors, ss->sector - sectors);
	}

	if (!maxcount)
	{
		hitssec[unresolved++] = ss - subsectors;
		return;
	}

	num = -1;

	if (maxcount == 2)
	{
		if (   sectors[secptr[0]].specFlags & SF_NOLOTEX
			&& sectors[secptr[0]].specFlags & SF_NOHITEX
			&& sectors[secptr[1]].specFlags & SF_NOLOTEX
			&& sectors[secptr[1]].specFlags & SF_NOHITEX
			&& !(sectors[secptr[0]].specFlags & SF_MOVSECTOR)
			&& !(sectors[secptr[1]].specFlags & SF_MOVSECTOR))
		{
			num = sectors[secptr[0]].floorheight > sectors[secptr[1]].floorheight ? 0 : 1;
		}
		else
		if (   seccount[0] == 1
			&& seccount[1] > 1
			&& segsec[0] == 1
			&& segsec[1] == 0)
		{
			num = 1;
		}
		else
		if (   sectors[secptr[0]].heightsec >= 0
			&& sectors[secptr[1]].heightsec <  0)
		{
			if (!origTEX[linptr[0]]) num = 0;
		}
		else
		if (   sectors[secptr[0]].heightsec <  0
			&& sectors[secptr[1]].heightsec >= 0)
		{
			if (!origTEX[linptr[1]]) num = 1;
		}
		else
		if (!segsec[0] ^ !segsec[1])
		{
			if (   sectors[secptr[0]].ceilingpic == skyflatnum
				&& sectors[secptr[1]].ceilingpic != skyflatnum
				&& sectors[secptr[0]].ceilingheight
				>  sectors[secptr[1]].ceilingheight)
			{
				num = 1;
			}
			else
			if (   sectors[secptr[0]].ceilingpic != skyflatnum
				&& sectors[secptr[1]].ceilingpic == skyflatnum
				&& sectors[secptr[0]].ceilingheight
				<  sectors[secptr[1]].ceilingheight)
			{
				num = 0;
			}
		}
		else if ((!segnobk[0] ^ !segnobk[1])
			   && !segsec[0] && !segsec[1])
		{
			if (   segnobk[0]
				&& sectors[secptr[0]].ceilingpic == skyflatnum
				&& sectors[secptr[1]].ceilingpic != skyflatnum
				&& sectors[secptr[0]].ceilingheight
				>  sectors[secptr[1]].ceilingheight)
			{
				num = 1;
			}
			else
			if (   segnobk[1]
				&& sectors[secptr[0]].ceilingpic != skyflatnum
				&& sectors[secptr[1]].ceilingpic == skyflatnum
				&& sectors[secptr[0]].ceilingheight
				<  sectors[secptr[1]].ceilingheight)
			{
				num = 0;
			}
		}
		else if (segsec[0] && segsec[1] && seccount[0] == 1 && seccount[1] == 1)
		{
			// This sets the rendering sector directly
			if (!origTEX[linptr[0]] && origTEX[linptr[1]] == 0x22)
			{
				num = 0;
			}
			else if (!origTEX[linptr[1]] && origTEX[linptr[0]] == 0x22)
			{
				num = 1;
			}
		}
	}

	if (num < 0)
	{
		for (i = 0; i < maxcount; i++)
		{
			if (segsec[i] && segnobk[i])
			{
				num = i; break;
			}
		}
	}

	if (num < 0)
	{
		for (i = k = 0; k < maxcount; k++)
		{
			if (!segsec[k])
				if (seccount[k] > i) num = k, i = seccount[k];
		}

		if (num < 0 && heightfix)
		{
			RDBOOL notset = false;

			for (i = 0; i < maxcount; i++)
			{
				if (segmtht[i] == DDMAXINT || segmtht[i] == DDMAXINT)
				{
					notset = true;
					break;
				}
			}

			if (!notset)
			{
				int ht = DDMAXINT;

				for (i = 0; i < maxcount; i++)
				{
					if (segmtht[i] < ht) num = i, ht = segmtht[i];
				}
				if (num >= 0) fixchange = true;
			}
		}

		if (num < 0)
		{
			for (i = 0, k = 0; k < maxcount; k++)
				if (seccount[k] > i) num = k, i = seccount[k];
		}

		if (num < 0) num = 0;
	}

	if (ss->sector - sectors != secptr[num]
		// test added $v113 (
		&& !(ss->sector->specFlags & SF_MOVFLOOR
		&& !(sectors[secptr[num]].specFlags & SF_MOVFLOOR)) )
	{
		RDBOOL dochange = true;

		// test added v220-5
		// Look for subsecs where the lines are all superimposed on the
		// first lines and eliminate them from any further tests. These
		// types will not have had a valid midpoint set so the first
		// test is to check for these types.
		if (   ss->ssflags & DDSUBF_MIDPOINT
			&& ss->linecount > 1)
		{
			seg_t	*sg = &segs[ss->firstline];
			line_t	*ld = sg->linedef;
			int		c = 0;

			if (sg->linedef && sg->backsector)
			{
				if (sg->linedef->specType >= R3D_40)
					goto end_test;

				sg++;

				for (i = 1; i < cnt; i++, sg++)
				{
					if (sg->linedef && sg->linedef->specType < R3D_40)
					{
						if (   sg->linedef == ld
							|| !VertexIsOnLine(sg->linedef->v1, ld)
							|| !VertexIsOnLine(sg->linedef->v2, ld))
						{
							goto exit_test;
						}
						c++;
					}
				}
			}

			if (c)
			{
				goto end_test;
			}
		}

exit_test:;
		ma_inf.dbcount[0]++;
		chkDiffsecs(ss);

		// v220-18 weed out nodes build faults
		if (ss->linecount > 3)
		{
			seg_t	*sg = &segs[ss->firstline];
			line_t	*ld = sg->linedef;

			for (i = 0; i < ss->linecount; i++, sg++)
			{
				if (   sg->frontsector == ss->sector
					&& sg->backsector - sectors == secptr[num])
				{
					dochange = false; break;
				}
			}
		}

		if (dochange)
		{
			// v220-16 test added
			if (   ss->linecount > 3
				&& !(ss->sector->specFlags & SF_MOVSECTOR)
				&& ssecValid(ss->sector, ss, true))
			{
				SetOverride(ss, ss->sector - sectors, 31004);
			}

			// v220-33 test added
			if (!(ss->ssflags & SSF_CONTAINER))
				ChangeSSECsec(ss, &sectors[secptr[num]], 31004);
		}
	}

end_test:;
	if (fixchange)
			ss->ssflags |= SSF_BASESSECFIX;

	if (maxcount == 6)
	{
		seg_t *sg = &segs[ss->firstline];

		for (i = 0; i < ss->linecount; i++, sg++)
		{
			if (sg->linedef) sg->linedef->lnflags2 |= LNF2_BADSSECLN;
		}
	}
exitinnerloops:;

	// v220-5 use the sg frontsector with the dominant length in certain cases
	if (   do_length_count
		&& longest
		&& longest->frontsector != ss->sector
		&& longest->frontsector->floorpic != ss->sector->floorpic
		&& longest->frontsector->floorheight == ss->sector->floorheight
		&& !(ss->sector->specFlags & SF_MOVSECTOR))
	{
		ChangeSSECsec(ss, longest->frontsector, 0);
	}
}

static sector_t *subfn_SupportLine(subsector_t *ss, int *heightsec)
{
	sector_t	*s = NULL;
	line_t		*ld;
	seg_t		*sg = &segs[ss->firstline];
	int			i, count = 0;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		if ((ld = sg->linedef) != NULL && LSS(ld))
		{
			if (hitline[ld - lines] == 1)
			{
				if (!s) s = ld->frontsector;
				if (s != ld->frontsector) return NULL;
			}
			else if (Chk0304Type(ld) == R3D_03)
			{
				if (ld->frontsector->heightsec >= 0)
					*heightsec = ld->frontsector->heightsec;
				count++;
			}
			else
				return NULL;
		}
	}
	return count ? s : NULL;
}

subsector_t *R_PointInSubsector (fixed_t x, fixed_t y);

static RDBOOL checkVirtual(subsector_t *ss)
{
	int i, c, d, ret = false;

	if (ss->linecount == 2)
	{
		seg_t *sg;

		c = -1, d = -1;

		for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
		{
			if (sg->linedef && sg->linedef->specType >= R3D_40) return false;	// v220-31 inhibit for R3D lines

			if (   sg->frontsector == sg->backsector
				&& !(sg->frontsector->specFlags & (SF_MOVSECTOR|SF_NOLOTEX|SF_NOHITEX)) )
			{
				if (ss->sector == sg->frontsector)
				{
					c = sg->frontsector - sectors;
				}
				else
				{
					d = sg->frontsector - sectors;
				}
			}
		}

		if (   c >= 0 && d >= 0
			&& sectors[d].floorheight < sectors[c].floorheight
			&& ss->sector - sectors != d)
		{
			SetOverride(ss, d, 31005);
			return true;
		}
		return false;
	}

	// v220-33 If an R3D line exists in a ssec where all other lines belong
	//         to another sector then the other sector should prevail.
	if (ss->linecount > 2)
	{
		sector_t	*os = NULL;
		seg_t		*sg;

		for (c = d = i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
		{
			if (sg->linedef && sg->linedef->specType >= R3D_40)
			{
				if (sg->linedef->frontsector != sg->linedef->backsector) { os = NULL; break; }
				if (ss->sector != sg->frontsector) c++;
			}
			else
			if (sg->linedef && sg->linedef->frontsector == sg->linedef->backsector)
			{
				// do nothing
			}
			else
			if (sg->frontsector != ss->sector)
			{
				if (!os) os = sg->frontsector, c++; else if ( os != sg->frontsector) { os = NULL; break; }
			}
			else
			{
				d++;
			}
		}

		if (os && c && !d) ChangeSSECsec(ss, os, 31006);
	}

	// v220-3 modified
	// Note: test depends on minisegs having been removed as sg->linedef is not checked
	if (ss->linecount > 2)
	{
		sector_t	*os = NULL;
		seg_t		*sg;

		for (c = d = i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
		{
			if (sg->linedef && sg->linedef->specType >= R3D_40) return false;	// v220-31 inhibit for R3D lines

			if (   sg->frontsector == sg->backsector
				&& ss->sector == sg->frontsector)
			{
				c++;
			}

			if (!sg->backsector && sg->frontsector != ss->sector)
			{
				os = sg->frontsector; d++;	// v220-3
			}
		}

		if (c > 0 && c < ss->linecount)
		{
			int	  ns = -1;

			if (os)	// v220-3
			{
				for (c = i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
				{
					if (os == sg->frontsector) c++;
				}

				if (c && c + d == ss->linecount)
				{
					ChangeSSECsec(ss, os, 31007);
					return true;
				}
			}

			for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
			{
				if (sg->frontsector == sg->backsector)
				{
					if (sg->frontsector != ss->sector)
					{
						return false;
					}
				}
				else
				{
					if (ns < 0)
						ns = sg->frontsector - sectors;
					else if (ns != sg->frontsector - sectors)
					{
						return false;
					}
					if (sg->backsector)
					{
						if (sg->backsector->floorheight != sg->frontsector->floorheight
							&& !(origTEX[sg->linedef - lines] & 0x11))
						{
							return false;
						}

						if (sg->backsector->ceilingheight != sg->frontsector->ceilingheight
							&& !(sg->frontsector->ceilingpic == skyflatnum
							&&   sg->backsector->ceilingpic  == skyflatnum)
							&& !(origTEX[sg->linedef - lines] & 0x44))
						{
							return false;
						}
					}
				}
			}

			if (ns >= 0 && ss->sector - sectors != ns)
			{
				SetOverride(ss, ns, 31008);
				return true;
			}
		}
	}
	return false;
}


static RDBOOL checkLogical(subsector_t *ss)
{
	sector_t	*newsec = NULL, *ovs = NULL;
	seg_t		*sg = &segs[ss->firstline];
	int			fht = ss->sector->floorheight;
	int			i;
	RDBOOL		retry = false;
	RDBOOL		liqflr = IsLiquidFloorFlat(ss->sector);

	if (ss->ssflags & SSF_BASESSECFIX) return false;

	// double floor move sec test removed v220-33

	for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		line_t		*ld = sg->linedef;

		if (ld)
		{
			if (   !ld->backsector
				|| Vtxs1(ld)
				|| (lxarray[ld - lines] & 3 && (origTEX[ld - lines] & 0x44))
				|| !sg->backsector	// can be even if ld has a backsector
				|| ld->specType >= R3D_40
				|| ld->frontsector->specFlags & SF_MOVSECTOR)
			{
				return false;
			}

			if (sg->frontsector != sg->backsector)
			{
				if (sg->frontsector->floorheight == sg->backsector->floorheight)
				{
					ovs = sg->frontsector;
					continue;
				}
				return false;
			}

			if ((origTEX[ld - lines] & 0x22) != 0x22) return false;

			if (sg->frontsector->floorheight > fht)
			{
				fht = sg->frontsector->floorheight;
				newsec = sg->frontsector;
			}
		}
	}

	if (newsec && newsec != ss->sector)
	{
		if (ovs && ovs->floorheight < newsec->floorheight)
			SetOverride(ss, ovs - sectors, 31009);

		ChangeSSECsec(ss, newsec, 31009);
		
		return true;
	}
	return false;
}

// Look for constructs that can upset
// ssec->sector assignment by nodes builders.
static void checkDisconnected()
{
	sector_t	*sec;
	line_t		*ld;
	int			i;

	for (i = 0, sec = sectors; i < numsectors; i++, sec++)
	{
		sector_t *s = NULL;
		int k, count = 0;

		if (secdone[i] ||
			sec->specFlags & SF_MOVSECTOR ||
			IsLiquidFloorFlat(sec))
			continue;

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];

			if (!(ld->ml_flags & ML_TWOSIDED) ||
				origTEX[ld - lines] ||
				vtxarray[ld->v1 - vertexes] < 2 ||
				vtxarray[ld->v2 - vertexes] < 2)
			{
				s = NULL;
				break;
			}

			if (ld->frontsector == ld->backsector)
			{
				int vf1 = vtxflags[ld->v1 - vertexes] == 3;
				int vf2 = vtxflags[ld->v2 - vertexes] == 3;

				if ((vf1 && vf2) || !(vf1 ^ vf2))
				{
					s = NULL;
					break;
				}
				count++;
			}
			else
			{
				if (s ||
					ld->frontsector == sec ||
					!Vtfl3(ld))
				{
					s = NULL;
					break;
				}
				s = ld->frontsector;
			}
		}

		if (!count || !s) continue;

		for (k = 0; k < sec->linecount; k++)
		{
			ld = sec->lines[k];
			Clr_allmtflags(ld, 23002);
			Settype(ld, R3D_01, 23004);
			Setspec(ld, s - sectors, 23026);
		}
		Setsecdone2(sec, 23005);
	}
}

// Check confusion between various same sided lines in one subsec
// e.g. line 1649 in Requiem map13
static RDBOOL CheckAssignment0(subsector_t *ss)
{
	sector_t *fsecptr =NULL;
	seg_t	 *sg = &segs[ss->firstline];
	int		 secn[6] = { -1, -1, -1, -1, -1, -1 };
	int		 i, k, count = 0, sscnt = 0;
	RDBOOL	 found = false;

	if (ss->sector->specFlags & SF_MOVSECTOR) return false;

	for (i = 0; i < ss->linecount; i++, sg++)
	{
		line_t	 *ld = sg->linedef;
		
		if (!ld) continue;

		if (!(ld->ml_flags & ML_TWOSIDED)
			|| ld->frontsector->specFlags & SF_MOVSECTOR
			|| ld->backsector->specFlags & SF_MOVSECTOR)
		{
			return false;
		}

		if (ld->frontsector == ld->backsector)
		{
			int sn = ld->frontsector - sectors;

			if (!(ld->lnflags2 & LNF2_OWNBOTHGTRTWO))
			{
				if (!fsecptr)
					fsecptr = sg->frontsector;
				else if (fsecptr != sg->frontsector)
					return false;
				if (ss->sector == fsecptr) return false;
			}

			for (k = 0; k < 6; k++)
			{
				if (secn[k] == sn) continue;

				if (secn[k] < 0)
				{
					secn[k] = sn; sscnt++;
				}
			}
		}
		else if (sg->sidedef->bottomtexture
			&& !sg->sidedef->toptexture
			&& sg->frontsector->floorheight < sg->backsector->floorheight)
		{
			if (!fsecptr)
				fsecptr = sg->frontsector;
			else if (fsecptr != sg->frontsector)
				return false;

			if (ss->sector == fsecptr) return false;
			found = true;
		}
		else
			return false;
	}

	if (found && sscnt && fsecptr)
	{
		ChangeSSECsec(ss, fsecptr, 31010);
	}
	return false;
}

// Check if ssec set to support line sec where all other lines are step lines
static RDBOOL CheckAssignment1(subsector_t *ss)
{
	sector_t *s1 = NULL, *s2 = NULL;
	line_t	 *ld;
	seg_t	 *sg;
	int		 i;

	if (!(ss->sector->specFlags & SF_MOVFLOOR)) return false;

	for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		if ((ld = sg->linedef) != NULL &&
			ld->specType < R3D_40)
		{
			if (ld->frontsector != ld->backsector)
				return false;

			if (ld->frontsector->specFlags & SF_MOVFLOOR)
			{
				if (!Vtxs1(ld) || origTEX[ld - lines])
					return false;

				if (!s1)
					s1 = ld->frontsector;
				else
					return false;

				if (s1 != ss->sector)
					return false;
			}
			else
			{
				if (origTEX[ld - lines] != 0x22)
					return false;

				if (!s2) s2 = sg->frontsector;

				if (s2 != sg->frontsector ||
					s2 == ss->sector)
					return false;
			}
		}
	}
	if (s1 && s2)
	{
		if (ss->ssflags & SSF_BASESSECFIX)
			SetOverride(ss, s2 - sectors, 31011);
		else
			ChangeSSECsec(ss, s2, 31011);
		return true;
	}
	return false;
}

// Sets support line ssecs (modified v220-18)
static RDBOOL CheckAssignment2(subsector_t *ss)
{
	extern byte walkCheck[];
	line_t	*ld, *ls = NULL, *support = NULL;
	seg_t	*sg;
	int		i;

	//---------------------------------

	if (ss->ssflags & SSF_NOCHANGE) return 0;
	if (ss->sector->specFlags & SF_MOVSECTOR) return 0;

	for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		if ( !(ld = sg->linedef) ) continue;

		if (ld->frontsector == ld->backsector)
		{
			if (   !i
				&& !origTEX[ld - lines]
				&& ld->tag
				&& ld->special > 0
				&& ld->special < 270
				&& walkCheck[ld->special] == 1)
			{
				return false;
			}

			// v220-3
			if (ld->frontsector->specFlags & SF_MOVFLOOR)
			{
				if (  (support && support->frontsector->floorheight > ld->frontsector->floorheight)
					|| ld->frontsector->floorheight > ss->sector->floorheight)
				{
					support = ld;
				}
			}
		}
		else if ( !(ld->ml_flags & ML_TWOSIDED) )
		{
			if (!ls && ss->sector == ld->frontsector)
				ls = ld;
		}
		else if (ss->sector == ld->frontsector || ss->sector == ld->backsector)
		{
			if (   !(ld->frontsector->specFlags & SF_MOVSECTOR)
				&& !(ld->backsector->specFlags  & SF_MOVSECTOR) )
			{
				return false;
			}
		}
	}

	if (ls && support)
	{
		SetOverride(ss, ss->sector - sectors, 31012);	// The rendering sector
		ss->sector = ls->frontsector;					// The logical sector

		return true;
	}

	return false;
}

// Some subsecs although having a line with no backsector, which should
// take precedence, contain a two sided line where the sg->frontsector
// should override the 1's line ('creative' maps again).
// E.g line loop in 10sector map17 with lines 3266, 3417, 3430, 4469
//       or loop in 10sector map29 with lines 1175, 1204, 1209, 1210
static RDBOOL CheckAssignment3(subsector_t *ss)
{
	sector_t	*chk = NULL;
	line_t		*lc = NULL;
	seg_t		*sg;
	int			i, flags, lncount = 0, s1count = 0, s2count = 0;

	//-------------------------------------------

	if (   ss->sector->ceilingpic != skyflatnum
		|| ss->sector->specFlags & SF_MOVSECTOR)
	{
		return false;
	}

	validcount++;

	// Search for a closed loop
	for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		line_t *ld = sg->linedef;

		if (!ld || ld->validcount == validcount) continue;

		ld->validcount = validcount;

		if (ld->ml_flags & ML_TWOSIDED)
		{
			if (   ld->frontsector == ld->backsector
				|| sg->frontsector == ss->sector
				|| sg->sidedef->midtexture)
			{
				return false;
			}

			if (!chk)
				chk = sg->frontsector;
			else if (chk != sg->frontsector)
			{
				return false;
			}

			if (   sg->sidedef->bottomtexture
				&& sg->sidedef->toptexture)
			{
				linearray[lncount++] = ld - lines;
				continue;
			}

			flags = LS_TestLineTex(ld);

			if (sg->sidedef->toptexture || !(flags & LT_HI))
			{
				return false;
			}
			
			s2count++;
		}
		else
		{
			if (ld->frontsector != ss->sector)
			{
				return false;
			}
			s1count++;

			if (s1count > 2) return false;

			if (!lc)
				lc = ld;
			else
			{
				// check where a second one sided line that they are not connected
				if (ld->v1->x == lc->v1->x && ld->v1->y == lc->v1->y) return false;
				if (ld->v1->x == lc->v2->x && ld->v1->y == lc->v2->y) return false;
				if (ld->v2->x == lc->v1->x && ld->v2->y == lc->v1->y) return false;
				if (ld->v2->x == lc->v2->x && ld->v2->y == lc->v2->y) return false;
			}
		}

		linearray[lncount++] = ld - lines;

		if (lncount > 4)
		{
			return false;
		}
	}

	if (lncount != 4 || !chk || !s1count || s2count != 2)
	{
		return false;
	}

	if (nf_chkLineListRing(lncount, NULL, 0, NULL) == lncount)
	{
		ChangeSSECsec(ss, chk, 31013);
		SetSSF_BASESSECFIX(ss, 31013);
		return true;
	}
	return false;
}

// Check for trigger lines in the wrong sector
static void assignFromSSLines()
{
	line_t	*ld;
	int		i, k, m;

	for (i = 0; i < numlines; i++)
	{
		if (   lines[i].special
			&& lines[i].tag
			&& lines[i].specType < R3D_40
			&& lines[i].frontsector == lines[i].backsector
			&& Vtxs1(&lines[i])
			&& !origTEX[i]
			&& !lxarray[i])
		{
			sector_t *os = NULL, *chk = lines[i].frontsector;
			ssline_t *sl = &sslines[i];
			int		mask = 0;	// Check ssecs are found both sides of the line

			hitssecptr = 0;

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];
				seg_t		*sg = &segs[ss->firstline];

				hitssec[hitssecptr++] = sl->list[k];

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if ( !(ld = sg->linedef) ) continue;
					// v220-16 ignore same sector lines
					if (ld->frontsector == ld->backsector) continue;

					if (ld - lines == i)
					{
						if (!(mask & 1) && sg->sidedef - sides == lines[i].sidenum[0]) mask |= 1;
						if (!(mask & 2) && sg->sidedef - sides == lines[i].sidenum[1]) mask |= 2;
					}
					else
					{
						if (!os)
						{
							os = sg->frontsector;
							if (os == chk) goto end;
						}
						else
							if (sg->frontsector != os) goto end;
					}
				}
			}

			if (os && mask == 3)
			{
				for (k = 0; k < hitssecptr; k++)
				{
					subsectors[hitssec[k]].ssflags &= ~SSF_BASESSECFIX;
					ChangeSSECsec(&subsectors[hitssec[k]], os, 31014);
				}
			}
		}
end:;
	}
}

// v220-10
static RDBOOL CheckAssignment4(subsector_t *ss)
{
	sector_t	*s = NULL;
	seg_t		*sg, *refsg;
	int			i;

	if (ss->linecount < 2) return false;

	for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
	{
		if (   ss->sector == sg->frontsector
			&& sg->backsector
			&& sg->backsector != sg->frontsector)
		{
			s = sg->backsector;
			refsg = sg;
			break;
		}
	}

	if (s)
	{
		for (i = 0, sg = &segs[ss->firstline]; i < ss->linecount; i++, sg++)
		{
			if (sg == refsg) continue;
			if (sg->length < 1.0f) return false;	// ignore short segs
			if (sg->backsector || sg->frontsector != refsg->backsector) return false;
		}

		ChangeSSECsec(ss, s, 31015);
		return true;
	}

	return false;
}

// v220-33 sslinering sectors check (type that's guaranteed to fuck up any nodes builder) ===========

typedef struct {
	fixed_t	bbox[4];
	int		secnum;
} sslinering_t;

static sslinering_t *sslinering;
static int	sslineringcount;
static int	maxsslineringcount;

// Check for a containing sector
static RDBOOL confirmNew(sector_t *nw)
{
	line_t	*ld;
	fixed_t	bbox[4];
	int		c, i, rc, lnum = -1;

	for (i = 0; i < sslineringcount; i++)
	{
		if (sslinering[i].secnum == nw - sectors) return true;
	}

	CLR_conarray();

	for (c = i = 0; i < nw->linecount; i++)
	{
		ld = nw->lines[i];
		if (ld->frontsector != ld->backsector)
		{
			lnum = ld - lines;
			conarray[lnum] = 1; c++;
		}
	}

	if (c > 2 && subfn_CheckForRing(conarray, nw, lnum, &rc, bbox, NULL, NULL) && rc == c)
	{
		if (sslineringcount >= maxsslineringcount)
		{
			maxsslineringcount += 16;
			sslinering = (sslinering_t*)M_Realloc(sslinering, maxsslineringcount * sizeof(sslinering_t), 39);
		}
		memcpy(sslinering[sslineringcount].bbox, bbox, sizeof(bbox));
		sslinering[sslineringcount++].secnum = nw - sectors;
		return true;
	}
	return false;
}

static void chkForContainingSecRef(sector_t *sc)
{
	int		i, lnum = -1;

	for (i = 0; i < sc->linecount; i++)
	{
		lnum = sc->lines[i] - lines;
		if (hitline[lnum] != 1) return;
	}

	int		bbox[4], k, rc;

	if (subfn_CheckForRing(hitline, sc, lnum, &rc, bbox, NULL, NULL) && rc == sc->linecount)
	{
		subsector_t **surrogate = (subsector_t**)M_Calloc(numsubsectors * sizeof(subsector_t**), 0);
		subsector_t	*ss;
		sector_t	*nw;
		line_t		*ld;
		seg_t		*sg;
		fixed_t		x, y;
		int			side;
		int			surrogate_count = 0;
		float		fbox[4];
		
		fbox[BOXLEFT ] = FIX2FLT(bbox[BOXLEFT ]);
		fbox[BOXRIGHT] = FIX2FLT(bbox[BOXRIGHT]);
		fbox[BOXBASE ] = FIX2FLT(bbox[BOXBASE ]);
		fbox[BOXTOP  ] = FIX2FLT(bbox[BOXTOP  ]);

		hitssecptr = 0;

		for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
		{
			if (ss->sector == sc)
			{
				if (   ss->midpoint.x > fbox[BOXLEFT ]
					&& ss->midpoint.x < fbox[BOXRIGHT]
					&& ss->midpoint.y > fbox[BOXBASE ]
					&& ss->midpoint.y < fbox[BOXTOP  ])
				{
					hitssec[hitssecptr++] = i;
					ss->ssflags |= (SSF_CONTAINER|SSF_BASESSECFIX);
				}
				else
				{
					surrogate[surrogate_count++] = ss;
				}
			}
		}

		for (i = 0; i < surrogate_count; i++)
		{
			ss = surrogate[i];

			for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
			{
				if (  (ld = sg->linedef)
					&& ld->frontsector != ld->backsector
					&& sg->frontsector != sc)
				{
					x = (fixed_t)(ss->midpoint.x*FRACUNIT);
					y = (fixed_t)(ss->midpoint.y*FRACUNIT);
					side = P_PointOnLineSide(x, y, ld);
					nw = side ? ld->backsector : ld->frontsector;

					if (nw != sc && confirmNew(nw)) goto end_check;
				}
			}
		}
end_check:;
		M_Free(surrogate);
	}
}

static sector_t *getSecSurrounding(sector_t *con)
{
	sector_t *os, *other = NULL;
	line_t	 *ld;
	int		 i;

	if (con->specFlags & SF_NOHITEX && con->ceilingpic != skyflatnum)
	{
		for (i = 0; i < con->linecount; i++)
		{
			ld = con->lines[i];

			if (hitline[ld - lines] == 4)
			{
				os = ld->backsector == con ? ld->frontsector : ld->backsector;
				if (!os || os == con) return NULL;	// should not happen;
				if (os->ceilingheight > con->ceilingheight) return NULL;
				if (!other) other = os; else if (other != os) return NULL;
			}
		}
	}

	return other;
}

static void subfn_RingSectors()
{
	line_t	*ld = lines;
	int		i, k, csnum, secid;

	CLR_hitline();
	sslinering = NULL;
	sslineringcount = maxsslineringcount = 0;

	for (i = csnum = 0; i < numlines; i++, ld++)
	{
		if (   ld->frontsector == ld->backsector
			&& (!ld->dx || !ld->dy)
			&& ld->frontsector->linecount > 3
			&& !(lxarray[i] & 0x3F)
			&& vtxarray[ld->v1 - vertexes] == 2
			&& vtxarray[ld->v2 - vertexes] == 2)
		{
			hitline[i] = 1;
			secid = ld->frontsector - sectors;
			for (k = 0; k < csnum; k++)
			{
				if (countsec[k] == secid) goto found_secid;
			}
			countsec[csnum++] = secid;
		}
found_secid:;
	}

	if (csnum)
	{
		subsector_t *ss;
		int			m;

		// Get line rings that may be in a ring sector
		for (i = 0; i < csnum; i++) chkForContainingSecRef(&sectors[countsec[i]]);

		// If rings were found then set the ssec->sec references accordingly
		for (i = 0; i < sslineringcount; i++)
		{
			sector_t *con = &sectors[sslinering[i].secnum];
			int c;

			CLR_hitline();

			for (c = k = 0; k < con->linecount; k++)
			{
				if (con->lines[k]->frontsector != con->lines[k]->backsector)
				{
					hitline[con->lines[k] - lines] = 4;
				}
				else
				{
					c++;
				}
			}

			if (LS_MarkSubsecsInAreaGeneral(con, hitline, 0, NULL, sslinering[i].bbox, -1, MSIA_NONE))
			{
				sector_t *os = getSecSurrounding(con);
				sector_t *cs = NULL;

				if (con->linecount == c)
				{
					for (m = 0; m <  hitssecptr; m++)
					{
						if (!cs) cs = subsectors[hitssec[m]].sector;
						else if (subsectors[hitssec[m]].sector != cs)
						{
							cs = NULL; break;
						}
					}
				}

				if (!cs) continue;
#ifdef _DEBUG
	pr_OFinsert("CHANGING ssline containing ssecs for S%i OS%i\n", con - sectors, os ? os - sectors : -1);
#endif
				for (m = 0; m <  hitssecptr; m++)
				{
					ss = &subsectors[hitssec[m]];

					if (os) Setr3dsslcsec(ss, os - sectors, 31016);

					if (ss->ssflags & SSF_CONTAINER)
					{
						if (ss->sector != con)
							Setr3dssfs(ss, con - sectors, 31016);
						continue;
					}

					if (ss->sector != con)
					{
#ifdef _DEBUG
	pr_OFinsert("  ssec %i changed from S%i to S%i\n", hitssec[m], ss->sector - sectors,con - sectors);
#endif
						ss->ssflags &= ~SSF_BASESSECFIX;
						ChangeSSECsec(ss, con, 31016);
						if (!secdone[ss->sector - sectors]) Setsecdone(ss->sector, 31016);
					}

					ss->ssflags |= (SSF_CONTAINER|SSF_BASESSECFIX);
				}
				if (!secdone[con - sectors]) Setsecdone(con, 31016);
			}
		}
	}

	M_Free(sslinering);
}

// end ssline sectors check ------------------------------------------==========================

// Note: many tests depends on minisegs having been removed as sg->linedef may not checked
void LS_CheckBaseSSECsectors()
{
	subsector_t *ss;
	int			i;

	// v220-33 First do stuff that's guaranteed to fuck the nodes builder
	subfn_RingSectors();

	unresolved = sgarraycnt = 0;
	badssecmessage = true;
	CLR_ssecarray();

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (!(ss->ssflags & SSF_NOTRACEN)) chkSSECassignments(ss);		// Main first pass nodes check
	}

	checkDisconnected();

	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		// v220-16 add test
		if (ss->linecount > 3 && !(ss->ssflags & SSF_BASESSECFIX))
		{
			sector_t *tf = NULL;
			seg_t	 *sg = &segs[ss->firstline];
			int		 k;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (!k)
				{
					if (!sg->backsector || ss->sector != sg->frontsector) break;
					if (sg->frontsector == sg->backsector) break;
					if (sg->frontsector->floorheight != sg->backsector->floorheight) break;
					if (sg->frontsector->ceilingheight != sg->backsector->ceilingheight) break;
				}
				else
				{
					if (ss->sector == sg->frontsector) { tf = NULL; break; }
					if (!sg->backsector && ss->sector == sg->frontsector) { tf = NULL; break; }
					if (!tf) tf = sg->frontsector; else if (tf != sg->frontsector) { tf = NULL; break; }
				}
			}

			if (tf && (tf->specFlags & 0x400 || (tf->specFlags & 0xC) != 0xC))	// v220-33
			{
				ChangeSSECsec(ss, tf, 31017);
				continue;
			}
		}

		// v220-16 add test
		if (ss->linecount > 3)
		{
			sector_t *tf = NULL, *tb = NULL;
			seg_t	 *sg = &segs[ss->firstline];
			int		 c, k;

			for (c = k = 0; k < ss->linecount; k++, sg++)
			{
				if (!sg->backsector) { c = 0; break; }
				if (sg->frontsector == sg->backsector)
				{
					if (k) { c = 0; break; }
					if (sg->frontsector != ss->sector) { c = 0; break; };
					c++; continue;
				}
				if (!k) { c = 0; break; }	// first seg must be same sided
				if (sg->frontsector == ss->sector) { c = 0; break; }
				if (!tf) tf = sg->frontsector; else if (tf != sg->frontsector) { c = 0; break; }
				if (!tb) tb = sg->backsector ; else if (tb != sg->backsector ) { c = 0; break; }
				if (tf->floorheight == tb->floorheight) { c = 0; break; }
			}

			if (c && tf)
			{
				int flags = ss->ssflags;

				if (flags & SSF_BASESSECFIX)
					ss->ssflags &= ~SSF_BASESSECFIX;

				ChangeSSECsec(ss, tf, 31018);

				if (flags & SSF_BASESSECFIX)
					ss->ssflags |= SSF_BASESSECFIX;

				continue;
			}
		}

		checkVirtual(ss);
		if (checkLogical(ss)) continue;
		if (CheckAssignment0(ss)){ ma_inf.dbcount[1]++; continue;}
		if (CheckAssignment1(ss)){ ma_inf.dbcount[2]++; continue;}
		if (CheckAssignment2(ss)){ ma_inf.dbcount[3]++; continue;}
		if (CheckAssignment3(ss)){ ma_inf.dbcount[4]++; continue;}
		if (CheckAssignment4(ss)){ ma_inf.dbcount[5]++; continue;}
	}
	
	assignFromSSLines();
}
