//-----------------------------------------------------------------------------
//
// Revision 2.1.0 Copyright (C) 2005 Jack Vermeulen
// Revision 2.2.0 Copyright (C) 2006 - 2012 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_verify.c
//
// Original idea by Jack Vermeulen (2005)
// Since then has been debugged, successively revised and extended by gmj
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  Modified to deal with conflicting lump types with the same name
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
//
// DESCRIPTION: inspects lump headers to determine their type
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <direct.h>

#include "r3d_compat.h"

#include "dd_types.h"
#include "dd_main.h"
#include "dd_share.h"
#include "dd_wad.h"
#include "r3d_defs.h"
#include "de_audio.h"
#include "de_graphics.h"
#include "con_console.h"
#include "s_main.h"
#include "m_misc.h"

#include "r3d_mus2mid.h"

// MACROS ------------------------------------------------------------------

// use half max patch width/height
#define MAXPATCHX  2048
#define MAXPATCHY  2048
#define MAXPATCHX1 2047
#define MAXPATCHY1 2047

#define SZHDRCACHE	128

// TYPES -------------------------------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

char *pngID = "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A";	// x89, "P","N","G" 0D 0A 1A 0A

// CODE --------------------------------------------------------------------

static RDBOOL VerifyYoffsets(char *RawImageData, int RawSize)
{
    patch_t* PicHeader = (patch_t*) RawImageData;

    int	CurrentX;
    int	offset;
    int	*Yoffsets = (int *)&RawImageData[0] + 2; // address 1st Y offset start =+8

    // quick test

    if (RawSize     < sizeof(PicHeader-28)
    ||  Yoffsets[0] < sizeof(PicHeader-28)
    ||  Yoffsets[0] >= RawSize )
    {
        return false;
    }

    // full test

    for (CurrentX = 0; CurrentX < PicHeader->width ; CurrentX++)
    {
        offset = Yoffsets[CurrentX];           

        if (offset >= RawSize
        ||  offset <  sizeof(PicHeader-28))
		{
			return false;
		}
    }

    return true;
}

static int graphicsINF(lumpinfo_t *linf, int w, int h, int *width, int *height, int type)
{
	if (width ) *width  = w;
	if (height) *height = h;

	linf->width  = w; linf->height = h;

	return (linf->lumptype = type);
}

// v220-13
static const char *mapchunks[] = {
	"THINGS",
	"LINEDEFS",
	"SIDEDEFS",
	"VERTEXES",
	"SEGS",
	"SSECTORS",
	"NODES",
	"SECTORS",
	"REJECT",
	"BLOCKMAP",
	"GL_VERT",
	"GL_SEGS",
	"GL_SSECT",
	"GL_NODES",
	NULL
};

// v220-13 known text files
static const char *txtfiles[] = {
	"ADLEVINF",
	"ANIMDEFS",
	"DD_DEFNS",
	"DEHACKED",
	"MAPINFO",
	"SCP_DEFN",
	"SCP_ZOFF",
	"SNDINFO",
	NULL
};

// v220-13
static RDBOOL checkTXT(byte *ptr, int size)
{
	int i;

	for (i = 0; i < size; i++)
	{
		if (ptr[i] > 0x7f) return false;

		if (ptr[i] < ' ')			// Space
		{
			if (ptr[i] == '\n') continue;	// Newline
			if (ptr[i] == '\r') continue;	// Carriage return
			if (ptr[i] == '\t') continue;	// Tab

			return false;
		}
	}

	return true;
}

// DOOM style graphics max 2048 x 2048 suffices
// NOTE: All types are just checked, headers are not actually set
//		 Only the first 128 bytes are read (or less if warranted) gmj
//       First revised b220-1
int DoomLumpIdentify(int lump, int *width, int *height, int type)
{
	wad_header_t	wadhdr;
	lumpinfo_t		*linf = NULL;
	int				filesize = 0, size;
	byte			hdrcache[SZHDRCACHE];
	int				i, w, h;
	void			*ptr = (void*)hdrcache;

	//----------------------------------

	if (width ) *width = 0;
	if (height) *height= 0;

	// linf size may have latterly been set to 0
	// so this must always be tested for (plus sets linf ptr).
	if (lump < 0 || lump >= numlumps)
		return LMP_UNDEF;	// could be a bad sprite lump or something

	linf = &lumpinfo[lump];

	if (linf->lumptype != LMP_UNDEF)	// already verified
	{
		if (width ) *width  = linf->width;
		if (height) *height = linf->height;
		return linf->lumptype;
	}

	if (type == DLI_ANY && linf->groupMarker >= LGT_FLATS && linf->groupMarker <= LGT_TX)
		type = DLI_GRAPHIC;

	// default
	linf->width  = 0;
	linf->height = 0;

	if (!linf->size)
	{
		linf->lumptype = LMP_MARKER;
		return LMP_MARKER;
	}
	
	filesize = linf->size;

	for (i = 0; mapchunks[i]; i++)
	{
		if (!stricmp(mapchunks[i], linf->name))
		{
			linf->lumptype = LMP_MAP;
			return LMP_MAP;
		}
	}

	for (i = 0; txtfiles[i]; i++)
	{
		if (!stricmp(txtfiles[i], linf->name))
		{
			linf->lumptype = LMP_TXT;
			return LMP_TXT;
		}
	}

	memset(hdrcache, 0, SZHDRCACHE);

	lseek(linf->handle, linf->position, SEEK_SET);

	// Only read the first 128 bytes if filesize > SZHDRCACHE
	size = (filesize > SZHDRCACHE) ? SZHDRCACHE : filesize;

	if ( read(linf->handle, ptr, size) != size)
	{
		Con_Error("DoomLumpIdentify: only partially read header for lump %i", lump);
	}

	if (M_ValidateID(linf->name))
	{
		linf->lumptype = LMP_MARKER;
		return LMP_MARKER;
	}

	if (linf->size < 12)
	{
		if (checkTXT((byte*)ptr, size))
		{
			linf->lumptype = LMP_TXT;
			return LMP_TXT;
		}

		return LMP_UNDEF;
	}

	///////////////////
	// WAV or MIDI
	///////////////////
	
	if (type == DLI_SOUND || type == DLI_ANY)
	{
		///////////////////
		// MUS (Doom type)
		//////////////////
		if (filesize > sizeof(MUShdr_t) + 3)
		{
			MUShdr_t musHdr;
			byte	 *mptr = (byte*)ptr;

			// v220-6 Check for mus lump formatted as a pwad
			//        e.g. diescum.wad. (Only known case - should I have bothered?)
			//
			// The assumption is that if "PWAD" are the first four bytes then the next
			// entries are formatted as a single directory entry. The following code
			// confirms whether the header bytes are a valid directory entry

			wadhdr = *(wad_header_t*) ptr;

			if (   !strncmp(wadhdr.type, "PWAD", 4)
				&& wadhdr.num_entries == 1
				&& wadhdr.dir_start < 128 - sizeof(MUShdr_t))
			{
				lump_header_t lhdr;
				char lname[8];

				lhdr = *(lump_header_t*) ((byte*)ptr + wadhdr.dir_start);

				memset(lname, 0, 8);
				strcpy(lname, linf->name);

				if (!strncmp(lhdr.name, lname, 8))
					mptr = (byte*)ptr + lhdr.start;
			}

			musHdr = *(MUShdr_t*) mptr;

			if (    DWORDSWAP(musHdr.ID) == MUS_HDR_ID
				&&  musHdr.muschannels 
				&&  musHdr.muschannels < MIDI_MAXNUMCHANS 
				&&  musHdr.scoreLen + musHdr.scoreStart <= filesize)
			{
				if (mptr != ptr)
				{
					linf->liflags |= LINF_LMPISWAD;
					Con_Message("WAD ERROR: mus lump %s is configured as a pwad.\n", linf->name);
				}

				return (linf->lumptype = LMP_MUS);
			}
		}

		if (linf->liflags & LINF_LMPISWAD) return LMP_IGNORE;

		if (filesize > sizeof(WAVEhdr_t)+1 )
		{
			WAVEhdr_t *WAVEHdr = (WAVEhdr_t*)ptr;

			if (   !strncmp(WAVEHdr->riff,"RIFF",4)
				&& !strncmp(WAVEHdr->wave,"WAVEfmt ",8) )
			{
				linf->lumptype = LMP_WAV;
				return LMP_WAV;  // WAV
			}
			
			if (!strncmp(WAVEHdr->riff,"MThd",4))
			{
				return (linf->lumptype = LMP_MIDI);
			}

			if (   !strncmp(WAVEHdr->riff,"RIFF",4)
				&& !strncmp(WAVEHdr->wave,"RMIDdata",8))
			{
				return (linf->lumptype = LMP_RMI);
			}
		}
		
		///////////////////
		// SND (Doom type)
		///////////////////
		
		if (filesize > sizeof(SOUNDHdr_t)+1)
		{
			SOUNDHdr_t	*SNDHdr	= (SOUNDHdr_t*)ptr;
			
			if (    SNDHdr->three == 0x0003		// DOOM sound lump
				&&  SNDHdr->smplrate >= 2000 
				&&  SNDHdr->smplrate <= 45000
				&&  SNDHdr->datasize >  0		// sound data size
				&&  SNDHdr->datasize <= filesize)
			{
				if (   SNDHdr->datasize < filesize - 8
					&& ((filesize - 8) & 0xFFFF) == SNDHdr->datasize)
				{
					pr_OFinsert("WARNING: sound %8s has a header entry of %i bytes.\n"
								"The true size of %i will be used\n",
								linf->name, SNDHdr->datasize, filesize - 8);

					linf->liflags |= LINF_TRUNCATED;
				}

				return (linf->lumptype = LMP_SND);
			}
		}
	}

	if (linf->liflags & LINF_LMPISWAD) return LMP_IGNORE;

	if (type == DLI_GRAPHIC || type == DLI_ANY)
	{
		// GRAPHICS

		// Note: TGA not used with lumps (would not be conducive to keeping wad sizes down)
		// It is supported elsewhere with models/skyboxes although they are normally
		// only ever used to allow external skybox sets to be imported - gmj

		//////////////////
		// PNG
		//////////////////
		
		if (filesize >= sizeof(png_t) + 64)	// also needs a palette, 64 should suffice
		{
			png_t	*PNGHdr = (png_t*)ptr;

			if (!memcmp(ptr, pngID, 8) && !memcmp(PNGHdr->chunkname, "IHDR", 4))
			{
				// image size is little endian so swap for Intel CPU's	
				return graphicsINF(linf,
					DWORDSWAP(PNGHdr->xsize),
					DWORDSWAP(PNGHdr->ysize),
					width, height, LMP_PNG);
			}
		}

		 ///////////////////
		 // v220-6 JPG
		 //////////////////

		if (JPG_Data(NULL, lump, &w, &h))
		{
			return graphicsINF(linf, w, h, width, height, LMP_JPG) ;
		}
		
		//////////////////
		// PCX (only because all the code was there)
		//////////////////
		
		if (   filesize > sizeof(pcx_t) + 768	// big enough
			&& PCX_HeaderGetInfo(ptr, &w, &h))
		{   
			return graphicsINF(linf, w, h, width, height, LMP_PCX) ;
		}

		 ///////////////////
		 // PATCH (Doom format)
		 //////////////////

		if (filesize > (sizeof(patch_t) - 28) )	// enough for a patch
		{
			patch_t *PicHdr	= (patch_t*)ptr;
			
			if (   PicHdr->width  > 0 && PicHdr->width  <= MAXPATCHX	//maximum X size should be xsize
				&& PicHdr->height > 0 && PicHdr->height <= MAXPATCHY	//maximum Y size should be ysize
				&& PicHdr->xoffset > -MAXPATCHX1				// x and y offset limits
				&& PicHdr->xoffset <  MAXPATCHX1				//
				&& PicHdr->yoffset > -MAXPATCHY1				//
				&& PicHdr->yoffset <  MAXPATCHY1				//
				&& PicHdr->columnofs[0] >= PicHdr->width * 4	// logical check 
				&& PicHdr->columnofs[0] <  filesize)
			{
				// We need more data for the next test, so just cache the lot
				if ( (PicHdr = (patch_t*)M_CacheLumpNum(lump, NULL, false)) )
				{
					if (VerifyYoffsets((char*)PicHdr, filesize))
					{
						graphicsINF(linf, PicHdr->width, PicHdr->height, width, height, LMP_PATCH);
						M_Free(PicHdr);
						return LMP_PATCH;
					}
					M_Free(PicHdr);
				}
			}
		}

		// v220-13 Could be a text file
		if (type == DLI_ANY)
		{
			byte *lptr = (byte*)M_CacheLumpNum(lump, NULL, false);
			int res = checkTXT(lptr, size);			;
			M_Free(lptr);
			
			if (res) return (linf->lumptype = LMP_TXT);
		}

		////////////////
		// FLAT safe default (even if it's not)
		////////////////
		
		// SKY and some others are not a consistent size

		if (filesize == 64000)
		{
			return graphicsINF(linf, 320, 200, width, height, LMP_FLAT320);
		}
		
		// FLATS that are within standard size ranges
		if (   filesize >= 0x001000		//   64 *   64 (4096)
			&& filesize <= 0x400000)	// 2048 * 2048
		{
			int sz = 0;

			if      (filesize == 0x400000) sz = 2048;
			else if (filesize == 0x100000) sz = 1024;
			else if (filesize == 0x040000) sz = 512;
			else if (filesize == 0x010000) sz = 256;
			else if (filesize == 0x004000) sz = 128;
			// Next valid size should be 0x1000 but thanks to the odd
			// badly constructed graphic this cannot be relied upon.
			else if (filesize <= 0x002000) sz = 64;

			if (sz)
				return graphicsINF(linf, sz, sz, width, height, LMP_FLAT);
		}

		// Hmmm.....
		return (linf->lumptype = LMP_UNDEF);
	}
	
	return (linf->lumptype = LMP_IGNORE);
}


