//-----------------------------------------------------------------------------
//
// Copyright (C) 2013 - 2020 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_wadanalyse.cpp	First release December 2013 (v220-22) gmj
//
// Revision 2.2.0 build 26 (v220-26) 05/Jul/2014 gmj
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
//   addtex and noclamp added
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
//   triggerDirect and secfixed added
// Revision 2.2.0 build 31 (v220-31) 14/Dec/2016 gmj
//   addlowtexsec, fixfloorheight, l3dref,
//   remtex, setceillightsec, setheightsec, fn_setssecind, sectortag added
//   setssec, addtex revised
//   Support for R3D_presetVars added
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//   preventGlow added to deal with map areas that are not glow
//               compatible through changing zbias criteria.
//   setssecsinlist, linemarksec added.
//	 support for R3D_NODEfix added
// Revision 2.2.0 build 34 (v220-34) 12/Nov/2018 gmj
//	setssecsinbox and setssecsinboxsec extended to allow a sector change to be made
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//	changessecsec, fracproject, fracprojectback, setspriteheight, fixsky,
//	forcesky, fixskynm,fixskykeep, forceskywall added
//
// DESCRIPTION: Designed for use where a function is particular to
//				just one map and where a simple set of external
//				instructions can be used to deal with it instead of
//				having to employ a more cumbersome and slow internal
//				function being applied to all maps.
//
//	PRELIMINARY VALUES
//
//	Note: fp numbers must contain a decimal point.
//	      E.g 2 will return 0, 2,0 should be used.
//		  PI, as a string, is not recognised.
//
//	[wadname]	Can only be set once as the script will not continue, after
//				being matched, after its following mapids have been processed.
//
//  mapid(s) { lumpvals keywords } must follow next where a mapid must match
//	the name used in a wad (e.g. MAP01 or E1M1 as appropriate).
//
//	lumpvals, used for verification, must precede any keywords for each mapid.
//	These are;
//
//	R3D_NODEfix	next four values are pre-node build values calculated from the
//				lump size divided by the relevant array size and must be;
//		numsectors, numlines, numsegs, numsides
//
//	R3D_presetVars next five values act as verification and must be
//		numsectors, numsubsectors, numlines, numsegs, numsides
//		and are those set before any other changes are made
//.
//  R3D_WADanalyse
//		numsectors, numsubsectors, numlines, numsegs, numsides
//		and are those set after other changes are made.
//
//	The values needed are printed to Risen3D.out if -devmap is used.
//
// KEYWORDS
//
// ****** The following can only be called from R3D_NODEfix ******
//
//	vertex <vx vy>
//		Change the numbered vertex's values
//
//	line <num> followed by side1 and side2 defs
//		The line num whose sides need to be changed
//	side1 <SECN> <lowtex name> <midtex name> <toptex name>
//		special name cases are NONE (where the name "-" will be used and NOCHANGE
//		side1's sidenum will remain unchanged
//
//	side2 <SAME, SAMEX or NEW> <SECN> <lowtex name> <midtex name> <toptex name>
//		Where a line has no sidedef one can be created using NEW.
//		Using SAMEX, which is treated as SAME, clears a line's twosided flag
//		all other parameters must be defined, however, even if these are unchanged..
//
//	A line def must immediately be followed by a side1 and then side2 def
//	existing sidedefs can neither be deleted nor replaced.
//
//	removemobj <mapnum>
//			allows a mobj (currently just one) to be inhibited from loading
//	respawnmobj <mapnum> <line triggers - max of three - if not needing all use -1>
//			if the specified line is triggerd than mobj will be respawned
//
// ****** The following two can be called from either R3D_presetVars or R3D_WADanalyse ******
//
//  changesecheight <secn> <floor 0 celing 1> <height>
//			change the sector's floor or ceiling height to height where height is in mapunits
//			Note: can also be called from R3D_WADanalyse
//	changesec <secn> <floortex> <ceilingtex>
//			change the floor and ceiling textures for sector secn
//			both textures must be sent even if one does not need changing
//
// ************************************* For use with R3D_presetVars;
//
//	SH		<secnum> <action>
//			where action is a keyword which determines what is to be done when
//			the player enters the sector defined by secnum;
//			Currently two action keyords are supported;
//			1)	SOUNDONCE <lumpname>
//				A wav sound is played once on entering.
//				It can be played again on re-entering after exiting the sector
//			2)	MUSICSTOP
//				Any music playing is stopped.
//
//	lineflags <lnum> <action>
//			change line flags;
//			action  = 0 clear upper and lower unpeg
//					= 1 set lower unpeg
//					= 2 set upper unpeg
//					= 3 set lower and upper unpeg
//					= 4 set blocking
//
//	fixsky	<skylump name> <rotation fp -pi to pi> [horizon offset fp -0.4 to 0.4]
//			Used for a map with 271/272 skychange specials where this causes problems
//			with Risen3D. This will fix the sky to one texture. Rotation and horizon
//			offset are not mandatory if not needed (the default is 0) but if horizon
//			is wanted rotation must be set even if only to 0.
//			NOTE: only one fixsky per map is allowed.
//	fixskynm as fixsky but used where a texture has an unwanted mirror duplicate of itself.	
//	fixskykeep <skylump name> <skylump name> <rotation fp -pi to pi> [horizon offset fp -0.4 to 0.4]
//			As fixsky but the second skylump name will not be replaced.
//			NOTE: fixskykeep must follow a fixsky def. Up to 16 are allowed.
//	linetag <lnum> <special> <tag>
//			used to change a line's special and/or tag
//	linex <linenum> <side> <integer value>
//			where side must be 0 or 1. The relevant sidenum is derived from the line and side
//			and its textureoffset set to the value sent shifted by FRACBITS.
//	liney <linenum> <side> <integer value>
//			where side must be 0 or 1. The relevant sidenum is derived from the line and side
//			and its rowoffset set to the value sent shifted by FRACBITS.
//	newwalltex <linenum> <side> <pos> <texname>
//			where side must be 0 or 1. The relevant sidenum is derived from the line and side
//			and the texname sent stored to pos where 0 = bottom,  1 = top  and 2 = midtex.
//			To clear an existing texture 0 or - can be used for the texname.
//	seclowoffset { list of space separated sector numbers }
//			Add -1 to the rowoffset for the lowtex for all lines belonging to sectors in the
//			list. Numbers must be contained inside curly braces.
//	sectorlight <snum> <val>
//			Change the sector's light level to val in range 0 to 255
//	sectortag <snum> <tag> OR <snumstart to snumend> <tag>
//			used to change a sector's tag OR the tags, where the tags are the same, for a sequence
//			where snumstart is the lowest and snumend the highest. All between and including from
//			the start to the end will have their sector tags changed to tag,.
//	sectorspecial <snum> <special>
//			used to change a sector's special
//
// ************************************* For use with R3D_WADanalyse;
//
//	addtex	<line num><side 0,1><pos bottom = 0, top = 1, midtex = 2><texname>
//			Used where missing textures cause HoMs missed by authors
//			through not being obvious using software renderers.
//	addlowtexsec  <secn><pos bottom = 0, top = 1, midtex = 2><texname>
//			Finds all lines in secn with missing lo tex and applies texname to any found,
//	remtex	<line num><side 0,1><pos bottom = 0, top = 1, midtex = 2>
//			Used to remove textures.
//	basesec <newsecn flags>
//			sets ssec floorsec and/or ceilsec to newsecn in a ssec group
//			set with a preceding largestloop command
//			sets floorsec if flags & 1, ceilsec if flags & 2
//	changessec <newsecn>
//			changes the sector for all ssecs set in a group to newsecn
//	changessecsec <ssec newsecn>
//			changes the ssec's sec to newsecn
//	changesseccsec <sec newsecn>
//			changes the ceilsec for all ssecs set in a sec group to newsecn
//	checksec <secn ch fh cpic fpic>
//			for validation: checks floorheight, ceilingheight, floorpic,
//			ceilingpic against values sent. Using this is not mandatory.
//	csec <newsecn refsec>
//			sets csec for all ssecs set in a group to newsecn if
//			refsec = -1 else only for those that match refsec.
//  customloop <secn> <number of lines in list> <line list>
//          set all ssecs inside a loop formed using the list.
//          secn should be the sector num of the sector facing inwards.
//	exclude <secn>
//			removes all ssecs referencing secn from a ssec group
//	fixfloorheight <secn> <height in mapunits>
//	fixsseccsec
//			Set the ssflag for all ssecs currently in the hitssec list to SSF_CSECFIXED
//	fn_setsamelinessec <linenum secn newsecn flags>
//			Finds ssecs bordering line sent in linenum. Checks line has same
//			sector both sides and is the same as that sent in secn. If any
//			ssecs' sector is equal to secn then it has its floorsec and/or
//			ceilsec set to newsecn. 
//			Sets floorsec if flags & 1, ceilsec if flags & 2
//	fn_setssec <secn newsecn flags>
//			Creates a ssec group from all ssecs belonging to lines in secn. If flag & 16 rejects any
//			ssec whose sector is not set to secn.
//			Depending on flags sets requested ssec parm to newsecn as follows:
//			sets floorsec if flags & 1, ceilsec if flags & 2, lfsec if flags & 4, lcsec if flags & 8
//	fn_setssecind <secn floorsec ceilsec lfsec lcsec>
//			creates a ssec group from all ssecs belonging to secn
//			Sets floorsec  ceilsec  lfsec lcsec to values sent. -1 should be used where no setting needed.
//	forcesky <secn>
//			sets flag for all ssecs set in a group to forcesky if
//			the ssec's sector matches secn..
//			The flag is then used to tell the renderer to draw a sky ceiling
//			flat if the ssec's ceiling sec height (which must have been
//			preset using fn_setssec) is >= to the ssec's sector ceiling
//			height.
//	forceskywall {list of lines]
//			Normally the analysis code will not draw sky walls with one's lines that form a ring.
//			Using this will override this forcing a sky wall to be drawn. Note that this only
//			affects one sided lines.
//	fracproject <line number>
//			Projects a line forwards from its current position when rendered.
//			Differs from normal projection in that it is not dependent on the user's look direction.
//			Used where a frig has had to be used in a map to stop a software renderer causing visible 
//			texture ceiling breakthrough where offset above its sector ceiling. This is normally not a
//			problem with OGL rendering unless a masked texture is less than 2 mapunits away from a high
//			texture where breakthrough is now visible through the high texture. In esence the best
//			solution would be to create a texture of the right size (i.e. with its obscured top row
//			removed) but the issue is vanishingly rare and this would be more complicated.
//	fracprojectback <line number>
//			As fracproject but projects backwards.
//	fsec <newsecn refsec>
//			sets fsec for all ssecs set in a group to newsecn if
//			refsec = -1 else only for those that match refsec.
//	l3dref <lnum val height>
//			Change 3d height for line lnum for val 0 - offset 1 - low  2 - mid
//	largestloop <secn loopnum>
//          finds the largest loop area with secn.
//          the number of loops that are expected to be found is sent in
//			loopnum (used for verification)
//			all ssecs inside the largest loop are placed in a ssec group
//	lcsec <newsecn refsec>
//			sets lcsec for all ssecs set in a group to newsecn if
//			refsec = -1 else only for those that match refsec.
//	lfsec <newsecn refsec>
//			sets lfsec for all ssecs set in a group to newsecn if
//			refsec = -1 else only for those that match refsec.
//	linemark <linenum R3D_type>
//			sets lines[linenum] to R3D_type and marks it for no further
//			analysis
//	linemarksec <linenum> <R3D_type>
//			sets lines in sec[linenum] to R3D_type and marks it for no further
//			analysis. Note: this ignores any prefixed flags that would otherwise
//			prevent th line type being changed (such as in linemark).
//  lineoffset <linenum side xoff yoff>
//          sets a line's texture offsets
//	marklines
//			marks all lines found in a ssec group to prevent further
//			analysis also sets line types
//	marksectors
//			marks all sectors found in a ssec group to prevent further
//			analysis
//	movemobj <mapnum x y>
//			where mapnum is the Tvalue and x and yo are the new coords in mapunits.
//			E.g.movemobj 340 56 -76 where in the map the mobj is shown as T340
//	noclamp <texture name>
//			used with a masked texture that should not be clamped
//			only possible if the texture has not been used as a midtexture
//	nomisslosec <secnum>
//			sets instruction not to render a missing hi tex
//	nomisshisec <secnum>
//			sets instruction not to render a missing lo tex
//	preventGlow { list of space separated sector numbers }
//			Used where a map structure is incompatible with applyying glow
//	radiusfix <doomednum>
//			Maps may have sprites with larger radii. These can be occluded
//			through not having been linked into a sector at the edge that
//			is the only one visible in the frustrum. This sets a flag used
//			by P_LinkThing to use the things true radius. In addition if the
//			sector it is spawned in is not a mov type then its position is
//			flagged fixed to prevent it from falling off of a dropoff.
//	secdone <secn>
//			marks secn and marks all lines found in a secn to prevent
//			further analysis
//			also sets line types
//  secfixed <secn>
//			Sets all lines in the sector as fixed to type R3D_36
//  setsec <secin secout flags>
//          if flags & 1 copy secin's floor values to floorsec  in all ssecs' referencing secout
//          if flags & 2 copy secin's ceiling values to ceilsec in all ssecs' referencing secout
//  setheightsec  <secn hsecn>
//			Set sectors[secn].heightsec to hsecn
//	setceillightsec <secn lsec>
//			Set sectors[secn].ceillightsec to lsec
//  setsecfloor <secin secout>
//          copy secin's floor values to lfsec in all ssecs' referencing secout
//	setsky	<secn texname rowoff type>
//			set type 0 for 271, 1 for 272
//	setssec <ssecnum secn flags>
//			set a single ssec floor/ceiling to secn.
//			sets floorsec if flags & 1, ceilsec if flags & 2, sector if fl & 4
//	setssecinlist <secn flags {space separated list of subsecs}>
//			set ssecs values in list according to flags to secn.
//			sets floorsec if flags & 1, lfsec if & 2, ceilsec if & 4, lcsec if & 8, sector if fl & 16,
//			if fl & 32 sets ors SSF_BASESSECFIX with ss->ssflags
//	setssecsinbox <x y length height losecfloor secfloor losecceil secceil (ssec sec but only if other 4 values are all set to -1)>
//			marks all subsecs in a box with limits determined x y width
//			and height where x,y is the bottom left hand corner. All
//			values must be in mapunits NOT fixed_t. width and height
//			must be positive. Range extends to +-32000.
//			If a substitute floorsec is needed then set secfloor to the
//			sec number wanted else set it to -1 (it must be sent)
//			If a substitute ceilsec is needed then set secceil to the
//			sec number wanted else set it to -1 (it must be sent)
//			If all values are set to -1 then a 5th can be added which will change the ssecs sector reference.
//	setssecsinboxsec <secn losecfloor secfloor losecceil secceil (ssec sec but only if other 4 values are all set to -1)>
//			As setssecsinbox but box limits are taken directly from the limits of secn
//	setnoocclude <secn>
//	setunlit <thing mapnum>
//	setspriteheight <doomednum height>
//			Set the sprite's height
//	ssvg <ssnum vnum>
//			On rare occasions a nodes builder will fail to provide a line referencing
//			a subsector. This has to be done manually.
//  triggerDirect <linen>
//			marks a line with a special that can never be crossed but which
//			needs to be triggered. If any defined are encountered then they
//			are actioned immediately. Cases are rare but AV map23 is a prime
//			example. Only lines with a walkover special are supported and
//			the number that can be added is limited to MAXDTNUM.
//	noMovUnset no parameter. If present inhibts the function LG_AnalyseMovTexUnset()
//
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include "r3d_MAINspec.h"
#include "m_misc.h"
#include "dd_wad.h"
#include "..\bsp\rd_shared.h"
#include "doom\p_local.h"
#include "dedfile.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

int W_CheckLumpNumForName(char *name);
int Add_NewTexture(char *name, int lumpnum, int *type);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern char		new_mapid[8];
extern byte		walkCheck[];
extern float	setskycolour;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

analysesides_t *analysesides;
int			maxannumsides;

int			hackflat;

nodefix_t	*nodefix;
int			numnodefix;
int			currentnodefix;

vertexfix_t	*vertexfix;
int			numvertexfix;
int			currentvertexfix;

wadani_t wadani;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static char		 *nextstr;

// CODE --------------------------------------------------------------------


// ===============================================================

static void freeSortbox(sortbox_t *sbx, int bc)
{
	int		i;

	for (i = 0; i < bc; i++) free(sbx[i].llist);
}

static int getSsecsInLoop(sector_t *sc, int ringnum)
{
	sortbox_t	sbx[MAXSORTBOXES];
	sortbox_t	*sbp;
	__int64		area;
	int			index, lnum;
	int			bc, rc;
	int			c, i, n;

	if (!sc) SR_SCPErr("sector unset with largestloop");

	memset(sbx, 0, sizeof(sortbox_t) * MAXSORTBOXES);

	CLR_hitline();

	for (c = i = 0; i < sc->linecount; i++)
	{
		if (sc->lines[i]->frontsector != sc->lines[i]->backsector)
		{
			n = sc->lines[i] - lines; hitline[n] = 1; c++;
		}
	}

	if (c < 3) return 0;

	bc = 0;

	while(1)
	{
		lnum = -1;

		for (i = 0; i < sc->linecount; i++)
		{
			n = sc->lines[i] - lines;
			if (hitline[n] == 1) { lnum = n; break; }
		}

		if (lnum < 0) break;

		if (bc == MAXSORTBOXES)	Con_Error("program error: boxSort max number of %i exceeded", MAXSORTBOXES);

		sbp = &sbx[bc];

		if (subfn_CheckForRing(hitline, sc, lnum, &rc, sbp->box, NULL, NULL) && rc > 2)
		{
			bc++;
			
			sbp->area = ((sbp->box[BOXRIGHT] - sbp->box[BOXLEFT]) >> FRACBITS)
					  * ((sbp->box[BOXTOP  ] - sbp->box[BOXBASE]) >> FRACBITS);
			sbp->listcnt = rc;
			sbp->llist = (int*)malloc(rc * sizeof(int));

			for (c = i = 0; i < sc->linecount; i++)
			{
				n = sc->lines[i] - lines;
				if (hitline[n] == 4) { hitline[n] = 0; sbp->llist[c++] = n; }
			}
		}
		else
		{
			hitline[lnum] = 0;
		}
	}

	if (bc != ringnum)
	{
		SR_SCPWarn("number of loops %i does not match number expected", bc);
		freeSortbox(sbx, bc);
		return 0;
	}

	for (index = 0, area = sbx[0].area, i = 1; i < bc; i++)
	{
		if (sbx[i].area > area) { area = sbx[i].area; index = i; }
	}

	CLR_hitline();

	sbp = &sbx[index];

	for (i = 0; i < sbp->listcnt; i++) hitline[sbp->llist[i]] = 4;

	if (LS_MarkSubsecsInAreaGeneral(sc, hitline, 0, NULL, sbp->box, -1, MSIA_NONE))
	{
		freeSortbox(sbx, bc);
		return hitssecptr;
	}

	SR_SCPWarn("loop failed with lnum %i", sbp->llist[0]);
	freeSortbox(sbx, bc);
	return 0;
}

static RDBOOL excludeSsecsInLoop(sector_t *exc)
{
	int c, i;
	int *ssecmem;

	if (!hitssecptr) return false;

	ssecmem = (int*)malloc(hitssecptr * sizeof(int));
	memcpy(ssecmem, hitssec, hitssecptr * sizeof(int));

	for (c = i = 0; i < hitssecptr; i++)
	{
		if (subsectors[hitssec[i]].sector != exc) ssecmem[c++] = hitssec[i];
	}

	memcpy(hitssec, ssecmem, c * sizeof(int));
	hitssecptr = c;
	free(ssecmem);
	return (c > 0);
}

static int wa_mtxline(line_t *ld)
{
	if (ld->backsector && ld->frontsector != ld->backsector)
	{
		if (   !(origTEX[ld - lines] & 0x11)
			&& ld->backsector->floorheight != ld->frontsector->floorheight
			&& !(ld->backsector->floorpic == skyflatnum
			&&   ld->frontsector->floorpic == skyflatnum))
		{
			return 1;
		}

		if (   !(origTEX[ld - lines] & 0x44)
			&& ld->backsector->ceilingheight != ld->frontsector->ceilingheight
			&& !(ld->backsector->ceilingpic == skyflatnum
			&&   ld->frontsector->ceilingpic == skyflatnum))
		{
			return 1;
		}
	}

	return 0;
}

static void getSTR(char *nextstr)
{
	if (!(nextstr = SR_NextSTR(false)) ) SR_SCPErr("End of file");
}

static int getVal(int *err)
{
	int		n, scperror;

	if (!(nextstr = SR_NextSTR(false)) ) SR_SCPErr("End of file");
	n = SR_NumForString(&scperror);
	if (scperror) { {if (err) *err = 1; } return 0; }
	if (err) *err = 0;
	return n;
}

static float getFloat(int *err)
{
	int		scperror;
	float	n;

	if (!(nextstr = SR_NextSTR(false)) ) SR_SCPErr("End of file");
	n = SR_FloatForString(&scperror);
	if (scperror) { {if (err) *err = 1; } return 0; }
	if (err) *err = 0;
	return n;
}

static int getLineIdx(int lcount)
{
	int		n, err;

	n = getVal(&err);
	if (err || n < 0 || n >= lcount) SR_SCPErr("Invalid line num");
	return n;
}

static int getSidenumFromLinenum(int lnum, int *side)
{
	int		k, n, err;

	if (lnum < 0 || lnum >= numlines) SR_SCPErr("Program error: Invalid line num %i sent", lnum);
	k = getVal(&err);	// Side required
	if (err || k < 0 || k > 1) SR_SCPErr("Side is invalid - must be 0 or 1");
	if ((n = lines[lnum].sidenum[k]) < 0 || n > numsides) SR_SCPErr("L%i side %i is invalid", lnum, k);
	if (side) *side = k;
	return n;
}

static int getSecIdx(RDBOOL allowminusone)
{
	int		n, err;

	n = getVal(&err);
	if (err) SR_SCPErr("Invalid sector num");
	if (n == -1 && allowminusone) return -1;
	if (n < 0 || n >= numsectors) SR_SCPErr("Invalid sector num");
	return n;
}

static int getSsecIdx(void)
{
	int		n, err;

	n = getVal(&err);
	if (err) SR_SCPErr("Invalid subsector num");
	if (n < 0 || n >= numsubsectors) SR_SCPErr("Invalid subsector num");
	return n;
}

static int getMU(RDBOOL positive)
{
	int		n, err;
	int		neg_limit = positive ? 0 : -32000;

	n = getVal(&err);
	if (err) SR_SCPErr("Invalid num");
	if (n < neg_limit || n > 32000) SR_SCPErr("Invalid num");
	return n;
}

static int getNextNumVal()
{
	int		n, err;

	n = getVal(&err);
	if (err) SR_SCPErr("Invalid num");
	return n;
}

static void setLineParms(line_t *ld, int num)
{
	byte txn;

	if (lineflags[ld - lines]) return;

	if (!(txn = origTEX[ld - lines]))
		Settype(ld, R3D_3F, num);
	else if (ld->frontsector == ld->backsector)
		Settype(ld, R3D_03, num);
	else if (ld->backsector
		&& txn & 0x22
		&& !(txn & 0x55)
		&& ld->frontsector->floorheight == ld->backsector->floorheight)
		Settype(ld, R3D_03, num);
	else if (wa_mtxline(ld))
		Settype(ld, R3D_00, num);
	else
		Settype(ld, R3D_36, num);

	SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, num);
}

static void setLineFixed(sector_t *sc)
{
	int i;

	for (i = 0; i < sc->linecount; i++)
	{
		line_t *ld = sc->lines[i];

		if (ld->ml_flags & ML_TWOSIDED)
		{
			ld->specType = R3D_36;
			Clr_allmtflags(ld, 52001);
		}

		SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 52001);
	}
}

static void subfn_setSsecsInBox(float *box)
{
	subsector_t	*ss;
	line_t		*ld;
	seg_t		*sg;
	int			a, c, i, k, n;

	hitssecptr = 0;

	// keep it simple - time's not an issue here
	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		if (   ss->midpoint.x > box[BOXLEFT ]
			&& ss->midpoint.x < box[BOXRIGHT]
			&& ss->midpoint.y > box[BOXBASE ]
			&& ss->midpoint.y < box[BOXTOP  ])
		{
			hitssec[hitssecptr++] = i;
		}
	}

	validcount++;

	for (c = i = 0; i < hitssecptr; i++)
	{
		ss = &subsectors[hitssec[i]];
		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if ( !(ld = sg->linedef) ) continue;
			if (ld->validcount == validcount) continue;
			ld->validcount = validcount;
			linearray[c++] = ld - lines;
		}
	}

	a = true;
	n = getSecIdx(true);

	if (n >= 0)
	{
		for (a = false, i = 0; i < hitssecptr; i++)
			Setr3dsslfsec(&subsectors[hitssec[i]], n, 52002);
	}

	n = getSecIdx(true);

	if (n >= 0)
	{
		for (a = false, i = 0; i < hitssecptr; i++)
			Setr3dssfs(&subsectors[hitssec[i]], n, 52003);
	}

	n = getSecIdx(true);

	if (n >= 0)
	{
		for (a = false, i = 0; i < hitssecptr; i++)
			Setr3dsslcsec(&subsectors[hitssec[i]], n, 52003);
	}

	n = getSecIdx(true);

	if (n >= 0)
	{
		for (a = false, i = 0; i < hitssecptr; i++)
			Setr3dsscs(&subsectors[hitssec[i]], n, 52004);
	}

	for (i = 0; i < c; i++)
		setLineParms(&lines[linearray[i]], 52005);

	if (a == true && (n = getSecIdx(true)) >= 0)
	{
		for (i = 0; i < hitssecptr; i++)
		{
			subsectors[hitssec[i]].ssflags &= ~SSF_BASESSECFIX;
			subsectors[hitssec[i]].sector = sectors+n;
			subsectors[hitssec[i]].ssflags |= SSF_BASESSECFIX;
		}
	}
}

static RDBOOL verifySSEClist()
{
	if (hitssecptr > numsubsectors) { SR_SCPWarn("Invalid loop value"); return false; }

	for (int i = 0; i < hitssecptr; i++)
	{
		if (hitssec[i] < 0 || hitssec[i] >= numsubsectors) { SR_SCPWarn("Invalid loop value");return false; }
	}
	return true;
}

static void setSSECvals(int type, RDBOOL dupval1)
{
	int		i, err = 0, refsec;
	int		n = getSecIdx(false);

	if (dupval1)
		refsec = n;
	else
		refsec = getVal(&err);

	if (err || (refsec != -1 && (refsec < 0 || refsec >= numsectors)))
		SR_SCPWarn("Invalid refsec\n");
	else
	{
		for (i = 0; i < hitssecptr; i++)
		{
			subsector_t *ss = &subsectors[hitssec[i]];

			if (refsec == -1 || ss->sector - sectors == refsec)
			{
				if (type == 1) Setr3dsscs(ss, n, 52006);
				else
				if (type == 2) Setr3dssfs(ss, n, 52006);
				else
				if (type == 3) Setr3dsslcsec(ss, n, 52006);
				else
				if (type == 4) Setr3dsslfsec(ss, n, 52006);
				else
				if (type == 5)
				{
					if (ss->ceilsec >= 0) ss->ssflags |= SSF_FORCESKY;
				}
			}
		}
	}
}

int setSsecList(sector_t *sc, int sconly)
{
	subsector_t	*ss;
	ssline_t	*sl;
	int			i, k;

	hitssecptr = 0;
	ss_validcount++;

	for (i = 0; i < sc->linecount; i++)
	{
		sl = &sslines[ (sc->lines[i] - lines) ];

		for (k = 0; k < sl->num; k++)
		{
			ss = &subsectors[sl->list[k]];

			if (ss->ssvalidcount == ss_validcount) continue;
			if (sconly && ss->sector != sc) continue;

			ss->ssvalidcount = ss_validcount;

			hitssec[hitssecptr++] = sl->list[k];
		}
	}

	return hitssecptr;
}

static RDBOOL subfn_parseCommands(int from)
{
	subsector_t	*ss;
	ssline_t	*sl;
	sector_t	*ch, *sc = NULL;
	line_t		*ld;
	side_t		*sd;
	seg_t		*sg;
	fixed_t		ibox[4];
	float		fbox[4];
	int			numssecs;
	int			err, sn, fl;
	int			c, i, k, n;

	hitssecptr = 0;

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!stricmp(nextstr, "}")) return true;

		if (from == 2)
		{
			uint	len;

// calls that can only be made from R3D_NODEfix
			if (!stricmp(nextstr, "vertex"))
			{
				int vx, vy;

				n = getVal(&err);
				if (n < 0 || n >= NFnumvertexes) SR_SCPErr("vertex %i is out of range", n);
				vx = getVal(&err); 
				if (vx < -32768 || vx > 32767) SR_SCPErr("vx %i is out of range",vx);
				vy = getVal(&err);
				if (vy < -32768 || vy > 32767) SR_SCPErr("vy %i is out of range",vy);

				numvertexfix++;
				vertexfix = (vertexfix_t*)W_ReallocCLR(vertexfix, sizeof(vertexfix_t)*(numvertexfix), LEVELID, 0);
				currentvertexfix++;

				vertexfix[currentvertexfix].vtxnum = n;
				vertexfix[currentvertexfix].vx = vx;
				vertexfix[currentvertexfix].vy = vy;
			}
			else
			if (!stricmp(nextstr, "line"))
			{
				numnodefix++;
				nodefix = (nodefix_t*)W_ReallocCLR(nodefix, sizeof(nodefix_t)*(numnodefix), LEVELID, 0);
				currentnodefix++;
				nodefix[currentnodefix].line = getLineIdx(NFnumlinedefs);		// linenum
				nodefix[currentnodefix].side1sec = -1;		// unset side1sec
				nodefix[currentnodefix].side2sec = -1;		// unset side2sec
			}
			else
			if (!stricmp(nextstr, "side1"))
			{
				if (currentnodefix < 0 || currentnodefix == numnodefix)
					SR_SCPErr("unset line %s", nextstr);
				n = getVal(&err);
				if (err) SR_SCPErr("Invalid sector num");
				if (n < 0 || n >= NFnumsectors) SR_SCPErr("Invalid sector num");
				nodefix[currentnodefix].side1sec = n;
				nextstr = SR_NextSTR(false);
				if (!nextstr) SR_SCPErr("End of file");
				len = strlen(nextstr); if (len > 8) len = 8;
				strncpy(nodefix[currentnodefix].sd1low, nextstr, len);
				nextstr = SR_NextSTR(false);
				if (!nextstr) SR_SCPErr("End of file");
				len = strlen(nextstr); if (len > 8) len = 8;
				strncpy(nodefix[currentnodefix].sd1mid, nextstr, len);
				nextstr = SR_NextSTR(false);
				if (!nextstr) SR_SCPErr("End of file");
				len = strlen(nextstr); if (len > 8) len = 8;
				strncpy(nodefix[currentnodefix].sd1hi, nextstr, len);
			}
			else
			if (!stricmp(nextstr, "side2"))
			{
				if (currentnodefix < 0 || currentnodefix == numnodefix)
					SR_SCPErr("unset line %s", nextstr);
				if (nodefix[currentnodefix].side1sec == -1)
					SR_SCPErr("unset side1 sector %s", nextstr);
				nextstr = SR_NextSTR(false);
				if (!nextstr) SR_SCPErr("End of file");
				if (!stricmp("SAME", nextstr))
				{
					n = getVal(&err);
					if (err) SR_SCPErr("Invalid sector num");
				}
				else
				if (!stricmp("SAMEX", nextstr))
				{
					n = getVal(&err);
					if (err) SR_SCPErr("Invalid sector num");
					nodefix[currentnodefix].ndflags |= ND_CLEAR2S;
				}
				else
				if (!stricmp("NEW", nextstr))
				{
					n = getVal(&err);
					if (err) SR_SCPErr("Invalid sector num");
					nodefix[currentnodefix].ndflags |= ND_NEWSIDE2;
				}
				else
					SR_SCPErr("unrecognised type %s", nextstr);

				if (n < 0 || n >= NFnumsectors) SR_SCPErr("Invalid sector num");
				nodefix[currentnodefix].side2sec = n;
				nextstr = SR_NextSTR(false);
				if (!nextstr) SR_SCPErr("End of file");
				len = strlen(nextstr); if (len > 8) len = 8;
				strncpy(nodefix[currentnodefix].sd2low, nextstr, len);
				nextstr = SR_NextSTR(false);
				if (!nextstr) SR_SCPErr("End of file");
				len = strlen(nextstr); if (len > 8) len = 8;
				strncpy(nodefix[currentnodefix].sd2mid, nextstr, len);
				nextstr = SR_NextSTR(false);
				if (!nextstr) SR_SCPErr("End of file");
				len = strlen(nextstr); if (len > 8) len = 8;
				strncpy(nodefix[currentnodefix].sd2hi, nextstr, len);
			}
			else
			{
				SR_SCPErr("unknown keyword %s", nextstr);
			}
		}
		else
// calls that can be made from R3D_presetVars or R3D_WADanalyse
		if ((from == 0 || from == 1) && !stricmp(nextstr, "changesecheight"))
		{
			sn = getSecIdx(false);
			k = getVal(NULL) != 0;	// floor 0 ceiling 1
			n = getVal(NULL) << FRACBITS;
			if (k)
				sectors[sn].ceilingheight = sectors[sn].smoothceil = n;
			else
				sectors[sn].floorheight = sectors[sn].smoothfloor = n;
		}
		else
		if ((from == 0 || from == 1) && !stricmp(nextstr, "changesec"))
		{
			sn = getSecIdx(false);
			if (!(nextstr = SR_NextSTR(false))) SR_SCPErr("End of file");
			if (stricmp(nextstr, "NONE"))
			{
				if ( (n = W_CheckLumpNumForName(nextstr)) < 0) SR_SCPErr("Invalid texture name \"%s\"", nextstr);
				sectors[sn].floorpic = n;

			}
			if (!(nextstr = SR_NextSTR(false))) SR_SCPErr("End of file");
			if (stricmp(nextstr, "NONE"))
			{
				if ( (n = W_CheckLumpNumForName(nextstr)) < 0) SR_SCPErr("Invalid texture name \"%s\"", nextstr);
				sectors[sn].ceilingpic = n;
			}
		}
		else
		if (!stricmp(nextstr, "exit"))
		{
			return false;
		}
		else
// calls that can only be made from R3D_presetVars
		if (from == 0)
		{
			if (!stricmp(nextstr, "fixsky") || !stricmp(nextstr, "fixskynm"))
			{
				extern float skyAngleGlobal;	// globally applied

				ded_mapinfo_t	*dmt = NULL;
				float			rotation;
				RDBOOL			typenm = !stricmp(nextstr, "fixskynm");
				RDBOOL			err;

				// -----------------------
				if (ma_inf.noskychange)
					SR_SCPErr("There must only be one fixsky or fixskynm def");

				if ( !(dmt = Def_GetMapInfo(new_mapid)) )
				{
					dmt = (ded_mapinfo_t*)DED_NewEntry( (void**) &defs.mapinfo,
							&defs.count.mapinfo, sizeof(ded_mapinfo_t));
				}

				if (dmt)
				{
					ded_skylayer_t	*sl = dmt->sky_layers;
					int				idx;

					strcpy(sl->texture, SR_NextSTR(true));

					ma_inf.noskychange = true;

					if (typenm)
					{
						for (i = 0; i < numtextures; i++)
						{
							idx = texturetranslation[i];

							if (!stricmp(sl->texture, texArray[idx]->name))
							{
								texArray[idx]->txf_flags |= TXF_DELMIRROR;
								break;
							}
						}
					}

					SR_NextSTR(true);
					rotation = SR_FloatForString(&err);
					if (err) { SR_KeepString(); continue; }	// must be next keyword

					skyAngleGlobal = rotation/PI;
					if (skyAngleGlobal >  1.0f) skyAngleGlobal =  1.0f;
					if (skyAngleGlobal < -1.0f) skyAngleGlobal = -1.0f;
					sl->skyAngleIncrement = skyAngleGlobal;

					SR_NextSTR(true);
					ma_inf.horizonOffset = SR_FloatForString(&err);
					if (err) { SR_KeepString(); continue; }	// must be next keyword

					if (FABS(ma_inf.horizonOffset) > 0.4f)
						SR_SCPErr("SKY OFFSET error: limits are not in  -0.4 to 0.4");

					strcpy(sl->flags, "slf_enable");
				}
				else if (ge_inf.devmap)
				{
					SR_SCPErr("FIXSKY error: %s not found", new_mapid);
				}
			}
			else
			if (!stricmp(nextstr, "fixskykeep"))
			{
				n = getVal(&err);
				if (err) SR_SCPErr("viewgroup number missing");

				if (ma_inf.skykeepcount < sizeof(ma_inf.skykeep)/(2 * sizeof(int)) )
				{
					if (!ma_inf.noskychange)
						SR_SCPErr("fixskykeep must be preceded by a fixsky def");
					ma_inf.skykeep[ma_inf.skykeepcount][0] = n;
					n = R_TextureNumForName(SR_NextSTR(true), true);
					if (n < 0) SR_SCPErr("viewgroup lump with name %s", nextstr);
					ma_inf.skykeep[ma_inf.skykeepcount++][1] = n;
				}
				else
				{
					Con_Error("ERROR: sizeof skykeep breached.\n\n"
						"Reduce number of fixskykeep entries for %s in R3D_presetVars\n", new_mapid);
				}
			}
			else
			if (!stricmp(nextstr, "SH"))	// v220-31 Sector Hooks
			{
				if (!SHarray) SHarray = (SH_t*)W_Calloc(numsectors * sizeof(SH_t), LEVELID, 0);

				sn = getSecIdx(false);		// sec num
				getSTR(nextstr);

				if (!stricmp(nextstr, "SOUNDONCE"))
				{
					getSTR(nextstr);

					if ( (k = R3D_AddNewSound(nextstr)) )
					{
						k |= HSF_NO_ATTENUATION;
						for (i = 1; i < SH_MAXSOUNDS; i++) { if (!SHcontainer[i] || SHcontainer[i] == k) break; }
						if (i == SH_MAXSOUNDS) i = 1;
						SHarray[sn].cp = i;
						SHcontainer[i] = k;
					}
					else
					{
						SR_SCPWarn("Invalid sound");
					}
				}
				else
				if (!stricmp(nextstr, "MUSICSTOP"))
				{
					SHarray[sn].fp |= SH_STOPMUSIC;
				}
			}
			else
			if (!stricmp(nextstr, "lineflags"))
			{
				n = getLineIdx(numlines);		// linenum
				ld = &lines[n];
				n = getVal(NULL);
				if (n & 1) ld->ml_flags |= ML_DONTPEGBOTTOM; else ld->ml_flags &= ~ML_DONTPEGBOTTOM;
				if (n & 2) ld->ml_flags |= ML_DONTPEGTOP;    else ld->ml_flags &= ~ML_DONTPEGTOP;
				if (n & 4) ld->ml_flags |= ML_BLOCKING;
			}
			else
			if (!stricmp(nextstr, "linetag"))
			{
				n = getLineIdx(numlines);		// linenum
				ld = &lines[n];
				n = getVal(NULL);
				ld->special = n;
				n = getVal(NULL);
				ld->tag = n;
				if (ld->special == 260)
				{
					if (ld->spflags & SPF_MIDTRANSLUCENT)	// specifically toggle mid translucent
						ld->spflags &= ~SPF_MIDTRANSLUCENT;
					else if (!ld->backsector)
						SR_SCPWarn("Translucency can only be applied to lines having a backsector");
					else
						ld->spflags |= SPF_MIDTRANSLUCENT;

					ld->special = 0;
				}
			}
			else
			if (!stricmp(nextstr, "linex"))
			{
				int lnum = getLineIdx(numlines);		// linenum

				n = getSidenumFromLinenum(lnum, NULL);
				c = getVal(NULL);				// x offset
				sides[n].textureoffset = c << FRACBITS;
			}
			else
			if (!stricmp(nextstr, "liney"))
			{
				int lnum = getLineIdx(numlines);		// linenum

				n = getSidenumFromLinenum(lnum, NULL);
				c = getVal(NULL);				// y offset
				sides[n].rowoffset = c << FRACBITS;
			}
			else
			if (!stricmp(nextstr, "newwalltex"))
			{
				int lnum = getLineIdx(numlines);				// linenum
				int snum = getSidenumFromLinenum(lnum, NULL);	// sidenum

				c = getVal(NULL);	// bottom 0, top 1, mid 2

				if (c < 0 || c > 2) SR_SCPErr("Invalid texture position num %i", c);

				if (!(nextstr = SR_NextSTR(false))) SR_SCPErr("End of file");

				if (!strcmp(nextstr, "0") || !strcmp(nextstr, "-"))
				{
					n = 0;
				}
				else
				if ( (n = R_CheckTextureNumForName(nextstr)) < 0)
				{
					int type;

					// v220-33 try the flats
					if ( (n = R_FlatNumForName(nextstr)) < 0 || (n = Add_NewTexture(nextstr, n, &type)) < 0)
						SR_SCPErr("Invalid texture or flat name \"%s\"", nextstr);
				}

				sd = &sides[snum];
				
				// v220-23 add texFlags changes
				if (c == 0) { sd->bottomtexture = n; if (n) sd->texFlags |= T_LOW; else sd->texFlags &= ~T_LOW; }
				if (c == 1) { sd->toptexture = n;    if (n) sd->texFlags |= T_HI ; else sd->texFlags &= ~T_HI ; }
				if (c == 2) { sd->midtexture = n;    if (n) sd->texFlags |= T_MID; else sd->texFlags &= ~T_MID; }
			}
			else
			if (!stricmp(nextstr, "seclowoffset"))
			{
				if (strcmp(SR_NextSTR(true), "{")) SR_SCPErr("missing {");

				while(strcmp( (nextstr = SR_NextSTR(true)), "}"))
				{
					n = SR_NumForString(&err);	// secnum
					if (err) n = -1;
					if (n < 0 || n >= numsectors) SR_SCPErr("Invalid sector num %i", n);
					sectors[n].specFlags2 |= SF2_NEGOFFSET;
				}
			}
			else
			if (!stricmp(nextstr, "sectorlight"))
			{
				n = getSecIdx(false);			// secnum
				k = getVal(NULL);				// light level
				if (k < 0) k = 0; else if (k > 255) k = 255;
				sectors[n].lightlevel = k;
			}
			else
			if (!stricmp(nextstr, "sectortag"))
			{
				n = getSecIdx(false);			// secnum
				if (!(nextstr = SR_NextSTR(false))) SR_SCPErr("End of file");

				if (!stricmp("TO", nextstr))
				{
					int e = getSecIdx(false);	// end sector
					int t = getVal(NULL);		// tag

					for (; n <= e; n++) sectors[n].tag = t;
				}
				else
				{
					int t = SR_NumForString(&err);	// tag
					if (err) t = -1;
					if (t < 0 || abs(t) > 65536) SR_SCPErr("Invalid tag %i", t);
					sectors[n].tag = t;
				}
			}
			else
			if (!stricmp(nextstr, "sectorspecial"))
			{
				n = getSecIdx(false);			// secnum
				sectors[n].special = getVal(NULL);	// special
			}
			else
			if (!stricmp(nextstr, "useR3Dskyoffsets"))
			{
				ma_inf.r3dskyoffsets = true;
			}
			else
			if (!stricmp(nextstr, "exit"))
			{
				return false;
			}
			else
			{
				SR_SCPErr("unknown keyword %s", nextstr);
			}
		}
// calls that can only be made from R3D_WADanalyse
		else if (from == 1)
		{
			if (!stricmp(nextstr, "noMovUnset"))
			{
				ma_inf.noMovUnset = true;
			}
			else if (!stricmp(nextstr, "addtex"))
			{
				int lnum = getLineIdx(numlines);		// linenum

				n = getSidenumFromLinenum(lnum, &k);
				sd = &sides[n];

				c = getVal(NULL);	// bottom 0, top 1, mid 2
				if (c < 0 || c > 2) SR_SCPErr("Invalid texture position num %i", c);

				if (!(nextstr = SR_NextSTR(false))) SR_SCPErr("End of file");
				if ( (n = R_CheckTextureNumForName(nextstr)) < 0) SR_SCPErr("Invalid texture name \"%s\"", nextstr);
				
				if (k) k = 4;
				if (c == 0) { sd->bottomtexture = n; origTEX[lnum] |= (1 << k); }
				if (c == 1) { sd->toptexture = n;    origTEX[lnum] |= (4 << k); }
				if (c == 2)	{ sd->midtexture = n;    origTEX[lnum] |= (2 << k); }
			}
			else
			if (!stricmp(nextstr, "addlowtexsec"))
			{
				n = getSecIdx(false);
				sc = &sectors[n];
				if (!(nextstr = SR_NextSTR(false))) SR_SCPErr("End of file");
				if ( (n = R_CheckTextureNumForName(nextstr)) < 0) SR_SCPErr("Invalid texture name \"%s\"", nextstr);

				for (i = 0; i < sc->linecount; i++)
				{
					ld = sc->lines[i];

					if (!ld->backsector) continue;
					if (ld->frontsector == ld->backsector) continue;
					if (ld->spflags & SPF_CONTROLLINE) continue;
					if (ld->frontsector->floorheight < ld->backsector->floorheight)
					{
						if (!sides[ld->sidenum[0]].bottomtexture)
						{
							sides[ld->sidenum[0]].bottomtexture = n;
							origTEX[ld - lines] |= 0x01;
						}
					}
					if (ld->frontsector->floorheight > ld->backsector->floorheight)
					{
						if (!sides[ld->sidenum[1]].bottomtexture)
						{
							sides[ld->sidenum[1]].bottomtexture = n;
							origTEX[ld - lines] |= 0x10;
						}
					}
				}
			}
			else
			if (!stricmp(nextstr, "remtex"))
			{
				int lnum = getLineIdx(numlines);		// linenum

				k = getVal(NULL) != 0;	// side 0 or 1
				if ( (n = lines[lnum].sidenum[k]) < 0 || n > numsides) SR_SCPErr("L%i side %i is invalid", lnum, k);

				sd = &sides[n];

				c = getVal(NULL);	// bottom 0, top 1, mid 2
				if (c < 0 || c > 2) SR_SCPErr("Invalid texture position num %i", c);
				
				if (k) k = 4;
				if (c == 0) { sd->bottomtexture = 0; origTEX[lnum] &= ~(1 << k); }
				if (c == 1) { sd->toptexture = 0;    origTEX[lnum] &= ~(4 << k); }
				if (c == 2)	{ sd->midtexture = 0;    origTEX[lnum] &= ~(2 << k); }
			}
			else
			if (!stricmp(nextstr, "basesec"))
			{

				sn = getSecIdx(false);
				fl = getSecIdx(false);

				if (verifySSEClist())
				{
					for (i = 0; i < hitssecptr; i++)
					{
						ss = &subsectors[hitssec[i]];

						if (fl & 1 && ss->floorsec != sn)
							Setr3dssfs(ss, sn, 52007);
						if (fl & 2 && ss->ceilsec != sn)
							Setr3dsscs(ss, sn, 52007);
					}
				}
			}
			else
			if (!stricmp(nextstr, "changessec"))
			{
				n = getSecIdx(false);

				if (verifySSEClist())
				{
					int flag;
					for (i = 0; i < hitssecptr; i++)
					{
						ss = &subsectors[hitssec[i]];

						flag = ss->ssflags & SSF_BASESSECFIX;

						if (flag) ss->ssflags &= ~SSF_BASESSECFIX;
						ChangeSSECsec(ss, &sectors[n], 52008);
						if (flag) ss->ssflags |= SSF_BASESSECFIX;
					}
				}
			}
			else
			if (!stricmp(nextstr, "changessecsec"))
			{
				n = getSsecIdx();
				sn = getSecIdx(false);
				subsectors[n].ssflags &= ~SSF_BASESSECFIX;
				ChangeSSECsec(&subsectors[n], &sectors[sn], 52024);
				subsectors[n].ssflags |= SSF_BASESSECFIX;
			}
			else
			if (!stricmp(nextstr, "changesseccsec"))
			{
				n = getSecIdx(false);	
				setSsecList(&sectors[n], true);

				int csec = getSecIdx(false);

				for (i = 0; i < hitssecptr; i++)
				{
					ss = &subsectors[hitssec[i]];

					ss->ceilsec = csec;
				}
			}
			else
			if (!stricmp(nextstr, "checksec"))
			{
				n = getSecIdx(false);
				ch = &sectors[n];

				n = 0;
				if ( (ch->ceilingheight >> 16) == getNextNumVal()) n++;
				if ( (ch->floorheight   >> 16) == getNextNumVal()) n++;
				if (   (nextstr = SR_NextSTR(false))
					&& !stricmp(nextstr, lumpinfo[ch->ceilingpic].name)) n++;
				if (   (nextstr = SR_NextSTR(false))
					&& !stricmp(nextstr, lumpinfo[ch->floorpic].name)) n++;

				if (n != 4)
				{
					SR_SCPWarn("Sector %i values do not match. Script aborted", ch - sectors);
					break;
				}
			}
			else
			if (!stricmp(nextstr, "csec"))
			{
				setSSECvals(1, false);
			}
			else
			if (!stricmp(nextstr, "customloop"))
			{
				n = getSecIdx(false);	// sector
				sc = &sectors[n];

				c = getVal(NULL);		// numlines

				if (c < 3 || c >= numlines)
				{
					SR_SCPWarn("Invalid number of lines %i. Script aborted", c);
					break;
				}

				CLR_vtxcheck();

				for (i = 0; i < c; i++)
				{
					n = getLineIdx(numlines);

					if (lines[n].frontsector!= sc && lines[n].backsector != sc)
					{
						SR_SCPWarn("Invalid line %i has no sector %i. Script aborted", n, sc - sectors);
						break;
					}

					linearray[i] = n;
					vtxcheck[lines[n].v1 - vertexes]++;
					vtxcheck[lines[n].v2 - vertexes]++;
				}

				M_ClearBox(ibox);

				// Check lines form a loop
				for (i = 0; i < c; i++)
				{
					n = linearray[i];

					if (vtxcheck[lines[n].v1 - vertexes] != 2 || vtxcheck[lines[n].v2 - vertexes] != 2)
					{
						c = 0; break;
					}

					M_AddLineToBox(ibox, &lines[n]);
				}

				if (!c || !LS_MarkSubsecsInAreaGeneral(sc, NULL, c, NULL, ibox, -1, MSIA_NONE))
				{
					SR_SCPWarn("customloop failed. Script aborted");
					break;
				}
			}
			else
			if (!stricmp(nextstr, "exclude"))
			{
				n = getSecIdx(false);
				if ( !(numssecs = excludeSsecsInLoop(&sectors[n])) )
				{
					break;
				}
			}
			else
			if (!stricmp(nextstr, "fn_setsamelinessec"))
			{
				int			lnum;

				lnum = getLineIdx(numlines);	// line num to check
				n = getSecIdx(false);	// sector num line belongs to
				ch = &sectors[n];
				n = getSecIdx(false);	// replacement sector num
				fl = getVal(&err);		// flags
				if (err) { SR_SCPWarn("Invalid flag value"); break; }

				ld = &lines[lnum];
				if (ld->frontsector != ch || ld->backsector != ch)
				{
					SR_SCPWarn("Invalid line %i", lnum);
				}
				else
				{
					sl = &sslines[lnum];

					for (i = 0; i < sl->num; i++)
					{
						if ( (ss = &subsectors[sl->list[i]])->sector == ch)
						{
							if (fl & 1 && ss->floorsec != n)
								Setr3dssfs(ss, n, 52009);
							if (fl & 2 && ss->ceilsec != n)
								Setr3dsscs(ss, n, 52009);
						}
					}
				}
			}
			else
			if (!stricmp(nextstr, "fn_setssec"))
			{
				n = getSecIdx(false);
				ch = &sectors[n];
				sn = getSecIdx(false);	// replacement sector num
				fl = getVal(&err);		// flags
				if (err) { SR_SCPWarn("Invalid flag value"); break; }

				setSsecList(ch, fl & 16);

				for (i = 0; i < hitssecptr; i++)
				{
					ss = &subsectors[hitssec[i]];

					if (fl & 1 && ss->floorsec != sn)
						Setr3dssfs(ss, sn, 52010);
					if (fl & 2 && ss->ceilsec != sn)
						Setr3dsscs(ss, sn, 52010);
					if (fl & 4 && ss->lfsec != sn)
						Setr3dsslfsec(ss, sn, 52010);
					if (fl & 8 && ss->lcsec != sn)
						Setr3dsslcsec(ss, sn, 52010);
				}
			}
			else
			if (!stricmp(nextstr, "fn_setssecind"))
			{
				int fs, cs, lfs, lcs;

				n = getSecIdx(false);
				ch = &sectors[n];
				fs = getSecIdx(true);	// replacement floorsec
				cs = getSecIdx(true);	// replacement ceilsec
				lfs = getSecIdx(true);	// replacement lfsec
				lcs = getSecIdx(true);	// replacement lcsec

				setSsecList(ch, true);

				for (i = 0; i < hitssecptr; i++)
				{
					ss = &subsectors[hitssec[i]];
					Setr3dssfs(ss, fs, 52011);
					Setr3dsscs(ss, cs, 52011);
					Setr3dsslfsec(ss, lfs, 52011);
					Setr3dsslcsec(ss, lcs, 52011);
				}

			}
			else if (!stricmp(nextstr, "fixfloorheight"))
			{
				int ht;

				n = getSecIdx(false);
				ht = getVal(NULL);	// height to set
				sectors[n].R3Dfloorheight = (ht << FRACBITS);
			}
			else if (!stricmp(nextstr, "fixsseccsec"))
			{
				if (verifySSEClist())
				{
					for (i = 0; i < hitssecptr; i++)
					{
						subsectors[hitssec[i]].ssflags |= SSF_CSECFIXED;
					}
				}
			}
			else
			if (!stricmp(nextstr, "forcesky"))
			{
				setSSECvals(5, true);
			}
			else
			if (!stricmp(nextstr, "fsec"))
			{
				setSSECvals(2, false);
			}
			else
			if (!stricmp(nextstr, "fracproject"))
			{
				n = getLineIdx(numlines);		// linenum
				lines[n].xtraflags |= XT_FRACPROJECT;
			}
			else
			if (!stricmp(nextstr, "fracprojectback"))
			{
				n = getLineIdx(numlines);		// linenum
				lines[n].xtraflags |= XT_FRACPROJECTBACK;
			}
			else
			if (!stricmp(nextstr, "l3dref"))
			{
				int	cht, ht;	// for current height and new height

				n = getLineIdx(numlines);		// linenum
				ld = &lines[n];
				if (ld->specType < R3D_40)
					SR_SCPErr("Line %i is not a 3D line", n);

				n = getVal(NULL);		// value to change
				if (n < 0 || n > 2)
					SR_SCPErr("Invalid val\n");

				ht = getVal(&err);	// new height to set
				if (err || ht < 0 || (!n && ht > 99999) || (n && ht > 999))
					SR_SCPErr("Invalid height\n");

				cht = n == 1 ? ld->d3d->lineheightlow : n == 2 ? ld->d3d->lineheightmid : 0;

				ht <<= FRACBITS;
				if      (n == 0) ld->d3d->offset3D = ht;
				else if (n == 1) { ld->d3d->lineheightlow = ht; ld->d3d->totalheight += (ht - cht); }
				else if (n == 2) { ld->d3d->lineheightmid = ht; ld->d3d->totalheight += (ht - cht); }
			}
			else
			if (!stricmp(nextstr, "largestloop"))
			{
				n = getSecIdx(false);
				sc = &sectors[n];
				if ( (n = getNextNumVal()) < 1) SR_SCPErr("Invalid loopcount num");
				if ( !(numssecs = getSsecsInLoop(sc, n)) )
				{
					break;
				}
			}
			else
			if (!stricmp(nextstr, "lcsec"))
			{
				setSSECvals(3, false);
			}
			else
			if (!stricmp(nextstr, "lfsec"))
			{
				setSSECvals(4, false);
			}
			else
			if (!stricmp(nextstr, "keyoverride"))
			{
				ma_inf.keyoverride = getLineIdx(numlines);		// linenum
			}
			else
			if (!stricmp(nextstr, "forceskywall"))
			{
				nextstr = SR_NextSTR(false);

				if (stricmp(nextstr, "{"))
					SR_SCPErr("missing opening {\n");

				while ( (nextstr = SR_NextSTR(true)) )
				{
					if (!stricmp(nextstr, "}") ) break;

					n = atoi(nextstr);

					if (n >= 0 && n < numlines)
						lines[n].ovflags |= OVF_NOOCCLUDE;

					if (lines[n].backsector)
						SR_SCPWarn("FORCESKYWALL: Line %i is not a one's line\n", n);
					else
						ma_inf.checkforcesky = true;
				}
			}
			else
			if (!stricmp(nextstr, "lineoffset"))
			{
				n = getLineIdx(numlines);		// linenum
				ld = &lines[n];
				n = getVal(NULL);		// side
				c = ld->sidenum[ (n > 0) ];
				if (c < 0) SR_SCPErr("Line %i has no sidenum for side %i", ld - lines, n > 0);
				n = getVal(NULL);		// xoff
				if (n < -2048) n = -2048;
				if (n >  2048) n =  2048;
				sides[c].textureoffset = n * FRACUNIT;
				n = getVal(NULL);		// yoff
				if (n < -2048) n = -2048;
				if (n >  2048) n =  2048;
				sides[c].rowoffset = n * FRACUNIT;

				// v220-33 These will be reloaded if a savegame is set
				if (ma_inf.annumsides >= maxannumsides)
				{
					maxannumsides += 32;
					analysesides = (analysesides_t*)W_Realloc(analysesides, maxannumsides * sizeof(analysesides_t), STATICID, 0);
				}
				analysesides[ma_inf.annumsides].sidenum = c;
				analysesides[ma_inf.annumsides].texoff = sides[c].textureoffset;
				analysesides[ma_inf.annumsides++].rowoff = sides[c].rowoffset;
			}
			else
			if (!stricmp(nextstr, "linemark"))
			{
				n = getLineIdx(numlines);		// linenum
				ld = &lines[n];
				n = getNextNumVal();	// type to be set to
				Settype(ld, (byte)n, 52012);
				SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 52012);
			}
			else
			if (!stricmp(nextstr, "nomisshisec"))
			{
				n = getSecIdx(false);		// secnum
				sc = &sectors[n];

				for (i = 0; i < sc->linecount; i++)
				{
					ld = sc->lines[i];
					ld->ml_flags |= ML_IGNOREMISSINGHI;
					SetLFG(ld, LFG_OVERRIDE, 52013);
				}
			}
			else
			if (!stricmp(nextstr, "nomisslosec"))
			{
				n = getSecIdx(false);		// secnum
				sc = &sectors[n];

				for (i = 0; i < sc->linecount; i++)
				{
					ld = sc->lines[i];
					ld->ml_flags |= ML_IGNOREMISSINGLO;
					SetLFG(ld, LFG_OVERRIDE, 52014);
				}
			}
			else
			if (!stricmp(nextstr, "linemarksec"))
			{
				n = getSecIdx(false);		// secnum
				ch = &sectors[n];
				n = getNextNumVal();		// type to be set to
				for (i = 0; i < ch->linecount; i++)
				{
					ld = ch->lines[i];
					if (ld->spflags & SPF_CONTROLLINE) continue;
					Absolutetype(ld, (byte)n, 52015);
					SetLFG(ld, LFG_NEWSET|LFG_FIXED|LFG_EXCLUDE|LFG_FIXABS|LFG_NOARBITER, 52015);
				}
			}
			else
			if (!stricmp(nextstr, "marklines"))
			{
				if (verifySSEClist())
				{
					validcount++;

					for (i = 0; i < hitssecptr; i++)
					{
						ss = &subsectors[hitssec[i]];
						sg = &segs[ss->firstline];

						for (k = 0; k < ss->linecount; k++, sg++)
						{
							if (!(ld = sg->linedef) || ld->validcount == validcount) continue;
							ld->validcount = validcount;
							setLineParms(ld, 52016);
						}
					}
				}
			}
			else
			if (!stricmp(nextstr, "marksectors"))
			{
				if (verifySSEClist())
				{
					int c;

					for (c = i = 0; i < hitssecptr; i++)
					{
						ss = &subsectors[hitssec[i]];
						addSec(ss->sector - sectors, &c);
					}

					for (i = 0; i < c; i++) SetSecidxDone(countsec[i], 52017);
				}
			}
			else
			if (!stricmp(nextstr, "movemobj"))
			{
				thinker_t	*thinker, *next;
				thinker = thinkercap.next;

				n = getVal(NULL);	// map number

				while(thinker != &thinkercap)
				{
					next = thinker->next;

					if ((void (__cdecl *)(void))thinker->function == (void (__cdecl *)(void))P_MobjThinker)
					{
						mobj_t *iter = (mobj_t *) thinker;

						if (MAPNUM(iter) == n)
						{
							iter->x = getVal(NULL)<<FRACBITS;
							iter->y = getVal(NULL)<<FRACBITS;
							break;
						}
					}
					thinker = next;
				}
			}
			else
			if (!stricmp(nextstr, "noclamp"))
			{
				nextstr = SR_NextSTR(false);
				if (!nextstr) SR_SCPErr("End of file");
				if ( (n = R_CheckTextureNumForName(nextstr)) < 0)
					SR_SCPErr("Invalid texture name \"%s\"", nextstr);
				texArray[n]->txf_flags |= TXF_NOCLAMP;
			}
			else
			if (!stricmp(nextstr, "preventGlow"))
			{
				if (strcmp(SR_NextSTR(true), "{")) SR_SCPErr("missing {");

				while(strcmp( (nextstr = SR_NextSTR(true)), "}"))
				{
					n = SR_NumForString(&err);	// secnum
					if (err) n = -1;
					if (n < 0 || n >= numsectors) SR_SCPErr("Invalid Sector num %i", n);

					for (i = 0; i < numsubsectors; i++)
					{
						if (subsectors[i].sector - sectors == n) subsectors[i].ssflags |= SSF_NOGLOW;
					}
				}
			}
			else
			if (!stricmp(nextstr, "radiusfix"))
			{
				thinker_t	*thinker, *next;
				thinker = thinkercap.next;

				c = 0;
				n = getVal(NULL);	// doomednum

				while(thinker != &thinkercap)
				{
					next = thinker->next;

					if ((void (__cdecl *)(void))thinker->function == (void (__cdecl *)(void))P_MobjThinker)
					{
						mobj_t *iter = (mobj_t *) thinker;

						if (iter->info->doomednum == n)
						{
							iter->flags3 |= MF3_USETRUERADIUS;
							if (!(iter->subsector->sector->specFlags & SF_MOVFLOOR))
								iter->flags2 |= MF2_SPECNOMOVE;
							c++;
						}
					}
					thinker = next;
				}
			}
			else
			if (!stricmp(nextstr, "removemobj"))
			{
				wadani.mapnummoremove = getVal(NULL);
			}
			else
			if (!stricmp(nextstr, "respawnmobj"))
			{
				wadani.mapnummorespawn = getVal(NULL);
				for (n = 0; n < 3; n++)
					wadani.triglinenum[n] = getVal(NULL);
			}
			else
			if (!stricmp(nextstr, "secdone"))
			{
				n = getSecIdx(false);
				SetSecidxDone(n, 52018);
				ch = &sectors[n];
				for (i = 0; i < ch->linecount; i++)
				{
					setLineParms(ch->lines[i], 52018);
				}
			}
			else
			if (!stricmp(nextstr, "secfixed"))
			{
				n = getSecIdx(false);
				SetSecidxDone(n, 52019);
				setLineFixed(&sectors[n]);
			}
			else
			if (!stricmp(nextstr, "setnoocclude"))
			{
				n = getSecIdx(false);
				SetNOocclude(&sectors[n], 52020);
			}
			else
			if (!stricmp(nextstr, "setfakecompat"))
			{
				compat[c_fake] = true;
			}
			else
			if (!stricmp(nextstr, "setsec"))
			{
				int in = getSecIdx(false);
				n = getSecIdx(false);
				fl = getSecIdx(false);

				sc = &sectors[n];

				for (i = 0; i < sc->linecount; i++)
				{
					sl = &sslines[sc->lines[i] - lines];

					for (k = 0; k < sl->num; k++)
					{
						ss = &subsectors[sl->list[k]];

						if (fl & 1 && ss->floorsec < 0 && ss->sector == sc)
							Setr3dssfs(ss, in, 52021);
						if (fl & 2 && ss->ceilsec < 0 && ss->sector == sc)
							Setr3dsscs(ss, in, 52021);
					}
				}
			}
			else
			if (!stricmp(nextstr, "setsecfloor"))
			{
				int in = getSecIdx(false);
				n = getSecIdx(false);

				sc = &sectors[n];

				for (i = 0; i < sc->linecount; i++)
				{
					sl = &sslines[sc->lines[i] - lines];

					for (k = 0; k < sl->num; k++)
					{
						ss = &subsectors[sl->list[k]];
						if (ss->lfsec < 0 && ss->sector == sc)
							Setr3dsslfsec(ss, in, 52022);
					}
				}
			}
			else
			if (!stricmp(nextstr, "setheightsec"))
			{
				int in = getSecIdx(false);
				n = getSecIdx(false);

				sectors[in].heightsec = n;
			}
			else
			if (!stricmp(nextstr, "setceillightsec"))
			{
				int in = getSecIdx(false);
				n = getSecIdx(false);

				sectors[in].ceilinglightsec = n;
			}
			else
			if (!stricmp(nextstr, "setsky"))
			{
				n = getSecIdx(false);
				sc = &sectors[n];
				nextstr = SR_NextSTR(false);
				if (!nextstr) SR_SCPErr("End of file");
				if ( (n = R_CheckTextureNumForName(nextstr)) < 0)
					SR_SCPErr("Invalid texture name \"%s\"", nextstr);
				sn = getVal(&err);		// rowoff
				if (err) { SR_SCPWarn("Invalid rowoff value"); break; }
				fl = getVal(&err);		// flags
				if (err) { SR_SCPWarn("Invalid flag value"); break; }
				sc->sidesky = n | PL_SKYFLAT;
				sc->sky_offy = sn * FRACUNIT;
				if (fl) sc->specFlags2 |= SF2_HASTYPE272;
			}
			else
			if (!stricmp(nextstr, "setssec"))
			{
				n = getSsecIdx();
				sn = getSecIdx(false);	// replacement sector num
				fl = getVal(&err);		// flags
				if (err) { SR_SCPWarn("Invalid flag value"); break; }
				ss = &subsectors[n];
				if (fl & 1) Setr3dssfs(ss, sn, 52023);
				if (fl & 2) Setr3dsscs(ss, sn, 52023);
				if (fl & 4)
				{
					ss->ssflags &= ~SSF_BASESSECFIX;
					ChangeSSECsec(ss, &sectors[sn], 52023);
					ss->ssflags |= SSF_BASESSECFIX;
				}
			}
			else
			if (!stricmp(nextstr, "setssecinlist"))
			{
				sn = getSecIdx(false);	// replacement sector num
				fl = getVal(&err);		// flags
				if (err) { SR_SCPWarn("Invalid flag value"); break; }

				if (strcmp(SR_NextSTR(true), "{")) SR_SCPErr("missing {");

				while(strcmp( (nextstr = SR_NextSTR(true)), "}"))
				{
					n = SR_NumForString(&err);	// ssecnum
					if (err) n = -1;
					if (n < 0 || n >= numsubsectors) SR_SCPErr("Invalid subsector num %i", n);
					ss = &subsectors[n];
					if (fl & 0x01) Setr3dssfs(ss, sn, 52024);
					if (fl & 0x02) Setr3dsslfsec(ss, sn, 52024);
					if (fl & 0x04) Setr3dsscs(ss, sn, 52024);
					if (fl & 0x08) Setr3dsslcsec(ss, sn, 52024);
					if (fl & 0x10)
					{
						ss->ssflags &= ~SSF_BASESSECFIX;
						ChangeSSECsec(ss, &sectors[sn], 52024);
						ss->ssflags |= SSF_BASESSECFIX;
					}
					if (fl & 0x20) ss->ssflags |= SSF_BASESSECFIX;
				}
			}
			else
			if (!stricmp(nextstr, "setssecsinbox"))
			{
				float	x, y, w, h;

				x = (float)getMU(false);
				y = (float)getMU(false);
				w = (float)getMU(true);
				h = (float)getMU(true);

				fbox[BOXLEFT ] = x;
				fbox[BOXRIGHT] = (x + w);
				fbox[BOXBASE ] = y;
				fbox[BOXTOP  ] = (y + h);

				subfn_setSsecsInBox(fbox);
			}
			else
			if (!stricmp(nextstr, "setssecsinboxsec"))
			{
				n = getSecIdx(true);
				memcpy(fbox, sectors[n].sibounds, sizeof(fbox));
#ifdef _DEBUG
				pr_OFinsert("S%i %f %f %f %f\n",
					n,
					fbox[BOXLEFT],
					fbox[BOXRIGHT],
					fbox[BOXTOP],
					fbox[BOXBASE]);
#endif

				subfn_setSsecsInBox(fbox);
			}
			else
			if (!stricmp(nextstr, "setunlit"))
			{
				thinker_t	*thinker, *next;
				thinker = thinkercap.next;

				n = getVal(false);	// mapnum

				while(thinker != &thinkercap)
				{
					next = thinker->next;

					if ((void (__cdecl *)(void))thinker->function == (void (__cdecl *)(void))P_MobjThinker)
					{
						mobj_t *iter = (mobj_t *) thinker;

						// Retrieve thing's map number
						if (MAPNUM(iter) == n)
						{
							iter->misc |= MJF_UNLIT;
							goto nextnum;
						}
					}
					thinker = next;
				}

				SR_SCPWarn("Thing not found\n");
			}
			else
			if (!stricmp(nextstr, "skyflatnum"))
			{
				nextstr = SR_NextSTR(false);

				skyflatnum = R_FlatNumForName(nextstr);
				setskycolour = getFloat(false);
			}
			else
			if (!stricmp(nextstr, "setspriteheight"))
			{
				thinker_t	*thinker, *next;
				thinker = thinkercap.next;

				int dmedn = getVal(false);	// doomednum

				n = getVal(false);	// height
				n <<= FRACBITS;		// fixed_t

				while(thinker != &thinkercap)
				{
					next = thinker->next;

					if ((void (__cdecl *)(void))thinker->function == (void (__cdecl *)(void))P_MobjThinker)
					{
						mobj_t *iter = (mobj_t *) thinker;

						if (iter->info->doomednum == dmedn)
						{
							if (iter->info->height != n) iter->info->height = n;
							iter->height = iter->info->height;
						}
					}
					thinker = next;
				}
			}
			else
			if (!stricmp(nextstr, "ssvg"))
			{
				n = getSsecIdx();

				subsectors[n].viewgroup = getVal(NULL);
			}
			else	// v220-29
			if (!stricmp(nextstr, "triggerDirect"))
			{
				n = getLineIdx(numlines);		// linenum

				if (   lines[n].special > 0
					&& lines[n].special < 270
					&& walkCheck[lines[n].special] == 1)
				{
					if (ge_inf.tdcount >= MAXTDNUM) goto nextnum;

					for (i = 0; i < ge_inf.tdcount; i++)
					{
						if (ge_inf.tdnum[i] == n) goto nextnum;
					}

					ge_inf.tdnum[ge_inf.tdcount++] = n;
				}
				else
				{
					SR_SCPWarn("L%i is not a standard walkover type", n);
				}
			}
			else
			if (!stricmp(nextstr, "exit"))
			{
				return false;
			}
			else
			{
				SR_SCPErr("unknown keyword %s", nextstr);
			}
		}
nextnum:;
	}

	return false;
}

void WA_CheckAnalyseFile(char *name, int from)
{
	extern char		actualwadname[MAXFNAME];

	char	buf[512];
	int		found = (from == 1 ? 8 : 0);
	uint	numchar;

	if (ge_inf.devmap)
	{
		if (from == 2)
			pr_OFinsert("\t%i %i %i %i\n", NFnumsectors, NFnumlinedefs, NFnumvertexes, NFnumsidedefs);
		else
			pr_OFinsert("\t%i %i %i %i %i\n", numsectors, numsubsectors, numlines, numsegs, numsides);
	}

	ge_inf.tdcount = 0;

	sprintf(buf, "%s%s", ddBinDir.path, name);

	if (!SR_LoadFile(buf, name, NULL))
	{
		Con_Error("ERROR: the file Risen3D\\bin\\%s is missing", name);
	}

recheck:;	
	nextstr = SR_NextSTR(false);

	if (_stricmp("Version", nextstr) || SR_NextNum() != 4)
	{
		Con_Error("ERROR: the %s file loaded is an old version\n"
				  "       Please install the latest version", name);
	}

	sprintf(buf, "[%s]", actualwadname);

	while(1)
	{
		if ( !(nextstr = SR_NextSTR(false)) ) break;

		if ( !(found & 1) )
		{
			if (!stricmp(nextstr, buf))
			{
				found |= 1;
			}
			else
			if (   nextstr[0] == '['
				&& nextstr[1] == '\\'
				&&  (numchar = strlen(nextstr)) > 3
				&& nextstr[numchar - 1] == ']')
			{
				numchar -= 3;	// Eliminate end ] from count

				if (strlen(actualwadname) >= numchar && !_strnicmp(actualwadname, nextstr + 2, numchar))
				{
					found |= 1;
				}
			}
			continue;
		}
		else if (*nextstr == '[')	// start of next definition
		{
			found = 0;
			continue;
		}

		if ( !(found & (2|4)) )
		{
			if (!stricmp(nextstr, new_mapid)) found |= 2;
			else
			if (found & 8 && !stricmp(nextstr, "mapall")) found |= 4;
			continue;
		}

		pr_OFinsert("Running %s script for %s %s\n", name, analysename, found & 4 ? "mapall" : new_mapid);

		while(1)
		{
			if (!strncmp(nextstr, "[", 1)) break;

			if (!stricmp(nextstr, "{"))
			{
				if (!(found & 4))
				{
					// Note: getVAL checks must be tested in same order as written in the R3D_ file
					if (from == 2)
					{
						if (   getVal(NULL) != NFnumsectors
							|| getVal(NULL) != NFnumlinedefs
							|| getVal(NULL) != NFnumvertexes
							|| getVal(NULL) != NFnumsidedefs)
						{
							if (ge_inf.devmap)
								SR_SCPErr("map parms for R3D_NODEfix don't match: script abandoned");
							else
								SR_SCPWarn("map parms don't match: script abandoned");
							goto end_script;
						}
					}
					else
					if (   getVal(NULL) != numsectors
						|| getVal(NULL) != numsubsectors
						|| getVal(NULL) != numlines
						|| getVal(NULL) != numsegs
						|| getVal(NULL) != numsides)
					{
						if (ge_inf.devmap)
							SR_SCPErr("map parms don't match: script abandoned");
						else
							SR_SCPWarn("map parms don't match: script abandoned");
						goto end_script;
					}
				}

				if (!subfn_parseCommands(from))
				{
					SR_SCPWarn("script abandoned");
					goto end_script;
				}
				if ( !(nextstr = SR_NextSTR(false))
					|| stricmp(nextstr, "{"))
				{
					if ((found & (4|8)) == (4|8))
					{
						found = 0; SR_RewindScript(); goto recheck;
					}
					goto end_script;
				}
			}
			else
				SR_SCPErr("opening { is missing");
		}
	}

end_script:;

	SR_EndScript();
}