//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jack Vermeulen, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_winit.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
//  screenfile message ammended
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
//  some reformatting, device context made global
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 17 (v220-17) 27/Oct/2012 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// DESCRIPTION: Set up main instance and window callbacks
//
//-----------------------------------------------------------------------------

//=========================================================================
//
// R3D_WINIT.C
//
// Win32 init.
// Create windows, load DLLs, setup APIs.
//
// Modified for R3D
//
//=========================================================================

// HEADER FILES ------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <io.h>
#include <direct.h>
#include <process.h>
#include <assert.h>
#include <math.h>
#include <stdio.h>

#include "resource.h"

#include "de_base.h"
#include "de_audio.h"
#include "de_console.h"
#include "de_system.h"
#include "de_play.h"
#include "de_refresh.h"
#include "de_network.h"
#include "de_misc.h"

#include "r3d_compat.h"
#include "r3d_sfx.h"

#include "gl_main.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	CloseSavefile(void);
void	getDesktopParms(int *width, int *height, int *freq, int *bpp);
RDBOOL	compareFULLSCREEN();
int		P_CountMobjThinkers(void);
void	closeWADhandles(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern FILE		*dbfile;		// debug file

extern int		check_called[7];
extern RDBOOL	ui_active;
extern int		con_error;
extern RDBOOL	gl_was_inited;
extern char		savepathfname[512];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

HWND		hWndMain;		// The main window handle
HINSTANCE	hInstApp;		// Instance handle to the application.
HDC			mainHdc;


FILE		*R3Doutfile;
char		LOGoutfile[MAXPATH];	// Path\Risen3D.out
char		DEBUGfile[MAXPATH];		// Path\R3Ddebug.out
char		screenfile[MAXPATH];
char		OsVersionString[128];
char		savefileDeletestr[1024];

RDBOOL		resolution_set;				// v220-3
RDBOOL		openDBFILE(void);
RDBOOL		waspaused;
RDBOOL		ScreenError;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static LPCTSTR lpszAppName = "Risen3D";
static char		R3Dcaption[32];
static WORD		original_gamma_ramp[256 * 3];
static WORD		gamma_ramp_support = false;
static HANDLE	hMutex;
static char		*errhndl = "Risen3D has detected an error and has self terminated.\n"
						   "Please report this to the Risen3D team together with a\n"
					       "copy of Risen3D.out (in the Risen3D folder).\n\n";

uint	errflags = MB_SETFOREGROUND|MB_OK|MB_SYSTEMMODAL|MB_ICONEXCLAMATION;

// CODE --------------------------------------------------------------------

//===========================================================================
// GL_MakeGammaRamp - jk
//	Gamma      - non-linear factor (curvature; >1.0 multiplies)
//	Contrast   - steepness
//	Brightness - uniform offset 
//===========================================================================
void W_MakeGammaRamp(float gamma, float contrast, float bright)
{
	WORD	ramp[3 * 256];
	double	ideal[256]; // After processing clamped to ushort.
	int		i;

	// Init the ramp as a line with the steepness defined by contrast.
	for (i = 0; i < 256; i++)
		ideal[i] = i*contrast - (contrast-1)*127;

	// Apply the gamma curve.
	if (gamma != 1)
	{
		double norm, gm = 1/gamma;

		if (gm < 0.1) gm = 0.1;
		norm = pow(255, gm - 1); // Normalizing factor.

		for (i = 0; i < 256; i++)
			ideal[i] = pow(ideal[i], gm) / norm;
	}

	// The last step is to add the brightness offset.
	for (i = 0; i < 256; i++) ideal[i] += bright * 128;

	// Clamp it and write the ramp table.
	for (i = 0; i < 256; i++)
	{
		ideal[i] *= 0x100; // Byte => word
		if(ideal[i] < 0) ideal[i] = 0;
		if(ideal[i] > 0xffff) ideal[i] = 0xffff;
		// RGB are each given the same values
		ramp[i] = ramp[i + 256] = ramp[i + 512] = (ushort) ideal[i];
	}

	update_gamma = false;
	SetDeviceGammaRamp(mainHdc, ramp);
}

void W_SetOriginalGammaRamp()
{
	if (gamma_ramp_support)
		SetDeviceGammaRamp(mainHdc, original_gamma_ramp);
}

// Just do the minimum required
void CrashShutdown()
{
	if (appStatus == 3) return;

	if (appStatus > 0)
	{
		CloseSavefile();	// In case of crash whilst open
	}

	if (dbfile)
	{
		fflush(dbfile);			// debug info file
		fclose(dbfile);
		dbfile = NULL;
	}

	if (R3Doutfile)
	{
		fflush(R3Doutfile);		// standard cfg file
		fclose(R3Doutfile);
		R3Doutfile = NULL;
	}

	appStatus = 3;
	
	if (ppt.viewplrmo)
		pr_OFinsert("Player position x %i y %i\n", ppt.viewplrmo->x >> 16, ppt.viewplrmo->y >> 16);

	SW_Shutdown();	// Kill the message window, if it exists.
	S_Shutdown();
	closeWADhandles();
	W_SetOriginalGammaRamp();

	if (!compareFULLSCREEN()) ChangeDisplaySettings(NULL, 0);	// Restore default mode

	pr_OFdirect("Crash shutdown sequence complete.\n");

	MessageBox(hWndMain, errhndl, "R3D ERROR HANDLER", errflags);
}

// v220-33 Functions that may be called on exit

RDBOOL printDBGfile;
RDBOOL printOutfile;
char *exitpath;

void fn2()
{
	if ((_access(exitpath, 4)) != -1)
		spawnlp(P_NOWAIT, exitpath, exitpath, "", 0);
}

// Used to give an informative message before exiting
static void ErrorBox(char *buff)
{
	MessageBox(NULL, buff, VERSIONTEXT, MB_OK | MB_ICONERROR);
	if (mainHdc) ReleaseDC(hWndMain, mainHdc);
	exit(0);
}

// end exit functions

// All shutdown function calls centralised - gmj
void EngineShutdown()
{
	FILE	*f;
	char	resfile[MAXPATH];
	int		current_appStatus = appStatus;

	if (appStatus == 3) return;

	appStatus = 3;

	if (!compat[c_nomusic]) Music_Stop();

	SW_Shutdown();	// Kill the message window, if it exists.
	CloseSavefile();

	if (current_appStatus == 2)
	{
		if (ppt.thinkers_have_run)
		{
			pr_OFdirect("\nIn the last game map:\n");
			pr_OFinsert("%10i mobj thinkers were active\n"		, P_CountMobjThinkers());
			pr_OFinsert("%10i mobj thinkers were created\n"		, ma_inf.mobjsCreated);
			pr_OFinsert("%10i mobj thinkers were resurrected\n"	, ma_inf.mobjsResurrected);
			pr_OFinsert("%10i mobj thinkers were destroyed\n"	, ma_inf.mobjsRemoved);
		}

		ZoomReset(true);	// to restore xhair before writing the cfg file
		Con_SaveDefaults();
	}

	W_SetOriginalGammaRamp();

#ifdef _DEBUG
	pr_OFinsert("%i %i %i %i %i %i %i\n",
		check_called[0],
		check_called[1],
		check_called[2],
		check_called[3],
		check_called[4],
		check_called[5],
		check_called[6]);
#endif

	pr_OFdirect("\nShutdown sequence initiated:\n");
	
    Sys_Shutdown();			pr_OFdirect(" A System shutdown\n");
	R3D_MAINfreeArrays();	pr_OFdirect(" B Main arrays released\n");
	B_Shutdown();			pr_OFdirect(" C Bindings released\n");
	Con_Shutdown();			pr_OFdirect(" D Shutdown console\n");
	F_ShutdownDFILES();		pr_OFdirect(" E Dfiles closed\n");
	closeWADhandles();		pr_OFdirect(" F WAD files closed\n");

	// Free memory arrays
	Def_Destroy(); pr_OFdirect(" H Definition arrays released\n");

	pr_OFdirect("\nReleasing win mem lists;\n");
	M_FreelumpCACHE();
	W_FreeALL(0);

	if (strstr(savefileDeletestr, "LOADGAME ERROR"))
	{
		if (strstr(savefileDeletestr, "Delete"))
		{
			char buff[1512];

			sprintf(buff, "%s\n%s\n", savefileDeletestr, savepathfname);

			if (MessageBox(NULL, buff, VERSIONTEXT, MB_TOPMOST|MB_YESNO|MB_ICONEXCLAMATION) == IDYES)			
			{
				if (remove(savepathfname))
				{
					Sys_AdvisoryHeader("ERROR: SAVEGAME DELETE FAILED ");
					pr_OFinsert("%s : %s\n", savepathfname, strerror(errno));
				}
				else
				{
					Sys_AdvisoryHeader("USER SAVEGAME DELETE ");
					pr_OFinsert("%sUser chose to delete file\n", buff);
				}
			}
			else
			{
				Sys_AdvisoryHeader("USER SAVEGAME LOAD ERROR ");
				pr_OFinsert("%sUser chose to keep file\n", buff);
			}
		}
		else {
			MessageBox(NULL, savefileDeletestr, VERSIONTEXT, MB_TOPMOST|MB_OK|MB_ICONEXCLAMATION);			
			Sys_AdvisoryHeader("USER SAVEGAME LOAD ERROR ");
			pr_OFinsert("%s\n", savefileDeletestr);
			if (savepathfname[0]) pr_OFinsert("%s\n", savepathfname);
		}

		Sys_AdvisoryHeader(NULL);
	}

	ArgShutdown();

	pr_OFdirect("Shutdown complete\n");

	if (mainHdc)
	{
		ReleaseDC(hWndMain, mainHdc);
		if (DestroyWindow(hWndMain)) hWndMain = NULL;
	}

	printDBGfile = dbfile && dbfile_active;
	printOutfile = R3Doutfile && !printDBGfile && opennotepad && gl_was_inited;

	if (dbfile)			// debug file
	{
		fflush(dbfile);
		fclose(dbfile);
		dbfile = NULL;
	}

	if (R3Doutfile)		// general info file
	{
		fflush(R3Doutfile);
		fclose(R3Doutfile);
		R3Doutfile = NULL;
	}

	// Don't call unless necessary - it can upset Radeon drivers
	if (!compareFULLSCREEN()) ChangeDisplaySettings(NULL, 0);	// Restore default mode

	sprintf(resfile, "%sR3Dres.out", ddBinDir.path);

	if (ScreenError)
	{
		if ( (_access(screenfile, 4)) != -1)
		{
			remove(screenfile);
			ErrorBox(
				"   #### RISEN3D ADVISORY ####\n\n"
				"PROBLEM: The last screen defaults saved cannot be used\n"
				"Either the file has been changed outside of Risen3D or\n"
				"a new monitor has been installed. Please restart to\n"
				"allow a new file to be created.\n\n"
				"###########################\n");
		}
		else
		{
			ErrorBox(
				"   #### RISEN3D ADVISORY ####\n\n"
				"PROBLEM: The requested settings have been rejected.\n"
				"Please install your monitor correctly and restart.\n\n"
				"###########################\n");
		}
	}
	else
	if (resolution_set)
	{
		// v220-3 dump current screen parms
		if ( (f = fopen(resfile, "wt")) != NULL)
		{
			fprintf(f, "%c %i x %i x %i * %i Hz",
				cMode.d_fullscreen ? 'F' : 'W',
				cMode.d_width,
				cMode.d_height,
				cMode.d_bpp,
				cMode.d_freq);
			fclose(f);
		}

		// v220-3 create screen_file
		if ( (f = fopen(screenfile, "wt")) != NULL)
		{
			fprintf(f,	"# This file has values validated for your system and must\n"
						"# neither be changed nor copied for use on a different system\n\n"
						"[SCREEN]\n"
						"width %i\n"
						"height %i\n"
						"freq %i\n"
						"fullscreen %i\n"
						"bpp %i\n"
						"xpos %i\n"
						"ypos %i\n"
						"monitor_w %i\n"
						"monitor_h %i\n"
						"use43 %i\n",
						cMode.d_width,
						cMode.d_height,
						cMode.d_freq,
						cMode.d_fullscreen,
						cMode.d_bpp,
						cMode.d_xpos,
						cMode.d_ypos,
						scr_max_width,
						scr_max_height,
						cfg.use43);
			
			fclose(f);
		}
	}

	if ((printDBGfile || printOutfile) && Dir_ChDir(&ddRuntimeDir))
	{
		// Default wordpad path
		const char *txtreader = "c:\\Program Files\\Windows NT\\Accessories\\wordpad.exe";

		if (printDBGfile)	// If anything was printed to it
		{
			if ((_access(txtreader, 4)) != -1)
				spawnlp(P_WAIT, txtreader, DEBUGfile, DEBUGfile, NULL);
			else
				spawnlp(P_WAIT, "notepad.exe", DEBUGfile, DEBUGfile, NULL);
		}

		if (printOutfile)
		{
			if ((_access(txtreader, 4)) != -1)
				spawnlp(P_WAIT, txtreader, LOGoutfile, LOGoutfile, NULL);
			else
				spawnlp(P_WAIT, "notepad.exe", LOGoutfile, LOGoutfile, NULL);
		}
	}

	exit(0);
}

//===========================================================================
// ErrorBoxAndQuit - should only be called after args have been inited
//                   and before a call is made to Sys_Init. gmj
//===========================================================================
void ErrorBoxAndQuit(char *err)
{
	ArgShutdown();
	if (R3Doutfile) fclose(R3Doutfile);
	ErrorBox(err);
}

void EarlyQuit()
{
	ArgShutdown();
	if (R3Doutfile) fclose(R3Doutfile);
	exit(0);
}

// R3D
static void CheckWINVer()
{
	OSVERSIONINFO OsInfo;

	char	*OsType = "Win95";	// default
	int		OsVersion;
	int		type;

	OsInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

	GetVersionEx (&OsInfo);

	type = OsInfo.dwPlatformId;
	OsVersion = OsInfo.dwBuildNumber;

	if (type == VER_PLATFORM_WIN32s)	// Exit
		ErrorBox("Win32s is not supported - At least Win98 is required\n");

	if (type == VER_PLATFORM_WIN32_WINDOWS && OsInfo.dwMinorVersion >= 10)
	{
		OsType = "ME";	// default

		if (OsInfo.dwMinorVersion < 90) OsType = "98";
		OsVersion &= 0xffff;
	}
	else if (type == VER_PLATFORM_WIN32_NT)
	{
		switch(OsInfo.dwMajorVersion)
		{
		case 0: case 1: case 2: case 3: case 4:
			OsType = "NT"; break;
		case 5:
			OsType = OsInfo.dwMinorVersion ? "XP" : "2000";
			break;
		case 6:
			OsType = OsInfo.dwMinorVersion ? "Windows 7" : "Vista";
			break;
		default:
			OsType = "Unknown";
			break;
		}
	}
	else	// Exit
		ErrorBox("Your OS is not supported - At least Win98 is required\n");

	sprintf(OsVersionString,
			"Windows %s %lu.%lu.%lu\n",
			OsType,
			OsInfo.dwMajorVersion, OsInfo.dwMinorVersion,
			OsVersion);
}

//=============================================================================
//  In debug mode return exception
//=============================================================================

static LONG WINAPI R3DExceptionHandler (LPEXCEPTION_POINTERS ExceptionInfo)
{
	if (R3Doutfile)	fflush(R3Doutfile);

#ifdef _DEBUG
	return EXCEPTION_CONTINUE_SEARCH;
#else
	if (appStatus != 3)
		CrashShutdown();
#endif
	exit(-1);
}

void pr_OFdirect(char *msg)
{
	if (R3Doutfile && msg) fprintf(R3Doutfile, msg);
}

// v220-6 print start msg (can be NULL)
//        print <rch> <num> of times 
//        print lfe number of linefeeds
void pr_OFrepeat(char *msg, char *rch, int num, int lfe)
{
	if (R3Doutfile)
	{
		int i;

		if (msg) fprintf(R3Doutfile, msg);
		if (rch) for (i = 0; i < num; i++) fprintf(R3Doutfile, rch);
		for (i = 0; i < lfe; i++) fprintf(R3Doutfile, "\n");
#ifdef _DEBUG
		fflush(R3Doutfile);
#else
		if (appStatus != 2 || con_error) fflush(R3Doutfile);
#endif
	}
}

void pr_OFinsert(char *message, ...)
{
	if (R3Doutfile)
	{
		char buffer[2048];
		va_list argptr;

		va_start(argptr, message);
		_vsnprintf(buffer, 2047, message, argptr);
		va_end(argptr);
		
		buffer[2047] = 0;
		fprintf(R3Doutfile, "%s", buffer);
#ifdef _DEBUG
		fflush(R3Doutfile);
#else
		if (appStatus != 2 || con_error) fflush(R3Doutfile);
#endif
	}
}

//===========================================================================
// WinMain - entry point
//===========================================================================
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, 
		LPSTR lpCmdLine, int nCmdShow)
{
	WNDCLASS	wc;
	int			x, y;

	char path[512];
	uint len;

	//----------------------------------

	// v220-13 prevent more than one instance of Risen3D from running
	SetLastError(0);
	if ( (hMutex = CreateMutex(NULL, true, lpszAppName)) == NULL)
		ErrorBox("Failed to create mutex");
	if (GetLastError() == ERROR_ALREADY_EXISTS)
		ErrorBox("Risen3D is already running");

	// Where are we?
	GetModuleFileName(hInstance, path, _MAX_PATH);
	Dir_FileDir(path, &ddBinDir);

	// Check Risen3D.exe is in correct location
	strcpy(path, ddBinDir.path);				// This will incude a trailing "\"

	if ( (len = strlen(path)) < 9 ||			// x:\a\bin\  minimum possible path
		_stricmp(path + len - 5, "\\bin\\"))
	{
		ErrorBox("Risen3D.exe not in \\bin directory");
	}

	// Store the runtime path (just knock off the bin\)
	path[len - 4] = 0;
	Dir_FileDir(path, &ddRuntimeDir);

	sprintf(path, "%sbin\\R3Dlauncher.exe", ddRuntimeDir.path);

	if ( (_access(path, 4)) == -1)
		ErrorBox("R3Dlauncher.exe not found in Risen3D\\bin directory");

	if (ArgInit() < 2)
	{
		ArgShutdown();
		ErrorBox("Insufficient Arguments sent\nCloseBox before continuing");
		exitpath = path;
		atexit(fn2);
		exit(0);
	}

	CheckWINVer();

	// Register a window class
	wc.style = 0;
	wc.lpfnWndProc = (WNDPROC) MainWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor (NULL , IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = lpszAppName;
	
	if (!RegisterClass((LPWNDCLASS)&wc))
		ErrorBox("Failed to register class");

	getDesktopParms(&x, &y, NULL, NULL);

	sprintf(R3Dcaption, "%s-%s", VERSIONTEXT, BUILDTEXT);

	// Create the main window
	hWndMain = CreateWindowEx(
			WS_EX_APPWINDOW,
			(LPCTSTR)lpszAppName,
			(LPCTSTR)R3Dcaption,
			WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_VISIBLE,
			(x - 320) / 2,
			(y - 200) / 2,
			320,
			200,
			NULL, NULL, hInstance, NULL);

	if (!hWndMain)
		ErrorBoxAndQuit("Failed to create window");

	// Make the instance handle global knowledge.
	hInstApp = hInstance;
	mainHdc = GetDC(hWndMain);

	if (GetDeviceGammaRamp(mainHdc, original_gamma_ramp))
		gamma_ramp_support = true;

	ShowWindow(hWndMain, SW_SHOW);
	UpdateWindow(hWndMain);

	// Set the font.
	SetMapMode(mainHdc, MM_TEXT);
	SelectObject(mainHdc, GetStockObject(SYSTEM_FIXED_FONT));

	// Messages will be saved to this file
	sprintf(LOGoutfile, "%sRisen3D.out" , ddRuntimeDir.path);
	sprintf(DEBUGfile , "%sR3Ddebug.out", ddRuntimeDir.path);

	if( (R3Doutfile = freopen(LOGoutfile, "w", stdout)) == NULL )
		ErrorBox("Failed to reassign stdout");

	// Initialise the win mem ptr arrays
	// Must be done before any chance of an R3D error exit
	W_InitWINM();

	_chdir(ddRuntimeDir.path);

	SetUnhandledExceptionFilter(R3DExceptionHandler); // jav

	sprintf(screenfile, "%sscreen_defaults", ddBinDir.path);

	pr_OFinsert("Command Line:\n%s\n\n", lpCmdLine);

	// Fire up the engine. The game loop acts as the message pump.
	DD_Main();

	return 0;
}

// v220-5 revised.
// v220-13 mustpause added as part of code change to allow
//         music to continue to play
// v220-31 deal with fullscreen issues with Win10
void setPauseState(RDBOOL setpause, RDBOOL mustpause)
{
	if (setpause)
	{
		if (paused) return;

	/*	if (ge_inf.realtekCompat)*/ Sfx_StopSound(0, NULL); // v220-33

		paused = true;
		waspaused = true;

		if (!ui_active || mustpause)
		{
			FModPause(true);
			MC_WinMidiPause(true);
		}
		// v220-31 only suspend repeating channels
		Sfx_SuspendRepeatingChannels();
		W_SetOriginalGammaRamp();
	}
	else
	{
		if (!paused) return;

		update_gamma = true;
		ShowWindow(hWndMain, SW_RESTORE);
		FModPause(false);
		MC_WinMidiPause(false);
		// v220-31 restart repeating channels
		Sfx_ReleaseRepeatingChannels();
		paused = false;
	}
}

// v220-5 revised.
LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	// v220-4 only check win pos etc. when loading is complete
	switch(msg)
	{
	case WM_CLOSE:	// ALT-F4
		if (yesno_exit > 0)
		{
			yesno_exit = -2;
			return 0;
		}

		if (appStatus == 2)
		{
			appTerminate = true;			// Exit will be from DD_GameLoop()
			return 0;
		}
		break;

	case WM_SETFOCUS:
		if (!yesno_exit && ge_inf.clear_screen & 1) ge_inf.clear_screen = 3;
		if (appStatus == 2)
			setPauseState(false, false);
		break;

	case WM_KILLFOCUS:
		if (appStatus == 2) setPauseState(true, false);
		break;

	case WM_ACTIVATE:
		if (!yesno_exit && ge_inf.clear_screen & 1) ge_inf.clear_screen = 3;
		if (appStatus == 2)
			waspaused = true;
		break;

	case WM_WINDOWPOSCHANGED:
		if (   appStatus == 2
			&& !cMode.w_checkwin
			&& !cMode.fullscreen)
		{
			cMode.w_checkwin = -1;
		}
		break;

	case WM_KEYDOWN:
		if (yesno_exit > 0)
		{
			if (wParam == 89)
			{
				yesno_exit = -1;
				return 0;
			}

			if (yesno_exit == 1 && wParam == 78)
			{
				yesno_exit = -2;
				return 0;
			}
		}

		if (   wParam == VK_PAUSE
			&& appStatus == 2
			&& ppt.thinkers_have_run)
		{
			setPauseState(!paused, true);
			return 0;
		}

		break;

	default:
		break;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}
