//-----------------------------------------------------------------------------
//
// Copyright (C) 2007 - 2008 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r3d_xssec.c	New file b211-0
//
// Revision R3D v2.1.1 build 0 04/Mar/2007 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
//
// DESCRIPTION:   Boundary checks.
//
//-----------------------------------------------------------------------------


// HEADER FILES ------------------------------------------------------------

#include <string.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_play.h"
#include "de_refresh.h"
#include "m_misc.h"
#include "r3d_MAINspec.h"
#include "r3d_edit.h"
#include "r3d_defs.h"
#include "r3d_utils.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

typedef struct {
	line_t		*plx;
	line_t		*ptest_line;
	RDBOOL		test_line_same;
	RDBOOL		test_line_tex;
	int			test_line_num;

	float		line_dist;
	fixed_t		refxc, refyc;
	vertex_t	*pva, *pvb;
	seg_t		*pseg;		// Only set with XS_SetBoundary()
	float		heightcmp;

} xssec_t;

static xssec_t xs;

// CODE --------------------------------------------------------------------

#pragma optimize("g", off)

static float lineDist(line_t *ld)
{
	double	lena = M_AccurateDistance(ld->v1->x - xs.refxc, ld->v1->y - xs.refyc);
	double	lenb = M_AccurateDistance(ld->v1->x - ld->v2->x, ld->v1->y - ld->v2->y);
	double	lenc = M_AccurateDistance(ld->v2->x - xs.refxc, ld->v2->y - xs.refyc);
	double	an, height;

	// Find the point on the (projected) line ld
	// that if drawn through refxc, refyc would be at 90deg to ld

	// height (of projected line) = 2 * sqrt(s(s - lena)(s - lenb)(s - lenc))/lenb
	// where s = 1/2(lena + lenb + lenc) and lenb is ld's length

	double	s = (lena + lenb + lenc) * 0.5;
	fixed_t	px, py, hx, hy, test_box[4];

	fixed_t	mvldx = meanV(ld->v1->x, ld->v2->x);
	fixed_t	mvldy = meanV(ld->v1->y, ld->v2->y);

	//-----------------------

	height = 2 * sqrt(s * (s - lena) * (s - lenb) * (s - lenc)) / lenb;

	if (!ld->dx)
		an = PI/2;
	else if (!ld->dy)
		an = 0;
	else
		an = atan((double)ld->dy/(double)ld->dx);

	hx = abs((int)(height * sin(an) * FRACUNIT));
	hy = abs((int)(height * cos(an) * FRACUNIT));

	// Check that the projected point from refxc, refyc lies on ld
	px = mvldx >= xs.refxc ? xs.refxc + hx : xs.refxc - hx;
	py = mvldy >= xs.refyc ? xs.refyc + hy : xs.refyc - hy;

	memcpy(test_box, ld->bbox, sizeof(test_box));

	// Introduce a small fiddle factor to avoid accuracy errors.
	test_box[BOXLEFT ] -= 2;
	test_box[BOXRIGHT] += 2;
	test_box[BOXBASE ] -= 2;
	test_box[BOXTOP  ] += 2;

	if (   px >= test_box[BOXLEFT ]
		&& px <= test_box[BOXRIGHT]
		&& py >= test_box[BOXBASE ]
		&& py <= test_box[BOXTOP  ])
	{
		return (float) height;
	}

	return (float) MIN_OF(lena, lenc);
}

static RDBOOL doorLine(line_t *ld)
{
	int txn;

	if (   !ld->backsector
		|| !((txn = origTEX[ld - lines]) & 0x22)
		|| txn & 0x55)
	{
		return false;
	}

	return ld->frontsector->specFlags & SF_MOVSECTOR
		|| ld->backsector->specFlags  & SF_MOVSECTOR;
}

static float unitDist(line_t *ld, vertex_t *vd, vertex_t *vt, vertex_t *vr)
{
	double	a, b;
	int		xa, ya, xb, yb;
	int		x1, y1, x2, y2;

	// Translate to 0, 0
	x1 = vd->x - vr->x; y1 = vd->y - vr->y;
	x2 = vt->x - vr->x; y2 = vt->y - vr->y;

	if ((!x1 && !y1) || (!x2 && !y2))
	{
		return 0;	// shouldn't happen
	}

	// Starting from the shared co-ord (now at 0,0)

	if (x1 && y1)
	{
		if (abs(x1) < abs(y1))
		{
			ya = y1 > 0 ? FRACUNIT : -FRACUNIT;
			// get the x value where y = 1
			xa = FixedMul(FixedDiv(FRACUNIT, abs(y1)), FixedDiv(x1, abs(y1)));
		}
		else
		{
			xa = x1 > 0 ? FRACUNIT : -FRACUNIT;
			// get the y value where x = 1
			ya = FixedMul(FixedDiv(FRACUNIT, abs(x1)), FixedDiv(y1, abs(x1)));
		}
	}
	else if (x1)
	{
		xa = x1 > 0 ? FRACUNIT : -FRACUNIT;
		ya = 0;
	}
	else 
	{
		ya = y1 > 0 ? FRACUNIT : -FRACUNIT;
		xa = 0;
	}

	if (x2 && y2)
	{
		if (abs(x2) < abs(y2))
		{
			yb = y2 > 0 ? FRACUNIT : -FRACUNIT;
			// get the x value where y = 1
			xb = FixedMul(FixedDiv(FRACUNIT, abs(y2)), FixedDiv(x2, abs(y2)));
		}
		else
		{
			xb = x2 > 0 ? FRACUNIT : -FRACUNIT;
			// get the y value where x = 1
			yb = FixedMul(FixedDiv(FRACUNIT, abs(x2)), FixedDiv(y2, abs(x2)));
		}
	}
	else if (x2)
	{
		xb = x2 > 0 ? FRACUNIT : -FRACUNIT;
		yb = 0;
	}
	else
	{
		yb = y2 > 0 ? FRACUNIT : -FRACUNIT;
		xb = 0;
	}

	a = (double)FIX2FLT(xa - xb);
	b = (double)FIX2FLT(ya - yb);

	return (float)sqrt(a*a + b*b);
}

//
// PIT_CrossSSEC
//
static RDBOOL PIT_CrossSSEC(line_t *ld, void *data)
{
	if (   !(ld->spflags & SPF_CONTROLLINE)		// control lines never define a boundary
		&& ld->specType < R3D_40				// nor 3D line types
		&& ld->frontsector != ld->backsector	// nor same sided lines
		&& !(linetable[ld - lines] & 2)			// nor diff sec of a particular type
		&& !doorLine(ld))						// nor midtex door lines
	{
		int side  = P_PointOnLineSide(xs.refxc, xs.refyc, ld);

		if (side && !ld->backsector) return true;

		if (xs.ptest_line->frontsector == xs.ptest_line->backsector)
		{
			if (linetable[ld - lines] == 4
				&& ld->frontsector->floorheight > xs.ptest_line->frontsector->floorheight)
			{
				return true;
			}
		}

		// Eliminate lines that are end on end (can't easily determine the side)
		if (   !(!xs.ptest_line->dx && !ld->dx && xs.ptest_line->v1->x == ld->v1->x)
			&& !(!xs.ptest_line->dy && !ld->dy && xs.ptest_line->v1->y == ld->v1->y)
			&& !(!ld->dx && xs.refxc == ld->v1->x)
			&& !(!ld->dy && xs.refyc == ld->v1->y))
		{
			// boundary line
			float calc_dist;

			if (ld->v1->x == xs.pva->x && ld->v1->y == xs.pva->y)
			{
				calc_dist = unitDist(ld, ld->v2, xs.pvb, ld->v1);
			}
			else if (ld->v2->x == xs.pva->x && ld->v2->y == xs.pva->y)
			{
				calc_dist = unitDist(ld, ld->v1, xs.pvb, ld->v2);
			}
			else if (ld->v1->x == xs.pvb->x && ld->v1->y == xs.pvb->y)
			{
				calc_dist = unitDist(ld, ld->v2, xs.pva, ld->v1);
			}
			else if (ld->v2->x == xs.pvb->x && ld->v2->y == xs.pvb->y)
			{
				calc_dist = unitDist(ld, ld->v1, xs.pva, ld->v2);
			}
			else
			{
				calc_dist = lineDist(ld);
			}

			if (!calc_dist) return true;

			if (calc_dist < xs.line_dist)
			{
				xs.plx = ld; xs.line_dist = xs.heightcmp = calc_dist;
			}
			else if (xs.plx && calc_dist == xs.heightcmp)
			{
				int		 sd1 = P_PointOnLineSide(xs.refxc, xs.refyc, xs.plx);
				int		 sd2 = P_PointOnLineSide(xs.refxc, xs.refyc, ld);
				sector_t *s1 = sd1 ? xs.plx->backsector : xs.plx->frontsector; 
				sector_t *s2 = sd2 ? ld->backsector : ld->frontsector; 
				
				if (s1 != s2)
				{
					xs.plx = NULL; xs.heightcmp = 0; 
				}
			}
		}
	}

	return true;
}

// XS_FindBoundary
//	Finds the first valid boundary line from the point x, y
//
static sector_t *XS_FindBoundary (line_t *lt, vertex_t *v1, vertex_t *v2, int *ref, int from)
{
	int		b[] = { 128, 256, 1024, 0 };
	int		i;

	xs.ptest_line     = lt;
	xs.test_line_num  = lt - lines;
	xs.test_line_tex  = origTEX[xs.test_line_num];
	xs.test_line_same = lt->frontsector == lt->backsector;

	xs.pva = v1, xs.pvb = v2;

	xs.plx = NULL;
	xs.heightcmp = 0;

	xs.refxc = meanV(v1->x, v2->x);
	xs.refyc = meanV(v1->y, v2->y);
	xs.line_dist = 1024.0f;

	validcount++;

	lt->validcount = validcount;

	// Fan out - with any luck we'll get the line straight away
	for (i = 0; b[i]; i++)
	{
		int bx, by, xl, xh, yl, yh;
		int offset = b[i] * FRACUNIT;

		xl = (xs.refxc - offset - bmaporgx)>>MAPBLOCKSHIFT;
		xh = (xs.refxc + offset - bmaporgx)>>MAPBLOCKSHIFT;
		yl = (xs.refyc - offset - bmaporgy)>>MAPBLOCKSHIFT;
		yh = (xs.refyc + offset - bmaporgy)>>MAPBLOCKSHIFT;

		for ( bx = xl ; bx <= xh ; bx++)
		{
			for (by = yl ; by <= yh ; by++)
			{
				if (!P_BlockLinesIterator(bx, by, PIT_CrossSSEC, NULL))
					return NULL;
			}
		}
		if (xs.plx) break;
	}

	if (ref) *ref = xs.plx ? xs.plx - lines : -1;

	if (!xs.plx) return NULL;

	return P_PointOnLineSide(xs.refxc, xs.refyc, xs.plx) ? xs.plx->backsector : xs.plx->frontsector;
}

// Only call if the line's front and back sectors are the same
static RDBOOL sameseclineValid(int lineidx)
{
	if (!(origTEX[lineidx] & 0x22)
		|| lines[lineidx].frontsector->ceilingheight
		-  lines[lineidx].frontsector->floorheight < 4*FRACUNIT ) return true;

	if ((origTEX[lineidx] & 0x22) != 0x22)
		return !(lines[lineidx].frontsector->specFlags & SF_MOVSECTOR);

	return texArray[sides[lines[lineidx].sidenum[0]].midtexture]->height
		== texArray[sides[lines[lineidx].sidenum[1]].midtexture]->height;
}

static RDBOOL chkLineContained(vertex_t *v1, vertex_t *v2, sector_t *sec)
{
	int i, N = 0, S = 0, E = 0, W = 0;
	int box[4], vbox[4];

	vbox[BOXLEFT ] = MIN_OF(v1->x, v2->x);
	vbox[BOXRIGHT] = MAX_OF(v1->x, v2->x);
	vbox[BOXBASE ] = MIN_OF(v1->y, v2->y);
	vbox[BOXTOP  ] = MAX_OF(v1->y, v2->y);

	M_ClearBox(box);

	for (i = 0; i < sec->linecount; i++)
	{
		line_t *ld = sec->lines[i];

		if (!(ld->spflags & SPF_CONTROLLINE))
		{
			int side = P_PointOnLineSide(xs.refxc, xs.refyc, ld);

			sector_t *s = side ? ld->backsector : ld->frontsector;

			if (s == sec)
			{
				if (sec->specFlags & SF_SECCLOSED
					&& nf_CheckBoxRing(sec, vbox, ld))
				{
					return true;
				}
				else
				{
					if      (ld->v1->x < xs.refxc || ld->v2->x < xs.refxc) E++, M_AddLineToBox(box, ld);
					else if (ld->v1->x > xs.refxc || ld->v2->x > xs.refxc) W++, M_AddLineToBox(box, ld);
					if      (ld->v1->y < xs.refyc || ld->v2->y < xs.refyc) S++, M_AddLineToBox(box, ld);
					else if (ld->v1->y > xs.refyc || ld->v2->y > xs.refyc) N++, M_AddLineToBox(box, ld);

					if (N && S && E && W && M_FitBox(vbox, box)) return true;
				}
			}
		}
	}
	return false;
}

void XS_setLineTable()
{
	line_t	*ld;
	int		i;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		if ( !(ld->ml_flags & ML_TWOSIDED)
			|| ld->spflags & SPF_LOL_MASK
			|| lineflags[i] & LFG_3DTYPE
			|| ld->specType >= R3D_40
			|| (origTEX[i] && Vtfl3(ld)))
		{
			linetable[i] = 1;
		}
		else if (ld->frontsector == ld->backsector)
		{
			if (vtoflags[ld->v1 - vertexes] && vtoflags[ld->v2 - vertexes])
			{
				linetable[i] = 1;
			}
			else
			{
				if ( !(origTEX[i] & 0x33)
					&& origTEX[i] & 0x44
					&& IsLiquidFloorFlat(ld->frontsector))
				{
					// do nothing;
				}
				else
				if (   lineflags[i] & LFG_NEWSET
					|| ld->frontsector->heightsec >= 0
					|| !sameseclineValid(i))
				{
					linetable[i] = 1;
				}
			}
		}
		else if (ld->frontsector->stptr >= 0
			  || ld->backsector->stptr  >= 0)
		{
			linetable[i] = 1;
		}
		else if (!origTEX[i]
			&& ((ld->frontsector->specFlags & 0xC) == 0xC
			||  (ld->backsector->specFlags  & 0xC) == 0xC)
			&& (Vtfl3(ld)
			|| (!(ld->frontsector->specFlags & SF_MOVSECTOR)
			&&  !(ld->backsector->specFlags  & SF_MOVSECTOR)
			&& SecCtrl(ld->frontsector - sectors)
			&& SecCtrl(ld->backsector - sectors)
			&& ld->frontsector->floorheight == ld->backsector->floorheight
			&& ld->frontsector->ceilingheight == ld->backsector->ceilingheight)) )
		{
			linetable[i] = 2;
		}
		else if ( (!(origTEX[i] & 0x55)	// v220-6 test modified
			|| (!(origTEX[i] & 0x33)
			&& (ld->frontsector->ceilingheight == ld->backsector->ceilingheight
			||  ld->frontsector->ceilingpic == skyflatnum && ld->backsector->ceilingpic == skyflatnum)))
			&& !Vtfl(ld)
			&& Vtxs2(ld)
			&& IsLiquidFloorFlat(ld->frontsector)
			&& ld->frontsector->floorpic == ld->backsector->floorpic
			&& ((ld->frontsector->specFlags & 0x1C) == 0x14
			||  (ld->backsector->specFlags  & 0x1C) == 0x14))
		{
			linetable[i] = 2;
		}
		// Check for type R3D_37
		else if (!origTEX[i]
			&& (ld->frontsector->specFlags & 0xC) == 0xC
			&& (ld->backsector->specFlags  & 0xC) == 0xC 
			&& ld->frontsector->floorheight == ld->backsector->floorheight
			&& ld->frontsector->ceilingheight != ld->backsector->ceilingheight
			&& ld->frontsector->ceilingpic == skyflatnum
			&& ld->backsector->ceilingpic == skyflatnum)
		{
			linetable[i] = 2;
		}
		else if (!(origTEX[i] & 0x55)
			  && ld->frontsector->floorpic   == ld->backsector->floorpic
			  && ld->frontsector->ceilingpic == ld->backsector->ceilingpic)
		{
			RDBOOL sameheights =
				   ld->frontsector->floorheight   == ld->backsector->floorheight
				&& ld->frontsector->ceilingheight == ld->backsector->ceilingheight;

			if (   sameheights
				&& ld->frontsector->lightlevel != ld->backsector->lightlevel
				&& !IsLiquidFloorFlat(ld->frontsector))
			{
				linetable[i] = (1|16);
			}
			else if (sameheights
				&& (ld->frontsector->specFlags & 0x3000C) == 0xC
				&& (ld->backsector->specFlags & 0x3000C) == 0xC)
			{
				// test added v220-9
				if (   ld->frontsector->lightlevel != ld->backsector->lightlevel
					&& ld->frontsector->ceilingpic == skyflatnum
					&& ld->backsector->ceilingpic == skyflatnum)
				{
					linetable[i] = 1;
				}
				else
				{
					linetable[i] = 2|32;
				}
			}
			else if (sameheights
				&& Vtxs2(ld)
				&& SecCtrl(ld->frontsector - sectors)
				&& SecCtrl(ld->backsector  - sectors)
				&& (ld->frontsector->specFlags & SF_MOVSECTOR) == SF_MOVFLOOR
				&& (ld->backsector->specFlags  & SF_MOVSECTOR) == SF_MOVFLOOR)
			{
				linetable[i] = 2;
			}
			else if (sameheights
				&& !vtxflags[ld->v1 - vertexes]
				&& !vtxflags[ld->v2 - vertexes]
				&& !SecCtrl(ld->frontsector - sectors)
				&& !SecCtrl(ld->backsector  - sectors)
				&& !(ld->frontsector->specFlags & SF_MOVSECTOR)
				&& !(ld->backsector->specFlags  & SF_MOVSECTOR))
			{
				RDBOOL s1 = (ld->frontsector->specFlags & 0xC) == 0xC;
				RDBOOL s2 = (ld->backsector->specFlags  & 0xC) == 0xC;

				if (s1 ^ s2)
					linetable[i] = 8;
				else
					linetable[i] = 4;
			}
			else
			{
				linetable[i] = 1;
			}
		}
		else
		{
			linetable[i] = 1;
		}
	}
}

sector_t *XS_GetBoundary(sector_t *sec, int *ref)
{
	sector_t	*s;
	line_t		*ld;
	int			i;

	for (i = 0; i < sec->linecount; i++)
	{
		ld = sec->lines[i];

		if (ld->spflags & SPF_CONTROLLINE) continue;

		if (   lineflags[ld - lines] & (LFG_NEWSET|LFG_3DTYPE)
			|| linetable[ld - lines] & (1|4))
		{
			return NULL;	// Invalid test line
		}

		xs.refxc = meanV(ld->v1->x, ld->v2->x);
		xs.refyc = meanV(ld->v1->y, ld->v2->y);

		s = XS_FindBoundary(ld, ld->v1, ld->v2, ref, 1);

		if (s) return s;
	}
	return NULL;
}

int XS_GetBoundarySecnum(sector_t *sec)
{
	sector_t *s = XS_GetBoundary(sec, NULL);

	return s ? s - sectors : -1;
}

static seg_t *segmem[50];
static int   segcounter;

static int getSegs(line_t *ld)
{
	subsector_t *ss;
	seg_t		*sg;
	ssline_t	*sl = &sslines[ld - lines];
	int			i, k;

	for (segcounter = i = 0; i < sl->num; i++)
	{
		ss = &subsectors[sl->list[i]];
		sg = &segs[ss->firstline];

		for (k = 0; k < ss->linecount; k++, sg++)
		{
			if (segcounter == 50) break;
			if (sg->linedef == ld) segmem[segcounter++] = sg;
		}
	}

	return segcounter;
}

void XS_SetLineSegs(line_t *ld)
{
	seg_t	*seg;
	int		i, k, m, n, sn, ref;
	
	sn = getSegs(ld);

	// Find the rendering sector that a line with the same sector each side should reference
	for (i = 0; i < sn; i++)
	{
		line_t		*lr   = NULL;
		sector_t	*sec  = NULL;
		RDBOOL		unset = true;

		seg = segmem[i];
		xs.pseg = seg;
		sec = XS_FindBoundary(ld, seg->v1, seg->v2, &ref, 2);

		if (sec && ref >= 0)
		{
			// Note that we have to test for this line as rejecting it prior
			// to running XS_FindBoundary can result in a totally wrong line
			// being selected as this one could well be the nearest one.
			// IOW we need to get the nearest line in all circumstances.
			if (sec && ref >= 0)
			{
				lr = &lines[ref];

				if (   (seg->v1->x != lr->v1->x && seg->v1->y != lr->v1->y)
					&& (seg->v1->x != lr->v2->x && seg->v1->y != lr->v2->y)
					&& (seg->v2->x != lr->v1->x && seg->v2->y != lr->v1->y)
					&& (seg->v2->x != lr->v2->x && seg->v2->y != lr->v2->y)
					&& P_PointOnLineSide(seg->v1->x, seg->v1->y, lr)
					!= P_PointOnLineSide(seg->v2->x, seg->v2->y, lr) )
				{
					ma_inf.boundssegfail++;
					continue;
				}
			}
		}

		if (lr)
		{
			ssline_t	*sl = &sslines[ld - lines];
			line_t		*lr = &lines[ref];

			//------------------------------------------

			if (lnsec[ld - lines] >= 0) Setspec(ld, -1, 35000);

			if (!Vtxs1(ld) && !chkLineContained(seg->v1, seg->v2, sec)) continue;

			SetLFG(ld, LFG_NEWSET, 35001);

			for (k = 0; k < sl->num; k++)
			{
				subsector_t *ss = &subsectors[sl->list[k]];
				seg_t		*sg = &segs[ss->firstline];

				if (ss->floorsec >= 0) continue;

				for (m = 0; m < ss->linecount; m++, sg++)
				{
					if (sg == seg)
					{
						if (sec != ld->frontsector && lr->frontsector != lr->backsector)
						{
							int		x = (int)(ss->midpoint.x * FRACUNIT);
							int		y = (int)(ss->midpoint.y * FRACUNIT);
							int		side = P_PointOnLineSide(x, y, lr);

							if (side)
							{
								if (lr->backsector != sec) break;
							}
							else
							{
								if (lr->frontsector != sec) break;
							}
						}
						
						if (   ss->floorsec < 0
							&& ss->ceilsec < 0
							&& ss->lfsec != sec - sectors
							&& ss->sector != sec)
						{
							sector_t *uppersec = NULL;

							// Check for an overriding upper sec
							if (ld->frontsector == ld->backsector)
							{
								seg_t *sgc = &segs[ss->firstline];

								for (n = 0; n < ss->linecount; n++, sgc++)
								{
									if (   sgc->sidedef
										&& sgc->frontsector != sec
										&& sgc->frontsector
										&& sgc->backsector
										&& sgc->sidedef->toptexture
										&& sgc->frontsector->ceilingheight > sgc->backsector->ceilingheight)
									{
										uppersec = sgc->frontsector; break;
									}
								}
							}

							if (uppersec)
							{
								Setr3dsscs(ss, uppersec - sectors, 35002);
								Setr3dssfs(ss, sec - sectors, 35002);
							}
							else
								SetOverride(ss, sec - sectors, 35002);
						}

						goto endloop;
					}
				}
			}
endloop:;
			xs.pseg = NULL;
		}
	}
}

void XS_GetBoundsFromLine(line_t *ld, int *losec, int *hisec, int *rf)
{
	sector_t *s;
	int		 ref;

	*losec = -1, *hisec = -1;
	if (rf) *rf = -1;

	if (   linetable[ld - lines] & (1|4)
		|| ld->spflags & (SPF_LINEONLINE|SPF_LINEONLINE2))
	{
		return;
	}

	if (   ld->frontsector == ld->backsector
		&& (vtxarray[ld->v1 - vertexes] == 1
		||  vtxarray[ld->v2 - vertexes] == 1) )
	{
		ssline_t *sl = &sslines[ld - lines];
		line_t	 *li;
		int		 i, k;

		for (i = 0; i < sl->num; i++)
		{
			subsector_t *ss = &subsectors[sl->list[i]];
			seg_t		*sg = &segs[ss->firstline];
			RDBOOL		same;

			for (k = 0; k < ss->linecount; k++, sg++)
			{
				if (  (li = sg->linedef)
					&& li != ld
					&& !Vtfl3(li)
					&& (!li->backsector
					|| (!(same = li->frontsector == li->backsector)
					&& !(sg->frontsector->specFlags & 0xC))
					|| (same && origTEX[li - lines] == 0x22)) )
				{
					if (!LS_AllSameSecLine(sg->frontsector))
					{
						*hisec = *losec = sg->frontsector - sectors;
						return;
					}
				}
			}
		}
	}

	s = XS_FindBoundary(ld, ld->v1, ld->v2, &ref, 3);

	if (   s && ref >= 0
		&& !LS_AllSameSecLine(s))
	{
		line_t	*lr = &lines[ref];

		*losec = s - sectors;
		*hisec = s - sectors;
		if (rf) *rf = ref;

		if (Vtfl3(lr) && lr->backsector)
		{
			int flags = LS_TestLineTex(lr);

			if (flags & LT_MISLO && flags & LT_DIFFLO)
			{
				*losec = lr->frontsector->floorheight > lr->backsector->floorheight ?
						lr->frontsector - sectors : lr->backsector - sectors;
			}

			if (flags & LT_MISHI)
			{
				*hisec = lr->frontsector->ceilingheight < lr->backsector->ceilingheight ?
						lr->frontsector - sectors : lr->backsector - sectors;
			}
		}
	}
}

#pragma optimize("", on)
