//-----------------------------------------------------------------------------
//
// Copyright (C) 1998 - 2005 id Software, Inc., Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r_things.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified/extended for R3D
//  main changes are for Core FX, DEH support, Thing visibility
//  and slopes.
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  old sort code removed
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 14 (v220-14) 01/May/2012 gmj
// Revision 2.2.0 build 21 (v220-21) 25/Jul/2013 gmj
// Revision 2.2.0 build 23 (v220-23) 10/Jan/2014 gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
// Revision 2.2.0 build 27 (v220-27) 09/Sep/2014 gmj
// Revision 2.2.0 build 28 (v220-28) 04/Oct/2014 gmj
// Revision 2.2.0 build 29 (v220-29) 22/Nov/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doom, Doomsday v1.7.8
//
//
// DESCRIPTION:   Preparation of data for rendering
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_network.h"
#include "de_refresh.h"
#include "de_play.h"
#include "de_render.h"
#include "de_graphics.h"
#include "de_misc.h"

#include "m_winmem.h"
#include "r3d_MAINspec.h"
#include "doom/info.h"
#include "doom/p_local.h"
#include "r3d_items.h"
#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

#define MAX_FRAMES			100
#define MAX_OBJECT_RADIUS	128
#define HALF_PI				1.57079632679f

// 10*BANG_180*BAMS_PI/65536 = 5*BAMS_PI
#define BAMS_FACTOR			15.7079632679

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

//lumobj_t	*getLuminous(int num); $halo removed
uint		Sys_GetRealTime (void);
RDBOOL		CheckAction(state_t *st);	// v220-3
int			C_CheckBspBox(fixed_t *bspbox);
int			dynsplnum = 0;
short		SectorDoomLight(short ll);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

RDBOOL IW_CheckSpritePalette(int lump);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern fixed_t	*amodelintensity;
extern RDBOOL	setNoFog;		// v220-3
extern int		maxmodelDefs;
extern RDBOOL	dehdebug;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// NOTE: 46341 sqrt(32768*32768*2) is actually the greatest distance possible
// (where 32768 = map blocks max 256 * block size 128) but 65536 will do
float	r_maxmodelz;
float	zdistlist[] = { 1000.f, 2000.f, 3000.f, 6000.f, 65536.f };
int		zdistptr = 3;	// default

int		thing_validcount;

// variables used to look up and range check thing_t sprites patches
spritedef_t		*sprites;
int				numsprites;

spritelump_t	*spritelumps;
int				numspritelumps;

vissprite_t		*vissprites;
vissprite_t		**vispointer;

uint			maxvissprites;
uint			visspritecnt;

int th_items[TH_NUMTYPES];	// b220-9

vissprite_t		vispsprites[DDMAXPSPRITES];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static spriteframe_t	sprtemp[MAX_FRAMES];

static sector_t	*chksec;
static mobj_t	*chkthing;

// sector specific list
typedef struct {
	mobj_t	**slist;
	int		maxnum;
} visseclist_t;

static visseclist_t	*visseclist;
static int			*vissec_count;

int LevelFullBright = false;

// CODE --------------------------------------------------------------------

void modelError(int lump, int num)
{
	char	buf[256];
	char	*name;

	if (lump >= 0 && lump < numlumps)
		name = lumpinfo[lump].name;
	else
		name = "unknown";

	if (num >= 0)
	{
		sprintf(buf,
		"Sprite \"%s\" with mapthing T%i has no model which is illegal for a scripted map\n",
			name, num);
		Con_Error("%s", buf);
		return;
	}
	else
		sprintf(buf,
		"Hudsprite \"%s\" has no model which is illegal for a scripted map\n",
			name);

	Con_ErrorD(buf);
}

//===============================================================================
// Required for R3D sky routine
//===============================================================================
int GetSpriteMaxHeight(uint sprite)
{
	int height = 0;

	if (sprite < (unsigned)numsprites)
	{
		spritedef_t	*sprdef = &sprites[sprite];
		int			i, k, h;

		for (i = 0; i < sprdef->numframes; i++)
		{
			spriteframe_t *sf = &sprdef->spriteframes[i];

			for (k = 0; k < 8; k++)
			{
				if (sf->lump[k] < 0) break;

				// v220-33 add yoffset
				h = spritelumps[sf->lump[k]].height + spritelumps[sf->lump[k]].slyoffset;
				if (h > height) height = h;
			}
		}
	}

	return height << FRACBITS;
}

// v220-9
static int SetSpriteMaxWidth(uint sprite)
{
	int width = 0, lump = 0;

	if (sprite < (unsigned)numsprites)
	{
		spritedef_t	*sprdef = &sprites[sprite];
		int			i, k, h, w;

		for (i = 0; i < sprdef->numframes; i++)
		{
			spriteframe_t *sf = &sprdef->spriteframes[i];

			for (k = 0; k < 8; k++)
			{
				if (sf->lump[k] < 0) break;

				lump = spritelumps[sf->lump[k]].lump;

				DoomLumpIdentify(lump, &w, &h, DLI_GRAPHIC);

				if (w > width) width = w;
			}
		}
	}

	return width << FRACBITS;
}

/*
===============================================================================

						INITIALISATION FUNCTIONS

 Sprite rotation 0 is facing the viewer, rotation 1 is one angle turn 
 CLOCKWISE around the axis. This is not the same as the angle, which 
 increases counter clockwise (protractor).

===============================================================================
*/

const char *flametype[] = {
	"YFLM", "TGFL", "TBFL", "TRFL",
	"SGFL", "SBFL", "SRFL", "CFLM",
	"EVIL", "HFLM", NULL
};

RDBOOL chkLinkedType(char *nptr)
{
	if (   loadModels
		&& nptr[5] && !nptr[6]
		&& nptr[5] == '0'
		&& nptr[4] >= 'A'
		&& nptr[4] <= 'D')
	{
		int  i;

		for (i = 0; flametype[i] != NULL; i++)
		{
			if (!strnicmp(nptr, flametype[i], 4))
				return true;	// match found
		}
	}
	return false;
}

//===========================================================================
// R_NewSpritePtr
//	Initialises and returns the new sprite pointer.
//===========================================================================
static int reloading;

static int R_NewSpritePtr(int lump, RDBOOL checkPWAD)
{
	int i, splnum;

	if (checkPWAD)
	{
		for (i = 0; i < numspritelumps; i++)
		{
			if (lumpinfo[spritelumps[i].lump].handle != IWADhandle && !stricmp(lumpinfo[spritelumps[i].lump].name, lumpinfo[lump].name))
			{
				if (reloading && dehdebug)
					pr_OFinsert("Found PWAD name %s %s\n", lumpinfo[spritelumps[i].lump].name, lumpinfo[lump].name);
				return i;
			}
		}
	}

	// Is this lump already entered?
	for (i = 0; i < numspritelumps; i++)
	{
		if (!stricmp(lumpinfo[spritelumps[i].lump].name, lumpinfo[lump].name)) return i;
	}

	if (reloading && dehdebug)
		pr_OFinsert("Found %s name %s\n", lumpinfo[lump].handle == IWADhandle ? "IWAD" : "PWAD", lumpinfo[lump].name);

	splnum = numspritelumps++;

	spritelumps = (spritelump_t*)W_ReallocCLR(spritelumps,
		sizeof(spritelump_t) * (numspritelumps), STATICID, 308);

	spritelumps[splnum].lump = lump;

	if (   !strnicmp(lumpinfo[lump].name, "PISG", 3)
		|| !strnicmp(lumpinfo[lump].name, "PISF", 3))
	{
		lumpinfo[lump].liflags |= LINF_PISTOL;
	}

	return splnum;
}

//===========================================================================
// R_InstallSpriteLump
//	 Local function for R_InitSprites.
//===========================================================================
static void R_InstallSpriteLump (int lump, unsigned frame, unsigned rotation, RDBOOL checkPWAD, RDBOOL flipped)
{
	int		r, splnum = -1;

	if (rotation > 8 || frame >= MAX_FRAMES || (splnum = R_NewSpritePtr(lump, checkPWAD)) < 0)
		Con_Error ("Program error [11]: lump %i rotation %i frame %i splnum %i", lump, rotation, frame, splnum);

	if (rotation == 0)
	{
		sprtemp[frame].rotate = 0;

		for (r = 0; r < 8; r++)
		{
			sprtemp[frame].lump[r] = splnum; // ptr to spritelump;
			sprtemp[frame].flip[r] = (byte) flipped;
		}
		return;
	}

	sprtemp[frame].rotate = 1;

	--rotation;
	sprtemp[frame].lump[rotation] = splnum; // ptr to spritelump;
	sprtemp[frame].flip[rotation] = (byte) flipped;

	return;
}

//===========================================================================
// R_InitSpriteDefs
//	Pass a null terminated list of sprite names (4 chars exactly) to be used
//	Builds the sprite rotation matrixes to account for horizontally flipped
//	sprites.  Will report an error if the lumps are inconsistant.
//
//	Sprite lump names are 4 characters for the actor, a letter for the frame,
//	and a number for the rotation, A sprite that is flippable will have an
//	additional letter/number appended.  The rotation character can be 0 to
//	signify no rotations.
//
//  Rewritten for R3D
//===========================================================================

static RDBOOL checkSpriteName(int lump)
{
	char *name = lumpinfo[lump].name;
	int	 len;

	if (   lumpinfo[lump].groupMarker != LGT_SPRITES
		|| ( (len = strlen(name)) != 6 && len != 8)
		|| (byte)name[4] < 'A'		// Validate the rotation/frame numbers bytes 4/5
		|| (byte)(name[4] - 'A') >= MAX_FRAMES
		|| (byte)name[5] < '0'
		|| (byte)name[5] > '8'
		|| (len == 8 &&				// If len 8 (flippable types)
		(  (byte)name[6] < 'A'		// Validate the rotation/frame numbers bytes 6/7
		|| (byte)(name[6] - 'A') >= MAX_FRAMES
		|| (byte)name[7] < '0'
		|| (byte)name[7] > '8'))
		|| CheckDEHThingSPR(name))	// Finally check if replaced by DEH
	{
		return false;
	}
	return true;
}

RDBOOL R_CheckForPWADSprites()
{
	char	*name;
	int		i, k;

	for (i = numlumps; --i >= 0;)
	{
		if (!pwadHandle(lumpinfo[i].handle)) continue;	// Not a pwad lump
		if (!checkSpriteName(i)) continue;	// Not a sprite lump

		name = lumpinfo[i].name;

		for (k = 0; k < numsprites; k++)
		{
			if (!strnicmp(sprnames[k].name, name, 4))
				return true;
		}
	}

	return false;
}

static RDBOOL CheckHiResSprite(char *name)
{
	int i;

	for (i = 0; i < numlumps; i++)
	{
		if (   pwadHandle(lumpinfo[i].handle)
			&& (lumpinfo[i].lumptype == LMP_PNG
			||  lumpinfo[i].lumptype == LMP_JPG)
			&& !stricmp(name, lumpinfo[i].name))
		{
			return true;
		}
	}

	return false;
}

#define		PWAD_LIST		0
#define		IWAD_LIST		1
#define		RWAD_LIST		2	// sprites in risen3d.wad
#define		WADLIST_TYPES	3

typedef struct {
	__int64	list[WADLIST_TYPES][MAX_FRAMES*8];	// __int64 to store up to 8 byte name
	int		lump[WADLIST_TYPES][MAX_FRAMES*8];
	int		listcounter[WADLIST_TYPES];
	int		maxframe;
} checksprite_t;

void R_InitSpriteDefs(void)
{
	int		lim, num, frame, rotation;
	int		i, k, m;

	// Allow models with duplicate sprites
	RDBOOL	hasmodels = CFX && !ge_inf.noobjects;	// noobjects is an edit mode option

	checksprite_t	*cs;
			
	byte	*amwdsflags;
	char	*name;

	//-------------------------------------

	numspritelumps = 0;

	if (ge_inf.codlev) ge_inf.addon_flags |= AF_COD;

	// Check that some sprites are defined.
	if (!numsprites) return;

	sprites = (spritedef_t*)W_Calloc(numsprites * sizeof(*sprites), STATICID, 114);

	cs = (checksprite_t*)W_Calloc(numsprites * sizeof(checksprite_t), STATICID, 318);

	// Set the maxframe sprite counter to -1
	for (i = 0; i < numsprites; i++) cs[i].maxframe = -1;

	// Scan the lumps validating all sprites including duplicates
	// Add lump sprites to the appropriate list
	// Scan backwards so that only the latest types are loaded
	for (i = numlumps; --i >= 0;)
	{
		checksprite_t	*cst;
		char			fullname[10];
		__int64			hname;			// To simplify the code
		uint			len;
		int				found, type;

		//------------------------------------------

		// Is this a sprite frame? (only standard LGT_SPRITES are searched)
		if (!checkSpriteName(i)) continue;	// Not a sprite lump

		name = lumpinfo[i].name;
		len = strlen(name);

		memset (fullname, 0, 9);
		strncpy(fullname, name, 8);
		_strupr (fullname);
		hname = *(__int64*) fullname;

		for (k = 0; k < numsprites; k++)
		{
			if (strnicmp(sprnames[k].name, name, 4)) continue;

			found = false;

			type =  (lumpinfo[i].handle == IWADhandle) ?
					IWAD_LIST :
					(lumpinfo[i].handle == RISEN3DWADhandle) ?
					RWAD_LIST : PWAD_LIST;

			if (   !dynsplnum
				&& type == RWAD_LIST
				&& !stricmp(sprnames[k].name, "DYNLD0"))
			{
				dynsplnum = k;
			}

			if (ge_inf.replaceall && type == PWAD_LIST) continue;

			cst = &cs[k];

			// Check if on list
			for (m = 0; m < cst->listcounter[type]; m++)
			{
				if (cst->list[type][m] == hname)
				{
					found = true; break;
				}
			}

			if (found) continue;

			// Check for hi-res sprite
			if (   type == IWAD_LIST
				&& CheckHiResSprite(lumpinfo[i].name))
			{
				sprites[k].sprflags |= SPR_HIRES;
			}

			if ( !(sprites[k].sprflags & SPR_NOMODEL) )
			{
				if (ge_inf.switch_flags & SWF_NC)
					sprites[k].sprflags |= SPR_NOMODEL;
				else if (type == IWAD_LIST	// v220-21 palette change detection
					&& !ge_inf.amwds
					&& IW_CheckSpritePalette(i))
				{
					for (int n = 0; n < numsprites; n++)
					{
						if (!strnicmp(sprnames[n].name, name, 4))
							sprites[n].sprflags |= SPR_NOMODEL;
					}
				}
			}

			// Add to list
			num = cst->listcounter[type];
			
			if (num < MAX_FRAMES*8)
			{
				int fnum = (int)((byte)fullname[4] - 'A');

				if (len == 8)
					fnum = MAX_OF(fnum, ((int)((byte)fullname[6] - 'A')));

				cst->lump[type][num] = i;
				cst->list[type][num] = hname;
				cst->listcounter[type]++;

				if (cst->maxframe < fnum) cst->maxframe = fnum;
			}
		}
	}

	amwdsflags = (byte*)W_Calloc(numsprites, STATICID, 0);

	// Check for an "add_on" lump
	if (!(ge_inf.switch_flags & (SWF_NC|SWF_XA)) && SR_LoadPWADLumpName("ADLEVINF"))
	{
		RDBOOL	nomodels = ArgExists("-xa");
		RDBOOL	allow = true;
		char	*nextstr;

		if (ge_inf.hasscp_defn)
		{
			allow = false;
			pr_OFinsert("R3D_ScriptedMap: ADLEVINF lump has been ignored.\n");
		}

		while(allow && (nextstr = SR_NextSTR(false)) )
		{
			if (!_stricmp(nextstr, "flags"))
			{
				nextstr = SR_NextSTR(true);
			}
			else // if (strlen(nextstr) == 4)
			{
				int	num = -1;

				for (i = 0; i < numsprites; i++)
				{
					if (!_stricmp(nextstr, sprnames[i].name))
					{
						num = i; break;
					}
				}

				if (num >= 0)
				{
					int flagval = SR_NextNum();

					if (   (!loadModels && flagval & 1)
						|| ( loadModels && flagval & 2) )
					{
						amwdsflags[num] = true;
					}
				}
				else
				{
					SR_SCPWarn("Entry not found - script ADLEVINF aborted");
					memset(amwdsflags, 0, numsprites);
					break;
				}
			}
		}

		SR_EndScript();
	}

	lim = ge_inf.hasscp_defn ? 0 : MIN_OF(numsprites, SPR_TNT1);

	if (ge_inf.replaceall)
	{
		ge_inf.spritemodels = true;

		for (i = 0; i < lim; i++) amwdsflags[i] = true;
	}
	else
	{
		if (ge_inf.devmap && dehdebug)
		{
			pr_OFdirect("\nList of sprites replaced in pwad\n");
			pr_OFrepeat(NULL, "=", 32, 1);
		}

		for (i = 0; i < lim; i++)
		{
			if (hasmodels && DEHCheck[i] & DEHF_USEMODEL)
			{
				ge_inf.spritemodels = amwdsflags[i] = true;
				sprites[i].sprflags &= ~(SPR_NOMODEL|SPR_HIRES);
				if (ge_inf.devmap && dehdebug)
					pr_OFinsert("ALLOW MODEL: sprite %s\n", sprnames[i].name);
			}
			else if (cs[i].listcounter[PWAD_LIST])
			{
				sprites[i].sprflags |= SPR_NOMODEL;
			}
		}

		if (ge_inf.devmap && dehdebug)
		{
			pr_OFinsert("End list ");
			pr_OFrepeat(NULL, "=", 23, 2);
		}
	}

	// If not directly specified in R3D_WADmodify then revert to the Force Model Override option
	if (!ge_inf.spritemodels && ge_inf.amwds)
	{
		// Check if any need to have models suppressed by comparing the lists
		for (i = 0; i < lim; i++)
		{
			checksprite_t	*cst = &cs[i];
			int				sizei, sizep;

			if ( !(num = cst->listcounter[IWAD_LIST]) ) continue;	// no iwad sprite (probably doom)
			if (sprites[i].sprflags & SPR_HIRES) continue;

#ifdef _DEBUG
			if (cst->listcounter[PWAD_LIST] == num)
				pr_OFinsert("%5i) R_InitSpriteDefs: lump iwad %i pwad %i\n", i, num, cst->listcounter[PWAD_LIST]);
#endif

			if (   !(DEHCheck[i] & DEHF_SPRITE)
				&& !amwdsflags[i]
				// v220-6 only allow certain actors
				&& (i == SPR_TROO || i == SPR_POSS|| i == SPR_SPOS || i == SPR_CPOS || i == SPR_SKEL) )
			{
				cst->listcounter[PWAD_LIST] = 0;
				continue;
			}

			// There will only be sprites on list 0 if they exist in
			// the pwad.  This means that none of the following will
			// be processed where a pwad has no sprites.
			if (cst->listcounter[PWAD_LIST] == num)
			{
				RDBOOL cfx_type = CFX && (i == SPR_BAL1 || i == SPR_BAL2 || i == SPR_BAL7);

				for (k = 0; k < num; k++)
				{
					int		found = -1;
					__int64	hname = cst->list[PWAD_LIST][k];

					sizep = lumpinfo[cst->lump[PWAD_LIST][k]].size;

					for (m = 0; m < num; m++)
					{
						if (cst->list[IWAD_LIST][m] == hname)
						{
							sizei = lumpinfo[cst->lump[IWAD_LIST][m]].size;
							found = m;
							break;
						}
					}

					if (found >= 0 && cfx_type) continue;

					// If the lump size is within an int's size then we assume they
					// are the same.  The need to test within an int is because
					// many of the original textures were stored to the exact
					// byte number whereas replicas have usually been int aligned.
					if (found < 0 || abs(sizep - sizei) > 4)
					{
						ge_inf.suppress_duplicate++;
						sprites[i].sprflags |= SPR_NOMODEL;
						break;
					}
				}
			}
			else if ((cst->listcounter[PWAD_LIST] || DEHCheck[i] & DEHF_SPRITE) && !amwdsflags[i])
			{
				sprites[i].sprflags |= SPR_NOMODEL;
			}
		}
	}

	// Sprites can now be loaded
	for (i = 0; i < numsprites; i++)
	{
		char	*spritename = sprnames[i].name;
		int		type = -1, maxframe;
		RDBOOL	specspr = chkLinkedType(spritename);
		RDBOOL	customspr = false;
		RDBOOL	has_rotate = false;
		byte	framenm[MAX_FRAMES];
		int		checkPWAD = false;
		int		typeset;
		int		reloadcount = 0;

		checksprite_t *cst = &cs[i];

		// check the frames that were found for completeness
		if (cst->maxframe < 0)
		{
			sprites[i].spriteframes = 0;
			sprites[i].numframes = 0;
			sprites[i].sprflags |= SPR_NOMODEL;
			continue;
		}

		if (!(DEHCheck[i] & DEHF_SPRITE))
		{
			if (cst->listcounter[RWAD_LIST])
			{
				type = RWAD_LIST;
			}
		}

		if (type == -1)	// Get the dominant type
		{
			for (k = 0; k < WADLIST_TYPES; k++)
			{
				if (cst->listcounter[k])	// pwads take precedence
				{
					type = k;
					break;
				}
			}
		}

		if (type < 0) continue;	// shouldn't happen
		typeset = type;
		reloading = false;

reload_type:;

		maxframe = cst->maxframe;
		memset(sprtemp, -1, sizeof(sprtemp));
		memset(framenm,  0, MAX_FRAMES);

		if (checkPWAD) type = IWAD_LIST;

		for (k = 0; k < cst->listcounter[type]; k++)		// load the sprite lumps
		{
			num = cst->lump[type][k];
			name = lumpinfo[num].name;

			if (specspr && chkLinkedType(lumpinfo[k].name))
				sprites[i].sprflags = SPR_LINKMODEL;

			frame = name[4] - 'A'; rotation = name[5] - '0';

			if (rotation) has_rotate = true;

			R_InstallSpriteLump(num, frame, rotation, checkPWAD, false);

			framenm[frame] = true;

			if(name[6])
			{
				frame = name[6] - 'A'; rotation = name[7] - '0';

				if (rotation) has_rotate = true;

				R_InstallSpriteLump(num, frame, rotation, checkPWAD, true);

				framenm[frame] = true;
			}
		}


		// Check all frames were found where loading from a pwad
		if (type == PWAD_LIST && maxframe >= 0)
		{
			for (k = 0; k <= maxframe; k++)
			{
				if (framenm[k]) continue;

				// Check on list 1
				for(m = 0; m < cst->listcounter[1]; m++)		// load the sprite lumps
				{
					num = cst->lump[1][m];
					name = lumpinfo[num].name;

					frame = name[4] - 'A'; rotation = name[5] - '0';

					if (!framenm[frame])
					{
						customspr = true;
						R_InstallSpriteLump(num, frame, rotation, false, false);
					}

					if(name[6])
					{
						frame = name[6] - 'A'; rotation = name[7] - '0';

						if (!framenm[frame])
						{
							customspr = true;
							R_InstallSpriteLump(num, frame, rotation, false, true);
						}
					}
				}
			}
		}

		checkPWAD = false;

		maxframe++;

		for (frame = 0; frame < maxframe ; frame++)
		{
			switch (sprtemp[frame].rotate)
			{
			case 0:		// only the first rotation is needed
				break;

			case -1:	// no rotations were found for that frame at all
				{
					if (!frame)
					{
						Con_Error("R_InitSprites: %s type %i has no patch for frame %i",
									spritename, type, frame + 1);
					}

					Con_Message("R_InitSprites: %s has no patch for frame %i\n"
								  "               Frame 0 will be substituted\n",
								spritename, frame + 1);

					customspr = true;
				}
				break;

			case 1:		// should have all 8 frames
				for (rotation = 1 ; rotation < 8 ; rotation++)
				{
					if (sprtemp[frame].lump[rotation] == -1)
					{
						if (type == PWAD_LIST && !reloadcount)
						{
							pr_OFinsert("R_InitSprites: Incomplete PWAD list IWAD sprites will be inserted\n");
							// insert missing sprites from iwad list
							customspr = false;
							if (ge_inf.amwds)
							{
								sprites[i].sprflags &= ~SPR_NOMODEL;
								amwdsflags[i] = true;
							}
							checkPWAD = true;
							reloading = true;
							reloadcount++;
							goto reload_type;
						}
						else
						{
							Con_Error("R_InitSprites: Sprite %s has a missing rotation", spritename);
						}
					}
				}

				break;
			}
		}

		if ((type == PWAD_LIST && !amwdsflags[i]) || customspr)
		{
			sprites[i].sprflags |= SPR_NOMODEL;
		}

		sprites[i].numframes = maxframe;

		// Allocate space for the frames present and copy sprtemp to it.
		sprites[i].spriteframes =
			(spriteframe_t*)W_Malloc(maxframe * sizeof(spriteframe_t), STATICID, 44);

		memcpy (sprites[i].spriteframes, sprtemp, 
			maxframe * sizeof(spriteframe_t));

		sprites[i].trueradius = SetSpriteMaxWidth(i)>>1;	// v220-9

		if (!has_rotate)
			sprites[i].sprflags |= SPR_ALIGNED;
	}

	W_Free(amwdsflags, STATICID, 0);
	W_Free(cs, STATICID, 0);
}

//===========================================================================
// R_GetSpriteInfo
//===========================================================================
void R_GetSpriteInfo(int sprite, int frame, spriteinfo_t *sprinfo, int num)
{
	spritedef_t		*sprdef;
	spriteframe_t	*sprframe;
	spritelump_t	*sprlump;

#ifdef RANGECHECK
	if( (unsigned) sprite >= (unsigned) numsprites)
		Con_Error( "R_GetSpriteInfo [%i]: invalid sprite number %i", num, sprite);
#endif

	sprdef = &sprites[sprite];

#ifdef RANGECHECK
	if((frame & FF_FRAMEMASK) >= sprdef->numframes)
		Con_Error( "R_GetSpriteInfo [%i]: invalid sprite frame %i : %i", num, sprite, frame);
#endif

	sprframe = &sprdef->spriteframes[ (frame & FF_FRAMEMASK) ];
	sprlump = spritelumps + sprframe->lump[0];

	sprinfo->numFrames	= sprdef->numframes;
	sprinfo->splnum		= sprframe->lump[0];
	sprinfo->realLump	= spritelumps[sprinfo->splnum].lump;
	sprinfo->flip		= sprframe->flip[0];
	sprinfo->xoffset	= sprlump->slxoffset;
	sprinfo->yoffset	= sprlump->slyoffset;
	sprinfo->width		= sprlump->width;
	sprinfo->height		= sprlump->height;
}

//===========================================================================
// R_MaxRadius v220-4
//	if (shadow) returns the integer shadow radius
//  else returns the fixed_t max mesh radius or sprite lump width/2
//===========================================================================
int R_MaxRadius(mobj_t *mo, RDBOOL shadow, int from)
{
	modeldef_t		*mf;

	if (mo->flags2 & MF2_DONTDRAW) return 0;

	// v220-16 check if it has any frames
	if (!sprites[mo->sprite].numframes)
	{
		pr_OFinsert("WAD ERROR: ref %i invalid sprite dmednum %i mapnum %i has no frames\n", from, mo->info->doomednum, MAPNUM(mo));
		mo->flags2 |= MF2_DONTDRAW;
		return 0;
	}

	// If models are being used, use the model's radius.
	if (   loadModels
		&& !(sprites[mo->sprite].sprflags & SPR_NOMODEL)
		&& (mf = GetStateModel(mo->state, mo->selector)) )
	{
		if (shadow) return (int)mf->visualradius;

		return mf->xyradius; // NOTE: fixed_t
	}

	// simplify v220-10
#ifdef RANGECHECK
	if( (unsigned) mo->sprite >= (unsigned) numsprites)
		Con_Error( "R_MaxRadius: from %i invalid sprite number %i", from, mo->sprite);
	if ((mo->frame & FF_FRAMEMASK) >= sprites[mo->sprite].numframes)
	{
		Con_Error( "R_MaxRadius: invalid sprite dmednum %i frame %i numframes %i: sprite %i frame&FF_FRAMEMASK %i",
			mo->info->doomednum, mo->frame, sprites[mo->sprite].numframes, mo->sprite, mo->frame&FF_FRAMEMASK);
	}
#endif

	return sprites[mo->sprite].trueradius; // NOTE: fixed_t
}

// v220-9 initialise r_maxmodelz
void R_InitZDist()
{
	if (zdistptr < 0) zdistptr = 0;
	if (zdistptr > MAXZLISTPTR) zdistptr = MAXZLISTPTR;
	r_maxmodelz = zdistlist[zdistptr];
}

//===========================================================================
// R_InitSprites
//===========================================================================
void R_InitSprites(void)
{
	spritelump_t *sl;
	int			 a, b, i, k, m, lump;

	R_InitSpriteDefs();

	// InitSpriteLumps
	for (i = 0, sl = spritelumps; i < numspritelumps; i++, sl++)
	{
		if (lumpinfo[sl->lump].groupMarker == LGT_SPRITES)
		{
			if (lumpinfo[sl->lump].size)
			{
				patch_t *patch = (patch_t*)M_lumpCACHE(sl->lump, NULL, MLC_NOLOCK, MLC_ERR);

				sl->width = patch->width;
				sl->height = patch->height;
				sl->slxoffset = patch->xoffset;
				sl->slyoffset = patch->yoffset;

				if (isarctic == 1)	// arctgfx1 wad
				{
					// Fir tree top PVISA0
					if (!strcmp(lumpinfo[spritelumps[i].lump].name, "PVISA0"))
						sl->slyoffset = 249;
				}

				if (isarctic == 2)	// arctgfx2 wad
				{
					if (   !strcmp(lumpinfo[spritelumps[i].lump].name, "HEADC0")
						|| !strcmp(lumpinfo[spritelumps[i].lump].name, "HDB2A0")
						|| !strcmp(lumpinfo[spritelumps[i].lump].name, "HDB3A0"))
					{
						sl->slyoffset = 58;
					}
					// Fir tree top VILE\0
					if (!strcmp(lumpinfo[spritelumps[i].lump].name, "VILE\\0"))
						sl->slyoffset = 249;
				}			
			}
			else if (!chkLinkedType(lumpinfo[sl->lump].name))
			{
				Con_Error("SPRITE lump %s has no data : type %s group %s",
					lumpinfo[sl->lump].name,
					typedesc[lumpinfo[sl->lump].lumptype]);
			}
		}
	}

	// v220-7 Check if an addon wad has been loaded with models enabled
	if (!loadModels || !(ge_inf.addon_handle > 0)) return;

	// v220-7 if a sprite is defined in the addon wad clear any SPR_NOMODEL flag
	for (i = 0; i < numsprites; i++)
	{
		if (sprites[i].sprflags & SPR_NOMODEL)
		{
			spritedef_t	*sprdef = &sprites[i];

			for (a = b = k = 0; k < sprdef->numframes; k++)
			{
				spriteframe_t *sf = &sprdef->spriteframes[k];

				for (m = 0; m < 8; m++)
				{
					if ( (lump = spritelumps[sf->lump[m]].lump) >= 0)
					{
						a++;

						if (lumpinfo[lump].handle == ge_inf.addon_handle) b++;
					}
				}
			}

			if (a && a == b) sprites[i].sprflags &= ~SPR_NOMODEL;
		}
	}
}	

/*
===============================================================================

							GAME FUNCTIONS

===============================================================================
*/

//===========================================================================
// R_ResetVisSprites b211-0
//	Called at game start.
//===========================================================================
void R_ResetVisSprites(uint num)
{
	visspritecnt = 0;
	maxvissprites = (num/1024) * 1024 + 1024;
	if (maxvissprites > 51200)	// Keep it half sensible
		Con_ErrorD("Max number of vissprites exceeded");
	vispointer   = (vissprite_t**)W_Calloc(maxvissprites * sizeof(vissprite_t* ), LEVELID, 23);
	vissprites   = (vissprite_t* )W_Calloc(maxvissprites * sizeof(vissprite_t  ), LEVELID, 24);
	visseclist   = (visseclist_t*)W_Calloc(numsectors    * sizeof(visseclist_t ), LEVELID, 25);
	vissec_count = (int*)W_Calloc(numsectors * sizeof(int), LEVELID, 26);
	thing_validcount = 1;
}

//===========================================================================
// R_NewVisSprite
//===========================================================================
vissprite_t *R_NewVisSprite(void)
{
	if (visspritecnt >= maxvissprites)
	{
		maxvissprites += 1024;

		if (maxvissprites > 51200)	// Keep it half sensible
			Con_ErrorD("Max number of vissprites exceeded");

		vissprites = (vissprite_t*)W_Realloc(vissprites,
			maxvissprites * sizeof(vissprite_t), LEVELID, 27);

		vispointer = (vissprite_t**)W_Realloc(vispointer,
			maxvissprites * sizeof(vissprite_t*), LEVELID, 28);
	}

	return &vissprites[visspritecnt++];
}

//===========================================================================
// R_ProjectPlayerSprites
//	Create vissprites for psprites having 3D models. 
//===========================================================================
void R_ProjectPlayerSprites(int *psptype)
{
	ddpsprite_t		*psp;
	vissprite_t		*vis;
	modeldef_t		*mf, *nextmf;
	mobj_t			dummy;
//	short			ll;
	int				i;
	
	*psptype = 0;

	if (/*   ppt.viewplayer->flags & DDPF_CAMERA	// Cameramen have no psprites
		||*/ inZoomMode < 0	// Neither models nor sprites rendered when zoom at max
		|| R3D_edit)
	{
		return;
	}

	for (i = 0, psp = ppt.viewplayer->psprites; i < DDMAXPSPRITES; i++, psp++)
	{
		vis = vispsprites + i;
		psp->flags &= ~DDPSPF_RENDERED;

		vis->visflags = VIS_NONE;

		if (!psp->stateptr) continue;

		mf = NULL;

		// NOTE: stateptr is the current state (psp->stateptr - states gives the idx)
		if (CFX)
		{
			// Is there a model for this frame?
			// Setup a dummy for the call to R_CheckModelFor.
			dummy.state = psp->stateptr;
			dummy.tics = psp->tics;
			dummy.selector = 0;

			vis->mo.inter = R_CheckModelFor(&dummy, &mf, &nextmf);
		}

		// If no model frame draw a 2D sprite (in Rend_DrawPlayerSprites).
		if (!mf)
		{
			if (R3D_ScriptedMap)
				modelError(spritelumps[psp->stateptr->sprite].lump, -1);

			if (inZoomMode)
			{
				// sprites not rendered in zoom mode
				*psptype = 0;	// In case it's been set
				return;
			}

			*psptype |= REND_PSPRITE;
			continue;
		}

		// Prevent the sprite being rendered
		psp->flags |= DDPSPF_RENDERED;
	
		// There are model psprites.
		*psptype |= REND_PMODEL;

		vis->visflags = VIS_MODEL|VIS_PSPR;
		vis->distance = 4.0f;
		vis->mo.vmoflags = HAS_FOGUNSET;
		vis->mo.mf = mf;
		vis->mo.nextmf = nextmf;
		vis->mo.flags = 0;
		vis->mo.gx = viewx;
		vis->mo.gy = viewy;
		vis->mo.v1[VX] = fvx;
		vis->mo.v1[VY] = fvz;

		vis->mo.gzt = vis->mo.gz = viewz;

		vis->mo.viewaligned = true; 
		vis->mo.secfloor = -1.0e6f;
		vis->mo.secceil = 1.0e6f;
		vis->mo.pclass = 0;

		// Offsets to rotation angles.
		vis->mo.v2[VX] = psp->x / PI - 90.0f;
		// 32 is the raised weapon height.
		vis->mo.v2[VY] = (32.0f - psp->y) / PI;

		// Real rotation angles.
		vis->mo.yaw = viewangle / (float) ANGMAX * -360.0f 
			+ vis->mo.v2[VX] + 90.0f;
		vis->mo.pitch = viewpitch * RATIO + vis->mo.v2[VY];
		vis->mo.zangle = 0;
		vis->mo.flip = false;
		vis->mo.intensity = 0;
		vis->mo.alpha = psp->alpha;
		memset(vis->mo.visoff, 0, sizeof(vis->mo.visoff));
	}
}

static RDBOOL modelOcclusionTest(mobj_t *thing, modeldef_t *mf, int *tr_orig)
{
	if (thing->flags2 & MF2_NOOCCLUDE) return true;

	if (abs(tr_orig[0]) < mf->xyradius && abs(tr_orig[1]) < mf->xyradius)
	{
		return true;
	}
	else
	{
		fixed_t	ofx = mf->offset[VX] ? (fixed_t)(mf->offset[VX] * FRACUNIT) : 0;
		fixed_t	ofy = mf->offset[VZ] ? (fixed_t)(mf->offset[VZ] * FRACUNIT) : 0;
		fixed_t	box[4];

		box[BOXLEFT ] = thing->x - mf->xyradius + ofx;
		box[BOXRIGHT] = thing->x + mf->xyradius + ofx;
		box[BOXBASE ] = thing->y - mf->xyradius + ofy;
		box[BOXTOP  ] = thing->y + mf->xyradius + ofy;

		return C_CheckBspBox(box);
	}
}

static RDBOOL spriteOcclusionTest(mobj_t *thing, int *tr_orig)
{
	fixed_t radius;

	if (thing->flags2 & MF2_NOOCCLUDE) return true;

	radius = sprites[thing->sprite].trueradius;

	if (abs(tr_orig[0]) < radius && abs(tr_orig[1]) < radius)
	{
		return true;
	}
	else
	{
		fixed_t	box[4];

		box[BOXLEFT ] = thing->x - radius;
		box[BOXRIGHT] = thing->x + radius;
		box[BOXBASE ] = thing->y - radius;
		box[BOXTOP  ] = thing->y + radius;

		return C_CheckBspBox(box);
	}
}

RDBOOL true2DLimit(double x, double y, double d)
{
	return x*x + y*y > d*d;
}

//===========================================================================
// R_ProjectSprite
//	Generates a vissprite for a thing if it might be visible.
//  Returns true if thing->corelink != NULL
//===========================================================================
static RDBOOL R_ProjectSprite (subsector_t *ss, short llevel, mobj_t *thing, mobj_t *linkflame)
{
	spritedef_t			*sprdef;
	vissprite_mobj_t	*vmo;
	vissprite_t			*vis;
	modeldef_t			*mf = NULL, *nextmf = NULL;
	subsector_t			*ssec;
	sector_t			*sect;
	mobj_t				*mo;

	fixed_t			tr_orig[2];
	float			v1[2], v2[2];
	float			thangle;	// rv = real value
	float			interp = -1.0f;
	float			distance;
	RDBOOL			align = false;
	RDBOOL			rend_lod = false;
	RDBOOL			occ_test;
	RDBOOL			flip;
	int				var, splnum, numf;

	//-------------------------------------------

	// Check a vissprite's needed.
	if (   thing->flags2 & (MF2_DONTDRAW|MF2_NORENDER|MF2_CORPSEFADE)
		|| thing->ddflags & DDMF_DONTDRAW)
	{ 
		return false;
	}

	occ_test = !linkflame; // if linkflame then model owning it has already passed the occlusion test

	// modified $v2.0.11
	if (  !(numf = sprites[thing->sprite].numframes)
		|| (thing->frame & FF_FRAMEMASK) >= numf )
	{
#ifdef RANGECHECK
		Con_Message ("INVALID sprite %i dmednum %i thing: tidnum %i flags %.4X [FM %i NF %i]\n",
		thing->sprite, thing->info->doomednum, TIDNUM(thing) ,thing->ddflags, thing->frame & FF_FRAMEMASK, numf);
#endif
		if (thing->sprite == SPR_PLAY)
			Con_ErrorD("TERMINAL ERROR: sprite with state SPR_PLAY is invalid");

		thing->flags2 |= MF2_DONTDRAW;
		return false;
	}

	// Transform the origin point.
	tr_orig[0] = thing->x - viewx;
	tr_orig[1] = thing->y - viewy;

	// Get distance from viewpoint
	v1[VX] = FIX2FLT(thing->x);
	v1[VY] = FIX2FLT(thing->y);
	distance = Rend_PointDist2D(v1);

	sprdef = &sprites[thing->sprite];

	// Check for a model.
	if (CFX && !(sprdef->sprflags & SPR_NOMODEL))
	{
		interp = R_CheckModelFor(linkflame ? linkflame : thing, &mf, &nextmf);

		if (  !mf
			|| mf->flags & MFF_NO_DISTANCE_CHECK
			|| thing->type > MT_MINE
			|| (R3D_ScriptedMap && (linkflame || thing->corelink)) )
		{
			goto end_lodtest;
		}

		// Switch to simpler (lod) model
		if (thing->lodstate > 0 && !(states[thing->lodstate].stflags & ST_NOMODEL))
		{
			// Lods always kick in at a distance of > 1000 (use true 2D distance not the viewplane distance)
			if ( true2DLimit((double)(fvx - v1[VX]), (double)(fvz - v1[VY]), 1000) )
			{
				modeldef_t *mfc;

				if (!modelOcclusionTest(thing, mf, tr_orig)) return false;

				// draw the standard flame with skullpile or firecan lod
				if (linkflame || thing->corelink)
				{
					if (   thing->corelink
						&& (thing->corelink->type == MT_YELLOWFLAMES
						||  thing->corelink->type == MT_SKULLFLAME))
					{
						if (linkflame) goto end_lodtest;
					}
					else
					{
						if (!linkflame) return true;
					}
				}

				occ_test = false;
				// Check if a simpler lod model exits
				if (  (mfc = GetStateModel(&states[thing->lodstate], linkflame ? linkflame->selector : thing->selector))
					&& mfc->sub[0].model)
				{
					interp = 1.0f;
					rend_lod = true;
					mf = mfc;
					nextmf = NULL;
				}
			}
		}
		else if (!R3D_ScriptedMap && distance > r_maxmodelz)
		{
			mf = nextmf = NULL; interp = -1.f;	// Use a sprite
		}
	}

end_lodtest:;

	// v220-24 add sprite occluded test
	if (!mf && !spriteOcclusionTest(thing, tr_orig)) return false;

	thangle = (float)(BANG2RAD(bamsAtan2((int)(FIX2FLT(tr_orig[1])*10),
									(int)(FIX2FLT(tr_orig[0])*10))) - (PI/2));

	if (!mf)
	{
		float				sinrv, cosrv;
		float				xof, wsp;
		spriteframe_t		*sprframe;

		sprframe = &sprdef->spriteframes[ (thing->frame & FF_FRAMEMASK) ];

		if (sprframe->rotate == 1)
		{	
			// Choose a different rotation based on player view.
			angle_t ang = R_PointToAngle (thing->x, thing->y);

			// (ANG45/2)*9 = 0x90000000
			unsigned rot = (ang - thing->angle + (unsigned)0x90000000)>>29;

			splnum = sprframe->lump[rot];
			flip   = (RDBOOL)(sprframe->flip[rot] != 0);
		}
		else
		{	
			// Use single rotation for all views.
			splnum = sprframe->lump[0];
			flip   = (RDBOOL)(sprframe->flip[0] != 0);
		}

		if (splnum < 0)
		{
			if (R3D_ScriptedMap)
				modelError(-1, MAPNUM(thing));	// even if error is normally allowed!

#ifdef RANGECHECK
			Con_Message ("INVALID sprite %i dmednum %i thing: tidnum %i flags %.4X [FM %i NF %i]\n",
			thing->sprite, thing->info->doomednum, TIDNUM(thing) ,thing->ddflags, thing->frame & FF_FRAMEMASK, numf);
#endif

			thing->flags2 |= MF2_DONTDRAW;
			return 0;
		}

		if (R3D_ScriptedMap)
			modelError(spritelumps[splnum].lump, MAPNUM(thing));	// sprites are definately not allowed!

		// Align to the view plane?
		if (thing->ddflags & DDMF_VIEWALIGN)
		{
			align = true;

			// The sprite should be fully aligned to view plane.
			sinrv = -FIX2FLT(viewcos);
			cosrv =  FIX2FLT(viewsin);
		}
		else
		{
			sinrv = (float)sin(thangle);
			cosrv = (float)cos(thangle);
		}
		
		xof = (float)spritelumps[splnum].slxoffset;
		wsp = (float)spritelumps[splnum].width;
		
		v1[VX] -= cosrv * xof; 
		v1[VY] -= sinrv * xof; 
		v2[VX] = v1[VX] + cosrv * wsp;
		v2[VY] = v1[VY] + sinrv * wsp;
	}
	// v220-8/v220-9 modified
	else if (occ_test
	//	&& !(thing->flags3 & MF3_AMODEL)	// v220-9 Now occluded unless flagged otherwise in the ded
		&& !modelOcclusionTest(thing, mf, tr_orig) )
	{
		return false;
	}

	//
	// Create a vissprite and store the information needed to render the Thing
	//
	vis = R_NewVisSprite();

	ssec = thing->subsector;
	sect = ssec->sector;
	
	vis->visflags = mf ? VIS_MODEL : VIS_SPRITE;
	vis->distance = distance;

	// b211-0 Make sure the flame is rendered before the model owning it
	//        Note: this is also true with lod torches. The distance
	//        change is immaterial; it's the suppression of dyn-lights
	//        that's required which setting VIS_FLAME achieves.
	if (linkflame)
	{
		vis->distance -= .01f;
		vis->visflags |= VIS_FLAME;
	}
	else if (mf && thing->flags3 & MF3_OFFSETMASK)
	{
		// b211-0 Check if offsetpos flag is set
		if (thing->flags3 & MF3_OFFSETPOS) vis->distance += 0.2f;
		else
		if (thing->flags3 & MF3_OFFSET5  ) vis->distance += 5.0f;
		else // must be MF3_OFFSET32
			vis->distance += 32.0f;
	}

	mo = thing;

	if (linkflame) thing = linkflame;

	vis->torchtic = thing->torchtic;
	vis->skintic  = thing->skintic;

	vmo = &vis->mo;

	// thing->light will be 0 unless set in DL_AddLuminous
	// The value is the list position + 1 (suppressed if LUMF_NOHALO is set)
//	vmo->light = getLuminous(thing->lightnum); $halo removed

	vmo->mf = mf;
	vmo->nextmf = nextmf;
	vmo->inter = interp;
	vmo->flags = thing->ddflags;
	vmo->think_id = thing->thinker.id;
	vmo->selector = thing->selector;
	vmo->gx = thing->x;
	vmo->gy = thing->y;
	vmo->gz = linkflame ? mo->z : (thing->flags2 & MF2_VDOOR) ?
		sect->ceilingheight : (thing->flags2 & MF2_GLUECEIL) ?
		sect->ceilingheight - thing->height : thing->z;

	vmo->gzt = vmo->gz;
	vmo->height = thing->z + thing->height;	// b211-0
	vmo->vmoflags = 0;
#ifdef _DEBUG
	vmo->thing_num = MAPNUM(thing);
#endif

	// v220-3
	if (whitefog)
	{
		if (ssec->ssflags & SSF_FOGUNSET)
			vmo->vmoflags |= HAS_FOGUNSET;
		else if (thing->type == MT_INS)
			vmo->vmoflags |= HAS_NOFOG;
	}

	if (thing->state - states == S_GIBS
		&& thing->info->radius <= 8*FRACUNIT)
	{
		vmo->vmoflags |= HAS_RADIUS8;
	}
	
	vmo->viewaligned = align;

	if (thing->flags3 & MF3_TRANSLUCENT)
	{
		vis->visflags |= VIS_TRANSSPR;
	}
	else if (thing->type == MT_SHADOWS)
	{
		if (!mf && !(hasDEH && DEHCheck[MT_SHADOWS] & DEHF_THING))
			vis->visflags |= VIS_SHADOW;
	}
	else if (thing->flags3 & MF3_STEALTH)
	{
		vis->visflags |= VIS_STEALTH;	// b211-0 stealth monster
	}
	else if (thing->flags & MF_SHOOTABLE
		&& thing->info->seestate
		&& !(thing->flags3 & (MF3_AMODEL|MF3_STATIC)) )
	{
		vis->visflags |= VIS_SHOOTABLE;	// v220-24 enemies
	}

	if (thing->misc & MJF_SCP_SPLASH)
		vmo->secfloor = FIX2FLT(thing->z);
	else
		vmo->secfloor = FIX2FLT(sect->smoothfloor);
		
	vmo->secceil = FIX2FLT(sect->smoothceil);

	// The start and end vertices.
	vmo->v1[VX] = v1[VX];
	vmo->v1[VY] = v1[VY];
	vmo->splnum = splnum;
	vmo->intensity = 0;

	if (thing->misc & MJF_UNLIT) vis->visflags |= VIS_NOTLIT;

	if (!mf)
	{
		// We must find the correct positioning using the
		// sector floor and ceiling heights as an aid.
		// b211-0 revised ->
		fixed_t	sprh = spritelumps[splnum].height << FRACBITS;
		fixed_t	toph;

		// Translation is restricted to sprites (rarely found in single player maps)
		if (thing->ddflags & DDMF_TRANSLATION)
			vmo->pclass = (thing->ddflags>>DDMF_CLASSTRSHIFT) & 0x3;
		else
			vmo->pclass = 0;

		// v220-13 Ignore litObjects if dehacked
		if (hasDEH && DEHCheck[thing->sprite] & DEHF_SPRITE)
			vis->visflags |= VIS_NOTLIT;

		vmo->gzt += ((fixed_t)spritelumps[splnum].slyoffset << FRACBITS);	// default
		toph = vmo->gzt;

		// Revised v220-13/v220-23/v220-27/v220-29
		if (thing->ddflags & DDMF_FITTOP)	// dehacked Things
		{
			int ht = sect->heightsec >= 0 ?
				MIN_OF(sectors[sect->heightsec].ceilingheight, sect->ceilingheight) : sect->ceilingheight;
			toph = ht + (spritelumps[splnum].height << FRACUNIT);
		}
		else if ( !(thing->flags2 & MF2_NOFITBOTTOM)
			&& toph - sprh < sect->floorheight)
		{
			toph = sect->floorheight + sprh;
		}

		vmo->gzt = toph;
		vmo->top = FIX2FLT(toph);
		vmo->v2[VX] = v2[VX];
		vmo->v2[VY] = v2[VY];
		vmo->flip   = flip;
	}
	else
	{
		if (mf->flags & MFF_DISABLE_Z_WRITE)
			vmo->vmoflags |= HAS_NOZ;

		if (thing->flags3 & MF3_NODYNLIGHT)
		{
			vmo->vmoflags |= HAS_NOLIGHT;
		}
		else if (thing->flags2 & MF2_EVENLIGHT)
		{
			vmo->vmoflags |= HAS_EVENLIGHT;
		}
		else if (thing->misc & MJMISC_LIT3D)
		{
			vmo->vmoflags |= HAS_3DLIGHT;	// v220-35
		}
		else if ((ge_inf.useSitters || thing->sprite == SPR_TFOG)
			&& (thing->info->seestate || thing->player || thing->flags & MF_CORPSE || thing->flags2 & (MF2_CORPSEFADE|MF2_TELEPCORPSE)) )
		{
			vmo->vmoflags |= HAS_EVENLIGHT;
		}
		else
		{
			vmo->vmoflags |= HAS_2DLIGHT;
		}

		// v220-6 don't use litskins with melee attacks
		if ( !(thing->flags2 & MF2_MELEE) )
		{
			// v220-4 set flag if the model Thing is attacking
			if (thing->state->stflags & ST_LITSKIN)
				vis->visflags |= VIS_LITSKIN;
			else
			if (thing->target && CheckAction(thing->state))
					vis->visflags |= VIS_ATKLIGHT;
		}

		if (R3D_ScriptedMap)
		{
			if (thing->scpzoff_ptr > 0)
			{	
				vmo->intensity = scpzoff[thing->scpzoff_ptr - 1].intensity;
			}
			else if (thing->flags3 & MF3_AMODEL && amodelintensity)
			{
				int thingnum = TIDNUM(thing) - 7000;

				if (thingnum >= 0 && thingnum < 700)
					vmo->intensity = amodelintensity[thingnum];
			}
		}

		vmo->pitch = 0;

		if (thing->flags3 & MF3_NOROTATE)
		{
			vmo->yaw = (thing->visangle << 16) / (float) ANGMAX * -360;
		}
		else
		{
			int	spin_flags;

			// Determine the rotation angles (in degrees).
			if (mf->sub[0].flags & MFF_ALIGN_YAW)
			{
				vmo->yaw = 90.0f - thangle / (float)PI * 180.0f;
			}
			else if ( (spin_flags = mf->sub[0].flags & MODEL_SPINMASK) )
			{
				// v220-5 double, half and quarter spin speed support
				float spin = ma_inf.global_yaw;

				if (spin_flags & MFF_HALFSPIN  ) spin *= 2.0f;
				if (spin_flags & MFF_SPIN      ) spin *= 4.0f;
				if (spin_flags & MFF_DOUBLESPIN) spin *= 8.0f;

				vmo->yaw = spin + (float)thing->startangle;
			}
			else if (mf->sub[0].flags & MFF_MOVEMENT_YAW)
			{
				vmo->yaw = BANG2DEG(bamsAtan2((int)(FIX2FLT(thing->momy) * -100.0f), 
					(int)(FIX2FLT(thing->momx) * 100.0f)));
			}
			else
			{
				vmo->yaw = (thing->visangle << 16) / (float) ANGMAX * -360.0f;
			}

			if (mf->sub[0].flags & MFF_ALIGN_PITCH)
			{
				float zd = (FIX2FLT(vmo->gz) + FIX2FLT(vmo->gzt)) * 5.0f;

				vmo->pitch = -BANG2DEG(bamsAtan2((int)(zd - fvy * 10.0f), (int)(distance * 10.0f)));			
			}
			else if (mf->sub[0].flags & MFF_MOVEMENT_PITCH)
			{
				vmo->pitch = BANG2DEG(bamsAtan2((int)(FIX2FLT(thing->momz) * 100.0f), 
					(int)(M_AccurateDistance(thing->momx, thing->momy) * 100.0f)));
			}
			else if (!rend_lod && thing->misc & MJF_FLAMEWALL)
			{
				// v220-5 adjust flame vertically
				binangle_t angle = bamsAtan2((viewy - thing->y) >> FRACBITS, 
									  (viewx - thing->x) >> FRACBITS) >> (ANGLETOFINESHIFT - FRACBITS);

				vmo->pitch = -20.0f;

				// Position correction for when pitch is applied
				if (mo->type >= MT_MISC41 && mo->type <= MT_MISC43)
				{
					vmo->v1[VX] -= FIX2FLT(FixedMul(0xF0000,finecosine[angle]));
					vmo->v1[VY] -= FIX2FLT(FixedMul(0xF0000,finesine  [angle]));
					vmo->gz += 0x3A000;
				}
				else
				{
					vmo->v1[VX] -= FIX2FLT(FixedMul(0x70000,finecosine[angle]));
					vmo->v1[VY] -= FIX2FLT(FixedMul(0x70000,finesine  [angle]));
					vmo->gz += 0x1C000;
				}
			}
		}

		// v220-4 if a model splash then mark it
		if (rv.waterplane != DDMININT && thing->type == MT_SPLASHBASE)
			vis->visflags |= VIS_SPLASH;

		if (thing->flags3 & MF3_NOGLIGHT) vis->visflags |= VIS_NOGLIGHT;
	}

	// v220-9 $translucent_water
	// v220-14 revised
	if (   !underwater	// The player that is (so must be looking down on this thing's water surface)
		&& ssec->ssflags & SSF_FAKETRANS
		&& (var = ssec->s_waterplane) > DDMININT
		&& !(vmo->vmoflags & HAS_NOZ)
		&& !(mf && thing->type == MT_SPLASHBASE))
	{
		// v220-5 use new zmin and zmax values
		fixed_t zlo = mf ? thing->z + mf->zoff + mf->zmin : thing->z;
		fixed_t	zhi = mf ? zlo + mf->zmax : vmo->height;

		if (zlo < var)
		{
			if (zhi >var)
				vmo->vmoflags |= HAS_WATERPLANE; // split by waterplane
			else
				vmo->vmoflags |= HAS_UNDERWATER; // totally underwater
		}
	}

	// R3D - MF2_CANSLOPE identifies a thing that should be orientated with the slope.
	// (e.g. deathstate or item that can be picked up etc.)
	if (mf && thing->flags2 & MF2_CANSLOPE && sect->slopeangles[0])
	{
		double yang = (vmo->yaw - 90) / 180 * PI;

		vmo->pitch +=  sect->slopeangles[0]*(float)cos(sect->slopeangles[1] + yang);
		vmo->zangle = -sect->slopeangles[0]*(float)sin(sect->slopeangles[1] + yang);
	}
	else
		vmo->zangle = 0;

	if (thing->rotateperiod)
	{
		uint  t = Sys_GetRealTime();
		vmo->rotangle = 360.0f * (t - thing->rotateprev)/thing->rotateperiod;

		if (vmo->rotangle >= 360.0f) thing->rotateprev = t;
	}
	else
		vmo->rotangle = 0;

	// Get light level (with linkflame use owner Thing)
	// v220a revised. Since things can be assigned to an overlapping sector
	//       then the pre-determined light level may need to be modified.
	if ((var = sect->heightsec) >= 0 || ssec->sector != ss->sector)
	{
		short	usell;

		if (ssec->floorsec >= 0)
			usell = sectors[ssec->floorsec].lightlevel;
		else if (ssec->lfsec >= 0)
			usell = sectors[ssec->lfsec].lightlevel;
		// v220-29 add adjustment for heightsec
		else if (var >= 0 && thing->z + thing->height < sectors[var].floorheight)
			usell = sectors[var].lightlevel;
		else
			usell = sect->lightlevel;

		llevel = R3D_ScriptedMap ? usell : SectorDoomLight(usell);
	}

	if (LevelFullBright || mo->frame & FF_FULLBRIGHT)
	{
		if (mf && !LevelFullBright && mo->flags & MF_SHOOTABLE)
		{
			vmo->lightlevel = -llevel;
		}
		else
			vmo->lightlevel = -1;
	}
	else
	{
		int ht = mf ? mf->zmax - mf->zmin : thing->height;

		// v220-9 Scripted light.
		if (thing->scplightlevel)
		{
			vmo->lightlevel = thing->scplightlevel;
		}
		// v220-5 If in a fake sector and under it's floorheight
		//        use the fake sector's light level
		else if (sect->heightsec >= 0
			&& thing->z + ht <= sectors[sect->heightsec].floorheight)
		{
			vmo->lightlevel = R3D_ScriptedMap ? sectors[sect->heightsec].lightlevel : SectorDoomLight(sectors[sect->heightsec].lightlevel);
		}
		else
		{
			vmo->lightlevel = llevel;
		}
	}

	// Set custom alpha
	// R3D Make sprite fog less opaque
	if (!mf && thing->type == MT_TFOG)
	{
		vmo->alpha = 0.5f;
	}
	// v220-3 why was this being shifted twice (first by DDMOBJ_SELECTOR_SHIFT and then by 5)?
	else if (thing->selector & R3DMOBJ_ALPHA_CHECK)
	{
		// The three highest bits of the selector are used for an alpha level.
		vmo->alpha = 1.0f - (float)((thing->selector >> R3DMOBJ_ALPHA_SHIFT) & 0x7) / 8.0f;
	}
	else
	{
		vmo->alpha = -1.0f;
	}

	// v220-9 map development aid
	if (ge_inf.devmap && (rend_lod || !linkflame))
	{
		if (mf)
		{
			if (rend_lod)
				th_items[TH_LOD]++;
			else
			if (mo->flags3 & MF3_AMODEL)
			{
				th_items[TH_AMODEL]++;
			}
			else
				th_items[TH_MODEL]++;
		}
		else
		{
			th_items[TH_SPRITE]++;
		}

		if (mo->flags & MF_SHOOTABLE)
		{
			th_items[TH_SHOOTABLE]++;
		}
		else
		{
			switch(mo->type)
			{
			case MT_MISC0:	// Green armour
			case MT_MISC1:	// Blue armour
			case MT_MISC10:	// Stim pack
				th_items[TH_PICKUPS]++;
				break;
			case MT_MISC4:	// Blue card key
			case MT_MISC5:	// Red card key
			case MT_MISC6:	// Yellow card key
			case MT_MISC7:	// Blue skull key
			case MT_MISC8:	// Red skull key
			case MT_MISC9:	// Yellow skull key
				th_items[TH_KEYS]++;
				break;
			case MT_BARREL:
			case MT_MISC29:	// Techlamp
			case MT_MISC30:	// Techlamp2
			case MT_MISC31:	// Column
			case MT_MISC32:	// Tall green column
			case MT_MISC33:	// Short green column
			case MT_MISC34:	// Tall red column
			case MT_MISC35:	// Short red column
			case MT_MISC36:	// Skull column
			case MT_MISC37:	// Heart column
			case MT_MISC38:	// Evil eye
			case MT_MISC39:	// Floating skull
			case MT_MISC48:	// Techpillar
			case MT_MISC77:	// Firecan
				th_items[TH_DECORATIONS]++;
				break;
			case MT_MISC40:	// Grey or torch tree
			case MT_MISC47:	// Stalactite
			case MT_MISC76:	// Big tree
			case MT_STALAG:	// Stalactite2 (grey)
				th_items[TH_TREES]++;
				break;
			case MT_MISC41:	// Blue torch
			case MT_MISC42:	// Green torch
			case MT_MISC43:	// Red torch
			case MT_MISC44:	// Short blue torch
			case MT_MISC45:	// Short green torch
			case MT_MISC46:	// Short blue torch
			case MT_MISC49:	// Candlestick
			case MT_MISC50:	// Candelabra
				th_items[TH_TORCHES]++;
				break;
			case MT_MISC51:	// Bloody twitch
			case MT_MISC52:	// Hanging1 arms out
			case MT_MISC53:	// One legged1 body
			case MT_MISC54:	// Hanging1 torso
			case MT_MISC55:	// Hanging2 torso
			case MT_MISC56:	// Hanging7 arms out
			case MT_MISC57:	// Hanging3 torso
			case MT_MISC58:	// One legged2 body
			case MT_MISC59:	// Hanging4 torso
			case MT_MISC60:	// Swaying2 body
			case MT_MISC78:	// Hanging body full2
			case MT_MISC79:	// Hanging body damaged 2
			case MT_MISC80:	// Hanging body head 2
			case MT_MISC81:	// Hanging body vertebra1
			case MT_MISC82:	// Hanging body vertebra2
			case MT_MISC83:	// Hanging body no vertebra
				th_items[TH_GORE]++;
				break;
			case MT_MISC61:	// Dead cacodemo
			case MT_MISC62:	// Dead player (green)
			case MT_MISC63:	// Dead trooper
			case MT_MISC64:	// Dead demon
				// MT_MISC65 Dead Lost Soul ignored
			case MT_MISC66:	// Dead imp
			case MT_MISC67:	// Dead sergeant
			case MT_MISC68:	// Guts and bones
			case MT_MISC69:	// Guts and bones2
			case MT_MISC70:	// Heads on stick
			case MT_MISC71:	// Pool of blood
			case MT_MISC72:	// Pole with skull
			case MT_MISC73:	// Pile of skulls
			case MT_MISC74:	// Impaled body
			case MT_MISC84:	// Pool of blood2
			case MT_MISC85:	// Small pool of blood
			case MT_MISC86:	// Brain and spine
				th_items[TH_DEAD]++;
				break;
			case MT_MISC2:	// Health bottle
			case MT_MISC3:	// Armour helmet
				th_items[TH_BON]++;
				break;
			case MT_MISC11:	// Medical pack
			case MT_MISC13:	// Berserk pack
			case MT_MISC24:	// Backpack
				th_items[TH_PACKS]++;
				break;
			case MT_MISC12:	// Soul sphere
			case MT_INV:	// Invulnerability
			case MT_INS:	// Blur sphere
			case MT_MEGA:	// Megasphere
				th_items[TH_SPHERES]++;
				break;
			case MT_MISC14:	// Radiation suit
			case MT_MISC15:	// Computer map
			case MT_MISC16:	// Light goggles
				th_items[TH_EQUIPMENT]++;
				break;
			case MT_CLIP:	// Bullets
			case MT_MISC17:	// Box of bullets
			case MT_MISC18:	// Rocket
			case MT_MISC19:	// Box of rockets
			case MT_MISC20:	// Energy cell
			case MT_MISC21:	// Energy pack
			case MT_MISC22:	// Shells
			case MT_MISC23:	// Box of shells
				th_items[TH_AMMO]++;
				break;
			case MT_MISC25:	// BFG
			case MT_CHAINGUN:
			case MT_MISC26:	// Chainsaw
			case MT_MISC27:	// Rocket launcher
			case MT_MISC28:	// Plasma gun
			case MT_SHOTGUN:
			case MT_SUPERSHOTGUN:
				th_items[TH_WEAPONS]++;
				break;
			}
		}
	}

	// Reset the visual offset.
	memset(vmo->visoff, 0, sizeof(vmo->visoff));

	// v220-5/v220-28 revised servo-smoothing
	if (!mo->modelboxnum && mo->state && mo->state->tics)
	{
		mposv_t *mpt = &mpos.mp[mo - mobjarray];

		if (   mo->state->tics < 0
			|| (mo->flags & MF_MISSILE)
			|| (!mo->dplayer && !(mo->flags2 & MF2_PUSHABLE) && !mo->info->speed) )
		{
			if (mpt->ox != mo->x) vmo->visoff[VX] = FIX2FLT((fixed_t)(mpt->ox*mpos.mt + mo->x*mpos.tm));
			if (mpt->oy != mo->y) vmo->visoff[VY] = FIX2FLT((fixed_t)(mpt->oy*mpos.mt + mo->y*mpos.tm));
			if (mpt->oz != mo->z) vmo->visoff[VZ] = FIX2FLT((fixed_t)(mpt->oz*mpos.mt + mo->z*mpos.tm));
		}
		else if (mo->tics > 0)	// note: mo->state->tics > 0 must also be true
		{
			float	mul = (float)mo->tics / (float)mo->state->tics;

			if (mo->srvo[VX]) vmo->visoff[VX] = FIX2FLT(mo->srvo[VX]<<8) * mul;
			if (mo->srvo[VY]) vmo->visoff[VY] = FIX2FLT(mo->srvo[VY]<<8) * mul;
			if (!mo->momz && mo->misc & MJF_ONLIFT)
			{
				if (mpt->oz != mo->z) vmo->visoff[VZ] = FIX2FLT((fixed_t)(mpt->oz*mpos.mt + mo->z*mpos.tm));
			}
			else
			{
				if (mo->srvo[VZ]) vmo->visoff[VZ] = FIX2FLT(mo->srvo[VZ]<<8) * mul;
			}
		}
	}

	if (mo->misc & MJMISC_NEGOFF)
		vmo->visoff[VZ] -= (mo->misc & MJMISC_NEGOFF) >> NEGOFF_SHIFT;

	// Glowing floor and ceiling.
	memset(vmo->floorglow, 0, sizeof(vmo->floorglow));
	memset(vmo->ceilglow , 0, sizeof(vmo->ceilglow ));

	if (!mf && useWallGlow && !(vmo->vmoflags & (HAS_WATERPLANE|HAS_UNDERWATER)) )
	{
		if (R_CheckFlatGlowFlag(sect->ceilingpic))
		{
			GL_GetFlatColor(sect->ceilingpic, vmo->ceilglow);
			for (int i = 0; i < 3; i++)
				vmo->ceilglow[i] = (byte)(vmo->ceilglow[i] * dlFactor);
			vmo->vmoflags |= HAS_CEILGLOW;
		}

		if (R_CheckFlatGlowFlag(sect->floorpic))
		{
			GL_GetFlatColor(sect->floorpic, vmo->floorglow);
			for (int i = 0; i < 3; i++)
				vmo->floorglow[i] = (byte)(vmo->floorglow[i] * dlFactor);
			vmo->vmoflags |= HAS_FLOORGLOW;
		}
	}

	return thing->corelink != NULL;
}

//===========================================================================
//  R_SetMobjSectorLists - b211-0
//   This helps to ensure that a sprite that should be visible is projected
//   Modified v220-5
//===========================================================================

static int PIT_ProjectList(sector_t *sec, mobj_t *mo)
{
	int				secn = sec - sectors;
	visseclist_t	*vsl = &visseclist[secn];

	if (vissec_count[secn] >= vsl->maxnum)
	{
		vsl->maxnum += 16;
		vsl->slist = (mobj_t**)W_Realloc(vsl->slist, vsl->maxnum * sizeof(mobj_t*), LEVELID, 29);
	}

	vsl->slist[vissec_count[secn]++] = mo;
	return true;
}

void R_SetMobjSectorLists()
{
	thinker_t	*th;
	sector_t	*sc;

	//-----------------------------------

	memset(vissec_count , 0, numsectors * sizeof(int));

	for (th = thinkercap.next; th != &thinkercap; th = th->next)
	{
		if (th->function == (void (__cdecl *)(void))P_MobjThinker)
		{
			mobj_t	*mo = (mobj_t *) th;

			if (   mo->flags & MF_NOSECTOR		// Only sector linked Things are drawn
				|| mo->flags2 & (MF2_DONTDRAW|MF2_NORENDER|MF2_CORPSEFADE)
				|| !M_GroupVis(mo->subsector))
			{
				continue;
			}
			
			sc = mo->subsector->sector;

			// v220-8 test moved from rend_main.c
			if (   sc->ceilingpic == skyflatnum
				&& sc->specFlags & SF_MOVFLOOR
				&& sc->smoothceil - sc->smoothfloor <= FRACUNIT)
			{
				// R3D - don't draw things in heightless mov floor sky sectors
				continue;
			}

			// flag set in ded or in source to indicate whether thing
			// should have all sectors spanned put in its list
			if (mo->flags2 & MF2_UNCLIPPED)
			{
				P_ThingSectorsIterator(mo, (int (__cdecl *)(sector_t*, void*))PIT_ProjectList, mo);
			}
			else
			{
				PIT_ProjectList(mo->subsector->sector, mo);
			}
		}
	}
}

//===========================================================================
//  R_AddSprites - Modifed for R3D
//===========================================================================
void R_AddSprites (subsector_t *ss, fixed_t fakefloor, short llevel)
{
	mobj_t		*thing;
	int			i;
	int			idx = ss->sector - sectors;

	for (i = 0; i < vissec_count[idx]; i++)
	{
		thing = visseclist[idx].slist[i];

		if (!thing)
		{
#ifdef _DEBUG
		Con_Error("Unset THING on visseclist slist");
#else
		continue;
#endif
		}

		if (thing->mjvalidcount != thing_validcount)
		{
			if (thing->subsector != ss && !(thing->flags2 & MF2_UNCLIPPED)) continue;

			thing->mjvalidcount = thing_validcount;		// b211-0

			// v220-31 add occlusion test
			if (fakefloor > DDMININT && ppt.view_subsector->sector == ss->sector)	// set to DDMININT if translucent
			{
				if (   (viewz >= fakefloor && thing->z + thing->height < fakefloor)
					|| (viewz <  fakefloor && thing->z >= fakefloor))
				{
					continue;
				}
			}

			if (   R_ProjectSprite (ss, llevel, thing, NULL)
				&& thing->corelink->thinker.function == (void (__cdecl *)(void))P_MobjLinkFlame)
			{
				R_ProjectSprite(ss, llevel, thing, thing->corelink);
			}
		}
	}
}

//===========================================================================
// R_SortVisSprites - b211-0 rewritten
//   It was found, where a large number of vissprites per frame are drawn,
//   that replacing the original R_SortSprite routine with the optimised
//   MSVC qsort routine speeded things up considerably.
//   Coincident sprites are also checked to help prevent texture "fighting"
//===========================================================================

static void setOffsets(vissprite_mobj_t *mx)
{
	mx->vmoflags |= HAS_OFFSET;
	mx->v1[VX] += FIX2FLT(FixedMul(0x20000, viewcos));
	mx->v1[VY] += FIX2FLT(FixedMul(0x20000, viewsin));
}

static int __cdecl visdistSORT(const void *e1, const void *e2)
{
	vissprite_t *m1 = (*(vissprite_t**) e1);
	vissprite_t *m2 = (*(vissprite_t**) e2);
	float diff = m1->distance - m2->distance;

	if (diff > 0) return -1;
	if (diff    ) return  1;

	if (m1->visflags == VIS_SPRITE && m2->visflags == VIS_SPRITE)
	{
		if (m1->mo.vmoflags & HAS_OFFSET) return -1;
		if (m2->mo.vmoflags & HAS_OFFSET) return  1;

		if (m1->mo.height > m2->mo.height)
		{
			setOffsets(&m1->mo); return 1;
		}
		else
		{
			setOffsets(&m2->mo); return -1;
		}
	}

	return 0;
}

void R_SortVisSprites(void)
{
	uint	i;

	for (i = 0; i < visspritecnt; i++) vispointer[i] = &vissprites[i];

	if (visspritecnt > 1)
		qsort((void *)vispointer, visspritecnt, sizeof(vissprite_t*), visdistSORT);
}

