//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2013 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: r_world.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified/extended for R3D
//  Made gl nodes dependent, revised vtx owner code
//  Removed polyobj code
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 1 14/Jun/2008 gmj (b220-1)
//  Div3d (was used in p_sight.c) removed
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//  R_SetupTRANSLL() removed
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:	Polygonise, set up nodepile, vertex owners etc.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_network.h"
#include "de_play.h"
#include "de_refresh.h"
#include "de_render.h"
#include "de_graphics.h"
#include "de_audio.h"
#include "de_misc.h"
#include "doom\p_local.h"

#include "r3d_MAINspec.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void P_CreateBlockmap(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL	resetNextViewer;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

vertexowner_t	*vertexowners;
nodeindex_t		*linelinks;			// indices to roots
byte			*vtoflags;	//b211-0
ushort			*vtocount;	//b211-0

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//==========================================================================
// init_VertexOwners
//	Generates an array of sector references for each vertex. The list 
//	includes all the sectors the vertex belongs to.
//
//  Rewritten to use single list instead of a separate allocation per
//  vtx. This should be more efficient and makes release simple - R3D
//==========================================================================

static void checkMaxPolySides(void)
{
	int c, i, num;
	int	chk = MAXPOLYSIDES/3 - 4;	// up to 3 textures/seg + margin
	
	for (c = i = 0; i < numsubsectors; i++)
	{
		if ( (num = subsectors[i].linecount) >= chk)
		{
			Con_Error("Error numverts %i subsec %i sec %i exceeds max allowed %i",
				num, i, subsectors[i].sector - sectors, chk);
		}

		subsectors[i].numverts = (short)num;
		if (num > c) c = num;
	}

	ma_inf.max_polynum = c;
}

static int *vtxlist;
static int vtxlistptr;

typedef struct {
	int list[MAX_SEC_PER_VTX];
} tmpVtxList_t;

static void subfn_SetOwnerNum(int idx, int secnum, tmpVtxList_t *templist)
{
	vertexowner_t *own = vertexowners + idx;
	int i;

	for (i = 0; i < own->num; i++)		// Check not in temp list
	{
		if (templist->list[i] == secnum) return;
	}

	if (own->num >= MAX_SEC_PER_VTX)
	{
		Con_Error("Program error:"
				  " array limit of %i sectors per vtx breached with vtx %i",
				  MAX_SEC_PER_VTX, idx);
	}

	templist->list[own->num++] = secnum;	// Add to temp list
	vtxlistptr++;

	if (own->num > ma_inf.max_ownersec[0])
	{
		ma_inf.max_ownersec[0] = own->num;
		ma_inf.max_ownersec[1] = idx;
	}
}

static int __cdecl vtxlistSorter(const void *n1, const void *n2)
{
	return *(int*)n1 > *(int*)n2 ? 1 : -1;
}

static void init_VertexOwners()
{
	tmpVtxList_t *templist;
	line_t		 *ld;
	int			 i, vertexnumcheck;

	//-------------------------------------

	vertexowners =
		(vertexowner_t*) W_Calloc(numvertexes * sizeof(vertexowner_t), LEVELID, 30);

//PASS 1) Initialise the templist array.

	templist = (tmpVtxList_t*)W_Malloc(numvertexes * sizeof(tmpVtxList_t), STATICID, 45);

	memset(templist, -1, numvertexes * sizeof(tmpVtxList_t));

	vtxlistptr = 0;

	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		int snum1 = ld->frontsector - sectors, snum2;
		int vnum1 = ld->v1 - vertexes;
		int vnum2 = ld->v2 - vertexes;

		subfn_SetOwnerNum(vnum1, snum1, &templist[vnum1]);
		subfn_SetOwnerNum(vnum2, snum1, &templist[vnum2]);

		if (ld->backsector)
		{
			if ( (snum2 = ld->backsector - sectors) != snum1)
			{
				subfn_SetOwnerNum(vnum1, snum2, &templist[vnum1]);
				subfn_SetOwnerNum(vnum2, snum2, &templist[vnum2]);
			}
		}
		else
		{
			vtoflags[vnum1] |= 2;
			vtoflags[vnum2] |= 2;
		}
	}

//PASS 2) Set up the own->list ptr into the list and the sector numbers in vtxlist

	vtxlist = (int*)W_Calloc(vtxlistptr * sizeof(int), LEVELID, 31);

	vertexnumcheck = vtxlistptr;
	vtxlistptr = 0;
	
	for (i = 0; i < numvertexes; i++)
	{
		int k, n = vertexowners[i].num;

		if (!n) continue;	// GL_VERT

		// order the temp list sectors in the list (to aid analysis)
		if (n > 1) qsort(templist[i].list, n, sizeof(int), vtxlistSorter);

		// Set owner's pointer position in list
		vertexowners[i].list = &vtxlist[vtxlistptr];

		// Transfer owner's sector numbers to the vtxlist
		for (k = 0; k < n; k++)
			vtxlist[vtxlistptr++] = templist[i].list[k];
	}

	W_Free(templist, STATICID, 514);

	if (vertexnumcheck != vtxlistptr)
		Con_ErrorD("Program error: vtx list count mismatch");

	// b211-0 Set the line's flags
	for (i = 0, ld = lines; i < numlines; i++, ld++)
	{
		int vnum1 = ld->v1 - vertexes;
		int vnum2 = ld->v2 - vertexes;

		if (ld->ml_flags & ML_TWOSIDED)
		{
			if (vertexowners[vnum1].num == 1 && vertexowners[vnum2].num == 1)
				ld->lnflags2 |= LNF2_OWNEQUALTOONE;

			if (vertexowners[vnum1].num > 2 && vertexowners[vnum2].num  > 2)
				ld->lnflags2 |= LNF2_OWNBOTHGTRTWO;
		}

		if (vertexowners[vnum1].num > 1 && vertexowners[vnum2].num > 1)
			ld->lnflags2 |= LNF2_OWNBOTHGTRONE;

		if (vertexowners[vnum1].num == 2 && vertexowners[vnum2].num == 2)
			ld->lnflags2 |= LNF2_OWNEQUALTOTWO;
		else
		if (vertexowners[vnum1].num > 2 || vertexowners[vnum2].num > 2)
			ld->lnflags2 |= LNF2_OWNGTRTHANTWO;

		if (   (vertexowners[vnum1].num == 2 && vertexowners[vnum2].num  > 2)
			|| (vertexowners[vnum1].num  > 2 && vertexowners[vnum2].num == 2))
		{
			ld->lnflags2 |= LNF2_OWNTWOANDPLUS;
		}
	}
}

// v220-22 SSF_OTHERSEC flag now redundant. This just sets SSF_REFONELINE
//         which is used in the automap with reveal 7
static void flagOneLineSSECS()
{
	subsector_t *ss;
	seg_t		*sg;
	int			i, k;

	// added v220-22
	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		line_t	*lc = NULL;
		int		num = ss->numverts;
		
		for (k = 0, sg = &segs[ss->firstline]; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef)
			{
				if      (!lc) lc = sg->linedef;
				else if (lc != sg->linedef) goto next_test;
			}
		}

		ss->ssflags |= SSF_REFONELINE;

next_test:;
	}
}

static void set_Subsectors()
{
	subsector_t *ss = subsectors;
	uint		size;
	seg_t		*sglist;
	int			segcount, c, i, k, mc;
	int			max_sseclinecount;

	//		a) set the ssec vtxs, bounding box and midpoint
	//		b) remove minisegs

	for (segcount = i = 0; i < numsubsectors; i++, ss++)
	{
		zvertex_t	*vtx;
		fixed_t		*box = ss->ssbox;
		seg_t		*sg = &segs[ss->firstline];
		int			num = ss->numverts;

		ss->verts = (zvertex_t*)W_Calloc(sizeof(zvertex_t) * num, LEVELID, 32);

		for (k = 0, vtx = ss->verts; k < num; k++, sg++, vtx++)
		{
			vtx->x = sg->fv1[VX];
			vtx->y = sg->fv1[VY];
			vtx->zptr = 0;

			if (!k)
			{
				ss->midpoint.x = vtx->x;
				ss->midpoint.y = vtx->y;
				box[BOXLEFT] = box[BOXRIGHT] = sg->v1->x;
				box[BOXBASE] = box[BOXTOP  ] = sg->v1->y;
			}
			else
			{
				ss->midpoint.x += vtx->x;
				ss->midpoint.y += vtx->y;
				if (sg->v1->x < box[BOXLEFT ]) box[BOXLEFT ] = sg->v1->x;
				if (sg->v1->y < box[BOXBASE ]) box[BOXBASE ] = sg->v1->y;
				if (sg->v1->x > box[BOXRIGHT]) box[BOXRIGHT] = sg->v1->x;
				if (sg->v1->y > box[BOXTOP  ]) box[BOXTOP  ] = sg->v1->y;
			}

			if (sg->linedef) segcount++;	// Count number of standard segs
		}

		ss->midpoint.x /= num;
		ss->midpoint.y /= num;
		ss->midpoint.zptr = 0;
	}

	// b211-0 minisegs are no longer required. As such they are now removed.
	// First check for subsecs that only reference one line
	flagOneLineSSECS();

	size = segcount*sizeof(seg_t);
	sglist = (seg_t*)M_Malloc(size, 0);
	max_sseclinecount = 0;

	for (i = mc = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		seg_t	*sg = &segs[ss->firstline];
		
		for (c = k = 0; k < ss->linecount; k++, sg++)
		{
			if (sg->linedef)
				memcpy(&sglist[ (mc + c++) ], sg, sizeof(seg_t));
		}

		ss->firstline = mc; ss->linecount = c; mc += c;
		if (c > max_sseclinecount) max_sseclinecount = c;
	}

	if (mc != segcount)	Con_Error("Program error: segcount [%i %i]", segcount, mc);
	if (ma_inf.minisegs != numsegs - segcount)	Con_Error("Program error: mini segcount [%i %i]", numsegs - segcount, ma_inf.minisegs);

	// Reset the segs list (will free up some space)
	segs = (seg_t*)W_Realloc(segs, size, LEVELID, 33);
	memcpy(segs, sglist, size);
	numsegs = segcount;

	M_Free(sglist);
}

static float TriangleArea(zvertex_t *o, zvertex_t *s, zvertex_t *t)
{
	fvertex_t	a, b;
	float		area;

	a.x = s->x - o->x, a.y = s->y - o->y;
	b.x = t->x - o->x, b.y = t->y - o->y;

	area = (a.x*b.y - b.x*a.y) / 2.0f;

	return area < 0 ? -area : area;
}

static int R_TestTriFan(subsector_t *ss, int base, int num)
{
#define TRIFAN_LIMIT	0.1f

	zvertex_t	*vtx;
	int			i;

	if (num == 3) return true;	// They're all valid.

	// Higher vertex counts need checking.
	for (i = 0, vtx = ss->verts; i < num - 2; i++)
	{
		int a = base + 1 + i;
		int b = a + 1;

		if (a >= num) a -= num;
		if (b >= num) b -= num;

		if (TriangleArea(&vtx[base], &vtx[a], &vtx[b]) <= TRIFAN_LIMIT)
		{
			return false;
		}
	}
	// Whole triangle fan checked out OK, must be good.
	return true;
}

//===========================================================================
// R_SubsectorPlanes
//===========================================================================
static void R_SubsectorPlanes(void)
{
//	zvertex_t	buf[MAXPOLYSIDES];
	subsector_t	*ss;
	int			i, k, size = sizeof(zvertex_t);
	
	for (i = 0, ss = subsectors; i < numsubsectors; i++, ss++)
	{
		zvertex_t	*vtx  = ss->verts;
		int			num   = ss->numverts;
		RDBOOL		valid = false;

		// Test until finding a good tri-fan base vertex
		// that doesn't generate zero-area triangles.
		for (k = 0; k < num; k++)
		{
			if (R_TestTriFan(ss, k, num))
			{
				// v220-22 Is this necessary when using gl nodes?
				// This code was previously used with normal nodes
				// that had been split by an internal routine.
			/*	if(k)	// If k is not the first vertex change the order
				{
					// Copy existing to the buffer
					memcpy(buf, vtx, size * num);

					// Change the order so that k comes first.
					memcpy(vtx, &buf[k], size * (num - k));
					memcpy(&vtx[num-k], buf, size * k);
				}*/
				valid = true;
				break;
			}
		}

		// If no match use the subsec midpoint as the base. It's always valid.
		if (!valid)
		{
			ss->ssflags |= DDSUBF_MIDPOINT;	// v220-22 used only by the analyser
			ma_inf.degenerate++;
		}
	}
}

// Initialize node piles and line rings.
void R_InitLinks()
{
	int d = 1024;				// Amount to increase linenode allocation by
	int lsz = numlines + d; 	// Some leeway as only 2's lines are linked
	int i;

	lsz = (lsz/d + 1) * d;		// make a multiple of d

	W_FreeALL(NODESID);			// Dump any nodes set from a previous map

	NP_Init(&thingnodes, 256);	// Initialise the thing nodes
	NP_Init(&linenodes , lsz);	// Initialise the line nodes

	// Allocate the rings.
	linelinks = (nodeindex_t*)W_Calloc(sizeof(nodeindex_t) * numlines, NODESID, 13);

	for (i = 0; i < numlines; i++)
	{
		if (lines[i].ml_flags & ML_TWOSIDED)
		{
			linelinks[i] = NP_New(&linenodes, NP_ROOT_NODE);
		}
	}
}

//==========================================================================
// R_SetupLevel
//	This routine is called to polygonize the current level
//  and set various key arrays.
//  Mapinfo must be set before calling.
//==========================================================================
void R_SetupLevel(void)
{
	static int firsttime = true;
	int i;

	// Order critical;

	// 1) Must be called before ResetArrays()
	P_CreateBlockmap();

	// 2) Must be called before any mobjs are spawned.
	ResetArrays(false);

	// 3) Check limits supported are OK and set the ssec->numverts
	checkMaxPolySides();

	// 4) compose the vertex owners array.
	init_VertexOwners();

	// 5) glnodes must be present
	set_Subsectors();

	if (firsttime)
	{
		RL_Init();			// Rendering lists.
		firsttime = false;
	}
	else
	{
		RL_DeleteLists();	// Delete all rendering lists
	}

	R_SubsectorPlanes();

	DL_InitBlockMap();	

	// scale parms from mapinfo values
	mapgravity = (fixed_t)(mapinfo->gravity * FRACUNIT);
	r_ambient = (int)(mapinfo->ambient * 255);

	if (r_ambient <   0) r_ambient = 0;
	if (r_ambient > 255) r_ambient = 255;
	
	Net_ClearCmds();

	// Set target heights of all planes.
	for(i = 0; i < numsectors; i++)
	{
		sectors[i].planes[PLN_FLOOR  ].target = sectors[i].floorheight;
		sectors[i].planes[PLN_CEILING].target = sectors[i].ceilingheight;
	}

	// Init and spawn all type-triggered particle generators.
	// Let's hope there aren't too many...
	P_SpawnTypeParticleGens();
	P_SpawnMapParticleGens();	// uses mapinfo

	// Make sure that the next frame doesn't use a filtered viewer.
	resetNextViewer = true;
}
