//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2010 Jaakko Kernen, Lee Killough, Tim Stump
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: rend_clip.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Minor changes made for R3D (oranges removed)
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
//   Nodes limit removed although previous fixed limit of 128
//   was unlikely to be exceeded. It's just more efficient.
//   C_CheckBspBox added. Note that use of the atantable has
//   been dropped as innaccuracies caused clipping problems.
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
//   Reformatted in parts.
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj
//   Lower precision angles suitable for clipping purposes added
//   (pseudoAngles)
//   Scheme first introduced by Tim Stump in his ZDoomGL port
//   to aid with speed.
//   Interestingly Tim Stump based his initial clipper code
//   on Jaako Kernen's original code.
//
// Based on: Doomsday v1.7.8
// Code reformatted in parts for Risen3D by G.Jackson
//
// DESCRIPTION:  Implement clipnodes
//   The idea is to keep track of occluded angles around the camera.
//   Since subsectors are rendered front-to-back, the occlusion lists
//   start by having the back range set (i.e. the portion outside the
//   frustrum) and eventually fill up to cover the whole 360 degrees
//   around the camera.

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_render.h"
#include "de_misc.h"
#include "p_maputil.h"
#include "r_main.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static int	C_SafeCheckRange(angle_t startAngle, angle_t endAngle);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

clipnode_t	*freenodes;	// v220-6 A list of unused nodes
clipnode_t	*clipnodes;	// The working nodes
clipnode_t	*cliphead;	// The head node.

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

static angle_t angleToPseudo(angle_t ang)
{
	double vecx = cos(ang * M_PI / ANG180);
	double vecy = sin(ang * M_PI / ANG180);
	double result = vecy / (fabs(vecx) + fabs(vecy));

	if (vecx < 0)
	{
		result = 2.f - result;
	}

	return (angle_t)(result * (1 << 30));
}

static angle_t R_PointToPseudoAngle (fixed_t x, fixed_t y)
{	
extern double		dbViewx;
extern double		dbViewy;

	if (x != viewx || y != viewy)
	{
		double vecx = (double)x - dbViewx;
		double vecy = (double)y - dbViewy;
		double result = vecy / (fabs(vecx) + fabs(vecy));

		if (vecx < 0) result = 2.0 - result;

		return (angle_t)(result * (1 << 30));
	}

	return 0;
}

//===========================================================================
//	Finds the first available freenode else creates a new one
//===========================================================================
static clipnode_t *C_GetNew()	
{
	// see if a free node is available
	if (freenodes)
	{
		clipnode_t *free = freenodes;
		freenodes = free->next;
		return free;
	}

	// None available so create a new one
	return (clipnode_t*)W_Malloc(sizeof(clipnode_t), STATICID, 46);
}

//===========================================================================
//	add a new node
//===========================================================================
static clipnode_t *C_NewRange(angle_t stAng, angle_t endAng)	
{
	clipnode_t * c = C_GetNew();

	c->start = stAng;
	c->end   = endAng;
	c->next  = c->prev=NULL;

	return c;
}

//===========================================================================
// return a node to the freenode pool
//===========================================================================
static void C_FreeNode(clipnode_t *node)
{
	node->next = freenodes;
	freenodes = node;
}

//===========================================================================
// remove the range sent
//===========================================================================
static void C_RemoveRange(clipnode_t *crange)
{
	if (crange == cliphead)
	{
		cliphead = cliphead->next;
	}
	else
	{
		if (crange->prev) crange->prev->next = crange->next;
		if (crange->next) crange->next->prev = crange->prev;
	}
	
	C_FreeNode(crange);
}

//===========================================================================
// 	Must only be called with startAngle < endAngle
//===========================================================================
static void C_AddRange(angle_t startAngle, angle_t endAngle)
{
	clipnode_t	*ci, *crange;

	// If there is no head, this will be the first range.
	if (!cliphead)
	{
		cliphead = C_NewRange(startAngle, endAngle);
		return;
	}

	// Now check if any of the old ranges are contained by the new one.
	for (ci = cliphead; ci;)
	{
		if (startAngle >= ci->start && endAngle <= ci->end)
			return;	// The new range already exists.

		if (ci->start >= startAngle && ci->end <= endAngle)
		{
			crange = ci;
			ci = ci->next;	
			C_RemoveRange(crange);
		}
		else
			ci = ci->next;
	}

	// Now it is possible that the new range overlaps one or two old ranges.
	// If two are overlapped, they are consecutive. First we'll try to find
	// a range that overlaps the beginning.
	for (crange = NULL, ci = cliphead; ci; ci = ci->next)
	{
		// In preparation of the next stage, find a good spot for the range.
		if(ci->start < endAngle) crange = ci; // Add after this one.

		if (ci->start >= startAngle && ci->start <= endAngle)
		{
			// New range's end and ci's beginning overlap. ci's end is outside.
			// Otherwise it would have been already removed.
			// It suffices to adjust ci.
			ci->start = startAngle;
			return;
		}

		// Check an overlapping end.
		if (ci->end >= startAngle && ci->end <= endAngle)
		{
			// Now it's possible that the ci->next's beginning overlaps the new
			// range's end. In that case there will be a merger.
			crange = ci->next;

			if(!crange)
			{
				ci->end = endAngle;
			}
			else
			{
				if(crange->start <= endAngle)
				{
					// A fusion will commence. ci will eat the new range 
					ci->end = crange->end;
					C_RemoveRange(crange);
				}
				else
				{
					ci->end = endAngle;	// Not overlapping.
				}
			}
			return;
		}
	}
	
	// If we get here then the range is disconnected from
	// the others and needs to be added to the list.
	if (!crange)
	{
		// We have a new head node.
		crange = cliphead;
		cliphead = C_NewRange(startAngle, endAngle);
		cliphead->next = crange;
		if(crange) crange->prev = cliphead;
	}
	else
	{
		// Add the new range after crange.
		ci = C_NewRange(startAngle, endAngle);
		ci->next = crange->next;
		if(ci->next) ci->next->prev = ci;
		ci->prev = crange;
		crange->next = ci;
	}
}

//===========================================================================
// Add a new range
//===========================================================================
static int C_SafeAddRange(angle_t startAngle, angle_t endAngle)
{
	// The range may wrap around.
	if (startAngle > endAngle)
	{
		// The range has to added in two parts.
		C_AddRange(startAngle, ANGMAX);
		C_AddRange(0, endAngle);
	}
	else
	{
		// Add the range as usual.
		C_AddRange(startAngle, endAngle);
	}

	return true;
}

//===========================================================================
//	Must only be called with startAngle < endAngle
//===========================================================================
static int C_IsRangeVisible(angle_t startAngle, angle_t endAngle)
{
	clipnode_t	*ci = cliphead;
	
	if (ci && !endAngle && !ci->start) return false;
	
	while (ci && ci->start < endAngle)
	{
		if (startAngle >= ci->start && endAngle <= ci->end)
			return false;

		ci = ci->next;
	}

	// No node fully contained the specified range.
	return true;
}

//===========================================================================
//	Returns 1 if the range is not entirely clipped.
//   v220-7 modified
//===========================================================================
static int C_SafeCheckRange(angle_t an1, angle_t an2)
{
	if (an1 > an2)
		return (C_IsRangeVisible(an1, ANGMAX) || C_IsRangeVisible(0, an2));
	
	if (an1 < an2)
		return C_IsRangeVisible(an1, an2);

	return true;
}

// Public functions ---------------------------------------------------------

//===========================================================================
// Clear all ranges
//===========================================================================
void C_ClearRanges(void)
{
	clipnode_t *cprev, *crange = cliphead;
	
	while (crange != NULL)
	{
		cprev = crange;
		crange = crange->next;
		C_FreeNode(cprev);
	}
	
	cliphead = NULL;
}

//===========================================================================
//	Add a segment relative to the current viewpoint.
//===========================================================================
void C_AddViewRelSeg(seg_t *sg)
{
	// If R3DSEGF_SHORTSEG is set then the seg will not be used for occlusion.
	if (sg->sgflags & R3DSEGF_SHORTSEG)
		return;
	
	C_SafeAddRange (R_PointToPseudoAngle(sg->v2->x, sg->v2->y), 
					R_PointToPseudoAngle(sg->v1->x, sg->v1->y));
}

//===========================================================================
//	Used with segs
//===========================================================================
int C_CheckViewSeg(seg_t *sg)
{
	return C_SafeCheckRange(
		R_PointToPseudoAngle(sg->v2->x, sg->v2->y),
		R_PointToPseudoAngle(sg->v1->x, sg->v1->y));
}

void C_SafeAddRangeRealAngles(angle_t startangle, angle_t endangle)
{
	C_SafeAddRange(angleToPseudo(startangle), angleToPseudo(endangle));
}

// array of order of possible edges.
// Note 0 = BOXTOP, 1 = BOXBASE, 2 = BOXLEFT, 3 = BOXRIGHT
static const int checkcoord[10][4] =
{
	{3,0,2,1},	//  0
	{3,0,2,0},	//  1
	{3,1,2,0},	//  2
	{0,0,0,0},	//  3 (never referenced)
	{2,0,2,1},	//  4
	{0,0,0,0},	//  5
	{3,1,3,0},	//  6
	{2,0,3,1},	//  7
	{2,1,3,1},	//  8
	{2,1,3,0}	//  9
};

//===========================================================================
// v220-6 Find the corners of the box that define the edges from current viewpoint.
//  Returns 1 if the subsector might be visible
//===========================================================================
int C_CheckBspBox(fixed_t *bspbox)
{
	// possible boxpos results 0, 1, 2, 4, 5, 6, 7, 8, 9
	int boxpos  = (viewx <= bspbox[BOXLEFT] ? 0 : viewx < bspbox[BOXRIGHT] ? 1 : 2)		// boxpos  = 0, 1 or 2
				+ (viewy >= bspbox[BOXTOP ] ? 0 : viewy > bspbox[BOXBASE ] ? 4 : 7);	//         + 0, 4 or 7

	if (boxpos != 5)
	{
		const int	*check = checkcoord[boxpos];
		angle_t		an0 = R_PointToPseudoAngle(bspbox[check[0]], bspbox[check[1]]);
		angle_t		an1 = R_PointToPseudoAngle(bspbox[check[2]], bspbox[check[3]]);

		return C_SafeCheckRange(an1, an0);
	}

	return true;
}
