//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: rend_dyn.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D mainly for slopes and using changes
//  made for edit mode that can help speed things up.
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 15 (v220-15) 08/Jun/2012 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Application of lit surfaces
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_render.h"
#include "de_refresh.h"
#include "de_play.h"
#include "de_graphics.h"
#include "de_misc.h"

#include "r_model.h"
#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "r3d_mapinfo.h"

// MACROS ------------------------------------------------------------------

enum { CLIP_TOP, CLIP_BASE, CLIP_LEFT, CLIP_RIGHT };

#define BLK_ROOT(bx, by)	(dlBlockLinks + bx + by*bmapwidth)

// Wall Surface Present flags (for DL_ProcessWallSeg).
#define WSP_BOTTOM		0x01
#define WSP_MIDDLE		0x02
#define WSP_MIDDLEH		0x04
#define WSP_MIDDLEL		0x08
#define WSP_TOP			0x10

#define LUM_FACTOR(dist)	(1.5f - 1.5f*(dist)/lum->radius)

#define LUM_INCREMENT	512
#define MISSILE_FACTOR	0.25f	// v220-31 used with LUMF_MISSILE
#define MISSILE_FACTOR1	0.5f	// v220-31 used with LUMF_MISSILE1

// TYPES -------------------------------------------------------------------

typedef struct
{
	RDBOOL		lightFloor, lightCeiling;
	rendpoly_t	*poly;
	subsector_t	*subsector;
	float		fceil, ffloor;
	float		oldfceil, oldffloor;
	int			segcount;
	seg_t		**segptr;
} flatitervars_t;
	
flatitervars_t	fi;

// Put globals in a structure - R3D
struct wallitervars_t
{
	int			do_floor;
	int			do_ceil;
	int			top;
	int			bottom;
	int			oldtop;
	int			oldbottom;
	int			glowhi;
	int			glowlo;
	short		floorpic;
	short		ceilpic;
} wiv;

typedef struct
{
	// Note: these must correspond with spritelump_t
	float xoffset;		// flarex
	float yoffset;		// flarey
	float size;			// lumsize

} lightconfig_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL		setNoFog;	// v220-03
extern dynphase_t	dynphase;	// v220-31

// PUBLIC DATA DEFINITIONS -------------------------------------------------

float		dlFactor = 0.7f;
int			dlInited = false;
int			useWallGlow = true;
int			glowHeight = 100;
int			dlMaxRad   = 360; // Dynamic lights maximum radius.
int			numLuminous = 0;
fixed_t		dlMaxRadft = 360 << FRACBITS; // Dynamic lights maximum radius.
fixed_t		dlMaxMissileRadft = 64 << FRACBITS; // max rad for missiles

// R3D: No longer used
//float		dlRadFactor = 3;
//int		maxDynLights = 0; $halo removed
//float		dlContract = 0.02f; // Almost unnoticeable... a hack, though.

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static lumobj_t	**dlSubLinks = NULL;
static lumobj_t	*luminousList = NULL;
static mobj_t lummy;

static int		maxLuminous = 0;

static lumobj_t	**dlBlockLinks = NULL;

static size_t	dlBlockLinksize = 0;
static size_t	dlSubLinksize = 0;

static fixed_t	glowhgt;	

static RDBOOL	hasfog;		// R3D

// v220-5 made global as memory is now dynamically allocated as needed
static zvertex_t	*inverts;
static zvertex_t	*verts;
static RDBOOL		*dynsides;

// CODE --------------------------------------------------------------------

static RDBOOL DL_ThingColor(lumobj_t *lum, DGLubyte *outRGB, float light)
{
	fixed_t	dist;
	int		i;

	if (light <= 0) return false;

	// v220-3 Attenuate light with dist
	dist = P_ApproxDistance3(viewx - lum->th->x, viewy - lum->th->y, viewz - lum->z);

	if (light > 1.f) light = 1.f;

	light *= hasfog ? DL_FOGFACTOR : dlFactor;	// This will reduce the value

	if (hasfog)
	{
		// reduce if distance > DL_FOGDIST
		if (dist > DL_FOGDIST) light *= DL_FOGDIST/(float)dist;
	}
	else if (light > 0.01f)
	{
		float	min = light * 0.25f;
		float	max = light;

		light = light - (FIX2FLT(dist) - 32.f) * ((1.f-light) / maxLightDistance);

		// Clamp it
		if      (light < min) light = min;
		else if (light > max) light = max;
	}

	if (light <= 0) return false;

	// Multiply with the light color.
	for (i = 0; i < 3; i++)
	{
		outRGB[i] = (DGLubyte) (light * lum->rgb[i]);
	}

	return true;
}

//===========================================================================
// DL_InitBlockMap
//===========================================================================
void DL_InitBlockMap(void)
{
	// init fixed values for lummy (dummy lum)
	if (!lummy.state)
		lummy.state = (state_t*)W_Calloc(sizeof(state_t), STATICID, 326);

	lummy.sprite = SPR_DYNL;
	lummy.radius = 16*FRACUNIT;
	lummy.height = 16*FRACUNIT;

	dlInited = false;
	// initialize the subsector links (root pointers).
	dlSubLinksize = sizeof(lumobj_t*) * numsubsectors;
	dlSubLinks = (lumobj_t**)W_Calloc(dlSubLinksize, LEVELID, 35);

	dlBlockLinksize = sizeof(lumobj_t*) * bmapwidth * bmapheight;
	dlBlockLinks = (lumobj_t**)W_Malloc(dlBlockLinksize, LEVELID, 47);

	// v220-31 Initialise luminous list. This can expand if needed
	if (!luminousList)
	{
		maxLuminous = LUM_INCREMENT;
		luminousList = (lumobj_t*)W_Realloc(luminousList, sizeof(lumobj_t) * maxLuminous, STATICID, 221);
	}
}

static RDBOOL DL_FinalizeSurface(rendpolyq_t *dlq, lumobj_t *lum)
{
	dlq->light	 = lum;
	dlq->texoffy = FIX2FLT(lum->z) + lum->center 
		+ lum->radius/DYN_ASPECT - dlq->top;

	if(dlq->texoffy < -dlq->top + dlq->bottom) return false;
	if(dlq->texoffy > dlq->texh) return false;
	if(dlq->top + dlq->texoffy - dlq->texh >= dlq->bottom) // Fits completely?
		dlq->bottom = dlq->top + dlq->texoffy - dlq->texh;
	if(dlq->texoffy < 0)
	{
		dlq->top += dlq->texoffy;
		dlq->texoffy = 0;
	}
	return true;
}


//===========================================================================
// DL_ProcessWallSeg
//===========================================================================

static void setWallOffsets(rendpolyq_t *dlq, int *type)
{
	float	uvecWall[2];
	int		c;

	*type = 0;

	// The wall vector.
	for(c = 0; c < 2; c++) 
		uvecWall[c] = (dlq->vertices[1].pos[c] - dlq->vertices[0].pos[c]) / dlq->length;

	// dlContract removed - R3D

	if (dlq->texoffx > 0)
	{
		for (c = 0; c < 2; c++)
			dlq->vertices[0].pos[c] += dlq->texoffx * uvecWall[c];

		if (dlq->texoffx + dlq->texw <= dlq->length) // Fits completely?
		{
			for(c = 0; c < 2; c++) 
				dlq->vertices[1].pos[c] = dlq->vertices[0].pos[c] + dlq->texw*uvecWall[c];

			dlq->length = (float)dlq->texw; *type = 1;
		}
		else	// Doesn't fit, must truncate.
		{
			dlq->length -= dlq->texoffx; *type = 2;
		}

		dlq->texoffx = 0;
	}
	else if (dlq->texoffx + dlq->texw <= dlq->length) // Fits completely?
	{
		// It goes off to the left.
		for (c = 0; c < 2; c++)
		{
			dlq->vertices[1].pos[c] = dlq->vertices[0].pos[c] + 
				(dlq->texw + dlq->texoffx) * uvecWall[c];
		}

		dlq->length = dlq->texw + dlq->texoffx; *type = 3;
	}
}

// R3D If fog is present projects and resizes glow planes to
//     help prevent texture fighting and edge artefacts
static void projectForwards(rendpolyq_t *dlq, seg_t *sg)
{
	binangle_t angle = bamsAtan2((sg->v2->y - sg->v1->y) >> FRACBITS, 
				(sg->v2->x - sg->v1->x) >> FRACBITS) >> (ANGLETOFINESHIFT - FRACBITS);
	float dx = FIX2FLT(FixedMul(0x2000, finesine  [angle]));
	float dy = FIX2FLT(FixedMul(0x2000, finecosine[angle]));

	// Offset
	dlq->vertices[0].pos[VX] += dx;
	dlq->vertices[0].pos[VY] -= dy;
	dlq->vertices[1].pos[VX] += dx;
	dlq->vertices[1].pos[VY] -= dy;
}

// R3D slope support added
static void DL_ProcessWallSeg(lumobj_t *lum, seg_t *seg)
{
	side_t	*ls = seg->sidedef;
	short	lsiflags = ls->lsiflags;
	int		present = 0;

	// If any textures on this seg have alpha or translucency then dyn lighting is not allowed.

	if ((lsiflags & (LS_HITEX |LS_HITRANS |LS_HIALPHA )) == LS_HITEX && ls->hh - ls->hl >= FRACUNIT)
	{
		present |= WSP_TOP;
	}

	if ((lsiflags & (LS_MIDTEX|LS_MIDTRANS|LS_MIDALPHA)) == LS_MIDTEX && ls->mh - ls->ml >= FRACUNIT)
	{
		present |= WSP_MIDDLE;
	}

	if (lsiflags & LS_MIDTEXH && ls->sh0 - ls->sl0 >= FRACUNIT)
	{
		present |= WSP_MIDDLEH;
	}

	if (lsiflags & LS_MIDTEXL && ls->sh1 - ls->sl1 >= FRACUNIT)
	{
		present |= WSP_MIDDLEL;
	}

	if ((lsiflags & (LS_LOWTEX|LS_LOWTRANS|LS_LOWALPHA)) == LS_LOWTEX && ls->lh - ls->ll >= FRACUNIT)
	{
		present |= WSP_BOTTOM;
	}

	if (!present) return;	// No surfaces to light

	// -------------------------------------

	rendpolyq_t	dlq;
	float		pntLight[2];
	int			type;


	dlq.type = rp_quad;
	dlq.numvertices = 2;
	dlq.vertices[0].pos[VX] = seg->fv1[VX];
	dlq.vertices[0].pos[VY] = seg->fv1[VY];
	dlq.vertices[1].pos[VX] = seg->fv2[VX];
	dlq.vertices[1].pos[VY] = seg->fv2[VY];

	pntLight[VX] = FIX2FLT(lum->th->x);
	pntLight[VY] = FIX2FLT(lum->th->y);
	
	float dist = ((dlq.vertices[0].pos[VY] - pntLight[VY]) * 
		(dlq.vertices[1].pos[VX] - dlq.vertices[0].pos[VX]) - 
		(dlq.vertices[0].pos[VX] - pntLight[VX]) * 
		(dlq.vertices[1].pos[VY] - dlq.vertices[0].pos[VY])) / seg->length;

	if (lum->flags & LUMF_MISSILE) dist *= MISSILE_FACTOR;
	else
	if (lum->flags & LUMF_MISSILE1) dist *= MISSILE_FACTOR1;

	// Is it close enough and on the right side?
	if (dist < 0 || dist > lum->radius) return; // Nope.

	dlq.light  = lum;
	dlq.texw   = (float)(lum->radius + lum->radius);	// twice the radius
	dlq.texh   = dlq.texw * DYN_ASPECTRCP;
	dlq.length = seg->length;
	
	// Do a scalar projection for the offset.
	dlq.texoffx = ((dlq.vertices[0].pos[VY] - pntLight[VY]) * 
		(dlq.vertices[0].pos[VY] - dlq.vertices[1].pos[VY]) - 
		(dlq.vertices[0].pos[VX] - pntLight[VX]) * 
		(dlq.vertices[1].pos[VX] - dlq.vertices[0].pos[VX])) / dlq.length - lum->radius;

	// There is no need to always draw the *whole* wall. Adjust the start
	// and end points so that only a relevant portion is included.
	if(dlq.texoffx > dlq.length) return; // Doesn't fit on the wall.
	if(dlq.texoffx < -dlq.texw ) return; // Ditto, but in the other direction.

	// Calculate the color of the light.
	if (!DL_ThingColor(lum, dlq.vertices[0].rgba, LUM_FACTOR(dist))) return;

	setWallOffsets(&dlq, &type);

	if (lum->flags & LUMF_TORCH)
	{
		int litnum = seg->sgflags & DDSEGF_ISLIT;

		if (litnum >= 8) return;	// v220-9 8 torch lights/seg max (was 6 although bug reduced it to 4)

		if (litnum && seg->frontsector == seg->backsector)
			return;

		seg->sgflags &= ~DDSEGF_ISLIT;	// bug fix - clear before resetting
		seg->sgflags |= ++litnum;
	}

	// -------------------------------------

	seg->sgflags |= DDSEGF_DLIGHT;

	// ------------------------- $slopes - R3D
	// Render the present surfaces.
	if (   present & (LS_MIDTEXL|LS_MIDTEXH)
		&& seg->backsector
		&& seg->linedef->ml_flags & (LNF2_HITRIWALL|LNF2_LOWTRIWALL) )
	{
		float zlo0 = ssecvtxz[seg->zptr1].z[0];
		float zhi0 = ssecvtxz[seg->zptr1].z[1];
		float zlo1 = ssecvtxz[seg->zptr2].z[0];
		float zhi1 = ssecvtxz[seg->zptr2].z[1];

		if (present & WSP_MIDDLEH && zhi0 != zhi1)	// slope trihi
		{
			dlq.top	   = FIX2FLT(MIN_OF(ls->sh0, wiv.top));
			dlq.bottom = FIX2FLT(MAX_OF(ls->sl0, wiv.bottom));

			if (DL_FinalizeSurface(&dlq, lum) && dlq.top > dlq.bottom)
			{
				float tht = FIX2FLT(seg->backsector->smoothceil);

				if (dlq.bottom < tht)
				{					
					// points along the line (introduces an adjustment which cannot be
					// emulated simply by reducing the z heights slightly)
					dlq.vertices[0].pos[VZ] = FIX2FLT(SL_XYHeight(seg->backsector->stptr,
						(int)(dlq.vertices[0].pos[VX]*FRACUNIT),
						(int)(dlq.vertices[0].pos[VY]*FRACUNIT), HISLOPE));
					dlq.vertices[1].pos[VZ] = FIX2FLT(SL_XYHeight(seg->backsector->stptr,
						(int)(dlq.vertices[1].pos[VX]*FRACUNIT),
						(int)(dlq.vertices[1].pos[VY]*FRACUNIT), HISLOPE));
					
					dlq.flags = RPF_TRIHI|RPF_LIGHT;

					// Note: zheights will be -ve so everything is reversed
					dlq.bottom = tht + MIN_OF(dlq.vertices[0].pos[VZ], dlq.vertices[1].pos[VZ]);

					if (hasfog) projectForwards(&dlq, seg);
					RL_AddPolyWall(&dlq, NULL);
				}
			}
		}

		if (present & WSP_MIDDLEL && zlo0 != zlo1)	// slope trilo
		{
			dlq.bottom = FIX2FLT(MAX_OF(ls->sl1, wiv.bottom));
			dlq.top = FIX2FLT(MIN_OF(ls->sh1, wiv.top));

			if (DL_FinalizeSurface(&dlq, lum) && dlq.top > dlq.bottom)
			{
				// $slopes - R3D
				float bht = FIX2FLT(seg->backsector->smoothfloor);

				if (dlq.top > bht)
				{
					float top = dlq.top;

					// points along the line (introduces an adjustment which cannot be
					// emulated simply by reducing the z heights slightly)
					dlq.vertices[0].pos[VZ] = FIX2FLT(SL_XYHeight(seg->backsector->stptr,
						(int)(dlq.vertices[0].pos[VX]*FRACUNIT),
						(int)(dlq.vertices[0].pos[VY]*FRACUNIT), LOSLOPE));
					dlq.vertices[1].pos[VZ] = FIX2FLT(SL_XYHeight(seg->backsector->stptr,
						(int)(dlq.vertices[1].pos[VX]*FRACUNIT),
						(int)(dlq.vertices[1].pos[VY]*FRACUNIT), LOSLOPE));
					
					dlq.flags = RPF_TRILOW|RPF_LIGHT;
					dlq.top = bht + MAX_OF(dlq.vertices[0].pos[VZ], dlq.vertices[1].pos[VZ]);
					dlq.texoffy -= dlq.top - top;

					if (hasfog) projectForwards(&dlq, seg);
					RL_AddPolyWall(&dlq, NULL);
				}
			}
		}

		dlq.vertices[0].pos[VZ] = 0.f;
		dlq.vertices[1].pos[VZ] = 0.f;
		goto bypass_middle;
	}

	// -------------------------- end slopes

	if (present & WSP_MIDDLE)
	{
		dlq.flags = RPF_LIGHT;
		dlq.top	   = FIX2FLT(MIN_OF(ls->mh, wiv.top));
		dlq.bottom = FIX2FLT(MAX_OF(ls->ml, wiv.bottom));

		if (DL_FinalizeSurface(&dlq, lum) && dlq.top > dlq.bottom)
		{
			dlq.flags = RPF_LIGHT;

			if (hasfog) projectForwards(&dlq, seg);
			RL_AddPolyWall(&dlq, NULL);
		}
	}
bypass_middle:;
	if (present & WSP_TOP)
	{
		dlq.top	   = FIX2FLT(MIN_OF(ls->hh, wiv.top));
		dlq.bottom = FIX2FLT(MAX_OF(ls->hl, wiv.bottom));

		if (DL_FinalizeSurface(&dlq, lum) && dlq.top > dlq.bottom)
		{
			dlq.flags = RPF_LIGHT;

			if (hasfog) projectForwards(&dlq, seg);
			RL_AddPolyWall(&dlq, NULL);
		}
	}

	if (present & WSP_BOTTOM)
	{
		dlq.top    = FIX2FLT(MIN_OF(ls->lh, wiv.top));
		dlq.bottom = FIX2FLT(MAX_OF(ls->ll, wiv.bottom));

		if (DL_FinalizeSurface(&dlq, lum) && dlq.top > dlq.bottom)
		{
			dlq.flags = RPF_LIGHT;

			if (hasfog) projectForwards(&dlq, seg);
			RL_AddPolyWall(&dlq, NULL);
		}
	}
}

static void DL_SetupGlowQuads(seg_t *seg, fixed_t segtop, fixed_t segbottom)
{
	// R3D: For top and bottom of glow panes only
	float	tbContract = 0.02f;

	rendpolyq_t glow;
	int		i, g;
//	float	factor = hasfog ? dlFactor * 0.4f : dlFactor * 0.5f;
	float factor = hasfog ? 0.12f : 0.3f; // b211-0
	
	// Check the heights.
	if (segtop <= segbottom) return;	// No height.

	glow.vertices[0].pos[VX] = seg->fv1[VX];
	glow.vertices[0].pos[VY] = seg->fv1[VY];
	glow.vertices[1].pos[VX] = seg->fv2[VX];
	glow.vertices[1].pos[VY] = seg->fv2[VY];

	for(g = 0; g < 2; g++)
	{
		if ( (!g && wiv.do_ceil) || (g && wiv.do_floor) )
		{
			fixed_t top = segtop, bottom = segbottom, limit;

			// Clip.						
			if (g) // Floor glow.
			{
				limit = wiv.glowlo + glowhgt;
				if (bottom > limit) continue;	// Too high.
				if (top > limit) top = limit;	// Clip the top.
			}
			else // Ceiling glow
			{
				limit = wiv.glowhi - glowhgt;
				if (top < limit) continue;			// Too low.
				if (bottom < limit) bottom = limit;	// Clip the bottom.
			}
			
			// $slopes - R3D
			if (!g && seg->linedef->lnflags2 & LNF2_HITRIWALL)
			{
				float	z0 = ssecvtxz[seg->zptr1].z[1];
				float	z1 = ssecvtxz[seg->zptr2].z[1];
				glow.vertices[0].pos[VZ] = z0;
				glow.vertices[1].pos[VZ] = z1;
				glow.flags = RPF_TRIHI;
				bottom = seg->backsector->ceilingheight + (int)(MIN_OF(z0,z1)*FRACUNIT);	
			}
			else
			if (g && seg->linedef->lnflags2 & LNF2_LOWTRIWALL)
			{
				float	z0 = ssecvtxz[seg->zptr1].z[0];
				float	z1 = ssecvtxz[seg->zptr2].z[0];
				glow.vertices[0].pos[VZ] = z0;
				glow.vertices[1].pos[VZ] = z1;
				glow.flags = RPF_TRILOW;
				top = seg->backsector->floorheight + (int)(MAX_OF(z0,z1)*FRACUNIT);	
			}
			else
			{
				glow.flags = 0;
			}

			glow.type = rp_quad;
			glow.numvertices = 2;
			glow.length = seg->length;
			glow.flags |= (RPF_WALL_GLOW | RPF_LIGHT);

			GL_GetFlatColor( g ? wiv.floorpic : wiv.ceilpic, glow.vertices[0].rgba);

			for (i = 0; i < 3; i++)
			{
				glow.vertices[0].rgba[i] =
					(byte) ((float)((byte)glow.vertices[0].rgba[i]) * factor);
			}

			// dlContract removed - R3D
			glow.top = FIX2FLT(top) - tbContract;
			glow.bottom = FIX2FLT(bottom) + tbContract;

			// texoffx will store the height.
			glow.texoffx = (float)(g ? -glowHeight : glowHeight);	// Flip for floor glow.

			if (g == 0) // Ceiling glow.
			{
				// Set texoffy so the texture top reaches ceiling height.
				glow.texoffy = FIX2FLT(wiv.glowhi - top);
			}
			else // Floor glow.
			{
				// Set texoffy so the texture bottom reaches floor height.
				glow.texoffy = FIX2FLT(bottom - wiv.glowlo);
			}

			// Mark the segment for dlight rendering.
			seg->sgflags |= DDSEGF_DLIGHT;

			RL_AddPolyWall(&glow, NULL);
		}
	}
}

// R3D: modifed for use with slopes
static void DL_ProcessWallGlow(seg_t *seg, subsector_t *ssec)
{
	side_t  *ls = seg->sidedef;
	short	lsiflags = ls->lsiflags;
	RDBOOL	hislope = R3D_slopetypes
		&& lsiflags & LS_HITEX
		&& (seg->frontsector->specFlags2 & SF2_HISLOPE
		|| (seg->backsector && seg->backsector->specFlags2 & SF2_HISLOPE));
	RDBOOL	loslope = R3D_slopetypes
		&&  underwater
		&& lsiflags & LS_LOWTEX
		&& seg->linedef->lnflags2 & LNF2_LOWTRIWALL;
	// If any textures on this seg have alpha then glow is not allowed.
	// This is set where a patched texture has mixed alpha and non-alpha tex's

	// Used to determine extent of height to be lit.
	glowhgt = glowHeight << FRACBITS;

	if (   lsiflags & LS_MIDTEX
		&& !(lsiflags & (LS_NOGLOW | LS_MIDTRANS | LS_MIDALPHA) )
		&& ls->mh > ls->ml)
	{
		DL_SetupGlowQuads(seg, ls->mh, ls->ml);
	}

	// Suppress glow being transferred to hi textures if underwater
	// glow is also not allowed on hi triwalls
	if (lsiflags & LS_HITEX && !hislope && !(underwater && ls->hl > ssec->s_waterplane))
	{
		if (!(lsiflags & LS_HIALPHA) && ls->hh > ls->hl)
		{
			DL_SetupGlowQuads(seg, ls->hh, ls->hl);
		}
	}

	// glow not allowed on lo triwalls when underwater.
	// To overcome this the intersection of a fake floor with a sloped
	// wall would have to be calculated, amongst other considerations,
	// and it's probably not worth the effort. 
	if (!loslope && lsiflags & LS_LOWTEX && ls->lh > ls->ll)
	{
		DL_SetupGlowQuads(seg, ls->lh, ls->ll);
	}
}
				
//===========================================================================
// DL_AddLuminous
//	Registers the given thing as a luminous, light-emitting object.
//	Note that this is called each frame for each luminous object!
//  plr: 1 if player, 2 if player weapon flash ele 0
//===========================================================================
static void DL_AddLuminous(mobj_t *thing, byte *strobergb, int plr)
{
	extern int		dynsplnum;

	ded_light_t		*def = NULL;
	lightconfig_t	cf;
	lumobj_t		*lum;
	spritedef_t		*sprdef   = &sprites[thing->sprite];
	spriteframe_t	*sprframe = &sprdef->spriteframes[thing->frame & FF_FRAMEMASK];
	float			n;
	int				i, splnum;

	if (plr)
	{
		splnum = dynsplnum;
	}
	else if (thing->sprite == SPR_DYNL)
	{
		splnum = sprframe->lump[4];
	}
	else
	{
		if (thing->type == MT_ROCKET)
		{
			splnum = dynsplnum;
		}
		else
		{
			splnum = sprframe->lump[0];
		}
	}

	// Ensure we have up-to-date texture information
	// Sets flarex, flarey, amplified color etc.
	if (!GL_PrepareSprite(splnum, NULL, 1)) return;

	// Check that the luminousList array is large enough
	if (numLuminous >= maxLuminous)
	{
		maxLuminous += LUM_INCREMENT;
		luminousList = (lumobj_t*)W_Realloc(luminousList, sizeof(lumobj_t) * maxLuminous, STATICID, 221);
	}

	lum = &luminousList[numLuminous++];

	if (plr == 2)
	{
		angle_t angle = (thing->angle >> ANGLETOFINESHIFT);
		fixed_t dx = FixedMul(15*FRACUNIT, finecosine[angle]);
		fixed_t dy = FixedMul(15*FRACUNIT, finesine[angle]);

		lum->th = &lummy;

		lummy.state = thing->state;
		lummy.state->light = thing->state ? thing->state->light : 0;
		lummy.x = thing->x + dx;
		lummy.y = thing->y + dy;
		lummy.z = thing->z;
//		lummy.momx = thing->momx;
//		lummy.momx = thing->momy;
//		lummy.momz = thing->momz;
		lummy.subsector = thing->subsector;

		lum->z = thing->z + ge_inf.flashoffset*FRACUNIT;
	}
	else
	{
		lum->th = thing;
		lum->z = thing->z;
	}

	lum->flags = 0;

	// Set the defaults (note correspondance between spritelump_t and lightconfig_t)
	memcpy(&cf, &spritelumps[splnum].flarex, 3*sizeof(float));

	// v220-6 replaced with above
/*	cf.xoffset = lum->xoff = spritelumps[splnum].flarex;
	cf.yoffset = spritelumps[splnum].flarey;
	cf.size    = spritelumps[splnum].lumsize;*/

	// X offset to the flare position.
	lum->xoff = cf.xoffset - spritelumps[splnum].width/2.0f;

	// Check if these are overriden by a ded definition
	if (plr == 1)	// torchlight
	{
		cf.size = 3.f;	// $r_ambient must be this size else sprite lighting goes wrong
		lum->flags |= LUMF_PLAYER;
	}
	else if (plr == 2)	// extralight
	{
		cf.size = 2.f;
	}
	else
	if (thing->state && thing->state->light)
	{
		def = (ded_light_t*) thing->state->light;

		if (def->size   ) cf.size = def->size;
		if (def->xoffset) cf.xoffset = lum->xoff = def->xoffset;
		if (def->yoffset) cf.yoffset = def->yoffset;		

		lum->flags |= def->flags;
	}
	else
	{
		// v220-6 hard-coded changes when using sprites
		if (thing->type == MT_MISC38)		// evil-eye
		{
			if (cf.size > 0.8f) cf.size = 0.8f;
		}
		else if (thing->sprite >= SPR_BKEY
			&& thing->sprite <= SPR_YSKU)	// keys
		{
			cf.size = 0.2f;
		}
	}

	// Set the radius
	// R3D was cf->size * 40 * dlRadFactor
	//     dlRadFactor is now fixed at 3
	lum->radius = (int)(cf.size * 120);
	
	// Keep the radius within sensible limits
	if (lum->radius > dlMaxRad) lum->radius = dlMaxRad;

	if (thing->flags3 & MF3_HASCENTRE)	//models only
	{
		// v220-3 use average of yoffset to stop bouncing torch light
		lum->center = thing->zcenter;
	}
	else
	{
		lum->center = (float)spritelumps[splnum].slyoffset - cf.yoffset;

		if ( !(thing->flags2 & MF2_NOFITBOTTOM) )
		{
			n = FIX2FLT(thing->z)
				+ spritelumps[splnum].slyoffset 
				- spritelumps[splnum].height
				- FIX2FLT(SL_PointHeight(thing->subsector->sector, thing->x, thing->y, LOSLOPE));

			if (n < 0) lum->center -= n;	// Must adjust.
		}
	}

//	if (plr == 1) lum->center += 10.f;

	// v220-31 support for extended dynlight types
	if (strobergb)
	{
		memcpy(lum->rgb, strobergb, 3);
	}
	else
	{
		// Does the mobj use a light scale?
		if (thing->ddflags & DDMF_LIGHTSCALE)
		{
			// Also reduce the size of the light according to 
			// the scale flags. *Won't affect the flare.*
			n = 1.0f - ((thing->ddflags & DDMF_LIGHTSCALE) >> DDMF_LIGHTSCALESHIFT)/4.0f;
			lum->radius = (int)(lum->radius * n);
		}

		// If there's a def then color comp is set using the def values
		if (def && (def->color[0] || def->color[1] || def->color[2]))
		{
			for (i = 0; i < 3; i++)
			{
				lum->rgb[i] = (byte) (255.0f * def->color[i]);
			}
		}
		else	// use the sprite's (amplified) color.
		{
			memcpy(lum->rgb, &spritelumps[splnum].color, 3);
		}
	}

	// R3D add flags

	// 1) Flag used to prevent too many torches illuminating a wall segment
	//    v220-4 code revised
	if (thing->flags2 & MF2_TORCH) lum->flags |= LUMF_TORCH;

	// 2) Flag used to identify missiles for parameter modification
	//	  a) illuminating another model if not near enough - see modelLighter()
	//	  b) modifying distance to surface for calculating travelling glow
	else if (thing->momx|thing->momy)
	{
		if (thing->type == MT_ROCKET || thing->type == MT_PLASMA)
		{
			lum->flags |= LUMF_MISSILE;
		}
		else if (thing->type == MT_TRACER
			|| thing->type == MT_BRUISERSHOT
			|| thing->type == MT_HEADSHOT
			|| thing->type == MT_TROOPSHOT)
		{
			lum->flags |= LUMF_MISSILE1;
		}
	}
}

//===========================================================================
// DL_LinkLuminous
//	Clears the dlBlockLinks and then links all the listed luminous objects.
//===========================================================================
static void DL_LinkLuminous()
{
	lumobj_t	**root, *lum = luminousList;
	int			i, bx, by;

	// Clear all the roots.
	memset(dlBlockLinks, 0, dlBlockLinksize);
	memset(dlSubLinks  , 0, dlSubLinksize  );
	
	for(i = 0; i < numLuminous; i++, lum++)
	{
		// Link this lumobj to the dlBlockLinks, if it can be linked.
		lum->next = NULL;

		bx = MAPBS(lum->th->x, bmaporgx);
		by = MAPBS(lum->th->y, bmaporgy);

		if(bx >= 0 && by >= 0 && bx < bmapwidth && by < bmapheight)
		{
			root = BLK_ROOT(bx, by);
			lum->next = *root;
			*root = lum;
		}

		// Link this lumobj into its subsector (always possible).
		root = &dlSubLinks[lum->th->subsector - subsectors];
		lum->ssnext = *root;
		*root = lum;
	}
}

static void intersectionVertex(zvertex_t *out, zvertex_t *a, zvertex_t *b,
						RDBOOL horizontal, float boundary)
{
	float dx = b->x - a->x, dy = b->y - a->y;

	if(horizontal)
	{
		out->y = boundary;
		out->x = a->x + (boundary - a->y) / dy * dx;
	}
	else
	{
		out->x = boundary;
		out->y = a->y + (boundary - a->x) / dx * dy;
	}
}


//===========================================================================
// Rend_SubsectorClipper
//	Clip the subsector to the rectangle. Returns the number of vertices in 
//	the out buffer.
//===========================================================================
static int Rend_SubsectorClipper
	(zvertex_t *out, subsector_t *sub, mobj_t *th, float radius)
{
	static zvertex_t	inverts[MAXPOLYSIDES];
	static zvertex_t	verts  [MAXPOLYSIDES]; // This many vertices, max!
	static RDBOOL		dynsides[MAXPOLYSIDES];
	zvertex_t	*vptr, *a, *b;
	fvertex_t	light[4];

	int			i, clip, aidx, num;
	RDBOOL		clip_needed[4];
	float		box[4];
	float		x = FIX2FLT(th->x);
	float		y = FIX2FLT(th->y);
	mobj_t		lightdummy;

	//-------------------------------------------

	if ( !(num = sub->numverts) ) return 0;

	// v220-33 This cannot work based on subsectors alone
	// Considered map design is required which is why this sepcific test
	// has been added to allow for specific areas to only use this.
	if (th->subsector->sector->special == 20)
	{
		memset(&lightdummy, 0, sizeof(mobj_t));
		lightdummy.x = (int)(sub->midpoint.x * FRACUNIT);
		lightdummy.y = (int)(sub->midpoint.y * FRACUNIT);
		lightdummy.z = sub->sector->floorheight;
		lightdummy.height = FRACUNIT;
		lightdummy.subsector = sub;

		if (!P_CheckSight(th, &lightdummy)) return 0;
	}

	for (i = 0; i < 4; i++) box[i] = FIX2FLT(sub->ssbox[i]);

	light[0].x = light[3].x = x - radius;
	light[0].y = light[1].y = y - radius;
	light[1].x = light[2].x = x + radius;
	light[2].y = light[3].y = y + radius;

	if (   box[BOXLEFT ] >= light[0].x
		&& box[BOXBASE ] >= light[0].y 
		&& box[BOXRIGHT] <= light[2].x
		&& box[BOXTOP  ] <= light[2].y)
	{
		// No clipping needed, the whole subsector is inside the light.
		memcpy(out, sub->verts, sizeof(zvertex_t) * num);
		return num;
	}

	// See which clips are needed.
	clip_needed[CLIP_LEFT ] = light[0].x > box[BOXLEFT ];
	clip_needed[CLIP_RIGHT] = light[2].x < box[BOXRIGHT];
	clip_needed[CLIP_TOP  ] = light[0].y > box[BOXBASE ];
	clip_needed[CLIP_BASE ] = light[2].y < box[BOXTOP  ];

	// Prepare the vertex array.
	memcpy(inverts, sub->verts, sizeof(zvertex_t) * sub->numverts);
	vptr = verts;

	// Horizontal clips first.
	for (clip = 0; clip < 4; clip++)
	{
		if (!clip_needed[clip]) continue;

		// First determine on which side each vertex is.
		for(i = 0; i < num; i++)
		{
			switch(clip)
			{
			case CLIP_TOP  : dynsides[i] = inverts[i].y >= light[0].y; break;
			case CLIP_BASE : dynsides[i] = inverts[i].y <= light[2].y; break;
			case CLIP_LEFT : dynsides[i] = inverts[i].x >= light[0].x; break;
			case CLIP_RIGHT: dynsides[i] = inverts[i].x <= light[2].x; break;
			}
		}

		for (i = 0; i < num; i++)
		{
			if (i) aidx = i - 1; else aidx = num - 1;

			a = inverts + aidx;
			b = inverts + i;

			if (dynsides[aidx] && dynsides[i])
			{
				// This edge is completely within the light, no need to clip it.
				memcpy(vptr++, a, sizeof(*a));
			}
			else if(dynsides[aidx])
			{	
				// Only the start point is inside.
				memcpy(vptr++, a, sizeof(*a));

				// Add an intersection vertex.
				intersectionVertex(vptr++, a, b,
					(RDBOOL)(clip == CLIP_TOP || clip == CLIP_BASE), 
					  clip == CLIP_TOP? light[0].y
					: clip == CLIP_BASE ? light[2].y
					: clip == CLIP_LEFT ? light[0].x
					: light[2].x);
			}
			else if(dynsides[i])
			{
				// First an intersection vertex.
				intersectionVertex(vptr++, a, b,
					(RDBOOL)(clip == CLIP_TOP || clip == CLIP_BASE), 
					  clip == CLIP_TOP  ? light[0].y
					: clip == CLIP_BASE ? light[2].y
					: clip == CLIP_LEFT ? light[0].x
					: light[2].x);
				// Then the end vertex, which is inside.
				memcpy(vptr++, b, sizeof(*b));
			}
		}
		// Prepare for another round.
		if ( !(num = vptr - verts) ) return 0;
		vptr = verts;
		memcpy(inverts, verts, sizeof(zvertex_t) * num);
	}

	memcpy(out, inverts, sizeof(zvertex_t) * num);
	return num;
}

//===========================================================================
// DL_LightIteratorFunc
//===========================================================================
static RDBOOL DL_LightIteratorFunc(lumobj_t *lum, RDBOOL s3dl) 
{
	static zvertex_t dynvtxs[MAXPOLYSIDES];

	seg_t			*seg;
	int				i, num_vertices;
	RDBOOL			doceil = fi.lightCeiling;
	RDBOOL			dofloor = fi.lightFloor;
	RDBOOL			restore = false;
	fixed_t			sscheit = fi.subsector->sector->ceilingheight;	// default

	if (doceil || dofloor)
	{
		float			ceiling, floor;
		float			applyCeiling = 0.f;
		float			applyFloor = 0.f;
		float			srcRadius;
		float			z = FIX2FLT(lum->z) + lum->center;	// Center the Z.

		if (lum->flags & LUMF_MISSILE && lum->th->momz < 0) z += 2.f;

		srcRadius = (float)(lum->radius)/4.f;

		if (srcRadius < 1.0f ) srcRadius = 1.0f;

		if (s3dl && fi.subsector->flats3DArray)
		{
			sector_t *rsec, *lsec;
			int		*flatslist = fi.subsector->flats3DArray;
			int		ceilht, floorht;
			int		nextcht, lastfht;	// Allow light to pass through masked flats
			int		i, count = *(int*)flatslist++;

			lastfht = floorht = fi.subsector->sector->floorheight;
			nextcht = sscheit;

			for (i = 0; i < count; i++)
			{
				lsec = &sectors[flatslist[0]];
				rsec = &sectors[flatslist[1]];	
				ceilht = rsec->smoothfloor + flatslist[2];

				if (z >= FIX2FLT(floorht) && z <= FIX2FLT(ceilht))
				{
					if (R_GetFlat(lsec->ceilingpic)->flflags & FLT_3DMASKED)
					{
						int *fl = flatslist;

						for (int k = i + 1; k < count; k++)
						{
							fl += FLIST_ENTRIES;

							if ( !(R_GetFlat(sectors[fl[0]].ceilingpic)->flflags & FLT_3DMASKED) )
								nextcht = sectors[fl[1]].smoothfloor + fl[2];
						}
					}
					else
					{
						nextcht = ceilht;
					}

					break;
				}

				floorht = ceilht + flatslist[3];

				if ( !(R_GetFlat(lsec->floorpic)->flflags & FLT_3DMASKED) )
					lastfht	= floorht;

				flatslist += FLIST_ENTRIES;
			}

			fi.ffloor  = floor = FIX2FLT(lastfht);
			wiv.bottom = lastfht;
			fi.fceil = ceiling = FIX2FLT(nextcht);
			wiv.top = nextcht;
			restore = true;
			doceil = nextcht < sscheit;
			dofloor = lastfht > fi.subsector->sector->floorheight;
		}
		else
		{
			if (fi.subsector->sector->heightsec >= 0)
			{
				fi.ffloor = FIX2FLT(MIN_OF(fi.subsector->sector->floorheight, sectors[fi.subsector->sector->heightsec].floorheight));
				restore = true;
			}

			floor = fi.ffloor;
			ceiling = fi.fceil;
		}

		// Determine on which side the light is for both the floor and ceiling.
		// $slopes Note: slopes are not used with flats so no need to test for flats3DArray
		// Where encountered the lum's sector is used to avoid adjacent slopes being clipped
		if (doceil && !(ceiling == FIX2FLT(sscheit) && wiv.ceilpic == skyflatnum))
		{
			if (fi.subsector->sector->specFlags2 & SF2_HISLOPE)
				ceiling = FIX2FLT(SL_PointHeight(lum->th->subsector->sector, lum->th->x, lum->th->y, HISLOPE));

			if (z < ceiling)
			{
				if (fi.subsector->s_waterplane > DDMININT)
				{
					if (z > fi.subsector->s_waterplane || ceiling >= fi.subsector->s_waterplane)
						applyCeiling = 1.f;
				}
				else
					applyCeiling = 1.f;
			}
			else if (z < ceiling + srcRadius)
				applyCeiling = 1.f - (z - ceiling)/srcRadius;
		}

		if (dofloor && !(floor == wiv.bottom && wiv.floorpic == skyflatnum))
		{
			if (fi.subsector->sector->specFlags2 & SF2_LOSLOPE)
				floor = FIX2FLT(SL_PointHeight(lum->th->subsector->sector, lum->th->x, lum->th->y, LOSLOPE));

			if (z > floor)
				applyFloor = 1.f;
			else if (z > floor - srcRadius)
				applyFloor = 1.f - (floor - z)/srcRadius;
		}

		if (applyCeiling <= 0 && applyFloor <= 0) goto exit;

		float	ffactor = 0.f, cfactor = 0.f;	// happy compiler (need not be set)
		float	cdiff = ceiling - z, fdiff = z - floor;

		if (lum->flags & LUMF_MISSILE) { cdiff *= MISSILE_FACTOR; fdiff *= MISSILE_FACTOR; }
		else
		if (lum->flags & LUMF_MISSILE1) { cdiff *= MISSILE_FACTOR1; fdiff *= MISSILE_FACTOR1; }

		// Check height differences are tolerable.
		if (applyCeiling > 0 && (cdiff >= lum->radius || (cfactor = LUM_FACTOR(cdiff)*applyCeiling) <= 0.f)) applyCeiling = 0;
		if (applyFloor   > 0 && (fdiff >= lum->radius || (ffactor = LUM_FACTOR(fdiff)*applyFloor  ) <= 0.f)) applyFloor = 0;
		if (applyCeiling <= 0 && applyFloor <= 0) goto exit;

		fi.poly->light = lum;

		// clipping the light's texture removes artefacts (ATM it does not clip walls however - too much overhead?)
		RDBOOL _cliplights = fi.subsector->sector->stptr < 0 	// Don't clip with sloped floors
			&& fi.subsector->numverts < (MAXPOLYSIDES/2);		// or overendowed subsecs

		if (_cliplights)
		{
			num_vertices = Rend_SubsectorClipper(dynvtxs, fi.subsector, lum->th, (float)(lum->radius));

			if (!num_vertices) goto exit;	// The light doesn't reach this subsector
		}

		if (applyCeiling > 0)
		{
			int flags = fi.subsector->sector->specFlags2 & SF2_HISLOPE ? RLPF_REVERSE|RLPF_TILT : RLPF_REVERSE;

			fi.poly->top = fi.fceil;

			if (hasfog) fi.poly->top -= .1f;	// Try to prevent texture fighting

			if (DL_ThingColor(lum, fi.poly->vertices[0].rgba, cfactor))
			{
				// We can add the light poly.
				if (_cliplights)
					RL_PrepareFlat(fi.poly, num_vertices, dynvtxs, flags, NULL);
				else
					RL_PrepareFlat(fi.poly, 0, 0, flags, fi.subsector);

				RL_AddPolyFlat(fi.poly);

				fi.subsector->ssflags |= DDSUBF_DLIGHT_CEILING;	// Mark the ceiling for dlighting
			}
		}

		if (applyFloor > 0)
		{
			int flags = fi.subsector->sector->specFlags2 & SF2_LOSLOPE ? RLPF_NORMAL|RLPF_TILT : RLPF_NORMAL;

			fi.poly->top = fi.ffloor;

			if (hasfog) fi.poly->top += .1f;	// Try to prevent texture fighting

			if (DL_ThingColor(lum, fi.poly->vertices[0].rgba, ffactor))
			{
				// Add the light quads.
				if (_cliplights)
					RL_PrepareFlat(fi.poly, num_vertices, dynvtxs, flags, NULL);
				else
					RL_PrepareFlat(fi.poly, 0, 0, flags, fi.subsector);

				RL_AddPolyFlat(fi.poly);

				fi.subsector->ssflags |= DDSUBF_DLIGHT_FLOOR;	// Mark the floor for dlighting
			}
		}
	}

exit:;

	 if (!s3dl)
	 {
		 for (i = 0; i < fi.segcount; i++)
		 {
			 if ( (seg = fi.segptr[i])->sidedef->lsiflags & LS_TEXPRESENT)
			 {
				 DL_ProcessWallSeg(lum, seg);
			 }
		 }
	 }

	 if (restore)
	 {
		 // restore values altered setting 3Dflats
		 wiv.top    = wiv.oldtop;
		 wiv.bottom = wiv.oldbottom;
		 fi.fceil   = fi.oldfceil;
		 fi.ffloor  = fi.oldffloor;
	 }
	 
	return true;
}

// Box contains the coordinates of the top left and bottom right corners.
static RDBOOL DL_BoxIterator(fixed_t *box, fixed_t *bbc, RDBOOL s3dl, RDBOOL (*func)(lumobj_t*, RDBOOL))
{
	int			x, y;
	lumobj_t	*lum;
	RDBOOL		oey, onedge;

	// The most typical box spans 3x3 blocks
	for (y = bbc[BOXBASE]; y <= bbc[BOXTOP]; y++)
	{
		oey = (y == bbc[BOXTOP] || y == bbc[BOXBASE]);

		for (x = bbc[BOXLEFT]; x <= bbc[BOXRIGHT]; x++)
		{
			// If the edge of the box is in this block, we must do an accurate check.
			onedge = (oey || x == bbc[BOXLEFT] || x == bbc[BOXRIGHT]);

			// Iterate the lights in this block.
			for (lum = *BLK_ROOT(x, y); lum; lum = lum->next)
			{
				if (!M_GroupVis(lum->th->subsector)) continue;

				if (onedge)
				{	// check it is fully contained in the block
					if (   lum->th->x >= box[BOXLEFT] && lum->th->x <= box[BOXRIGHT]
						&& lum->th->y <= box[BOXTOP ] && lum->th->y >= box[BOXBASE ])
					{
						if(!func(lum, s3dl)) return false;
					}
				}
				else if(!func(lum, s3dl)) return false;
			}
		}
	}
	return true;
}

// FIXME: Currently limited to 1 floor and 1 ceiling which limits what's possible with 3D floors
// v220-35 SPAWNVDOOR ceiling glow inhibit test moved
void DL_ProcessSubsector(rendpoly_t *poly, subsector_t *ssec, int segcount, seg_t **segptr)
{
	sector_t	*fs = rv.sect, *temp;
	seg_t		*seg;
	fixed_t		box[4], bbc[4];	// box block coordinates
	fixed_t		dlm;
	RDBOOL		noceil = ssec->sector->ceiloffset != DDMAXINT;	// no ceiling glow with SPAWNVDOOR
	fixed_t		tempfloor, tempceil;
	int			i;

	hasfog = whitefog && (!outsidefog || !setNoFog);

	if (!R3D_ScriptedMap)
	{
		tempfloor = fs->specFlags & SF_NOLOTEX && ssec->dlfloorheight > fs->smoothfloor ?
			ssec->dlfloorheight : fs->smoothfloor;
		tempceil = fs->specFlags & SF_NOHITEX && ssec->dlceilheight < fs->smoothceil ?
			ssec->dlceilheight : fs->smoothceil;

		// v220-31
		if (   ssec->floorsec >= 0											// If floorsec set and
			&& (fs->specFlags & (SF_NOLOTEX|SF_SSLINETRUE)) == SF_NOLOTEX	// the current sector has no low textures and
			&& sectors[ssec->floorsec].smoothfloor < viewz)					// is lower than the current viewheight..	
		{
			temp = &sectors[ssec->floorsec];

			if (temp->heightsec >= 0 && sectors[temp->heightsec].floorheight < fs->floorheight)
				tempfloor = sectors[fs->heightsec].floorheight;
			else
				tempfloor = temp->smoothfloor;
		}
		
		if (cfg.useTorchLight && fs->heightsec < 0)
		{
			if (   ssec->lcsec >= 0
				&& viewz < sectors[ssec->lcsec].ceilingheight)
			{
				noceil = true;
			}
			else
			if (   ssec->ceilsec >= 0
				&& viewz < sectors[ssec->ceilsec].ceilingheight
				&& viewz > sectors[ssec->ceilsec].floorheight)
			{
				temp = &sectors[ssec->ceilsec];
				wiv.ceilpic  = temp->ceilingpic;
				tempceil = temp->smoothceil;
			}
		}
	}
	else
	{
		tempfloor = fs->smoothfloor;
		tempceil = fs->smoothceil;
	}

	wiv.glowhi = wiv.top = tempceil;
	wiv.glowlo = wiv.bottom = tempfloor;

	// Currently only one floor level can have glow
	// Floor glow uses the visible floor/ceilings stored
	// in ssec->ssfloorpic, ssec->ssceilpic. This allows
	// for fake sector floor glow if present.
	wiv.floorpic = ssec->ssfloorpic;
	wiv.ceilpic  = ssec->ssceilpic;
	wiv.do_floor = R_CheckFlatGlowFlag(wiv.floorpic);
	wiv.do_ceil  = R_CheckFlatGlowFlag(wiv.ceilpic );

	wiv.oldtop    = wiv.top;
	wiv.oldbottom = wiv.bottom;

	fi.poly = poly;
	fi.subsector = ssec;

	fi.fceil  = FIX2FLT(wiv.top);
	fi.ffloor = FIX2FLT(wiv.bottom);
	fi.oldfceil  = fi.fceil;
	fi.oldffloor = fi.ffloor;

	// Check if the floor and/or ceiling lighting can be skipped.
	fi.lightFloor   = (wiv.floorpic != skyflatnum) || ssec->flats3DArray;
	fi.lightCeiling = !noceil && ((wiv.ceilpic  != skyflatnum) || ssec->flats3DArray);

	if (   cfg.useTorchLight
		&& !ssec->flats3DArray
		&& fi.lightCeiling
		&& ssec->sector->heightsec >= 0
		&& (ssec->sector->specFlags & 0xC) == 0xC
		&& sectors[ssec->sector->heightsec].ceilingpic == wiv.ceilpic
		&& sectors[ssec->sector->heightsec].ceilingheight > ssec->sector->ceilingheight)
	{
		wiv.glowhi = wiv.top = sectors[ssec->sector->heightsec].ceilingheight;
		fi.fceil  = FIX2FLT(wiv.top);
	}

	fi.segcount = segcount;
	fi.segptr = segptr;

	// View height might prevent us from seeing the floor/ceiling light.
	if (ssec->sector->stptr < 0)
	{
		if(viewz < wiv.bottom) fi.lightFloor   = false;
		if(viewz > wiv.top   ) fi.lightCeiling = false;
	}

	poly->flags = RPF_LIGHT;
	
	if (numLuminous)	// v220-33 accomodate race haxard
	{
		// Prepare the bounding box.
		dlm = dlMaxRad * FRACUNIT;

		box[BOXLEFT ] = M_AddLIMIT(ssec->ssbox[BOXLEFT ], -dlm);
		box[BOXBASE ] = M_AddLIMIT(ssec->ssbox[BOXBASE ], -dlm);
		box[BOXRIGHT] = M_AddLIMIT(ssec->ssbox[BOXRIGHT],  dlm);
		box[BOXTOP  ] = M_AddLIMIT(ssec->ssbox[BOXTOP  ],  dlm);

		bbc[BOXLEFT ] = MAPBS(box[BOXLEFT ], bmaporgx);
		bbc[BOXRIGHT] = MAPBS(box[BOXRIGHT], bmaporgx);
		bbc[BOXBASE ] = MAPBS(box[BOXBASE ], bmaporgy);
		bbc[BOXTOP  ] = MAPBS(box[BOXTOP  ], bmaporgy);

		// Check lights are inside the map
		if (   bbc[BOXRIGHT] >= 0
			&& bbc[BOXTOP  ] >= 0
			&& bbc[BOXLEFT ] < bmapwidth
			&& bbc[BOXBASE ] < bmapheight)
		{
			// The box lies within valid blocks.
			// Clip the box to fit inside the blockmap.
			if(bbc[BOXLEFT ] < 0) bbc[BOXLEFT] = 0;
			if(bbc[BOXBASE ] < 0) bbc[BOXBASE] = 0;
			if(bbc[BOXRIGHT] >= bmapwidth ) bbc[BOXRIGHT] = bmapwidth  - 1;
			if(bbc[BOXTOP  ] >= bmapheight) bbc[BOXTOP  ] = bmapheight - 1;

			if (ssec->flats3DArray)
				DL_BoxIterator(box, bbc, true, DL_LightIteratorFunc);

			DL_BoxIterator(box, bbc, false, DL_LightIteratorFunc);
		}
	}

	// ================== WALL/FLOOR TEXTURE GLOW ===============

	// v220-9 turn off wall-glow if fog is to be drawn
	if (hasfog || !useWallGlow || !BDMFX) return;

	wiv.bottom = ssec->dlfloorheight;
	wiv.top = ssec->dlceilheight;
	wiv.do_floor = R_CheckFlatGlowFlag(wiv.floorpic);
	wiv.do_ceil  = R_CheckFlatGlowFlag(wiv.ceilpic );

	// NOTE: with stacked 3D floors only one glow floor is supported
	//       If one is found lower than the viewz then this is selected
	if (fi.subsector->flats3DArray && !wiv.do_floor)
	{
		sector_t *rsec, *lsec;
		int		*flatslist = fi.subsector->flats3DArray;
		int		ceilht, floorht;
		int		i, count = *(int*)flatslist++;

		floorht = wiv.bottom;

		for (i = 0; i < count; i++)
		{
			lsec = &sectors[flatslist[0]];
			rsec = &sectors[flatslist[1]];	
			ceilht = rsec->smoothfloor + flatslist[2];

			if (fvy >= FIX2FLT(floorht) && fvy <= FIX2FLT(ceilht))
			{
				wiv.top = ceilht;
				wiv.ceilpic = lsec->ceilingpic;
				break;
			}

			floorht = ceilht + flatslist[3];
			wiv.floorpic = lsec->floorpic;
			wiv.bottom = floorht;

			// break on lowest floor found with a glow texture
			if (viewz > floorht && R_CheckFlatGlowFlag(wiv.floorpic))
			{
				wiv.do_floor = true;
				break;
			}

			flatslist += FLIST_ENTRIES;
		}
	}

	// Check glowing planes.
	if (wiv.do_floor || wiv.do_ceil)
	{
		wiv.glowhi = wiv.top;
		wiv.glowlo = wiv.bottom;

		// The wall segments.
		for (i = 0; i < segcount; i++)
		{
			if ( (seg = fi.segptr[i])->sidedef->lsiflags & LS_TEXPRESENT)
			{
				DL_ProcessWallGlow(seg, ssec);
			}
		}
	}
}

//===========================================================================
// DL_InitForNewFrame
//  Creates the dynlight links by removing everything
//  and then linking this frame's luminous objects.
//===========================================================================
void DL_InitForNewFrame()
{
	extern int	lastlight;				// v220-31
	byte strobergb[3];

	dlInited = true;
	numLuminous = 0;	// Clear the luminous object list.

	if (cfg.useTorchLight && lastlight > 10)
	{
		strobergb[0] = strobergb[1] = strobergb[2] = lastlight;	// $r_ambient
		DL_AddLuminous(ppt.viewplrmo, (byte*)strobergb, 1);
	}

	if (ppt.viewplayer->extralight)// && ppt.viewer_fake_sector < 0)
	{
		if (extralight == 2)
			strobergb[0] = strobergb[1] = strobergb[2] = ge_inf.flashintensity;
		else
			strobergb[0] = strobergb[1] = strobergb[2] = ge_inf.flashintensity/2;

		DL_AddLuminous(ppt.viewplrmo, (byte*)strobergb, 2);
	}

	if (!useDynLights)
	{
		if (numLuminous) DL_LinkLuminous();
		return;
	}

	thinker_t	*th;
	mobj_t		*iter;

	// Note: MF2_NOTLIT is set if the thing is not sector linked - R3D
	for (th = thinkercap.next; th != &thinkercap ; th = th->next)
	{
		if (   th->function == (void (__cdecl *)(void))P_MobjThinker
			&& M_GroupVis( (iter = (mobj_t*)th)->subsector) )
		{
			if (iter->flags2 & (MF2_NORENDER|MF2_CORPSEFADE|MF2_NOTLIT)) continue;
		//	if (iter == ppt.viewplrmo) continue;

			if (iter->flags & MF_MISSILE && (iter->momx|iter->momy))
			{
				if (iter->type == MT_ARACHPLAZ || iter->type == MT_PLASMA) continue;
			}

			if (   iter->ddflags & DDMF_ALWAYSLIT
				|| (iter->frame & FF_FULLBRIGHT && !(iter->ddflags & DDMF_DONTDRAW)) )
			{
				if (iter->flags3 & MF3_DYNLIGHT)
				{
					byte level = iter->type == MT_STROBE5 ? (byte)(dynphase.val) :
								 iter->type == MT_STROBEX ? (byte)iter->torchtic : 255;

					if (level)	// no point rendering it if 0
					{
						strobergb[0] = ((iter->inverseAlpha & 0x00FF0000) >> 16) & level;	// R
						strobergb[1] = ((iter->inverseAlpha & 0x0000FF00) >>  8) & level;	// G
						strobergb[2] = ((iter->inverseAlpha & 0x000000FF)      ) & level;	// B

						if (iter->floorz < DDMAXINT)
						{
							iter->z = iter->floorz; DL_AddLuminous(iter, (byte*)strobergb, false);
						}
						if (iter->ceilingz > DDMININT)
						{
							iter->z = iter->ceilingz; DL_AddLuminous(iter, (byte*)strobergb, false);
						}
					}
				}
				else
				{
					DL_AddLuminous(iter, NULL, false);
				}
			}
		}
	}

	// Link (or just clear) the luminous objects into the blockmap.
	DL_LinkLuminous();
}

// Calls func for all luminous objects within the specified range from (x,y).
RDBOOL DL_RadiusIterator(fixed_t ix, fixed_t iy, fixed_t radius, int pspr, 
						  RDBOOL (*func)( lumobj_t*, fixed_t ))
{
	int			x, y, xl, xh, yl, yh;
	fixed_t		dist;
	lumobj_t	*lum;
	
	xl = MAPBS3(ix, bmaporgx, -radius);
	xh = MAPBS3(ix, bmaporgx,  radius);
	yl = MAPBS3(iy, bmaporgy, -radius);
	yh = MAPBS3(iy, bmaporgy,  radius);

	// We can't go outside the blockmap.
	if (xl < 0) xl = 0;
	if (yl < 0) yl = 0;
	if (xh >= bmapwidth ) xh = bmapwidth  - 1;
	if (yh >= bmapheight) yh = bmapheight - 1;

	for (y = yl; y <= yh; y++)	// Traverse the blocks.
	{
		for (x = xl; x <= xh; x++)
		{
			for (lum = *BLK_ROOT(x, y); lum; lum = lum->next)
			{
				if (pspr && lum->flags & LUMF_PLAYER) continue;

				dist = P_ApproxDistance(lum->th->x - ix, lum->th->y - iy);

				if (dist <= radius && !func(lum, dist))
				{
					return false;						
				}
			}
		}
	}
	return true;
}
