//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 G Jackson, Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: rend_list.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D
//  Added code to deal with translucent walls
//  Added new list for alpha flats
//  Added code for triangular side walls created with a slope.
// Revision 2.1.1 build 0 07/Dec/2006 gmj (b211-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 8 (v220-8) 14/Aug/2010 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 10 (v220-10) 25/May/2011 gmj
// Revision 2.2.0 build 16 (v220-16) 28/Sep/2012 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 22 (v220-22) 14/Dec/2013 gmj
// Revision 2.2.0 build 31 (v220-31) 11/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Risen3D rendering Lists
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_render.h"
#include "de_play.h"
#include "de_refresh.h"
#include "de_graphics.h"

#include "r3d_OGLfunc.h"
#include "r3d_compat.h"
#include "r3d_mapinfo.h"
#include "r3d_MAINspec.h"
#include "r3d_items.h"

// MACROS ------------------------------------------------------------------

#define RL_HASH_SIZE		64
#define RL_MASK_SIZE		16

// The max depends on the max number of textures likely
// to be encountered as each list generated is a sequence
// of flats or walls sharing the same texture.
#define MAX_RLISTS			1024
#define MAX_MLISTS			 128

#define DYNLITBIAS		0.0002

#define IFCOL	if(with_col)
#define IFDET	if(with_det)
#define IFTEX	if(with_tex)
#define IFTC	if(with_tex || with_det)

// Rendering List 'has' flags.
#define RLHAS_DLIT		0x1		// Primitives with RPF_DLIT.
#define RLHAS_DETAIL	0x2		// ...with RPF_DETAIL.

// List identifiers.
typedef enum {
	LID_SKYMASK,				// Draw only into Z-buffer.
	LID_NORMAL,					// Normal walls and planes
	LID_NORMAL_DLIT,			// Normal, DLIT with no textures.
	LID_DLIT_NOTEX,				// Just DLIT with no textures (automatical).
	LID_DLIT_TEXTURED,			// DLIT with multiplicative blending.
	LID_DYNAMIC_LIGHTS,
	LID_DETAILS,
	LID_SHADOWS,
	LID_MASKEDFLAT,
	LID_MASKEDWALL,
	LID_TRANS
} lidtype_e;

// Lists for skymask.
enum
{
	RLSKY_FLATS,
	RLSKY_WALLS,
	NUM_RLSKY
};

// Lists for dynamic lights.
enum
{
	RLDYN_FLATS,
	RLDYN_WALLS,
	RLDYN_GLOW,
	NUM_RLDYN
};

// TYPES -------------------------------------------------------------------

#pragma pack(4)

// Primitives are the actual datablocks written to rendering lists.
// Each primhdr begins a block of polygon data that ends up as one or
// more triangles on the screen. This data is internal to the engine
// and can be modified at will. The rendering DLL only receives 
// requests to render sets of primitives from a set of vertices.
typedef struct primhdr_s
{
	// RL_AddPoly expects that size is the first thing in the header.
	// Must be an offset since the list is sometimes reallocated.
	int		size;				// Size of this primitive (zero = n/a).
	void	*ptr;
	
	// Generic data common to all polys.
	uint	flags;				// RPF_*.
	ushort	type;	
	float	texoffx, texoffy;	// Texture coordinates (not normalized).

	lumobj_t	*light;			// For RPF_LIGHT polygons.
	int			shadowradius;	// For RPF_SHADOW polygons.

} primhdr_t;

// The primitive data follows immediately after the header.

// Primvertices are used by the datablocks to store vertex data.
// 3D vertex.
typedef struct primvertex2_s
{
	float	pos[3];				// X, Y, Z coordinates.
	byte	color[4];			// Color of the vertex (RGBA).
	float	dist;				// Distance to the vertex.
} primvertex3_t;

typedef struct texcoord_s
{
	float	s, t;
} texcoord_t;

typedef struct color3_s
{
	float	rgb[3];
} color3_t;

// Data for a quad (wall segment).
typedef struct primquad_s
{
	float	length;				// Length of the wall segment.
	float	top, bottom;		// Top and bottom heights.
	primvertex3_t vertices[2];	// Start and end vertex.
} primquad_t;

// Data for a divquad (wall segment with divided sides).
typedef struct primdivquad_s
{
	primquad_t quad;			// The data for a normal quad.
	int		numdivs[2];			// Number of divisions for start and end.
	float	divs[1];			// Really [numdivs[0] + numdivs[1]].
	// The division heights follow:	first numdivs[0] floats and then
	// numdivs[1] floats.
} primdivquad_t;

// Data for a flat (planes).
typedef struct primflat_s
{
	float	z;					// Height.
	int		numvertices;		// Number of vertices for the poly.		
	primvertex3_t vertices[1];	// Really [numvertices].
} primflat_t;

// The rendering list.
typedef struct rendlist_s
{
	struct rendlist_s *next;
	rendlisttype_e type;		// Quads or flats?
	DGLuint	tex;				// The name of the texture for this list.
	float	texw, texh;			// Width and height of the texture.
	detailinfo_t *detail;		// Detail texture name and dimensions.
	int		size;				// Number of bytes allocated for the data.
	byte	*data;				// Data for a number of polygons (The List).
	byte	*cursor;			// A pointer to data, for reading/writing.
	primhdr_t *last;			// Pointer to the last primitive (or NULL).
	int		has;				
} rendlist_t;

typedef struct listhash_s {
	rendlist_t *first, *last;
} listhash_t;

// v220-3
struct vp_t {
	uint		split_flags;
	byte		calc_col[2][4];
	float		texyu, texyv;
	float		height;
	RDBOOL		calc_tex;
	rendpolyq_t *poly;
} vp;

#pragma pack()

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

vissprite_t *R_NewVisSprite(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// v220-3
extern RDBOOL	setNoFog;
extern byte		fogcolour[4];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// Rendering parameters for detail textures.
float		detailFactor = .5f;
float		detailMaxDist = 256.0f;
float		detailScale = 4.0f;
int			hasmask, transfloor;
int			swappartvis;		// v220-31

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// colour look-up for glow
static byte			glow_lookup[256];	// b211-0

static listhash_t	list_quad_hash[RL_HASH_SIZE];
static listhash_t	list_flat_hash[RL_HASH_SIZE];
static listhash_t	list_mask_hash[RL_MASK_SIZE];
static listhash_t	mask_hash[RL_MASK_SIZE];
static rendlist_t	mask_rlists[NUM_RLSKY];
static rendlist_t	dl_rlists[NUM_RLDYN];
static rendlist_t	shadow_rlist;

static rendlist_t	*ptr_mask_rlists[NUM_RLSKY];
static rendlist_t	*ptr_dl_rlists[NUM_RLDYN];
static rendlist_t	*ptr_shadow_rlist[1] = { &shadow_rlist };

// The rendering state for RL_Draw(Div)Quad and RL_DrawFlat.
static int with_tex, with_col, with_det;
static float rl_texw, rl_texh;
static detailinfo_t *rl_detail;

static color3_t		rldet_color[MAXPOLYSIDES];

static int		Realloc_Addition;
static int		Max_Poly_Size;

static int		max_vpf;
static float	start_val, end_val;
static RDBOOL	chkfog;	// v220-3
static GLclampd	ZBIAS;	// v220-33


// CODE --------------------------------------------------------------------

// Used with fog walls. Negative returns are allowed
float fLength2D(float *c)
{	
	return (fvz - c[VY])*viewsidex - (fvx - c[VX])*viewsidey;
}

// v220-3 Split lengths > 16.f in length having alpha, translucent or fog
//        textures to overcome out of order line rendering problems
static void addVSprite(float *va, float *vb, float length)
{
	uint		visptr = visspritecnt;
	vissprite_t *vis = R_NewVisSprite();
	float	    texoffx, vdis[2];
	float		x, y;

	th_items[TH_WALLMASKED]++;

	vdis[VX] = (va[VX] + vb[VX]) * 0.5f;
	vdis[VY] = (va[VY] + vb[VY]) * 0.5f;

	vis->distance    = Rend_PointDist2D(vdis);
	vis->visflags    = vp.split_flags;
	vis->mo.vmoflags = 0;
	vis->mo.mf = NULL;

	vis->wall.top    = vp.poly->top;
	vis->wall.bottom = vp.poly->bottom;
	vis->wall.vertices[0].pos[VX] = va[VX];
	vis->wall.vertices[0].pos[VY] = va[VY];
	vis->wall.vertices[1].pos[VX] = vb[VX];
	vis->wall.vertices[1].pos[VY] = vb[VY];

	if (!vp.calc_tex)	// fogwall
	{
		int i;

		if (rv.waterplane > DDMININT)
		{
			if (vis->wall.bottom < rv.fpwaterplane)
			{
				// No fog walls underwater
				if (vis->wall.top <= rv.fpwaterplane)
				{
					visspritecnt = visptr;	// reset the counter
					return;
				}

				vis->wall.bottom = rv.fpwaterplane;
			}
		}

		for (i = 0; i < 2; i++)
		{
			float d = fLength2D(vis->wall.vertices[i].pos) - fogval.fogstart;

			if (d <= 0)
			{
				fogcolour[3] = 0;
			}
			else if (d < fogval.fogend)
			{
				// reduce the alpha with decreasing distance
				// use a sin function
				float val = (1.5f + d/fogval.fogend)*PI;

				fogcolour[3] = (byte)((float)(sin(val) + 1)*127.0f);
			}
			else
				fogcolour[3] = 0xFF;

			memcpy(&vis->wall.vertices[i].color, fogcolour, 4);
		}

		return;
	}

	x = vp.poly->vertices[0].pos[VX] - va[VX];
	y = vp.poly->vertices[0].pos[VY] - va[VY];

	if (x < 0) x = -x; if (y < 0) y = -y;

	if      (!x) texoffx = y;
	else if (!y) texoffx = x;
	else texoffx = (float)sqrt((double)(x*x + y*y));

	texoffx += vp.poly->texoffx;

	memcpy(vis->wall.vertices[0].color, vp.calc_col[0], 4);
	memcpy(vis->wall.vertices[1].color, vp.calc_col[1], 4);

	vis->wall.texture = vp.poly->tex;
	x =  (float)vp.poly->texw;
	vis->wall.texc[0][VX] = texoffx / x;
	vis->wall.texc[1][VX] = (texoffx + length) / x;
	vis->wall.texc[0][VY] = vp.texyu;

	// Is this wall cut by a water flat?
	if (rv.waterplane > DDMININT)
	{
		if (   vis->wall.top    > rv.fpwaterplane
			&& vis->wall.bottom < rv.fpwaterplane)
		{
			vissprite_t *nvis = R_NewVisSprite();	// Lower part

			th_items[TH_WALLMASKED]++;

			// Reget vis using visptr as the vissprites
			// address could change when getting nvis
			vis = &vissprites[visptr];

			// do the top first
			vis->wall.bottom = rv.fpwaterplane;
			vis->wall.texc[1][VY] = vp.texyu + (vp.poly->top - rv.fpwaterplane) / vp.height;

			// The lower one will need further tex co-ord adjustments
			nvis->distance    = vis->distance;
			nvis->visflags    = vp.split_flags;
			nvis->mo.vmoflags |= HAS_UNDERWATER;
			nvis->mo.mf = NULL;

			memcpy(&nvis->wall, &vis->wall, sizeof(vissprite_wall_t));

			// Adjust values for lower half of texture
			nvis->wall.top = rv.fpwaterplane;
			nvis->wall.bottom = vp.poly->bottom;
			nvis->wall.texc[0][VY] = (vp.poly->texoffy + vp.poly->top - rv.fpwaterplane) / vp.height;
			nvis->wall.texc[1][VY] = nvis->wall.texc[0][VY] + (rv.fpwaterplane - vp.poly->bottom) / vp.height;

			return;
		}
		
		if (vis->wall.bottom < rv.fpwaterplane)
			vis->mo.vmoflags |= HAS_UNDERWATER;
	}

	vis->wall.texc[1][VY] = vp.texyv;
}

// v220-3
static void segSplitter(float *va, float *vb, float length)
{
	if (length > 16.0f)
	{
		float vn[2];

		// Calculate the centre point
		vn[VX] = (va[VX] + vb[VX]) * 0.5f;
		vn[VY] = (va[VY] + vb[VY]) * 0.5f;

		length /= 2.0f;

		segSplitter(va, vn, length);
		segSplitter(vn, vb, length);
	}
	else
		addVSprite(va, vb, length);
}

//===========================================================================
// RL_AddVISWallPoly
//	Translucent/masked support for R3D
//  v220-3 seg split added to aid with out of order line rendering
//===========================================================================
static void RL_AddVISWallPoly(rendpolyq_t *poly, seg_t *sg)
{
	vp.poly = poly;
	vp.split_flags = VIS_WALL;

	if (poly->flags & RPF_FOGWALL)
	{
		if (fogval.fogend <= fogval.fogstart) return;

		vp.calc_tex = false;
		vp.split_flags |= VIS_FOGWALL;
	}
	else
	{
		// The vertex colours each end are copied from only one of the
		// original vertices to prevent banding.
		// Strictly speaking the split vertices' light levels should be
		// calculated individually but it's just not worth the overhead.
		memcpy(&vp.calc_col[0], poly->vertices[0].rgba, 3);
		memcpy(&vp.calc_col[1], poly->vertices[0].rgba, 3);

		if (poly->flags & RPF_TRANS)
		{
			vp.split_flags |= VIS_TRANS;
			vp.calc_col[0][3] = vp.calc_col[1][3] = 0x80;
		}
		else
			vp.calc_col[0][3] = vp.calc_col[1][3] = 0xFF;

		if (poly->flags & RPF_NOALPHACUT)
			vp.split_flags |= VIS_NOALPHACUT;

		vp.calc_tex = true;
		vp.height = (float)poly->texh;
		vp.texyu  = poly->texoffy / vp.height;
		vp.texyv  = vp.texyu + (poly->top - poly->bottom) / vp.height;

		if (outsidefog && setNoFog) vp.split_flags |= VIS_NOFOG;
	}

	// For ordering purposes see if the seg needs to be split

	// v220-10 unless inhibited split trans, blended alpha or fog textures
	if ( !(poly->flags & RPF_NOSPLIT)
		&& vp.split_flags & (VIS_FOGWALL|VIS_TRANS|VIS_NOALPHACUT))
	{
		segSplitter(poly->vertices[0].pos, poly->vertices[1].pos, poly->length);
	}
	else
	{
		addVSprite(poly->vertices[0].pos, poly->vertices[1].pos, poly->length);
	}
}

//===========================================================================
// RL_VertexColors
//	Color distance attenuation, extralight, fixedcolormap.
//  v220-5 simplified. Torch also removed - fixedcolormap is always 0
// v220-31 support for map torchlight option added (plus light adjustment
//         for masked textures as these cannot be dynlit).
// v220-33 modified to take account of sector->rgb having been scrapped 
//===========================================================================
void RL_VertexColors(rendpoly_vertex_t *vtx, int numvertices, int lightlevel, RDBOOL masked)
{
	if (lightlevel <   0) lightlevel =   0;
	if (lightlevel > 255) lightlevel = 255;

	float	light   = lightlevel * reciprocal255;
	float	minimum = light*light + (light - .63f) * 0.5f;
	float	real;

	if (minimum < 0.f) minimum = 0.f;

	if (cfg.useTorchLight && numvertices != 2)
	{
		if (masked)
		{
			if (numvertices > 2) { if (lightlevel < 128) light = 128.f * reciprocal255; }	// $r_ambient
			else if (lightlevel < 96) light = 96.f * reciprocal255;
		}

		light *= light;
	}

	for (int i = 0; i < numvertices; i++, vtx++)
	{
		if (cfg.setVtxLight || numvertices == 2)
		{
			real = light - (vtx->vpdist - 32.f)/maxLightDistance * (1.f - light);
		}
		else
		{
			real = light;
		}

		if (real < minimum) real = minimum; // Clamp it.

		// Clamp the final light.
		if (real <= 0.f)
		{
			memset(vtx->rgba, 0, 3);
		}
		else if (real >= 1.f)
		{
			memset(vtx->rgba, 0xFF, 3);
		}
		else
		{
			memset(vtx->rgba, (int)(0xFF * real), 3);
		}
	}
}

//===========================================================================
// RL_PrepareFlat
// v220-22 Eliminate unnecessary ssec DDSUBF_MIDPOINT calcs
//===========================================================================
void RL_PrepareFlat(rendpoly_t *poly, int numvrts, zvertex_t *vrts, 
					int flags, subsector_t *subsector)
{
	rendpoly_vertex_t *rpv;
	zvertex_t	*vtx;
	int			lightlevel = poly->vertices[0].rgba[CR];
	RDBOOL		dir    = (flags & RLPF_REVERSE) > 0;
	RDBOOL		tilted = (flags & RLPF_TILT) > 0;
	RDBOOL		unlit  = !(poly->flags & RPF_LIGHT);

	//-------------------------------------

	if (subsector)				// Take the subsector's vertices.
	{
		numvrts = subsector->numverts;
		vrts = subsector->verts;
	}

	poly->type = rp_flat;		// Indicate this is a flat
	rpv = poly->vertices;
	vtx = vrts;

	poly->numvertices = numvrts;

	// The first vertex is always the same: vertex zero.
	rpv->pos[VX] = vtx->x;
	rpv->pos[VY] = vtx->y;
	rpv->pos[VZ] = tilted ? ssecvtxz[vtx->zptr].z[dir] : 0;

	vtx++;
	numvrts--;

	// Normal polys.
	if (unlit) rpv->vpdist = Rend_PointDist2D(rpv->pos);

	if (dir) vtx += (numvrts - 1);

	// Add the rest of the vertices.
	for (; numvrts > 0; numvrts--, (dir ? vtx-- : vtx++))
	{
		rpv++;

		rpv->pos[VX] = vtx->x;
		rpv->pos[VY] = vtx->y;
		rpv->pos[VZ] = tilted ? ssecvtxz[vtx->zptr].z[dir] : 0;

		// Normal polys.
		if (unlit) rpv->vpdist = Rend_PointDist2D(rpv->pos);
	}
	
	// Calculate the color for each vertex.
	if ( !(poly->flags & (RPF_LIGHT|RPF_SKY_MASK|RPF_SKY_MASKLO)) ) 
	{
		RL_VertexColors(poly->vertices, poly->numvertices, lightlevel, poly->flags & RPF_MASKED);
	}
}

// The Rendering Lists -----------------------------------------------------

//===========================================================================
// RL_Init
//	Called once from R_SetupLevel
//===========================================================================
void RL_Init(void)
{
	float	scale = 255.0f/355.0f;
	int		i;

	Max_Poly_Size = 4 + sizeof(primhdr_t)
					+ sizeof(primflat_t)
					+ sizeof(primvertex3_t)*(MAXPOLYSIDES-1);

	Realloc_Addition = Max_Poly_Size * 16; // bytes

	// b211-0 Set the glow lookup table
	// v220-33 keep byte 0 at 0 so no glow is applied with a light level of 0? If so set i to 1	$killglow
	for (i = 0; i < 256; i++)
	{
		int v = (int)((float)(i + 100)*scale);
		if (v > 255) v = 255;
		glow_lookup[i] = (byte)v;
	}

	for(i = 0; i < NUM_RLSKY; i++)
	{
		ptr_mask_rlists[i] = &mask_rlists[i];
	}

	for(i = 0; i < NUM_RLDYN; i++)
	{
		ptr_dl_rlists[i] = &dl_rlists[i];
	}

	memset(mask_rlists  , 0, sizeof(mask_rlists));
	memset(dl_rlists    , 0, sizeof(dl_rlists));
	memset(&shadow_rlist, 0, sizeof(shadow_rlist));

	// Types for the 'static' lists.
	mask_rlists[RLSKY_WALLS].type = rl_quads;
	mask_rlists[RLSKY_FLATS].type = rl_flats;
	dl_rlists[RLDYN_WALLS].type   = rl_quads;
	dl_rlists[RLDYN_GLOW].type    = rl_quads;
	dl_rlists[RLDYN_FLATS].type   = rl_flats;
	shadow_rlist.type = rl_flats;
}

//===========================================================================
// RL_DestroyList
//===========================================================================
static void RL_DestroyList(rendlist_t *rl)
{
	// All the list data will be destroyed.
	if (rl->data) free(rl->data);

	rl->data = NULL;
	rl->cursor = NULL;
	rl->detail = NULL;
	rl->last = NULL;
	rl->size = 0;
	rl->has = false;
}

//===========================================================================
// RL_DeleteLists
//	All lists will be destroyed.
//===========================================================================
void RL_DeleteLists(void)
{
	rendlist_t	*list, *next;
	int			i;

	// Delete all normal lists.
	for(i = 0; i < RL_HASH_SIZE; i++)
	{
		for(list = list_quad_hash[i].first; list; list = next)
		{
			next = list->next;
			RL_DestroyList(list);
			free(list); list = NULL;
		}
	}

	memset(list_quad_hash, 0, sizeof(list_quad_hash));

	for(i = 0; i < RL_HASH_SIZE; i++)
	{
		for(list = list_flat_hash[i].first; list; list = next)
		{
			next = list->next;
			RL_DestroyList(list);
			free(list); list = NULL;
		}
	}

	memset(list_flat_hash, 0, sizeof(list_flat_hash));

	for(i = 0; i < RL_MASK_SIZE; i++)
	{
		for(list = list_mask_hash[i].first; list; list = next)
		{
			next = list->next;
			RL_DestroyList(list);
			free(list); list = NULL;
		}
	}

	memset(list_mask_hash, 0, sizeof(list_mask_hash));

	// Delete all 3D flat mask and flat trans list.
	if (hasmask || transfloor)
	{
		for(i = 0; i < RL_MASK_SIZE; i++)
		{
			for(list = mask_hash[i].first; list; list = next)
			{
				next = list->next;
				RL_DestroyList(list);
				free(list); list = NULL;
			}
		}
	}

	memset(mask_hash, 0, sizeof(mask_hash));
	hasmask = transfloor = false;

	for(i = 0; i < NUM_RLSKY; i++) RL_DestroyList(&mask_rlists[i]);
	for(i = 0; i < NUM_RLDYN; i++) RL_DestroyList(&dl_rlists[i]);
	RL_DestroyList(&shadow_rlist);
}

//===========================================================================
// RL_RewindList
//	Set the R/W cursor to the beginning.
//===========================================================================
static void RL_RewindList(rendlist_t *rl)
{
	rl->cursor = rl->data;
	rl->last = NULL;
	rl->has = 0;
}

//===========================================================================
// RL_ClearLists
//	Called before rendering a frame.
//===========================================================================
void RL_ClearLists(void)
{
	int	i;
	rendlist_t *list;

	for(i = 0; i < RL_HASH_SIZE; i++)
	{
		for(list = list_quad_hash[i].first; list; list = list->next)
			RL_RewindList(list);
		for(list = list_flat_hash[i].first; list; list = list->next)
			RL_RewindList(list);
	}

	for(i = 0; i < RL_MASK_SIZE; i++)
	{
		for(list = list_mask_hash[i].first; list; list = list->next)
			RL_RewindList(list);
		for(list = mask_hash[i].first; list; list = list->next)
			RL_RewindList(list);
	}

	for(i = 0; i < NUM_RLSKY; i++) RL_RewindList(&mask_rlists[i]);
	for(i = 0; i < NUM_RLDYN; i++) RL_RewindList(&dl_rlists[i]);
	RL_RewindList(&shadow_rlist);

	// Dynamic light textures.
	dl_rlists[RLDYN_WALLS].tex = dltexname;
	dl_rlists[RLDYN_FLATS].tex = dltexname;
	dl_rlists[RLDYN_GLOW].tex = glowtexname;
}

//===========================================================================
// RL_GetWallListFor
//  Deals with rl_quads
//===========================================================================
rendlist_t *RL_GetWallListFor(rendpolyq_t *poly)
{
	listhash_t *hash;
	rendlist_t *dest;
	
	if (poly->type == rp_flat)
		Con_Error("Program error [12]: poly type %i", poly->type);

	// Check for specialized rendering lists first.
	if (poly->flags & (RPF_SKY_MASK|RPF_SKY_MASKLO))
	{
		return &mask_rlists[RLSKY_WALLS];
	}
	if (poly->flags & RPF_WALL_GLOW)
	{
		return &dl_rlists[RLDYN_GLOW];
	}
	if (poly->flags & RPF_LIGHT) // Dynamic lights?
	{
		return &dl_rlists[RLDYN_WALLS];
	}

	// Find a suitable list.
	if (poly->flags & RPF_MASKED)
		hash = &list_mask_hash[ (unsigned) poly->tex % RL_MASK_SIZE ];
	else
		hash = &list_quad_hash[ (unsigned) poly->tex % RL_HASH_SIZE ];

	for (dest = hash->first; dest; dest = dest->next)
	{
		if (dest->tex == poly->tex) return dest;
	}

	// Create a new list.
	dest = (rendlist_t*)calloc(1, sizeof(rendlist_t));

	if(hash->last) hash->last->next = dest;
	hash->last = dest;
	if(!hash->first) hash->first = dest;

	// Init the info.
	dest->type   = rl_quads;
	dest->tex    = poly->tex;
	dest->texw   = poly->texw;
	dest->texh   = poly->texh;
	if (!(poly->flags & RPF_MASKED)) dest->detail = poly->detail;

	return dest;
}

//===========================================================================
// RL_GetFlatListFor
//  Deals with rl_flats
//===========================================================================
rendlist_t *RL_GetFlatListFor(rendpoly_t *poly)
{
	rendlist_t *dest;
	listhash_t *hash;

	if (poly->type != rp_flat)
		Con_Error("Program error [13]: poly type %i", poly->type);
	
	// Check for specialized rendering lists first.
	if(poly->flags & RPF_SHADOW)
	{
		return &shadow_rlist;
	}
	if(poly->flags & (RPF_SKY_MASK|RPF_SKY_MASKLO))
	{
		return &mask_rlists[RLSKY_FLATS];
	}
	if(poly->flags & RPF_LIGHT) // Dynamic lights?
	{
		return &dl_rlists[RLDYN_FLATS];
	}

	// Find a suitable list. 
	if (poly->flags & (RPF_MASKED|RPF_TRANS))
	{
		if (poly->flags & RPF_MASKED) hasmask = true;
		if (poly->flags & RPF_TRANS ) transfloor = true;

		hash = &mask_hash[ (unsigned)poly->tex % RL_MASK_SIZE ];
	}
	else
	{
		hash = &list_flat_hash[ (unsigned) poly->tex % RL_HASH_SIZE ];
	}

	for(dest = hash->first; dest; dest = dest->next)
	{
		if (dest->tex == poly->tex) return dest;
	}

	// Create a new list.
	dest = (rendlist_t*)calloc(1, sizeof(rendlist_t));

	if(hash->last) hash->last->next = dest;
	hash->last = dest;
	if(!hash->first) hash->first = dest;

	// Init the info.
	dest->type   = rl_flats;
	dest->tex    = poly->tex;
	dest->texw   = poly->texw;
	dest->texh   = poly->texh;
	dest->detail = poly->detail;

	return dest;
}

//===========================================================================
// RL_AddPolyWall
//	Adds the given poly onto the correct list.
//===========================================================================
void RL_AddPolyWall(rendpolyq_t *poly, seg_t *sg)
{
	rendpoly_vertex_t	*vtx;
	primvertex3_t		*pv2;
	rendlist_t			*li;
	primquad_t			*pq;
	primhdr_t			*hdr;
	int					i, k, m;

	if (poly->type != rp_quad && poly->type != rp_divquad)	// Type check
		Con_Error("Program error [14]: poly type %i", poly->type);

	if (sg && (poly->flags & (RPF_TRANS|RPF_NOALPHACUT|RPF_FOGWALL)) )
	{
		RL_AddVISWallPoly(poly, sg);
		return;
	}

	// Find/create a rendering list for the polygon.
	li = RL_GetWallListFor(poly);

	// First check that the data buffer of the list is large enough.
	if (li->cursor - li->data > li->size - Max_Poly_Size)
	{
		// Allocate more memory for the data buffer.
		int cursor_offset = li->cursor - li->data;
		int last_offset = (byte*) li->last - li->data;

		li->size  += Realloc_Addition;
		li->data   = (byte*)realloc(li->data, li->size);
		li->cursor = li->data + cursor_offset;
		li->last   = (primhdr_t*) (li->data + last_offset);
	}

	// This becomes the new last primitive.
	li->last = hdr = (primhdr_t*) li->cursor;
	li->cursor += sizeof(*hdr);
	hdr->ptr = NULL;

	if (outsidefog && setNoFog)
		hdr->flags = poly->flags|RPF_NOFOG;
	else
		hdr->flags = poly->flags;

	// Check if a detail tex is needed
	if (   r_detail
		&& li->detail
		&& !whitefog
		&& !(poly->flags & (RPF_WALL_GLOW | RPF_SKY_MASK | RPF_SKY_MASKLO | RPF_LIGHT | RPF_SHADOW | RPF_MASKED))
		&& (poly->vertices[0].vpdist < detailMaxDist
		||  poly->vertices[1].vpdist < detailMaxDist))
	{		
		hdr->flags |= RPF_DETAIL;	// Eligible for a detail texture.
		li->has |= RLHAS_DETAIL;
	}

	hdr->type  = (ushort)poly->type;
	hdr->texoffx = poly->texoffx;
	hdr->texoffy = poly->texoffy;
	// The light data is needed for calculating texture coordinates 
	// for RPF_LIGHT polygons (note: RPF_SHADOW is only used with flats
	hdr->light = poly->light;

	if(poly->flags & RPF_DLIT) li->has |= RLHAS_DLIT;

	pq = (primquad_t*) li->cursor;
	pq->top = poly->top;
	pq->bottom = poly->bottom;
	pq->length = poly->length;

	for(i = 0, vtx = poly->vertices, pv2 = pq->vertices; i < 2; i++, vtx++, pv2++)
	{
		pv2->pos[VX] = vtx->pos[VX];
		pv2->pos[VY] = vtx->pos[VY];

		if (poly->flags & (RPF_TRILOW|RPF_TRIHI))
		{
			pv2->pos[VZ] = vtx->pos[VZ];
		}

		pv2->dist = vtx->vpdist;
		if(poly->flags & RPF_WALL_GLOW)
		{
			pv2->color[CR] = glow_lookup[vtx->rgba[CR]];
			pv2->color[CG] = glow_lookup[vtx->rgba[CG]];
			pv2->color[CB] = glow_lookup[vtx->rgba[CB]];
		}
		else
		{
			pv2->color[CR] = vtx->rgba[CR];
			pv2->color[CG] = vtx->rgba[CG];
			pv2->color[CB] = vtx->rgba[CB];
		}
		pv2->color[CA] = 255;
	}

	// Divquads need the division info.
	if (poly->type == rp_divquad)
	{
		primdivquad_t *pdq = (primdivquad_t*) pq;

		for(i = 0, k = 0; i < 2; i++)
		{
			pdq->numdivs[i] = poly->divs[i].num;

			for(m = 0; m < pdq->numdivs[i]; m++)
				pdq->divs[k++] = poly->divs[i].pos[m];
		}
		// Move the cursor forward the necessary amount.
		li->cursor += sizeof(*pdq) + sizeof(float)*(k - 1);
	}
	else
	{
		// Move the cursor forward the necessary amount.
		li->cursor += sizeof(*pq);
	}
	
	// The primitive has been written, update the size.
	hdr->size = li->cursor - (byte*)hdr;

	// A following primhdr will cause this to be set to the required
	// cursor position else 0 will be taken as a list terminator.  
	*(int*) li->cursor = 0;
}

//===========================================================================
// RL_AddPolyFlat
//	Adds the given poly onto the correct list.
//===========================================================================
void RL_AddPolyFlat(rendpoly_t *poly)
{
	rendpoly_vertex_t *vtx;
	rendlist_t		*li;
	primhdr_t		*hdr;
	primvertex3_t	*pv2;
	primflat_t		*pf;
	int				i;
	byte			op;

	//--------------------------------

	if (poly->type != rp_flat)	// Type check
		Con_Error("Program error [15]: poly type %i", poly->type);

	// Find/create a rendering list for the polygon.
	li = RL_GetFlatListFor(poly);

	// First check that the data buffer of the list is large enough.
	if (li->cursor - li->data > li->size - Max_Poly_Size)
	{
		// Allocate more memory for the data buffer.
		int cursor_offset = li->cursor - li->data;
		int last_offset = (byte*) li->last - li->data;

		li->size += Realloc_Addition;
		li->data = (byte*)realloc(li->data, li->size);
		li->cursor = li->data + cursor_offset;
		li->last = (primhdr_t*) (li->data + last_offset);
	}

	if ( !(poly->flags & (RPF_PLANE_GLOW | RPF_LIGHT | RPF_SHADOW)) ) th_items[TH_FLOORTEX]++;

	// This becomes the new last primitive.
	li->last = hdr = (primhdr_t*) li->cursor;
	li->cursor += sizeof(*hdr);

	hdr->ptr = NULL;

	if (outsidefog && setNoFog)
		hdr->flags = poly->flags|RPF_NOFOG;
	else
		hdr->flags = poly->flags;

	// Check if a detail texture is needed
	if (   r_detail
		&& li->detail
		&& !whitefog
		&& !(poly->flags & (RPF_PLANE_GLOW | RPF_SKY_MASK | RPF_SKY_MASKLO | RPF_LIGHT | RPF_SHADOW)))
	{
		if (poly->flags & RPF_PLRINSSEC)
		{
			hdr->flags |= RPF_DETAIL;	// Eligible for a detail texture.
			li->has |= RLHAS_DETAIL;
		}
		else
		{
			// Check if within the minimum distance
			for (i = 0; i < poly->numvertices; i++)
			{
				if (poly->vertices[i].vpdist < detailMaxDist)
				{
					hdr->flags |= RPF_DETAIL;	// Eligible for a detail texture.
					li->has |= RLHAS_DETAIL;
					break;
				}
			}
		}
	}

	hdr->type    = rp_flat;
	hdr->texoffx = poly->texoffx;
	hdr->texoffy = poly->texoffy;
	
	// The light data is needed with RPF_SHADOW and RPF_LIGHT
	// for calculating texture coordinates 
	hdr->light = poly->light;

	pf = (primflat_t*) li->cursor;
	pf->numvertices = poly->numvertices;
	pf->z = poly->top;

	if (water_opacity >= 0xFF)
		poly->flags &= ~RPF_TRANS;

	op = poly->flags & RPF_TRANS ? (byte)water_opacity : 0xFF;

	for(i = 0, vtx = poly->vertices, pv2 = pf->vertices; 
		i < pf->numvertices; i++, vtx++, pv2++)
	{
		pv2->pos[VX] = vtx->pos[VX];
		pv2->pos[VY] = vtx->pos[VY];
		pv2->pos[VZ] = vtx->pos[VZ] + poly->top;

		pv2->dist = vtx->vpdist;

		if(poly->flags & RPF_PLANE_GLOW)
		{
			pv2->color[CR] = glow_lookup[vtx->rgba[CR]];
			pv2->color[CG] = glow_lookup[vtx->rgba[CG]];
			pv2->color[CB] = glow_lookup[vtx->rgba[CB]];
		}
		else
		{
			pv2->color[CR] = vtx->rgba[CR];
			pv2->color[CG] = vtx->rgba[CG];
			pv2->color[CB] = vtx->rgba[CB];
		}

		pv2->color[CA] = op;
	}

	// Move the cursor forward.
	li->cursor += sizeof(*pf) + sizeof(*pv2)*(pf->numvertices - 1);
	
	// The primitive has been written, update the size.
	hdr->size = li->cursor - (byte*)hdr;

	// A following primhdr will cause this to be set to the required
	// cursor position else 0 will be taken as a list terminator.  
	*(int*) li->cursor = 0;

	if(poly->flags & RPF_DLIT) li->has |= RLHAS_DLIT;
}

//===========================================================================
// RL_QuadTexCoords
//===========================================================================
void RL_QuadTexCoords(primhdr_t *prim, primquad_t *quad, texcoord_t *tex, float yoff)
{
	if(prim->flags & RPF_LIGHT)
	{
		// Wallglow needs different texture coordinates.
		if(prim->flags & RPF_WALL_GLOW)
		{
			// The glow texture is uniform along the S axis.
			tex[0].s = 0;
			tex[1].s = 1;
			// texoffx contains the glow height.
			if(prim->texoffx > 0)
			{
				tex[0].t = prim->texoffy/prim->texoffx;
				tex[1].t = tex[0].t + (quad->top - quad->bottom)/prim->texoffx;
			}
			else
			{
				tex[1].t = -prim->texoffy/prim->texoffx;
				tex[0].t = tex[1].t - (quad->top - quad->bottom)/prim->texoffx;
			}
		}
		else if (prim->light->radius)	// A regular dynamic light.
		{
			float dlsize = prim->light->radius * 2.0f;
			float muldl  = DYN_ASPECT/dlsize;

			tex[0].s = -prim->texoffx/dlsize;
			tex[0].t = (prim->texoffy + yoff) * muldl;
			tex[1].s = tex[0].s + quad->length/dlsize;
			tex[1].t = tex[0].t + (quad->top - quad->bottom) * muldl;
		}
	}
	else
	{
		tex[0].s = prim->texoffx/rl_texw;
		tex[0].t = (prim->texoffy + yoff)/rl_texh;
		tex[1].s = tex[0].s + quad->length/rl_texw;
		tex[1].t = tex[0].t + (quad->top - quad->bottom)/rl_texh;
	}
}

//===========================================================================
// RL_QuadDetailTexCoords
//===========================================================================
void RL_QuadDetailTexCoords
	(primhdr_t *prim, primquad_t *quad, texcoord_t *tex, float yoff)
{
	float mul = rl_detail->scale * detailScale;
	int i;

	tex[0].s = prim->texoffx/rl_detail->width;
	tex[0].t = (prim->texoffy + yoff)/rl_detail->height;
	tex[1].s = tex[0].s + quad->length/rl_detail->width;
	tex[1].t = tex[0].t + (quad->top - quad->bottom)/rl_detail->height;
	for(i = 0; i < 2; i++) 
	{
		tex[i].s *= mul;
		tex[i].t *= mul;
	}
}

//===========================================================================
// RL_DetailDistFactor
//===========================================================================
static float RL_DetailDistFactor(primvertex3_t *vertices, int idx)
{
	float f = 1 - vertices[idx].dist 
		/ (rl_detail->maxdist? rl_detail->maxdist : detailMaxDist);
	// Clamp to [0,1].
	if(f > 1) f = 1;
	if(f < 0) f = 0;
	return f;
}

//===========================================================================
// RL_DetailColor
//===========================================================================
static void RL_DetailColor(color3_t *col, primvertex3_t *vertices, int num) 
{
	float mul;
	int i, k;

	// Also calculate the color for the detail texture.
	// It's <vertex-color> * <distance-factor> * <detail-factor>.
	for(i = 0; i < num; i++)
	{
		mul = RL_DetailDistFactor(vertices, i) 
			* detailFactor 
			* rl_detail->strength
			* reciprocal255;	// Could be incorporated into detailFactor

		for(k = 0; k < 3; k++)
			col[i].rgb[k] = vertices[i].color[k] * mul;
	}
}	

//===========================================================================
// RL_DrawQuad
//  Support for sloped sector wall triangles added - R3D
//===========================================================================
void RL_DrawQuad(primhdr_t *prim)
{
	primquad_t		*quad = (primquad_t*) ((byte*)prim + sizeof(*prim));
	primvertex3_t	*vtx = quad->vertices;
	color3_t		detail_color[2];
	texcoord_t		texcoord[2];

	// Dynamic light polygons are the same color all over.
	if(prim->flags & RPF_LIGHT) glColor3ubv(vtx[0].color);

	// Slope triangles --------------------------------------------

	if (prim->flags & (RPF_TRILOW|RPF_TRIHI))
	{
		primquad_t	qd;
		float	z0 = vtx[0].pos[VZ];
		float	z1 = vtx[1].pos[VZ];
		float	yoff;
		RDBOOL	lowquad = false;

		if (z0 == z1) return;

		glBegin(GL_TRIANGLES);

		memcpy(&qd, quad, sizeof(primquad_t));

		// Dynamic light polygons are the same color all over.
		if (prim->flags & RPF_LIGHT) glColor3ubv(vtx[0].color);

		if (prim->flags & RPF_TRILOW)
		{
			if (z0 > z1)
			{
				qd.bottom = qd.top - z0 + z1;

				// Calculate normal texture coordinates.
				IFTEX	RL_QuadTexCoords(prim, &qd, texcoord, 0);

				// Calculate detail texture coordinates and detail color.
				IFDET {
					RL_QuadDetailTexCoords(prim, &qd, texcoord, 0);
					RL_DetailColor(detail_color, vtx, 2);
				}

				// The origin vertex.
				IFCOL	glColor3ubv(vtx[1].color);
				IFTC	glTexCoord2f(texcoord[1].s, texcoord[1].t);
				IFDET	glColor3fv(detail_color[1].rgb);
				glVertex3f(vtx[1].pos[VX], qd.bottom, vtx[1].pos[VY]);

				IFCOL	glColor3ubv(vtx[0].color);
				IFDET	glColor3fv(detail_color[0].rgb);
				IFTC	glTexCoord2f(texcoord[0].s, texcoord[1].t);
				glVertex3f(vtx[0].pos[VX], qd.bottom, vtx[0].pos[VY]);
				
				IFCOL	glColor3ubv(vtx[0].color);
				IFDET	glColor3fv(detail_color[0].rgb);
				IFTC	glTexCoord2f(texcoord[0].s, texcoord[0].t);
				glVertex3f(vtx[0].pos[VX], qd.top, vtx[0].pos[VY]);

				// If required set up the low quad
				if (qd.bottom > quad->bottom)
				{
					yoff = z0 - z1;
					qd.top = qd.bottom, qd.bottom = quad->bottom;
					lowquad = true;
				}
			}
			else
			{
				qd.bottom = qd.top - z1 + z0;

				// Calculate normal texture coordinates.
				IFTEX	RL_QuadTexCoords(prim, &qd, texcoord, 0);

				// Calculate detail texture coordinates and detail color.
				IFDET {
					RL_QuadDetailTexCoords(prim, &qd, texcoord, 0);
					RL_DetailColor(detail_color, vtx, 2);
				}

				// The origin vertex.
				IFCOL	glColor3ubv(vtx[0].color);
				IFTC	glTexCoord2f(texcoord[0].s, texcoord[1].t);
				IFDET	glColor3fv(detail_color[0].rgb);
				glVertex3f(vtx[0].pos[VX], qd.bottom, vtx[0].pos[VY]);
				
				IFCOL	glColor3ubv(vtx[1].color);
				IFDET	glColor3fv(detail_color[1].rgb);
				IFTC	glTexCoord2f(texcoord[1].s, texcoord[0].t);
				glVertex3f(vtx[1].pos[VX], qd.top, vtx[1].pos[VY]);

				IFCOL	glColor3ubv(vtx[1].color);
				IFDET	glColor3fv(detail_color[1].rgb);
				IFTC	glTexCoord2f(texcoord[1].s, texcoord[1].t);
				glVertex3f(vtx[1].pos[VX], qd.bottom, vtx[1].pos[VY]);

				// If required set up the low quad
				if (qd.bottom > quad->bottom)
				{
					qd.top = qd.bottom, qd.bottom = quad->bottom;
					yoff = z1 - z0;
					lowquad = true;
				}
			}
		}
		else	// RPF_TRIHI (z0/z1 heights are negative)
		{
			yoff = quad->top - quad->bottom;

			if (z0 > z1)	// vtx 0 has a single coord
			{
				// We have to draw an inverted triangle to invert the y-coords
				qd.top = qd.bottom;
				qd.bottom = qd.top + z0 - z1;

				if (!(prim->flags & RPF_LIGHT)) yoff += z0 - z1;

				// Calculate normal texture coordinates.
				IFTEX	RL_QuadTexCoords(prim, &qd, texcoord, yoff);

				// Calculate detail texture coordinates and detail color.
				IFDET {
					RL_QuadDetailTexCoords(prim, &qd, texcoord, yoff);
					RL_DetailColor(detail_color, vtx, 2);
				}

				// The origin vertex.
				IFCOL	glColor3ubv(vtx[0].color);
				IFDET	glColor3fv(detail_color[0].rgb);
				IFTC	glTexCoord2f(texcoord[0].s, texcoord[1].t);
				glVertex3f(vtx[0].pos[VX], qd.bottom, vtx[0].pos[VY]);

				IFCOL	glColor3ubv(vtx[1].color);
				IFTC	glTexCoord2f(texcoord[1].s, texcoord[1].t);
				IFDET	glColor3fv(detail_color[1].rgb);
				glVertex3f(vtx[1].pos[VX], qd.bottom, vtx[1].pos[VY]);

				IFCOL	glColor3ubv(vtx[1].color);
				IFDET	glColor3fv(detail_color[1].rgb);
				IFTC	glTexCoord2f(texcoord[1].s, texcoord[0].t);
				glVertex3f(vtx[1].pos[VX], qd.top, vtx[1].pos[VY]);
				
				// If required set up the high quad
				if (quad->top > qd.bottom)
				{
					qd.top = quad->top;
					yoff = (prim->flags & RPF_LIGHT) ? 0 : z0 - z1;
					lowquad = true;
				}
			}
			else	// vtx 1 has a single coord
			{
				// We have to draw an inverted triangle to invert the y-coords
				qd.top = qd.bottom;
				qd.bottom = qd.top + z1 - z0;

				if (!(prim->flags & RPF_LIGHT)) yoff += z1 - z0;

				// Calculate normal texture coordinates.
				IFTEX	RL_QuadTexCoords(prim, &qd, texcoord, yoff);

				// Calculate detail texture coordinates and detail color.
				IFDET {
					RL_QuadDetailTexCoords(prim, &qd, texcoord, yoff);
					RL_DetailColor(detail_color, vtx, 2);
				}

				// The origin vertex.
				IFCOL	glColor3ubv(vtx[1].color);
				IFTC	glTexCoord2f(texcoord[1].s, texcoord[1].t);
				IFDET	glColor3fv(detail_color[1].rgb);
				glVertex3f(vtx[1].pos[VX], qd.bottom, vtx[1].pos[VY]);

				IFCOL	glColor3ubv(vtx[0].color);
				IFDET	glColor3fv(detail_color[0].rgb);
				IFTC	glTexCoord2f(texcoord[0].s, texcoord[0].t);
				glVertex3f(vtx[0].pos[VX], qd.top, vtx[0].pos[VY]);
				
				IFCOL	glColor3ubv(vtx[0].color);
				IFDET	glColor3fv(detail_color[0].rgb);
				IFTC	glTexCoord2f(texcoord[0].s, texcoord[1].t);
				glVertex3f(vtx[0].pos[VX], qd.bottom, vtx[0].pos[VY]);

				// If required set up the high quad
				if (quad->top > qd.bottom)
				{
					qd.top = quad->top;
					yoff = (prim->flags & RPF_LIGHT) ? 0 : z1 - z0;
					lowquad = true;
				}
			}
		}

		glEnd();
		
		// If a triangle wall has been split by the nodesbuilder then
		// a rectangle will have been formed which needs to be drawn.
		// The wall can be imagined as a triangle on a rectangle.
		// This is the case where the triangle's upper/lower limits
		// are different from the sent quad's.
		if (lowquad)
		{
			glBegin(GL_QUADS);

			// Calculate normal texture coordinates.
			IFTEX	RL_QuadTexCoords(prim, &qd, texcoord, yoff);

			// Calculate detail texture coordinates and detail color.
			IFDET {
				RL_QuadDetailTexCoords(prim, &qd, texcoord, yoff);
				RL_DetailColor(detail_color, vtx, 2);
			}

			IFCOL	glColor3ubv(vtx[0].color);
			IFDET	glColor3fv(detail_color[0].rgb);

			IFTC	glTexCoord2f(texcoord[0].s, texcoord[1].t);	
			glVertex3f(vtx[0].pos[VX], qd.bottom, vtx[0].pos[VY]);
				
			IFTC	glTexCoord2f(texcoord[0].s, texcoord[0].t);
			glVertex3f(vtx[0].pos[VX], qd.top, vtx[0].pos[VY]);

			IFCOL	glColor3ubv(vtx[1].color);
			IFDET	glColor3fv(detail_color[1].rgb);

			IFTC	glTexCoord2f(texcoord[1].s, texcoord[0].t);
			glVertex3f(vtx[1].pos[VX], qd.top, vtx[1].pos[VY]);
			
			IFTC	glTexCoord2f(texcoord[1].s, texcoord[1].t);
			glVertex3f(vtx[1].pos[VX], qd.bottom, vtx[1].pos[VY]);

			glEnd();
		}
		return;
	}

	// Standard rectangles --------------------------------------------

	// Calculate normal texture coordinates.
	IFTEX RL_QuadTexCoords(prim, quad, texcoord, 0);

	// Calculate detail texture coordinates and detail color.
	IFDET {
		RL_QuadDetailTexCoords(prim, quad, texcoord, 0);
		RL_DetailColor(detail_color, vtx, 2);
	}

	if (shortListQuads) glBegin(GL_QUADS);

	// Start side.
	IFCOL	glColor3ubv(vtx[0].color);
	IFDET	glColor3fv(detail_color[0].rgb);

	IFTC	glTexCoord2f(texcoord[0].s, texcoord[1].t);
			
	glVertex3f(vtx[0].pos[VX], quad->bottom, vtx[0].pos[VY]);
		
	IFTC	glTexCoord2f(texcoord[0].s, texcoord[0].t);
			
	glVertex3f(vtx[0].pos[VX], quad->top, vtx[0].pos[VY]);
		
	// End side.
	IFCOL	glColor3ubv(vtx[1].color);
	IFDET	glColor3fv(detail_color[1].rgb);


	IFTC	glTexCoord2f(texcoord[1].s, texcoord[0].t);
			
	glVertex3f(vtx[1].pos[VX], quad->top, vtx[1].pos[VY]);
	
	IFTC	glTexCoord2f(texcoord[1].s, texcoord[1].t);
		
	glVertex3f(vtx[1].pos[VX], quad->bottom, vtx[1].pos[VY]);

	if (shortListQuads) glEnd();
}

//===========================================================================
// RL_DrawDivQuad
//	DivQuads are rendered as two triangle fans.
//===========================================================================
void RL_DrawDivQuad(primhdr_t *prim)
{
	primdivquad_t	*divquad = (primdivquad_t*) ((byte*)prim + sizeof(*prim));
	primvertex3_t	*vtx = divquad->quad.vertices;
	float			zpos[2] = { divquad->quad.top, divquad->quad.bottom }, z;
	float			*divpos[2];
	color3_t		detail_color[2];
	texcoord_t		texcoord[2];
	int				i, side, other;

	//-----------------------------------------------

	divpos[0] = divquad->divs;
	divpos[1] = divquad->divs + divquad->numdivs[0];

	// Calculate texture coordinates.
	IFTEX RL_QuadTexCoords(prim, &divquad->quad, texcoord, 0);

	IFDET { 
		RL_QuadDetailTexCoords(prim, &divquad->quad, texcoord, 0);
		RL_DetailColor(detail_color, vtx, 2);
	}

	// Dynamic light polygons are the same color all over.
	if(prim->flags & RPF_LIGHT) glColor3ubv(vtx[0].color);

	// We'll render two fans.
	glBegin(GL_TRIANGLE_FAN);

	// A more general algorithm is used for divquads.
	for(side = 0; side < 2; side++)	// Left->right is side zero.
	{
		other = !side;

		// The origin vertex.
		IFCOL	glColor3ubv(vtx[side].color);
		IFTC	glTexCoord2f(texcoord[side].s, texcoord[side].t);
		IFDET	glColor3fv(detail_color[side].rgb);
				glVertex3f(vtx[side].pos[VX], zpos[side], vtx[side].pos[VY]);

		for(i = 0; i <= divquad->numdivs[other] + 1; i++)
		{
			// The vertex on the opposite side.
			IFCOL	glColor3ubv(vtx[other].color);
			IFDET	glColor3fv(detail_color[other].rgb);
			if(!i) // The top/bottom vertex.
			{
				IFTC glTexCoord2f(texcoord[other].s, texcoord[side].t);
				glVertex3f(vtx[other].pos[VX], zpos[side], vtx[other].pos[VY]);
			}
			else if(i == divquad->numdivs[other] + 1) // The bottom/top vertex.
			{
				IFTC glTexCoord2f(texcoord[other].s, texcoord[other].t);
				glVertex3f(vtx[other].pos[VX], zpos[other], vtx[other].pos[VY]);
			}
			else // A division vertex.
			{
				z = divpos[other][i - 1];
				IFTC { 
					// Calculate the texture coordinate.
					glTexCoord2f(texcoord[other].s, (z - divquad->quad.bottom) / 
						(divquad->quad.top - divquad->quad.bottom)
						* (texcoord[0].t - texcoord[1].t) + texcoord[1].t);
				}					
				glVertex3f(vtx[other].pos[VX], z, vtx[other].pos[VY]);
			}
		}
	}

	glEnd();
}

//===========================================================================
// RL_DrawFlat
//===========================================================================
void RL_DrawFlat(primhdr_t *prim)
{
	uint	Sys_GetRealTime(void);

	primflat_t		*flat = (primflat_t*) ((byte*)prim + sizeof(*prim));
	primvertex3_t	*vtx = flat->vertices;
	float			dlsize, dloffx, dloffy;
	int				i;
	//-------------------------------

	// Dynamic light polygons are the same color all over.
	if (prim->flags & RPF_LIGHT) 
	{
		lumobj_t *lum = prim->light;

		// For light polys the texture offset is global.
		dloffx = FIX2FLT(lum->th->x) + lum->radius;
		dloffy = FIX2FLT(lum->th->y) + lum->radius;
		dlsize = lum->radius*2.0f;

		glColor3ubv(vtx[0].color);
	}

	// Shadow polygons use only one color, too.
	if(prim->flags & RPF_SHADOW)
	{
		prim->shadowradius = (int)prim->light;	// v220-8
		glColor3ubv(vtx[0].color);
	}

	// Detail texture coordinates and color.
	IFDET RL_DetailColor(rldet_color, vtx, flat->numvertices);

	// In a fan all the triangles share the first vertex.
	glBegin(GL_TRIANGLE_FAN);

	for(i = 0; i < flat->numvertices; i++, vtx++)
	{
		if(prim->flags & RPF_LIGHT)
		{
			glTexCoord2f ( (dloffx - vtx->pos[VX])/dlsize,
							(dloffy - vtx->pos[VY])/dlsize ); 
		}
		else if(prim->flags & RPF_SHADOW)
		{
			glTexCoord2f((prim->texoffx - vtx->pos[VX])/prim->shadowradius,
				(prim->texoffy - vtx->pos[VY])/prim->shadowradius);
		}
		else
		{
			IFCOL {
				if (prim->flags & RPF_TRANS)
					glColor4ubv(vtx->color);
				else
					glColor3ubv(vtx->color);
			}
			IFTEX {
				float s = ( vtx->pos[VX] + prim->texoffx)/rl_texw;
				float t = (-vtx->pos[VY] - prim->texoffy)/rl_texh;

				glTexCoord2f(s, t);
			}
			IFDET { 
				glColor3fv(rldet_color[i].rgb);
				glTexCoord2f((vtx->pos[VX] + prim->texoffx)/rl_detail->width
					* detailScale * rl_detail->scale, 
					(-vtx->pos[VY] - prim->texoffy)/rl_detail->height
					* detailScale * rl_detail->scale);
			}
		}

		glVertex3f(vtx->pos[VX], vtx->pos[VZ], vtx->pos[VY]);
	}

	glEnd();
}

//===========================================================================
// RL_DoList
//	This is the worker routine, a helper for RL_RenderList (who does all
//	the thinking and sets up and restores the DGL state).
//===========================================================================
void RL_DoList(lidtype_e lid, rendlist_t *li)
{
	byte		*cursor;
	RDBOOL		blank_dlight = false;
	primhdr_t	*prim, *firstdq = NULL, *prevdq = NULL;
	RDBOOL		isquads;
	RDBOOL		onlydetail = (lid==LID_DETAILS);
	RDBOOL		skipdlit = (lid==LID_NORMAL_DLIT);
	RDBOOL		onlydlit = (lid==LID_DLIT_NOTEX || lid==LID_DLIT_TEXTURED);
	RDBOOL		fogdlit  = chkfog && lid==LID_DLIT_TEXTURED;

	// Is there any point in processing this list?
	if (   (onlydlit   && !(li->has & RLHAS_DLIT  ))
		|| (onlydetail && !(li->has & RLHAS_DETAIL))
		|| (onlydetail && !(rl_detail = li->detail)))
	{
		return;
	}

	// What data to include in the drawing?
	// These control the operation of the primitive drawing routines.
	switch(lid)
	{
	case LID_NORMAL_DLIT:
	case LID_MASKEDFLAT:
	case LID_MASKEDWALL:
	case LID_TRANS:
	case LID_NORMAL:
		with_tex = with_col = true;
		with_det = false;
		GL_Bind(li->tex);
		rl_texw = (float)(li->texw);
		rl_texh = (float)(li->texh);
		break;

	case LID_DLIT_NOTEX:
		with_col = true;
		with_tex = with_det = false;
		break;

	case LID_DLIT_TEXTURED:
		with_tex = true;
		with_det = with_col = false;
		GL_Bind(li->tex);
		rl_texw = (float)(li->texw);
		rl_texh = (float)(li->texh);
		break;

	case LID_DYNAMIC_LIGHTS:
		GL_Bind(li->tex);
		with_tex = true;
		with_col = with_det = false;
		break;

	case LID_DETAILS:
		// li->detail can't be NULL at this stage.
		GL_Bind(li->detail->tex);
		with_det = true;		
		with_col = with_tex = false;
		break;

	default:
		Con_Error("Invalid list %i\n", lid);
	}

	// Does the list contain flats or quads?
	isquads = (li->type == rl_quads);
		
	// Step #1: Render normal quads or flats. If DLIT primitives are
	// encountered, they are skipped and marked for later rendering.
	if (isquads && !shortListQuads) glBegin(GL_QUADS);

	for(cursor = li->data;;)
	{
		// Get primitive at cursor, move cursor to the next one.
		prim = (primhdr_t*) cursor;

		if(!prim->size) break;	// This is the last.
		cursor += prim->size;	// Advance cursor to the next primitive.

		if (lid == LID_TRANS)
		{ 
			if (prim->flags & RPF_TRANS) RL_DrawFlat(prim);
 			continue;
		}

		if (lid == LID_MASKEDFLAT)
		{
			if (prim->flags & RPF_MASKED) RL_DrawFlat(prim);
 			continue;
		}

		// Should we render only the DLIT primitives?
		if (   (onlydlit   && !(prim->flags & RPF_DLIT  ))
			|| (fogdlit    && !(prim->flags & RPF_NOFOG ))
			|| (onlydetail && !(prim->flags & RPF_DETAIL)) )
		{
			continue;
		}

		// What do we have here?
		if (prim->type == rp_divquad) 
		{
			// There are divquads to render. We'll draw them in Step #2.
			// Let's set up a simple linked list through all divquads.
			prim->ptr = NULL; 
			if (!firstdq) firstdq = prim;
			if (prevdq  ) prevdq->ptr = prim;
			prevdq = prim;
		}

		if (   skipdlit
			&& prim->flags & RPF_DLIT
			&& !(outsidefog && !(prim->flags & RPF_NOFOG)))
		{
			// The real texture is multiplied on the lit surface during a later stage.
			blank_dlight = true;
			continue;
		}

		if (prim->type == rp_divquad) continue;

		if (chkfog && prim->flags & (RPF_NOFOG | RPF_MASKED | RPF_TRANS))
			glDisable(GL_FOG);

		// Draw the primitive.
		if(isquads)
			RL_DrawQuad(prim);
		else
		{
			RL_DrawFlat(prim); // Flats are triangle fans.
		}

		if (chkfog && prim->flags & (RPF_NOFOG | RPF_MASKED | RPF_TRANS))
			glEnable(GL_FOG);
	}

	if (isquads && !shortListQuads) glEnd();

// Step #2: Divided walls are drawn separately 
	if(firstdq)
	{
		for(prim = firstdq; prim; prim = (primhdr_t*)prim->ptr) 
		{
			// Don't draw DLIT divquads.
			if (   prim->flags & RPF_DLIT
				&& skipdlit
				&& !(chkfog && !(prim->flags & RPF_NOFOG)))
			{
				continue;
			}

			if (chkfog && prim->flags & RPF_NOFOG)
				glDisable(GL_FOG);

			RL_DrawDivQuad(prim);

			if (chkfog && prim->flags & RPF_NOFOG)
				glEnable(GL_FOG);
		}
	}

	// Start a 2nd pass, of sorts.
	if (blank_dlight)
	{
		glDisable(GL_TEXTURE_2D);
		RL_DoList(LID_DLIT_NOTEX, li);
		glEnable(GL_TEXTURE_2D);
	}
}

//===========================================================================
// 
//===========================================================================
void RL_DoBaseList(lidtype_e lid, rendlist_t *li)
{
	byte		*cursor;
	primhdr_t	*prim;
	RDBOOL		isquads = (li->type == rl_quads);

	// What data to include in the drawing?
	switch(lid)
	{
	case LID_SKYMASK:
		with_tex = with_col = with_det = false;
		break;

	case LID_SHADOWS:
		with_tex = with_col = true;
		with_det = false;
		GL_Bind(GL_PrepareLightTexture());
		break;

	default:
		Con_Error("Invalid list %i\n", lid);
	}

	if (isquads && !shortListQuads) glBegin(GL_QUADS);

	for (cursor = li->data; ;cursor += prim->size)
	{
		if (!(prim= (primhdr_t*) cursor)->size) break;

		// Draw the primitive.
		if (isquads)
			RL_DrawQuad(prim);
		else
			RL_DrawFlat(prim);
	}

	if (isquads && !shortListQuads) glEnd();
}

//===========================================================================
// RL_RenderLists
//	Renders the given lists. RL_DoList does the actual work, we just set
//	up and restore the DGL state here.
//===========================================================================
static void RL_RenderLists(lidtype_e lid, rendlist_t **lists, int num)
{
	int		i;

	if (!num) return;
	
	// If there are just a few empty lists then it
	// is pointless to set and restore the state.
	// except for these which should always have something
	if (lid != LID_NORMAL && lid != LID_NORMAL_DLIT)
	{
		for(i = 0; i < num; i++) if(lists[i]->last) break;
		if(i == num) return; // Nothing to do!
	}

	// Setup the state.
	switch(lid)
	{
	case LID_NORMAL:
		// Disable alpha blending with normal textures
		glDisable(GL_BLEND);
		break;

	case LID_NORMAL_DLIT:
		glDepthRange(ZBIAS, 1);
		// Disable alpha blending with normal textures
		glDisable(GL_BLEND);
		break;

	case LID_MASKEDWALL:
	case LID_MASKEDFLAT:
		glAlphaFunc(GL_GEQUAL, 0.5f);
		glDisable(GL_BLEND);
		if (chkfog) glDisable(GL_FOG);
		break;

	case LID_TRANS:	// waterplanes
		break;

	case LID_DLIT_TEXTURED:
		glDepthFunc(GL_LEQUAL);	
		// Multiply src and dest colors.
		glBlendFunc(GL_ZERO, GL_SRC_COLOR);
		glColor3f(1.0f, 1.0f, 1.0f);
		break;

	case LID_DYNAMIC_LIGHTS:
		if(whitefog)
		{
			glDisable(GL_FOG);
			chkfog = false;
		}
		// This'll allow multiple light quads to be rendered on top of 
		// each other.
		glDepthMask(GL_FALSE);
		glDepthFunc(GL_LEQUAL);
		// The source is added to the destination.
		glBlendFunc(GL_ONE, GL_ONE); 
		break;

	case LID_DETAILS:
		{
			static GLfloat	midGray[4] = { .5f, .5f, .5f, 1 };
			glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, midGray);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
			glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB , GL_INTERPOLATE_ARB);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB , GL_TEXTURE);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB , GL_CONSTANT_ARB);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
			glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB , GL_PRIMARY_COLOR_ARB);
			glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, GL_SRC_COLOR);
			glBlendFunc(GL_DST_COLOR, GL_SRC_COLOR);
			glDepthFunc(GL_LEQUAL);
		}
		break;
	default:
		Con_Error("Invalid list %i\n", lid);
	}

	// Render each of the provided lists.
	for(i = 0; i < num; i++)
	{
		if (lists[i]->last) RL_DoList(lid, lists[i]);
	}


	// Restore state.
	switch(lid)
	{
	case LID_NORMAL:
		glEnable(GL_BLEND);
		break;

	case LID_NORMAL_DLIT:
		glDepthRange(0, 1);
		glEnable(GL_BLEND);
		break;

	case LID_MASKEDWALL:
	case LID_MASKEDFLAT:
		glAlphaFunc(GL_GREATER, 0);
		glEnable(GL_BLEND);
		if (chkfog) glEnable(GL_FOG);
		break;

	case LID_TRANS:
		break;

	case LID_DLIT_TEXTURED:
		glDepthFunc(GL_LESS);	
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		break;

	case LID_DYNAMIC_LIGHTS:
		if(whitefog)
		{
			glEnable(GL_FOG);
			chkfog = outsidefog;
		}
		glDepthMask(GL_TRUE);
		glDepthFunc(GL_LESS);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		break;

	case LID_DETAILS:
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDepthFunc(GL_LESS);
		break;
	default:
		Con_Error("Invalid list %i\n", lid);
	}
}

//===========================================================================
// 
//===========================================================================
void RL_RenderLID_SKYMASK(rendlist_t **lists)
{
	// Setup the state.
	glDisable(GL_TEXTURE_2D);
	glBlendFunc(GL_ZERO, GL_ONE);	// disable color buffer writes.

	if (lists[0]->last) RL_DoBaseList(LID_SKYMASK, lists[0]); // Sky flats
	if (lists[1]->last) RL_DoBaseList(LID_SKYMASK, lists[1]); // Sky walls

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_TEXTURE_2D);
}

//===========================================================================
//
//===========================================================================
void RL_RenderLID_SHADOWS(rendlist_t **lists)
{
	if (lists[0]->last)
	{
		glDepthRange(ZBIAS, 1);
		if(whitefog) glDisable(GL_FOG);
		glDepthMask(GL_FALSE);
		glDepthFunc(GL_LEQUAL);
		glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);

		RL_DoBaseList(LID_SHADOWS, lists[0]);

		glDepthMask(GL_TRUE);
		if(whitefog) glEnable(GL_FOG);
		glDepthFunc(GL_LESS);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDepthRange(0, 1);
	}
}

//===========================================================================
// RL_RenderAllLists
//===========================================================================
void RL_RenderAllLists()
{
	// Multiplicative lights?
	RDBOOL		muldyn = !whitefog || outsidefog;

	// Pointers to all the rendering lists.
	rendlist_t	*it;
	rendlist_t	*rlists[MAX_RLISTS];
	rendlist_t	*rmasklists[MAX_MLISTS];
	rendlist_t	*rwallmasklists[MAX_MLISTS];
	int			numrlists = 0;
	int			nummasklists = 0;
	int			numwallmasklists = 0;
	int			i;

	//------------------------------------

	ZBIAS = (GLclampd)cfg.zbias;	// In case changed in control panel
//	glDepthRange(0, 1);

	chkfog = outsidefog;

	// Collate the rendering lists.
	for (i = 0; i < RL_HASH_SIZE; i++)
	{
		for (it = list_quad_hash[i].first; it; it = it->next)
		{
			if (numrlists == MAX_RLISTS)
				Con_ErrorD("RL_RenderAllLists: Ran out of MAX_RLISTS for quads");

			if (it->last != NULL) rlists[ numrlists++ ] = it;
		}

		for (it = list_flat_hash[i].first; it; it = it->next)
		{
			if (numrlists == MAX_RLISTS)
				Con_ErrorD("RL_RenderAllLists: Ran out of MAX_RLISTS for flats");

			if (it->last != NULL) rlists[ numrlists++ ] = it;
		}
	}

	for (i = 0; i < RL_MASK_SIZE; i++)
	{
		for (it = mask_hash[i].first; it; it = it->next)
		{
			if (nummasklists == MAX_MLISTS)
				Con_ErrorD("RL_RenderAllLists: Ran out of MAX_MLISTS");

			if (it->last != NULL) rmasklists[ nummasklists++ ] = it;
		}

		// v220-10 masked walls rendered directly
		//         They were previously put on the vissprite list
		for (it = list_mask_hash[i].first; it; it = it->next)
		{
			if (nummasklists == MAX_MLISTS)
				Con_ErrorD("RL_RenderAllLists: Ran out of MAX_MLISTS");

			if (it->last != NULL) rwallmasklists[ numwallmasklists++ ] = it;
		}
	}

	// v220-4 All player view rendering starts here

	glClear(GL_DEPTH_BUFFER_BIT);		// This is all the clearing we'll do.

	glMatrixMode(GL_MODELVIEW);			// Set up the model view matrix.
	glRotatef(vpitch, 1.0f, 0.0f, 0.0f);
	glRotatef(vang  , 0.0f, 1.0f, 0.0f);

	// The sky might be visible. If so render the needed hemispheres.
	if (skyhemispheres) Rend_RenderSky();				

	// v220-9 Translate after rendering the hemispheres
	glScalef(1.0f, 1.2f, 1.0f);			// Doom aspect ratio
	glTranslatef(-fvx, -fvy, -fvz);

	if (skyhemispheres) RL_RenderLID_SKYMASK(ptr_mask_rlists);

	// Render the real surfaces of the visible world.
	RL_RenderLists(muldyn? LID_NORMAL_DLIT : LID_NORMAL, rlists, numrlists);
	RL_RenderLists(LID_MASKEDWALL, rwallmasklists, numwallmasklists);

	// Render object shadows.
	RL_RenderLID_SHADOWS(ptr_shadow_rlist);

	// Render dynamic lights.
	RL_RenderLists(LID_DYNAMIC_LIGHTS, ptr_dl_rlists, NUM_RLDYN);

	if (cfg.glflush) glFlush();

	// Apply the dlit pass?
	if(muldyn)
	{
		RL_RenderLists(LID_DLIT_TEXTURED, rlists, numrlists);
		if (cfg.glflush) glFlush();
	}

	// Render the detail texture pass?
	if(r_detail && has_detail) RL_RenderLists(LID_DETAILS, rlists, numrlists);

	// Render 3D masked flats
	if (hasmask) RL_RenderLists(LID_MASKEDFLAT, rmasklists, nummasklists);

	if (transfloor)
	{
		// v220-4 Render vissprites/particles checking for waterplanes
		//        Strictly speaking the particles should be distance ordered
		//        with the sprites and both rendered on the same list but the
		//        overhead is not justified given it works well enough in the
		//        main without doing this.

		// set order depending on viewpoint
		int n = underwater ? -1 : 1;

		if (swappartvis)
		{
			PG_Render(-n);											// Draw particles  +above -under water
			Rend_Draw_vissprites(-n);								// draw vissprites +above -under water
			RL_RenderLists(LID_TRANS, rmasklists, nummasklists);	// draw translucent flats
			PG_Render(n);											// Draw particles  +above -under water
			Rend_Draw_vissprites(n);								// draw vissprites +above -under water
		}
		else
		{
			Rend_Draw_vissprites(-n);								// draw vissprites +above -under water
			PG_Render(-n);											// Draw particles  +above -under water
			RL_RenderLists(LID_TRANS, rmasklists, nummasklists);	// draw translucent flats
			Rend_Draw_vissprites(n);								// draw vissprites +above -under water
			PG_Render(n);											// Draw particles  +above -under water
		}
	}
	else
	{
		// Normal pass
		if (swappartvis)
		{
			PG_Render(0);				// Draw all particles.
			Rend_Draw_vissprites(0);	// Draw all vissprites
		}
		else
		{
			Rend_Draw_vissprites(0);	// Draw all vissprites
			PG_Render(0);				// Draw all particles.
		}
	}
}

int CCmdSwapRenderingOrder(int argc, char **argv)
{
	swappartvis = !swappartvis;
	Con_FPrintf(CBLF_CYAN, "vissprite/particle rendering order : %s\n", swappartvis ? "SWAPPED" : "NORMAL");
	return true;
}