//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2018 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: rend_model.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified as required for R3D
//  missileBlend removed as always on
//  LUMF_MISSILE flag added
//  RENDER_AS_TRIANGLES code deleted
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 24 (v220-24) 20/Mar/2014 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 33 (v220-33) 08/Jul/2018 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//	shininess distance test removed
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  MD2/DMD renderer
//
//-----------------------------------------------------------------------------

//** Note: Light vectors and triangle normals are considered to be
//** in a totally separate, right-handed coordinate system.

//** There is some more confusion with Y and Z axes as the game uses 
//** Z as the vertical axis whereas this is the Y axis (Z is depth)
//** as used by the rendering code and model definitions.

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_render.h"
#include "de_play.h"
#include "de_refresh.h"
#include "de_graphics.h"
#include "de_misc.h"

#include "m_winmem.h"
#include "rend_main.h"
#include "r3d_OGLfunc.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

#define ASINF(x)	(float)asin(x)
#define ATAN2F(y,x)	(float)atan2((double)y, (double)x)
#define DOT3LIGHT(a, b)	(a[0]*a[0]*b + a[1]*a[1]*b + a[2]*a[2]*b)
#define INVDOT3LIGHT(a)	(-a[0]*a[0] + -a[1]*a[1] + -a[2]*a[2])
#define INVDOTPROD2(a, b)	(-a[0]*b[0] + a[1]*b[1])
//#define INVDOTPROD3(a, b)	(-a[0]*b[0] + a[1]*b[2] + a[2]*b[1])

// TYPES -------------------------------------------------------------------

typedef struct {
	int			used;
	fixed_t		dist;		// Only an approximation.
	lumobj_t	*lum;
	float		vector[3];	// Light direction vector.
	float		color[3];	// How intense the light is (0..1, RGB).
	float		intensity;

} mlight_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

mvertex_t	*mvertices;
int			maxnumModelVtx;

float	set_intensity  = 0;
int		modelLights = 6;
extern float	evenLightFactor;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// Add 1 for world light
static mlight_t mlights[MAX_MODEL_LIGHTS + 1];

static int worldLightSet;

// FIXME Currently the light direction is fixed
static float worldLightVec[3]   = { .267261f, .534522f, .801783f };

// Parameters for the modelLighter.
static vissprite_t *mlSpr;

// CODE --------------------------------------------------------------------

static void scaleAmbientRgb(float *out, byte *in, float mul)
{
	int i;
	
	if (mul < 0) mul = 0; else if (mul > 1) mul = 1;

	for (i = 0; i < 3; i++)
	{
		float d = in[i] * mul * reciprocal255;

		if (out[i] < d) out[i] = d;
	}
}

static void scaleFloatRgb(float *out, byte *in, float mul)
{
	memset(out, 0, sizeof(float)*3);
	scaleAmbientRgb(out, in, mul);
}

static fixed_t	z_centre;
static int		hasnolight;
static int		lightlistfull;
static int		lightLimit;

//===========================================================================
// modelLighter
//===========================================================================
static RDBOOL modelLighter(lumobj_t *lum, fixed_t xy_dist)
{
	mlight_t	*light;

	fixed_t		z_dist = z_centre - (lum->z + (int)(FRACUNIT*lum->center));
	fixed_t		dist = P_ApproxDistance(xy_dist, z_dist);
	fixed_t		maxDist = 0;
	float		intensity;

	int			i, newIndex = 0;

	//---------------------------------------

	// see if light can be skipped
	if (   !lum->radius			// nothing to light
		|| !dist				// too near
	//	||  dist > dlMaxRadft	// too distant

		// if it's a missile and outside a different radius
		|| ((lum->flags & (LUMF_MISSILE|LUMF_MISSILE1)) && dist > dlMaxMissileRadft)

		// if not to be lit other than by a missile
		|| (!(lum->flags & (LUMF_MISSILE|LUMF_MISSILE1)) && hasnolight)

		// if intensity is too low
		|| (intensity = 2.0f - FIX2FLT(dist)/(float)lum->radius) <= 0)
	{
		return true;
	}

	if (!lightlistfull)
	{
		// If the list is not exhausted just add the light to the list.
		for (i = worldLightSet, light = mlights + i; i < lightLimit; i++, light++)
		{
			if (!light->used)
			{
				// Add the light to the list
				light->lum = lum;
				light->dist = dist;
				light->used = true;
				light->intensity = intensity;
				return true;
			}
		}
	}
	
	lightlistfull = true;

	// The list is full so get the farthest light on the current list
	// (skip the worldlight)
	for (i = worldLightSet, light = mlights + i; i < lightLimit; i++, light++)
	{
		if (light->dist > maxDist)
		{
			maxDist = light->dist;
			newIndex = i;
		}
	}

	// See if this lumobj is close enough to make it to the list.
	if (dist < maxDist)
	{
		// The new light is closer. Replace the old max.
		mlights[newIndex].lum = lum;
		mlights[newIndex].dist = dist;
		mlights[newIndex].intensity = intensity;
	}

	return true;
}

//===========================================================================
// R_GetVisibleFrame
//===========================================================================
dmd_frame_t *R_GetVisibleFrame(modeldef_t *mf, int subnumber, int mobjid, int from)
{
	model_t	*mdl = modellist[mf->sub[subnumber].model];
	int		index = mf->sub[subnumber].frame;

	if (!mdl->info.numFrames) return NULL;
	
	if (mf->flags & MFF_IDFRAME)
		index += mobjid % mf->sub[subnumber].framerange;

	if (index >= mdl->info.numFrames)
	{
		Con_Error("R_GetVisibleFrame [%i]: Frame index out of bounds.\n"
			"  mobjid %i index %i numframes %i\n"
			"  Model: path %s : name %s)",
			from,
			mobjid, index, mdl->info.numFrames,
			mdl->fullpathName ? mdl->fullpathName : "none",
			mdl->modelName    ? mdl->modelName    : "");
	}
	return mdl->frames + index;
}

extern int dlInited;

//===========================================================================
// R_RenderModel
//  R3D - modified as needed
//  Note: a sub model's existence must have been pre-checked
//===========================================================================

void Rend_RenderModel(vissprite_t *spr, modeldef_t *mf,	submodeldef_t *smf, int number, RDBOOL psprite)
{
	modeldef_t		*nextmf;
	model_t			*mdl;

	dmd_frame_t		*frame;
	dmd_frame_t		*nextframe = NULL;
	dmd_vertex_t	*vtx, *nextvtx;

	mvertex_t		*mVtx;
	mlight_t		*light;

	RDBOOL			lightVertices;

	float			shininess;
	float			*shinycol;

	float			ambientColor[4] = { 1.0f, 1.0f, 1.0f, 0 };
	float			modelCenter[3];

	float			alpha, yawAngle, pitchAngle;
	float			cosrot, sinrot;
	float			inv, interpos, endpos;
	float			sh_ang, sh_pnt, *nptr;
	float			mdl_nyaw, mdl_npitch;
	float			dist;

	int				activeLod = 0;
	int				mflags;

	int				pass, i, k, c;
	int				numVertices;
	int				useskin = -1;

	ushort			rendFlags = 0;

	//----------------------------------------------

	if (   !(mdl = modellist[smf->model])
		|| !(frame = R_GetVisibleFrame(mf, number, spr->mo.think_id, 5)) )
	{
		Con_ErrorD("Model fault [2]: models have not been properly installed");
	}

	mflags = smf->flags;
	if (mflags & MFF_DISABLE_Z_WRITE) rendFlags |= RF_NODEPTH;

	nextmf = spr->mo.nextmf;
	shininess = mf->def->sub[number].shiny;
	shinycol = mf->def->sub[number].shinycolor;
	lightVertices = dlInited && !(spr->visflags & (VIS_FLAME|VIS_STEALTH));
	lightLimit = 0;

	// Distance reduces shininess (if not truly shiny).
	if (   shininess > 0.0f && shininess < 1.0f
		&& spr->distance > 512)
	{
		shininess = 0.0f;	// Too far away.
	}

	// Init. Set the texture for pass 0.
	if (shininess < 1.0f)
	{
		if (mflags & MFF_SKINTRANS)
		{
			// use translation (overrides all other selection types)
			useskin = (spr->mo.flags & DDMF_TRANSLATION) >> DDMF_TRANSSHIFT;
		}
		else
		{
			if (mflags & MFF_SELSKIN)
			{
				i = (spr->mo.selector >> DDMOBJ_SELECTOR_SHIFT)
					& mf->def->sub[number].selskinbits[0];	// Selskin mask

				c = mf->def->sub[number].selskinbits[1];	// Selskin shift

				if(c > 0) i >>= c; else i <<= -c;
				if(i > 7) i = 7;							// Maximum number of skins for selskin.
				if(i < 0) i = 0;							// Improbable (impossible?), but doesn't hurt.

				useskin = mf->def->sub[number].selskins[i];
			}
			else
			{
				useskin = smf->skin;
			}
			
			// Is there a skin range for this frame?
			// (During model setup skintics and/or skinrange are set to > 0)
			if (smf->skinrange > 1)
			{
				// What rule to use for determining the skin?
				if (spr->torchtic || spr->skintic)
				{
					float skintics = mf->skintics + (float)spr->skintic/35.0f;

					useskin +=
						(int)((float)(leveltime + (int)spr->torchtic)/skintics) % smf->skinrange;
				}
				else
				{
					useskin +=
						(mflags & MFF_IDSKIN ? spr->mo.think_id : leveltime/mf->skintics) % smf->skinrange;
				}
			}
		}

		// v220-4 Use attack skins
		if (   mflags & MFF_ATTACKSKIN				// if attack skin is asked for
			&& spr->visflags & VIS_LITSKIN			// and in attack state
			&& useskin < 4							// and current skin index is less than 4
			&& mdl->info.numSkins > useskin + 4)	// and attack skins are available
		{
			useskin += 4;							// then add 4 to get the attack skin index
		}

		if (useskin >= mdl->info.numSkins) useskin %= mdl->info.numSkins;
	}

	if (useskin < 0 && shininess <= 0) return;	// Hmm..

	interpos = spr->mo.inter;

	// Scale interpos. Intermark becomes zero and endmark becomes one.
	// (Full sub-interpolation!) But only do it for the standard interrange.
	// If a custom one is defined, don't touch interpos.
	if (mf->interrange[0] == 0 && mf->interrange[1] == 1)
	{
		if(mf->internext) 
			endpos = mf->internext->intermark;
		else
			endpos = 1;

		interpos = (interpos - mf->intermark) / (endpos - mf->intermark);		
	}

	// Check for possible interpolation.
	if (   spr->mo.nextmf
		&& !(smf->flags & MFF_DONT_INTERPOLATE)
		&& spr->mo.nextmf->sub[number].model == smf->model)
	{
		nextmf = spr->mo.nextmf;
		nextframe = R_GetVisibleFrame(nextmf, number, spr->mo.think_id, 6);
		nextvtx = nextframe->vertices;
	}

	// Light level/alpha
	if (spr->visflags & VIS_STEALTH)	// b211-0 Takes precedence
	{
		alpha = .33f;
	}
	else
	{
		// Submodels can also define a custom Transparency level.

		if (spr->mo.flags & DDMF_BRIGHTSHADOW || mflags & MFF_BRIGHTSHADOW)
		{
			alpha = .80f;	//204;	// 80 %.
			rendFlags |= RF_SPECULAR;
		}
		else if(mflags & MFF_BRIGHTSHADOW2)
		{
			alpha = 1 - smf->alpha/255.0f; // NOTE: CustomAlpha
			rendFlags |= RF_SPECULAR;
		}
		else if(mflags & MFF_DARKSHADOW)
		{
			alpha = 1 - smf->alpha/255.0f; // NOTE: CustomAlpha
			rendFlags |= RF_DARK;
		}
		else if (spr->mo.flags & DDMF_SHADOW || mflags & MFF_SHADOW2)
			alpha = .2f; 
		else if(spr->mo.flags & DDMF_ALTSHADOW || mflags & MFF_SHADOW1)
			alpha = .62f;	
		else
			alpha = 1 - smf->alpha/255.0f; // NOTE: CustomAlpha

		// More custom alpha?
		if(spr->mo.alpha >= 0) alpha *= spr->mo.alpha;
	}

	if (   whitefog
		&& !(spr->mo.vmoflags & HAS_FOGUNSET)
		&& (mflags & (MFF_BRIGHTSHADOW2|MFF_DARKSHADOW)
		|| spr->mo.vmoflags & HAS_NOFOG)
		&& spr->distance > 100.0f)
	{
		alpha *= (100.0f/spr->distance);
	} 

	if (alpha <= 0.0f) return; // Fully transparent.
	if (alpha >  1.0f) alpha = 1.0f;

	yawAngle = spr->mo.yaw;
	pitchAngle = spr->mo.pitch;

	// v220-5 used for lighting
	z_centre = ((spr->mo.gz + spr->mo.gzt) >> 1) + mf->zmax + mf->zoff;

	// Coordinates to the center of the model.
	modelCenter[VX] = FIX2FLT(spr->mo.gx) + mf->offset[VX] + spr->mo.visoff[VX];
	modelCenter[VY] = FIX2FLT(spr->mo.gy) + mf->offset[VZ] + spr->mo.visoff[VY];
	modelCenter[VZ] = FIX2FLT( (spr->mo.gz + spr->mo.gzt) >> 1 )
										  + mf->offset[VY] + spr->mo.visoff[VZ];

	ambientColor[3] = alpha;

	// ambientColors are preset to 1.0f. These defaults values
	// are used if the following test fails.
	if ( !(mflags & MFF_FULLBRIGHT || spr->mo.lightlevel == -1) )
	{
		rendpoly_vertex_t tempvtx;
		float	factor = reciprocal255;
		int		ll;

		// v220-33 adjust values for pspr lighting
		if (spr->visflags & VIS_PSPR)
		{
			ll = (r_ambient > ppt.viewer_lightlevel) ?
						r_ambient : ppt.viewer_lightlevel;

			if (cfg.useTorchLight && ll < 156) ll = 156;

			tempvtx.vpdist = 0;
		}
		else
		{
			// v220-31 alternative sheme preferred by testers
			if (spr->mo.lightlevel < 0) 
			{
				ll = -spr->mo.lightlevel + 80;

				if (ll < 0) ll = 128; else if (ll > 255) ll = 255;

				tempvtx.vpdist = 0;
			}
			else
			{
				factor = 1.f/150.f;
				tempvtx.vpdist = spr->distance; //Rend_PointDist2D(spr->mo.v1);
				ll = (r_ambient > spr->mo.lightlevel) ?
							r_ambient : spr->mo.lightlevel;
			}
		}

		RL_VertexColors(&tempvtx, 1, ll, false);

		// This way the distance darkening has an effect.
		for (i = 0; i < 3; i++)
			ambientColor[i] = (float)((byte)tempvtx.rgba[i]) * factor;

		worldLightSet = lightlistfull = 0;

		if (lightVertices)
		{
			memset(mlights, 0, sizeof(mlights));

			if ( !(spr->visflags & VIS_NOGLIGHT) )
			{
				// The model should be lit with world light. 
				memcpy(mlights->vector, worldLightVec, sizeof(mlights->vector));

				mlights->used = 1; worldLightSet = 1;

				// Copy half the actual ambient light that hits the
				// object to prevent it swamping any dynlights.
				for (i = 0; i < 3; i++)
					mlights[0].color[i] = (ambientColor[i] *= 0.5f);
			}

			// Rotate the light direction to model space.
			M_RotateVector(mlights[0].vector, -yawAngle, -pitchAngle);

			if ( !(spr->visflags & (VIS_SPLASH|VIS_NOTLIT)) )
			{				
				lightLimit = worldLightSet + modelLights;

				// Add extra light using dynamic lights.
				if (useDynLights && modelLights)
				{
					lumobj_t *lum;
					float	 intensity, lightCenter[3];

					mlSpr = spr;
					hasnolight = spr->mo.vmoflags & HAS_NOLIGHT;

					// Find the nearest sources of light. They will be used to
					// light the vertices a bit later. First initialize the array.
					if (spr->visflags & (VIS_PSPR|VIS_SHOOTABLE))
						DL_RadiusIterator(spr->mo.gx, spr->mo.gy, 320*FRACUNIT, spr->visflags & VIS_PSPR, modelLighter);
					else
						DL_RadiusIterator(spr->mo.gx, spr->mo.gy, dlMaxRadft, false, modelLighter);

					// Calculate the light vectors
					for (i = worldLightSet, light = mlights + i; i < lightLimit; i++, light++)
					{
						if (!light->used) break;

						lum = light->lum;
						intensity = light->intensity;

						// The center of the light source.
						lightCenter[VX] = Q_FIX2FLT(lum->th->x);
						lightCenter[VY] = Q_FIX2FLT(lum->th->y);

						if (lum->flags&LUMF_PLAYER)
							intensity = 1.f - M_AccurateDistanceFP(lightCenter[VX] - spr->mo.v1[VX], lightCenter[VY] - spr->mo.v1[VY])/(float)lum->radius;
						else if (spr->mo.intensity)
							intensity *= FIX2FLT(spr->mo.intensity);
						else if (set_intensity)
							intensity *= set_intensity;

						if (intensity > 1.0f) intensity = 1.0f;

						intensity *= reciprocal255;

						// v220-24 centre offset used when adjusting light position for
						//         map architecture/amodels. This should be ignored if
						//         negative for hud weapon/enemies
						if (spr->visflags & (VIS_PSPR|VIS_SHOOTABLE))
						{
							lightCenter[VZ] = Q_FIX2FLT(lum->z);
							if (lum->center > 0) lightCenter[VZ] += lum->center;
						}
						else
							lightCenter[VZ] = Q_FIX2FLT(lum->z) + lum->center;

						// Calculate the normalized direction vector pointing out of the model.
						// v220-4 Adjust vector if it's a thing's attack light (but only for bullet firing)
						if (   spr->visflags & VIS_ATKLIGHT
							&& lum->th->target
							&& lum->th->thinker.id == spr->mo.think_id)
						{
							float dx = Q_FIX2FLT(lum->th->target->x) - (lightCenter[VX] + mf->visualradius);
							float dy = Q_FIX2FLT(lum->th->target->y) - (lightCenter[VY] + mf->visualradius);
							float dz = Q_FIX2FLT(lum->th->target->z) - lightCenter[VZ];
							float d  = (float)sqrt((double)(dx*dx + dy*dy + dz*dz));

							light->vector[VX] = dx/d; light->color[VX] = lum->rgb[VX] * intensity;
							light->vector[VY] = dy/d; light->color[VY] = lum->rgb[VY] * intensity;
							light->vector[VZ] = dz/d; light->color[VZ] = lum->rgb[VZ] * intensity;
						}
						else if (light->dist)	// should not be 0 but better to make sure
						{
							dist = FIX2FLT(light->dist);

							for (c = 0; c < 3; c++) 
							{
								light->vector[c] = (lightCenter[c] - modelCenter[c]) / dist;	// direction
								light->color[c] = lum->rgb[c] * intensity;						// color of the light.
							}
						}

						// We must transform the light vector to model space.
						M_RotateVector(light->vector, -yawAngle, -pitchAngle);
					}
				}
			}
		}
	}

	// Should we prepare for shiny rendering?
	if (shininess > 0.0f)
	{
		float delta[3];

		// Calculate normalized (0,1) model yaw and pitch.
		// Note: viewaligned is only true with hud weapons
		mdl_nyaw   = (spr->mo.viewaligned ? spr->mo.v2[VX] : yawAngle   ) / 360.0f;
		mdl_npitch = (spr->mo.viewaligned ? spr->mo.v2[VY] : pitchAngle ) / 360.0f;
		
		// Are they in range?
		while(mdl_nyaw   < 0.0f) mdl_nyaw++;
		while(mdl_nyaw   > 1.0f) mdl_nyaw--;
		while(mdl_npitch < 0.0f) mdl_npitch++;
		while(mdl_npitch > 1.0f) mdl_npitch--;

		delta[VX] = modelCenter[VX] - fvx;
		delta[VY] = modelCenter[VY] - fvz;
		delta[VZ] = modelCenter[VZ] - fvy;
		sh_ang = ATAN2F(delta[VZ], M_ApproxDistancef(delta[VX], delta[VY])) / PI + 0.5f; // sh_ang is [0,1]
		sh_pnt = ATAN2F(delta[VY], delta[VX]) / (2*PI);
	}

	// Prepare the vertices (calculate interpolated coordinates,
	// lighting, texture coordinates, normals).
	numVertices = mdl->info.numVertices;

	// Calculate a suitable LOD factor (Level of Detail)
	if (mdl->vertexUsage && mdl->info.numLODs > 1)
	{
		extern float lodFactor;

		// Determine the LOD we will be using.
		if ( (activeLod = (int) (lodFactor * spr->distance)) >= mdl->info.numLODs)
			activeLod = mdl->info.numLODs - 1;
	}

	inv = 1.0f - interpos;

	if (spr->mo.rotangle)
	{
		float rotyaw =  -spr->mo.rotangle /180.0f * PI;

		cosrot = (float)cos(rotyaw); 
		sinrot = (float)sin(rotyaw);
	}
	else
	{
		float rotyaw =  -spr->mo.rotangle /180.0f * PI;

		cosrot = (float)cos(rotyaw); 
		sinrot = (float)sin(rotyaw);
	}

	// Calculate interpolated vertices, shiny texcoords and lighting.
	for (i = 0, mVtx = mvertices, vtx = frame->vertices; i < numVertices; i++, mVtx++, vtx++)
	{
		// Is this vertex used at this LOD?
		if (activeLod && !(mdl->vertexUsage[i] & (1 << activeLod)))
			continue; // Not used; can be skipped.

		// Copy the surface normals at this vertex.
		// Note: vtx VY, VZ are transposed
		mVtx->normal[VX] = vtx->normal[VX];
		mVtx->normal[VY] = vtx->normal[VZ];
		mVtx->normal[VZ] = vtx->normal[VY];

		if (nextframe)
		{
			nextvtx = nextframe->vertices + i; //glc->vertexIndex;
			nptr = nextvtx->normal;

			if (vtx->normal[VX] != nptr[VX]
			 || vtx->normal[VY] != nptr[VY]
			 || vtx->normal[VZ] != nptr[VZ])
			{
				// Interpolate to next position.
				// Note: nptr VY, VZ are transposed
				mVtx->normal[VX] = mVtx->normal[VX]*inv + nptr[VX]*interpos;
				mVtx->normal[VY] = mVtx->normal[VY]*inv + nptr[VZ]*interpos;
				mVtx->normal[VZ] = mVtx->normal[VZ]*inv + nptr[VY]*interpos;

				// Normalize it (approximately).
				dist = M_ApproxDistancef(mVtx->normal[VZ],
					M_ApproxDistancef(mVtx->normal[VX], mVtx->normal[VY]));

				if (dist)
				{
					dist = 1.0f/dist;
					mVtx->normal[VX] *= dist;
					mVtx->normal[VY] *= dist;
					mVtx->normal[VZ] *= dist;
				}
			}

			// The interpolated vertex coordinates.
			// Note: vtx VY, VZ are transposed
			mVtx->vertex[VX] = vtx->vertex[VX] * inv + nextvtx->vertex[VX] * interpos;
			mVtx->vertex[VY] = vtx->vertex[VZ] * inv + nextvtx->vertex[VZ] * interpos;
			mVtx->vertex[VZ] = vtx->vertex[VY] * inv + nextvtx->vertex[VY] * interpos;
		}
		else
		{
			// Note: vtx VY, VZ are transposed
			mVtx->vertex[VX] = vtx->vertex[VX];
			mVtx->vertex[VY] = vtx->vertex[VZ];
			mVtx->vertex[VZ] = vtx->vertex[VY];


			// v220-5 Note: only applies to single frame models
			if (spr->mo.rotangle)
			{
				float vec[2];

				// Compensation for doom's 1:1.2 ratio must
				// be applied before rotating the vertices...
				mVtx->vertex[VX] *= RATIO_FACTOR;
				mVtx->vertex[VZ] *= RATIO_FACTOR;

				vec[VX] = mVtx->vertex[VX];
				vec[VY] = mVtx->vertex[VY];

				mVtx->vertex[VX] = vec[VX] *  cosrot + vec[VY] * sinrot;
				mVtx->vertex[VY] = vec[VX] * -sinrot + vec[VY] * cosrot;

				vec[VX] = mVtx->normal[VX];
				vec[VY] = mVtx->normal[VY];

				mVtx->normal[VX] = vec[VX] *  cosrot + vec[VY] * sinrot;
				mVtx->normal[VY] = vec[VX] * -sinrot + vec[VY] * cosrot;

				// Normalize it (approximately).
				dist = M_ApproxDistancef(mVtx->normal[VZ],
					M_ApproxDistancef(mVtx->normal[VX], mVtx->normal[VY]));

				if (dist)
				{
					dist = 1.0f/dist;
					mVtx->normal[VX] *= dist;
					mVtx->normal[VY] *= dist;
					mVtx->normal[VZ] *= dist;
				}
			}
		}		
		
		// Shiny texture coordinates, too, if they're needed.
		if(shininess > 0.0f)
		{
			// Calculate cylindrically mapped texcoords.
			// Quite far from perfect but very nice anyway.
			float sh_u = ATAN2F(mVtx->normal[VZ], mVtx->normal[VX]) / (2.0f*PI) - mdl_nyaw;
			float sh_v = ASINF(-mVtx->normal[VY])/PI + 0.5f - mdl_npitch;

			// This'll hide the wrap-around behind the model.
			// Works more often than not.
			while(sh_u > sh_pnt    ) sh_u--;
			while(sh_u < sh_pnt - 1) sh_u++;

			mVtx->shinytexcoord[VX] = sh_u + sh_u - sh_pnt;
			mVtx->shinytexcoord[VY] = sh_v + sh_v - sh_ang;
		}
		
		// Also calculate the light level at the vertex?
		if (lightVertices) 
		{
			// Begin with total darkness.
			float dot, vtxLight[3] = { 0, 0, 0};
			
			// v220-3 Add light from any other sources
			// light->used == 0 indicates end of list
			for (k = 0, light = mlights; k < lightLimit && light->used; k++, light++)
			{
				// v220-35 psprite always evenly lit
				if (psprite || spr->mo.vmoflags & HAS_EVENLIGHT)
				{
					if (dot >= evenLightFactor)
						dot = evenLightFactor * DOT3LIGHT(light->vector, evenLightFactor);
					else
						dot = evenLightFactor * DOT3LIGHT(light->vector, 1.f);
				}
				else if (spr->mo.vmoflags & HAS_3DLIGHT)
				{
					dot = DOTPROD3(light->vector, mVtx->normal);
				}
				else if (spr->mo.vmoflags & HAS_2DLIGHT)	// 2D vector reintroduced v220-35
				{
					dot = DOTPROD2(light->vector, vtx->normal);
					dot *= 1.25f; // Looks-good factor :-)
				}
				else	// default
				{
					dot = DOTPROD3(light->vector, mVtx->normal);
				}

				if(dot <= 0.0f) continue; // No light or from the wrong side.
				if(dot >  1.0f) dot = 1.0f;

				for(c = 0; c < 3; c++)
					vtxLight[c] += dot * (float)light->color[c];
			}
			
			for (k = 0; k < 3; k++)
			{
				// Check ambient
				if (vtxLight[k] < ambientColor[k]) 
					vtxLight[k] = ambientColor[k];

				// This is the final color.
				mVtx->color[k] = vtxLight[k];
			}
		}
	}

	// We can now render the model
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	// $translucent_water or z write disabled in general
	if (rendFlags & RF_NODEPTH) glDepthMask(GL_FALSE);

	// Change the blending mode?
	if (rendFlags & RF_SPECULAR)
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);					// Bright.
		rendFlags |= RF_BFNSET;
	}
	else if (rendFlags & RF_DARK)
	{
		glBlendFunc(GL_DST_COLOR, GL_ONE_MINUS_SRC_ALPHA);	// Dark.
		rendFlags |= RF_BFNSET;
	}

	// Render the model in two passes:
	// Pass 0 = regular
	// Pass 1 = shiny skin

	if (useskin >= 0)
	{
		pass = 0;
		GL_Bind(GL_PrepareSkin(mdl, useskin, &rendFlags));

		if (rendFlags == RF_ALPHA && alpha == 1.0f && !(mflags & MFF_ALPHANORMAL))
		{
			glAlphaFunc(GL_GEQUAL, 0.5f);
			glDisable(GL_BLEND);
			rendFlags |= (RF_ALPHAFN|RF_NOBLEND);

			if (mflags & MFF_NOCULL)
			{
				glDisable(GL_CULL_FACE);
				rendFlags |= RF_NOCULL;
			}
		}
	}
	else
	{
		pass = 1;
	}

	// Model space -> World space
	// Note: spr->mo.gz will have any adjusted smooth-z value
	glTranslatef(spr->mo.v1[VX] + mf->offset[VX] + spr->mo.visoff[VX], 
		FIX2FLT(spr->mo.gz) + mf->offset[VY] + spr->mo.visoff[VZ],
		spr->mo.v1[VY] + mf->offset[VZ] + spr->mo.visoff[VY]);

	// Model rotation (do not change the order of setting rotations)
	// Note: viewaligned is only true with hud weapons
	glRotatef(spr->mo.viewaligned ? spr->mo.v2[VX] : yawAngle  , 0.0f, 1.0f, 0.0f);
	glRotatef(spr->mo.viewaligned ? spr->mo.v2[VY] : pitchAngle, 0.0f, 0.0f, 1.0f);

	if (spr->mo.zangle)
		glRotatef(spr->mo.zangle, 1.0f, 0.0f, 0.0f); // R3D - For dead things/pickups on a slope

	// Scaling and model space offset.
	glTranslatef(smf->offset[VX], smf->offset[VY], smf->offset[VZ]);

	if (spr->mo.vmoflags & HAS_RADIUS8)
		glScalef(mf->scale[VX]*0.3f, mf->scale[VY]*0.3f, mf->scale[VZ]*0.3f);
	else
		glScalef(mf->scale[VX], mf->scale[VY], mf->scale[VZ]);

	for (; pass < 2; pass++)
	{
		byte	*ptr;
		int		count;
		RDBOOL	opaque = alpha == 1.0f;

		if (pass) // Initialize the shiny skin pass.
		{
			if (shininess <= 0.0f) break;

			// We're rendering on top of pass zero (usually).
			glDepthFunc(GL_LEQUAL);
			rendFlags |= RF_EQUAL;

			// Set blending mode, two choices: reflected and specular.
			if (mflags & MFF_SHINY_SPECULAR)
			{
				if (!(rendFlags & RF_SPECULAR))
				{
					glBlendFunc(GL_SRC_ALPHA, GL_ONE);	// Specular.
					rendFlags |= RF_BFNSET;
				}
			}
			else if (rendFlags & RF_BFNSET)
			{
				rendFlags &= ~RF_BFNSET;
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	// Reflective.	
			}

			// Change the texture.
			GL_Bind(GL_PrepareShinySkin(mf, number));

			if (mflags & MFF_SHINY_LIT) 
			{
				glColor4f(ambientColor[0] * shinycol[0],
					ambientColor[1] * shinycol[1], 
					ambientColor[2] * shinycol[2], shininess);
			}
			else
			{
				glColor4f(shinycol[0], shinycol[1], shinycol[2], shininess);
			}
		}
		else if (!lightVertices)
		{
			glColor4fv(ambientColor);
		}

		// Render as strips/fans
		// The texture offsets are contained in the glCommands list.
		ptr = (byte*) mdl->lods[activeLod].glCommands;

		while (*ptr)
		{
			count = *(int*) ptr;

			ptr += 4;

			// The polarity of count determines the primitive
			glBegin(count < 0 ? GL_TRIANGLE_FAN : GL_TRIANGLE_STRIP);

			if (count < 0) count = -count;

			while (count--)
			{	
				dmd_glCommandVertex_t *glc = (dmd_glCommandVertex_t*) ptr;

				ptr += sizeof(dmd_glCommandVertex_t);

				mVtx = mvertices + glc->vertexIndex;

				if (pass)
				{	// shiny tex co-ords
					glTexCoord2fv(mVtx->shinytexcoord);
				}
				else
				{	// Regular pass: just use the texcoord at the vertex.
					glTexCoord2fv(&glc->s);

					if (lightVertices)	// Set the color.
					{
						if (opaque)
							glColor3fv(mVtx->color);
						else
							glColor4f(mVtx->color[0],
									  mVtx->color[1],
									  mVtx->color[2],
									  alpha);
					}
				}

				glVertex3fv(mVtx->vertex);
			}

			glEnd();
		}
	}
	
	if (rendFlags & RF_BFNSET ) glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	if (rendFlags & RF_ALPHAFN) glAlphaFunc(GL_GREATER, 0);
	if (rendFlags & RF_NOBLEND) glEnable(GL_BLEND);
	if (rendFlags & RF_EQUAL  ) glDepthFunc(GL_LESS);
	if (rendFlags & RF_NOCULL ) glEnable(GL_CULL_FACE);
	if (rendFlags & RF_NODEPTH) glDepthMask(GL_TRUE);

	// Restore renderer state.
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}


