//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: rend_particle.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D, moved to main engine
//  Replaced zone memory with win mem [PARTCLID]
//  Added fix to prevent PG_LinkPtcGen returning an invalid slot
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
//   Particles separated out onto a rendering list to enable
//   order of rendering with trans flats to be determined.
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
//
// DESCRIPTION:
//	Rendering of particle generators.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <math.h>
#include <stdlib.h>

#include "de_base.h"
#include "de_console.h"
#include "de_render.h"
#include "de_play.h"
#include "de_refresh.h"
#include "de_graphics.h"
#include "de_misc.h"

#include "r3d_MAINspec.h"
#include "r3d_slopes.h"
#include "r3d_items.h"
#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

// Point + custom textures.
#define NUM_TEX_NAMES	(1 + MAX_PTC_TEXTURES)
#define PTC_INCREMENT	256
#define MAX_PGPARTS		4096
#define MIN_PGDIST		FRACUNIT * 5

// We can link 64 generators each into four sectors before
// running out of pglinks. 
#define PG_MAX			4 * MAX_ACTIVE_PTCGENS

// TYPES -------------------------------------------------------------------

typedef struct pglink_s
{
	struct pglink_s	*next;
	ptcgen_t *gen;
} pglink_t;

typedef struct
{
	ushort	gen;		// Index of the generator (active_ptcgens)
	ushort	index;
	fixed_t	distance;
} porder_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void P_FreePtcGen(ptcgen_t *gen);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern ptcgen_t *active_ptcgens[MAX_ACTIVE_PTCGENS];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

DGLuint			ptctexname[NUM_TEX_NAMES];
float			rend_particle_diffuse = 3.0f;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static pglink_t **pglinks;	// Array of pointers to pglinks in pgstore.
static pglink_t pgstore[PG_MAX];
static porder_t	pgtorder[MAX_PGPARTS];

static int		pgcursor;
static int		numparts;
static RDBOOL	haspoints[NUM_TEX_NAMES + 1];
static RDBOOL	haslines, hasnoblend, hasblend;
static float	viewupvec[3], viewsidevec[3];

// CODE --------------------------------------------------------------------

//===========================================================================
// PG_InitTextures
//===========================================================================
void PG_InitTextures(void)
{
	byte	*image, *data;
	size_t	size;
	int		i;

	//	The particle texture is a modification of the dynlight texture.
	if (!ptctexname[0])
	{
		// generate the texture
		data = (byte*)M_CacheLumpName("DLIGHT", &size, true);

		if (size != 64*64) Con_ErrorD("Invalid DLIGHT tex");
		image = (byte*)M_Calloc(size*4, 319);

		// Clear the texture name array.
		memset(ptctexname, 0, sizeof(ptctexname));

		// Mipmap it down to 32x32 and create an alpha mask.
		memcpy(image, data, size);
		GL_DownMipmap32(image, 64, 64, 1);
		memcpy(image + 32 * 32, image, 32 * 32);
		memset(image, 255, 32 * 32);		

		// No further mipmapping or resizing is needed, upload directly.
		// The zeroth texture is the default: a blurred point.
		GL_NewTexture(&ptctexname[0]);
		GL_TexImage(GL_LUMINANCE_ALPHA, 32, 32, 0, image);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

		M_Free(data);
		M_Free(image);
	}

	// Load any custom particle textures. They are loaded from the 
	// data\particles texture directory and are named "ParticleNN.png".
	// where NN is the range 00 to 31.
	// Note: ptctexname[0] is used to store the DLIGHT texture.
	for(i = 1; i < NUM_TEX_NAMES; i++)
	{
		char filename[MAXPATHNAMESIZE];
		int width, height, pixelsize;

		if (!ptctexname[i])
		{
			// Try to load the texture.
			sprintf(filename, "%sdata\\particles\\Particle%02i.png", ddRuntimeDir.path, i - 1);

			if ( (image = PNG_Load(filename, 0, &width, &height, &pixelsize, 6)) == NULL)
			{
				if (verbose) pr_OFinsert("PG_InitTextures: %s not found.\n", filename);
				continue;
			}

			// Create a new texture and upload the image.
			GL_NewTexture(&ptctexname[i]);

			if(verbose) 
			{
				pr_OFinsert("PG_InitTextures: Texture %02i: %i * %i * %i\n", 
					i, width, height, pixelsize);
			}

			GL_TexImage(pixelsize == 4 ? GL_RGBA : GL_RGB, width, height, 0, image);

			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

			// Free the buffer.
			M_Free(image);
		}
	}
}

//===========================================================================
// PG_ShutdownTextures
//===========================================================================
void PG_ShutdownTextures(void)
{
	int i;

	for (i = 0; i < NUM_TEX_NAMES; i++)
		GL_DeleteTexIndex(&ptctexname[i], "ptctexname");
}

//===========================================================================
// PG_InitForLevel
//===========================================================================
void PG_InitForLevel(void)
{
	P_InitParticleThinkers();

	pgcursor = 0;
	pglinks = (pglink_t**)W_Calloc(sizeof(pglink_t*) * numsectors, PARTCLID, 124);
	memset(active_ptcgens, 0, sizeof(active_ptcgens));
}

//===========================================================================
// PG_LinkPtcGen
//  Modified to correct problem when links are exhausted - R3D
//===========================================================================
static RDBOOL PG_LinkPtcGen(ptcgen_t *gen, int ssecidx)
{
	pglink_t	*link;
	int			si;

#ifdef _DEBUG
	// v220-4  Check for corruption
	if (ssecidx < 0 || ssecidx >= numsubsectors) return false;
#endif

	si = subsectors[ssecidx].sector - sectors;

	// Check one is not already set...
	for (link = pglinks[si]; link; link = link->next)
	{
		if (link->gen == gen) return true; // R3D
	}

	// We need a new PG link.
	if (pgcursor < PG_MAX)	// Check the limit has not been reached
	{
		link = &pgstore[ pgcursor++ ];
		link->gen   = gen;
		link->next  = pglinks[si];
		pglinks[si] = link;
		return true; // R3D
	}

	return false;	// Out of links - R3D
}

// Calculate the unit vectors used with particles
// The vectors are in the left-handed DGL coordinate system
// (same as in the game, but Y and Z have been swapped).
static void calc_ViewVectors()
{
	float viewfrontvec[3];
	float yawRad   = (float)viewangle/(float)ANGMAX * 2.0f * PI;
	float pitchRad = viewpitch * RATIO /180.0f * PI;

	float cpr = (float)cos(pitchRad);
	float cyr = (float)cos(yawRad);
	float spr = (float)sin(pitchRad);
	float syr = (float)sin(yawRad);

	//---------------------

	// The front vector.
	viewfrontvec[VX] = cyr * cpr;
	viewfrontvec[VZ] = syr * cpr;
	viewfrontvec[VY] = spr;

	// The up vector.
	viewupvec[VX] = -cyr * spr;
	viewupvec[VZ] = -syr * spr;
	viewupvec[VY] = cpr;

	// The side vector is the cross product of the front and up vectors.
	M_CrossProduct(viewfrontvec, viewupvec, viewsidevec);
}

//===========================================================================
// PG_InitForNewFrame
//	Init all active particle generators for a new frame.
//	Modified to correct problem when links are exhausted - R3D
//  Also see PG_LinkPtcGen above.
//===========================================================================
void PG_InitForNewFrame(void)
{
	ptcgen_t *gen;
	int		 i, k;

	//----------------------

	if (!CFX) return;
	
	calc_ViewVectors();	// Set the particle vectors

	// Clear the PG links.
	memset(pglinks, 0, sizeof(pglink_t*) * numsectors);
	pgcursor = 0;

	// Clear all visibility flags.
	for (i = 0; i < MAX_ACTIVE_PTCGENS; i++)
	{
		if (  (gen = active_ptcgens[i]) != NULL
			&& gen->thinker.function == (void (__cdecl *)(void))P_PtcGenThinker)
		{
			gen->flags &= ~PGF_VISIBLE;

			for(k = 0; k < gen->count; k++)
			{
				if (gen->ptcs[k].stage >= 0)
				{
					if ( !PG_LinkPtcGen(gen, gen->ptcs[k].subsector - subsectors) )
					{
						// v220-31 No link found, so null active_ptcgens[i]
						active_ptcgens[i] = NULL;
						goto next_ptcgen;
					}
				}
			}

			if (gen->flags & PGF_SUSPEND)
			{
				// R3D Preset the suspended flag for suspend types.
				//     It'll be cleared if visible
				gen->flags |= PGF_SUSPENDED;
			}
		}
next_ptcgen:;
	}
}

//===========================================================================
// PG_SectorIsVisible
//	The given sector is visible. All PGs in it should be rendered.
//	Scans PG links.
//===========================================================================
void PG_SectorIsVisible(int secidx)
{
	pglink_t *it = pglinks[secidx];

	for(; it; it = it->next)
	{
		it->gen->flags |= PGF_VISIBLE;
		it->gen->flags &= ~PGF_SUSPENDED;
	}
}

//===========================================================================
// PG_Sorter
//	Sorts in descending order.
//===========================================================================
static int __cdecl PG_Sorter(const void *pt1, const void *pt2)
{
	fixed_t diff = ((porder_t*)pt1)->distance - ((porder_t*)pt2)->distance;

	return diff > 0 ? -1 : diff < 0 ? 1 : 0;
}

//===========================================================================
// PG_ListVisibleParticles
//	Returns true if there are particles to render.
//===========================================================================
static int PG_ListVisibleParticles(void)
{
	ded_ptcgen_t	*def;
	ptcgen_t		*gen;
	particle_t		*pt;

	fixed_t			dist, maxdist; 
	int				i, p, stagetype;
	byte			visp[MAX_ACTIVE_PTCGENS];

	//------------------------------------

	haslines = hasblend = hasnoblend = false;
	memset(haspoints, 0, sizeof(haspoints));
	memset(visp, 0, MAX_ACTIVE_PTCGENS);

	// First count how many particles are in the visible generators.
	for (i = 0, numparts = 0; i < MAX_ACTIVE_PTCGENS; i++)
	{
		if (  (gen = active_ptcgens[i])
			&& gen->count
			&& gen->flags & PGF_VISIBLE
			&& !(gen->flags & PGF_SUSPENDED))
		{
			numparts += gen->count;
			visp[i] = true;
		}
	}

	if (numparts <= 0) return false;

	// Prepare the order list and particle type
	for (i = 0, numparts = 0; i < MAX_ACTIVE_PTCGENS; i++)
	{
		if (!visp[i]) continue;

		gen = active_ptcgens[i];
		def = gen->def;
		maxdist = (fixed_t)(def->maxdist * (float)FRACUNIT);

		for (p = 0, pt = gen->ptcs; p < gen->count; p++, pt++)
		{
			if (pt->stage < 0) continue;

			if ( !(gen->flags & PGF_BUBBLES) )
			{
				// Is the particle's sector visible?
				if (   !(pt->subsector->sector->siflags & SIF_VISIBLE)
					|| !M_GroupVis(pt->subsector))
				{
					continue; // No; this particle can't be seen.
				}
			}

			// Check the distance
			// v220-35 Note with weapon smoke looking straight up down will reduce the distance to 0 but will be in the frustrum
			if (!(gen->flags & PGF_SMOKE) && (dist = P_ApproxDistance(viewx - pt->pos[VX], viewy - pt->pos[VY])) < MIN_PGDIST)
				continue;	// pointless - as will be clipped by the frustrum

			pgtorder[numparts].gen      = (ushort)i;
			pgtorder[numparts].index    = p;
			pgtorder[numparts].distance = dist;

			if (!(gen->flags & PGF_BUBBLES))
			{
				if (maxdist && pgtorder[numparts].distance > maxdist)	// Too far
				{
					continue;
				}
			}

			stagetype = gen->stages[pt->stage].type;

			if(stagetype == PTC_POINT)
			{
				haspoints[0] = true;
			}
			else if(stagetype == PTC_LINE)
			{
				haslines = true;
			}
			else if(stagetype >= PTC_TEXTURE
				&& stagetype < PTC_TEXTURE + MAX_PTC_TEXTURES)
			{
				haspoints[stagetype - PTC_TEXTURE + 1] = true;
			}

			if (gen->flags & PGF_ADD_BLEND)
				hasblend = true;
			else
				hasnoblend = true;

			if (++numparts >= MAX_PGPARTS) goto sortparts;
		}
	}

	// This is the real number of possibly visible particles.
	if (!numparts) return false;	// Particles all too distant

sortparts:;

	// qsort the order list back->front
	qsort(pgtorder, numparts, sizeof(*pgtorder), PG_Sorter);
	return true;
}

typedef struct {
	float	col[4];
	float	vtx[4][3];
	float	z;
	float	waterz;
	ushort	isrendered;
} pointvar_t;

typedef struct {
	pointvar_t	*ptv;
	int			texnum;	// first pass
	int			count;
	int			maxcount;
} pointlist_t;

static pointlist_t pntlist[2][NUM_TEX_NAMES + 1];
static int index[2];

//===========================================================================
// v220-4 PG_PrepareList - replaces PG_RenderParticles
//  Prepares rendering lists for particles.
//  No rendering is done here.
//===========================================================================
static RDBOOL PG_PrepareList(int rtype, int list_num)
{
	ded_ptcstage_t	*dst, *next_dst;
	particle_t		*pt;
	ptcstage_t		*st;
	ptcgen_t		*gen;
	fixed_t			fixline[2];
	float			leftoff[3], rightoff[3], center[3];
	float			line[2], projected[2];
	float			size, mark, inv_mark;
	float			dist, maxdist;
	RDBOOL			hastex;
	int				idx, i, c;

	//------------------------------------

	idx = index[list_num];

	pntlist[list_num][idx].count = 0;

	// viewsidevec points to the left.
	for(i = 0; i < 3; i++)
	{
		leftoff[i] = viewupvec[i] + viewsidevec[i];
		rightoff[i] = viewupvec[i] - viewsidevec[i];
	}

	// Should we use a texture?
	if (rtype == PTC_POINT)
		pntlist[list_num][idx].texnum = 0;
	else if (rtype >= PTC_TEXTURE && rtype < PTC_TEXTURE + MAX_PTC_TEXTURES)
		pntlist[list_num][idx].texnum = rtype - PTC_TEXTURE + 1;
	else
		pntlist[list_num][idx].texnum = -1;

	if (pntlist[list_num][idx].maxcount < numparts)
	{
		pntlist[list_num][idx].maxcount = numparts;
		pntlist[list_num][idx].ptv = (pointvar_t*)W_Realloc(pntlist[list_num][idx].ptv, numparts * sizeof(pointvar_t), STATICID, 0);
	}

	hastex = pntlist[list_num][idx].texnum >= 0;

	for (i = 0; i < numparts; i++)
	{
		pointvar_t	*ptr;
		RDBOOL		isplane = false;
		RDBOOL		iswall = false;
		float		adjust = 0.5f;

		gen = active_ptcgens[pgtorder[i].gen];
		pt = gen->ptcs + pgtorder[i].index;

		// Only render one type of particles.
		if (    gen->stages[pt->stage].type != rtype		// types must match
			|| !(gen->flags & PGF_ADD_BLEND)^(!list_num))	// blended types only go on list 1, unblended only on list 0
		{
			continue;
		}

		dst = &gen->def->stages[pt->stage];
		st = &gen->stages[pt->stage];

		// Is there a next stage for this particle?
		if (   pt->stage >= MAX_PTC_STAGES - 1
			|| !gen->stages[pt->stage + 1].type)
		{
			// There is no "next stage". Use the current one.
			next_dst = gen->def->stages + pt->stage;
		}
		else
			next_dst = gen->def->stages + (pt->stage + 1);
		
		// Where is intermark?
		inv_mark = pt->tics / (float) dst->tics;
		mark = 1 - inv_mark;		
		
		// Calculate size and color.
		size = P_GetParticleRadius(dst     , pgtorder[i].index) * inv_mark 
			 + P_GetParticleRadius(next_dst, pgtorder[i].index) * mark;

		if (!size) continue; // Won't be visible

		ptr = &pntlist[list_num][idx].ptv[pntlist[list_num][idx].count];

		for (c = 0; c < 4; c++)
		{
			ptr->col[c] = dst->color[c]*inv_mark + next_dst->color[c]*mark;

			if (c == 3) break;	// Alpha will be adjusted next

			if (st->flags & PTCF_BRIGHT) continue; 

			// This is a simplified version of sectorlight (no distance attenuation).
			ptr->col[c] *= pt->subsector->sector->lightlevel * reciprocal255;
		}

		if (hastex)
		{
			sector_t *s;

			if (st->flags & PTCF_PLANE_FLAT && (s = pt->subsector->sector))
			{
				if ( !(isplane = s->floorheight + 0x200 >= pt->pos[VZ])
					&& s->ceilingheight - 0x200 <= pt->pos[VZ])
				{
					isplane = true; adjust = -adjust;
				}
			}
			
			if (st->flags & PTCF_WALL_FLAT && pt->contact && !pt->mov[VX] && !pt->mov[VY])
				iswall = true;
		}

		// See if the alpha needs adjusting
		maxdist = gen->def->maxdist;
		dist = FIX2FLT(pgtorder[i].distance);

		if (maxdist && dist > maxdist*.75f)	// Far diffuse?
			ptr->col[3] *= 1 - (dist - maxdist*.75f) / (maxdist*.25f);

		// Test for archvile flame (gnf_diffusenr flag in ded should only be set with this)
		if (gen->flags & PGF_DIFFUSE_NEAR && rend_particle_diffuse > 1.6f)	// R3D
		{
			if (dist < 1.6f * size)
				ptr->col[3] -= 1 - dist/(1.6f * size);
		}
		else if (!iswall && !isplane && dist < rend_particle_diffuse * size)
			ptr->col[3] -= 1 - dist/(rend_particle_diffuse * size);
	
		if (ptr->col[3] <= 0) continue;		// Fully transparent

		center[VX] = FIX2FLT(pt->pos[VX]);
		center[VZ] = FIX2FLT(pt->pos[VY]);

		if (isplane && pt->subsector->sector->specFlags2 & SF2_LOSLOPE && !(gen->flags & PGF_BUBBLES))
			center[VY] = FIX2FLT(SL_PointHeight(pt->subsector->sector, pt->pos[VX], pt->pos[VY], LOSLOPE) + 0x200);
		else
			center[VY] = FIX2FLT(P_GetParticleZ(pt));

		// If started on a 3D flat check it is still within bounds
		if (center[VY] < pt->lowlimit)
			center[VY] = ptr->z = pt->lowlimit;

		ptr->z = center[VY];

		if (!(gen->flags & PGF_BUBBLES))
			ptr->waterz = FIX2FLT(pt->subsector->s_waterplane);
		else
			ptr->waterz = 32767.0f;

		ptr->isrendered = false;

		// Set the vertices
		if (hastex)
		{
			float val;

			if (isplane)	// particle to be rendered on a plane	
			{
				ptr->vtx[0][VX] = ptr->vtx[3][VX] = center[VX] - size;
				ptr->vtx[0][VZ] = ptr->vtx[1][VZ] = center[VZ] - size;
				ptr->vtx[1][VX] = ptr->vtx[2][VX] = center[VX] + size;
				ptr->vtx[2][VZ] = ptr->vtx[3][VZ] = center[VZ] + size;
				
				if (pt->subsector->sector->specFlags2 & SF2_LOSLOPE)
				{
					ptr->vtx[0][VY] = SL_PointHeightFP(pt->subsector->sector, ptr->vtx[0][VX], ptr->vtx[0][VZ], LOSLOPE) + adjust;
					ptr->vtx[1][VY] = SL_PointHeightFP(pt->subsector->sector, ptr->vtx[1][VX], ptr->vtx[1][VZ], LOSLOPE) + adjust;
					ptr->vtx[2][VY] = SL_PointHeightFP(pt->subsector->sector, ptr->vtx[2][VX], ptr->vtx[2][VZ], LOSLOPE) + adjust;
					ptr->vtx[3][VY] = SL_PointHeightFP(pt->subsector->sector, ptr->vtx[3][VX], ptr->vtx[3][VZ], LOSLOPE) + adjust;
				}
				else
				{
					ptr->vtx[0][VY] =  ptr->vtx[2][VY] = center[VY] + adjust;
					ptr->vtx[1][VY] =  ptr->vtx[3][VY] = center[VY] + adjust;
				}
			}
			else if (iswall)	// Flat against a wall, then?
			{
				// There will be a slight approximation on the XY plane since
				// the particles aren't that accurate when it comes to wall
				// collisions.

				// Calculate a new center point (project onto the wall). Also
				// move 1 unit away from the wall to avoid the worst Z-fighting.

				fixline[VX] = pt->contact->dx;
				fixline[VY] = pt->contact->dy;
								
				M_ProjectPointOnLinef(pt->pos, &pt->contact->v1->x, fixline, 1.0f, projected);
				M_LineUnitVector(pt->contact, line);

				ptr->vtx[0][VX] = ptr->vtx[1][VX] = projected[VX] - (val = size*line[VX]);
				ptr->vtx[2][VX] = ptr->vtx[3][VX] = projected[VX] + val;
				ptr->vtx[0][VZ] = ptr->vtx[1][VZ] = projected[VY] - (val = size*line[VY]);
				ptr->vtx[2][VZ] = ptr->vtx[3][VZ] = projected[VY] + val;

				ptr->vtx[0][VY] = ptr->vtx[3][VY] = center[VY] - size;
				ptr->vtx[1][VY] = ptr->vtx[2][VY] = center[VY] + size;
			}
			else
			{
				ptr->vtx[0][VX] = center[VX] + (val = size*leftoff[VX]);
				ptr->vtx[2][VX] = center[VX] - val;
				ptr->vtx[0][VY] = center[VY] + (val = size*leftoff[VY]/1.2f);
				ptr->vtx[2][VY] = center[VY] - val;
				ptr->vtx[0][VZ] = center[VZ] + (val = size*leftoff[VZ]);
				ptr->vtx[2][VZ] = center[VZ] - val;

				ptr->vtx[1][VX] = center[VX] + (val = size*rightoff[VX]);
				ptr->vtx[3][VX] = center[VX] - val;
				ptr->vtx[1][VY] = center[VY] + (val = size*rightoff[VY]/1.2f);
				ptr->vtx[3][VY] = center[VY] - val;
				ptr->vtx[1][VZ] = center[VZ] + (val = size*rightoff[VZ]);
				ptr->vtx[3][VZ] = center[VZ] - val;
			}
		}
		else // It's a line.
		{
			ptr->vtx[0][VX] = center[VX];
			ptr->vtx[0][VY] = center[VY];
			ptr->vtx[0][VZ] = center[VZ];

			ptr->vtx[1][VX] = center[VX] - FIX2FLT(pt->mov[VX]);
			ptr->vtx[1][VY] = center[VY] - FIX2FLT(pt->mov[VZ]);
			ptr->vtx[1][VZ] = center[VZ] - FIX2FLT(pt->mov[VY]);
		}
		pntlist[list_num][idx].count++;
	}

	return pntlist[list_num][idx].count > 0;
}

static void rendParticles(int index, int list_num, int pos)
{
	int i, n;

	// Render the particles
	if (pntlist[list_num][index].texnum >= 0)
	{
		glDepthMask(GL_FALSE);
		glDepthFunc(GL_LEQUAL);
		glDisable(GL_CULL_FACE);
		GL_Bind(ptctexname[pntlist[list_num][index].texnum]);
		glBegin(GL_QUADS); // A waste of vertices and triangles, but...
	}
	else
	{
		glDisable(GL_TEXTURE_2D);	// Lines don't use textures.
		glBegin(GL_LINES);
	}
	
	n = pntlist[list_num][index].count;

	for (i = 0; i < n; i++)
	{
		pointvar_t *ptr = &pntlist[list_num][index].ptv[i];

		if (ptr->isrendered) continue;

		if (pos > 0 && ptr->z <  ptr->waterz) continue;
		if (pos < 0 && ptr->z >= ptr->waterz) continue;

		ptr->isrendered = true;

		th_items[TH_PTARCL]++;

		glColor4fv(ptr->col);

		if (pntlist[list_num][index].texnum >= 0)
		{
			glTexCoord2i(0, 0); glVertex3fv(ptr->vtx[0]);
			glTexCoord2i(1, 0); glVertex3fv(ptr->vtx[1]);
			glTexCoord2i(1, 1); glVertex3fv(ptr->vtx[2]);
			glTexCoord2i(0, 1); glVertex3fv(ptr->vtx[3]);
		}
		else
		{
			glVertex3fv(ptr->vtx[0]);
			glVertex3fv(ptr->vtx[1]);
		}
	}

	glEnd();

	if (pntlist[list_num][index].texnum >= 0) 
	{
		glEnable(GL_CULL_FACE);
		glFrontFace(GL_CW);
		glDepthMask(GL_TRUE);
		glDepthFunc(GL_LESS);
	}
	else
	{
		glEnable(GL_TEXTURE_2D);
	}
}

//===========================================================================
// PG_RenderPass - blended pass goes onto list 1
//===========================================================================
void PG_RenderPass(int list_num)
{
	int i;

	if (haslines && PG_PrepareList(PTC_LINE, list_num))
		index[list_num]++;

	for (i = 0; i < NUM_TEX_NAMES; i++)
	{
		if (   haspoints[i]
			&& PG_PrepareList(
				i ? PTC_TEXTURE + i - 1 : PTC_POINT, 
				list_num))
		{
			index[list_num]++;
		}
	}
}

void PG_PreparePG(void)
{
	index[0] = index[1] = 0;

	if (   r_use_particles
		&& PG_ListVisibleParticles())
	{
		if (hasnoblend) PG_RenderPass(0);
		if (hasblend  ) PG_RenderPass(1);
	}
}


//===========================================================================
// PG_Render
//	Render all the visible particle generators. 
//	We must render all particles ordered back->front, or otherwise 
//	particles from one generator will obscure particles from another.
//	This would be especially bad with smoke trails.
//===========================================================================
void PG_Render(int pos)
{
	int i;

	if (index[0])
	{
		for (i = 0; i < index[0]; i++) rendParticles(i, 0, pos);
	}

	if (index[1])
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);	// Set blending mode.

		for (i = 0; i < index[1]; i++) rendParticles(i, 1, pos);
	
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	// Restore blending mode.
	}
}

// v220-4 Used to clear flags to allow particles to be rendered when in pause mode
void PG_PtcClearRender()
{
	int i, k, n;

	for (i = 0; i < index[0]; i++)
	{
		n = pntlist[0][i].count;

		for (k = 0; k < n; k++) pntlist[0][i].ptv[k].isrendered = false;
	}

	for (i = 0; i < index[1]; i++)
	{
		n = pntlist[1][i].count;

		for (k = 0; k < n; k++) pntlist[1][i].ptv[k].isrendered = false;
	}
}
