//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2009 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: rend_shadow.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified for R3D
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Shadow rendering
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "de_base.h"
#include "de_refresh.h"
#include "de_render.h"
#include "de_play.h"
#include "de_misc.h"

#include "doom\p_local.h"

#include "r3d_MAINspec.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern rendpoly_t	sspoly;					// v220-5

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		shadowMaxRad = 80;
int		shadowMaxDist = 1000;
float	shadowFactor = 0.5f;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

//===========================================================================
// Rend_ProcessThingShadow
//  Modified for z-smoothing - R3D
//  Note: shadows are completely disabled if a map has slopes
//  v220-5 change to use global sspoly
//  v220-6 reformatted slightly
//===========================================================================
static void Rend_ProcessThingShadow(mobj_t *mo)
{
	float		height, moh, halfmoh, color, pos[2];
	sector_t	*sec = mo->subsector->sector;
	int			radius, i, z;

	// shadows on slopes are not currently supported
	if (sec->specFlags2 & SF2_LOSLOPE) return;

	// Is this too far away to bother?
	pos[VX] = FIX2FLT( mo->x );
	pos[VY] = FIX2FLT( mo->y );

	if(Rend_PointDist2D(pos) > shadowMaxDist) return;

	// Apply a Short Range Visual Offset?
	if (mo->state && mo->tics >= 0)
	{
		float mul = mo->tics / (float) mo->state->tics;				
		pos[VX] += FIX2FLT(mo->srvo[VX] << 8) * mul;
		pos[VY] += FIX2FLT(mo->srvo[VY] << 8) * mul;
	}

	// Check the height.
	z = mo->subsector->dlfloorheight;// + SL_XYHeight(sec->stptr, mo->x, mo->y, LOSLOPE);
	if (mo->z < z) return;		// below visible floor - R3D

	height = FIX2FLT(mo->z - z);
	moh = FIX2FLT( mo->height );
	if(!moh) moh = 1;
	if(height > moh) return;	// Too high above floor.
	
	// Calculate the strength of the shadow.
	color = shadowFactor * sec->lightlevel/255.0f;
	halfmoh = moh*0.5f;
	if(height > halfmoh) color *= 1 - (height - halfmoh)/halfmoh;
	if(whitefog) color *= 0.5f;
	if(color <= 0) return;		// Can't be seen.
	if(color > 1.0f) color = 1.0f;

	// Calculate the radius of the shadow.
	radius = R_MaxRadius(mo, true, 4);
	if(!radius) return;
	if(radius > shadowMaxRad) radius = shadowMaxRad;

	// Prepare the poly.
	// memset(&poly, 0, sizeof(poly));
	sspoly.type = rp_flat;
	sspoly.flags = RPF_SHADOW;
	sspoly.light = (lumobj_t*) (radius*2); // hijack for radius (Yuk..)
	sspoly.texoffx = pos[VX] + radius;
	sspoly.texoffy = pos[VY] + radius;
	sspoly.top = FIX2FLT(z) + 0.5f;	// original bias was 0.1f
	sspoly.numvertices = 4;
	sspoly.detail = NULL;

	sspoly.vertices[0].pos[VX] = sspoly.vertices[3].pos[VX] = pos[VX] - radius;
	sspoly.vertices[0].pos[VY] = sspoly.vertices[1].pos[VY] = pos[VY] + radius;
	sspoly.vertices[1].pos[VX] = sspoly.vertices[2].pos[VX] = pos[VX] + radius;
	sspoly.vertices[2].pos[VY] = sspoly.vertices[3].pos[VY] = pos[VY] - radius;

	for(i = 0; i < 4; i++)
	{
		sspoly.vertices[i].pos[VZ] = 0;	// Slopes not supported so z always 0
		// Shadows are shades of grey
		memset(sspoly.vertices[i].rgba, (int)(color * 255), 3);
	}

	RL_AddPolyFlat(&sspoly);
}

//===========================================================================
// Rend_RenderShadows
//===========================================================================
void Rend_RenderShadows(void)
{
	mobj_t	 *mo;
	int		 i;
	
	validcount++;

	if (!BDMFX || !cfg.useShadows || R3D_slopetypes) return;

	// Check all mobjs in all visible sectors.
	for (i = 0; i < numsectors; i++)
	{
		if (!(sectors[i].siflags & SIF_VISIBLE)) continue;

		for (mo = sectors[i].thinglist; mo; mo = mo->snext)
		{
			// Should this mobj have a shadow?
			// NOTE: certain types have already had the flag set.
			//       see PH_CheckMobjFitsSector()
			// R3D - MF2_NOSHADOW can be ded set to selectively disable a mo type
			if (mo->validcount != validcount)
			{
				mo->validcount = validcount;

				if (   !(mo->flags2 & (MF2_CORPSEFADE|MF2_NOSHADOW))
					&& !(mo->frame & FF_FULLBRIGHT))
				{
					// check for other things where shadow is not wanted
					if (mo->ddflags & (DDMF_DONTDRAW|DDMF_ALWAYSLIT))
						mo->flags2 |= MF2_NOSHADOW;
					else
						Rend_ProcessThingShadow(mo);
				}
			}
		}
	}
}
