//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: rend_sky.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Sky rotation added
//  Simple sky support commented out
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified for R3D
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
//  Support for sky boxes added
//  Skytop colours now supported with sky from linedef
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 9 (v220-9) 18/Dec/2010 gmj
// Revision 2.2.0 build 18 (v220-18) 05/Jan/2013 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Sky rendering
//
//-----------------------------------------------------------------------------

//** This version supports only two sky layers.
//** More would be a waste of resources.

// HEADER FILES ------------------------------------------------------------

#include <direct.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "de_base.h"
#include "de_console.h"
#include "de_render.h"
#include "de_graphics.h"

#include "r3d_compat.h"
#include "m_winmem.h"

#include "rend_main.h"
#include "r3d_OGLfunc.h"

// MACROS ------------------------------------------------------------------

// v220-3 skybox
#define SZBOX   1024.0f	// 32766/2
#define TCTL	0.0014f
#define TCTH	0.9986f

// v220-9 this expands the sky 'sphere' as the Intel graphics chip fails
//        if the vertexes' positions are too close to the viewpoint.
#define EXPAND_FACTOR	1024.f

// Sky dome used with doom sky textures
// 3 and 24 are considered to be optimal - gmj
#define NUMROWS			12
#define NUMROWSFP		12.0f
#define NUMCOLUMNS		24
#define NUMCOLUMNSFP	24.0f

#define SKYVTX_IDX(c, r)	( (r)*NUMCOLUMNS + (c)%NUMCOLUMNS )


// TYPES -------------------------------------------------------------------

// v220-3 skybox ------------------------

typedef struct {
	float vt[2];
} vttex_t;

static vttex_t vt[4] = {
	{ TCTL, TCTL },
	{ TCTH, TCTL },
	{ TCTH, TCTH },
	{ TCTL, TCTH }
};

static short vtxsign[8][3] = {
	{  1,  1,  1 },	// 0 - top left back
	{ -1,  1,  1 },	// 1 - top right back
	{ -1, -1,  1 },	// 2 - top right front
	{  1, -1,  1 },	// 3 - top left front
	{  1,  1, -1 },	// 4 - base left back
	{ -1,  1, -1 },	// 5 - base right back
	{ -1, -1, -1 },	// 6 - base right front
	{  1, -1, -1 }	// 7 - base left front
};

typedef struct {
	int vn[4];
} vtxnum_t;

static vtxnum_t vn[6] = {
	{ 0, 1, 2, 3 },		// top texture
	{ 0, 3, 7, 4 },		// wall left
	{ 1, 0, 4, 5 },		// wall back
	{ 2, 1, 5, 6 },		// wall right
	{ 3, 2, 6, 7 },		// wall front
	{ 7, 6, 5, 4 }		// base texture
};

typedef struct {
	uint	txnum;
//	float	width, height;
} skyboxtex_t;

static char *skytexnamepng[6] = {
	"side6.png", "side5.png", "side4.png",
	"side3.png", "side2.png", "side1.png"
};

static char *skytexnamejpg[6] = {
	"side6.jpg", "side5.jpg", "side4.jpg",
	"side3.jpg", "side2.jpg", "side1.jpg"
};

static char *skytex_tga[6] = {
	"_up.tga", "_bk.tga", "_lf.tga",
	"_ft.tga", "_rt.tga", "_dn.tga"
};

static char *skytex_jpg[6] = {
	"_up.jpg", "_bk.jpg", "_lf.jpg",
	"_ft.jpg", "_rt.jpg", "_dn.jpg"
};

static char *wadname[] = {
	"Doom1", "Doom2", "tnt", "plutonia"
};

// end skybox --------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void Rend_SkyDetail(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern byte			topLineRGB[3];
extern texscale_t	*ctex;
extern char			skyboxname[34];
extern int			num_skytop_colors;
extern int			GLOBAL_SKY_SCROLL_LINE;	// v220-35


// PUBLIC DATA DEFINITIONS -------------------------------------------------

skylayer_t	skyLayers[MAXSKYLAYERS];
int			skyhemispheres;
int			r_skybox;				// v220-3
int			has_skybox;				// v220-9
float		setskycolour = -1.f;	// v220-35
char		defsky[7] = "F_SKY1";	
char		*globalskytex = defsky;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// The texture offset to be applied to the texture coordinates in SkyVertex().

static struct sky_t {
	fadeout_t	*currentFO;
	float		maxSideAngle;
	float		horizonOffset;
	float		rotateAngle;
	float		lowersky_offset;
	float		skyfactor;
	float		last_horizon;
	float		horizon_offset;
	int			active_layer;
	int			skylump;
	int			current_skyvarnum;

	RDBOOL		yflip;
	float		sky_Yscroll;
	float		memh, memw;

} skv;


typedef struct {
	float x,y,z;
} skyvertex_t;		// Vertices for a hemisphere.

skyvertex_t skyVerts[NUMCOLUMNS * (NUMROWS + 1)];

// v220-3 skybox
static skyboxtex_t skyboxtex[MAXSKYLAYERS][6];
static skyvertex_t skyboxvtx[8];


// CODE --------------------------------------------------------------------


// v220-3 delete skybox and/or skycap textures
//        the doom sky textures are stored directly to the flats array
void Rend_DeleteSkyAddOnTexures(RDBOOL skycaps)
{
	int		i, k;

	for (i = 0; i < MAXSKYLAYERS; i++)
	{
		if (   skycaps
			&& skyLayers[i].texture > 0
			&& texArray[skyLayers[i].texture]->gltex)
		{
			GL_DeleteTexIndex(&texArray[skyLayers[i].texture]->gltex, "skycap");
		}

		for (k = 0; k < 6; k++)
		{
			if (skyboxtex[i][k].txnum)
				GL_DeleteTexIndex(&skyboxtex[i][k].txnum, "skyboxtex");
		}
	}
}

// v220-31 support added for GLDEFS lump
static int loadGLDEFSsky(int lnum, char *skyname)
{
	char	*nextstr;
	RDBOOL	fliptop = false;
	RDBOOL	flipbottom = false;	// needed?

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, "Skybox"))
		{				
			nextstr = SR_NextSTR(false);
			if (stricmp(nextstr, skyname)) return false;

			while( (nextstr = SR_NextSTR(false)) )
			{
				if (!_stricmp(nextstr, "fliptop")) fliptop = true;
				if (!_stricmp(nextstr, "flipbottom")) flipbottom = true;
				if (!_stricmp(nextstr, "{")) goto loadglsky;
			}
		}
	}

	return false;

loadglsky:;

	int		slump[6] = { -1, -1, -1, -1, -1, -1 };
	char	*ptr;
	int		lump;
	int		typedefn;

	while( (nextstr = SR_NextSTR(false)) )
	{
		if (!_stricmp(nextstr, "}")) goto setglsky;
		if ( (lump = W_CheckNumForName(nextstr)) < 0) return false;

		ptr = &nextstr[ (strlen(nextstr) - 2) ];

		if (!_stricmp(ptr, "UP")) slump[0] = lump;
		if (!_stricmp(ptr, "BK")) slump[1] = lump;
		if (!_stricmp(ptr, "LF")) slump[2] = lump;
		if (!_stricmp(ptr, "FT")) slump[3] = lump;
		if (!_stricmp(ptr, "RT")) slump[4] = lump;
		if (!_stricmp(ptr, "DN")) slump[5] = lump;
	}

	return false;	// If we get here then there's a script error

setglsky:;

	int		i, w, h, cw, ch;
	int		pixSize;
	byte	*buffer;
	
	if (fliptop) slump[5] = slump[0];	// fliptop but could there be a flipbottom?
	else
	if (flipbottom) { slump[0] = slump[5]; fliptop = true; }

	for (i = 0; i < 6; i++)
	{
		if (slump[i] >= 0)
		{
			typedefn = DoomLumpIdentify(slump[0], &cw, &ch, DLI_GRAPHIC);

			if (cw < 64 || cw != ch || (typedefn != LMP_JPG && typedefn != LMP_PNG) ) break;

			if (i == 5 && fliptop)
			{
				skyboxtex[lnum][5].txnum = skyboxtex[lnum][0].txnum;
			}
			else
			{
				if (typedefn == LMP_JPG)
					buffer = JPG_Load(NULL, slump[i], &w, &h, &pixSize, 12);
				else	// must be png
					buffer = PNG_Load(NULL, slump[i], &w, &h, &pixSize, 12);

				if (!buffer) Con_Error("Failed to load skylump\n");

				// Upload it.
				GL_PrepareSkyBox(buffer, w, h, pixSize == 4, &skyboxtex[lnum][i].txnum);
			//	skyboxtex[lnum][i].width  = (float)w;
			//	skyboxtex[lnum][i].height = (float)h;

				M_Free(buffer);
			}
		}
	}

	if (i == 6) return true;
	
	while (--i >= 0)
	{
		if (skyboxtex[lnum][i].txnum)
			GL_DeleteTexIndex(&skyboxtex[lnum][i].txnum, "skyboxtex");
	}

	return false;
}

// v220-3 load skybox textures
static RDBOOL Rend_InitSkyBox(skylayer_t *slayer)
{
	extern	char	*customTexPath;

	char	fname[MAXPATH];
	char	*skyname;
	int		i, c, lnum = skyLayers - slayer;

	//------------------------------

	Rend_DeleteSkyAddOnTexures(false);

	if (lnum < 0 || lnum >= MAXSKYLAYERS)
		Con_Error("Skybox: unset layer");

	if (!(ge_inf.use_sky_box & GE_SKYBOXCHECK) ) goto check_general;

	if (!_stricmp(skyboxname, "none")) skyboxname[0] = 0;

	// v220-4 Add command line skyboxname support (where *_up, *_dn etc, is used)
	// Check for a match in the skyboxbyname folder. jpg takes precedence
	if (skyboxname[0])
	{
		int  width, height, pixSize;
		char buf[512];
		char **tagptr;

		sprintf(fname, "%sskyboxbyname", ddRuntimeDir.path);
		tagptr = skytex_jpg;

		for (c = i = 0; i < 6; i++)
		{
			sprintf(buf, "%s\\%s%s", fname, skyboxname, tagptr[i]);

			if ( (_access(buf, 4)) == -1)
			{
				c++; break;
			}
		}

		if (c)
		{
			tagptr = skytex_tga;

			for (c = i = 0; i < 6; i++)
			{
				sprintf(buf, "%s\\%s%s", fname, skyboxname, tagptr[i]);

				if ( (_access(buf, 4)) == -1)
				{
					c++; break;
				}
				else if (!TGA_GetSize(buf, &width, &height))
				{
					Con_Message("Skybox texture \"%s\" unsupported format\n", buf);
					c++; break;
				}
			}
		}

		if (c)
		{
			Con_Message("Skybox texture by name not found or incomplete set\n", buf);
			goto check_general;
		}

		for (i = 0; i < 6; i++)
		{
			byte	*buffer;

			sprintf(buf, "%s\\%s%s", fname, skyboxname, tagptr[i]);

			if (tagptr == skytex_jpg)
			{
				buffer = JPG_Load(buf, 0, &width, &height, &pixSize, 11);
			}
			else
			{
				if (!(buffer = TGA_LoadFile888(buf, &width, &height, &pixSize)))
					Con_Error("Failed to load file %s\n", buf);
			}

			// Upload it.
			GL_PrepareSkyBox(buffer, width, height, pixSize == 4, &skyboxtex[lnum][i].txnum);
		//	skyboxtex[lnum][i].width  = (float)width;
		//	skyboxtex[lnum][i].height = (float)height;

			M_Free(buffer);
		}

		Con_Message("loadsky success %s.%s\n", skyboxname, tagptr == skytex_jpg ? "jpg" : "tga");
		goto sky_loaded;
	}

check_general:;

	skyname = (char*)&texArray[slayer->texture]->name[0];	// e.g. SKY1

	if (skyname && ge_inf.use_sky_box & GE_SKYBOXFORCE)
	{
		if (!stricmp(skyname, "RSKY1")) skyname = "SKY1";
		if (!stricmp(skyname, "RSKY2")) skyname = "SKY2";
		if (!stricmp(skyname, "RSKY3")) skyname = "SKY3";
	}

	if (strlen((const char *)skyname) != 4)
	{
		if (ge_inf.use_sky_box & GE_SKYBOXCHECK)
			Con_Message("Unrecognised skybox name \"%s\"\n", skyname);
		return false;
	}

	if (SR_LoadPWADSCPLump(W_CheckPWADNumForName("GLDEFS")) )
	{
		i = loadGLDEFSsky(lnum, skyname);
		SR_EndScript();
		if (i) goto sky_loaded;
	}

	if ( !(ge_inf.use_sky_box & GE_SKYBOXCHECK) ) return false;

	// If not found search the generic skyboxes folder
	if (customTexPath)
	{
		// custom textures. If any exist these are always loaded
		sprintf(fname, "%s\\%s", customTexPath, skyname);
	}

	if (!customTexPath || (_access(fname, 0)) == -1)
	{
		if (gamemission > 3)
		{
			Con_Message("Skybox: unknown iwad");
			return false;
		}

		sprintf(fname, "%sData\\Skyboxes\\%s\\%s",
					ddRuntimeDir.path, wadname[gamemission], skyname);

		if ( (_access(fname, 0)) == -1)
		{
			Con_Message("SKY BOX loader; no valid folder found\n");
			return false;
		}

		if (!(ge_inf.use_sky_box & GE_SKYBOXFORCE) && W_CheckPWADtexture((char*)skyname))
		{
			Con_Message("SKYBOX: PWAD has custom %s: -usb ignored\n", skyname);
			return false;
		}
	}

	if (_chdir(fname)) return false;

	if ( (_access("cube.png", 4)) != -1)	// All six faces share the same texture
	{
		int  width, height, pixSize;
		byte *buffer;

		buffer = PNG_Load("cube.png", 0, &width, &height, &pixSize, 7);
		
		// Upload it.
		GL_PrepareSkyBox(buffer, width, height, pixSize == 4, &skyboxtex[lnum][0].txnum);
	//	skyboxtex[lnum][0].width = (float)width;
	//	skyboxtex[lnum][0].height = (float)height;

		M_Free(buffer);
	}
	else
	{
		int c, type = LMP_JPG;

		for (c = i = 0; i < 6; i++)
		{
			if ( (_access(skytexnamejpg[i], 4)) == -1)
			{
				c++; break;
			}
		}

		if (c)
		{
			type = LMP_PNG;

			for (c = i = 0; i < 6; i++)
			{
				if ( (_access(skytexnamepng[i], 4)) == -1)
				{
					c++; break;
				}
			}
		}

		if (c)
		{
			Con_Message("Skybox textures missing\n");
			return false;
		}

		for (i = 0; i < 6; i++)
		{
			int  width, height, pixSize;
			byte *buffer;

			if (type == LMP_PNG)
				buffer = PNG_Load(skytexnamepng[i], 0, &width, &height, &pixSize, 8);
			else
				buffer = JPG_Load(skytexnamejpg[i], 0, &width, &height, &pixSize, 8);
			
			// Upload it.
			GL_PrepareSkyBox(buffer, width, height, pixSize == 4, &skyboxtex[lnum][i].txnum);
		//	skyboxtex[lnum][i].width = (float)width;
		//	skyboxtex[lnum][i].height = (float)height;

			M_Free(buffer);
		}
	}
	
	_chdir(ddRuntimeDir.path);

sky_loaded:;
	
	// set the vertex values
	for (i = 0; i < 8; i++)
	{
		skyboxvtx[i].x = SZBOX * (float)vtxsign[i][0];
		skyboxvtx[i].y = SZBOX * (float)vtxsign[i][1];
		skyboxvtx[i].z = SZBOX * (float)vtxsign[i][2];
	}

	return true;
}

// Calculate the vertex and texture coordinates.
//  v220-5 xflip added
// v220-35 revised
static void SkyVertex(int r, int c, RDBOOL yflip)
{
	skyvertex_t *svtx = skyVerts + SKYVTX_IDX(c, r);

	if (skv.yflip) c = -c;	// The sign of c sets y (upper or lower hemi) flip

	// And the texture coordinates.
	// v220-5 sky scroll added
	if(!yflip)	// Flipped Y is for the lower hemisphere.
	{
		glTexCoord2f(skv.skyfactor * c / NUMCOLUMNSFP, (r - skv.sky_Yscroll) / NUMROWSFP);
	}
	else
	{
		glTexCoord2f(skv.skyfactor * c / NUMCOLUMNSFP, (NUMROWS - r - skv.sky_Yscroll) / NUMROWSFP);
	}
	// Also the color.
	if (setskycolour >= 0)
	{
		if(r==0)
			glColor4f(setskycolour,setskycolour,setskycolour,0);
		else
			glColor3f(setskycolour,setskycolour,setskycolour);
	}
	else
	if(skv.currentFO->use)
	{
		if(r==0) glColor4f(1.0f,1.0f,1.0f,0); else glColor3f(1.0f,1.0f,1.0f);
	}
	else
	{
		if(r==0) glColor3f(0,0,0); else glColor3f(1.0f,1.0f,1.0f);
	}
	// And finally the vertex itself.
	glVertex3f(svtx->x, yflip ? -svtx->y - skv.lowersky_offset: svtx->y, svtx->z);	// v220-9
}

static void CapSideVertex(int r, int c, RDBOOL yflip)
{
	// Look up the precalculated vertex.
	skyvertex_t *svtx = skyVerts + SKYVTX_IDX(c, r);

	glVertex3f(svtx->x, yflip ? -svtx->y - skv.lowersky_offset: svtx->y, svtx->z);	// v220-9
}

//===========================================================================
// Rend_SkyRenderer
//	Hemi is Upper or Lower. Zero is not acceptable.
//	The current texture is used. SKYHEMI_NO_TOPCAP can be used.
//===========================================================================
static void Rend_SkyRenderer(int hemi)
{
	int		r, c;
	RDBOOL	yflip = (hemi & SKYHEMI_LOWER) > 0;

	// The top row (row 0) is the one that's faded out.
	// There must be at least 4 columns. The preferable number
	// is 4n, where n is 1, 2, 3... There should be at least
	// two rows because the first one is always faded.
	if(hemi & SKYHEMI_JUST_CAP)
	{
		glDisable(GL_TEXTURE_2D);

		// Use the appropriate color.
		if (skv.currentFO->use)
			glColor3fv(skv.currentFO->rgb);
		else
			glColor3f(0,0,0);

		// Draw the cap.
		glBegin(GL_TRIANGLE_FAN);
		for (c = 0; c < NUMCOLUMNS; c++) CapSideVertex(0, c, yflip);
		glEnd();

		// If we are doing a colored fadeout...
		if(hemi & SKYHEMI_FADEOUT_BG)
		{
			// We must fill the background for the top row since it'll
			// be partially translucent.
			glBegin(GL_TRIANGLE_STRIP);
			CapSideVertex(0, 0, yflip);

			for (c = 0; c < NUMCOLUMNS; c++)
			{
				CapSideVertex(1, c  , yflip);	// One step down.
				CapSideVertex(0, c+1, yflip);	// And one step right.
			}

			CapSideVertex(1, c, yflip);
			glEnd();
		}
		glEnable(GL_TEXTURE_2D);
		return;
	}

	glBegin(GL_QUAD_STRIP);

	for (r = 0; r < NUMROWS; r++)
	{
		for (c = 0; c <= NUMCOLUMNS; c++)
		{
			SkyVertex(r  , c, yflip);
			SkyVertex(r+1, c, yflip);
		}
	}

	glEnd();
}

//===========================================================================
// Rend_SkyRenderHemisphere
//===========================================================================
static void Rend_RenderSkyHemisphere(skylayer_t	*slayer, uint texn, int whichHemi)
{
	GL_SetSkyTopColor(slayer);

	// First render the cap and the background for fadeouts, if needed.
	// The color for both is the current fadeout color.
	Rend_SkyRenderer(whichHemi | SKYHEMI_JUST_CAP | (skv.currentFO->use? SKYHEMI_FADEOUT_BG : 0));

	GL_Bind(texn);
	Rend_SkyRenderer(whichHemi);
}

// v220-5 yflip and sky rotation added
// v220-35 modified
void Rend_RenderSky()
{
	int		i, n;

	// IS there a sky to be rendered?
	if (!skyhemispheres) return;

	skv.active_layer = -1;

	for (i = 0; i < MAXSKYLAYERS; i++)
	{
		if (skyLayers[i].flags & SLF_ENABLED)
		{
			skv.active_layer = i;
			break;
		}
	}

	if (skv.active_layer < 0) return;

	// skies must not be written into the depth buffer
	glDepthMask(GL_FALSE);

	// v220-3 skybox renderer
	//   Note: if whitefog then don't use skybox
	if (r_skybox && !whitefog && skyLayers[skv.active_layer].flags & SLF_SKYBOX)
	{
		if (!skyboxtex[skv.active_layer][0].txnum)
		{
			Rend_InitSkyBox(&skyLayers[skv.active_layer]);
			globalskytex = defsky;
		}

		for (i = 0; i < 6; i++)
		{
			if (!skyboxtex[skv.active_layer][1].txnum)
				GL_Bind(skyboxtex[skv.active_layer][0].txnum);
			else
				GL_Bind(skyboxtex[skv.active_layer][i].txnum);

			glBegin(GL_QUADS);

			glColor3f(1.0f,1.0f,1.0f);

			for (n = 0; n < 4; n++)
			{
				glTexCoord2f(vt[n].vt[0], vt[n].vt[1]);
				glVertex3f (skyboxvtx[vn[i].vn[n]].x,
							skyboxvtx[vn[i].vn[n]].z,
							skyboxvtx[vn[i].vn[n]].y);
			}

			glEnd();
		}
	}
	else
	{
		static int hadskyline = false;
		skylayer_t	*slayer = &skyLayers[skv.active_layer];
		int		idx, printparms = false;

		//---------------------------------------

		if (slayer->texture < 0)
			Con_ErrorD("Rend_RenderSkyHemisphere: Sky layer without a texture");

		int		ht = texArray[ (idx = texturetranslation[slayer->texture]) ]->height;
		int svnum = false;

		globalskytex = texArray[idx]->name ? texArray[idx]->name : defsky;

		// Setup the matrix.
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();	// In case it's rotated

		// The texture is actually loaded when an update is done.
		uint	texn = GL_PrepareSky(slayer->texture);

		// skies 1024 wide or more are considered to stretch round the entire sky
		skv.skyfactor = ctex->rendw <= 512.f ? 1024.f / ctex->rendw : 1.f;

		if (   skv.memh != ctex->rendh
			|| skv.memw != ctex->rendw
			|| (hadskyline^(GLOBAL_SKY_SCROLL_LINE >= 0)) )
		{
			skv.memh = ctex->rendh, skv.memw = ctex->rendw;
			hadskyline = (GLOBAL_SKY_SCROLL_LINE >= 0);
			printparms = true;
		}

		if (ht > (int)skv.memh)
		{
			glScalef(1.0f, 1.2f * ht/skv.memh, 1.0f);			// Doom aspect ratio
		}
		else
		{
			glScalef(1.0f, 1.0f, 1.0f);					// Doom aspect ratio
		}

		skv.rotateAngle = 0;	// doom default

		if (current_skyvarnum >= 0)
		{
			skv.rotateAngle = skyvars[current_skyvarnum].rotate_angle;
		}

		if (skyAngleGlobal)
		{
			skv.rotateAngle = skyAngle;
		}

		if (GLOBAL_SKY_SCROLL_LINE >= 0)
		{
			skv.sky_Yscroll =
				4.f*FIX2FLT(-sides[lines[GLOBAL_SKY_SCROLL_LINE].sidenum[0]].rowoffset)/skv.memh;
		}
		else
			skv.sky_Yscroll = 0;
		
		idx = texturetranslation[slayer->texture];
		
		if (   texArray[idx]->patches[0].patch != skv.skylump
			|| skv.current_skyvarnum != current_skyvarnum)
		{
			fixed_t	yOffset = slayer->sky_offy;
			float	yoffFactor = .49f;	// set to default;
			
			svnum = skv.current_skyvarnum != current_skyvarnum;

			skv.current_skyvarnum = current_skyvarnum;

			if (skv.skylump != texArray[idx]->patches[0].patch)
				printparms = true;

			skv.skylump = texArray[idx]->patches[0].patch;

			if (lumpinfo[skv.skylump].lumptype == LMP_PNG)
			{
				skv.horizon_offset = -.07f;
			}
			else if (texArray[idx]->patchcount == 4 && ht == 240)	//composite skies
			{
				yoffFactor = 2.0f;
				skv.horizon_offset = -.18f;
			}
			else if (texArray[idx]->patchcount == 1	// eliminate composite skies
				&& (ht >= 240 && texArray[idx]->width >=  512))
			{
				yoffFactor *= 1.5f;
				skv.horizon_offset = -.12f;
			}
			else
			{
				skv.horizon_offset = -.105f;	// default
			}

			if (yOffset)
			{
				skv.horizon_offset += 0.32f * (float)((yOffset>>16) % ht)/(float)ht;
				// sky vertex array based on texh of 256.
				skv.maxSideAngle = PI/2.f * .735f;
			}
			else if (lumpinfo[skv.skylump].lumptype == LMP_PNG)
				skv.maxSideAngle = PI/2.f * ctex->rendh/(256.f * 1.2f);
			else
				skv.maxSideAngle = PI/2.f * ctex->rendh/256.f;

			if (ctex->rendw > 1024.f)
			{
				skv.maxSideAngle *= (1024.f/ctex->rendw);
				if (ctex->rendh > 256.f)
					skv.maxSideAngle *= (256.f/ctex->rendh);
			}

			if (skv.last_horizon != skv.horizon_offset)
			{
				skv.last_horizon = skv.horizon_offset;

				if (ma_inf.horizonOffset)
					skv.horizonOffset = (float)(PI/2 * ma_inf.horizonOffset);
				else
					skv.horizonOffset = (float)(PI/2 * skv.horizon_offset);

				skv.lowersky_offset = (float)(-sin(skv.horizonOffset)*2)*EXPAND_FACTOR;	// v220-9

				Rend_SkyDetail();	// Recalculate the vertices.

				skv.yflip = (slayer->flags & SLF_SKYFLIPY) ? true : false;
				printparms = true;
			}
		}

		skv.currentFO = &slayer->fadeout;

		// v220-9 simplify
		// Draw the upper hemispheres.
		// Rotate sky - R3D
		if (skv.rotateAngle) glRotatef(skv.rotateAngle, 0, 1, 0);
		Rend_RenderSkyHemisphere(slayer, texn, SKYHEMI_UPPER);

		glFrontFace(GL_CCW);

		// Draw the lower hemispheres.
		Rend_RenderSkyHemisphere(slayer, texn, SKYHEMI_LOWER);

		glPopMatrix();
		glFrontFace(GL_CW);
	}

	glDepthMask(GL_TRUE);
}

//===========================================================================
// Rend_SkyDetail
//  A hemisphere comprise of NUMCOLUMNS by NUMROWS.
//  This requires NUMCOLUMNS * (NUMROWS + 1) skyVerts
//  The array set is used for both hemis
//===========================================================================
static void Rend_SkyDetail()
{
	int	c, r;
	
	// Calculate the vertices' x,y,z positions
	for (r = 0; r <= NUMROWS; r++)
	{
		for (c = 0; c < NUMCOLUMNS; c++)
		{
			skyvertex_t	*svtx = skyVerts + SKYVTX_IDX(c, r);
			float	topAngle = (float)c/NUMCOLUMNSFP * 2*PI;
			float	sideAngle = skv.horizonOffset + skv.maxSideAngle*(NUMROWS-r)/NUMROWSFP;
			float	realRadius = (float)(cos(sideAngle));

			// v220-9 use EXPAND_FACTOR
			svtx->x = EXPAND_FACTOR * (float)(realRadius * cos(topAngle));
			svtx->y = EXPAND_FACTOR * (float)(sin(sideAngle));  // The height.
			svtx->z = EXPAND_FACTOR * (float)(realRadius * sin(topAngle));
		}
	}
}

//===========================================================================
// Init_SkyParams v220-5
//   Init with default values
//===========================================================================
void Init_SkyParams(void)
{
	int i, k;

	num_skytop_colors = 0;
	has_skybox = false;

	// Note: ded values are ignored
	skv.last_horizon = 16.0f;	// Any value outside the range will do
	skv.horizon_offset = -.105f;
	skv.skylump = 0;				// As yet unset
	skv.maxSideAngle = (float)(PI/2 * .49f);
	skv.horizonOffset = (float)(PI/2 * skv.horizon_offset);
	skv.lowersky_offset = (float)(-sin(skv.horizonOffset)*2)/EXPAND_FACTOR;	// v220-9

	// Initialize the layers.
	for (i = 0; i < MAXSKYLAYERS; i++)
	{
		skylayer_t *sl = &skyLayers[i];

		sl->texture = -1; // No texture.
		sl->fadeout.limit = .3f;
		sl->flags &= ~SLF_SKYFLIPY;	// v220-5

		k = Def_EvalFlags(mapinfo->sky_layers[i].flags);

		if (k & SLF_ENABLED)
		{
			sl->flags |= SLF_ENABLED;
			sl->texture = R_TextureNumForName(mapinfo->sky_layers[i].texture, true);

			if ( (r_skybox = Rend_InitSkyBox(sl)) )
			{
				sl->flags |= SLF_SKYBOX;
				has_skybox = true;
			}
		}
		else
		{
			sl->flags &= ~SLF_ENABLED;
		}
	}
}

int CCmdSkyBox(int argc, char **argv)
{
	if(argc != 2)
	{
		Con_Message( "Usage: skybox skyname\n");
	}
	else if (skyLayers[skv.active_layer].flags & SLF_SKYBOX)
	{
		if (strlen(argv[1]) < 32)
		{
			strcpy(skyboxname, argv[1]);
			Rend_InitSkyBox(&skyLayers[skv.active_layer]);
		}
		else
		{
			Con_Message( " error: skyname length exceeds max allowed of 32\n");
		}
	}
	else
	{
		Con_Message(" skybox can only be changed if the current active layer uses a skybox\n");
	}

	return true;
}
