//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2008 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: s_cache.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
//  Resampling removed.  Sounds now played at their true rate
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Sound sample cache
//
//-----------------------------------------------------------------------------

//** To play a sound:
//**  1) Figure out the ID of the sound.
//**  2) Call Sfx_Cache() to get a sfxsample_t.
//**  3) Pass the sfxsample_t to Sfx_StartSound().

// HEADER FILES ------------------------------------------------------------

#include <string.h>

#include "de_base.h"
#include "de_system.h"
#include "de_console.h"
#include "de_refresh.h"
#include "de_audio.h"
#include "de_misc.h"

#include "doom\sndidx.h"
#include "m_winmem.h"
#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// The cached samples are stored in a hash. When a sample is purged, its
// data will stay in the hash (sample lengths needed by the Logical Sound 
// Manager).
#define CACHE_HASH_SIZE		64

#define PURGE_TIME			(10 * TICSPERSEC)

#define MEAN_CACHE_MByte	0x1000000	// 16MB

// Convert an unsigned byte to signed short (for resampling).
#define U8_S16(b)			(((byte)(b) - 0x80) << 8)

#define DOOM_SND_ID			3

// Time interval before deleting unaccessed sound from cache
#define DEADSOUNDTIME		TICSPERSEC * 60 * 4	 // 4 minutes.

// TYPES -------------------------------------------------------------------

typedef struct sfxcache_s {
	struct	sfxcache_s *next, *prev;
	int		hits;
	uint	lastused;			// Tic the sample was last hit.
	struct	sfxsample_s sample;
} sfxcache_t;

typedef struct cachehash_s {
	sfxcache_t *first, *last;
} cachehash_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static cachehash_t scHash[CACHE_HASH_SIZE];

// CODE --------------------------------------------------------------------

#pragma optimize("g",off)

//===========================================================================
// Sfx_InitCache
//===========================================================================
void Sfx_InitCache(void)
{
	// The cache is empty in the beginning.
	memset(scHash, 0, sizeof(scHash));
}

//===========================================================================
// Sfx_CacheHash
//===========================================================================
static cachehash_t *Sfx_CacheHash(int id)
{
	return &scHash[ (uint) id % CACHE_HASH_SIZE ];
}

//===========================================================================
// Sfx_Uncache
//  Calls to BEGIN_COP and END_COP must be made by the calling routine.
//===========================================================================
static void Sfx_Uncache(sfxcache_t *node)
{
	cachehash_t *hash;

	// Reset all channels loaded with this sample.
	Sfx_UnloadSoundID(node->sample.id);

	// Unlink the node.
	hash = Sfx_CacheHash(node->sample.id);
	if(hash->last  == node) hash->last  = node->prev;
	if(hash->first == node) hash->first = node->next;
	if(node->next) node->next->prev = node->prev;
	if(node->prev) node->prev->next = node->next;

	// Free all memory allocated for the node.
	M_Free(node->sample.sdata);
	M_Free(node);
}

//===========================================================================
// Sfx_ShutdownCache
//===========================================================================
void Sfx_ShutdownCache(void)
{
	int i;

	BEGIN_COP;

	// Uncache all the samples in the cache.
	for(i = 0; i < CACHE_HASH_SIZE; i++)
	{
		while(scHash[i].first) Sfx_Uncache(scHash[i].first);
	}

	memset(scHash, 0, sizeof(scHash));

	END_COP;
}

//===========================================================================
// Sfx_GetCached
//	If the sound is cached, return a pointer to it.
//===========================================================================
static sfxcache_t *Sfx_GetCached(int id)
{
	int			soundId = id & ~HSF_FLAG_MASK;
	sfxcache_t	*it = (Sfx_CacheHash(soundId))->first;

	for (; it; it = it->next)
	{
		if (it->sample.id == soundId) return it;
	}
	return NULL;
}


//===========================================================================
// Sfx_CacheInsert
//	Caches a copy of the given sample. If it's already in the cache and has 
//	the same format, nothing is done. Returns a pointers to the cached 
//	sample. Always returns a valid cached sample.
//  v220-4 resampling removed
//===========================================================================
static sfxcache_t *Sfx_CacheInsert(sfxsample_t *sample)
{
	cachehash_t		*hash;
	sfxcache_t		*node;
	sfxsample_t		cached;

	cached.id    = sample->id;
	cached.group = sample->group;
	cached.rate  = sample->rate;
	cached.size  = sample->numsamples * sample->bytesper;
	cached.bytesper = sample->bytesper;
	cached.numsamples = sample->numsamples;

	if ( (node = Sfx_GetCached(sample->id)) )	// Check if already cached
	{
		if (   node->sample.sdata
			&& node->sample.sdata == sample->sdata
			&& node->sample.size == sample->size
			&& node->sample.rate == sample->rate)
		{
			return node; // The format matches
		}

		// The sound is in the wrong format so stop all sounds using
		// the node, destroy the sample and load the required type
		Sfx_UnloadSoundID(node->sample.id);
		M_Free(node->sample.sdata);
	}
	else
	{
		// Create a new node and link it in.
		node = (sfxcache_t*)M_Calloc(sizeof(sfxcache_t), 320);
		hash = Sfx_CacheHash(sample->id);

		if(hash->last)
		{
			hash->last->next = node;
			node->prev = hash->last;
		}

		hash->last = node;

		if (!hash->first) hash->first = node;
	}

	// Do the resampling, if necessary.
	cached.sdata = M_Malloc(cached.size, 96);

	memcpy(cached.sdata, sample->sdata, sample->size);

	// Hits keep count of how many times the cached sound has been played.
	// The purger will remove samples with the lowest hitcount first.
	node->hits = 0;
	memcpy(&node->sample, &cached, sizeof(cached));
	return node;
}

//===========================================================================
// Sfx_PurgeCache
//  Called periodically by S_Ticker(). If the cache is too large, samples
//  not in use with the lowest hitcount will be uncached.
//  R3D - Sounds that repeat are not purged.
//  Note: no sounds can be played until this routine has been called as a
//  call to END_COP is needed to set allowrefresh to true.
//===========================================================================
void Sfx_PurgeCache(void)
{
	sfxcache_t	*it, *next, *lowest;

	int		i, totalsize = 0;
	uint	nowtime = (uint)((float)Sys_GetRealTime() * 0.035f);

	static int lastpurge = 0;

	//-----------------------------------------

	if(!sfx_avail) return;

	// Is it time for a purge?
	if (nowtime - lastpurge < PURGE_TIME) return;	//No

	BEGIN_COP;

	lastpurge = nowtime;

	// Count the total size of the cache excluding repeating sounds.
	for(i = 0; i < CACHE_HASH_SIZE; i++)
	{
		for(it = scHash[i].first; it; it = next)
		{
			next = it->next;

			if (nowtime - it->lastused > DEADSOUNDTIME)
			{
				Sfx_Uncache(it);
			}
			else
			{
				totalsize += it->sample.size + sizeof(*it);
			}
		}
	}

	// Check cache is within the required maxsize limit.
	// If > then reduce it by deleting sounds not being accessed
	while(totalsize > MEAN_CACHE_MByte) 
	{
		lowest = NULL;

		for(i = 0; i < CACHE_HASH_SIZE; i++)
		{
			for(it = scHash[i].first; it; it = it->next)
			{
				// If the sample is playing we won't remove it now.
				if (Sfx_CountPlaying(it->sample.id, false))
				{
					continue;
				}
				
				// This sample could be removed, let's check the hits.
				if (!lowest || it->hits < lowest->hits)
				{
					lowest = it;
				}
			}
		}

		if (!lowest) break; // No more samples to remove.

		// Remove this cached sample.
		totalsize -= lowest->sample.size + sizeof(*lowest);
		Sfx_Uncache(lowest);
	}	

	END_COP;
}

//===========================================================================
// Sfx_GetCacheInfo
//	Return the number of bytes and samples cached.
//===========================================================================
void Sfx_GetCacheInfo(uint *cacheBytes, uint *sampleCount)
{
	sfxcache_t *it;
	uint size = 0, count = 0;
	int i;

	for(i = 0; i < CACHE_HASH_SIZE; i++)
	{
		for(it = scHash[i].first; it; it = it->next, count++)
			size += it->sample.size;
	}

	if(cacheBytes) *cacheBytes = size;
	if(sampleCount) *sampleCount = count;
}

//===========================================================================
// Sfx_CacheHit
//===========================================================================
RDBOOL Sfx_CacheHit(uint id)
{
	sfxcache_t *node = Sfx_GetCached(id);

	if (!node) return false;

	if (node->hits < 0x7fffffff)	// R3D make no assumptions
		node->hits++;

	node->lastused = Sys_GetTime();
	return true;
}

//===========================================================================
// Sfx_Cache
//	Returns a pointer to the cached copy of the sample. Give this pointer
//	to Sfx_StartSound(). Returns NULL if the sound ID is invalid.
//===========================================================================
sfxsample_t *Sfx_Cache(uint id)
{
	byte		*sp, *ptr;
	sfxcache_t	*node;
	sfxsample_t	samp;				// Temporary data container 
	RDBOOL		needFree = false;
	int			lumpsize, type;
	int			soundId = id & ~HSF_FLAG_MASK;
	sfxinfo_t	*si;

	if (  !sfx_avail
		|| soundId <= 0
		|| soundId >= defs.count.sounds.num
		|| sounds[soundId].flags & SFX_BADSOUND
		|| sounds[soundId].lumpnum < 0)
	{
		return NULL;
	}

	// Check if already cached
	if ( (node = Sfx_GetCached(soundId)) != NULL)
	{
		return &node->sample;
	}

	si = &sounds[soundId];

	// R3D - add lump verification - flag bad sounds
	if ( (lumpsize = W_LumpLength(si->lumpnum)) < 64)
	{
		Con_Message("Sound lumpname %s contains no data.\n", si->lumpname);
		si->flags |= SFX_BADSOUND;
		return NULL;
	}

	type = DoomLumpIdentify(si->lumpnum, NULL, NULL, DLI_SOUND);

	if (type != LMP_SND && type != LMP_WAV)
	{
		Con_Message("Sfx_Cache: unrecognised sound format %i - lump \"%s\"\n",
			type, si->lumpname);

		si->flags |= SFX_BADSOUND;
		return NULL;
	}
	// end verification R3D

	// Init the sample data. A copy of 'samp' will be placed in the cache.
	memset(&samp, 0, sizeof(samp));

	samp.id    = soundId;
	samp.group = si->group;
	samp.sdata = NULL;

	sp = ptr = (byte*)M_CacheLumpNum(si->lumpnum, (uint*)&lumpsize, true);

	if (type == LMP_SND)	// A DOOM format sound
	{
		SOUNDHdr_t	*SNDHdr = (SOUNDHdr_t*)sp;

		samp.rate       = SNDHdr->smplrate;
		samp.numsamples = SNDHdr->datasize;
		samp.sdata      = sp + sizeof(SOUNDHdr_t);

		if (lumpinfo[si->lumpnum].liflags & LINF_TRUNCATED)
			samp.numsamples = lumpinfo[si->lumpnum].size - 8;

		samp.bytesper = 1;	// always a single byte (i.e. 8-bit sound)
	}
	else					// must be a WAV type
	{
		// Load as WAV, then.
		WAV_MemoryLoad(sp, sp + lumpsize, &samp);

		if (!samp.sdata)
		{
			si->flags |= SFX_BADSOUND;	// R3D - flag bad sounds

			Con_Message("Sfx_Cache: WAV data in lump %s is bad.\n", si->lumpname);

			M_Free(ptr);
			return NULL;
		}
		
		needFree = true;
	}

	samp.size = samp.bytesper * samp.numsamples;

	// v220-4 - filter out sounds with a rate > 44100
	if (samp.rate > 44100 || samp.size > 0x700000)
	{
		if(needFree) M_Free(samp.sdata);

		si->flags |= SFX_BADSOUND;	// R3D - flag bad sounds

		if (samp.rate > 44100)
			Con_Message("Sound rate of %i for sound %s exceeds max supported of 44100\n",
				samp.rate, si->lumpname);
		else
			Con_Message("Sound size of %i for sound %s exceeds max allowed of 7Mb\n",
				samp.size, si->lumpname);

		M_Free(ptr);
		return NULL;
	}

	// Insert a copy of this into the cache.
	node = Sfx_CacheInsert(&samp);

	// Check if samp.sdata should be reclaimed
	if(needFree) M_Free(samp.sdata);

	M_Free(ptr);
	
	if (   soundId == r3ds.stnmov_id
		&& type == LMP_SND
		&& (1000*node->sample.numsamples) / node->sample.rate < 300 /*mS*/)
	{
		// In order to lessen channel hogging then one strategy used here is
		// to make a sound that is normally played once but repeatedly called
		// into a sound that actually repeats (thus using just one channel).
		// A case in point is DSSTNMOV. In order to differentiate calls to
		// play this repeatedly (from the plats code) from a genuine one off
		// and also to ensure that where replaced the original is actually
		// replaced then the container r3ds.stnmov_id is used.
		// The short duration of stnmov may, however, not be refreshed soon
		// enough.  Either the refresh thread can be called more frequently
		// or the sound size increased to ensure it is not completed before
		// refreshing is done.  Doubling the sound size has been adopted
		// because, on balance, it is the more efficient solution - R3D

		byte *data = (byte*)M_Malloc(2*node->sample.size, 0);

		memcpy(data, node->sample.sdata, node->sample.size);
		memcpy(data + node->sample.size, node->sample.sdata, node->sample.size);

		M_Free(node->sample.sdata);
		node->sample.sdata = data;
		node->sample.numsamples *= 2;
		node->sample.size *= 2;
	}

	return &node->sample;
}

void Sfx_PreCache()
{
	sfxsample_t *sample;
	sfxinfo_t	*si;
	int			i, n, type;
	int			totalsize = 0, num = 0;

	if (!sfx_avail) return;

	for (i = 0, si = sounds; i < count_sounds.num; i++, si++)
	{
		if (   (n = si->lumpnum) > 0
			&& !(si->flags & SFX_BADSOUND)
			&& lumpinfo[n].size
			&& ((type = DoomLumpIdentify(n, NULL, NULL, DLI_SOUND)) == LMP_SND || type == LMP_WAV)
			&& !Sfx_GetCached(i)
			&& totalsize < MEAN_CACHE_MByte
			&& (sample = Sfx_Cache(i))
			&& sample->size)
		{
			totalsize += sample->size; num++;
		}
	}

	if (num)
		pr_OFinsert("  Pre cached %3i sounds   : cache used %.2f MB\n", num, float(totalsize)/1000000.f);
}

//===========================================================================
// Sfx_GetSoundLength
//	Returns the length of the sound (in milliseconds).
//===========================================================================
uint Sfx_GetSoundLength(uint id)
{
	sfxsample_t *sample = Sfx_Cache(id);

	if(sample)
		return (1000 * sample->numsamples) / sample->rate;

	return 0;	// No idea.
}