//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2009 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: s_sfx.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  Modified for R3D
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
//  Modified to get rid of looping buffers where the sound is
//  only played once. sidrv function array introduced as a pre-
//  cursor to allow the support of a different sound API (OAL
//  for example).
//  Note: not all changes marked as too extensive (search sidrv).
// Revision 2.2.0 build 5 (v220-5) 20/Dec/2009 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  sound effects handler
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <math.h>
#include <eax.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_play.h"
#include "de_refresh.h"
#include "de_graphics.h"
#include "de_audio.h"
#include "de_misc.h"
#include "r3d_OGLfunc.h"
#include "r3d_compat.h"
#include "r3d_sfx.h"
#include "sound_var.h"

// MACROS ------------------------------------------------------------------

#define SFX_LOWEST_PRIORITY		-1000
#define UPDATE_TIME				(2.0/TICSPERSEC)	// ~57mS
#define LOOKDIR2DEG(x)			((x) * 85.0/110.0)

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

int		Def_GetSoundNum(char *id);
RDBOOL	Sfx_SetNodeSuspend(int id, mobj_t *origin);
RDBOOL	Sfx_DelNodeSuspend(int id, mobj_t *origin);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern sidriver_t sid3d;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

sfxchannel_t	snd_channels[SFX_MAX_CHANNELS];
r3dsound_t		r3ds;
int				num_channels = SFX_MAX_CHANNELS;

RDBOOL	sfx_avail = false;
RDBOOL	sfx_3d	  = false;
RDBOOL	allowrefresh = false;
float	sfx_reverb_strength = 1.0f;

sidriver_t sidrv;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

unsigned long	refreshthreadid;
unsigned long	refreshparam;

static mobj_t	*listener;
static int		sfx_dedicated_2d;

static int refresh_handle;
static RDBOOL refreshing;

static RDBOOL refmonitor = 0;
static sfxchannel_t *chan_is_available[3];

// b220-0 These are used for channel limitation with repeating sounds
static float distances[SFX_MAX_CHANNELS];

// CODE --------------------------------------------------------------------

#pragma optimize("g",off)

//===========================================================================
// Sfx_ChannelRefreshThread
//	This is a high-priority thread that periodically checks if the 
//	channels need to be updated with more data. The thread terminates
//	when it notices that the channels have been destroyed.
//	The Sfx driver maintains a 250ms buffer for each channel, which 
//	means the refresh must be done often enough to keep them filled.
//   Note: wait times must take into account the duration of
//         the shortest sound likely to looped (e.g. DSSTNMOV)
// v220-4 This is only started with DSound
//===========================================================================
static int Sfx_ChannelRefreshThread(void *parm)
{
	sfxchannel_t *ch;
	int			 i;

	// We'll continue looping until the Sfx module is shut down.
	while (sfx_avail)
	{
		// The bit is swapped on each refresh (debug info).
		refmonitor ^= 1;

		if (allowrefresh)
		{
			refreshing = true;

			// Do the refresh.
			for (i = 0, ch = snd_channels; i < num_channels; i++, ch++)
				sidrv.SoundRefresh(ch);

			refreshing = false;
		}

		Sys_Sleep(50);	// wait for allowrefresh.
	}

	return 0;	// Exiting with 0 will cause this thread to end.
}

//===========================================================================
// Sfx_AllowRefresh
//	Enabling refresh is simple: the refresh thread is resumed.
//	When disabling refresh, first make sure a new refresh doesn't
//	begin (using allowrefresh). We still have to see if a refresh
//	is being made and wait for it to stop. Then we can suspend the 
//	refresh thread.
//===========================================================================
void Sfx_AllowRefresh(RDBOOL allow)
{
	if (   !refresh_handle
		|| allowrefresh == allow) // No change.
	{
		return;
	}

	allowrefresh = allow;

	// If we're denying refresh then ensure, if it's currently
	// running, that we first allow it to complete.
	if (!allow)
	{
		while (refreshing) Sys_Sleep(1);
	}

	Sys_SuspendThread(refresh_handle, !allow);
}

//===========================================================================
// Sfx_StopSoundGroup
//	Stop all sounds of the group. If an emitter is specified, only its
//	sounds are checked.
//===========================================================================
void Sfx_StopSoundGroup(int group, mobj_t *emitter)
{
	sfxchannel_t *ch;
	int i;

	for(i = 0, ch = snd_channels; i < num_channels; i++, ch++)
	{
		if (  !(ch->buffer.SFXBFflags & SFXBF_PLAYING)
			||  ch->buffer.sample->group != group
			|| (emitter && ch->emitter != emitter))
		{
			continue;
		}
	
		sidrv.SoundStop(ch);	// This channel must stop.
	}
}

//===========================================================================
// Sfx_CheckChannel
//  An id and origin must be sent
//===========================================================================
RDBOOL Sfx_CheckChannel(int id, mobj_t *origin)
{
	sfxchannel_t *ch = snd_channels;
	int i, sound_id = id & ~HSF_FLAG_MASK;

	for (i = 0; i < num_channels; i++, ch++)
	{
		if (   ch->emitter == origin
			&& ch->buffer.sample
			&& ch->buffer.sample->id == sound_id)
		{
			return (ch->buffer.SFXBFflags & SFXBF_PLAYING) > 0;
		}
	}
	return false;
}

//===========================================================================
// Sfx_StopSound
//===========================================================================
int Sfx_StopSound(int id, mobj_t *emitter)
{
	sfxchannel_t	*ch;
	int				i, stopCount = 0;
	int				soundid = id & ~HSF_FLAG_MASK;

	if(!sfx_avail) return 0;

	for(i = 0, ch = snd_channels; i < num_channels; i++, ch++)
	{
		if (   !(ch->buffer.SFXBFflags & SFXBF_PLAYING)
			|| (soundid && ch->buffer.sample->id != soundid)
			|| (emitter && ch->emitter != emitter))
		{
			continue;
		}

		// Can it be stopped?
		if (ch->buffer.SFXBFflags & SFXBF_DONT_STOP)
		{
			// The emitter might get destroyed...
			ch->emitter = NULL;
			ch->SFXCFflags |= (SFXCF_NO_UPDATE|SFXCF_NO_ORIGIN);
			continue;
		}

		if (emitter && ch->buffer.SFXBFflags & (SFXBF_SUSPEND|SFXBF_REPEAT))
			Sfx_DelNodeSuspend(ch->buffer.sample->id & HSF_FLAG_MASK, emitter);

		// This channel must be stopped!
		sidrv.SoundStop(ch);
		stopCount++;
	}

	return stopCount;
}

//===========================================================================
// Sfx_SuspendChannel b220-0 for used with suspended sounds
//  Suspends channel if suspend == true else restarts channel
//===========================================================================
RDBOOL Sfx_SuspendChannel(int id, mobj_t *emitter, RDBOOL suspend)
{
	sfxchannel_t *ch;
	int			 i;
	int			 soundid = id & ~HSF_FLAG_MASK;

	if (!id || id >= count_sounds.num || !sfx_avail) return false;

	for(i = 0, ch = snd_channels; i < num_channels; i++, ch++)
	{
		if (   ch->buffer.sample
			&& ch->emitter == emitter
			&& ch->buffer.sample->id == soundid)
		{
			if (suspend)
			{
				// Check if already suspended
				if (ch->buffer.SFXBFflags & SFXBF_SUSPEND)
				{
					Sfx_SetNodeSuspend(soundid, emitter);
					return true;
				}
	
				ch->buffer.SFXBFflags |= SFXBF_SUSPEND;

				sidrv.SoundStop(ch);

				return true;
			}
			else if (ch->buffer.SFXBFflags & SFXBF_SUSPEND)
			{
				Sfx_DelNodeSuspend(soundid, emitter);
				sidrv.SoundRestart(ch);
			}
		}
	}
	return false;
}

//===========================================================================
// Sfx_UnloadSoundID
//	The specified sample will soon no longer exist. All channel buffers
//	loaded with the sample will be reset.
//===========================================================================
void Sfx_UnloadSoundID(int id)
{
	sfxchannel_t *ch;
	int i;
	
	if(!sfx_avail) return;

	BEGIN_COP;
	for(i = 0, ch = snd_channels; i < num_channels; i++, ch++)
	{
		if (   ch->buffer.sample
			&& ch->buffer.sample->id == id)
		{
			if (ch->buffer.SFXBFflags & (SFXBF_SUSPEND|SFXBF_REPEAT))
				Sfx_DelNodeSuspend(ch->buffer.sample->id & HSF_FLAG_MASK, ch->emitter);

			sidrv.SoundReset(ch);	// Stop and unload.
		}
	}
	END_COP;
}

//===========================================================================
// Sfx_CountPlaying
//	Returns the number of channels the sound is playing on.
//===========================================================================
int Sfx_CountPlaying(int id, RDBOOL notlistener)
{
	sfxchannel_t *ch = snd_channels;
	int i, c;
	
	if (!sfx_avail) return 0;

	for (i = c = 0; i < num_channels; i++, ch++)
	{
		if (notlistener && ppt.viewplrmo == ch->emitter) continue;

		if (   !ch->buffer.sample				// No loaded sample.
			|| ch->buffer.sample->id != id)	// Not the requested sample.
		{
			continue;
		}

		if (ch->buffer.SFXBFflags & SFXBF_PLAYING)
			c++;
	}
	return c;
}

//===========================================================================
// Sfx_Priority
//	The priority of a sound is affected by distance, volume and age.
//===========================================================================
static float Sfx_Priority(mobj_t *emitter, float *fixpos,
						  float volume, float attenuation,
						  uint starttic, int chan_num)
{
    // In five seconds all priority of a sound is gone.
    float   timeoff = 1000 * ((uint)Sys_GetTimef() - starttic) / (5.0f * TICSPERSEC);
    float   orig[3];

    if (!listener || (!emitter && !fixpos))
    {
        // The sound does not have an origin.
        return 1000 * volume - timeoff;
    }

    // The sound has an origin, base the points on distance.
    if (emitter)
    {
        orig[VX] = FIX2FLT(emitter->x);
        orig[VY] = FIX2FLT(emitter->y);
        orig[VZ] = FIX2FLT(emitter->z);
    }
    else
    {
        // No emitter mobj, use the fixed source position.
        memcpy(orig, fixpos, sizeof(orig));
    }


	if (   snd_channels[chan_num].buffer.SFXBFflags & SFXBF_DONT_STOP
		|| snd_channels[chan_num].emitter == ppt.viewplrmo)
		distances[chan_num] = 0;
	else
		distances[chan_num] = P_MobjSoundDistancef(NULL, orig, attenuation);

    return 1000 * volume - distances[chan_num] * 0.5f - timeoff;
}

///===========================================================================
// Sfx_ChannelPriority
//===========================================================================
static float Sfx_ChannelPriority(int chan_num)
{
	sfxchannel_t *ch  = snd_channels + chan_num;
	float		 *pos = NULL;
	
	if ( !(ch->buffer.SFXBFflags & SFXBF_PLAYING) )
	{
		distances[chan_num] = 32768.0f;
		return SFX_LOWEST_PRIORITY;
	}
	
	// ch->pos is set to emitter->xyz during updates.
	if (!(ch->SFXCFflags & SFXCF_NO_ORIGIN)) pos = ch->nowpos;
	
	return Sfx_Priority(ch->emitter, pos,
		ch->volume, ch->attenuation, ch->chstarttime, chan_num);
}

//===========================================================================
// Sfx_ChannelUpdate
//	Updates the channel buffer's properties based on 2D/3D position 
//	calculations. Listener might be NULL. Sounds emitted from the listener
//	object are considered to be inside the listener's head.
//===========================================================================
static void Sfx_ChannelUpdate(sfxchannel_t *ch)
{
	RDBOOL		mobjThinker = false;
	RDBOOL		isListener = false;
	RDBOOL		ambient = ch->SFXCFflags & SFXCF_AMBIENT;
	float		chvolume;
	float		vec[3];

	if (!sfx_avail || ch->SFXCFflags & SFXCF_NO_UPDATE) return;

	// Frequency is common to both 2D and 3D sounds.
	sidrv.SoundSet(ch, SFXBP_FREQUENCY, ch->frequency);

	// Copy the emitter's position (if any), to the pos coord array.
	if (ch->emitter)
	{
		if ( (isListener = (ch->emitter == listener)) )
		{
            vec[VX] = vec[VY] = vec[VZ] = 0;
		}
		else
		{
			vec[VX] = FIX2FLT( ch->emitter->x );
			vec[VY] = FIX2FLT( ch->emitter->y );
			vec[VZ] = FIX2FLT( ch->emitter->z );
		}

		mobjThinker = ch->emitter->thinker.function == (void (__cdecl *)(void))P_MobjThinker;

		// If this is a mobj, center the Z pos.
		if (listener && !isListener && mobjThinker)
		{
		//	vec[VY] += FIX2FLT( ch->emitter->height >> 1 );
			vec[VZ] += FIX2FLT(listener->height - 0x50000);
		}
	}
	else	// must be fixed pos
	{
		vec[VX] = ch->realpos[VX];
		vec[VY] = ch->realpos[VY];
		vec[VZ] = ch->realpos[VZ];
	}

	if (isListener)
	{
		ch->nowpos[VX] = FIX2FLT(listener->x);
		ch->nowpos[VY] = FIX2FLT(listener->y);
		ch->nowpos[VZ] = FIX2FLT(listener->z);
	}
	else
	{
		ch->nowpos[VX] = vec[VX];
		ch->nowpos[VY] = vec[VY];
		ch->nowpos[VZ] = vec[VZ];
	}


	chvolume = ch->volume * sfx_volume/255.0f;

    if (ch->buffer.SFXBFflags & SFXBF_3D)
    {
        if (isListener || ch->SFXCFflags & SFXCF_NO_ORIGIN)
        {
			// Emitted by the listener object or unattenuated.
			// Go to relative position mode; pos = (0,0,0)
			sidrv.SoundSet(ch, SFXBP_RELATIVE_MODE, 0);
			sidrv.SoundSetv(ch, SFXBP_POSITION, vec);
			// Volume is affected only by maxvol.
			sidrv.SoundSet(ch, SFXBP_VOLUME, chvolume);
        }
        else
        {
			float distance = P_MobjSoundDistancef(NULL, vec, ch->attenuation);
			float v = 1.0f;

			if (ch->attenuation && ch->attenuation != 1.0f)
			{
				float dist1 = SOUND_INTER_DISTANCE;
				float dist2 = SOUND_MAX_DISTANCE;
				float atten;

				// v220-35 change range according to attenuation
				if (ch->attenuation > 1.0f)
				{
					atten = 2.f/(ch->attenuation + 1.f);
				}
				else
				{
					atten = 1.f/ch->attenuation;
				}

				dist1 *= atten;
				dist2 *= atten;
				distance *= atten;

				sidrv.SoundSet(ch, SFXBP_MAX_DISTANCE, dist2);

				// 3D buffers refuse to roll-off the volume to 0 irrespective of the distance.
				if (distance > dist1)
					v = (dist2 - distance)/(dist2 - dist1);
			}
			else
			{
				sidrv.SoundSet(ch, SFXBP_MAX_DISTANCE, SOUND_MAX_DISTANCE);
				// 3D buffers refuse to roll-off the volume to 0 irrespective of the distance.
				if (distance > SOUND_INTER_DISTANCE)
					v = (SOUND_MAX_DISTANCE - distance)/(SOUND_MAX_DISTANCE - SOUND_INTER_DISTANCE);
			}

		//	if (ch->SFXCFflags & SFXCF_AMBIENT)
			{
				chvolume = chvolume * v;
				if (chvolume > 1.f) chvolume = 1.f;
				if (chvolume < 0.f) chvolume = 0.f;
			}

			sidrv.SoundSet(ch, SFXBP_VOLUME, chvolume);
			sidrv.SoundSet(ch, SFXBP_NORMAL_MODE, 0);
			sidrv.SoundSetv(ch, SFXBP_POSITION, vec);
        }

        if (   mobjThinker
			&& !isListener
			&& !(ch->buffer.SFXBFflags & SFXBF_REPEAT)
			&& ch->attenuation)
        {
			// Only doppler shift missiles the normal value - for any other sound
			// it is reduced by a further factor of 4 (else it sounds terrible)
			int factor = ch->emitter->flags & MF_MISSILE ? FRACBITS : FRACBITS + 2;

            vec[VX] = (float)(ch->emitter->momx>>factor) * TICSPERSEC;
            vec[VY] = (float)(ch->emitter->momy>>factor) * TICSPERSEC;
            vec[VZ] = (float)(ch->emitter->momz>>factor) * TICSPERSEC;
        }
        else
        {
			// fixedPos and listener speeds are always taken as zero.
            vec[VX] = vec[VY] = vec[VZ] = 0;   // Not moving.
        }

        sidrv.SoundSetv(ch, SFXBP_VELOCITY, vec);
    }
	else // This is a 2D buffer.
	{
		if (isListener || ch->SFXCFflags & SFXCF_NO_ORIGIN)
		{
			sidrv.SoundSet(ch, SFXBP_VOLUME, ch->volume * sfx_volume/255.0f);
			sidrv.SoundSet(ch, SFXBP_PAN, 0.0f);
		}
		else
		{
			float distance = P_MobjSoundDistancef(NULL, vec, ch->attenuation);
			float pan, angle, dist;

			//-----------------------------------

			angle = pan = 0.0f;	// unless decided otherwise

			if (distance <= SOUND_MIN_DISTANCE)
			{
				dist = 1.0f;
			}
			else if (distance >= SOUND_MAX_DISTANCE)
			{
				dist = 0;
			}
			else
			{
				float normdist = ((distance - SOUND_MIN_DISTANCE) / (SOUND_MAX_DISTANCE - SOUND_MIN_DISTANCE));

				dist = .125f/(.125f + normdist) * (1.0f - normdist);
			}

			// Pan
			if (listener && dist && distance > 16.0f)
			{
				// Calculate angle from listener to emitter.
				angle = (R_PointToAngle2(listener->x, listener->y,
					(fixed_t)(vec[VX] * FRACUNIT),
					(fixed_t)(vec[VY] * FRACUNIT)) 
					- listener->angle) / (float) ANGMAX * 360;

				// We want a signed angle.
				if (angle > 180) angle -= 360;

				// Set panning values
				if (angle <= 90 && angle >= -90)
				{
					// Front half.
					pan = -angle/90.0f;
				}
				else 
				{
					// Back half.
					pan = (angle + (angle > 0 ? -180.0f : 180.0f)) / 90.0f;

					// Dampen sounds coming from behind
					if (dist < 1.0f)
						dist *= (1 + (pan > 0 ? pan : -pan)) / 2;
				}
			}

			sidrv.SoundSet(ch, SFXBP_VOLUME, ch->volume * dist * sfx_volume/255.0f);
			sidrv.SoundSet(ch, SFXBP_PAN   , pan);
		}
	}
}

//===========================================================================
// Sfx_ListenerUpdate
//===========================================================================
static void Sfx_ListenerUpdate(void)
{
	float vec[NUM_REVERB_DATA];

	//-----------------------

	if (   !sfx_avail
		|| !sfx_3d
		|| !sfx_volume		// No volume means no sound.
		|| !listener
		|| !listener->dplayer)		// Set by calling function Sfx_Update()
	{
		return;
	}

	// Update the listener mobj. Position at eye-level.
	vec[VX] = FIX2FLT( listener->x );
	vec[VY] = FIX2FLT( listener->y );
	vec[VZ] = FIX2FLT( listener->z + listener->height - 0x50000 );

	sidrv.SoundListenerv(SFXLP_POSITION, vec);

	vec[0] = listener->angle / (float) ANGMAX * 360;
	vec[1] = (float)LOOKDIR2DEG(listener->dplayer->lookdir);

	// Convert to radians
	vec[0] *= PI/180;	// yaw
	vec[1] *= PI/180;	// pitch

	sidrv.SoundListenerv(SFXLP_ORIENTATION, vec);

	// Velocity. The unit is world distance units per second.
	vec[VX] = FIX2FLT( listener->momx ) * TICSPERSEC;
	vec[VY] = FIX2FLT( listener->momy ) * TICSPERSEC;
	vec[VZ] = FIX2FLT( listener->momz ) * TICSPERSEC;

	sidrv.SoundListenerv(SFXLP_VELOCITY, vec);

	// Reverb effects.
	if (underwater)
	{
		vec[SRD_VOLUME]  = 1.0f;
		vec[SRD_PRESET]  = EAX_ENVIRONMENT_UNDERWATER;
		vec[SRD_DECAY]   = EAXLISTENER_DEFAULTDECAYTIME;
		vec[SRD_DAMPING] = EAXLISTENER_MINDECAYTIME;
	}
	else
	{
		memcpy(vec, listener->subsector->sector->reverb, sizeof(vec));
	}

	vec[SRD_VOLUME] *= sfx_reverb_strength;

	sidrv.SoundListenerv(SFXLP_REVERB, vec);

	// Update all listener properties.
	sidrv.SoundListener(SFXLP_UPDATE, 0);
}

//===========================================================================
// Sfx_ListenerReverbDoppler
//===========================================================================
static void Sfx_ListenerReverbDoppler(void)
{
	float rev[4] = { 0, 0, 0, 0 };

	listener = NULL;

	if(!sfx_avail) return;

	sidrv.SoundListenerv(SFXLP_REVERB, rev);
	sidrv.SoundListener (SFXLP_UPDATE, 0);

	// While we are at it init the doppler default
	// these will remain at the default values
	sidrv.SoundListener(SFXLP_UNITS_PER_METER, 30.0f);
	sidrv.SoundListener(SFXLP_DOPPLER, 0.2f);	// was cfg.doppler
}

//===========================================================================
// Sfx_ChannelStop
//	Stops the sound playing on the channel.
//	Just stopping a buffer doesn't affect refresh.
//===========================================================================
static void Sfx_ChannelStop(sfxchannel_t *ch)
{
	ch->emitter = NULL;
	ch->buffer.SFXBFflags &= ~SFXBF_SUSPEND;
	sidrv.SoundStop(ch);
}

//===========================================================================
// Sfx_ChannelFindVacant
//===========================================================================
static sfxchannel_t *Sfx_ChannelFindVacant(RDBOOL play3d, sfxsample_t *sample, uint nowtime)
{
	sfxchannel_t *ch;
	int i = play3d ? sfx_dedicated_2d : 0;
	int e = play3d ? num_channels : sfx_dedicated_2d;

	for (ch = snd_channels + i; i < e; i++, ch++)
	{
		if (ch->buffer.SFXBFflags & SFXBF_PLAYING)
		{
			continue;
		}

		if (ch->buffer.sample)
		{
			// If type then a buffer with the same rate/bytes is required 
			if (   ch->buffer.sample->id  == sample->id
				&& ch->buffer.rate  == sample->rate
				&& ch->buffer.bytes == sample->bytesper)
			{
				return ch;	// Meets the required conditions
			}

			if (  !chan_is_available[1]
				&& ch->buffer.rate  == sample->rate
				&& ch->buffer.bytes == sample->bytesper)
			{
				chan_is_available[1] = ch;
			}
			else if (!chan_is_available[0])
			{
				chan_is_available[0] = ch;
			}
		}
		else if (!chan_is_available[2])
			chan_is_available[2] = ch;
	}
	return NULL;
}

//===========================================================================
// Sfx_StartSound
//	Used by the high-level sound interface to play sounds on this system.
//	If emitter is NULL, fixedpos is checked for a position. If both emitter
//	and fixedpos are NULL, then the sound is played as centered 2D. Freq is
//	relative and modifies the sample's rate. Returns true if a sound is 
//	started.
//
//	The 'sample' pointer must be persistent. No copying is done here.
//===========================================================================
int Sfx_StartSound(sfxsample_t *sample,
				   float volume,
				   float freq,
				   mobj_t *emitter,
				   float *fixedpos,
				   float attenuation,
				   float dist,
				   int flags)
{
	sfxchannel_t	*ch, *selch = NULL;
    float			myprio, lowprio = 0, channel_prios[SFX_MAX_CHANNELS];
	int				i, count, maxchan, start, end;
	uint			nowtime;
	RDBOOL			have_channel_prios = false;
	RDBOOL			play3d = sfx_3d && !(flags & SFX_MENUONLY);

	uint			SXBFflags = 0;

	//------------------------------------------

	if (flags & SFX_REPEAT   ) SXBFflags |= SFXBF_REPEAT;
	if (flags & SFX_DONT_STOP) SXBFflags |= SFXBF_DONT_STOP;
	if (!attenuation         ) SXBFflags |= SFXBF_NOATTN;
	if (play3d && gamestate == GS_LEVEL) SXBFflags |= SFXBF_3D;

	play3d = play3d && SXBFflags & SFXBF_3D;

	start = play3d ? sfx_dedicated_2d : 0;
	end   = play3d ? num_channels : sfx_dedicated_2d;

	// Pick a channel for the sound. We will do our best to play the sound,
	// cancelling existing ones if need be. The best choice would be a 
	// free channel already loaded with the sample, in the correct format
	// and mode.
	if (!attenuation && flags & SFX_REPEAT)
	{
		i = start;

		// Check for global sounds.
		for (ch = snd_channels + i; i < end; i++, ch++)
		{
			if (   ch->buffer.sample
				&& ch->buffer.sample->id == sample->id
				&& ch->buffer.SFXBFflags & SFXBF_PLAYING)
			{
				if (!ch->attenuation)
					return true;		// Already playing

				// Stop any attenuated types playing if we're
				// going to start an unattenuated sound
				if (ch->attenuation) Sfx_ChannelStop(ch);
			}
		}
	}

	if (flags & SFX_HALFATTN) attenuation = 0.5f;

    // Calculate the new sound's priority.
    nowtime = (uint)Sys_GetTimef();
	myprio = 1000.f * volume - dist * 0.5f;

	// Check that there aren't too many channels playing this sample.
	// b220-0 Repeating sounds are allowed just two instances apart
	//        from sfx_stnmov which is allowed 4 (set in audio.ded).

	maxchan = (sounds + sample->id)->sfxmaxchan;

	if (emitter == ppt.viewplrmo)	// hi priority
		maxchan = -1;
	else if (flags & SFX_REPEAT)	// lo priority
		maxchan = maxchan < 1 ? 2 : maxchan;
	else if (maxchan < 1)	// standard priority
		maxchan = 8;

	maxchan = play3d ? maxchan : sfx_dedicated_2d;

	if (maxchan > 0)
	{
		// b220-0 The decision to stop channels is based on priorities
		// and/or distance. For repeating sounds just distance is used.
		for (i = start; i < end; i++)
			channel_prios[i] = Sfx_ChannelPriority(i);

		have_channel_prios = true;
		count = Sfx_CountPlaying(sample->id, true);

		while (count >= maxchan)
		{
			float testdist = dist;

			// Stop the lowest priority sound of the playing instances,
			// again noting sounds that are more important than us.
			for (selch = NULL, i = start; i < end; i++)
			{
				ch = &snd_channels[i];

				if (   ch->buffer.SFXBFflags & SFXBF_PLAYING
					&& ch->buffer.sample->id == sample->id)
				{
					if (flags & SFX_REPEAT)
					{
						if (distances[i] > testdist)
						{
							selch = ch;
							testdist = distances[i];
						}
					}
					else
					{
						if (   myprio >= channel_prios[i]
							&& (!selch || channel_prios[i] <= lowprio))
						{
							selch = ch;
							lowprio = channel_prios[i];
						}
					}
				}
			}

			if (!selch)
			{
				// The new sound can't be played because we were unable
				// to stop enough channels to accommodate the limitation.
				return false;
			}

			// Stop this one.
			Sfx_ChannelStop(selch);
			count--;
		}
	}

	// Hitcount tells how many times the cached sound has been used.
	Sfx_CacheHit(sample->id);

	BEGIN_COP;

	// If selch == NULL then the great channel hunt begins.
	// Note in all cases the buffer type must match; i.e. 2D or 3D
	// First look through the stopped channels
	if (!selch)
	{
		memset(chan_is_available, 0, sizeof(chan_is_available));

		// First look for a perfect match.
		selch = Sfx_ChannelFindVacant(play3d, sample, nowtime);

		// If not set then check the chan_is_available array
		if (   !selch
			&& !(selch = chan_is_available[2])
			&& !(selch = chan_is_available[1])
			&& !(selch = chan_is_available[0]))
		{
			// hard cheddar
		}
	}

	if (!selch)
	{
		sfxchannel_t *prioch = NULL;

		// Fourth try: find and stop a channel based on priorities
        if (!have_channel_prios)
		{
			for (i = start; i < end; i++)
				channel_prios[i] = Sfx_ChannelPriority(i);
		}

		// We want the lowest priority sound so all channels must be checked
		for (i = start; i < end; i++)
		{
			ch = &snd_channels[i];

			// Getting here should mean the following fails - insurance only
			if (   !ch->buffer.sample
				|| !(ch->buffer.SFXBFflags & SFXBF_PLAYING))
			{
				selch = ch;
				break;
			}

            // Are we more important than this sound?
            // We want to choose the lowest priority sound.
            if (   myprio >= channel_prios[i]
				&& (!prioch || channel_prios[i] <= lowprio))
            {
                prioch = ch;
                lowprio = channel_prios[i];
            }
		}

        // If a good low-priority channel was found, use it.
        if (prioch)
        {
            selch = prioch;
            Sfx_ChannelStop(selch);
        }
	}

	if (!selch) 
	{
		END_COP;
		return false; // We couldn't find a suitable channel. For shame.
	}

	if ( (selch->buffer.SFXBFflags & SFXBF_3D)^(SXBFflags & SFXBF_3D) )
	{
		END_COP;
		Con_Error("Program error: channel [%i] has the wrong format %i %i\n",
			selch - snd_channels, selch->buffer.SFXBFflags & SFXBF_3D, SXBFflags & SFXBF_3D);
	}

	// Does our channel need to be reformatted?
	if (   selch->buffer.sample != sample
		|| selch->buffer.SFXBFflags & SFXBF_RELOAD
		|| ((selch->buffer.SFXBFflags & SFXBF_REPEAT) ^ (SXBFflags & SFXBF_REPEAT)) )
	{
		uint	size;
		int		err;
		
		if (SXBFflags & SFXBF_REPEAT)
			size = sample->rate * sample->bytesper / 2;	// DS buffer limited to holding 500mS worth of data
		else
			size = sample->size;

		sidrv.SoundDestroyBuffer(selch);

		// Create a new buffer with the correct format.
		err = sidrv.SoundCreateBuffer(
				selch,
				play3d ? SFXBF_3D : 0,
				sample->bytesper,
				sample->rate,
				size);

		if (err)
		{
			END_COP;

			if (err == 1)
				Con_ErrorD("SoundCreateBuffer D3D: error");
			if (err == 2)
				Con_ErrorD(
				"System failure: D3D Failed to create buffer\n"   
				"Ensure no changes have been made to dlls in Risen3D\\bin\n"  
				" Your sound hardware may need upgrading");
			if (err == 3)
				Con_ErrorD("ERROR D3D: tried to create 3D buffer with no ds_listener");
			if (err == 4)
				Con_ErrorD("ERROR D3D: Failed to get 3D interface:\n"
						  " Your sound hardware may need upgrading");

			Con_Error("SoundCreateBuffer: error %i unknown\n", err);
		}

		SXBFflags |= SFXBF_RELOAD;
	}

	// Set buffer flags.
	selch->buffer.SFXBFflags = SXBFflags;
	selch->SFXCFflags = 0;

	if (flags & SFX_AMBIENT)
		selch->SFXCFflags |= SFXCF_AMBIENT;

	// Init the channel information.
	selch->volume = volume;
	selch->frequency = freq;
	selch->attenuation = attenuation;
	selch->chstarttime = nowtime;
	selch->emitter = emitter;	// even if NULL

	if (!attenuation || (!emitter && !fixedpos))
	{
		selch->SFXCFflags |= SFXCF_NO_ORIGIN;
	}
	else if (!emitter && fixedpos)
	{
		memcpy(selch->realpos, fixedpos, sizeof(selch->realpos));
	}

	// Load in the sample. Must load prior to setting properties, because
	// the driver might actually create the real buffer only upon loading.
	if (   selch->buffer.SFXBFflags & SFXBF_RELOAD
		&& !(sidrv.SoundLoad(selch, sample)) )
	{
		END_COP;
		return false;
	}

	if (!selch->buffer.sample)
	{
		END_COP;
		Con_Error("Error ch %i : buffer had no sample", selch - snd_channels);
	}

	if (play3d)	// 3D sounds need a few extra properties set up.
	{
		// Set this buffer's default min/max distances.
		// They could be changed later
		sidrv.SoundSet(selch, SFXBP_MIN_DISTANCE, SOUND_MIN_DISTANCE);
		sidrv.SoundSet(selch, SFXBP_MAX_DISTANCE, SOUND_MAX_DISTANCE);
	}

	Sfx_ChannelUpdate(selch);				// Update channel properties.
	sidrv.SoundListener(SFXLP_UPDATE, 0);	// Commit deferred properties
	sidrv.SoundPlay(selch);					// Start playing.

	END_COP;

	return true;
}

void Sfx_LineSounds(void)
{
	if (leveltime&0x1f) return;

	for (int i = 0; i < MAXLINESOUNDS; i++)
	{
		if (soundstarts[i])
			S_SoundRepeat(soundstarts[i]|(HSF_REPEAT|HSF_NO_ATTENUATION), NULL, 1.0f, 0);
	}
}

void Sfx_SuspendRepeatingChannels()
{
	if (!sfx_avail) return;

	sfxchannel_t	*ch;
	int				i;

	BEGIN_COP;

	for (i = 0, ch = snd_channels; i < num_channels; i++, ch++)
	{
		if ( (ch->buffer.SFXBFflags & (SFXBF_PLAYING|SFXBF_REPEAT|SFXBF_SUSPEND)) == (SFXBF_PLAYING|SFXBF_REPEAT))
		{
			ch->buffer.SFXBFflags |= SFXBF_SUSPEND;
			sidrv.SoundStop(ch);
		}
	}

	END_COP;
}

void Sfx_ReleaseRepeatingChannels()
{
	if (!sfx_avail) return;

	sfxchannel_t	*ch;
	int				i;

	BEGIN_COP;

	for (i = 0, ch = snd_channels; i < num_channels; i++, ch++)
	{
		if ( (ch->buffer.SFXBFflags & (SFXBF_PLAYING|SFXBF_REPEAT|SFXBF_SUSPEND)) == (SFXBF_PLAYING|SFXBF_REPEAT|SFXBF_SUSPEND))
		{
			ch->buffer.SFXBFflags &= ~SFXBF_SUSPEND;
			sidrv.SoundPlay(ch);
		}
	}

	END_COP;
}

//===========================================================================
// Sfx_Update
//	Update channel and listener properties.
//===========================================================================
static void Sfx_Update(void)
{
	int i;
	sfxchannel_t *ch;

	// Set the listener
	listener = ppt.viewplrmo;
	
	// Update channels.
	for (i = 0, ch = snd_channels; i < num_channels; i++, ch++)
	{
		if (   ch->buffer.sample
			&& ch->buffer.SFXBFflags & SFXBF_PLAYING)
		{
			Sfx_ChannelUpdate(ch);
		}
	}

	// Update listener.
	Sfx_ListenerUpdate();
}

//===========================================================================
// Sfx_StartFrame
//	Periodical routines: channel updates, cache purge, cvar checks.
//===========================================================================
void Sfx_StartFrame(void)
{
	static uint lastupdate = 0;

	if (sfx_avail)
	{
		uint nowtime = Sys_GetRealTime();

		// Should we purge the cache (to conserve memory)?
		Sfx_PurgeCache();

		// Is it time to do a channel update?
		if (nowtime - lastupdate >= TICSPERSEC)
		{
			lastupdate = nowtime;
			Sfx_Update();	
		}
	}
}

//===========================================================================
// Sfx_ShutdownChannels
//	Frees all memory allocated for the channels. 
//===========================================================================
static void Sfx_ShutdownChannels(void)
{
	int i;

	if (!num_channels) return;

	if (allowrefresh)
	{
		allowrefresh = false;			// Stop refreshing.
		while (refreshing) Sys_Sleep(10);	// let thread finish
	}

	for (i = 0; i < num_channels; i++)
	{
		if (snd_channels[i].buffer.ptr)
		{
			Sfx_ChannelStop(&snd_channels[i]);
			sidrv.SoundDestroyBuffer(&snd_channels[i]);
		}
	}

	num_channels = 0;
}

//===========================================================================
// Sfx_StartRefresh
//	Start the channel refresh thread.
//	It self terminates when channels are destroyed on exit.
//===========================================================================
static void Sfx_StartRefresh(void)
{
	refreshing = false;

	// Create a high-priority thread for the channel refresh.
	// Use 1MB for the stack (see note in Sys_StartThread)
	refresh_handle = Sys_StartThread(Sfx_ChannelRefreshThread,
											0x100000L, &refreshthreadid, &refreshparam);

	if(!refresh_handle)
		Con_ErrorD("Sfx_StartRefresh: Failed to start refresh");

	SetThreadPriority((HANDLE)refresh_handle, THREAD_PRIORITY_HIGHEST);
	allowrefresh = true;
}

static void checkR3DSplashSounds(int sound_id, char *shouldbe)
{
	if (   sound_id > 0
		&& sound_id < count_sounds.num
		&& lumpinfo[sounds[sound_id].lumpnum].handle != RISEN3DWADhandle)
	{
		int		i;

		for (i = 0; i < numlumps; i++)
		{
			if (   lumpinfo[i].handle == RISEN3DWADhandle
				&& !_stricmp(lumpinfo[i].name, shouldbe))
			{
				Con_Printf("%s has been changed and is being reset\n", shouldbe);
				sounds[sound_id].lumpnum = i;
				break;
			}
		}
	}
}

static int getR3Dsound(char *sndname)
{
	int id = Def_GetSoundNum(sndname);

	if (id <= 0 || sounds[id].flags & SFX_BADSOUND)
		Con_Error("SOUND ERROR %s: Risen3D.wad or audio.ded wrong version or corrupted\n", sndname);

	return id;
}

void Sfx_Init(void)
{
	// Check if sound has been disabled with a command line option.
	if (ArgExists("-nosfx") || compat[c_nosound])
	{
		Con_Message("Sfx_Init: Disabled.by user\n");
		num_channels = 0;
		return;
	}
	
	// Set num_channels to value requested (64 - 128)
	if (ArgCheckWith("-sfxchan", 1))
	{
		num_channels = atoi(ArgNext());

		if (num_channels < 64 || num_channels >  SFX_MAX_CHANNELS)
			num_channels = SFX_MAX_CHANNELS;
	}

	// Initialise the sound interface
	pr_OFdirect("Initialising D3D_Sound.\n");

	if (D3D_SoundInit())
	{
		D3D_DSoundSetPrimaryFormat();
		sidrv = sid3d;	
	}
	else
	{
		pr_OFinsert("Sfx_Init: Driver init failed. Sfx is disabled.\n");
		num_channels = 0;
		return;
	}

	sfx_avail = true;

	// b211-0 Get id's of Risen3D.wad sounds
	r3ds.stnmov_id  = getR3Dsound("STNMOV");
	r3ds.pstop_id   = r3ds.stnmov_id;	// for the moment
	r3ds.zoomin_id  = getR3Dsound("ZOOMIN");
	r3ds.zoomout_id = getR3Dsound("ZOOMOUT");
	r3ds.r3dp1_id   = getR3Dsound("punch");	// doom default
	r3ds.uwater_id  = getR3Dsound("PWATER");
	r3ds.breath_id  = getR3Dsound("PBREATH");
	r3ds.drown_id   = getR3Dsound("PDROWN");

	if (BDMFX)
	{
		r3ds.fakewt_id  = getR3Dsound("FAKEWT");
		r3ds.jumpdn_id  = getR3Dsound("JUMPDOWN");
		r3ds.wade_id    = getR3Dsound("PWADE");

		r3ds.bardrop_id = getR3Dsound("BARRELDROP");

		r3ds.wprdy1_id  = getR3Dsound("WEAPRDY1");
		r3ds.wprdy2_id  = getR3Dsound("WEAPRDY2");
		r3ds.wprdy3_id  = getR3Dsound("WEAPRDY3");
		r3ds.xtrabfg_id = getR3Dsound("EXTRABFG");
		r3ds.r3dp1_id   = getR3Dsound("PUNCHNMY");
		r3ds.r3dp2_id   = getR3Dsound("PUNCHAIR");
		r3ds.ricoh1_id  = getR3Dsound("RICOH1");
		r3ds.ricoh2_id  = getR3Dsound("RICOH2");
	}

	if (!hasDEHactors)
	{
		checkR3DSplashSounds(sfx_gloop, "DSGLOOP");
		checkR3DSplashSounds(sfx_sludg, "DSSLUDG");
		checkR3DSplashSounds(sfx_burn , "DSBURN" );
	}

	// Set the channel flags
	{
		sfxchannel_t *ch;
		int i;

		sfx_dedicated_2d = sfx_3d ? 8 : num_channels;

		for (i = 0, ch = snd_channels; i < num_channels; i++, ch++)
		{
			ch->buffer.SFXBFflags = (i >= sfx_dedicated_2d) ? SFXBF_3D : 0;
		}
	}

	pr_OFinsert("DS_DSound initialized : 3D sound %s\n",
		ArgExists("-sfx_3d") ? sfx_3d ? "enabled" : "requested but not supported" : "not requested");

	pr_OFinsert("Channels assigned : ");

	if (sfx_dedicated_2d < num_channels)
		pr_OFinsert("%i channels 2D : %i channels 3D\n", sfx_dedicated_2d, num_channels - sfx_dedicated_2d);
	else
		pr_OFinsert("%i channels 2D (stereo only)\n", num_channels);

	pr_OFdirect("Initialising sound cache.\n");

	Sfx_InitCache();				// Init the sample cache.
	Sfx_ListenerReverbDoppler();	// Init reverb and doppler shift
	Sfx_StartRefresh();				// Start the D3D refresh thread.
}

void Sfx_Doppler()
{
	if (sfx_avail)
		sidrv.SoundListener(SFXLP_DOPPLER, 0.2f);	// was cfg.doppler
}

//===========================================================================
// Sfx_Shutdown
//	Shut down the whole Sfx module: drivers, channel buffers and the cache.
//===========================================================================
void Sfx_Shutdown(void)
{
	if(!sfx_avail) return;		// Not initialized.

	// Prevent terminating click
	sidrv.SoundListener(SFXLP_GAIN, 0.0f);

	Sfx_StopSound(0, NULL);		// Stop all sounds.
	Sfx_ShutdownChannels();		// Destroy channels.
	Sfx_ShutdownCache();		// Destroy the sample cache.

	sidrv.SoundShutdown();		// Finally, close the driver.
	sfx_avail = false;			// This will cause the refresh thread
								// to terminate.
}

//===========================================================================
// Sfx_LevelChange
//	Must be done before the level is changed
// (from P_SetupLevel, via	S_LevelChange).
//===========================================================================
void Sfx_LevelChange(void)
{
	int				i;

	if (!sfx_avail) return;

	BEGIN_COP;

	for (i = 0; i < num_channels; i++)
		Sfx_ChannelStop(snd_channels + i);

	END_COP;

	listener = NULL;	// stop reverb
}

//===========================================================================
// Sfx_Reset
//	Stop all channels, clear the cache.
//===========================================================================
void Sfx_Reset(void)
{
	if(!sfx_avail) return;

	Sfx_LevelChange();
	Sfx_ShutdownCache();	// Free all samples.
}

char *getSndName(sfxsample_t *s)
{
	return sounds[s->id].lumpname;
}

// v220-19 returns true if any sound is being played by the emitter
RDBOOL SFXBF_playing(mobj_t *emitter)
{
	sfxchannel_t	*ch;
	int				i;

	if (sfx_avail)
	{
		for(i = 0, ch = snd_channels; i < num_channels; i++, ch++)
		{
			if (   ch->emitter == emitter
				&& ch->buffer.SFXBFflags & SFXBF_PLAYING)
			{
				return true;
			}
		}
	}

	return false;
}

//===========================================================================
// Sfx_DebugInfo v220-35 reconfigured for 128 channels and to id ambients
//===========================================================================
void Sfx_DebugInfo()
{
	extern int		pu_counter;
	extern int		currentFont;
	static int		maxp = 0;
	sfxchannel_t	*ch;
	char			buf[200], nbuf[400];
	int				i, p, x, y;
	int				lh;
	uint			cachesize, ccnt;
	int				idx = currentFont;
	int				w = current_w, h = current_h;

	pu_counter = 0;	// disable pickup display
	FR_SetFont(0);
	lh = FR_TextHeight("W") - 3;

	// Go into screen projection mode.
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	if (w < 866) w = 866;
	if (h < 560) h = 560;
	glOrtho(0, w, h, 0, -1, 1);

	for (i = p = 0, ch = snd_channels; i < num_channels; i++, ch++)
		if (ch->buffer.SFXBFflags & SFXBF_PLAYING) p++;

	if (p > maxp) maxp = p;

	// Sample cache information.	
	Sfx_GetCacheInfo(&cachesize, &ccnt);
	sprintf(buf, "%c Cached:%i (%i)  playing %i maxplaying %i",
		refmonitor ? '!' : ' ', cachesize, ccnt, p, maxp);
	glColor3f(1.f, 1.f, 0.f);
	FR_TextOut(buf, 10, 10);

	// Print info about each channel.
	for (i = 0, ch = snd_channels; i < num_channels; i++, ch++)
	{
		if (i < 64)
		{
			x = 5;
			y = lh*(i + 2);
		}
		else
		{
			x = current_w/2;
			y = lh*(i - 64 + 2);
		}

		if (ch->buffer.SFXBFflags & SFXBF_PLAYING)
			glColor4f(1.f, 1.f, 1.f, .7f);
		else 
			glColor4f(1.f, 1.f, .0f, .7f);

		sprintf(buf, "%3i: %c%c%c v=%3.1f f=%3.3f attn = %.3f x%6i y%6i dur %.2f",
			i, 
			!(ch->SFXCFflags & SFXCF_NO_ORIGIN) ? 'O' : '.',
			!(ch->buffer.SFXBFflags & SFXBF_NOATTN) ? 'A' : '.',
			ch->emitter ? 'E' : '.',
			ch->volume,
			ch->frequency,
			ch->attenuation,
			ch->emitter ? ch->emitter->x>>16 : 0,
			ch->emitter ? ch->emitter->y>>16 : 0,
			ch->buffer.sample ? (float)ch->buffer.sample->numsamples/(float)ch->buffer.sample->rate : 0);
		// split to make more readable
		sprintf(nbuf, "%s  %c%c%c%c%c%c id %8s=%4i sz=%6i b=%i rt=%5i", buf,
			ch->buffer.SFXBFflags & SFXBF_3D     ? '3' : '.',
			ch->buffer.SFXBFflags & SFXBF_PLAYING? 'P' : '.',
			ch->buffer.SFXBFflags & SFXBF_REPEAT ? 'R' : '.',
			ch->buffer.SFXBFflags & SFXBF_SUSPEND? 'S' : '.',
			ch->SFXCFflags & SFXCF_AMBIENT? 'A' : '.',
			ch->SFXCFflags & SFXCF_NO_ORIGIN? 'X' : '.',
			ch->buffer.sample? sounds[ch->buffer.sample->id].lumpname : "",
			ch->buffer.sample? ch->buffer.sample->id : 0,
			ch->buffer.sample? ch->buffer.sample->size : 0,
			ch->buffer.sample? ch->buffer.bytes : 0,
			ch->buffer.sample? ch->buffer.rate : 0);

		FR_TextOut(nbuf, x, y);
	}

	// Restore original matrix.
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	if (idx != currentFont) FR_SetFont(idx);
}

int CCmdPlaySound(int argc, char **argv)
{
	int			sound_id;
	float		vol = 1.0f;

	switch(argc)
	{
	case 3:
		vol = (float)atof(argv[2]);
		if (vol > 1 || vol < 0) break;
	case 2:
		sound_id = Def_GetSoundNum(argv[1]);

		if (sound_id <= 0)
		{
			Con_Printf("\n  %s %s no definition found.\n",
				argv[0], argv[1]);
			return true;
		}

		S_LocalSoundAtVolume(sound_id, NULL, vol);
		return true;
	default:
		break;
	}
	Con_Printf("Usage:\n  %s id [vol]\n"
		"  vol range 0 to 1 (default 1)\n",
		argv[0]);
	return true;
}
