//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 Jaakko Kernen, Jack Vermeulen
// Copyright (C) 2010 G. Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: sys_filein.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
//  conversion to file handles vs FILE*
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
//  File write routines added
//  
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  File input : Can read from real files or WAD lumps
//               File output: Can write to files
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

//#include <stdio.h>
#include <io.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>  // 12-2-2004 for io using handles
#include <sys/stat.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_misc.h"

#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

#define MAX_LUMPDIRS	1024
#define MAX_FILES		64

// TYPES -------------------------------------------------------------------

typedef struct
{
	char lump[9];
	byte align;
	char *path;		// Full path name.

} lumpdirec_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

static void F_ShutdownDirec(void);

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static lumpdirec_t	direc[MAX_LUMPDIRS + 1];
static DFILE		dfiles[MAX_FILES];

// CODE --------------------------------------------------------------------

//===========================================================================
// F_MatchName
//	Returns true if the string matches the pattern. 
//	This is a case-insensitive test.
//	I do hope this algorithm works like it should...
//===========================================================================
static int F_MatchName(const char *string, const char *pattern)
{
	const char *in = string, *st = pattern;

	while(*in)
	{
		if(*st == '*') 
		{
			st++; 
			continue;
		}
		if(*st != '?' 
			&& (tolower( (unsigned char) *st) 
				!= tolower( (unsigned char) *in)))
		{
			// A mismatch. Hmm. Go back to a previous *.
			while(st >= pattern && *st != '*') st--;
			if(st < pattern) return false; // No match!
			// The asterisk lets us continue.
		}
		// This character of the pattern is OK.
		st++; 
		in++;
	}

	// Match is good if the end of the pattern was reached.

	while(*st == '*') st++; // Skip remaining asterisks.

	return *st == 0;
}

//===========================================================================
// F_SkipSpace
//	Skips all whitespace except newlines.
//===========================================================================
static char *F_SkipSpace(char *ptr)
{
	while(*ptr && *ptr != '\n' && isspace(*ptr)) ptr++;
	
    return ptr;
}

//===========================================================================
// F_FindNewline
//===========================================================================
static char *F_FindNewline(char *ptr)
{
	while(*ptr && *ptr != '\n') ptr++;
	
    return ptr;
}

//===========================================================================
// F_GetDirecIdx
//===========================================================================
static int F_GetDirecIdx(char *exact_path)
{
	int i;

	for(i = 0; direc[i].path; i++)
    {
		if(!_stricmp(direc[i].path, exact_path))
			return i;
    }
	return -1;
}

//===========================================================================
// F_AddDirec
//	The path names can converted to full paths before adding to the table.
//===========================================================================
static void F_AddDirec(char *lumpname, char *symbolic_path)
{
	lumpdirec_t	*ld;
	char		path[MAXPATH], *full;
	int			i;

	if(!lumpname[0] || !symbolic_path[0]) return;

	for(i = 0; direc[i].path && i < MAX_LUMPDIRS; i++);

	if(i == MAX_LUMPDIRS) 
	{
		// FIXME: Why no dynamic allocation? (1024 should be enough)
		Con_Error("F_AddDirec: Not enough direcs (%s)", symbolic_path);
	}
	ld = direc + i;

	// Convert the symbolic path into a real path.

	for(i = 0; symbolic_path[i]; i++) // Slashes to backslashes.
    {
		if(symbolic_path[i] == '/') symbolic_path[i] = '\\';
    }

	if(symbolic_path[0] == '\\')
    {
		sprintf(path, "%s", symbolic_path + 1);
    }
	else
    {
		sprintf(path, "%s%s", ddRuntimeDir.path, symbolic_path);
    }

	// Since the basepath might be relative, let's explicitly make
	// the path absolute. _fullpath will allocate memory, which will
	// be freed in F_ShutdownDirec() (also via F_InitDirec()).
	full = _fullpath(0, path, 0);

	// If this path already exists, we'll just update the lump name.
	i = F_GetDirecIdx(full);
	if(i >= 0)
	{
		// Already exists!
		M_Free(full);
		ld = direc + i;
		full = ld->path;
	}
	ld->path = full;
	strcpy(ld->lump, lumpname);
	_strupr(ld->lump);

	if(verbose) 
	{
		pr_OFinsert("F_AddDirec: %s -> %s\n", ld->lump, ld->path);
	}
}

//===========================================================================
// F_ParseDirecData
//	LUMPNAM0 \Path\In\The\Base.ext
//	LUMPNAM1 Path\In\The\RuntimeDir.ext
//	 :
//===========================================================================
static void F_ParseDirecData(char *buffer)
{
	char	*ptr = buffer, *end;
	char	name[9], path[MAXPATH];
	int		len;
	
	for (; *ptr; )
	{
		ptr = F_SkipSpace(ptr);

		if (!*ptr) break;

		if (*ptr == '\n') 
		{
			ptr++; // Advance to the next line.
			continue;
		}

		// We're at the lump name.
		end = M_FindWhite(ptr);

		if (!*end) break;

		if ((len = end - ptr) > 8) len = 8;

		memset(name, 0, sizeof(name));
		strncpy(name, ptr, len);
		ptr = F_SkipSpace(end);

		if (!*ptr || *ptr == '\n') continue; // Missing filename?

		// We're at the path name.
		end = F_FindNewline(ptr);

		// Get rid of extra whitespace.
		while(end > ptr && isspace(*end)) end--;

		end = M_FindWhite(end);
		len = end - ptr;

		if (len > 259) len = 259;	// jav 12-2-2004 safety

		memset(path, 0, sizeof(path));
		strncpy(path, ptr, len);
		F_AddDirec(name, path);
		ptr = end;
	}
}

//===========================================================================
// F_InitDirec
//	Initialize the WAD/dir translations. Called after WADs have been read.
//===========================================================================
void F_InitDirec(void)
{
	int   i, len1;
	byte  *buf1;   // jav
	byte  *buf2;   // jav

	// Free old paths, if any.
	
    F_ShutdownDirec();
	memset(direc, 0, sizeof(direc));
	
	// Add the contents of all DD_DIREC lumps

	for(i = 0; i < numlumps; i++)
    {
		if(!strnicmp(lumpinfo[i].name, "DD_DIREC", 8))
		{
			// Make a copy of it so we can make it end in a null

			len1 = W_LumpLength(i);
			buf1 = (byte*)M_Malloc(len1 + 1, 98);
            buf2 = (byte*)M_CacheLumpNum(i, NULL, false);    // jav 12-2-2004

            if (buf2)
            {
             	memcpy(buf1,buf2, len1); 
                M_Free(buf2); 
            }
			
            buf1[len1] = 0;

			F_ParseDirecData((char*)buf1);

			M_Free(buf1);
		}
    }
}

//===========================================================================
// F_ShutdownDirec
//===========================================================================
static void F_ShutdownDirec(void)
{
	int i;

	for(i = 0; direc[i].path; i++) 
	{	
        M_Free(direc[i].path); // Allocated by _fullpath. jav function checks for null
        direc[i].path = NULL;  // jav 12-2-2004
    } 
}

//===========================================================================
// F_InitDFILES (just in case.. R3D)
//===========================================================================
void F_InitDFILES(void)
{
	memset(dfiles, 0, sizeof(dfiles));
	memset(direc , 0, sizeof(direc ));
}

//===========================================================================
// F_Close
//===========================================================================
void F_Close(DFILE *file)
{
	if (!file || !file->flags.open) return;

	if (file->flags.file)
	{
		_close(file->handle); // jav 12-2-2004
	}
	else if (file->data)	// Free the stored data.
	{
		M_Free(file->data);
	}
	memset(file, 0, sizeof(DFILE));
}

//===========================================================================
// F_ShutdownDFILES
//===========================================================================
void F_ShutdownDFILES(void)
{
	int i;

	F_ShutdownDirec();

	for(i = 0; i < MAX_FILES; i++) F_Close(dfiles + i);
}

//===========================================================================
// F_Access
//	Returns true if the file can be opened for reading.
//===========================================================================
int F_Access(char *path)
{
	DFILE *file;

    Dir_ConvertSlashes(path);    // jav 12-23-2004
    
    file = F_Open(path, "f");
	
	if(!file) 
    {
        return false;
    }
	
    F_Close(file);

	return true;
}

//===========================================================================
// F_GetFreeFile
//===========================================================================
static DFILE *F_GetFreeFile(void)
{
	int i;

	for (i = 0; i < MAX_FILES; i++)
	{
		if (!dfiles[i].flags.open) 
		{
			memset(dfiles + i, 0, sizeof(DFILE));
			return dfiles + i;
		}
	}

	Con_ErrorD("F_GetFreeFile : No free files");

	return NULL;
}

//===========================================================================
// F_OpenLump
//  b211-0 modified to handle lumps directly
//===========================================================================
DFILE *F_OpenLump(int lump)
{
	byte  *lumpdata;
	DFILE *file;
	
	if (   lump < 0
		|| lumpinfo[lump].size <= 0)
	{
		return NULL;			// Nothing to read
	}
	
	if ( (file = F_GetFreeFile()) )
	{
		// Init and load in the lump data.
		file->flags.open = true;		// it's open
		file->flags.file = false;		// not a file
		file->flags.write = false;		// it cannot be written to
		file->size       = lumpinfo[lump].size;		// size of lump
		file->data		 = M_Malloc(file->size, 99);
		file->pos        = (char*)file->data;
		
		lumpdata = (byte*)M_CacheLumpNum(lump, NULL, false);
		memcpy(file->data, lumpdata, file->size); 
		M_Free(lumpdata); 
		
		return file;
	}

	return NULL;
}

//===========================================================================
// F_OpenData
//  v220-6 modified to handle data directly
//===========================================================================
DFILE *F_OpenData(byte *data, uint size)
{
	DFILE *file;
	
	if ( (file = F_GetFreeFile()) )
	{
		file->flags.open = true;	// it's open
		file->flags.file = false;	// not a file
		file->flags.write = false;	// it cannot be written to
		file->size       = size;	// size of lump
		file->data		 = data;
		file->pos        = (char*)file->data;
		
		return file;
	}
	
	return NULL;
}

//===========================================================================
// F_OpenFile
//===========================================================================
DFILE *F_OpenFile(const char *path, const char *mymode)
{
	DFILE *file;
	
    int   oflag = _O_RDONLY;  // jav 12-2-2004 read only

	if( (file = F_GetFreeFile()) == NULL ) return NULL;
	
	if(strchr(mymode, 't')) oflag |= _O_TEXT;
	if(strchr(mymode, 'b')) oflag |= _O_BINARY;

	file->handle = _open(path, oflag);
    if (file->handle == -1) return NULL;
	
    file->flags.open = true;	// it's open
	file->flags.file = true;	// it's a file
	file->flags.write = false;	// it cannot be written to
	return file;
}

//===========================================================================
// F_OpenFileForWrite v220-6
//  The contents of any existing file are destroyed
//  Kept separate to avoid unintentional writes to other files
//  Only mode t (text) and b (binary) are suppoprted
//===========================================================================
DFILE *F_OpenFileForWrite(const char *path, const char *mymode)
{
	DFILE *file;
	
    int   oflag = _O_WRONLY|_O_CREAT|_O_TRUNC ;  // write only

	if( (file = F_GetFreeFile()) == NULL ) return NULL;

	if(strchr(mymode, 't')) oflag |= _O_TEXT;
	if(strchr(mymode, 'b')) oflag |= _O_BINARY;

	file->handle = _open(path, oflag, _S_IREAD | _S_IWRITE);
    if (file->handle == -1) return NULL;
	
    file->flags.open = true;	// it's open
	file->flags.file = true;	// it's a file
	file->flags.write = true;	// it can be written to
	return file;
}

//===========================================================================
// F_FileWrite v220-6
//  Automatically terminates if there's an error unless from < 0 where
//	false is returned if the bytes written does not equal the size sent.
//  If there is no error true is returned
//===========================================================================
int F_FileWrite(void *data, uint length, DFILE *file, int from)
{
	if (file && file->flags.open)
	{
		uint count = 0;

		// Must be an actual FILE which can be written to
		if (   !file->flags.open
			|| !file->flags.file
			|| !file->flags.write)
		{
			return 0;
		}

		if(length)
			count = write(file->handle, data, length);

		if (count != length)
		{
			F_Close(file);
			if (from < 0) return false;
			Con_Error("Failed to write data to DFILE [from %i]", from);
		}

		return true;
	}

	return false;
}

//===========================================================================
// F_Open
//	Opens the given file (will be translated) or lump for reading.
//	"t" = text mode (with real files, lumps are always binary)
//	"b" = binary
//	"f" = must be a real file
//	"w" = file must be in a WAD
//===========================================================================
DFILE *F_Open(char *path, const char *mode)
{
	char trans[MAXPATH]; // relative path created
	char full [MAXPATH]; // jav 12-2-2004 
	int i;
	
	if (!mode)
		Con_Error("ERROR: F_Open mode unset [mode %s]", mode ? mode : "");
	
	// Make it a full path.
	
	M_TranslatePath(path, trans); // create relative path
	_fullpath(full, trans, 259);  // back to absolute path
	
	// Lumpdirecs take precedence.
	
	if(!strchr(mode, 'f')) // Doesn't need to be a real file?
	{
		for(i = 0; direc[i].path; i++)
		{	
			if(!_stricmp(full, direc[i].path))
			{
				int lump = W_CheckNumForName( (char*)direc[i].lump );

				return F_OpenLump(lump);
			}
		}
	}

	// Try to open as a real file, then
	
	return F_OpenFile(full, mode);
}

//===========================================================================
// F_Read
//	Returns the number of bytes read (up to 'count').
//===========================================================================
int F_Read(void *dest, int length, DFILE *file)
{
	int bytesleft;

	if (!file || !file->flags.open) return 0;

    // if an actual FILE 

	if(file->flags.file) 
	{
        int count= read(file->handle, dest, length);  // 12-2-2004

    	if (count <= 0) file->flags.eof = true;   // ignore error for now
		
        return count;
	}
	
    // Is there enough room in the file?
	
    bytesleft = file->size - (file->pos - (char*) file->data);
	
    if(length > bytesleft) 
	{
		//count = bytesleft;
		length = bytesleft;
		file->flags.eof = true;
	}

	if(length)
	{
		memcpy(dest, file->pos, length);
		file->pos += length;
	}
	return length;
}

//===========================================================================
// F_GetC
//===========================================================================
int F_GetC(DFILE *file)
{
	unsigned char ch = 0;

	if (!file || !file->flags.open) return 0;

	F_Read(&ch, 1, file);
	return ch;
}

//===========================================================================
// F_Tell
//===========================================================================
int F_Tell(DFILE *file)
{
	if (!file || !file->flags.open) return 0;

	if (file->flags.file)  
    {
       return tell(file->handle); // for actual file
    }

	return file->pos - (char*) file->data;
}

//===========================================================================
// F_Seek
//	Returns the current position in the file, before the move, as an offset
//	from the beginning of the file.
//===========================================================================
int F_Seek(DFILE *file, int offset, int origin)
{
	int oldpos = F_Tell(file);

	if (!file || !file->flags.open) return 0;

    file->flags.eof = false;

	if(file->flags.file)
    {
		lseek(file->handle, offset, origin);
    }
	else
	{
		if(origin == SEEK_SET)
		   file->pos = (char*) file->data + offset;
		else 
        if(origin == SEEK_END)
		   file->pos = (char*) file->data + (file->size + offset);
		else 
        if(origin == SEEK_CUR)
		   file->pos += offset;
	}

	return oldpos;
}

//===========================================================================
// F_Rewind
//===========================================================================
void F_Rewind(DFILE *file)
{
	F_Seek(file, 0, SEEK_SET);
}

//===========================================================================
// F_ForAll
//	Parm is passed on to the callback, which is called for each file
//	matching the filespec. Both DD_DIREC and the real files are scanned.
//===========================================================================
int F_ForAll(char *filespec, int parm, f_forall_func_t func)
{
	long        hFile;
	struct      _finddata_t finddata;
	directory_t specdir;
	char        filename[MAXPATHNAMESIZE];
	int         i, ret = false;
	
	Dir_FileDir(filespec, &specdir);
	
	// Check through the dir/WAD direcs.
	
	_fullpath(filename, filespec, 259);
	
	for(i = 0; direc[i].path; i++)
	{
		if(F_MatchName(direc[i].path, filename))
		{
			if(!func(direc[i].path, parm)) return false;
		}
	}

	// b211-0 Try searching for the real file
	if((hFile = _findfirst(filespec, &finddata)) != -1)
	{
		// Compile the full pathname of the found file
		strcpy(filename, specdir.path);
		strcat(filename, finddata.name);
		
		// We're only interested in the first file found
		ret = func(filename, parm); // true if found
		
		_findclose( hFile );		// Search no more
	}
	return ret;
}

