//-----------------------------------------------------------------------------
//
// Copyright (C) 2012-2013 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: sys_fmod.c
//
// DESCRIPTION:  FMOD Music Driver
//
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
//
// v220-16 built using FMOD 17/12/12 4.44.03 - Stable branch update
// FMOD Sound System, copyright  Firelight Technologies Pty, Ltd., 1994-2011
//  
//  http://www.fmod.org
//
//  FMOD is used for playing standard OGG, MPEG, MOD, IT, S3M, XM music files.
//  (NOTE: 'music'  wav files can be played/tested with the console pm command
//          but game wav sound effects are always played using Risen3D's own
//          DSound code).
//
// NOTE;
//  Risen3D uses DSOUND (directly - not through FMOD) for sound effects.
//  The reason is because it has always done so and I see no good reason
//  to abandon code that works well and has been thoroughly tested.
//
//  Risen3D uses WinMM  (directly - not through FMOD) to play Midi's.
//  This is because there is a problem in some cases with the quality of
//  MIDI sound produced. (E.g. playing the Doom E1M5 midi using FMOD sounds
//  particularly bad).
//  Support for playing CD's has been removed because the FMOD cdplayer
//  could not be made to work in XP (including their own sample cdplayer code)
//  A WinMM player could have been included but this was decided against
//  because there is no way of setting the volume in Vista (and on) using the
//  scripted WinMM commands. (Which probably means there's a problem in any
//  event given Vista is cursed when it comes to dealing with sound).
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <windows.h>
#include <stdio.h>

#include <fmod.h>
#include <fmod_errors.h>

#include "r3d_compat.h"

#include "de_audio.h"
#include "con_console.h"
#include "dd_main.h"
#include "dd_wad.h"
#include "sys_system.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

char	*getFNAME(char *path);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern RDBOOL	ui_active;
extern char		szFNAME[42];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

unsigned long	fmodthreadid;
unsigned long	fmodparam;

static FMOD_SYSTEM			*fmsystem;
static FMOD_SOUND			*fmsound;
static FMOD_CHANNEL			*fmchannel;
static FMOD_RESULT			result;
static unsigned int			version;


static int		fmod_handle;	// plays using an independant thread
static int		end_index;
static int		pause_change;
static int		pause_state;
static uint		lengthMs;
static RDBOOL	fmod_inited;
static char		looping;
static char		*strfm_type;
static char		*lmpName;
static float fmex_vol = -1.f;


static RDBOOL fmodrefresh, fmodrefreshing;

// Note: this may change with a later FMOD release
char *strtype[] = { 
	"unknown",		//  0
	"AIFF",
	"ASF",
	"AT3",
	"CDDA",
	"DLS",
	"FLAC",
	"FSB",
	"GCADPSM",
	"IT",
	"MIDI",			// 10
	"MOD",
	"MPEG",
	"OGGVORBIS",
	"PLAYLIST",
	"RAW",
	"S3M",
	"SF2",
	"USER",
	"WAV",
	"XM",			// 20
	"XMA",
	"VAG",
	"AUDIOQUEUE",
	"XWMA",
	"BCWAV",
	"AT9",
	"MAX",
	NULL
};

// CODE --------------------------------------------------------------------

#pragma optimize("g",off)

/* FMOD callbacks

	Fmod seeks to the end and reads because it is looking for ID3V1 tags.  
	Use FMOD_IGNORETAGS in FMOD_System_CreateStream if you don't like this behaviour.
*/

static FMOD_RESULT F_CALLBACK myopen(const char *name, int unicode, unsigned int *filesize, void **handle, void **userdata)
{
    if (name)
    {
        FILE *fp;

        fp = fopen(name, "rb");
        if (!fp)
        {
            return FMOD_ERR_FILE_NOTFOUND;
        }

        fseek(fp, 0, SEEK_END);
        *filesize = ftell(fp);
        fseek(fp, 0, SEEK_SET);

        *userdata = (void *)0x12345678;
        *handle = fp;
    }

    return FMOD_OK;
}

static FMOD_RESULT F_CALLBACK myclose(void *handle, void *userdata)
{
    if (!handle) return FMOD_ERR_INVALID_PARAM;

    fclose((FILE *)handle);

    return FMOD_OK;
}

static FMOD_RESULT F_CALLBACK myread(void *handle, void *buffer, unsigned int sizebytes, unsigned int *bytesread, void *userdata)
{
    if (!handle)  return FMOD_ERR_INVALID_PARAM;

    if (bytesread)
    {
        *bytesread = (int)fread(buffer, 1, sizebytes, (FILE *)handle);
    
        if (*bytesread < sizebytes)
        {
            return FMOD_ERR_FILE_EOF;
        }
    }

    return FMOD_OK;
}

static FMOD_RESULT F_CALLBACK myseek(void *handle, unsigned int pos, void *userdata)
{
    if (!handle) return FMOD_ERR_INVALID_PARAM;

    fseek((FILE *)handle, pos, SEEK_SET);

    return FMOD_OK;
}

//===========================================================================

static void ERRCHECK(FMOD_RESULT result, int from)
{
    if (result != FMOD_OK)
		pr_OFinsert("FMOD error [from %i]: (%d) %s\n",
		from, result, FMOD_ErrorString(result));
}

void FMOD_AllowRefresh(RDBOOL allow)
{
	if (!fmod_handle || fmodrefresh == allow) return;

	fmodrefresh = allow;

	// If we're denying refresh then ensure, if it's
	// currently running, that its allowed to complete.
	if (!allow)
	{
		while (fmodrefreshing) Sys_Sleep(1);
	}

	if (allow)
		SetThreadPriority((HANDLE)fmod_handle, THREAD_PRIORITY_HIGHEST);
	else
		SetThreadPriority((HANDLE)fmod_handle, THREAD_PRIORITY_LOWEST);
}

//===========================================================================
// FModStop
//===========================================================================
void FModStop()
{
	if (!fmod_inited || !fmsound) return;

	FMOD_AllowRefresh(false);
	result = FMOD_Sound_Release(fmsound);
	ERRCHECK(result, 0);
	fmsound = NULL;
}

//===========================================================================
// FModShutdown
//===========================================================================
void FModShutdown(void)
{
	if (!fmod_inited) return;
	
	FModStop();

	FMOD_System_Close(fmsystem);
	FMOD_System_Release(fmsystem);

	fmod_inited = false;	// cause termination of thread
	FMOD_AllowRefresh(true);

	pr_OFdirect("FMOD  shutdown\n");
}

//===========================================================================
// FModVolumeSet
//===========================================================================
void FModVolumeSet(int vol)
{
	if (!fmod_inited) return;

	fmex_vol = (float)vol/255.f;
}

static FMOD_BOOL valid_error(FMOD_RESULT error, int num)
{
	if (error == FMOD_OK) return false;
	if (error == FMOD_ERR_INVALID_HANDLE) return false;
	if (error == FMOD_ERR_CHANNEL_STOLEN) return false;
	ERRCHECK(error, num);
	return true;
}

static int fmod_ThreadPlay()
{
	if (!fmod_inited) return 0; // Thread will end

	if (fmsound)
	{
		if (fmchannel && !pause_state)
		{
			if (fmex_vol >= 0)
			{
				if (FMOD_Channel_SetVolume(fmchannel, fmex_vol) == FMOD_OK)
					fmex_vol = -1.f;
			}
		}

		if (pause_change && fmchannel)
		{
			FMOD_Channel_SetPaused(fmchannel, (FMOD_BOOL) pause_state);
			pause_change = false;
		}

		FMOD_System_Update(fmsystem);

		if (/*!midi_selected &&*/ fmchannel)
		{
			static uint lastms = 0;
			int		fmplaying;
			int		fmpaused;

			if (   valid_error(FMOD_Channel_IsPlaying(fmchannel, &fmplaying), /*ERRCHECK*/1)
				|| valid_error(FMOD_Channel_GetPaused(fmchannel, &fmpaused) , /*ERRCHECK*/2) )
			{
				return 0;
			}

			music_status = (fmplaying || fmpaused) ? 1 : 0;

			if (ui_active && fmchannel)
			{
				uint	ms = 0;

				if (fmplaying && !fmpaused)
				{
					if (valid_error(FMOD_Channel_GetPosition(fmchannel, &ms, FMOD_TIMEUNIT_MS), /*ERRCHECK*/3) )
						return 0;

					lastms = ms;

					if (musicpos != 0xFFFFFFFF)
					{
						ms = musicpos;

						if (ms > lengthMs - 1) ms = lengthMs - 1;

						if (!sld_musicpos.button[0] && !sld_musicpos.button[1] && !sld_musicpos.button[2])
						{
							FMOD_Channel_SetPosition(fmchannel, ms, FMOD_TIMEUNIT_MS);
							musicpos = 0xFFFFFFFF;
						}
					}
				}
				else
				{
					ms = lastms;
					musicpos = 0xFFFFFFFF;
				}

				sld_musicpos.value = (float)ms;

				sprintf(music_playing,
					"%s %s: %02d:%02d / %02d:%02d  [%s]",
					lmpName,
					strfm_type,
					ms / 1000 / 60,
					ms / 1000 % 60,
					lengthMs / 1000 / 60,
					lengthMs / 1000 % 60,
					fmpaused ? "Paused" : fmplaying ? looping ? "Looping" : "Play Once" : "Completed");
			}
		}
	}

	return 1;
}

int FMOD_Thread(void *parm)
{
	while(fmod_inited)
	{
		if (fmodrefresh)
		{
			fmodrefreshing = true;

			if (!fmod_ThreadPlay()) break;

			fmodrefreshing = false;

		}

		Sleep(10);
	}

	fmodrefreshing = false;

	return (0);	// thread ends
}

RDBOOL FModInit()
{
	pr_OFinsert("FMOD Sound System, copyright  Firelight Technologies Pty, Ltd., 1994-2011\n");

	//  Create a System object and initialize.
	result = FMOD_System_Create(&fmsystem);
	if (result != FMOD_OK) { ERRCHECK(result, 4); pr_OFdirect("FMOD failed to initialise\n"); return false; }

	FMOD_System_GetVersion(fmsystem, &version);
	if (result != FMOD_OK) { ERRCHECK(result, 5); pr_OFdirect("FMOD failed to get version\n"); return false; }

	if (version < FMOD_VERSION)
	{
		pr_OFinsert("Error!  You are using an old version of FMOD %08x.  This program requires %08x\n", version, FMOD_VERSION);
		return false;
	}

	pr_OFinsert("FMOD version %08x\n", version);

	result = FMOD_System_SetOutput(fmsystem, FMOD_OUTPUTTYPE_DSOUND);
	if (result != FMOD_OK) { ERRCHECK(result, 6); pr_OFdirect("FMOD failed to initialise\n"); return false; }

	result = FMOD_System_Init(fmsystem, 1, FMOD_INIT_NORMAL, NULL);
	if (result != FMOD_OK) { ERRCHECK(result, 7); pr_OFdirect("FMOD failed to initialise\n"); return false; }

	result = FMOD_System_SetFileSystem(fmsystem, myopen, myclose, myread, myseek, 0, 0, 2048);
	if (result != FMOD_OK) { ERRCHECK(result, 8); pr_OFdirect("FMOD failed to initialise\n"); return false; }

	result = FMOD_System_SetStreamBufferSize(fmsystem, 16*1024, FMOD_TIMEUNIT_MS);
	if (result != FMOD_OK) { ERRCHECK(result, 9); pr_OFdirect("FMOD failed to initialise\n"); return false; }

	while (strtype[++end_index]);	// set limit of strtype

	fmodrefreshing = false;
	fmodrefresh = false;
	fmod_inited = true;
	fmsound = NULL;

	// Create a high-priority thread for the channel refresh.
	// Use 1MB for the stack (Too much? This is the smallest
	// size that can be assigned without fiddling around with
	// compiler directives. Or too little?).
	fmod_handle = Sys_StartThread(FMOD_Thread, 0x100000L, &fmodthreadid, &fmodparam);

	if (!fmod_handle)
	{
		fmod_inited = false;

		if (fmsystem)
		{
			FMOD_System_Close(fmsystem);
			FMOD_System_Release(fmsystem);
			fmsystem = NULL;
		}

		Con_Error("Failed to start FMOD thread");

		fmod_inited = false;

		return false;
	}
	
	SetThreadPriority((HANDLE)fmod_handle, THREAD_PRIORITY_LOWEST);
	pr_OFinsert("Initialised  : FMOD %s\n", ge_inf.useFMODsoftware ? "Using FMOD software mixer" : "");

	return true;
}

//===========================================================================
// FModPause
//===========================================================================
void FModPause(int pause)
{
	pause_state = pause;
	pause_change = true;
}

void resetMapMusicThinkerIden(void);

//===========================================================================
// FModPlay
//===========================================================================
int FModPlay(char *file, char *mname, int lumpnum, char *calledfrom, char *calledfromfrom)
{
	FMOD_SOUND_TYPE		fm_type;
	int					verr;
	uint				flags;
	char				path[512];

	if (!fmod_inited) return false;
#ifdef _DEBUG
	Con_Printf("file %s mname %s lumpnum %i name %s called from %s fromfrom %s\n", file, mname, lumpnum, lumpnum > 0 ? lumpinfo[lumpnum].name : "" , calledfrom, calledfromfrom);
#endif

	if (!calledfromfrom || stricmp(calledfromfrom, "Music_PlayLump"))
	{
		resetMapMusicThinkerIden();
	}

	looping = global_looped;

	if (file)
	{
		strcpy(path, file);
		lmpName = mname;
	}
	else
	{
		uint size;
		byte *data;
		FILE *f;

		sprintf(path, "%sr3dtemp", ddRuntimeDir.path);

		if ( (f = fopen( path, "wb" )) == NULL) return false;

		data = (byte*)M_CacheLumpNum(lumpnum, &size, false);
		fwrite(data, size, 1, f);
		fclose(f);

		M_Free(data);

		lmpName = lumpinfo[lumpnum].name;
	}

	if (!looping && ma_inf.loadgame)
	{
		return false;
	}

	flags = looping ? FMOD_LOOP_NORMAL : FMOD_LOOP_OFF;
	
	if (ge_inf.useFMODsoftware)
		flags |= FMOD_SOFTWARE;
	else
		flags |= FMOD_HARDWARE;
	
	result = FMOD_System_CreateStream(fmsystem, path, flags | FMOD_2D | FMOD_ACCURATETIME, 0, &fmsound);
	if (result != FMOD_OK) { ERRCHECK(result, 10); return false; }
	
	result = FMOD_System_PlaySound(fmsystem, FMOD_CHANNEL_FREE, fmsound, 0, &fmchannel);
	if (result != FMOD_OK) { ERRCHECK(result, 11); FMOD_Sound_Release(fmsound); fmsound = NULL; return false; }
	
	result = FMOD_Sound_GetFormat(fmsound, &fm_type, NULL, NULL, NULL);
	if (result != FMOD_OK) { ERRCHECK(result, 12); FMOD_Sound_Release(fmsound); fmsound = NULL; return false; }
	
	// Trust no-one
	if (fm_type >= end_index)
	{
		pr_OFinsert("FMOD type %i exceeds entries %i\n", fm_type, end_index + 1);
		FMOD_Sound_Release(fmsound);
		fmsound = NULL;
		return false;
	}

	// v220-20
	if (midi_selected && ge_inf.usefmodmidi)
	{
		if (fm_type != FMOD_SOUND_TYPE_MIDI)
		{
			FMOD_Sound_Release(fmsound);
			fmsound = NULL;
			return false;
		}
	}
	else
	if (   fm_type != FMOD_SOUND_TYPE_MPEG
		&& fm_type != FMOD_SOUND_TYPE_OGGVORBIS
		&& fm_type != FMOD_SOUND_TYPE_IT
		&& fm_type != FMOD_SOUND_TYPE_MOD
		&& fm_type != FMOD_SOUND_TYPE_S3M
		&& fm_type != FMOD_SOUND_TYPE_XM
		&& fm_type != FMOD_SOUND_TYPE_WAV)	// purely for wav music (not sfx)
	{
		pr_OFinsert("FMOD type %s not supported in Risen3D\n", strtype[fm_type]);
		FMOD_Sound_Release(fmsound);
		fmsound = NULL;
		return false;
	}
	
	strfm_type = strtype[fm_type];	
	verr = valid_error(FMOD_Sound_GetLength(fmsound, &lengthMs, FMOD_TIMEUNIT_MS), /*ERRCHECK*/13);
	
	if (verr || lengthMs < 1000)
	{
		if (!verr)
			Con_Message("FMOD type %s ignored. Length %imS < 1 second\n", mname, lengthMs);
		FMOD_Sound_Release(fmsound);
		fmsound = NULL;
		return false;
	}

	sld_musicpos.max = (float)lengthMs;

	fmex_vol = ext_volume / 255.f;
	FMOD_AllowRefresh(true);
	
	return true;
}
