//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: sys_input.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified as needed for R3D
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
//  made compatible for c++
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Keyboard, mouse and joystick input using DirectInput
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include <windows.h>
#include <dinput.h>
#include <assert.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_misc.h"

#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

#define INV(x, axis)	(joyInverseAxis[axis]? -x : x)

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

RDBOOL	joyInited = false;
RDBOOL	mouseInited = false;

int		joydevice = 0;			// Joystick index to use.
int		joyInverseAxis[8];		// Axis inversion (default: all false).

cvar_t inputCVars[] =
{
	"i_JoyDevice",		CVF_HIDE|CVF_NO_ARCHIVE|CVF_NO_MAX|CVF_PROTECTED, CVT_INT,	&joydevice,		0,	0,	"ID of joystick to use (if more than one).",
	"i_UseJoystick",	CVF_HIDE|CVF_NO_ARCHIVE,	CVT_BYTE,	&cfg.joyEnable,	0,	1,	"1=Enable joystick input.",
//--------------------------------------------------------------------------
	"input-joy-device",	CVF_NO_MAX|CVF_PROTECTED,	CVT_INT,	&joydevice,		0,	0,	"ID of joystick to use (if more than one).",
	"input-joy-x-inverse",			0,	CVT_INT,	&joyInverseAxis[0],	0, 1,	"1=Inverse joystick X axis.",
	"input-joy-y-inverse",			0,	CVT_INT,	&joyInverseAxis[1],	0, 1,	"1=Inverse joystick Y axis.",
	"input-joy-z-inverse",			0,	CVT_INT,	&joyInverseAxis[2],	0, 1,	"1=Inverse joystick Z axis.",
	"input-joy-rx-inverse",			0,	CVT_INT,	&joyInverseAxis[3],	0, 1,	"1=Inverse joystick RX axis.",
	"input-joy-ry-inverse",			0,	CVT_INT,	&joyInverseAxis[4],	0, 1,	"1=Inverse joystick RY axis.",
	"input-joy-rz-inverse",			0,	CVT_INT,	&joyInverseAxis[5],	0, 1,	"1=Inverse joystick RZ axis.",
	"input-joy-slider1-inverse",	0,	CVT_INT,	&joyInverseAxis[6], 0, 1,	"1=Inverse joystick slider 1.",
	"input-joy-slider2-inverse",	0,	CVT_INT,	&joyInverseAxis[7], 0, 1,	"1=Inverse joystick slider 2.",
	NULL
};

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static RDBOOL	I_initOk = false;		

static LPDIRECTINPUT8 dInput;
static LPDIRECTINPUTDEVICE8 didKeyb;
static LPDIRECTINPUTDEVICE8 didMouse;
static LPDIRECTINPUTDEVICE8 didJoy;
static DIDEVICEINSTANCE firstJoystick;

static int		counter;

// CODE --------------------------------------------------------------------

//===========================================================================
// I_ErrorMsg
//===========================================================================
const char *I_ErrorMsg(HRESULT hr)
{
	return hr == DI_OK? "OK"
		: hr == DIERR_GENERIC? "Generic error"
		: hr == DI_PROPNOEFFECT? "Property has no effect"
		: hr == DIERR_INVALIDPARAM? "Invalid parameter"
		: hr == DIERR_NOTINITIALIZED? "Not initialized"
		: hr == DIERR_UNSUPPORTED? "Unsupported"
		: hr == DIERR_NOTFOUND? "Not found"
		: "?";
}

//===========================================================================
// I_SetProperty
//===========================================================================
HRESULT I_SetProperty(void *dev, REFGUID property, DWORD how,
						 DWORD obj, DWORD data)
{
	DIPROPDWORD dipdw;	

	dipdw.diph.dwSize = sizeof(dipdw);
	dipdw.diph.dwHeaderSize = sizeof(dipdw.diph);
	dipdw.diph.dwObj = obj;
	dipdw.diph.dwHow = how;
	dipdw.dwData = data;
	return IDirectInputDevice8_SetProperty( (LPDIRECTINPUTDEVICE8) dev, 
		property, &dipdw.diph);
}

//===========================================================================
// I_SetRangeProperty
//===========================================================================
HRESULT I_SetRangeProperty(void *dev, REFGUID property, DWORD how,
							  DWORD obj, int min, int max)
{
	DIPROPRANGE dipr;

	dipr.diph.dwSize = sizeof(dipr);
	dipr.diph.dwHeaderSize = sizeof(dipr.diph);
	dipr.diph.dwObj = obj;
	dipr.diph.dwHow = how;
	dipr.lMin = min;
	dipr.lMax = max;
	return IDirectInputDevice8_SetProperty( (LPDIRECTINPUTDEVICE8) dev,
		property, &dipr.diph);
}

//===========================================================================
// I_InitMouse
//===========================================================================
void I_InitMouse(void)
{
	HRESULT hr;

	if (ge_inf.nomouse || didMouse) return;

	hr = dInput->CreateDevice(GUID_SysMouse, &didMouse, 0);

	if(FAILED(hr))
	{
		Con_Message("I_InitMouse: failed to create device (0x%x).\n", hr);
		return;
	}
	
	// Set data format.
	hr = IDirectInputDevice8_SetDataFormat(didMouse, &c_dfDIMouse2);

	if(FAILED(hr))
	{
		Con_Message("I_InitMouse: failed to set data format (0x%x).\n", hr);
		goto kill_mouse;
	}

	// Set behaviour.
	hr = IDirectInputDevice8_SetCooperativeLevel(didMouse, hWndMain, 
		DISCL_FOREGROUND | DISCL_EXCLUSIVE);

	if(FAILED(hr))
	{
		Con_Message("I_InitMouse: failed to set co-op level (0x%x).\n", hr);
		goto kill_mouse;
	}

	// Acquire the device.
	IDirectInputDevice8_Acquire(didMouse);
	
	mouseInited = true;

	// Init was successful.
	return;

kill_mouse:;
	IDirectInputDevice8_Release(didMouse);
	didMouse = 0;
}

//===========================================================================
// I_JoyEnum
//===========================================================================
BOOL CALLBACK I_JoyEnum(LPCDIDEVICEINSTANCE lpddi, void *ref)
{
	// The first joystick is used by default.
	if(!firstJoystick.dwSize)
		memcpy(&firstJoystick, lpddi, sizeof(*lpddi));

	if(counter == joydevice)
	{
		// We'll use this one.
		memcpy(ref, lpddi, sizeof(*lpddi));
		return DIENUM_STOP;
	}
	counter++;
	return DIENUM_CONTINUE;
}

//===========================================================================
// I_InitJoystick
//===========================================================================
void I_InitJoystick(void)
{
	DIDEVICEINSTANCE ddi;
	HRESULT hr;

	int i, joyProp[] =
	{
		DIJOFS_X, DIJOFS_Y, DIJOFS_Z,
		DIJOFS_RX, DIJOFS_RY, DIJOFS_RZ,
		DIJOFS_SLIDER(0), DIJOFS_SLIDER(1)
	};
	const char *axisName[] =
	{
		"X", "Y", "Z", "RX", "RY", "RZ", "Slider 1", "Slider 2"
	};

	//---------------------------------------

	// ddi will contain info for the joystick device.
	memset(&firstJoystick, 0, sizeof(firstJoystick));
	memset(&ddi, 0, sizeof(ddi));
	counter = 0;

	hr = dInput->CreateDevice(GUID_Joystick, &didJoy, 0);

	if(FAILED(hr))
	{
		Con_Message("I_InitJoystick: no joystick devices found.\n");
		return;
	}

	// Find the joystick we want by doing an enumeration.
	dInput->EnumDevices(DI8DEVCLASS_GAMECTRL, I_JoyEnum, &ddi, DIEDFL_ALLDEVICES);

	// Was the joystick we want found?
	if(!ddi.dwSize) 
	{
		// Use the default joystick.
		if(!firstJoystick.dwSize) return; // Not found.
		Con_Message("I_InitJoystick: joydevice = %i, out of range.\n", 
			joydevice);
		// Use the first joystick that was found.
		memcpy(&ddi, &firstJoystick, sizeof(ddi));
	}

	// Show some info.
	Con_Message("I_InitJoystick: %s\n", ddi.tszProductName);

	// Create the joystick device.
	hr = dInput->CreateDevice(ddi.guidInstance, &didJoy, 0);
	if(FAILED(hr))
	{
		Con_Message("I_InitJoystick: failed to create device (0x%x).\n", hr);
		ge_inf.nojoystick = 2;
		return;
	}

	// Set data format.
	if(FAILED(hr = IDirectInputDevice8_SetDataFormat(didJoy, &c_dfDIJoystick)))
	{
		Con_Message("I_InitJoystick: failed to set data format (0x%x).\n", hr);
		goto kill_joy;
	}

	// Set behaviour.
	if(FAILED(hr = IDirectInputDevice8_SetCooperativeLevel(didJoy, hWndMain,
		DISCL_NONEXCLUSIVE | DISCL_FOREGROUND)))
	{
		Con_Message("I_InitJoystick: failed to set co-op level (0x%x: %s).\n", 
			hr, I_ErrorMsg(hr));
		goto kill_joy;
	}

	// Set properties.
	for(i = 0; i < sizeof(joyProp)/sizeof(joyProp[0]); i++)
	{
		if(FAILED(hr = I_SetRangeProperty(didJoy, DIPROP_RANGE, DIPH_BYOFFSET,
			joyProp[i], IJOY_AXISMIN, IJOY_AXISMAX)))
		{
			if(verbose) pr_OFinsert("I_InitJoystick: failed to set %s "
				"range (0x%x: %s).\n", axisName[i], hr, I_ErrorMsg(hr));
			//goto kill_joy;
		}
	}

	// Set no dead zone.
	if(FAILED(hr = I_SetProperty(didJoy, DIPROP_DEADZONE, DIPH_DEVICE, 0, 0)))
	{
		Con_Message("I_InitJoystick: failed to set dead zone (0x%x: %s).\n", 
			hr, I_ErrorMsg(hr));
	}

	// Set absolute mode.
	if(FAILED(hr = I_SetProperty(didJoy, DIPROP_AXISMODE, DIPH_DEVICE,
		0, DIPROPAXISMODE_ABS)))
	{
		Con_Message("I_InitJoystick: failed to set absolute axis mode (0x%x: %s).\n",
			hr, I_ErrorMsg(hr));
	}

	// Acquire it.
	IDirectInputDevice8_Acquire(didJoy);

	joyInited = true;

	// Initialization was successful.
	return;

kill_joy:;
	IDirectInputDevice8_Release(didJoy);
}

//===========================================================================
// I_Init
//	Initialize input. Returns true if successful.
//===========================================================================
int I_Init(void)
{
	HRESULT hr;

	// We'll create the DirectInput object. The only mandatory
	// input device is the keyboard. The others are optional.
	dInput = NULL;

	if (FAILED(hr = CoCreateInstance(
					CLSID_DirectInput8,
					NULL,
					CLSCTX_INPROC_SERVER,
					IID_IDirectInput8,
					(void**)&dInput)) 
		|| FAILED(hr = dInput->Initialize(
					hInstApp, 
					DIRECTINPUT_VERSION)))
	{
		Con_Message("I_Init: DirectInput 8 init failed (0x%x).\n", hr);
		return false;
	}

	if(!dInput)
	{
		Con_Message("I_Init: DirectInput init failed.\n");
		return false;
	}

	// Create the keyboard device.
	hr = dInput->CreateDevice(GUID_SysKeyboard, &didKeyb, 0);
	if(FAILED(hr))
	{
		Con_Message("I_Init: failed to create keyboard device (0x%x).\n", hr);
		return false;
	}

	// Setup the keyboard input device.
	hr = IDirectInputDevice8_SetDataFormat(didKeyb, &c_dfDIKeyboard);
	if(FAILED(hr))
	{
		Con_Message("I_Init: failed to set keyboard data format (0x%x).\n", hr);
		return false;
	}

	// Set behaviour.
	hr = IDirectInputDevice8_SetCooperativeLevel(didKeyb, hWndMain, 
		DISCL_FOREGROUND | DISCL_NONEXCLUSIVE | DISCL_NOWINKEY);
	if(FAILED(hr))
	{
		Con_Message("I_Init: failed to set keyboard co-op level (0x%x).\n", hr);
		return false;
	}

	// The input buffer size.
	hr = I_SetProperty(didKeyb, DIPROP_BUFFERSIZE, DIPH_DEVICE, 0, KEYBUFSIZE);
	if(FAILED(hr))
	{
		Con_Message("I_Init: failed to set keyboard buffer size (0x%x).\n", hr);
		return false;
	}
	
	// Acquire the keyboard.
	IDirectInputDevice8_Acquire(didKeyb);
	I_initOk = true;

	// Create the mouse and joystick devices.
	// We can live with init failing for these

	if (!ge_inf.nomouse)
	{
		I_InitMouse();
		if (!mouseInited) ge_inf.nomouse = 2;
	}

	if (!ge_inf.nojoystick)
	{
		I_InitJoystick();
		if (!joyInited) ge_inf.nojoystick = 2;
	}

	return true;
}

//===========================================================================
// I_Shutdown
//===========================================================================
void I_Shutdown(void)
{
	if (I_initOk)
	{
		I_initOk = false;

		// Release all the input devices.
		if (mouseInited)
		{
			IDirectInputDevice8_Unacquire(didMouse);
			IDirectInputDevice8_Release  (didMouse);
			mouseInited = false;
		}
		if (joyInited  )
		{
			IDirectInputDevice8_Unacquire(didJoy);
			IDirectInputDevice8_Release  (didJoy);
			joyInited   = false;
		}

		IDirectInputDevice8_Unacquire(didKeyb);
		IDirectInputDevice8_Release  (didKeyb);

		// Release DirectInput.
		IDirectInput8_Release(dInput);
	}
}

//===========================================================================
// I_GetKeyEvents
//===========================================================================
int I_GetKeyEvents(keyevent_t *evbuf)
{
	DIDEVICEOBJECTDATA keyData[KEYBUFSIZE];
	HRESULT		hr;
	DWORD		i, dwItems;

	if (!I_initOk) return 0;
	
	// Try two times to get the data.
	for (i = 1; i >= 0; i--)
	{
		dwItems = KEYBUFSIZE;

		hr = IDirectInputDevice8_GetDeviceData(didKeyb, 
			sizeof(DIDEVICEOBJECTDATA), keyData, &dwItems, 0);

		if (SUCCEEDED(hr)) break;
		
		// No tries left? The operation is a failure.
		if (!i) return 0;		

		// Try to reacquire.
		IDirectInputDevice8_Acquire(didKeyb);
	}

	// Get the events.
	for (i = 0; i < dwItems; i++)
	{
		evbuf[i].event = (keyData[i].dwData & 0x80)? IKE_KEY_DOWN : IKE_KEY_UP;
		evbuf[i].code = (unsigned char) keyData[i].dwOfs;
	}
	return (int)i;
}

//===========================================================================
// I_GetMouseState
//===========================================================================
void I_GetMouseState(mousestate_t *state, RDBOOL flush)
{
	DIMOUSESTATE2	mstate;
	HRESULT			hr;
	int				i;

	memset(state, 0, sizeof(*state));

	// Has the mouse been initialized?
	if (!mouseInited) return;

	// Try to get the mouse state.
	for (i = 1; i >= 0; i--)
	{
		hr = IDirectInputDevice8_GetDeviceState(didMouse, 
			sizeof(mstate), &mstate);

		if(SUCCEEDED(hr)) break;

		if(!i) return;	// Out of tries

		IDirectInputDevice8_Acquire(didMouse);	// Try to reacquire.
	}

	if (flush) return;

	// Fill in the state structure.
	state->x = mstate.lX;
	state->y = mstate.lY;
	state->z = mstate.lZ;
	
	// The buttons bitfield is ordered according to the numbering.
	for(i = 0; i < NUM_MOUSE_BUTTONS; i++)
	{
		if(mstate.rgbButtons[i] & 0x80) state->buttons |= 1 << i;
	}
}

//===========================================================================
// I_GetJoystickState
//===========================================================================
void I_GetJoystickState(joystate_t *state)
{
	DIJOYSTATE	dijoy;
	HRESULT		hr;
	int			i;
	int			pov;

	memset(state, 0, sizeof(joystate_t));

	// Initialization has not been done.
	if (!joyInited  || !cfg.joyEnable) return;

	// Some joysticks need to be polled.
	IDirectInputDevice8_Poll(didJoy);

	for (i = 0; i < 2; i++)
	{
		hr = IDirectInputDevice8_GetDeviceState(didJoy, 
			sizeof(dijoy), &dijoy);

		if (SUCCEEDED(hr)) break;

		if (i) return;	// Out of tries

		IDirectInputDevice8_Acquire(didJoy);	// Try to reacquire.
	}

	state->axis[0] = INV(dijoy.lX, 0);
	state->axis[1] = INV(dijoy.lY, 1);
	state->axis[2] = INV(dijoy.lZ, 2);

	state->rotAxis[0] = INV(dijoy.lRx, 3);
	state->rotAxis[1] = INV(dijoy.lRy, 4);
	state->rotAxis[2] = INV(dijoy.lRz, 5);

	state->slider[0] = INV(dijoy.rglSlider[0], 6);
	state->slider[1] = INV(dijoy.rglSlider[1], 7);

	for(i = 0; i < IJOY_MAXBUTTONS; i++)
		state->buttons[i] = (dijoy.rgbButtons[i] & 0x80) != 0;

	pov = dijoy.rgdwPOV[0];

	if((pov & 0xffff) == 0xffff)
		state->povAngle = IJOY_POV_CENTER;
	else
		state->povAngle = pov / 100.0f;
}