//-------------------------------------------------------------------------------------------------
//
// Copyright (C) 2012 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: sys_midi.c
//
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
// Revision 2.2.0 build 20 (v220-20) 31/Mar/2013 gmj
// Revision 2.2.0 build 26 (v220-26) 05/Jul/2014 gmj
//
// With acknowlegement to MSVC 1998 sample midi player
//
// DESCRIPTION:  Windows Midi/Riff File Sequencer
//
//-------------------------------------------------------------------------------------------------

// HEADER FILES -----------------------------------------------------------------------------------

#include <windows.h>
#include <mmsystem.h>
#include <malloc.h>
#include <math.h>
#include <stdio.h>

#include "m_winmem.h"
#include "s_mus.h"
#include "m_args.h"
#include "m_misc.h"
#include "r3d_compat.h"

// MACROS ----------------------------------------------------------------------------------------

#ifndef MOD_WAVETABLE
	#define MOD_WAVETABLE 6
#endif
#ifndef MOD_SWSYNTH
	#define MOD_SWSYNTH   7
#endif

#define SEQ_NOFILE			0
#define SEQ_OPENED			1
#define SEQ_PREPARSED		2
#define SEQ_PLAYING			3
#define SEQ_PAUSED			4
#define SEQ_STOPPING		5
#define SEQ_RESET			6

#define MAX_TICKS			0xFFFFFFFF
#define C_TEMPO_MAP_CHK		16

#define C_MIDI_BUFFERS		4		// number of buffers to assign
#define C_MIDI_BUFFERSIZE	1024	// size of (each) buffer

#define SQF_EOF				0x0001
#define SQF_INSERTSYSEX		0x0002

// default if no tempo events exist
#define MIDI_DEFAULT_TEMPO	(500000L)

#define MIDI_TEMPO			0x51
#define MIDI_EOT			0x2F
#define MIDI_MSG			0x80
#define MIDI_CTRLCHANGE		0xB0
#define MIDI_PROGCHANGE		0xC0
#define MIDI_SYSEX			0xF0
#define MIDI_SYSEXEND		0xF7
#define MIDI_META			0xFF

#define EK_UNSET	0xFF

// Midi File Sequencer flags
enum {
	MFS_SUCCESS = 0,
	MFS_NOMEMORY,
	MFS_BADFILE,
	MFS_BADTRACK,
	MFS_TRACKEND ,
	MFS_EOF,
	MFS_CLOSED,
} mfs_e;

// TYPES ------------------------------------------------------------------------------------------

typedef HANDLE	HMFS;
typedef DWORD	MFSRESULT;
typedef DWORD	TICKS;

// Used to track events
// Where not seen then values will be set to EK_UNSET
typedef struct
{
    // Meta events.
    BYTE        rbTempo[3];
    // MIDI channel messages.
    BYTE        rbProgram[16];
    BYTE        rbControl[16*120];

} eventkey_t;

typedef struct
{
	UINT		uState;			// Sequencer state
	TICKS		tcLength;		// Length of longest track

	HMFS		hMfs;			// Handle to open file
	HMIDIOUT	hmidi;			// Handle to open MIDI device
	DWORD		dwTimeDiv;		// time division

	// Info for stream buffers
	BYTE*		pbAlloc;		// initial allocation
	PMIDIHDR	pmhFree;		// free list
	PMIDIHDR	pmhPreparse;	// main buffer..
	UINT		midiBuffers;	// buffers in use

	TICKS		tcbase;			// Where playback started from in stream
	TICKS		tcEnd;			// Where playback should end

	WORD		looped;			// set if file should loop
	WORD		wEof;			// set if file eof has been reached

} sequencer_t;

typedef struct
{
	TICKS	tcTempo;
	DWORD	msBase;
	DWORD	dwTempo;

} tempo_t;

typedef struct
{
	TICKS	tcPosition;
	DWORD	cbLeft;
	DWORD	idxTrack;
	BYTE	*pbData;
	BYTE	bEvent;

	DWORD	fdwTrack;
	DWORD	cbLength;

} track_t;

typedef struct
{
	TICKS	tcDelta;
	BYTE	bEv[3];
	BYTE	align;

	DWORD	cbParm;
	BYTE	*pbParm;

} mevent_t;

typedef struct
{
	FOURCC	fourccType;
	DWORD	dwLength;

} chunkhdr_t;

typedef struct
{
	BYTE		*pbFileData;
	DWORD		szFileData;
	HTASK		htask;

	TICKS		tcPosition;
	TICKS		tcLength;
	DWORD		dwFormat;
	DWORD		dwTracks;
	DWORD		dwTimeDiv;
	WORD		wfMFS;

	DWORD		cTempoMap;
	DWORD		cTempoMapAlloc;
	tempo_t		*pTempoMap;

	DWORD		dwPendingEvent;
	DWORD		cbPendingEvent;
	BYTE		*pbPendingEvent;

} mfs_t;

#pragma pack(2)	// match file

typedef struct tag_filehdr
{
	WORD	wFormat;
	WORD	wTracks;
	WORD	wDivision;

} filehdr_t;

#pragma pack()

// EXTERNAL FUNCTION PROTOTYPES -------------------------------------------------------------------

void	Con_Error(char *error, ...);
void	Con_Message(char *message, ...);
void	pr_OFinsert(char *message, ...);
void	pr_OFdirect(char *msg);

// PUBLIC FUNCTION PROTOTYPES ---------------------------------------------------------------------

char	lastmidifname[256];
short	iMidiFlags;

// PRIVATE FUNCTION PROTOTYPES --------------------------------------------------------------------

static MFSRESULT	mfsReadEvents(HMFS hMfs, PMIDIHDR pmh, TICKS tcMax);
static MFSRESULT	mfsInsertParmData(TICKS tcDelta, PMIDIHDR pmh);
static MFSRESULT	mfsSeek(TICKS tcPosition, PMIDIHDR pmh);
static MFSRESULT	mfsNextEvent(mevent_t *pevent, TICKS tcMax);

static void CALLBACK seqCallback(HMIDISTRM hms, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2);

// EXTERNAL DATA DECLARATIONS ---------------------------------------------------------------------

// Needed by MC_WinMidiTime()
extern BOOL		paused;
extern BOOL		ui_active;
extern int		mididev;

// Used for setting midi device in ui
#include "ui_main.h"
extern uidata_listitem_t *lstit_mididev;
extern uidata_list_t lst_mididev;


// PUBLIC DATA DEFINITIONS ------------------------------------------------------------------------

char		*midi_mname;
uint		numMidiDev;
int			mididev;
int			midi_error;

// PRIVATE DATA DEFINITIONS -----------------------------------------------------------------------

static WORD			midi_available;
static DWORD		midilengthMs;
static DWORD		uDeviceID = 0;

// for volune setting
static BOOL			wmex = false;
static DWORD		dwVolume;
static mfs_t		pMfs[1];
static track_t		*tracks;
static eventkey_t	evk;
static sequencer_t	sequencer;
static uint			dwAlloc;
static bool			seqActive;	// v220-26


// Number of bytes used by an event.
// Note: only top nibble of an event byte used for look-up
static UINT evLength[] =
{
	// 0x to 7x are not status bytes
	0, 0, 0, 0, 0, 0, 0, 0,
	3,	// 8x	Note off
	3,	// 9x	Note on
	3,	// Ax	Poly pressure
	3,	// Bx	Control change
	2,	// Cx	Program change
	2,	// Dx	Chan pressure
	3,	// Ex	Pitch bend change
	0,	// Fx	SysEx
};

// CODE -------------------------------------------------------------------------------------------

#pragma optimize("g",off)

// Called once on startup
static bool seqAllocBuffers()
{
	MIDIHDR		*pmh, *pmhfree;
	DWORD		dwSzBuffer;
	BYTE		*pbWork;
	UINT		i;

	// ------------------------------------

	sequencer.uState   = SEQ_NOFILE;
	sequencer.pmhFree  = NULL;
	sequencer.pbAlloc  = NULL;
	sequencer.hMfs     = (HMFS)NULL;
	
	// First allocate the total buffer memory required
	dwSzBuffer = sizeof(MIDIHDR) + C_MIDI_BUFFERSIZE;
	dwAlloc    = dwSzBuffer * C_MIDI_BUFFERS;

	pbWork = (BYTE*)calloc(dwAlloc, 1);

	if (!pbWork) return false;

	sequencer.pbAlloc = pbWork;

	// Initialize all MIDIHDR's and put them into a free list
	for (i = 0, pmhfree = NULL; i < C_MIDI_BUFFERS; i++)
	{
		pmh = (PMIDIHDR)pbWork;

		pmh->lpNext           = pmhfree;
		pmh->lpData           = (char*)(pbWork + sizeof(MIDIHDR));
		pmh->dwBufferLength   = C_MIDI_BUFFERSIZE;
		pmh->dwBytesRecorded  = 0;
		pmh->dwUser           = (DWORD)&sequencer;
		pmh->dwFlags          = 0;
		
		pmhfree = pmh;
		pbWork += dwSzBuffer;
	}
	
	sequencer.pmhFree = pmhfree;
	return true;
}

// Determine, if functioning, the current playback position
static DWORD seqTime()
{	
	MMRESULT	mmr;
	MMTIME		mmt;

	if (   sequencer.uState != SEQ_PLAYING
		&& sequencer.uState != SEQ_PAUSED)
	{
		return 0;
	}

	mmt.wType = TIME_TICKS;
	mmr = midiStreamPosition((HMIDISTRM)sequencer.hmidi, &mmt, sizeof(mmt));

	if (mmr != MMSYSERR_NOERROR) return sequencer.tcbase;
	
	return mmt.u.ticks + sequencer.tcbase;
}

// stop playback, if functioning, of an instance.
static void seqStop()
{
	if (midi_available)
	{
		while(seqActive) Sleep(1);

		if (   sequencer.uState != SEQ_PLAYING
			&& sequencer.uState != SEQ_PAUSED)
		{
			return;	// already stopped
		}
		
		sequencer.uState = SEQ_STOPPING;
		
		if (!midiStreamStop((HMIDISTRM)sequencer.hmidi))
		{
			while (sequencer.midiBuffers) Sleep(0);
		}
	}
}

static MMRESULT seqPreparse()
{
	MFSRESULT			mfsrc;
	MMRESULT			mmrc = MMSYSERR_NOERROR;
	MIDIPROPTIMEDIV		mptd;
	MIDIHDR				*pmh = NULL;
	DWORD				dwSzTotalBuffer = C_MIDI_BUFFERS * C_MIDI_BUFFERSIZE;

	// ------------------------------------
	
	if (   sequencer.uState != SEQ_OPENED
		&& sequencer.uState != SEQ_PREPARSED)
	{
		return MCIERR_UNSUPPORTED_FUNCTION;
	}
	
	if (sequencer.hmidi)
	{
		sequencer.uState = SEQ_RESET;
		midiOutReset(sequencer.hmidi);
		
		while (sequencer.midiBuffers) Sleep(0);
	}
	
	sequencer.midiBuffers = 0;
	
	// Open the MIDI device and set the time division.
	// NOTE: the device will already be open if looping
	if (!sequencer.hmidi)
	{
		if ((mmrc = midiStreamOpen(
			(LPHMIDISTRM)&sequencer.hmidi,
			(LPUINT)&uDeviceID,
			1,
			(DWORD)seqCallback,
			0,
			CALLBACK_FUNCTION) ) != MMSYSERR_NOERROR)
		{
			sequencer.hmidi = NULL;
			goto preparse_abort;
		}
		
		mptd.cbStruct  = sizeof(mptd);
		mptd.dwTimeDiv = sequencer.dwTimeDiv;

		if ((mmrc = midiStreamProperty(
			(HMIDISTRM)sequencer.hmidi,
			(LPBYTE)&mptd,
			MIDIPROP_SET|MIDIPROP_TIMEDIV)) != MMSYSERR_NOERROR)
		{
			midiStreamClose((HMIDISTRM)sequencer.hmidi);
			sequencer.hmidi = NULL;
			goto preparse_abort;
		}
	}

	mmrc = MMSYSERR_NOERROR;

	if (!sequencer.pmhPreparse)
	{
		sequencer.pmhPreparse = (PMIDIHDR)malloc(dwSzTotalBuffer);
		sequencer.pmhPreparse->dwBufferLength = dwSzTotalBuffer - sizeof(MIDIHDR);
	}

	sequencer.pmhFree						= NULL;
	sequencer.pmhPreparse->lpNext           = sequencer.pmhFree;
	sequencer.pmhPreparse->lpData           = (char*)sequencer.pmhPreparse + sizeof(MIDIHDR);
	sequencer.pmhPreparse->dwUser           = (DWORD)&sequencer;
	sequencer.pmhPreparse->dwBytesRecorded  = 0;
	sequencer.pmhPreparse->dwFlags          = 0;

	mfsrc = mfsSeek(sequencer.tcbase, sequencer.pmhPreparse);

	if (mfsrc != MFS_SUCCESS)
	{
		mmrc = MCIERR_OUT_OF_MEMORY;
		goto preparse_abort;
	}

	if ((mmrc = midiOutPrepareHeader(sequencer.hmidi, sequencer.pmhPreparse, sizeof(MIDIHDR)))
		!= MMSYSERR_NOERROR)
	{
		goto preparse_abort;
	}
	
	++sequencer.midiBuffers;
	
	if ((mmrc = midiStreamOut((HMIDISTRM)sequencer.hmidi, sequencer.pmhPreparse, sizeof(MIDIHDR)))
		!= MMSYSERR_NOERROR)
	{
		--sequencer.midiBuffers;
		goto preparse_abort;
	}
	
	sequencer.wEof = FALSE;

	while (sequencer.pmhFree)
	{
		pmh = sequencer.pmhFree;
		sequencer.pmhFree = pmh->lpNext;
		mfsrc = mfsReadEvents(sequencer.hMfs, pmh, sequencer.tcEnd);
		
		if (mfsrc != MFS_SUCCESS && mfsrc != MFS_EOF)
		{
			mmrc = MMSYSERR_READERROR;
			goto preparse_abort;
		}
		
		if ((mmrc = midiOutPrepareHeader(sequencer.hmidi, pmh, sizeof(*pmh)))
			!= MMSYSERR_NOERROR)
		{
			goto preparse_abort;
		}
		
		if ((mmrc = midiStreamOut((HMIDISTRM)sequencer.hmidi, pmh, sizeof(*pmh)))
			!= MMSYSERR_NOERROR)
		{
			goto preparse_abort;
		}
		
		++sequencer.midiBuffers; 
		
		if (mfsrc == MFS_EOF)
		{
			// Note: at this point mmrc will be set to MMSYSERR_NOERROR
			break;
		}
	}
	
preparse_abort:;

	if (mmrc == MMSYSERR_NOERROR)
		sequencer.uState = SEQ_PREPARSED;
	else
		sequencer.uState = SEQ_OPENED;
	
	return mmrc;
}

static DWORD mfsTicksToMillisecs(HMFS hMfs, TICKS tkOffset)
{
	tempo_t		*pTempo;
	UINT		idx;
	UINT		uSMPTE;
	DWORD		dwTicksPerSec;

	// ------------------------------------

	if (!hMfs) return 0;

	if (tkOffset > pMfs->tcLength)
	{
		tkOffset = pMfs->tcLength;
	}

	// SMPTE time uses linear conversion (tempo map not needed)
	// Note: 30-Drop is meaningless since we're not converting
	// to a format which needs it.

	if (pMfs->dwTimeDiv & 0x8000)
	{
		uSMPTE = -(int)(char)((pMfs->dwTimeDiv >> 8)&0xFF);
		
		if (uSMPTE == 29) uSMPTE = 30;
		
		dwTicksPerSec = (DWORD)uSMPTE *
			(DWORD)(BYTE)(pMfs->dwTimeDiv & 0xFF);
		
		return MULDIV32(tkOffset, 1000L, dwTicksPerSec);
	}
       
	//	Get the nearest tick position using the tempo map
	pTempo = pMfs->pTempoMap;

	if (!pTempo) return 0;

	for (idx = 0; idx < pMfs->cTempoMap; idx++, pTempo++)
	{
		if (tkOffset < pTempo->tcTempo)
			break;
	}

	pTempo--;

	// pTempo is the tempo map entry preceding the requested tick offset.

	return pTempo->msBase +
		MULDIV32(tkOffset - pTempo->tcTempo,
				 pTempo->dwTempo,
				 1000L * pMfs->dwTimeDiv);   
}

// If file was opened unqueue buffers 
static MMRESULT seqCloseFile()
{
	PMIDIHDR	pmh;

	if (sequencer.uState != SEQ_OPENED)
		return MCIERR_UNSUPPORTED_FUNCTION;

	for (pmh = sequencer.pmhFree; pmh; pmh = pmh->lpNext)
		midiOutUnprepareHeader(sequencer.hmidi, pmh, sizeof(*pmh));

	if (sequencer.pmhPreparse)
		midiOutUnprepareHeader(sequencer.hmidi, sequencer.pmhPreparse,
															sizeof(*sequencer.pmhPreparse));

	if (sequencer.hmidi != NULL)
	{
		midiStreamClose((HMIDISTRM)sequencer.hmidi);
		sequencer.hmidi = NULL;
	}

	sequencer.hMfs = (HMFS)NULL;
	sequencer.uState = SEQ_NOFILE;

	return MMSYSERR_NOERROR;
}

static DWORD mfsGetDelta(BYTE *pbData, DWORD dwLeft, DWORD *pDw)
{
	BYTE	b;
	DWORD	dwUsed  = 0;

	*pDw = 0;

	do
	{
		if (!dwLeft) return 0;

		dwLeft--;
		dwUsed++;
		b = *pbData++;
		*pDw = (*pDw << 7) | (b & 0x7F);
		
	} while (b&0x80);

	return dwUsed;
}

// Get the file header and build a track index.
static MFSRESULT  mfsBuildFileIndex()
{
	chunkhdr_t	*pCh;
	filehdr_t	*pFh;
	track_t		*trk;
	tempo_t		*pTempo;
	mevent_t	event;

	MFSRESULT	mfsrc;
	DWORD		idx;
	DWORD		dwLeft;
	DWORD		dwLength;
	BYTE		*pbData;
	BOOL		tempoMapMemorySet = false;
			
	// ------------------------------------
	// MIDI data image is already in pMfs->pbData
	// (extracted from RIFF header if exists).
	
	// Validate MIDI header
	dwLeft = pMfs->szFileData;
	pbData = pMfs->pbFileData;
	
	if (dwLeft < sizeof(chunkhdr_t))
		return MFS_BADFILE;
	
	pCh = (chunkhdr_t*)pbData;
	
	dwLeft -= sizeof(chunkhdr_t);
	pbData += sizeof(chunkhdr_t);
	
	if (pCh->fourccType != mmioFOURCC('M','T','h','d'))
		return MFS_BADFILE;
	
	dwLength = DWORDSWAP(pCh->dwLength);
	if (dwLength < sizeof(filehdr_t) || dwLength > dwLeft)
		return MFS_BADFILE;
	
	pFh = (filehdr_t*)pbData;
	
	dwLeft -= dwLength;
	pbData += dwLength;
	
	pMfs->dwFormat  = (DWORD)(WORDSWAP(pFh->wFormat));
	pMfs->dwTracks  = (DWORD)(WORDSWAP(pFh->wTracks));
	pMfs->dwTimeDiv = (DWORD)(WORDSWAP(pFh->wDivision));

	if (!pMfs->dwTracks || pMfs->dwTracks > 64) return MFS_BADFILE;

	tracks = (track_t*)W_Realloc(tracks, pMfs->dwTracks * sizeof(track_t), STATICID, 0);
	
	//  Inspect track headers as a preliminary integrity check.	
	for (idx = 0, trk = tracks; idx < pMfs->dwTracks; idx++, trk++)
	{
		if (dwLeft < sizeof(chunkhdr_t)) return MFS_BADFILE;
		
		pCh = (chunkhdr_t*)pbData;
		
		if (pCh->fourccType != mmioFOURCC('M','T','r','k'))
			return MFS_BADFILE;
		
		dwLeft -= sizeof(chunkhdr_t);
		pbData += sizeof(chunkhdr_t);
		
		trk->idxTrack = (DWORD)(pbData - pMfs->pbFileData);
		trk->cbLength = DWORDSWAP(pCh->dwLength);
		
		if (trk->cbLength > dwLeft) return MFS_BADFILE;
		
		dwLeft -= trk->cbLength;
		pbData += trk->cbLength;
	}
	
	// Track headers check out OK so now preparse to;
	// 1) Build tempo map for fast ticks to position conversion
	// 2) Validate all track events
	// 3) Determine file's tick length

	pMfs->tcPosition = 0;
	pMfs->wfMFS &= ~SQF_EOF;
	
	for (trk = tracks, idx = pMfs->dwTracks; idx--; trk++)
	{
		trk->tcPosition	= 0;
		trk->cbLeft		= trk->cbLength;
		trk->pbData		= pMfs->pbFileData + trk->idxTrack;
		trk->bEvent		= 0;
		trk->fdwTrack	= 0;
	}
	
	while (MFS_SUCCESS == (mfsrc = mfsNextEvent((mevent_t*)&event, MAX_TICKS)))
	{
		if (MIDI_META == event.bEv[0] && MIDI_TEMPO == event.bEv[1])
		{
			BOOL tNext;

			if (event.cbParm != 3) return MFS_BADFILE;
			
			if (pMfs->cTempoMap >= pMfs->cTempoMapAlloc)
			{
				UINT sz;
				
				pMfs->cTempoMapAlloc += C_TEMPO_MAP_CHK;
				sz = (UINT)(pMfs->cTempoMapAlloc*sizeof(tempo_t));
				
				pMfs->pTempoMap = (tempo_t*)W_Realloc(pMfs->pTempoMap, sz, STATICID, 0);

				if (!pMfs->pTempoMap) return MFS_NOMEMORY;

				tempoMapMemorySet = true;
			}
			
			tNext = pMfs->cTempoMap;
			pTempo = &pMfs->pTempoMap[pMfs->cTempoMap++];
			
			if (tNext)
			{
				pTempo->tcTempo = pMfs->tcPosition;

				pTempo->msBase = (pTempo-1)->msBase +
					MULDIV32(pTempo->tcTempo - ((pTempo-1)->tcTempo),
					(pTempo-1)->dwTempo,
					1000L * pMfs->dwTimeDiv);
			}
			else
			{				
				pTempo->tcTempo = 0;
				pTempo->msBase  = 0;
				pTempo->dwTempo = MIDI_DEFAULT_TEMPO;
			}
			

			pTempo->dwTempo =
				(((DWORD)event.pbParm[0])<<16)|
				(((DWORD)event.pbParm[1])<< 8)|
				( (DWORD)event.pbParm[2]);
		}
	}
	
	if (!tempoMapMemorySet)
	{
		pMfs->pTempoMap = (tempo_t*)W_Realloc(pMfs->pTempoMap, sizeof(tempo_t), STATICID, 0);
		
		if (!pMfs->pTempoMap) return MFS_NOMEMORY;
		
		pMfs->cTempoMap = 1;
		pMfs->cTempoMapAlloc = 1;
		
		pMfs->pTempoMap->tcTempo = 0;
		pMfs->pTempoMap->msBase  = 0;
		pMfs->pTempoMap->dwTempo = MIDI_DEFAULT_TEMPO;
	}
	
	if (mfsrc == MFS_EOF || mfsrc == MFS_SUCCESS)
	{
		pMfs->tcLength = pMfs->tcPosition;
		mfsrc = MFS_SUCCESS;
	}
	
	return mfsrc;
}

static MFSRESULT  mfsNextEvent(mevent_t *pEvent, TICKS	tcMax)
{
	track_t	*trk;
	track_t	*trkFound;
	TICKS	tcEventDelta;
	TICKS	tcRelTime;
	TICKS	tcMinRelTime;
	BYTE	bEvent;
	DWORD	idx;
	DWORD	dwTotal;
	DWORD	dwNDB;
	DWORD	cbEvent;
	UINT	uVal;

	// ------------------------------------

	if (pMfs->wfMFS & SQF_EOF) return MFS_EOF;

	trkFound = NULL;
	tcMinRelTime = MAX_TICKS;

	for (trk = tracks, idx = pMfs->dwTracks; idx--; trk++)
	{
		if (trk->fdwTrack & SQF_EOF) continue;
		
		if (!mfsGetDelta(trk->pbData, trk->cbLeft, (DWORD*)&tcEventDelta))
		{
			return MFS_BADFILE;
		}
		
		tcRelTime = trk->tcPosition + tcEventDelta - pMfs->tcPosition;
		
		if (tcRelTime < tcMinRelTime)
		{
			tcMinRelTime = tcRelTime;
			trkFound = trk;
		}
	}
	
	if (!trkFound)
	{
		pMfs->wfMFS |= SQF_EOF;
		return MFS_EOF;
	}
	
	trk = trkFound;

	if (pMfs->tcPosition + tcMinRelTime >= tcMax)
	{
		return MFS_TRACKEND;
	}
		
	trk->pbData += (dwNDB = mfsGetDelta(trk->pbData, trk->cbLeft, (DWORD*)&tcEventDelta));
	trk->cbLeft -= dwNDB;

	// Check at least 3 bytes are left as an EOF meta
	// event needs this and one has yet to be seen.
	// This also covers remaining events (which need 1 to 3 bytes)
	if (trk->cbLeft < 3) return MFS_BADFILE;
	
	trk->tcPosition += tcEventDelta;
	pEvent->tcDelta = trk->tcPosition - pMfs->tcPosition;
	pMfs->tcPosition = trk->tcPosition;
	
	bEvent = *trk->pbData++;
	
	if (MIDI_MSG > bEvent)
	{
		if (!trk->bEvent) return MFS_BADFILE;
		
		dwTotal = 1;
		pEvent->bEv[0] = trk->bEvent;
		pEvent->bEv[1] = bEvent;

		uVal = (trk->bEvent >> 4) & 0x0F;

		if (evLength[uVal] == 3)
		{
			pEvent->bEv[2] = *trk->pbData++;
			dwTotal++;
		}
	}
	else if (MIDI_SYSEX > bEvent)
	{
		trk->bEvent = bEvent;
		
		dwTotal = 2;
		pEvent->bEv[0] = bEvent;
		pEvent->bEv[1] = *trk->pbData++;

		uVal = (bEvent >> 4) & 0x0F;

		if (evLength[uVal] == 3)
		{
			pEvent->bEv[2] = *trk->pbData++;
			dwTotal++;
		}
	}
	else
	{
		trk->bEvent = 0;

		if (MIDI_META == bEvent)
		{
			pEvent->bEv[0] = MIDI_META;

			if (MIDI_EOT == (pEvent->bEv[1] = *trk->pbData++))
			{
				trk->fdwTrack |= SQF_EOF;
			}
			
			dwTotal = 2;
		}
		else if (MIDI_SYSEX == bEvent || MIDI_SYSEXEND == bEvent)
		{
			pEvent->bEv[0] = bEvent;
			dwTotal = 1;
		}
		else
		{
			return MFS_BADFILE;
		}
		
		if (0 == (dwNDB = mfsGetDelta(trk->pbData, trk->cbLeft - 2, (DWORD*)&cbEvent)))
		{
			return MFS_BADFILE;
		}
		
		trk->pbData  += dwNDB;
		dwTotal      += dwNDB;
		
		if (dwTotal + cbEvent > trk->cbLeft)
		{
			return MFS_BADFILE;
		}

		pEvent->cbParm  = cbEvent;
		pEvent->pbParm = trk->pbData;

		trk->pbData += cbEvent;
		dwTotal     += cbEvent;
	}

	if (trk->cbLeft < dwTotal) return MFS_BADFILE;

	trk->cbLeft -= dwTotal;

	return MFS_SUCCESS;
}

static setDWORD(DWORD v1, DWORD v2, DWORD v3, DWORD v4)
{
	return v1|(v2 << 8)|(v3 << 16)|(v4 << 24);
}

static MFSRESULT mfsReadEvents(HMFS hMfs, PMIDIHDR pmh, TICKS tcMax)
{
	mevent_t	event;
	MFSRESULT	mfsrc;
	DWORD		*pdw;
	
	if (!hMfs) return MFS_CLOSED;

	//	Read track events and pass them into the buffer in polymsg format.
	//	A SysEx or meta event > than the buffer's limit gets split.

	pmh->dwBytesRecorded = 0;
	
	if (   pMfs->dwPendingEvent
		&& (mfsrc = mfsInsertParmData((TICKS)0, pmh)) != MFS_SUCCESS)
	{
		return mfsrc;
	}
	
	pdw = (DWORD*)(pmh->lpData + pmh->dwBytesRecorded);
	
	if (pMfs->wfMFS & SQF_EOF)
	{
		return MFS_EOF;
	}
	
	while(1)
	{
		if (pmh->dwBytesRecorded > pmh->dwBufferLength)	// shouldn't happen
		{
			pMfs->wfMFS |= SQF_EOF;
			break;
		}
		
		// If there's no room for the event then break.  We need 2 DWORD's for
		// the final event and at least 2 DWORD's for any event to be stored.

		if (pmh->dwBufferLength - pmh->dwBytesRecorded < 4*sizeof(DWORD))
		{
			break;
		}
		
		mfsrc = mfsNextEvent((mevent_t*)&event, tcMax);

		if (mfsrc != MFS_SUCCESS)
		{
			// mfsNextEvent doesn't set this because it's also used by mfsSeek
			// which has to distinguish between reaching the seek point of the
			// entire file and not the end of any one track.

			if (mfsrc == MFS_TRACKEND) pMfs->wfMFS |= SQF_EOF;
			
			break;
		}

		// NOTE; MIDI_EOT == event.bEv[1] is ignored here.
		// mfsReadNextEvent() takes care of track merging.
		
		if (event.bEv[0] < MIDI_SYSEX)
		{
			*pdw++ = (DWORD)event.tcDelta;
			*pdw++ = 0;
			*pdw++ = setDWORD(event.bEv[0], event.bEv[1], event.bEv[2], MEVT_SHORTMSG);
			pmh->dwBytesRecorded += 3*sizeof(DWORD);
		}
		else if (event.bEv[0] != MIDI_META)
		{
			//	Must be F0, F7 sysex or unrecognised FF meta

			pMfs->cbPendingEvent = event.cbParm;
			pMfs->pbPendingEvent = event.pbParm;
			pMfs->wfMFS &= ~SQF_INSERTSYSEX;
			
			if (MIDI_SYSEX == event.bEv[0])
			{
				pMfs->wfMFS |= SQF_INSERTSYSEX;
				++pMfs->cbPendingEvent;
				pMfs->dwPendingEvent = ((DWORD)MEVT_LONGMSG) << 24;
			}
			else if (MIDI_SYSEXEND == event.bEv[0])
			{
				pMfs->dwPendingEvent = ((DWORD)MEVT_LONGMSG) << 24;
			}
			
			if ( (mfsrc = mfsInsertParmData(event.tcDelta, pmh)) != MFS_SUCCESS)
				return mfsrc;
			
			pdw = (DWORD*)(pmh->lpData + pmh->dwBytesRecorded);
		}
		else if (event.bEv[1] == MIDI_TEMPO)
		{
			if (event.cbParm != 3) return MFS_BADFILE;
			
			*pdw++ = (DWORD)event.tcDelta;
			*pdw++ = 0;
			*pdw++ = setDWORD(event.pbParm[2], event.pbParm[1], event.pbParm[0], MEVT_TEMPO);
			pmh->dwBytesRecorded += 3*sizeof(DWORD);
		}
	}

	return (pMfs->wfMFS & SQF_EOF) ? MFS_EOF : MFS_SUCCESS;
}

static MFSRESULT  mfsInsertParmData(TICKS tcDelta, PMIDIHDR pmh)
{
	DWORD		dwLength;
	DWORD		dwRounded;
	DWORD*		pdw;
	
	// ------------------------------------

	if (pmh->dwBufferLength < pmh->dwBytesRecorded)	// shouldn't happen
	{
		return MFS_BADFILE;
	}
	
	// Check there's space for 4 DWORD's (tcDelta + stream-id + event + data)

	if (pmh->dwBufferLength - pmh->dwBytesRecorded < 4*sizeof(DWORD))
	{
		if (!tcDelta) return MFS_SUCCESS;
		
		//  Getting here with a real delta means the file's faulty
		return MFS_BADFILE;
	}
	
	pdw = (DWORD*)(pmh->lpData + pmh->dwBytesRecorded);
	
	dwLength = pmh->dwBufferLength - pmh->dwBytesRecorded - 3*sizeof(DWORD);
	dwLength = min(dwLength, pMfs->cbPendingEvent);
	
	*pdw++ = (DWORD)tcDelta;
	*pdw++ = 0;
	*pdw++ = (pMfs->dwPendingEvent & 0xFF000000) | (dwLength & 0x00FFFFFF);
	
	dwRounded = (dwLength + 3) & (~3);
	
	if (pMfs->wfMFS & SQF_INSERTSYSEX)
	{
		*((BYTE*)pdw) = MIDI_SYSEX;
		pdw = (DWORD*)((BYTE*)pdw + 1);	// increment pdw by 1

		pMfs->wfMFS &= ~SQF_INSERTSYSEX;
		--dwLength;
		--pMfs->cbPendingEvent;
	}
	
	if (dwLength & 0x80000000) return MFS_BADFILE;
	
	memcpy(pdw, pMfs->pbPendingEvent, dwLength);

	if ((pMfs->cbPendingEvent -= dwLength) == 0)
		pMfs->dwPendingEvent = 0;
	
	pmh->dwBytesRecorded += 3*sizeof(DWORD) + dwRounded;
	
	return MFS_SUCCESS;
}

static MFSRESULT mfsSeek(TICKS tcPosition, PMIDIHDR pmh)
{
	mevent_t	event;
	track_t		*ptrk;

	MFSRESULT	mfsrc;
	DWORD		*pdw;
	UINT		idx;
	UINT		idxChannel;
	UINT		idxController;
	BYTE		bEvent;

	// ------------------------------------

	memset(&evk, EK_UNSET, sizeof(evk));

	pMfs->tcPosition = 0;
	pMfs->wfMFS &= ~SQF_EOF;

	for (ptrk = tracks, idx = pMfs->dwTracks; idx--; ptrk++)
	{
		ptrk->tcPosition = 0;
		ptrk->cbLeft	 = ptrk->cbLength;
		ptrk->pbData	 = pMfs->pbFileData + ptrk->idxTrack;
		ptrk->bEvent	 = 0;
		ptrk->fdwTrack	 = 0;
	}
	
	while ( (mfsrc = mfsNextEvent((mevent_t*)&event, tcPosition)) == MFS_SUCCESS)
	{
		if ( (bEvent = event.bEv[0]) == MIDI_META)
		{
			if (event.bEv[1] == MIDI_TEMPO)
			{
				if (event.cbParm != sizeof(evk.rbTempo))
					return MFS_BADFILE;
				
				memcpy(evk.rbTempo, event.pbParm, event.cbParm);
			}
		}
		else switch(bEvent & 0xF0)
		{
		case MIDI_PROGCHANGE:
			evk.rbProgram[bEvent & 0x0F] = event.bEv[1];
			break;
			
		case MIDI_CTRLCHANGE:
			evk.rbControl[(((WORD)bEvent & 0x0F)*120) + event.bEv[1]] =
				event.bEv[2];
			break;
		}
	}
	
	if (mfsrc != MFS_TRACKEND) return mfsrc;

	// Build pmh from keyframe
	pmh->dwBytesRecorded = 0;
	pdw = (DWORD*)pmh->lpData;

	// Tempo change event?
	if (   evk.rbTempo[0] != EK_UNSET
		|| evk.rbTempo[1] != EK_UNSET
		|| evk.rbTempo[2] != EK_UNSET)
	{
		if (pmh->dwBufferLength - pmh->dwBytesRecorded < 3*sizeof(DWORD))
			return MFS_NOMEMORY;

		*pdw++ = 0;
		*pdw++ = 0;
		*pdw++ = setDWORD(evk.rbTempo[2], evk.rbTempo[1], evk.rbTempo[0], MEVT_TEMPO);
		pmh->dwBytesRecorded += 3*sizeof(DWORD);
    }

	// Program change events?
	for (idx = 0; idx < 16; idx++)
	{
		if (evk.rbProgram[idx] != EK_UNSET)
		{
			if (pmh->dwBufferLength - pmh->dwBytesRecorded < 3*sizeof(DWORD))
				return MFS_NOMEMORY;
			
			*pdw++ = 0;
			*pdw++ = 0;
			*pdw++ = setDWORD(MIDI_PROGCHANGE|idx, evk.rbProgram[idx], 0, MEVT_SHORTMSG);
			pmh->dwBytesRecorded += 3*sizeof(DWORD);
		}
	}

	// Controller events?
	for (idx = idxChannel = 0; idxChannel < 16; idxChannel++)
	{
		for (idxController = 0; idxController < 120; idxController++)
		{
			if (evk.rbControl[idx] != EK_UNSET)
			{
				if (pmh->dwBufferLength - pmh->dwBytesRecorded < 3*sizeof(DWORD))
					return MFS_NOMEMORY;
				
				*pdw++ = 0;
				*pdw++ = 0;
				*pdw++ = setDWORD(MIDI_CTRLCHANGE|idxChannel, idxController,
												evk.rbControl[idx], MEVT_SHORTMSG);
				pmh->dwBytesRecorded += 3*sizeof(DWORD);
			}
			
			idx++;
		}
	}

	if (!sequencer.tcbase)
	{
		// If the whole track is being played then the track
		// positions can be safely aligned. This is not absolutely
		// necessary but at least means if something odd happens
		// following a seek that the track should play OK.
		for (ptrk = tracks, idx = pMfs->dwTracks; idx--; ptrk++)
		{
			ptrk->tcPosition = tcPosition;
		}
	}

	return MFS_SUCCESS;
}

static void CALLBACK seqCallback(HMIDISTRM hms, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
	PMIDIHDR		pmh = (PMIDIHDR)dw1;
	MFSRESULT		mfsrc;

	// ------------------------------------
	
	if (uMsg != MOM_DONE) return;
	
	seqActive = true;
	
	if (!sequencer.midiBuffers) goto close_down;

	--sequencer.midiBuffers;
	
	if (sequencer.uState == SEQ_RESET)
	{
		// Recollect buffers from MMSYSTEM
		if (pmh != sequencer.pmhPreparse)
		{
			pmh->lpNext = sequencer.pmhFree;
			sequencer.pmhFree = pmh;
		}
		
		seqActive = false;
		return;
	}

	if ( (sequencer.uState == SEQ_STOPPING) || sequencer.wEof)
	{
		// Reached EOF, put the buffer back on the free list 
		if (pmh != sequencer.pmhPreparse)
		{
			pmh->lpNext = sequencer.pmhFree;
			sequencer.pmhFree = pmh;
		}
		
		if (midiOutUnprepareHeader(sequencer.hmidi, pmh, sizeof(*pmh)) != MMSYSERR_NOERROR)
		{
			seqActive = false;
			return;
		}

		if (!sequencer.midiBuffers)
		{
			// All events played
			// Either free the device or replay the file
			// NOTE: hanging notes at the end are ignored

			if (   sequencer.looped
				&& sequencer.hmidi
				&& sequencer.uState != SEQ_STOPPING)
			{
				sequencer.uState = SEQ_OPENED;
				sequencer.tcbase = 0;

				if (   seqPreparse() != MMSYSERR_NOERROR
					|| sequencer.uState != SEQ_PREPARSED
					|| midiStreamRestart((HMIDISTRM)sequencer.hmidi) != MMSYSERR_NOERROR)
				{
					goto close_down;
				}
	
				sequencer.uState = SEQ_PLAYING;
			}
			else
			{
close_down:;
				midiStreamClose((HMIDISTRM)sequencer.hmidi);
				
				sequencer.hmidi = NULL;
				sequencer.uState = SEQ_OPENED;
			}
		}
	}
	else
	{
		// Not EOF yet; attempt to fill another buffer
		if ( (mfsrc = mfsReadEvents(sequencer.hMfs, pmh, sequencer.tcEnd)) != MFS_SUCCESS)
		{
			if (mfsrc == MFS_EOF)
				sequencer.wEof = TRUE;
			else
			{
				sequencer.uState = SEQ_STOPPING;
				seqActive = false;
				return;
			}
		}
		
		++sequencer.midiBuffers;

		if (midiStreamOut((HMIDISTRM)sequencer.hmidi, pmh, sizeof(*pmh)) != MMSYSERR_NOERROR)
		{                
			--sequencer.midiBuffers;
			sequencer.uState = SEQ_STOPPING;
		}
	}

	seqActive = false;
}

void MC_SetMidiDevs()
{
	DWORD		n;

	numMidiDev = midiOutGetNumDevs();

	if (numMidiDev > 10) numMidiDev = 10;

	pr_OFinsert("Midi devices available; %i\n", numMidiDev);

	if (!numMidiDev) {mididev = 0; return; }
	if (mididev >= (int)numMidiDev) mididev = 0;

	if (compat[c_nosound] || compat[c_nomusic]) return;

	lstit_mididev = (uidata_listitem_t*)W_Calloc(numMidiDev * sizeof(uidata_listitem_t), STATICID, 0);
	lst_mididev.items = lstit_mididev;
	lst_mididev.numtodisplay = (numMidiDev > 4) ? 4 : numMidiDev;
	lst_mididev.count = numMidiDev;

	// List/store devices
	for (n = 0; n < numMidiDev; n++)
	{
		MIDIOUTCAPS caps;
		MMRESULT	res = midiOutGetDevCaps (n, &caps, sizeof(caps));

		if (res != MMSYSERR_NOERROR)
		{
			Con_Error(	"An error has occurred whilst trying to read the midi\n"
						"devices. Tick nomusic before restarting\n");
		}

		pr_OFinsert(" %i) ", n);

		if (caps.wTechnology == MOD_MIDIPORT)  pr_OFdirect(" MIDIPORT  "); 
		else
		if (caps.wTechnology == MOD_SYNTH)     pr_OFdirect(" SYNTH     ");
		else
		if (caps.wTechnology == MOD_SQSYNTH)   pr_OFdirect(" SQSYNTH   ");
		else
		if (caps.wTechnology == MOD_FMSYNTH)   pr_OFdirect(" FMSYNTH   ");
		else
		if (caps.wTechnology == MOD_MAPPER)    pr_OFdirect(" MAPPER    ");
		else
		if (caps.wTechnology == MOD_WAVETABLE) pr_OFdirect(" WAVETABLE ");
		else
		if (caps.wTechnology == MOD_SWSYNTH)   pr_OFdirect(" SWSYNTH   ");
		else
											   pr_OFdirect(" unknown   ");

#ifdef _DEBUG
		if (caps.dwSupport & MIDICAPS_CACHE)		pr_OFdirect(" CACHE   ");
		if (caps.dwSupport & MIDICAPS_LRVOLUME)		pr_OFdirect(" LRVOLUME");
		if (caps.dwSupport & MIDICAPS_STREAM)		pr_OFdirect(" STREAM  ");
		if (caps.dwSupport & MIDICAPS_VOLUME)		pr_OFdirect(" VOLUME  ");
#endif

		pr_OFinsert("\"%s\"\n", caps.szPname);
		strncpy(lstit_mididev[n].text, caps.szPname, 32);
		lstit_mididev[n].data = (int)n;
	}
}

// ===========================================
// Public Functions to access midi playback
// ===========================================

void MC_SetMidiDevice()
{
	MIDIOUTCAPS	caps;
	int			md = mididev;

	// defaults
	iMidiFlags = 1;
	uDeviceID = 0;
	mididev = 0;

	if (!numMidiDev) return;

	if (md <= 0 || md >= (int)numMidiDev) md = 0;

	if (midiOutGetDevCaps((DWORD)md, &caps, sizeof(caps)) == MMSYSERR_NOERROR)
	{
		uDeviceID = (DWORD)md;

		iMidiFlags = 0;
		mididev = md;

		// Volume cannot be changed when using a midi port
		 if ( !(caps.dwSupport & (MIDICAPS_VOLUME|MIDICAPS_LRVOLUME)) )
			 iMidiFlags |= 1;
		// GSReset only any use with a midiport 
		if (caps.wTechnology != MOD_MIDIPORT)
			 iMidiFlags |= 2;
	}
}

// To be called only once on start-up
bool MC_MidiInit()
{
	if (ge_inf.usefmodmidi) return 0;

	pr_OFdirect("Initialising : WinMM midi player\n");

	if (!numMidiDev) { midi_available = false; return false; }

	return (midi_available = seqAllocBuffers());
}

void MC_MidiStop()
{
	if (midi_available)
		seqStop();
}

void MC_MidiShutdown()
{
	if (midi_available)
	{
		if (sequencer.hmidi)
		{
			seqStop();
			midiStreamClose((HMIDISTRM)sequencer.hmidi);
			sequencer.hmidi = NULL;
		}
		else
		{
			while(seqActive) Sleep(1);
		}

		if (sequencer.pmhPreparse)
		{
			free(sequencer.pmhPreparse);
			sequencer.pmhPreparse = NULL;
		}

		if (sequencer.pbAlloc)
		{
			free(sequencer.pbAlloc);
			sequencer.pbAlloc = NULL;
		}

		midi_available = false;

		pr_OFdirect("WINMM midi shutdown\n");
	}
}

void MC_WinMidiPause(int pause)
{
	if (midi_available)
	{
		while(seqActive) Sleep(1);

		if (pause && sequencer.uState == SEQ_PLAYING)
		{
			if (!midiStreamPause((HMIDISTRM)sequencer.hmidi))
			{
				char *pos;

				if ( (pos = strstr(music_playing, "[")) )
				{
					*pos = 0;
					strcat(music_playing, "[paused]");
				}
					
				sequencer.uState = SEQ_PAUSED;
			}
		}
		else if (!pause && sequencer.uState == SEQ_PAUSED)
		{
			int i;

			for (i = 0; i < 4; i++)	// Might need a kick up the bum
			{
				if (!midiStreamRestart((HMIDISTRM)sequencer.hmidi))
				{			
					sequencer.uState = SEQ_PLAYING;
					break;
				}

				Sleep(40);	// wait awhile
			}
		}
	}
}

void MC_MidiVolumeSet(int vol)
{
	DWORD	dwVol;

	if (!(iMidiFlags & 1))
	{
		if (vol <   0) vol =   0;
		if (vol > 255) vol = 255;

		dwVol = (0xFF * vol)|0xFF;
		dwVolume = dwVol|(dwVol<<16);
		wmex = true;
	}
}

void MC_WinMidiTime()
{
	if (!midi_available || midi_error) return;

	if (!sequencer.hmidi)
	{
		if (midi_selected && midilengthMs)	// i.e. was playing
		{
			music_status = 0;
			musicpos = 0xFFFFFFFF;
			sld_musicpos.max = 0.f;
			sld_musicpos.value = 0.f;
			sprintf(music_playing,
				"%s MIDI: 00:00 / %02d:%02d  [Completed]",
				midi_mname,
				midilengthMs / 1000 / 60,
				midilengthMs / 1000 % 60);
			midilengthMs = 0;
		}

		return;
	}

	music_status = (sequencer.uState == SEQ_PLAYING);

	if (!music_status) return;

	if (wmex)
	{
		if (midiOutSetVolume(sequencer.hmidi, dwVolume) == MMSYSERR_NOERROR )
			wmex =false;
		else
			Con_Error("Midi error");
	}

	if (ui_active)
	{
		TICKS	tcPos;
		DWORD	ms;

		if (musicpos != 0xFFFFFFFF)
		{
			tcPos = musicpos;

			if (tcPos > sequencer.tcLength) tcPos = sequencer.tcLength;

			if (!sld_musicpos.button[0] && !sld_musicpos.button[1] && !sld_musicpos.button[2])
			{
				seqStop();
				sequencer.uState = SEQ_OPENED;
				sequencer.tcbase = musicpos;

				if (   seqPreparse() != MMSYSERR_NOERROR
					|| sequencer.uState != SEQ_PREPARSED
					|| midiStreamRestart((HMIDISTRM)sequencer.hmidi) != MMSYSERR_NOERROR)
				{
					Con_Error("WIN MIDI preparse error");	// should not happen
				}
	
				sequencer.uState = SEQ_PLAYING;
				musicpos = 0xFFFFFFFF;
			}
		}
		else
		{
			tcPos = seqTime();
			sld_musicpos.value = (float)tcPos;
		}

		ms = mfsTicksToMillisecs(sequencer.hMfs, tcPos);

		sprintf(music_playing,
			"%s MIDI: %02d:%02d / %02d:%02d  [%s]",
			midi_mname,
			ms / 1000 / 60,
			ms / 1000 % 60,
			midilengthMs / 1000 / 60,
			midilengthMs / 1000 % 60,
			sequencer.looped ? "Looping" : "Play Once");
	}
}

// The sequencer must have been stopped prior to calling.
// The file must have been pre-validated as a midi/rmid
int MC_PlayMidiFile(char *filename, char *mname)
{
	midi_error = 0;

	if (midi_available)
	{
		HMMIO			hmmio = NULL;
		MFSRESULT		mfsrc = MFS_SUCCESS;
		MMIOINFO		mmioinfo;
		MMCKINFO		ckRIFF;
		MMCKINFO		ckDATA;
		
		// ------------------------------------

		musicpos = 0xFFFFFFFF;
		midilengthMs = 0;
		sld_musicpos.max = 0.f;

		while(seqActive) Sleep(1);

		// If already open start a new instance
		if (sequencer.hmidi)
		{
			midiStreamClose((HMIDISTRM)sequencer.hmidi);
			sequencer.hmidi = NULL;
		}

		memset(sequencer.pbAlloc, 0, dwAlloc);

		sequencer.looped = global_looped != 0;
		sequencer.uState = SEQ_NOFILE;
		sequencer.hMfs = NULL;

		if (!sequencer.looped && ma_inf.loadgame)
		{
			return false;
		}
		
		midi_error = 1;	// Until proven otherwise

		// Verify that the file can be opened or created
		memset(&mmioinfo, 0, sizeof(mmioinfo));
		hmmio = mmioOpen(filename, &mmioinfo, MMIO_READ|MMIO_ALLOCBUF);

		if (!hmmio) return false;

		memset(pMfs, 0, sizeof(mfs_t));
		pMfs->wfMFS = 0;
		pMfs->pTempoMap = NULL;

		memset(&ckRIFF, 0, sizeof(ckRIFF));

		if (   mmioDescend(hmmio, &ckRIFF, NULL, MMIO_FINDRIFF) == MMSYSERR_NOERROR
			&& ckRIFF.fccType == mmioFOURCC('R','M','I','D'))
		{
			ckDATA.ckid = mmioFOURCC('d','a','t','a');
			
			if (mmioDescend(hmmio, &ckDATA, &ckRIFF, MMIO_FINDCHUNK) == MMSYSERR_NOERROR)
			{
				pMfs->szFileData = ckDATA.cksize;
			}
			else
			{
				mmioClose(hmmio, 0);
				return false;
			}
		}
		else
		{
			mmioSeek(hmmio, 0L, SEEK_SET);
			pMfs->szFileData = mmioSeek(hmmio, 0L, SEEK_END);
			mmioSeek(hmmio, 0L, SEEK_SET);
		}

		pMfs->pbFileData = (BYTE*)W_Realloc(pMfs->pbFileData, pMfs->szFileData, STATICID, 0);

		// Load the file into memory. 
		if (pMfs->szFileData != (DWORD)mmioRead(hmmio, (char*)pMfs->pbFileData, pMfs->szFileData))
		{
			mmioClose(hmmio, 0);
			return false;
		}

		mmioClose(hmmio, 0);
		
		if ( (mfsrc = mfsBuildFileIndex()) != MFS_SUCCESS)
		{
		//	Con_Message("MIDI \"%s\" has errors\n", mname);
			return false;
		}
		
		// The file's now open
		midi_error = 2;	// Until proven otherwise

		strcpy(lastmidifname, filename);
		midi_mname = mname;

		sequencer.hMfs		= (HMFS)pMfs;
		sequencer.dwTimeDiv	= pMfs->dwTimeDiv;
		sequencer.tcbase	= 0;
		sequencer.tcLength	= pMfs->tcLength;
		sequencer.tcEnd		= pMfs->tcLength;
	
		sequencer.uState = SEQ_OPENED;

		if (   seqPreparse() != MMSYSERR_NOERROR
			|| sequencer.uState != SEQ_PREPARSED)
		{
			return false;
		}

		// Values needed for file info display
		midilengthMs = mfsTicksToMillisecs(sequencer.hMfs, sequencer.tcLength);
		sld_musicpos.max = (float)sequencer.tcLength;
		
		if (midiStreamRestart((HMIDISTRM)sequencer.hmidi) != MMSYSERR_NOERROR)
		{
			seqCloseFile();
			return false;
		}

		midi_error = 0;
		sequencer.uState = SEQ_PLAYING;
		return true;
	}
	else if (ge_inf.usefmodmidi)	// v220-20
	{
int FModPlay(char *file, char *mname, int lumpnum, char *calledfrom, char *calledfromfrom);

		return FModPlay(filename, mname, 0, "MC_PlayMidiFile", "");
	}

	return false;
}

void MC_ChangeMidiDev()
{
	if (!midi_available) return;

	MC_MidiStop();
	MC_SetMidiDevice();
	MC_PlayMidiFile(lastmidifname, midi_mname);
}
