//-----------------------------------------------------------------------------
//
// Copyright (C) 2006 - 2012 G Jackson, Jaakko Kernen
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: sys_sfxd_ds.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
//  Modified for R3D
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
//  Modified to get rid of looping buffers where the
//  sound is only played once.
//  Note: not all changes marked as too extensive (also see s_chan.h)
// Revision 2.2.0 build 13 (v220-13) 15/Mar/2012 gmj
//  made compatible for c++
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  DirectSound Sfx Driver, with EAX 2.0
//  Low-level implementation, with manual sound
//  streaming by the  sound refresh thread.
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#define DIRECTSOUND_VERSION 0x0800

#include <dsound.h>
#include <initguid.h>
#include <eax.h>
#include <math.h>

#include "r3d_compat.h"
#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_misc.h"
#include "sys_sfxd.h"
#include "s_sfx.h"

// MACROS ------------------------------------------------------------------

// DSound buffer pointers
#define DSBuf(ch)		((LPDIRECTSOUNDBUFFER8  ) ch->buffer.ptr  )
#define DSBuf3(ch)		((LPDIRECTSOUND3DBUFFER8) ch->buffer.ptr3d)

#define EAXSUP			(KSPROPERTY_SUPPORT_GET|KSPROPERTY_SUPPORT_SET)

// TYPES -------------------------------------------------------------------

// v220-4
sidriver_t sid3d = {

	DS_SoundCreateBuffer,
	DS_SoundDestroyBuffer,
	DS_SoundRefresh,
	DS_SoundLoad,
	DS_SoundPlay,
	DS_SoundReset,
	DS_SoundStop,
	DS_SoundRestart,
	DS_SoundSet,
	DS_SoundSetv,
	DS_SoundListener,
	DS_SoundListenerv,
	DS_SoundShutdown
};

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern HWND		hWndMain;
extern RDBOOL	sfx_3d;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static IDirectSound8			*dsound = NULL;
static IDirectSoundBuffer		*primary = NULL;
static IDirectSound3DListener8	*ds_listener = NULL;
static IKsPropertySet			*eax_listener = NULL;

static HRESULT	hr;

// CODE --------------------------------------------------------------------

#pragma optimize("g",off)

//===========================================================================
//
//===========================================================================
static void ds_DSoundError(char *msg)
{
	Con_Message("ds_DSoundError: %s [%x]\n", msg, hr);
}

//===========================================================================
//
//===========================================================================
static IDirectSoundBuffer8 *ds_createBuffer8(DSBUFFERDESC *desc)
{
	IDirectSoundBuffer  *d3dbuf;
	IDirectSoundBuffer8 *d3dbuf8;

	// Try to create a secondary buffer with the requested properties.
	if(FAILED(hr = dsound->CreateSoundBuffer(desc, &d3dbuf, NULL)))
	{
		return NULL;
	}
	
	// Obtain the DirectSoundBuffer8 interface.
	if(FAILED(hr = d3dbuf->QueryInterface(IID_IDirectSoundBuffer8, (void**)&d3dbuf8)))
	{
		d3dbuf8 = NULL;
	}

	// Release the old interface, we don't need it.
	IDirectSoundBuffer_Release(d3dbuf);
	return d3dbuf8;
}

//===========================================================================
//
//===========================================================================
static IDirectSound3DBuffer8 *ds_DSound8Get3D(IDirectSoundBuffer8 *d3dbuf8)
{
	IDirectSound3DBuffer8 *d3dbuf3d;

	if (!d3dbuf8) return NULL;

	// Query the 3D interface.
	if (FAILED(hr = d3dbuf8->QueryInterface(IID_IDirectSound3DBuffer8, (void**)&d3dbuf3d)))
	{
		ds_DSoundError("Failed to get 3D interface.");
		d3dbuf3d = NULL;
	}
	return d3dbuf3d;
}

//===========================================================================
//	Returns true if getting and setting the property is supported.
//===========================================================================
static RDBOOL ds_EAXHasSupport(int property)
{
	ULONG support = 0;
	boolean has_support;
	
	if (!eax_listener) return false;
	
	has_support = SUCCEEDED(hr = eax_listener->QuerySupport(DSPROPSETID_EAX_ListenerProperties, property, &support))
		&& (support & EAXSUP) == EAXSUP;

	if(verbose) pr_OFinsert("ds_EAXHasSupport: Property %i => %s\n",
		property, has_support? "Yes" : "No");

	return has_support;
}

//===========================================================================
//	Not a driver per se, but part of the DSound driver.
//	Returns true if EAX properties are available.
//===========================================================================
static RDBOOL ds_EAXInit(void)
{
	IDirectSoundBuffer8   *dummy;
	IDirectSound3DBuffer8 *dummy3d = NULL;
	WAVEFORMATEX format;
	DSBUFFERDESC desc;

	// Configure the temporary buffer.
	memset(&desc, 0, sizeof(desc));
	desc.dwSize = sizeof(desc);
	desc.dwFlags = (DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLVOLUME | DSBCAPS_STATIC | DSBCAPS_CTRL3D);

	desc.dwBufferBytes = 44100;
	desc.lpwfxFormat = &format;
	
	desc.guid3DAlgorithm = DS3DALG_HRTF_FULL;

	memset(&format, 0, sizeof(format));
	format.wFormatTag = WAVE_FORMAT_PCM;
	format.nChannels = 1;
	format.nSamplesPerSec  = 44100;
	format.wBitsPerSample  = 2 * 8;
	format.nBlockAlign     = 2;
	format.nAvgBytesPerSec = 88200;
	
	// Create the temporary buffer.
	if( !(dummy = ds_createBuffer8(&desc)) )
		return false;

	// Get the 3D interface.
	if( !(dummy3d = ds_DSound8Get3D(dummy)) )
	{
		IDirectSound8_Release(dummy);
		return false;
	}

	// Query the property set interface.
	if (FAILED(hr = dummy3d->QueryInterface(IID_IKsPropertySet, (void**)&eax_listener)))
	{
		eax_listener = NULL;
	}
	else
	// Check for EAX support.
	if (   !ds_EAXHasSupport(DSPROPERTY_EAXLISTENER_ENVIRONMENT)
		|| !ds_EAXHasSupport(DSPROPERTY_EAXLISTENER_ROOM)
		|| !ds_EAXHasSupport(DSPROPERTY_EAXLISTENER_DECAYTIME)
		|| !ds_EAXHasSupport(DSPROPERTY_EAXLISTENER_ROOMHF)
		|| !ds_EAXHasSupport(DSPROPERTY_EAXLISTENER_REVERB)
		|| !ds_EAXHasSupport(DSPROPERTY_EAXLISTENER_ROOMROLLOFFFACTOR))
	{
		Con_Message("ds_EAXInit: Required EAX support missing.\n");
		IKsPropertySet_Release(eax_listener);
		eax_listener = NULL;
	}

	IDirectSoundBuffer8_Release(dummy3d);
	IDirectSoundBuffer8_Release(dummy);

	return eax_listener != NULL;
}

//===========================================================================
//
//===========================================================================
static void ds_EAXCommitDeferred(void)
{
	if(!eax_listener) return;

	eax_listener->Set(DSPROPSETID_EAX_ListenerProperties,
		DSPROPERTY_EAXLISTENER_COMMITDEFERREDSETTINGS, 
		NULL, 0, NULL, 0);
}

//===========================================================================
//	Convert linear volume 0..1 to logarithmic -10000..0.
//===========================================================================
static int ds_DSoundLinLog(float vol)
{
	int ds_vol;

	if(vol <= 0) return DSBVOLUME_MIN;
	if(vol >= 1) return DSBVOLUME_MAX;
	
	// Straighten the volume curve.
	ds_vol = (int)(100 * 20 * log10(vol));
	if(ds_vol < DSBVOLUME_MIN) ds_vol = DSBVOLUME_MIN;
	if(ds_vol > DSBVOLUME_MAX) ds_vol = DSBVOLUME_MAX;
	return ds_vol;
}

//===========================================================================
//
//===========================================================================
static void dS_EAXSetdw(DWORD prop, int value)
{
	hr = eax_listener->Set(DSPROPSETID_EAX_ListenerProperties,
		prop | DSPROPERTY_EAXLISTENER_DEFERRED, NULL, 0, 
		&value, sizeof(DWORD));
}

//===========================================================================
//
//===========================================================================
static void ds_EAXSetf(DWORD prop, float value)
{
	eax_listener->Set(DSPROPSETID_EAX_ListenerProperties,
		prop | DSPROPERTY_EAXLISTENER_DEFERRED, NULL, 0, 
		&value, sizeof(float));
}

//===========================================================================
//	Linear multiplication for a linear property.
//===========================================================================
static void ds_EAXMulf(DWORD prop, float mul, float min, float max)
{
	DWORD retBytes;
	float value;

	if(SUCCEEDED(hr = eax_listener->Get(DSPROPSETID_EAX_ListenerProperties,
		prop, NULL, 0, 
		&value, sizeof(value), &retBytes)))
	{
		value *= mul;
		if(value < min) value = min;
		if(value > max) value = max;
		ds_EAXSetf(prop, value);
	}
}

//===========================================================================
//	Linear multiplication for a logarithmic property.
//===========================================================================
static void ds_EAXMuldw(DWORD prop, float mul)
{
	DWORD	retBytes;
	LONG	value;

	if(SUCCEEDED(hr = eax_listener->Get(DSPROPSETID_EAX_ListenerProperties,
		prop, NULL, 0, 
		&value, sizeof(value), &retBytes)))
	{
		dS_EAXSetdw(prop, ds_DSoundLinLog((float)(pow(10, value/2000.0f) * mul)));
	}
}

//===========================================================================
//	Values use SRD_* for indices.
//===========================================================================
static void ds_DSoundListenerEnvironment(float *rev)
{
	// This can only be done if EAX is available.
	if (!eax_listener || !ds_listener) return;
	
	// Set the environment. Other properties are updated automatically.
	dS_EAXSetdw(DSPROPERTY_EAXLISTENER_ENVIRONMENT, (DWORD)rev[SRD_PRESET]);

	// General reverb volume adjustment.
	dS_EAXSetdw(DSPROPERTY_EAXLISTENER_ROOM, ds_DSoundLinLog(rev[SRD_VOLUME]));

	// Reverb decay.
	ds_EAXMulf(DSPROPERTY_EAXLISTENER_DECAYTIME, rev[SRD_DECAY], 
		EAXLISTENER_MINDECAYTIME, EAXLISTENER_MAXDECAYTIME);

	// Reverb damping.
	ds_EAXMuldw(DSPROPERTY_EAXLISTENER_ROOMHF, rev[SRD_DAMPING]);

	// A slightly increased roll-off.
	ds_EAXSetf(DSPROPERTY_EAXLISTENER_ROOMROLLOFFFACTOR, 1.3f);
}

//===========================================================================
//	Convert linear pan -1..1 to logarithmic -10000..10000.
//===========================================================================
static int ds_DSoundLogPan(float pan)
{
	if(pan >=  1.0f) return DSBPAN_RIGHT;
	if(pan <= -1.0f) return DSBPAN_LEFT;
	if(pan ==  0) return DSBPAN_CENTER;
	if(pan >   0) return (int)(-100 * 20 * log10(1 - pan));
	return (int)(100 * 20 * log10(1 + pan));	// pan < 0
}

//	End of static functions
//===========================================================================


//===========================================================================
//	Init DirectSound, start playing the primary buffer. Returns true
//	if successful.
//===========================================================================
RDBOOL D3D_SoundInit(void)
{
	DSBUFFERDESC desc;
	
	ds_listener = NULL;

	// Initialise the DirectSound object
	hr = CoCreateInstance(CLSID_DirectSound8, NULL, 
		CLSCTX_INPROC_SERVER, IID_IDirectSound8, (void**)&dsound);

	if ( FAILED(hr) )
	{
		ds_DSoundError("Failed to create the DS8 instance.");

		Sys_MessageBox("ERROR: 2D sound has failed to initialise.\n\n"
			"This can be caused if an incompatible dsound.dll is present in\n"
			"the Risen3D\\bin folder. If so delete any that may be present.\n", true);

		Sys_Quit(false);
	}

	// Initialize the DSound instance.
	if (FAILED(hr = IDirectSound8_Initialize(dsound, NULL)))
	{
		ds_DSoundError("Failed to init DS8");
		return false;
	}

	// Set cooperative level.
	if (FAILED(hr = IDirectSound_SetCooperativeLevel(dsound,
		hWndMain, DSSCL_PRIORITY)))
	{
		ds_DSoundError("Failed to set cooperative level");
		return false;
	}

	memset(&desc, 0, sizeof(desc));
	desc.dwSize = sizeof(desc);

	sfx_3d = ArgExists("-sfx_3d");

	if (sfx_3d)
	{
		// Create the primary buffer and try to initialize the 3D ds_listener.
		// If it succeeds, 3D sounds can be played.
		desc.dwFlags = DSBCAPS_PRIMARYBUFFER | DSBCAPS_CTRL3D;

		if (FAILED(hr = IDirectSound_CreateSoundBuffer(dsound, &desc, 
			&primary, NULL)))
		{
			ds_DSoundError("3D not available.");
			sfx_3d = false;
		}
		else	// Get the ds_listener.
		if(FAILED(hr = primary->QueryInterface(IID_IDirectSound3DListener8, (void**)&ds_listener)))
		{
			sfx_3d = false;
			ds_DSoundError("3D ds_listener not available.");
		}
		// initialise 3D eax properties.
		if (sfx_3d && !ds_EAXInit())
		{
			sfx_3d = false;
			ds_listener = NULL;
			Con_Message("ds_DSoundError: Failed to initiialise 3D effects.\n");
		}
	}
	
	if (!sfx_3d)
	{
		// 3D not wanted or not available? Create a 2D primary buffer instead.
		desc.dwFlags = DSBCAPS_PRIMARYBUFFER;

		if (FAILED(hr = IDirectSound_CreateSoundBuffer(dsound, &desc,
			&primary, NULL)))
		{
			ds_DSoundError("Failed to create 2D primary buffer.");
			return false;
		}
	}

	// Supposedly can be a bit more efficient not to stop the
	// primary buffer when there are no secondary buffers playing.
	IDirectSoundBuffer_Play(primary, 0, 0, DSBPLAY_LOOPING);

	return true;
}


//===========================================================================
//	Called using Listener().
//===========================================================================
void D3D_DSoundSetPrimaryFormat()
{
	WAVEFORMATEX wave;

	wave.wFormatTag      = WAVE_FORMAT_PCM; 
	wave.nChannels       = 2; 
	wave.nSamplesPerSec  = 44100; 
	wave.nBlockAlign     = wave.nChannels * 2;
	wave.nAvgBytesPerSec = wave.nSamplesPerSec * wave.nBlockAlign; 
	wave.wBitsPerSample  = 2 * 8;
	wave.cbSize = 0;

	IDirectSoundBuffer_SetFormat(primary, &wave);
}



//===========================================================================
//	Routines called using the sound interface driver.
//===========================================================================

//===========================================================================
//
//===========================================================================
int DS_SoundCreateBuffer(sfxchannel_t *ch, int flags, int bytes, int rate, uint size)
{
	IDirectSoundBuffer8		*buf_object8;
	IDirectSound3DBuffer8	*buf_object3d = NULL;
	WAVEFORMATEX			format;
	DSBUFFERDESC			desc;

	// Setup the buffer descriptor.
	memset(&desc, 0, sizeof(desc));
	desc.dwSize = sizeof(desc);
	desc.dwFlags = DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLVOLUME | DSBCAPS_STATIC	
		| (flags & SFXBF_3D ? DSBCAPS_CTRL3D | DSBCAPS_MUTE3DATMAXDISTANCE : DSBCAPS_CTRLPAN);

	// Set buffer size
	desc.dwBufferBytes = size; // 250ms buffer if repeating else size of data
	desc.lpwfxFormat = &format;

	// Control the selection with a Property!
	desc.guid3DAlgorithm = flags & SFXBF_3D ? DS3DALG_HRTF_FULL : DS3DALG_DEFAULT;

	// And the wave data format.
	format.wFormatTag = WAVE_FORMAT_PCM;
	format.nChannels = 1;
	format.nSamplesPerSec = rate;
	format.wBitsPerSample = bytes * 8;
	format.nBlockAlign = bytes;	// i.e. (format.nChannels * format.wBitsPerSample) / 8;
	format.nAvgBytesPerSec = rate * bytes;
	format.cbSize = 0;

	if (!(buf_object8 = ds_createBuffer8(&desc))) return 2;

	// How about a 3D interface?
	// If we don't have the ds_listener we cannot create 3D buffers. The
	// SFXBF_3D flag should be unset if ds_listener is unavailable.
	if (flags & SFXBF_3D)
	{
		if (!ds_listener) return 3;
		if (!(buf_object3d = ds_DSound8Get3D(buf_object8))) return 4;
	}

	// Setup the buffer object
	memset(&ch->buffer, 0, sizeof(sfxbuffer_t));

	ch->buffer.SFXBFflags = flags;
	ch->buffer.d3dsize = size;
	ch->buffer.ptr   = buf_object8;
	ch->buffer.ptr3d = buf_object3d;
	ch->buffer.bytes = bytes;
	ch->buffer.rate  = rate;
	ch->buffer.freq  = rate;	// Modified by calls to DS_DSoundSet(SFXBP_FREQUENCY).

	return 0;
}

//===========================================================================
//
//===========================================================================
void DS_SoundDestroyBuffer(sfxchannel_t *ch)
{
	uint flag = ch->buffer.SFXBFflags & SFXBF_3D;

	if (DSBuf3(ch)) IDirectSoundBuffer8_Release(DSBuf3(ch));
	if (DSBuf(ch) ) IDirectSoundBuffer8_Release(DSBuf(ch));

	memset(&ch->buffer, 0, sizeof(sfxbuffer_t));
	ch->buffer.SFXBFflags = flag;
}

//===========================================================================
//  Buffer streamer. Called by the Sfx refresh thread.
//   Non-repeating buffers not playing are stopped
//   If required a repeating buffer has its data updated.
//===========================================================================
RDBOOL DS_SoundRefresh(sfxchannel_t *ch)
{
	DWORD	bytes[2], dose, fill;
	DWORD	playcursor;
	void	*data[2];
	int		write_bytes, i;

	// Can only be done if there is a sample and the buffer is playing.
	if (   !ch->buffer.sample
		|| !(ch->buffer.SFXBFflags & SFXBF_PLAYING)
		|| ch->buffer.SFXBFflags & SFXBF_SUSPEND)
	{
		return false;
	}

	if ( !(ch->buffer.SFXBFflags & SFXBF_REPEAT) )
	{
		DWORD status;

		if (SUCCEEDED(hr = IDirectSoundBuffer_GetStatus(DSBuf(ch), &status))
			&& !(status & DSBSTATUS_PLAYING) )
		{
			DS_SoundStop(ch);
			return false;
		}
		return true;
	}

	// Get the cursor range in the buffer.
	if (FAILED(hr = IDirectSoundBuffer_GetCurrentPosition(DSBuf(ch),
		&playcursor, NULL)))
	{
		return false;	// Not much else can be done
	}

	// Calculate how many bytes we must write
	// (from buffer cursor up to play cursor).
	if (ch->buffer.cursor < playcursor)
		write_bytes = playcursor - ch->buffer.cursor;
	else // Play has looped back to the beginning.
		write_bytes = ch->buffer.d3dsize - ch->buffer.cursor + playcursor;

	// Try to lock the region, restoring if failed.
	for (i = 0; i < 2; i++)
	{
		hr = IDirectSoundBuffer8_Lock(DSBuf(ch),
			ch->buffer.cursor, write_bytes, 
			&data[0], &bytes[0], 
			&data[1], &bytes[1], 0);

		if (SUCCEEDED(hr)) break;

		if (!i && hr == DSERR_BUFFERLOST)
		{
			IDirectSoundBuffer8_Restore(DSBuf(ch));
			continue;	// Try again
		}
		return false;	// Beyond redemption
	}

	// Copy in two parts: as much sample data as we've got, and then zeros.
	for (i = 0; i < 2 && data[i]; i++)
	{
		// The dose is limited to the number of bytes we can write to this
		// pointer and the number of bytes we've got left.
		dose = MIN_OF(bytes[i], ch->buffer.sample->size - ch->buffer.written);

		if (dose)
		{
			// Copy from the sample data and advance cursor & written.
			memcpy((byte*) data[i], 
				(byte*) ch->buffer.sample->sdata + ch->buffer.written, dose);
			ch->buffer.written += dose;
			ch->buffer.cursor += dose;
		}

		if (dose < bytes[i])
		{
			// Repeating samples just rewind the 'written' counter when the
			// end is reached.
			if ( !(ch->buffer.SFXBFflags & SFXBF_REPEAT) )
			{
				byte fd = 0;

				if (ch->buffer.bytes == 1)
				{
					int  ptr = ch->buffer.sample->numsamples - 1;
					byte *sd = (byte*)ch->buffer.sample->sdata;
					fd = sd[ptr];
				}

				// The whole block was not filled. Write zeros in the rest.
				fill = bytes[i] - dose;

				// Filling an 8-bit sample with zeroes produces a nasty click.
				// Use the final sound byte instead
				memset((byte*) data[i] + dose, fd, fill);
				ch->buffer.cursor += fill;
			}
		}

		// Wrap the cursor back to the beginning if needed. The wrap can only
		// happen after the first write, really (where the buffer "breaks").
		if (ch->buffer.cursor >= ch->buffer.d3dsize)
			ch->buffer.cursor -= ch->buffer.d3dsize;
	}

	// And we're done! Unlock and get out of here.
	IDirectSoundBuffer8_Unlock(DSBuf(ch), 
		data[0], bytes[0], 
		data[1], bytes[1]);

	// If the buffer is in repeat mode, go back to the beginning once the
	// end has been reached.
	if (   ch->buffer.SFXBFflags & SFXBF_REPEAT
		&& ch->buffer.written == ch->buffer.sample->size)
	{
		ch->buffer.written = 0;
	}

	return true;
}

//===========================================================================
//	Prepare the buffer for playing a sample by filling the buffer with 
//	as much sample data as fits. The pointer to sample is saved, so
//	the caller must not free it while the sample is loaded.
//===========================================================================
RDBOOL DS_SoundLoad(sfxchannel_t *ch, struct sfxsample_s *sample)
{
	static	uint phased_increment = 0;

	void	*data;
	DWORD	locked_bytes, wrote_bytes;

	// Try to lock the buffer.
	if (FAILED(hr = IDirectSoundBuffer8_Lock(DSBuf(ch), 
		0, 0, &data, &locked_bytes, NULL, NULL, DSBLOCK_ENTIREBUFFER)))
	{
		return false; // Couldn't lock!
	}

	// Write as much data as we can.
	wrote_bytes = MIN_OF(locked_bytes, sample->size);
	memcpy(data, sample->sdata, wrote_bytes);
	// If the buffer's not filled set the cursor to the end
	ch->buffer.cursor = wrote_bytes < locked_bytes ? wrote_bytes : 0;

	IDirectSoundBuffer8_Unlock(DSBuf(ch), data, locked_bytes, NULL, 0);

	// Now the buffer is ready for playing.
	ch->buffer.sample = sample;	
	ch->buffer.written = wrote_bytes;
	ch->buffer.SFXBFflags &= ~SFXBF_RELOAD;

	if (   ch->buffer.SFXBFflags & SFXBF_REPEAT
		&& ch->buffer.SFXBFflags & SFXBF_3D)
	{
		DWORD dwNewPosition = phased_increment % (ch->buffer.written/2);

		// b220-0 Try and eliminate horrible 3D buffer phasing sound between 
		//        adjacent repeating buffers by getting them out of phase.
		IDirectSoundBuffer8_SetCurrentPosition( DSBuf(ch), dwNewPosition);

		phased_increment += 0x7175;
	}
	else
		IDirectSoundBuffer8_SetCurrentPosition(DSBuf(ch), 0);

	return true;
}

//===========================================================================
// DS_DSoundPlay
//===========================================================================
void DS_SoundPlay(sfxchannel_t *ch)
{
	uint flags = ch->buffer.SFXBFflags & SFXBF_REPEAT;

	if ( SUCCEEDED(hr = IDirectSoundBuffer8_Play(
		DSBuf(ch),
		0, 0,
		flags ? DSBPLAY_LOOPING : 0)))
	{
		// The buffer is now playing.
		ch->buffer.SFXBFflags |= SFXBF_PLAYING;
	}
}

//===========================================================================
//	Stops the buffer and makes it forget about its sample.
//===========================================================================
void DS_SoundReset(sfxchannel_t *ch)
{
	if (!ch->buffer.sample) return;	// Insurance

	IDirectSoundBuffer_Stop(DSBuf(ch));

	ch->buffer.sample = NULL;
	ch->buffer.SFXBFflags = (ch->buffer.SFXBFflags & SFXBF_3D) ? SFXBF_3D : 0;
}

//===========================================================================
//
//===========================================================================
void DS_SoundStop(sfxchannel_t *ch)
{
	if (!ch->buffer.sample) return;

	IDirectSoundBuffer_Stop(DSBuf(ch));

	if (ch->buffer.SFXBFflags & SFXBF_SUSPEND) return;

	ch->buffer.SFXBFflags = (ch->buffer.SFXBFflags & SFXBF_3D) ?
							SFXBF_3D|SFXBF_RELOAD : SFXBF_RELOAD;
}

//===========================================================================
//	b220-0 Restarts a suspended sound.
//===========================================================================
void DS_SoundRestart(sfxchannel_t *ch)
{
	ch->chstarttime = (uint)((float)Sys_GetRealTime() * 0.035f);	// used for priority

	if (FAILED(hr = IDirectSoundBuffer8_Play(
		DSBuf(ch),
		0, 0,
		ch->buffer.SFXBFflags & SFXBF_REPEAT ? DSBPLAY_LOOPING : 0)))
	{
		ch->buffer.SFXBFflags &= ~SFXBF_PLAYING;
	}

	ch->buffer.SFXBFflags &= ~SFXBF_SUSPEND;
}

//===========================================================================
//
//===========================================================================
void DS_SoundSet(sfxchannel_t *ch, int property, float value)
{
	switch(property)
	{
	case SFXBP_VOLUME:	// if negative, interpreted as attenuation
		{
			long val;
			
			if (value == 0)
				val = (long)(-10000);
			else if (value < 0)
				val = (long)((-1 - value) * 10000);	// log for attenuation
			else
				val = (long)ds_DSoundLinLog(value);	// lin for volume.

			IDirectSoundBuffer8_SetVolume(DSBuf(ch), val);
		}
		break;

	case SFXBP_FREQUENCY:
		{
			uint f = (uint)(ch->buffer.rate * value);
			// Don't set redundantly.
			if(f != ch->buffer.freq)
			{
				ch->buffer.freq = f;
				IDirectSoundBuffer8_SetFrequency(DSBuf(ch), f);
			}
		}
		break;

	case SFXBP_PAN:
		IDirectSoundBuffer8_SetPan(
			DSBuf(ch), 
			ds_DSoundLogPan(value));
		break;

	case SFXBP_MIN_DISTANCE:
		if(!DSBuf3(ch)) return;
		IDirectSound3DBuffer_SetMinDistance(
			DSBuf3(ch),
			value,
			DS3D_DEFERRED);
		break;

	case SFXBP_MAX_DISTANCE:
		if(!DSBuf3(ch)) return;
		IDirectSound3DBuffer_SetMaxDistance(
			DSBuf3(ch),
			value,
			DS3D_DEFERRED);
		break;

	case SFXBP_NORMAL_MODE:
		if(!DSBuf3(ch)) return;
		IDirectSound3DBuffer_SetMode(
			DSBuf3(ch), 
			DS3DMODE_NORMAL,
			DS3D_DEFERRED);
		break;

	case SFXBP_RELATIVE_MODE:
		if(!DSBuf3(ch)) return;
		IDirectSound3DBuffer_SetMode(
			DSBuf3(ch), 
			DS3DMODE_HEADRELATIVE,
			DS3D_DEFERRED);
		break;
	
	default:
		break;
	}
}

//===========================================================================
//	Coordinates specified in world coordinate system, converted to DSound's: 
//	+X to the right, +Y up and +Z away (Y and Z swapped, i.e.).
//===========================================================================
void DS_SoundSetv(sfxchannel_t *ch, int property, float *values)
{
	if(!DSBuf3(ch)) return;

	switch(property)
	{
	case SFXBP_POSITION:
		IDirectSound3DBuffer_SetPosition(
			DSBuf3(ch),
			values[VX], values[VZ], values[VY],
			DS3D_DEFERRED);
		break;

	case SFXBP_VELOCITY:
		IDirectSound3DBuffer_SetVelocity(
			DSBuf3(ch),
			values[VX], values[VZ], values[VY],
			DS3D_DEFERRED);
		break;
	
	default:
		break;
	}
}

//===========================================================================
//	Call SFXLP_UPDATE at the end of every channel update.
//===========================================================================
void DS_SoundListener(int property, float value)
{
	if (!ds_listener) return;

	switch(property)
	{
	case SFXLP_UPDATE:
		// Commit any deferred settings.
		IDirectSound3DListener_CommitDeferredSettings(ds_listener);
		ds_EAXCommitDeferred();
		break;

	case SFXLP_UNITS_PER_METER:
		IDirectSound3DListener_SetDistanceFactor(ds_listener, 1.0f/value, 
			DS3D_IMMEDIATE);
		break;

	case SFXLP_DOPPLER:
		IDirectSound3DListener_SetDopplerFactor(
			ds_listener,
			value, 
			DS3D_IMMEDIATE);
		break;
	
	case SFXLP_GAIN:
		break;
	
	default:
		break;
	}
}

//===========================================================================
//
//===========================================================================
void DS_SoundListenerv(int property, float *values)
{
	if(!ds_listener) return;

	switch(property)
	{
	
	case SFXLP_POSITION:
		IDirectSound3DListener_SetPosition(
			ds_listener, 
			values[VX], values[VZ], values[VY],
			DS3D_DEFERRED);
		break;

	case SFXLP_VELOCITY:
		IDirectSound3DListener_SetVelocity(
			ds_listener, 
			values[VX], values[VZ], values[VY],
			DS3D_DEFERRED);
		break;

	case SFXLP_ORIENTATION:
		{
			float	cos_yaw   = (float)cos(values[0]);
			float	cos_pitch = (float)cos(values[1]);
			float	sin_yaw   = (float)sin(values[0]);
			float	sin_pitch = (float)sin(values[1]);

			// The front vector points in the direction of the listener's nose, and
			// the top vector points out the top of the listener's head. By default,
			// the front vector is (0,0,1.0) and the top vector is (0,1.0,0).

			IDirectSound3DListener_SetOrientation(ds_listener,
				 cos_yaw * cos_pitch, sin_pitch,  sin_yaw * cos_pitch,	// front
				-cos_yaw * sin_pitch, cos_pitch, -sin_yaw * sin_pitch,	// top
				DS3D_DEFERRED);
		}
		break;

	case SFXLP_REVERB:
		ds_DSoundListenerEnvironment(values);
		break;

	default:
		break;
	}
}

//===========================================================================
//	Shut everything down.
//===========================================================================
void DS_SoundShutdown(void)
{
	if (!dsound) return;

	if (eax_listener)
		IKsPropertySet_Release(eax_listener);

	IDirectSound_Release(dsound);
	dsound = NULL;
	primary = NULL;
	ds_listener = NULL;
	eax_listener = NULL;
}
