//-----------------------------------------------------------------------------
//
// Copyright (C) 2005 Jaakko Kernen
// Copyright (C) 2006 - 2016 G Jackson
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
//  02111-1307, USA.
//
// $Log: sys_input.c
//
// Revision 2.1.0 build 62 04/Oct/2006 gmj
// Revision 2.2.0 build 0 20/May/2008 gmj (b211-0 / b220-0)
// Revision 2.2.0 build 3 (v220-3) 17/Dec/2008 gmj
// Revision 2.2.0 build 4 (v220-4) 23/Jul/2009 gmj
// Revision 2.2.0 build 6 (v220-6) 19/Mar/2010 gmj
// Revision 2.2.0 build 7 (v220-7) 05/Jul/2010 gmj
// Revision 2.2.0 build 31 (v220-31) 22/Dec/2016 gmj
//
// Based on: Doomsday v1.7.8
//
// DESCRIPTION:  Message boxes, shutdown routines
//
//-----------------------------------------------------------------------------

// HEADER FILES ------------------------------------------------------------

#include <windows.h>
#include <io.h>
#include <process.h>

#include "de_base.h"
#include "de_console.h"
#include "de_system.h"
#include "de_network.h"
#include "de_refresh.h"
#include "de_graphics.h"
#include "de_audio.h"
#include "de_misc.h"

#include "r3d_compat.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void	PH_ResetChanges(void);
void	prDisplayMsg(char *msg, char *response);
void	printTexReport(void);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int		systics = 0;	// System tics (every game tic).
RDBOOL	bypassAdvisory = false;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static size_t heap;
static RDBOOL coinit;

// CODE --------------------------------------------------------------------

//==========================================================================
// Sys_Init
//	Initialize machine state.
//==========================================================================
void Sys_Init(void)
{
	HRESULT		hr;

	if ( (hr = CoInitialize(NULL)) != S_OK)
		Con_ErrorD("Sys_Init: CoInitialize failed");

	coinit = true;

	Sys_InitTimer();

	if (!I_Init())
		Con_ErrorD("Sys_Init: failed to initialize DirectInput");

	Con_Message("Sys_Init     : keyboard[Y]");
	Con_Message("   mouse[%s]", mouseInited ? "Y" : "N");
	Con_Message("   joystick[%s]\n", joyInited ? "Y" : "N");
}

//==========================================================================
// Sys_Shutdown
//	Return to default system state.
//==========================================================================
void Sys_Shutdown(void)
{
	S_Shutdown();
	Sys_ShutdownTimer();
	GL_Shutdown();
	I_Shutdown();
	if (coinit) CoUninitialize();
}

int FR_CriticalMessage(const char *msg, int exit_type);
extern RDBOOL critical_inuse;

//===========================================================================
// Sys_CriticalMessage
//===========================================================================
int Sys_CriticalMessage(char *msg, RDBOOL printtofile)
{
	int ret;

	if (gl_inited)
	{
		if (critical_inuse) return 0;
		ret = FR_CriticalMessage(msg, 0);
	}
	else
	{
		char buf[300];
		char msgbuf[1024];

		strcpy(msgbuf, msg);
		strcat(msgbuf, "\nContinue ?");

		GetWindowText(hWndMain, buf, 300);

		ret = (MessageBox(hWndMain, msgbuf, buf,
			MB_SETFOREGROUND|MB_TOPMOST|MB_NOFOCUS|MB_YESNO|MB_SYSTEMMODAL|MB_ICONEXCLAMATION) 
			== IDYES);
	}

	bypassAdvisory = !ret;

	if (printtofile)
		prDisplayMsg(msg, ret ? "Y" : "N");

	MessagePump();
	return ret;
}

//===========================================================================
// Sys_MessageBox
//===========================================================================
void Sys_MessageBox(const char *msg, RDBOOL iserror)
{
	if (gl_inited)
	{
		if (critical_inuse) return;
		FR_CriticalMessage(msg, 1);
	}
	else
	{
		char title[300];

		GetWindowText(hWndMain, title, 300);

		// v220-3 MB_SYSTEMMODAL flag added
		MessageBox(hWndMain, msg, title,	// v220-6 MB_TOPMOST added
			MB_SETFOREGROUND|MB_TOPMOST|MB_NOFOCUS|MB_SYSTEMMODAL|MB_OK|
			(iserror? MB_ICONERROR : MB_ICONINFORMATION));
	}

	MessagePump();
}

//===========================================================================
// Sys_Sleep
//===========================================================================
void Sys_Sleep(int millisecs)
{
	Sleep(millisecs);
}

// v220-4
void Sys_AdvisoryHeader(char *msg)
{
	if (msg)
		pr_OFrepeat(msg, "=", 64 - strlen(msg), 2);
	else
		pr_OFrepeat(NULL, "=", 64, 2);
}

//==========================================================================
// Sys_Quit - Rewritten for R3D
//	Shut everything down and quit the program
//==========================================================================
void Sys_Quit(int spawnfile)
{
	extern FILE		*R3Doutfile;

	if (R3Doutfile) fflush(R3Doutfile);

	if (appStatus == 3)	exit(-1);

	if (appStatus == 2) PH_ResetChanges();	// R3D

	Edit_ResetVars();
	
	printTexReport();

	if (!bypassAdvisory)
	{
		// v220-4 print out textures not found in the wad(s)
		if ((ArgExists("-listloresfm") && R3D_ScriptedMap && unknownname.count > 0)
		|| (!R3D_ScriptedMap && unknownname.count - unknownname.flat_count > 4))
		{
			int	 i, k;
			
			Sys_AdvisoryHeader("SYSTEM ADVISORY ");

			pr_OFinsert("WAD TEXTURE ERROR REPORT for %s:\n"
					"%4i pnames, tex or flat names have no corresponding lump\n"
					"entries. Check all resources shipped by the pwad's author\n"
					"have been loaded. All loaded files are shown near the top\n"
					"in this file. Also check any readme shipped with the pwad\n"
					"to see if another file must be downloaded and installed.\n\n"
					"A small number missing means it is probably just a map\n"
					"authoring error requiring no further action. A large number\n"
					"may mean the wrong iwad is being used if no other faults\n"
					"can be found.\n\n"
					"The entries at fault are:\n\n",
					mainPwadname,
					unknownname.count);

			for (i = 0, k = 1; i < unknownname.count; i++)
			{
				pr_OFinsert("%4i) %s \"%s\"\n",
					k++,
					unknownname.flat_substituted[i] ? "FLAT   " : "TEXTURE", 
					&unknownname.name[i * 9]);
			}

			addPnameErrorToList(NULL);

			Sys_AdvisoryHeader(NULL);
		}
		else
			addPnameErrorToList(NULL);
	}

	if (!spawnfile) opennotepad = false;

	EngineShutdown();
}

// v220-31
void Sys_GeneralMessage(char *msg)
{
	if (!Sys_CriticalMessage(msg, true))
	{
		bypassAdvisory = true;
		Sys_Quit(0);
	}
}

// v220-31
void SCP_ZoffWarn(int mapnum, int thingtype)
{
	if (R3D_scpzoff || ge_inf.devmap)
	{
		char msg[256];

		sprintf(msg, "WARNING: T%i type %i not found in a SCP_ZOFF script\n", mapnum, thingtype);

		if (ge_inf.devmap)
			Sys_GeneralMessage(msg);
		else
			pr_OFdirect(msg);
	}
}

//===========================================================================
// Sys_StartThread
//	Returns a handle to the started thread.
//===========================================================================
int Sys_StartThread(systhreadfunc_t startpos, uint stacksize,
											unsigned long *id, void *parm)
{
	// If stacksize is less than 1MB then the stacksize of the calling
	// thread is used.  The linker may support STACKSIZE to allow a size
	// of less than 1MB to actually be used.
													
	// v220-7 use LPTHREAD_START_ROUTINE for consistency
	HANDLE handle = CreateThread(NULL,	// handle sharing not wanted
		stacksize,
		*(LPTHREAD_START_ROUTINE*)&startpos,
		parm,
		0,								// Run immediately on creation
		id);

	if (!handle)
	{
		Con_Message("Sys_StartThread: Failed to start new thread (%x).\n",
			GetLastError());
		return 0;
	}

	return (int) handle;
}

//===========================================================================
// Sys_SuspendThread
//	Suspends or resumes the execution of a thread.
//===========================================================================
void Sys_SuspendThread(int handle, RDBOOL dopause)
{
	if (dopause)
		SuspendThread( (HANDLE) handle);
	else
		ResumeThread( (HANDLE) handle);
}
