//------------------------------------------------------------------------
// MAIN : Main program for glBSP
//------------------------------------------------------------------------
//
//  GL-Friendly Node Builder (C) 2000-2002 Andrew Apted
//
//  Based on `BSP 2.3' by Colin Reed, Lee Killough and others.
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  Revision 2.2.0 build 6 (v220-6) 19/Apr/2010 gmj
//   Modified for internal use. See the bsp_ReadMe.txt in the bsp dir
//
//------------------------------------------------------------------------
 
#include "rd_system.h"

#include <stdio.h>
#include <io.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>
#include <assert.h>
#include <time.h>

#include "rd_blockmap.h"
#include "rd_level.h"
#include "rd_node.h"
#include "rd_seg.h"
#include "rd_structs.h"
#include "rd_util.h"
#include "rd_wad.h"

#include "r3d_arrays.h"

void pr_OFinsert(char *message, ...);

const nodebuildinfo_t *cur_info = NULL;
const nodebuildfuncs_t *cur_funcs = NULL;
volatile nodebuildcomms_t *cur_comms = NULL;

extern BOOL doing_normal;

static char glbsp_message_buf[1024];


const nodebuildinfo_t default_buildinfo =
{
	NULL,    // input_file
	NULL,    // output_file
	
	DEFAULT_FACTOR,  // factor
	
	false,   // mini_warnings
	false,   // pack_sides
	false,   // choose_fresh
	
	false,   // no_normal
	false,   // force_normal
	false,   // gwa_mode
	false,   // keep_sect
	false,   // no_prune
};

const nodebuildcomms_t default_buildcomms =
{
	NULL,   // message
	false   // cancelled
};

/* ----- option parsing ----------------------------- */

static void SetMessage(const char *str)
{
	GlbspFree(cur_comms->message);
	
	cur_comms->message = GlbspStrDup(str);
}

glbsp_ret_e GlbspCheckInfo(
			nodebuildinfo_t *info,
			volatile nodebuildcomms_t *comms,
			int type,
			int factor,
			int noprune,
			char *infile,
			char *outfile)
{
	cur_comms = comms;
	SetMessage(NULL);
	
	info->input_file  = infile;
	info->output_file = outfile;

	if ( !(_access(info->output_file, 4)) )
		remove(info->output_file);

	// Note: no_prune and pack_sides cannot be used together
	info->mini_warnings = false;
	info->pack_sides = false;
	info->choose_fresh = false;

	if (type)
	{
		info->force_normal = true;
		info->gwa_mode = false;
	}
	else
	{
		info->force_normal = false;
		info->gwa_mode = true;
	}

	info->factor = factor;	//DEFAULT_FACTOR;	// range 0 to 32
	info->no_prune = noprune;
	
	if (info->gwa_mode && info->force_normal)
		// force_normal used, but GWA files don't have normal nodes
		info->force_normal = false;
	
	if (info->no_normal && info->force_normal)
		// force_normal and no_normal cannot be used together
		info->force_normal = false;
	
	return GLBSP_E_OK;
}


/* ----- memory functions --------------------------- */

const char *GlbspStrDup(const char *str)
{
	if (! str)
		return NULL;
	
	return UtilStrDup(str);
}

void GlbspFree(const char *str)
{
	if (! str)
		return;
	
	UtilFree((char *) str);
}


/* ----- build nodes for a single level --------------------------- */

static glbsp_ret_e HandleLevel(void)
{
	superblock_t	*seg_list;
	nodb_t			*root_node;
	nodb_t			*root_stale_node;
	subsec_t		*root_sub;
	
	glbsp_ret_e ret;

	BOOL	rebuild = false;
	
	if (cur_comms->cancelled)
		return GLBSP_E_Cancelled;
	
	LoadLevel();
	
	InitBlockmap();
	
	// create initial segs
	seg_list = CreateSegs();
	
	root_stale_node = (num_stale_nodes == 0) ? NULL : 
						LookupStaleNode(num_stale_nodes - 1);
	
	// recursively create nodes
	ret = BuildNodes(seg_list, &root_node, &root_sub, 0, root_stale_node);

	FreeSuper(seg_list);
	
	if (ret == GLBSP_E_OK)
	{
		BOOL extended = false;

		ClockwiseBspTree(root_node);
		
		// Check for type of format required
		if ((num_normal_vert + num_gl_vert > 0xFFFE)
					|| num_complete_seg > 0xFFFE)
		{
			extended = true;
		}

		if (extended && !doing_normal && !ma_inf.zdnormal)
		{
			pr_OFinsert("\nGL 4 nodes required - normal nodes must be rebuilt\n"
					  "Build cancelled.");

			ret = GLBSP_E_NeedsNormal;
		}
		else
		{
/*  $glbsp_nomsg
			pr_OFinsert("Built %d NODES, %d SSECTORS, %d SEGS, %d VERTEXES\n",
				num_nodes, num_subsecs, num_segs, num_normal_vert + num_gl_vert);
			
			if (root_node)
				pr_OFinsert("Heights of left and right subtrees = (%d,%d)\n",
				ComputeHeight(root_node->r.node), ComputeHeight(root_node->l.node));
*/

			SaveLevel(root_node, extended);
		}
	}
	
	FreeLevel();
	FreeQuickAllocCuts();
	FreeQuickAllocSupers();

	return ret;
}

const nodebuildfuncs_t *funcs;

/* ----- main routine -------------------------------------- */

glbsp_ret_e GlbspBuildNodes(const nodebuildinfo_t *info,
		volatile nodebuildcomms_t *comms)
{
	glbsp_ret_e ret = GLBSP_E_OK;
	
	cur_info  = info;
	cur_funcs = funcs;
	cur_comms = comms;
	
	total_big_warn = total_small_warn = 0;
	
	// clear cancelled flag
	comms->cancelled = false;
	
	// sanity check
	if (!cur_info->input_file  || cur_info->input_file[0] == 0 ||
		!cur_info->output_file || cur_info->output_file[0] == 0)
	{
		pr_OFinsert("INTERNAL ERROR: Missing in/out filename !");
		return GLBSP_E_BadArgs;
	}
	
	// opens and reads directory from the input wad
	ret = ReadWadFile(cur_info->input_file);
	
	if (ret != GLBSP_E_OK)
		return ret;
	
	if (CountLevels() <= 0)
	{
		CloseWads();	
		return GLBSP_E_Unknown;
	}
	
	// loop over each level in the wad
	while (FindNextLevel())
	{
		if ((ret = HandleLevel()) != GLBSP_E_OK)
			break;
	}
	
	// writes all the lumps to the output wad
	if (ret == GLBSP_E_OK)
	{
		ret = WriteWadFile(cur_info->output_file);
		
		pr_OFinsert("Serious warnings: %d\n", total_big_warn);
		pr_OFinsert("Minor warnings: %d\n", total_small_warn);
		
		if (ReportFailedLevels()) ret = GLBSP_E_Overflow;
	}
	
	// close wads and free memory
	CloseWads();
	
	cur_info  = NULL;
	cur_comms = NULL;
	cur_funcs = NULL;
	
	return ret;
}

