/*
    Reads wad files, builds nodes, and saves new wad files.
    Copyright (C) 2002-2006 Randy Heit

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	$log: zd_processor.cpp
	First modified for use with Risen3D v220-14 02/Apr/2012 gmj
// Revision 2.2.0 build 35 (v220-35) 06/Dec/2021 gmj
    	Add R3D xline and nodes fix support
*/

#include "zd_processor.h"

extern void ShowView (FLevel *level);

enum
{
	// Thing numbers used in Hexen maps
    PO_HEX_ANCHOR_TYPE = 3000,
    PO_HEX_SPAWN_TYPE,
    PO_HEX_SPAWNCRUSH_TYPE,

    // Thing numbers used in Doom and Heretic maps
    PO_ANCHOR_TYPE = 9300,
    PO_SPAWN_TYPE,
    PO_SPAWNCRUSH_TYPE,
	PO_SPAWNHURT_TYPE
};

int second_pass; // v220-14 used for progress indicator

// --------------------------------------------------

FLevel::FLevel ()
{
	memset (this, 0, sizeof(*this));
}

FLevel::~FLevel ()
{
	if (Vertices)		delete[] Vertices;
	if (Subsectors)		delete[] Subsectors;
	if (Segs)			delete[] Segs;
	if (Nodes)			delete[] Nodes;
	if (Blockmap)		delete[] Blockmap;
	if (Reject)			delete[] Reject;
	if (GLSubsectors)	delete[] GLSubsectors;
	if (GLSegs)			delete[] GLSegs;
	if (GLNodes)		delete[] GLNodes;
	if (GLPVS)			delete[] GLPVS;
	if (OrgSectorMap)	delete[] OrgSectorMap;
}

FProcessor::FProcessor (FWadReader &inwad, int lump)
:
  Wad (inwad), Lump (lump)
{
	pr_OFinsert ("----%s----\n", Wad.LumpName (Lump));

	LoadVertices ();
	LoadLines ();
	LoadSides ();
	LoadSectors ();

	if (Level.NumLines() == 0 || Level.NumVertices == 0 || Level.NumSides() == 0 || Level.NumSectors() == 0)
	{
		pr_OFinsert ("   Map is incomplete\n");
	}
	else
	{
		Level.R3D_NODESfix ();	// R3Dgwa v220-35
		// Removing extra vertices is done by the node builder.
		Level.RemoveExtraLines ();

		if (!NoPrune)
		{
			Level.RemoveExtraSides ();
			Level.RemoveExtraSectors ();
		}

		Level.FindMapBounds ();
	}
}

void FProcessor::LoadLines ()
{
	int NumLines;

	MapLineDef *ml;
	ReadMapLump<MapLineDef> (Wad, "LINEDEFS", Lump, ml, NumLines);

	Level.Lines.Resize(NumLines);
	for (int i = 0; i < NumLines; ++i)
	{
		Level.Lines[i].v1 = LittleShort(ml[i].v1);
		Level.Lines[i].v2 = LittleShort(ml[i].v2);
		Level.Lines[i].flags = LittleShort(ml[i].flags);
		Level.Lines[i].sidenum[0] = LittleShort(ml[i].sidenum[0]);
		Level.Lines[i].sidenum[1] = LittleShort(ml[i].sidenum[1]);
		if (Level.Lines[i].sidenum[0] == NO_MAP_INDEX) Level.Lines[i].sidenum[0] = NO_INDEX;
		if (Level.Lines[i].sidenum[1] == NO_MAP_INDEX)
		{
			Level.Lines[i].sidenum[1] = NO_INDEX;
			if ( !(ml[i].flags & 0x0001) )
			{
			//	pr_OFinsert("L%i\n", i);
				ml[i].flags |= 0x0001;	// IMPASSABLE
				Level.Lines[i].flags = LittleShort(ml[i].flags);
			}
		}

		// Store the special and tag in the args array so we don't lose them
		Level.Lines[i].special = 0;
		Level.Lines[i].args[0] = LittleShort(ml[i].special);
		Level.Lines[i].args[1] = LittleShort(ml[i].tag);
	}
	delete[] ml;
}

void FProcessor::LoadVertices ()
{
	MapVertex *verts;
	ReadMapLump<MapVertex> (Wad, "VERTEXES", Lump, verts, Level.NumVertices);
	int i, k;

	Level.Vertices = new WideVertex[Level.NumVertices];

	for (i = 0; i < Level.NumVertices; ++i)
	{
		Level.Vertices[i].x = LittleShort(verts[i].x) << FRACBITS;
		Level.Vertices[i].y = LittleShort(verts[i].y) << FRACBITS;
		Level.Vertices[i].index = 0; // we don't need this value for non-UDMF maps

		// R3Dgwa - v220-35 check for any vertex changes
		for (k = 0; k < numvertexfix; k++)
		{
			if (vertexfix[k].vtxnum == i)
			{
				Level.Vertices[i].x = LittleShort(vertexfix[k].vx) << FRACBITS;
				Level.Vertices[i].y = LittleShort(vertexfix[k].vy) << FRACBITS;
				break;
			}
		}
	}
}

void FProcessor::LoadSides ()
{
	MapSideDef *Sides;
	int NumSides;
	ReadMapLump<MapSideDef> (Wad, "SIDEDEFS", Lump, Sides, NumSides);

	Level.Sides.Resize(NumSides);
	for (int i = 0; i < NumSides; ++i)
	{
		Level.Sides[i].textureoffset = Sides[i].textureoffset;
		Level.Sides[i].rowoffset = Sides[i].rowoffset;
		memcpy(Level.Sides[i].toptexture, Sides[i].toptexture, 8);
		memcpy(Level.Sides[i].bottomtexture, Sides[i].bottomtexture, 8);
		memcpy(Level.Sides[i].midtexture, Sides[i].midtexture, 8);

		Level.Sides[i].sector = LittleShort(Sides[i].sector);
		if (Level.Sides[i].sector == NO_MAP_INDEX) Level.Sides[i].sector = NO_INDEX;
	}
	delete [] Sides;
}

void FProcessor::LoadSectors ()
{
	MapSector *Sectors;
	int NumSectors;

	ReadMapLump<MapSector> (Wad, "SECTORS", Lump, Sectors, NumSectors);
	Level.Sectors.Resize(NumSectors);

	for (int i = 0; i < NumSectors; ++i)
	{
		Level.Sectors[i].data = Sectors[i];
	}
}

void FLevel::FindMapBounds ()
{
	fixed_t minx, maxx, miny, maxy;

	minx = maxx = Vertices[0].x;
	miny = maxy = Vertices[0].y;

	for (int i = 1; i < NumVertices; ++i)
	{
			 if (Vertices[i].x < minx) minx = Vertices[i].x;
		else if (Vertices[i].x > maxx) maxx = Vertices[i].x;
			 if (Vertices[i].y < miny) miny = Vertices[i].y;
		else if (Vertices[i].y > maxy) maxy = Vertices[i].y;
	}

	MinX = minx;
	MinY = miny;
	MaxX = maxx;
	MaxY = maxy;
}

// R3Dgwa v220-35 Change line textures and sector references
void FLevel::R3D_NODESfix ()
{
	if (!currentnodefix) return;

	static char *nderr = "NODESfix: script error ";
	static char *ndabort = ", script abandoned\n";
	int i;

	for (i = 0; i <= currentnodefix; i++)
	{
		int lnum = nodefix[i].line;

		if (nodefix[i].side1sec < 0)
		{
			pr_OFinsert("%sside1 must be given a valid sector%s", nderr, ndabort);
			return;
		}

		if (nodefix[i].ndflags & ND_NEWSIDE2 && nodefix[i].side2sec < 0)
		{
			pr_OFinsert("%sside2 must be given a valid sector%s\n", nderr, ndabort);
			return;
		}

		if (   (Lines[lnum].sidenum[0] != NO_INDEX && nodefix[i].ndflags & ND_NEWSIDE1)
			|| (Lines[lnum].sidenum[1] != NO_INDEX && nodefix[i].ndflags & ND_NEWSIDE2))
		{
			pr_OFinsert("%san existing sidedef cannot be replaced%s\n", nderr, ndabort);
			return;
		}
	}

	for (i = 0; i <= currentnodefix; i++)
	{
		int sec1, sec2;
		int sd1, sd2;
		int lnum = nodefix[i].line;

		sd1 = Lines[lnum].sidenum[0];
		sd2 = Lines[lnum].sidenum[1];

		if (sd1 == NO_INDEX)
		{
			pr_OFinsert("%sNODES FIX not possible as line %i has no side 0%s\n", nderr, lnum, ndabort);
			continue;
		}

		if (nodefix[i].ndflags & ND_NEWSIDE1 || nodefix[i].ndflags & ND_NEWSIDE2)
		{
			pr_OFinsert("%sNODES FIX new sides with L%i not currently supported with zdbsp nodes%s\n", nderr, lnum, ndabort);
			continue;
		}

		sec1 = nodefix[i].side1sec;
		sec2 = nodefix[i].side2sec;

		if (nodefix[i].ndflags & ND_CLEAR2S)
		{
			Lines[lnum].flags &= ~0x0204;	// ML_TWOSIDED|ML_PASSUSE
			Lines[lnum].flags |= 0x0001;	// ML_BLOCKING
		}

		if (nodefix[i].ndflags & ND_SWAPSIDES)
		{
			if (sd2 == NO_INDEX)	// nodefix[i].ndflags |= ND_NEWSIDE1;
			{
				pr_OFinsert("%sNODES FIX not possible as line %i has no side 0%s\n", nderr, lnum, ndabort);
				continue;
			}
			Lines[lnum].sidenum[0] = sd2;
			Lines[lnum].sidenum[1] = sd1;
		}
// FIXME R3D code added for glbsp needs to be changed, if possible, for zdbsp
/*		else
		{
			sd1 = nodefix[i].ndflags & ND_NEWSIDE1 ? NewSidedef() : line->right;
			sd2 = nodefix[i].ndflags & ND_NEWSIDE2 ? NewSidedef() : line->left;
		}

		sd1->sector = LookupSector(nodefix[i].side1sec); 

		if (sd2)
			sd2->sector = nodefix[i].side2sec >= 0 ? LookupSector(nodefix[i].side2sec) : NULL;

		if (nodefix[i].ndflags & ND_NEWSIDE2)
		{
			if (!sd2) FatalError("NODESfix: newside asked for but side2 is unset");
			sd2->index = num_sidedefs - 1;
			sd2->sector->ref_count++;
			line->left = sd2;
			line->flags |= LINEFLAG_TWO_SIDED;
			line->two_sided = true;
			line->left->ref_count++;
			line->left->sector = sd2->sector;
		}

		if (nodefix[i].ndflags & ND_NEWSIDE1)
		{
			sd1->index = num_sidedefs - 1;
			sd1->sector->ref_count++;
			line->right = sd1;
			line->right->ref_count++;
			line->right->sector = sd1->sector;
		}

		line->right->sector = sd1->sector;
*/

		Sides[Lines[lnum].sidenum[0]].sector = sec1;
		if (sd2 != NO_INDEX && sec2 >= 0)
			Sides[Lines[lnum].sidenum[1]].sector = sec2;

		// Set the textures *****************************
		if (stricmp(nodefix[i].sd1hi, "NOCHANGE"))	// i.e. NOT NOCHANGE
		{
			if (stricmp(nodefix[i].sd1hi, "NONE"))	// i.e. NOT NONE
				strncpy(Sides[Lines[lnum].sidenum[0]].toptexture, nodefix[i].sd1hi, 8);
			else
				strcpy(Sides[Lines[lnum].sidenum[0]].midtexture, "-");
		}
		
		if (stricmp(nodefix[i].sd1mid, "NOCHANGE"))
		{
			if (stricmp(nodefix[i].sd1hi, "NONE"))
				strncpy(Sides[Lines[lnum].sidenum[0]].midtexture, nodefix[i].sd1mid, 8);
			else
				strcpy(Sides[Lines[lnum].sidenum[0]].midtexture, "-");
		}
		
		if (stricmp(nodefix[i].sd1low, "NOCHANGE"))
		{
			if (stricmp(nodefix[i].sd1low, "NONE"))
				strncpy(Sides[Lines[lnum].sidenum[0]].bottomtexture, nodefix[i].sd1low, 8);
			else
				strcpy(Sides[Lines[lnum].sidenum[0]].midtexture, "-");
		}
		
		if (sd2 != NO_INDEX)
		{
			if (stricmp(nodefix[i].sd2hi, "NOCHANGE"))	// i.e. NOT NOCHANGE
			{
				if (stricmp(nodefix[i].sd1hi, "NONE"))	// i.e. NOT NONE
					strncpy(Sides[Lines[lnum].sidenum[1]].toptexture, nodefix[i].sd2hi, 8);
				else
					strcpy(Sides[Lines[lnum].sidenum[0]].midtexture, "-");
			}
			
			if (stricmp(nodefix[i].sd2mid, "NOCHANGE"))
			{
				if (stricmp(nodefix[i].sd2mid, "NONE"))
					strncpy(Sides[Lines[lnum].sidenum[1]].midtexture, nodefix[i].sd2mid, 8);
				else
					strcpy(Sides[Lines[lnum].sidenum[0]].midtexture, "-");
			}
			
			if (stricmp(nodefix[i].sd2low, "NOCHANGE"))
			{
				if (stricmp(nodefix[i].sd2low, "NONE"))
					strncpy(Sides[Lines[lnum].sidenum[1]].bottomtexture, nodefix[i].sd2low, 8);
				else
					strcpy(Sides[Lines[lnum].sidenum[0]].midtexture, "-");
			}
		}
		// ************************************************
	}
}

// R3Dgwa v220-35 replicate xline support first added to R3Dglbsp
void FLevel::RemoveExtraLines ()
{
	int i, newNumLines;

// R3Dgwa xline --------------?
	xlinemem_t	*lm;
	BYTE		*lmem = NULL;
	int			num_xlfound = 0;

	if (use_xl_lines)
	{
		lmem = (BYTE*)calloc(NumLines(), sizeof(BYTE));
		
		int			*cptr = (int *)calloc(NumLines(), sizeof(int));
		int			*lptr = (int *)calloc(NumLines(), sizeof(int));
		int			i, m, n;
		int			L1s, L1e, L2s, L2e;

		// identify same sector lines
		for (i = m = n = 0; i < NumLines(); ++i)
		{
			if (   Vertices[Lines[i].v1].x != Vertices[Lines[i].v2].x
				|| Vertices[Lines[i].v1].y != Vertices[Lines[i].v2].y)
			{
				if (   Lines[i].sidenum[0] != NO_INDEX
					&& Lines[i].sidenum[1] != NO_INDEX
					&& Sides[Lines[i].sidenum[0]].sector == Sides[Lines[i].sidenum[1]].sector
					&& !Sectors[Sides[Lines[i].sidenum[0]].sector].data.special
					&& Sides[Lines[i].sidenum[0]].bottomtexture[0] == '-'
					&& Sides[Lines[i].sidenum[0]].toptexture[0] == '-'
					&& Sides[Lines[i].sidenum[1]].bottomtexture[0] == '-'
					&& Sides[Lines[i].sidenum[1]].toptexture[0] == '-')
				{
					if (   Sectors[Sides[Lines[i].sidenum[0]].sector].data.tag
						&& Sides[Lines[i].sidenum[0]].midtexture[0] == '-'
						&& Sides[Lines[i].sidenum[1]].midtexture[0] == '-')
					{
						cptr[n++] = i;
					}
					else
					if (   !(Lines[i].flags & 0x0243)	// NOT ML_PASSUSE/ML_SOUNDBLOCK/ML_BLOCKMONSTERS/ML_BLOCKING
						&& !Lines[i].special	// sector tag
						&& !Sectors[Sides[Lines[i].sidenum[0]].sector].data.tag
						&& Sides[Lines[i].sidenum[0]].midtexture[0] != '-'
						&& Sides[Lines[i].sidenum[1]].midtexture[0] != '-'
						&& Sides[Lines[i].sidenum[0]].rowoffset == Sides[Lines[i].sidenum[1]].rowoffset
						&& !strcmp(Sides[Lines[i].sidenum[0]].midtexture, Sides[Lines[i].sidenum[1]].midtexture))
					{
						lptr[m++] = i;
					}
				}
			}
		}

		if (m > 1 && n > 1)
		{
			typedef struct { float	bbox[4]; } fbox_t;

			fbox_t		*fbox1, *fbox2;
			float		*b1, *b2;
			int			j;

			fbox1 = (fbox_t*)malloc(m * sizeof(fbox_t));
			fbox2 = (fbox_t*)malloc(n * sizeof(fbox_t));

			// create bounding boxes for lines
			for (i = 0; i < m; i++)
			{
				L1s = Vertices[Lines[lptr[i]].v1].x, L1e = Vertices[Lines[lptr[i]].v2].x;
				L2s = Vertices[Lines[lptr[i]].v1].y, L2e = Vertices[Lines[lptr[i]].v2].y;

				fbox1[i].bbox[BOXTOP   ] = L2s < L2e ? L2e : L2s;
				fbox1[i].bbox[BOXBOTTOM] = L2e < L2s ? L2e : L2s;
				fbox1[i].bbox[BOXLEFT  ] = L1s < L1e ? L1s : L1e;
				fbox1[i].bbox[BOXRIGHT ] = L1e < L1s ? L1s : L1e;
			}

			for (i = 0; i < n; i++)
			{
				L1s = Vertices[Lines[cptr[i]].v1].x, L1e = Vertices[Lines[cptr[i]].v2].x;
				L2s = Vertices[Lines[cptr[i]].v1].y, L2e = Vertices[Lines[cptr[i]].v2].y;

				fbox2[i].bbox[BOXTOP   ] = L2s < L2e ? L2e : L2s;
				fbox2[i].bbox[BOXBOTTOM] = L2e < L2s ? L2e : L2s;
				fbox2[i].bbox[BOXLEFT  ] = L1s < L1e ? L1s : L1e;
				fbox2[i].bbox[BOXRIGHT ] = L1e < L1s ? L1s : L1e;
			}

			for (i = 0; i < m; i++)
			{
				b1 = fbox1[i].bbox; 

				for (j = 0; j < n; j++)
				{
					b2 = fbox2[j].bbox;

					// check if an lptr line overlaps a cptr line
					if (   b1[BOXLEFT  ] > b2[BOXRIGHT] || b1[BOXRIGHT] < b2[BOXLEFT  ]
						|| b1[BOXBOTTOM] > b2[BOXTOP  ] || b1[BOXTOP  ] < b2[BOXBOTTOM])
					{
						continue;
					}

					// midtex line crosses a textureless line
					lmem[lptr[i]] = true;
					num_xlfound++;
					break;
				}
			}

			free(fbox1);
			free(fbox2);
		}

		free(cptr);
		free(lptr);

		if (num_xlfound)
		{
			xlinemem = (xlinemem_t*)W_Calloc(num_xlfound * sizeof(xlinemem_t), LEVELID, 0);
			xlisrendered = (unsigned char*)W_Malloc(num_xlfound, LEVELID, 56);
		}
	}
// ---------------------------<

	// Extra lines are those with 0 length. Collision detection against
	// one of those could cause a divide by 0, so it's best to remove them.

	for (i = newNumLines = 0; i < NumLines(); ++i)
	{
		if (Vertices[Lines[i].v1].x != Vertices[Lines[i].v2].x ||
			Vertices[Lines[i].v1].y != Vertices[Lines[i].v2].y)
		{
// R3Dgwa xline --------------?
			// R3Dgwa remove lines identified as crossing lines
			//        after storing them to the xlinemem_t array
			if (num_xlfound)
			{
				// R3Dgwa - remove xline sector lines
				if (lmem[i])
				{
					if (num_xlsame < num_xlfound)	// insurance
					{
						lm = &xlinemem[num_xlsame++];

						strncpy(lm->midtex, Sides[Lines[i].sidenum[0]].midtexture, 8);
						lm->midtex[8] = 0;
						lm->startx = FIX2FLT(Vertices[Lines[i].v1].x);
						lm->starty = FIX2FLT(Vertices[Lines[i].v1].y);
						lm->endx = FIX2FLT(Vertices[Lines[i].v2].x);
						lm->endy = FIX2FLT(Vertices[Lines[i].v2].y);
						lm->y_off = (float)Sides[Lines[i].sidenum[0]].rowoffset;
						lm->x_off[0] = (float)Sides[Lines[i].sidenum[0]].textureoffset;
						lm->x_off[1] = (float)Sides[Lines[i].sidenum[1]].textureoffset;
						lm->secn = Sides[Lines[i].sidenum[1]].sector;
						lm->flags = Lines[i].flags;

						continue;
					}
				}
			}
// ---------------------------<

			if (i != newNumLines)
			{
				Lines[newNumLines] = Lines[i];
			}

			++newNumLines;
		}
	}

	if (newNumLines < NumLines())
	{
		int diff = NumLines() - newNumLines - num_xlfound;

		if (num_xlfound) pr_OFinsert ("   Removed %d xlines.\n", num_xlfound);
		if (diff) pr_OFinsert ("   Removed %d line%s with 0 length.\n", diff, diff > 1 ? "s" : "");
	}

	if (lmem) free(lmem);
	Lines.Resize(newNumLines);
}

void FLevel::RemoveExtraSides ()
{
	BYTE *used;
	int *remap;
	int i, newNumSides;
	int haserrors = 0;

	// Extra sides are those that aren't referenced by any lines.
	// They just waste space, so get rid of them.
	int NumSides = this->NumSides();

	used = new BYTE[NumSides];
	memset (used, 0, NumSides*sizeof(*used));
	remap = new int[NumSides];

	// Mark all used sides
	for (i = 0; i < NumLines(); ++i)
	{
		if (Lines[i].sidenum[0] != NO_INDEX)
		{
			used[Lines[i].sidenum[0]] = 1;
		}
		else
		{
			if (!haserrors)
				pr_OFinsert ("--- ZDSP fatal errors (missing front sidedefs) lines:\n");
			pr_OFinsert("%7i", i);
			if (!(++haserrors % 8)) pr_OFinsert ("\n");
		}

		if (Lines[i].sidenum[1] != NO_INDEX)
		{
			used[Lines[i].sidenum[1]] = 1;
		}
	}

	if (haserrors)
		pr_OFinsert("\n--- End list ---\n");

	// Shift out any unused sides
	for (i = newNumSides = 0; i < NumSides; ++i)
	{
		if (used[i])
		{
			if (i != newNumSides)
			{
				Sides[newNumSides] = Sides[i];
			}
			remap[i] = newNumSides++;
		}
		else
		{
			remap[i] = NO_INDEX;
		}
	}

	if (newNumSides < NumSides)
	{
		int diff = NumSides - newNumSides;

		pr_OFinsert ("   Removed %d unused sidedef%s.\n", diff, diff > 1 ? "s" : "");
		Sides.Resize(newNumSides);

		// Renumber side references in lines
		for (i = 0; i < NumLines(); ++i)
		{
			if (Lines[i].sidenum[0] != NO_INDEX)
			{
				Lines[i].sidenum[0] = remap[Lines[i].sidenum[0]];
			}
			if (Lines[i].sidenum[1] != NO_INDEX)
			{
				Lines[i].sidenum[1] = remap[Lines[i].sidenum[1]];
			}
		}
	}
	delete[] used;
	delete[] remap;
}

void FLevel::RemoveExtraSectors ()
{
	BYTE *used;
	DWORD *remap;
	int i, newNumSectors;

	// Extra sectors are those that aren't referenced by any sides.
	// They just waste space, so get rid of them.

	NumOrgSectors = NumSectors();
	used = new BYTE[NumSectors()];
	memset (used, 0, NumSectors()*sizeof(*used));
	remap = new DWORD[NumSectors()];

	// Mark all used sectors
	for (i = 0; i < NumSides(); ++i)
	{
		if ((DWORD)Sides[i].sector != NO_INDEX)
		{
			used[Sides[i].sector] = 1;
		}
		else
		{
			pr_OFinsert ("   Sidedef %d needs a front sector.\n", i);
		}
	}

	// Shift out any unused sides
	for (i = newNumSectors = 0; i < NumSectors(); ++i)
	{
		if (used[i])
		{
			if (i != newNumSectors)
			{
				Sectors[newNumSectors] = Sectors[i];
			}
			remap[i] = newNumSectors++;
		}
		else
		{
			remap[i] = NO_INDEX;
		}
	}

	if (newNumSectors < NumSectors())
	{
		int diff = NumSectors() - newNumSectors;
		pr_OFinsert ("   Removed %d unused sector%s.\n", diff, diff > 1 ? "s" : "");

		// Renumber sector references in sides
		for (i = 0; i < NumSides(); ++i)
		{
			if ((DWORD)Sides[i].sector != NO_INDEX)
			{
				Sides[i].sector = remap[Sides[i].sector];
			}
		}
		// Make a reverse map for fixing reject lumps
		OrgSectorMap = new DWORD[newNumSectors];
		for (i = 0; i < NumSectors(); ++i)
		{
			if (remap[i] != NO_INDEX)
			{
				OrgSectorMap[remap[i]] = i;
			}
		}

		Sectors.Resize(newNumSectors);
	}

	delete[] used;
	delete[] remap;
}

void FProcessor::Write (FWadWriter &out)
{
	if (Level.NumLines() == 0 || Level.NumSides() == 0 || Level.NumSectors() == 0 || Level.NumVertices == 0)
	{
		return;
	}

	bool	gl5 = false;

	FNodeBuilder *builder = NULL;
	
	if ( !(builder = new FNodeBuilder (Level, Wad.LumpName (Lump), BuildGLNodes)) )
		Con_Error("ZDBSP: Not enough memory to build nodes!");

	delete[] Level.Vertices;
	builder->GetVertices (Level.Vertices, Level.NumVertices);

	second_pass = 0;

	if (BuildGLNodes)
	{
		builder->GetVertices (Level.GLVertices, Level.NumGLVertices);
		builder->GetGLNodes (Level.GLNodes, Level.NumGLNodes,
			Level.GLSegs, Level.NumGLSegs,
			Level.GLSubsectors, Level.NumGLSubsectors);

		// Now repeat the process to obtain regular nodes
		delete builder;

		second_pass = 100;

		if ( !(builder = new FNodeBuilder (Level, Wad.LumpName (Lump), false)) )
			Con_Error("ZDBSP: Not enough memory to build regular nodes!");

		delete[] Level.Vertices;
		builder->GetVertices (Level.Vertices, Level.NumVertices);
	}

	builder->GetNodes (Level.Nodes, Level.NumNodes,
			Level.Segs, Level.NumSegs,
			Level.Subsectors, Level.NumSubsectors);

	delete builder;
	builder = NULL;

	FBlockmapBuilder bbuilder (Level);
	WORD *blocks = bbuilder.GetBlockmap (Level.BlockmapSize);
	Level.Blockmap = new WORD[Level.BlockmapSize];
	memcpy (Level.Blockmap, blocks, Level.BlockmapSize*sizeof(WORD));

	Level.RejectSize = (Level.NumSectors()*Level.NumSectors() + 7) / 8;
	Level.Reject = NULL;
	
	if (Level.GLNodes != NULL )
	{
		gl5 =  (Level.NumGLVertices > 32767)
			|| (Level.NumGLSegs > 65534)
			|| (Level.NumGLNodes > 32767)
			|| (Level.NumGLSubsectors > 32767);

	}
	
	out.CopyLump (Wad, Lump);
	WriteLines (out);
	WriteSides (out);
	WriteVertices (out, Level.NumVertices);

	out.CreateLabel ("SEGS");
	out.CreateLabel ("SSECTORS");
	out.CreateLabel ("NODES");

	WriteSectors (out);

	if (Level.GLNodes != NULL)
	{
		char glname[9];
		glname[0] = 'G';
		glname[1] = 'L';
		glname[2] = '_';
		glname[8] = 0;
		strncpy (glname+3, Wad.LumpName (Lump), 5);
		out.CreateLabel (glname);
		WriteGLVertices (out, gl5);
		WriteGLSegs (out, gl5);
		WriteGLSSect (out, gl5);
		WriteGLNodes (out, gl5);
	}
}

MapNodeEx *FProcessor::NodesToEx (const MapNode *nodes, int count)
{
	if (count == 0)
	{
		return NULL;
	}

	MapNodeEx *Nodes = new MapNodeEx[Level.NumNodes];
	int x;

	for (x = 0; x < count; ++x)
	{
		WORD child;
		int i;

		for (i = 0; i < 4+2*4; ++i)
		{
			*((WORD *)&Nodes[x] + i) = LittleShort(*((WORD *)&nodes[x] + i));
		}
		for (i = 0; i < 2; ++i)
		{
			child = LittleShort(nodes[x].children[i]);
			if (child & NF_SUBSECTOR)
			{
				Nodes[x].children[i] = child + (NFX_SUBSECTOR - NF_SUBSECTOR);
			}
			else
			{
				Nodes[x].children[i] = child;
			}
		}
	}
	return Nodes;
}

MapSubsectorEx *FProcessor::SubsectorsToEx (const MapSubsector *ssec, int count)
{
	if (count == 0)
	{
		return NULL;
	}

	MapSubsectorEx *out = new MapSubsectorEx[Level.NumSubsectors];
	int x;

	for (x = 0; x < count; ++x)
	{
		out[x].numlines = LittleShort(ssec[x].numlines);
		out[x].firstline = LittleShort(ssec[x].firstline);
	}
	
	return out;
}

MapSegGLEx *FProcessor::SegGLsToEx (const MapSegGL *segs, int count)
{
	if (count == 0)
	{
		return NULL;
	}

	MapSegGLEx *out = new MapSegGLEx[count];
	int x;

	for (x = 0; x < count; ++x)
	{
		out[x].v1 = LittleShort(segs[x].v1);
		out[x].v2 = LittleShort(segs[x].v2);
		out[x].linedef = LittleShort(segs[x].linedef);
		out[x].side = LittleShort(segs[x].side);
		out[x].partner = LittleShort(segs[x].partner);
	}

	return out;
}

void FProcessor::WriteVertices (FWadWriter &out, int count)
{
	int i;
	WideVertex *vertdata = Level.Vertices;

	short *verts = new short[count * 2];

	for (i = 0; i < count; ++i)
	{
		verts[i*2] = LittleShort(vertdata[i].x >> FRACBITS);
		verts[i*2+1] = LittleShort(vertdata[i].y >> FRACBITS);
	}
	out.WriteLump ("VERTEXES", verts, sizeof(*verts)*count*2);
	delete[] verts;

	if (count >= 32768)
	{
		pr_OFinsert ("   VERTEXES is past the normal limit. (%d vertices)\n", count);
	}
}

void FProcessor::WriteLines (FWadWriter &out)
{
	int i;

	MapLineDef *ld = new MapLineDef[Level.NumLines()];

	for (i = 0; i < Level.NumLines(); ++i)
	{
		ld[i].v1 = LittleShort(WORD(Level.Lines[i].v1));
		ld[i].v2 = LittleShort(WORD(Level.Lines[i].v2));
		ld[i].flags = LittleShort(WORD(Level.Lines[i].flags));
		ld[i].sidenum[0] = LittleShort(WORD(Level.Lines[i].sidenum[0]));
		ld[i].sidenum[1] = LittleShort(WORD(Level.Lines[i].sidenum[1]));
		ld[i].special = LittleShort(WORD(Level.Lines[i].args[0]));
		ld[i].tag = LittleShort(WORD(Level.Lines[i].args[1]));
	}
	out.WriteLump ("LINEDEFS", ld, Level.NumLines()*sizeof(*ld));
	delete[] ld;
}

void FProcessor::WriteSides (FWadWriter &out)
{
	int i;
	MapSideDef *Sides = new MapSideDef[Level.NumSides()];

	for (i = 0; i < Level.NumSides(); ++i)
	{
		Sides[i].textureoffset = Level.Sides[i].textureoffset;
		Sides[i].rowoffset = Level.Sides[i].rowoffset;
		memcpy(Sides[i].toptexture, Level.Sides[i].toptexture, 8);
		memcpy(Sides[i].bottomtexture, Level.Sides[i].bottomtexture, 8);
		memcpy(Sides[i].midtexture, Level.Sides[i].midtexture, 8);
		Sides[i].sector = LittleShort(Level.Sides[i].sector);
	}
	out.WriteLump ("SIDEDEFS", Sides, Level.NumSides()*sizeof(*Sides));
	delete[] Sides;
}

void FProcessor::WriteSectors (FWadWriter &out)
{
	int i;
	MapSector *Sectors = new MapSector[Level.NumSectors()];

	for (i = 0; i < Level.NumSectors(); ++i)
	{
		Sectors[i] = Level.Sectors[i].data;
	}

	out.WriteLump ("SECTORS", Sectors, Level.NumSectors()*sizeof(*Sectors));
}

void FProcessor::WriteSegs (FWadWriter &out)
{
	int i;
	MapSeg *segdata;

	assert(Level.NumVertices < 65536);

	segdata = new MapSeg[Level.NumSegs];

	for (i = 0; i < Level.NumSegs; ++i)
	{
		segdata[i].v1 = LittleShort(WORD(Level.Segs[i].v1));
		segdata[i].v2 = LittleShort(WORD(Level.Segs[i].v2));
		segdata[i].angle = LittleShort(Level.Segs[i].angle);
		segdata[i].linedef = LittleShort(Level.Segs[i].linedef);
		segdata[i].side = LittleShort(Level.Segs[i].side);
		segdata[i].offset = LittleShort(Level.Segs[i].offset);
	}
	out.WriteLump ("SEGS", segdata, sizeof(*segdata)*Level.NumSegs);

	if (Level.NumSegs >= 65536)
	{
		pr_OFinsert ("   SEGS is too big for any port. (%d segs)\n", Level.NumSegs);
	}
	else if (Level.NumSegs >= 32768)
	{
		pr_OFinsert ("   SEGS is too big for vanilla Doom and some ports. (%d segs)\n", Level.NumSegs);
	}
}

void FProcessor::WriteSSectors (FWadWriter &out) const
{
	WriteSSectors2 (out, "SSECTORS", Level.Subsectors, Level.NumSubsectors);
}

void FProcessor::WriteSSectors2 (FWadWriter &out, const char *name, const MapSubsectorEx *subs, int count) const
{
	int i;
	MapSubsector *ssec;

	ssec = new MapSubsector[count];

	for (i = 0; i < count; ++i)
	{
		ssec[i].firstline = LittleShort((WORD)subs[i].firstline);
		ssec[i].numlines = LittleShort((WORD)subs[i].numlines);
	}
	out.WriteLump (name, ssec, sizeof(*ssec)*count);
	delete[] ssec;

	if (count >= 65536)
	{
		pr_OFinsert ("   %s is too big. (%d subsectors)\n", name, count);
	}
}

void FProcessor::WriteSSectors5 (FWadWriter &out, const char *name, const MapSubsectorEx *subs, int count) const
{
	int i;
	MapSubsectorEx *ssec;

	ssec = new MapSubsectorEx[count];

	for (i = 0; i < count; ++i)
	{
		ssec[i].firstline = LittleLong(subs[i].firstline);
		ssec[i].numlines = LittleLong(subs[i].numlines);
	}
	out.WriteLump (name, ssec, sizeof(*ssec)*count);
	delete[] ssec;
}

void FProcessor::WriteNodes (FWadWriter &out) const
{
	WriteNodes2 (out, "NODES", Level.Nodes, Level.NumNodes);
}

void FProcessor::WriteNodes2 (FWadWriter &out, const char *name, const MapNodeEx *zaNodes, int count) const
{
	int i, j;
	short *onodes, *nodes;

	nodes = onodes = new short[count * sizeof(MapNode)/2];

	for (i = 0; i < count; ++i)
	{
		short *inodes = (short *)&zaNodes[i];
		for (j = 0; j < 4+2*4; ++j)
		{
			nodes[j] = LittleShort(inodes[j]);
		}
		nodes += j;
		for (j = 0; j < 2; ++j)
		{
			DWORD child = zaNodes[i].children[j];
			if (child & NFX_SUBSECTOR)
			{
				*nodes++ = LittleShort(WORD(child - (NFX_SUBSECTOR + NF_SUBSECTOR)));
			}
			else
			{
				*nodes++ = LittleShort((WORD)child);
			}
		}
	}
	out.WriteLump (name, onodes, count * sizeof(MapNode));
	delete[] onodes;

	if (count >= 32768)
	{
		pr_OFinsert ("   %s is too big. (%d nodes)\n", name, count);
	}
}

void FProcessor::WriteNodes5 (FWadWriter &out, const char *name, const MapNodeEx *zaNodes, int count) const
{
	int i, j;
	MapNodeEx *const nodes = new MapNodeEx[count * sizeof(MapNodeEx)];

	for (i = 0; i < count; ++i)
	{
		const short *inodes = &zaNodes[i].x;
		short *coord = &nodes[i].x;
		for (j = 0; j < 4+2*4; ++j)
		{
			coord[j] = LittleShort(inodes[j]);
		}
		for (j = 0; j < 2; ++j)
		{
			nodes[i].children[j] = LittleLong(zaNodes[i].children[j]);
		}
	}
	out.WriteLump (name, nodes, count * sizeof(MapNodeEx));
	delete[] nodes;
}

void FProcessor::WriteGLVertices (FWadWriter &out, bool v5)
{
	int i, count = (Level.NumGLVertices - Level.NumOrgVerts);
	WideVertex *vertdata = Level.GLVertices + Level.NumOrgVerts;

	fixed_t *verts = new fixed_t[count*2+1];
	char *magic = (char *)verts;
	magic[0] = 'g';
	magic[1] = 'N';
	magic[2] = 'd';
	magic[3] = v5 ? '5' : '2';

	for (i = 0; i < count; ++i)
	{
		verts[i*2+1] = LittleShort(vertdata[i].x);
		verts[i*2+2] = LittleShort(vertdata[i].y);
	}
	out.WriteLump ("GL_VERT", verts, sizeof(*verts)*(count*2+1));
	delete[] verts;

	if (count > 65536)
	{
		pr_OFinsert ("   GL_VERT is too big. (%d GL vertices)\n", count/2);
	}
}

void FProcessor::WriteGLSegs (FWadWriter &out, bool v5)
{
	if (v5)
	{
		WriteGLSegs5 (out);
		return;
	}
	int i, count;
	MapSegGL *segdata;

	count = Level.NumGLSegs;
	segdata = new MapSegGL[count];

	for (i = 0; i < count; ++i)
	{
		if (Level.GLSegs[i].v1 < (DWORD)Level.NumOrgVerts)
		{
			segdata[i].v1 = LittleShort((WORD)Level.GLSegs[i].v1);
		}
		else
		{
			segdata[i].v1 = LittleShort(0x8000 | (WORD)(Level.GLSegs[i].v1 - Level.NumOrgVerts));
		}
		if (Level.GLSegs[i].v2 < (DWORD)Level.NumOrgVerts)
		{
			segdata[i].v2 = (WORD)LittleShort(Level.GLSegs[i].v2);
		}
		else
		{
			segdata[i].v2 = LittleShort(0x8000 | (WORD)(Level.GLSegs[i].v2 - Level.NumOrgVerts));
		}
		segdata[i].linedef = LittleShort((WORD)Level.GLSegs[i].linedef);
		segdata[i].side = LittleShort(Level.GLSegs[i].side);
		segdata[i].partner = LittleShort((WORD)Level.GLSegs[i].partner);
	}
	out.WriteLump ("GL_SEGS", segdata, sizeof(MapSegGL)*count);
	delete[] segdata;

	if (count >= 65536)
	{
		Con_Error ("GL_SEGS >= 65536.\nNo Doom Port can run with this (%d GL segs)\n", count);
	}
}

struct MapSegGL5Ex
{
	DWORD	v1;
	DWORD	v2;
	WORD	linedef;
	WORD	side;
	DWORD	partner;
};

void FProcessor::WriteGLSegs5 (FWadWriter &out)
{
	int i, count;
	MapSegGL5Ex *segdata;	// $Risen3D fix bug with array size

	count = Level.NumGLSegs;
	segdata = new MapSegGL5Ex[count];

	for (i = 0; i < count; ++i)
	{
		if (Level.GLSegs[i].v1 < (DWORD)Level.NumOrgVerts)
		{
			segdata[i].v1 = LittleLong(Level.GLSegs[i].v1);
		}
		else
		{
			segdata[i].v1 = LittleLong(0x80000000u | ((int)Level.GLSegs[i].v1 - Level.NumOrgVerts));
		}
		if (Level.GLSegs[i].v2 < (DWORD)Level.NumOrgVerts)
		{
			segdata[i].v2 = LittleLong(Level.GLSegs[i].v2);
		}
		else
		{
			segdata[i].v2 = LittleLong(0x80000000u | ((int)Level.GLSegs[i].v2 - Level.NumOrgVerts));
		}
		segdata[i].linedef = LittleShort(Level.GLSegs[i].linedef);
		segdata[i].side = LittleShort(Level.GLSegs[i].side);
		segdata[i].partner = LittleLong(Level.GLSegs[i].partner);
	}
	out.WriteLump ("GL_SEGS", segdata, sizeof(MapSegGL5Ex)*count);
	delete[] segdata;
}

void FProcessor::WriteGLSSect (FWadWriter &out, bool v5)
{
	if (!v5)
	{
		WriteSSectors2 (out, "GL_SSECT", Level.GLSubsectors, Level.NumGLSubsectors);
	}
	else
	{
		WriteSSectors5 (out, "GL_SSECT", Level.GLSubsectors, Level.NumGLSubsectors);
	}
}

void FProcessor::WriteGLNodes (FWadWriter &out, bool v5)
{
	if (!v5)
	{
		WriteNodes2 (out, "GL_NODES", Level.GLNodes, Level.NumGLNodes);
	}
	else
	{
		WriteNodes5 (out, "GL_NODES", Level.GLNodes, Level.NumGLNodes);
	}
}
